#include "AdaToCppFuncs.H"
#include "S_ScenEl.H"
#include "FreeDynamicAttributes.H"
#include "F_ProxyItem.H"
#include "EntityProxy.H"
#include "ExtrapolateMotion.H"
#include "event.H"
#include "rollbackStreams.H"

extern C_SIMOBJ *CurrentSimobj;
extern C_EVENT *CurrentEvent;
extern FREE_DYNAMIC_ATTRIBUTES *FreeDynamicAttributes;

static OBJECT_PROXY *CurrentObjectProxy = NULL;
static F_PROXY_ITEM *CurrentProxyItem = NULL;

/*
*************************************************************************
* SetExtrapolateMotion : Sets the dynamic position for the object proxy
* using an extrapolation equation.
*************************************************************************
*/

void SetExtrapolatedMotion(double StartTime, double EndTime,
			   double x0[3], double v0[3], double a0[3],
			   int PositionType) {

  EXTRAPOLATE_MOTION *ExtrapolateMotion;
  DYNAMIC_POSITION_ATTRIBUTE *Position;

  Position = ((S_ScenEl *)CurrentSimobj)->GetPosition();
  while ((ExtrapolateMotion = (EXTRAPOLATE_MOTION *)Position->get_top())) {
    *Position -= ExtrapolateMotion;
     RB_FREE_DELETE(FreeDynamicAttributes, ExtrapolateMotion);
  }

  ExtrapolateMotion = (EXTRAPOLATE_MOTION *)
    RB_FREE_NEW(FreeDynamicAttributes, EXTRAPOLATE_MOTION_ID);

  ExtrapolateMotion->Init(StartTime, EndTime, x0, v0, a0, PositionType);
  *Position += ExtrapolateMotion;

}

/*
*************************************************************************
* GetNumProxies : Returns the number of remote object proxies
*************************************************************************
*/

int GetNumProxies() {
  S_ScenEl *ScenEl = (S_ScenEl *)CurrentSimobj;
  C_XQUEUE *RemoteObjectProxies = ScenEl->GetRemoteObjectProxies();
  int length = RemoteObjectProxies->get_length();
  return length;
}

/*
*************************************************************************
* GoToFirstProxy : Goes to the first object proxy (i.e., sets the Current
* Object Proxy pointer to the first one)
*************************************************************************
*/

void GoToFirstProxy() {
  S_ScenEl *ScenEl = (S_ScenEl *)CurrentSimobj;
  C_XQUEUE *RemoteObjectProxies = ScenEl->GetRemoteObjectProxies();
  if (RemoteObjectProxies->get_length() == 0)
  {
    CurrentProxyItem = NULL;
    return;
  }
  CurrentProxyItem = (F_PROXY_ITEM *)RemoteObjectProxies->get_top();
  CurrentObjectProxy = CurrentProxyItem->GetObjectProxy();
}

/*
*************************************************************************
* GoToNextProxy : Goes to the next object proxy (i.e., sets the Current
* Object Proxy pointer to the next one in the list)
*************************************************************************
*/

void GoToNextProxy() {
  S_ScenEl *ScenEl = (S_ScenEl *)CurrentSimobj;
  C_XQUEUE *RemoteObjectProxies = ScenEl->GetRemoteObjectProxies();

  if (CurrentProxyItem == RemoteObjectProxies->get_bot()) return;

  CurrentProxyItem = (F_PROXY_ITEM *)CurrentProxyItem->get_link();
  CurrentObjectProxy = CurrentProxyItem->GetObjectProxy();
}

/*
*************************************************************************
* GetProxyName : Returns the name of the current proxy
*************************************************************************
*/

char *GetProxyName() {
  return CurrentObjectProxy->GetObjectName();
}

/*
*************************************************************************
* Returns the position, velocity, and acceleration of the current object
* proxy
*************************************************************************
*/

void GetProxyPosVelAcc(double time,
		       double x[3],
		       double v[3],
		       double a[3],
		       int coordinateSystem) {
  ENTITY_PROXY *EntityProxy = (ENTITY_PROXY *)CurrentObjectProxy;
  EntityProxy->Get_Position(x, v, a, coordinateSystem, time);
}

/*
*************************************************************************
* GetProxyAlive : Returns a flag telling if the remote proxy is alive
*************************************************************************
*/

int GetProxyAlive() {
  ENTITY_PROXY *EntityProxy = (ENTITY_PROXY *)CurrentObjectProxy;
  return EntityProxy->GetAlive();
}

/*
*************************************************************************
* GetAlive : Sets the flag for my proxy saying that I am alive
*************************************************************************
*/

int GetAlive() {
  S_ScenEl *ScenEl = (S_ScenEl *)CurrentSimobj;
  return ScenEl->GetAlive();
}

/*
*************************************************************************
* SetAlive : Sets the flag for my proxy saying that I am alive
*************************************************************************
*/

void SetAlive() {
  S_ScenEl *ScenEl = (S_ScenEl *)CurrentSimobj;
  ScenEl->SetAlive();
}

/*
*************************************************************************
* ResetAlive : Sets the flag for my proxy saying that I am alive
*************************************************************************
*/

void ResetAlive() {
  S_ScenEl *ScenEl = (S_ScenEl *)CurrentSimobj;
  ScenEl->ResetAlive();
}

/*
*************************************************************************
* RB_PrintString : Prints the string that is passed in after the event
* is committed
*************************************************************************
*/

void RB_PrintString(char *s) {
  RB_cout << s;
}
