// TestOPSM.C is a test program that tests if the object proxy state manager
// is working.

#include "SpStdlib.H"
#include "SpIostream.H"
#include "SpFstream.H"
#include "SpStrstream.H"
#include "SpIomanip.H"
#include "SpSignal.H"

#include "StateManager.H"
#include "error_signals.H"

STATE_MANAGER *StateManager = NULL;

/* This sigintfunc() call is to make sure that, when the OPSM is
** killed by pressing <ctrl>-c, that it will delete the StateManager.
** THis has the effect of sending a hangup message to the HostRouter
** and is the "best" way to hang up the line presently.  Simply
** include it as a default and ignore it.
*/

void sigintfunc(int) {
  cout << "Interupt" << endl;
  signal(SIGINT, sigintfunc);
  if (StateManager) delete StateManager;
  StateManager = NULL;
  exit(1);
}

void delete_StateManager() { delete StateManager; }

/*
** Enumerate the types of events you anticipate.  We generally finish the
** list by using N_EVENTS so that this # always has the correct number of
** events
*/

enum {
  DETECT_OBJECT_EVENT_ID,
  REFLECT_ATTRIBUTES_ID,
  RESCHEDULED_EVENT_ID,
  N_EVENTS
};

/* This event is goign to be scheduled by the OPSM after a detect
** object event has occured
*/

class RESCHEDULED_EVENT : public STATE_MANAGER_EVENT {
  private:
  protected:
  public:

    RESCHEDULED_EVENT(){;}

    virtual void Process() {
      cout << "Hello from RESCHEDULED_EVENT at time " << time_tag << endl;
    }

};

/*
** This is the event that will be launched when we first detect an object.
** You will see later on how we are able to link this event to that message.
*/

class DETECT_OBJECT_EVENT : public STATE_MANAGER_EVENT {
  private:
  protected:
  public:

    DETECT_OBJECT_EVENT(){;}

  /*
  ** Every event must have a process() method call.  Notice that this event
  ** begins scheduling the event RESCHEDULED_EVENT at 5 second intervals
  ** after this event has been processed.
  */
    virtual void Process() {
      cout << "Hello from DETECT_OBJECT_EVENT at time " << time_tag << endl;
      cout <<" We just got this detection of object id "<<GlobalId<<endl;
      RESCHEDULED_EVENT *ReschedEvent = new RESCHEDULED_EVENT();
      /*
      ** Sometimes you can get negative times as small as -1x10^20
      */
      if (time_tag < 0.0){
	ReschedEvent->ScheduleEvent(0);
      }
      else{
	ReschedEvent->ScheduleEvent(time_tag + 5);
      }
    }

};

/*
** This event will be launched every time that we get updated information
** about a given object
*/

class REFLECT_ATTRIBUTES : public STATE_MANAGER_EVENT {
  private:
  protected:
  public:

    REFLECT_ATTRIBUTES(){;}

    virtual void Process() {
      cout << "Hello from REFLECT_ATTRIBUTES at time " << time_tag << endl;
      cout <<" The object with ID "<<GlobalId<<" just changed something"<<endl;
    }

};

/* These are simply function calls that return some new events.  These
** are used so that SPEEDES can put them on free lists and not spend a
** great deal of time newing memory and deleting memory.
*/

void *CreateDETECT_OBJECT_EVENT(int n) {return new DETECT_OBJECT_EVENT[n];}
void *CreateREFLECT_ATTRIBUTES(int n) {return new REFLECT_ATTRIBUTES[n];}
void *CreateRESCHEDULED_EVENT(int n) {return new RESCHEDULED_EVENT[n];}

int main() {
  DATA_PARSER Parser("speedes.par");
  DATA_PARSER OPSM_Parser("OPSM.par");
  double Time = 0;
  OBJECT_PROXY *ObjectProxy;
  EM_HOST_USER * Em;

  // register sigintfunc to be call if a <ctrl>-c is ever pressed.
  signal(SIGINT, sigintfunc);

  // Also register a function for error signals like segmentation fault
  ErrorSignals(delete_StateManager);

  OPSM_Parser.GoTo("parameters", NULL);

  //If we are running the OPSM in playback mode
  if (OPSM_Parser.GetLogical("ReadFromFile")){
    StateManager = new STATE_MANAGER(OPSM_Parser.GetString("InFileName"));
  }
  else{
    // Create a new STATE_MANGER from the parser and don't lag the simulation
    // by more than 10 seconds.
    StateManager = new STATE_MANAGER(&Parser, 10.0);
  }


  // Grab the EM_HOST_USER for this OPSM
  Em = StateManager->GetEmHostUser();

  //If we are supposed to record the results of this simulation for later
  //playback
  if (OPSM_Parser.GetLogical("WriteToFile")){
    Em->RecordInputMessages(OPSM_Parser.GetString("OutFileName"));
  }


  //Subscribe to all SBIRS objects
  StateManager->SubscribeType("SBIRS", NULL);

  //Subscribe to the "Random Motion 33" object
  StateManager->SubscribeObject("Random Motion 33", NULL);

  // Tell the OPSM how many event types may be scheduled.
  StateManager->DefineNumberOfEvents(N_EVENTS);
  
  // Tell the OPSM to launch a DETECT_OBJECT_EVENT whenever a message with
  // the name "DISCOVER_OBJECT is sent from SPEEDES.  SPEEDES will
  // automatically send this message if you have subscribed to an object. The
  // number 100 means that the OPSM should make a freelist with 100 of these
  // events.
  StateManager->DefineEvent("DISCOVER_OBJECT",
			    DETECT_OBJECT_EVENT_ID,
			    CreateDETECT_OBJECT_EVENT,
			    sizeof(DETECT_OBJECT_EVENT),
			    100);

  // Tell the OPSM to launch a REFLECT_ATTRIBUTES event whenever a message
  // of name "REFLECT_ATTRIBUTES" is received.  Again, SPEEDES will
  // automatically send you this message when an object, to which you have
  // subscribed, changes some of it's attributes.
  StateManager->DefineEvent("REFLECT_ATTRIBUTES",
			    REFLECT_ATTRIBUTES_ID,
			    CreateREFLECT_ATTRIBUTES,
			    sizeof(REFLECT_ATTRIBUTES),
			    100);

  // Tell the OPSM to launch a RESCHEDULED_EVENT event whenever a message
  // of name "RESCHEDULED_EVENT" is received.  Again, SPEEDES will
  // automatically send you this message when an object, to which you have
  // subscribed, changes some of it's attributes.
  StateManager->DefineEvent("RESCHEDULED_EVENT",
			    RESCHEDULED_EVENT_ID,
			    CreateRESCHEDULED_EVENT,
			    sizeof(RESCHEDULED_EVENT),
			    100);

  // While SPEEDES is executing.  Note that if SPEEDES exits while we
  // are in some of the calls such as GoToTime, we will never break
  // out of those calls if they are blocking.
  while (Em->SpeedesExecuting()) {

    // Try to advance time by one second.  If SPEEDES has not yet
    // reached that time, this call will block until that has
    // happened.  If SPEEDES exits while we are in this call, we will
    // block indefinately.
    StateManager->GoToTime(Time+1);

    // Find out what time we are actually at
    Time = StateManager->GetCurrentTime();
    cout << "Current Time = " << StateManager->GetCurrentTime()
	 << ", Granted Time = " << StateManager->GetGrantedTime()
	 << endl;

    /*
    ** The StateManager can roll foward or backward in time.  Here we are
    ** going back to time 0 to find some information at that point.  You
    ** can think of the OPSM as a big tape recording of the simulation up
    ** to the current time.
    */
    StateManager->GoToTime(0.0);
    cout << "Number of Object Proxies at time 0 = "
	 << StateManager->GetNproxies() << endl;

    StateManager->GoToTime(Time/2);
    cout << "Number of Object Proxies at time " << Time << " = "
	 << StateManager->GetNproxies() << endl;

    // Here we actually get the object proxy for one of the objects
    // and print out some information about that object.  Notice that
    // we are going to print out information about that object at time
    // Time/2, not the current simulation time.
    ObjectProxy = StateManager->GetProxy("Random Motion 33");
    if (ObjectProxy) ObjectProxy->PrintStream();

    // Here we return to the present.
    StateManager->GoToTime(Time);
    cout << "Number of Object Proxies at time " << Time << " = "
	 << StateManager->GetNproxies() << endl;

    // Grab the object proxy for the same object and print out it's
    // value at the current time.
    ObjectProxy = StateManager->GetProxy("Random Motion 33");
    if (ObjectProxy) ObjectProxy->PrintStream();

    // Print out how many RANDOM_MOTION objects there are at our time
    cout << "Number of RANDOM MOTION = "
	 << StateManager->GetNproxies("RANDOM_MOTION") << endl;

    // Get the first object proxy of the type SBIRS
    ObjectProxy = StateManager->GetFirstProxy("SBIRS");

    // Now iterate through all the SBIRS objects and print out the
    // information about them
    while (ObjectProxy) {
      ObjectProxy->PrintStream();
      ObjectProxy = StateManager->GetNextProxy();
    }


  }

  //Delete the StateManager so that we hang up politely with the HostRouter
  delete StateManager;

  return 0;
}
