// H_ChangeScriptedMotion.C is an event handler that allows an object to
// be redirected in its motion (possibly from an external module). It receives
// the name of the location is should go to in its data.

#include "H_ChangeScriptedMotion.H"
#include "DataParser.H"
#include "GreatCircle.H"
#include "ConstMotion.H"
#include "Attribute.H"
#include "event.H"
#include "S_MovingEntity.H"

#include "FreeDynamicAttributes.H"
extern FREE_DYNAMIC_ATTRIBUTES *FreeDynamicAttributes;
extern DATA_PARSER *LocationParser;
extern C_EVENT *CurrentEvent;

void H_CHANGE_SCRIPTED_MOTION::Process(char *eventName, SIMTIME &timeTag,
				       char *data, int dataBytes,
				       int externalId) {
  double X[3], V[3];
  double Lat2, Lon2, Alt2;
  double EndTime;
  double Speed;
  GREAT_CIRCLE *GreatCircle;
  CONSTANT_MOTION *ConstantMotion;

//  RB_cout << "Change scripted motion event handler: " << timeTag << endl;
//  RB_cout << "Received new location " << data << endl;

/*
** Get our current position and speed. Convert this into earth coordinates
** so that we can obtain our lat, lon, and alt.
*/

  (*Position)(timeTag, ECR, X, V);
  Speed = Magnitude(V);

  if (Speed < 1.0e-20) {
    S_MOVING_ENTITY *MovingEntity = (S_MOVING_ENTITY *)SimObj;
    Speed = MovingEntity->GetVmax();
    if (Speed == 0.0) {
//      RB_cout << "Ignoring Change Scripted Motion since Speed = 0" << endl;
    }
  }

  (*Position)(timeTag, EARTH, X);

//  RB_cout << "Current Position = ["
//       << X[0] << ","
//       << X[1] << ","
//       << X[2] << "]"
//       << endl;

/*
** Using the location parser (which contains all of the named locations
** we can go to), find the location that was passed in through the data to
** this event handler. Then convert the location into a lat, lon, and alt.
*/

  LocationParser->GoTo(data, NULL);

  Lat2 = (LocationParser->GetFloat("LatDeg")
	+ LocationParser->GetFloat("LatMin") / 60.0
	+ LocationParser->GetFloat("LatSec") / 3600.0)
        * (PI/180);
  if (!LocationParser->GetLogical("North")) Lat2 = -Lat2;

  Lon2 = (LocationParser->GetFloat("LonDeg")
	+ LocationParser->GetFloat("LonMin") / 60.0
	+ LocationParser->GetFloat("LonSec") / 3600.0)
        * (PI/180);
  if (!LocationParser->GetLogical("East")) Lon2 = -Lon2;

  Alt2 = 0.0;

/*
** If we are already where the command told us to go, then ignore the command
*/

  double temp[2];
  temp[0] = Lat2 - X[0];
  temp[1] = Lon2 - X[1];
  if (temp[0]*temp[0] + temp[1]*temp[1] < 1e-20)  {
//    RB_cout << "Ignoring Change Scripted Motion since already there..."
//	    << endl;
    return;
  }

/*
** Create a great circle object which will start at our current location
** at our current time and then go to the destination using the current
** velocity.
*/

  GreatCircle = (GREAT_CIRCLE *)
    RB_FREE_NEW(FreeDynamicAttributes, GREAT_CIRCLE_ID);

  GreatCircle->init_Vconstant(X[0], X[1], Alt2, timeTag-0.001,
			      Lat2, Lon2, Alt2, Speed);

/*
** Remove all of the current dynamic position items and put them back in their
** appropriate free lists. Note, that this must be done in a rollbackable
** manner (i.e., the RB_FREE_DELETE function).
*/

  DYNAMIC_POSITION_ITEM *dpi = (DYNAMIC_POSITION_ITEM *)Position->get_top();
  while (Position->get_top()) {
//  cout << "Removing motion with start time " << dpi->GetStartTime() << endl;
    *Position -= dpi;
    RB_FREE_DELETE(FreeDynamicAttributes, dpi);
    dpi = (DYNAMIC_POSITION_ITEM *)Position->get_top();
  }

/*
** Add the great circle motion to the dynamic position attribute
*/

//  RB_cout << "Adding GreatCircle with time interval ["
//          << GreatCircle->GetStartTime()
//          << ", " << GreatCircle->GetEndTime() << "]"
//          << endl;
  *Position += GreatCircle;

/*
** Get the end time of the great circle motion we just created and use that
** to set up a constant motion to describe its location after it arrives.
** Then, add it to the dynamic position attribute as well.
*/

  EndTime = GreatCircle->GetEndTime();

  ConstantMotion = (CONSTANT_MOTION *)
    RB_FREE_NEW(FreeDynamicAttributes, CONSTANT_MOTION_ID);
  ConstantMotion->SetEARTH();
  ConstantMotion->SetStartTime(EndTime);
  ConstantMotion->SetEndTime(1.0e20);
  ConstantMotion->SetPosition(Lat2, Lon2, Alt2);
  *Position += ConstantMotion;

/*
** Print out some information about events that this event automatically
** generated. Normally, applications don't do this kind of thing. Later,
** we should delete this.
*/

#ifdef COMMENTOUT
  C_QUEUE *qmess = CurrentEvent->get_mque();
  C_HOLDER *holder = (C_HOLDER *)qmess->get_top();

  for (int i=0; i<qmess->get_length(); i++) {
    char *message = holder->get_buff();
    C_HEADER *header = (C_HEADER *)message;
    cout << "Header: Simtime " << header->time_tag << endl;
    cout << "Header: EVENT_TYPE " << header->evtype << endl;
    cout << "Header: OBJECT_TYPE " << header->obtype << endl;
    cout << "Header: LOCAL_ID " << header->objid << endl;
    cout << "Header: NODE " << holder->get_node() << endl;
    holder = (C_HOLDER *)holder->get_link();
  }
#endif

//  cout << "Done changing scripted motion" << endl;

}

