#define LatLonConstants_Here
#include "LatLonConstants.H"
#include "LatLon.H"

//Auxiliary Routines///////////////////////////////////////////////////
//As long as cycleLength > 0, FMOD returns a nonnegative number less than
//cycleLength.  Really, all bets are off if someone submits a negative or
//zero cycleLength.
double FMOD(double a, double cycleLength)
{
	if(cycleLength > 0.0){
		if(a >= 0.0) return fmod(a, cycleLength);
		return cycleLength - fmod(-a, cycleLength);
	}
	if(cycleLength < 0.0) return -FMOD(-a, -cycleLength);
	return 0.0;
}

double NormLon(double lon)
{
//Return something bigger than -180 and at most 180.
	lon = FMOD(lon, 360.0);
	if(lon <= 180.0) return lon;
	return lon - 360.0;
}

double aTrigd(double y, double x, int which)
{
//	which:
//  1 means look at x,    find acosd.
//  2 means look at y,    find asind.
//  3 means look at both, find atan2d.
	double a;
	if(which==1){
		if(fabs(x) <= 0.75) return acos_degrees(x);
		a = 1.0 - x*x;
		if(a < 0.0) return 0.0;
		a = asin_degrees(sqrt(a));
		return x >= 0.0? a: 180.0 - a;
	}
	if(which==2){
		if(fabs(y) <= 0.75) return asin_degrees(y);
		a = 1.0 - y*y;
		if(a < 0.0) return 90.0;
		a = acos_degrees(sqrt(a));
		return y >= 0.0? a: -a;
	}
	a = sqrt(x*x + y*y);
	if(a == 0.0) return 0.0;
	x /= a;
	y /= a;
	if(fabs(x) < fabs(y)) return y >= 0.0? acos_degrees(x): -acos_degrees(x);
	return x >= 0.0? asin_degrees(y): 180.0  - asin_degrees(y);
}

//Vect3/////////////////////////////////////////////////////////////
Vect3::Vect3(): x(0.0), y(0.0), z(0.0) {}

Vect3::Vect3(double xx, double yy, double zz, int normalize):
	x(xx), y(yy), z(zz)
{
	if(!normalize) return;
	double size = VSize();
	if(size < VERYSMALL) return;
	x /= size; y /= size; z /= size;
}

Vect3::Vect3(const LatLon &ltLn)
{
	double r = cos_degrees(ltLn.Lat());
	x = r*cos_degrees(ltLn.Lon());
	y = r*sin_degrees(ltLn.Lon());
	z = sin_degrees(ltLn.Lat());
}

Vect3::Vect3(const LatLonAlt &ltLnAlt)
{
	double r = cos_degrees(ltLnAlt.ll.Lat());
	double scale = 1.0 + ltLnAlt.alt/FT_PER_E_RAD;
	LatLon ltLn = ltLnAlt.ll;
	x = r*scale*cos_degrees(ltLn.Lon());
	y = r*scale*sin_degrees(ltLn.Lon());
	z = scale*sin_degrees(ltLn.Lat());
}

//The following could be inlined, but then we'd need <math.h>
//in latLon.hh.  So we define the function here.
double Vect3::VSize() const
{
	return sqrt(x*x + y*y + z*z);
};

/*
ostream& operator<<(ostream& out, const Vect3 &v)
{
//	return out <<strFormat("(%8.4lf,%8.4lf,%8.4lf)",v.X(), v.Y(), v.Z());
  out << "(" << v.X() << "," << v.Y() << "," << v.Z() << ")";
  return out;
}
*/

//RangeAzimuth/////////////////////////////////////////////////////////
RangeAzimuth::RangeAzimuth(): range(0), azimuth(0) {}
RangeAzimuth::RangeAzimuth(double r, double a): azimuth(a)
{
	range = FMOD(r, NMI_PER_E_CIRC);
	if(range > NMI_PER_E_CIRC/2.0){
		azimuth += 180.0;
		range = NMI_PER_E_CIRC - range;
	}
	azimuth = FMOD(azimuth, 360.0);
}

/*
ostream& operator<<(ostream& out, const RangeAzimuth &ra) {
//  return out << strFormat("(%8.4lf,%8.4lf)", ra.Range(), ra.Azimuth());
  out << "(" << ra.Range() << "," << ra.Azimuth() <<")";
  return out;
}
*/

//LatLon///////////////////////////////////////////////////////////
LatLon::LatLon(double lt, double ln): lat(FMOD(lt, 360.0)), lon(ln)
{
	if(lat <= 90.0);
	else if(lat < 270.0){
		lat = 180.0 - lat;
		lon += 180.0;
	}
	else lat -= 360.0;
	lon = NormLon(lon);
}

LatLon::LatLon(const Vect3 &v): lat(0.0), lon(0.0)
{
//Normalize to make sure we're on the earth's surface.
	Vect3 vv(v.X(), v.Y(), v.Z(), 1);
	if(vv.VSize() < VERYSMALL) return;
	lat = aTrigd(vv.Z(), 0.0, 2);
	if(vv.X()*vv.X() + vv.Y()*vv.Y() < VERYSMALL) lon = 0.0;
	else lon = NormLon(aTrigd(vv.Y(), vv.X(), 3));
}

int
LatLon::operator==(const LatLon &ltLn) const
{
	Vect3 v1(*this), v2(ltLn);
	return (v1-v2).VSize() < VERYSMALL;
}

const LatLon LatLon::operator+(const RangeAzimuth &ra) const
{
	Vect3 b1(*this), b3(LatLon(lat + 90.0, lon)), b2(b3*b1);
	Vect3 b23(cos_degrees(90.0 - ra.Azimuth())*b2 + sin_degrees(90.0 - ra.Azimuth())*b3);
	double range_in_e_rad = ra.Range()/NMI_PER_E_RAD;
	return cos(range_in_e_rad)*b1 + sin(range_in_e_rad)*b23;
}
/*
ostream& operator<<(ostream& out, const LatLon &ltLn) {
//  return out << strFormat("(%8.4lf,%8.4lf)", ltLn.Lat(), ltLn.Lon());
  out << "(" << ltLn.Lat() << "," << ltLn.Lon() << ")";
  return out;
}
*/

//LatLon///////////////////////////////////////////////////////////
LatLonAlt::LatLonAlt(const LatLon &llIn, double altIn):
	ll(llIn), alt(altIn)
{};

LatLonAlt::LatLonAlt(const Vect3 &v):
	ll(v), alt((v.VSize()-1)*FT_PER_E_RAD)
{};

/*
ostream& operator<<(ostream& out, const LatLonAlt &ltLnAlt)
{
	out <<"[" <<ltLnAlt.ll <<"," <<ltLnAlt.alt <<"]";
	return out;
};
*/

//Global Functions using LatLons//////////////////////////////////////////////
RangeAzimuth rangeAzimuthFromFirstToSecond(
	const LatLon &latLon0, const LatLon &latLon1
)
{
//Change coordinates.
	const Vect3 b1(latLon0);
	const Vect3 b3(LatLon(latLon0.Lat()+90,latLon0.Lon()));
	const Vect3 b2(b3*b1);
	const Vect3 v1(latLon1);

//The cosine between latLon1 and latLon0 gives range information.  First check
//to see if range is 0.0 or half-of-earth.
	double theCosine = (v1,b1);
	if(theCosine > 1.0 - VERYSMALL) return RangeAzimuth(0.0, 0.0);
	if(theCosine < -1.0 + VERYSMALL){
		return RangeAzimuth(NMI_PER_E_CIRC/2.0, 0.0);
	}

//Range is acos times the earth's radius.
//Azimuth is gotten by considering the projection of v1 onto the plane
//containing b2 and b3, and taking the angle of this point with b3.
	return RangeAzimuth(
		aTrigd(0.0, theCosine, 1) * LATLON_PI/180.0 * NMI_PER_E_RAD,
		90.0 - aTrigd((v1,b3), (v1,b2), 3)
		);
}

RangeAzimuth rangeAzimuthToArc(
	const LatLon &startingLatLon,
	const LatLon &endpoint0, const LatLon &endpoint1
)
{
//To find the smallest distance, find the largest cosine.
//Start by considering the endpoints.
	Vect3 v0=endpoint0, v1=endpoint1, w=startingLatLon;
	double c0w = (v0,w);
	double c1w = (v1,w);
	double biggestCosine = c0w>c1w? c0w: c1w;
	LatLon tgt = c0w>c1w? endpoint0: endpoint1;
	
//Within the great circle containing the arc, find the extreme cosines
//to startingLatLon.  Start by parametrizing the great circle containing
//the two endpoints.  To do this, we need two orthonormal vectors v0 and v11,
//so that the parametrization of the great circle is
//cos(th)*v0 + sin(th)*v11.
//We have v0.  To find v11, go in the direction of v1, but go 1/4 of the
//way around the world.
	RangeAzimuth ra01 = rangeAzimuthFromFirstToSecond(endpoint0,endpoint1);
	Vect3 v11 =
		Vect3(endpoint0 + RangeAzimuth(NMI_PER_E_CIRC/4.0, ra01.Azimuth()));
//Let q(theta) = [cos(theta)*v0 + sin(theta)*v11] dot w,
//and find the values of theta where dq/d(theta)=0.  This gives the
//values of theta that are closest to and farthest from w.
//Letting alpha=(v0,w) and beta=(v11,w), the critical values of theta satisfy
//-sin(theta)*alpha + cos(theta)*beta = 0, or tan(theta) = beta/alpha.
	double alpha = (v0,w), beta = (v11,w);
	double theta = FMOD(aTrigd(beta, alpha, 3), 360.0);
//If this critical point is not in the arc, consider the other one.
//If neither critical point is in the arc, use the preferred endpoint.
	if(60.0 * theta >= ra01.Range()) theta = FMOD(theta+180.0, 360.0);
	if(60.0 * theta >= ra01.Range()){
		return rangeAzimuthFromFirstToSecond(startingLatLon, tgt);
	}
//At this point, theta is a critical point and it's within the arc.
	Vect3 vx = cos_degrees(theta)*v0 + sin_degrees(theta)*v11;
	double cx2 = (vx,w);
	if(cx2>biggestCosine){
		return rangeAzimuthFromFirstToSecond(startingLatLon, LatLon(vx));
	}
	return rangeAzimuthFromFirstToSecond(startingLatLon, tgt);
}

RangeAzimuth twoDRangeAzimuthToArc(
	const LatLon &startingLatLon,
	const LatLon &endpoint0, const LatLon &endpoint1
)
{
	double midLat =
		(startingLatLon.Lat() + endpoint0.Lat() + endpoint1.Lat())/3.0;
	double nmiPerDegLon = 60.0 * cos_degrees(midLat);
	Vect3 vStart(
		nmiPerDegLon*startingLatLon.Lon(), 60.0 * startingLatLon.Lat(), 0.0
		);
	Vect3 v0(nmiPerDegLon*endpoint0.Lon(), 60.0 * endpoint0.Lat(), 0.0);
	Vect3 v1(nmiPerDegLon*endpoint1.Lon(), 60.0 * endpoint1.Lat(), 0.0);
	Vect3 d = v1-v0;
	double size=d.VSize();
	if(size<=0) d=Vect3(1,0,0);
	else d = d*(1/size);
	double coord = ((vStart-v0),d);
	if(coord<0) coord=0;
	else if(coord>size) coord=size;
	Vect3 offset = vStart - (v0 + coord*d);
	return RangeAzimuth(offset.VSize(), 90 - aTrigd(vStart.Y(), vStart.X(), 3));
}

void getRngeAzElAng(
	double lat1, double lon1, double alt1,
	double lat2, double lon2, double alt2,
	double &range, double &elAngle, double &azimuth
	)
{
	LatLon ll1(lat1, lon1), ll2(lat2, lon2);

	double z1 = alt1/FT_PER_E_RAD + 1.0;
	double z2 = alt2/FT_PER_E_RAD + 1.0;
	Vect3 v1 = z1*ll1, v2 = z2*ll2, vDiff = v2-v1;
	double distance = vDiff.VSize();

//Get the azimuth; one can ignore altitude for this.  Pick up range as well,
//converting it to earth radii for convenience.
	RangeAzimuth ra = rangeAzimuthFromFirstToSecond(ll1,ll2);
	azimuth = ra.Azimuth();
	double r = ra.Range()/NMI_PER_E_RAD;
	if(r < VERYSMALL){
		azimuth = 0.0;
		range = 0.0;
		elAngle = z1>z2? -90.0: 90.0;
		return;
	}
	if(r > LATLON_PI - VERYSMALL){
		azimuth = 0.0;
		range = NMI_PER_E_CIRC/2.0;
		elAngle = -90.0;
		return;
	}
//Get the elevation angle; Go through the earth if necessary.
	double theCosine = (-1.0*v1, vDiff)/(v1.VSize()*vDiff.VSize());
	if(theCosine > 1.0 - VERYSMALL) elAngle = 0.0;
	else if(theCosine < -1.0 + VERYSMALL) elAngle = 180.0;
	else elAngle = aTrigd(0.0, theCosine, 1);
	elAngle -= 90.0;

//Range: If the distance from the first to its horizon is larger than
//the distance to the second, or -elAngle < angle-to-horizon,
//use the straight line.
	double alpha1 = aTrigd(0.0, 1.0/z1, 1);
	double los1 = sqrt(z1*z1 - 1.0);
	if(los1 > distance || -elAngle < alpha1) range=distance;
	else{
		double alpha2 = aTrigd(0.0, 1.0/z2, 1);
		double los2 = sqrt(z2*z2 - 1.0);
//Convert angles to radians.
		double angleInRadians = (alpha1 + alpha2) / NMI_PER_E_RAD;
		range = los1+los2 + (r - angleInRadians);
	}
	range *= NMI_PER_E_RAD;
}

void getRelElAz(
	double az1, double elAng1, double az2, double elAng2,
	double &relEl12, double &relAz12
	)
{
//Rotate so that the first azimuth is 0.
	az2 = FMOD(az2-az1, 360.0);
//Form an orthonormal basis.  b1 points east, b2 points north and upward as per
//the elAng1, and b3 is the cross product in the direction of the
//relative angle of elevation.
	Vect3 b1(1,0,0), b2(0,cos_degrees(elAng1), sin_degrees(elAng1)), b3(b1*b2);

	double c=cos_degrees(elAng2);
	Vect3 u(c*cos_degrees(90.0 - az2), c*sin_degrees(90.0 - az2), sin_degrees(elAng2));
	double x=(u,b1), y=(u,b2), z=(u,b3), magXY=sqrt(x*x + y*y);
	relEl12 = aTrigd(z,0,2);
	if(magXY < VERYSMALL) relAz12=az2;
	else relAz12 = FMOD(90.0 - aTrigd(y/magXY,x/magXY,3), 360.0);
}

double whenHitAzimuth(
	const LatLonAlt &lla0, const LatLonAlt &lla1, const LatLonAlt &lla2,
	double azimuthToHit, int &returnCode
)
{
//Form (lat0,lon0)'s basis and compute coordinates of segment's endpoints.
//Then set u = endpt1 wrt (b1,b2,b3), and d = diff wrt (b1,b2,b3).
	Vect3 b1(lla0.ll);
	Vect3 b3(LatLon(lla0.ll.Lat()+90.0, lla0.ll.Lon()));
	Vect3 b2(b3*b1);
	Vect3 endpt1(lla1);
	Vect3 endpt2(lla2);
	Vect3 diff = endpt2-endpt1;
	Vect3 u((endpt1,b1),(endpt1,b2),(endpt1,b3));
	Vect3 d((diff,b1),(diff,b2),(diff,b3));

//We need (uy,uz) + p*(dy,dz) = q*(c,s), where
//(c,s) = (cos_degrees(90.0 - azimuthToHit), sin_degrees(90.0 - azimuthToHit)).  We solve this
//simple linear system with Cramer's rule.
	double uy = u.Y(), uz = u.Z(), dy = d.Y(), dz = d.Z();
	double c = cos_degrees(90.0 - azimuthToHit), s = sin_degrees(90.0 - azimuthToHit);
	double determinant = -dy*s - (-dz*c);

//If the determinant is zero, then (uy,uz) must be a positive multiple
//of (c,s).  If it's a negative multiple, then move along (dy,dz) until
//we get to 0.
	if(fabs(determinant) < FAIRLYSMALL){
		if(fabs(uy*s - uz*c) > FAIRLYSMALL){
			returnCode = 1001;
			return FLT_MAX;
		}
		if(fabs(uy) < FAIRLYSMALL && fabs(uz) < FAIRLYSMALL){
			returnCode = 1;
			return 0.0;
		}
		if(uy > 0.0 || uz > 0.0){
			returnCode = 2;
			return 0.0;
		}
		if(fabs(dy) < FAIRLYSMALL && fabs(dz) < FAIRLYSMALL){
			returnCode = 1002;
			return FLT_MAX;
		}
		double p;
		if(fabs(dy) >= FAIRLYSMALL) p = uy/dy;
		else p = uz/dz;
		returnCode = 3;
		return p;
	}
//After repulsing the zero-determinant dragon, solve for p and q.
	double p = (uy*s - c*uz)/determinant;
	double q = (-dy*uz - uy*(-dz))/determinant;
	if(q >= 0) returnCode = 5;
	else returnCode = 1003;
	return p;
}

void whenHitElAng(
	const LatLonAlt &lla0, const LatLonAlt &lla1, const LatLonAlt &lla2,
	double elAngToHit, double &p1, double &p2, int &returnCode
)
{
//Form (lat0,lon0)'s basis and compute coordinates of segment's endpoints.
//Then set u1 = endpt1 wrt (b1,b2,b3), u2 = endpt2 wrt (b1,b2,b3) and
//d = u2 - u1;
	Vect3 b1(lla0.ll);
	Vect3 b3(LatLon(lla0.ll.Lat()+90.0, lla0.ll.Lon()));
	Vect3 b2(b3*b1);

	Vect3 endpt1(lla1), endpt2(lla2);
	Vect3 u1((endpt1,b1),(endpt1,b2),(endpt1,b3));
	Vect3 u2((endpt2,b1),(endpt2,b2),(endpt2,b3));
	Vect3 d = u2 - u1;

//We need p for which u1 + p*d has the right elAng to (lat0,lon0,alt0).
//Defining P(p) to be  u1 + p*d, we are looking for p that satisfies:
//	(P(p).X() - alt0ER)/
//	sqrt((P(p).X() - alt0ER)^2 + P(p).Y()^2 + P(p).Z()^2) =
//		sin_degrees(elAngToHit).
//Set up and solve this quadratic in p.
//Define some convenience variables, and then a, b, c,
//where the quadratic is  a*p^2 + b*p + c = 0.
//Then just use the quadratic formula.
	double alt0ER = 1.0 + lla0.alt/FT_PER_E_RAD;
	double ux = u1.X();
	double uy = u1.Y();
	double uz = u1.Z();
	double dx = d.X();
	double dy = d.Y();
	double dz = d.Z();
	double sindElAng = sin_degrees(elAngToHit);
	double xi = sindElAng*sindElAng;
	double ximinus = 1.0 - xi;

	if(fabs(dx) < FAIRLYSMALL && fabs(ux-alt0ER) < FAIRLYSMALL){
		p1 = 0.0;
		p2 = 1.0;
		returnCode = 4;
	}
	else{
		double a = ximinus*dx*dx - xi*(dy*dy + dz*dz);
		double b = ximinus*2*dx*(ux - alt0ER) - 2*xi*(uy*dy + uz*dz);
		double c = ximinus*(ux-alt0ER)*(ux-alt0ER) - xi*(uy*uy + uz*uz);
		if(fabs(a) < FAIRLYSMALL){
			if(fabs(b) < FAIRLYSMALL){
				if(fabs(c) < FAIRLYSMALL){
					returnCode = 5;
					p1 = p2 = 0.0;
				}
				else{
					returnCode = 3006;
					p1 = p2 = FLT_MAX;
					return;
				}
			}
			else{
				returnCode = 7;
				p1 = p2 = -c/b;
			}
		}
		else{
			double discriminant, rootDiscriminant;
			if(fabs(b) > FAIRLYSMALL){
				discriminant = b*b - 4.0*a*c;
				if(fabs((4.0*a*c/(b*b)) - 1.0) < FAIRLYSMALL){
					discriminant = 0.0;
				}
			}
			else discriminant = -4.0*a*c;
			if(discriminant < 0.0){
				returnCode = 3008;
				p1 = p2 = FLT_MAX;
				return;
			}
			else{
				rootDiscriminant = sqrt(discriminant);
				returnCode = 9;
			}
			p1 = (-b + rootDiscriminant) / (2.0*a);
			p2 = (-b - rootDiscriminant) / (2.0*a);
		}
	}

//Make sure the roots solve the original equation.  Accuracy is hard to detect.
	double numer1 = ux + p1*dx - alt0ER;
	double numer2 = ux + p2*dx - alt0ER;
	double y1 = uy + p1*dy;
	double y2 = uy + p2*dy;
	double z1 = uz + p1*dz;
	double z2 = uz + p2*dz;
	double denom1 = sqrt(numer1*numer1 + y1*y1 + z1*z1);
	double denom2 = sqrt(numer2*numer2 + y2*y2 + z2*z2);
	double check1 = numer1 / denom1 - sindElAng;
	double check2 = numer2 / denom2 - sindElAng;
	if(fabs(check1) > 0.001) returnCode += 1000;
	if(fabs(check2) > 0.001) returnCode += 2000;
	return;
}

LatLonAlt endAt(LatLonAlt lla, double range, double azimuth, double elAng)
{
	Vect3 b1(lla.ll), b3(LatLon(lla.ll.Lat()+90,lla.ll.Lon())), b2(b3*b1);
	Vect3 v(lla.alt/FT_PER_E_RAD + 1.0, 0.0, 0.0);
	Vect3 u(
		sin_degrees(elAng),
		cos_degrees(elAng)*cos_degrees(90.0 - azimuth),
		cos_degrees(elAng)*sin_degrees(90.0 - azimuth)
		);
	Vect3 d = (range/NMI_PER_E_RAD) * u;
	Vect3 endVect = v + d;
	Vect3 w = endVect.X() * b1 + endVect.Y() * b2 + endVect.Z() * b3;
	LatLon ll(w);
	double alt = (w.VSize() - 1.0) * FT_PER_E_RAD;
	return LatLonAlt(ll,alt);
}

