// TestOPSM.C is a test program that tests if the object proxy state manager
// is working.

#include <stdlib.h>
#include <stream.h>
#include <signal.h>

#include "StateManager.H"
#include "StopWatch.H"

//#define NOISY

STATE_MANAGER *StateManager = NULL;

/* This sigintfunc() call is to make sure that, when the OPSM is
** killed by pressing <ctrl>-c, that it will delete the StateManager.
** THis has the effect of sending a hangup message to the HostRouter
** and is the "best" way to hang up the line presently.  Simply
** include it as a default and ignore it.
*/

void sigintfunc(int) {
  cout << "Interupt" << endl;
  signal(SIGINT, sigintfunc);
  if (StateManager) delete StateManager;
  StateManager = NULL;
  exit(1);
}

/*
** Enumerate the types of events you anticipate.  We generally finish the
** list by using N_EVENTS so that this # always has the correct number of
** events
*/

enum {
  DETECT_OBJECT_EVENT_ID,
  REFLECT_ATTRIBUTES_ID,
  RESCHEDULED_EVENT_ID,
  N_EVENTS
};

/* This event is goign to be scheduled by the OPSM after a detect
** object event has occured
*/

class RESCHEDULED_EVENT : public STATE_MANAGER_EVENT {
  private:
  protected:
  public:

    RESCHEDULED_EVENT(){;}

    virtual void Process() {
      cout << "Hello from RESCHEDULED_EVENT at time " << time_tag << endl;
    }

};

/*
** This is the event that will be launched when we first detect an object.
** You will see later on how we are able to link this event to that message.
*/

class DETECT_OBJECT_EVENT : public STATE_MANAGER_EVENT {
  private:
  protected:
  public:

    DETECT_OBJECT_EVENT(){;}

  /*
  ** Every event must have a process() method call.  Notice that this event
  ** begins scheduling the event RESCHEDULED_EVENT at 5 second intervals
  ** after this event has been processed.
  */
    virtual void Process() {
      cout << "Hello from DETECT_OBJECT_EVENT at time " << time_tag << endl;
      cout <<" We just got this detection of object id "<<GlobalId<<endl;
      RESCHEDULED_EVENT *ReschedEvent = new RESCHEDULED_EVENT();
      /*
      ** Sometimes you can get negative times as small as -1x10^20
      */
      if (time_tag < 0.0){
	ReschedEvent->ScheduleEvent(0);
      }
      else{
	ReschedEvent->ScheduleEvent(time_tag + 5);
      }
    }

};

/*
** This event will be launched every time that we get updated information
** about a given object
*/

class REFLECT_ATTRIBUTES : public STATE_MANAGER_EVENT {
  private:
  protected:
  public:

    REFLECT_ATTRIBUTES(){;}

    virtual void Process() {
      cout << "Hello from REFLECT_ATTRIBUTES at time " << time_tag << endl;
      cout <<" The object with ID "<<GlobalId<<" just changed something"<<endl;
    }

};

/* These are simply function calls that return some new events.  These
** are used so that SPEEDES can put them on free lists and not spend a
** great deal of time newing memory and deleting memory.
*/

void *CreateDETECT_OBJECT_EVENT(int n) {return new DETECT_OBJECT_EVENT[n];}
void *CreateREFLECT_ATTRIBUTES(int n) {return new REFLECT_ATTRIBUTES[n];}
void *CreateRESCHEDULED_EVENT(int n) {return new RESCHEDULED_EVENT[n];}

int main() {
  DATA_PARSER Parser("speedes.par");
  double Time = 0;
  OBJECT_PROXY *ObjectProxy;
  EM_HOST_USER * Em;
  STOPWATCH StopWatch;

  StopWatch.UseWallTime();

/*
** register sigintfunc to be call if a <ctrl>-c is ever pressed.
*/

  signal(SIGINT, sigintfunc);

/*
** Create a new STATE_MANGER from the parser and don't lag the simulation
** by more than 10 seconds.
*/

  StateManager = new STATE_MANAGER(&Parser, 10.0);

/*
** Grab the EM_HOST_USER for this OPSM
*/

  Em = StateManager->GetEmHostUser();

/*
** Subscribe to all ship objects
*/

  StateManager->SubscribeType("Ship", NULL);
//  StateManager->SubscribeObject("Ship_0", NULL);

/*
** Tell the OPSM how many event types may be scheduled.
*/

  StateManager->DefineNumberOfEvents(N_EVENTS);

/*
** Plug in the events (these are just test events right now)
*/

  StateManager->DefineEvent("DISCOVER_OBJECT",
			    DETECT_OBJECT_EVENT_ID,
			    CreateDETECT_OBJECT_EVENT,
			    sizeof(DETECT_OBJECT_EVENT),
			    100);

  StateManager->DefineEvent("REFLECT_ATTRIBUTES",
			    REFLECT_ATTRIBUTES_ID,
			    CreateREFLECT_ATTRIBUTES,
			    sizeof(REFLECT_ATTRIBUTES),
			    100);

  StateManager->DefineEvent("RESCHEDULED_EVENT",
			    RESCHEDULED_EVENT_ID,
			    CreateRESCHEDULED_EVENT,
			    sizeof(RESCHEDULED_EVENT),
			    100);

/*
** Get the initial time set up
*/

  StateManager->GoToTime(Time);
  Time = StateManager->GetGrantedTime();

/*
** Loop forever until SPEEDES exits
*/

  while (Em->SpeedesExecuting()) {

    StopWatch.Reset();
    StopWatch.StartTimer();

/*
** Print out how many object proxies there are at our time
*/

#ifdef NOISY
    cout << endl;
    cout << "-------------------------------------------------------"
	 << "-------------------------------------------------------" << endl;
    cout << endl;
    cout << "Number of Proxies at time " << Time << " = "
	 << StateManager->GetNproxies() << endl;
#else
    cout << "Time = " << Time << endl;
#endif

/*
** Get the first object proxy of the type Ship
*/

    ObjectProxy = StateManager->GetFirstProxy("Ship");

/*
** Now iterate through all the Ship objects and print out the
** information about them
*/

    while (ObjectProxy) {
#ifdef NOISY
      ObjectProxy->PrintStream();
#endif
      ObjectProxy = StateManager->GetNextProxy();
    }

/*
** The StateManager can roll foward or backward in time.  Here we are
** going back to time 0 to find some information at that point.  You
** can think of the OPSM as a big tape recording of the simulation up
** to the current time.
*/

    StateManager->GoToTime(0.0);
    StateManager->GoToTime(Time/2);

/*
** Advance time by spinning until the actual time really does advance. In other
** applications, they might not want to wait here (e.g., graphics that need
** to respond to mouse clicks, etc.).
*/

//#define GOTOTIME
//#define ADVANCE_REALTIME
#define ADVANCE_FAST_AS_POSSIBLE

#ifdef GOTOTIME
    sleep(1);
    Time += StopWatch.StopTimer();
    StateManager->GoToTime(Time);
#endif

#ifdef ADVANCE_REALTIME
    sleep(1);
    double Time1 = Time;
    while (Time <= Time1) {
      Time = StateManager->AdvanceTime(Time + StopWatch.StopTimer());
    }
#endif

#ifdef ADVANCE_FAST_AS_POSSIBLE
    sleep(1);
    Time = StateManager->AdvanceTime();
#endif

  }

/*
** Delete the StateManager so that we hang up politely with the HostRouter
*/

  delete StateManager;

}
