// O_Ship.C source file

#include "O_Ship.H"
#include "ScheduleEvent.H"
#include "E_CheckRange.H"

O_Ship *Ship_OBJECT_MANAGER = NULL;

/******************************************************************************
* O_Ship constructor                                                          *
******************************************************************************/
O_Ship::O_Ship() {
  int i,j;
  int offset;
  void *Mark;

/*
** Set the global variable for the ship manager so that it can be used by
** others
*/

  Ship_OBJECT_MANAGER = this;

/*
** Create the parser that reads in the Ship.par file. This will define a
** list of objects (ships) that need to be created. The Ships.par file also
** describes what components are loaded up on each ship.
*/

  cout << "Reading in Ship.par file" << endl;
  Parser = new DATA_PARSER("Ship.par");
  Mark = Parser->GoTo("Objects", NULL);

/*
** We will card-deal the ships to the different nodes. The deal_me function
** takes a total number of objects (N_TOT ships) and figures out how many
** local objects my node needs to create. The offset is how many objects
** should be skipped by my node before creating the first one. Each node
** will have a different offset. For example, if there are 10 nodes, and
** from previous objects managers, the last simobj was created on node 4,
** then the offsef for node 5 would be zero. Likewise, node 6 would have
** an offset of 1, etc.. Node 4 would have an offset of 9. The offset is
** calculated by the deal-me function. We will create our ships by using
** the offset to find out what our first object is, and then create every
** N_NODES one thereafter.
*/

  N_TOT = Parser->GetNumReferences(Mark, "OBJECT");
  N_LOC = deal_me(N_TOT, offset);

/*
** Create an array of Ships. Remember that since we are creating them as
** an array, we won't be able to do dynamic object deletion of individual
** ships. This should be changed later if we need to create and delete
** ships dynamically. When creating the ships, SPEEDES (in the simobj base
** class) automatically generates a LOCAL_ID and GLOBAL_ID and stores them
** in the simobj base class.
*/

  LocalObjects = new S_Ship[N_LOC];

  cout << "NODE " << NODE << " Created "
       << N_LOC << " out of " << N_TOT << " Ship objects" << endl;

/*
** Work through the parser and create my ships (using the offset and then
** creating every N_NODES one thereafter). The ships get initialized by
** setting a unique seed that is dependent on the replication number (which
** is defined in O_PARAMETERS - which we have inherited from) and the index
** into the ships that are listed in the Objects group in the par file.
** We also set the object's name and then call its init function (where
** we pass in the parser that has already been set up to be in the correct
** ship's set.
*/

  j=0;
  Parser->GoToFirstReference(Mark, "OBJECT");

  for (i=0; i<N_TOT; i++) {
    if ((i-offset+N_NODES)%N_NODES == 0) {
      LocalObjects[j].set_SEED(10000*Replication + 12345 + i);
      LocalObjects[j].set_NAME(Parser->GetReferenceName());
      LocalObjects[j].init(Parser);
      j++;
    }
    Parser->GoToNextReference(Mark, "OBJECT");
  }

/*
** Tell SPEEDES how many simobjs we have created (that's what the reserve
** method does - it creates an array of points to simobjs to store my ships).
** Then, we pass in each ship. Note that we don't have to tell SPEEDES what
** our local Id is because SPEEDES automatically set the local Id for the
** ship. When we are done plugging in our simobjs, we need to update the
** TOTAL_OBJECTS variable so that SPEEDES knows how many total objects have
** been created (this may be automated by SPEEDES later).
*/

  reserve_n_objects(N_LOC);
  for (i=0; i<N_LOC; i++) define_object(&LocalObjects[i]);
  TOTAL_OBJECTS += N_TOT;

/*
** Tell SPEEDES to make our ships interactive. What the set_interact method
** does is to tell the Host Router the names of our ships.
*/

  set_interact();


  cout << "Done constructing Ships" << endl;

#ifdef VALIDATE_DDM
  if (N_NODES==1){
    SpSimTime event_time(0.0);
    event_time.SetPriority(MaxPriority1);
    ScheduleEvent(event_time, CheckRange, get_OBJECT_TYPE(), get_LOCAL_ID(), get_NODE());
  }
#endif
}

/******************************************************************************
* init_events: Initializes the events for my objects                          *
******************************************************************************/
void O_Ship::init_events() {

/*
** Schedule all of the initial events for each ship. Currently, we are
** scheduling the BROADCAST_PROXY event which broadcasts my proxy to all of
** the entities. This will not be required once the declaration management
** and data distribution management services are completed. We also call the
** InitEvents method for each ship (which initializes events for the
** different components on our ship).
*/

  cout << "Initializing events for Ship objects" << endl;

  for (int i=0; i<N_LOC; i++) {
    LocalObjects[i].InitEvents();
  }

}

/******************************************************************************
* terminate: executed at the end of the simulation                            *
******************************************************************************/
void O_Ship::terminate(double /*tend*/) {
  cout << "Terminating Ship Object Manager" << endl;
}
