// P_RadarScan.C

#include "P_RadarScan.H"

#include "S_Entity.H"
#include "EntityProxy.H"
#include "process.H"
#include "RB.H"
#include "F_ProxyItem.H"

extern int DebugRadarScan;

extern int FindRemoteProxy(S_ENTITY *entity, F_PROXY_ITEM *proxy_item);

#define _DEBUG

/*
** Local variables (LV) that are persistent with this process are defined
** here. The PROCESS_LV and BEGIN_PROCESS macros really define the variables
** in a locally defined class called LV. This is why you need to use the LV->
** sytax for accessing these variables.
*/

PROCESS_LV (P_RadarScan, process)
RB_int HappyVariable;
int LoopForever;
RADAR_COMPONENT *RadarComponent;
double RadarRange;
double ScanTime;
RB_BINARY_TREE *DetectableProxies;
RB_BINARY_TREE Detections;
SpCounterSem *NumDetectableProxies;
RADAR_SENSOR_MODEL *RadarSensorModel;
RB_int seed;
BEGIN_PROCESS (P_RadarScan, process, 2)

/*
** These variables are defined as temporary. You should not assume their
** values are set correctly after waking up after time has elapsed. The
** reason we define temporary variables is for efficiency. Loop variables,
** for example, which do not pass time in the loop are much more efficient
** than their rollbackable counterparts.
*/

S_ENTITY *Entity;
ENTITY_PROXY *EntityProxy;
ENTITY_PROXY *RemoteEntityProxy;
double X1[3], V1[3], A1[3];
double X2[3], V2[3], A2[3];
double R[3];
double Range;
F_PROXY_ITEM *ProxyItem;
M_RadarScan *radarScan;
int SuccessFlag;
RADAR_DETECTION *radarDetection;
int sd;

/*
** The first thing we do when entering into a process is to access whatever
** data was passed in through the MESSAGE and store it in the LV variables.
** This is only done once, when the process is first entered. After than,
** there is no user data passed in the message.
*/

radarScan = (M_RadarScan *)MESSAGE;
LV->RadarComponent = radarScan->RadarComponent;
LV->RadarRange = (LV->RadarComponent)->GetRange();
LV->DetectableProxies = (LV->RadarComponent)->GetDetectableProxies();
LV->NumDetectableProxies = (LV->RadarComponent)->GetNumDetectableProxies();
LV->ScanTime = (LV->RadarComponent)->GetScanTime();
LV->RadarSensorModel = (LV->RadarComponent)->GetRadarSensorModel();

LV->RadarComponent = radarScan->RadarComponent;

LV->seed = SIMOBJ->get_SEED() + (LV->RadarComponent)->GetSensorId();

/*
** Here we loop forever (note that the LoopForever variable must be a of the
** LV type since it is used even when time passes.
*/

LV->LoopForever = 1;
while (LV->LoopForever) {

/*
** Here we wait for the NumDetectableProxies (a counter semaphore) to be non
** zero. What is really happening here is we go to sleep until there is at
** least one detectable proxy in our field of view. The RadarPrecomp process
** is responsible for controling this counter semaphore and populating the
** Detectable Proxy list.
*/

  WAIT_FOR_COUNTER(2, *(LV->NumDetectableProxies), -1, SuccessFlag);

/*
** First get the position, velocity, and acceleration of my object
** (done through my proxy)
*/

  Entity = (S_ENTITY *)SIMOBJ;

  EntityProxy = (ENTITY_PROXY *)Entity->GetObjectProxy();
  EntityProxy->Get_Position (X1, V1, A1, ECR, TIME_TAG);

/*
** Loop through all of the remote proxies and determine their position,
** velocity, and acceleration.
*/

  ProxyItem = (F_PROXY_ITEM *)(LV->DetectableProxies)->GetFirstElement();
  while (ProxyItem) {
    if (!FindRemoteProxy(Entity, ProxyItem)){
      ProxyItem = (F_PROXY_ITEM *) ++(*(LV->DetectableProxies));
      continue;
    }
    RemoteEntityProxy = (ENTITY_PROXY *)ProxyItem->GetObjectProxy();
    RemoteEntityProxy->Get_Position (X2, V2, A2, ECR, TIME_TAG);

/*
** Compute the range between myself and the remote entity.
*/

    Difference(X1,X2,R);
    Range = Magnitude(R);

#ifdef COMMENTOUT
    RB_cout << (LV->RadarComponent)->GetRadarType() << ": "
	    << "Range between "
	    << EntityProxy->GetName()
	    << " and "
	    << RemoteEntityProxy->GetName()
	    << " = " << Range
	    << " at time " << TIME_TAG.GetTime()
	    << endl;
#endif

    if (Range < LV->RadarRange) {

/*
** Here is where we might feed this into a sensor model to form detections
** (or we might collect them into another list and then feed the whole list
** into the sensor).
*/

      radarDetection = RB_NEW_RADAR_DETECTION();

      radarDetection->SetTime(TIME_TAG);

      radarDetection->SetTrueSensorPosition(X1);
      radarDetection->SetTrueSensorVelocity(V1);

      radarDetection->SetTrueTargetPosition(X2);
      radarDetection->SetTrueTargetVelocity(V2);

      radarDetection->SetTargetCrossSection(1.0);

      radarDetection->SetGlobalId(RemoteEntityProxy->GetGlobalId());
      radarDetection->SetNode(RemoteEntityProxy->GetNode());
      radarDetection->SetType(RemoteEntityProxy->GetObjectType());
      radarDetection->SetLocalId(RemoteEntityProxy->GetLocalId());

      (LV->Detections).Insert(radarDetection, radarDetection->GetGlobalId());

    }else{

/*
** Print out a warning message that for some reason, we have a proxy in
** our detectable proxy list that shouldn't be there. If a precomputation
** is not done exactly right (it currently isn't), then we might get an
** occasional warning.
*/

      if (DebugRadarScan) {
	RB_cout << "Warning, object "
		<< RemoteEntityProxy->GetName()
		<< " is out of range" << endl;
      }

    }

    ProxyItem = (F_PROXY_ITEM *) ++(*(LV->DetectableProxies));
  }

/*
** Feed the detections into our radar sensor model
*/

  sd = LV->seed;
  (LV->RadarSensorModel)->FormDetections(&(LV->Detections), sd);
  LV->seed = sd;

/*
** For now, delete the detections. Later, we will want to feed these to the
** tracking components.
*/

#ifdef COMMENTOUT
  RB_cout << "Detecting " << (LV->Detections).GetNumElements()
	  << " entities" << endl;
#endif

  while (radarDetection = (RADAR_DETECTION *)
	 (LV->Detections).RemoveFirstElement()) {
    RB_DELETE_RADAR_DETECTION(radarDetection);
  }

/*
** Now we are done with our sensor scan, so wait the appropriate amount of
** time before doing another scan. In this case, we are waiting a constant
** amount of time each time, using the sensor's constant scan time.
*/

  WAIT(1,LV->ScanTime);

}

END_PROCESS (P_RadarScan, process)
