// RadarSensor.C is a radar model for generic radar sensors

#include "RadarSensorModel.H"
#include "random.H"
#include "Convert.H"
#include "VectorOperations.H"

RADAR_SENSOR_MODEL::RADAR_SENSOR_MODEL() {
  Range = 0.0;
  ScanTime = 0.0;
  SigmaRange = 0.0;
  SigmaRangeRate = 0.0;
  SigmaAngle = 0.0;
}

void RADAR_SENSOR_MODEL::FormDetections(RB_BINARY_TREE *detections,
					int &seed) {
  RADAR_DETECTION *radarDetection;
  C_RANDOM random;
  random.set_seed(seed);
  double *sensorPosition;
  double *targetPosition;
  double *sensorVelocity;
  double *targetVelocity;
  double losVector[3];
  double VelocityVector[3];
  double u1[3];
  double u2[3];
  double u3[3];
  double range;
  double rangeRate;
  double deltaRange;
  double deltaRangeRate;
  double deltaAngle;

/*
** Loop over all of the detections (which are initialized with truth) and
** compute detections using random numbers to smear the data.
*/

  radarDetection = (RADAR_DETECTION *)detections->GetFirstElement();
  while (radarDetection) {

/*
** Get the position of the sensor and the target
*/

    sensorPosition = radarDetection->GetTrueSensorPosition();
    targetPosition = radarDetection->GetTrueTargetPosition();

/*
** Compute the line of sight vector and then the range. In this model, we
** always detect the object if it is in range, so we set the valid detection
** flag. Later, we should add signal return criteria using the radar equation
** along with random numbers (since it should really be statistical)
*/

    Difference(targetPosition, sensorPosition, losVector);
    range = Magnitude(losVector);
    Normalize(losVector);

    if ((range < Range) && (0.001 < range)) {

/*
** This is the case where we have a valid detection so we need to smear
** the detection measurement according to the parameters that describe this
** radar.
*/

      radarDetection->SetGoodDetection();

      random.set_gauss(SigmaRange, 0.0);
      deltaRange = random.get_random_gauss();
      range += deltaRange;
      if (range < 0.0) range = 0.0;

//      cout << "Delta Range = " << deltaRange << endl;

      radarDetection->SetMeasuredRange(range);
      radarDetection->SetSigmaRange(SigmaRange);

/*
** Compute the range rate
*/

      sensorVelocity = radarDetection->GetTrueSensorVelocity();
      targetVelocity = radarDetection->GetTrueTargetVelocity();

      Difference(targetVelocity, sensorVelocity, VelocityVector);
      rangeRate = DotProduct(VelocityVector, losVector);
      random.set_gauss(SigmaRangeRate, 0.0);
      deltaRangeRate = random.get_random_gauss();
      rangeRate += deltaRangeRate;

//      cout << "Delta Range Rate = " << deltaRangeRate << endl;

      radarDetection->SetMeasuredRangeRate(rangeRate);
      radarDetection->SetSigmaRangeRate(SigmaRangeRate);

/*
** Compute the normalized directional vector as u1 (note, u2 and u3 are
** the other vectors that form an orthogonal coordinate system.
*/

      VectorCopy(u1, losVector);
      FormCoordinateSystem(u1, u2, u3);

      random.set_float_limits(0.0, PI);
      deltaAngle = random.get_random_float();
      Rotate1(u1, u2, u3, deltaAngle);

      random.set_gauss(SigmaAngle, 0.0);
      deltaAngle = random.get_random_gauss();

      Rotate2(u1, u2, u3, deltaAngle);

      if (!CheckCoordinateSystem(u1, u2, u3)) {
	cout << "Bad Coordinate System after rotation" << endl;
      }

      radarDetection->SetMeasuredDirectionVector(u1);
      radarDetection->SetSigmaAngle(SigmaAngle);

//      radarDetection->Print(cout);

    }else{

/*
** This is the case where we do not have a valid detection
*/

      radarDetection->ResetGoodDetection();

    }

/*
** Get the next detection
*/

    radarDetection = (RADAR_DETECTION *) ++(*detections);

  }

/*
** Return the seed
*/

  seed = random.get_seed();

}
