// SpEnterExitRange.C file provides a function that computes when two objects
// are within a specified range. Four states can be returned from this
// function:
//
// (1) RANGE_NO_SOLUTION - objects do not come within specified range ever
// (2) RANGE_COMPUTE_LATER - if there is a solution, it is far into the future
// (3) RANGE_ENTER - the approaching objects come within the specified range
// (4) RANGE_EXIT - the objects are moving out of range

#include "SpEnterExitRange.H"

//#define DEBUG_ENTER_EXIT_RANGE

void SpEnterExitRange(EntityProxy *entityProxy1,
			EntityProxy *entityProxy2,
			double maxSpeed1,
			double maxSpeed2,
			double startTime,
			double range,
			double maxPredictAheadTime,
			double &time,
			int &flag) {

  double R1[3], V1[3], A1[3];
  double R2[3], V2[3], A2[3];
  double R[3], V[3], A[3];
  double loBound1, hiBound1;
  double loBound2, hiBound2;
  double hiBound, loBound;
  double distance;
  double jumpAhead;
  double a,b,c;
  double disc;
  double dt1, dt2;
  int status1, status2;

  double maxSpeed = maxSpeed1 + maxSpeed2;
  double maxTime = startTime + maxPredictAheadTime;
  time = startTime + 0.001;
  int numIterations = 0;

#ifdef DEBUG_ENTER_EXIT_RANGE
  cout << ">>> SpEnterExitRange start time " << time << " for "
       << entityProxy1->GetObjectName() << " & "
       << entityProxy2->GetObjectName() << endl;
#endif

  while (1) {

    if (numIterations > 1000) {
      flag = RANGE_COMPUTE_LATER;
      return;
    }

    numIterations++;

    entityProxy1->Get_Position(R1, V1, A1, ECR, time);
    entityProxy2->Get_Position(R2, V2, A2, ECR, time);

    Difference(R1, R2, R);
    distance = Magnitude(R);

    jumpAhead = fabs(distance-range) / maxSpeed;
    time += jumpAhead;

    if (maxTime < time) {
      if (time >= 1.0e20) {
	flag = RANGE_NO_SOLUTION;
      }else{
	flag = RANGE_COMPUTE_LATER;
      }
      return;
    }

    if (jumpAhead < maxPredictAheadTime) {

      status1 = entityProxy1->GetPositionTimes(time, loBound1, hiBound1);
      if (!status1) {
	cout << "Error, could not find entity 1 position time" << time << endl;
	entityProxy1->PrintStream();
      }

      status2 = entityProxy2->GetPositionTimes(time, loBound2, hiBound2);
      if (!status2) {
	cout << "Error, could not find entity 2 position time" << time << endl;
	entityProxy2->PrintStream();
      }

      loBound = time;

      if (hiBound1 < hiBound2) {
	hiBound = hiBound1;
      }else{
	hiBound = hiBound2;
      }

      while (1) {

	if (numIterations > 1000) {
	  flag = RANGE_ERROR;
	  return;
	}

	numIterations++;

	entityProxy1->Get_Position(R1, V1, A1, ECR, time);
	entityProxy2->Get_Position(R2, V2, A2, ECR, time);

	Difference(R1, R2, R);
	Difference(V1, V2, V);
	Difference(A1, A2, A);

	a = DotProduct(A,R) + DotProduct(V,V);
	b = 2*DotProduct(V,R);
	c = DotProduct(R,R) - range*range;

	disc = (b*b) - (4*a*c);

	if (0 < disc) {
	  disc = sqrt(disc);
	  if (0 < a) {
	    dt1 = (-b - disc) / (2*a);
	    dt2 = (-b + disc) / (2*a);
	  }else{
	    dt1 = (-b + disc) / (2*a);
	    dt2 = (-b - disc) / (2*a);
	  }

	  if ((loBound <= time+dt1) && (time+dt1 < hiBound)) {
	    time += dt1;
	    flag = RANGE_ENTER;
	    if (fabs(dt1) < 0.000001) {
#ifdef DEBUG_ENTER_EXIT_RANGE
	      cout << ">>> Found enter at time " << time
		   << " with " << numIterations << " iterations " << endl;
#endif
	      return;
	    }
	  }else{
	    if ((loBound <= time+dt2) && (time+dt2 < hiBound)) {
	      time += dt2;
	      flag = RANGE_EXIT;
	      if (fabs(dt2) < 0.000001) {
#ifdef DEBUG_ENTER_EXIT_RANGE
		cout << ">>> Found exit at time " << time
		     << " with " << numIterations << " iterations" << endl;
#endif
		return;
	      }
	    }else{
	      time = hiBound;
	      break;
	    }
	  }

	}else{
	  time = hiBound;
	  break;
	}

      }

    }

  }

}


