// VectorOperations.C file defines more vector operations. These should be
// merged into the Convert.C file later on

#include <math.h>
#include <iostream.h>

#include "VectorOperations.H"
#include "Convert.H"

void FormCoordinateSystem(double u1[3], double u2[3], double u3[3]) {
  int i,j;

/*
** First, make sure that the first vector (which is the input vector) is
** normalized.
*/

  Normalize(u1);

/*
** Now, find the index with the smallest value for the u1 vector and use
** that to define a vector which will be used in the cross product. Note,
** that we call this temporary vector u2 even though we will recompute u2
** later on.
*/

  j = 0;
  for (i=1; i<3; i++) {
    if (fabs(u1[i]) < fabs(u1[j])) {
      j = i;
    }
  }

  for (i=0; i<3; i++) u2[i] = 0.0;
  u2[j] = 1;

/*
** Compute the cross product between u1 and u2 which will provide a vector
** is guaranteed to be perpendicular to u1. We then normalize u2.
*/

  CrossProduct(u1, u2, u3);
  Normalize(u3);

/*
** Finally, compute the real u2 as the cross product between u3 and u1. By
** definitiion, u2 should be normalized.
*/

  CrossProduct(u3, u1, u2);

/*
** Verify that we did this right.
*/

  if (!CheckCoordinateSystem(u1, u2, u3)) {
    cout << "Bad Coordinate System" << endl;
  }

}


int CheckCoordinateSystem(double u1[3], double u2[3], double u3[3]) {
  double mag;
  double dot;
  double cross[3];
  int status = 1;

/*
** First, make sure that the unit vectors are normalized
*/

  mag = Magnitude(u1);
  if ((mag < 0.999999) || (1.000001 < mag)) {
    cout << "Error, bad coordinate system: |u1| = " << mag;
    status = 0;
  }

  mag = Magnitude(u2);
  if ((mag < 0.999999) || (1.000001 < mag)) {
    cout << "Error, bad coordinate system: |u2| = " << mag;
    status = 0;
  }

  mag = Magnitude(u3);
  if ((mag < 0.999999) || (1.000001 < mag)) {
    cout << "Error, bad coordinate system: |u3| = " << mag;
    status = 0;
  }

/*
** Make sure that the dot products are zero
*/

  dot = DotProduct(u1, u2);
  if ((dot < -0.000001) || (0.000001 < dot)) {
    cout << "Error, bad coordinate system: u1.u2 = " << dot;
    status = 0;
  }

  dot = DotProduct(u2, u3);
  if ((dot < -0.000001) || (0.000001 < dot)) {
    cout << "Error, bad coordinate system: u2.u3 = " << dot;
    status = 0;
  }

  dot = DotProduct(u3, u1);
  if ((dot < -0.000001) || (0.000001 < dot)) {
    cout << "Error, bad coordinate system: u3.u1 = " << dot;
    status = 0;
  }

/*
** Finally, check the cross products to make sure that the vectors are
** orthogonal and follow the right hand rule (i.e, u1 X u2 = u3)
*/

  CrossProduct(u1,u2,cross);
  mag = DotProduct(cross,u3);
  if ((mag < 0.999999) || (1.000001 < mag)) {
    cout << "Error, bad coordinate system: (u1 X u2).u3 = " << mag;
    status = 0;
  }

  return status;

}

/*
** Rotate coordinate system about one of the axis
*/

void Rotate1(double u1[3], double u2[3], double u3[3], double angle) {
  double v2[3];
  double v3[3];
  double cosAngle;
  double sinAngle;
  int i;

  cosAngle = cos(angle);
  sinAngle = sin(angle);

  for (i=0; i<3; i++) {
    v2[i] = u2[i]*cosAngle + u3[i]*sinAngle;
    v3[i] = u3[i]*cosAngle - u2[i]*sinAngle;
  }

  VectorCopy(u2, v2);
  VectorCopy(u3, v3);
}

void Rotate2(double u1[3], double u2[3], double u3[3], double angle) {
  Rotate1(u2, u3, u1, angle);
}

void Rotate3(double u1[3], double u2[3], double u3[3], double angle) {
  Rotate1(u3, u1, u2, angle);
}
