// servman.C method file

#include "SpStdio.H"
#include "servman.H"
#include "servobj.H"
#include "parser.H"
#include "ScheduleEvent.H"
#include "SpBasetype.H"

EVENT_TYPE (SERVE_EVENT);

/************************************************************************
* C_SERVMAN : construct a servman object				*
************************************************************************/
C_SERVMAN::C_SERVMAN() {
  if (SpPoDataBase::DoNotAllocateMemoryInConstructor){
    cout<<"Creating a servman with persistence"<<endl;
  }
  else{
    C_BASETYPE *parameters;
    C_SERVOBJ *servobj;
    int i;
    char *s;
    int temp1;
    int temp2;
    int gid;
    int nd;
    int lid;

// initialize the object manager with data from the input parser file

    C_PARSER parser("server.par");

    parameters = parser.get_basetype("parameters");
    n_servers = parameters->get_int("n_servers");
    n_customers_server = parameters->get_int("n_customers_server");
    fully_interconnected = parameters->get_logical("fully_interconnected");
    beta_n1_serve = parameters->get_int("beta_n1_serve");
    beta_n2_serve = parameters->get_int("beta_n2_serve");
    beta_n1_travel = parameters->get_int("beta_n1_travel");
    beta_n2_travel = parameters->get_int("beta_n2_travel");
    const_serve_time = parameters->get_logical("const_serve_time");
    const_travel_time = parameters->get_logical("const_travel_time");
    serve_time = parameters->get_float("serve_time");
    travel_time = parameters->get_float("travel_time");

    // initialize the random number generators (there is some work being done
    // here in the random number generators to efficiently compute beta
    // distributions. It's best to do this once here and then pass the random
    // number generators to each server in the initialization loop. We are
    // keeping things general in case we want to modify this program later to
    // let each server have its own statistics.

    random_serve.set_beta(beta_n1_serve, beta_n2_serve);
    random_travel.set_beta(beta_n1_travel, beta_n2_travel);

    // figure out how many objects are for my node

    N_TOT = n_servers;
    N_LOC = (N_TOT-NODE) / N_NODES;
    if ((N_TOT-NODE) % N_NODES) N_LOC++;

    // create the local objects for my node

    servobj = PO_NEW_ARRAY_C_SERVOBJ(N_LOC);

    // calculate the id of my first server (assuming circular connectivity)
    // this is kind of tricky but it seems to work

    temp1 = N_TOT/N_NODES;
    temp2 = N_TOT%N_NODES;
    if (NODE <= temp2) {
      starting_server = (temp1+1)*NODE;
    }else{
      starting_server = (temp1+1)*temp2 + temp1*(NODE-temp2);
    }

    // initialize my local objects

    for (i=0; i<N_LOC; i++) {

      gid = starting_server + i;		// server global id

      servobj[i].set_GLOBAL_ID(gid);	// set the server's global id

      servobj[i].set_SEED(gid+12345);	// set the server's seed

      s = new char[20];
      sprintf(s,"SERVER%10.10d",gid);
      servobj[i].set_NAME(s);		// set the server's name

      servobj[i].set_random_generators(&random_serve, &random_travel);

      get_nodid(gid, nd, lid);
      if ((nd != NODE) || (lid != i)) {
	fprintf(stderr, "Error, bad node id algorithm\n");
      }

    }

    //...... tell SPEEDES about the serv objects

    TOTAL_OBJECTS += N_TOT;
    reserve_n_objects(N_LOC);
    for (i=0; i<N_LOC; i++) define_object(&servobj[i]);
    set_interact();

    if (N_LOC) {
      cout << "SERVMAN created objects " << servobj[0].get_GLOBAL_ID()
	   << " to " << servobj[N_LOC-1].get_GLOBAL_ID()
	   << " (" << N_TOT << " total) on node " << NODE << "\n";
    }else{
      cout << "SERVMAN created 0 objects (" << N_TOT
	   << " total) on node " << NODE << "\n";
    }
  }
}


void C_SERVMAN::ResetObjectForRestart()
{
  random_serve.reset();
  random_serve.set_beta(beta_n1_serve, beta_n2_serve);
  random_travel.reset();
  random_travel.set_beta(beta_n1_travel, beta_n2_travel);
}


/************************************************************************
* init_events : initialize events 				        *
************************************************************************/
void C_SERVMAN::init_events() {
  int i,j;

  for (i=0; i<N_LOC; i++) {
    for (j=0; j<n_customers_server; j++) {
      schedule(0.0, SERVE_EVENT, OBJECT_TYPE, i, NODE);
    }
  }

}

/************************************************************************
* get_nodid : get the node and local id of a given server	        *
************************************************************************/
void C_SERVMAN::get_nodid(int server, int &nd, int &lid) {
  int temp1;
  int temp2;
  int temp3;

  temp1 = N_TOT/N_NODES;
  temp2 = N_TOT%N_NODES;

  if (server < (temp1+1)*temp2) {
    nd = server/(temp1+1);
    lid = server%(temp1+1);
  }else{
    temp3 = server-(temp1+1)*temp2;
    nd = temp2 + temp3/temp1;
    lid = temp3%temp1;
  }

}

/************************************************************************
* terminate : termination method 				        *
************************************************************************/
void C_SERVMAN::terminate(double tend) {
  int i;
  double wait_time;
  int n_served;
  int temp;
  C_SERVOBJ *servobj;

// print out the event memory statistics for this run

  if (NODE == 0) {
    printf("terminating qnet servers at time %f\n",tend);
  }

// get the average wait time for a customer in the network

  wait_time = 0.0;
  n_served = 0;

  for (i=0; i<N_LOC; i++) {

// get each of my local objects from the base class object manager with the
// get_obj(i) call.

    servobj = (C_SERVOBJ *)get_obj(i);
    wait_time += servobj->get_wait_time();
    n_served += servobj->get_n_served();

  }

// combine global results (note that we must use integers for this)

  temp = int(wait_time);
  temp = SpComm_GlobalSum(temp);
  n_served = SpComm_GlobalSum(n_served);

// print out the statistics

  if (NODE == 0) {
    printf("Total number of customers served = %d\n", n_served);
    printf("Average wait time = %f\n", double(temp)/double(n_served));
  }

}
