// dummy_event.C method file

#include "SpStdio.H"

#include "dummy_event.H"
#include "dummy_event_mess.H"

/************************************************************************
*									*
*  Overview of event processing in SPEEDES				*
*									*
*  Events in SPEEDES are processed in multiple stages. The first thing	*
*  that happens is a message is fed into the virtual function called	*
*  init(). Events must grab any data that is associated with this	*
*  message and store it inside the event object. The message will be	*
*  discarded by SPEEDES after this method is called. When it is time	*
*  to process the event (which might be much later than the init()	*
*  method is called), SPEEDES calls your temp_process() method. You	*
*  do your basic event processing here but you must be careful to only	*
*  change the state using rollback queue operations that are supplied	*
*  by SPEEDES (see below). Simple assignments which change the state	*
*  of your simulation object should be defered to the exchange() method	*
*  which is called right after the temp_process	method. In the exchange	*
*  method, you will call user defined exchange methods to swap values	*
*  with the simulation object. Rollbacks are supported by SPEEDES by	*
*  calling exchange again (which swaps values back) and then by undoing	*
*  the rollback queue operations that you used in temp_process. When	*
*  the event is committed after GVT is updated, you have one more pass	*
*  to process the event through your perm_process() method. This allows	*
*  you to send data to the outside world without having to worry about	*
*  rollbacks. Be careful however to not use any state variables in the	*
*  simulation object during perm_process because other events may have	*
*  been	processed in the mean time. The cleanup() method is always	*
*  called only durring garbage collection even if the event is rolled	*
*  back. One final method that you can define for your event is called	*
*  lazy(). This method is called only if you make the call, set_lazy(),	*
*  in the constructor of your object. When you make this call, it	*
*  tells SPEEDES that your event is participating in lazy cancellation.	*
*  If your event is rolled back, instead of sending antimessages, etc.	*
*  right away, SPEEDES waits and just before reprocessing the event,	*
*  your lazy method will be invoked. In your lazy method, you have to	*
*  check if the critical state variables that are now in the simulation	*
*  object are different than they were when you first processed the	*
*  event. If they haven't changed (or if you can determine that your	*
*  event will get the same answer anyway) then lazy should return	*
*  the value 1 which means that SPEEDES will roll your event forward	*
*  without having to reprocess the event. Otherwise, you return the	*
*  value 0 which means that antimessages will be sent to chase down	*
*  bad messages (if Time Warp or Breathing Time Warp are used) and the	*
*  event will be reprocessed.
*									*
/************************************************************************
*									*
* Variables provided by SPEEDES that are local to this event		*
*									*
*  SpSimObj *SIMOBJ;			// generic simulation object	*
*  int OBJECT_TYPE;			// simobj object type		*
*  int EVENT_TYPE;			// event type			*
*  int GLOBAL_ID;			// global id for simobj		*
*  int TIME_TAG;			// event time tag		*
*  int LOCAL_ID;			// local ID of simobj		*
*									*
*  static int NODE;			// node				*
*  static int N_NODES;			// number of nodes		*
*									*
*  static SpRbMgr *rbq;			// roll back queue manager	*
*  static C_RANDOM *RANDOM;		// random number generator	*
*  static C_HOST_USER *HOST_USER;	// host user object		*
*									*
*  static int CANCEL;			// cancel event type		*
*  static int CREATE;			// create object event type	*
*  static int DESTROY;			// destroy object event type	*
*									*
*************************************************************************
*									*
* Some useful methods provided by SPEEDES				*
*									*
*  void set_lazy();			// turn on lazy cancellation	*
*  int event_type(char *object_name);	// gets type of event by name	*
*  int object_type(char *object_name);	// gets type of object by name	* 
*  SpSimObjMgr *get_manager(int object_type);	// gets object manager	* 
*									*
*************************************************************************
*									*
* Rollback queue sample operations					*
*									*
*  C_XQUEUE *xqueue;		// xqueue from SPEEDES			*
*  C_XHASH *xhash;		// xhash from SPEEDES			*
*  C_XPRIORITY *xpriority;	// xpriority from SPEEDES		*
*  C_SQ_ITEM *sq_item;		// item from SPEEDES			*
*  SpCancelHandle *ch		// cancel handle from SPEEDES		*
*  C_OBJ *obj;			// user object				*
*  char s[80];			// string				*
*  char *buff;			// user buffer				*
*  char *buff1;			// user buffer				*
*  char *buff2;			// user buffer				*
*  int size;			// number of bytes			*
*  int type;			// object type for freelist		*
*  int var1;			// user variable			*
*  double var2;			// user variable			*
*									*
*  EXCHANGE(var1, var2);		  // exchange two variables	*
*  var1 RB= var2;			  // simple assignment		*
*									*
*  char *buff = RB_NEW_ARRAY_char(size);  // called during event        *
*                                         // processing-creates new mem *
*  RB_DELETE_ARRAY_char(buff);            // deletes memory		*
*									*
*  chunk *c = RB_NEW_chunk(new chunk(chunk_size))  // creates single    *
*                                              // object of type chunk  *
*  RB_DELETE_SINGLE_chunk(c);             // c must be single chunk     *
*									*
*  RB_DEFINE_FREE(freelist);		  // define the free list	*
*  obj = (C_OBJ *)RB_FREE_NEW(type);	  // new object from free list	*
*  RB_FREE_DELETE(obj);			  // puts object in free list	*
*									*
*  RB_MEMCPY(buff1, buff2, size);	  // memory copy		*
*  RB_XMEM(buff1, buff2, size);		  // memory exchange		*
*									*
*  ch = RB_NEW_CANCEL_HANDLE();		  // new cancel handle		*
*  RB_DELETE_CANCEL_HANDLE(ch);		  // delete cancel handle	*
*									*
*  RB_PRINT(stderr, s);			  // print a string		*
*									*
*  RB_INSERT(xqueue, sq_item);		  // insert into xqueue		*
*  RB_PUSH(xqueue, sq_item);		  // same as insert		*
*  sq_item = RB_REMOVE(xqueue);		  // remove top item		*
*  sq_item = RB_REMOVE(xqueue, i);	  // remove item with id=i	*
*  sq_item = RB_POP(xqueue);	 	  // remove top item		*
*  *xqueue += sq_item;			  // add to xqueue		*
*  *xqueue -= sq_item;			  // remove from xqueue		*
*									*
*  RB_INSERT(xhash, sq_item);		  // insert into xhash		*
*  sq_item = RB_REMOVE(xhash, i);	  // remove from xhash		*
*  *xhash += sq_item;			  // add to xhash		*
*  *xhash -= sq_item;			  // remove from xhash		*
*									*
*  RB_INSERT(xpriority, sq_item);	  // insert into xpriority	*
*  sq_item = RB_REMOVE(xpriority);	  // remove from xpriority	*
*  *xpriority += sq_item;		  // add to xpriority		*
*  *xpriority -= sq_item;		  // does not work right now	*
*									*
*************************************************************************
*									*
* Event scheduling and cancelling methods				*
*									*
*	void *schedule(	double time,					*
*			int event_type,					*
*			int object_type,				*
*			int object_id,					*
*			int object_node	);				*
*									*
*	void *schedule(	SpCancelHandle *cancel_handle,			*
*			double time,					*
*			int event_type,					*
*			int object_type,				*
*			int object_id,					*
*			int object_node	);				*
*									*
*	void *schedule(	double time,					*
*			int event_type,					*
*			int object_type,				*
*			int object_id,					*
*			int object_node					*
*			int mess_size					*
*			char *&buff	);				*
*									*
*	void *schedule(	SpCancelHandle *cancel_handle,			*
*			double time,					*
*			int event_type,					*
*			int object_type,				*
*			int object_id,					*
*			int object_node					*
*			int mess_size					*
*			char *&buff	);				*
*									*
*	void cancel( SpCancelHandle *cancel_handle);			*
*									*
*  time:		time tag for the event being scheduled		*
*  event_type:		event type being scheduled			*
*  object_type:		type of simulation object			*
*  object_id:		local id of the simulation object	 	*
*  object_node:		node that the simulation object is on		*
*  cancel_handle:	cancel handle for user cancellation		*
*  mess_size:		size of user data for variable sized messages	*
*  buff:		pointer to start of user data			*
*									*
*  Note: "schedule" returns a pointer to the message that will generate	*
*	 the event. The message will not be sent until after the event	*
*	 is completed - even in Time Warp. This message will be created	*
*	 dynamically if it has a variable length. All fixed length	*
*	 messages are generated from a built-in free list supported by	*
*	 SPEEDES. Fixed	length messages are therefore never deleted and	*
*	 are reused over and over. If you have data associated with	*
*	 your event-message, then you will have to fill in the		*
*	 variables after scheduling the event. Users can cancel events	*
*	 through the cancel handle object. When the event is first	*
*	 scheduled, users supply a cancel handle (SPEEDES does not	*
*	 create the memory for cancel handles for you automatically	*
*	 when you schedule an event). SPEEDES will fill in the details	*
*	 it needs to be able to cancel the event later when you call	*
*	 cancel.							*
*									*
************************************************************************/

int C_DUMMY_EVENT::done = 0;
int C_DUMMY_EVENT::DUMMY_EVENT = -1;

/************************************************************************
* C_DUMMY_EVENT : construct a dummy_event event object			*
************************************************************************/
C_DUMMY_EVENT::C_DUMMY_EVENT() {

  if (!done) {
    done = 1;
    DUMMY_EVENT = event_type("DUMMY_EVENT");
  }

//  set_lazy();		// enable lazy cancellation (default aggressive)

}

/************************************************************************
* init : initialize the event		                                *
************************************************************************/
void C_DUMMY_EVENT::init(SpMsg *header) {
  DUMMY_EVENT_MESS *dummy_event_mess;

  dummy_event_mess = (DUMMY_EVENT_MESS *)header;

  value1 = dummy_event_mess->value1;
  value2 = dummy_event_mess->value2;

}

/************************************************************************
* temp_process : temporarily process event				*
************************************************************************/
void C_DUMMY_EVENT::temp_process() {
  char s[80];

//...... sample code that gets the seed from the simulation object and
//......  then generates a random number.

  lazy_seed = SIMOBJ->get_SEED();

  seed = lazy_seed;
  RANDOM->set_seed(SEED);
  random_number = RANDOM->get_random_float();

//...... print out the seed (it is actually only printed when the
//...... event gets committed).

  sprintf(s,"My new random number is %f\n",random_number);
  RB_PRINT(stdout,s);

//...... schedule this event again one time unit into the future
//...... for the same simulation object.

  schedule(TIME_TAG+1.0, DUMMY_EVENT, OBJECT_TYPE, LOCAL_ID, NODE);

//...... get the new seed to set up for the exchange.

  seed = RANDOM->get_seed();

}

/************************************************************************
* exchange : exchange state variables                                   *
************************************************************************/
void C_DUMMY_EVENT::exchange() {

//...... swap seeds to update the simulation object with the new seed.

  SIMOBJ->exchange_SEED(seed);

}

/************************************************************************
* perm_process : permanently process event                              *	
************************************************************************/
void C_DUMMY_EVENT::perm_process() {

//...... print out a message here.

  print("Event processed at time %f\n",TIME_TAG);

}

/************************************************************************
* cleanup : clean up data structures used				*
************************************************************************/
void C_DUMMY_EVENT::cleanup() {

//...... normally nothing to clean up. Any data that you might
//...... dynamically allocate during init (for example, you might
//...... want to copy your message into a new buffer during init)
//...... should be cleaned up here.

}

/************************************************************************
* lazy : lazy cancellation routine (0=reprocess, 1=rollforward)         *
************************************************************************/
int C_DUMMY_EVENT::lazy() {

  if (lazy_seed != SIMOBJ->get_SEED()) return 0;

  return 1;

}
