/*****************************************************************************/ 
/*                               CodeGen.c                                   */
/*****************************************************************************/

/*****************************************************************************/

#include "SpIostream.H"
#include "IDAST.H"
#include "SpCtype.H"
#include "SpStrstream.H"
#include "TypeEnv.H"		// for EnumValDecl
#include "CodeGen.H"
#include "SpDynPtrArray.H"

static const AtomicType aAnyType(Type::anyType);
static const AtomicType aTaggedType(Type::taggedType);
static const AtomicType aBoolType(Type::boolType);
static const AtomicType aIntType(Type::intType);
static const AtomicType aAnyObjType(Type::anyObjType);

const int PiIndent::IndentSpacing = 3;

ostream& operator<< (ostream &out, PiIndent &indent)
{
   char indentBuff[64];
   indent.CopyOverIndentBuffer(indentBuff);
   return out << indentBuff;
}

/***************************************************************************/
static void litString(const char* str, ostream& out) {

    out << '"';
    while(*str != 0) {
        if (isgraph(*str) || *str == ' ') {
            if (*str == '"' || *str == '\\')
                out << '\\';
            out << *str;
            }
        else if (*str == '\n')
            out << "\\n";
        else {
            out << '\\';
            out << (((unsigned int) *str) / 64);
            out << (((unsigned int) *str) % 64) / 8;
            out << (((unsigned int) *str)) % 8;
            }
        str++;
        }
    out << '"';
}

/**************************************************************************/
/* Since we have to map (mangle) the names.  This function does this.
   (the maping is pretty arbitrary, as long it is consistant and
    one to one */

void toCname(const char* idName, ostream& out) {

   // JB: skip name mangling
   out << idName;

   /* JBco:
    while (*idName != 0) {
        if (isalnum(*idName) && *idName != '_')
            out << *idName;
        else switch(*idName) {
            case '!':
                out << "_EXC_";
                break;
            case '#':
                out << "_HASH_";
                break;
            case '@':
                out << "_AT_";
                break;
            case '$':
                out << "_DOL_";
                break;
            case '^':
                out << "_UP_";
                break;
            case '&':
                out << "_AMP_";
                break;
            case '%':
                out << "_PER_";
                break;
            case '+':
                out << "_PLUS_";
                break;
            case '-':
                out << "_MINUS_";
                break;
            case '*':
                out << "_TIMES_";
                break;
            case '/':
                out << "_DIV_";
                break;
            case '<':
                out << "_LSS_";
                break;
            case '>':
                out << "_GTR_";
                break;
            case '=':
                out << "_EQL_";
                break;
            case '_':
                out << "__";
                break;
            default:
                out << "_" << (unsigned int) *idName << "_";
            }
        idName++;
        }
   */
}

/**************************************************************************/
const char* CodeGen::canonname(const Decl* decl, char* buff, int len) {

    ostrstream name(buff, len);
    canonName(decl, name);
    name << '\0';
    buff[len-1] = 0;    // make sure the string is terminated on overflow
    return(name.str());
} 

/**************************************************************************/
Bool inheritsInterface(const ObjectType* obj) {

    if (obj->getDecl()->isInterface())
	return(TRUE);

    const PiList* supers = obj->getSupers();
    while (supers != 0) {
        ObjectType* super = (ObjectType *) supers->first();
	if (inheritsInterface(super))
	    return(TRUE);
        supers = supers->rest();
        }
    return(FALSE);
}

/**************************************************************************/
/* generate the '::' prefix for every object that is encloses this 
   declaration in its scope */

static void genObjPrefix(const Decl* decl, ostream& out) {


    if (decl == 0) return;
    if (decl->toObjectDecl() == 0) return;
    
    genObjPrefix(decl->getEnvDecl(), out);
    out << decl->getName() << "::";
}

/**************************************************************************/
void CodeGen::canonName(const Decl* decl, ostream& out) {

   // JB: don't mangle names
   out << decl->getName();

   /* JBco:
   decl = decl->normDecl();
    if (decl->isInterface()) {
	if (decl->toEnumValDecl() != 0)
	    genObjPrefix(decl->getEnvDecl()->getEnvDecl(), out);
	else if (decl->toVarDecl() == 0 && decl->toFunDecl() == 0)
	    genObjPrefix(decl->getEnvDecl(), out);
	if ((*decl->getName()) != '_' && ispunct(*decl->getName())) 
	    out << "operator";
	out << decl->getName();
	return;
	}
    if (decl->getEnvDecl() != 0) {
            // don't produce prefix if this is instance variable or ftn.
        if (decl->getEnvDecl()->toObjectDecl() != 0 &&
           (decl->declKind() == Decl::dVarDecl ||
           (decl->declKind() == Decl::dFunDecl &&
            decl->toFunDecl()->getFunType()->getKind() != FunType::ftn))) {
            const ObjectType* obj = decl->getEnvDecl()->toObjectDecl()
					->getTypeRef();
	    if (inheritsInterface(obj)) {
			// no name mangling for things that inherit from
			// interface objects 
		out << decl->getName();
		return;
		}
            out << "I";
	    }
        else
            canonName(decl->getEnvDecl(), out);
	}
    else
	out << "I";    // EVERY user identifer begins with this prefix
    out << "_";
    toCname(decl->getName(), out);
   */
}

/**************************************************************************/
unsigned int declHash(const Decl* decl) {

    return(((unsigned int) decl) / 4);
}

/**************************************************************************/
int declCmp(const Decl* decl1, const Decl* decl2) {
    if (decl1 > decl2) return(1);
    if (decl1 < decl2) return(-1);
    return(0);
}

void CodeGen::CollectObjMgrAndEventNamesAndPrint(const PiList* decls, ostream& out,
						 enum cond_t cond)
{
   int j;
   static SpDynPtrArray* objMgr;
   static int simObjNum;
   PiList *evts;
   PiList *alreadyDeclaredObjMgrTypes = NULL;
   
   if (cond == FIRST_COLLECT)
   {
      simObjNum = 0;
      objMgr = new SpDynPtrArray;
   }
   switch (cond)
   {
      case PRINT_DEFINES:
      case PRINT_SCHEDULE_DECLARES:
	 if (cond == PRINT_DEFINES)
	 {
	    for (j=0; j<simObjNum; j++)
	    {
	       char *name = ((objMgrCont_t *) (void *) (*objMgr)[j])->Name;
	       int numSimObjs =
		 ((objMgrCont_t *) (void *) (*objMgr)[j])->TotSimObj;
	       if (!numSimObjs) numSimObjs = 1;
	       out << "DEFINE_SCATTER_DECOMP_OBJECT_MANAGER(O_"
		   << name << ", " << name << ", "
		   << numSimObjs
		   << ", IMPORT);" << endl;
	    }
	 }
	 evts = AsyncStmt::globalAsyncStmtPiList;
	 if (evts) out << endl;
	 while (evts)
	 {
	    const ApplyExp    *app = ((AsyncStmt*) evts->first())->getApply();
	    const FunType  *method = app->getDecl()->toFunDecl()->getFunType();
	    const char  *className = method->getObj()->getName();
	    const char *methodName = app->getName()->getIdent();
	    const PiList       *args = method->getArgs();
	    if (cond == PRINT_DEFINES)
	    {
	       out << "DEFINE_EVENT_METHOD_"
		   << args->len() // number of args
		   << "_ARG("
		   << className << "_" << methodName // event name
		   << ", 10, " // free list granularity
		   << className
		   << ", "
		   << methodName;
	    }
	    else // cond == PRINT_SCHEDULE_DECLARES
	    {
	       int found = 0;
	       if (alreadyDeclaredObjMgrTypes)
		 for (int i = 0; i < alreadyDeclaredObjMgrTypes->len(); i++)
		 {
		    if (((char*)(alreadyDeclaredObjMgrTypes->first())) == className)
		    {
		       found = 1;
		       break;
		    }
		    alreadyDeclaredObjMgrTypes->rest();
		 }
	       if (!found)
	       {
		  alreadyDeclaredObjMgrTypes =
		    new PiList((char *) className, alreadyDeclaredObjMgrTypes);
		  out << "OM_TYPE(O_"
		      << className
		      << ");"
		      << endl;
	       }
	       out << "SCHEDULE_"
		   << args->len() // number of args
		   << "_ARG("
		   << className << "_" << methodName; // event name
	    }
	    if (args)
	    {
	       out << ", ";
	       while (args)
	       {
		  FunArgDecl* arg = (FunArgDecl* ) args->first();
		  genType(arg->getType(), "", out, 0, FALSE, BY_VALUE);
		  if (arg->getKind() != FunArgDecl::in) out << "&";
		  args = args->rest();
		  if (args != 0) out << ", ";
	       }
	    }
	    out << ");" << endl;
	    evts = evts->rest();
	 }
	 break;
      case PRINT_MAIN:
	 out << "int main (int argc, char **argv)"
	     << endl
	     << "{"
	     << endl;
	 for (j=0; j<simObjNum; j++)
	   out <<  "   PLUG_IN_OBJECT_MANAGER(O_"
	       << ((objMgrCont_t *) (void *) (*objMgr)[j])->Name
	       << ");"
	       << endl;
	 evts = AsyncStmt::globalAsyncStmtPiList;
	 if (evts) out << endl;
	 while (evts)
	 {
	    const char *methodName =
	      ((AsyncStmt*) evts->first())->getApply()->getName()->getIdent();
	    const char  *className =
	      ((AsyncStmt*) evts->first())->getApply()->getDecl()->toFunDecl()
	      ->getFunType()->getObj()->getName();
	    out << "   PLUG_IN_EVENT("
		<< className << "_" << methodName // event name
		<< ")"
		<< endl;
	    evts = evts->rest();
	 }
	 out << endl
	     << "   ExecuteSpeedes(argc, argv);"
	     << endl
	     << "   return 0;"
	     << endl
	     << "}"
	     << endl
	     << endl;
	 delete objMgr;
	 break;
      case FIRST_COLLECT:
      case INTERIM_COLLECT:
	 while (decls)
	 {
	    Decl* decl = (Decl *) decls->first();
	    if (decl->declKind() == Decl::dObjectDecl)
	    {
	       const ObjectType* objType = decl->toObjectDecl()->getTypeRef();
	       if (objType->isSimObj())
		 (*objMgr)[simObjNum++] =
		   new objMgrCont_t(objType->getName(), objType->getTotSimObj());
	    }
	    decls = decls->rest();
	 }
	 break;
      default:
	 cout << "Switch error" << endl;
	 exit (-1);
   }
}

void CodeGen::CollectDefines(const PiList* decls, ostream& out)
{
   cond_t condition = FIRST_COLLECT;
   
   while (decls != 0)
   {
      Decl* decl = (Decl *) decls->first();
      switch(decl->declKind())
      {
	 case Decl::dModDecl:
	    CollectObjMgrAndEventNamesAndPrint(decl->toModDecl()->getDecls(),
					       out, condition);
	    condition = INTERIM_COLLECT;
	    break;
	 case Decl::dObjectDecl:
	    {
	       const ObjectDecl* objDecl = decl->toObjectDecl();
	       CollectObjMgrAndEventNamesAndPrint(objDecl->getTypeRef()->
						  getDecls(), out, condition);
	    }
	    condition = INTERIM_COLLECT;
	    break;
	 case Decl::dCommentDecl:
	    if (decl->toCommentDecl()->getDecl() != 0)
	    {
	       PiList *decls =
		 new PiList((void *) decl->toCommentDecl()->getDecl(), NULL);
	       CollectObjMgrAndEventNamesAndPrint(decls, out, condition);
	       delete decls;
	       condition = INTERIM_COLLECT;
	    }
	    break;
	 case Decl::dAFrameDecl:
	    CollectObjMgrAndEventNamesAndPrint(decl->toAFrameDecl()->getDecls(),
					       out, condition);
	    break;
	 case Decl::dAssocInstDecl:
	    CollectObjMgrAndEventNamesAndPrint(decl->toAssocInstDecl()->
					       getDecls(), out, condition);
	    condition = INTERIM_COLLECT;
	    break;
      }
      decls = decls->rest();
   }
}

void CodeGen::genDefines(const PiList*, ostream& out)
{
   CollectObjMgrAndEventNamesAndPrint(NULL, out, PRINT_DEFINES);
}

void CodeGen::genScheduleDeclares(const PiList*, ostream& out)
{
   CollectObjMgrAndEventNamesAndPrint(NULL, out, PRINT_SCHEDULE_DECLARES);
}

void CodeGen::genMain(const PiList*, ostream& out)
{
   CollectObjMgrAndEventNamesAndPrint(NULL, out, PRINT_MAIN);
}

/**************************************************************************/
/* Generate the code for 'decls' and a main program consisting of
   'startStmt'.  If 'startStmt' is null then no main program is created.
   'stdDecls' is a list of declarations which will generate code ONLY
   if they are dependant on the startStmt (or decls if startStmt is null) */

void CodeGen::genCode(const PiList* decls, const PiList* stdDecls, ostream& out,
		enum FileType_t fileType, const Stmt* startStmt,
		enum CodeGen_t type)
{
   /* JBco:  if (deps != 0) delete deps;
    deps = new DeclSet(declCmp, declHash, 8192);
    if (startStmt != 0) 
	getDepStmt(startStmt, deps);
    else 
	getDepDecls(decls, deps);
   */
   
   PiIndent indent;

   if (fileType == _HFILE || type == SINGLE_FILE)
     out << "#include \"IMPORT.H\"" << endl << endl;
   if (fileType == _HFILE)
   {
      genDeclObjs(stdDecls, out, indent);
      genDecls   (stdDecls, out, indent);
      genDeclObjs(decls, out, indent);
      genDecls   (decls, out, indent);
   }
   else // fileType == CFILE
   {
      CollectDefines(decls, out);
      if (type == SINGLE_FILE || type == NORMAL)
      {
	 if (type == SINGLE_FILE)
	 {
	    genDeclObjs(stdDecls, out, indent);
	    genDecls   (stdDecls, out, indent);
	    genDeclObjs(decls, out, indent);
	    genDecls   (decls, out, indent);
	    genDefines (decls, out);
	 }
	 else genScheduleDeclares(decls, out); // type == NORMAL
	 out << endl;

	 genDeclDefs(stdDecls, out, indent);
	 genDeclDefs(decls, out, indent);
      }
      else
      {
	 genDefines (decls, out); // type == MAIN
	 out << endl;
      }
      if (type == SINGLE_FILE || type == MAIN) genMain (decls, out);
   }
   
   if (startStmt != 0)
   {
      out << "\n/* The main program */\n\n";
      out << "void mainBody() {\n";
      genStmt(startStmt, out, indent);
      out << "}\n\n";
       
      out << "main(int argc, char* argv[]) {\n";
      out << "    setArgs(argc, argv);\n";
      out << "    RUN_MAIN(mainBody);\n";
      out << "    return(0);\n}\n";
    }
   
   // JBco:    delete deps;
   //deps = 0;
}

/**************************************************************************/
void CodeGen::genName(const AtomicType* type, ostream& out, PiIndent) {

    switch(type->getKind()) {
        case Type::boolType:
            out << "Bool";
            break;
        case Type::intType:
            out << "Int";
            break;
        case Type::charType:
            out << "Char";
            break;
        case Type::realType:
            out << "Real";
            break;
        case Type::stringType:
            out << "String";
            break;
        case Type::listType:
            out << "PiList";
            break;
        case Type::taggedType:
            out << "Tagged";
            break;
        default:
            out << "\n #error bad atomic type\n";
        }
}

/**************************************************************************/
void CodeGen::genTypeTag(const Type* type, const Exp* exp, ostream& out,
			 PiIndent indent) {

    type = type->normType();
    const AtomicType* atomicType = type->toAtomicType();
    if (atomicType != 0) {
        out << "typeTag(";
        genExp(exp, &aAnyType, out, indent);
        out << ")";
        return;
        }
    const ObjectType* objectType = type->toObjectType();
    if (objectType != 0) {
        out << "typeTagObj(((void*)";
        genExp(exp, &aAnyType, out, indent);
        out << "), &";
        canonName(objectType->getDecl(), out);
        out << "_typeTAG";
        out << ")";
        return;
        }
    const EnumType* enumType = type->toEnumType();
    if (enumType != 0) {
        out << "typeTagEnum(((int)";
        genExp(exp, &aAnyType, out, indent);
        out << "), &";
        canonName(enumType->getDecl(), out);
        out << "_typeTAG";
        out << ")";
        return;
        }
    out << "\n #error trying to tag an illegal type\n";
}

/**************************************************************************/
void CodeGen::genTypeUnTag(const Type* type, const Exp* exp, ostream& out,
			   PiIndent indent) {

    type = type->normType();
    const AtomicType* atomicType = type->toAtomicType();
    if (atomicType != 0) {
	if (atomicType->getKind() == Type::anyObjType) {
	    out << "unTagObject(";
	    genExp(exp, &aAnyType, out, indent);
	    out << ", 0)";
	    return;
	    }
	if (atomicType->getKind() == Type::anyEnumType) {
	    out << "unTagEnum(";
	    genExp(exp, &aAnyType, out, indent);
	    out << ", 0)";
	    return;
	    }
        out << "unTag";
        genName(atomicType, out, indent);
        out << "(";
        genExp(exp, &aAnyType, out, indent);
        out << ")";
        return;
        }
    const ObjectType* objectType = type->toObjectType();
    if (objectType != 0) {
        out << "((";
        canonName(objectType->getDecl(), out);
        out << "*)";
        out << "unTagObject(";
        genExp(exp, &aAnyType, out, indent);
        out << ", &";
        canonName(objectType->getDecl(), out);
        out << "_typeTAG";
        out << "))";
        return;
        }
    const EnumType* enumType = type->toEnumType();
    if (enumType != 0) {
        out << "((";
        canonName(enumType->getDecl(), out);
        out << "&)";
        out << "unTagEnum(";
        genExp(exp, &aAnyType, out, indent);
        out << ", &";
        canonName(enumType->getDecl(), out);
        out << "_typeTAG";
        out << "))";
        return;
        }
    out << "\n #error trying to untag an illegal type\n";
}

/**************************************************************************/
void CodeGen::genTypeCheck(const Type* type, const Exp* exp, ostream& out,
			   PiIndent indent) {

    type = type->normType();
    const AtomicType* atomicType = type->toAtomicType();
    if (atomicType != 0) {
	if (atomicType->getKind() == Type::anyObjType) {
	    out << "isObject(";
	    genExp(exp, &aAnyType, out, indent);
	    out << ", 0)";
	    return;
	    }
	if (atomicType->getKind() == Type::anyEnumType) {
	    out << "isEnum(";
	    genExp(exp, &aAnyType, out, indent);
	    out << ", 0)";
	    return;
	    }
        out << "is";
        genName(atomicType, out, indent);
        out << "(";
        genExp(exp, &aAnyType, out, indent);
        out << ")";
        return;
        }
    const ObjectType* objectType = type->toObjectType();
    if (objectType != 0) {
        out << "isObject(";
        genExp(exp, &aAnyType, out, indent);
        out << ", &";
        canonName(objectType->getDecl(), out);
        out << "_typeTAG";
        out << ")";
        return;
        }
    const EnumType* enumType = type->toEnumType();
    if (enumType != 0) {
        out << "isEnum(";
        genExp(exp, &aAnyType, out, indent);
        out << ", &";
        canonName(enumType->getDecl(), out);
        out << "_typeTAG";
        out << ")";
        return;
        }
    out << "\n #error trying to typecheck an illegal type\n";
}

/**************************************************************************/
void CodeGen::genDeclObj(const Decl* decl, ostream& out, PiIndent indent)
  {
     
     switch(decl->declKind()) {
	case Decl::dModDecl:
	    genDeclObjs(decl->toModDecl()->getDecls(), out, indent);
	    break;
	case Decl::dObjectDecl: {
	    const ObjectDecl* objDecl = decl->toObjectDecl();

	    /* JBco, add back in when adding run time type checking
	    out << "extern OBJECTTAG "; // for run time type checking
	    canonName(objDecl, out);
	    out << "_typeTAG;\n";   // the address of this is the unique
	    out << "class ";
	    canonName(objDecl, out);
	    out << ";\n";
	    */
	    
	    genDeclObjs(objDecl->getTypeRef()->getDecls(), out, indent);
	    } break;
	case Decl::dCommentDecl:
	    if (decl->toCommentDecl()->getDecl() != 0) 
		 genDeclObj(decl->toCommentDecl()->getDecl(), out, indent);
	    break;
	case Decl::dAFrameDecl:
	    genDeclObjs(decl->toAFrameDecl()->getDecls(), out, indent);
	    break;
	case Decl::dAssocInstDecl:
	    genDeclObjs(decl->toAssocInstDecl()->getDecls(), out, indent);
	    break;
	case Decl::dImportDecl:
	case Decl::dEnumDecl:
	case Decl::dTypeDecl:
	case Decl::dConstDecl:
	case Decl::dVarDecl:
	case Decl::dFunDecl:
	case Decl::dFunArgDecl:
	case Decl::dFlashBackDecl:
	case Decl::dRenameDecl:
	case Decl::dEnumValDecl:
	case Decl::dAssocDecl:
	case Decl::dAssocArgDecl:
	    break;
	}
}

/********************************************************************/
void CodeGen::genDeclObjs(const PiList* decls, ostream& out, PiIndent indent) {

    while (decls != 0) {
        Decl* decl = (Decl *) decls->first();
        genDeclObj(decl, out, indent);
        decls = decls->rest();
        }
}

/**************************************************************************/
void CodeGen::genDecl(const Decl* decl, ostream& out, PiIndent indent,
		      Bool overRide, Bool rollbackable)
{

	// except for objects, interface declarations do not generate code
    if (decl->isInterface() != 0)
	return;
    switch(decl->declKind()) {
	case Decl::dModDecl:
	    genModDecl(decl->toModDecl(), out, indent);
	    break;
	case Decl::dEnumDecl:
	    genEnumDecl(decl->toEnumDecl(), out, indent);
	    break;
	case Decl::dObjectDecl:
	   /* JBco:
	   if (deps != 0 && deps->find(decl) == 0) {
		out << "// omitting decl for "<< decl->getName() <<"\n";
		return;
		}
	   */
	    genObjectDecl(decl->toObjectDecl(), out, indent);
	    break;
	case Decl::dTypeDecl:
	   genTypeDecl(decl->toTypeDecl(), out, indent);
	    break;
	case Decl::dConstDecl:
	    genConstDecl(decl->toConstDecl(), out, indent);
	    break;
	case Decl::dVarDecl:
	    genVarDecl(decl->toVarDecl(), out, indent, rollbackable);
	    break;
	case Decl::dFunDecl:
	    genFunDecl(decl->toFunDecl(), out, indent, overRide);
	    break;
	case Decl::dFunArgDecl:
	    genFunArgDecl(decl->toFunArgDecl(), out, indent);
	    break;
	case Decl::dCommentDecl:
	    if (decl->toCommentDecl()->getDecl() != 0)
		 genDecl(decl->toCommentDecl()->getDecl(), out, indent);
	    break;
	case Decl::dAFrameDecl:
	    out<<"/* Declarations for AFrame: "<<decl->getName()<<" */\n\n";
	    genDecls(decl->toAFrameDecl()->getDecls(), out, indent);
	    break;
	case Decl::dAssocInstDecl: 
	    genAssocInstDecl(decl->toAssocInstDecl(), out, indent);
	    break;
	case Decl::dAssocDecl:
	case Decl::dAssocArgDecl:
	case Decl::dImportDecl:
	case Decl::dFlashBackDecl:
	case Decl::dRenameDecl:
	case Decl::dEnumValDecl:
		// these dont generate code
	    break;
	}
}

/*********************************************************************/
void CodeGen::genAssocInstDecl(const AssocInstDecl* decl, ostream& out,
			       PiIndent indent) {

    out << "/* Declaration for Assoc Instance: " << decl->getName() <<" */\n";
    int arity = decl->getArgs()->len();
    if (arity > 5) {
	out << "\n #error associations must obey  arity <= 5\n";
	return;
	}

    out << "Assoc ";
    canonName(decl, out);
    out << "(" << arity << ");\n\n";

    const PiList* ftnPtr = decl->getAssocDecls();
    while (ftnPtr) {
        FunDecl* ftn = (FunDecl*) ftnPtr->first();
	char buff[1024];
	ostrstream name(buff, 1024);
	canonName(ftn, name);
	name << '\0';

	out << "inline ";
	genFunType(ftn->getFunType(), name.str(), out, indent);
	out << " {\n";
	if (strcmp(ftn->getName(), "index") == 0) {
	    out << "    return(";
	    canonName(decl, out);
	    out << ".index(I_keyCol, I_keyVal, I_retCol, I_idx));\n";
	    }
        else if (strcmp(ftn->getName(), "remove") == 0) {
	    if(ftn->getFunType()->getArgs()->len()==2){
	        out << "    return(";
	        canonName(decl, out);
	        out << ".remove(I_keyCol, I_keyVal));\n";
	        }
	    else{
	        out << "    return(";
	        canonName(decl, out);
	        out << ".remove(I_keyCol1, I_keyVal1, I_keyCol2, I_keyVal2));\n";
	        }
            }
	else if (strcmp(ftn->getName(), "indexAll") == 0) {
	    out << "    return(";
	    canonName(decl, out);
	    out << ".indexAll(I_keyCol, I_keyVal, I_retCol));\n";
	    }
	else if (strcmp(ftn->getName(), "index1") == 0) {
	    Type* type2 = (Type*) decl->getArgs()->rest()->first();
	    out << "    TAGGED I_X = ";
	    canonName(decl, out);
	    out << ".index1(I_v, I_idx);\n";
	    out << "    return(";
	    VarDecl aVarDecl("X", type2, Decl::publicVis);
	    Ident aIdent("X");
	    IdentExp aVarExp(&aIdent, &aVarDecl);
	    genTypeUnTag(type2, &aVarExp, out, indent);
	    out << ");\n";
	    }
	else if (strcmp(ftn->getName(), "index2") == 0) {
	    Type* type1 = (Type*) decl->getArgs()->first();
	    out << "    ";
	    out << "TAGGED I_X = ";
	    canonName(decl, out);
	    out << ".index2(I_v, I_idx);\n";
	    out << "return(";
	    VarDecl aVarDecl("X", type1, Decl::publicVis);
	    Ident aIdent("X");
	    IdentExp aVarExp(&aIdent, &aVarDecl);
	    genTypeUnTag(type1, &aVarExp, out, indent);
	    out << ");\n";
	    }
	else if (strcmp(ftn->getName(), "add") == 0) {
	    out << "    ";
	    canonName(decl, out);
	    out << ".add(";
	    int i=0; 
	    for(;;) {
		out << "I_" << (char) ('A' + i);
		i++;
		if (i >= arity) break;
		out << ", ";
		}
	    out << ");\n";
	    }
	out << "}\n";
        ftnPtr = ftnPtr->rest();
        }
    genDecls(decl->getDecls(), out, indent);
}

/*********************************************************************/
void CodeGen::genModDecl(const ModDecl* decl, ostream& out,
			 PiIndent indent)
{

   // JBco: out << "/* Declarations for Module: "<< decl->getName()<<" */\n\n";
    const PiList* interface = decl->getInterface();
    if (interface != 0) {
       /* JBco:
	if (deps != 0 && deps->find(decl) == 0) {
	    out << "// omitting decl for "<< decl->getName() <<"\n";
	    return;
	    }
       */
	while(interface != 0)
	{
	    PiList* tuple  = (PiList*) interface->first();
	    char* tag  = (char*) tuple->first();
	    char* val  = (char*) tuple->rest()->first();
	    if (strcmp(tag, "header") == 0) out << "#include <" << val << ">\n";
	    /* JBco {
	       out << "#ifndef I_";
	       toCname(val, out);
	       out << "\n#define I_";
	       toCname(val, out);
	       out << " 1\n";
	       out << "#include <" << val << ">\n";
	       out << "#endif\n";
	       }*/
	    interface = interface->rest();
	}
	}

    genDecls(decl->getDecls(), out, indent);
    out << endl;
}

/*********************************************************************/
void CodeGen::genEnumDecl(const EnumDecl* decl, ostream& out, PiIndent) {
    out << "extern ENUMTAG ";     // make the type identifier for this class
    canonName(decl, out);
    out << "_typeTAG;\n";      // the address of this is the unique
                                // identifier for this object
    out << "enum ";
    canonName(decl, out);
    out << " { ";

    const PiList* enumPtr = decl->getTypeRef()->getEnums();
    while (enumPtr) {
        EnumValDecl* enum_ = (EnumValDecl*) enumPtr->first();
        canonName(enum_, out);
        enumPtr = enumPtr->rest();
        if (enumPtr != 0)
            out << ", ";
        }
    out << " };\n";
}

/*********************************************************************/
void CodeGen::genObjectDecl(const ObjectDecl* objDecl, ostream& out,
			    PiIndent indent) {

   const ObjectType* obj = objDecl->getTypeRef();
   // JBco: out << "/* Declaration for Object: " << obj->getName() << " */\n\n";

   // do all the non-method declarations
   const PiList* decls = obj->getDecls();
   while (decls != 0)
   {
      const Decl* decl = (Decl *) decls->first();
      while (decl->toCommentDecl() && decl->toCommentDecl()->getDecl())
	decl = decl->toCommentDecl()->getDecl();
      if (decl->declKind() != Decl::dVarDecl &&
	  (decl->declKind() != Decl::dFunDecl ||
	   decl->toFunDecl()->getFunType()->getKind() == FunType::ftn))
	genDecl(decl, out, indent);
      decls = decls->rest();
   }
   
   const char *className = obj->getDecl()->getName();
   
   out << "class "
       << className;
   if (obj->isSimObj() || obj->getSupers())
   {
      out << ": ";
      const PiList* aSupers = obj->getSupers();

      if (obj->isSimObj())
      {
	 out << "public S_SpProc";
	 if (aSupers) out << ", ";
      }
      while (aSupers)
      {
	 ObjectType* super = (ObjectType *) aSupers->first();
	 out << " public ";
	 canonName(super->getDecl(), out);
	 aSupers = aSupers->rest();
	 if (aSupers) out << ", ";
      }
   }
   out << endl
       << "{"
       << endl
       << " public:"
       << endl;
   // only do the method or variable declarations
   decls = obj->getDecls();
   indent++;

   while (decls != 0)
   {
      const Decl* decl = (Decl *) decls->first();
      while (decl->toCommentDecl() && decl->toCommentDecl()->getDecl())
	decl = decl->toCommentDecl()->getDecl();
      if (decl->declKind() == Decl::dVarDecl ||
	  (decl->declKind() == Decl::dFunDecl &&
	   decl->toFunDecl()->getFunType()->getKind() != FunType::ftn))
	genDecl(decl, out, indent, FALSE, TRUE /*rollbackable */);
      decls = decls->rest();
   }

   genDecls(obj->getExternalMeths(), out, indent, 0, TRUE);

   // JBco, put back in when adding in DOME
   //out << indent << "static clause_table* KB();" << endl;
   out << "};"
       << endl
       << endl;

   indent--;
   
   const PiList* declPtr = obj->getDecls();
   while (declPtr != 0)
   {
      const Decl* decl = (Decl*) declPtr->first();
      const FunDecl* funDecl = decl->normDecl()->toFunDecl();
      if (funDecl != 0 && funDecl->isCalledAsync())
	genAsyncDecl(funDecl, out, indent);
      declPtr = declPtr->rest();
   }
   
    declPtr = obj->getExternalMeths();
    while (declPtr != 0)
    {
       const Decl* decl = (Decl*) declPtr->first();
       const FunDecl* funDecl = decl->normDecl()->toFunDecl();
       if (funDecl != 0 && funDecl->isCalledAsync())
	 genAsyncDecl(funDecl, out, indent);
       declPtr = declPtr->rest();
    }
}

/*********************************************************************/
void CodeGen::genTypeDecl(const TypeDecl* decl, ostream& out, PiIndent indent)
{
    char buff[1024];
    out << "typedef ";
    genType(decl->getTypeRef(), canonname(decl, buff, 1024), out, indent);
    out << ";\n";
}

/*********************************************************************/
void CodeGen::genConstDecl(const ConstDecl* decl, ostream& out,
			   PiIndent indent)
{
    char buff[1024];
    out << "const ";
    genType(decl->getType(), canonname(decl, buff, 1024), out, indent);
    out << " = ";
    genExp(decl->getValue(), decl->getType(), out, indent);
    out << ";\n";
}

/*********************************************************************/
void CodeGen::genVarDecl(const VarDecl* decl, ostream& out, PiIndent indent,
			 Bool rollbackable)
{
   if (decl->getType()->toObjectType())
     if (decl->getType()->toObjectType()->isSimObj()) return; // JB
   
    char buff[1024];
    strcpy(buff, decl->getName());
    genType(decl->getType(), buff, out, indent, rollbackable,
	    decl->isByPointer() ? BY_POINTER : BY_VALUE);
    out << ";\n";
}

/*********************************************************************/
void CodeGen::genFunDecl(const FunDecl* fun, ostream& out,
			 PiIndent indent, Bool overRide)
{   
   if (fun->isExternal() && !overRide) return; // declaration handled by object

   char buff[1024];
   ostrstream name(buff, 1024);

   out << indent;
   if (fun->getFunType()->getKind() == FunType::constructor)
   {
      if (fun->getFunType()->getObj()->isSimObj()) name << "void Init";
      else name << fun->getFunType()->getObj()->getDecl()->getName();
   }
   else if (fun->getFunType()->getKind() == FunType::destructor)
   {
      out << "virtual ";
      name << "~";
      canonName(fun->getFunType()->getObj()->getDecl(), name);
   }
   else
   {
      if (fun->getFunType()->getKind() != FunType::ftn) out << "virtual ";
      canonName(fun, name);
   }
   name << '\0';
   
   PiIndent noIndent;
   genFunType(fun->getFunType(), name.str(), out, noIndent);
   if (fun->isDeferedMeth()) out << "=0";
   out << ";" << endl;
}

/*********************************************************************/
void CodeGen::genFunArgDecl(const FunArgDecl* decl, ostream& out,
			    PiIndent indent) {

    char buff[1024];
    ostrstream name(buff, 1024);
    if (decl->getKind() != FunArgDecl::in)
	name << "&"; 
   canonName(decl, name);
    name << '\0';

    genType(decl->getType(), name.str(), out, indent, FALSE, BY_REFERENCE);
}

/********************************************************************/
void CodeGen::genDecls(const PiList* decls, ostream& out, PiIndent indent,
		       int skipSpaces, Bool overRide)
{
   while (decls != 0)
   {
      Decl* decl = (Decl *) decls->first();
      genDecl(decl, out, indent, overRide);
      if (skipSpaces)
	out << "/*-----------------------------------------------*/"
	    << endl;
      decls = decls->rest();
   }
}

/**************************************************************************/
void CodeGen::genDeclDef(const Decl* decl, ostream& out, PiIndent indent) {

    if (decl->isInterface() != 0 && decl->toObjectDecl() == 0
	    && decl->toCommentDecl() == 0)
	return;
    switch(decl->declKind()) {
	case Decl::dModDecl:
	   // JBco:
	   // out << "/* Definitions for Module: "<< decl->getName()<< " */\n\n";
	    genDeclDefs(decl->toModDecl()->getDecls(), out, indent);
	    out << "\n";
	    break;
	case Decl::dEnumDecl:
	    genEnumDeclDef(decl->toEnumDecl(), out, indent);
	    break;
	case Decl::dObjectDecl:
	   /* JBco:
	    if (deps != 0 && deps->find(decl) == 0) {
		out << "// omitting def for "<< decl->getName() <<"\n";
		return;
		}
	   */
	   // JBco:
	   // out << "/* Definitions for Object: "<< decl->getName()<< " */\n\n";
	   genObjectDeclDef(decl->toObjectDecl(), out, indent);
	   break;
	case Decl::dFunDecl:
	   /* JBco:
	    if (deps != 0 && deps->find(decl) == 0) {
		out << "// omitting def for "<< decl->getName() <<"\n";
		return;
		}
	    */
	   genFunDeclDef(decl->toFunDecl(), out, indent);
	    break;
	case Decl::dCommentDecl:
	    if (decl->toCommentDecl()->getDecl() != 0)
		 genDeclDef(decl->toCommentDecl()->getDecl(), out, indent);
	    break;
	case Decl::dAFrameDecl:
	    out<<"/* Definitions for AFrame: "<<decl->getName()<<" */\n\n";
	    genDeclDefs(decl->toAFrameDecl()->getDecls(), out, indent);
	    break;
	case Decl::dAssocInstDecl:
	    out<<"/* Definitions for Assoc Inst: "<<decl->getName()<<" */\n\n";
	    genDeclDefs(decl->toAssocInstDecl()->getDecls(), out, indent);
	    break;
	case Decl::dAssocDecl:
	case Decl::dAssocArgDecl:
	case Decl::dImportDecl:
	case Decl::dTypeDecl:
	case Decl::dConstDecl:
	case Decl::dVarDecl:
	case Decl::dFunArgDecl:
	case Decl::dFlashBackDecl:
	case Decl::dRenameDecl:
	case Decl::dEnumValDecl:
	    break;
	}
}

/**************************************************************************/
void CodeGen::genEnumDeclDef(const EnumDecl* decl, ostream& out,
			     PiIndent)
{

    out << "char* "; 
    canonName(decl, out);
    out << "_typeNames[] = {\n"; 

    const PiList* enumPtr = decl->getTypeRef()->getEnums();
    while (enumPtr) {
        EnumValDecl* enum_ = (EnumValDecl*) enumPtr->first();
	out << "\"" << enum_->getName() << "\", ";
        enumPtr = enumPtr->rest();
        }
	    
     out << "0 };\n";
    out << "ENUMTAG "; 
    canonName(decl, out);
    out << "_typeTAG = { TagEnum, ";
    canonName(decl, out);
    out << "_typeNames };\n";

}

/**************************************************************************/
void CodeGen::genObjectDeclDef(const ObjectDecl* objDecl, ostream& out,
			       PiIndent indent) {

    const ObjectType* obj = objDecl->getTypeRef();
// JBco:    out << "/* Definitions for Object: " << obj->getName() << " */\n\n";

    /* JBco, add back in when adding run time type checking
    out << "SuperInfo ";        // for run time type checking 
    canonName(obj->getDecl(), out);	// the address of this is unique
    out << "_typeSupers[] = {\n";

    const PiList* aSupers = obj->getSupers();
    while (aSupers != 0) {
        ObjectType* super = (ObjectType *) aSupers->first();
	out << "    { &";
        canonName(super->getDecl(), out);
	out << "_typeTAG, superOffset(";
	canonName(obj->getDecl(), out);
        out << ", ";
        canonName(super->getDecl(), out);
	out << ") },\n";
        aSupers = aSupers->rest();
        }
    out << "    { 0, 0 } };\n";
    out << "OBJECTTAG ";        // for run time type checking 
    canonName(obj->getDecl(), out);	// the address of this is unique
    out << "_typeTAG = { TagObject, ";
    canonName(obj->getDecl(), out);	// the address of this is unique
    out << "_typeSupers };\n\n";
    */
    
    if (objDecl->isInterface() != 0)  // check if this is a foreign C++ object
	return;

    genDeclDefs(obj->getDecls(), out, indent);

    // Generate the Knowledge base for this object
    
    int curName=0;
    PiList* clausePtr = (PiList*) obj->getClauses();
    while (clausePtr != 0) {
        Clause* clause = (Clause*) clausePtr->first();
        genAuxClause(clause, curName, out, indent);
        clausePtr = clausePtr->rest();
        }
    curName=0;

    /* JBco: put back in when adding DOME
    out << "clause_table* ";
    canonName(obj->getDecl(), out);
    out << "::KB()"
	<< endl
	<< "{"
	<< endl;
    indent++;
    if (obj->getClauses() != 0)
    {
       out << indent
	   << "static clause_table* ret = 0;"
	   << endl
	   << indent
	   << "if (ret != 0) return(ret);"
	   << endl
	   << indent
	   << "ret = nilClauseTable();"
	   << endl;
       PiList* clausePtr = (PiList*) obj->getClauses();
       while (clausePtr != 0)
       {
	  Clause* clause = (Clause*) clausePtr->first();
	  out << indent
	      << "addClause(ret, "
	      << endl;
	  genClause(clause, curName, out, indent);
	  out << ");"
	      << endl;
	  clausePtr = clausePtr->rest();
       }
       out << indent
	   << "return(ret);"
	   << endl;
    }
    else out << indent
	     << "return(0);"
	     << endl;

    indent--;
    out << "}"
	<< endl
	<< endl;
    */
}

/**************************************************************************/
void CodeGen::genFunDeclDef(const FunDecl* fun, ostream& out, PiIndent indent) {

    if (fun->getBody() == 0 && fun->getFunType()->getKind() != FunType::query)
        return;
    if (fun->isDeferedMeth())
        return;

    char buff[1024];
    ostrstream name(buff, 1024);


    if (fun->getFunType()->getKind() != FunType::ftn)
    {
       if (fun->getFunType()->getKind() == FunType::constructor &&
	   fun->getFunType()->getObj()->isSimObj()) name << "void ";
       name << fun->getFunType()->getObj()->getDecl()->getName() << "::";
    }

    if (fun->getFunType()->getKind() == FunType::constructor)
    {
       if (fun->getFunType()->getObj()->isSimObj()) name << "Init";
       else name << fun->getFunType()->getObj()->getDecl()->getName();
    }
    else if (fun->getFunType()->getKind() == FunType::destructor)
       name << "~" << fun->getFunType()->getObj()->getDecl()->getName();
    else name << fun->getName();
    name << '\0';

    if (fun->getFunType()->getKind() == FunType::query)
    {
       genFunType(fun->getFunType(), name.str(), out, indent);
       out << "{\n\n";
       const FunType* funType = fun->getFunType();
       genProverCall(funType, fun->getName(), out, indent);
       out << "\n}\n\n";
    }
    else
    {
       char buff[2048];
       ostrstream funHead(buff, 2048);
       genFunType(fun->getFunType(), name.str(), funHead, indent);
       funHead << '\0';
       genMethodDef(fun->getBody(), funHead.str(), out, indent.SendIncrement());
    }

    if (fun->isCalledAsync())
      genAsyncDef(fun, out, indent.SendIncrement());
}

/***************************************************************************/
void CodeGen::genProverCall(const FunType* fun, const char* name,
			    ostream& out, PiIndent indent){

    out << "return(prove(";
    canonName(fun->getObj()->getDecl(), out);
    out << "::KB(),";	
    out << "domePred(" << '"' << name << '"' << ",\n";

    int parenCtr = 0;
    const PiList* argPtr = fun->getArgs();
    while (argPtr) {  
	FunArgDecl* arg = (FunArgDecl *) argPtr->first();

	out << "domeCons(";
	parenCtr++;
	Ident idIdent((char *) arg->getName());
	IdentExp id(&idIdent, arg);
	if (arg->getKind() == FunArgDecl::in) {
	    out << "dome_";
	genTypeTag(arg->getType(), &id, out, indent);
	    }
	else 
	    genDomeVar(arg->getType(), arg->getName(), &id, out, indent);
	out << ",\n";
	argPtr = argPtr->rest();
	}
    out << "0";
    while(parenCtr > 0) {
	out << ")";
	--parenCtr;
	}

    out << "), this));";
}

/**************************************************************************/
void CodeGen::genDomeVar(const Type* type, const char* name, 
			 const Exp* exp, ostream& out, PiIndent indent) {

   type = type->normType();
    const AtomicType* atomicType = type->toAtomicType();
    if (atomicType != 0) {
        out << "domeVar(";
        out << '"' << name << '"' << ", ";
        genExp(exp, &aAnyType, out, indent);
        out << ")";
        return;
        }
    const ObjectType* objectType = type->toObjectType();
    if (objectType != 0) {
        out << "domeVar(" << '"' << name << '"';
        out << ", ((void*)";
        genExp(exp, &aAnyType, out, indent);
        out << "), &";
        canonName(objectType->getDecl(), out);
        out << "_typeTAG";
        out << ")";
        return;
        }
    const EnumType* enumType = type->toEnumType();
    if (enumType != 0) {
        out << "domeVar(" << '"' << name << '"';
        out << ", ((int&)";
        genExp(exp, &aAnyType, out, indent);
        out << "), &";
        canonName(enumType->getDecl(), out);
        out << "_typeTAG";
        out << ")";
        return;
        }
    out << "\n #error trying to tag an illegal type\n";
}

/********************************************************************/
void CodeGen::genDeclDefs(const PiList* decls, ostream& out, PiIndent indent) {

    while (decls != 0) {
        Decl* decl = (Decl *) decls->first();
        genDeclDef(decl, out, indent);
        decls = decls->rest();
        }
}

/********************************************************************/
void CodeGen::genType(const Type* type, const char* cName, ostream& out,
		      PiIndent indent, Bool rollbackable,
		      qualifier_t qualifier)
{
   out << indent;
    type = type->normType();
    switch(type->typeKind()) {
	case Type::tIdentType: 
	    canonName(type->toIdentType()->getDef(), out);
	    if (type->toObjectType() != 0) // is this an object type?
	    {
	       if (qualifier == BY_REFERENCE) out << "&";
	       else if (qualifier == BY_POINTER) out << "*";
	    }
	    break;
	case Type::tArrayType: 
	    if (type->toArrayType()->getBound() == 0) {
		out << "VAR_ARRAY<";
		genType(type->toArrayType()->getElemType(), "", out, indent);
		out << ">" << cName;
		}
	    else {
		out << "ARRAY<";
		genType(type->toArrayType()->getElemType(), "", out, indent);
		out << ", (";
		genExp(type->toArrayType()->getBound(), &aAnyType, out, indent);
		out << ")>" << cName;
		}
	    break;
	case Type::tFunType: {
	    const FunType* asFunType = type->toFunType();
	    ostrstream nameStream;
	    nameStream << "(";
	    if (asFunType->getObj() != 0) {
		canonName(asFunType->getObj()->getDecl(), nameStream);
		nameStream << "::";
		}
	    nameStream << "*" << cName << ")" << '\0';
	    char* name = nameStream.str();
	    genFunType(type->toFunType(), name, out, indent);
	    delete [] name;
	    } break;
	case Type::tAtomicType:
	    genAtomicType(type->toAtomicType(), cName, out, indent,
			  rollbackable);
	    break;
	case Type::tFutureType:
	    out << "FUTURE<";
	    genType(type->toFutureType()->getSubType(), "", out, indent);
	    out << ">" << cName;
	    break;
       case Type::tObjectType:
	  if (!strcmp (type->toObjectType()->getDecl()->getName(), "SEMAPHORE"))
	    out << "LogicalSemaphore";
	  else out << type->toObjectType()->getDecl()->getName();
	  out << " ";
	  if (qualifier == BY_REFERENCE) out << "&";
	  else if (qualifier == BY_POINTER) out << "*";
	  out << cName;
	  break;
       case Type::tEnumType: 
	  canonName(type->toEnumType()->getDecl(), out);
	  break;
       case Type::tSynType: 	// should never happen
	  break;
    }
}

/********************************************************************/
void CodeGen::genFunType(const FunType* type, const char* cName, 
			 ostream& out, PiIndent indent)
{
   ostrstream funAndArgs;

   funAndArgs << cName << "(";
   const PiList* argPtr = type->getArgs();
   while (argPtr)
   {
      FunArgDecl* arg = (FunArgDecl* ) argPtr->first();
      genDecl(arg, funAndArgs, indent);
      argPtr = argPtr->rest();
      if (argPtr != 0) funAndArgs << ", ";
   }
   funAndArgs << ")";
   funAndArgs << '\0';

   char *str = funAndArgs.str();
   if (type->getKind() == FunType::constructor || 
       type->getKind() == FunType::destructor) out << str;
   else genType(type->getRetType(), str, out, indent);
   delete [] str;
}

/********************************************************************/
void CodeGen::genAtomicType(const AtomicType* type, const char* cName, 
			    ostream& out, PiIndent, Bool rollbackable)
{
   
   switch (type->getKind()) {
      case Type::boolType      :

	// The code below used to use the C trinary operator which
	// involves a question mark and colon symbol, but Insure++
	// cannot handle the trinary operator, so I changed it to use
	// the C "if" keyword.
      case Type::intType       : if (rollbackable)
				   out << "RB_int";
				   else out << "int"    ; break; // JB
      case Type::realType      : if (rollbackable)
				   out << "RB_double";
				   else out << "double" ; break; // JB
      case Type::charType      : out << "char"       ; break; // JB
      case Type::taggedType    : out << "TAGGED"     ; break;
      case Type::stringType    : out << "STRING"     ; break;
      case Type::listType      : out << "LIST"       ; break;
      case Type::voidType      : out << "void"       ; break;
      case Type::anyObjType    : out << "void*"      ; break;
      case Type::anyFutureType : out << "ANYFUTURE"  ; break;
	 
	 // this is not really correct, but it is pretty close
      case Type::anyType: out << "void*"      ; break;
	 
      case Type::unknownObjType:	// this is probably an error
	 
      case Type::objectType:        // these are definately errors.
      case Type::arrayType:
      case Type::funType:
      case Type::enumType:
      case Type::setType:
      default:
	 out << "\n #error bad type reference\n";
	 break;
   }
   if (strcmp(cName, "")) out << " ";
   out << cName;
}

/********************************************************************/
void CodeGen::
genMethodDef(const Method* method, const char* funHead, ostream& out,
	     PiIndent indent)
{
    switch(method->methodKind())
    {
       case Method::mDeferedBody:
	  break;
       case Method::mMethodBody:
	  {
	     const MethodBody* body = method->toMethodBody();
	     
	     PiList* varDecls=0;
	     PiList* otherDecls=0;
	     const PiList* decls = body->getDecls();
	     while (decls != 0)
	     {
		Decl* decl = (Decl *) decls->first();
		while(decl->toCommentDecl() && decl->toCommentDecl()->getDecl())
		  decl = (Decl*) decl->toCommentDecl()->getDecl();
		if (decl->declKind() == Decl::dVarDecl)
		  varDecls = new PiList(decl, varDecls);
		else
		  otherDecls = new PiList(decl, otherDecls);
		decls = decls->rest();
	     }
	     genDeclObjs(otherDecls, out, indent);
	     genDecls(otherDecls, out, indent);
	     genDeclDefs(otherDecls, out, indent);
	     if (otherDecls != 0) delete otherDecls;
	     out << funHead;
	     if (body->getConstrInits() != 0)
	     {
		out << " :\n";
		const PiList* inits = body->getConstrInits();
		while (inits)
		{
		   NewExp* init = (NewExp *) inits->first();
		   genNewExp(init, out, indent);
		   inits = inits->rest();
		   if (inits != 0)
		     out << ", ";
		}
	     }
	     out << endl;
	     if (body->getTotReentryLabels())
	     {
		out << "P_VAR" << endl;
		genDecls(varDecls, out, indent);
		out << "P_BEGIN("
		    << body->getTotReentryLabels()
		    << ")"
		    << endl;
		genStmts(body->getStmts(), out, indent);
		out << "P_END"
		    << endl
		    << endl;
	     }
	     else
	     {
		out << "{";
		genDecls(varDecls, out, indent);
		out << endl;
		genStmts(body->getStmts(), out, indent);
		out << "}\n\n";
	     }
	     if (varDecls != 0) delete varDecls;
	  }
	  break;
    }
}

/********************************************************************/
void CodeGen::genExp(const Exp* exp, const Type* expectType, ostream& out,
		     PiIndent indent, qualifier_t qualifier)
{

    Bool parens = FALSE;
    expectType = expectType->normType();
    if (expectType->isKind(Type::taggedType)) {
	if (!exp->getType()->normType()->isKind(Type::taggedType)) {
	    genTypeTag(exp->getType(), exp, out, indent); 
	    return;
	    }
	}
    else if (expectType->toArrayType() != 0) {
	const ArrayType* asArray = exp->getType()->normType()->toArrayType();
	if (asArray != 0 && expectType->toArrayType()->getBound() == 0 && 
	    asArray->getBound() != 0) {
	    genType(expectType, "", out, indent);
	    out << "(" << asArray->getEvalBound() << ", ";
	    parens = TRUE;
	    }
	}

    /* JBco:
    if (expectType->toFutureType() == 0 &&
	exp->getType()->normType()->toFutureType() != 0 &&
        !expectType->isKind(Type::anyFutureType)) {
	out << "FUTURE_VAL(";
	parens = TRUE;
	}
    */
	
    switch(exp->expKind()) {
	case Exp::eAsExp: {
	    const AsExp* asAsExp = exp->toAsExp();
	    genExp(asAsExp->getBody(), asAsExp->getType(), out, indent);
	    } break;
	case Exp::eIsAExp: 
	    genIsAExp(exp->toIsAExp(), out, indent);
	    break;
	case Exp::eCastExp: 
	    genCastExp(exp->toCastExp(), out, indent);
	    break;
	case Exp::eSelfExp:
	   out << "this";
	    break;
	case Exp::eIdentExp:
	   {
	      const Decl* def = exp->toIdentExp()->getDef();
	      // if defined in an object, means it is an instance variable
	      def = def->normDecl();
	      if (def->getEnvDecl()!=0 && def->getEnvDecl()->toObjectDecl()!=0)
	      {
		 if (def->toVarDecl() != 0)
		 {
		    if (qualifier == BY_REFERENCE)
		      out << "*" << endl;
		    out << "this->";
		 }
		 else
		 {
		    // FIX I suspect that this is probably not a
		    // robust way of implementing method pointers
		    const FunDecl* asFun = def->toFunDecl();
		    if (asFun != 0 && asFun->getFunType()->getObj() != 0)
		    {
		       out << "&";
		       canonName(asFun->getFunType()->getObj()->getDecl(),out);
		       out << "::";
		    }
		 }
	      }
	      const char *name = def->getName();
	      if (!strcmp(name, "SEMAPHORE")) out << "LogicalSemaphore";
	      else    out << name;
	   }
	   break;
	case Exp::eArrayExp: {
	   
	    const ArrayExp* array = exp->toArrayExp();
	    genExp(array->getArray(), &aAnyType, out, indent);
	    out << "[";
	    genExp(array->getIndex(), &aAnyType, out, indent);
	    out << "]";
	    } break;
	case Exp::eApplyExp:
	    genApplyExp(exp->toApplyExp(), out, indent);
	    break;
	case Exp::eNullExp:
	    out << "0";
	    break;
	case Exp::eIntExp:
	   //out << "((int) "; // JB
	   out << exp->toIntExp()->getValue();
	   //out << ")";
	    break;
	case Exp::eRealExp:
	   // out << "((double) "; // JB
	    out << exp->toRealExp()->getValue();
	    // out << ")";
	    break;
	case Exp::eStringExp:
	   //out << "((STRING) ";
	    litString(exp->toStringExp()->getValue(), out);
	    // out << ")";
	    break;
	case Exp::eCharExp:
	   //out << "((char) "; // JB
	    out << (int) exp->toCharExp()->getValue();
	    //out << ")";
	    break;
	case Exp::eBoolExp:
	   //out << "((BOOL) ";
	    out << exp->toBoolExp()->getValue();
	    // out << ")";
	    break;
	case Exp::eVarExp:
	   // JBco: out << "I_V_";
	    out << exp->toVarExp()->getVar();
	    break;
	case Exp::eNewExp:
	    out << "new";
	    genNewExp(exp->toNewExp(), out, indent);
	    break;
	}
    if (parens) out << ")";	
}

/***************************************************************************/
/* Note that this function does NOT print the actual 'new' keyword */

void CodeGen::genNewExp(const NewExp* exp, ostream& out, PiIndent indent)
{ 

    if (exp->getNewArgs() != 0) {
	out << "(";
	genExps(exp->getNewArgs(), 
		exp->getNewDecl()->getFunType()->getArgs(), out, ", ", indent);
	out << ")";
	}
    out << " ";
    const ObjectType* object = exp->getType()->normType()->toObjectType();
    if (object == 0) 
	out << "\n #error new of non-objects is not implmented\n";

    if (!strcmp(object->getDecl()->getName(), "SEMAPHORE"))
      out << "LogicalSemaphore";
    else out << object->getDecl()->getName();
    out << "(";
    if (object->getDecl()->isInterface() != 0) 
	genToCExps(exp->getArgs(), 
		   exp->getConstrDecl()->getFunType()->getArgs(), out, ", ",
		   indent);
    else 
	genExps(exp->getArgs(), 
		exp->getConstrDecl()->getFunType()->getArgs(), out, ", ",
		indent);
    out << ")";
}

/***************************************************************************/
void CodeGen::genIsAExp(const IsAExp* exp, ostream& out, PiIndent indent)
{

    if (exp->getSubType() == 0) {	// handle IsA VAR expression
   	out << "isVariable(";
	genExp(exp->getBody(), &aAnyType, out, indent);
	out << ")";
	return;
	}
    genTypeCheck(exp->getSubType(), exp->getBody(), out, indent);
}

/***************************************************************************/
void CodeGen::genCastExp(const CastExp* exp, ostream& out, PiIndent indent)
{

    out << "(";
    if (exp->getBodyType()->isKind(Type::taggedType) && 
       !exp->getType()->isKind(Type::taggedType))
	genTypeUnTag(exp->getType(), exp->getBody(), out, indent);
    else if (!exp->getBodyType()->isKind(Type::taggedType) &&
	      exp->getType()->isKind(Type::taggedType))
	genTypeTag(exp->getBodyType(), exp->getBody(), out, indent);
    else
	genExp(exp->getBody(), &aTaggedType, out, indent);
    out << ")";
}

/***************************************************************************/
void CodeGen::genApplyExp(const ApplyExp* exp, ostream& out, PiIndent indent)
{ 

    const FunType* funType = exp->getDecl()->getType()->toFunType();
    Bool needConv = (funType->getRetType()->normType()->isKind(Type::stringType)
		     && exp->getDecl()->isInterface());
    if (needConv)
	out << "((STRING) ";

    genApply_Exp(exp, out, indent);
    if (needConv)
	out << ")";
}
    
/***************************************************************************/
void CodeGen::genApply_Exp(const ApplyExp* exp, ostream& out, PiIndent indent)
{    
   const FunType* funType = exp->getDecl()->getType()->toFunType();
   const char *funName = exp->getDecl()->getName();

   if (!strcmp(funName, "ID")) out << "get_NUM_ID()";
   else if (!strcmp(funName, "TIME")) out << "GetTime().GetTime()";
   else if (!strcmp(funName, "out")) out << "cout";
   else
   {
      if (
	  !strcmp(funName, "<<")  ||
	  !strcmp(funName, ">>")  ||
	  !strcmp(funName, "<")  ||
	  !strcmp(funName, ">")  ||
	  !strcmp(funName, "<=") ||
	  !strcmp(funName, ">=") ||
	  !strcmp(funName, "==") ||
	  !strcmp(funName, "!=") ||
	  !strcmp(funName, "+")  ||
	  !strcmp(funName, "-")  ||
	  !strcmp(funName, "/")  ||
	  !strcmp(funName, "%")  ||
	  !strcmp(funName, "++") ||
	  !strcmp(funName, "--") ||
	  !strcmp(funName, "=")  ||
	  !strcmp(funName, "+=") ||
	  !strcmp(funName, "-=") ||
	  !strcmp(funName, "/=") ||
	  !strcmp(funName, "*=") ||
	  !strcmp(funName, "%=") ||
	  !strcmp(funName, "*")  )
      {
	 char sep[8] = " ";
	 strcat(sep, funName);
	 if (!strcmp(funName, "=")) strcat(sep, "=");
	 strcat(sep, " ");
	 genExps(exp->getArgs(), funType->getArgs(), out, sep, indent);      
      }
      else
      {
	 if (exp->getKind() == FunType::constructor) 
	   canonName(funType->getObj()->getDecl(),out);
	 else if (exp->getKind() == FunType::ftn)
	 {
	    if (exp->getDecl()->toFunDecl() != 0)
	    {
	       if (!strcmp(exp->getDecl()->getName(), "wait"))
		 out << "WAIT";
	       else if (!strcmp(exp->getDecl()->getName(), "waitUntil"))
		 out << "WAIT_UNTIL";
	       else if (!strcmp(exp->getDecl()->getName(), "waitFor"))
		 out << "WAIT_FOR";
	       else out << exp->getDecl()->getName();
	    }
	    else
	    {
	       out << "(*";
	       canonName(exp->getDecl(), out);
	       out << ")";
	    }
	 }
	 else
	 {
	    if (exp->getDecl()->toFunDecl() != 0)
	    {
	       if (exp->getObj()->expKind() != Exp::eSelfExp)
	       {
		  out << "(";
		  genExp(exp->getObj(), funType->getObj(), out, indent);
		  out << ")";
		  // JB:
		  const VarDecl *asVarDecl =
		    exp->getObj()->toIdentExp()->getDef()->toVarDecl();
		  if (asVarDecl && asVarDecl->isByPointer()) out << "->";
		  else out << ".";
	       }
	       if (exp->getFromObj() != 0)
	       {
		  canonName(exp->getFromObj()->getDecl(), out);
		  out << "::";
	       }
	       canonName(exp->getDecl(), out);
	    }
	    else // it is a function argument, must dereference
	    {	 
	       out << "((";
	       genExp(exp->getObj(), funType->getObj(), out, indent);
	       out << ")->*" << exp->getDecl()->getName() << ")";
	    }
	 }
	 out << "(";
	 if (exp->getReentryLabel())
	   out << exp->getReentryLabel() << ", ";
	 if (exp->getDecl()->isInterface())
	   genToCExps(exp->getArgs(), funType->getArgs(), out, ", ", indent);
	 else 
	   genExps(exp->getArgs(), funType->getArgs(), out, ", ", indent);
	 out << ")";
      }
   }
}

/***************************************************************************/
void CodeGen::genAsyncStmt(const AsyncStmt* stmt, ostream& out,
			   PiIndent indent)
{ 
   const Decl* funDecl = stmt->getApply()->getDecl();
   const FunType* funType = funDecl->getType()->toFunType();
   const char *className = funType->getObj()->getName();
   const char *methodName = funDecl->getName();

   out << "Schedule";
   if (className[0] >= 'a' && className[0] <= 'z') out << "_";
   out << className << "_" << methodName << "(";
   if (stmt->isWaitUntilTimeToTell())
     genExp(stmt->getWaitValExp(), &aAnyType, out, indent);
   else if (stmt->isWaitToTell())
   {
      out << "GetTime() + ";
      genExp(stmt->getWaitValExp(), &aAnyType, out, indent);
   }
   else out << "GetTime()";
   out << ", GetObjHandle(";
   if (stmt->getApply()->getObj()->expKind() == Exp::eSelfExp) out << ")";
   else
   {
      out << "O_" << className << ", ";
      genExp(stmt->getDestEnumIdExp(), &aAnyType, out, indent);
      out << ")";
   }
   
   if (stmt->getApply()->getArgs() != 0) out << ", ";
   genExps(stmt->getApply()->getArgs(), funType->getArgs(), out, ", ", indent);
   out << ");" << endl;
   
   /* JBco (the obsolete thread system)

    const Decl* funDecl = stmt->getApply()->getDecl();
    const FunType* funType = funDecl->getType()->toFunType(); 
    out << "    taskCreate(";
    canonName(funType->getObj()->getDecl(), out);
    out << "_";
    canonName(funDecl, out);
    out << "_task";
    out << ",\n";

    out << "        new ";
    canonName(funType->getObj()->getDecl(), out);
    out << "_";
    canonName(funDecl, out);
    out << "_args(";
    genExp(stmt->getApply()->getObj(), funType->getObj(), out, indent);

    if (stmt->getApply()->getArgs() != 0)
	out << ", ";
    genExps(stmt->getApply()->getArgs(), funType->getArgs(), out, ", ", indent);

    if (stmt->getLhs() != 0) {
	out << ", &(";
	genExp(stmt->getLhs(), stmt->getLhs()->getType(), out, indent);
	out << ")";
	}
    out << "));\n";
   */
}

/********************************************************************/
void CodeGen::genExps(const PiList* exps, const PiList* expectTypes, 
    ostream& out, const char* sep, PiIndent indent, qualifier_t qualifier)
{
    while (exps != 0)
    {
	Exp* exp = (Exp *) exps->first();
	const Type* expectType = ((FunArgDecl*)expectTypes->first())->getType();

	// JB: dereference if by pointer and calling a method
	if (exp->toIdentExp())
	{
	   const VarDecl *asVarDecl = exp->toIdentExp()->getDef()->toVarDecl();
	   if (asVarDecl && asVarDecl->isByPointer()) out << "*";
	}

	genExp(exp, expectType, out, indent, qualifier);
	exps = exps->rest();
	expectTypes = expectTypes->rest();
	if (exps != 0) out << sep;
    }
}

/********************************************************************/
void CodeGen::genToCExps(const PiList* exps, const PiList* expectTypes, 
    ostream& out, const char* sep, PiIndent indent)
{
    while (exps != 0)
    {
	Exp* exp = (Exp *) exps->first();
        const FunArgDecl* argDecl = (FunArgDecl*) expectTypes->first();
        genToCExp(exp, argDecl, out, indent);
	exps = exps->rest();
	expectTypes = expectTypes->rest();
	if (exps != 0) out << sep;
    }
}

/********************************************************************/
/* do any type conversion necessary to convert IMPORT types to C types */

void CodeGen::genToCExp(const Exp* exp,const FunArgDecl* argDecl,
			ostream& out, PiIndent indent){

    const Type* expectType = argDecl->getType();
    if (expectType->isKind(Type::stringType)) {
	if (exp->toStringExp() != 0)
	    litString(exp->toStringExp()->getValue(), out);
	else if (exp->toApplyExp() != 0) {
	    const ApplyExp* asApply = exp->toApplyExp();
	    const FunType* funType = asApply->getDecl()->getType()->toFunType();
	    Bool needConv = (funType->getRetType()->normType()->
		isKind(Type::stringType) && asApply->getDecl()->isInterface());
	    if (needConv)
		genApply_Exp(asApply, out, indent);
	    else {
		if (argDecl->getKind() == FunArgDecl::in) {
		    out << "toChars(";
		    genExp(exp, expectType, out, indent);
		    out << ")";
		    }
		else
		    genExp(exp, expectType, out, indent);
		}
	    }
	else {
	    if (argDecl->getKind() == FunArgDecl::in) {
		out << "toChars(";
		genExp(exp, expectType, out, indent);
		out << ")";
		}
	    else
		genExp(exp, expectType, out, indent);
	    }
	}
    else if (expectType->normType()->toObjectType() != 0 &&
	     argDecl->getKind() != FunArgDecl::in) {
	    // out parameters for objects are passed by reference
	out << "*(";
        genExp(exp, expectType, out, indent);
	out << ")";
	}
    else
        genExp(exp, expectType, out, indent);
}

/********************************************************************/
void CodeGen::genStmt(const Stmt* stmt, ostream& out, PiIndent indent)
{
   // JB indenting:
   if (stmt->stmtKind() != Stmt::sCommentStmt) out << indent;
   
    switch(stmt->stmtKind()) {
	case Stmt::sAssignStmt:
	    genAssignStmt(stmt->toAssignStmt(), out, indent);
	    break;
	case Stmt::sIfStmt:
	    genIfStmt(stmt->toIfStmt(), out, indent);
	    break;
	case Stmt::sWhileStmt:
	    genWhileStmt(stmt->toWhileStmt(), out, indent);
	    break;
	case Stmt::sRepeatStmt:
	    genRepeatStmt(stmt->toRepeatStmt(), out, indent);
	    break;
	case Stmt::sForStmt:
	    genForStmt(stmt->toForStmt(), out, indent);
	    break;
	case Stmt::sLoopStmt:
	    genLoopStmt(stmt->toLoopStmt(), out, indent);
	    break;
	case Stmt::sReturnStmt:
	    genReturnStmt(stmt->toReturnStmt(), out, indent);
	    break;
	case Stmt::sExitStmt:
	    genExitStmt(stmt->toExitStmt(), out, indent);
	    break;
	case Stmt::sExpStmt:
	    genExp(stmt->toExpStmt()->getExp(), &aAnyType, out, indent);
	    out << ";\n";
	    break;
	case Stmt::sDisposeStmt:
	    genDisposeStmt(stmt->toDisposeStmt(), out, indent);
	    break;
	case Stmt::sBlockStmt:
	    genBlockStmt(stmt->toBlockStmt(), out, indent);
	    break;
	case Stmt::sCaseStmt:
	    genCaseStmt(stmt->toCaseStmt(), out, indent);
	    break;
	case Stmt::sAsyncStmt:
	    genAsyncStmt(stmt->toAsyncStmt(), out, indent);
	    break;
	case Stmt::sCommentStmt:
	    if (stmt->toCommentStmt()->getStmt() != 0)
		 genStmt(stmt->toCommentStmt()->getStmt(), out, indent);
	    break;
	}
}

/***************************************************************************/
void CodeGen::genAssignStmt(const AssignStmt* stmt, ostream& out,
			    PiIndent indent)
{
   genExp(stmt->getLhs(), stmt->getLhs()->getType(), out, indent);
   out << " = ";
   genExp(stmt->getRhs(), stmt->getLhs()->getType(), out, indent);
   out << ";\n";
}

/***************************************************************************/
void CodeGen::genIfStmt(const IfStmt* stmt, ostream& out, PiIndent indent)
{ 
    out << "if (";
    genExp(stmt->getCond(), &aBoolType, out, indent);
    out << ")\n"
	<< indent
	<< "{" << endl;
    genStmts(stmt->getThen(), out, indent.SendIncrement());
    out << indent
	<< "}\n";
    if (stmt->getElse())
    {
       out << indent
	   << "else\n"
	   << indent
	   << "{"
	   << endl;
       genStmts(stmt->getElse(), out, indent.SendIncrement());
       out << indent
	   << "}"
	   << endl;
    } 
}

/***************************************************************************/
void CodeGen::genWhileStmt(const WhileStmt* stmt, ostream& out,
			   PiIndent indent)
{ 

    out << "while (";
    genExp(stmt->getCond(), &aBoolType, out, indent);
    out << ")\n";
    out << indent;
    out << "{\n";
    
    genStmts(stmt->getBody(), out, indent.SendIncrement());
    out << indent;
    out << "}\n";
}

/***************************************************************************/
void CodeGen::genRepeatStmt(const RepeatStmt* stmt, ostream& out,
			    PiIndent indent) 
{ 
   out << "do" << endl;
   out << indent;
   out << "{\n";
   genStmts(stmt->getBody(), out, indent.SendIncrement());
   out << indent;
   out << "} while (";
   genExp(stmt->getCond(), &aBoolType, out, indent);
   out << ");\n"; 
}

/***************************************************************************/
void CodeGen::genForStmt(const ForStmt* stmt, ostream& out,
			 PiIndent indent) 
{
    out << "for (";
    genExp(stmt->getIndex(), &aIntType, out, indent); out << " = "; 

    genExp(stmt->getFrom(), &aIntType, out, indent); out << "; ";
    genExp(stmt->getIndex(), &aIntType, out, indent); 
    if (stmt->getDir() >= 0)
	out << " <= ";  
    else
	out << " >= ";  
    genExp(stmt->getTo(), &aIntType, out, indent); out << "; ";
    genExp(stmt->getIndex(), &aIntType, out, indent); 
    out << " = ";
    genExp(stmt->getIndex(), &aIntType, out, indent); 
    if (stmt->getDir() >= 0)
	out << " + ";
    else 
	out << " - ";
    out << " ("; 
    if (stmt->getBy()) 
	genExp(stmt->getBy(), &aIntType, out, indent);
    else 
	out << "1";
    out << "))\n";
    out << indent;
    out << "{\n";
    genStmts(stmt->getBody(), out, indent.SendIncrement());
    out << indent;
    out << "}\n"; 
}

/***************************************************************************/
void CodeGen::genLoopStmt(const LoopStmt* stmt, ostream& out,
			  PiIndent indent)
{ 
    out << "while(1)\n";
    out << indent;
    out << "{\n";
    genStmts(stmt->getBody(), out, indent.SendIncrement());
    out << indent;
    out << "}\n";
}

/***************************************************************************/
void CodeGen::genReturnStmt(const ReturnStmt* stmt, ostream& out,
			    PiIndent indent)
{ 
    if (stmt->getVal() != 0) {
        out << "return (";
	genExp(stmt->getVal(), stmt->getFunRetType(), out, indent);
        out << ");\n";
        }
    else out << "return;" << endl;
}

/***************************************************************************/
void CodeGen::genExitStmt(const ExitStmt*, ostream& out, PiIndent)
{
   /*
    out << "\n#ifdef IN_CASE_STMT";
    out << "\n #error " << '"' << "Unimplemented: EXIT inside of CASE" << '"';
    out << "\n#endif\n";
   */
   out << "break;"
       << endl;
}

/***************************************************************************/
void CodeGen::genBlockStmt(const BlockStmt* stmt, ostream& out,
			   PiIndent indent)
{   
    // Generate declarations
    genDecls(stmt->getDecls(), out, indent);

    // Generate the statements
    genStmts(stmt->getBody(), out, indent.SendIncrement());
    out << indent;
    out << "}\n";
}

/***************************************************************************/
void CodeGen::genCaseStmt(const CaseStmt* stmt, ostream& out,
			  PiIndent indent)
{
    out << "\n#define IN_CASE_STMT\n";
    out << "switch (";
    genExp(stmt->getExp(), &aAnyType, out, indent);
    out << ")\n";
    out << indent;
    out << "{\n";
    
    const PiList* whens = stmt->getWhens();
    while (whens != 0) {
	When* when = (When*) whens->first();
        const PiList* exps = when->whenLabels;
	if (exps != 0) {
	    while (exps != 0) {
		Exp* exp = (Exp*) exps->first();
		out << indent;
		out << "case ";
		genExp(exp, &aAnyType, out, indent);
		out << " : \n";
		exps = exps->rest();
		}
	    genStmts(when->stmts, out, indent.SendIncrement());
	    out << indent;
	    out << "break;\n";
	    }
	whens = whens->rest();
	} 

    out << indent;
    out << "default : ";
    genStmts(stmt->getOtherwise(), out, indent.SendIncrement());
    out << "break;\n";
    out << indent;
    out << "}\n";
    out << "#undef IN_CASE_STMT\n";
}

/***************************************************************************/
void CodeGen::genDisposeStmt(const DisposeStmt* stmt, ostream& out,
			     PiIndent indent)
{ 
    out << "delete ";
    genExp(stmt->getBody(), &aAnyObjType, out, indent);
    out << ";\n";
}

/********************************************************************/
void CodeGen::genStmts(const PiList* stmts, ostream& out, PiIndent indent)
{
    while (stmts)
    {
       Stmt* stmt = (Stmt *) stmts->first();
       genStmt(stmt, out, indent);
       stmts = stmts->rest();
    }
}

/***************************************************************************/
void CodeGen::genClause(const Clause* clause, int& curName, ostream& out,
			PiIndent indent) {

    switch (clause->clauseKind()) {
	case Clause::cCommentClause:
	    if (clause->toCommentClause()->getClause() != 0)
		genClause(clause->toCommentClause()->getClause(),
			  curName, out, indent);
	    break;
	case Clause::cPropClause:
	    out << "domeClause(";
	    genTerm(clause->toPropClause()->getGoal(), curName, out, indent);
	    out << ",\n    ";
	    genTerms(clause->toPropClause()->getTerms(), curName, out, indent);
	    out << ")";
	    break;
	}
}

/********************************************************************/
void CodeGen::genTerm(const Term* term, int& curName, ostream& out,
		      PiIndent indent) {

    switch(term->termKind()) {
	case Term::tFunctorTerm: {
	    const FunctorTerm* fTerm = term->toFunctorTerm();
	    out << "domePred(";
	    out << '"' << fTerm->getFunctor() << '"';
	    out << ",\n    ";
	    genTerms(fTerm->getArgs(), curName, out, indent);
	    out << ")";
	    } break;
	case Term::tVarTerm:
	    out << "domeVar(";
	    out << '"' << term->toVarTerm()->getVar() << '"';
	    out << ")";
	    break;
	case Term::tConstTerm: {
	    const ConstTerm* cTerm = term->toConstTerm();
	    out << "dome_";
	genTypeTag(cTerm->getType(), cTerm->getExp(), out, indent);
	    } break;
	case Term::tImportTerm:
	    out << "domeImportCall(";
	    canonName(term->toImportTerm()->getObj()->getDecl(), out);
	    out << "_importCall" << curName;
	    out << ")";
	    curName++;
	    break;
	case Term::tListTerm: {
	    const ListTerm* lTerm = term->toListTerm();
	    genListTerm(lTerm, curName, out, indent);
	    } break;
	}
}

/***************************************************************************/
void CodeGen::genListTerm(const ListTerm* term, int& curName, ostream& out,
			  PiIndent indent) {

    int parenCtr = 0;
    const PiList* termPtr = (PiList*) term->getFirstTerms();
    while (termPtr != 0) {
        Term* term = (Term*) termPtr->first();
        out << "domeCons(";
        genTerm(term, curName, out, indent);
        out << ", ";
        termPtr = termPtr->rest();
        parenCtr++;
        }
    if (term->getLast() != 0)
        genTerm(term->getLast(), curName, out, indent);
    else
        out << "0";

    while(parenCtr > 0) {
        out << ")";
        --parenCtr;
        }
}

/********************************************************************/
void CodeGen::genTerms(const PiList* terms, int& curName, ostream& out,
		       PiIndent indent) {

    int parenCtr = 0;
    const PiList* termPtr = (PiList*) terms;
    while (termPtr != 0) {
        Term* term = (Term*) termPtr->first();
        out << "domeCons(";
        genTerm(term, curName, out, indent);
        out << ",\n    ";
        termPtr = termPtr->rest();
        parenCtr++;
        }
    out << "0";
    while(parenCtr > 0) {
        out << ")";
        --parenCtr;
        }
}

/***************************************************************************/
void CodeGen::genAuxClause(const Clause* clause, int& curName,
			   ostream& out, PiIndent indent) {

    switch (clause->clauseKind()) {
	case Clause::cCommentClause:
	    if (clause->toCommentClause()->getClause() != 0)
		genAuxClause(clause->toCommentClause()->getClause(),
			     curName, out, indent);
	    break;
	case Clause::cPropClause:
	    genAuxTerm(clause->toPropClause()->getGoal(), curName, out, indent);
	    genAuxTerms(clause->toPropClause()->getTerms(), curName, out, indent);
	    break;
	}
}

/********************************************************************/
void CodeGen::genAuxTerm(const Term* term, int& curName, ostream& out,
			 PiIndent indent) {

    switch(term->termKind()) {

	case Term::tFunctorTerm:
	    genAuxTerms(term->toFunctorTerm()->getArgs(), curName, out, indent);
	    break;
	case Term::tImportTerm:
	    genImportAuxTerm(term->toImportTerm(), curName, out, indent);
	    break;
	case Term::tListTerm:
	    genAuxTerms(term->toListTerm()->getFirstTerms(), curName, out, indent);
	    break;
	case Term::tVarTerm:
	case Term::tConstTerm:
	    break;
	}
}

/***************************************************************************/
void CodeGen::genImportAuxTerm(const ImportTerm* term, int& curName, 
			       ostream& out, PiIndent indent) {

    /*******************************************************/
	// first generate the body of the import Term.
    out << "inline Bool ";
    canonName(term->getObj()->getDecl(), out); 
    out << "_importCallBody" << curName << "("; 
    canonName(term->getObj()->getDecl(), out); 
    out << "* const self, int I_INDEX";

    const PiList* varPtr = (PiList*) term->getVars();
    while (varPtr != 0) {
        VarExp* var = (VarExp*) varPtr->first();

	out << ", TAGGED& ";
        genExp(var, &aAnyType, out, indent);

        varPtr = varPtr->rest();
        }
    out << ") {\n";

    varPtr = (PiList*) term->getVars();
    out << "if (!(1";
    while (varPtr != 0) {
        VarExp* var = (VarExp*) varPtr->first();
	if (var->getType() != 0) {
	    out << "&& ";
	genTypeCheck(var->getType(), var, out, indent);
	    }
        varPtr = varPtr->rest();
        }
    out << ")) return(0);\n";

    out << "#define this self\n";
    genStmts(term->getStmts(), out, indent);

    out << "return(1);\n";
    out << "#undef this\n";
    out << "}\n\n";


    /*******************************************************/
	// now generate the interface function 

    out << "Bool ";
    canonName(term->getObj()->getDecl(), out); 
    out << "_importCall" << curName;
    out << "(subst* aBind, int level, int I_INDEX, void* context) {\n";

    varPtr = (PiList*) term->getVars();
    while (varPtr != 0) {
        VarExp* var = (VarExp*) varPtr->first();
	out << "    TAGGED ";
	genExp(var, &aAnyType, out, indent);
	out << ";\n";
        out << "    fetch(aBind, " << '"' << var->getVar() << '"' << ", level,";
	genExp(var, &aAnyType, out, indent);
	out << ");\n";
        varPtr = varPtr->rest();
        }

    out << "    Bool ret=\n      ";
    canonName(term->getObj()->getDecl(), out); 
    out << "_importCallBody" << curName;
    out << "((";
    canonName(term->getObj()->getDecl(), out); 
    out << "*) context, I_INDEX";

    varPtr = (PiList*) term->getVars();
    while (varPtr != 0) {
        VarExp* var = (VarExp*) varPtr->first();
	out << ", ";
	genExp(var, &aAnyType, out, indent);
        varPtr = varPtr->rest();
        }
    out << ");\n";

    varPtr = (PiList*) term->getVars();
    while (varPtr != 0) {
        VarExp* var = (VarExp*) varPtr->first();
        out << "    bind(aBind, " << '"' << var->getVar() << '"' << ", level,";
	genExp(var, &aAnyType, out, indent);
	out << ");\n";
        varPtr = varPtr->rest();
        }
    out << "    return(ret);\n";
    out << "}\n\n";

	// FIX	this is dangerous since the coorepondance is so fragile.
	// 	get off one count and the matchup fails
    curName++;
}

/********************************************************************/
void CodeGen::genAuxTerms(const PiList* terms, int& curName, ostream& out,
			  PiIndent indent) {

    while (terms) {
        Term* term = (Term *) terms->first();
        genAuxTerm(term, curName, out, indent);
        terms = terms->rest();
        }
}

/********************************************************************/
void CodeGen::genAsyncDecl(const FunDecl*, ostream&, PiIndent) {

   /* JBco (the obsolete thread system)
   
	// Define the structure for passing args to the task
    out << "struct ";
    canonName(fun->getFunType()->getObj()->getDecl(), out);
    out << "_";
    canonName(fun, out);
    out << "_args {\n    ";

    canonName(fun->getFunType()->getObj()->getDecl(), out);
    out << "_";
    canonName(fun, out);
    out << "_args(";
    canonName(fun->getFunType()->getObj()->getDecl(), out);
    out << "* object, ";
    const PiList* argPtr = fun->getFunType()->getArgs();
    while (argPtr) {
	FunArgDecl* arg = (FunArgDecl* ) argPtr->first();
	genDecl(arg, out, indent);
	out << ", ";
	argPtr = argPtr->rest();
	} 
    if (fun->getFunType()->getRetType()->isKind(Type::voidType))
	out << "ANYFUTURE";
    else {
	out << "FUTURE<";
	genType(fun->getFunType()->getRetType(), "", out, indent);
	out << " >";
	}
    out << "* ret=0)\n         : object_(object)";
    argPtr = fun->getFunType()->getArgs();
    while (argPtr) {
	out << ", ";
	FunArgDecl* arg = (FunArgDecl* ) argPtr->first();
	canonName(arg, out);
	out << "_(";
	canonName(arg, out);
	out << ")";
	argPtr = argPtr->rest();
	} 
    out << ", ret_(ret) {}\n    ";

    canonName(fun->getFunType()->getObj()->getDecl(), out);
    out << "* object_;\n    ";
    if (fun->getFunType()->getRetType()->isKind(Type::voidType))
	out << "ANYFUTURE";
    else {
	out << "FUTURE<";
	genType(fun->getFunType()->getRetType(), "", out, indent);
	out << " >";
	}
    out << "* ret_;\n";

    argPtr = fun->getFunType()->getArgs();
    while (argPtr) {
	FunArgDecl* arg = (FunArgDecl* ) argPtr->first();
	out << "    ";
	char buff[1024];
	ostrstream name(buff, 1024);
	if (arg->getKind() != FunArgDecl::in)
	    name << "&";
	canonName(arg, name);
	name << "_" << '\0';
	genType(arg->getType(), name.str(), out, indent);
	out << ";\n";
	argPtr = argPtr->rest();
	} 
    out << "    };\n\n";

	// Define the task itself
    out << "static void ";
    canonName(fun->getFunType()->getObj()->getDecl(), out);
    out << "_";
    canonName(fun, out);
    out << "_task(void* context);\n\n";
   */
}

/********************************************************************/
void CodeGen::genAsyncDef(const FunDecl*, ostream&, PiIndent) {

   /* JBco (the obsolete thread system)
    out << "static void ";
    canonName(fun->getFunType()->getObj()->getDecl(), out);
    out << "_";
    canonName(fun, out);
    out << "_task(void* context) {\n";

    canonName(fun->getFunType()->getObj()->getDecl(), out);
    out << "_";
    canonName(fun, out);
    out << "_args* args = (";
    canonName(fun->getFunType()->getObj()->getDecl(), out);
    out << "_";
    canonName(fun, out);
    out << "_args*) context;\n";
 
    out << "    ";
    if (!fun->getFunType()->getRetType()->isKind(Type::voidType)) {
	genType(fun->getFunType()->getRetType(), "ret", out, indent);
	out << " = ";
	}

    out << "args->object_->";
    canonName(fun, out);
    out << "(";
    const PiList* argPtr = fun->getFunType()->getArgs();
    while (argPtr) {
	FunArgDecl* arg = (FunArgDecl* ) argPtr->first();
	out << "args->";
	canonName(arg, out);
	out << "_";
	argPtr = argPtr->rest();
	if (argPtr != 0)
	    out << ", ";
	} 
    out << ");\n";

    out << "    if (args->ret_ != 0)\n";
    if (!fun->getFunType()->getRetType()->isKind(Type::voidType)) 
	out << "        *args->ret_ = ret;\n";
    else
	out << "        args->ret_->set();\n";
    out << "    delete args;\n";
    out << "}\n";
    */
}

