/*****************************************************************************/ 
/*                             IDAST.h                                       */
/*****************************************************************************/

/* IDAST.h contains the description of the abstract syntax tree for the
   imort language */

/* AUTHOR: Vance Morrison
   DATE:  10/20/93		 */
/*****************************************************************************/

#ifndef _IDAST_h
#define _IDAST_h 1

#include "PiTime.H"
#include "PiList.H"
#include "SpIostream.H"

#ifndef Bool
#define Bool int
#endif
#ifndef TRUE
#define TRUE 1
#define FALSE 0
#endif

class TypeScope;	// defined elsewhere
class ModScope;		// defined elsewhere
class TypeStatus;	// defined elsewhere
class TypeEnv;		// defined elsewhere

/*****************************************************************************/
/* forward delcarations of all classes defined in this file */
class Node;
class Decl;
class Type;
class Exp;
class Stmt;
class Term;
class Ident;
class Method;
class Clause;

class TimeSpec;
class LitTimeSpec;
class IdentTimeSpec;

class RenameDecl;
class FlashBackDecl;
class ModDecl;
class ImportDecl;
class EnumDecl;
class EnumValDecl;
class ObjectDecl;
class TypeDecl;
class ConstDecl;
class VarDecl;
class FunDecl;
class FunArgDecl;
class CommentDecl;

class DeferedBody;
class MethodBody;

class IdentType;
class FunType;
class ArrayType;
class FutureType;
class AtomicType;
class EnumType;
class ObjectType;
class SynType;

class NewExp;
class AsExp;
class IsAExp;
class CastExp;
class SelfExp;
class IdentExp;
class ArrayExp;
class ApplyExp;
class NullExp;
class IntExp;
class RealExp;
class StringExp;
class CharExp;
class BoolExp;
class VarExp;

class AssignStmt;
class IfStmt;
class WhileStmt;
class RepeatStmt;
class ForStmt;
class LoopStmt;
class ReturnStmt;
class ExitStmt;
class ExpStmt;
class DisposeStmt;
class BlockStmt;
class CaseStmt;
class AsyncStmt;
class CommentStmt;

class PropClause;
class CommentClause;

class FunctorTerm;
class VarTerm;
class ConstTerm;
class ImportTerm;
class ListTerm;

class AFrameDecl;
class AssocInstDecl;
class AssocDecl;
class AssocArgDecl;

void unlinkNodePiList(PiList* elems);
void destroyStrPiList(PiList* elems);
char* str(const char* s, const void* where);
PiList* copyDecls(const PiList* elems, const void* where);
PiList* copyStmts(const PiList* elems, const void* where);
PiList* copyExps(const PiList* elems, const void* where);


/*************************************************************************/
class Comment {
public:
    enum CommentKind_ { prefix, postfix, replace, passThrough };
};
typedef Comment::CommentKind_ CommentKind;

/*************************************************************************/
class Node {
public:
    Node() : refCount(1) {}
    void link() { refCount++; }
    void unlink() { --refCount; if(refCount <=0) destroy(); }
    virtual ~Node() {}

    enum NodeKinds { nDecl, nType, nExp, nTerm, nStmt , nMethod,
                     nTimeSpec, nIdent, nClause };

    virtual NodeKinds nodeKind() const=0;
    virtual const Decl* toDecl() const { return(0); }
    virtual const Type* toType() const { return(0); }
    virtual const Exp* toExp() const { return(0); }
    virtual const Term* toTerm() const { return(0); }
    virtual const Stmt* toStmt() const { return(0); }
    virtual const Method* toMethod() const { return(0); }
    virtual const TimeSpec* toTimeSpec() const { return(0); }
    virtual const Ident* toIdent() const { return(0); }
    virtual const Clause* toClause() const { return(0); }

protected:
    virtual void destroy()=0;	// delete and unlink all its decendants 

private:
    int refCount;		// reference counter
};

/*************************************************************************/
class Type : public Node {
public:
    enum TypeKinds { tIdentType, tArrayType, tAtomicType, tFunType, 
		     tObjectType, tEnumType, tSynType, tFutureType };

    enum TypeKind { anyType=0, // any type
	boolType=1, charType=2, intType=3, realType=4, stringType=5,
	listType=6, arrayType=7, funType=8, 
	anyEnumType=10,    // ENUM type (can be any enumeration */
	anyObjType=11,     // OBJECT type (the type of the arg to DISPOSE) */
	anyFutureType=12,  // futures without a value */
	unknownObjType=13, // the type the NULL operator returns */
	taggedType=19, 	   // tagged type can be any type SMALLER than its value

	    /* these are specialty types used for internal compiler use */
	setType=20, 	  // things that can be intantiated at run time
			  // like objects, enums, INTEGER ... but notMODULE
	modType=21, 	  // things that can be used before '.' operator
	enumType=22, 	  // the type of an enumeration DECLARATION
	objectType=23,    // the type of an object DECLARATION
	moduleType=24, 	  // the type of a module DECLARATION 
	aframeType=30,	  // the type of a aframe DECLARATION
	assocType=31,	 // the type of a association DECLARATION
        assocInstType=32, // the type of a association instantiation DECLARATION
	voidType=33,    /* return type of functions that do not return values */
	varType=40 	/* signifies a DOME variable */
	};

    virtual TypeKinds typeKind() const=0;
    virtual const IdentType* toIdentType() const 	{ return(0); }
    virtual const ArrayType* toArrayType() const 	{ return(0); }
    virtual const FutureType* toFutureType() const 	{ return(0); }
    virtual const AtomicType* toAtomicType() const 	{ return(0); }
    virtual const FunType* toFunType() const 		{ return(0); }
    virtual const ObjectType* toObjectType() const 	{ return(0); }
    virtual const EnumType* toEnumType() const 		{ return(0); }
    virtual const SynType* toSynType() const 		{ return(0); }

	// where is a pointer that indicates which heap to allocate
	// the copy.  The copy will be in the same heap as where 
	// (where = 0 implies the normal transient heap)
    virtual Type* copy(const void* where=0) const=0;
    virtual Bool isKind(TypeKind) const			{ return(FALSE); }
    virtual Bool indexAble() const 			{ return(FALSE); }
    virtual Bool canBe(const Type* t) const=0;
    virtual void typeType(const TypeEnv* env, TypeStatus* status)=0;

        // Needed by Node
    NodeKinds nodeKind() const { return(nType); }
    const Type* toType() const { return(this); }

	// normalize a type.  (dereference top level identifiers)
    const Type* normType() const; 	
};

/*************************************************************************/
/* things common to both type and variable declarations */

class Decl : public Node {
public:
    enum Visibility_ { defaultVis, publicVis, privateVis, cppVis };
    Decl() : envDecl(0) {}

    enum DeclKinds { dModDecl, dImportDecl, dEnumDecl, dObjectDecl, 
	dTypeDecl, dConstDecl, dVarDecl, dFunDecl, dFunArgDecl,
	dFlashBackDecl, dRenameDecl, dEnumValDecl, dCommentDecl,
	dAFrameDecl, dAssocInstDecl, dAssocDecl, dAssocArgDecl };

    virtual DeclKinds declKind() const=0; 
    virtual const ModDecl* toModDecl() const { return(0); }
    virtual const ImportDecl* toImportDecl() const { return(0); }
    virtual const EnumDecl* toEnumDecl() const { return(0); }
    virtual const ObjectDecl* toObjectDecl() const { return(0); }
    virtual const TypeDecl* toTypeDecl() const { return(0); }
    virtual const ConstDecl* toConstDecl() const { return(0); }
    virtual const VarDecl* toVarDecl() const { return(0); }
    virtual const FunDecl* toFunDecl() const { return(0); }
    virtual const FunArgDecl* toFunArgDecl() const { return(0); }
    virtual const FlashBackDecl* toFlashBackDecl() const { return(0); }
    virtual const RenameDecl* toRenameDecl() const { return(0); }
    virtual const CommentDecl* toCommentDecl() const { return(0); }
    virtual const AFrameDecl* toAFrameDecl() const { return(0); }
    virtual const AssocInstDecl* toAssocInstDecl() const { return(0); }
    virtual const AssocDecl* toAssocDecl() const { return(0); }
    virtual const AssocArgDecl* toAssocArgDecl() const { return(0); }    
    virtual const EnumValDecl* toEnumValDecl() const { return(0); }    

	// Object types have to be added to the environment first, before 
	// any typechecking so that mutually recursive objects work.
    virtual void addObjDeclToScope(TypeScope*, TypeStatus*) {};
	// typeDecl type checks a delcaration (but not bodies of functions)
	// does NOT change the environment (that is the next step
	// 'sup' is the superEnvironment (the one containing it) if that
	// makes sence (it doesn't if the decl is in a block for example)
    virtual void typeDecl(const TypeEnv* env, TypeStatus* status,
			  const Decl* sup=0)=0;
	// as a seperate step, the declarations are added to a scope
    virtual void addDeclToScope(TypeScope*, TypeStatus*);
	// finally the bodies of functions are type checked.  This also
	// does not effect the environment
    virtual void typeDeclBody(const TypeEnv* env, TypeStatus* status)=0;

    virtual const char* getName() const=0;
    virtual Visibility_ getVisibility() const=0;
	// gets type for the name (for of OBJ is a OBJECT this returns setType)
	// this is what overloading is done on.
    virtual const Type* getType() const=0;	
	// IF name is a type variable, return the type structure for it
    virtual const Type* getTypeDef() const { return(0); }
	// returns subscope IF the declaration is a Module like structure
    virtual const TypeScope* getScope() const { return(0); }
	// where is a pointer that indicates which heap to allocate
	// the copy.  The copy will be in the same heap as where 
	// (where = 0 implies the normal transient heap)
    virtual Decl* copy(const void* where=0) const=0;

    const PiList* isInterface() const;
    const Decl* getEnvDecl() const 	   { return(envDecl); }
    const Decl* normDecl() const;

        // Needed by Node
    virtual NodeKinds nodeKind() const { return(nDecl); }
    const Decl* toDecl() const { return(this); }

protected:			// this stuff added by type checker
    const Decl* envDecl;	// If the environment in which this is
				// declared is named, then this is the
				// declaration of it. 
};

typedef Decl::Visibility_ Visibility;

/*************************************************************************/
class Exp : public Node {
public:
    enum ExpKinds { eAsExp, eIsAExp, eCastExp, eSelfExp, 
	eIdentExp, eArrayExp, eApplyExp, eNullExp, eIntExp, 
	eRealExp, eStringExp, eCharExp, eBoolExp, eVarExp, eNewExp }; 

    virtual ExpKinds expKind() const=0;
    virtual const AsExp* toAsExp() const { return(0); }
    virtual const IsAExp* toIsAExp() const { return(0); }
    virtual const CastExp* toCastExp() const { return(0); }
    virtual const SelfExp* toSelfExp() const { return(0); }
    virtual const IdentExp* toIdentExp() const { return(0); }
    virtual const ArrayExp* toArrayExp() const { return(0); }
    virtual const ApplyExp* toApplyExp() const { return(0); }
    virtual const NullExp* toNullExp() const { return(0); }
    virtual const IntExp* toIntExp() const { return(0); }
    virtual const RealExp* toRealExp() const { return(0); }
    virtual const StringExp* toStringExp() const { return(0); }
    virtual const CharExp* toCharExp() const { return(0); }
    virtual const BoolExp* toBoolExp() const { return(0); }
    virtual const VarExp* toVarExp() const { return(0); }
    virtual const NewExp* toNewExp() const { return(0); }

	// where is a pointer that indicates which heap to allocate
	// the copy.  The copy will be in the same heap as where 
	// (where = 0 implies the normal transient heap)
    virtual Exp* copy(const void* where=0) const=0;
	// ownership of the returned object is NOT transfered.  Thus
	// the type is read only and is only guarenteed to exist as
	// long as the AST for the code is still 'alive'
	// 'context' is the best guess as to what the type is (often type ANY)
	// If 'varExps' is not zero, the list of DOME variables used 
	// within the expression is returned there.  
    virtual const Type* typeExp(const TypeEnv* env, TypeStatus* status,
	const Type* context, PiList** varExps=0)=0;
    virtual Bool isLValue() const { return(FALSE); }
	// if the expression is a constant, return its value
    virtual const Exp* getConstant() const { return(0); }
	// only valid after the node has been typechecked.  
    virtual const Type* getType() const=0;	

        // Needed by Node
    NodeKinds nodeKind() const { return(nExp); }
    const Exp* toExp() const { return(this); }
};

/*****************************************************************************/
/* a term is a DOME expression */

class Term : public Node {
public:
    enum TermKinds { tFunctorTerm, tVarTerm, tConstTerm, 
	tImportTerm, tListTerm };

    virtual TermKinds termKind() const=0;
    virtual const FunctorTerm* toFunctorTerm() const { return(0); }
    virtual const VarTerm* toVarTerm() const { return(0); }
    virtual const ConstTerm* toConstTerm() const { return(0); }
    virtual const ImportTerm* toImportTerm() const { return(0); }
    virtual const ListTerm* toListTerm() const { return(0); }

    virtual void typeTerm(const TypeEnv* env, TypeStatus* status)=0;
	// where is a pointer that indicates which heap to allocate
	// the copy.  The copy will be in the same heap as where 
	// (where = 0 implies the normal transient heap)
    virtual Term* copy(const void* where=0) const=0;

        // Needed by Node
    NodeKinds nodeKind() const { return(nTerm); }
    const Term* toTerm() const { return(this); }
};

/*************************************************************************/
class Stmt : public Node {
public:
    enum StmtKinds { sAssignStmt, sIfStmt, sWhileStmt, sRepeatStmt,
	sForStmt, sLoopStmt, sReturnStmt, sExitStmt, sExpStmt,
	sDisposeStmt, sBlockStmt, sCaseStmt, sAsyncStmt, sCommentStmt };

    virtual StmtKinds stmtKind() const=0;
    virtual const AssignStmt* toAssignStmt() const { return(0); }
    virtual const IfStmt* toIfStmt() const { return(0); } 
    virtual const WhileStmt* toWhileStmt() const { return(0); }
    virtual const RepeatStmt* toRepeatStmt() const { return(0); }
    virtual const ForStmt* toForStmt() const { return(0); }
    virtual const LoopStmt* toLoopStmt() const { return(0); }
    virtual const ReturnStmt* toReturnStmt() const { return(0); }
    virtual const ExitStmt* toExitStmt() const { return(0); }
    virtual const ExpStmt* toExpStmt() const { return(0); }
    virtual const DisposeStmt* toDisposeStmt() const { return(0); }
    virtual const BlockStmt* toBlockStmt() const { return(0); }
    virtual const CaseStmt* toCaseStmt() const { return(0); }
    virtual const CommentStmt* toCommentStmt() const { return(0); }
    virtual const AsyncStmt* toAsyncStmt() const { return(0); }

	// If 'varExps' is not zero, the list of DOME variables used 
	// within the expression is returned there.  
    virtual void typeStmt(const TypeEnv* env, TypeStatus* status, 
	PiList** varExps=0)=0;
	// where is a pointer that indicates which heap to allocate
	// the copy.  The copy will be in the same heap as where 
	// (where = 0 implies the normal transient heap)
    virtual Stmt* copy(const void* where=0) const=0;

        // Needed by Node
    NodeKinds nodeKind() const { return(nStmt); }
    const Stmt* toStmt() const { return(this); }
};

/*************************************************************************/
class Method : public Node {
public:
    enum MethodKinds { mDeferedBody, mMethodBody };

    virtual MethodKinds methodKind() const=0;
    virtual const DeferedBody* toDeferedBody() const { return(0); }
    virtual const MethodBody* toMethodBody() const { return(0); }

	// where is a pointer that indicates which heap to allocate
	// the copy.  The copy will be in the same heap as where 
	// (where = 0 implies the normal transient heap)
    virtual Method* copy(const void* where=0) const=0;
    virtual void typeMethod(const TypeEnv* env, TypeStatus* status, 
	FunType* type, const Decl* sup)=0;
    virtual void typeMethodBody(const TypeEnv*env, TypeStatus*status)=0;

        // Needed by Node
    NodeKinds nodeKind() const { return(nMethod); }
    const Method* toMethod() const { return(this); }
};

/****************************************************************************/
class TimeSpec : public Node {
public:
    enum CmpResult { TIME_SPEC_UNKNOWN, LESS, NOW, GREATER };
    virtual CmpResult cmp(Time* time) const=0;
	// where is a pointer that indicates which heap to allocate
	// the copy.  The copy will be in the same heap as where 
	// (where = 0 implies the normal transient heap)
    virtual TimeSpec* copy(const void* where=0) const=0;

        // Needed by Node
    NodeKinds nodeKind() const { return(nTimeSpec); }
    const TimeSpec* toTimeSpec() const { return(this); }
};

/*****************************************************************************/
class Clause : public Node {
public:
    enum ClauseKinds { cPropClause, cCommentClause };

    virtual ClauseKinds clauseKind() const=0; 
    virtual const PropClause* toPropClause() const { return(0); }
    virtual const CommentClause* toCommentClause() const { return(0); }
	// where is a pointer that indicates which heap to allocate
	// the copy.  The copy will be in the same heap as where 
	// (where = 0 implies the normal transient heap)
    virtual Clause* copy(const void* where=0) const=0;

    NodeKinds nodeKind() const { return(nClause); }
    const Clause* toClause() const { return(this); }

    virtual void typeClause(const TypeEnv* env, TypeStatus* status)=0;
};

/*************************************************************************/
class Ident : public Node {
public:
    Ident(char* aIdent, Ident* aMod=0) : ident(aIdent), mod(aMod) {}

	// Needed by Node
    NodeKinds nodeKind() const { return(nIdent); }
    const Ident* toIdent() const { return(this); }
    void destroy() {
	delete [] ident; 
	if (mod != 0) mod->unlink(); 
	delete this; 
	}

	// specific to Ident
    const Decl* typeIdent(const TypeEnv* env, TypeStatus* status,
	const Type* context);
    Ident* copy(const void* where=0) const;
    const Ident* getMod() const { return(mod); }
    const char* getIdent() const { return(ident); }

private:
    char* ident;
    Ident* mod;		// optional 0 means current environment
};

/*************************************************************************/
class AtomicType : public Type {
public:
    AtomicType(TypeKind aKind) : kind(aKind) {}

	// Needed by Node
    void destroy() { delete this; }

	// Needed by Type
    TypeKinds typeKind() const 	      		{ return(tAtomicType); }
    const AtomicType* toAtomicType() const 	{ return(this); }
    Bool indexAble() const;
    Bool canBe(const Type* t) const;
    Bool isKind(TypeKind aKind) const { return(aKind == kind); }
    void typeType(const TypeEnv*, TypeStatus*) {}
    Type* copy(const void* where=0) const;

	// Specific to AtomicType 
    TypeKind getKind() const { return(kind); }
private:
    TypeKind kind;
};

/*************************************************************************/
class FunType : public Type {
public:
    enum FunKind_ { ftn, ask, tell, constructor, destructor, query };
    FunType(FunKind_ aKind, PiList* aArgs, Type* aRetType, Type* aObjType=0)
	: kind(aKind), args(aArgs), retType(aRetType), objType(aObjType), 
	  obj(0) {}

	// Needed by Node
    void destroy() { unlinkNodePiList(args); retType->unlink(); 
		     if (objType != 0) objType->unlink(); delete this; }

	// Needed by Type
    TypeKinds typeKind() const 	      		{ return(tFunType); }
    const FunType* toFunType() const 		{ return(this); }
    Bool canBe(const Type* t) const;
    void typeType(const TypeEnv* env, TypeStatus* status);
    Type* copy(const void* where=0) const;

    FunKind_ getKind() const		{ return(kind); }
    const PiList* getArgs() const		{ return(args); }
    const Type* getRetType() const	{ return(retType); }
    const Type* getObjType() const	{ return(objType); }
    const ObjectType* getObj() const	{ return(obj); }

private:
    FunKind_ kind;
    PiList* args;			// of FunArgDecl
    Type* retType;
    Type* objType;		// optional object type 

private:			// this stuff is added by the typechecker 
    const ObjectType* obj;	// the object this is a method of
};

typedef FunType::FunKind_ FunKind;

/*************************************************************************/
class IdentType : public Type {
public:
    IdentType(Ident* aIdent) : ident(aIdent), def(0) {}
    IdentType(Ident* aIdent, const Decl* aDef) : ident(aIdent), def(aDef) {}

	// Needed by Node
    void destroy() { ident->unlink(); delete this; }

	// Needed by Type
    const IdentType* toIdentType() const     { return(this); }
    TypeKinds typeKind() const 	     	     { return(tIdentType); }
    Bool indexAble() const           { return(def->getTypeDef()->indexAble()); }
    Bool isKind(TypeKind aKind) const { return(def->getType()->isKind(aKind)); }
    Bool canBe(const Type* t) const 	{ return(def->getTypeDef()->canBe(t)); }
    void typeType(const TypeEnv* env, TypeStatus* status);
    Type* copy(const void* where=0) const;

	// Specific to IdentType
    const Decl* getDef() const { return(def); }
    const Ident* getIdent() const { return(ident); }

private:
    Ident* ident;

private:			// this stuff is added by the typechecker 
    const Decl* def;
};

/*************************************************************************/
class ArrayType : public Type {
public:
    ArrayType(Type* aElemType, Exp* aBound) 
	: elemType(aElemType), bound(aBound), indexType(0), evalBound(0) {}

	// Needed by Node
    void destroy() { elemType->unlink(); 
	if (bound != 0) bound->unlink(); delete this; }

	// Needed by Type
    TypeKinds typeKind() const 	      		{ return(tArrayType); }
    const ArrayType* toArrayType() const 	{ return(this); }
    Bool canBe(const Type* t) const;
    void typeType(const TypeEnv* env, TypeStatus* status);
    Type* copy(const void* where=0) const;

	// ArrayType specific routines
    const Type* getIndexType() const 	{ return(indexType); }
    const Type* getElemType() const 	{ return(elemType); }
    const Exp* getBound() const 	{ return(bound); }
    int getEvalBound() const 		{ return(evalBound); }

private:

private:
    Type* elemType;
    Exp* bound;

private:	// this stuff is added by the typechecker 
    const Type* indexType;
    int evalBound;
};

/*************************************************************************/
class FutureType : public Type {
public:
    FutureType(Type* aSubType) : subType(aSubType) {}

	// Needed by Node
    void destroy() { subType->unlink(); delete this; }

	// Needed by Type
    TypeKinds typeKind() const 	      		{ return(tFutureType); }
    const FutureType* toFutureType() const 	{ return(this); }
    Bool canBe(const Type* t) const;
    Bool indexAble() const                   { return(subType->indexAble()); }
    Bool isKind(TypeKind aKind) const	     { return(subType->isKind(aKind)); }
    void typeType(const TypeEnv* env, TypeStatus* status);
    Type* copy(const void* where=0) const;

	// FutureType specific routines
    const Type* getSubType() const 	{ return(subType); }

private:

private:
    Type* subType;
};

/*************************************************************************/
class SynType : public Type {
public:
    SynType(char* aName, Type* aTypeRef)
	: name(aName), typeRef(aTypeRef) {}

	// Needed by Node
    void destroy() { delete [] name; typeRef->unlink(); delete this; }

        // Needed by Type
    TypeKinds typeKind() const 	     	     { return(tSynType); }
    Bool indexAble() const                   { return(typeRef->indexAble()); }
    Bool isKind(TypeKind aKind) const	     { return(typeRef->isKind(aKind)); }
    Bool canBe(const Type* t) const          { return(typeRef->canBe(t)); }
    void typeType(const TypeEnv* env, TypeStatus* status);
    Type* copy(const void* where=0) const;
    const SynType* toSynType() const 	     { return(this); }

	// Specific to SynType
    const char* getName() const 	     { return(name); }
    const Type* getTypeRef() const 	     { return(typeRef); }

private:
    char* name;
    Type* typeRef;
};

/*************************************************************************/
class EnumType : public Type {
public:
    EnumType(char* aName, PiList* aEnums) : name(aName), enums(aEnums) {}

	// Needed by Node
    void destroy() { delete [] name; unlinkNodePiList(enums); delete this; }

	// Needed by Type 
    TypeKinds typeKind() const 	      		{ return(tEnumType); }
    const EnumType* toEnumType() const 		{ return(this); }
    Bool canBe(const Type* t) const;
    Bool indexAble() const 		  	{ return(TRUE); }
    void typeType(const TypeEnv*, TypeStatus*) {}
    Type* copy(const void* where=0) const;

	// Specific to EnumDecl
    const char* getName() const  { return(name); }
    const PiList* getEnums() const { return(enums); }
    const EnumDecl* getDecl() const { return(decl); }
    friend class EnumDecl;

private:
    char* name;	
    PiList* enums;	// of EnumValDecl

private:			// this stuff is added by the typechecker 
    const EnumDecl* decl;	// declaration for this obj
};

/*************************************************************************/
class ObjectType : virtual public Type {
public:
    ObjectType(char* aName, PiList* aSupers, PiList* aDecls, PiList* aClauses,
	       Bool simobj = FALSE, int totSimobj = 0);
    ~ObjectType();

	// Needed by Node
    void destroy() { 
	delete [] name;
	unlinkNodePiList(decls);
	unlinkNodePiList(supers);
	unlinkNodePiList(clauses);
	delete this; }

	// Needed by Type 
    TypeKinds typeKind() const 	      		{ return(tObjectType); }
    const ObjectType* toObjectType() const 	{ return(this); }
    Bool indexAble() const			{ return(FALSE); }
    Bool canBe(const Type* t) const;
    void typeType(const TypeEnv* env, TypeStatus* status);
    Type* copy(const void* where=0) const;

	// ObjectType specific routines
    const char* getName() const		{ return(name); }
    const PiList* getDecls() const	{ return(decls); }
    const PiList* getSuperNames() const	{ return(supers); }
    const PiList* getSupers() const	{ return(superClasses); }
    const PiList* getClauses() const	{ return(clauses); }
    PiList* abstractClass() const		{ return(deferedMeths); }
    const TypeScope* getScope() const	{ return(scope); }
    const PiList* getExternalMeths() const	{ return(externalMeths); }
    const ObjectDecl* getDecl() const 	{ return(declm); }
    void typeBody(const TypeEnv* env, TypeStatus* status);

    Bool isSimObj() const         { return(simObj); }
    int getTotSimObj() const	{ return(totSimObj); }

    friend class ObjectDecl;
    friend class FunDecl; 

private:
    char* name;	
    PiList* decls;	// of Decl
    PiList* supers;	// of Ident
    PiList* clauses;	// of Clause

private:			// this stuff is added by the typechecker 
    PiList* superClasses;		// of ObjectType  
    PiList* deferedMeths;		// deferred methods
    TypeScope* scope;		// associates names with declaration in mod
    ObjectDecl* declm;		// declaration for this obj
    PiList* externalMeths;	// of Decl // method describe outside my body
    Bool simObj;                // JB
    int totSimObj;              // JB (applies only if simObj==TRUE)
};

/*************************************************************************/
class CommentDecl : public Decl {
public:
    CommentDecl(CommentKind aKind, char* aMsg, Decl* aDecl) 
	: kind(aKind), msg(aMsg), decl(aDecl) {}

       // Need by Node
    void destroy() { if (decl != 0) decl->unlink(); 
	delete [] msg; delete this; }

        // needed by Decl
    DeclKinds declKind() const             { return(dCommentDecl); }
    virtual const CommentDecl* toCommentDecl() const { return(this); }
    const char* getName() const            
	{ if (decl != 0) return(decl->getName()); else return("<comment>"); }
    const Type* getType() const
	{ if (decl != 0) return(decl->getType()); else return(0); }
    void typeDecl(const TypeEnv* env, TypeStatus* status, const Decl* sup=0)
	{ if (decl != 0) decl->typeDecl(env, status, sup); }
    void addDeclToScope(TypeScope* scope, TypeStatus* status)
	{ if (decl != 0) decl->addDeclToScope(scope, status); }
    void typeDeclBody(const TypeEnv* env, TypeStatus* status)
	{ if (decl != 0) decl->typeDeclBody(env, status); }
    void addObjDeclToScope(TypeScope* scope, TypeStatus* status)
	{ if (decl != 0) decl->addObjDeclToScope(scope, status); }
    Visibility getVisibility() const
	{ return((decl != 0) ? decl->getVisibility() : publicVis); }
    Decl* copy(const void* where=0) const;

	// specific to CommentDecl
    CommentKind getKind() const { return(kind); }
    const Decl* getDecl() const { return(decl); }
    const char* getMsg() const { return(msg); }

private:
    CommentKind kind;
    char* msg;		// the comment message (complete with comment operator)
    Decl* decl;		// is NULL for a replace comment
};

/*************************************************************************/
class ModDecl : public Decl {
public:
    ModDecl(char* aName, PiList* aDecls, PiList* aInterface, Visibility aVis) 
	: vis(aVis), name(aName), decls(aDecls), interface(aInterface), 
	scope(0) {}
    ~ModDecl();

	// Need by Node
    void destroy();

	// needed by Decl
    DeclKinds declKind() const 		   { return(dModDecl); }
    const ModDecl* toModDecl() const 	   { return(this); }
    void typeDecl(const TypeEnv* env, TypeStatus* status, const Decl* sup=0);
    void typeDeclBody(const TypeEnv* env, TypeStatus* status);
    Visibility getVisibility() const   	   { return(vis); }
    const Type* getType() const;
    const char* getName() const		   { return(name); }
    const TypeScope* getScope() const	   { return(scope); }
    Decl* copy(const void* where=0) const;

	// ModDecl specific routines
    const PiList* getDecls() const	   { return(decls); }
    const PiList* getInterface() const	{ return(interface); }

private:
    Visibility vis;
    char* name;
    PiList* decls;		// of Decl
    PiList* interface;		// of PiList* (char* char* tuples)

private:  			// this stuff is added by the typechecker 
    TypeScope* scope;	// associates names with declaration in mod
};

/***************************************************************************/
class FlashBackDecl : public Decl {
public:
    FlashBackDecl(Decl* aDecl, TimeSpec* aTime, char* aWork=0) : 
	decl(aDecl), time(aTime), work(aWork) {}

       // Need by Node
    void destroy() { time->unlink(); decl->unlink(); 
	if (work != 0) delete [] work; delete this; }

        // needed by Decl
    DeclKinds declKind() const             { return(dFlashBackDecl); }
    virtual const FlashBackDecl* toFlashBackDecl() const { return(this); }
    const char* getName() const            { return(decl->getName()); }
    const Type* getType() const            { return(decl->getType()); }
    void typeDecl(const TypeEnv* env, TypeStatus* status, const Decl* sup=0);
    void addDeclToScope(TypeScope* scope, TypeStatus* status) {
	decl->addDeclToScope(scope, status); }
    void typeDeclBody(const TypeEnv* env, TypeStatus* status);
    void addObjDeclToScope(TypeScope* scope, TypeStatus* status);
    Visibility getVisibility() const   	   { return(decl->getVisibility()); }
    Decl* copy(const void* where=0) const;

	// specific to FlashBackDecl
    const Decl* getDecl() const { return(decl); }

private:
    char* work;		// working version name NULL means current one
    TimeSpec* time;
    Decl* decl;
};

/***************************************************************************/
class RenameDecl : public Decl {
public:

    RenameDecl(char* aName, Ident* aIdent, Type* aType, Visibility aVis)
	: vis(aVis), name(aName), ident(aIdent), type(aType), 
	  origDecl(0) {}

       // Need by Node
    void destroy() { 
	delete [] name; 
	ident->unlink(); 
	if (type != 0) type->unlink();
	delete this; 
	}

        // needed by Decl
    DeclKinds declKind() const             { return(dRenameDecl); }
    virtual const RenameDecl* toRenameDecl() const { return(this); }
    const char* getName() const            { return(name); }
    const Type* getType() const            { 
	if (origDecl == 0) return(0);
	else return(origDecl->getType()); }
    const Type* getTypeDef() const         { 
	if (origDecl == 0) return(0);
	else return(origDecl->getTypeDef()); }
    void typeDecl(const TypeEnv* env, TypeStatus* status, const Decl* sup=0);
    void typeDeclBody(const TypeEnv*, TypeStatus*) {}
    Visibility getVisibility() const   	   { return(vis); }
    Decl* copy(const void* where=0) const;

	// specific to RenameDecl
    Ident* getIdent() const { return(ident); }
    const Decl* getOrigDecl() const { return(origDecl); }

private:
    Visibility vis;
    char* name;			// new name for this Assoc
    Ident* ident;		// old name for this Assoc
    Type* type;			// type of old name (optional)

private:                        // this stuff is added by the typechecker
    const Decl* origDecl;	// Ident points to this declaration
};

/*************************************************************************/
class ImportDecl : public Decl {
public:
    ImportDecl(Ident* aIdent, Visibility aVis)
	: vis(aVis), ident(aIdent), modDecl(0) {}

	// Needed by Node
    void destroy() { 
	ident->unlink(); 
	delete this; }

	// Needed by Decl
    DeclKinds declKind() const 			{ return(dImportDecl); }
    const ImportDecl* toImportDecl() const 	{ return(this); }

    void typeDecl(const TypeEnv* env, TypeStatus* status,const Decl* sup=0);
    void addDeclToScope(TypeScope* scope, TypeStatus* status);
    void typeDeclBody(const TypeEnv*, TypeStatus*) {}
    const Type* getType() const;
    const char* getName() const  { return(""); }
    Visibility getVisibility() const   	   	{ return(vis); }
    Decl* copy(const void* where=0) const;

	// ImportDecl specific functions
    const Ident* getIdent() const { return(ident); }
    const Decl* getModDecl() const { return(modDecl); }
    
private:
    Visibility vis;
    Ident* ident;

private:			// this stuff is added by the typechecker 
    const Decl* modDecl;	// declaration for 'ident'
};

/*************************************************************************/
class ConstDecl : public Decl {
public:
    ConstDecl(char* aName, Exp* aValue, Visibility aVis)
	: vis(aVis), name(aName), value(aValue), type(0) {}

	// Needed by Node
    void destroy() { delete [] name; value->unlink(); delete this; }

	// Needed by Decl
    DeclKinds declKind() const 	{ return(dConstDecl); }
    void typeDecl(const TypeEnv* env, TypeStatus* status,const Decl* sup=0);
    void typeDeclBody(const TypeEnv*, TypeStatus*) {}
    const Type* getType() const { return(type); }
    const char* getName() const  { return(name); }
    const ConstDecl* toConstDecl() const { return(this); }
    Visibility getVisibility() const  	   { return(vis); }
    Decl* copy(const void* where=0) const;

	// Specific to ConstDecl
    const Exp* getValue() const { return(value); }

private:
    Visibility vis;
    char* name;
    Exp* value;

private:		// this stuff is added by the typechecker 
    const Type* type;	// type of above expression
};

/*************************************************************************/
class TypeDecl : virtual public Decl {
public:
    TypeDecl(char* aName, Type* aTypeRef, Visibility aVis) 
	: vis(aVis), name(aName), typeRef(aTypeRef) {}

    void destroy() { 
	delete [] name;
	typeRef->unlink(); 
	delete this; }

	// Needed by Decl
    DeclKinds declKind() const { return(dTypeDecl); }
    const TypeDecl* toTypeDecl() const { return(this); }
    void typeDecl(const TypeEnv* env, TypeStatus* status,const Decl* sup=0);
    void typeDeclBody(const TypeEnv*, TypeStatus*) {}
    const Type* getType() const;
    const char* getName() const  { return(name); }
    Visibility getVisibility() const { return(vis); }
    Decl* copy(const void* where=0) const;
    const Type* getTypeDef() const 		{ return(typeRef); }

	// specific to TypeDecl
    const Type* getTypeRef() const	{ return(typeRef); }

private:
    Visibility vis;
    char* name;	
    Type* typeRef;
};

/*************************************************************************/
class VarDecl : public Decl {
public:
    VarDecl(char* aName, Type* aType, Visibility aVis, Bool aByPointer = FALSE) 
       :vis(aVis), name(aName), type(aType), byPointer(aByPointer) {}

	// Needed by Node
    void destroy() { delete [] name; type->unlink(); delete this; }

	// Needed by Decl
    DeclKinds declKind() const 		{ return(dVarDecl); }
    const VarDecl* toVarDecl() const 	{ return(this); }
    void typeDecl(const TypeEnv* env, TypeStatus* status,const Decl* sup=0);
    void typeDeclBody(const TypeEnv*, TypeStatus*) {}
    const Type* getType() const { return(type); }
    const char* getName() const  { return(name); }
    Visibility getVisibility() const { return(vis); }
    Decl* copy(const void* where=0) const;
   Bool isByPointer() const  { return(byPointer); } // JB

private:
    Visibility vis;
    char* name;	
    Type* type;
   Bool byPointer; // JB
};

/*************************************************************************/
class FunDecl : public Decl {
public:
    FunDecl(char* aName, FunType* aType, Visibility aVis, Method* aBody=0) 
	: vis(aVis), name(aName), type(aType), body(aBody), asyncRet(0),
	  external(0) {}
    ~FunDecl() { if (asyncRet != 0) asyncRet->unlink(); }

	// Needed by Node
    void destroy();

	// Needed by Decl
    DeclKinds declKind() const 		{ return(dFunDecl); }
    const FunDecl* toFunDecl() const 	{ return(this); }
    void typeDecl(const TypeEnv* env, TypeStatus* status,const Decl* sup=0);
    void typeDeclBody(const TypeEnv* env, TypeStatus* status);
    const Type* getType() const  { return(type); }
    const char* getName() const  { return(name); }
    Visibility getVisibility() const   	   { return(vis); }
    Bool isDeferedMeth() const	{ if (body == 0) return(FALSE);
				  else return(body->toDeferedBody() != 0); }
    Decl* copy(const void* where=0) const;

	// Specific to FunDecl
    const FunType* getFunType() const { return(type); }
    const Method* getBody() const { return(body); }
    Bool isCalledAsync() const { return(asyncRet != 0); }
    Bool isExternal() const { return(external); }
    const Type* retType(Bool async) const;

private:
    Visibility vis;
    char* name;	
    FunType* type;
    Method* body;

private:		// this stuff is added by the typechecker 
    Type* asyncRet;	// non-zero means that this method is called async
			// this value is its return value in that case.
    Bool external;	// this is a declaration of a method outside the
			// body of the object
};

/*************************************************************************/
class EnumValDecl : public Decl {
public:
    EnumValDecl(char* aName)
        : name(aName), type(0) {}
 
        // Needed by Node
    void destroy() { delete [] name; delete this; }
 
        // Needed by Decl
    DeclKinds declKind() const { return(dEnumValDecl); };
    const EnumValDecl* toEnumValDecl() const { return(this); }    
    void typeDecl(const TypeEnv*, TypeStatus*, const Decl* sup=0);
    void typeDeclBody(const TypeEnv*, TypeStatus*) {}
    const Type* getType() const { return(type); }
    const char* getName() const  { return(name); }
    Visibility getVisibility() const { return(publicVis); }
    Decl* copy(const void* where=0) const;

private:
    char* name; 

private: 	// this stuff added by typechecer
    const Type* type;
};

/*************************************************************************/
class EnumDecl : virtual public Decl {
public:
    EnumDecl(char* aName, EnumType* aTypeRef, Visibility aVis) : 
	vis(aVis), name(aName), typeRef(aTypeRef) {}

    void destroy() { 
	delete [] name;
	typeRef->unlink(); 
	delete this; }

	// Needed by Decl
    DeclKinds declKind() const  		{ return(dEnumDecl); }
    const EnumDecl* toEnumDecl() const 		{ return(this); }
    void typeDecl(const TypeEnv* env, TypeStatus* status, const Decl* sup=0);
    void addDeclToScope(TypeScope* scope, TypeStatus* status);
    void typeDeclBody(const TypeEnv*, TypeStatus*) {}
    const Type* getType() const;
    const char* getName() const  		{ return(name); }
    Visibility getVisibility() const   	   	{ return(vis); }
    Decl* copy(const void* where=0) const;
    const Type* getTypeDef() const 		{ return(typeRef); }
	
	// specific to EnumDecl
    const EnumType* getTypeRef() const		{ return(typeRef); }

private:
    Visibility vis;
    char* name;	
    EnumType* typeRef;
};

/*************************************************************************/
class ObjectDecl : virtual public Decl {
public:
    ObjectDecl(char* aName, ObjectType* aTypeRef, Visibility aVis) 
	: vis(aVis), name(aName), typeRef(aTypeRef) {}
    void destroy() { 
	delete [] name;
	typeRef->unlink(); 
	delete this; }

	// Needed by Decl
    DeclKinds declKind() const 			{ return(dObjectDecl); }
    const ObjectDecl* toObjectDecl() const 	{ return(this); }
    void typeDecl(const TypeEnv* env, TypeStatus* status,const Decl* sup=0);
    void addObjDeclToScope(TypeScope* scope, TypeStatus* status);
    void addDeclToScope(TypeScope*, TypeStatus*);
    void typeDeclBody(const TypeEnv* env, TypeStatus* status);
    const Type* getType() const;
    const char* getName() const  		{ return(name); }
    Visibility getVisibility() const   	   	{ return(vis); }
    const TypeScope* getScope() const	{ return(typeRef->getScope()); }
    Decl* copy(const void* where=0) const;
    const Type* getTypeDef() const 		{ return(typeRef); }

	// specific to ObjectDecl
    const ObjectType* getTypeRef() const 	{ return(typeRef); }
    
private:
    Visibility vis;
    char* name;	
    ObjectType* typeRef;
};

/*************************************************************************/
class FunArgDecl : public Decl {
public:
    enum ArgType { in, out, inout };
    FunArgDecl(ArgType aKind, char* aName, Type* aType) 
        :  name(aName), kind(aKind), type(aType) {}

	// Needed by Node
    void destroy() { 
	if (name != 0) delete [] name; type->unlink(); delete this; }

	// Needed by Decl
    const FunArgDecl* toFunArgDecl() const { return(this); }
    DeclKinds declKind() const { return(dFunArgDecl); }
    void typeDecl(const TypeEnv* env, TypeStatus* status,const Decl* sup=0);
    void typeDeclBody(const TypeEnv*, TypeStatus*) {}
    const Type* getType() const { return(type); }
    const char* getName() const { if (name != 0) return(name); else return("");}
    Visibility getVisibility() const   	   { return(privateVis); }
    Decl* copy(const void* where=0) const;

	// Specific to FunArgDecl
    ArgType getKind() const { return(kind); }

private:
    ArgType kind;
    char* name;	
    Type* type;
};

/***************************************************************************/
/************************   ASSOCIATIONS   *********************************/
/***************************************************************************/

/* RAW: 26 Jan 1995 */

class AFrameDecl : public Decl {
public:
    AFrameDecl(char* aName, PiList* aDecls, Visibility aVis) 
	: vis(aVis), name(aName), decls(aDecls), scope(0) {}
    ~AFrameDecl();

	// needed by Node
    void destroy() { delete [] name; unlinkNodePiList(decls); delete this; }

	// needed by Decl
    DeclKinds declKind() const 		   { return(dAFrameDecl); }
    const Type* getType() const;
    const char* getName() const  	   { return(name); }
    const AFrameDecl* toAFrameDecl() const { return(this); }
    Visibility getVisibility() const   	   { return(vis); }
    const TypeScope* getScope() const	   { return(scope); }
    void typeDecl(const TypeEnv* env, TypeStatus* status,const Decl* sup=0);
    void typeDeclBody(const TypeEnv*, TypeStatus*);
    Decl* copy(const void* where=0) const;

	// specific to AFrameDecl
    const PiList* getDecls() const	   { return(decls); }
    
private:
    Visibility vis;
    char* name;	
    PiList* decls;	// of Decl

private:
    TypeScope* scope;	// associates names with declaration in mod
};

/*************************************************************************/
/* RAW: 26 Jan 1995 */

class AssocDecl : public Decl {
public:
    AssocDecl(char* aName, PiList* aArgs, PiList* aSuperIdents, PiList* aReqs, 
	    PiList* aSpecs, PiList* aRoles, Visibility aVis)
	: vis(aVis), name(aName), roles(aRoles), args(aArgs), 
	  superIdents(aSuperIdents), reqs(aReqs), specs(aSpecs), 
	  supers(0), testArgs(0), assocDecls(0), scope(0),
	  assocEnv(0) {}
    ~AssocDecl();

	// Needed by Node
    void destroy() { 
	delete [] name;
	unlinkNodePiList(args);
	unlinkNodePiList(supers);
	unlinkNodePiList(reqs);
	unlinkNodePiList(specs);
	unlinkNodePiList(roles); 
	delete this; }

	// Needed by Decl
    DeclKinds declKind() const 			{ return(dAssocDecl); }
    const AssocDecl* toAssocDecl() const 	{ return(this); }
    void typeDecl(const TypeEnv* env, TypeStatus* status,const Decl* sup=0);
    void typeDeclBody(const TypeEnv*, TypeStatus*);
    const Type* getType() const;
    const char* getName() const  		{ return(name); }
    Visibility getVisibility() const   	   	{ return(vis); }
    Decl* copy(const void* where=0) const;

	// AssocDecl specific routines
    const PiList* getArgs() const		{ return(args); }
    const PiList* getSuperIdents() const	{ return(superIdents); }
    const PiList* getReqs() const		{ return(reqs); }
    const PiList* getSpecs() const	{ return(specs); }
    const PiList* getRoles() const	{ return(roles); }
    const PiList* getSupers() const	{ return(supers); }
    const TypeEnv* getEnv() const	{ return(assocEnv); }

private:
    Visibility vis;
    char* name;	
    PiList* args;		// of AssArgDecl;
    PiList* superIdents;	// of Ident
    PiList* reqs;		// of Decl
    PiList* specs;	// of Clause
    PiList* roles;	// of Decl

private:			// this stuff is added by the typechecker 
    PiList* testArgs;		// of Type (argument list to typecheck with)
    PiList* assocDecls;		// decls for the assoc itself (index, add etc)
    PiList* supers;		// of AssocDecl  
    TypeScope* scope;  		// associates names with declaration in mod
    TypeEnv* assocEnv;		// the environment of the assoc
};

/***************************************************************************/
/* RAW: 26 Jan 1995 */
class AssocInstDecl : public Decl {
public:
    AssocInstDecl(char* aName,Ident* aIdent, PiList* aArgs, Visibility aVis):
      vis(aVis), name(aName), ident(aIdent), args(aArgs),
      scope(0), decls(0), assocDecls(0), 
      argDecls(0), reqDecls(0), assoc(0) {}
    ~AssocInstDecl();

	// Needed by Node
    void destroy() { delete [] name; ident->unlink(); 
		     unlinkNodePiList(args); delete this; }

	// Needed by Decl
    DeclKinds declKind() const  		{ return(dAssocInstDecl); }
    const AssocInstDecl* toAssocInstDecl() const 	{ return(this); }
    void typeDecl(const TypeEnv* env, TypeStatus* status, const Decl* sup=0);
    void typeDeclBody(const TypeEnv*, TypeStatus*);
    const Type* getType() const;
    const char* getName() const  	  { return(name); }
    Visibility getVisibility() const   	   { return(vis); }
    const TypeScope* getScope() const	   { return(scope); }
    Decl* copy(const void* where=0) const;

	// Specific to AssocInstDecl
    const Ident* getIdent() const	{ return(ident); }
    const PiList* getArgs() const		{ return(args); }
    const PiList* getDecls() const	{ return(decls); }
    const PiList* getAssocDecls() const	{ return(assocDecls); }
    const AssocDecl* getAssoc() const	{ return(assoc); }

private:
    Visibility vis;
    char* name;	
    Ident* ident;
    PiList* args;		// of Type

private:
    PiList* assocDecls;		// decls for the assoc itself (index, add etc)
    PiList* argDecls;		// decls which instantiate the assoc args
    PiList* reqDecls;		// decls which instantiate the req declarations
    PiList* decls;		// the association decls instantiated
    const AssocDecl* assoc;	// association I am instantiating
    TypeScope* scope;	// associates names with declaration in mod
};

/*************************************************************************/
/* RAW: 26 Jan 1995 */
class RangeSpec {
public:
    RangeSpec(int* aLValue=0,int* aUValue=0) 
	: lValue(aLValue), uValue(aUValue) {}

	// Specific to RangeSpec
    RangeSpec* copy(const void* where=0) const;
    void destroy() { 
	if (lValue != 0) delete lValue; 
	if (uValue != 0) delete uValue; delete this; }
    int getUValue() const { 
	if (uValue != 0) return(*uValue);
	else if (uValue == 0 && lValue != 0) return(*lValue);
	else return(-1);
	}
    int getLValue() const {
	if (lValue != 0) return(*lValue); else return(0); }

private:
    int* lValue; // range lower value
    int* uValue; // range upper value
};

/*************************************************************************/
class AssocArgDecl : public Decl {
public:
    AssocArgDecl(char* aName, Type* aType, RangeSpec* aRange) 
        : name(aName), type(aType), range(aRange) {}

	// Needed by Node
    void destroy() { if (name != 0) delete [] name; 
	type->unlink(); 
	if (range != 0) range->destroy(); 
	delete this; }

	// Needed by Decl
    const AssocArgDecl* toAssocArgDecl() const { return(this); }
    DeclKinds declKind() const { return(dAssocArgDecl); }
    void typeDecl(const TypeEnv* env, TypeStatus* status, const Decl* sup=0);
    void typeDeclBody(const TypeEnv*, TypeStatus*) {};
    const Type* getType() const;
    const Type* getTypeDef() const { return(type); }
    const char* getName() const { return(name); }
    Visibility getVisibility() const { return(privateVis); }
    Decl* copy(const void* where=0) const;

private:
    char* name;		
    Type* type;
    RangeSpec* range;             // two integers; -1 means "unspecified"
};

/***************************************************************************/
/************************   END ASSOCIATIONS   *****************************/
/***************************************************************************/

/*************************************************************************/
class NewExp : public Exp {
public:
    NewExp(PiList* aNewArgs, Type* aType, PiList* aArgs) 
	: newArgs(aNewArgs), type(aType), args(aArgs),
	  constrDecl(0), newDecl(0) {}

	// Needed by Node
    void destroy() { 
	type->unlink();
	unlinkNodePiList(args); 
	unlinkNodePiList(newArgs); 
	delete this; }

	// Needed by Exp
    ExpKinds expKind() const { return(eNewExp); }
    const NewExp* toNewExp() const { return(this); }
    const Type* typeExp(const TypeEnv* env, TypeStatus* status,
	const Type* context, PiList** varExps=0);
    Exp* copy(const void* where=0) const;
    const Type* getType() const { return(type); }

	// Specific to NewExp
    const PiList* getArgs() const	{ return(args); }
    const PiList* getNewArgs() const { return(newArgs); }
    const FunDecl* getConstrDecl() const { return(constrDecl); }
    const FunDecl* getNewDecl() const { return(newDecl); }
    const Type* typeNewExp(const TypeEnv* env,TypeStatus* stat,PiList** varExps);

private:
    PiList* newArgs;	// of Exp
    Type* type;
    PiList* args;		// of Exp

private:	// this stuff added by typeChecker
    const FunDecl* constrDecl;	// declaration of constructor being called
    const FunDecl* newDecl;	// declaration of NEW operator
};

/*************************************************************************/
class AsExp : public Exp {
public:
    AsExp(Exp* aBody, Type* aType) : body(aBody), type(aType) {}

	// Needed by Node
    void destroy() { body->unlink(); type->unlink();  delete this; }

	// Needed by Exp
    ExpKinds expKind() const { return(eAsExp); }
    const AsExp* toAsExp() const { return(this); }
    const Type* typeExp(const TypeEnv* env, TypeStatus* status,
	const Type* context, PiList** varExps=0);
    Exp* copy(const void* where=0) const;
    const Type* getType() const { return(type); }

	// Specific to AsExp
    const Exp* getBody() const { return(body); }

private:
    Exp* body;
    Type* type;
};

/*************************************************************************/
class IsAExp : public Exp {
public:
    IsAExp(Exp* aBody, Type* aSubType) : body(aBody), subType(aSubType) {}

	// Needed by Node
    void destroy() { body->unlink(); 
	if (subType != 0) subType->unlink();  delete this; }

	// Needed by Exp
    ExpKinds expKind() const { return(eIsAExp); }
    const IsAExp* toIsAExp() const { return(this); }
    const Type* typeExp(const TypeEnv* env, TypeStatus* status,
	const Type* context, PiList** varExps=0);
    Exp* copy(const void* where=0) const;
    const Type* getType() const;

	// Specific to IsAExp 
    const Exp* getBody() const { return(body); }
    const Type* getSubType() const { return(subType); }

private:
    Exp* body;
    Type* subType;
};

/*************************************************************************/
class CastExp : public Exp {
public:
    CastExp(Exp* aBody, Type* aType) 
	: body(aBody), type(aType), bodyType(0) {}

	// Needed by Node
    void destroy() { body->unlink(); type->unlink();  delete this; }

	// Needed by Exp
    ExpKinds expKind() const { return(eCastExp); }
    const CastExp* toCastExp() const { return(this); }
    const Type* typeExp(const TypeEnv* env, TypeStatus* status,
	const Type* context, PiList** varExps=0);
    Exp* copy(const void* where=0) const;
    const Type* getType() const { return(type); }

	// Specific to CastExp
    const Exp* getBody() const { return(body); }
    const Type* getBodyType() const { return(bodyType); }

private:
    Exp* body;
    Type* type;

private:	// this stuff is added by the typechecker 
    const Type* bodyType;
};

/*************************************************************************/
class SelfExp : public Exp {
public:
    SelfExp() : type(0) {}

	// Needed by Node
    void destroy() { delete this; }

	// Needed by Exp
    ExpKinds expKind() const { return(eSelfExp); }
    const SelfExp* toSelfExp() const { return(this); }
    const Type* typeExp(const TypeEnv* env, TypeStatus* status,
	const Type* context, PiList** varExps=0);
    Exp* copy(const void* where=0) const;
    const Type* getType() const { return(type); }

private:	// this stuff is added by the typechecker 
    const Type* type;
};

/*************************************************************************/
class IdentExp : public Exp {
public:
   IdentExp(Ident* aIdent, const Decl* aDef=0) : ident(aIdent), def(aDef) {}

	// Needed by Node
    void destroy() 	    { ident->unlink(); delete this; }

	// Needed by Exp
    ExpKinds expKind() const { return(eIdentExp); }
    const IdentExp* toIdentExp() const { return(this); }
    const Type* typeExp(const TypeEnv* env, TypeStatus* status,
	const Type* context, PiList** varExps=0);
    Bool isLValue() const;
    const Exp* getConstant() const;
    Exp* copy(const void* where=0) const;
    const Type* getType() const { return(def->getType()); }

	// Specific to IdentExp
    const Decl* getDef() const	{ return(def); }
    const Ident* getIdent() const { return(ident); }
    
 private:
   Ident* ident;

 private:	// this stuff is added by the typechecker 
   const Decl* def;
};

/*************************************************************************/
class ArrayExp : public Exp {
public:
    ArrayExp(Exp* aArray, Exp* aIndex) : array(aArray), index(aIndex),
	type(0) {}

	// Needed by Node
    void destroy() { array->unlink(); index->unlink(); delete this; }

	// Needed by Exp
    ExpKinds expKind() const { return(eArrayExp); }
    const ArrayExp* toArrayExp() const { return(this); }
    const Type* typeExp(const TypeEnv* env, TypeStatus* status,
	const Type* context, PiList** varExps=0);
    Bool isLValue() const { return(TRUE); }
    Exp* copy(const void* where=0) const;
    const Type* getType() const { return(type); }

	// Specific to ArrayExp
    const Exp* getArray() const { return(array); }
    const Exp* getIndex() const { return(index); }

private:
    Exp* array;
    Exp* index;

private:			// this stuff is added by the typechecker 
     const Type* type;
};

/*************************************************************************/
class ApplyExp : public Exp {
public:
    ApplyExp(FunKind aKind, Exp* aObj, Ident* aName, PiList* aArgs, Ident* aFrom):
	kind(aKind), obj(aObj), name(aName), args(aArgs), from(aFrom),
	decl(0), fromObj(0), ReentryLabel(0) {}
    ApplyExp(Ident* aName, PiList* aArgs) : 
	kind(FunType::ftn), obj(0), name(aName), args(aArgs), from(0),
	decl(0), fromObj(0), ReentryLabel(0) {}

	// Needed by Node
    void destroy() { 
	name->unlink(); 
	if (obj != 0) obj->unlink();
	if (from != 0) from->unlink();
	unlinkNodePiList(args); 
	delete this; }

	// Needed by Exp
    ExpKinds expKind() const { return(eApplyExp); }
    const ApplyExp* toApplyExp() const { return(this); }
    const Type* typeExp(const TypeEnv* env, TypeStatus* status,
	const Type* context, PiList** varExps=0);
    Exp* copy(const void* where=0) const;
    const Type* getType() const;

	// Specific to ApplyExp
    FunKind getKind() const { return(kind); }
    const Exp* getObj() const { return(obj); }
    const Ident* getName() const { return(name); }
    const PiList* getArgs() const { return(args); }
    const Decl* getDecl() const { return(decl); }
    const Ident* getFrom() const { return(from); }
    const ObjectType* getFromObj() const { return(fromObj); }
    int getReentryLabel() const { return(ReentryLabel); }

   static int CumReentryLabel;  // JB

 private:
   FunKind kind;
   Exp* obj;	// object of the method being called (not used for 'ftn' type)
   Ident* name; // name of method being called
   PiList* args;	// args of method being called

   Ident* from;	// optional From clause

 private:  			// this stuff is added by the typechecker 
   const Decl* decl;		// declaration of method being called
   int ReentryLabel;            // JB

   const ObjectType* fromObj;	// type for 'from'
};

/*************************************************************************/
class NullExp : public Exp {
public:
    NullExp() {}

	// Needed by Node
    void destroy() { delete this; }

	// Needed by Exp
    ExpKinds expKind() const { return(eNullExp); }
    const NullExp* toNullExp() const { return(this); }
    const Type* typeExp(const TypeEnv* env, TypeStatus* status,
	const Type* context, PiList** varExps=0);
    Bool isLValue() const { return(FALSE); }
    const Exp* getConstant() const { return(this); }
    Exp* copy(const void* where=0) const;
    const Type* getType() const;

private:
};

/*************************************************************************/
class IntExp : public Exp {
public:
    IntExp(int* aValue) : value(aValue) {}

	// Needed by Node
    void destroy() { delete value; delete this; }

	// Needed by Exp
    ExpKinds expKind() const { return(eIntExp); }
    const IntExp* toIntExp() const { return(this); }
    const Type* typeExp(const TypeEnv* env, TypeStatus* status,
	const Type* context, PiList** varExps=0);
    Bool isLValue() const { return(FALSE); }
    const Exp* getConstant() const { return(this); }
    Exp* copy(const void* where=0) const;
    const Type* getType() const;

	// Specific to IntExp
    int getValue() const { return(*value); }
private:
    int* value;
};

/*************************************************************************/
class RealExp : public Exp {
public:
    RealExp(double* aValue) : value(aValue) {}

	// Needed by Node
    void destroy() { delete value; delete this; }

	// Needed by Exp
    ExpKinds expKind() const { return(eRealExp); }
    const RealExp* toRealExp() const { return(this); }
    const Type* typeExp(const TypeEnv* env, TypeStatus* status,
	const Type* context, PiList** varExps=0);
    Bool isLValue() const { return(FALSE); }
    const Exp* getConstant() const { return(this); }
    Exp* copy(const void* where=0) const;
    const Type* getType() const;

	// Specific to RealExp
    double getValue() const { return(*value); }
private:
    double* value;
};

/*************************************************************************/
class StringExp : public Exp {
public:
    StringExp(char* aValue) : value(aValue) {}

	// Needed by Node
    void destroy() { delete [] value; delete this; }

	// Needed by Exp
    ExpKinds expKind() const { return(eStringExp); }
    const StringExp* toStringExp() const { return(this); }
    const Type* typeExp(const TypeEnv* env, TypeStatus* status,
	const Type* context, PiList** varExps=0);
    Bool isLValue() const { return(FALSE); }
    const Exp* getConstant() const { return(this); }
    Exp* copy(const void* where=0) const;
    const Type* getType() const;

	// Specific to StringExp
    const char* getValue() const { return(value); }

private:
    char* value;
};

/*************************************************************************/
class CharExp : public Exp {
public:
    CharExp(char* aValue) : value(aValue) {}

	// Needed by Node
    void destroy() { delete value; delete this; }

	// Needed by Exp
    ExpKinds expKind() const { return(eCharExp); }
    const CharExp* toCharExp() const { return(this); }
    const Type* typeExp(const TypeEnv* env, TypeStatus* status,
	const Type* context, PiList** varExps=0);
    Bool isLValue() const { return(FALSE); }
    const Exp* getConstant() const { return(this); }
    Exp* copy(const void* where=0) const;
    const Type* getType() const;

	// Specific to CharExp
    char getValue() const { return(*value); }

private:
    char* value;	// note this is NOT a string, but a single char
};

/*************************************************************************/
class BoolExp : public Exp {
public:
    BoolExp(Bool aVal) : val(aVal) {}

	// Needed by Node
    void destroy() { delete this; }

	// Needed by Exp
    ExpKinds expKind() const { return(eBoolExp); }
    const BoolExp* toBoolExp() const { return(this); }
    const Type* typeExp(const TypeEnv* env, TypeStatus* status,
	const Type* context, PiList** varExps=0);
    Bool isLValue() const { return(FALSE); }
    const Exp* getConstant() const { return(this); }
    Exp* copy(const void* where=0) const;
    const Type* getType() const;

	// Specific to BoolExp
    Bool getValue() const { return(val); }
private:
    Bool val;
};

/*****************************************************************************/
class VarExp : public Exp {
public:
     VarExp(char* aVar) : var(aVar), type(0) {}

	// Needed by Node
     void destroy() {
        delete [] var;
        delete this; }

	// Needed by Exp
    ExpKinds expKind() const { return(eVarExp); }
    const VarExp* toVarExp() const { return(this); }
    Bool isLValue() const { return(TRUE); }
    const Type* typeExp(const TypeEnv* env, TypeStatus* status,
	const Type* context, PiList** varExps=0);
    Exp* copy(const void* where=0) const;
    const Type* getType() const;

	// specific to VarExp
    const char* getVar() const { return(var); }

private:
     char* var;

private:			// this stuff is added by the typechecker 
     const Type* type;
};

/***************************************************************************/
class CommentStmt : public Stmt {
public:
    CommentStmt(CommentKind aKind, char* aMsg, Stmt* aStmt) 
	: kind(aKind), msg(aMsg), stmt(aStmt) {}

       // Need by Node
    void destroy() { if (stmt != 0) stmt->unlink(); 
	delete [] msg; delete this; }

        // needed by Stmt
    StmtKinds stmtKind() const             { return(sCommentStmt); }
    const CommentStmt* toCommentStmt() const { return(this); }
    Stmt* copy(const void* where=0) const;

    void typeStmt(const TypeEnv* env, TypeStatus* status, PiList** varExps=0)
	{ if (stmt != 0) stmt->typeStmt(env, status, varExps); }

	// specific to CommentStmt
    CommentKind getKind() const { return(kind); }
    const Stmt* getStmt() const { return(stmt); }
    const char* getMsg() const { return(msg); }

private:
    CommentKind kind;
    char* msg;		// the comment message (complete with comment operator)
    Stmt* stmt;		// is NULL for a replace comment
};

/*************************************************************************/
class AsyncStmt : public Stmt
{
 public:
    enum WaitKind { awNone, awWait, awWaitUntil };

   AsyncStmt(Exp* aLhs, ApplyExp* aApply, Exp *aDestEnumIdExp = 0,
	     enum WaitKind aWaitKind = awNone, Exp *aWaitValExp = 0);
   
	// Needed by Node
    void destroy() { if (lhs != 0) lhs->unlink(); apply->unlink(); delete this;}

	// Needed by Stmt
    StmtKinds stmtKind() const { return(sAsyncStmt); }
    const AsyncStmt* toAsyncStmt() const { return(this); }
    void typeStmt(const TypeEnv* env, TypeStatus* status, PiList** varExps=0);
    Stmt* copy(const void* where=0) const;

	// Accessors
    const Exp* getLhs() const { return lhs; }
    const ApplyExp* getApply() const { return apply; }
    const Exp* getDestEnumIdExp() const { return destEnumIdExp; } // JB
    int isWaitToTell() const { return (kind == awWait); } // JB
    int isWaitUntilTimeToTell()
       const { return (kind == awWaitUntil); } // JB
   const Exp *getWaitValExp() const { return waitValExp; } // JB

private:
   Exp *lhs;
   ApplyExp *apply;
   Exp *destEnumIdExp; // JB: enum id for scheduling desination simobj's event
   enum WaitKind kind; // JB: how to interpret waitVal
   Exp *waitValExp;    // JB: the "WAIT" or "WAIT UNTIL" value, if applicable
 public:
   static PiList *globalAsyncStmtPiList; // JB: for collecting DEFINE_EVENT info
};

/*************************************************************************/
class AssignStmt : public Stmt {
public:
    AssignStmt(Exp* aLhs, Exp* aRhs) : lhs(aLhs), rhs(aRhs), 
	lhsType(0), rhsType(0) {}

	// Needed by Node
    void destroy() { lhs->unlink(); rhs->unlink(); delete this; }

	// Needed by Stmt
    StmtKinds stmtKind() const { return(sAssignStmt); }
    const AssignStmt* toAssignStmt() const { return(this); }
    void typeStmt(const TypeEnv* env, TypeStatus* status, PiList** varExps=0);
    Stmt* copy(const void* where=0) const;

	// Accessors
    const Exp* getLhs() const { return(lhs); }
    const Exp* getRhs() const { return(rhs); }
    const Type* getLhsType() const { return(lhsType); }
    const Type* getRhsType() const { return(rhsType); }
private:
    Exp* lhs;
    Exp* rhs;

private:	// this stuff is added by the typechecker 
    const Type* lhsType;
    const Type* rhsType;
};

/*************************************************************************/
class IfStmt : public Stmt {
public:
    IfStmt(Exp* aCond, PiList* aThen, PiList* aElse)
	: cond(aCond), then(aThen), else_(aElse) {}

	// Needed by Node
    void destroy() { 
	cond->unlink();
	unlinkNodePiList(then); 
	unlinkNodePiList(else_); 
	delete this; }

	// Needed by Stmt
    StmtKinds stmtKind() const { return(sIfStmt); }
    const IfStmt* toIfStmt() const { return(this); } 
    void typeStmt(const TypeEnv* env, TypeStatus* status, PiList** varExps=0);
    Stmt* copy(const void* where=0) const;

	// Accessors
    const Exp* getCond() const  { return(cond); }
    const PiList* getThen() const { return(then); }
    const PiList* getElse() const { return(else_); }
private:
    Exp* cond;
    PiList* then;		// of Stmt
    PiList* else_;	// of Stmt
};

/*************************************************************************/
class WhileStmt : public Stmt {
public:
    WhileStmt(Exp* aCond, PiList* aBody) : cond(aCond), body(aBody) {}

	// Needed by Node
    void destroy() { cond->unlink(); unlinkNodePiList(body); delete this; }

	// Needed by Stmt
    StmtKinds stmtKind() const { return(sWhileStmt); }
    const WhileStmt* toWhileStmt() const { return(this); }
    void typeStmt(const TypeEnv* env, TypeStatus* status, PiList** varExps=0);
    Stmt* copy(const void* where=0) const;

	// Accessors
    const Exp* getCond() const  { return(cond); }
    const PiList* getBody() const { return(body); }

private:
    Exp* cond;
    PiList* body;		// of Stmt
};

/*************************************************************************/
class RepeatStmt : public Stmt {
public:
    RepeatStmt(Exp* aCond, PiList* aBody) : cond(aCond), body(aBody) {}

	// Needed by Node
    void destroy() { cond->unlink(); unlinkNodePiList(body); delete this; }

	// Needed by Stmt
    StmtKinds stmtKind() const { return(sRepeatStmt); }
    const RepeatStmt* toRepeatStmt() const { return(this); }
    void typeStmt(const TypeEnv* env, TypeStatus* status, PiList** varExps=0);
    Stmt* copy(const void* where=0) const;

	// Accessors
    const Exp* getCond() const  { return(cond); }
    const PiList* getBody() const { return(body); }
private:
    Exp* cond;
    PiList* body;		// of Stmt
};

/*************************************************************************/
class ForStmt : public Stmt {
public:
    ForStmt(Exp* aIndex, Exp* aFrom, int aDir, Exp* aTo, Exp* aBy, PiList* aBody) 
	:index(aIndex), from(aFrom), dir(aDir), to(aTo), by(aBy), body(aBody) {}

	// Needed by Node
    void destroy() { 
	index->unlink(); 
	from->unlink(); 
	to->unlink(); 
	if (by != 0) by->unlink(); 
	unlinkNodePiList(body); 
	delete this; }

	// Needed by Stmt
    StmtKinds stmtKind() const { return(sForStmt); }
    const ForStmt* toForStmt() const { return(this); }
    void typeStmt(const TypeEnv* env, TypeStatus* status, PiList** varExps=0);
    Stmt* copy(const void* where=0) const;

	// Accessors
    const Exp* getIndex() const { return(index); }
    const Exp* getFrom() const  { return(from); }
    int getDir() const    	{ return(dir); }
    const Exp* getTo() const    { return(to); }
    const Exp* getBy() const    { return(by); }
    const PiList* getBody() const { return(body); }
private:
    Exp* index;
    Exp* from;
    int dir;		// postive means counting up, negative counting down
    Exp* to;
    Exp* by;
    PiList* body;		// of Stmt
};

/*************************************************************************/
class LoopStmt : public Stmt {
public:
    LoopStmt(PiList* aBody) : body(aBody) {}

	// Needed by Node
    void destroy() { unlinkNodePiList(body); delete this; }

	// Needed by Stmt
    StmtKinds stmtKind() const { return(sLoopStmt); }
    const LoopStmt* toLoopStmt() const { return(this); }
    void typeStmt(const TypeEnv* env, TypeStatus* status, PiList** varExps=0);
    Stmt* copy(const void* where=0) const;

	// Accessors
    const PiList* getBody() const { return(body); }
private:
    PiList* body;		// of Stmt
};

/*************************************************************************/
class ReturnStmt : public Stmt {
public:
    ReturnStmt(Exp* aVal) : val(aVal), funRetType(0) {}

	// Needed by Node
    void destroy() { if (val != 0) val->unlink(); delete this; }

	// Needed by Stmt
    StmtKinds stmtKind() const { return(sReturnStmt); }
    const ReturnStmt* toReturnStmt() const { return(this); }
    void typeStmt(const TypeEnv* env, TypeStatus* status, PiList** varExps=0);
    Stmt* copy(const void* where=0) const;

	// Accessors
    const Exp* getVal() const { return(val); }
    const Type* getFunRetType() const { return(funRetType); }
private:
    Exp* val;	

private:	// this stuff is added by the typechecker 
    const Type* funRetType;
};

/*************************************************************************/
class ExitStmt : public Stmt {
public:
    ExitStmt() {}

	// Needed by Node
    void destroy() { delete this; }

	// Needed by Stmt
    StmtKinds stmtKind() const { return(sExitStmt); }
    const ExitStmt* toExitStmt() const { return(this); }
    void typeStmt(const TypeEnv* env, TypeStatus* status, PiList** varExps=0);
    Stmt* copy(const void* where=0) const;

private:
};

/*************************************************************************/
class ExpStmt : public Stmt {
public:
    ExpStmt(Exp* aExp) : exp(aExp) {}

	// Needed by Node
    void destroy() { exp->unlink(); delete this; }

	// Needed by Stmt
    StmtKinds stmtKind() const { return(sExpStmt); }
    const ExpStmt* toExpStmt() const { return(this); }
    void typeStmt(const TypeEnv* env, TypeStatus* status, PiList** varExps=0);
    Stmt* copy(const void* where=0) const;

	// Specific to ExpStmt
    const Exp* getExp() const { return(exp); }

private:
    Exp* exp;
};

/*************************************************************************/
class DisposeStmt : public Stmt {
public:
    DisposeStmt(Exp* aBody) : body(aBody) {}

	// Needed by Node
    void destroy() { 
	body->unlink(); 
	delete this; }

	// Needed by Stmt
    StmtKinds stmtKind() const { return(sDisposeStmt); }
    const DisposeStmt* toDisposeStmt() const { return(this); }
    void typeStmt(const TypeEnv* env, TypeStatus* status, PiList** varExps=0);
    Stmt* copy(const void* where=0) const;

	// Specific to DisposeStmt
    const Exp* getBody() const { return(body); }

private:
    Exp* body;
};

/*************************************************************************/
class BlockStmt : public Stmt {
public:
    BlockStmt(PiList* aDecls, PiList* aBody) : decls(aDecls), body(aBody) {}

	// Needed by Node
    void destroy() { 
	unlinkNodePiList(decls); 
	unlinkNodePiList(body); 
	delete this; }

	// Needed by Stmt
    StmtKinds stmtKind() const { return(sBlockStmt); }
    const BlockStmt* toBlockStmt() const { return(this); }
    void typeStmt(const TypeEnv* env, TypeStatus* status, PiList** varExps=0);
    Stmt* copy(const void* where=0) const;

	// Specific to BlockStmt
    const PiList* getDecls() const { return(decls); }
    const PiList* getBody() const	{ return(body); }
private:
    PiList* decls;	// of Decl
    PiList* body;		// of Stmt
};

/*************************************************************************/
class CaseStmt;
/* the when class can be thought of as part of the CaseStmt class */
class When {
public:
    When(PiList* aWhenLabels, PiList* aStmts)
	: whenLabels(aWhenLabels), stmts(aStmts) {}

    friend class CaseStmt;
    void destroy() { unlinkNodePiList(whenLabels); 
		     unlinkNodePiList(stmts); delete this; }
    When* copy(const void* where=0) const;

public:
    PiList* whenLabels;	// of Exp (constant enumerable type)
    PiList* stmts;	// of Stmt (default case)
};

/*************************************************************************/
class CaseStmt : public Stmt {
public:
    CaseStmt(Exp* aExp, PiList* aWhens, PiList* aOtherwise)
	: expm(aExp), whens(aWhens), otherwise(aOtherwise) {}

	// Needed by Node
    void destroy();

	// Needed by Stmt
    StmtKinds stmtKind() const { return(sCaseStmt); }
    const CaseStmt* toCaseStmt() const { return(this); }
    void typeStmt(const TypeEnv* env, TypeStatus* status, PiList** varExps=0);
    Stmt* copy(const void* where=0) const;

	// Specific to CaseStmt
    const Exp* getExp() const { return(expm); }
    const PiList* getWhens() const { return(whens); }
    const PiList* getOtherwise() const { return(otherwise); }

private:
    Exp* expm;
    PiList* whens;	// of When
    PiList* otherwise;	// of Stmt
};

/*****************************************************************************/
class FunctorTerm : public Term {
public:
     FunctorTerm(char* aFunctor, PiList* aArgs = 0) 
	: functor(aFunctor), args(aArgs) {}

	// Needed by Node
     void destroy() {
        delete [] functor;
        unlinkNodePiList(args);
        delete this; }

	// Needed by Term
    TermKinds termKind() const { return(tFunctorTerm); }
    const FunctorTerm* toFunctorTerm() const { return(this); }
    void typeTerm(const TypeEnv* env, TypeStatus* status);
    Term* copy(const void* where=0) const;

	// Specific to FunctorTerm
     const char* getFunctor() const { return(functor); }
     const PiList* getArgs() const { return(args); }
private:
     char* functor;
     PiList* args;	// of Term
};

/*****************************************************************************/
class VarTerm : public Term {
public:
     VarTerm(char* aVar) : var(aVar) {}

	// Needed by Node
     void destroy() {
        delete [] var;
        delete this; }

	// Needed by Term
    TermKinds termKind() const { return(tVarTerm); }
    const VarTerm* toVarTerm() const { return(this); }
    void typeTerm(const TypeEnv*, TypeStatus*) {}
    Term* copy(const void* where=0) const;

	// specific to VarTerm
    const char* getVar() const { return(var); }

private:
     char* var;
};

/*****************************************************************************/
class ConstTerm : public Term {
public:
     ConstTerm(Exp* aExp) : exp(aExp), type(0) {}

	// Needed by Node
     void destroy() {
        exp->unlink();
        delete this; }

	// Needed by Term
    TermKinds termKind() const { return(tConstTerm); }
    const ConstTerm* toConstTerm() const { return(this); }
    void typeTerm(const TypeEnv* env, TypeStatus* status);
    Term* copy(const void* where=0) const;

	// specific to ConstTerm
    const Exp* getExp() const { return(exp); }
    const Type* getType() const { return(type); }

private:
     Exp* exp;

private:			// this stuff is added by the typechecker 
     const Type* type;
};

/*****************************************************************************/
class ImportTerm : public Term {
public:
    ImportTerm(PiList* aStmts) : stmts(aStmts), obj(0), vars(0), index(0) {}
    ~ImportTerm() { if (vars != 0) delete vars; 
		    if (index != 0) delete index; }
    

	// Needed by Node
     void destroy() {
	unlinkNodePiList(stmts);
        delete this; }

	// Needed by Term
    TermKinds termKind() const { return(tImportTerm); }
    const ImportTerm* toImportTerm() const { return(this); }
    void typeTerm(const TypeEnv* env, TypeStatus* status);
    Term* copy(const void* where=0) const;

	// specific to ImportTerm
    const PiList* getStmts() const { return(stmts); }
    const ObjectType* getObj() const { return(obj); }
    const PiList* getVars() const { return(vars); }
    const VarDecl* getIndex() const { return(index); }

private:
    PiList* stmts;	// of Stmt

private:			// this stuff is added by the typechecker 
    const ObjectType* obj;			// object for this term
    PiList* vars;		// of VarExp		(dome vars used in stmts)
    VarDecl* index;		// the variable declaration for variable INDEX
};

/*****************************************************************************/
class ListTerm : public Term {
public:
     ListTerm(PiList* aFirstTerms, VarTerm* aLast) 
	: firstTerms(aFirstTerms), last(aLast) {}

	// Needed by Node
     void destroy() {
        unlinkNodePiList(firstTerms);
	if (last != 0) last->unlink();
        delete this; }

	// Needed by Term
    TermKinds termKind() const { return(tListTerm); }
    const ListTerm* toListTerm() const { return(this); }
    void typeTerm(const TypeEnv* env, TypeStatus* status);
    Term* copy(const void* where=0) const;

	// specific to ListTerm
    const PiList* getFirstTerms() const { return(firstTerms); }
    const VarTerm* getLast() const { return(last); }

private:
     PiList* firstTerms;	// of Term
     VarTerm* last;
};

/*****************************************************************************/
class PropClause : public Clause {
public:
    PropClause(PiList* aTerms, FunctorTerm* aGoal) : terms(aTerms), goal(aGoal) {}

	// Needed by Node
    void destroy() { unlinkNodePiList(terms); goal->unlink(); delete this; }

	// Needed by Clause
    ClauseKinds clauseKind() const { return(cPropClause); }
    const PropClause* toPropClause() const { return(this); }
    void typeClause(const TypeEnv* env, TypeStatus* status);
    Clause* copy(const void* where=0) const;

	// Specific to PropClause
    const PiList* getTerms() const { return(terms); }
    const FunctorTerm* getGoal() const { return(goal); }

private:
     PiList* terms;	// of Term
     FunctorTerm* goal;
};

/***************************************************************************/
class CommentClause : public Clause {
public:
    CommentClause(CommentKind aKind, char* aMsg, Clause* aClause) 
	: kind(aKind), msg(aMsg), clause(aClause) {}

       // Need by Node
    void destroy() { if (clause != 0) clause->unlink(); 
	delete [] msg; delete this; }

        // needed by Clause
    ClauseKinds clauseKind() const             { return(cCommentClause); }
    virtual const CommentClause* toCommentClause() const { return(this); }
    void typeClause(const TypeEnv* env, TypeStatus* status)
	{ if (clause != 0) clause->typeClause(env, status); }
    Clause* copy(const void* where=0) const;

	// specific to CommentClause
    CommentKind getKind() const { return(kind); }
    const Clause* getClause() const { return(clause); }
    const char* getMsg() const { return(msg); }

private:
    CommentKind kind;
    char* msg;		// the comment message (complete with comment operator)
    Clause* clause;	// is NULL for a replace comment
};

/*************************************************************************/
class DeferedBody : public Method {
public:
    DeferedBody() {}

	// Needed by Node
    void destroy() { delete this; }

	// Needed by Method 
    MethodKinds methodKind() const { return(mDeferedBody); };
    const DeferedBody* toDeferedBody() const { return(this); }
    void typeMethod(const TypeEnv*, TypeStatus*, FunType*, const Decl*) {}
    void typeMethodBody(const TypeEnv*, TypeStatus*) {}
    Method* copy(const void* where=0) const;
};

/*************************************************************************/
class MethodBody : public Method {
 public:
   MethodBody(PiList* aDecls, PiList* aStmts, PiList* inits=0) 
      : decls(aDecls), stmts(aStmts), constrInits(inits), localScope(0),
	TotReentryLabels(0) {}
   ~MethodBody();

	// Needed by Node
    void destroy() {
	unlinkNodePiList(constrInits); 
	unlinkNodePiList(decls); 
	unlinkNodePiList(stmts); 
	delete this; 
	}

	// Needed by Method 
    MethodKinds methodKind() const { return(mMethodBody); };
    const MethodBody* toMethodBody() const { return(this); }
    void typeMethod(const TypeEnv* env, TypeStatus* status, FunType* type,
	const Decl* sup);
    void typeMethodBody(const TypeEnv* env, TypeStatus* status);
    Method* copy(const void* where=0) const;

	// Specific to MethodBody
    const PiList* getDecls() const { return(decls); }
    const PiList* getStmts() const { return(stmts); }
    const PiList* getConstrInits() const { return(constrInits); }
    int getTotReentryLabels() const { return(TotReentryLabels); }

 private:
   PiList* decls;	// of Decls
   PiList* stmts;	// of Stmt
   PiList* constrInits;	  // of ApplyExp (only used if it is a construtor)
 private:		  // this stuff is added by the typechecker 
   TypeScope* localScope; // local scope for decls.
   int TotReentryLabels;  // JB
};

/****************************************************************************/
class IdentTimeSpec : public TimeSpec {
public:
    IdentTimeSpec(char* aName) : name(aName) {} 

       // Need by Node
    void destroy() { delete [] name; delete this; }

	// needed by TimeSpec
    CmpResult cmp(Time* time) const;
    TimeSpec* copy(const void* where=0) const;

private:
    char* name;
};

/****************************************************************************/
class LitTimeSpec : public TimeSpec {
public:
    LitTimeSpec(Time aTicks) : ticks(aTicks) {} 

       // Need by Node
    void destroy() { delete this; }

	// needed by TimeSpec
    CmpResult cmp(Time* time) const;
    TimeSpec* copy(const void* where=0) const;

private:
    Time ticks;	
};

#endif

