/**************************************************************************/
/*                           IDFormat.c    IDFormat.c                     */
/**************************************************************************/

/* This file defines the formater (Pretty Printer) for the Import/Dome
   Language
*/

/* AUTHOR: Vance Morrison
   DATE  : 9/19/94                  */
/**************************************************************************/

#include "SpIostream.H"		// FIX for debugging 
#include "IDFormat.H"
#include "IDAST.H"
#include "SpCtype.H"
#include "SpString.H"


/**************************************************************************/
static void litString(const char* str, char term, Formatter& out) {

    while(*str != 0) {
        if (isgraph(*str) || *str == ' ') {
            if (*str == term || *str == '\\')
                out << '\\';
            out << *str;
            }
        else if (*str == '\n')
            out << "\\n";
        else {
            out << '\\';
            out << (char) (((unsigned int) *str) / 64);
            out << (char) (((unsigned int) *str) % 64) / 8;
            out << (char) (((unsigned int) *str)) % 8;
            }
        str++;
        }
}

/**************************************************************************/
static void asIdent(const char* str, Formatter& out) {

    static const char* keywords[] = {
	"ABOUT", "AND", "AM", "ANYOBJ", "ARRAY", "AT", "ASK", "BEGIN",
	"BLOCK", "BOOLEAN", "BY", "CASE", "CAST", "CHAR", "CONST",
	"CONSTRUCTOR", "DESTRUCTOR", "DISPOSE", "DO", "DOWNTO", "DOME",
	"DURATION", "ELSE", "ELSEIF", "END", "EXIT", "EXPORT", "FALSE", "FOR",
	"FROM", "FUNCTION", "IF", "IMPORT", "IN", "INIT", "INOUT", "INTEGER",
	"INTERFACE", "INTERRUPT", "ISA", "IS", "LIST", "LOOP", "METHOD",
	"MOD", "MODULE", "NEW", "NOT", "NULL", "OBJECT", "OF", "OR", 
	"OTHERWISE", "OUT", "PM", "PRIVATE", "PUBLIC", "QUERY", "REAL", 
	"REPEAT", "RETURN", "SUBCLASS", "SELF", "STRING", "TAGGED", 
	"TELL", "THEN", "TO", "TRUE", "TYPE", "UNTIL", "VAR", "WHEN", 
	"WHILE", "VIEWED", 0 };

    if (!(isalpha(*str) || *str == '_')) {
	out.canBreak(FALSE);
	out << "OP("; litString(str, ')', out); out << ")";
	out.canBreak(TRUE);
	return;
	}

    register const char* ptr = str;
    while (*ptr != 0) {
	if (!(isalnum(*ptr) || *ptr == '_')) {
	    out.canBreak(FALSE);
	    out << "OP("; litString(str, ')', out); out << ")";
	    out.canBreak(TRUE);
	    return;
	    }
	ptr++;
	}
    
    register const char** key = keywords;
    while (*key != 0) {
	if (strcmp(*key, str) == 0) {
	    out << "OP(" << str << ")";
	    return;
	    }
	key++;
	}

    out << str;
}

/**************************************************************************/
Bool IDFormat::isInfix(const Ident* id,int args,int& prec,int& assoc) const {

    const int leftAssoc = 1;
    const int rightAssoc = -1;

    if (id->getMod() != 0) {
	return(FALSE);
	}
    assoc = leftAssoc;
    if (id->getIdent()[1] == 0)	{	// on character operator
	if (args == 2) {		// binary operators;
	    switch(*id->getIdent()) {
		case '=' :
		case '>' :
		case '<' :
		    prec = 50;
		    return(TRUE);
		case '+' :
		case '-' :
		    prec = 60;
		    return(TRUE);
		case '*' :
		case '/' :
		    prec = 70;
		    return(TRUE);
		case '#' :
		case '^' :
		    prec = 80;
		    assoc = rightAssoc;
		    return(TRUE);
		case '@' :
		case '%' :
		case '&' :
		    prec = 90;
		    return(TRUE);
		}
	    return(FALSE);
	    }
	else if (args == 1) {
	    assoc = rightAssoc;
	    switch(*id->getIdent()) {
		case '-' :
		case '!' :
		case '$' :
		    prec = 100;
		    return(TRUE);
		}
	    return(FALSE);
	    }
	else 
	    return(FALSE);
	}

    // multicharacter symbolic tokens.
    if (args == 2) {		// binary operators;
	switch(*id->getIdent()) {
	    case '<' :
		if (strcmp(id->getIdent(), "<<") == 0) {
		    prec = 10;
		    return(TRUE);
		    }
		if (strcmp(id->getIdent(), "<>") == 0) {
		    prec = 50;
		    return(TRUE);
		    }
		if (strcmp(id->getIdent(), "<=") == 0) {
		    prec = 50;
		    return(TRUE);
		    }
		return(FALSE);
	    case '>' :
		if (strcmp(id->getIdent(), ">>") == 0) {
		    prec = 10;
		    return(TRUE);
		    }
		if (strcmp(id->getIdent(), ">=") == 0) {
		    prec = 50;
		    return(TRUE);
		    }
		return(FALSE);
	    case 'O' :
		if (strcmp(id->getIdent(), "OR") == 0) {
		    prec = 20;
		    return(TRUE);
		    }
		return(FALSE);
	    case 'A' :
		if (strcmp(id->getIdent(), "AND") == 0) {
		    prec = 30;
		    return(TRUE);
		    }
		return(FALSE);
	    case '=' :
		if (strcmp(id->getIdent(), "==") == 0) {
		    prec = 50;
		    return(TRUE);
		    }
		return(FALSE);
	    case 'M' :
		if (strcmp(id->getIdent(), "MOD") == 0) {
		    prec = 70;
		    return(TRUE);
		    }
		return(FALSE);
	    }
	return(FALSE);
	}
     else if (args == 1) {
	if (strcmp(id->getIdent(), "NOT") == 0) {
	    prec = 40;
	    assoc = rightAssoc;
	    return(TRUE);
	    }
	return(FALSE);
	}
    return(FALSE);
}

/**************************************************************************/
FormatNode* IDFormat::copy() {

    return(new IDFormat(*this));
}

/**************************************************************************/
void IDFormat::format(Formatter& out) const {

    if (myNode == 0) return;
    switch(myNode->nodeKind()) {
        case Node::nDecl: 
	    formatDecl(myNode->toDecl(), out);
	    break;
        case Node::nType: 
	    formatType(myNode->toType(), out);
	    break;
        case Node::nExp: 
	    formatExp(myNode->toExp(), out);
	    break;
        case Node::nTerm: 
	    formatTerm(myNode->toTerm(), out);
	    break;
        case Node::nStmt: 
	    formatStmt(myNode->toStmt(), out);
	    break;
        case Node::nMethod: 
	    formatMethod(myNode->toMethod(), out);
	    break;
        case Node::nTimeSpec: 
	    formatTimeSpec(myNode->toTimeSpec(), out);
	    break;
        case Node::nIdent: 
	    formatIdent(myNode->toIdent(), out);
	    break;
        case Node::nClause: 
	    formatClause(myNode->toClause(), out);
	    break;
	}
}

/**************************************************************************/
void IDFormat::formatDecl(const Decl* decl, Formatter& out) const {

    if (decl->declKind() == Decl::dCommentDecl) {
	formatCommentDecl(decl->toCommentDecl(), out);
	return;
	}

    switch (decl->getVisibility()) {
	case Decl::publicVis:
	    out << "PUBLIC ";
	    break;
	case Decl::privateVis:
	    out << "PRIVATE ";
	    break;
	case Decl::defaultVis:
        case Decl::cppVis:
	    break;
	}

    switch(decl->declKind()) {
	case Decl::dModDecl:
	    formatModDecl(decl->toModDecl(), out);
	    break;
	case Decl::dEnumDecl:
	    formatEnumDecl(decl->toEnumDecl(), out);
	    break;
	case Decl::dObjectDecl:
	    formatObjectDecl(decl->toObjectDecl(), out);
	    break;
	case Decl::dTypeDecl:
	    out << decl->toTypeDecl()->getName() << " = ";
	    formatType(decl->toTypeDecl()->getTypeRef(), out);
	    out << ";\n";
	    break;
	case Decl::dConstDecl:
	    out << decl->toConstDecl()->getName() << " := ";
	    formatExp(decl->toConstDecl()->getValue(), out);
	    out << ";\n";
	    break;
	case Decl::dVarDecl:
	    out << decl->toVarDecl()->getName() << " : ";
	    formatType(decl->toVarDecl()->getType(), out);
	    out << ";\n";
	    break;
	case Decl::dFunDecl:
	    formatFunDecl(decl->toFunDecl(), out);
	    break;
	case Decl::dFunArgDecl:
	    formatFunArgDecl(decl->toFunArgDecl(), out);
	    break;
	case Decl::dImportDecl:
	    out << "IMPORT ";
	    formatIdent(decl->toImportDecl()->getIdent(), out);
	    out << ";\n";
	    break;
        case Decl::dFlashBackDecl:
            formatFlashBackDecl(decl->toFlashBackDecl(), out);
            break;
        case Decl::dRenameDecl:
            formatRenameDecl(decl->toRenameDecl(), out);
            break;
	case Decl::dEnumValDecl:
	    out << decl->toEnumValDecl()->getName();
            break;
	case Decl::dCommentDecl:
	    // should never happen
	    break;
	}
}

/**************************************************************************/
void IDFormat::formatDeclHead(const Decl* decl, Formatter& out) const {

    if (decl->declKind() == Decl::dCommentDecl) {
	decl = decl->toCommentDecl()->getDecl();
	if (decl == 0) return;
	}

    switch (decl->getVisibility()) {
	case Decl::publicVis:
	    out << "PUBLIC ";
	    break;
	case Decl::privateVis:
	    out << "PRIVATE ";
	    break;
	case Decl::defaultVis:
        case Decl::cppVis:
	    break;
	}

    switch(decl->declKind()) {
	case Decl::dModDecl:
	    out << "MODULE " << decl->getName() << ";\n";
	    break;
	case Decl::dEnumDecl:
	    formatEnumDecl(decl->toEnumDecl(), out);
	    break;
	case Decl::dObjectDecl:
	    out << "OBJECT " << decl->getName() << ";\n";
	    break;
	case Decl::dTypeDecl:
	    out << decl->toTypeDecl()->getName() << " = ";
	    formatType(decl->toTypeDecl()->getTypeRef(), out);
	    out << ";\n";
	    break;
	case Decl::dConstDecl:
	    out << decl->toConstDecl()->getName() << " := ";
	    formatExp(decl->toConstDecl()->getValue(), out);
	    out << ";\n";
	    break;
	case Decl::dVarDecl:
	    out << decl->toVarDecl()->getName() << " : ";
	    formatType(decl->toVarDecl()->getType(), out);
	    out << ";\n";
	    break;
	case Decl::dFunDecl:
	    formatFunDeclHead(decl->toFunDecl(), out);
	    out << ";\n";
	    break;
	case Decl::dFunArgDecl:
	    break;
	case Decl::dImportDecl:
	    out << "IMPORT ";
	    formatIdent(decl->toImportDecl()->getIdent(), out);
	    out << ";\n";
	    break;
        case Decl::dFlashBackDecl:
            formatFlashBackDecl(decl->toFlashBackDecl(), out);
            break;
        case Decl::dRenameDecl:
            formatRenameDecl(decl->toRenameDecl(), out);
            break;
	case Decl::dEnumValDecl:
            break;
	case Decl::dCommentDecl:
	    break;
	}
}

/**************************************************************************/
void IDFormat::formatCommentDecl(const CommentDecl* node,Formatter& out) const {

    switch(node->getKind()) {
	case Comment::replace:
	    out << node->getMsg() << "\n";
	    break;
	case Comment::prefix:
	    out.canBreak(FALSE);
	    out << node->getMsg();
	    out.canBreak(TRUE);
	    formatDecl(node->getDecl(), out);
	    break;
	case Comment::postfix:
	    formatDecl(node->getDecl(), out);
	    out.canBreak(FALSE);
	    out << node->getMsg();
	    out.canBreak(TRUE);
	    break;
        case Comment::passThrough:
	    break;
	}
}

/**************************************************************************/
void IDFormat::formatEnumDecl(const EnumDecl* node, Formatter& out) const {

    out << node->getName() << " = (";
    const EnumType* enumType = node->getTypeRef();

    const PiList* enumPtr = enumType->getEnums();
    while (enumPtr != 0) {
	EnumValDecl* enumVal = (EnumValDecl*) enumPtr->first();
	formatDecl(enumVal, out);
	enumPtr  = enumPtr->rest();
	if (enumPtr != 0)
	    out << ", ";
	}
    out << ");\n";
}

/**************************************************************************/
void IDFormat::formatModDecl(const ModDecl* node, Formatter& out) const {

    out << "MODULE " << node->getName();
    const PiList* ptrInterface = node->getInterface();
    if (ptrInterface  != 0) {
	out << " INTERFACE(" ;
	while (ptrInterface != 0) {
	    PiList* tuple = (PiList*) ptrInterface->first();
	    out << (char*) tuple->first() << " : ";
	    out << "\"" << (char*) tuple->rest()->first() << "\"";
	    ptrInterface  = ptrInterface->rest();
	    if (ptrInterface != 0)
		out << ", ";
	}
	out << ")";
    }
    out << "\n" ;
    out.shiftIn();
    formatDecls(node->getDecls(), out);
    out.shiftOut();
    out << "END MODULE;\n";
}

/**************************************************************************/
void IDFormat::formatObjectDecl(const ObjectDecl* node, Formatter& out) const {

    out << node->getName() << " = OBJECT" ;
    const ObjectType* objType = node->getTypeRef();

    const PiList* ptrSuper = objType->getSuperNames();
    if (ptrSuper  != 0) {
        out << "(" ;
        while (ptrSuper != 0) {
            Ident* super = (Ident*) ptrSuper->first();
	    formatIdent(super, out);
            ptrSuper  = ptrSuper->rest();
            if (ptrSuper != 0)
		out << ", ";
        }
        out << ")";
    }
    out << "\n";

    out.shiftIn();
    formatDecls(objType->getDecls(), out);
    out.shiftOut();
     
    if (objType->getClauses() != 0) {
	out << "DOME BEGIN\n" ;
	out.shiftIn();
	formatClauses(objType->getClauses(), out);
	out.shiftOut();
	out << "END DOME;\n" ;
	}

    out << "END OBJECT;\n" ;
}

/**************************************************************************/
void IDFormat::formatFunDeclHead(const FunDecl* node, Formatter& out) const {

    const FunType* funType = node->getFunType();
    if (funType->getKind() == FunType::ftn)
       out << "FUNCTION ";
    else {
	switch(funType->getKind()) {
	    case FunType::ask:
	       out << "ASK ";
	       break;
	    case FunType::tell:
	       out << "TELL ";
	       break;
	    case FunType::constructor:
	       out << "CONSTRUCTOR ";
	       break;
	    case FunType::destructor:
	       out << "DESTRUCTOR ";
	       break;
	    case FunType::query:
	       out << "QUERY ";
	       break;
	    case FunType::ftn:
	       break;
	    }
	if (funType->getObjType() != 0)
	    formatType(funType->getObjType(), out);
	else if (funType->getObj() != 0)
	    formatType(funType->getObj(), out);
	out << " METHOD ";
	}

    out << node->getName();
    out << "(";
    const PiList* argPtr = funType->getArgs();
    while (argPtr != 0) {
        FunArgDecl* arg = (FunArgDecl*) argPtr->first();
	formatFunArgDecl(arg, out);
        argPtr = argPtr->rest();
        if (argPtr != 0)
            out << ";";
        }

    out << ")";
    if (funType->getRetType() != 0 && 
	!funType->getRetType()->isKind(Type::voidType)) {
	out << " : ";
	formatType(funType->getRetType(), out);
	}
}

/**************************************************************************/
void IDFormat::formatFunDecl(const FunDecl* node, Formatter& out) const {

    formatFunDeclHead(node, out);
    formatMethod(node->getBody(), out);
}

/**************************************************************************/
void IDFormat::formatFunArgDecl(const FunArgDecl* node, Formatter& out) const {

    switch(node->getKind()) {
	case FunArgDecl::in        : out << "IN";              break;
	case FunArgDecl::out       : out << "OUT";             break;
	case FunArgDecl::inout     : out << "INOUT";           break;
	}

    if (*node->getName() != 0) {
	out << " ";
	out << node->getName();
	out << " : ";
	}
    else
	out << ":";
    formatType(node->getType(), out);
}

/**************************************************************************/
void IDFormat::formatFlashBackDecl(const FlashBackDecl*,
        Formatter&) const {

        // FIX
}

/**************************************************************************/
void IDFormat::formatRenameDecl(const RenameDecl* node, Formatter& out) const {

    out << node->getName() << " == ";
    formatIdent(node->getIdent(), out);
    if (node->getType() != 0) {
	out << "::";
	formatType(node->getType(), out);
	}
    out << ";\n";
}

/**************************************************************************/
static Decl::DeclKinds declKind(const Decl* decl) {

    switch(decl->declKind()) {
        case Decl::dFlashBackDecl:
	    return(declKind(decl->toFlashBackDecl()->getDecl()));
	case Decl::dCommentDecl:
	    if (decl->toCommentDecl()->getDecl() != 0)
		return(declKind(decl->toCommentDecl()->getDecl()));
	    return(decl->declKind());
	default:
	    return(decl->declKind());
	}
};

/**************************************************************************/
void IDFormat::formatDecls(const PiList* decls, Formatter& out) const {
    
    out.listBegin();
    while (decls) {
        Decl* decl = (Decl *) decls->first();
	switch(declKind(decl)) {
	    case Decl::dModDecl:
	    case Decl::dObjectDecl:
	    case Decl::dFunDecl:
		out << "\n";
	    }
	out.subObj((IDFormat) decl);
        decls = decls->rest();
        }
    out.listEnd();
}

/**************************************************************************/
void IDFormat::formatType(const Type* type, Formatter& out) const {

    switch(type->typeKind()) {
	case Type::tIdentType:
	    formatIdent(type->toIdentType()->getIdent(), out);
	    break;
	case Type::tArrayType:
	    out << "ARRAY[";
	    if (type->toArrayType()->getBound() != 0)
		formatExp(type->toArrayType()->getBound(), out);
	    out << "] OF ";
	    formatType(type->toArrayType()->getElemType(), out);
	    break;
	case Type::tAtomicType:
	    formatAtomicType(type->toAtomicType(), out);
	    break;
	case Type::tFunType:
	    formatFunType(type->toFunType(), out);
	    break;
	case Type::tObjectType:
	    out << type->toObjectType()->getName();
	    break;
	case Type::tEnumType:
	    out << type->toEnumType()->getName();
	    break;
	case Type::tSynType:
	    out << type->toSynType()->getName();
	    break;
	case Type::tFutureType:
	    out << "FUTURE(";
	    formatType(type->toFutureType()->getSubType(), out);
	    out << ")";
	    break;
	}
}

/**************************************************************************/
void IDFormat::formatFunType(const FunType* type, Formatter& out) const {

    if (type->getKind() == FunType::ftn)
       out << "FUNCTION ";
    else {
	switch(type->getKind()) {
	    case FunType::ask:
	       out << "ASK ";
	       break;
	    case FunType::tell:
	       out << "TELL ";
	       break;
	    case FunType::constructor:
	       out << "CONSTRUCTOR ";
	       break;
	    case FunType::destructor:
	       out << "DESTRUCTOR ";
	       break;
	    case FunType::query:
	       out << "QUERY ";
	       break;
	    case FunType::ftn:
	       break;
	    }
	if (type->getObjType() != 0)
	    formatType(type->getObjType(), out);
	else if (type->getObj() != 0)
	    formatType(type->getObj(), out);
	out << " METHOD ";
	}

    out << "(";
    const PiList* argPtr = type->getArgs();
    while (argPtr != 0) {
        FunArgDecl* arg = (FunArgDecl*) argPtr->first();
	formatFunArgDecl(arg, out);
        argPtr = argPtr->rest();
        if (argPtr != 0)
            out << "; ";
        }

    out << ")";
    if (type->getRetType() != 0 && !type->getRetType()->isKind(Type::voidType)) {
	out << " : ";
	formatType(type->getRetType(), out);
	}
}

/**************************************************************************/
void IDFormat::formatAtomicType(const AtomicType* type, Formatter& out) const {

    switch(type->getKind()) {
        case Type::anyType:           out <<  "ANY";	      break;
        case Type::boolType:          out <<  "BOOLEAN";      break;
        case Type::charType:          out <<  "CHAR";         break;
        case Type::intType:           out <<  "INTEGER";      break;
        case Type::realType:          out <<  "REAL";         break;
        case Type::stringType:        out <<  "STRING";       break;
        case Type::listType:          out <<  "LIST";         break;

        case Type::anyEnumType:       out <<  "ENUM";         break;
        case Type::anyObjType:        out <<  "OBJECT";       break;
        case Type::anyFutureType:     out <<  "FUTURE";       break;
        case Type::taggedType:        out <<  "TAGGED";       break;

        case Type::setType:           out <<  "TYPE";         break;
        case Type::modType:           out <<  "MODULE";       break;
        case Type::assocType:         out <<  "ASSOC";        break;
        case Type::assocInstType:     out <<  "ASSOC_INST";   break;
        case Type::aframeType:        out <<  "AFRAME";       break;
        case Type::objectType:        out <<  "OBJECT_TYPE";  break;
        case Type::enumType:          out <<  "ENUM_TYPE";    break;
        default:                                              break;
	}
}

/**************************************************************************/
void IDFormat::formatExp(const Exp* exp, Formatter& out) const {

    switch(exp->expKind()) {
	case Exp::eAsExp: 
            formatExp(exp->toAsExp()->getBody(), out);
	    out << "::";
            formatType(exp->toAsExp()->getType(), out);
	    break;
	case Exp::eIsAExp: 
            formatExp(exp->toIsAExp()->getBody(), out);
            out << " ISA ";
            formatType(exp->toIsAExp()->getSubType(), out);
	    break;
	case Exp::eCastExp: 
            formatExp(exp->toCastExp()->getBody(), out);
            out << " CAST ";
            formatType(exp->toCastExp()->getType(), out);
	    break;
	case Exp::eSelfExp:
	    out << "SELF";
	    break;
	case Exp::eIdentExp: 
	    formatIdent(exp->toIdentExp()->getIdent(), out);
	    break;
	case Exp::eArrayExp:
            formatExp(exp->toArrayExp()->getArray(), out);
	    out << "[";
            formatExp(exp->toArrayExp()->getIndex(), out);
	    out << "]";
	    break;
	case Exp::eApplyExp:
	    formatApplyExp(exp->toApplyExp(), out);
	    break;
	case Exp::eNullExp:
	    out << "NULL";
	    break;
	case Exp::eIntExp:
	    out << exp->toIntExp()->getValue();
	    break;
	case Exp::eRealExp:
	    out << exp->toRealExp()->getValue();
	    break;
	case Exp::eStringExp:
	    out.canBreak(FALSE);
	    out << "\"";
	    litString(exp->toStringExp()->getValue(), '"', out);
	    out << "\"";
	    out.canBreak(TRUE);
	    break;
	case Exp::eCharExp: {
	    out.canBreak(FALSE);
	    char str[2];
	    str[0] = exp->toCharExp()->getValue();
	    str[1] = 0;
	    out << "'";
	    litString(str, '\'', out);
	    out << "'";
	    out.canBreak(TRUE);
	    } break;
	case Exp::eBoolExp:
	    if (exp->toBoolExp()->getValue())
		out << "TRUE";
	    else
		out << "FALSE";
	    break;
	case Exp::eVarExp:
	    out << "?";
	    asIdent(exp->toVarExp()->getVar(), out);
	    break;
	case Exp::eNewExp:
	    formatNewExp(exp->toNewExp(), out);
	    break;
	}
}

/**************************************************************************/
void IDFormat::formatNewExp(const NewExp* exp, Formatter& out) const {

    out << "NEW";
    if (exp->getNewArgs() != 0) {
	out << "(";
	formatExps(exp->getNewArgs(), out);
	out << ")";
	}
    out << " ";

    formatType(exp->getType(), out);
    if (exp->getArgs() != 0) {
	out << "(";
	formatExps(exp->getArgs(), out);
	out << ")";
	}
}

/**************************************************************************/
void IDFormat::formatExps(const PiList* exps, Formatter& out) const {
    
    out.listBegin();
    while (exps) {
        Exp* exp = (Exp *) exps->first();
	out.subObj((IDFormat) exp);
        exps = exps->rest();
	if (exps != 0)
	    out << ", ";
        }
    out.listEnd();
}

/**************************************************************************/
void IDFormat::formatApplyExp(const ApplyExp* apply, Formatter& out)const {

    int opPrec;
    int assoc;
    const PiList* args = apply->getArgs();
    if (apply->getKind() == FunType::ftn) {
	if (isInfix(apply->getName(), args->len(), opPrec, assoc)) {
	    if (opPrec < precm) 
		out << "(";
	    if (args->len() == 1) {
		out << apply->getName()->getIdent();
		out << " ";
		IDFormat arg((Exp*) args->first(), opPrec);
		out.subObj(arg);
		}
	    else if (args->len() == 2) {
		IDFormat arg1((Exp*) args->first(), opPrec-assoc);
		IDFormat arg2((Exp*) args->rest()->first(), opPrec+assoc);
		out.subObj(arg1);
		out << " ";
		out << apply->getName()->getIdent();
		out << " ";
		out.subObj(arg2);
	        }
	    if (opPrec < precm) 
		out << ")";
	    }
	else {
	    formatIdent(apply->getName(), out);
	    out << "(";
	    formatExps(args, out);
	    out << ")";
	    }
	return;
	}

    switch(apply->getKind()) {
	case FunType::ask: 
	    out << "ASK ";
	    formatExp(apply->getObj(), out);
	    out << " TO ";
	    break;
	case FunType::tell: 
	    out << "TELL ";
	    formatExp(apply->getObj(), out);
	    out << " TO ";
	    break;
	case FunType::query: 
	    out << "QUERY ";
	    formatExp(apply->getObj(), out);
	    out << " ABOUT ";
	    break;
	case FunType::constructor: 	// this is a constructor init
		// do nothing
	    break;
	case FunType::ftn: 
	case FunType::destructor: 
		// should never happen
	    return;
	}
    formatIdent(apply->getName(), out);
    out << "(";
    formatExps(apply->getArgs(), out);
    out << ")";
}

/**************************************************************************/
void IDFormat::formatTerm(const Term* term, Formatter& out) const {

    switch(term->termKind()) {
	case Term::tFunctorTerm: 
	    out << term->toFunctorTerm()->getFunctor();
	    if (term->toFunctorTerm()->getArgs() != 0) {
		out << "(";
		formatTerms(term->toFunctorTerm()->getArgs(), ", ", out);
		out << ")";
		}
	    break;
	case Term::tVarTerm:
	    out << "?" << term->toVarTerm()->getVar();
	    break;
	case Term::tConstTerm: 
	    out << "<| ";
	    formatExp(term->toConstTerm()->getExp(), out);
            out << " |>";
	    break;
	case Term::tImportTerm:
	    out << "<<| ";
	    formatStmts(term->toImportTerm()->getStmts(), out);
	    out  << " |>>";
	    break;
	case Term::tListTerm:
	    out << "[";
	    formatTerms(term->toListTerm()->getFirstTerms(), ", ", out);
	    if (term->toListTerm()->getLast() != 0) {
		out << " | ";
		formatTerm(term->toListTerm()->getLast(), out);
		}
	    out << "]";
	    break;
	}
}

/**************************************************************************/
void IDFormat::formatTerms(const PiList* terms, 
    const char* sep, Formatter& out) const {
    
    out.listBegin();
    while (terms) {
        Term* term = (Term *) terms->first();
	out.subObj((IDFormat) term);
        terms = terms->rest();
	if (terms != 0)
	    out << sep;
        }
    out.listEnd();
}

/**************************************************************************/
void IDFormat::formatStmt(const Stmt* stmt, Formatter& out) const {

    switch(stmt->stmtKind()) {
	case Stmt::sAssignStmt:
	    formatAssignStmt(stmt->toAssignStmt(), out);
	    break;
	case Stmt::sIfStmt:
	    formatIfStmt(stmt->toIfStmt(), out);
	    break;
	case Stmt::sWhileStmt:
	    formatWhileStmt(stmt->toWhileStmt(), out);
	    break;
	case Stmt::sRepeatStmt:
	    formatRepeatStmt(stmt->toRepeatStmt(), out);
	    break;
	case Stmt::sForStmt:
	    formatForStmt(stmt->toForStmt(), out);
	    break;
	case Stmt::sLoopStmt:
	    out << "LOOP\n";
	    out.shiftIn();
	    formatStmts(stmt->toLoopStmt()->getBody(), out);
	    out.shiftOut();
	    out << "END LOOP;\n";
	    break;
	case Stmt::sReturnStmt: {
	    const Exp* val = stmt->toReturnStmt()->getVal();
	    out << "RETURN";
	    if (val != 0) {
	 	out << "(";
		out.subObj((IDFormat) val);
	 	out << ")";
		}
	    out << ";\n";
	    } break;
	case Stmt::sExitStmt:
	    out << "EXIT;\n";
	    break;
	case Stmt::sExpStmt:
	    formatExp(stmt->toExpStmt()->getExp(), out);
	    out << ";\n";
	    break;
	case Stmt::sDisposeStmt:
	    out << "DISPOSE ";
	    formatExp(stmt->toDisposeStmt()->getBody(), out);
	    out << ";\n";
	    break;
	case Stmt::sBlockStmt:
	    formatBlockStmt(stmt->toBlockStmt(), out);
	    break;
	case Stmt::sCaseStmt:
	    formatCaseStmt(stmt->toCaseStmt(), out);
	    break;
        case Stmt::sAsyncStmt:
	    break;
	case Stmt::sCommentStmt:
	    formatCommentStmt(stmt->toCommentStmt(), out);
	    break;
	}
}

/**************************************************************************/
void IDFormat::formatStmts(const PiList* stmts, Formatter& out) const {
    
    out.listBegin();
    while (stmts) {
        Stmt* stmt = (Stmt *) stmts->first();
	out.subObj((IDFormat) stmt);
        stmts = stmts->rest();
        }
    out.listEnd();
}

/**************************************************************************/
void IDFormat::formatCommentStmt(const CommentStmt* node,Formatter& out) const {

    switch(node->getKind()) {
	case Comment::replace:
	    out << node->getMsg() << "\n";
	    break;
	case Comment::prefix:
	    out.canBreak(FALSE);
	    out << node->getMsg();
	    out.canBreak(TRUE);
	    formatStmt(node->getStmt(), out);
	    break;
	case Comment::postfix:
	    formatStmt(node->getStmt(), out);
	    out.canBreak(FALSE);
	    out << node->getMsg();
	    out.canBreak(TRUE);
	    break;
        case Comment::passThrough:
	    break;
	}
}

/**************************************************************************/
void IDFormat::formatAssignStmt(const AssignStmt* node, Formatter& out) const {

    formatExp(node->getLhs(), out);
    out << " := ";
    formatExp(node->getRhs(), out);
    out << ";\n";
}

/**************************************************************************/
void IDFormat::formatIfStmt(const IfStmt* node, Formatter& out) const {

    out << "IF ";
    formatExp(node->getCond(), out);
    out << " THEN\n";
    out.shiftIn();
    formatStmts(node->getThen(), out);
    out.shiftOut();
    if (node->getElse() != 0) {
	if (node->getElse()->rest() == 0 &&
	   ((Stmt*) node->getElse()->first())->stmtKind() == Stmt::sIfStmt) {
	    out << "ELSE";	// merges with IF KeyWord!
	    formatStmt((Stmt*) node->getElse()->first(), out);
	    return;
	    };
	out << "ELSE\n";
	out.shiftIn();
	formatStmts(node->getElse(), out);
	out.shiftOut();
	}
    out << "END IF;\n";
}

/**************************************************************************/
void IDFormat::formatWhileStmt(const WhileStmt* node, Formatter& out) const {

    out << "WHILE ";
    formatExp(node->getCond(), out);
    out << " DO\n";
    out.shiftIn();
    formatStmts(node->getBody(), out);
    out.shiftOut();
    out << "END WHILE;\n";
}

/**************************************************************************/
void IDFormat::formatRepeatStmt(const RepeatStmt* node, Formatter& out) const {

    out << "REPEAT\n";
    out.shiftIn();
    formatStmts(node->getBody(), out);
    out.shiftOut();
    out << "UNTIL ";
    formatExp(node->getCond(), out);
    out << "END REPEAT;\n";
}

/**************************************************************************/
void IDFormat::formatForStmt(const ForStmt* node, Formatter& out) const {

    out << "FOR ";
    formatExp(node->getIndex(), out);
    out << " := ";
    formatExp(node->getFrom(), out);
    if (node->getDir() > 0)
	out << " TO ";
    else 
	out << " DOWNTO ";
    formatExp(node->getTo(), out);

    if (node->getBy() != 0) {
	out << " BY ";
	formatExp(node->getBy(), out);
	}
    out << " DO\n";

    out.shiftIn();
    formatStmts(node->getBody(), out);
    out.shiftOut();
    out << "END FOR;\n";
}

/**************************************************************************/
void IDFormat::formatBlockStmt(const BlockStmt* node, Formatter& out) const {

    out << "BLOCK\n";
    out.shiftIn();
    formatDecls(node->getDecls(), out);
    out.shiftOut();
    out << "BEGIN\n";
    out.shiftIn();
    formatStmts(node->getBody(), out);
    out.shiftOut();
    out << "END BLOCK;\n";
}

/**************************************************************************/
void IDFormat::formatCaseStmt(const CaseStmt* node, Formatter& out) const {

    out << "CASE\n";
    formatExp(node->getExp(), out);
    out.shiftIn();

    const PiList* whenPtr = node->getWhens();
    while (whenPtr != 0) {
	When* when = (When*) whenPtr->first();
	out << "WHEN ";
	formatExps(when->whenLabels, out);
	out << ":\n";
	out.shiftIn();
	formatStmts(when->stmts, out);
	out.shiftOut();
	whenPtr  = whenPtr->rest();
	}

    if (node->getOtherwise() != 0)
    out << "OTHERWISE\n";
    out.shiftIn();
    formatStmts(node->getOtherwise(), out);
    out.shiftOut();

    out.shiftOut();
    out << "END CASE;\n";
}

/**************************************************************************/
void IDFormat::formatMethod(const Method* method, Formatter& out) const {

    if (method == 0) {
	out << ";\n";
        return;
	}
    switch(method->methodKind()) {
	case Method::mDeferedBody:
	    out << " SUBCLASS;\n";
	    return;
	case Method::mMethodBody: 
	    break;
	}
    const MethodBody* body = method->toMethodBody();
    out << "\n";

    if (body->getConstrInits() != 0) {
	out << "INIT ";
	out.shiftIn();

	const PiList* initPtr = body->getConstrInits();
	while (initPtr != 0) {
	    NewExp* init = (NewExp*) initPtr->first();
	    formatType(init->getType(), out);
	    out << "(";
	    formatExps(init->getArgs(), out);
	    out << ")";
	    initPtr = initPtr->rest();
	    if (initPtr != 0)
		out << ";";
	    }
	out << ";\n";
	out.shiftOut();
	}

    out.shiftIn();
    formatDecls(body->getDecls(), out);
    out.shiftOut();

    out << "BEGIN\n";
    out.shiftIn();
    formatStmts(body->getStmts(), out);
    out.shiftOut();
    out << "END METHOD;\n";
}

/**************************************************************************/
void IDFormat::formatTimeSpec(const TimeSpec*, 
	Formatter&) const {

	// FIX 
}

/**************************************************************************/
void IDFormat::formatIdent(const Ident* ident, Formatter& out) const {

    if (ident->getMod() != 0) {
	formatIdent(ident->getMod(), out);
	out << ".";
	}
    asIdent(ident->getIdent(), out);
}

/**************************************************************************/
void IDFormat::formatClause(const Clause* clause, Formatter& out) const {

	// FIX FIX FIX
    switch (clause->clauseKind()) {
        case Clause::cCommentClause:
	    formatCommentClause(clause->toCommentClause(), out);
            break;
        case Clause::cPropClause: {
	    const PropClause* propClause = clause->toPropClause();
	    if (propClause->getTerms() != 0) {
		out << "IF ";
		formatTerms(propClause->getTerms(), " AND ", out);
		out << "THEN  ";
		}

	    formatTerm(propClause->getGoal(), out);
            } break;
        }
}

/**************************************************************************/
void IDFormat::formatCommentClause(const CommentClause* node,
    Formatter& out) const {

    switch(node->getKind()) {
	case Comment::replace:
	    out << node->getMsg() << "\n";
	    break;
	case Comment::prefix:
	    out.canBreak(FALSE);
	    out << node->getMsg();
	    out.canBreak(TRUE);
	    formatClause(node->getClause(), out);
	    break;
	case Comment::postfix:
	    formatClause(node->getClause(), out);
	    out.canBreak(FALSE);
	    out << node->getMsg();
	    out.canBreak(TRUE);
	    break;
        case Comment::passThrough:
	    break;
	}
}

/**************************************************************************/
void IDFormat::formatClauses(const PiList* clauses, Formatter& out) const {
    
    out.listBegin();
    while (clauses) {
        Clause* clause = (Clause *) clauses->first();
	out.subObj((IDFormat) clause);
        clauses = clauses->rest();
        }
    out.listEnd();
}


/*****************************************************************************/
#include "IndentFormat.H"

/*****************************************************************************/
void print(const Node* node, ostream& out) {
 
    IndentFormatter fout(out);
    IDFormat format(node);
    format.format(fout);
}

/*****************************************************************************/
void printDecls(const PiList* decls, ostream& out) {
 
    while(decls != 0) {
	const Decl* decl = (Decl*) decls->first();
	print(decl, out);
	decls = decls->rest();
	out << "\n";
	}
}

/*****************************************************************************/
void printExps(const PiList* exps, ostream& out) {
 
    while(exps != 0) {
	const Exp* exp = (Exp*) exps->first();
	print(exp, out);
	exps = exps->rest();
	out << "\n";
	}
}

/*****************************************************************************/
void printStmts(const PiList* stmts, ostream& out) {
 
    while(stmts != 0) {
	const Stmt* stmt = (Stmt*) stmts->first();
	print(stmt, out);
	stmts = stmts->rest();
	out << "\n";
	}
}


