/****************************************************************************/ 
/*			 Lexer.c	 Lexer.c			    */ 
/****************************************************************************/ 
/*  Copyright (C) 1992,	 Vance Morrison	  */ 
/*  All rights reserved, see Copyright file for more information */ 

/* describes a simple lexer for C sytle programming languages.	It has
   been my experience that this simple lexer will handle a very broad
   family of languages.	 Basically it has built into it the standard
   definition of identifiers, numbers (strings of digits) and quoted
   expressions (single and double).  All single symbolic characters are
   also tokens.	 In addition the user can declare that certain identifiers
   and certain strings of symbolic characters (like -> := etc) are also
   tokens */
 
/* Author Vance Morrison 
   Date	  10/22/93		 */ 
/****************************************************************************/ 
 
#include "SpIostream.H"	    // for debugging 
#include "SpCtype.H" 
#include "SpString.H"
#include "CharStream.H" 
#include "CharArrayStream.H" 
#include "Lex.H" 

/****************************************************************************/ 
class KeyWord {
public:
    ~KeyWord();

    friend  void addVal(KeyWord*& table, const char* id, int val);
    int findVal(const char* id, int n);
    const char* findKey(int val);
    friend class Lexer;

private:
    KeyWord(const char* aName, int aValue);
    KeyWord* higher;	/* binary tree */
    KeyWord* lower;	/* binary tree */
    char* name;		/* the name of the keyword */
    int value;		/* its value */
};

/****************************************************************************/ 
KeyWord::KeyWord(const char* aName, int aValue) 
    :value(aValue),higher(0),lower(0) {
    name = new char[strlen(aName)+1];
    strcpy(name, aName);
}

/****************************************************************************/ 
KeyWord::~KeyWord() { 
    delete [] name; 
    if (higher != 0) 
	delete higher; 
    if (lower != 0) 
	delete lower; 
}

/****************************************************************************/ 
void addVal(KeyWord*& table, const char* id, int val) {

    register KeyWord** tree = &table;
    while (*tree != 0) {
	int cmp = strcmp(id, (*tree)->name);
	if (cmp < 0)
	    tree = &(*tree)->lower;
	else if (cmp > 0)
	    tree = &(*tree)->higher;
	else {
	    (*tree)->value = val;
	    return;
	    }
	}
    *tree = new KeyWord(id, val);
}

/****************************************************************************/ 
/* find the first keyword that matches 'id' up to length 'n' and returns
   its value.  Returns 0 (invalid as a keyword) if it is not found */

int KeyWord::findVal(const char* id, int n) {

    register KeyWord* tree = this;
    while (tree != 0) {
	// case insensitive keywords int cmp = strncasecmp(id, tree->name, n);
	int cmp = strncmp(id, tree->name, n);
	if (cmp < 0)
	    tree = tree->lower;
	else if (cmp > 0)
	    tree = tree->higher;
	else {
	    return(tree->value);
	    }
	}
    return(0);
}

/****************************************************************************/ 
/* looks for the token value 'val' and returns the string associated with 
   it.	returns 0 if it is not found */

const char* KeyWord::findKey(int tok) {

    if (this == 0)
	return(0);
    if (value == tok)
	return(name);
    const char* ret = lower->findKey(tok);
    if (ret != 0)
	return(ret);
    return(higher->findKey(tok));
}

/****************************************************************************/ 
Lexer::Lexer(InStream& aStream) : in(&aStream), keyWords(0), symbols(0) { 

    maxComment = 2000000000;	/* make the maximum comment size infinity */
    if (in->isEnd()) 
	look = 0;
    else 
	look = in->cur();
} 

/****************************************************************************/ 
Lexer::~Lexer() {

   if (keyWords != 0)
       delete keyWords;
   if (symbols != 0)
       delete symbols;
}
 
/****************************************************************************/ 
void Lexer::addKeyWord(const char* key, int value) {
   addVal(keyWords, key, value);
}

/****************************************************************************/ 
void Lexer::addSymbol(const char* key, int value) {
   addVal(symbols, key, value);
}

/****************************************************************************/ 
void Lexer::next() { 
    if (in->isEnd()) {
	look = 0;
	return;
	}
    cursor.pos++; 
    if (look == '\n') { 
	cursor.row++; 
	cursor.col = 1; 
	} 
    else 
	cursor.col++; 
    in->next();
    if (in->isEnd()) {
	look = 0;
	return;
	}
    look = in->cur();
} 

/****************************************************************************/ 
int Lexer::getIdent(OutStream& out) {

    char keyWordBuff[256];
    char* ptr = keyWordBuff;

    while (isalnum(look) || look == '_') { 
	if (ptr >= &keyWordBuff[255])
	   break;
	*ptr++ = look;
	next(); 
	} 
    *ptr = 0;

    if (strcmp(keyWordBuff, "OP") == 0 && look == '(') {  // arbitary ID name
	if (getQuote(out, ')') == 0)
	    return(ID);
	else
	    return(BADSYMBOL);
	}

    ptr = keyWordBuff;		// copy the buffer out
    while (*ptr != 0) {
	out.put(*ptr);
	ptr++;
	} 
	
    int ret = keyWords->findVal(keyWordBuff, 257);
    if (ret <= 0)
	ret = ID; 
    return(ret);
}
 
/****************************************************************************/ 
int Lexer::getNum(OutStream& out) {

    if (look == '0') {
	out.put(look); 
	next(); 
	if (look == 'x' || look == 'X') {
	    out.put(look); 
	    next(); 
	    while (isxdigit(look)) { 
		out.put(look); 
		next(); 
		} 
	    return(NUM); 
	    } 
	}
    while (isalnum(look)) { 
	out.put(look); 
	next(); 
	} 
    return(NUM); 
}

/****************************************************************************/ 
/* gets a quoted thing.  'quote' is the ending character.  Returns 0 on
   success */
int Lexer::getQuote(OutStream& out, char quote) {

    next(); 	// skip begining quote
    Bool escape = FALSE; 
    while (escape || look != quote) { 
	if(look <= 0) 
	    return(-1); 
	if (escape) {
	    escape = FALSE; 
	    if (look == '\n') {	    // skip line until next non-space char
		while(isspace(look))
		    next(); 
		continue;	// I hate using these, goes to top of while
		}
	    if (look == 'n')
		out.put('\n');
	    else if (look == 't')
		out.put('\t');
	    else if (look == 'b')
		out.put('\b');
	    else if (look == 'r')
		out.put('\r');
	    else if (look == 'f')
		out.put('\f');
	    else if (isdigit(look)) {
		char num = (look - '0');
		next(); if (!isdigit(look)) return(-1);
		num = (look - '0') + num * 8;
		next(); if (!isdigit(look)) return(-1);
		num = (look - '0') + num * 8;
		out.put(num);
		}
	    else
		out.put(look); 	// handles \" and \\ escapes
	    }
	else if (look == '\\') 
	    escape = TRUE; 
	else if (look == '\n') 
	    return(-1); 
	else
	    out.put(look);
	next(); 
	} 
    next(); 	// skip ending quote
    return(0);	// return success 
}

/****************************************************************************/ 
int Lexer::getSymbol(OutStream& out) {

    char symBuff[256];
    int	 idx = 0;

    if (look == '\\') {	    // could be a pre comment 
	symBuff[idx++] = look;
	out.put(look);
	next();
	if (look ==  '/') {			// line style pre comments 
	    while(look != '\n') {
		if (look == 0)
		    return(BADCOMMENT);
		if (idx < maxComment)
		    out.put(look);
		next();
		idx++;
		}
	    out.put(look);
	    next();
	    return(PRECOMMENT);
	    } 
	}
    if (look == '<') {	    // could be a post comment 
	symBuff[idx++] = look;
	out.put(look);
	next();
	if (look ==  '-') {
	    symBuff[idx++] = look;
	    out.put(look);
	    next();
	    idx++;
	    if (look ==  '-') {			// line style comments 
		while(look != '\n') {
		    if (look == 0)
			return(BADCOMMENT);
		    if (idx < maxComment)
			out.put(look);
		    next();
		    idx++;
		    }
		out.put(look);
		next();
		return(POSTCOMMENT);
		} 
	    }
	}
    else if (look == '{' /* } */ ) {
	out.put(look);
	next(); 
	idx++;
	int nesting = 0;
	for(;;) { 
	    if (look == 0) 
		return(BADCOMMENT);  
	    if (look == '{')
		nesting++;
	    else if (look == '}') {
		--nesting; 
		if (nesting < 0)
		    break;
		}
	    if (idx < maxComment)
		out.put(look);
	    next(); 
	    idx++;
	    } 
	if (idx < maxComment)
	    out.put(look);
	next(); 
	return(COMMENT);
	} 

    while (ispunct(look)) {
	symBuff[idx] = look;
	if (idx > 0 && !symbols->findVal(symBuff, idx+1))
	    break;
	idx++;
	if (idx >= 255)
	    return(BADSYMBOL);
	out.put(look);
	next();
	}

    if (idx == 1)
	return(*symBuff);
    symBuff[idx] = 0;
    int ret = symbols->findVal(symBuff, idx+1);
    if (ret <= 0)
	ret = BADSYMBOL;
    return(ret);
}

/****************************************************************************/ 
int Lexer::get(OutStream& out, TermExtent* extent) {

    while(isspace(look))    // skip space 
	next(); 
     if (extent != 0)
	extent->begin = cursor;

    int ret;
    if (look == 0) 
	ret = END; 
    else if (isalpha(look) || look == '_')
	ret = getIdent(out);
    else if (isdigit(look)) 
	ret = getNum(out);
    else if (look == '"') 
	ret = (getQuote(out, '"') == 0) ? QUOTE : BADQUOTE;
    else if (look == '\'')
	ret = (getQuote(out, '\'') == 0) ? SQUOTE : BADQUOTE;
    else if (ispunct(look)) 
	ret = getSymbol(out);
    else
	ret = BADSYMBOL;

     if (extent != 0)
	extent->end = cursor;
    return(ret);
} 

/****************************************************************************/ 

char* Lexer::printTok(int tok, char* buff, int len) {

    StrOutStream out(buff, len);
    printTok(tok, out);
    return(out.val());
}

/****************************************************************************/ 
/* prints a human-readable version of 'tok' on 'out */
void Lexer::printTok(int tok, OutStream& out) {

    switch(tok) { 
	case END: 
	    out.write("END", 3);
	    return;
	case ID: 
	    out.write("ID", 2);
	    return;
	case NUM: 
	    out.write("NUM", 3);
	    return;
	case QUOTE: 
	    out.write("QUOTE", 5);
	    return;
	case SQUOTE: 
	    out.write("SQUOTE", 6);
	    return;
	case OVERFLOW: 
	    out.write("OVERFLOW", 8);
	    return;
	case BADQUOTE: 
	    out.write("BADQUOTE", 8);
	    return;
	case COMMENT: 
	    out.write("COMMENT", 7); 
	    return;
	case POSTCOMMENT: 
	    out.write("POSTCOMMENT", 8); 
	    return;
	case PRECOMMENT: 
	    out.write("PRECOMMENT", 9); 
	    return;
	case BADCOMMENT: 
	    out.write("BADCOMMENT", 10); 
	    return;
	case BADSYMBOL: 
	    out.write("BADSYMBOL", 9); 
	    return;
	default:  {
	    if (tok < 128 && ispunct(tok)) {
		out.put('\''); 
		char c = tok;
		out.put(c); 
		out.put('\''); 
		return;
		}
	    const char* str = keyWords->findKey(tok);
	    if (str == 0)
		str = symbols->findKey(tok);
	    if (str != 0) 
		out.write(str, strlen(str));
	    else
		out.write("<UNKNOWN TOKEN>", 15);
	    }
	} 
}

/**********************************************************************/ 
#ifdef TEST 
#include "IOStream.H"

main() { 
    IOInStream in(cin);
    char buff[80];
    StrOutStream tokStr(buff, 80);

    Lexer lexer(in);
    lexer.addKeyWord("hello", 512);
    lexer.addKeyWord("there", 513);
    lexer.addKeyWord("in", 514);
    lexer.addKeyWord("while", 515);

    lexer.addSymbol("->", 515);
    lexer.addSymbol("--->", 516);
    lexer.addSymbol("+++", 517);
    lexer.addSymbol("::", 518);
    lexer.addSymbol(">>", 519);
    lexer.addSymbol("<>", 520);

    for(;;) { 
	tokStr.reset();
	int tok = lexer.get(tokStr); 
	cout << "TOKEN '" << tokStr.val() << "' is a ";

	tokStr.reset();
	lexer.printTok(tok, tokStr);
	cout << tokStr.val() << "\n"; 
 
	if (tok == Lexer::END) 
	    break; 
	} 
} 

#endif 
 
