/**************************************************************************/ 
/*			   Lexer.h    Lexer.h				  */ 
/**************************************************************************/ 
/*  Copyright (C) 1992,	 Vance Morrison	  */ 
/*  All rights reserved, see Copyright file for more information */ 
 
/* describes a simple lexer for C sytle programming languages.	It has
   been my experience that this simple lexer will handle a very broad
   family of languages.	 Basically it has built into it the standard
   definition of identifiers, numbers (strings of digits) and quoted
   expressions (single and double).  All single symbolic characters are
   also tokens.	 In addition the user can declare that certain identifiers
   and certain strings of symbolic characters (like -> := etc) are also
   tokens */
 
/* AUTHOR: Vance Morrison  
   DATE	 : 5/23/91		    */ 
/**************************************************************************/ 
 
#ifndef _Lex_H 
#define _Lex_H

class KeyWord;

/**************************************************************************/ 
struct TermPos {
    TermPos() : pos(0), row(1), col(1) {}

    int pos;		// character position in input stream, starts at 0
    int row;		// rows start at 1
    int col;		// columns start at 1
};

/**************************************************************************/ 
/* an TermExtent is simply logs the begining and end of a term */
struct TermExtent {
    TermPos begin;	// first position of the Term
    TermPos end;	// postition immediately after the term
};

class InStream;
class OutStream;

/**************************************************************************/ 
class Lexer { 
public: 
    /* make a Lexer on the input stream.  Note that Lexer does
       not become owner of 'in' (that is Lexer does not delete it) */
    Lexer(InStream& in);
    ~Lexer();

    /* AlphaNumeric keywords tokens need to be declared to the lexer */
    /* note that key is copied, so it need to persist */
    /* 'value' must be >= 256 */
    void addKeyWord(const char* key, int value);

    /* MultCharacter symbolic tokens also need to be declared */
    /* note that key is copied, so it need to persist */
    /* 'value' must be >= 256 */
    void addSymbol(const char* symbol, int value);

    /* fills in 'pos' with the  current position in the input stream */
    void curPos(TermPos* pos)	{ *pos = cursor; };

    /* gets the next token and returns its value.  Also writes the
       characters associated with the token on the stream 'out. 
       if 'exent' is given, it fills updates it with the currect
       extent of the returned token */
    int get(OutStream& out, TermExtent* extent=0);

    /* prints a human-readable version of 'tok' on 'out */
    void printTok(int tok, OutStream& out);

    /* for those who just want a quick printout */
    char* printTok(int tok, char* buff, int len);

    /* we throw away chacters after this */
    void setMaxComment(int size)	    { maxComment = (size>3)?size:3; };

    /* Token values < 0 are reserved by the lexer and should not
       be used as a keyword or symbol toke (using).  In addition
       token values 0 < 256 are reserved for single character tokens.
       Thus user defined tokens must have values over 255.  This
       fits with YACC very well */
    enum Tokens { 
	END = 0, 
	    // 1-255 are reserved for single character tokens 
	ID = -1,
	NUM = -2,
	QUOTE = -3, 
	SQUOTE = -4,
	COMMENT = -5,
	POSTCOMMENT = -6,
	PRECOMMENT = -7,
	 
	OVERFLOW = -100,		// error conditions  
	BADQUOTE = -101, 
	BADSYMBOL = -102,
	BADCOMMENT = -103 
    }; 

private:
    int getIdent(OutStream& out);
    int getNum(OutStream& out);
    int getQuote(OutStream& out, char quote);
    int getSymbol(OutStream& out);

private:
    InStream* in; 
    char look; 
    TermPos cursor;
    int maxComment;	    /* throw away comment text longer then this */
 
    void next(); 
    KeyWord* keyWords;
    KeyWord* symbols;	    /* multi-character symbolic keywords */
}; 
 
#endif 
 
