/*****************************************************************************/
/*                             TypeCheck.cxx                                 */
/*****************************************************************************/

/* TypeCheck does all the non-context free checking of the Import language.
   As a side effect it also fills in some information in the parse tree */

/* AUTHOR: Vance Morrison
   DATE:  10/20/93		*/
/*****************************************************************************/

#include "SpIostream.H"		// FIX for debugging only
#include "IDAST.H"
#include "TypeEnv.H"
#include "TypeStatus.H"
#include "NewWhere.H"

// JB: for collecting process model label info:
int ApplyExp::CumReentryLabel = 0;

// JB: for collecting DEFINE_EVENT() macro info:
PiList* AsyncStmt::globalAsyncStmtPiList = 0;


static const AtomicType aAnyType(Type::anyType);
static const AtomicType aIntType(Type::intType);
static const AtomicType aRealType(Type::realType);
static const AtomicType aStringType(Type::stringType);
static const AtomicType aSetType(Type::setType);
static const AtomicType aModType(Type::modType);
static const AtomicType aUnknownObjType(Type::unknownObjType);
static const AtomicType aCharType(Type::charType);
static const AtomicType aBoolType(Type::boolType);
static const AtomicType aPiListType(Type::listType);
static const AtomicType aTaggedType(Type::taggedType);
static const AtomicType aAnyObjType(Type::anyObjType);
static const AtomicType aAnyFutureType(Type::anyFutureType);
static const AtomicType aAssocType(Type::assocType);
static const AtomicType aObjectType(Type::objectType);

static void addObjDeclsToScope(PiList* decls,TypeScope* scope,TypeStatus* status);
static void addDeclsToScope(const PiList* decls, TypeScope* scope,
	TypeStatus* status);
static void typeDeclBodies(PiList* decls, const TypeEnv* env, TypeStatus* status);
static void typeAndAddDecls(PiList* decls, const TypeEnv* env, TypeScope* scope, 
	TypeStatus* status, const Decl* sup=0);
static void typeStmts(PiList* stmts, const TypeEnv* env, TypeStatus* status,
	PiList** varExps);
static void typeTerms(PiList* stmts, const TypeEnv* env, TypeStatus* status);
static Decl* definedBy(const char* name, const Type* type, PiList* decls);
static FunType* funTypeFromArgs(FunKind kind, const Type* obj, const PiList* args,
    const TypeEnv* env, TypeStatus* status, PiList** varExps);
static void delFunTypeFromArgs(FunType* fun);
static PiList* genAssocDecls(const PiList* args, const TypeEnv* env,
	TypeStatus* status, const Decl* sup);

/*****************************************************************************/
void Decl::addDeclToScope(TypeScope* scope, TypeStatus* status) {

    if(scope->addDecl(this) != 0) {
	status->setIdMultDefined(this, getName(), getType());
	return;
	}
}

/*****************************************************************************/
void ImportDecl::addDeclToScope(TypeScope* aScope, TypeStatus*) {

    aScope->merge(modDecl->getScope());
}

/*****************************************************************************/
void ObjectDecl::addDeclToScope(TypeScope* scope, TypeStatus*) {

	// The declaration for the object itself was done in addObjDeclToScope
	// merge in the public scope of the object.  
    scope->merge(getScope());
}

/*****************************************************************************/
void EnumDecl::addDeclToScope(TypeScope* scope, TypeStatus* status) {

    const PiList* enumPtr = typeRef->getEnums();
    while (enumPtr != 0) {
        Decl* enum_ = (Decl*) enumPtr->first();
	if(scope->addDecl(enum_) != 0) {
	    status->setIdMultDefined((Decl*) this, enum_->getName(), typeRef);
	    return;
	    }
        enumPtr = enumPtr->rest();
        }

    Decl::addDeclToScope(scope, status);
}

/*****************************************************************************/
void RenameDecl::typeDecl(const TypeEnv* env, TypeStatus* status, 
    const Decl* sup){

    envDecl = sup;
    if (type != 0) {
	type->typeType(env, status);
	if (!status->ok()) return;
	}
    origDecl = ident->typeIdent(env, status,
				(type!=0) ? type : (Type *) &aAnyType);
    if (!status->ok()) return;
    origDecl = origDecl->normDecl();
}

/*****************************************************************************/
void FlashBackDecl::typeDecl(const TypeEnv* env, TypeStatus* status, 
    const Decl* sup){

	// FIX 
    envDecl = sup;
    decl->typeDecl(env, status, sup);
}

/*****************************************************************************/
void FlashBackDecl::typeDeclBody(const TypeEnv* env, TypeStatus* status) {

	// FIX 
    decl->typeDeclBody(env, status);
}

/*****************************************************************************/
void FlashBackDecl::addObjDeclToScope(TypeScope* scope, TypeStatus* status)  {

	// FIX 
    decl->addObjDeclToScope(scope, status);
}

/*****************************************************************************/
const Decl* Ident::typeIdent(const TypeEnv* env, TypeStatus* status, 
	const Type* context) {

    const Decl* ret = 0;
    if (mod != 0)
    {
       const Decl* modDecl = mod->typeIdent(env, status, &aModType);
       if (!status->ok()) return(0);
       modDecl = modDecl->normDecl();
       if (modDecl->toObjectDecl() != 0)
       {
	  const ObjectType* obj = modDecl->toObjectDecl()->getTypeRef();
	  ret = obj->getScope()->findIdent(ident, context, 
					   obj == env->getSelfType());
       }
       else ret = modDecl->getScope()->findIdent(ident, context);
    }
    else ret = env->findIdent(ident, context);

    if (ret == 0)
    {
       status->setIdentNotFound(this, this, context, env);
       return 0;

       /* JB idea (maybe):
       PiList *empty = new PiList(NULL, NULL);
       ObjectType *cppType = new ObjectType(ident, empty, empty, empty);
       ObjectDecl *cppObj  = new ObjectDecl(ident, cppType, Decl::cppVis);
       cout<< "Object '" << ident << "' added to current scope" << endl;
       return(cppObj);
       */
    }
    else if (ret == TypeScope::ambiguousDecl()) {
	status->setIdentAmbiguous(this, this, context, env);
	return(0);
	}
    else if (ret == TypeScope::privateDecl()) {
	status->setIdentPrivate(this, this, context, env);
	return(0);
	}
    return(ret);
}

/*****************************************************************************/
void ModDecl::typeDecl(const TypeEnv* env, TypeStatus* status, const Decl* sup) {

    envDecl = sup;
    if (scope != 0)
	scope->unlink();
    scope = new(this, "TypeScope") TypeScope;

    addObjDeclsToScope(decls, scope, status);
    if (!status->ok()) return;

    typeAndAddDecls(decls, env, scope, status, this);
    if (!status->ok()) return;
    
        // we disallow variable declarations (global vars) in modules
    const PiList* declPtr = decls;
    while (declPtr != 0) {
        const Decl* decl = (Decl*) declPtr->first();

        const VarDecl* varDecl = decl->normDecl()->toVarDecl();
	if (varDecl != 0 && decl->getVisibility() == Decl::publicVis)
	{
	   status->setVarIllegal(varDecl);
	   return;
	}
        declPtr = declPtr->rest();
        }
}

/*****************************************************************************/
void ModDecl::typeDeclBody(const TypeEnv* env, TypeStatus* status) {

    TypeEnv myTypeEnv(scope, env);
    typeDeclBodies(decls, &myTypeEnv, status);
}

/*****************************************************************************/
void AFrameDecl::typeDecl(const TypeEnv* env, TypeStatus* status, const Decl* sup) {

    envDecl = sup;
    if (scope != 0)
	scope->unlink();
    scope = new(this, "TypeScope") TypeScope;

    addObjDeclsToScope(decls, scope, status);
    if (!status->ok()) return;

    typeAndAddDecls(decls, env, scope, status, this);
    if (!status->ok()) return;
    
        // we disallow variable declarations (global vars) in modules
    const PiList* declPtr = decls;
    while (declPtr != 0) {
        const Decl* decl = (Decl*) declPtr->first();

        const VarDecl* varDecl = decl->normDecl()->toVarDecl();
           if (varDecl != 0 && decl->getVisibility() == Decl::publicVis) {
                status->setVarIllegal(varDecl);
		return;
		}
        declPtr = declPtr->rest();
        }
}

/*****************************************************************************/
void AFrameDecl::typeDeclBody(const TypeEnv* env, TypeStatus* status) {

    TypeEnv myTypeEnv(scope, env);
    typeDeclBodies(decls, &myTypeEnv, status);
}

/*****************************************************************************/
void ImportDecl::typeDecl(const TypeEnv* env, TypeStatus* status, 
    const Decl* sup) {

    envDecl = sup;
    modDecl = ident->typeIdent(env, status, &aModType);
    if (!status->ok()) return;
    modDecl = modDecl->normDecl();
}

/*****************************************************************************/
void ConstDecl::typeDecl(const TypeEnv* env, TypeStatus* status, 
    const Decl* sup) {

    envDecl = sup;
    type = value->typeExp(env, status, &aAnyType);
    if (!status->ok()) return;
    if (type->isKind(Type::voidType)) {
	status->setNeedValue(value);
        return;
	}
    if (value->getConstant() == 0) {
	status->setNotConstant(value);
	return;
	}
}

/*****************************************************************************/
void TypeDecl::typeDecl(const TypeEnv* env,TypeStatus* status, const Decl* sup){

    envDecl = sup;
    typeRef->typeType(env, status);
    if (!status->ok())
	return;
}

/*****************************************************************************/
void VarDecl::typeDecl(const TypeEnv* env, TypeStatus* status, const Decl* sup){

    envDecl = sup;
    type->typeType(env, status);
    if (!status->ok())
	return;
}

/*****************************************************************************/
void FunArgDecl::typeDecl(const TypeEnv* env, TypeStatus* status, 
    const Decl* sup) {

    envDecl = sup;
    type->typeType(env, status);
    if (!status->ok())
	return;
}

/*****************************************************************************/
void FunDecl::typeDecl(const TypeEnv* env, TypeStatus* status, const Decl* sup){

    envDecl = sup;
    type->typeType(env, status);
    if (!status->ok()) return;
    if (body != 0) body->typeMethod(env, status, type, this);
}

/*****************************************************************************/
void FunDecl::typeDeclBody(const TypeEnv* env, TypeStatus* status) {

    external = 0;
    ObjectType* obj = (ObjectType*) type->getObj();
    if (obj != 0) {
	external = (env->getSelfType() != obj);
	if (external)
	    obj->externalMeths = new(obj,"PiList") PiList(this, obj->externalMeths);
	}

    if (body != 0) {
	if (external) {
		// create the scope of the external object (note only the
		// PUBLIC scope is available)
	    TypeEnv myEnv((TypeScope*) obj->getScope(), env, FALSE);
	    body->typeMethodBody(&myEnv, status);
	    }
	else
	    body->typeMethodBody(env, status);
	}
}

/*****************************************************************************/
void EnumDecl::typeDecl(const TypeEnv* env,TypeStatus* status, const Decl* sup){

    envDecl = sup;
    typeRef->decl = this;
    const PiList* enumPtr = typeRef->getEnums();
    while (enumPtr != 0) {
        Decl* enum_ = (Decl*) enumPtr->first();
        enum_->typeDecl(env, status, this);
        enumPtr = enumPtr->rest();
        }
}

/*****************************************************************************/
void EnumValDecl::typeDecl(const TypeEnv*, TypeStatus*, 
    const Decl* sup) {

    envDecl = sup;
    type = sup->toEnumDecl()->getTypeRef();
}

/*****************************************************************************/
void ObjectDecl::addObjDeclToScope(TypeScope* scope, TypeStatus* status) {

    if(scope->addDecl(this) != 0) {
        status->setIdMultDefined(this, getName(), getType());
	return;
	}
}

/*****************************************************************************/
void ObjectDecl::typeDecl(const TypeEnv* env, TypeStatus* status,
    const Decl*sup) {

    envDecl = sup;
    typeRef->declm = this;
    typeRef->typeType(env, status);
}

/*****************************************************************************/
void ObjectType::typeType(const TypeEnv* env, TypeStatus* status) {

    if (superClasses != 0) {
	delete superClasses;
	superClasses = 0;
	}
    if (deferedMeths != 0) {
	delete deferedMeths;
	deferedMeths = 0;
	}
    if (externalMeths != 0) {
	delete externalMeths;
	externalMeths = 0;
	}
    scope->clear();
	
	// inherit methods from your superclasses
    PiList* superPtr = supers;
    while (superPtr != 0) {
	Ident* super = (Ident*) superPtr->first();
	const Decl* decl = super->typeIdent(env, status, &aObjectType);
	if (!status->ok()) return;
	const ObjectType* sClass=decl->getTypeDef()->normType()->toObjectType();
	scope->merge(sClass->getScope());
	superClasses = new(this, "PiList") 
	    PiList((ObjectType*) sClass, superClasses);
	superPtr = superPtr->rest();
	}

    addObjDeclsToScope(decls, scope, status);
    if (!status->ok()) return;

    typeAndAddDecls(decls, env, scope, status, declm);
    if (!status->ok()) return;
    
	// make sure that no instance variables are in here.
    const PiList* declPtr = decls;	// add my defered methods to the list
    while (declPtr != 0) {
	Decl* decl = (Decl*) declPtr->first();
	declPtr = declPtr->rest();	// important that this be before remove
	if (decl->normDecl()->toVarDecl() != 0 && 
	    decl->getVisibility() == Decl::publicVis) {
                status->setPublicVarIllegal(decl);
		return;
		}
	}

    if(scope->addDecl(declm) != 0) {	// add self to scope 
        status->setIdMultDefined(declm, declm->getName(), declm->getType());
	return;
	}

	// update `deferedMethod', first all defered methods of superclasses
    const PiList* superClassPtr = superClasses;
    while (superClassPtr != 0) {
	const ObjectType* superClass = (ObjectType*) superClassPtr->first();
	PiList* defPtr = superClass->deferedMeths;
	while (defPtr != 0) {
	    Decl* decl = (Decl*) defPtr->first();
	    if (definedBy(decl->getName(), decl->getType(), decls) == 0)
		deferedMeths = new(this, "PiList") PiList(decl, deferedMeths);
	    defPtr = defPtr->rest();
	    }
	superClassPtr = superClassPtr->rest();
	}

    declPtr = decls;	// add my defered methods to the list
    while (declPtr != 0) {
	Decl* decl = (Decl*) declPtr->first();
	const FunDecl* asFunDecl = decl->normDecl()->toFunDecl();
	if (asFunDecl != 0 && asFunDecl->isDeferedMeth())
	    deferedMeths = new(this, "PiList") PiList(decl, deferedMeths);
	declPtr = declPtr->rest();
	}
}

/*****************************************************************************/
void ObjectDecl::typeDeclBody(const TypeEnv* env, TypeStatus* status) {

    typeRef->typeBody(env, status);
}

/*****************************************************************************/
void ObjectType::typeBody(const TypeEnv* env, TypeStatus* status) {

    TypeEnv myTypeEnv(scope, env);
    typeDeclBodies(decls, &myTypeEnv, status);

	// typeCheck the DOME code
    PiList* clausePtr = clauses;
    while (clausePtr != 0) {
	Clause* clause = (Clause*) clausePtr->first();
	clause->typeClause(&myTypeEnv, status);
	if (!status->ok()) {
	    return;
	    }
	clausePtr = clausePtr->rest();
	}
}

/*****************************************************************************/
void AssocDecl::typeDecl(const TypeEnv* env, TypeStatus* status,
    const Decl*sup) {

    envDecl = sup;
	// FIX deal with super associations
    if (assocEnv != 0)
	assocEnv->destroy();
    assocEnv = env->copy(this);
    if (scope != 0)
	scope->unlink();
    scope = new(this, "TypeScope") TypeScope();

    if (testArgs != 0)
	unlinkNodePiList(testArgs);
    testArgs = 0;
    PiList* argPtr = args;
    while (argPtr != 0) {
	AssocArgDecl* arg = (AssocArgDecl*) argPtr->first();
	arg->typeDecl(env, status, sup);
	if (!status->ok()) return;

	Decl* argDecl = 0;
	const Type* argType = arg->getTypeDef();
	if (argType->isKind(Type::anyObjType)) 
	    argDecl = new(this, "ObjectDecl") ObjectDecl(str(arg->getName(), 
		this), new(this, "ObjectType") ObjectType(str(arg->getName(), 
		this),0,0,0), Decl::privateVis);
	else
	    argDecl = new(this, "TypeDecl") TypeDecl(str(arg->getName(), this),
		argType->copy(this), Decl::privateVis);

	testArgs = new(this, "PiList") PiList(argDecl, testArgs);
	argPtr = argPtr->rest();
	}

    addObjDeclsToScope(testArgs, scope, status);
    if (!status->ok()) return;
    addObjDeclsToScope(reqs, scope, status);
    if (!status->ok()) return;
    addObjDeclsToScope(roles, scope, status);
    if (!status->ok()) return;

    typeAndAddDecls(testArgs, env, scope, status, this);
    if (!status->ok()) return;
    typeAndAddDecls(reqs, env, scope, status, this);
    if (!status->ok()) return;
    typeAndAddDecls(roles, env, scope, status, this);

    PiList* argTypes = 0;
    argPtr = args;
    while (argPtr != 0) {
	AssocArgDecl* arg = (AssocArgDecl*) argPtr->first();
	argTypes = new(this, "PiList") PiList((Type*) arg->getTypeDef(), argTypes);
	argPtr = argPtr->rest();
	}
    if (assocDecls != 0)
	unlinkNodePiList(assocDecls);
    assocDecls = genAssocDecls(argTypes, env, status, this);
    delete argTypes;

    addDeclsToScope(assocDecls, scope, status);
    if (!status->ok()) return;
}

/*****************************************************************************/
void AssocDecl::typeDeclBody(const TypeEnv* env, TypeStatus* status) {

    TypeEnv myTypeEnv(scope, env);
    typeDeclBodies(roles, &myTypeEnv, status);
}

/*****************************************************************************/
void AssocInstDecl::typeDecl(const TypeEnv* env, TypeStatus* status, 
	const Decl* sup) {

    envDecl = sup;
    if (scope != 0)
	scope->unlink();
    scope = new(this, "TypeScope") TypeScope();
	
    const Decl* templ = ident->typeIdent(env, status, &aAssocType);
    if (!status->ok()) return;
    assoc = templ->normDecl()->toAssocDecl();

    unlinkNodePiList(decls);
    decls = copyDecls(assoc->getRoles(), this);

    unlinkNodePiList(argDecls);
    argDecls = 0;
    const PiList* formArgPtr = assoc->getArgs();
    PiList* actArgPtr = args;
    while (actArgPtr != 0 && formArgPtr != 0) {
	Type* actArg = (Type*) actArgPtr->first();
	AssocArgDecl* formArg = (AssocArgDecl*) formArgPtr->first();
	actArg->link();
	Decl* argDecl = new(this, "TypeDecl")  TypeDecl(str(formArg->getName(),
	     this), actArg, Decl::privateVis);
	argDecls = new(this, "PiList") PiList(argDecl, argDecls);
	actArgPtr = actArgPtr->rest();
	formArgPtr = formArgPtr->rest();
	}
    if (actArgPtr != 0 || formArgPtr != 0) {
	status->setBadNumArgs(this);
	return;
	}

    typeAndAddDecls(argDecls, env, scope, status, this);
    if (!status->ok()) {
	status->setBadInstance(this, getName());  // FIX poor error message
	return;
	}

    addObjDeclsToScope(decls, scope, status);
    typeAndAddDecls(decls, assoc->getEnv(), scope, status, this);
    if (!status->ok()) {
	status->setBadInstance(this, getName());
	return;
	}

    TypeEnv reqEnv(scope, env);
    unlinkNodePiList(reqDecls);
    reqDecls = 0;
    const PiList* reqPtr = assoc->getReqs();
    while (reqPtr != 0) {
	Decl* req = (Decl*) reqPtr->first();
	Decl* reqDecl = new(this, "RenameDecl") RenameDecl(
	    str(req->getName(), this), new(this, "Ident") 
	    Ident(str(req->getName(), this)), 
	    req->getType()->copy(this), Decl::publicVis);
	reqDecls = new(this, "PiList") PiList(reqDecl, reqDecls);
	reqPtr = reqPtr->rest();
	}

    typeAndAddDecls(reqDecls, env, scope, status, this);
    if (!status->ok()) {
	status->setBadInstance(this, getName());
	return;
	}

    unlinkNodePiList(assocDecls);
    assocDecls = genAssocDecls(args, env, status, this);
    addDeclsToScope(assocDecls, scope, status);
    if (!status->ok()) return;
}

/*****************************************************************************/
void AssocInstDecl::typeDeclBody(const TypeEnv*, TypeStatus* status) {

    TypeEnv instEnv(scope, assoc->getEnv());
    typeDeclBodies(decls, &instEnv, status);
    if (!status->ok()) {
	status->setBadInstance(this, getName());	// FIX error message
	return;
	}
}

/*****************************************************************************/
void AssocArgDecl::typeDecl(const TypeEnv* env, TypeStatus* status, 
	const Decl* sup) {

    envDecl = sup;
    type->typeType(env, status);
    if (!status->ok()) return;
}

/*****************************************************************************/
void SynType::typeType(const TypeEnv* env, TypeStatus* status) {

    typeRef->typeType(env, status);
    if (!status->ok()) return;
}

/*****************************************************************************/
void IdentType::typeType(const TypeEnv* env, TypeStatus* status)
{
    def = ident->typeIdent(env, status, &aSetType);
    if (!status->ok()) return;
    def = def->normDecl();	// resolve Renamings, FlashBacks etc
}

/*****************************************************************************/
void FunType::typeType(const TypeEnv* env, TypeStatus* status)
{
    PiList* argPtr = args;
    while (argPtr != 0) {
	FunArgDecl* arg = (FunArgDecl*) argPtr->first();
	((Type*) arg->getType())->typeType(env, status);
	if (!status->ok()) return;
	argPtr = argPtr->rest();
	}
    retType->typeType(env, status);
    if (!status->ok()) return;

    obj = 0;					// set the obj type
    if (kind != ftn) {
	if (objType != 0) {
	    objType->typeType(env, status);
	    if (!status->ok()) return;
	    obj = objType->normType()->toObjectType();
	    if (obj == 0) {
		status->setNotObject(objType);
		return;
		}
	    }
	else {
	    obj = env->getSelfType();
	    if (obj == 0) {
		status->setSelfNotFound(this);
		return;
		}
	    }

	if ((kind == destructor || kind == constructor) && 
	   !retType->isKind(voidType)) {
	    status->setRetNotExpected(retType);
	    return;
	    }

	if (kind == destructor && argPtr != 0) {
	    status->setArgNotExpected(this);
	    return;
	    }
	}
}

/*****************************************************************************/
void FutureType::typeType(const TypeEnv* env, TypeStatus* status) {

    subType->typeType(env, status);
}

/*****************************************************************************/
void ArrayType::typeType(const TypeEnv* env, TypeStatus* status) {

    evalBound = 0;
    if (bound == 0)
	indexType = &aIntType;
    else {
	indexType = bound->typeExp(env, status, &aAnyType);
	if (!status->ok()) return;
	const Exp* boundExp = bound->getConstant();
	if (boundExp == 0) {
	    status->setNotConstant(bound);
	    return;
	    }
	const IntExp* asIntExp = boundExp->toIntExp();
	if (asIntExp == 0) {
	    status->setTypeConflict(bound, indexType, &aIntType);
	    return;
	    }
	evalBound = asIntExp->getValue();
	}
    elemType->typeType(env, status);
    if (!status->ok())
	return;
}

/*****************************************************************************/
void MethodBody:: typeMethod(const TypeEnv* env, TypeStatus* status, 
    FunType* type, const Decl* sup) {

    if (localScope != 0) localScope->unlink();
    localScope = new(this, "TypeScope") TypeScope(type->getRetType());

    addDeclsToScope(type->getArgs(), localScope, status);
    if (!status->ok()) return;
    addObjDeclsToScope(decls, localScope, status);
    if (!status->ok()) return; 
    typeAndAddDecls(decls, env, localScope, status, sup);
    if (!status->ok()) return;
}

/*****************************************************************************/
void MethodBody::
typeMethodBody(const TypeEnv* env, TypeStatus* status) {

    TypeEnv myTypeEnv(localScope, env);		// new scope
    typeDeclBodies(decls, &myTypeEnv, status);
    if (!status->ok()) return; 

// FIX should make sure that only direct superclasses are called
    PiList* initPtr = constrInits;
    while (initPtr != 0) {
	NewExp* init = (NewExp*) initPtr->first();
	init->typeNewExp(&myTypeEnv, status, 0);
        if (!status->ok()) return; 
	initPtr = initPtr->rest();
	}
    typeStmts(stmts, &myTypeEnv, status, 0);

    // JB:
    TotReentryLabels = ApplyExp::CumReentryLabel;
    ApplyExp::CumReentryLabel = 0;
};

/*****************************************************************************/
void IfStmt::
typeStmt(const TypeEnv* env, TypeStatus* status, PiList** varExps) {

    const Type* condType = cond->typeExp(env, status, &aBoolType, varExps);
    if (!status->ok()) return;
    if (!condType->canBe(&aBoolType)) {
	status->setTypeConflict(cond, condType, &aBoolType);
	return;
	}
    typeStmts(then, env, status, varExps);
    if (!status->ok())
	return;
    typeStmts(else_, env, status, varExps);
}

/*****************************************************************************/
void WhileStmt::
typeStmt(const TypeEnv* env, TypeStatus* status, PiList** varExps) {

    const Type* condType = cond->typeExp(env, status, &aBoolType, varExps);
    if (!status->ok()) return;
    if (!condType->canBe(&aBoolType)) {
	status->setTypeConflict(cond, condType, &aBoolType);
	return;
	}

    typeStmts(body, env, status, varExps);
}

/*****************************************************************************/
void RepeatStmt::
typeStmt(const TypeEnv* env, TypeStatus* status, PiList** varExps) {

    const Type* condType = cond->typeExp(env, status, &aBoolType, varExps);
    if (!status->ok()) return;
    if (!condType->canBe(&aBoolType)) {
	status->setTypeConflict(cond, condType, &aBoolType);
	return;
	}

    typeStmts(body, env, status, varExps);
}

/*****************************************************************************/
void LoopStmt::
typeStmt(const TypeEnv* env, TypeStatus* status, PiList** varExps) {
    typeStmts(body, env, status, varExps);
}

/*****************************************************************************/
void ForStmt::
typeStmt(const TypeEnv* env, TypeStatus* status, PiList** varExps) {

    index->typeExp(env, status, &aIntType, varExps);
    if (!status->ok()) return;
    if (!index->isLValue()) {
	status->setNotLvalue(index);
	return;
	}

    const Type* fromType = from->typeExp(env, status, &aIntType, varExps);
    if (!status->ok()) return;
    const Type* toType = to->typeExp(env, status, &aIntType, varExps);
    if (!status->ok()) return;

    if (by != 0) {
	const Type* byType = by->typeExp(env, status, &aIntType, varExps);
	if (!status->ok()) return;
	}
    typeStmts(body, env, status, varExps);
}

/*****************************************************************************/
void ReturnStmt::
typeStmt(const TypeEnv* env, TypeStatus* status, PiList** varExps) {

    funRetType = env->getRetType();
    if (val == 0) {
	if (!funRetType->isKind(Type::voidType))
	    status->setNeedRetVal(this, funRetType);
	return;
	}
    else {
	if (funRetType->isKind(Type::voidType)) {
	    status->setBadRetExp(val);
	    return;
	    }
	}
    const Type* retType = val->typeExp(env, status, &aAnyType, varExps);
    if (!status->ok()) return;
    if (!retType->canBe(funRetType)) {
	status->setTypeConflict(val, retType, funRetType);
	return;
	}
}

/*****************************************************************************/
void ExitStmt::typeStmt(const TypeEnv*, TypeStatus*, PiList**) {

   // Don't need to do anything.
}

/*****************************************************************************/
void BlockStmt::
typeStmt(const TypeEnv* env, TypeStatus* status, PiList** varExps) {

    TypeScope* localScope = new(this, "TypeScope") TypeScope();	
    TypeEnv myTypeEnv(localScope, env);		// new scope
    localScope->unlink(); 

    addObjDeclsToScope(decls, localScope, status);
    if (!status->ok()) return; 
    typeAndAddDecls(decls, env, localScope, status);
    if (!status->ok()) return; 
    typeDeclBodies(decls, &myTypeEnv, status);
    if (!status->ok()) return; 

    typeStmts(body, &myTypeEnv, status, varExps);
}


/*****************************************************************************/
void CaseStmt::
typeStmt(const TypeEnv* env, TypeStatus* status, PiList** varExps) {

    const Type* caseExpType = expm->typeExp(env, status, &aAnyType, varExps);
    if (!status->ok()) return;

	// check to make sure that expression is of a valid case type
    if (!caseExpType->indexAble()) {
	status->setNotIndexAble(this, caseExpType);
	return;
	}
	
    PiList* whenPtr = whens;
    while (whenPtr != 0) {
	const When* when = (When*) whenPtr->first();
	const PiList* exps = when->whenLabels;
	while (exps != 0) {
	    Exp* exp = (Exp*) exps->first();
	    const Type* expType = exp->typeExp(env, status, &aAnyType, varExps);
	    if (!status->ok()) return;
	    if (!expType->canBe(caseExpType)) {
		status->setTypeConflict(exp, expType, caseExpType);
		return;
		}
	    if (exp->getConstant() == 0) {
		status->setNotConstant(exp);
		return;
		}
	    exps = exps->rest();
	    }
	typeStmts(when->stmts, env, status, varExps);
	if (!status->ok())
	    return;
	whenPtr = whenPtr->rest();
	}
    if (otherwise != 0) 
	typeStmts(otherwise, env, status, varExps);
}

/*****************************************************************************/
void AssignStmt::
typeStmt(const TypeEnv* env, TypeStatus* status, PiList** varExps) {

    lhsType = lhs->typeExp(env, status, &aAnyType, varExps);
    if (!status->ok()) return;
    
    rhsType = rhs->typeExp(env, status, &aAnyType, varExps);
    if (!status->ok()) return;

    if (!rhsType->canBe(lhsType)) {
	status->setTypeConflict(rhs, rhsType, lhsType);
	return;
	}
    if (!lhs->isLValue()) {
	status->setNotLvalue(lhs);
	return;
	}
}

/*****************************************************************************/
AsyncStmt::AsyncStmt(Exp* aLhs, ApplyExp* aApply, Exp *aDestEnumIdExp, 
		     enum WaitKind aWaitKind, Exp *aWaitValExp):
   lhs(aLhs), apply(aApply), destEnumIdExp(aDestEnumIdExp), kind(aWaitKind),
   waitValExp(aWaitValExp)
{
   // JB this constructor adds this to the event method list is appropriate

   if (!globalAsyncStmtPiList) globalAsyncStmtPiList = new PiList(this, 0);
   PiList *listPtr = globalAsyncStmtPiList;
   int needToAdd = 1;
   while (listPtr)
   {
      if (!strcmp(((AsyncStmt*)(listPtr->first()))->
		  getApply()->getName()->getIdent(),
		  aApply->getName()->getIdent()))
      {
	 needToAdd = 0;
	 break;
      }
      listPtr = listPtr->rest();
   }
   if (needToAdd)
   {
      PiList *newPiList = new PiList(this, globalAsyncStmtPiList);
      globalAsyncStmtPiList = newPiList;
   }
   // JB FIX: could put code here to, on a source file basis, make list of 
   //         declare statements needed at the top of the file
}
   

void AsyncStmt::
typeStmt(const TypeEnv* env, TypeStatus* status, PiList** varExps) {

    const Type* applyType = apply->typeExp(env, status, &aAnyType, varExps);
    if (!status->ok()) return;

    // JB
    const Type *ret;
    if (destEnumIdExp)
    {
       ret = destEnumIdExp->typeExp(env, status, &aIntType, varExps);
       if (!status->ok()) return;
       if (!ret->toAtomicType() || !ret->toAtomicType()->isKind(Type::intType))
       {
	  status->setTypeConflict(destEnumIdExp, ret, &aIntType);
	  return;
       }
    }
    if (waitValExp)
    {
      ret = waitValExp->typeExp(env, status, &aRealType, varExps);
      if (!status->ok()) return;
      if (!ret->toAtomicType() ||
	  (!ret->toAtomicType()->isKind(Type::intType) &&
	   !ret->toAtomicType()->isKind(Type::realType)))
      {
	 status->setTypeConflict(waitValExp, ret, &aRealType);
	 return;
      }
    }
    
    if (lhs != 0) {
	const Type* lhsType = lhs->typeExp(env, status, &aAnyType, varExps);
	if (!status->ok()) return;
	if (!lhs->isLValue()) {
	    status->setNotLvalue(lhs);
	    return;
	    }
	if (!lhsType->canBe(&aAnyFutureType)) {
	    status->setTypeConflict(lhs, lhsType, &aAnyFutureType);
	    return;
	    }
	if (applyType->isKind(Type::voidType) && 
	    !lhsType->isKind(Type::anyFutureType)) {
	    status->setTypeConflict(lhs, applyType, lhsType);
	    return;
	    }
	else if (!applyType->canBe(lhsType)) {
	    status->setTypeConflict(apply, applyType, lhsType);
	    return;
	    }
	}
}

/*****************************************************************************/
void ExpStmt::
typeStmt(const TypeEnv* env, TypeStatus* status, PiList** varExps) {

    exp->typeExp(env, status, &aAnyType, varExps);
}

/*****************************************************************************/
void DisposeStmt::
typeStmt(const TypeEnv* env, TypeStatus* status, PiList** varExps) {

    const Type* bodyType = body->typeExp(env, status, &aAnyType, varExps);
    if (!status->ok()) return;
    if (!bodyType->canBe(&aAnyObjType)) {
	status->setTypeConflict(body, bodyType, &aAnyObjType);
	return;
	}
}

/*****************************************************************************/
const Type* NewExp::typeExp(const TypeEnv* env, TypeStatus* status,
        const Type*, PiList** varExps) {

    typeNewExp(env, status, varExps);
    if (!status->ok()) return(0);

    const ObjectType* asObj = type->normType()->toObjectType();
    PiList* deferedMeths = asObj->abstractClass();
    if (deferedMeths != 0) {
        status->setDeferedMeths(this, deferedMeths);
        return(0);
        }

    return(type);
}

/*****************************************************************************/
const Type* NewExp::typeNewExp(const TypeEnv* env, TypeStatus* status,
        PiList** varExps) {

    type->typeType(env, status);
    if (!status->ok()) return(0);

    const ObjectType* asObj = type->normType()->toObjectType();
    if (asObj == 0) {
        status->setNotObject(this);
        return(0);
        }

    FunType* constrType = 
	funTypeFromArgs(FunType::constructor, type, args, env, status, varExps);
    if (!status->ok()) return(0);

	// look up constructor in object as well as the environment
    TypeEnv objEnv((TypeScope*) asObj->getScope(), env, FALSE);
    const Decl* ret= objEnv.findIdent("", constrType);
    if (ret == 0) {
	status->setConstrNotFound(this, type, constrType, env);
	delFunTypeFromArgs(constrType);
	return(0);
	}
    else if (ret == TypeScope::ambiguousDecl()) {
	status->setConstrAmbiguous(this, type, constrType, env);
	delFunTypeFromArgs(constrType);
	return(0);
	}
    else if (ret == TypeScope::privateDecl()) {
	status->setConstrPrivate(this, type, constrType, env);
	delFunTypeFromArgs(constrType);
	return(0);
	}
    delFunTypeFromArgs(constrType);
    constrDecl = ret->normDecl()->toFunDecl();

    if (newArgs != 0) {
	FunType* newType = 	
	    funTypeFromArgs(FunType::ftn, 0, newArgs, env, status, varExps);
	if (!status->ok()) return(0);
	    ret = env->findIdent("NEW", newType);
	if (ret == 0 || ret == TypeScope::ambiguousDecl()) {
	    status->setBadNewArgs(this, newType);
	    delFunTypeFromArgs(newType);
	    return(0);
	    }
	delFunTypeFromArgs(newType);
	}
    newDecl = ret->normDecl()->toFunDecl();
    return(type);
}

/*****************************************************************************/
const Type* ApplyExp::typeExp(const TypeEnv* env, TypeStatus* status, 
			      const Type*, PiList** varExps)
{

   // JB:
   if (
       !strcmp(name->getIdent(), "wait") ||
       !strcmp(name->getIdent(), "waitFor") ||
       !strcmp(name->getIdent(), "waitUntil")
       )
     ReentryLabel = ++CumReentryLabel;
   
    Bool calledAsync = (kind == FunType::tell);
    FunKind applyKind = calledAsync ? FunType::ask : kind;

    const Type* objType = 0;
    if (obj != 0)
    {
        objType = obj->typeExp(env, status, &aAnyType, varExps);
        if (!status->ok()) return(0);

	objType = objType->normType();
	const ObjectType* asObj = objType->toObjectType();
	if (asObj == 0)
	{
	    status->setNotObject(obj);
	    return(0);
	}
    }
    FunType* callType = 
	funTypeFromArgs(applyKind, objType, args, env, status, varExps);
    if (!status->ok()) return(0);

    decl  = name->typeIdent(env, status, callType);    
    delFunTypeFromArgs(callType);
    if (!status->ok()) return(0);
    decl = decl->normDecl();

    fromObj =0;
    if (from != 0)
    {
        const Decl* fromDecl = from->typeIdent(env, status, &aObjectType);
        if (!status->ok()) return(0);
	fromDecl = fromDecl->normDecl();
	fromObj = fromDecl->toObjectDecl()->getTypeDef()->toObjectType();
	if (!objType->canBe(fromObj))
	{	// is it a superclass of myself
		// this is not really a good error message, but it is
		// probably close enough
	   status->setIdentNotFound(from, from, &aObjectType, env);
	   return(0);
	}
    }
    
    const FunType* funType = decl->getType()->toFunType();
    const PiList* formArgPtr = funType->getArgs();
    const PiList* argPtr = args;
    while (formArgPtr != 0) {
	FunArgDecl* formArg = (FunArgDecl*) formArgPtr->first();
	Exp* arg = (Exp*) argPtr->first();
	if ((formArg->getKind() != FunArgDecl::in)) {
	    if (!arg->isLValue()) {
		status->setNotLvalue(arg);
		return(0);
		}
		// can't pass a non-tagged to a function requiring a
		// tag as in IN-OUT parameter.
	    if (!arg->getType()->isKind(Type::taggedType) &&
		formArg->getType()->isKind(Type::taggedType)) {
		status->setNotLvalue(arg);
		return(0);
		}
	    }
	formArgPtr = formArgPtr->rest();
	argPtr = argPtr->rest();
	}

	// FIX this is a kludge, we should not assume that an function ID
	// points back to a FunDecl (it could be an arg)
    const FunDecl* funDecl = decl->toFunDecl();
    if (funDecl != 0)
    {
       if (funDecl->retType(calledAsync)->toAtomicType())
       return(funDecl->retType(calledAsync));
    }
    
    return(decl->getType()->toFunType()->getRetType());
}

/*****************************************************************************/
const Type* AsExp::typeExp(const TypeEnv* env, TypeStatus* status, 
    const Type*, PiList** varExps) {

    type->typeType(env, status);
    if (!status->ok()) return(0);
    const Type* ret = body->typeExp(env, status, type, varExps);
    if (!status->ok()) return(0);
    if (!ret->canBe(type)) {
	status->setTypeConflict(body, ret, type);
	return(0);
	}
    return(type);
}

/*****************************************************************************/
const Type* CastExp::typeExp(const TypeEnv* env, TypeStatus* status, 
    const Type*, PiList** varExps) {

    type->typeType(env, status);
    if (!status->ok()) return(0);
    bodyType = body->typeExp(env, status, &aAnyType, varExps);
    if (!status->ok()) return(0);
    if (!bodyType->isKind(Type::taggedType)) {
	status->setTypeConflict(this, bodyType, type);
	return(0);
	}
    return(type);
}

/*****************************************************************************/
const Type* IsAExp::typeExp(const TypeEnv* env, TypeStatus* status, 
    const Type*, PiList** varExps) {

    const Type* bodyType = body->typeExp(env, status, &aAnyType, varExps);
    if (!status->ok()) return(0);
    if (subType != 0) {	// subType = 0 means IsA VAR
	subType->typeType(env, status);
	if (!status->ok()) return(0);
	}

    if (!bodyType->canBe(&aTaggedType)) {
	status->setTypeConflict(body, bodyType, &aTaggedType);
	return(0);
	}

    return(&aBoolType);
}

/*****************************************************************************/
const Type* IdentExp::typeExp(const TypeEnv* env, TypeStatus* status, 
    const Type* context, PiList**) {

    def = ident->typeIdent(env, status, context);
    if (!status->ok())
    {
       return (0);
    }
    def = def->normDecl();
    return(def->getType());
}

/*****************************************************************************/
const Type* SelfExp::typeExp(const TypeEnv* env, TypeStatus* status, 
    const Type*, PiList**) {

    type = env->getSelfType();
    if (type == 0) {
        status->setSelfNotFound(this);
	return(0);
	}
    return(type);
}

/*****************************************************************************/
const Type* ArrayExp::typeExp(const TypeEnv* env, TypeStatus* status, 
    const Type*, PiList** varExps) {

    const Type* indexType = index->typeExp(env, status, &aAnyType, varExps);
    if (!status->ok()) return(0);
    
    if (!indexType->indexAble()) {
	status->setNotIndexAble(this, indexType);
	return(0);
	}
	
    const Type* expType = array->typeExp(env, status, &aAnyType, varExps);
    if (!status->ok()) return(0);

    const ArrayType* arrayType = expType->normType()->toArrayType();
    if (arrayType == 0) {
	status->setNotArray(array);
	return(0);
	}

    if (!indexType->canBe(arrayType->getIndexType())) {
	status->setTypeConflict(this, indexType, arrayType->getIndexType());
	return(0);
	}
	
    type = arrayType->getElemType();
    return(type);
}

/*****************************************************************************/
const Type* VarExp::typeExp(const TypeEnv*, TypeStatus* status, 
    const Type* context, PiList** varExps) {

    if (varExps == 0) {
	status->setDomeIllegal(this);
	return(0);
	}
    type = context;
    
    PiList* varPtr = *varExps;
    while (varPtr != 0) {
	VarExp* varExp = (VarExp*) varPtr->first();
	if (strcmp(varExp->var, var) == 0)	// already exists 
	    return(&aTaggedType);
	varPtr = varPtr->rest();
	}

	// does not exist, add to list.
    *varExps = new(this, "PiList") PiList(this, *varExps);
    return(&aTaggedType);
}

/*****************************************************************************/
const Type* IntExp::typeExp(const TypeEnv*, TypeStatus*, const Type*, PiList**) {

    return(&aIntType);
}

/*****************************************************************************/
const Type* RealExp::typeExp(const TypeEnv*, TypeStatus*, const Type*, PiList**) {

    return(&aRealType);
}

/*****************************************************************************/
const Type* StringExp::typeExp(const TypeEnv*,TypeStatus*,const Type*, PiList**) {

    return(&aStringType);
}

/*****************************************************************************/
const Type* CharExp::typeExp(const TypeEnv*, TypeStatus*, const Type*, PiList**) {

    return(&aCharType);
}

/*****************************************************************************/
const Type* NullExp::typeExp(const TypeEnv*, TypeStatus*, const Type*, PiList**) {

    return(&aUnknownObjType);	
}

/*****************************************************************************/
const Type* BoolExp::typeExp(const TypeEnv*, TypeStatus*, const Type*, PiList**) {

    return(&aBoolType);
}

/*****************************************************************************/
void PropClause::typeClause(const TypeEnv* env, TypeStatus* status) {

    goal->typeTerm(env, status);
    if (!status->ok()) return;
    typeTerms(terms, env, status);
}

/*****************************************************************************/
void ImportTerm::typeTerm(const TypeEnv* env, TypeStatus* status) {

	// create the local scope for the variable 'INDEX'
    if (index != 0) delete index;
    index = new(this, "VarDecl") 
	VarDecl("INDEX", (AtomicType*) &aIntType, Decl::publicVis);
    
    TypeScope* localScope = new(this, "TypeScope") TypeScope(&aBoolType);
    localScope->addDecl(index);
    TypeEnv myEnv(localScope, env);
    localScope->unlink(); 

    if (vars != 0) delete vars;
    vars = 0;

    typeStmts(stmts, &myEnv, status, &vars);
    if (!status->ok()) return; 
    obj = env->getSelfType();
}

/*****************************************************************************/
void FunctorTerm::typeTerm(const TypeEnv* env, TypeStatus* status) {
    typeTerms(args, env, status);
}

/*****************************************************************************/
void ListTerm::typeTerm(const TypeEnv* env, TypeStatus* status) {
    typeTerms(firstTerms, env, status);
}

/*****************************************************************************/
void ConstTerm::typeTerm(const TypeEnv* env, TypeStatus* status) {
    type = exp->typeExp(env, status, &aAnyType); 
    if (!status->ok()) return;
    if (exp->getConstant() == 0) {
	status->setNotConstant(exp);
	return;
	}
    if (!type->canBe(&aTaggedType)) {
	status->setTypeConflict(this, type, &aTaggedType);
	return;
	}
}

/*****************************************************************************/
static void typeDeclBodies(PiList* decls, const TypeEnv* env, TypeStatus* status){

    PiList* declPtr = decls;
    while (declPtr != 0) {
	Decl* decl = (Decl*) declPtr->first();
	decl->typeDeclBody(env, status);
	if (!status->ok())
	    return;
	declPtr = declPtr->rest();
	}
}

/*****************************************************************************/
static void typeAndAddDecls(PiList* decls, const TypeEnv* env, TypeScope* scope, 
    TypeStatus* status, const Decl* sup) {

    TypeEnv myTypeEnv(scope, env);		// new scope
    PiList* declPtr = decls;
    while (declPtr != 0) {
	Decl* decl = (Decl*) declPtr->first();
	decl->typeDecl(&myTypeEnv, status, sup);
	if (!status->ok()) return;
	decl->addDeclToScope(scope, status);
	if (!status->ok()) return;
	declPtr = declPtr->rest();
	}
}

/*****************************************************************************/
static void addObjDeclsToScope(PiList* decls, TypeScope* scope, 
    TypeStatus* status) {

    PiList* declPtr = decls;
    while (declPtr != 0) {
	Decl* decl = (Decl*) declPtr->first();
	decl->addObjDeclToScope(scope, status);
	if (!status->ok())
	    return;
	declPtr = declPtr->rest();
	}
}

/*****************************************************************************/
static void addDeclsToScope(const PiList* decls, TypeScope* scope, 
    TypeStatus* status) {

    const PiList* declPtr = decls;
    while (declPtr != 0) {
	Decl* decl = (Decl*) declPtr->first();
	decl->addDeclToScope(scope, status);
	if (!status->ok())
	    return;
	declPtr = declPtr->rest();
	}
}

/*****************************************************************************/
static void typeStmts(PiList* stmts, const TypeEnv* env, TypeStatus* status,
	PiList** varExps) {

    PiList* stmtPtr = stmts;
    while (stmtPtr != 0) {
	Stmt* stmt = (Stmt*) stmtPtr->first();
	stmt->typeStmt(env, status, varExps);
	if (!status->ok())
	    return;
	stmtPtr = stmtPtr->rest();
	}
}

/*****************************************************************************/
static void typeTerms(PiList* terms, const TypeEnv* env, TypeStatus* status) {

    PiList* termPtr = terms;
    while (termPtr != 0) {
	Term* term = (Term*) termPtr->first();
	term->typeTerm(env, status);
	if (!status->ok())
	    return;
	termPtr = termPtr->rest();
	}
}

/*****************************************************************************/
/* returns a declaration in 'decls' that will define (override) the identifier 
   represented by 'name':'type'.  Returns NULL if not found */

static Decl* definedBy(const char* name, const Type* type, PiList* decls) {

    PiList* declPtr = decls;
    while (declPtr != 0) {
	Decl* decl = (Decl*) declPtr->first();
	if (strcmp(name, decl->getName()) == 0) {
	    if (type->canBe(decl->getType())) 	// note backwards from normal
		return(decl);
	    }
	declPtr = declPtr->rest();
	}
    return(0);
}

/*****************************************************************************/
/* clean up a function argument list */

static void delFunArgs(PiList* argPtr) {

    while (argPtr != 0) {
        FunArgDecl* arg = (FunArgDecl*) argPtr->first();
	delete arg;
        argPtr = argPtr->rest();
        }
}

/*****************************************************************************/
/* compute the function type that the arguments have */

static FunType* funTypeFromArgs(FunKind kind, const Type* obj, const PiList* args,
    const TypeEnv* env, TypeStatus* status, PiList** varExps)
{ 
    PiList* argTypes = 0;
    const PiList* argPtr = args;
    
    while (argPtr != 0)
    {
        Exp* arg = (Exp*) argPtr->first();
        const Type* argType = arg->typeExp(env, status, &aAnyType, varExps);
	if (!status->ok())
	{
	   delFunArgs(argTypes);
	   return(0);
	}
        FunArgDecl* funArg = new FunArgDecl(FunArgDecl::out,0,(Type*) argType);
        argTypes = new PiList(funArg, argTypes);
        argPtr = argPtr->rest();
    }
    argTypes = reverse(argTypes);
    return(new FunType(kind, argTypes, (Type*) &aAnyType, (Type*) obj));
}

/*****************************************************************************/
/* clean up the function type created from funTypeFromArgs */

static void delFunTypeFromArgs(FunType* fun) {

    delFunArgs((PiList*) fun->getArgs());
    delete (PiList*) fun->getArgs();
    delete fun;
}

/*****************************************************************************/
/* an association as a set of built in functions associated with it.  This
   routine returns declarations for these built in functions */

static PiList* genAssocDecls(const PiList* args, const TypeEnv* env,
	TypeStatus* status, const Decl* sup) {

    PiList* ret = 0;

    PiList* addArgs = 0;
    const PiList* argPtr = args;
    char s[2]; s[0] = 'A'; s[1] = 0;
    while (argPtr != 0) {
        Type* arg = (Type*) argPtr->first();
	/*** FIX add should have types of association not TAGGED 
	addArgs = new(sup, "PiList") PiList(new(sup, "FunArgDecl") 
	    FunArgDecl(FunArgDecl::in, str(s, sup),arg->copy(sup)), addArgs);
	********/
	addArgs = new(sup, "PiList") PiList(new(sup, "FunArgDecl") 
	    FunArgDecl(FunArgDecl::in, str(s, sup),
	    new(sup, "AtomicType") AtomicType(Type::taggedType)), addArgs);
        argPtr = argPtr->rest();
	s[0]++;
        }
    addArgs = reverse(addArgs);
    FunDecl* addDecl = new(sup, "FunDecl") FunDecl(str("add", sup), 
	new(sup, "FunType") FunType(FunType::ftn, addArgs,
	new(sup, "AtomicType") AtomicType(Type::voidType),0), Decl::publicVis);
    addDecl->typeDecl(env, status, sup);
    ret = new(sup, "PiList") PiList(addDecl, ret);

    FunDecl* idx = new(sup, "FunDecl") FunDecl(str("index", sup), 
	new(sup, "FunType") FunType(FunType::ftn, new(sup, "PiList") PiList(
	    new(sup,"FunArgDecl") FunArgDecl(FunArgDecl::in,str("keyCol",sup),
		new(sup, "AtomicType") AtomicType(Type::intType)),
		    new(sup, "PiList") PiList(
	    new(sup,"FunArgDecl") FunArgDecl(FunArgDecl::in,str("keyVal",sup),
		new(sup, "AtomicType") AtomicType(Type::taggedType)),
		    new(sup, "PiList") PiList(
	    new(sup,"FunArgDecl") FunArgDecl(FunArgDecl::in,str("retCol",sup),
		new(sup, "AtomicType") AtomicType(Type::intType)),
		    new(sup, "PiList") PiList(
	    new(sup, "FunArgDecl") FunArgDecl(FunArgDecl::in, str("idx",sup), 
		new(sup, "AtomicType") AtomicType(Type::intType)), 0)))),
	    new(sup, "AtomicType") AtomicType(Type::taggedType)),
	Decl::publicVis);
    idx->typeDecl(env, status, sup);
    ret = new(sup, "PiList") PiList(idx, ret);

    FunDecl* idxAll = new(sup, "FunDecl") FunDecl(str("indexAll", sup), 
	new(sup, "FunType") FunType(FunType::ftn, new(sup, "PiList") PiList(
	    new(sup,"FunArgDecl") FunArgDecl(FunArgDecl::in,str("keyCol",sup),
		new(sup, "AtomicType") AtomicType(Type::intType)),
		    new(sup, "PiList") PiList(
	    new(sup,"FunArgDecl") FunArgDecl(FunArgDecl::in,str("keyVal",sup),
		new(sup, "AtomicType") AtomicType(Type::taggedType)),
		    new(sup, "PiList") PiList(
	    new(sup,"FunArgDecl") FunArgDecl(FunArgDecl::in,str("retCol",sup),
		new(sup, "AtomicType") AtomicType(Type::intType)), 0))),
	    new(sup, "AtomicType") AtomicType(Type::listType)),
	Decl::publicVis);
    idxAll->typeDecl(env, status, sup);
    ret = new(sup, "PiList") PiList(idxAll, ret);

    if (args->len() == 2) {
	Type* type1 = (Type*) args->first();
	Type* type2 = (Type*) args->rest()->first();

	/*** FIX index1 should be defined as below, where arg is not a TAGGED 
	FunDecl* idx1 = new(sup, "FunDecl") FunDecl(str("index1", sup), 
	    new(sup, "FunType") FunType(FunType::ftn, new(sup, "PiList") PiList(
		new(sup, "FunArgDecl") FunArgDecl(FunArgDecl::in, 
		    str("v", sup), type1->copy(sup)), new(sup, "PiList") PiList(
		new(sup, "FunArgDecl") FunArgDecl(FunArgDecl::in,
			str("idx",sup),
		    new(sup, "AtomicType") AtomicType(Type::intType)), 0)),
		type2->copy(sup)), Decl::publicVis);
	**** END FIX *****/
	FunDecl* idx1 = new(sup, "FunDecl") FunDecl(str("index1", sup), 
	    new(sup, "FunType") FunType(FunType::ftn, new(sup, "PiList") PiList(
		new(sup, "FunArgDecl") FunArgDecl(FunArgDecl::in, 
		    str("v", sup), new(sup, "AtomicType") 
		    AtomicType(Type::taggedType)), new(sup, "PiList") PiList(
		new(sup, "FunArgDecl") FunArgDecl(FunArgDecl::in,
			str("idx",sup),
		    new(sup, "AtomicType") AtomicType(Type::intType)), 0)),
		type2->copy(sup)), Decl::publicVis);
	idx1->typeDecl(env, status, sup);
	ret = new(sup, "PiList") PiList(idx1, ret);

	/*** FIX index2 should be defined as below, where arg is not a TAGGED 
	FunDecl* idx2 = new(sup, "FunDecl") FunDecl(str("index2", sup), 
	    new(sup, "FunType") FunType(FunType::ftn, new(sup, "PiList") PiList(
		new(sup, "FunArgDecl") FunArgDecl(FunArgDecl::in, 
		    str("v", sup), type2->copy(sup)), new(sup, "PiList") PiList(
		new(sup, "FunArgDecl") FunArgDecl(FunArgDecl::in, 
			str("idx",sup),
		    new(sup, "AtomicType") AtomicType(Type::intType)), 0)),
		type1->copy(sup)), Decl::publicVis);
	**** END FIX *****/
	FunDecl* idx2 = new(sup, "FunDecl") FunDecl(str("index2", sup), 
	    new(sup, "FunType") FunType(FunType::ftn, new(sup, "PiList") PiList(
		new(sup, "FunArgDecl") FunArgDecl(FunArgDecl::in, 
		    str("v", sup), new(sup, "AtomicType") 
		    AtomicType(Type::taggedType)), new(sup, "PiList") PiList(
		new(sup, "FunArgDecl") FunArgDecl(FunArgDecl::in, 
			str("idx",sup),
		    new(sup, "AtomicType") AtomicType(Type::intType)), 0)),
		type1->copy(sup)), Decl::publicVis);
	idx2->typeDecl(env, status, sup);
	ret = new(sup, "PiList") PiList(idx2, ret);
      }

    FunDecl* rmv1 = new(sup, "FunDecl") FunDecl(str("remove", sup), 
	new(sup, "FunType") FunType(FunType::ftn, new(sup, "PiList") PiList(
	    new(sup,"FunArgDecl") FunArgDecl(FunArgDecl::in,str("keyCol",sup),
		new(sup, "AtomicType") AtomicType(Type::intType)),
		    new(sup, "PiList") PiList(
	    new(sup,"FunArgDecl") FunArgDecl(FunArgDecl::in,str("keyVal",sup),
		new(sup, "AtomicType") AtomicType(Type::taggedType)),0)),
	    new(sup, "AtomicType") AtomicType(Type::voidType)),
	Decl::publicVis);
    rmv1->typeDecl(env, status, sup);
    ret = new(sup, "PiList") PiList(rmv1, ret);

    FunDecl* rmv2 = new(sup, "FunDecl") FunDecl(str("remove", sup), 
	new(sup, "FunType") FunType(FunType::ftn, new(sup, "PiList") PiList(
	    new(sup,"FunArgDecl") FunArgDecl(FunArgDecl::in,str("keyCol1",sup),
		new(sup, "AtomicType") AtomicType(Type::intType)),
		    new(sup, "PiList") PiList(
	    new(sup,"FunArgDecl") FunArgDecl(FunArgDecl::in,str("keyVal1",sup),
		new(sup, "AtomicType") AtomicType(Type::taggedType)),
		    new(sup, "PiList") PiList(
	    new(sup,"FunArgDecl") FunArgDecl(FunArgDecl::in,str("keyCol2",sup),
		new(sup, "AtomicType") AtomicType(Type::intType)),
		    new(sup, "PiList") PiList(
	    new(sup,"FunArgDecl") FunArgDecl(FunArgDecl::in,str("keyVal2",sup),
		new(sup, "AtomicType") AtomicType(Type::taggedType)),0)))),
	    new(sup, "AtomicType") AtomicType(Type::voidType)),
	Decl::publicVis);
    rmv2->typeDecl(env, status, sup);
    ret = new(sup, "PiList") PiList(rmv2, ret);
    return(ret);
}
    
