/*****************************************************************************/
/*                               TypeEnv.cxx                                 */
/*****************************************************************************/

/* TypeEnv defines a `TypeEnv' which defines an environment that allows
   the types of variables to be looked up */

/* AUTHOR: Vance Morrison
   DATE:  11/2/93              */
/*****************************************************************************/

#include "SpIostream.H"
#include "TypeEnv.H"
#include "SpHashTree.H"
#include "NewWhere.H"

/*****************************************************************************/
/* helper function needed to delete all the lists in the id table */

/*
static void* delPiList(const char*, const PiList* value, void*) {
    delete (PiList*) value;
    return(0);
}
*/

/*****************************************************************************/
/* delete all the lists (but not what is in the lists) in the table */

TypeScope::~TypeScope() {
    clear();
    PiList* mergePtr = merges;
    while (mergePtr != 0) {
        TypeScope* merge = (TypeScope*) mergePtr->first();
	merge->unlink();
        mergePtr = mergePtr->rest();
        }

    delete merges;
}

/*****************************************************************************/
void TypeScope::merge(const TypeScope* scope) { 
    merges = new(this, "PiList") PiList((void*) scope, merges); 
    scope->link();
}

/*****************************************************************************/
const Decl* TypeScope::addDecl(const Decl* decl)
{
   const char* name = decl->getName();
   PiList* look = (PiList *) idTable.Remove((char *) name);
   
   // check for multiply defined definitions
   PiList* declPtr = look;
   while (declPtr != 0)
   {    
      Decl* dec = (Decl*) declPtr->first();
      if (dec->getType()->canBe(decl->getType()) &&
	   decl->getType()->canBe(dec->getType()))
	return(dec);
      declPtr = declPtr->rest();
   }
   idTable.Insert(new(this, "PiList") PiList((Decl*) decl, look),(char *) name);
   return(0);
}

/*****************************************************************************/
const Decl* TypeScope::findIdent(const char* name, const Type* type,
	Bool canSeePrivate) const {

    const Decl* ret = findIdent_(name, type, canSeePrivate);
    const PiList* mergePtr = merges;
    while ((ret == 0 || ret == privateDecl()) && mergePtr != 0) {
        TypeScope* merge = (TypeScope *) mergePtr->first();
	    // import NEVER sees private
        const Decl* mergeRet = merge->findIdent(name, type, FALSE); 
	if (mergeRet != 0)
	   ret = mergeRet;
        mergePtr = mergePtr->rest();
        }
    return(ret);
}

/*****************************************************************************/
const Decl* TypeScope::findIdent_(const char* name, const Type* type,
				  Bool canSeePrivate) const
{   
    PiList* look = (PiList *) ((SpHashTree *)&idTable)->Find((char *) name);
    if (look == 0) return 0;

    Decl* ret = 0;
    Decl* evenPrivate = 0;
    PiList* declPtr = look;
    while (declPtr != 0) {
        Decl* decl = (Decl*) declPtr->first();
	if (decl->getType()->canBe(type)) {
	    evenPrivate = decl;
	    if (canSeePrivate || decl->getVisibility() == Decl::publicVis) {
		if (ret == 0) 
		    ret = decl;
		else {
		    if (ret->getType()->canBe(decl->getType()))
			ret = decl;
		    else if (!decl->getType()->canBe(ret->getType())) {
			    // at this point we have found two incompatable
			    // types which match 'type'  We need to see if
			    // there is a third match which is better than
			    // either of the two current matches.
			PiList* bestPtr = declPtr->rest();
			for(;;) {
			    if (bestPtr == 0) 
				return(ambiguousDecl()); // no better match
			    Decl* bestDecl = (Decl*) bestPtr->first();
			    if (bestDecl->getType()->canBe(type) &&
				ret->getType()->canBe(bestDecl->getType()) &&
				decl->getType()->canBe(bestDecl->getType())) {
				ret = bestDecl;
				break;
				}
			    bestPtr = bestPtr->rest();
			    }
			}
		    }
		}
	    }
        declPtr = declPtr->rest();
        }
    if (ret == 0 && evenPrivate != 0)	// was it sucess if privates included?
	return(privateDecl());
    return(ret);
}

/*****************************************************************************/
void TypeEnv::destroy() {

    if (superEnv != 0) 
	((TypeEnv*) superEnv)->destroy();
    delete this;
}

/*****************************************************************************/
TypeEnv* TypeEnv::copy(void* where) const {
	// FIX should we be copying work and time?
   TypeEnv* env = 0;
   if (superEnv != 0)
	env = superEnv->copy(where);
   return(new(where, "TypeEnv") TypeEnv(thisScope, env, canSeePrivate));
}

/*****************************************************************************/
const ObjectType* TypeEnv::getSelfType() const {

    const ObjectType* ret = 0;
    if (thisScope != 0)
	ret = thisScope->getSelfType();
    if (ret == 0 && superEnv != 0)
	ret = superEnv->getSelfType();
    return(ret);
}

/*****************************************************************************/
const Type* TypeEnv::getRetType() const {

    const static AtomicType aVoidType(FunType::voidType);
    const Type* ret = 0;
    if (thisScope != 0)
	ret = thisScope->getRetType();
    if (ret == 0 && superEnv != 0)
	ret = superEnv->getRetType();
    if (ret == 0)
	ret = &aVoidType;
    return(ret);
}

/*****************************************************************************/
const Decl* TypeEnv::addDecl(const Decl* decl) {

    if (thisScope != 0) 
	return(thisScope->addDecl(decl));
    else if (superEnv != 0)
	return(((TypeEnv*) superEnv)->addDecl(decl));
    else	// cant add to environment
	return((const Decl*) 1);	// FIX
}

/*****************************************************************************/
const Decl* TypeEnv::findIdent(const char* name, const Type* type) const {

    const Decl* ret = 0;
    if (thisScope != 0)
	ret = thisScope->findIdent(name, type, canSeePrivate);
    if ((ret == 0 || ret == TypeScope::privateDecl()) && superEnv != 0) {
	const Decl* subRet = superEnv->findIdent(name, type);
	if (subRet != 0)
	   ret = subRet;
	}
    return(ret);
}

#include "IDFormat.H"
#include "IndentFormat.H"

/*****************************************************************************/
/* helper function needed to print all the entries in the table */

/*
static void* printPiList(const char* key, const PiList* value, void* context) {
    Formatter& out = *((Formatter*) context);

    PiList* declPtr = (PiList*) value;
    IDFormat format;
    while (declPtr != 0) {
        Decl* decl = (Decl*) declPtr->first();
	format.formatDeclHead(decl, out);
        declPtr = declPtr->rest();
        }
    return(context);
}
*/

/*****************************************************************************/
void TypeEnv::print(ostream& out) const {

    if (thisScope != 0) {
	IndentFormatter fout(out, 80);
	thisScope->print(fout);
	}
    if (superEnv != 0)
	superEnv->print(out);
}

/*****************************************************************************/
void TypeScope::print(ostream& out) const {

    IndentFormatter fout(out, 80);
     print(fout);
}

/*****************************************************************************/
/* this function is meant only for ease of calling from gdb */

void print(const TypeScope* scope, ostream& out) {
    scope->print(out);	
}

/*****************************************************************************/
void TypeScope::print(Formatter& out) const {

    out << "BEGIN SCOPE {\n";
    IDFormat format;
    out.shiftIn();
    // JBco:    ((TreeTableObj<char, PiList>&) idTable).forAll(printPiList, (Formatter*) &out);
    out.shiftOut();
    const PiList* mergePtr = merges;
    while (mergePtr != 0) {
        TypeScope* merge = (TypeScope *) mergePtr->first();
	out << "IMPORTING\n";
	out.shiftIn();
	merge->print(out);
	out.shiftOut();
        mergePtr = mergePtr->rest();
        }

    if (retType != 0 && retType->isKind(Type::voidType)) {
	out << "    RETURN = ";
	format.formatType(retType, out);
	out << "\n";
	}
    if (selfType != 0) {
	out << "    SELF = ";
	format.formatType(selfType, out);
	out << "\n";
	}
    out << "} END SCOPE\n";
}
