/*****************************************************************************/
/*                               TypeEnv.h                                   */
/*****************************************************************************/

/* TypeEnv defines a `TypeEnv' which defines an environment that allows
   the types of variables to be looked up */

/* AUTHOR: Vance Morrison
   DATE:  11/2/93              */
/*****************************************************************************/

#ifndef _TypeEnv_h
#define _TypeEnv_h

#include "IDAST.H"
#include "SpHashTree.H"
#include "SpString.H"
#include "SpIostream.H"

class TypeScope;
class Formatter;

/*****************************************************************************/
/* a scope allows quick lookup of variable declaration as well as information
   on the current object and return type */

class TypeScope {
public:
    static const Decl* ambiguousDecl() { return((const Decl*) 1); }
    static const Decl* privateDecl()   { return((const Decl*) 2); }

    TypeScope() : refs(1),retType(0), selfType(0), idTable(1234), merges(0) {}
    TypeScope(const Type* aRetType) : refs(1), retType(aRetType), selfType(0), 
	idTable(1234), merges(0) {}
    TypeScope(const ObjectType* aSelfType) : refs(1), retType(0), 
	selfType(aSelfType), idTable(1234), merges(0) {}
    ~TypeScope();

    void merge(const TypeScope* scope);
    const Decl*  addDecl(const Decl* decl);
    const Decl* findIdent(const char* name, const Type* type=0, 
			  Bool canSeePrivate=FALSE) const;
    const ObjectType* getSelfType() const 	{ return(selfType); }
    const Type* getRetType() const		{ return(retType); }
   void clear()
      { // JB: Plug this leak
	 idTable.Reset();
      }
    void print(ostream& out) const;
    void print(Formatter& out) const;
    void link()	const				{ ((TypeScope*) this)->refs++;}
    void unlink()				{ --refs; if (refs <= 0) 
						  delete this; }
private:
    const Decl* findIdent_(const char* name, const Type* type=0,
			   Bool canSeePrivate=FALSE) const;
private:
    int refs;		// reference count
    const Type* retType;
    const ObjectType* selfType;
    SpHashTree idTable;
    PiList* merges;	// of const TypeScope (to be merged with this one)
};

/*****************************************************************************/
/* a TypeEnv is really just a list (stack) of scopes */

class TypeEnv {
public:
    TypeEnv(TypeScope *aScope, const TypeEnv* aEnv, Bool aCanSeePrivate=TRUE)
        : thisScope(aScope),  superEnv(aEnv), canSeePrivate(aCanSeePrivate) 
	{ aScope->link(); }
    ~TypeEnv() { if (thisScope != 0) thisScope->unlink(); }

    TypeEnv* copy(void* where) const;
    void destroy();
    const Decl* addDecl(const Decl* decl);
    const Decl* findIdent(const char* name, const Type* type=0) const;
    const ObjectType* getSelfType() const;
    const Type* getRetType() const;
    void print(ostream& out) const;
    void merge(const TypeScope* scope) { thisScope->merge(scope); }

private:
    TypeScope* thisScope;       // 0 means no scope associated with this
    Bool canSeePrivate;		// can we see private variables in 'thisScope'
    const TypeEnv* superEnv;
};

#endif
