/*****************************************************************************/
/*                               TypeStatus.cxx                              */
/*****************************************************************************/

/* TypeStatus defines a `TypeStatus' which defines a place to put error messages
   when code is typechecked */

/* AUTHOR: Vance Morrison
   DATE:  11/2/93              */
/*****************************************************************************/

#include "SpIostream.H"
#include "TypeStatus.H"
#include "TypeEnv.H"
#include "IDFormat.H"
#include "IndentFormat.H"
#include "SpStrstream.H"

/*****************************************************************************/
TypeStatus::TypeStatus() : error(0), at(0), env(0), out(*new ostrstream) {

}

/*****************************************************************************/
TypeStatus::~TypeStatus() {
   if (env != 0)
	env->destroy();
   delete (ostrstream*) &out;
}

/*****************************************************************************/
char* TypeStatus::msg() {
   out << (char) '\0'; 	// terminate the string (I thing str() should do this)
   return(out.str()); 
}

/*****************************************************************************/
void TypeStatus::setIdentAmbiguous(const void* where, const Ident*name,
    const Type*type, const TypeEnv* aEnv){
    error = 5;  at = where; env = aEnv->copy(this);

    out << "Identifier '";
    print(name, out); out << "' of type '";
    if (type != 0)
	print(type, out);
    else
	out << "UNKNOWN";
     out << "' is ambiguous"; 
}

/*****************************************************************************/
void TypeStatus::setIdentNotFound(const void*where,const Ident* name,
    const Type*type, const TypeEnv* aEnv){
    error = 6;  at = where; env = aEnv->copy(this);

    out << "Identifier '";
    print(name, out); out << "' of type '";
    if (type != 0)
	print(type, out);
    else
	out << "UNKNOWN";
    out << "' not found"; 
}

/*****************************************************************************/
void TypeStatus::setSelfNotFound(const void* where) {
    error = 7;  at = where; out << "SELF not defined in this context"; 
}

/*****************************************************************************/
void TypeStatus::setIdMultDefined(const void* where, const char* name,
    const Type* type) {
    error = 8;  at = where; 

    out << "Identifier '" << name << "' of type '";
    if (type != 0)
	print(type, out);
    else
	out << "UNKNOWN";
    out << "' already defined"; 
}

/*****************************************************************************/
void TypeStatus::setNotIndexAble(const void* where, const Type* type) {
    error = 10;  at = where; 

    out << "Type '";
    print(type, out); 
    out << "' not indexable"; 
}

/*****************************************************************************/
void TypeStatus::setTypeConflict(const void* where, const Type* actualType, 
    const Type* neededType) {
    error = 11;  
    at = where; 

    out << "Type mismatch need type '";
    print(neededType, out);
    out << "' but got type '";
    print(actualType, out);
    out << "'";
}

/*****************************************************************************/
void TypeStatus::setNeedRetVal(const Stmt* where, const Type* retType) {
    error = 12;  at = where; 

    out << "Method needs a return value of type ";
    print(retType, out); 
}

/*****************************************************************************/
void TypeStatus::setBadRetExp(const void* where) {
    error = 13;  at = where; 

    out << "No return value is needed"; 
}

/*****************************************************************************/
void TypeStatus::setNotConstant(const void* where) {
    error = 14;  at = where; 

    out << "Expression not a constant"; 
}

/*****************************************************************************/
void TypeStatus::setNotLvalue(const void* where) {
    error = 15;  at = where; 

    out << "Expression not a assignable value"; 
}

/*****************************************************************************/
void TypeStatus::setNotArray(const void* where) {
    error = 16;  at = where; 

    out << "Expression not an array\n"; 
}

/*****************************************************************************/
void TypeStatus::setArgNotExpected(const void* where) {
    error = 17;  at = where; 

    out << "Argument not expected\n"; 
}

/*****************************************************************************/
void TypeStatus::setNotObject(const void* where) {
    error = 19;  at = where; 

    out << "Expression not an object type"; 
}

/*****************************************************************************/
void TypeStatus::setNotVar(const void* where) {
    error = 20;  at = where; 

    out << "Expression must be a variable"; 
}

/*****************************************************************************/
void TypeStatus::setNeedValue(const void* where) {
    error = 21;  at = where; 

    out << "Expression can not be of void type"; 
}

/*****************************************************************************/
void TypeStatus::setDeferedMeths(const void* where, PiList* meths) {
    error = 22;  
    at = where; 

    out << "Object can not be created because the methods '";
    PiList* declPtr = meths;
    while (declPtr != 0) {
        Decl* decl = (Decl*) declPtr->first();
	out << decl->getName();
        declPtr = declPtr->rest();
	if (declPtr != 0) 	
	    out << ", ";
        }
   out << "' are abstract";
}

/*****************************************************************************/
void TypeStatus::setDomeIllegal(const void* where) {
    error = 23;  at = where; 

    out <<  "Dome Variable is Illegal here"; 
}

/*****************************************************************************/
void TypeStatus::setVarIllegal(const void* where) {
    error = 24;  at = where; 

    out <<  "Module level (global) Variables are illegal"; 
}

/*****************************************************************************/
void TypeStatus::setBadNewArgs(const void* where, const Type* newType) {
    error = 26;  
    at = where; 
    
    out << "New operator with args  '";
    print(newType,  out);

    out << "' not found";
}

/*****************************************************************************/
void TypeStatus::setBadNumArgs(const void* where) {
    error = 27;  at = where; 
    out <<  "Wrong number of arguments"; 
}

/*****************************************************************************/
void TypeStatus::setBadOut(const void* where, int i) {
    error = 28;  at = where; 

    out << "Automatic conversion of OUT parameters is not permitted" <<
    " (for parameter " << i << ")";
}

/*****************************************************************************/
void TypeStatus::setConstrNotFound(const void* where, const Type* type,
    const Type* constrType, const TypeEnv* aEnv) {
    error = 29;  at = where; env = aEnv->copy(this);

    out << "Constructor for type '";
    print(type, out); 
    out << "' of type '";
    print(constrType, out); 
    out << "' not found"; 
}

/*****************************************************************************/
void TypeStatus::setConstrAmbiguous(const void* where, const Type* type,
    const Type* constrType, const TypeEnv* aEnv) {
    error = 29;  at = where; env = aEnv->copy(this);

    out << "Constructor for type '";
    print(type, out); 
    out << "' of type '";
    print(constrType, out); 
    out << "' is ambiguous"; 
}

/*****************************************************************************/
void TypeStatus::setBadInstance(const void* where, const char* name) {

    error = 30;  at = where; 

    out << "\nError Instanitating association '" << name << "'";
}

/*****************************************************************************/
void TypeStatus::setRetNotExpected(const void* where) {
    error = 31;  at = where; 

    out << "Return types are not allowed on constructors and destructors"; 
}

/*****************************************************************************/
void TypeStatus::setPublicVarIllegal(const void* where) {
    error = 32;  at = where; 

    out <<  "Instance Variables may not be made public"; 
}

/*****************************************************************************/
void TypeStatus::setConstrPrivate(const void* where, const Type* type,
    const Type* constrType, const TypeEnv* aEnv) {
    error = 33;  at = where; env = aEnv->copy(this);

    out << "Constructor for type '";
    print(type, out); 
    out << "' of type '";
    print(constrType, out); 
    out << "' only matches PRIVATE declarations\n";
}

/*****************************************************************************/
void TypeStatus::setIdentPrivate(const void* where, const Ident*name,
    const Type*type, const TypeEnv* aEnv){
    error = 34;  at = where; env = aEnv->copy(this);

    out << "Identifier '";
    print(name, out); out << "' of type '";
    if (type != 0)
	print(type, out);
    else
	out << "UNKNOWN";
    out << "' only matches PRIVATE declarations\n";
}

