// This file must be included AFTER the declaration of a variable
// named container, whose type is a container class which has
// functions such as Insert, and Remove, etc.  Also optionally define
// __ITERATE_IN_ORDER__.  See the other files in this directory for
// examples.

// This test works by keeping an array of key value pairs.  Whenever
// an operation is performed on the container we are testing, the same
// operation is performed on the array.  Of course using an array has
// O(n), whereas the container we are testing probably has more like
// O(log(n)).  So this array system is very slow, but that is OK since
// it is only in test code.  The test does a container lookup every
// once in a while, and it knows what to expect since it has kept
// track of what is in the container.  It keeps track of that in the
// array I mentioned.

#include "SpIostream.H"
#include "SpMath.H"
#include "SpAssert.H"
#include "SpRandom.H"
#include "SpPoDataBase.H"
#ifndef __SPEEDES_NT__
#include <values.h>
#else
#define INTBITS 32
#define MININT      (1 << (INTBITS - 1))
#define MAXINT      (~MININT)
#endif

#define db_var(var) cout << (#var) << " " << setprecision(40) << (var) << endl;

SpRandom Random;

double Max(double x, double y) { return x > y ? x : y; }

int within_round_off_precision(double x, double y, double epsilon) {
   int result;
   double difference = fabs(x - y);
   if (x == 0 || y == 0)
     result = difference <= epsilon;
   else
     result = difference <= epsilon * Max(fabs(x), fabs(y));
   if (!result)
   {
     cout << "within_round_off_precision found differecnes:\n";
     db_var(x);
     db_var(y);
   }
   return result;
}

struct KeyValue
{
    double key;
    void *value;
};

ostream &operator<<(ostream &os, const KeyValue &kv) {
  os << "key = " << kv.key << ", " << "value = " << kv.value;
  return os;
}

int operator==(const KeyValue &a, const KeyValue &b) {
  return a.value == b.value
    && within_round_off_precision(a.key, b.key, 1e-30)
    && a.key == b.key;
}

const int MaxNum = 2000;
KeyValue KeyValueArray[MaxNum];
int CurrentNum = 0;

int seen_p[MaxNum], order[MaxNum];

double LargeNumber;

int GetFirstIndexOfARandomKey()
{
  const int random_index = Random.GenerateInt(0, CurrentNum-1);
  assert(random_index < CurrentNum);
  assert(random_index >= 0);
  const double random_key = KeyValueArray[random_index].key;
  for (int first_index_of_a_random_key = 0;
       first_index_of_a_random_key < CurrentNum;
       first_index_of_a_random_key++)
  {
    // Should I call within_round_off_precision here?
    if (KeyValueArray[first_index_of_a_random_key].key == random_key)
      return first_index_of_a_random_key;
  }
  assert(0);
  return -1;
}

void Add()
{
  assert(CurrentNum < MaxNum);

  double key;
  void *value;
  if (CurrentNum == 0 || Random.GenerateDouble(0, 1) < .5)
  {
    double m;
    if (Random.GenerateDouble(0, 1) < .5)
      m = LargeNumber;
    else
      m = 260;
    key = Random.GenerateDouble(-m, m);
    value = (void*) Random.GenerateInt(0, 20000);
  }
  else
  {
    const int index = Random.GenerateInt(0, CurrentNum-1);
    assert(index >= 0);
    assert(index < CurrentNum);
    key = KeyValueArray[index].key;
    if (Random.GenerateDouble(0, 1) < .5)
      value = KeyValueArray[index].value;
    else
      value = (void*) Random.GenerateInt(0, 20000);
  }
  
  container.Insert(value, key);

  int j;
  for (j = CurrentNum; j > 0 && key < KeyValueArray[j - 1].key; j--)
    KeyValueArray[j] = KeyValueArray[j - 1];
  
  KeyValueArray[j].value = value;
  KeyValueArray[j].key   = key;

#ifdef debug_RB_Containers2
  cout << "Adding " << KeyValueArray[j] << endl;
#endif

  ++CurrentNum;
}

void Remove()
{
  assert(CurrentNum >= 1);
  const int index_to_remove = GetFirstIndexOfARandomKey();
#ifdef debug_RB_Containers2
  cout << "Removing " << KeyValueArray[index_to_remove] << endl;
#endif
  void *value = container.Remove(KeyValueArray[index_to_remove].key);
  if (value != KeyValueArray[index_to_remove].value)
  {
    cerr << "The tree returned a value of " << value << endl;
    assert(0);
  }
  --CurrentNum;
  for (int i = index_to_remove; i < CurrentNum; i++)
    KeyValueArray[i] = KeyValueArray[i+1];
}

void TestIterator()
{
  // Test the built-in iterator, and also see if there is a key and
  // element in the container that should NOT be there or if it does
  // not contain a key and element which it should contain.

  int i;

  for (i = 0; i < CurrentNum; i++)
    seen_p[i] = 0;
  
#ifdef debug_RB_Containers2
  cout << "   Iterating in increasing order" << endl;
#endif

  int num_container_iterations = 0;
  void *value;
  for (value = container.GetFirstElement();
       value;
       value = container.GetNextElement())
  {
    assert(value == container.GetCurrentElement());
    KeyValue got;
    got.key   = container.GetCurrentKey();
    got.value = container.GetCurrentElement();
      
#ifdef debug_RB_Containers2
    cout << "      Currently iterating on " << got << endl;
#endif
    
    for (i = 0; i < CurrentNum; i++)
      if (!seen_p[i] && got == KeyValueArray[i])
	goto next_container_iteration;

    cerr << "I got " << got
	 << " but I think that it should NOT be in the container.";
    assert(0);

    next_container_iteration:
    seen_p[i] = 1;
#ifdef __ITERATE_IN_ORDER__
    assert(i == num_container_iterations);
#endif
    assert(num_container_iterations < CurrentNum);
    order[num_container_iterations++] = i;
  }
  assert(num_container_iterations == CurrentNum);

#ifdef debug_RB_Containers2
  cout << "   Iterating in decreasing order" << endl;
#endif

  // Now that the above code tested iterating forwards, we now test
  // iterating backwards in the code below.  Also test that the order
  // is the exact oposite from if you iterate forwards.
  i = CurrentNum;
  for (value = container.GetLastElement();
       value;
       value = container.GetPreviousElement())
  {
    --i;
    assert(i >= 0);

    KeyValue got;
    got.key   = container.GetCurrentKey();
    got.value = container.GetCurrentElement();
    
    assert(KeyValueArray[order[i]] == got);
  }
  assert(i == 0);
}

void Lookup()
{
  assert(CurrentNum == container.GetNumElements());

  container.CheckIntegrity();
  
  // Test what happens when we look for a key that is not there:
  double key;
  choose_another_key:
  key = Random.GenerateDouble(-LargeNumber, LargeNumber);
  for (int i = 0; i < CurrentNum; i++)
    if (KeyValueArray[i].key == key)
      goto choose_another_key;
  assert(container.Find(key) == NULL);

  // Test what happens when we look for a key that IS there:
  if (CurrentNum >= 1)
  {
#ifdef debug_RB_Containers2
    cout << "Beginning FindNext test" << endl;
#endif
    int array_index = GetFirstIndexOfARandomKey();
    const double key = KeyValueArray[array_index].key;
    for(void *iterator_return_value = container.Find(key);
	iterator_return_value;
	iterator_return_value = container.FindNext())
    {
      KeyValue got;
      got.key   = container.GetCurrentKey();
      got.value = container.GetCurrentElement();
#ifdef debug_RB_Containers2
      db_var(got);
#endif
      
      assert(got.key   == key);
      assert(got.value == iterator_return_value);
	
      assert(array_index < CurrentNum);
      assert(got == KeyValueArray[array_index++]);
    }
    if (array_index < CurrentNum && key == KeyValueArray[array_index].key)
    {
      db_var(key);
      db_var(KeyValueArray[array_index].key);
      assert(0);
    }
  }
}    

void Test()
{
  for (int i = 0; i < 50; i++)
    container.Insert(((void*)(100+i)), 1);
  assert(((void*)100) == container.Find(1));

  container.Reset();

#ifdef __INSURE__
  const int maxNumItemsToAdd = 40;
  const float probabilityOfAddingToHiGrid = .8;
  const int numCyclesToTest = 3;
#else
  const int maxNumItemsToAdd = 300;
  const float probabilityOfAddingToHiGrid = .8;
  const int numCyclesToTest = 4;
#endif
  
  const float
    probability_of_removing_from_HiGrid = 1 - probabilityOfAddingToHiGrid;
  float current_P_add = probabilityOfAddingToHiGrid;
  int current_num_cycles = 0, current_target = maxNumItemsToAdd;
  int test_reset_p = 0;
  int num_resets_tested = 0;
  for (;;)
  {
    switch (Random.GenerateInt(1, 3)) {
      case 1:
	container.SetNormalTreeMode();
	break;
      case 2:
	container.SetBalancedTreeMode();
	break;
      case 3:
	container.SetSplayTreeMode();
	break;
      default:
	assert(0);
    }
    
    if (!test_reset_p &&
	Random.GenerateDouble(0, 1) < (1.0 / maxNumItemsToAdd))
    {
#ifdef debug_RB_Containers2
      cout << "Resetting." << '\n';
#endif
      container.Reset();
      CurrentNum = 0;
      
      test_reset_p = 1;
      ++num_resets_tested;
    }

    if (Random.GenerateDouble(0, 1) < .3)
      TestIterator();

    Lookup();

    if (Random.GenerateDouble(0, 1) <= current_P_add || CurrentNum == 0)
      Add();
    else
      Remove();
    
    if (CurrentNum == current_target)
    {
      if (current_target == 0)
      {
	if ((++current_num_cycles) == numCyclesToTest)
	  break;
	current_target = maxNumItemsToAdd;
	test_reset_p = 0;
	current_P_add = probabilityOfAddingToHiGrid;
      }
      else
      {
	current_target = 0;
	current_P_add = probability_of_removing_from_HiGrid;
      }
    }
  }
  assert(num_resets_tested >= numCyclesToTest / 2);
}

int main()
{
  extern SpPoDataBase* DefaultDataBase;
  DefaultDataBase = new SpPoDataBase;
  LargeNumber = MAXDOUBLE / 10.0;

  Test();
  
  return 0;
}
