// TestRandom.C file tests the random number generator

#include "SpIostream.H"
#include "Random.H"

void PrintTestStatus(char *testName, int errorFlag) {
  if (errorFlag) {
    cout << "***Failed: " << testName << endl;
  }else{
    cout << "   Passed: " << testName << endl;
  }
}

void TestRandom(char *randomType, SpBaseRandom *random) {
  int errorFlag;
  int i;
  double Umin;
  double Umax;
  double U;
  double loDouble;
  double hiDouble;
  int Imin;
  int Imax;
  int I;
  int loInt;
  int hiInt;
  int B;
  int B0;
  int B1;
  double timeConstant;
  double Sum;
  double Sum2;
  double Num;
  double Average;
  double StandardDeviation;
  double mean;
  double sigma;
  double V[3];
  double Mag;
  int n1;
  int n2;

  cout << "Testing: " << randomType << endl;

/*
** Test genereration of uniform random numbers to make sure that values are
** between [0,1]. An error occurs if a number is out of range or if the
** minimum and maximum numbers are not close enough to [0,1].
*/

  errorFlag = 0;
  Umin = 1.0;
  Umax = 0.0;
  Sum = 0;
  Num = 10000;

  for (i=0; i<Num; i++) {
    U = random->GenerateUniform();
    if (U < Umin) Umin = U;
    if (U > Umax) Umax = U;
    Sum += U;
  }

  if (Umin < 0.0) {
    cout << "Error, GenerateUniform Negative Number " << Umin << endl;
    errorFlag = 1;
  }

  if (0.01 < Umin) {
    cout << "Error, GenerateUniform Bad Range " << Umin << endl;
    errorFlag = 1;
  }

  if (1.0 < Umax) {
    cout << "Error, GenerateUniform Greater Than One " << Umax << endl;
    errorFlag = 1;
  }

  if (Umax < 0.99) {
    cout << "Error, GenerateUniform Bad Range " << Umax << endl;
    errorFlag = 1;
  }

  Average = Sum/Num;
  if (0.01 < fabs(Average-0.5)) {
    cout << "Error, GenerateUniform Bad Average " << Average << endl;
    errorFlag = 1;
  }

  PrintTestStatus("Uniform Distribution", errorFlag);

/*
** Test generation of bounded random doubles. This test is exactly like
** the uniform random number test except that the bounds are not [0,1].
*/

  errorFlag = 0;
  Umax = loDouble = 2.7;
  Umin = hiDouble = 3.14;
  Sum = 0;
  Num = 10000;

  for (i=0; i<Num; i++) {
    U = random->GenerateDouble(loDouble, hiDouble);
    if (U < Umin) Umin = U;
    if (U > Umax) Umax = U;
    Sum += U;
  }

  if (Umin < loDouble) {
    cout << "Error, GenerateDouble Low Bound " << Umin << endl;
    errorFlag = 1;
  }

  if (0.01 < (Umin-loDouble)/(hiDouble-loDouble)) {
    cout << "Error, GenerateDouble Bad Range " << Umin << endl;
    errorFlag = 1;
  }

  if (hiDouble < Umax) {
    cout << "Error, GenerateUniform Hi Bound " << Umax << endl;
    errorFlag = 1;
  }

  if (0.01 < (hiDouble-Umax)/(hiDouble-loDouble)) {
    cout << "Error, GenerateDouble Bad Range " << Umax << endl;
    errorFlag = 1;
  }

  Average = Sum/Num;
  if (0.01 < fabs(Average - ((hiDouble-loDouble)/2 + loDouble))) {
    cout << "Error, GenerateDouble Bad Average " << Average << endl;
    errorFlag = 1;
  }

  PrintTestStatus("Double Bounded Distribution", errorFlag);


/*
** Test generation of bounded random integers. This test is exactly like
** the bounded random doubles test except that the bounds are integers.
*/

  errorFlag = 0;
  Imax = loInt = 28;
  Imin = hiInt = 35;
  Sum = 0;
  Num = 10000;

  for (i=0; i<Num; i++) {
    I = random->GenerateInt(loInt, hiInt);
    Sum += I;
    if (I < Imin) Imin = I;
    if (I > Imax) Imax = I;
  }

  if (Imin < loInt) {
    cout << "Error, GenerateInt Low Bound " << Imin << endl;
    errorFlag = 1;
  }

  if (0.01 < double(Imin-loInt) / double(hiInt-loInt)) {
    cout << "Error, GenerateInt Bad Range " << Imin << endl;
    errorFlag = 1;
  }

  if (hiInt < Imax) {
    cout << "Error, GenerateInt Hi Bound " << Imax << endl;
    errorFlag = 1;
  }

  if (0.01 < double(hiInt-Imax) / double(hiInt-loInt)) {
    cout << "Error, GenerateInt Bad Range " << Umax << endl;
    errorFlag = 1;
  }

  Average = Sum/Num;
  if (0.1 < (fabs(Average - (double(hiInt-loInt)/2.0 + loInt)))
      / double(hiInt-loInt)) {
    cout << "Error, GenerateInt Bad Average " << Average << endl;
    errorFlag = 1;
  }

  PrintTestStatus("Int Bounded Distribution", errorFlag);

/*
** Test generation of random bits. This test "flips coins" and verifies
** that the number of 0's is roughly equal to the number of 1's..
*/

  errorFlag = 0;
  B0 = 0;
  B1 = 0;

  for (i=0; i<50000; i++) {
    B = random->GenerateBit();
    if ((B < 0) || (B > 1)) {
      cout << "Error, GenerateBit is bad" << B << endl;
      errorFlag = 1;
    }
    if (B) {
      B1++;
    }else{
      B0++;
    }
  }

  if (0.01 < fabs(double(B1-B0))/ double(B1+B0)) {
    cout << "Error, GenerateBit Biased " << B0 << " " << B1 << endl;
    errorFlag = 1;
  }

  PrintTestStatus("Bit Distribution", errorFlag);

/*
** Test generation of random exponential distributions. This test generates
** a number of exponential random numbers and makes sure that the statistics
** are correct..
*/

  errorFlag = 0;
  Sum = 0;
  Num = 10000;
  timeConstant = 5.0;

  for (i=0; i<Num; i++) {
    U = random->GenerateExponential(timeConstant);
    if (U < 0) {
      cout << "Error, GenerateExponential Is Negative" << U << endl;
      errorFlag = 1;
    }
    Sum += U;
  }

  Average = Sum / Num;

  if (0.1 < fabs(Average-timeConstant)/timeConstant) {
    cout << "Error, GenerateExponential Has Bad Average " << Average << endl;
    errorFlag = 1;
  }

  PrintTestStatus("Exponential Distribution", errorFlag);

/*
** Test generation of random gaussian distributions. This test generates
** a number of gaussian random numbers and makes sure that the statistics
** are correct..
*/

  errorFlag = 0;
  Sum = 0;
  Sum2 = 0;
  Num = 10000;
  timeConstant = 5.0;
  mean = 4.6;
  sigma = 1.4;

  for (i=0; i<Num; i++) {
    U = random->GenerateGaussian(mean, sigma);
    Sum += U;
    Sum2 += U*U;
  }

  Average = Sum / Num;
  StandardDeviation = sqrt((Sum2 - Average*Sum)/Num);

  if (0.1 < fabs(Average-mean)) {
    cout << "Error, GenerateGaussian Has Bad Average " << Average << endl;
    errorFlag = 1;
  }

  if (0.1 < fabs(StandardDeviation-sigma)) {
    cout << "Error, GenerateGaussian Has Bad Sigma " << StandardDeviation
	 << endl;
    errorFlag = 1;
  }

  PrintTestStatus("Gaussian Distribution", errorFlag);

/*
** Test generation of random vectors.
*/

  errorFlag = 0;
  Sum = 0;
  Sum2 = 0;
  Num = 10000;
  mean = 13.3;
  sigma = 1.5;
  for (i=0; i<Num; i++) {
    random->GenerateVector(V, mean, sigma);
    Mag = sqrt(V[0]*V[0]+V[1]*V[1]+V[2]*V[2]);
    Sum += Mag;
    Sum2 += Mag*Mag;
  }

  Average = Sum / Num;
  StandardDeviation = sqrt((Sum2 - Average*Sum)/Num);

  if (0.1 < fabs(Average-mean)) {
    cout << "Error, GenerateVector Has Bad Average " << Average << endl;
    errorFlag = 1;
  }

  if (0.1 < fabs(StandardDeviation-sigma)) {
    cout << "Error, GenerateVector Has Bad Sigma " << StandardDeviation
	 << endl;
    errorFlag = 1;
  }

  PrintTestStatus("Vector", errorFlag);

/*
** Test generation of random beta distribution.(Flat)
*/

  errorFlag = 0;
  Sum = 0;
  Sum2 = 0;
  Num = 10000;
  n1 = 0;
  n2 = 0;
  for (i=0; i<Num; i++) {
    U = random->GenerateBeta(n1,n2);
    Sum += U;
    Sum2 += U*U;
  }

  Average = Sum / Num;
  StandardDeviation = sqrt((Sum2 - Average*Sum)/Num);

  if (0.1 < fabs(Average-0.5)) {
    cout << "Error, GenerateBeta Has Bad Average " << Average << endl;
    errorFlag = 1;
  }

  if (0.1 < fabs(StandardDeviation-1/sqrt(12.0))) {
    cout << "Error, GenerateBeta Has Bad Sigma " << StandardDeviation
	 << endl;
    errorFlag = 1;
  }

  PrintTestStatus("Beta Flat(0,0)", errorFlag);

/*
** Test generation of random beta distribution.(Bell)
*/

  errorFlag = 0;
  Sum = 0;
  Sum2 = 0;
  Num = 10000;
  n1 = 10;
  n2 = 10;
  for (i=0; i<Num; i++) {
    U = random->GenerateBeta(n1,n2);
    Sum += U;
    Sum2 += U*U;
  }

  Average = Sum / Num;
  StandardDeviation = sqrt((Sum2 - Average*Sum)/Num);

  if (0.1 < fabs(Average-0.5)) {
    cout << "Error, GenerateBeta Has Bad Average " << Average << endl;
    errorFlag = 1;
  }

  if (0.1 < fabs(StandardDeviation-sqrt(1.0/(8.0*n1+12.0)))) {
    cout << "Error, GenerateBeta Has Bad Sigma " << StandardDeviation
	 << endl;
    errorFlag = 1;
  }

  PrintTestStatus("Beta Bell(10,10)", errorFlag);


/*
** Test Erf(x) method
*/

  errorFlag = 0;
  U = random->Erf(1.8);
  if (0.01 < fabs(U-0.4641)) {
    cout << "Error, Bad Erf " << U << endl;
    errorFlag = 1;
  }

  PrintTestStatus("Error Function", errorFlag);

/*
** Test InvErf(x) method
*/

  errorFlag = 0;
  U = random->InvErf(0.4641);
  if (0.01 < fabs(U-1.8)) {
    cout << "Error, Bad InvErf " << U << endl;
    errorFlag = 1;
  }

  PrintTestStatus("Inverse Error Function", errorFlag);

/*
** Test Factorial(x) method
*/

  errorFlag = 0;
  I = int(random->Factorial(5));
  if (I != 120) {
    cout << "Error, Bad Factorial " << I << endl;
    errorFlag = 1;
  }

  PrintTestStatus("Factorial", errorFlag);

/*
** Verify that the seed can be set and retrieved correctly
*/

  errorFlag = 0;
  (*random) = 23;

  if (*random != 23) {
    cout << "Error, bad = operator" << endl;
    errorFlag = 1;
  }

  (*random)++;
  if (*random != 24) {
    cout << "Error, bad postfix ++ operator" << endl;
    errorFlag = 1;
  }

  --(*random);
  if (*random != 23) {
    cout << "Error, bad prefix -- operator" << endl;
    errorFlag = 1;
  }

  I = (*random)--;
  if (I != 23) {
    cout << "Error, bad postfix -- operator" << endl;
    errorFlag = 1;
  }

  if (*random != 22) {
    cout << "Error, bad postfix -- operator" << endl;
    errorFlag = 1;
  }

  (*random) <<= 1;
  if (*random != 44) {
    cout << "Error, bad <<= operator" << endl;
    errorFlag = 1;
  }

  (*random) >>= 1;
  if (*random != 22) {
    cout << "Error, bad >>= operator" << endl;
    errorFlag = 1;
  }

  (*random) *= 3;
  if (*random != 66) {
    cout << "Error, bad *= operator" << endl;
    errorFlag = 1;
  }

  (*random) /= 2;
  if (*random != 33) {
    cout << "Error, bad /= operator" << endl;
    errorFlag = 1;
  }

  (*random) |= 2;
  if (*random != 35) {
    cout << "Error, bad |= operator" << endl;
    errorFlag = 1;
  }

  (*random) &= 2;
  if (*random != 2) {
    cout << "Error, bad &= operator" << endl;
    errorFlag = 1;
  }

  (*random) ^= 3;
  if (*random != 1) {
    cout << "Error, bad ^= operator" << endl;
    errorFlag = 1;
  }

  PrintTestStatus("Operator", errorFlag);

}

int main() {
  SpFastRandom fastRandom;
  SpRandom random;

  TestRandom("Fast Random", &fastRandom);
  TestRandom("Slow Random", &random);

  return 0;
}
