// SpDDMRegion.C

#include "S_SpHLA.H"
#include "SpDDMRegion.H"
#include "SpDDMDim.H"
#include "SpSpace.H"
#include "SpDDMSpace.H"
#include "RB_SpBinaryTree.H"
#include "RB_SpDefineClass.H"

RB_DEFINE_CLASS(SpDDMEnumDimension);
RB_DEFINE_CLASS(SpDDMDoubleDimension);
RB_DEFINE_CLASS(SpDDMRegion);
RB_DEFINE_CLASS(SpPublishRegion);
RB_DEFINE_CLASS(SpSubscribeRegion);

SpDDMRegion::SpDDMRegion(void)
  : Name((char *) 0), space((SpDDMSpace *) 0), Dimensions(new RB_SpBinaryTree) 
{
}

SpDDMRegion::SpDDMRegion(SpDDMSpace *_space, const char *name)
  : Name((char*) 0), space(_space), Dimensions(new RB_SpBinaryTree) {
  int len = strlen(name);
  Name = (len>0) ? new char [len+1] : (char *) 0;
  if (len>0) strcpy(Name, name);
}

SpDDMRegion::~SpDDMRegion() {
  SpDDMDim *dim;

  while ((dim = (SpDDMDim *) Dimensions->RemoveFirstElement())){
    switch (dim->GetType()){
      case SpDimLookupItem::Double:
        delete (SpDDMDoubleDimension *) dim;
        break;
      case SpDimLookupItem::Enum:
        delete (SpDDMEnumDimension *) dim;
        break;
      default:
        abort();
    }
  }
  delete Dimensions;
  if (Name!=(char *) 0) delete [] Name;
}

void
SpDDMRegion::DeleteDimension(SpDDMDim *dimension){

  switch (dimension->GetType()){
    case SpDimLookupItem::Double:
      RB_DELETE_SpDDMDoubleDimension((SpDDMDoubleDimension *) dimension);
      break;
    case SpDimLookupItem::Enum:
      RB_DELETE_SpDDMEnumDimension((SpDDMEnumDimension *) dimension);
      break;
    default:
      abort();
  }
}

void
SpDDMRegion::CreateDimension(const char *_dim_name, double _lo, double _hi, ...){
  va_list va;
  va_start(va, _hi);
  ModifyDimension(_dim_name, _lo, _hi, va);
  va_end(va);
}

void
SpDDMRegion::ModifyDimension(const char *_dim_name, double _lo, double _hi, ...){
  va_list va;
  va_start(va, _hi);
  ModifyDimension(_dim_name, _lo, _hi, va);
  va_end(va);
}

void
SpDDMRegion::ModifyDimension(const char *_dim_name, double _lo, double _hi, va_list &va){
  char *dim_name;
  double dim_hi, dim_lo, dim_res;

  dim_res = _hi - _lo;
  if (!UpdateDimension(_dim_name, _lo, _hi, dim_res))
    CreateDoubleDimension(_dim_name, _lo, _hi, dim_res);

  for (; (dim_name=va_arg(va, char*)) ;){
    dim_hi  = va_arg(va, double);
    dim_lo  = va_arg(va, double);
    dim_res = (dim_hi - dim_lo) / 2.0;
    if (!UpdateDimension(dim_name, dim_lo, dim_hi, dim_res))
      CreateDoubleDimension(dim_name, dim_lo, dim_hi, dim_res);
  }
}

void 
SpDDMRegion::CreateDimension(const char *_dim_name, const char *_enum_name, ...){
  va_list va;
  SpDDMEnumDimension *enum_dimension = CreateEnumDimension(_dim_name);
  
  va_start(va, _enum_name);
  enum_dimension->Add(_enum_name, va);
  va_end(va);
}

void 
SpDDMRegion::ModifyDimension(const char *_dim_name, const char *_enum_name, ...){
  va_list va;
  SpDDMEnumDimension *enum_dimension;

  enum_dimension = (SpDDMEnumDimension *) Dimensions->Find((char *) _dim_name);
  if (enum_dimension==(SpDDMEnumDimension *) 0) enum_dimension = CreateEnumDimension(_dim_name);

  va_start(va, _enum_name);
  enum_dimension->Add(_enum_name, va);
  va_end(va);
}

SpDDMEnumDimension * 
SpDDMRegion::CreateEnumDimension(const char *_dim_name){

  if (!ValidateEnumDimension(_dim_name)) abort();

  SpDDMDim *dimension = (SpDDMDim *) Dimensions->Remove((char *) _dim_name);
  if (dimension) DeleteDimension(dimension);

  SpDDMEnumDimension *enum_dimension = new SpDDMEnumDimension(_dim_name);
  enum_dimension = RB_NEW_SpDDMEnumDimension(enum_dimension);
  Dimensions->Insert(enum_dimension, (char *) enum_dimension->GetName());

  return (enum_dimension);
}

int
SpDDMRegion::ValidateEnumDimension(const char *_dim_name){
  int found;
  SpDimensionTypes type;
  int ok = 1;

  space->GetSpace()->FindDimensionType(_dim_name, found, type);

  if (!found){
    cerr << "UNKNOWN DIMENSION " << _dim_name << " IN SPACE " << space->GetName() << endl;
    ok = 0;
  }

  if (type!=SpDimLookupItem::Enum){
    cerr << "DIMENSION " << _dim_name << " IN SPACE " << space->GetName()
         << " is not a ENUM dimension." << endl;
    ok = 0;
  }

  return (ok);
}

int
SpDDMRegion::ValidateDoubleDimension(const char *_dim_name){
  int found;
  SpDimensionTypes type;
  int ok = 1;

  space->GetSpace()->FindDimensionType(_dim_name, found, type);

  if (!found){
    cerr << "UNKNOWN DIMENSION " << _dim_name << " IN SPACE " << space->GetName() << endl;
    ok = 0;
  }

  if (type!=SpDimLookupItem::Double){
    cerr << "DIMENSION " << _dim_name << " IN SPACE " << space->GetName() 
         << " is not a LO/HI/RES dimension." << endl;
    ok = 0;
  }
  return (ok);
}

void 
SpDDMRegion::DeleteDimension(const char *name){
  SpDDMDim *dimension = (SpDDMDim *) Dimensions->Remove((char *) name);

  if (dimension==(SpDDMDim *) 0) return;

  DeleteDimension(dimension);
}

void
SpDDMRegion::CreateDoubleDimension(const char *dim_name, double lo, double hi, double res){
  if (!ValidateDoubleDimension(dim_name)) abort();

  SpDDMDoubleDimension *double_dimension = new SpDDMDoubleDimension(dim_name, lo, hi, res);
  Dimensions->Insert(RB_NEW_SpDDMDoubleDimension(double_dimension), 
                    (char *) double_dimension->GetName());
}

int
SpDDMRegion::UpdateDimension(const char *dim_name, double lo, double hi, double res){
  SpDDMDoubleDimension *double_dimension;
  SpDDMDim *dimension = (SpDDMDim *) Dimensions->Find((char *) dim_name);

  if (dimension==(SpDDMDim *)0) return (0);

  if (dimension->GetType()!=SpDimLookupItem::Double){
    cerr << "DIMENSION " << dim_name << " IN SPACE " << space->GetName()
         << " is not a LO/HI/RES dimension." << endl;
    abort();
  }

  double_dimension = (SpDDMDoubleDimension *) dimension;

  double_dimension->SetLo(lo);
  double_dimension->SetHi(hi);
  double_dimension->SetRes(res);

  return (1);
}

SpDDMDim *
SpDDMRegion::FindDimension(const char *dim_name){
  return ((SpDDMDim *) Dimensions->Find((char *) dim_name));
}

/////////////////////////////////////////////////////////////////////////////////////////

SpPublishRegion::SpPublishRegion() {}

SpPublishRegion::SpPublishRegion(SpDDMSpace *_space, const char *name)
  : SpDDMRegion(_space, name) {
  const char *class_name;
  static const char *dim_name = "CLASS NAME";
  SpObjTemplate *obj_template = _space->GetSimObj()->GetObjectProxy()->GetObjectTemplate();

  class_name = obj_template->ObjProxyGetClassName();
  CreateDimension(dim_name, class_name, 0);

  SpDDMDim *dimension = FindDimension(dim_name);
  if (dimension->GetType()==SpDimLookupItem::Enum){
    SpDDMEnumDimension *enum_dimension = (SpDDMEnumDimension *) dimension;
    while ((obj_template = obj_template->GetParent())){
      class_name = obj_template->ObjProxyGetClassName();
      enum_dimension->Add(class_name, (const char *) 0);
    }
  }
}

SpPublishRegion::~SpPublishRegion() {}

/////////////////////////////////////////////////////////////////////////////////////////

SpSubscribeRegion::SpSubscribeRegion() {}

SpSubscribeRegion::SpSubscribeRegion(SpDDMSpace *_space, const char *name)
  : SpDDMRegion(_space, name), class_name_dimension((SpDDMEnumDimension *) 0) {}
 
SpSubscribeRegion::~SpSubscribeRegion() {}

void
SpSubscribeRegion::AddClass(const char *class_name, ...){
  va_list va;

  if (class_name_dimension==(SpDDMEnumDimension *) 0)
    class_name_dimension = CreateEnumDimension("CLASS NAME");

  va_start(va, class_name);
  class_name_dimension->Add(class_name, va);
  va_end(va);
}

void
SpSubscribeRegion::RemoveClass(const char *class_name, ...){
  va_list va;

  if (class_name_dimension==(SpDDMEnumDimension *) 0) return;

  va_start(va, class_name);
  class_name_dimension->Remove(class_name, va);
  va_end(va);
}

/////////////////////////////////////////////////////////////////////////////////////////


void RB_Register_SpDDMRegion()
{
  RB_REGISTER_CLASS_SpDDMRegion();
}

void RB_Register_SpPublishRegion()
{
  RB_REGISTER_CLASS_SpPublishRegion();
}

void RB_Register_SpSubscribeRegion()
{
  RB_REGISTER_CLASS_SpSubscribeRegion();
}

RB_SpBinaryTree &SpDDMRegion::GetDimensions(void)
{
  return (*Dimensions);
}

int SpDDMRegion::GetNumDimensions(void)
{
  return (Dimensions->GetNumElements());
}
