// SpSubscribeSpace.C file

#include "SpSubscribeSpace.H"
#include "S_SpHLA.H"
#include "RB_Types.H"
#include "SpDDMGrid.H"
#include "SpDDMRegion.H"
#include "H_SpAddProxySpace.H"
#include "H_SpRemoveProxySpace.H"
#include "F_SpProxyItem.H"
#include "RB_SpBinaryTree.H"
#include "RB_SpDefineClass.H"
 
RB_DEFINE_CLASS(SpSubscribeRegion);
RB_DEFINE_CLASS(SpSubscribeProxy);
RB_DEFINE_CLASS(SpSubscribeGrid);

//#define DDM_DEBUG

SpSubscribeProxy::SpSubscribeProxy(void)
  : ProxyItem((F_SpProxyItem *) 0), GlobalId(-1) {
  NumGrid  = 0;
}

SpSubscribeProxy::SpSubscribeProxy(F_SpProxyItem *proxyItem)
  : ProxyItem(proxyItem), GlobalId(proxyItem->GetGlobalId()) {
  NumGrid = 0;
}

SpSubscribeSpace::SpSubscribeSpace(void)
  : UpdateSemaphore(0), NewObjectProxyCallback(0), RemoveObjectProxyCallback(0),
    Regions(new RB_SpBinaryTree), Grids(new RB_SpBinaryTree),
    SubscribeObjectProxies(new RB_SpBinaryTree){

  AddEventHandlerMethod("SubscribeSpace Add Proxy Ptr",    AddRemoteProxy,    this);
  AddEventHandlerMethod("SubscribeSpace Remove Proxy Ptr", RemoveRemoteProxy, this);
}

SpSubscribeSpace::SpSubscribeSpace(S_SpHLA *hla, const SpSpace *space)
  : SpDDMSpace(hla, space), UpdateSemaphore(0), 
    NewObjectProxyCallback(0), RemoveObjectProxyCallback(0),
    Regions(new RB_SpBinaryTree), Grids(new RB_SpBinaryTree),
    SubscribeObjectProxies(new RB_SpBinaryTree){
  AddEventHandlerMethod("SubscribeSpace Add Proxy Ptr",    AddRemoteProxy,    this);
  AddEventHandlerMethod("SubscribeSpace Remove Proxy Ptr", RemoveRemoteProxy, this);
}

SpSubscribeSpace::~SpSubscribeSpace(void){
  SpSubscribeGrid *grid;
  SpSubscribeRegion *region;
  SpSubscribeProxy *subscribe_proxy;

  while (grid = (SpSubscribeGrid *) Grids->RemoveFirstElement())
    delete grid;
  delete Grids;

  while (region = (SpSubscribeRegion *) Regions->RemoveFirstElement())
    delete region;
  delete Regions;

  while (subscribe_proxy = (SpSubscribeProxy *) SubscribeObjectProxies->RemoveFirstElement())
    delete subscribe_proxy;
  delete SubscribeObjectProxies;
}

SpSubscribeGrid *
SpSubscribeSpace::FindGrid(int global_id){
  return ((SpSubscribeGrid *) Grids->Find(global_id));
}

void
SpSubscribeSpace::AddGrid(SpSubscribeGrid *grid){
  if (Grids->Find(grid->GetGlobalId())){
    cerr << "SpSubscribeSpace - grid already exist : " << grid->GetGlobalId() << endl;
    abort();
  }
  Grids->Insert(grid, grid->GetGlobalId());
}

int
SpSubscribeSpace::DeleteGrid(int global_id){
  int deleted = 0;
  SpSubscribeGrid *grid = (SpSubscribeGrid *) Grids->Remove(global_id);
  if (grid){
    RB_DELETE_SpSubscribeGrid(grid);
    deleted = 1;
  }
  return (deleted);
}

void
SpSubscribeSpace::Update() {
  SetUpdateSemaphore(1);
}

SpSubscribeRegion *
SpSubscribeSpace::CreateRegion(const char *name) {
  SpSubscribeRegion *region = FindRegion((char *) name);

  if (region==(SpSubscribeRegion *) 0){
    region = RB_NEW_SpSubscribeRegion(new SpSubscribeRegion(this, name));
    Regions->Insert(region, (char *) region->GetName());
  }

  return (region);
}

SpSubscribeRegion *
SpSubscribeSpace::FindRegion(const char *name) {
  return ((SpSubscribeRegion *) Regions->Find((char *) name));
}

int
SpSubscribeSpace::DeleteRegion(const char *name) {
  int deleted = 0;
  SpSubscribeRegion *region = (SpSubscribeRegion *) Regions->Remove((char *) name);
  if (region){
    RB_DELETE_SpSubscribeRegion(region);
    deleted = 1;
  }
  return (deleted);
}

void
SpSubscribeSpace::Add(const SpSimTime & /* time */,
                      int grid_id, F_SpProxyItem *item, int &added, int &notify){

  SpSubscribeProxy *subscribe_proxy;
  SpSubscribeGrid *grid = FindGrid(grid_id);

  added  = 0;
  notify = 0;

  if (grid==(SpSubscribeGrid *) 0){
#ifdef DDM_DEBUG
    cout << time << " SpSubscribeSpace add proxy " << item->GetGlobalId() 
         << " to grid " << grid_id << " not found" << endl;
#endif
    return;
  }
  if (!grid->Add(item)){
#ifdef DDM_DEBUG
    cout << time << " SpSubscribeSpace add proxy " << item->GetGlobalId() 
         << " to grid " << grid_id << " duplicate" << endl;
#endif
    return;
  } 

  subscribe_proxy = (SpSubscribeProxy *) SubscribeObjectProxies->Find(item->GetGlobalId());
  if (subscribe_proxy==(SpSubscribeProxy *) 0){
    subscribe_proxy = RB_NEW_SpSubscribeProxy(new SpSubscribeProxy(item));
    SubscribeObjectProxies->Insert(subscribe_proxy, subscribe_proxy->GetGlobalId());
  }
  subscribe_proxy->IncrNumGrid();

  added  = 1;
  notify = (subscribe_proxy->GetNumGrid()==1) ? 1 : 0;
}

void
SpSubscribeSpace::Remove(const SpSimTime & /* time */, 
                         int grid_id, int global_id, int &removed, int &notify){
  int num_grid;
  SpSubscribeProxy *subscribe_proxy;
  SpSubscribeGrid *grid = FindGrid(grid_id);

  removed = 0;
  notify  = 0;

  if (grid==(SpSubscribeGrid *) 0) return;
  if (!grid->Remove(global_id))    return;

  subscribe_proxy = (SpSubscribeProxy *) SubscribeObjectProxies->Find(global_id);
  num_grid        = subscribe_proxy->DecrNumGrid();

  if (num_grid<=0){
    subscribe_proxy = (SpSubscribeProxy *) SubscribeObjectProxies->Remove(global_id);
    RB_DELETE_SpSubscribeProxy(subscribe_proxy);
  }

  removed = 1;
  notify  = (num_grid<=0) ? 1 : 0;
}

void 
SpSubscribeSpace::RegisterNewObjectProxyCallback(void (*NewObjProxy) 
                                                      (SpSubscribeSpace *, F_SpProxyItem *)){
  NewObjectProxyCallback = NewObjProxy;
}

void 
SpSubscribeSpace::RegisterRemoveObjectProxyCallback(void (*RemoveObjProxy) 
                                                         (SpSubscribeSpace *, int)){
  RemoveObjectProxyCallback = RemoveObjProxy;
}

void
SpSubscribeSpace::Debug(ostream & os){

  os << "\t" << "sim obj         : " 
     << GetSimObj()->get_NAME() << " (" << GetSimObj()->get_GLOBAL_ID() << ")" << endl;
  os << "\t" << "sub space name  : " << GetName() << endl;
  os << "\t" << "num regions     : " << Regions->GetNumElements() << endl;
  os << "\t" << "num grids       : " << Grids->GetNumElements() << endl;
  os << "\t" << "num proxies     : " << SubscribeObjectProxies->GetNumElements() << endl;
  SpSubscribeGrid *grid = (SpSubscribeGrid *) Grids->GetFirstElement();
  while (grid){
    grid->Debug(os);
    grid = (SpSubscribeGrid *) Grids->GetNextElement();
  }
  os << endl;
}

void
SpSubscribeSpace::Validate(SIMTIME &time){
  const int subscriber_global_id = GetSimObj()->get_GLOBAL_ID();

  SpSubscribeGrid *grid = (SpSubscribeGrid *) Grids->GetFirstElement();
  while (grid){
    grid->Validate(time, subscriber_global_id, GetName());
    grid = (SpSubscribeGrid *) Grids->GetNextElement();
  }
}

void 
SpSubscribeSpace::AddRemoteProxy(char *, SpSimTime &time, char *data, int, int) {
  SpSubscribeSpaceAddProxy *add_proxy = (SpSubscribeSpaceAddProxy *) data;
  int *grid_id = (int *) (data + sizeof(SpSubscribeSpaceAddProxy));
  int added, notify;

  if (add_proxy->GetSpaceId()!=GetId()) return;

  for (int i=0; i<add_proxy->GetNumGrid(); i++){
    Add(time, grid_id[i], add_proxy->GetProxy(), added, notify);
    if (!added) continue;
#ifdef DDM_DEBUG
    cout << time << " simobj " << GetSimObj()->get_GLOBAL_ID()
         << " add proxy " << add_proxy->GetProxy()->GetName()
         << " grid " << grid_id[i] << endl;
#endif
    add_proxy->IncrNumGridAdd();
    if (notify && NewObjectProxyCallback)
      NewObjectProxyCallback(this, add_proxy->GetProxy());
  }
}

void 
SpSubscribeSpace::RemoveRemoteProxy(char *, SpSimTime &time, char *data, int, int) {
  SpSubscribeSpaceRemoveProxy *remove_proxy = (SpSubscribeSpaceRemoveProxy *) data;
  int *grid_id = (int *) (data + sizeof(SpSubscribeSpaceRemoveProxy));
  int removed, notify;

  if (remove_proxy->GetSpaceId()!=GetId()) return; 

  for (int i=0; i<remove_proxy->GetNumGrid(); i++){
    Remove(time, grid_id[i], remove_proxy->GetGlobalId(), removed, notify);
    if (!removed) continue;
#ifdef DDM_DEBUG
    cout << time << " simobj " << GetSimObj()->get_GLOBAL_ID() 
         << " rem proxy " << remove_proxy->GetGlobalId() 
         << " grid " << grid_id[i] << " removed " << removed << endl; 
#endif
    remove_proxy->IncrNumGridRemove();
    if (notify && RemoveObjectProxyCallback)
      RemoveObjectProxyCallback(this, remove_proxy->GetGlobalId());
  }
}

int
SpSubscribeSpace::GetNumGrids(void) const{
  return ((Grids) ? Grids->GetNumElements() : 0);
}

void RB_Register_SpSubscribeProxy()
{
  RB_REGISTER_CLASS_SpSubscribeProxy();
}
