#ifndef __SpDim__
#define __SpDim__

class MAIN_PUBSUB_ITEM;
class SpDimInfo;
class SpBasetype;

#include "SpAssert.H"
#include "SpStdlib.H"

#include "SpDimLookupItem.H"
#include "SpHiGrid.H"

class SpDataParser;
class HiGridRegion;
class SpHiGridRegionVec;

class SpDim
{
  protected:
    // The descendants' contructors should call set_NIdsInDim
    SpDim(SpDataParser &dataParser);

    SpDim(const char *name, int distribute);

  public:
    //   This function will put into
    // DimInfo[number_of_dims_initialized] the lo and hi in this
    // dimension of the SpHiGrid corresponding to IdOfGridInDimension.
    // This function may even assign additional elements of
    // DimInfo.  It might do this if it wants the SpHiGrids to use more
    // than one dimension to model this SpDim.  For example,
    // the SpLatLngDim is a single SpDim that can cover the
    // entire surface of the earth, but SpLatLngDim::InitSpHiGrid
    // adds in *two* dimensions to DimInfo, one for latitude and one
    // for longitude.  In any case this function should increment
    // number_of_dims_initialized by the number of elements of DimInfo
    // it chooses to assign.  That way, other SpDim objects
    // will know where to begin writing into DimInfo.
    //   Of course, in addition to Lo and Hi, this function should
    // initialize the name, resolutions, etc.
    virtual void InitSpHiGrid(int IdOfGridInDimension,
			     SpDimInfo DimInfo[MAX_DIMENSIONS],
			     int &number_of_dims_initialized) const = 0;

    //   This function will calculate all of the ids in this SpDim
    // which correspond to grids whose area overlaps the area of
    // dimensionLookupItem.  It will put these ids in the
    // SpHiGridRegionVec ids_of_grids_in_dimension that it
    // dynamically allocates (i.e. using new).  In addition to the id
    // field of HiGridRegion, it should also set the other fields like
    // GridCovered_p, and it should call set_DimRegion for each HiGrid
    // dimension that it wants to add, in order to supply the lo, hi,
    // and resolution which will be copied later into a
    // MAIN_PUBSUB_ITEM, which the user of O_SpGrid might pass to
    // SpHiGrid::AddPublisher or SpHiGrid::AddSubscriber.
    //   When this function is called, the dimensionLookupItem will
    // have already been verified to be of the correct child type, and
    // so this function may immediately cast to that type and use this
    // variable.
    //   This function returns TRUE iff dimensionLookupItem fits
    // entirely within the min and max of this dimension.  If any part
    // of dimensionLookupItem falls outside this dimension, then this
    // function should return FALSE.
    virtual int HiGrid_Lookup
      (const SpDimLookupItem *dimensionLookupItem,
       SpHiGridRegionVec* &ids_of_grids_in_dimension) const = 0;

    // The descendents of this class should define this function to
    // return the number of HiGrid Dimensions that correspond to
    // themselves.  Most of the time this function should just return
    // 1, but some dimensions might return more than 1.  For example,
    // SpLatLngDim returns 2 since it adds two Dimensions to
    // the SpHiGrid objects, one for latitude and one for longitude.
    virtual int get_NumberOfHiGridDims() const = 0;
    
    const char *get_DimName() const { return DimName; }
    int get_NIdsInDim() const { return NIdsInDim; }
    virtual SpDimLookupItem::DimensionTypes get_DimType() const=0;

    void print(ostream &os) const;

    // This function should return a new descendent of
    // SpDimLookupItem which has its DontCare_p flag set to TRUE
    // and whose get_DimType function returns the same thing as the
    // get_DimType function in the decendent of this SpDim class
    // returns.
    virtual SpDimLookupItem *GetDontCareLookupItem() const = 0;

  protected:
    static void get_resolution
      (SpDataParser &dataParser, int &nresolutions, double* &resolution);

    char *SpHiGrid_DimName(const char *ChildName) const;
    int distribute_p() const { return Distribute_p; }

    void set_NIdsInDim(int nIdsInDim)
    { assert(nIdsInDim > 0); NIdsInDim = nIdsInDim; }
    
    SpDim()
      : DimName     (0),
	Distribute_p(0)
    {abort();}
    
  private:
    SpDim (SpDim&)
      : DimName     (0),
	Distribute_p(0)
    {abort();}
    
    virtual void operator=(SpDim&) {abort();}

    virtual void print_work(ostream &os) const = 0;

    char *DimName;

    int NIdsInDim; // Number of sections this dimension wants distributed

    const int Distribute_p;
};

inline ostream & operator<<(ostream &os, const SpDim &dimension)
{
  dimension.print(os);
  return os;
}

typedef SpDim Dimension;

#endif
