#include "SpAssert.H"
#include "SpIostream.H"
#include "SpFstream.H"
#include "SpStrstream.H"
#include "SpIomanip.H"
#include "SpStdlib.H"
#include "SpDoubleDim.H"
#include "SpDataParser.H"
#include "SpDimInfo.H"
#include "SpPubSubItem.H"
#include "SpHiGridRegionVec.H"
#include "SpDataParserTypes.H"

SpDoubleDim::SpDoubleDim(SpDataParser &dataParser)
  : SpDim(dataParser),
    DimensionLo(dataParser.GetFloat("Lo")),
    DimensionHi(dataParser.GetFloat("Hi"))
{
  assert(DimensionHi > DimensionLo);
  
  get_resolution(dataParser, Nresolutions, Resolution);
    
  if (distribute_p())
  {
    const double diff = DimensionHi - DimensionLo;
    set_NIdsInDim(int(0.5 + diff / Resolution[0]));
    Resolution[0] = diff / get_NIdsInDim();
  } else
    set_NIdsInDim(1);
}

void SpDoubleDim::InitSpHiGrid(int IdOfGridInDimension,
				  SpDimInfo DimInfo[MAX_DIMENSIONS],
				  int &number_of_dims_initialized) const
{
  DimInfo[number_of_dims_initialized].SetDimName(SpHiGrid_DimName("Double"));
  DimInfo[number_of_dims_initialized].SetDimNum(number_of_dims_initialized);

  DimInfo[number_of_dims_initialized].SetLo(GridLo(IdOfGridInDimension));
  DimInfo[number_of_dims_initialized].SetHi(GridHi(IdOfGridInDimension));

  int startingIdOfResInDim = distribute_p() ? 1 : 0;

  DimInfo[number_of_dims_initialized]
    .SetNresolutions(Nresolutions - startingIdOfResInDim);

  for (int IdOfResInDim = startingIdOfResInDim;
       IdOfResInDim < Nresolutions;
       IdOfResInDim++)
  {
    DimInfo[number_of_dims_initialized].SetResolution
      (IdOfResInDim - startingIdOfResInDim, Resolution[IdOfResInDim]);
  }
  
  ++number_of_dims_initialized;
}

int SpDoubleDim::HiGrid_Lookup
  (const SpDimLookupItem *dimensionLookupItem,
   SpHiGridRegionVec* &ids_of_grids_in_dimension) const
{
  int lookup_fits_in_dimension_p;
  
  const SpDoubleLookupItem *doubleLookupItem
    = (const SpDoubleLookupItem *) dimensionLookupItem;
  
  if (doubleLookupItem->get_Hi() < DimensionLo ||
      doubleLookupItem->get_Lo() > DimensionHi)
  {
    ids_of_grids_in_dimension = new SpHiGridRegionVec(0, 0);
    lookup_fits_in_dimension_p = 0;
  }
  else
  {
    int LoBin = id_of_grid_in_dimension(doubleLookupItem->get_Lo());
    int HiBin = id_of_grid_in_dimension(doubleLookupItem->get_Hi());

    ids_of_grids_in_dimension = new SpHiGridRegionVec(HiBin - LoBin + 1, 1);
    for (int id_in_dim = LoBin; id_in_dim <= HiBin; id_in_dim++)
    {
      ids_of_grids_in_dimension
	->set_next()
	.set_id(id_in_dim)
	.set_gridCovered_p(doubleLookupItem->get_Lo() <= GridLo(id_in_dim) &&
			   GridHi(id_in_dim) <= doubleLookupItem->get_Hi())
	.set_DimRegion(0,
		       doubleLookupItem->get_Lo(),
		       doubleLookupItem->get_Hi(),
		       doubleLookupItem->get_Res());
    }
    
    lookup_fits_in_dimension_p = (doubleLookupItem->get_Lo() >= DimensionLo &&
				  doubleLookupItem->get_Hi() <= DimensionHi);
  }

  return lookup_fits_in_dimension_p;
}

void SpDoubleDim::print_work(ostream &os) const
{
  os << "         DimensionLo = " << DimensionLo << endl
     << "         DimensionHi = " << DimensionHi << endl
     << "         Resolutions: [";

  for (int k=0; k<Nresolutions; k++) {
    os << Resolution[k];
    if (k != Nresolutions-1) os << ", ";
  }

  os << "]" << endl;
}

int SpDoubleDim::id_of_grid_in_dimension(double low_or_hi) const
{
  double
    HmL = DimensionHi - DimensionLo,
    diff = low_or_hi - DimensionLo,
    fraction = diff / HmL;

  int id_of_grid_in_dimension = int(get_NIdsInDim() * fraction);

  if (id_of_grid_in_dimension >= get_NIdsInDim())
    id_of_grid_in_dimension = get_NIdsInDim() - 1;

  if (id_of_grid_in_dimension < 0)
    id_of_grid_in_dimension = 0;

  assert(ValidIdOfGridInDimension(id_of_grid_in_dimension));
  return id_of_grid_in_dimension;
}

double SpDoubleDim::GridLo(int id_of_grid_in_dimension) const
{
  assert(ValidIdOfGridInDimension(id_of_grid_in_dimension));
  double Lo = DimensionLo + id_of_grid_in_dimension * Resolution[0];
  return Lo;
}

double SpDoubleDim::GridHi(int id_of_grid_in_dimension) const
{
  assert(ValidIdOfGridInDimension(id_of_grid_in_dimension));
  double Hi;
  // This if is needed for when distribute_p() is 0
  if (id_of_grid_in_dimension == get_NIdsInDim() - 1)
    Hi = DimensionHi;
  else
    Hi = DimensionLo + (id_of_grid_in_dimension + 1) * Resolution[0];

  return Hi;
}

int SpDoubleDim::ValidIdOfGridInDimension
  (int id_of_grid_in_dimension) const
{
  return (0 <= id_of_grid_in_dimension &&
              id_of_grid_in_dimension < get_NIdsInDim());
}
