#include "SpIostream.H"
#include "SpFstream.H"
#include "SpStrstream.H"
#include "SpIomanip.H"
#include "SpAssert.H"
#include "SpEnumDim.H"
#include "SpDataParser.H"
#include "SpDimInfo.H"
#include "SpHiGridRegionVec.H"

//#define DEBUG_SpaceMan

const int NumBitsPerByte = 8, MaxNumEnums = NumBitsPerByte * sizeof(BitVector);

SpEnumDim::SpEnumDim(SpDataParser &DataParser)
    : SpDim(DataParser), NEnums(DataParser.GetNtypes("ENUM"))
{
  assert(NEnums > 0);

  if (NEnums > MaxNumEnums)
  {
    cerr << "Error: SpDim \"" << get_DimName() << "\" has " << NEnums
	 << " enums, which is more\nthan the maximum of " << MaxNumEnums
	 << endl;
    abort();
  }

  Enums = new const char*[NEnums];

  void *enum_mark = DataParser.GetMark();
  
  int IdInDimension = 0;
  for (void *not_done_p = DataParser.GoToFirstElement(enum_mark, "ENUM");
       not_done_p;
       not_done_p = DataParser.GoToNextElement(enum_mark, "ENUM"))
  {
    Enums[IdInDimension] = DataParser.GetReferenceName();
    for (int i = 0; i < IdInDimension; i++)
    {
      if (0 == strcmp(Enums[IdInDimension], Enums[i]))
      {
	cerr << "Error: Enum SpDim \"" << get_DimName()
	     << "\" has two enums named \"" << Enums[i] << "\"." << endl;
	abort();
      }
    }
    ++IdInDimension;
  }
  assert(IdInDimension == NEnums);

  set_NIdsInDim(distribute_p() ? NEnums : 1);
}

SpEnumDim::SpEnumDim(const char *name, int distribute,
                             int n, const char** &value)
  : SpDim(name, distribute), NEnums(n) {

  Enums = (n>0) ? new const char* [n] : (const char **) 0;
  for (int i=0; i<n; i++)
    Enums[i] = value[i];

  set_NIdsInDim(distribute_p() ? NEnums : 1);
}

void SpEnumDim::InitSpHiGrid(int IdOfGridInDimension,
				SpDimInfo DimInfo[MAX_DIMENSIONS],
				int &number_of_dims_initialized) const
{
  /*
  const int size = 500;
  char s[size];
  char *name;
  if (distribute_p())
  {
    assert(strlen(Enums[IdOfGridInDimension]) + 100 < size);
    sprintf(s, "Enum %s", Enums[IdOfGridInDimension]);
    name = s;
  }
  else
    name = "Enum";
    
  DimInfo[number_of_dims_initialized].SetDimName(SpHiGrid_DimName(name)); 
  */
  
  ostrstream name_ostrstream;
  
  name_ostrstream << "Enum";
  
  if (distribute_p())
    name_ostrstream << ' ' << Enums[IdOfGridInDimension];
  
  name_ostrstream << ends;
  
  const char * const name = name_ostrstream.str();
  
  DimInfo[number_of_dims_initialized].SetDimName(SpHiGrid_DimName(name));

  DimInfo[number_of_dims_initialized].SetDimNum(number_of_dims_initialized);

  DimInfo[number_of_dims_initialized].SetNresolutions(1);
  DimInfo[number_of_dims_initialized].SetResolution(0, 1);

  DimInfo[number_of_dims_initialized].SetLo(IdOfGridInDimension);
  DimInfo[number_of_dims_initialized].SetHi(distribute_p() ?
					    IdOfGridInDimension + 1 :
					    NEnums);

  ++number_of_dims_initialized;
}

int SpEnumDim::HiGrid_Lookup
  (const SpDimLookupItem *dimensionLookupItem,
   SpHiGridRegionVec* &ids_of_grids_in_dimension) const
{
  const SpEnumLookupItem *enumLookupItem
    = (const SpEnumLookupItem *) dimensionLookupItem;

  const BitVector one = 1;
  BitVector enumBitVector = enumLookupItem->get_EnumBitVector();
  
  ids_of_grids_in_dimension = new SpHiGridRegionVec(NEnums, 1);

#ifdef DEBUG_SpaceMan
  cout << "\nEnumDimension::HiGrid_Lookup, name = " << get_DimName() << endl;
#endif

  for (int IdInDimension = 0; IdInDimension < NEnums; IdInDimension++)
  {
    if (enumBitVector & one)
    {
      ids_of_grids_in_dimension
	->set_next()
	.set_DimRegion(0, IdInDimension, IdInDimension, 1)
      	.set_id(distribute_p() ? IdInDimension : 0)

        // Notice that I always set gridCovered_p to TRUE if it is
        // distributed, since, if the bit is set, then it has to refer
        // to the whole grid since the grid contains only one enum.
        // But if it is not distributed then the grid contains all
        // enums, but at this point we only know that this single enum
        // is set, which means we don't know if all of the other enums are
        // set, so we write in FALSE.
	.set_gridCovered_p(distribute_p());

#ifdef DEBUG_SpaceMan
      cout << Enums[IdInDimension] << endl;
#endif
    }
    enumBitVector >>= 1;
  }

  // Optimization: If we are not distributing, then there is one grid
  // containing all enums, and if all of the enums are set then we can
  // set this one grid to be covered.  But first we clear out all of
  // the other regions added by the previous loop, since they are not
  // needed now that we are saying the one and only grid is covered.
  if (!distribute_p() && ids_of_grids_in_dimension->entries() == NEnums)
  {
    ids_of_grids_in_dimension
      ->clear()
      .set_next()
      .set_DimRegion(0, 0, 0, -1)
      .set_id(0)
      .set_gridCovered_p(1);
  }
  
  // There should be no more 1 bits in enumBitVector.  If there are then
  // the caller probably made a mistake, since there is no enum in the
  // .par file corresponding to the bit that the caller set.  For most
  // dimensions it is easy to imagine why the caller might pass in a
  // lookup item that is out of range of the space.  The most common
  // example is when the .par file specifies the boundaries of the
  // space, but a ship goes outside those boundaries.  But I cannot
  // think of a likely example of a enum being out of bounds.  Maybe the
  // author of the .par file thought only certain enums would be
  // used, but the caller used one that the .par author did not think
  // of.  But I think that is less likely than the theory that the
  // caller made a mistake, and so I assert.
  assert(enumBitVector == 0);

  return (enumBitVector == 0);
}

void SpEnumDim::print_work(ostream &os) const
{
  const int idW = 18, nameW = 15;

  os.setf(ios::left);
  
  os << "         "
     << setw(idW) << "Id in dimension"
     << setw(nameW) << "Enum name" << '\n';

  for (int id_of_enum_in_dim = 0;
       id_of_enum_in_dim < NEnums;
       id_of_enum_in_dim++)
  {
    os << "         "
       << setw(idW) << id_of_enum_in_dim
       << setw(nameW) << Enums[id_of_enum_in_dim] << '\n';
  }
}
