#ifndef __SpEnumLookupItem__
#define __SpEnumLookupItem__

#include "SpAssert.H"
#include "SpDimLookupItem.H"

class SpSpaceLookupItem;

typedef unsigned long int BitVector;

class SpEnumDim;

//   An object of this class stores data describing which enums the
// user wants looked up.  The user can use some easy to use functions
// which accept strings of the enum names to assign, set or clear a
// variable number of enums.  It is slow because of the calls to
// strcmp, and the variable argument usage, but that's OK for
// initialization code since initialization code is done only once and
// takes up only a small portion of the overall run time.  But when
// the simulation is started, you probably want to avoid the extra
// time, so you can use the fast function assign_id_in_dim.  To get
// the ids to pass to this function you can use the function
// name_to_id_in_dim.  Since that function is slow also because of
// strcmp, you can call that function in initialization code and store
// the return value for later passing to assign_id_in_dim.
//   Whenever an object of this class is constructed, it has only
// Don't Cares, which means that all enums are looked up.  As soon as
// you specify the lookup status on any enum, all enums that you never
// mentioned will not be looked up.  This may be confusing since the
// first enum mention changes the status of all other enums to being
// looked NOT up.  But I think it is still good since if you specify
// nothing then it is better to have everything looked up instead of
// nothing looked up.  That is in keeping with the meaning of a
// Don't Care.  But if you do start mentioning specific enums, then
// it would be bad to lookup even the ones you left out, since if you
// mention one thing you want, then you probably will mention everything
// else you want too.
class SpEnumLookupItem : public SpDimLookupItem
{
  public:
    // Use this constructor is you don't know what id_in_dim various
    // enums have.  By passing in a spaceLookupItem, this class will
    // be able to know what enum strings are in this dimension which
    // allows you to use functions which accept this info like
    // assign_enums, set_enums, clear_enums, and name_to_id_in_dim.
    // Unfortunately this constructor is slower than the next one.
    SpEnumLookupItem(const char * const lookupItemName,
		   const SpSpaceLookupItem &spaceLookupItem)
      : SpDimLookupItem(lookupItemName, 1), Enums(NULL), EnumBitVector(0)
	{ is_for_SpaceLookupItem(spaceLookupItem); }

    // Use this constructor if you know what ids you want to set.
    // Maybe you know this because you constructed another
    // SpEnumLookupItem with the previous constructor and then called
    // name_to_id_in_dim on it, and so now that you already know the ids
    // you don't need to use the above slower constructor anymore.
    SpEnumLookupItem(const char * const lookupItemName)
      : SpDimLookupItem(lookupItemName, 1), Enums(NULL), EnumBitVector(0)
	{}

    // If you used the previous constructor, but now you think you
    // need to use member functions that need Enums strings, then just
    // pass this function the spaceLookupItem into which you will
    // insert this object.  But call this function BEFORE you insert
    // this object into a SpSpaceLookupItem.
    SpEnumLookupItem &is_for_SpaceLookupItem
      (const SpSpaceLookupItem &spaceLookupItem);

    // The following functions change only the specified enums or ids.
    // If an enum or id is not specified in a call then that call will
    // leave the value the same for that unspecified enum or id.

    // The next argument after enum_name should be of type int and
    // should be 0 if you do NOT want the preceding name to be looked
    // up or a 1 if you DO want it looked up.  After you list all of
    // the pairs pass one final argument of NULL to indicate that
    // there are no other arguments to be read.
    void assign_enums(const char *enum_name, ...);

    // For the following two functions pass as many strings as you like,
    // and then pass NULL.  The set_enums causes the passed in
    // enums to be looked up, and clear_enums causes them to NOT be
    // looked up.
    void set_enums(const char *enum_name, ...);
    void clear_enums(const char *enum_name, ...);

    // Return the id_in_dim corresponding to the passed in string enum
    // name.  You can pass the return value of this function to
    // assign_id_in_dim in order to set the enum corresponding to the
    // string name.
    int name_to_id_in_dim(const char *enum_name) const;

    // Assign the enum having id_in_dim.  It will be assigned value.
    // The other enums will be unchanged.  You can figure out what id
    // you want by using the function name_to_id_in_dim.
    SpEnumLookupItem &assign_id_in_dim(int id_in_dim, int value);
    
    BitVector get_EnumBitVector() const
	{ assert(!get_DontCare_p()); return EnumBitVector; }

    // SpEnumDim::GetDontCareLookupItem is probably the only
    // function that needs to call this:
    SpEnumLookupItem(const char * const lookupItemName,
		   const char * const *enums, int nEnums)
      : SpDimLookupItem(lookupItemName, 1),
	Enums(enums), NEnums(nEnums), EnumBitVector(0) {}

  private:
    SpEnumLookupItem() {abort();}
    
    SpEnumLookupItem(SpEnumLookupItem&) {abort();}

    DimensionTypes get_DimType() const { return Enum; }

    void print_work(ostream &os) const;

    // We need this so that we can give the string names of the enums
    // to the user of this SpEnumLookupItem object.  This array is not
    // supposed to hold the names of only the enums that should be
    // looked up.  Rather it holds the names of all possible enums in
    // the dimension.  The EnumBitVector member variable is what holds data
    // describing what should be looked up.
    const char * const *Enums;
    int NEnums;
    BitVector EnumBitVector;

};

typedef SpEnumLookupItem EnumLookupItem;

#endif
