#ifndef __SpLatLngDim__
#define __SpLatLngDim__

#include "SpDim.H"
class HiGridRegion;
#include "SpLatLngLookupItem.H"

//   This dimension is is an efficient dimension for the earth.  If
// you did the straightforward thing and not use this dimension, but
// instead use two DoubleDimensions one for latitude and one for
// longitude, then the HiGrids near the pole would be smaller then
// near the equator since longitude lines get closer as they approach
// the pole.  The smaller grids near the pole would mean they do less
// work since less number of ships will tend to be in a smaller area.
// These grids doing less work means we will have bad load balancing.
// This dimension solves this problem by having one dimension instead
// of two separate ones for latitude and longitude.  By making it one
// dimension, I can have the grids near the poles be roughly the same
// size as near the equator.
//   The user specifies a lo and hi for latitude and a lo and hi for
// longitude, and the region is divided into ConicLayers.  A
// ConicLayer is a volume of space between two cones.  The cones
// always have there vertex at the center of the earth.  The axis of
// these cones is always the axis of the earth.  The height of the
// cones is infinity.  The only thing different about
// the various cones is the angle between the axis and a line in the
// cone leaving the vertex.  I.e. the only difference is how sharp the
// cones point is and whether the code points toward the north or
// south pole.  This SpLatLngDim in divided up by several of these
// cones, and the volume of space between the cones is called a
// ConicLayer.  The angular difference between two adjacent cones
// is ConicLayerDeltaRadLat.  This angle is calculated based on the
// resolution specified by the user in the .par file.
//   Each ConicLayer covers a different latitude range.  The number of
// these ConicLayers is stored in the variable
// SpLatLngDim::NConicLayers.  ConicLayer number 0 covers the
// latitudes closest to -pi/2, and ConicLayer number
// SpLatLngDim::NConicLayers - 1 covers latitudes closest to pi/2.
//   Each ConicLayer is divided into a number of sections indicated by
// ConicLayer.NIdsInConicLayer.  Each layer could have a different
// number for this variable.  This number will be less for layers
// covering latitudes closer to pi/2 or -pi/2 and greater for layers
// covering latitudes closer to 0.  Doing this means that we get a
// more constant area obtained by intersecting a grid in this
// dimension with the surface of the earth.
//   Notice that a grid in this dimension is
// shaped roughly like a four sided pyramid with the vertex of the
// pyramid at the center of the earth
//   Most of this code considers this Dimension to run from
// DimensionRadLngLo to DimensionRadLngHi + 2*pi if DimensionRadLngLo
// is more than DimensionRadLngHi!  Function
// DimensionRadLngHi_plus_maybe_two_pi implements this concept.  But
// this is just for internal calculations and you only need to be
// aware of it if you want to understand the internal code in this
// class.

enum IntersectStatus
{
  lookup_is_less_than_dimension,
  lookup_is_more_than_dimension,
  lookup_intersects_but_does_not_fit_in_dimension,
  lookup_fits_in_dimension
};

class SpLatLngDim : public SpDim
{
  public:
    SpLatLngDim(SpDataParser &dataParser);

  private:
    // In the following variable names, "Rad" means radians, "Lat"
    // means Latitude, and "Lng" means Longitude
    double
      DimensionRadLatLo, 
      DimensionRadLatHi,

      DimensionRadLngLo, 
      DimensionRadLngHi;

    double DimensionCoversSouthPole_p, DimensionCoversNorthPole_p;

    // True iff DimensionRadLngLo == -pi && DimensionRadLngHi == pi
    double DimensionCoversAllLng_p;

    // Each ConicLayer spans this many latitude radians
    double ConicLayerDeltaRadLat;

    class ConicLayer
    {
      private:
	
	ConicLayer (ConicLayer&) {abort();}
    
	void operator=(ConicLayer&) {abort();}

      public:
	ConicLayer() {}

	// Each grid in this ConicLayer spans this many longitude radians:
	double DeltaRadLng;
    
	// number of ids (grids) in this ConicLayer:
	int NIdsInConicLayer;

	// number of ids in all previous ConicLayers in the ConicLayers array:
	int CumNIdsInConicLayer;
    } *ConicLayers;

    int NConicLayers; // number of elements in array ConicLayers;
    
    double *KmResolution; // Resolutions in kilometers
    int Nresolutions; // number of elements in array Resolution

    // This the radius of the middle of the band formed by the
    // intersection of the earth's surface with the specified
    // ConicLayer.
    double Km_radius_of_band(const int id_of_ConicLayer_in_dim) const;
    
    // Arguments lo and hi will be assigned with the values from the
    // section indicated by the string in the second argument.
    void get_lo_and_hi
      (SpDataParser &DataParser, char * const Latitude_or_Longitude,
       double &lo, double &hi);

    // This function should be passed an angle in degrees.  But this
    // function changes it to radians.  This function verifys that the
    // angle is within the normal range for latitude or longitude.
    void CheckAndChangeAngle(const char * const Latitude_or_Longitude,
			     double &angle) const;

    // This function says that it maybe adds two_pi.  To be more
    // specific it adds it iff DimensionRadLngHi < DimensionRadLngLo
    double DimensionRadLngHi_plus_maybe_two_pi() const;

    double DimensionLongitudeRange() const;

    void InitSpHiGrid(int IdOfGridInDimension,
		     SpDimInfo DimInfo[MAX_DIMENSIONS],
		     int &number_of_dims_initialized) const;

    IntersectStatus DoIntersection
      (const int id_of_only_partially_covered_lo_ConicLayer_in_dim,
       const int id_of_only_partially_covered_hi_ConicLayer_in_dim,
       const double lng_lookup_lo,
       const double lng_lookup_hi,
       SpHiGridRegionVec* const ids_of_grids_in_dimension) const;

    int HiGrid_Lookup
      (const SpDimLookupItem *dimensionLookupItem,
       SpHiGridRegionVec* &ids_of_grids_in_dimension) const;

    void VerifyCorrect_two_pi_Multiples
      (const double integer_times_two_pi,
       const double lng_lookup_lo, const double lng_lookup_hi) const;

    void print_work(ostream &os) const;

    // Conversion functions:

    // You may think that only ints will be sent as input, but actually,
    // if you send an integer + .5 then you get back a latitude in the
    // middle of the ConicLayer, so I want to allow for that input by making
    // the input type be double instead of int.
    double id_of_ConicLayer_in_dim_to_RadLatLo
      (const double id_of_ConicLayer_in_dim) const;

    // This function determines which ConicLayer contains rad_lat.  It might
    // return an invalid index into the array ConicLayers.  If it returns a
    // negative number then rad_lat is south the dimension, or if it
    // returns something greater than or equal to NConicLayers, then rad_lat
    // is north of the Dimension.
    int RadLat_to_id_of_ConicLayer_in_dim(const double rad_lat) const;
    
    // This function returns the id of the grid in the conicLayer
    // specified by the first argument.  The returned id corresponds
    // to a grid containing the longitude in the second argument.  If
    // it returns a negative number then rad_lng is west of the
    // dimension, or if it returns something greater than or equal to
    // conicLayer.NIdsInConicLayer, then rad_lng is to the east of the
    // Dimension.
    int RadLng_to_id_of_grid_in_conicLayer
      (const ConicLayer &conicLayer, const double rad_lng) const;

    // This function calculates the Maximum Number Of Regions that the
    // HiGrid_Lookup function will need to return.  It is the maximum
    // that should be returned via the call by reference argument
    // named number_of_ids in the HiGrid_Lookup function.  The
    // SpDimLookupItem might intersect all grids which accounts
    // for the first operand to the plus operator below.  The second
    // operand is because for each ConicLayer, both ends of the lookup
    // might cover a grid in this ConicLayer, but yet the middle might not
    // be covered, which means that we need to return two
    // HiGridRegions for this grid, one of which accounts for one end
    // of the lookup and the other which accounts for the other end of
    // the lookup.  You may think that only one region is sufficient
    // and that the region would just be specified by the lo and hi of
    // the lookup.  But you do not want to cover the middle of the
    // grid and leave the ends uncovered.  Instead you want to cover
    // the ends and leave the middle uncovered, because this lookup
    // wraps almost all of the way around the earth and the small part
    // of the earth that the lookup does not cover is the middle part
    // of the grid.
    int MaximumNumberOfRegions() const
	{ return get_NIdsInDim() + NConicLayers; }

    SpDimLookupItem *GetDontCareLookupItem() const
	{ return new SpLatLngLookupItem; }

    SpDimLookupItem::DimensionTypes get_DimType() const
	{ return SpDimLookupItem::LatLng; }

    int get_NumberOfHiGridDims() const { return 2; }

    SpLatLngDim(SpLatLngDim&) {abort();}
};

typedef SpLatLngDim LatLngDimension;

#endif
