#include "SpStdlib.H"
#include "SpAssert.H"
#include "SpHiGrid.H"
#include "SpObsObjHandle.H"
#include "SpSpace.H"
#include "SpSpaceSubSet.H"
#include "SpDimLookupItem.H"
#include "SpDim.H"
#include "SpDataParser.H"
#include "SpSpaceLookupItem.H"
#include "SpHiGridRegionVec.H"
#include "SpOS.H"

void SpSpace::InitSpace(SpDataParser &DataParser, const int idOfSpaceInMan)
{
  IdOfSpaceInMan = idOfSpaceInMan;
  
  const char * const type = DataParser.GetType();
  assert(type != NULL);
  if (strcasecmp(type, "SPACE"))
  {
    cerr << "Error: Found type \"" << type << "\" in InterestSpaces section.\n"
	 << "Only SPACEs are allowed in the InterestSpaces section." << endl;
    abort();
  }

  void *SpaceMark = DataParser.GetMark();

  SpaceName = DataParser.GetReferenceName();
  NSpaceSubSets = DataParser.GetNumElements(SpaceMark, "Theater");

  const char * const SpaceLevelDimensionsNamePrefix
    = "Space Level Dimensions of ";
  char * const SpaceLevelDimensionsName
    = new char[strlen(SpaceLevelDimensionsNamePrefix) +
	      strlen(SpaceName) + 1];
  sprintf(SpaceLevelDimensionsName, "%s%s", SpaceLevelDimensionsNamePrefix,
	  SpaceName);

  SpaceLevelDimensions.InitSpaceSubSet
    (SpaceLevelDimensionsName, NULL, DataParser, 1, NSpaceSubSets, 1);

  assert(NSpaceSubSets > 0 || SpaceLevelDimensions.get_Ndims() > 0);
  
  const int cumNIdsInDim = SpaceLevelDimensions.get_NIdsInSpaceSubSet();
  
  NumDimRegions = SpaceLevelDimensions.get_NumDimRegions();
  
  if (NSpaceSubSets == 0)
  {
    SpaceSubSets = NULL;
    NIdsInSpace = cumNIdsInDim;
  }
  else
  {
    SpaceSubSets = new SpSpaceSubSet[NSpaceSubSets];
    NIdsInSpace = 0;
  
    int IdOfSpaceSubSetInSpace = 0;
    for (void *not_done_p = DataParser.GoToFirstElement
	   (SpaceMark, "Theater");
	 not_done_p;
	 not_done_p = DataParser.GoToNextElement(SpaceMark, "Theater"))
    {
      const char * const type2 = DataParser.GetType();
      assert(type2 != NULL);
      assert(0 == strcasecmp(type2, "Theater"));
      
      const char * const SpaceSubSetName = DataParser.GetReferenceName();
      char * const SpaceNameForHiGrid
	= new char[strlen(SpaceName) + strlen(SpaceSubSetName) + 4];
      sprintf(SpaceNameForHiGrid, "%s (%s)", SpaceName, SpaceSubSetName);
  
      SpaceSubSets[IdOfSpaceSubSetInSpace].InitSpaceSubSet
	(SpaceSubSetName, SpaceNameForHiGrid, DataParser, cumNIdsInDim, 0);

      SpaceSubSets[IdOfSpaceSubSetInSpace]
	.verify_that_dimensions_match_between_SpaceSubSets(SpaceSubSets[0]);

      NIdsInSpace
	+= SpaceSubSets[IdOfSpaceSubSetInSpace].get_NIdsInSpaceSubSet();
      
      ++IdOfSpaceSubSetInSpace;
    }
    assert(IdOfSpaceSubSetInSpace == NSpaceSubSets);
    NumDimRegions += SpaceSubSets[0].get_NumDimRegions();
  }

  // This is kind of a hack.  I use the GetDontCareLookupItem
  // function merely for the purpose of making it easy to compare
  // dimension names to look for duplicates.
  int NDimLookupItems;
  SpDimLookupItem **DimLookupItems;
  GetDontCareLookupItems(NDimLookupItems, DimLookupItems);
  for (int i = 0; i < NDimLookupItems; i++)
  {
    for (int j = i + 1; j < NDimLookupItems; j++)
    {
      if (0 == strcmp(DimLookupItems[i]->get_LookupItemName(),
		      DimLookupItems[j]->get_LookupItemName()))
      {
	cerr << "Error: SpSpace \"" << SpaceName
	     << "\" has two dimensions named \""
	     << DimLookupItems[i]->get_LookupItemName() << "\"." << endl;
	abort();
      }
    }
    delete DimLookupItems[i];
  }
  delete [] DimLookupItems;

  //if (get_node() == 0)// SpComm
  if (SpComm_GetNodeId() == 0)
    cout << "Space \"" << SpaceName << "\" has "
	 << NIdsInSpace << " HiGrid objects." << endl;
}

void SpSpace::GetDontCareLookupItems
  (int &NDimLookupItems, SpDimLookupItem ** &DimLookupItems) const
{
  const int Ndims = SpaceLevelDimensions.get_Ndims()
    + ((NSpaceSubSets == 0) ? 0 : SpaceSubSets[0].get_Ndims());

  DimLookupItems = new SpDimLookupItem *[Ndims];

  NDimLookupItems = 0;

  SpaceLevelDimensions.GetDontCareLookupItems(DimLookupItems, NDimLookupItems);

  if (NSpaceSubSets >= 1)
    SpaceSubSets[0].GetDontCareLookupItems(DimLookupItems, NDimLookupItems);

  assert(NDimLookupItems == Ndims);
}

void SpSpace::InitSpHiGrid
  (const int IdOfGridInSpace, SpHiGrid * const HiGrid) const
{
  assert(0 <= IdOfGridInSpace && IdOfGridInSpace < NIdsInSpace);
  
  // Notice that the number of SpDimInfos I create is NOT the total
  // number of Dimensions in this space =
  // SpaceLevelDimensions.get_Ndims() + SpaceSubSets->get_Ndims.  It
  // is MAX_DIMENSIONS which is defined in SpHiGrid.H.  It is possible
  // for a SpHiGrid to have more dimensions than the total number of
  // Dimensions in this space.  For example, SpSpaceSubSet::Dimensions
  // will sometimes hold a SpLatLngDim, but that one
  // SpLatLngDim will correspond to two SpHiGrid dimensions.  One
  // of those SpHiGrid dimensions will be called "Latitude" and the
  // other will be called "Longitude".  This is because it makes sense
  // to treat positions on the earth's surface as one dimension when
  // distributing SpHiGrids to nodes, but by the time SpHiGrid deals
  // with position, the grid it needs to divide up is roughly a square
  // and so it makes more sense for it to use two dimensions.
  SpDimInfo Dim_Info[MAX_DIMENSIONS];

  int number_of_dims_initialized = 0;
  // Jacob should consider passing only Dim_Info, to
  // SpSpaceSubSet::InitSpHiGrid.  I.e. do not pass
  // number_of_dims_initialized.  Then SpSpaceSubSet::InitSpHiGrid can
  // dereference Dim_Info, and write into it, and after writing into
  // it, it does code that looks like "++Dim_Info".
  // SpDim::InitSpHiGrid can also do "++Dim_Info".  And I can find
  // out how many elements were written by subtracting the beginning
  // pointer I originally passed in from the ending pointer .  But
  // will I then have to divide that difference by sizeof(SpDimInfo)?
  // If so, then maybe this is getting to be too kludgey.

  
  int IdOfGridInSpaceSubSet;
  const char *spaceNameForHiGrid;
  if (NSpaceSubSets > 0)
  {
    int cumNIdsInSpaceSubSet = 0, IdOfSpaceSubSetInSpace;
    for (IdOfSpaceSubSetInSpace = 0; ;IdOfSpaceSubSetInSpace++)
    {
      assert(IdOfSpaceSubSetInSpace < NSpaceSubSets);

      cumNIdsInSpaceSubSet +=
	SpaceSubSets[IdOfSpaceSubSetInSpace].get_NIdsInSpaceSubSet();

      if (IdOfGridInSpace < cumNIdsInSpaceSubSet)
	break;
    }

    cumNIdsInSpaceSubSet -=
      SpaceSubSets[IdOfSpaceSubSetInSpace].get_NIdsInSpaceSubSet();

    IdOfGridInSpaceSubSet = IdOfGridInSpace - cumNIdsInSpaceSubSet;

    SpaceLevelDimensions.InitSpHiGrid
      (IdOfGridInSpaceSubSet, Dim_Info, number_of_dims_initialized);

    SpaceSubSets[IdOfSpaceSubSetInSpace].InitSpHiGrid
      (IdOfGridInSpaceSubSet, Dim_Info, number_of_dims_initialized);

    spaceNameForHiGrid
      = SpaceSubSets[IdOfSpaceSubSetInSpace].spaceNameForHiGrid();
  }
  else
  {
    IdOfGridInSpaceSubSet = IdOfGridInSpace;

    SpaceLevelDimensions.InitSpHiGrid
      (IdOfGridInSpaceSubSet, Dim_Info, number_of_dims_initialized);

    spaceNameForHiGrid = SpaceName;
  }

  // It should be 0 since it had its end repeatedly chopped off.
  // There was an amount chopped off for each dimension.  The total of
  // these chops should have brought the number down to 0.  The
  // chopping was done by a division operation in the function
  // SpSpaceSubSet::InitSpHiGrid
  assert(IdOfGridInSpaceSubSet == 0);

  HiGrid->init((char*) spaceNameForHiGrid, IdOfSpaceInMan,
	       number_of_dims_initialized, Dim_Info);
}

int SpSpace::HiGrid_Lookup
  (const SpSpaceLookupItem &spaceLookupItem,
   SpHiGridRegionVec * &output_ids_of_grids_in_SpSpace) const
{
  SpHiGridRegionVec *SpaceLevel_ids_of_grids_in_SpSpaceSubSet;
  int lookup_fits_in_SpaceLevelDimensions_p;
  
  // Assign the previous two variables:
  if (SpaceLevelDimensions.get_Ndims() == 0)
  {
    (SpaceLevel_ids_of_grids_in_SpSpaceSubSet
     = new SpHiGridRegionVec(1, 0))
      ->set_next()
      .set_id(0)
      .set_gridCovered_p(1);

    lookup_fits_in_SpaceLevelDimensions_p = 1;
  }
  else
  {
    // Do the SpaceLevelLookupItems:
    SpHiGridRegionVec input_ids_of_grids_in_SpSpaceSubSet(1, 0);

    input_ids_of_grids_in_SpSpaceSubSet
      .set_next()
      .set_id(0)
      .set_gridCovered_p(1);
  
    lookup_fits_in_SpaceLevelDimensions_p = SpaceLevelDimensions
      .HiGrid_Lookup(0,
		     spaceLookupItem,
		     &input_ids_of_grids_in_SpSpaceSubSet,
		     SpaceLevel_ids_of_grids_in_SpSpaceSubSet,
		     1);
  }
  
  int return_value;
  
  if (NSpaceSubSets == 0)
  {
    output_ids_of_grids_in_SpSpace = SpaceLevel_ids_of_grids_in_SpSpaceSubSet;

    return_value = lookup_fits_in_SpaceLevelDimensions_p;
  }
  else
  {
    // Do the SpaceSubSetLevelLookupItems:
    output_ids_of_grids_in_SpSpace
      = new SpHiGridRegionVec(NIdsInSpace, NumDimRegions);

    int lookup_fits_in_some_SpSpaceSubSet_p = 0, cumNIdsInSpaceSubSet = 0;

    for (int IdOfSpaceSubSetInSpace = 0;
	 IdOfSpaceSubSetInSpace < NSpaceSubSets;
	 cumNIdsInSpaceSubSet
	   += SpaceSubSets[IdOfSpaceSubSetInSpace++].get_NIdsInSpaceSubSet())
    {
      SpHiGridRegionVec *output_ids_of_grids_in_SpSpaceSubSet;

      lookup_fits_in_some_SpSpaceSubSet_p
	= SpaceSubSets[IdOfSpaceSubSetInSpace].HiGrid_Lookup
	(SpaceLevelDimensions.get_Ndims(),
	 spaceLookupItem,
	 SpaceLevel_ids_of_grids_in_SpSpaceSubSet,
	 output_ids_of_grids_in_SpSpaceSubSet,
	 SpaceLevelDimensions.get_NIdsInSpaceSubSet())
	|| lookup_fits_in_some_SpSpaceSubSet_p;

      SpHiGridRegionVecReader iterator(output_ids_of_grids_in_SpSpaceSubSet);
      while (++iterator)
      {
	const int
	  id_of_grid_in_SpSpaceSubSet = iterator.id(),

	  id_of_grid_in_space
	  = id_of_grid_in_SpSpaceSubSet + cumNIdsInSpaceSubSet;

	output_ids_of_grids_in_SpSpace
	  ->set_next()
	  .copy_grid(iterator)
	  .set_id(id_of_grid_in_space);
      }

      delete output_ids_of_grids_in_SpSpaceSubSet;
    }
    assert(cumNIdsInSpaceSubSet == NIdsInSpace);
    
    delete SpaceLevel_ids_of_grids_in_SpSpaceSubSet;

    return_value =
      lookup_fits_in_SpaceLevelDimensions_p &&
      lookup_fits_in_some_SpSpaceSubSet_p;
  }

  return return_value;
}

int 
SpSpace::HasTheaterDimension(void) const{
  return (NSpaceSubSets>0);
}

void 
SpSpace::FindDimensionType(const char *name,
                        int &found, 
                        SpDimLookupItem::DimensionTypes &type) const{
  const SpDim *dimension;

  found = 0;

  dimension = SpaceLevelDimensions.FindDimension(name);
  if (dimension){
    type  = dimension->get_DimType();
    found = 1;
  }
  for (int i=0; ( (i<NSpaceSubSets) && (!found) ); i++){
    dimension = SpaceSubSets[i].FindDimension(name);
    if (!dimension) continue;
    type  = dimension->get_DimType();
    found = 1;
    break; 
  }
}

void SpSpace::print(ostream &os) const
{
  os << "Space \"" << SpaceName << "\" has "
     << NIdsInSpace << " HiGrid objects." << endl;

  os << "   SpaceLevelDimensions: " << endl;

  os << SpaceLevelDimensions;
  
  os << "   SpaceSubSets (theater level): " << endl;

  for (int id_of_SpSpaceSubSet_in_SpSpace = 0;
       id_of_SpSpaceSubSet_in_SpSpace < NSpaceSubSets;
       id_of_SpSpaceSubSet_in_SpSpace++)
  {
    os << SpaceSubSets[id_of_SpSpaceSubSet_in_SpSpace];
  }
}
