#ifndef __SpSpaceLookupItem__
#define __SpSpaceLookupItem__

#include "O_SpGrid.H"

class MAIN_PUBSUB_ITEM;
#include "SpDimLookupItem.H"

//   This class contains data for what the user wants to lookup.  An
// object of this class is passed to O_SpGrid::HiGrid_Lookup.
//   We want to avoid doing lots of string comparisons.  For example,
// we want to minimize the number of times part of the code passes a
// string containing the name of the dimension or space, etc.  Jeff
// Steinman thought we could minimize string comparisons by requiring
// integer ids to be passed instead of strings.  But we don't want the
// programmer who used this code to have to know what number to
// choose.  So Jeff Steinman thought we could have that programmer
// pass a string to a function whose only purpose was to return the
// corresponding id, and that the programmer would then store this id,
// and pass it to the other functions that need it.  I made this idea
// more elaborate by having an object of this class returned given a
// string name rather than returning an integer id.  For example you
// construct this class by passing a string and that stores the id of
// the space in this class.  You can also pass in strings for
// dimension names to specify the lookup region for that dimension,
// and it stores your SpDimLookupItem in the same order in an
// array as the order that the SpSpace object stores it.  But the
// programmer doesn't have to know what order that is.  Furthermore,
// because the initial storage is done in order, it reduces time to do
// a lookup, since the lookup functions do not have to verify that the
// correct dimension names and types are there in the correct order.
// Not only, do lookups NOT have to verify that strings match, they also
// do NOT have to verify that ids match!
class SpSpaceLookupItem
{
  public:
    // Pass in the name of the space for which you want to do a
    // lookup.  The constructed object will have "Don't Cares" for all
    // dimensions.  For all dimensions you care about, you should call
    // InsertDimLookupItem to replace the "Don't Care" with that
    // dimension's lookup region.  Then you pass this SpSpaceLookupItem
    // object to O_SpGrid::HiGrid_Lookup to do the lookup or you could
    // call HiGrid_Lookup on this SpSpaceLookupItem object.
    SpSpaceLookupItem(const char * const spaceName)
      : SpaceName(spaceName) { Initialize(); }

    // This constructor is used to supply some initial
    // DimensionLookupItems.  Pass as many as you want.  Each one
    // will be passed to InsertDimLookupItem.  The last argument
    // should be NULL.
    SpSpaceLookupItem(const char * const spaceName,
		    SpDimLookupItem *dimLookupItem, ...);

    ~SpSpaceLookupItem();
    
    // Note that this function stores the pointer passed in, not a
    // copy of that to which it points.  This function will delete the
    // passed in pointer if another insert is done or if this
    // SpSpaceLookupItem object is deleted.
    void InsertDimLookupItem(SpDimLookupItem * const dimLookupItem);

    const SpDimLookupItem *operator[](const int index) const;
    int get_NDimLookupItems() const { return NDimLookupItems; }

    int dim_name_and_type_to_id
      (const char * const dim_name,
       SpDimLookupItem::DimensionTypes dim_type) const;

    int HiGrid_Lookup(C_QUEUE * const ids_of_grids_in_speedes) const
	{ return SpSpace_Man->HiGrid_Lookup(*this, ids_of_grids_in_speedes); }

    int get_IdOfSpaceInManager() const { return IdOfSpaceInManager; }
    
    const char *get_SpaceName() const { return SpaceName; }

    static void set_SpSpace_Man(const O_SpGrid * const space_Man);

  private:
    SpSpaceLookupItem(SpSpaceLookupItem&) : SpaceName(NULL) {abort();}
    void operator= (SpSpaceLookupItem&) {abort();}

    void Initialize()
	{ SpSpace_Man->InitSpaceLookupItem(SpaceName, IdOfSpaceInManager,
					 NDimLookupItems, DimLookupItems); }

    int IdOfSpaceInManager;

    const char * const SpaceName;
    
    int NDimLookupItems;
    SpDimLookupItem **DimLookupItems;

    static const O_SpGrid *SpSpace_Man;
};

ostream & operator<<(ostream &os, const SpSpaceLookupItem &spaceLookupItem);

typedef SpSpaceLookupItem SpaceLookupItem;

#endif
