#include "SpStdlib.H"
#include "SpAssert.H"
#include "SpSpaceSubSet.H"
#include "SpDataParser.H"
#include "SpSpaceLookupItem.H"

#include "SpAltitudeDim.H"
#include "SpEnumDim.H"
#include "SpDoubleDim.H"
#include "SpLatLngDim.H"

#include "SpHiGridRegionVec.H"
#include "SpOS.H"

#include "SpObjTemplate.H"
#include "S_SpHLA.H"

void SpSpaceSubSet::InitSpaceSubSet(const char * const spaceSubSetName,
				  const char * const spaceNameForHiGrid,
				  SpDataParser &DataParser,
				  int cumNIdsInDim,
				  int NTheatersAtDataParserPosition,
                                  int addObjClassDim)
{
  SpaceSubSetName = spaceSubSetName;
  SpaceNameForHiGrid = spaceNameForHiGrid;
  NIdsInSpaceSubSet = cumNIdsInDim;

  void *SpaceSubSetMark = DataParser.GetMark();

  Ndims = DataParser.GetNumElements(SpaceSubSetMark)
    - NTheatersAtDataParserPosition;

    // implicit add : obj class dimension
  if (addObjClassDim) Ndims++;  

  Dimensions = (Ndims>0) ? new SpDim* [Ndims] 
                         : (SpDim **) 0;

  int IdOfDimInSpaceSubSet = 0;
  for (void *not_done_p = DataParser.GoToFirstElement(SpaceSubSetMark);
       not_done_p;
       not_done_p = DataParser.GoToNextElement(SpaceSubSetMark))
  {
    const char * const DimensionType = DataParser.GetType();

    if (DimensionType == NULL)
    {
      const char * const DimensionName = DataParser.GetReferenceName();
      assert(DimensionName != NULL);

      if (0 == strcasecmp(DimensionName, "LatLng"))
        Dimensions[IdOfDimInSpaceSubSet] = new SpLatLngDim(DataParser);

      else if (0 == strcasecmp(DimensionName, "Altitude"))
        Dimensions[IdOfDimInSpaceSubSet] = new SpAltitudeDim(DataParser);
    }
    else if (0 == strcasecmp(DimensionType, "Dimension"))
      Dimensions[IdOfDimInSpaceSubSet] = new SpDoubleDim(DataParser);

    else if (0 == strcasecmp(DimensionType, "EnumType"))
      Dimensions[IdOfDimInSpaceSubSet] = new SpEnumDim(DataParser);

    else if (0 == strcasecmp(DimensionType, "Theater"))
      continue; // Ignore Theaters; function SpSpace::SpSpace will handle them

    else
    {
      cerr << "Error: Found unknown type \"" << DimensionType << "\" in \""
           << SpaceSubSetName << '"' << endl;
      abort();
    }
    
    NIdsInSpaceSubSet *= Dimensions[IdOfDimInSpaceSubSet]->get_NIdsInDim();
    ++IdOfDimInSpaceSubSet;
  }

    // add : obj class dimension
  if (addObjClassDim){
    int i;
    SpObjTemplate *obj;
    C_QUEUE *object_templates    = S_SpHLA::GetObjectFactory()->GetObjTemplate();
    const int num_class_template = object_templates->length();
    const char **template_name   = (num_class_template>0) ? new const char * [num_class_template]
                                                          : (const char **) 0;
    obj = (SpObjTemplate *) object_templates->get_top();
    i   = 0;
    while (i < num_class_template){
      template_name[i++] = obj->GetName();
      obj = (SpObjTemplate *) object_templates->next(obj);
    }
    Dimensions[IdOfDimInSpaceSubSet] = new SpEnumDim("CLASS NAME", 0,
                                                     num_class_template, template_name);
    NIdsInSpaceSubSet *= Dimensions[IdOfDimInSpaceSubSet++]->get_NIdsInDim();
  
    delete [] template_name;
  }
}

void SpSpaceSubSet::verify_that_dimensions_match_between_SpaceSubSets
  (const SpSpaceSubSet &other_SubSet) const
{
  if (Ndims != other_SubSet.Ndims)
  {
    cerr << "Error: Theaters in the same SpSpace must have the same number of\n"
	 << "dimensions, but Theater \"" << SpaceSubSetName << "\" has "
	 << Ndims << " dimensions,\nand Theater \""
	 << other_SubSet.SpaceSubSetName << "\" has " << other_SubSet.Ndims
	 << " dimensions." << endl;
    abort();
  }
  
  for (int id_of_dim_in_SpSpaceSubSet = 0;
       id_of_dim_in_SpSpaceSubSet < Ndims;
       id_of_dim_in_SpSpaceSubSet++)
  {
    const SpDimLookupItem::DimensionTypes
      read_type =
      get_Dimension(id_of_dim_in_SpSpaceSubSet)->get_DimType(),

      expected_type = other_SubSet.
      get_Dimension(id_of_dim_in_SpSpaceSubSet)->get_DimType();
	
    const char
      * const     read_name =
      get_DimensionName(id_of_dim_in_SpSpaceSubSet),

      * const expected_name = other_SubSet.
      get_DimensionName(id_of_dim_in_SpSpaceSubSet);
	
    if (read_type != expected_type
	|| 0 != strcmp(read_name, expected_name))
    {
      cerr << "Error: Both the type and the name should match, between "
	   << "Theaters, but\ndimension number " << id_of_dim_in_SpSpaceSubSet
	   << " (numbering starts at 0) in Theater \""
	   << other_SubSet.get_SpaceSubSetName() << "\" is of\ntype \""
	   << expected_type << "\" and named \"" << expected_name
	   << "\" but dimension number " << id_of_dim_in_SpSpaceSubSet
	   << " in\nTheater \"" << get_SpaceSubSetName() << "\" is of type \""
	   << read_type << "\" and named \"" << read_name << "\"."
	   << endl;
      abort();
    }

    // We will want to have the following assert as long as the
    // SpEnumLookupItem uses a bit vector to specify the
    // categories.  The reason is that some one might attempt to
    // have two SpaceSubSets with the same enum name but with
    // different enums, but the bit vector method does not allow for
    // specifying different enums since, bit numbers will be re-used
    // in the two categories.
    if (read_type == SpDimLookupItem::Enum)
    {
      fflush(stdout);
      fflush(stderr);
  
      cerr << "Error: Theater \"" << other_SubSet.get_SpaceSubSetName()
	   << "\" has a dimension named \"" << read_name << "\" of\n"
	"type \"" << read_type << "\", but that type is only allowed "
	"at the SpSpace\nLevel, (it is NOT allowed in Theaters)." << endl;

      abort();
    }
  }
}

void SpSpaceSubSet::InitSpHiGrid
  (int &IdOfGridInSpaceSubSet,
   SpDimInfo * const Dim_Info, int &number_of_dims_initialized) const
{
  for (int IdOfDimInSpaceSubSet = 0;
       IdOfDimInSpaceSubSet < Ndims;
       IdOfDimInSpaceSubSet++)
  {
    const int NIdsInDim = Dimensions[IdOfDimInSpaceSubSet]->get_NIdsInDim();

    const int IdOfGridInDimension = IdOfGridInSpaceSubSet % NIdsInDim;

    assert(0 <= IdOfGridInDimension && IdOfGridInDimension < NIdsInDim);
    Dimensions[IdOfDimInSpaceSubSet]->InitSpHiGrid(IdOfGridInDimension,
						  Dim_Info,
						  number_of_dims_initialized);
    assert(number_of_dims_initialized < MAX_DIMENSIONS);

    IdOfGridInSpaceSubSet /= NIdsInDim;
  }
}

ostream & operator<<(ostream &os, const SpSpaceSubSet &spaceSubSet)
{
  os << "   SpaceSubSet \"" << spaceSubSet.get_SpaceSubSetName() << "\" has "
       << spaceSubSet.get_NIdsInSpaceSubSet() << " HiGrid objects." << endl;

  for (int IdOfDimInSpaceSubSet = 0;
       IdOfDimInSpaceSubSet < spaceSubSet.get_Ndims();
       IdOfDimInSpaceSubSet++)
  {
    os << *spaceSubSet.get_Dimension(IdOfDimInSpaceSubSet);
  }

  return os;
}

int SpSpaceSubSet::HiGrid_Lookup
  (const int number_of_previous_dimensions,
   const SpSpaceLookupItem &spaceLookupItem,
   const SpHiGridRegionVec * const input_ids_of_grids_in_SpSpaceSubSet,
   SpHiGridRegionVec * &output_ids_of_grids_in_SpSpaceSubSet,
   int cumNIdsInDim) const
{
  // For each dimension I copy data from one block of memory to 
  // another block of memory.  I refer to the blocks using the pointer
  // variables named source and target.  After I copy from source to
  // target, I re-assign the addresses of source and target so that in
  // the next iteration I read from the block to which the previous
  // iteration wrote.

  const SpHiGridRegionVec *source = input_ids_of_grids_in_SpSpaceSubSet;
  SpHiGridRegionVec *last_written_target = NULL;
  
  int lookup_fits_in_SpSpaceSubSet_p = 1;
  
  for (int IdOfDimInSpaceSubSet = 0;
       IdOfDimInSpaceSubSet < Ndims;
       IdOfDimInSpaceSubSet++)
  {
    const SpDim * const dimension = Dimensions[IdOfDimInSpaceSubSet];
    const SpDimLookupItem * const dimLookupItem
      = spaceLookupItem[number_of_previous_dimensions + IdOfDimInSpaceSubSet];

    SpHiGridRegionVec *ids_of_grids_in_dimension;
    int lookup_fits_in_dimension_p;

    // Assign the previous two variables.
    if (dimLookupItem->get_DontCare_p())
    {
      ids_of_grids_in_dimension	= new SpHiGridRegionVec
	(dimension->get_NIdsInDim(), dimension->get_NumberOfHiGridDims());
      for (int number_of_overlapping_ids_in_dimension = 0;
	   number_of_overlapping_ids_in_dimension < dimension->get_NIdsInDim();
	   ++number_of_overlapping_ids_in_dimension)
      {
	ids_of_grids_in_dimension->set_next()
	  .set_id(number_of_overlapping_ids_in_dimension)
	  .set_gridCovered_p(1);

	for (int i = 0; i < dimension->get_NumberOfHiGridDims(); i++)
	  ids_of_grids_in_dimension->set_DimRegion(i, 0, 0, -1);
      }
      lookup_fits_in_dimension_p = 1;
    }
    else
    {
      lookup_fits_in_dimension_p = dimension->HiGrid_Lookup
	(dimLookupItem, ids_of_grids_in_dimension);
    }

    lookup_fits_in_SpSpaceSubSet_p
      = lookup_fits_in_SpSpaceSubSet_p && lookup_fits_in_dimension_p;

    // Now duplicate the regions we have so far adding in the new info
    // from ids_of_grids_in_dimension that we just got:
    SpHiGridRegionVec *target = new SpHiGridRegionVec
      (ids_of_grids_in_dimension->entries() *
       source                   ->entries(),
       ids_of_grids_in_dimension->get_NumDimRegions() +
       source                   ->get_NumDimRegions());

    SpHiGridRegionVecReader
      current_dim_iterator(ids_of_grids_in_dimension);
    while (++current_dim_iterator)
    {
      const int current_dim_grid_id_multiple
	= current_dim_iterator.id() * cumNIdsInDim;

      // Copy each element of source into target adding in
      // current_dim_grid_id_multiple, and the other HiGridRegion
      // fields.
      SpHiGridRegionVecReader source_iterator(source);
      while (++source_iterator)
      {
	target
	  ->set_next()

	  .SetRegionsToConcatenationOf(source_iterator, current_dim_iterator)

	  .set_id(source_iterator.id() + current_dim_grid_id_multiple)

	  .set_gridCovered_p(        source_iterator.gridCovered_p()
			     && current_dim_iterator.gridCovered_p());
      }
    }
      
    cumNIdsInDim *= dimension->get_NIdsInDim();

    delete ids_of_grids_in_dimension;

    delete last_written_target;
    last_written_target = target;
    source              = target;
  }
  assert(cumNIdsInDim == NIdsInSpaceSubSet);

  output_ids_of_grids_in_SpSpaceSubSet = last_written_target;

  assert(output_ids_of_grids_in_SpSpaceSubSet
	 != input_ids_of_grids_in_SpSpaceSubSet);

  return lookup_fits_in_SpSpaceSubSet_p;
}

const char *SpSpaceSubSet::get_DimensionName(int IdOfDimInSpaceSubSet) const
{
  assert(IdOfDimInSpaceSubSet < Ndims);
  return Dimensions[IdOfDimInSpaceSubSet]->get_DimName();
}

const SpDim *SpSpaceSubSet::get_Dimension(int IdOfDimInSpaceSubSet) const
{
  assert(IdOfDimInSpaceSubSet < Ndims);
  return Dimensions[IdOfDimInSpaceSubSet];
}

void SpSpaceSubSet::GetDontCareLookupItems
  (SpDimLookupItem ** &DimLookupItems,
   int &id_of_SpDimLookupItem_in_SpSpaceLookupItem) const
{
  for (int id_of_SpDim_in_spaceSubSet = 0;
       id_of_SpDim_in_spaceSubSet < Ndims;
       id_of_SpDim_in_spaceSubSet++)
  {
    DimLookupItems[id_of_SpDimLookupItem_in_SpSpaceLookupItem++]
      = Dimensions[id_of_SpDim_in_spaceSubSet]->GetDontCareLookupItem();
  }
}

int SpSpaceSubSet::get_NumDimRegions() const
{
  int total = 0;

  for (int i = 0; i < Ndims; i++)
    total += Dimensions[i]->get_NumberOfHiGridDims();

  return total;
}

#include "SpDimLookupItem.H"

const SpDim *
SpSpaceSubSet::FindDimension(const char *name) const{
  for (int i=0; i<Ndims; i++)
    if (strcmp(Dimensions[i]->get_DimName(), name)==0)
      return (Dimensions[i]);
  return ((const SpDim *) 0);
}
