#ifndef __SpSpaceSubSet__
#define __SpSpaceSubSet__

#include "SpIostream.H"
#include "SpFstream.H"
#include "SpStrstream.H"
#include "SpIomanip.H"
#include "SpStdlib.H"

class SpHiGrid;
class C_QUEUE;
class SpSpaceLookupItem;
class SpDim;
class SpDataParser;
class SpDimInfo;
class SpHiGridRegionVec;
class SpDim;
class SpDimLookupItem;

//   This class holds a subset of a space.  Since a space is made up
// of dimensions, this class is also made up of dimensions, although
// the min, max, etc stored in these dimensions is a subset of the
// allowable values for those dimensions.
//   Given a point in a dimension, the dimension code knows how to
// find the id of the grid in that dimension which contains the given
// point, and also how to do the inverse operations of finding the min
// and max of an id in that dimension.
//   This class knows how to combine those ids in dimensions, into one
// id in this SpSpaceSubSet, and this class also knows how to do the
// inverse operation of breaking up a single id in this SpSpaceSubSet in
// to the multiple ids in the dimensions that this SpSpaceSubSet holds.
//   Here is an example of how SpSpaceSubSet does these two operations:
//   I will use an example in which all of the dimensions have 10 ids
// each.  I.e. they each distribute 10 grids.  I.e. they each divide
// their range into 10 sections.  It is not required that they all be
// 10; it would be OK for one dimension to have 3 ids, another to have
// 29, and another to have 12, but for my example in order to make it
// easy to understand, I will use 10 ids for each of three dimensions
// that this SpSpaceSubSet holds.
//   I will show how to combine the following ids in dimension:
//
// IdOfDimInSpaceSubSet    IdOfGridInDimension
// 0                       3
// 1                       7
// 2                       1
//
//   In the above example, the left column is the index into the array
// SpSpaceSubSet::Dimensions.  The second column is the id in dimension.
// For this example which has 10 ids for each dimension, all of the
// numbers in the second column would have to go from 0 to 9.  Nothing
// else would be allowed.
//   The three ids in dimensions shown above would be combined into an
// id in this SpSpaceSubSet of 173.  Note that this number 173 is stored
// in a single object of type int.
//   It is obvious how we got from the table to the single integer of
// 173.  And of course it is equally obvious how to perform the
// inverse operations.  The hard part is writing formulas to do the
// conversion - that is formulas that work for any example even one
// where the number of ids in the dimensions are 3, 29, and 12 instead
// of 10, 10, and 10.  But if you keep this more simple example in
// mind as you read the code in this class, it shouldn't be too hard
// to understand the formulas, and so I will not bother trying to
// explain them here.
class SpSpaceSubSet
{
  public:
    //   The current state of object in the argument called DataParser
    // should be such that dimensions can be immediately read from it.
    // I.e. it should be at a SpSpace level in which case the space
    // level dimensions will be read from it, or it should be at a
    // Theater level so that everything (which should only be
    // dimensions) at that level can be read from it.
    //   The argument, cumNIdsInDim, is the number of grids all
    // previous dimensions want distributed.  An example of these
    // previous dimensions is the space level dimensions.
    //    The argument, NTheatersAtDataParserPosition is the number of
    // theaters at the current position of the argument DataParser.
    // It could be non-zero if DataParser is at a SpSpace level, but it
    // should be 0 if DataParser is at a Theater level.
    void InitSpaceSubSet(const char * const spaceSubSetName,
			 const char * const spaceNameForHiGrid,
			 SpDataParser &DataParser, int cumNIdsInDim,
			 int NTheatersAtDataParserPosition,
                         int addObjClassDim=0);

    SpSpaceSubSet() {}
    
    // This function initializes Dim_Info starting at index
    // number_of_dims_initialized with los and his, etc., which
    // correspond to IdOfGridInSpaceSubSet.  It increments
    // number_of_dims_initialized by however many elements of Dim_Info
    // it assigns.
    void InitSpHiGrid
      (int &IdOfGridInSpaceSubSet,
       SpDimInfo * const Dim_Info, int &number_of_dims_initialized) const;

    //   This function will calculate all of the ids in this
    // SpSpaceSubSet which correspond to grids whose area overlaps the
    // area of spaceLookupItem.  input_ids_of_grids_in_SpSpaceSubSet
    // contains cumulative ids of grids for dimensions held by some
    // other SpSpaceSubSet.  This function will duplicate those
    // cumulative ids and add on the appropriate multiples of ids for
    // the dimensions that this SpSpaceSubSet holds.  This function will
    // dynamically (i.e. using new) allocate a SpHiGridRegionVec and
    // put the new ids in it.  The address of this SpHiGridRegionVec
    // is returned by reference in argument
    // output_ids_of_grids_in_SpSpaceSubSet.  The caller is responsible
    // for deleting this SpHiGridRegionVec.
    //   The argument, cumNIdsInDim, is the number of grids all
    // previous dimensions want distributed.  An example of these
    // previous dimensions is the space level dimensions.
    //   This function returns TRUE iff spaceLookupItem fits entirely
    // within the boundaries of this SpSpaceSubSet.  If any part of
    // spaceLookupItem falls outside this SpSpaceSubSet, then this
    // function should return FALSE.
    int HiGrid_Lookup
      (const int number_of_previous_dimensions,
       const SpSpaceLookupItem &spaceLookupItem,
       const SpHiGridRegionVec * const input_ids_of_grids_in_SpSpaceSubSet,
       SpHiGridRegionVec * &output_ids_of_grids_in_SpSpaceSubSet,
       int cumNIdsInDim) const;
  
    const char *get_SpaceSubSetName() const { return SpaceSubSetName; }
    const char *spaceNameForHiGrid() const { return SpaceNameForHiGrid; }
    int get_NIdsInSpaceSubSet() const { return NIdsInSpaceSubSet; }

    int get_Ndims() const { return Ndims; }

    const char *get_DimensionName(int IdOfDimInSpaceSubSet) const;

    const SpDim *get_Dimension(int IdOfDimInSpaceSubSet) const;
    
    void verify_that_dimensions_match_between_SpaceSubSets
      (const SpSpaceSubSet &first_SpSpaceSubSet) const;

    void GetDontCareLookupItems
      (SpDimLookupItem ** &DimLookupItems,
       int &id_of_SpDimLookupItem_in_SpSpaceLookupItem) const;

    int get_NumDimRegions() const;

    const SpDim *FindDimension(const char *name) const; 
    
  private:
    SpSpaceSubSet   (SpSpaceSubSet&) {abort();}
    void operator=(SpSpaceSubSet&) {abort();}

    const char *SpaceSubSetName;

    const char *SpaceNameForHiGrid;

    // Number of grids this SpSpaceSubSet wants distributed:
    int NIdsInSpaceSubSet;      

    int Ndims;
    SpDim ** Dimensions;
};

ostream & operator<<(ostream &os, const SpSpaceSubSet &spaceSubSet);

typedef SpSpaceSubSet SpaceSubSet;

#endif
