#include "SpMath.H"
#include "SpAssert.H"
#include "SpIostream.H"
#include "SpFstream.H"
#include "SpStrstream.H"
#include "SpIomanip.H"
#include "SpGeoFunctions.H"
#include "SpDataParser.H"
#include "SpDimInfo.H"
#include "SpPubSubItem.H"
#include "SpHiGridRegionVec.H"
#include "SpDataParserTypes.H"
#include "SpLatLngDim.H"

//#define DEBUG_SpaceMan

SpLatLngDim::SpLatLngDim(SpDataParser &dataParser)
  : SpDim(dataParser)
{
  get_resolution(dataParser, Nresolutions, KmResolution);

  get_lo_and_hi(dataParser, "Latitude",	DimensionRadLatLo, DimensionRadLatHi);

  if (DimensionRadLatLo > DimensionRadLatHi)
  {
    cerr << "Error: In the Latitude section of dimension " << get_DimName()
	 << ",\nLo = " << RAD_to_DEG(DimensionRadLatLo)
	 << " is greater than Hi = " 
	 << RAD_to_DEG(DimensionRadLatHi)
	 << '.' << endl;
    abort();
  }
  
  get_lo_and_hi(dataParser, "Longitude", DimensionRadLngLo, DimensionRadLngHi);

  const double rad_lat_diff = DimensionRadLatHi - DimensionRadLatLo;

  if (distribute_p())
  {
    const double rad_angle_subtended_by_KmResolution_0
      = KmResolution[0] / radius_of_earth;

    NConicLayers = int(0.5 + rad_lat_diff
		       / rad_angle_subtended_by_KmResolution_0);
    assert(NConicLayers >= 0);
    
    // The reason I do not output an error message when NConicLayers
    // is 0, is that NIdsInConicLayer for some ConicLayers might NOT
    // be zero.  I.e. the user can enter only one resolution, and it
    // might need to be big enough to avoid too many divisions of the
    // longitude, and that may mean it is so big that the above
    // formula evaluates to 0.
    if (NConicLayers == 0)
      NConicLayers = 1;
  }
  else
    NConicLayers = 1;
  
  ConicLayers = new ConicLayer[NConicLayers];
    
  // Next, we fill in the previous array with the number of ids
  // (grids) in each ConicLayer.  ConicLayers which contain latitudes
  // closer to the equator are bigger than ConicLayers containing
  // latitudes closer to the poles.  I mentioned in the .H file that
  // the grids are shaped in the form of pyramids.  We want the
  // intersection of this pyramid with the surface of the earth to be
  // roughly a square rather than a long but thin rectangle, and all
  // of the grids should be about the same size.  Therefore more grids
  // can fit into the bigger ConicLayers near the equator than can fit
  // into the smaller ConicLayers near the pole, which means there
  // will be a different number of grids in each ConicLayer.  Now we
  // calculate that number for each ConicLayer.
    
  ConicLayerDeltaRadLat = rad_lat_diff / NConicLayers;

  const double
    ConicLayerDeltaKm = ConicLayerDeltaRadLat * radius_of_earth,
    dimensionLongitudeRange = DimensionLongitudeRange();
  
  int nIdsInDim = 0;
  for (int id_of_ConicLayer_in_dim = 0;
       id_of_ConicLayer_in_dim < NConicLayers;
       id_of_ConicLayer_in_dim++)
  {
    if (distribute_p())
    {
      const double DeltaRadLng
	= ConicLayerDeltaKm / Km_radius_of_band(id_of_ConicLayer_in_dim);
      
      ConicLayers[id_of_ConicLayer_in_dim].NIdsInConicLayer
        = int(0.5 + dimensionLongitudeRange / DeltaRadLng);

      assert(ConicLayers[id_of_ConicLayer_in_dim].NIdsInConicLayer >= 0);
      
      // The reason I do not output an error message when NIdsInConicLayer
      // is 0, is that NConicLayers might NOT be zero.  I.e. the user
      // can enter only one resolution, and it might need to be big
      // enough to avoid too many divisions of the latitude, and that
      // may mean it is so big that the above formula evaluates to 0.
      if (ConicLayers[id_of_ConicLayer_in_dim].NIdsInConicLayer == 0)
	ConicLayers[id_of_ConicLayer_in_dim].NIdsInConicLayer = 1;
    }
    else
      ConicLayers[id_of_ConicLayer_in_dim].NIdsInConicLayer = 1;
    
    ConicLayers[id_of_ConicLayer_in_dim].DeltaRadLng = dimensionLongitudeRange
      / ConicLayers[id_of_ConicLayer_in_dim].NIdsInConicLayer;

    ConicLayers[id_of_ConicLayer_in_dim].CumNIdsInConicLayer = nIdsInDim;
      
    nIdsInDim += ConicLayers[id_of_ConicLayer_in_dim].NIdsInConicLayer;
  }
  set_NIdsInDim(nIdsInDim);
}

double SpLatLngDim::Km_radius_of_band
  (const int id_of_ConicLayer_in_dim) const
{
  const double rad_lat_in_middle_of_band
    = id_of_ConicLayer_in_dim_to_RadLatLo(id_of_ConicLayer_in_dim + 0.5);
  
  // This is the radius of a circle made of points having latitude
  // rad_lat_in_middle_of_band.  In general, the center of that
  // circle is NOT the center of the earth, but it is always on
  // the axis of rotation of the earth.
  return get_distance_from_axis
    (rad_lat_in_middle_of_band, radius_of_earth);
}

void SpLatLngDim::get_lo_and_hi
  (SpDataParser &DataParser, char * const Latitude_or_Longitude,
   double &lo, double &hi)
{
  void *DimensionMark = DataParser.GetMark();
  DataParser.Find(DimensionMark, Latitude_or_Longitude);
  lo = DataParser.GetFloat("Lo");
  hi = DataParser.GetFloat("Hi");
  DataParser.SetMark(DimensionMark);

  if (0 == strcmp(Latitude_or_Longitude, "Latitude"))
  {
    DimensionCoversSouthPole_p = (lo == -90);
    DimensionCoversNorthPole_p = (hi == 90);
  }
  else if (0 == strcmp(Latitude_or_Longitude, "Longitude"))
    DimensionCoversAllLng_p = (lo == -180 && hi == 180);
  
  CheckAndChangeAngle(Latitude_or_Longitude, lo);
  CheckAndChangeAngle(Latitude_or_Longitude, hi);
}

void SpLatLngDim::CheckAndChangeAngle
  (const char * const Latitude_or_Longitude, double &angle) const
{
  double max_magnitude;
  
  if (0 == strcmp(Latitude_or_Longitude, "Latitude"))
    max_magnitude = 90;
  else if (0 == strcmp(Latitude_or_Longitude, "Longitude"))
    max_magnitude = 180;
  else
  {
    cerr << "SpLatLngDim::CheckAndChangeAngle was not passed Latitude"
      " or Longitude" << endl;
    abort();
  }

  if (angle > max_magnitude || angle < -max_magnitude)
  {
    cerr << Latitude_or_Longitude << ' ' << angle << " read in dimension "
	 << get_DimName() << " is\nout of range; it should be between "
	 << -max_magnitude << " degrees and " << max_magnitude << "degrees."
	 << endl;
    abort();
  }
  angle = DEG_to_RAD(angle);
}

double SpLatLngDim::DimensionRadLngHi_plus_maybe_two_pi() const
{
  double lng_hi = DimensionRadLngHi;
  if (DimensionRadLngHi < DimensionRadLngLo)
    lng_hi += TWO_PI;
  return lng_hi;
}

double SpLatLngDim::DimensionLongitudeRange() const
{
  return DimensionRadLngHi_plus_maybe_two_pi() - DimensionRadLngLo;
}

void SpLatLngDim::InitSpHiGrid(int IdOfGridInDimension,
				  SpDimInfo DimInfo[MAX_DIMENSIONS],
				  int &number_of_dims_initialized) const
{
  const int
    lat_dim = number_of_dims_initialized,
    lng_dim = number_of_dims_initialized + 1;

  DimInfo[lat_dim].SetDimName(SpHiGrid_DimName("Latitude"));
  DimInfo[lat_dim].SetDimNum(lat_dim);

  DimInfo[lng_dim].SetDimName(SpHiGrid_DimName("Longitude"));
  DimInfo[lng_dim].SetDimNum(lng_dim);

  int
    lo_id_of_ConicLayer_in_dim = 0,
    hi_id_of_ConicLayer_in_dim = NConicLayers - 1,
    mid_id_of_ConicLayer_in_dim;

  // binary search:
  for (;;)
  {
    assert(hi_id_of_ConicLayer_in_dim >= lo_id_of_ConicLayer_in_dim);
    mid_id_of_ConicLayer_in_dim = (lo_id_of_ConicLayer_in_dim +
				   hi_id_of_ConicLayer_in_dim) / 2;
    if (IdOfGridInDimension
	< ConicLayers[mid_id_of_ConicLayer_in_dim].CumNIdsInConicLayer)
      hi_id_of_ConicLayer_in_dim = mid_id_of_ConicLayer_in_dim - 1;
    else if (IdOfGridInDimension
	     >= ConicLayers[mid_id_of_ConicLayer_in_dim].CumNIdsInConicLayer
	     +  ConicLayers[mid_id_of_ConicLayer_in_dim].   NIdsInConicLayer)
      lo_id_of_ConicLayer_in_dim = mid_id_of_ConicLayer_in_dim + 1;
    else
      break;
  }
  assert(ConicLayers[mid_id_of_ConicLayer_in_dim].CumNIdsInConicLayer
	 <= IdOfGridInDimension && IdOfGridInDimension
	 < ConicLayers[mid_id_of_ConicLayer_in_dim].CumNIdsInConicLayer
	 + ConicLayers[mid_id_of_ConicLayer_in_dim].   NIdsInConicLayer);

  DimInfo[lat_dim].SetLo
    (id_of_ConicLayer_in_dim_to_RadLatLo(mid_id_of_ConicLayer_in_dim));
  DimInfo[lat_dim].SetHi
    (id_of_ConicLayer_in_dim_to_RadLatLo(mid_id_of_ConicLayer_in_dim + 1.0));

  const ConicLayer &conicLayer
    = ConicLayers[mid_id_of_ConicLayer_in_dim];

  const int IdOfGridInConicLayer
    = IdOfGridInDimension - conicLayer.CumNIdsInConicLayer;

  const double
    GridRadLngLo =
    DimensionRadLngLo + conicLayer.DeltaRadLng * IdOfGridInConicLayer,
    
    GridRadLngHi =
    DimensionRadLngLo + conicLayer.DeltaRadLng * (IdOfGridInConicLayer + 1);
  
  // There could be a grid which lies on the line of longitude of pi,
  // which means the lo of this grid is less than pi and the hi is
  // more than pi, but longitudes only go up to pi, so does that
  // mean that the hi of this grid should be reduced by 2 * pi?  NO!
  // Because then I would be creating a HiGrid with a lo that is more
  // than its hi.  Even though things like this make sense when dealing
  // with angles, the HiGrid does NOT know we are dealing with angles.
  // It thinks we are dealing with plain old real numbers, so it would
  // be bad to subtract 2 * pi from the hi of a grid.  So this means
  // we might have some grids with a lo in the normal longitude range
  // of -pi to pi, and with a hi out of that range.  And we might have
  // some grids with BOTH the hi and the lo more than pi.  When we
  // *PRINT* the HiGrid statistics out we might want to subtract 2 * pi,
  // but we must *STORE* the hi as a number greater than pi.

  DimInfo[lng_dim].SetLo(GridRadLngLo);
  DimInfo[lng_dim].SetHi(GridRadLngHi);

  const int startingIdOfResInDim = distribute_p() ? 1 : 0;
  
  DimInfo[lat_dim].SetNresolutions(Nresolutions - startingIdOfResInDim);
  DimInfo[lng_dim].SetNresolutions(Nresolutions - startingIdOfResInDim);

  const double km_radius_of_band
    = Km_radius_of_band(mid_id_of_ConicLayer_in_dim);
  
  for (int IdOfResInDim = startingIdOfResInDim;
       IdOfResInDim < Nresolutions;
       IdOfResInDim++)
  {
    DimInfo[lat_dim].SetResolution
      (IdOfResInDim - startingIdOfResInDim,
       KmResolution[IdOfResInDim] / radius_of_earth);

    DimInfo[lng_dim].SetResolution
      (IdOfResInDim - startingIdOfResInDim,
       KmResolution[IdOfResInDim] / km_radius_of_band);
  }
  
  number_of_dims_initialized += 2;
}

static inline int max(int a, int b)
{
  if (a > b)
    return a;
  else
    return b;
}

static inline int min(int a, int b)
{
  if (a < b)
    return a;
  else
    return b;
}

IntersectStatus SpLatLngDim::DoIntersection
  (const int id_of_only_partially_covered_lo_ConicLayer_in_dim,
   const int id_of_only_partially_covered_hi_ConicLayer_in_dim,
   const double lng_lookup_lo,
   const double lng_lookup_hi,
   SpHiGridRegionVec* const ids_of_grids_in_dimension) const
{
  assert(lng_lookup_hi >= lng_lookup_lo);
  
  if (lng_lookup_hi < DimensionRadLngLo)
    return lookup_is_less_than_dimension;
  
  if (lng_lookup_lo > DimensionRadLngHi_plus_maybe_two_pi())
    return lookup_is_more_than_dimension;

  assert(ids_of_grids_in_dimension != NULL);
  
  const int
    intersect_lo_id_of_ConicLayer_in_dim
    = max(id_of_only_partially_covered_lo_ConicLayer_in_dim, 0),

    intersect_hi_id_of_ConicLayer_in_dim
    = min(id_of_only_partially_covered_hi_ConicLayer_in_dim, NConicLayers - 1);

  const int
    /*lat_dim = 0,*/
    lng_dim = 1;

  for (int id_of_ConicLayer_in_dim = intersect_lo_id_of_ConicLayer_in_dim;
       id_of_ConicLayer_in_dim <= intersect_hi_id_of_ConicLayer_in_dim;
       id_of_ConicLayer_in_dim++)
  {
    //   First I will discuss the case of a fully covered ending
    // ConicLayer.  An ending ConicLayer is either the highest
    // numbered ConicLayer intersected by the lookup or it is the
    // lowest numbered ConicLayer intersected by the lookup.  For this
    // case I'm discussing, the lookup border must exactly lie on the border
    // of the ConicLayer.  Here's why: If it lied farther into the
    // ConicLayer, then the ConicLayer would not be fully covered,
    // which contradicts the case I am currently discussing.  If the
    // lookup lied outside of the ConicLayer then it would not be an
    // ending ConicLayer which would contradict the case I am
    // currently discussing.  So the lookup must lie exactly on the
    // border of the ConicLayer.  Now suppose that I use a lookup item
    // that is an infinitesimal amount smaller than it actually is.
    // That would cause the ConicLayer to not be fully covered, but
    // that's OK since I only lowered the lookup size by an
    // infinitesimal amount.  The military does not even know sensor
    // lookup sizes exactly anyway.  Furthermore, an error of using
    // FALSE when I should use TRUE for covered_p is a safe
    // conservative error, since it will only mean extra execution
    // time, but there will never be a subscriber that should be
    // informed but will not be.  Therefore, the following code will
    // record that all fully covered ending ConicLayers are NOT fully
    // covered.
    //   For the case of a genuine not-fully covered ending
    // ConicLayer, obviously the code should record it is not covered.
    //   So for ANY ending ConicLayer, the code should record that it
    // is not fully covered.
    // id_of_only_partially_covered_lo_ConicLayer_in_dim and
    // id_of_only_partially_covered_hi_ConicLayer_in_dim are ending
    // ConicLayers, so if id_of_ConicLayer_in_dim is equal to one of
    // those varibles, then the inequality operator will return FALSE.
    //   Since the middle ConicLayers are not intersected, they must
    // be fully covered, and the code will do this.
    const int lat_covered_p
      = (id_of_only_partially_covered_lo_ConicLayer_in_dim
	 < id_of_ConicLayer_in_dim && id_of_ConicLayer_in_dim
	 < id_of_only_partially_covered_hi_ConicLayer_in_dim);

    const ConicLayer &conicLayer = ConicLayers[id_of_ConicLayer_in_dim];
      
    // The reason the following variable names say
    // "only_partially_covered" is for a similar reason given in
    // the previous comment for why ending ConicLayers are only
    // partially covered.
    const int id_in_ConicLayer_of_only_partially_covered_lo_grid
      = RadLng_to_id_of_grid_in_conicLayer(conicLayer, lng_lookup_lo);
    const int id_in_ConicLayer_of_only_partially_covered_hi_grid
      = RadLng_to_id_of_grid_in_conicLayer(conicLayer, lng_lookup_hi);

    assert(id_in_ConicLayer_of_only_partially_covered_hi_grid >=
	   id_in_ConicLayer_of_only_partially_covered_lo_grid);
  
    const int
      intersect_lo_id_in_ConicLayer
      = max(id_in_ConicLayer_of_only_partially_covered_lo_grid, 0),
      
      intersect_hi_id_in_ConicLayer
      = min(id_in_ConicLayer_of_only_partially_covered_hi_grid,
	    conicLayer.NIdsInConicLayer - 1);

    for (int id_in_ConicLayer = intersect_lo_id_in_ConicLayer;
	 id_in_ConicLayer <= intersect_hi_id_in_ConicLayer;
	 id_in_ConicLayer++)
    {
      const int lng_covered_p
	= (id_in_ConicLayer_of_only_partially_covered_lo_grid
	   < id_in_ConicLayer && id_in_ConicLayer
	   < id_in_ConicLayer_of_only_partially_covered_hi_grid);

      ids_of_grids_in_dimension
	->set_next()
	.set_id(conicLayer.CumNIdsInConicLayer + id_in_ConicLayer)
	.set_gridCovered_p(lat_covered_p && lng_covered_p)
	.set_Lo(lng_dim, lng_lookup_lo)
	.set_Hi(lng_dim, lng_lookup_hi);
    }
  }

  return (DimensionRadLngLo <= lng_lookup_lo
	  && lng_lookup_hi <= DimensionRadLngHi)
    ? lookup_fits_in_dimension
    : lookup_intersects_but_does_not_fit_in_dimension;
}

int SpLatLngDim::HiGrid_Lookup
  (const SpDimLookupItem *dimensionLookupItem,
   SpHiGridRegionVec* &ids_of_grids_in_dimension) const
{
  ids_of_grids_in_dimension = new SpHiGridRegionVec
    (MaximumNumberOfRegions(), get_NumberOfHiGridDims());

  const SpLatLngLookupItem * const latLngLookupItem
    = (const SpLatLngLookupItem *) dimensionLookupItem;

  const double km_distance_from_center_of_lookup_to_center_of_earth
    = radius_of_earth + latLngLookupItem->get_KmAlt();

  double delta_rad_lat, delta_rad_lng;
  
  //   Note that the following function call finds the maximum number
  // of longitude radians that the sphere covers in the eastern
  // direction.  But if the ConicLayer only covers the very top or bottom
  // of the sphere, then within just this ConicLayer the sphere covers a
  // smaller number of longitude radians.  But it might be difficult
  // to calculate this (maybe some day when we have more time we
  // should attempt it) so I just calculate the bigger number which
  // makes HiGrid_Lookup return more grids than it really needs to
  // return, which makes things slower, but at least correct results
  // will still be obtained, since HiGrid_Lookup will still definitely
  // return all grids that are covered.
  //   Actually, it might not speed things up to return less grids,
  // because to figure out how to return less grids, takes time too!
  get_delta_rad_angles_of_sphere
    (latLngLookupItem->get_RadLat(), latLngLookupItem->get_KmSizeRadius(),
     km_distance_from_center_of_lookup_to_center_of_earth,
     delta_rad_lat, delta_rad_lng);

  const double
    lat_lookup_lo
    = latLngLookupItem->get_RadLat() - delta_rad_lat,

    lat_lookup_hi
    = latLngLookupItem->get_RadLat() + delta_rad_lat;

  const int
    id_of_only_partially_covered_lo_ConicLayer_in_dim
    = RadLat_to_id_of_ConicLayer_in_dim(lat_lookup_lo),
    
    id_of_only_partially_covered_hi_ConicLayer_in_dim
    = RadLat_to_id_of_ConicLayer_in_dim(lat_lookup_hi);

  int lookup_fits_in_some_multiple_p;
  if (delta_rad_lng >= PI)
  {
    lookup_fits_in_some_multiple_p
      = (DoIntersection
	 (id_of_only_partially_covered_lo_ConicLayer_in_dim,
	  id_of_only_partially_covered_hi_ConicLayer_in_dim,
	  // 100 is a lot more than 2*pi so this range will definitely
	  // cover everything:
	  -100,
	  100,
	  ids_of_grids_in_dimension) == lookup_fits_in_dimension);
  }
  else
  {
    const double
      lng_lookup_lo = latLngLookupItem->get_RadLng() - delta_rad_lng,
      lng_lookup_hi = latLngLookupItem->get_RadLng() + delta_rad_lng;

    //   When crossing the longitude line of pi (AKA -pi), you go from
    // pi to -pi or from -pi to pi.  That is a discontinuous jump.  The
    // lookup might cover this discontinuous longitude line.  The
    // dimension might also cover this discontinuous longitude line.
    //   This discontinuity gives me headaches, and so in order to not
    // have to think about it, here is what I will do: The discontinuity
    // for the dimension occurs iff (DimensionRadLngHi <
    // DimensionRadLngLo).  So if (DimensionRadLngHi <
    // DimensionRadLngLo), I will add 2*pi to DimensionRadLngHi and will
    // think of this sum as my new longitude hi.  This new longitude hi
    // is more than the lo, which makes things simpler.  Of course now
    // the hi is out of the normal longitude range of -pi to pi, but I
    // can handle that as explained below.
    //   The above code always gives me a lng_lookup_lo less than a
    // lng_lookup_hi, although it does still sometimes give me
    // longitudes out of the normal longitude range.  So now I have two
    // regions, the dimension and the lookup both with hi more than lo.
    // Therefore I am back to the simple case which is almost equivalent
    // to the class SpDoubleDim.  The reason it is only almost
    // instead of completely equivalent is because I am dealing with
    // angles, and two angles might refer to the same place on the earth
    // but yet be different numbers.  I.e. they may differ by an integer
    // multiple of 2*pi.  So two different numbers might refer to the
    // same angle and I need to return the grid in which they lie.
    //   The way that I handle this is to do the ordinary class
    // SpDoubleDim code but do it more than once.  Each time I do it
    // I do it with a different integer multiple of 2*pi added onto the
    // lookup.  This way I can be sure that at least one of the integer
    // multiples will have the same number if there truly is an
    // intersection.  The most number of integer
    // multiples I have to try is 2.  I.e. the most number of times I
    // have to do the SpDoubleDim code is 2.  I won't bother trying
    // to formally prove it, but I do have some asserts that should
    // catch any counter example.  I have to figure out which 2
    // integer multiples are the relevant ones.  I.e. what two integers
    // would I use?  That's also pretty easy.  One integer is 0 and the
    // other is either 1 or -1.  After I figure out which two integers
    // it is, I have code that tries out two more integers, the ones
    // surrounding the two correct one.  For example, if the two correct
    // integers are -1 and 0, then I try out -2 and 1 just to make sure
    // that -2 means that ALL of the lookup is less than ALL of the
    // dimension, and that 1 means that ALL of the lookup is more than
    // ALL of the dimension.

    const double plus_or_minus_two_pi
      = (lng_lookup_hi - DimensionRadLngLo > TWO_PI)
      ? -TWO_PI
      :  TWO_PI;

#ifdef DEBUG_SpaceMan
    cout << "plus_or_minus_two_pi is "
	 << ((plus_or_minus_two_pi > 0) ? "positive" : "negative") << endl;
#endif
      
    const IntersectStatus
      lookup_multiple_IntersectStatus	= DoIntersection
      (id_of_only_partially_covered_lo_ConicLayer_in_dim,
       id_of_only_partially_covered_hi_ConicLayer_in_dim,
       lng_lookup_lo + plus_or_minus_two_pi,
       lng_lookup_hi + plus_or_minus_two_pi,
       ids_of_grids_in_dimension),
	
      lookup_no_multiple_IntersectStatus = DoIntersection
      (id_of_only_partially_covered_lo_ConicLayer_in_dim,
       id_of_only_partially_covered_hi_ConicLayer_in_dim,
       lng_lookup_lo,
       lng_lookup_hi,
       ids_of_grids_in_dimension);
      
#ifdef DEBUG_SpaceMan
    VerifyCorrect_two_pi_Multiples
      (-plus_or_minus_two_pi, lng_lookup_lo, lng_lookup_hi);

    VerifyCorrect_two_pi_Multiples
      (2 * plus_or_minus_two_pi, lng_lookup_lo, lng_lookup_hi);
#endif

    lookup_fits_in_some_multiple_p =
      lookup_fits_in_dimension == lookup_multiple_IntersectStatus ||
      lookup_fits_in_dimension == lookup_no_multiple_IntersectStatus;
  }

  double rad_lat_resolution, rad_lng_resolution;
  
  get_delta_rad_angles_of_sphere
    (latLngLookupItem->get_RadLat(),
     latLngLookupItem->get_KmResolutionRadius(),
     km_distance_from_center_of_lookup_to_center_of_earth,
     rad_lat_resolution, rad_lng_resolution);

  // The multiplications by 2 below are because
  // get_delta_rad_angles_of_sphere return half the number of radians
  // in which the sphere exists, but for resolution, I want the full
  // number of radians.
  rad_lat_resolution *= 2;
  rad_lng_resolution *= 2;

  const int
    lat_dim = 0,
    lng_dim = 1;
  
  SpHiGridRegionVecModifier iterator(ids_of_grids_in_dimension);
  while (++iterator)
  {
    iterator
      .set_DimRegion(lat_dim, lat_lookup_lo, lat_lookup_hi, rad_lat_resolution)
      .set_Res(lng_dim, rad_lng_resolution);
  }
  
  return
    // If lat_lookup_lo is less than the latitude of -pi/2
    // at the south pole then the lookup just goes south, hits the
    // pole, and then wraps around the pole, then continues on North,
    // so if the dimension covers the south pole then the lookup
    // really does NOT run past the southern border of the dimension.
    // A similar argument can be made for the northern border of the
    // dimension.
    (DimensionCoversSouthPole_p || DimensionRadLatLo <= lat_lookup_lo) &&
    (DimensionCoversNorthPole_p || lat_lookup_hi <= DimensionRadLatHi)

    // If during one call to DoIntersection, the lookup ran off of the
    // end of the dimension, it might immediately move into the other
    // end of the dimension.  This could only happen if the
    // dimension's two ends met.  And that could only happen if the
    // dimension wrapped all of the way around the earth.  The
    // DimensionCoversAllLng_p variable says whether or not the
    // dimension wraps all of the way around the earth.  So if
    // DimensionCoversAllLng_p were false then if the lookup ran off
    // of one end of the dimension, it would immediately hit space not
    // covered by the dimension, which means the lookup does not fit
    // into the dimension.  If DimensionCoversAllLng_p were TRUE then
    // there is no longitude that the lookup could cover that's not
    // covered by the dimension, so the longitude part of the lookup
    // would definitely be covered.  The following code expresses all
    // of this.
    && (DimensionCoversAllLng_p	|| lookup_fits_in_some_multiple_p);
}

#ifdef DEBUG_SpaceMan
void SpLatLngDim::VerifyCorrect_two_pi_Multiples
  (const double integer_times_two_pi,
   const double lng_lookup_lo, const double lng_lookup_hi) const
{
  const double
    test_lng_lookup_lo = lng_lookup_lo + integer_times_two_pi,
    test_lng_lookup_hi = lng_lookup_hi + integer_times_two_pi;

  const IntersectStatus expected_IntersectStatus
    =  integer_times_two_pi < 0
    ? lookup_is_less_than_dimension
    : lookup_is_more_than_dimension;

  SpHiGridRegionVec *np = NULL;
  assert(DoIntersection(0, 0, test_lng_lookup_lo, test_lng_lookup_hi, np)
	 == expected_IntersectStatus);
}
#endif

static void s_rad_width(ostream &os, const char *s,
				    double rad, int width)
{
  os << s << setw(width) << RAD_to_DEG(rad) << " degrees";
}

void SpLatLngDim::print_work(ostream &os) const
{
  os << resetiosflags(ios::left);
  s_rad_width(os, "         Latitude:   Lo = ", DimensionRadLatLo, 7);
  s_rad_width(os, ",   Hi = ", DimensionRadLatHi, 7);
  s_rad_width(os, "\n         Longitude:  Lo = ", DimensionRadLngLo, 7);
  s_rad_width(os, ",   Hi = ", DimensionRadLngHi, 7);
  s_rad_width(os, "\n         This Dimension spans ",
	      DimensionRadLatHi - DimensionRadLatLo, 0);
  s_rad_width(os, " of latitude, and ", DimensionLongitudeRange(), 0);
  os << "\n          of longitude."
        "\n         Number of ConicLayers = " << NConicLayers;
  s_rad_width(os, ".  Each Layer spans ", ConicLayerDeltaRadLat, 0);
  os << " of latitude.\n";

  os.setf(ios::left);

  const int
    WIndex = 19,
    WNIdsInConicLayer = 26;

  os << "         "
     << setw(WIndex)       << "ConicLayer index"
     << setw(WNIdsInConicLayer) << "Num grids in ConicLayer"
     << "Longitude degrees spanned"
     << '\n';

  for (int id_of_ConicLayer_in_dim = 0;
       id_of_ConicLayer_in_dim < NConicLayers;
       id_of_ConicLayer_in_dim++)
  {
    os << "         " << setw(WIndex) << id_of_ConicLayer_in_dim
       << setw(WNIdsInConicLayer)
       << ConicLayers[id_of_ConicLayer_in_dim].NIdsInConicLayer
       << RAD_to_DEG(ConicLayers[id_of_ConicLayer_in_dim].DeltaRadLng)
       << '\n';
  }
  
  os << "         Resolutions (kilometers): [";

  for (int k=0; k<Nresolutions; k++) {
    os << KmResolution[k];
    if (k != Nresolutions-1) os << ", ";
  }

  os << "]" << endl;
}

// Conversion functions:

double SpLatLngDim::id_of_ConicLayer_in_dim_to_RadLatLo
  (const double id_of_ConicLayer_in_dim) const
{
  return DimensionRadLatLo + id_of_ConicLayer_in_dim * ConicLayerDeltaRadLat;
}

inline int dimension_point_to_id(const double dimension_point,
				 const double dimension_lo,
				 const double grid_size)
{
  assert(grid_size > 0);
  
  const double diff = dimension_point - dimension_lo;
 
  if (diff < 0)
    return -1; // this is needed since int(-.8) == 0
  else
    return int(diff / grid_size);
}

int SpLatLngDim::RadLat_to_id_of_ConicLayer_in_dim
  (const double rad_lat) const
{
  return
    dimension_point_to_id(rad_lat, DimensionRadLatLo, ConicLayerDeltaRadLat);
}

int SpLatLngDim::RadLng_to_id_of_grid_in_conicLayer
  (const ConicLayer &conicLayer, const double rad_lng) const
{
  return
    dimension_point_to_id(rad_lng, DimensionRadLngLo, conicLayer.DeltaRadLng);
}
