// SpExportAttribute.H Header file

#ifndef Attribute_Object
#define Attribute_Object

#include "SpAssert.H"
#include "xlist.H"
#include "SpDynItem.H"
#include "SpProxyDefines.H"
#include "SpConvert.H"
#include "SpObjProxyMsg.H"

#include "RB_Types.H"
#include "RB_UpdateXqueue.H" //proxy: add

/*************************************************************************
* enum: ATTRIBUTE_TYPE
* Summary: enums of different attribute type
* Description: each attribute type must be enumerated so that we can
*              identify them by lookup.  The last enum must be
*              N_ATTRIBUTE_TYPES so we know how many there are
*************************************************************************/
enum ATTRIBUTE_TYPE {

  STATIC_INT_ATTRIBUTE_TYPE,      // attribute type of integer
  STATIC_FLOAT_ATTRIBUTE_TYPE,    // attribute type of float (doubles)
  STATIC_LOGICAL_ATTRIBUTE_TYPE,  // attribute type of logicals (booleans)
  STATIC_POSITION_ATTRIBUTE_TYPE, // attribute type of static positions

  DYNAMIC_INT_ATTRIBUTE_TYPE,     // attribute type of dynamic (time dependant) int
  DYNAMIC_FLOAT_ATTRIBUTE_TYPE,   // attribute type of dynamic (time dependant) floats
  DYNAMIC_LOGICAL_ATTRIBUTE_TYPE, // attribute type of dynamic (time dependant) logical
  DYNAMIC_OBJECT_ATTRIBUTE_TYPE,  // attribute type of dynamic (time dependant) object
  DYNAMIC_POSITION_ATTRIBUTE_TYPE,// attribute type of dynamic (time dependant) position

  STRING_ATTRIBUTE_TYPE,          // attribute type of string

  OBJECT_ATTRIBUTE_TYPE,          // attribute type of objects
  LIST_ATTRIBUTE_TYPE,            // attribute type for lists

  N_ATTRIBUTE_TYPES               // total number of attribute types.

};

ostream & operator << (ostream &os, const ATTRIBUTE_TYPE type);

enum {
  LOGICAL_FALSE=0,
  LOGICAL_TRUE=1
};
//static const int LOGICAL_TRUE=1;
//static const int LOGICAL_FALSE=0;

int MapAttribute(char *name);
char *MapAttribute(int index);

class SpObjProxy;
class LIST_ATTRIBUTE;
class OBJECT_ATTRIBUTE;

union ATTRIBUTE_LIST_ELEMENT 
{
    ATTRIBUTE_LIST_ELEMENT ():ObjectAttribute(NULL) {}
    ATTRIBUTE_LIST_ELEMENT (OBJECT_ATTRIBUTE *p) : ObjectAttribute(p) {}
    ATTRIBUTE_LIST_ELEMENT (SpDynItem *p) : DynamicItem(p) {}
    OBJECT_ATTRIBUTE *ObjectAttribute;
    SpDynItem *DynamicItem;
    int PurifyPaddingToRemoveUMR;
};

/************************************************************************
 * Class:BASE_ATTRIBUTE
 *
 * Summary: BASE_ATTRIBUTE is the class out of which all attribute
 *          classes are derived.
 *
 * Description: *_ATTRIBUTE are the objects within your SIMOBJ which
 *               have the ability to be monitored by other classes
 *               through object proxies.  These objects are inherently
 *               roll backable.  You should never use this class
 *               directly but should use the derived classes such as
 *               INT_ATTRIBUTE, DYNAMIC_ATTRIBUTE, STRING_ATTRIBUTE
 *               and so on.
 *
 *               In general, you should treat attribute object just
 *               like * you would treat any STATE_* members.  They are
 *               fully * roll-backable and have the added feature of
 *               being * publishable.  The two main derived types of
 *               the * BASE_ATTRIBUTE class are STATIC_*_ATTRIBUTE and
 *               * DYNAMIC_*_ATTRIBUTE. STATIC_*_ATTRIBUTE are
 *               attributes * that have a single non-time dependent *
 *               value. DYNAMIC_*_ATTRIBUTE types have values at
 *               different * points in time.  You can query a STATIC
 *               version by simply * asking for its value.  For a
 *               DYNAMIC type you have to * supply it a time at which
 *               you want to know the value
 *
 *
 * Examples: The best present examples of the use of these attribute
 *           types is in demos/BmdSim.  Many of the sim objects in
 *           this demo use attribute classes in order to broadcast
 *           information.  In particular, the RanMot class uses a
 *           DYNAMIC_ATTRIBUTE in order to change motion from one
 *           great circle to another at given points in time.  in the
 *           file S_RanMot.C, you can see in the initmotion() method
 *           how position (the DYNAMIC_POSITION_ATTRIBUTE) is modified.
 *
************************************************************************/
class BASE_ATTRIBUTE {

public:



  /************************************************************************
   * Method: BASE_ATTRIBUTE()
   * Summary: Default BASE_ATTRIBUTE constructor.
   * Description: 
   * Return: a new BASE_ATTRIBUTE
   *************************************************************************/
  BASE_ATTRIBUTE() {
    AttributeIndex = 0;
    ArrayIndex = 0;
    InObjectProxy = NULL;
  }



  /************************************************************************
   * Method: ~BASE_ATTRIBUTE()
   * Summary: BASE_ATTRIBUTE destructor
   * Description: 
   * Return: none
   *************************************************************************/
  virtual ~BASE_ATTRIBUTE() {;}



  /************************************************************************
   * Method: ScheduleUpdate
   * Summary: Schedules the update of an attribute
   * Description: This is the guts method that actually reflects the
   *              update of an attribute type.  Many attributes such as
   *              INT_ATTRIBUTE can be modified through operator
   *              overloading.  This overloaded function usually calls
   *              ScheduleUpdate to announce the updated values to the
   *              other objects and their object proxies through an
   *              UpdateAttribute event.  This is not a user level
   *              function.
   * Return: void.
   *************************************************************************/
  void ScheduleUpdate
    (ADD_REMOVE,
     ATTRIBUTE_LIST_ELEMENT  //InOut: Element to be added/removed
                                    //from the attribute
     );

  void ScheduleUpdate(){ScheduleUpdate(NEITHER,(OBJECT_ATTRIBUTE*)NULL);}

  /************************************************************************
   * Method: SetAttributeIndex(int)
   * Summary: Sets the attribute index
   * Description: 
   * Return: void
   *************************************************************************/
  void SetAttributeIndex(int i //In: AttributeIndex is set to this value
			 ) {AttributeIndex = i;}


  /************************************************************************
   * Method: GetAttributeIndex()
   * Summary: Returns the attribute index
   * Description: 
   * Return: Returns the attribute index
   *************************************************************************/
  int GetAttributeIndex() {return AttributeIndex;}



  /************************************************************************
   * Method: SetArrayIndex(int)
   * Summary: Sets the array index
   * Description: 
   * Return: void
   *************************************************************************/
  void SetArrayIndex(int i //In: ArrayIndex is set to this value
		     ) {ArrayIndex = i;}


  /************************************************************************
   * Method: GetArrayIndex()
   * Summary: Returns the array index
   * Description: 
   * Return: Returns the array index
   *************************************************************************/
  int GetArrayIndex() {return ArrayIndex;}



  /************************************************************************
   * Method: SetInObjectProxy(SpObjProxy*)
   * Summary: Sets the InObjectProxy to the passed value o
   * Description: 
   * Return: 
   *************************************************************************/
  virtual void SetInObjectProxy(SpObjProxy *o //In: InObjectProxy is set
				//    to this value
				) {InObjectProxy = o;}


  /************************************************************************
   * Method: GetInObjectProxy()
   * Summary: Returns the InObjectProxy
   * Description: 
   * Return: Returns the InObjectProxy
   *************************************************************************/
  SpObjProxy *GetInObjectProxy() {return InObjectProxy;}



  /************************************************************************
   * Method: GetType()
   * Summary: Pure virtual function that returns the type of attribute
   * Description: Each derived attribute type must implement this method
   *              so that that type can be queried as to it's enumerated
   *              type.all of these types are enumerated at the top of
   *              this header file
   * Return: the type of attribute on which this method is called
   *************************************************************************/
  virtual int GetType() = 0;


  /************************************************************************
   * Method: GetSize()
   * Summary: Pure virtual function that returns the size of the attribute
   * Description: Each derived attribute type must implement this method
   *              so that that type can be queried as to it's size.  The
   *              reason this is important is that we may need to send
   *              this type to another node and, therefore, need to know
   *              it's size so that we can pack the message.
   * Return: the size of the attribute on which this method is called
   *************************************************************************/
  virtual int GetSize() = 0;

  // Arrays of children are bad, but we do not have time to change
  // this program to get rid of that problem, so this function
  // is the solution.  This function should be defined by all
  // descendents which are in arrays in some part of the program.
  // To delete the array of children call this function on the first
  // element (element index 0).


  /************************************************************************
   * Method: delete_vector_with_child_type
   * Summary: Pure virtual that deletes a vector of objects using the child type.
   * Description: Arrays of children are bad, but we do not have time to
   *              change this program to get rid of that problem, so
   *              this function is the solution.  This function should
   *              be defined by all descendents which are in arrays in
   *              some part of the program.  To delete the array of
   *              children call this function on the first ement
   *              (element index 0).
   * Return: void
   *************************************************************************/
  virtual void delete_vector_with_child_type() = 0;



  /************************************************************************
   * Method: outstream(ostream&)
   * Summary: prints 0 to the outstream.  I don't know if this is still needed
   * Description: 
   * Return: void
   *************************************************************************/
  void outstream(ostream &out //In: Stream to which we will print
		 ) {outstream(out, 0);}


  /************************************************************************
   * Method: outstream(ostream&, int)
   * Summary: Pure virtual for printing an attribute to a stream
   * Description: Each attribute must be printable to a stream.  The
   *              first parameter is the stream to which we want to
   *              print while the second parameter is a "level" of
   *              verbosness for which we should print.
   * Return: void
   *************************************************************************/
  virtual void outstream(ostream &out, //In: Stream to which we should print
			 int           //In: Verbocity for the printing
			 ) {out << "Undefined";}


  /************************************************************************
   * Method: GetMessageSize()
   * Summary: Returns the size of the object proxy message size
   * Description: returns the size of the Object Proxy Message
   *              size. This is used for packing an object proxy into *
   *              a message for transmision.  For example, a ATTRIBUTE_INT
   *              is actually significantly larger than the data that
   *              needs to be transmitted (a single int)
   * Return: size of object to be transmitted.
   *************************************************************************/
  virtual int GetMessageSize() {return 0;}

  /************************************************************************
   * Method: FillMessage
   * Summary: packs a message with Object Proxy data
   * Description: char * is a buffer that is assumed to contain enough
   *              room for the entire info that needs to be sent.  This
   *              info is then extracted at the destination using the
   *              EmptyMessage method call.
   * Return: void
   *************************************************************************/
  virtual void FillMessage(char * //InOut: Buffer to be filled with data
			   ) {
    cout << "Error (BASE_ATTRIBUTE): Fill Message should never be called"
	 << endl;
  }
  /************************************************************************
   * Method: EmptyMessage
   * Summary: removes updated object proxy info from a message
   * Description: When we receive an update of object proxy info, it
   *              comes in a char string that was filled by FillMessage.
   *              We extract the appropriate info and update our values.
   * Return: void
   *************************************************************************/
  virtual void EmptyMessage(const char * //In: buffer containing updated attributes
			    ) {
    cout << "Error (BASE_ATTRIBUTE): Empty Message should never be called"
	 << endl;
  }


  /************************************************************************
   * Method: GetDepth
   * Summary: Determines the depth in terms of object proxies
   * Description: 
   * Return: returns the depth in terms of object proxies
   *************************************************************************/
  virtual int GetDepth();


  /************************************************************************
   * Method: PackLevelInfo
   * Summary: packs level info to be sent to another node
   * Description: packs up the level info buffer that is used for
   *              sending updates out to other nodes. It gives the info
   *              needed for finding the right object proxy to modify
   *              for the attribute
   * Return: pointer to the ProxyLevelInfo with the char * Buff passed in
   *************************************************************************/
  virtual char *PackLevelInfo(const char *Buff,   //InOut: buffer to be packed
			      //       and sent
			      int NlevelInfo//in: In: Level of the info.
			      );

  /************************************************************************
   * Method: Copy(BASE_ATTRIBUTE *)
   * Summary: Copies the attribute information from the passed in parameter
   *          onto this.
   * Description: 
   * Return: void
   *************************************************************************/
  virtual void Copy(BASE_ATTRIBUTE *a) {
    assert(this->GetType() == a->GetType());
    memcpy((char *)this, (char *)a, GetSize());
  }


protected:

  int AttributeIndex;
  int ArrayIndex;
  SpObjProxy *InObjectProxy;

private:
};

/************************************************************************
 * Class: STATIC_INT_ATTRIBUTE
 *
 * Summary: publishable integer attribute for use in object proxies
 *          
 * Description: If you want to publish non time dependant integer data
 *              about your object, this is the class to use.  Whenever
 *              this class is updated in value, that update is then
 *              sent to everyone who is subscribing to your object
 *              proxy.  These updates are automatically taken care of
 *              through the use of operator overloading.
 *
 * Examples: An example of the use of STATIC_INT_ATTRIBUTE can be found
 *           in S_RanMot.H in demos/BmdSim.
************************************************************************/
class STATIC_INT_ATTRIBUTE : public BASE_ATTRIBUTE {

public:

  /*************************************************************************
   * Method: STATIC_INT_ATTRIBUTE()
   * Summary: default constructor for a STATIC_INT_ATTRIBUTE
   * Description: The default constructor sets the value of an attribute
   *              to zero.  Notice that C++ does not initialize local
   *              variables or those allocated by new.  If you design
   *              your code assuming these objects are initialized to
   *              zero and then rewrite your code using normal integers,
   *              you could get different behavior so be careful of this.
   * Return: a new STATIC_INT_ATTRIBUTE
   *************************************************************************/
  STATIC_INT_ATTRIBUTE() {attribute = 0;}


  /*************************************************************************
   * Method: GetValue()
   * Summary: Returns the integer value of this attribute
   * Description: returns the integer value of the attribute
   * Return: Returns the integer value of this attribute
   *************************************************************************/
  int GetValue() {return attribute;}



  /*************************************************************************
   * Method: operator int()
   * Summary: Returns the integer value of this attribute
   * Description: Returns the integer value of this attribute
   * Return: Returns the integer value of this attribute
   *************************************************************************/
  operator int() const {return attribute;}


  /*************************************************************************
   * Method: assign(int)
   * Summary: updates the value of the attribute
   * Description: This schedules an update for all the subscribed object
   *              proxies and then updates the value.  This new value is
   *              then returned.
   * Return: 
   *************************************************************************/
  int assign(int a //In: New value for the attribute
	     ) {
    ScheduleUpdate();
    return (attribute = a);
  }



  /*************************************************************************
   * Method: operator = ()
   * Summary: assigns the integer value to this attribute
   * Description: This schedules an update for all the subscribed object
   *              proxies and then updates the value.  This new value is
   *              then returned.
   * Return: the new integer value
   *************************************************************************/
  int operator = (int a) {return assign(a);}


  /*************************************************************************
   * Method: operator ++()
   * Summary: Increments the value of a value and returns the new value
   * Description: Increments the value of a value and returns the new value.
   *              Through the assign call, the object proxes are updated as well
   * Return: the updated value (attribute + 1)
   *************************************************************************/
  int operator ++ () {return assign(attribute + 1);}


  /*************************************************************************
   * Method: operator ++()
   * Summary: Increments the value of a value and returns the old value
   * Description: Increments the value of a value and returns the old
   *              value.  Through the assign call, the object proxes are
   *              updated as well.  Notice that this is the postfix
   *              operator
   * Return: the old value of the attribute
   *************************************************************************/
  int operator ++ (int) {
    int v = attribute; assign(attribute + 1); return v;}


  /*************************************************************************
   * Method: operator --()
   * Summary: Decrements the value of a value and returns the new value
   * Description: Increments the value of a value and returns the new
   *              value.  Through the assign call, the object proxes are
   *              updated as well.  Notice that this is the prefix
   *              operator
   * Return: the new value (attribute - 1)
   *************************************************************************/
  int operator -- () {return assign(attribute - 1);}


  /*************************************************************************
   * Method: operator --()
   * Summary: Decrements the value of a value and returns the old value
   * Description: Increments the value of a value and returns the old
   *              value.  Through the assign call, the object proxes are
   *              updated as well.  Notice that this is the postfix
   *              operator
   * Return: the current value (attribute)
   *************************************************************************/
  int operator -- (int) {
    int v = attribute; assign(attribute - 1); return v;}

  /*************************************************************************
   * Method: operator +=(int)
   * Summary: Increments attribute by the passed value and returns new value
   * Description: Increments attribute by the passed value and returns
   *              new value Through the assign call, the object proxes
   *              are updated as well.
   * Return: the new value (attribute + a)
   *************************************************************************/
  int operator += (int a //In: value to be added to attribute
		   ) {return assign(attribute + a);}


  /*************************************************************************
   * Method: operator -=(int)
   * Summary: Decrements attribute by the passed value and returns new value
   * Description: Decrements attribute by the passed value and returns
   *              new value Through the assign call, the object proxes
   *              are updated as well.
   * Return: 
   *************************************************************************/
  int operator -= (int a //In: value to be subtracted from attribute
		   ) {return assign(attribute - a);}


  /*************************************************************************
   * Method: operator *=(int)
   * Summary: Multiplies attribute by the passed value and returns new value
   * Description: Multiplies attribute by the passed value and returns
   *              new value Through the assign call, the object proxes
   *              are updated as well. 
   * Return: The attribute multiplied by the passed in value
   *************************************************************************/
  int operator *= (int a //In:  amount by which to multiply the attribute
		   ) {return assign(attribute * a);}


  /*************************************************************************
   * Method: operator /=(int)
   * Summary: Divides attribute by the passed value and returns new value
   * Description: Divides attribute by the passed value and returns
   *              new value Through the assign call, the object proxes
   *              are updated as well. 
   * Return: The attribute divided by the passed in value
   *************************************************************************/
  int operator /= (int a //In: ammount by which to divide the attribute
		   ) {return assign(attribute / a);}


  /*************************************************************************
   * Method: operator %=(int)
   * Summary: Multiplies attribute mod the passed value and returns new value
   * Description: Multiplies attribute mod the passed value and returns
   *              new value Through the assign call, the object proxes
   *              are updated as well. 
   * Return: The attribute mod the passed in value
   *************************************************************************/
  int operator %= (int a //In: Amount by which to mod the attribute
		   ) {return assign(attribute % a);}


  /*************************************************************************
   * Method: operator >>=(int)
   * Summary: Right shifts attribute by the passed value and returns new value
   * Description: Right shifts attribute by the passed value and returns
   *              new value Through the assign call, the object proxes
   *              are updated as well. 
   * Return: Value of the attribute after being shifted right.
   *************************************************************************/
  int operator >>= (int a //In: ammount to shift attribute
		    ) {return assign(attribute >> a);}


  /*************************************************************************
   * Method: operator <<=(int)
   * Summary: shifts the attribute left by passed in vale and return new value
   * Description: left shifts attribute by the passed value and returns
   *              new value Through the assign call, the object proxes
   *              are updated as well.  
   * Return: Value of attribute after shifted left by passed in value
   *************************************************************************/
  int operator <<= (int a //In: ammount to shift attribute left
		    ) {return assign(attribute << a);}


  /*************************************************************************
   * Method: operator ^=(int)
   * Summary: not's attribute with passed in value and returns new value
   * Description: not's attribute with the passed value and returns
   *              new value Through the assign call, the object proxes
   *              are updated as well.  
   * Return: Value of attribute after being not'ed with the passed in value
   *************************************************************************/
  int operator ^= (int a //In: Value with which to 'not' the attribute
		   ) {return assign(attribute ^ a);}


  /*************************************************************************
   * Method: operator &=(int)
   * Summary: and's attribute with passed in value and returns new value
   * Description: and's attribute with the passed value and returns
   *              new value Through the assign call, the object proxes
   *              are updated as well.  
   * Return: Value of attribute after being and'ed with the passed in value
   *************************************************************************/
  int operator &= (int a //In: Value with which to and the attribute
		   ) {return assign(attribute & a);}


  /*************************************************************************
   * Method: operator |=(int)
   * Summary: or's attribute with the passed value and returns new value
   * Description: or's attribute with the passed value and returns
   *              new value Through the assign call, the object proxes
   *              are updated as well
   * Return: value of attribute after being or'ed with the passed value
   *************************************************************************/
  int operator |= (int a //In: value with which to or the attribute
		   ) {return assign(attribute | a);}



  /*************************************************************************
   * Method: GetType()
   * Summary: Returns the type of this attribute: STATIC_INT_ATTRIBUTE_TYPE
   * Description:Returns the type of this attribute: STATIC_INT_ATTRIBUTE_TYPE
   * Return: Returns the type of this attribute: STATIC_INT_ATTRIBUTE_TYPE
   *************************************************************************/
  virtual int GetType() {return STATIC_INT_ATTRIBUTE_TYPE;}


  /*************************************************************************
   * Method: GetSize()
   * Summary: Returns the size of a STATIC_INT_ATTRIBUTE
   * Description:Returns the size of a STATIC_INT_ATTRIBUTE
   * Return: Returns the size of a STATIC_INT_ATTRIBUTE
   *************************************************************************/
  virtual int GetSize() {return sizeof(STATIC_INT_ATTRIBUTE);}


  /*************************************************************************
   * Method: outstream(ostream&, int)
   * Summary: prints our attribute to the passed in stream
   * Description: prints our attribute to the passed in stream
   * Return: void
   *************************************************************************/
  virtual void outstream(ostream &out, //In: stream to which we should print
			 int) {out << attribute;}



  /*************************************************************************
   * Method: GetMessageSizeStatic()
   * Summary: Returns the size of an INT_ATTRIBUTE message
   * Description: Returns the size of an INT_ATTRIBUTE message
   * Return: Returns the size of an INT_ATTRIBUTE message
   *************************************************************************/
  static int GetMessageSizeStatic() {return sizeof(INT_ATTRIBUTE_MESSAGE_INFO);}


  /*************************************************************************
   * Method: FillMessage
   * Summary: Fills passed in buffer with a INT_ATTRIBUTE_MESSAGE
   * Description: Fills passed in buffer with a INT_ATTRIBUTE_MESSAGE
   * Return: void
   *************************************************************************/
  virtual void FillMessage(char *buff  //InOut: Buffer to contain the int
			               //       attribute message
			   ) { FillMessageWith(buff, attribute); }


  /*************************************************************************
   * Method: EmptyMessage
   * Summary: Removes int attribute data from the passed in buff (message)
   * Description: Removes int attribute data from the passed in buff
   *              (message).  This data is then used to update the
   *              values of our int attribute
   * Return: void
   *************************************************************************/
  virtual void EmptyMessage(const char *buff);

  virtual int GetMessageSize() {return GetMessageSizeStatic();}

  static void FillMessageWith(char *buff, int value)
  { ((INT_ATTRIBUTE_MESSAGE_INFO *)buff)->SetValue(value); }
    
protected:

private:

  void delete_vector_with_child_type() { delete [] this; }

  STATE_INT attribute;

};

typedef STATIC_INT_ATTRIBUTE INT_ATTRIBUTE;

/************************************************************************
 * Class: STATIC_FLOAT_ATTRIBUTE
 *
 * Summary: publishable float (double) attribute for use in object proxies
 *          
 * Description: If you want to publish non time dependant float data
 *              about your object, this is the class to use.  Whenever
 *              this class is updated in value, that update is then
 *              sent to everyone who is subscribing to your object
 *              proxy.  These updates are automatically taken care of
 *              through the use of operator overloading.
 *
 * Examples: An example of the use of STATIC_FLOAT_ATTRIBUTE can be found
 *           in S_RanMot.H in demos/BmdSim.
************************************************************************/
class STATIC_FLOAT_ATTRIBUTE : public BASE_ATTRIBUTE {

public:

  /*************************************************************************
   * Method: STATIC_FLOAT_ATTRIBUTE()
   * Summary: Standard constructor for a STATIC_FLOAT_ATTRIBUTE
   * Description: The standard constructor initializes the value to
   *              0. You should be aware of this when converting
   *              SPEEDES code to non SPEEDES code.  The default
   *              constructor for a double (float) does not initialize
   *              values to 0.
   * Return: 
   *************************************************************************/
  STATIC_FLOAT_ATTRIBUTE() {attribute = 0.0;}

  /*************************************************************************
   * Method: GetValue()
   * Summary: returns the value of the float attribute
   * Description: returns the value of the float attribute
   * Return: returns the value of the float attribute
   *************************************************************************/
  double GetValue() {return attribute;}

  /*************************************************************************
   * Method: operator double ()
   * Summary: returns the double value of the float attribute.
   * Description: This operator allows the STATIC_FLOAT_ATTRIBUTE to behave
   *              like a normal double.
   * Return: the double value of the float attribute
   *************************************************************************/
  operator double() const {return attribute;}

  /*************************************************************************
   * Method: assign(double)
   * Summary: schedules an update for every object proxy and updates the value
   * Description: Whenever the value of a STATIC_INT_ATTRIBUTE is
   *              changed, this method is called which schedules
   *              updates for every object proxy in the simulation and
   *              then changes the value of the attribute.
   * Return: the new value of the attribute
   *************************************************************************/
  double assign(double a //In: the new value for the attribute
		) {
    ScheduleUpdate();
    return (attribute = a);
  }


  /*************************************************************************
   * Method: operator =(double)
   * Summary: assigns a double value to a float attribute
   * Description: Through operator overloading, we can update all the
   *              object proxies as well as change the value of the
   *              actual variable.
   * Return: The new value to allow for chaining of assignments
   *************************************************************************/
  double operator = (double a //In: the new value for the attribute
		     ) {return assign(a);}

  /*************************************************************************
   * Method: operator ++()
   * Summary: prefix increment operator
   * Description: prefix increments the value of the float attribute and
   *              sends this new value to the object proxies.
   * Return: the incremented value of the attribute
   *************************************************************************/
  double operator ++ () {return assign(attribute + 1.0);}

  /*************************************************************************
   * Method: operator ++(int)
   * Summary: postfix increment operator
   * Description: postfix increments the value of the float attribute and
   *              sends this new value to the object proxies.
   * Return: the old value of the attribute
   *************************************************************************/
  double operator ++ (int) {
    double v = attribute; assign(attribute + 1.0); return v;}

  /*************************************************************************
   * Method: operator --()
   * Summary: prefix decrement operator
   * Description: prefix decrements the value of hte float attribute and
   *              sends this new value to the object proxies.
   * Return: the decremented value of the attribute.
   *************************************************************************/
  double operator -- () {return assign(attribute - 1.0);}

  /*************************************************************************
   * Method: operator --(int)
   * Summary: postfix decrement operator
   * Description: postfix decrements the value of hte float attribute and
   *              sends this new value to the object proxies
   * Return: the old value of hte attribute.
   *************************************************************************/
  double operator -- (int) {
    double v = attribute; assign(attribute - 1.0); return v;}

  /*************************************************************************
   * Method: operator +=(double)
   * Summary: increment value of attribute by passed in value
   * Description: this operator increments the value of the float attribute
   *              by the passed in value and returns this new value.  This
   *              new value is also sent to all the object proxies.
   * Return: the new value of the float attribute
   *************************************************************************/
  double operator += (double a //In: The amount by which to increase
		               //    the attribute
		      ) {return assign(attribute + a);}

  /*************************************************************************
   * Method: operator -=(double)
   * Summary: decrements the value of the attribute by the passed in value
   * Description: This operator decrements the value of the float attribute
   *              by the passed in value and then returns the new value.
   *              This new value is then sent to all the object proxies.
   * Return: the new value of the float attribute.
   *************************************************************************/
  double operator -= (double a //In:: The amount by which to increase
		               //    the attribute
		      ) {return assign(attribute - a);}

  /*************************************************************************
   * Method: operator *=(double))
   * Summary: multiplies the value of the attribute by the passed in value
   * Description: This operator multiplies the value of the float attribute
   *              by the passed in value and then returns the new value.
   *              This new value is then sent to all the object proxies
   * Return: the new value of the float attribute
   *************************************************************************/
  double operator *= (double a) {return assign(attribute * a);}

  /*************************************************************************
   * Method: operator /=(double)
   * Summary: divides the value of the attribute by the passed in value
   * Description: This operator divides the value of the float attribute
   *              by the passed in value and then returns the new value.
   *              This new value is then sent to all the object proxies
   * Return: the new value of the float attribute
   *************************************************************************/
  double operator /= (double a) {return assign(attribute / a);}


  /*************************************************************************
   * Method: GetType()
   * Summary: returns the type of this attribute: STATIC_FLOAT_ATTRIBUTE_TYPE
   * Description: Each attribute type is defined by a differnt enum so that
   *              we can construct them from buffers which can tell us
   *              their type.
   * Return: returns the type of this attribute: STATIC_FLOAT_ATTRIBUTE_TYPE
   *************************************************************************/
  virtual int GetType() {return STATIC_FLOAT_ATTRIBUTE_TYPE;}

  /*************************************************************************
   * Method: GetSize()
   * Summary: returns the size of this attribute
   * Description: 
   * Return: returns the size of this attribute
   *************************************************************************/
  virtual int GetSize() {return sizeof(STATIC_FLOAT_ATTRIBUTE);}

  /*************************************************************************
   * Method: outstream(ostream &, int)
   * Summary: prints this attribute to the passed in stream
   * Description: 
   * Return: void
   *************************************************************************/
  virtual void outstream(ostream &out, int) {out << attribute;}


  /*************************************************************************
   * Method: GetMessageSize()
   * Summary: returns the size of a FLOAT_ATTRIBUTE_MESSAGE
   * Description: When sending messages, we want to know how big they are.
   *              this method allows us to get the size of the message.
   * Return: returns the size of a FLOAT_ATTRIBUTE_MESSAGE
   *************************************************************************/
  virtual int GetMessageSize() {return sizeof(FLOAT_ATTRIBUTE_MESSAGE_INFO);}

  /*************************************************************************
   * Method: FillMessage
   * Summary: Fills the buffer buff with a float attribute message
   * Description: When an attribute is updated, we want to announce
   *              this new value to the object proxies.  This is done
   *              through a message and this method fills that message
   * Return: void
   *************************************************************************/
  virtual void FillMessage(char *buff  //InOut: Buffer to contain the message
			   );

  /*************************************************************************
   * Method: EmptyMessage
   * Summary: Removes the float attribute data and updates our value
   * Description: When an attribute within an object proxy is changed,
   *              a message is sent to everyone, announcing this change.
   *              this method is how we extract the data
   * Return: void
   *************************************************************************/
  virtual void EmptyMessage(const char *buff //In: The buffer containing the new
			               //    attribute information
			    );

protected:

private:
  void delete_vector_with_child_type() { delete [] this; }

  STATE_DOUBLE attribute;

};

typedef STATIC_FLOAT_ATTRIBUTE FLOAT_ATTRIBUTE;

/************************************************************************
 * Class: STATIC_LOGICAL_ATTRIBUTE
 *
 * Summary: publishable logical attribute for use in object proxies
 *          
 * Description: If you want to publish non time dependant logical data
 *              about your object, this is the class to use.  Whenever
 *              this class is updated in value, that update is then
 *              sent to everyone who is subscribing to your object
 *              proxy.  These updates are automatically taken care of
 *              through the use of operator overloading.
 *
 * Examples: An example of the use of STATIC_LOGICAL_ATTRIBUTE can be
 *           found in S_RanMot.H in demos/BmdSim.
************************************************************************/
class STATIC_LOGICAL_ATTRIBUTE : public BASE_ATTRIBUTE {
public:


  /*************************************************************************
   * Method: STATIC_LOGICAL_ATTRIBUTE()
   * Summary: Default constructor for a STATIC_LOGICAL_ATTRIBUTE
   * Description: The default constructor initializes the value of the
   *              logical to true. Notice that this may not be the case for
   *              logicals in non-SPEEDES code.
   * Return: a new STATIC_LOGICAL_ATTRIBUTE
   *************************************************************************/
  STATIC_LOGICAL_ATTRIBUTE() {attribute = LOGICAL_FALSE;}

  /*************************************************************************
   * Method: GetValue()
   * Summary: Returns the value of the attribute
   * Description: Returns the value of the attribute
   * Return: Returns the value of the attribute
   *************************************************************************/
  int GetValue() {return attribute;}


  /*************************************************************************
   * Method: operator int()
   * Summary: integer constructor from an attribute.
   * Description: This operator allows a STATIC_LOGICAL_ATTRIBUTE to
   *              behave exactly like a normal boolean value.  That is, it
   *              returns an integer value which can be tested against 0
   *              for truth or false.
   * Return: the integer value of the attribute
   *************************************************************************/
  operator int() const {return attribute;}

  /*************************************************************************
   * Method: assign(int)
   * Summary: updates the value of a STATIC_LOGICAL_ATTRIBUTE
   * Description: Whenever a STATIC_LOGICAL_ATTRIBUTE is updated
   *              through any public method, this method is called to
   *              change the value of the attribute as well as scheule
   *              all the object proxies to be updated
   * Return: the new value of the attribute
   *************************************************************************/
  int assign(int a //In: the new value of the attribute
	     ) {
    ScheduleUpdate();
    return (attribute = a);
  }


  /*************************************************************************
   * Method: operator =(int)
   * Summary: assigns the passed in value to the attribute
   * Description: Just like the operator int(), the intent of this
   *              method is to allow a STATIC_LOGICAL_ATTRIBUTE to
   *              operate exactly as one would expect a boolean to
   *              behave.
   * Return: the new value of the STATIC_LOGICAL_ATTRIBUTE
   *************************************************************************/
  int operator = (int a //In: the new value of the attribute
		  ) {return assign(a);}


  /*************************************************************************
   * Method: GetType()
   * Summary: returns the type of this attribute: STATIC_LOGICAL_ATTRIBUTE_TYPE
   * Description: We need to be able to construct or identify these
   *              attributes and use enum's along with these methods to
   *              identify the tyeps
   * Return: the type of this attribute: STATIC_LOGICAL_ATTRIBUTE_TYPE
   *************************************************************************/
  virtual int GetType() {return STATIC_LOGICAL_ATTRIBUTE_TYPE;}

  /*************************************************************************
   * Method: GetSize()
   * Summary: Returns the size of this attribute
   * Description: We need to be able to get attributes size through a
   *              virtual method.
   * Return: the size of a STATIC_LOGICAL_ATTRIBUTE
   *************************************************************************/
  virtual int GetSize() {return sizeof(STATIC_LOGICAL_ATTRIBUTE);}

  /*************************************************************************
   * Method: outstream(ostream &, int)
   * Summary: prints a STATIC_LOGICAL_ATTRIBUTE to the passed in stream
   * Description: This virutal method allows us to print out the attribute
   * Return: void
   *************************************************************************/
  virtual void outstream(ostream &out, int) {
    if (attribute) {
      out << "LOGICAL_TRUE";
    }else{
      out << "LOGICAL_FALSE";
    }
  }


  /*************************************************************************
   * Method: GetMessageSize()
   * Summary: Returns the size of a message for a STATIC_LOGICAL_ATTRIBUTE
   * Description: Returns the size of a message for a STATIC_LOGICAL_ATTRIBUTE
   * Return: Returns the size of a message for a STATIC_LOGICAL_ATTRIBUTE
   *************************************************************************/
  virtual int GetMessageSize() {
    return sizeof(LOGICAL_ATTRIBUTE_MESSAGE_INFO);
  }

  /*************************************************************************
   * Method: FillMessage
   * Summary: Fills a message with the STATIC_LOGICAL_ATTRIBUTE value
   * Description: Fills a message with the STATIC_LOGICAL_ATTRIBUTE
   *              value so that we can send that message to an object
   *              proxy for an update
   * Return: void
   *************************************************************************/
  virtual void FillMessage(char *buff  //InOut: Buffer to contain the message
			   );

  /*************************************************************************
   * Method: EmptyMessage
   * Summary: Remove STATIC_LOGICAL_ATTRIBUTE values from a message and
   *          updates our own values.
   * Description: Remove STATIC_LOGICAL_ATTRIBUTE values from a message and
   *              updates our own values
   * Retuvoidrn: 
   *************************************************************************/
  virtual void EmptyMessage(const char *buff //In: buffer containing the new values
			    );


protected:

private:
  void delete_vector_with_child_type() { delete [] this; }

  STATE_INT attribute;

};

typedef STATIC_LOGICAL_ATTRIBUTE LOGICAL_ATTRIBUTE;

/************************************************************************
 * Class: STRING_ATTRIBUTE
 *
 * Summary: publishable string attribute for use in object proxies
 *          
 * Description: If you want to publish non time  dependant string data
 *              about your object, this is the class to use.  Whenever
 *              this class is updated in value, that update is then
 *              sent to everyone who is subscribing to your object
 *              proxy.  These updates are automatically taken care of
 *              through the use of operator overloading.
 *
 * Examples: An example of the use of STRING_ATTRIBUTE can be
 *           found in S_RanMot.H in demos/BmdSim.
************************************************************************/
class STRING_ATTRIBUTE : public BASE_ATTRIBUTE {

public:


  /*************************************************************************
   * Method: STRING_ATTRIBUTE()
   * Summary: Default constructor, sets the string to NULL
   * Description: Default constructor, sets the string to NULL.
   * Return: a new STRING_ATTRIBUTE
   *************************************************************************/
  STRING_ATTRIBUTE() {attribute = NULL;}

  /*************************************************************************
   * Method: ~STRING_ATTRIBUTE()
   * Summary: destructor for a STRING_ATTRIBUTE
   * Description: The destructor for a STRING_ATTRIBUTE frees any memory that
   *              had been allocated for storage.
   * Return: void
   *************************************************************************/
  virtual ~STRING_ATTRIBUTE() {
  }


  /*************************************************************************
   * Method: GetValue()
   * Summary: returns the char * string value for the STRING_ATTRIBUTE
   * Description: returns the char * string value for the STRING_ATTRIBUTE
   * Return: returns the char * string value for the STRING_ATTRIBUTE
   *************************************************************************/
  char *GetValue() {return attribute;}
  operator char *() const {return attribute;}


  /*************************************************************************
   * Method: assign(char *)
   * Summary: Assigns a new value to the STRING_ATTRIBUTE
   * Description: Whenever a new value is assigned to a string attribute, 
   *              an update is scheduled for all the object proxies and the
   *              new value is assigned to the string.
   * Return: new value of the STRING_ATTRIBUTE
   *************************************************************************/
  char *assign(char *a) {
    ScheduleUpdate();
    return (attribute = a);
  }


  /*************************************************************************
   * Method: operator =(char *)
   * Summary: Assigns a string to a STRING_ATTRIBUTE
   * Description: When one assigns a string to a STRING_ATTRIBUTE, the string
   *              is copied so that one does not need to concern ones self
   *              with freeing or deleting the memory on the right hand side
   *              of the assignment.  As with all other ATTRIBUTES, this
   *              always schedules an update for the object proxies.
   * Return: the char * string that was allocated for thie STRING_ATTRIBUTE
   *************************************************************************/
  char *operator = (char *a //In: The string we want assigned to this attribute
		    ) {return assign(a);}

  /*************************************************************************
   * Method: operator ==(char *)
   * Summary: compares the passed in string with the attribute
   * Description: This operator returns 1 if the strings are equal and 0
   *              otherwise.
   * Return: This operator returns 1 if the strings are equal and 0 otherwise
   *************************************************************************/
  int operator == (char *a //In: The string  you wish to compare to
		           //    this attribute
		   ) {return !strcmp(a,attribute);}

  /*************************************************************************
   * Method: operator !=(char *)
   * Summary: compares the passed in string with this attribute
   * Description: This operator returns 0 if the strings are equal and 0
   *              otherwise
   * Return: This operator returns 0 if the strings are equal and 1 otherwise
   *************************************************************************/
  int operator != (char *a //In: The string  you wish to compare to
		           //    this attribute
		   ) {return strcmp(a,attribute);}


  /*************************************************************************
   * Method: GetType()
   * Summary: Returns the type of this attribute: STRING_ATTRIBUTE_TYPE
   * Description: Each derived ATTRIBUTE type has a virtual function call
   *              to identify it's type.  All the types are listed at the
   *              top of this header file.
   * Return: The type of this attribute: STRING_ATTRIBUTE_TYPE
   *************************************************************************/
  virtual int GetType() {return STRING_ATTRIBUTE_TYPE;}

  /*************************************************************************
   * Method: GetSize()
   * Summary: Returns the size of this attribute.
   * Description: Notice that this returns the size of the attribute and
   *              that means without including the size of the string, just
   *              the size of the attribute with the char *
   * Return: Returns the size of this attribute
   *************************************************************************/
  virtual int GetSize() {return sizeof(STRING_ATTRIBUTE);}

  /*************************************************************************
   * Method: outstream(ostream &, int)
   * Summary: prints this attribute to the passed in stream
   * Description: prints this attribute to the passed in stream
   * Return: void 
   *************************************************************************/
  virtual void outstream(ostream &out, //In: The stream to which we should
			               //    print this string attribute
			 int);

  /*************************************************************************
   * Method: GetMessageSize()
   * Summary: returns the size of message required to send this attribute
   * Description: STRING_ATTRIBUTE is different from many of the
   *              attributes like the INT_ATTRIBUTE since the
   *              STRING_ATTRIBUTE has a dynamic size.  This method
   *              lets us know how much room is really needed to send
   *              this attribute to another node in the case of a
   *              scheduled update.
   * Return: returns the size of message required to send this attribute
   *************************************************************************/
  virtual int GetMessageSize() { return GetMessageSize(attribute); }

  /*************************************************************************
   * Method: FillMessage
   * Summary: Fills the char * buff with this attribute as a message
   * Description: Once we have called GetMessageSize() to find out how
   *              much memory is required to send this attribute, we
   *              can allocate that memory and then pass it to this
   *              method to fill in that buffer.
   * Return: 
   *************************************************************************/
  virtual void FillMessage(char *buff  //InOut: Buffer to contain this
			               //       attribute as a message
			   ) { FillMessageWith(buff, attribute); }

  /*************************************************************************
   * Method: EmptyMessage
   * Summary: copies the attribute information out of the buffer and
   *          into this attribute
   * Description: When an attribute has been changed, a scheduleupdate
   *              event is scheduled which updates all the object
   *              proxies.  The update is carried in a message filled
   *              by the method FillMessage and this method is used to
   *              unpack that message
   * Return: void
   *************************************************************************/
  virtual void EmptyMessage(const char *buff);

  static int GetMessageSize(char *new_value);
    
  static void FillMessageWith(char *buff, char *value)
  { ((STRING_ATTRIBUTE_MESSAGE_INFO *)buff)->SetValue(value); }
    
  /*************************************************************************
   * Method: Copy(BASE_ATTRIBUTE*)
   * Summary: Copies the passed in attribute to this attribute
   * Description: Copies the passed in attribute to this attribute
   * Return: void
   *************************************************************************/
  virtual void Copy(BASE_ATTRIBUTE *a) {
    if (a) {
      STRING_ATTRIBUTE *sa = (STRING_ATTRIBUTE *)a;
      char *s = *sa;
      attribute.Copy(s);
    }
  }

protected:

private:
  void delete_vector_with_child_type() { delete [] this; }
  
  STATE_STRING attribute;
  
};


/************************************************************************
 * Class: STATIC_POSITION_ATTRIBUTE
 *
 * Summary: publishable position attribute for use in object proxies
 *          
 * Description: If you want to publish non time dependant position
 *              data about your object, this is the class to use.
 *              Whenever this class is updated in value, that update
 *              is then sent to everyone who is subscribing to your
 *              object proxy.  These updates are automatically taken
 *              care of through the use of operator overloading.
 *
 * Examples: An example of the use of STATIC_POSITION_ATTRIBUTE can be
 *           found in S_RanMot.H in demos/BmdSim.
************************************************************************/
typedef class STATIC_POSITION_ATTRIBUTE : public BASE_ATTRIBUTE {

public:


  /*************************************************************************
   * Method: STATIC_POSITION_ATTRIBUTE()
   * Summary: Default constructor for a STATIC_POSITION_ATTRIBUTE
   * Description: The default constructor sets the position of the
   *              object to zero and sets the position type to EARTH
   *              (latitude, longitude and altitude)
   * Return: a new STATIC_POSITION_ATTRIBUTE
   *************************************************************************/
  STATIC_POSITION_ATTRIBUTE() {
    Position[0] = 0.0;
    Position[1] = 0.0;
    Position[2] = 0.0;
    PositionType = EARTH;
  }


  /*************************************************************************
   * Method: SetEarth(double, double, double)
   * Summary: Sets a position in EARTH coordinates
   * Description: Sets the latitude, longitude and altitude to the
   *              passed in values as also sets the position type to
   *              EARTH. This is not error checked so, if you are in
   *              EARTH coordinates, it will happily allow you to make
   *              your position 74922.76 degrees longitude 87823
   *              degrees latitude, and an altitude of -5692.
   * Return: void
   *************************************************************************/
  void SetEARTH(double Lat, //In: New latitude
		double Lon, //In: New longitude
		double Alt  //In: New altitude
		) {
    Position[0] = Lat;
    Position[1] = Lon;
    Position[2] = Alt;
    PositionType = EARTH;
  }


  /*************************************************************************
   * Method: SetEDI(double[3])
   * Summary: Sets a new position in ECI coordinates
   * Description: Sets a new position in Earth Centered Inertial (ECI)
   *              coordinate system.
   * Return: void
   *************************************************************************/
  void SetECI(double x[3] //In: New position in ECI coordinates
	      ) {
    assign(x);
    PositionType = ECI;
  }


  /*************************************************************************
   * Method: SetEDR(double[3])
   * Summary: Sets a new position in ECR coordinates
   * Description: Sets a new position in Earth Centered Rotating (ECR)
   *              coordinate system.
   * Return: void
   *************************************************************************/
  void SetECR(double x[3]//In: New position in ECI coordinates
	      ) {
    assign(x);
    PositionType = ECR;
  }


  /*************************************************************************
   * Method: GetPositionType()
   * Summary: Returns the type of coordinate system that is currently in use
   * Description: Returns the type of coordinate system that is currently
   *              in use.  The possibilities are:
   *                     EARTH: Latitude, Longitude, Altitude
   *                     ECR:   X, Y, Z in a rotating earth coordinate system
   *                     ECI:   X, Y, Z in an inertial coordinate system
   * Return: Returns the type of coordinate system that is currently in use
   *************************************************************************/
  int GetPositionType() {return PositionType;}


  /*************************************************************************
   * Method: GetEarth(double&, double&, double&, double)
   * Summary: Computes the position in the EARTH coordinate system
   * Description: No matter what coordinate system is used to store
   *              the position, this returns the new position in the
   *              three parameters.  The time (t) parameter is ignored unless
   *              we are currently in ECI coordinate system in which case
   *              t is ?????????? (time of day?).
   * Return: void
   *************************************************************************/
  void GetEARTH(double &Lat, //Out: Latitude coordinate
		double &Lon, //Out: Longitude coordinate
		double &Alt, //Out: Altitude coordinate
		double t=0.0 //In:  Seconds after midnight
		);

  /*************************************************************************
   * Method: GetECI(double[3], double)
   * Summary: Computes the position in ECI coordinates
   * Description: Computes the position in ECI coordinates and returns
   *              those values in the first parameter.  Notice that
   *              this does not change the stored state.  For this
   *              reason, if you are going to be making successive
   *              calls to GetECI, it may be wise to call SetECI
   *              first or just save the values and not call the method
   *              again.
   * Return: void
   *************************************************************************/
  void GetECI(double x[3], //Out: Position in ECI coordinates
	      double t=0.0 //In:  seconds after midnight
	      );

  /*************************************************************************
   * Method: GetECR(double[3], double)
   * Summary: Computes the position in ECR coordinates
   * Description: Computes the position in ECR coordinates and returns
   *              those values in the first parameter.  Notice that
   *              this does not change the stored state.  For this
   *              reason, if you are going to be making successive
   *              calls to GetECR, it may be wise to call SetECR
   *              first or just save the values and not call the method
   *              again.
   * Return: void
   *************************************************************************/
  void GetECR(double x[3], //Out: Position in ECI coordinates
	      double t=0.0 //In:  seconds after midnight
	      );


  /*************************************************************************
   * Method: GetValue(double[3], int, double)
   * Summary: Computes and returns position in requested coordinate system
   * Description: Computes the position in the passed coordinate
   *              system and returns those values in the first
   *              parameter.  Notice that this does not change the
   *              stored state.  For this reason, if you are going to
   *              be making successive calls to GetValue, it may be
   *              wise to call SetECR or SetECI (whichever you are
   *              going to be using most frequently) first or just
   *              save the values and not call the method again.
   * Return: void
   *************************************************************************/
  void GetValue(double x[3], //Out: Coordinates in requested system
		int PT,      //In:  Coordinate system in which to
		             //     evaluate the position.  Options are:
                             //     EARTH: Latitude, Longitude, Altitude
                             //     ECR: X, Y, Z in a rotating earth
		             //         coordinate system
		             //     ECI: X, Y, Z in an inertial
	                     //          coordinate system
		double t=0.0 //In:  Seconds after midnight
		);

  /*************************************************************************
   * Method: GetValue()
   * Summary: Returns a pointer to the stored value of the position
   * Description: The position is stored as an array and this simply
   *              returns a pointer to that array.  BE VERY CAREFUL!!
   *              any changes you make to that array are not reflected
   *              in the object proxy!!!!  The only way to make
   *              changes that are reflected in the object proxies is
   *              to use the STATIC_POSITION_ATTRIBUTE::Set****()
   *              methods.
   * Return: A pointer to the present stored position
   *************************************************************************/
  STATE_DOUBLE *GetValue() {return Position;}

  /*************************************************************************
   * Method: operator [] (int)
   * Summary: Returns the i'th coordinate
   * Description: Returns the i'th coordinate value.  This is not
   *              error checked to make sure you request a valid
   *              number.  If you ask for the 973rd coordinate, you
   *              get what you deserve.
   * Return: Returns the i'th coordinate value
   *************************************************************************/
  double operator [] (int i //In: Which coordinate you want.
		      ) {return Position[i];}


  /*************************************************************************
   * Method: assign(double[3])
   * Summary: Sets the position to the passed in values
   * Description: Lets you change your position to the requested
   *              values.  This is not error checked so, if you are in
   *              EARTH coordinates, it will happily allow you to make
   *              your position 74922.76 degrees longitude 87823
   *              degrees latitude, and an altitude of -5692.
   * Return: A pointer to the stored values.
   *************************************************************************/
  STATE_DOUBLE *assign(double x[3]) {
    ScheduleUpdate();
    Position[0] = x[0];
    Position[1] = x[1];
    Position[2] = x[2];
    return Position;
  }


  /*************************************************************************
   * Method: operator =(double[3])
   * Summary: Allows assignment of a position through the '=' operator
   * Description: Allows assignment of a position through the '='
   *              operator.  As with the assign() method, there is no
   *              error checking.
   * Return: A pointer to the stored values.
   *************************************************************************/
   STATE_DOUBLE *operator = (double x[3]) {return assign(x);}


  /*************************************************************************
   * Method: GetType()
   * Summary: Returns the type of this attribute:
   *          STATIC_POSITION_ATTRIBUTE_TYPE
   * Description: Each type is enumerated so we can properly find them and
   *              the enumeration can be found at the top of SpExportAttribute.H.
   * Return: Returns the type of this attribute: STATIC_POSITION_ATTRIBUTE_TYPE
   *************************************************************************/
  virtual int GetType() {return STATIC_POSITION_ATTRIBUTE_TYPE;}

  /*************************************************************************
   * Method: GetSize()
   * Summary: Returns the size of a STATIC_POSITION_ATTRIBUTE
   * Description: Returns the size of a STATIC_POSITION_ATTRIBUTE
   * Return: Returns the size of a STATIC_POSITION_ATTRIBUTE
   *************************************************************************/
  virtual int GetSize() {return sizeof(STATIC_POSITION_ATTRIBUTE);}

  virtual void outstream(ostream &out, int);
  virtual int GetMessageSize() {
    return sizeof(POSITION_ATTRIBUTE_MESSAGE_INFO);
  }

  virtual void FillMessage(char *buff);
  virtual void EmptyMessage(const char *buff);

protected:

private:
  void delete_vector_with_child_type() { delete [] this; }

  STATE_DOUBLE Position[3];	// Position in ECI, ECR, or EARTH coordinates
  STATE_INT PositionType;	// ECI, ECR, or EARTH flag

} POSITION_ATTRIBUTE;












/************************************************************************
 * Class: OBJECT_ATTRIBUTE
 *
 * Summary: publishable object attribute for use in object proxies
 *          
 * Description: If you want to publish a non time dependant object
 *              within your object, this is the class to use.
 *              Whenever this class is updated in value, that update
 *              is then sent to everyone who is subscribing to your
 *              object proxy.
 *
 * Examples: An example of the use of OBJECT_ATTRIBUTE can be
 *           found in S_RanMot.H in demos/BmdSim.
 *
 *           Another example is this.  Suppose you had the class:
 *
 *           class STANDARD_INFO {
 *           public:
 *             char * GetName();
 *             void SetName(char *Name);
 *             double GetStartTime();
 *             void SetStartTime(double startTime);
 *             STANDARD_INFO()
 *             {
 *               Setname(NULL);
 *               SetStartTime(-INFINITY)
 *             };
 *
 *           private:
 *             char * Name;
 *             double StartTime;
 *             double TempVariable;
 *           };
 *
 *           You could turn this into a publishable class fairly easily
 *           by just changing it to:
 *           class STANDARD_INFO : public OBJECT_ATTRIBUTE{
 *           public:
 *             char * GetName();
 *             void SetName(char *Name);
 *             double GetStartTime();
 *             void SetStartTime(double startTime);
 *             STANDARD_INFO()
 *             {
 *               Setname(NULL);
 *               SetStartTime(-INFINITY);
 *               DEFINE_ATTRIBUTE(StartTime, "StartTime");
 *               DEFINE_ATTRIBUTE(Name, "Name");
 *               SetClassName("STANDARD_INFO");
 *             }
 *
 *           private:
 *             STRING_ATTRIBUTE Name;
 *             FLOAT_ATTRIBUTE StartTime;
 *             double TempVariable;
 *           };
 *           RB_DEFINE_CLASS(STANDARD_INFO);
 *
 *           The major changes you may notice are to change standard
 *           variables to ATTRIBUTE variables, to inherit from
 *           OBJECT_ATTRIBUTE, and to add DEFINE_ATTRIBUTE() calls to
 *           the constructor.  These DEFINE_ATTRIBUTE() calls allow
 *           SPEEDES to know what your class looks like and gives
 *           SPEEDES a way to know how to package up this attribute
 *           for updates.  The addition of the
 *           RB_DEFINE_CLASS(STANDARD_INFO) allows SPEEDES to know how
 *           to roll back your class as well as delete roll back data
 *           properly when events are committed.
 *
 *           If a LIST_ATTRIBUTE object is really a LIST_ATTRIBUTE and
 *           NOT a BASE_DYNAMIC_ATTRIBUTE, then it contains nothing
 *           but OBJECT_ATTRIBUTEs.  If it is a
 *           BASE_DYNAMIC_ATTRIBUTE, then it contains nothing but
 *           DYNAMIC_ITEMs.  An OBJECT_ATTRIBUTE is not necessarily in
 *           a LIST_ATTRIBUTE.  OBJECT_ATTRIBUTEs have a list id which
 *           indicates where in they may exist within a given dynamic
 *           attribute.
 *
************************************************************************/
class OBJECT_ATTRIBUTE : public BASE_ATTRIBUTE {

public:


  /*************************************************************************
   * Method: OBJECT_ATTRIBUTE()
   * Summary: Default constructor for an OBJECT_ATTRIBUTE
   * Description: It is highly recommended that you overload this
   *              method to initialize your own objects.  That is, you
   *              will want to add in DEFINE_ATTRIBUTE() calls for
   *              each of your own attributes as well as initialize
   *              any variables you need to initialize.
   * Return: a new OBJECT_ATTRIBUTE
   *************************************************************************/
  OBJECT_ATTRIBUTE(char *className=NULL); //proxy:  added arg

  /*************************************************************************
   * Method: ~OBJECT_ATTRIBUTE()
   * Summary: destructor for an OBJECT_ATTRIBUTE
   * Description: The default destructor partially destroy's your
   *              object attribute but you may want to clean up any
   *              memory you may have allocated within your object.
   * Return: void
   *************************************************************************/
  ~OBJECT_ATTRIBUTE();


  /*************************************************************************
   * Method: SetInObjectProxy(SpObjProxy*)
   * Summary: Sets attribute as being in an object proxy
   * Description: Once we create an attribute, it must be inserted
   *              into an object proxy (this is done for you).  This
   *              method performs that task.
   * Return: 
   *************************************************************************/
  virtual void SetInObjectProxy(SpObjProxy *o //In: Object proxy within which
				                //    this attribute is
				                //    contained.
				);

  /*************************************************************************
   * Method: SetInListAttribute(LIST_ATTRIBUTE*)
   * Summary: Sets attribute as being in the passed in LIST_ATTRIBUTE
   * Description: If you create an attribute list of object attributes,
   *              the list and this attribute must know about eachother.
   *              This method lets the object know about the list that
   *              contains it.
   * Return: 
   *************************************************************************/
  void SetInListAttribute(LIST_ATTRIBUTE *l //In: LIST_ATTRIBUTE which contains
			                    //    this OBJECT_ATTRIBUTE
			  );

  /*************************************************************************
   * Method: GetInListAttribute()
   * Summary: Returns the InListAttribute for this object.
   * Description: If this object is contained in a list, we can grab that
   *              list using this method.
   * Return: The InListAttribute for this object (the list that
   *         contains this object)
   *************************************************************************/
  LIST_ATTRIBUTE *GetInListAttribute() {return InListAttribute;}


  /*************************************************************************
   * Method: GetListId()
   * Summary: returns the list id for this object attribute
   * Description: Returns where in a list attribute this attribute resides
   * Return: Returns where in a list attribute this attribute resides
   *************************************************************************/
  int GetListId() {return ListId;}


  /*************************************************************************
   * Method: SetStartTime(double)
   * Summary: Sets the time this attribute starts
   * Description: OBJECT_ATTRIBUTES can be valid only for a given period
   *              of time.  One can set the start time and the end
   *              time using this method and SetEndTime().
   * Return: void
   *************************************************************************/
  //void SetStartTime(double t //In: Start time for this attribute
  //		    ) {StartTime = t;}

  /*************************************************************************
   * Method: SetEndTime(double)
   * Summary: Sets the time this attribute ends
   * Description: OBJECT_ATTRIBUTES can be valid only for a given period
   *              of time.  One can set the start time and the end
   *              time using this method and SetStartTime().
   * Return: 
   *************************************************************************/
  //void SetEndTime(double t //In: The time at which this attribute becomes
  //		           //    invalid.
  //		  ) {EndTime = t;}

  /*************************************************************************
   * Method: GetStartTime()
   * Summary: returns the time this attribute becomes valid
   * Description: OBJECT_ATTRIBUTES have a period of time for which
   *              they can be valid.  This method returns the time that
   *              the object is valid.
   * Return: returns the time this attribute becomes valid
   *************************************************************************/
  //double GetStartTime() {return StartTime;}

  /*************************************************************************
   * Method: GetendTime()
   * Summary: returns the time this attribute stops being valid
   * Description: OBJECT_ATTRIBUTES have a period of time for which
   *              they can be valid.  This method returns the time that
   *              the object stops being valid.
   * Return: returns the time this attribute stops being valid
   *************************************************************************/
  //double GetEndTime() {return EndTime;}


  /*************************************************************************
   * Method: SetClassName(char *)
   * Summary: Sets the name of this particular class
   * Description: Once you create an object attribute, it must be
   *              named so that we can find it.  This sets the name of
   *              the class.  It is highly recommended that you set
   *              the class name to the true class name.
   * Return: 
   *************************************************************************/
  void SetClassName(char *ObjectType //In: Name to which we should set the
		                     //    class.
		    );

  /*************************************************************************
   * Method: ObjProxyGetClassName
   * Summary: Each class is named,  This returns that name.
   * Description: Each class is named,  This returns that name
   * Return: Each class is named,  This returns that name
   *************************************************************************/
  char *ObjProxyGetClassName();

  /*************************************************************************
   * Method: SetObjectName(char *)
   * Summary: Sets the object name of the object proxy within this class.
   * Description: Sets the object name of the object proxy within this class.
   * Return: Sets the object name of the object proxy within this class.
   *************************************************************************/
  void SetObjectName(char *ObjectName //In: Name for the object proxy within
		                      //    this class.
		     );

  /*************************************************************************
   * Method: GetObjectName()
   * Summary: Returns the name of the object proxy within this class.
   * Description: Returns the name of the object proxy within this class.
   * Return: Returns the name of the object proxy within this class.
   *************************************************************************/
  char *GetObjectName();


  /*************************************************************************
   * Method: GetValue()
   * Summary: Returns the object proxy within this class
   * Description: Each OBJECT_ATTRIBUTE contains an object proxy and this
   *              method returns that object proxy. This is the same as
   *              GetObjectProxy()
   * Return: Returns the object proxy within this class
   *************************************************************************/
  SpObjProxy *GetValue() {return ObjectProxy;}

  /*************************************************************************
   * Method: GetObjectProxy()
   * Summary: Returns the object proxy within this class
   * Description: Each OBJECT_ATTRIBUTE contains an object proxy and this
   *              method returns that object proxy.  This is the same as
   *              GetValue()
   * Return: 
   *************************************************************************/
  SpObjProxy *GetObjectProxy() {return ObjectProxy;}


  /*************************************************************************
   * Method: GetType()
   * Summary: Returns the type of this attribute: OBJECT_ATTRIBUTE_TYPE
   * Description: Each attribute has a special type (all enumerated at
   *              the top of this header file) and that type is
   *              returned here.
   * Return: Returns the type of this attribute: OBJECT_ATTRIBUTE_TYPE
   *************************************************************************/
  virtual int GetType() {return OBJECT_ATTRIBUTE_TYPE;}

  /*************************************************************************
   * Method: GetSize()
   * Summary: Returns the size of an OBJECT_ATTRIBUTE
   * Description: Returns the size so that we can construct a buffer that
   *              contains one of these objects.
   * Return: Returns the size of an OBJECT_ATTRIBUTE
   *************************************************************************/
  virtual int GetSize() {return sizeof(OBJECT_ATTRIBUTE);}

  /*************************************************************************
   * Method: ostream(ostream&, int)
   * Summary: Prints the object proxy within this attribute to the stream.
   * Description: Prints the object proxy within this attribute to the stream.
   * Return: void
   *************************************************************************/
  virtual void outstream(ostream &out, int level);


  /*************************************************************************
   * Method: GetMessageSize()
   * Summary: Returns the size of message required to contain this attribute.
   * Description: If we change the value of an OBJECT_ATTRIBUTE, we
   *              need to announce this update to the object proxies.
   *              This method tells us the size of buffer required to
   *              contain that message.
   * Return: Returns the size of message required to contain this attribute.
   *************************************************************************/
  virtual int GetMessageSize();

  /*************************************************************************
   * Method: FillMessage
   * Summary: Fills a buffer with this attribute for sending.
   * Description: Once we know the size of buffer required for sending
   *              this attribute (from GetMessageSize()), we can then
   *              allocate the memory ourselves and call FillMessage
   *              with that buffer to be filled.
   * Return: void
   *************************************************************************/
  virtual void FillMessage(char *buff  //InOut: Buffer to be filled with our
			               //       attribute's value
			   );

  /*************************************************************************
   * Method: EmptyMessage
   * Summary: Updates our attribute using the value passed in the buffer.
   * Description: Once we have filled a message with the FIllMessage()
   *              call, we then can send that buffer to another object
   *              proxy as their update.  This method unpacks that
   *              buffer and updates the values of our attribute.
   * Return: void
   *************************************************************************/
  virtual void EmptyMessage(const char *buff //In: Buffer containing updated values
			               //    for our attribute
			    );


  /*************************************************************************
   * Method: GetDepth()
   * Summary: Determines the depth in terms of Object Proxies
   * Description: Determines the depth in terms of Object Proxies
   * Return: Feturns the depth in terms of Object Proxies
   *************************************************************************/
  virtual int GetDepth();

  /*************************************************************************
   * Method: PackLevelInfo
   * Summary: packs up the level info buffer.
   * Description: packs up the level info buffer that is used for
   *              sending updates out to other nodes. It gives the
   *              info needed for finding the right object proxy to
   *              modify for the attribute
   * Return: a pointer to the the start of the level info within the passed in
   *         buffer
   *************************************************************************/
  virtual char *PackLevelInfo(const char *Buff,    //In: Buffer to contain the
			                     //    level info
			      int NlevelInfo //In: Number of levels of info
			      );

  /*************************************************************************
   * Method: Copy(BASE_ATTRIBUTE*)
   * Summary: Copies the passed in attribute over this attribute
   * Description: Copies the passed in attribute over this attribute
   * Return: void
   *************************************************************************/
  virtual void Copy(BASE_ATTRIBUTE *a //In: Attribute to be copied over
		                      //    this attribute
		    );

  int  GetActive(void) const { return (Active); }
  void SetActive(int v) { Active = v; }

protected:

    /*
    I commented this and related functions out since people will
    usually not need times in their OBJECT_ATTRIBUTEs.
    OBJECT_ATTRIBUTEs should be like INT_ATTRIBUTEs except that
    OBJECT_ATTRIBUTEs are merely a collection of other Attributes.
    Since INT_ATTRIBUTEs do not have times, then neither should
    OBJECT_ATTRIBUTEs!  It does make sense for DYNAMIC_ITEMs to have
    times since they are specifically designed to change continually,
    but usually this will not be the case with OBJECT_ATTRIBUTEs.
      
  STATE_DOUBLE StartTime;
  STATE_DOUBLE EndTime;
    */

  int ListId;
  static int Counter;

  SpObjProxy *ObjectProxy;
  LIST_ATTRIBUTE *InListAttribute;

private:
  void delete_vector_with_child_type() { delete [] this; }

  LOGICAL_ATTRIBUTE Active;

};


/************************************************************************
 * Class: LIST_ATTRIBUTE
 *
 * Summary: publishable list attribute for use in object proxies
 *          
 * Description: If you want to publish a non time dependant list
 *              within your object, this is the class to use.
 *              Whenever this class is updated in value, that update
 *              is then sent to everyone who is subscribing to your
 *              object proxy.  These updates are automatically taken
 *              care of through the use of operator overloading.
 *
 * Examples: An example of the use of LIST_ATTRIBUTE can be found in
 *           S_RanMot.H in demos/BmdSim.
************************************************************************/
class LIST_ATTRIBUTE : public BASE_ATTRIBUTE {

public:

  /*************************************************************************
   * Method: LIST_ATTRIBUTE()
   * Summary: Default constructor for a list attribute
   * Description: Default constructor does nothing additional
   * Return: a new LIST_ATTRIBUTE
   *************************************************************************/
  LIST_ATTRIBUTE() {;}

  /*************************************************************************
   * Method: ~LIST_ATTRIBUTE()
   * Summary: Destructor for a list attribute.
   * Description: empties the list, deletes all members and returns
   * Return: void
   *************************************************************************/
  ~LIST_ATTRIBUTE();

  /*************************************************************************
   * Method: GetValue()
   * Summary: Returns the list within this attribute.
   * Description: Returns the list within this attribute.
   * Return: Returns the list within this attribute.
   *************************************************************************/
  RB_list *GetValue() {return &list;}

  /*************************************************************************
   * Method: get_length()
   * Summary: Returns the length of this list.
   * Description: Just like a normal list (RB_list), this returns the
   *              length of the list. Same as GetLength().
   * Return: Returns the length of this list.
   *************************************************************************/
  int get_length() {return list.get_length();}

  /*************************************************************************
   * Method: get_length()
   * Summary: Returns the length of this list.
   * Description: Just like a normal list (RB_list), this returns the
   *              length of the list. Same as GetLength().
   * Return: Returns the length of this list.
   *************************************************************************/
  int GetLength() {return list.get_length();}


  /*************************************************************************
   * Method: operator () (char *)
   * Summary: returns the OBJECT_ATTRIBUTE with the passed in name
   * Description: When one has a list of OBJECT_ATTRIBUTEs and wants
   *              to find the attribute with a particular name, this
   *              operator will walk through the list until it finds
   *              the matching OBJECT_ATTRIBUTE.  If there is not any
   *              object in the list that matches the passed in name,
   *              NULL is returned.
   * Return: The OBJECT_ATTRIBUTE matching the passed in name
   *************************************************************************/
  OBJECT_ATTRIBUTE *operator() (char *n //In: Name of the attribute you want
				        //    to find in this list
				);

  /*************************************************************************
   * Method: operator () (double)
   * Summary: returns the OBJECT_ATTRIBUTE that is valid for the
   *          passed in time.
   * Description: Since object attributes can be valid at different
   *              points in time, this operator finds the one in our
   *              list that is valid at that time.  For example, we
   *              may have a list attribute that contains
   *              object_attributes that are valid for the times
   *              0<-->10, 10<-->30, and 30<-->10000.  If we call this
   *              operator with the value 3, it returns the first
   *              object attribute, and f we call it with value 50, it
   *              returns the third.  If there happen to be two items
   *              in the list with overlapping times and you call this
   *              operator with a time in that overlap, then the
   *              OBJECT_ATTRIBUTE closest to the top of the list will
   *              be returned.  If there is not any attribute valid for
   *              that time, NULL will be returned.
   * Return: the OBJECT_ATTRIBUTE valid for the passed in time.
   *************************************************************************/
  //OBJECT_ATTRIBUTE *operator() (double t //In: Time for which you want a valid
  //				         //    OBJECT_ATTRIBUTE
  //				);

  /*************************************************************************
   * Method: FindObjectAttribute()
   * Summary: Finds the object attribute in the list with the passed in id.
   * Description: When an object attribute is added to the list with a
   *              given id, that allows us to find it by matching it
   *              with the id in the passed value.  If no attribute in
   *              the list has a matching id, then we return NULL.
   * Return: 
   *************************************************************************/
  OBJECT_ATTRIBUTE *FindObjectAttribute(int ListId //In: Id for the attribute
					           //    we wish to find.
					);

  /*************************************************************************
   * Method: top()
   * Summary: Returns a pointer to the top of the list.
   * Description: Returns a pointer to the top of the list.
   * Return: Returns a pointer to the top of the list.
   *************************************************************************/
  void *top() {return list.get_top();}

  /*************************************************************************
   * Method: bot()
   * Summary: Returns a pointer to the bottom of the list.
   * Description: Returns a pointer to the bottom of the list.
   * Return: Returns a pointer to the bottom of the list.
   *************************************************************************/
  void *bot() {return list.get_bot();}


  /*************************************************************************
   * Method: get_top()
   * Summary: Returns a pointer to the top of the list.
   * Description: Returns a pointer to the top of the list.
   * Return: Returns a pointer to the top of the list.
   *************************************************************************/
  void *get_top() {return list.get_top();}

  /*************************************************************************
   * Method: get_bot()
   * Summary: Returns a pointer to the bottom of the list.
   * Description: Returns a pointer to the bottom of the list.
   * Return: Returns a pointer to the bottom of the list.
   *************************************************************************/
  void *get_bot() {return list.get_bot();}


  /*************************************************************************
   * Method: operator ++()
   * Summary: Returns a pointer to the next item in the list.
   * Description: This is done by moving a current pointer down the
   *              list to the next item.  If we are at the bottom of
   *              the list or the list is empty, then we return NULL.
   * Return: Returns a pointer to the next item in the list
   *************************************************************************/
  void *operator ++ () {return (++list);}

  /*************************************************************************
   * Method: operator --()
   * Summary: Returns a pointer to the previous item in the list.
   * Description: This is done by moving a current pointer up the
   *              list to the previous item.  If we are at the top of
   *              the list or the list is empty, then we return NULL.
   * Return: Returns a pointer to the previous item in the list.
   *************************************************************************/
  void *operator -- () {return (--list);}

  // Pass this function something returned from new without brackets.

  /*************************************************************************
   * Method: operator += (OBJECT_ATTRIBUTE*)
   * Summary: Adds an item to the bottom of a list.
   * Description: Adds an item to the bottom of a list.  This also
   *              schedules update for all the other object
   *              proxies. 
   *              WARNING:  
   *              the item you are adding to the list MUST be 
   *	          allocated individually ie: v = new CHILD_OBJ_ATT
   *    	  (NOT v = new CHILD_OBJ_ATT[n])  This
   *              is because ~LIST_ATTRIBUTE does individual deletes
   *      	  (ie delete v) of the items on the list.
   * Return: returns a pointer to this item within the list.
   *************************************************************************/
  void * operator += (OBJECT_ATTRIBUTE *v //In: New object to be added to
		                          //    the bottom of the list
		      ) {
    v->SetInListAttribute(this);

    //proxy: start new code
    // Need to give this OBJ_ATT a ptr to the updateList, and ditto for
    // all proxies contained within
    SpObjProxy /* *objProxy,*/ *containingProxy = GetInObjectProxy();
    if (containingProxy!=NULL) {
      InsertUpdateList(v, containingProxy);
    }
    //proxy: end new code

    ScheduleUpdate(OBJECT_PROXY_ADD, v);
    return (list += v);
  }


  /*************************************************************************
   * Method: operator -=(void *)
   * Summary: Removes an item from the list.
   * Description: When you want to remove an item from the list, -=
   *              item will accomplish that for you. If the passed in
   *              item cannot be found in the list, an error message
   *              is printed and we continue.  The item that is
   *              removed is pushed back on the free list using
   *              FreeList->push_item_bot(item)
   * Return: void
   *************************************************************************/
  void operator -= (OBJECT_ATTRIBUTE *v //In: Item to be removed from the list
		    ) {
    ScheduleUpdate(REMOVE, v);
    list -= v;
  }


  /*************************************************************************
   * Method: GetType()
   * Summary: Returns the type of this attribute: LIST_ATTRIBUTE_TYPE
   * Description: Each attribute has a unique type (listed at the top
   *              of this header file) and this method returns that
   *              type.
   * Return: Returns the type of this attribute: LIST_ATTRIBUTE_TYPE
   *************************************************************************/
  virtual int GetType() {return LIST_ATTRIBUTE_TYPE;}

  /*************************************************************************
   * Method: GetSize()
   * Summary: Returns the size of this attribute.
   * Description: Returns the size of this attribute.
   * Return: Returns the size of this attribute.
   *************************************************************************/
  virtual int GetSize() {return sizeof(LIST_ATTRIBUTE);}

  /*************************************************************************
   * Method: outstream(ostream &, int)
   * Summary: Prints this attribute to the passed in stream.
   * Description: This operator will print out this list attribute to
   *              the passed in stream.  Since it is a list, we do
   *              this by calling the virtual function outstream() on
   *              each OBJECT_ATTRIBUTE member of this list.
   * Return: void
   *************************************************************************/
  virtual void outstream(ostream &out, //In: Stream to which we should print
			 int level=0   //In: level of printing to perform.
			 );

  /*************************************************************************
   * Method: GetMessageSize()
   * Summary: Returns the size of message required to contain this attribute.
   * Description: If we change the value of an LIST_ATTRIBUTE, we
   *              need to announce this update to the object proxies.
   *              This method tells us the size of buffer required to
   *              contain that message.
   * Return: Returns the size of message required to contain this attribute.
   *************************************************************************/
  virtual int GetMessageSize();

  /*************************************************************************
   * Method: FillMessage
   * Summary: Fills a buffer with this attribute for sending.
   * Description: Once we know the size of buffer required for sending
   *              this attribute (from GetMessageSize()), we can then
   *              allocate the memory ourselves and call FillMessage
   *              with that buffer to be filled.
   * Return: void
   *************************************************************************/
  virtual void FillMessage(char *buff  //InOut: Buffer that will contain this
			               //       attribute
			   );

  /*************************************************************************
   * Method: EmptyMessage
   * Summary: Updates our attribute using the value passed in the buffer.
   * Description: Once we have filled a message with the FIllMessage()
   *              call, we then can send that buffer to another object
   *              proxy as their update.  This method unpacks that
   *              buffer and updates the values of our attribute.
   * Return: void
   *************************************************************************/
  virtual void EmptyMessage(const char *buff //In: Buffer containing updated values
			               //    for our attribute
			    );

  /*************************************************************************
   * Method: Copy(BASE_ATTRIBUTE*)
   * Summary: Copies the passed in attribute over this attribute
   * Description: Copies the passed in attribute over this attribute
   * Return: void
   *************************************************************************/
  virtual void Copy(BASE_ATTRIBUTE *a//In: Attribute to be copied over
		                      //    this attribute
		    );

protected:
  void push_top(void *v) {list.push_top(v);}
  void push_bot(void *v) {list.push_bot(v);}

  //proxy: add
  //If possible, take the updateList ptr from the proxy that contains "addedItem"
  // and insert it into addItem's proxy (and all subsidiary proxies below).
  void InsertUpdateList(OBJECT_ATTRIBUTE *addedItem,
			SpObjProxy *containingProxy);
  RB_list list;

private:
  void delete_vector_with_child_type() { delete [] this; }

};

/************************************************************************
 * Class: BASE_DYNAMIC_ATTRIBUTE
 *
 * Summary: This is the type from which all dynamic types are derived.
 *          
 * Description: You may have already seen hints of the dynamic type
 *              from the list type.  A dynamic type is a type which,
 *              rather than having a single static value, may have
 *              different values at different points in time.  For an
 *              example, you may have an object that moves on a fixed
 *              path such as an elliptical orbit.  This path could be
 *              represented using an equation of motion which would be
 *              a type derived from the BASE_DYNAMIC_ATTRIBUTE type.
 *              there are several equations of motion included with
 *              SPEEDES along with many other dynamic types.
 *
 * Examples: There are several uses of the BASE_DYNAMIC_ATTRIBUTE in
 *           the classes following the BASE_DYNAMIC_ATTRIBUTE.  These
 *           include the the DYNAMIC_FLOAT_ATTRIBUTE,
 *           DYNAMIC_POSITION_ATTRIBUTE, and the
 *           DYNAMIC_LOGICAL_ATTRIBUTE
************************************************************************/
class BASE_DYNAMIC_ATTRIBUTE : public LIST_ATTRIBUTE {

public:

  /*************************************************************************
   * Method: BASE_DYNAMIC_ATTRIBUTE
   * Summary: Default constructor of a BASE_DYNAMIC_ATTRIBUTE
   * Description: The default constructor does nothing other than
   *              instansiate it's members.
   * Return: a new BASE_DYNAMIC_ATTRIBUTE
   *************************************************************************/
  BASE_DYNAMIC_ATTRIBUTE() {;}

  /*************************************************************************
   * Method: ~BASE_DYNAMIC_ATTRIBUTE
   * Summary: Destructor of a BASE_DYNAMIC_ATTRIBUTE
   * Description: The destructor goes through the list of items, pops
   *              them off and deletes them.
   * Return: void
   *************************************************************************/
  ~BASE_DYNAMIC_ATTRIBUTE();


  /*************************************************************************
   * Method: FindDynamicItem(double)
   * Summary: Returns the SpDynItem first valid for the given time.
   * Description: Since the BASE_DYNAMIC_ATTRIBUTE class inherits from
   *              the LIST_ATTRIBUTE type, it has a list which can
   *              store a variety of DYNAMIC_ATTRIBUTEs.  Each of
   *              these attributes has a start time and an end time.
   *              These are the times for which a dynamic item is
   *              considered to be valid and this returns the one that
   *              is valid for the passed in time.  If there is no
   *              item in the list that is valid for the passed in
   *              time, an error message is printed and NULL is
   *              returned.  Be careful to pass in a double value!
   *              Otherwise, this method could get confused with the
   *              method FindDynamicItem(int) which performs a
   *              different service.
   * Return: SpDynItem valid for that time or NULL if it does not exist
   *************************************************************************/
  SpDynItem *FindDynamicItem(double t //In: Time for which you want to
				         //    find the valid SpDynItem
				);

  /*************************************************************************
   * Method: FindDynamicItem(int)
   * Summary: Finds the item with the given id.
   * Description: Just as you can search on time in the list in the
   *              BASE_DYNAMIC_ATTRIBUTE class, you can also search on
   *              the integer id of the items in that list.  This
   *              returns the item matching the passed in ID or, if
   *              that item is not in the list, prints an error
   *              message and returns NULL.
   * Return: SpDynItem matching the passed ID or NULL if it does not exist
   *************************************************************************/
  SpDynItem *FindDynamicItem(int i //In: ID of the item you want to find
				      //    in the list.
				);

  /*************************************************************************
   * Method: GetTimeInterval(double time, double &startTime, double &endTime)
   * Summary: Returns the start time and end time for the segment in the
   *          dynamic attribute
   * Description: Finds the segment in the dynamic attribute that is bounded
   *              by the time argument passed in and then returns the start
   *              and end times of the segment.
   * Return: start and end times for the segment
   *************************************************************************/

  int GetTimeInterval(double time,
		      double &startTime,
		      double &endTime) {
    SpDynItem *DynamicItem = FindDynamicItem(time);
    if (DynamicItem) {
      startTime = DynamicItem->GetStartTime();
      endTime = DynamicItem->GetEndTime();
      return 1;
    }else{
      startTime = GetStartTime();
      endTime = GetEndTime();
      return 0;
    }
  }

  /*************************************************************************
   * Method: GetTimeInterval(double &startTime, double &endTime)
   * Summary: Returns the start time and end time for the dynamic attribute
   * Description: Finds the start and end times for the dynamic attribute.
   * Return: start and end times
   *************************************************************************/

  void GetTimeInterval(double &startTime,
				double &endTime
			      ) {
    startTime = GetStartTime();
    endTime = GetEndTime();
  }

  /*************************************************************************
   * Method: GetStartTime()
   * Summary: Returns the first time for which any item in the list is valid
   * Description: If you want to find the earliest time that anything in the
   *              list may be valid, this is the method to use.
   * Return: Returns the first time for which any item in the list is valid
   *************************************************************************/
  double GetStartTime();

  /*************************************************************************
   * Method: GetEndTime()
   * Summary: Returns the last time for which any item in the list is valid
   * Description: If you want to find the latest that anything in the
   *              list may be valid, this is the method to use.
   * Return: Returns the last time for which any item in the list is valid
   *************************************************************************/
  double GetEndTime();

  /*************************************************************************
   * Method: GetType
   * Summary: Returns the type of this class: -1 since it should never be used
   * Description: Returns the type of this class.  Since this class
   *              should only be used to build other classes it
   *              returns the value -1.
   * Return: Returns the type of this class: -1 since it should never be used
   *************************************************************************/
  virtual int GetType() {return -1;}

  /*************************************************************************
   * Method: GetSize()
   * Summary: Returns the size of this class
   * Description: Returns the size of this class
   * Return: Returns the size of this class
   *************************************************************************/
  virtual int GetSize() {return sizeof(BASE_DYNAMIC_ATTRIBUTE);}


  /*************************************************************************
   * Method: outstream(ostream &,int)
   * Summary: Prints the class to the passed in stream
   * Description: Since the BASE_DYNAMIC_ATTRIBUTE type is a list of
   *              SpDynItem types, it goes to each item in the list
   *              and prints the triplet [type,start_time,end_time].
   *              The second argument tells how many items to print
   *              per line.
   * Return: void
   *************************************************************************/
  virtual void outstream(ostream &out, int level);

  /*************************************************************************
   * Method: GetMessageSize
   * Summary: Returns the size of message required to contain this type
   * Description: If a BASE_DYNAMIC_ATTRIBUTE needs to be sent to
   *              another node, this routine returns the size of
   *              message required to contain this particular instance
   * Return: Returns the size of message required to contain this type
   *************************************************************************/
  virtual int GetMessageSize();

  /*************************************************************************
   * Method: FillMessage
   * Summary: Fills the buffer with the list of dynamic items
   * Description: After you find out how much message is required to
   *              send this attribute from one node to another, you
   *              can construct that buffer and provide it here.  The
   *              second argument is the index of this attribute.
   * Return: void
   *************************************************************************/
  virtual void FillMessage(char *buff  //InOut: Buffer to contain the message
			   );

  /*************************************************************************
   * Method: EmptyMessage
   * Summary: Extracts the SpDynItem list from the buffer and
   *          inserts it into our item.
   * Description: Once we have been told about a
   *              BASE_DYNAMIC_ATTRIBUTE type, we can extract it using
   *              this call.
   * Return: void
   *************************************************************************/
  virtual void EmptyMessage(const char *buff //In: Buffer containg dynamic items
			    );

  /*************************************************************************
   * Method: operator += (SpDynItem*)
   * Summary: Adds another SpDynItem to the list
   * Description: This schedules an update for all of the object
   *              proxies and then pushes the passed in DYANIC_ITEM
   *              onto the bottom of the list.  It then returns the
   *              item.  You should make sure to pass this function
   *              something that was returned from
   *              FreeDynamicAttributes->new_object
   * Return: Adds another SpDynItem to the list
   *************************************************************************/
  void * operator += (SpDynItem *v //In: Item to be added to the list
		      ) {
    ScheduleUpdate(OBJECT_PROXY_ADD, v);
    return (list += v);
  }


  /*************************************************************************
   * Method: operator -=(void *)
   * Summary: Remove an item from the list
   * Description: Removes an item from the list, schedules an update
   *              for the object proxies and then pushes this item
   *              back on the free list.
   * Return: void
   *************************************************************************/
  void operator -= (SpDynItem *v //In: Item to be removed from the list
		    ) {
    ScheduleUpdate(REMOVE, v);
    list -= v;
  }


  /*************************************************************************
   * Method: Copy(BASE_ATTRIBUTE*)
   * Summary: Copies the passed in BASE_ATTRIBUTE over this object
   * Description: This casts the passed in attribute to an object
   *              attribute, memcpy's it over our attribute, and
   *              clones it's object proxy.
   * Return: 
   *************************************************************************/
  virtual void Copy(BASE_ATTRIBUTE *a//In: Object to be copied over this.
		    );

protected:

private:
  void delete_vector_with_child_type() {
    cerr << "You should not have constructed an object of"
      " type BASE_DYNAMIC_ATTRIBUTE." << endl;
    abort();
  }

};

/************************************************************************
 * Class: DYNAMIC_FLOAT_ATTRIBUTE
 *
 * Summary: A time dependant attribute double/float type.
 * 
 * Description: If you want to have a double or float type that has
 *              dynamic value (values that change over time), then
 *              this is the correct type to use.  Unlike the
 *              STATIC_FLOAT_ATTRIBUTE, it is not a drop in
 *              replacement for a normal double or float.  That is
 *              because the accessors for dynamic types require that
 *              you provide a time at which you want the value.
 *
 * Examples: None of the demo code currently has examples of the use
 *           of this type
************************************************************************/
class DYNAMIC_FLOAT_ATTRIBUTE : public BASE_DYNAMIC_ATTRIBUTE {

public:


  /*************************************************************************
   * Method: DYNAMIC_FLOAT_ATTRIBUTE()
   * Summary: Default constructor for a DYNAMIC_FLOAT_ATTRIBUTE
   * Description: The default constructor simply instantiates the
   *              members which means that the list of values is empty
   * Return: 
   *************************************************************************/
  DYNAMIC_FLOAT_ATTRIBUTE() {;}

  /*************************************************************************
   * Method: GetValue(double)
   * Summary: Returns the value of this attribute at the passed in time.
   * Description: When you want to find the value of any dynamic type,
   *              you need to provide SPEEDES with the time at which
   *              you want the value.  This attribute return the value
   *              -INFINITY if you request a value for which there is
   *              not a valid time.  Be warned that this is a valid
   *              number and you need to check if it makes sense in
   *              your situation.  We hope to remedy this in the
   *              future by adding a global variable that can be
   *              checked for an error condition if -INFINITY is
   *              returned.
   * Return: the value of the variable at the passed in time or
   *         -INFINITY if the time is invalid
   *************************************************************************/
  double GetValue(double t //In: The time for which you want the value of
		           //    this dynamic float.
		  ) {return (*this)(t);}


  /*************************************************************************
   * Method: operator () (double)
   * Summary: Returns the value of this attribute at the passed in time.
   * Description: When you want to find the value of any dynamic type,
   *              you need to provide SPEEDES with the time at which
   *              you want the value.  This attribute return the value
   *              -INFINITY if you request a value for which there is
   *              not a valid time.  Be warned that this is a valid
   *              number and you need to check if it makes sense in
   *              your situation.  We hope to remedy this in the
   *              future by adding a global variable that can be
   *              checked for an error condition if -INFINITY is
   *              returned.
   * Return: the value of the variable at the passed in time or
   *         -INFINITY if the time is invalid
   *************************************************************************/
  double operator () (double t //In: The time for which you want the value of
		               //    this dynamic float.
		      );

  /*************************************************************************
   * Method: operator () (double, double &)
   * Summary: DO NOT USE
   * Description: DO NOT USE
   * Return: -1
   *************************************************************************/
  double operator () (double t, double &V);

  /*************************************************************************
   * Method: operator () (double, double &, double &)
   * Summary: DO NOT USE
   * Description: DO NOT USE
   * Return: -1
   *************************************************************************/
  double operator () (double t, double &V, double &A);

  /*************************************************************************
   * Method: GetType()
   * Summary: Returns the type of this attribute: DYNAMIC_FLOAT_ATTRIBUTE_TYPE
   * Description: Each type is enumerated so we can properly find them and
   *              the enumeration can be found at the top of SpExportAttribute.H.
   * Return: Returns the type of this attribute: DYNAMIC_FLOAT_ATTRIBUTE_TYPE
   *************************************************************************/
  virtual int GetType() {return DYNAMIC_FLOAT_ATTRIBUTE_TYPE;}

  /*************************************************************************
   * Method: GetSize()
   * Summary: Returns the size of a DYNAMIC_FLOAT_ATTRIBUTE
   * Description: Returns the size of a DYNAMIC_FLOAT_ATTRIBUTE
   * Return: Returns the size of a DYNAMIC_FLOAT_ATTRIBUTE
   *************************************************************************/
  virtual int GetSize() {return sizeof(DYNAMIC_FLOAT_ATTRIBUTE);}

protected:

private:
  void delete_vector_with_child_type() { delete [] this; }

};

/************************************************************************
 * Class: DYNAMIC_INT_ATTRIBUTE
 *
 * Summary: A time dependant attribute integer type.
 * 
 * Description: If you want to have an integer type that has dynamic
 *              value (values that change over time), then this is the
 *              correct type to use.  Unlike the STATIC_INT_ATTRIBUTE,
 *              it is not a drop in replacement for a normal integer.
 *              That is because the accessors for dynamic types
 *              require that you provide a time at which you want the
 *              value.
 *
 * Examples: None of the demo code currently has examples of the use
 *           of this type
************************************************************************/
class DYNAMIC_INT_ATTRIBUTE : public BASE_DYNAMIC_ATTRIBUTE {

public:


  /*************************************************************************
   * Method: DYNAMIC_INT_ATTRIBUTE()
   * Summary: Default constructor for a DYNAMIC_INT_ATTRIBUTE
   * Description: The default constructor simply instantiates the
   *              member variables which means that the list of values
   *              is empty.
   * Return: a new DYNAMIC_INT_ATTRIBUTE
   *************************************************************************/
  DYNAMIC_INT_ATTRIBUTE() {;}

  /*************************************************************************
   * Method: GetValue(double)
   * Summary: Returns the value of this attribute at the passed in time.
   * Description: When you want to find the value of any dynamic type,
   *              you need to provide SPEEDES with the time at which
   *              you want the value.  This attribute return the value
   *              -1 if you request a value for which there is not a
   *              valid time.  Be warned that this is a valid number
   *              and you need to check if it makes sense in your
   *              situation.  We hope to remedy this in the future by
   *              adding a global variable that can be checked for an
   *              error condition if -1 is returned.
   * Return: the value of the variable at the passed in time or
   *         -1 if the time is invalid
   *************************************************************************/
  int GetValue(double t//In: The time for which you want the value of
		       //    this dynamic integer.
		  ) {return (*this)(t);}


  /*************************************************************************
   * Method: operator () (double)
   * Summary: Returns the value of this attribute at the passed in time.
   * Description: When you want to find the value of any dynamic type,
   *              you need to provide SPEEDES with the time at which
   *              you want the value.  This attribute return the value
   *              -1 if you request a value for which there is not a
   *              valid time.  Be warned that this is a valid number
   *              and you need to check if it makes sense in your
   *              situation.  We hope to remedy this in the future by
   *              adding a global variable that can be checked for an
   *              error condition if -1 is returned.
   * Return: the value of the variable at the passed in time or
   *         -1 if the time is invalid
   *************************************************************************/
  int operator () (double t//In: The time for which you want the value of
		           //    this dynamic integer.
		  ) {
    DYNAMIC_INT_ITEM *item = (DYNAMIC_INT_ITEM *)FindDynamicItem(t);
    if (item) {
      return (*item)(t);
    }else{
      return -1;
    }
  }


  /*************************************************************************
   * Method: operator () (double, int &)
   * Summary: Returns the value of this attribute at the passed in time.
   * Description: This operator is identical to that of operator () (double)
   *              except that the second parameter is set to 0.
   * Return: the value of the variable at the passed in time or
   *         -1 if the time is invalid
   *************************************************************************/
  int operator () (double t,//In: The time for which you want the value of
		            //    this dynamic integer.
		   int &V   //Out: Value is set to zero
		   ) {
    V = 0;
    DYNAMIC_INT_ITEM *item = (DYNAMIC_INT_ITEM *)FindDynamicItem(t);
    if (item) {
      return (*item)(t);
    }else{
      return -1;
    }
  }


  /*************************************************************************
   * Method: operator () (double, int &, int &)
   * Summary: Returns the value of this attribute at the passed in time.
   * Description: This operator is identical to that of operator () (double)
   *              except that the second  and third parameters are set to 0.
   * Return: the value of the variable at the passed in time or
   *         -1 if the time is invalid
   *************************************************************************/
  int operator () (double t, //In: The time for which you want the value of
		             //    this dynamic integer.
		   int &V,   //Out: Value is set to zero
		   int &A    //Out: Value is set to zero
		   ) {
    V = 0;
    A = 0;
    DYNAMIC_INT_ITEM *item = (DYNAMIC_INT_ITEM *)FindDynamicItem(t);
    if (item) {
      return (*item)(t);
    }else{
      return -1;
    }
  }

  /*************************************************************************
   * Method: GetType()
   * Summary: Returns the type of this attribute: DYNAMIC_INT_ATTRIBUTE_TYPE
   * Description: Each type is enumerated so we can properly find them and
   *              the enumeration can be found at the top of SpExportAttribute.H.
   * Return: Returns the type of this attribute: DYNAMIC_INT_ATTRIBUTE_TYPE
   *************************************************************************/
  virtual int GetType() {return DYNAMIC_INT_ATTRIBUTE_TYPE;}

  /*************************************************************************
   * Method: GetSize()
   * Summary: Returns the size of a DYNAMIC_INT_ATTRIBUTE
   * Description: Returns the size of a DYNAMIC_INT_ATTRIBUTE
   * Return: Returns the size of a DYNAMIC_INT_ATTRIBUTE
   *************************************************************************/
   virtual int GetSize() {return sizeof(DYNAMIC_INT_ATTRIBUTE);}

protected:

private:
  void delete_vector_with_child_type() { delete [] this; }

};

/************************************************************************
 * Class: DYNAMIC_LOGICAL_ATTRIBUTE
 *
 * Summary: A time dependant attribute integer type.
 * 
 * Description: If you want to have an boolean type that has dynamic
 *              value (values that change over time), then this is the
 *              correct type to use.  Unlike the
 *              STATIC_LOGICAL_ATTRIBUTE, it is not a drop in
 *              replacement for a normal boolean.  That is because the
 *              accessors for dynamic types require that you provide a
 *              time at which you want the value.
 *
 * Examples: None of the demo code currently has examples of the use
 *           of this type
************************************************************************/
class DYNAMIC_LOGICAL_ATTRIBUTE : public BASE_DYNAMIC_ATTRIBUTE {

public:

  /*************************************************************************
   * Method: DYNAMIC_LOGICAL_ATTRIBUTE()
   * Summary: Default constructor for a DYNAMIC_LOGICAL_ATTRIBUTE
   * Description: The default constructor simply instantiates the
   *              member variables which means that the list of values
   *              is empty.
   * Return: a new DYNAMIC_LOGICAL_ATTRIBUTE
   *************************************************************************/
  DYNAMIC_LOGICAL_ATTRIBUTE() {;}
  int GetValue(double t) {return (*this)(t);}

  /*************************************************************************
   * Method: GetValue(double)
   * Summary: Returns the value of this attribute at the passed in time.
   * Description: When you want to find the value of any dynamic type,
   *              you need to provide SPEEDES with the time at which
   *              you want the value.  This attribute return the value
   *              -1 if you request a value for which there is not a
   *              valid time.  Be warned that this is a valid number
   *              and you need to check if it makes sense in your
   *              situation.  We hope to remedy this in the future by
   *              adding a global variable that can be checked for an
   *              error condition if -1 is returned.
   * Return: the value of the variable at the passed in time or
   *         -1 if the time is invalid
   *************************************************************************/
  int operator () (double t//In: The time for which you want the value of
		            //    this dynamic integer.
		   ) {
    DYNAMIC_LOGICAL_ITEM *item = (DYNAMIC_LOGICAL_ITEM *)FindDynamicItem(t);
    if (item) {
      return (*item)(t);
    }else{
      return -1;
    }
  }

  /*************************************************************************
   * Method: operator () (double, int &)
   * Summary: Returns the value of this attribute at the passed in time.
   * Description: This operator is identical to that of operator () (double)
   *              except that the second parameter is set to 0.
   * Return: the value of the variable at the passed in time or
   *         -1 if the time is invalid
   *************************************************************************/
  int operator () (double t,//In: The time for which you want the value of
		            //    this dynamic integer.
		   int &V   //Out: Value is set to zero
		   ) {
    V = 0;
    DYNAMIC_LOGICAL_ITEM *item = (DYNAMIC_LOGICAL_ITEM *)FindDynamicItem(t);
    if (item) {
      return (*item)(t);
    }else{
      return -1;
    }
  }

  /*************************************************************************
   * Method: operator () (double, int &, int &)
   * Summary: Returns the value of this attribute at the passed in time.
   * Description: This operator is identical to that of operator () (double)
   *              except that the second  and third parameters are set to 0.
   * Return: the value of the variable at the passed in time or
   *         -1 if the time is invalid
   *************************************************************************/
  int operator () (double t,//In: The time for which you want the value of
		            //    this dynamic integer.
		   int &V,  //Out: Value is set to zero
		   int &A   //Out: Value is set to zero
		   ) {
    V = 0;
    A = 0;
    DYNAMIC_LOGICAL_ITEM *item = (DYNAMIC_LOGICAL_ITEM *)FindDynamicItem(t);
    if (item) {
      return (*item)(t);
    }else{
      return -1;
    }
  }


  /*************************************************************************
   * Method: GetType()
   * Summary: Returns the type of this attribute:
   *          DYNAMIC_LOGICAL_ATTRIBUTE_TYPE
   * Description: Each type is enumerated so we can properly find them and
   *              the enumeration can be found at the top of SpExportAttribute.H.
   * Return: Returns the type of this attribute: DYNAMIC_LOGICAL_ATTRIBUTE_TYPE
   *************************************************************************/
  virtual int GetType() {return DYNAMIC_LOGICAL_ATTRIBUTE_TYPE;}

  /*************************************************************************
   * Method: GetSize()
   * Summary: Returns the size of a DYNAMIC_LOGICAL_ATTRIBUTE
   * Description: Returns the size of a DYNAMIC_LOGICAL_ATTRIBUTE
   * Return: Returns the size of a DYNAMIC_LOGICAL_ATTRIBUTE
   *************************************************************************/
  virtual int GetSize() {return sizeof(DYNAMIC_LOGICAL_ATTRIBUTE);}

protected:

private:
  void delete_vector_with_child_type() { delete [] this; }

};


/************************************************************************
 * DYNAMIC_OBJECT_ATTRIBUTE						*
 ************************************************************************/
class DYNAMIC_OBJECT_ATTRIBUTE : public BASE_DYNAMIC_ATTRIBUTE {

private:
  void delete_vector_with_child_type() { delete [] this; }

protected:

public:

  /*************************************************************************
   * Method: DYNAMIC_OBJECT_ATTRIBUTE()
   * Summary: Default constructor for a DYNAMIC_OBJECT_ATTRIBUTE
   * Description: The default constructor simply instantiates the
   *              member variables which means that the list of values
   *              is empty.
   * Return: a new DYNAMIC_OBJECT_ATTRIBUTE
   *************************************************************************/
  DYNAMIC_OBJECT_ATTRIBUTE() {;}

  /*************************************************************************
   * Method: GetValue(double)
   * Summary: Returns the value of this attribute at the passed in time.
   * Description: When you want to find the value of any dynamic type,
   *              you need to provide SPEEDES with the time at which
   *              you want the value.  This attribute return the value
   *              NULL if you request a value for which there is not a
   *              valid time.
   * Return: the value of the variable at the passed in time or
   *         NULL if the time is invalid
   *************************************************************************/
  DYNAMIC_OBJECT *GetValue(double t) {return (*this)(t);}

  /*************************************************************************
   * Method: operator () (double)
   * Summary: Returns the value of this attribute at the passed in time.
   * Description: When you want to find the value of any dynamic type,
   *              you need to provide SPEEDES with the time at which
   *              you want the value.  This attribute return the value
   *              NULL if you request a value for which there is not a
   *              valid time.
   * Return: the value of the variable at the passed in time or
   *         NULL if the time is invalid
   *************************************************************************/
  DYNAMIC_OBJECT *operator () (double t);

  /*************************************************************************
   * Method: GetType()
   * Summary: Returns the type of this attribute: DYNAMIC_OBJECT_ATTRIBUTE_TYPE
   * Description: Each type is enumerated so we can properly find them and
   *              the enumeration can be found at the top of SpExportAttribute.H.
   * Return: Returns the type of this attribute: DYNAMIC_OBJECT_ATTRIBUTE_TYPE
   *************************************************************************/
  virtual int GetType() {return DYNAMIC_OBJECT_ATTRIBUTE_TYPE;}

  /*************************************************************************
   * Method: GetSize()
   * Summary: Returns the size of a DYNAMIC_OBJECT_ATTRIBUTE
   * Description: Returns the size of a DYNAMIC_OBJECT_ATTRIBUTE
   * Return: Returns the size of a DYNAMIC_OBJECT_ATTRIBUTE
   *************************************************************************/
  virtual int GetSize() {return sizeof(DYNAMIC_OBJECT_ATTRIBUTE);}

};


/************************************************************************
 * Class: DYNAMIC_POSITION_ATTRIBUTE
 *
 * Summary: A time dependant attribute position type.
 * 
 * Description: Position attributes have many additional features that
 *              are different from other ATTRIBUTE types.
 *              Essentially, position attributes are equations of
 *              motion over given periods of time.  If you think of a
 *              plane flying from Colorado Springs to SanDiego.  It
 *              could have several different equations of motion for
 *              this trip.  There would be an equation both for
 *              takeoff and landing as well as several equations
 *              corresponding to course changes throughout the route.
 *
 *              At each point in the trip, a
 *              DYNAMIC_POSITION_ATTRIBUTE allows you to find the
 *              position (in one of three coordinate systems), the
 *              velocity, and the acceleration all at a specific point
 *              in time.  These values are then computed through the
 *              equations of motion that you have provided or used
 *              (SPEEDES provides several motion types). This results
 *              in a significant savings in the amount of data that
 *              must be passed.  It also distributes the work of
 *              computing positions since different objects on
 *              different nodes may be interested in the position at
 *              different times.
 *
 * Examples: None of the demo code currently has examples of the use
 *           of this type
************************************************************************/
class DYNAMIC_POSITION_ATTRIBUTE : public BASE_DYNAMIC_ATTRIBUTE {

public:

  DYNAMIC_POSITION_ATTRIBUTE() {;}
  void GetValue(double t, int PT, double x[3]) {(*this)(t, PT, x);}


  /*************************************************************************
   * Method: operator () (double, int, double[3]);
   * Summary: Computes the position for the given time and coordinate system
   * Description: 
   * Return: void
   *************************************************************************/
  void operator () (double t,   //In: Time at which the position should
		                //    be computed
		    int PT,     //In: Coordinate system in which to
		                //    evaluate the position.  Options are:
                                //    EARTH: Latitude, Longitude, Altitude
                                //    ECR: X, Y, Z in a rotating earth
		                //         coordinate system
		                //    ECI: X, Y, Z in an inertial
		                //         coordinate system

		    double x[3]// Out: Vector to be filled with the position
		    );

  /*************************************************************************
   * Method: operator () (double, int, double[3], double[3]);
   * Summary: Computes the position and velocity for the given time
   *          and coordinate system
   * Description: 
   * Return: void
   *************************************************************************/
  void operator () (double t,   //In: Time at which the position should
		                //    be computed
		    int PT,     //In: Coordinate system in which to
		                //    evaluate the position.  Options are:
                                //    EARTH: Latitude, Longitude, Altitude
                                //    ECR: X, Y, Z in a rotating earth
		                //         coordinate system
		                //    ECI: X, Y, Z in an inertial
		                //         coordinate system

		    double x[3],// Out: Vector to be filled with the position
		    double v[3]// Out: Vector to be filled with the velocity
		    );

  /*************************************************************************
   * Method: operator () (double, int, double[3], double[3], double[3])
   * Summary: Computes the position, velocity, and acceleration for
   *          the given time and coordinate system
   * Description: 
   * Return: void
   *************************************************************************/
  void operator () (double t,   //In: Time at which the position should
		                //    be computed
		    int PT,     //In: Coordinate system in which to
		                //    evaluate the position.  Options are:
                                //    EARTH: Latitude, Longitude, Altitude
                                //    ECR: X, Y, Z in a rotating earth
		                //         coordinate system
		                //    ECI: X, Y, Z in an inertial
		                //         coordinate system

		    double x[3],// Out: Vector to be filled with the position
		    double v[3],// Out: Vector to be filled with the velocity
		    double a[3] // Out: Vector to be filled with the
		                //      acceleration
		    );

  /*************************************************************************
   * Method: CheckMotion()
   * Summary: Checks the motion for realistic derivative values.
   * Description: Once you have defined a new motion type, it is
   *              important that this motion type provide realistic
   *              values for the first two derivatives: Velocity and
   *              Acceleration.  It does this by approximating the
   *              first two derivatives with a finite difference
   *              method and then checks to see if the computed values
   *              match the approximated values to within a reasonable
   *              tolerance level.  If the error is too large, the
   *              status code returned is incremented and an error
   *              message with the error is printed.
   *
   *              The motion is only checked over individual segments.
   *              It is not checked to make sure that the transition
   *              from one motion segment to another segment is
   *              continuous to any degree.
   *
   * Return: 1 for every inconsistency it finds
   *************************************************************************/
  int CheckMotion();

  /*************************************************************************
   * Method: CheckTime()
   * Summary: Checks the motion for continuous time.
   * Description: Once you have defined a dynamic position attribute,
   *              it is important that it provides a position for
   *              every time between it's start time and it's end
   *              time. If it is not continuous, it prints an error
   *              message and returns a non-zero status.  It does not
   *              check to make sure that the velocity and
   *              acceleration are continuous.
   * Return: 1 for every gap in the continuity
   *************************************************************************/
  int CheckTime();

  /*************************************************************************
   * Method: GetType()
   * Summary: Returns the type of this attribute:
   *          DYNAMIC_POSITION_ATTRIBUTE_TYPE
   * Description: Each type is enumerated so we can properly find them and
   *              the enumeration can be found at the top of SpExportAttribute.H.
   * Return: Returns the type of this attribute:
   *         DYNAMIC_POSITION_ATTRIBUTE_TYPE
   *************************************************************************/
  virtual int GetType() {return DYNAMIC_POSITION_ATTRIBUTE_TYPE;}

  /*************************************************************************
   * Method: GetSize()
   * Summary: Returns the size of a DYNAMIC_POSITION_ATTRIBUTE
   * Description: Returns the size of a DYNAMIC_POSITION_ATTRIBUTE
   * Return: Returns the size of a DYNAMIC_POSITION_ATTRIBUTE
   *************************************************************************/
  virtual int GetSize() {return sizeof(DYNAMIC_POSITION_ATTRIBUTE);}

protected:

private:
  void delete_vector_with_child_type() { delete [] this; }

};




#endif

