// SpExtrapolate.H header files defines the extrapolation function. The
// equation is X = X0 + V0(t) + A0(t*t)/2 + J0(t*t*t)/6. Note
// that J0 is the jerk term.

#ifndef Extrap_def
#define Extrap_def

#include "SpStdarg.H"
#include "SpMath.H"

#include "SpDynItem.H"
#include "SpNetTypes.H"

class DYNAMIC_EXTRAPOLATE : public DYNAMIC_FLOAT_ITEM {

  private:

  protected:

    NET_DOUBLE X0;
    NET_DOUBLE V0;
    NET_DOUBLE A0;
    NET_DOUBLE J0;

  public:

  DYNAMIC_EXTRAPOLATE() {;} // Constructor doesn't do anything

/*
** Initialize routine using X0 and V0 (a0 and j0 are optional)
*/

    void Init(double ts, double te, double x0, double v0,
	      double a0=0.0, double j0=0.0) {
      StartTime = ts;
      EndTime = te;
      X0 = x0;
      V0 = v0;
      A0 = a0;
      J0 = j0;
    }

/*
** Compute the value as a function of time
*/

    virtual double operator () (double t) {
      double T = t - StartTime;
      double T2 = T*T;
      double T3 = T*T2;

      double X = X0 + V0*T + A0*T2/2 + J0*T3/6;

      return X;
    }

/*
** Compute the value and first derivative as functions of time
*/

    virtual double operator () (double t, double &V) {
      double T = t - StartTime;
      double T2 = T*T;
      double T3 = T*T2;

      double X = X0 + V0*T + A0*T2/2 + J0*T3/6;
      V = V0 + A0*T + J0*T2/2;

      return X;
    }

/*
** Compute the value, with first, and second derivatives as functions of time
*/

    virtual double operator () (double t, double &V, double &A) {
      double T = t - StartTime;
      double T2 = T*T;
      double T3 = T*T2;

      double X = X0 + V0*T + A0*T2/2 + J0*T3/6;
      V = V0 + A0*T + J0*T2/2;
      A = A0 + J0*T;

      return X;
    }


/*
** return a string describing the dynamic float type
*/

    virtual char *GetType() {return "Dynamic Extrapolation Type";}

};

#endif
