// CircularOrbit.C file

#include "SpCircularOrbit.H"

void CIRCULAR_ORBIT::init(double I, double A, double Altitude,
                            double P, double Time) {

  T0 = Time;
  StartTime = -1.0e20;
  EndTime = 1.0e20;

  Phase = P;
  Amplitude = RE+Altitude;
  Omega = sqrt(GM/Amplitude)/Amplitude;

  double CosA = cos(A);
  double SinA = sin(A);

  double CosI = cos(I);
  double SinI = sin(I);

  Ux[0] = CosA;
  Ux[1] = SinA;
  Ux[2] = 0.0;

  Uy[0] = -SinA*CosI;
  Uy[1] = CosA*CosI;
  Uy[2] = SinI;

  Uz[0] = SinA*SinI;
  Uz[1] = -CosA*SinI;
  Uz[2] = CosI;

}

int CIRCULAR_ORBIT::CheckUnitVectors() {
  int ReturnCode = 1;
  double X[3];

/*
** Check that the magnitudes are one
*/

  VectorCopy(X, Ux);
  if ((Magnitude(X) > 1.000001) || (Magnitude(X) < 0.999999)) {
    ReturnCode = 0;
    cout << "Bad Ux Magnitude" << endl;
  } 

  VectorCopy(X, Uy);
  if ((Magnitude(X) > 1.000001) || (Magnitude(X) < 0.999999)) {
    ReturnCode = 0;
    cout << "Bad Uy Magnitude" << endl;
  }

  VectorCopy(X, Uz);
  if ((Magnitude(X) > 1.000001) || (Magnitude(X) < 0.999999)) {
    ReturnCode = 0;
    cout << "Bad Uz Magnitude" << endl;
  }

/*
** Check that the dot products with each other are zero
*/

  if ((DotProduct(Ux, Uz) > 0.000001) || (DotProduct(Ux,Uz) < -0.000001)) {
    ReturnCode = 0;
    cout << "Bad Ux DOT Uz" << endl;
  }

  if ((DotProduct(Ux, Uy) > 0.000001) || (DotProduct(Ux,Uy) < -0.000001)) {
    ReturnCode = 0;
    cout << "Bad Ux DOT Uy" << endl;
  }

  if ((DotProduct(Uy, Uz) > 0.000001) || (DotProduct(Uy,Uz) < -0.000001)) {
    ReturnCode = 0;
    cout << "Bad Ux DOT Uz" << endl;
  }

/*
** Check Cross Products
*/

  CrossProduct(Ux, Uy, X);
  if ((DotProduct(Uz, X) > 1.000001) || (DotProduct(Uz, X) < 0.999999)) {
    ReturnCode = 0;
    cout << "Bad Ux CROS Uy" << endl;
  }

  CrossProduct(Uy, Uz, X);
  if ((DotProduct(Ux, X) > 1.000001) || (DotProduct(Ux, X) < 0.999999)) {
    ReturnCode = 0;
    cout << "Bad Uy CROS Uz" << endl;
  }

  CrossProduct(Uz, Ux, X);
  if ((DotProduct(Uy, X) > 1.000001) || (DotProduct(Uy, X) < 0.999999)) {
    ReturnCode = 0;
    cout << "Bad Uz CROS Ux" << endl;
  }

  return ReturnCode;

}

void CIRCULAR_ORBIT::operator () (double t, int PT, double X[3]) {
  double X0[3];
  double Y0[3];
  double dt = t-T0;
  double Angle = Omega*dt + Phase;
  double CosAngle = cos(Angle);
  double SinAngle = sin(Angle);

  VectorCopy(X0, Ux);
  VectorCopy(Y0, Uy);

  ScalarMultiply(X0, (CosAngle*Amplitude));
  ScalarMultiply(Y0, (SinAngle*Amplitude));
  Sum(X0, Y0, X);
  
  Convert(X, PositionType, X, PT, t);

}

void CIRCULAR_ORBIT::operator () (double t, int PT, double X[3], double V[3]) {
  double X0[3];
  double Y0[3];
  double dt = t-T0;
  double Angle = Omega*dt + Phase;
  double CosAngle = cos(Angle);
  double SinAngle = sin(Angle);

  VectorCopy(X0, Ux);
  VectorCopy(Y0, Uy);

  ScalarMultiply(X0, (CosAngle*Amplitude));
  ScalarMultiply(Y0, (SinAngle*Amplitude));
  Sum(X0, Y0, X);
  
  CrossProduct(Uz, X, V);
  ScalarMultiply(V, Omega);

  Convert(X, V, PositionType, X, V, PT, t);

}

void CIRCULAR_ORBIT::operator () (double t, int PT, double X[3],
				double V[3], double A[3]) {
  double X0[3];
  double Y0[3];
  double dt = t-T0;
  double Angle = Omega*dt + Phase;
  double CosAngle = cos(Angle);
  double SinAngle = sin(Angle);

  VectorCopy(X0, Ux);
  VectorCopy(Y0, Uy);

  ScalarMultiply(X0, (CosAngle*Amplitude));
  ScalarMultiply(Y0, (SinAngle*Amplitude));
  Sum(X0, Y0, X);

  CrossProduct(Uz, X, V);
  ScalarMultiply(V, Omega);

  VectorCopy(A, X);
  Normalize(A);
  ScalarMultiply(A, -GM/(Amplitude*Amplitude));

  Convert(X, V, A, PositionType, X, V, A, PT, t);

}
