// SpExtrapolateMotion.H header file defines the extrapolation motion type. The
// equation of motion is X = X0 + V0(t) + A0(t*t)/2 + J0(t*t*t)/6. Note
// that J0 is the jerk term.

#ifndef ExtrapMot_def
#define ExtrapMot_def

#include "SpStdarg.H"
#include "SpMath.H"

#include "SpDynItem.H"
#include "SpNetTypes.H"

class EXTRAPOLATE_MOTION : public DYNAMIC_POSITION_ITEM {

  private:

  protected:

    NET_DOUBLE X0[3];
    NET_DOUBLE V0[3];
    NET_DOUBLE A0[3];
    NET_DOUBLE J0[3];

  public:

  EXTRAPOLATE_MOTION() {;} // Constructor doesn't do anything

/*
** Initialize routine using X0 and V0
*/

    void Init(double ts, double te,
	      double x0[3], double v0[3],
	      int PT) {
      int i;

      StartTime = ts;
      EndTime = te;

      for (i=0; i<3; i++) {
	X0[i] = x0[i];
	V0[i] = v0[i];
	A0[i] = 0.0;
	J0[i] = 0.0;
      }

      PositionType = PT;

      if (PositionType == EARTH) {
	cout << "Error, Extrapolate motion must be ECR or ECI" << endl;
      }

    }

/*
** Initialize routine using X0, V0, and A0
*/

    void Init(double ts, double te,
	      double x0[3], double v0[3], double a0[3],
	      int PT) {
      int i;

      StartTime = ts;
      EndTime = te;

      for (i=0; i<3; i++) {
	X0[i] = x0[i];
	V0[i] = v0[i];
	A0[i] = a0[i];
	J0[i] = 0.0;
      }

      PositionType = PT;

      if (PositionType == EARTH) {
	cout << "Error, Extrapolate motion must be ECR or ECI" << endl;
      }

    }

/*
** Initialize routine using X0, V0, A0, and J0 (the Jerk term)
*/

    void Init(double ts, double te,
	      double x0[3], double v0[3], double a0[3], double j0[3],
	      int PT) {
      int i;

      StartTime = ts;
      EndTime = te;

      for (i=0; i<3; i++) {
	X0[i] = x0[i];
	V0[i] = v0[i];
	A0[i] = a0[i];
	J0[i] = j0[i];
      }

      PositionType = PT;

      if (PositionType == EARTH) {
	cout << "Error, Extrapolate motion must be ECR or ECI" << endl;
      }

    }

/*
** Compute the position as a function of time
*/

    virtual void operator () (double t, int PT, double X[3]) {
      double T1 = t - StartTime;
      double T2 = T1*T1;
      double T3 = T1*T2;
      int i;

      for (i=0; i<3; i++) {
	X[i] = X0[i] + V0[i]*T1 + A0[i]*T2/2 + J0[i]*T3/6;
      }
      
      Convert(X, PositionType, X, PT, t);
    }

/*
** Compute the position and velocity as functions of time
*/

    virtual void operator () (double t, int PT, double X[3], double V[3]) {
      double T1 = t - StartTime;
      double T2 = T1*T1;
      double T3 = T1*T2;
      int i;

      for (i=0; i<3; i++) {
	X[i] = X0[i] + V0[i]*T1 + A0[i]*T2/2 + J0[i]*T3/6;
	V[i] = V0[i] + A0[i]*T1 + J0[i]*T2/2;
      }
      
      Convert(X, V, PositionType, X, V, PT, t);
    }

/*
** Compute the position, velocity, and accelaration as functions of time
*/

    virtual void operator () (double t, int PT, double X[3], double V[3],
			      double A[3]) {
      double T1 = t - StartTime;
      double T2 = T1*T1;
      double T3 = T1*T2;
      int i;

      for (i=0; i<3; i++) {
	X[i] = X0[i] + V0[i]*T1 + A0[i]*T2/2 + J0[i]*T3/6;
	V[i] = V0[i] + A0[i]*T1 + J0[i]*T2/2;
	A[i] = A0[i] + J0[i]*T1;
      }
      
      Convert(X, V, A, PositionType, X, V, A, PT, t);
    }

/*
** return a string describing the motion type
*/

    virtual char *GetType() {return "Extrapolation Motion Type";}

};

#endif
