// SpGreatCircle.C file

#include "SpGreatCircle.H"

double GREAT_CIRCLE::init_Vconstant(
	   double LatInit, double LonInit, double AltInit, double TimeInit,
	   double LatFin, double LonFin, double AltFin,
	   double V) {

  double Temp1[3];
  //double Temp2[3];
  double Uf[3];
  double Period;
  double R0;

/*
** First, construct the unit vectors U0 and U1 in ECR coordinates that are
** in the plane of the initial and final vectors. U0 is the vector of the
** starting position. U1 is perpendicular to U0.
*/

  Temp1[0] = LatInit;
  Temp1[1] = LonInit;
  Temp1[2] = AltInit;

  EARTH_to_ECR(Temp1, U0);
  Normalize(U0);

  Temp1[0] = LatFin;
  Temp1[1] = LonFin;
  Temp1[2] = AltFin;

  EARTH_to_ECR(Temp1, Uf);
  Normalize(Uf);
  CrossProduct(U0, Uf, Temp1);
  CrossProduct(Temp1, U0, U1);
  Normalize(U1);

/*
** Theta is the angle between the starting and final position vectors.
** DeltaAltitude is the difference between final and initial altitudes.
*/

  double dp = DotProduct(U0, Uf);
  if (dp > 1)
    dp = 1;
  
  Theta = acos(dp);
  DeltaAltitude = AltFin - AltInit;
  R0 = RE+AltInit;
  Alt0 = AltInit;

  if (DeltaAltitude == 0.0) {
    Period = Theta*R0/V;
  }else{
    Period = (fabs(DeltaAltitude)/V)
      *sqrt(1.0+Squared(Theta/log(1.0+DeltaAltitude/R0)));
  }

  StartTime = TimeInit;
  EndTime = StartTime + Period;

/*
** Velocity components (perpendicular and parallel to ground)
*/

  Vperp = DeltaAltitude / Period;
  Vpar = sqrt(V*V-Vperp*Vperp);

  return EndTime;

}

double GREAT_CIRCLE::init_Tconstant(
	   double LatInit, double LonInit, double AltInit, double TimeInit,
	   double LatFin, double LonFin, double AltFin, double TimeFin) {

  double Temp1[3];
  //double Temp2[3];
  double Uf[3];
  double V;
  double Period;
  double R0;

/*
** First, construct the unit vectors U0 and U1 in ECR coordinates that are
** in the plane of the initial and final vectors. U0 is the vector of the
** starting position. U1 is perpendicular to U0.
*/

  Temp1[0] = LatInit;
  Temp1[1] = LonInit;
  Temp1[2] = AltInit;

  EARTH_to_ECR(Temp1, U0);
  Normalize(U0);

  Temp1[0] = LatFin;
  Temp1[1] = LonFin;
  Temp1[2] = AltFin;

  EARTH_to_ECR(Temp1, Uf);
  Normalize(Uf);
  CrossProduct(U0, Uf, Temp1);
  CrossProduct(Temp1, U0, U1);
  Normalize(U1);

/*
** Theta is the angle between the starting and final position vectors.
** DeltaAltitude is the difference between final and initial altitudes.
*/

  double dp = DotProduct(U0, Uf);
  if (dp > 1)
    dp = 1;
  
  Theta = acos(dp);
  DeltaAltitude = AltFin - AltInit;
  R0 = RE+AltInit;
  Alt0 = AltInit;
  Period = TimeInit - TimeFin;

  if (DeltaAltitude == 0.0) {
    V = Theta*R0/Period;
  }else{
    V = (DeltaAltitude/Period)*
      sqrt(1.0+Squared(Theta/log(1.0+DeltaAltitude/R0)));
  }

  StartTime = TimeInit;
  EndTime = TimeFin;

/*
** Velocity components (perpendicular and parallel to ground)
*/

  Vperp = DeltaAltitude / Period;
  Vpar = sqrt(V*V-Vperp*Vperp);

  return V;

}

void GREAT_CIRCLE::operator () (double t, int PT, double X[3]) {
  double dt = t-StartTime;
  double Theta1;
  double CosTheta;
  double SinTheta;
  double Mag;

  if (t > EndTime) {
    cout << "Warning (Great Circle): Time greater than end time" << endl;
  }

  if (DeltaAltitude == 0.0) {
    Theta1 = Theta*dt/(EndTime-StartTime);
  }else{
    Theta1 = (Vpar/Vperp)*(log(1.0+dt*Vperp/(RE+Alt0)));
  }

  CosTheta = cos(Theta1);
  SinTheta = sin(Theta1);
  Mag = RE + Alt0 + DeltaAltitude*dt/(EndTime-StartTime);

  X[0] = (U0[0]*CosTheta + U1[0]*SinTheta) * Mag;
  X[1] = (U0[1]*CosTheta + U1[1]*SinTheta) * Mag;
  X[2] = (U0[2]*CosTheta + U1[2]*SinTheta) * Mag;

  Convert(X, PositionType, X, PT, t);

}

void GREAT_CIRCLE::operator () (double t, int PT, double X[3], double V[3]) {
  double U[3];
  double Temp1[3];
  double Temp2[3];

  (*this)(t, ECR, X);
  VectorCopy(U,X);
  Normalize(U);

  V[0] = U[0]*Vperp;
  V[1] = U[1]*Vperp;
  V[2] = U[2]*Vperp;

  if (StartTime != t) {

    CrossProduct(U0,U,Temp1);
    CrossProduct(Temp1,U,Temp2);
    Normalize(Temp2);
  }

  else
    VectorCopy(Temp2, U1);

  V[0] += Temp2[0]*Vpar;
  V[1] += Temp2[1]*Vpar;
  V[2] += Temp2[2]*Vpar;

  Convert(X, V, PositionType, X, V, PT, t);

}

void GREAT_CIRCLE::operator () (double t, int PT, double X[3],
				double V[3], double A[3]) {
  double U[3];
  double N[3];
  double Temp1;
  double Temp2;
  double R;

  (*this)(t, ECR, X, V);

  VectorCopy(U,X);
  Normalize(U);

  VectorCopy(N,V);
  Normalize(N);

  R = Magnitude(X);

  Temp1 = Vperp*Vpar/R;
  Temp2 = Vpar*Vpar/R;
  A[0] = Temp1*N[0] - Temp2*U[0];
  A[1] = Temp1*N[1] - Temp2*U[1];
  A[2] = Temp1*N[2] - Temp2*U[2];

  Convert(X, V, A, PositionType, X, V, A, PT, t);

}
