#ifndef LoiterMotion_Objects
#define LoiterMotion_Objects

#include "SpStdarg.H"
#include "SpMath.H"

#include "SpDynItem.H"
#include "SpNetTypes.H"

class LOITER_MOTION : public DYNAMIC_POSITION_ITEM {
  private:
  protected:
    NET_DOUBLE Ux[3];
    NET_DOUBLE Uy[3];
    NET_DOUBLE Uz[3];
    NET_DOUBLE Radius;
    NET_DOUBLE Speed;
    NET_DOUBLE Altitude;
  public:

    LOITER_MOTION() {;}

    void Init(double ts, double te,
	      double x0, double x1, double x2,
	      double r, double s, int PT) {
      StartTime = ts;
      EndTime = te;
      Uz[0] = x0;
      Uz[1] = x1;
      Uz[2] = x2;
      Radius = r;
      if (Radius < 0.001) Radius = 0.001;
      Speed = s;
      if (PT == EARTH) {
	PositionType = ECR;
	Convert(Uz, EARTH, Uz, ECR);
      }else{
	PositionType = PT;
      }
      Altitude = Magnitude(Uz) - RE;
      MakeUnitVectors();
    }

    void Init(double ts, double te,
	      double x[3],
	      double r, double s, int PT) {
      StartTime = ts;
      EndTime = te;
      Uz[0] = x[0];
      Uz[1] = x[1];
      Uz[2] = x[2];
      Radius = r;
      if (Radius < 0.001) Radius = 0.001;
      Speed = s;
      if (PT == EARTH) {
	PositionType = ECR;
	Convert(Uz, EARTH, Uz, ECR);
      }else{
	PositionType = PT;
      }
      Altitude = Magnitude(Uz) - RE;
      MakeUnitVectors();
    }

    void MakeUnitVectors() {
      double U[3];
      Normalize(Uz);
      U[0] = 0.0;
      U[1] = 0.0;
      U[2] = 1.0;
      CrossProduct(Uz, U, Uy);
      Normalize(Uy);
      CrossProduct(Uy, Uz, Ux);
    }

    virtual void operator () (double t, int PT, double X[3]) {
      double t0 = t - StartTime;
      double T = Radius/Speed;
      double t1 = t0-T;
      double Mag = RE+Altitude;
      double Phi;
      double SinPhi;
      double CosPhi;
      double SinTheta;
      double CosTheta;
      double Rperp;
      double Rz;

      if (t1 < 0.0) {
	Phi = t0*Speed/Mag;
	SinPhi = sin(Phi);
	CosPhi = cos(Phi);
	X[0] = Mag*(Uz[0]*cos(Phi) + Ux[0]*sin(Phi));
	X[1] = Mag*(Uz[1]*cos(Phi) + Ux[1]*sin(Phi));
	X[2] = Mag*(Uz[2]*cos(Phi) + Ux[2]*sin(Phi));
      }else{
	Phi = T*Speed/Mag;
	SinPhi = sin(Phi);
	CosPhi = cos(Phi);
	Rperp = Mag*SinPhi;
	Rz = Mag*CosPhi;
	SinTheta = sin(t1*Speed/Rperp);
	CosTheta = cos(t1*Speed/Rperp);
	X[0] = Rperp*(Ux[0]*CosTheta + Uy[0]*SinTheta) + Rz*Uz[0];
	X[1] = Rperp*(Ux[1]*CosTheta + Uy[1]*SinTheta) + Rz*Uz[1];
	X[2] = Rperp*(Ux[2]*CosTheta + Uy[2]*SinTheta) + Rz*Uz[2];
      }

      Convert(X, PositionType, X, PT, t);
    }

    virtual void operator () (double t, int PT, double X[3], double V[3]) {
      double t0 = t - StartTime;
      double T = Radius/Speed;
      double t1 = t0-T;
      double Mag = RE+Altitude;
      double Phi;
      double SinPhi;
      double CosPhi;
      double SinTheta;
      double CosTheta;
      double Rperp;
      double Rz;

      if (t1 < 0.0) {
	Phi = t0*Speed/Mag;
	SinPhi = sin(Phi);
	CosPhi = cos(Phi);

	X[0] = Mag*(Uz[0]*CosPhi + Ux[0]*SinPhi);
	X[1] = Mag*(Uz[1]*CosPhi + Ux[1]*SinPhi);
	X[2] = Mag*(Uz[2]*CosPhi + Ux[2]*SinPhi);

	V[0] = Speed*(-Uz[0]*SinPhi + Ux[0]*CosPhi);
	V[1] = Speed*(-Uz[1]*SinPhi + Ux[1]*CosPhi);
	V[2] = Speed*(-Uz[2]*SinPhi + Ux[2]*CosPhi);

      }else{
	Phi = T*Speed/Mag;
	SinPhi = sin(Phi);
	CosPhi = cos(Phi);
	Rperp = Mag*SinPhi;
	Rz = Mag*CosPhi;
	SinTheta = sin(t1*Speed/Rperp);
	CosTheta = cos(t1*Speed/Rperp);

	X[0] = Rperp*(Ux[0]*CosTheta + Uy[0]*SinTheta) + Rz*Uz[0];
	X[1] = Rperp*(Ux[1]*CosTheta + Uy[1]*SinTheta) + Rz*Uz[1];
	X[2] = Rperp*(Ux[2]*CosTheta + Uy[2]*SinTheta) + Rz*Uz[2];

	V[0] = Speed*(-Ux[0]*SinTheta + Uy[0]*CosTheta);
	V[1] = Speed*(-Ux[1]*SinTheta + Uy[1]*CosTheta);
	V[2] = Speed*(-Ux[2]*SinTheta + Uy[2]*CosTheta);

      }
      Convert(X, V, PositionType, X, V, PT, t);
    }

    virtual void operator () (double t, int PT, double X[3], double V[3],
			      double A[3]) {
      double t0 = t - StartTime;
      double T = Radius/Speed;
      double t1 = t0-T;
      double Mag = RE+Altitude;
      double Phi;
      double SinPhi;
      double CosPhi;
      double SinTheta;
      double CosTheta;
      double Rperp;
      double Rz;

      if (t1 < 0.0) {
	Phi = t0*Speed/Mag;
	SinPhi = sin(Phi);
	CosPhi = cos(Phi);

	X[0] = Mag*(Uz[0]*CosPhi + Ux[0]*SinPhi);
	X[1] = Mag*(Uz[1]*CosPhi + Ux[1]*SinPhi);
	X[2] = Mag*(Uz[2]*CosPhi + Ux[2]*SinPhi);

	V[0] = Speed*(-Uz[0]*SinPhi + Ux[0]*CosPhi);
	V[1] = Speed*(-Uz[1]*SinPhi + Ux[1]*CosPhi);
	V[2] = Speed*(-Uz[2]*SinPhi + Ux[2]*CosPhi);

	A[0] = -(Speed*Speed/Mag) * (Uz[0]*CosPhi + Ux[0]*SinPhi);
	A[1] = -(Speed*Speed/Mag) * (Uz[1]*CosPhi + Ux[1]*SinPhi);
	A[2] = -(Speed*Speed/Mag) * (Uz[2]*CosPhi + Ux[2]*SinPhi);

      }else{
	Phi = T*Speed/Mag;
	SinPhi = sin(Phi);
	CosPhi = cos(Phi);
	Rperp = Mag*SinPhi;
	Rz = Mag*CosPhi;
	SinTheta = sin(t1*Speed/Rperp);
	CosTheta = cos(t1*Speed/Rperp);

	X[0] = Rperp*(Ux[0]*CosTheta + Uy[0]*SinTheta) + Rz*Uz[0];
	X[1] = Rperp*(Ux[1]*CosTheta + Uy[1]*SinTheta) + Rz*Uz[1];
	X[2] = Rperp*(Ux[2]*CosTheta + Uy[2]*SinTheta) + Rz*Uz[2];

	V[0] = Speed*(-Ux[0]*SinTheta + Uy[0]*CosTheta);
	V[1] = Speed*(-Ux[1]*SinTheta + Uy[1]*CosTheta);
	V[2] = Speed*(-Ux[2]*SinTheta + Uy[2]*CosTheta);

	A[0] = -(Speed*Speed/Rperp) * (Ux[0]*CosTheta + Uy[0]*SinTheta);
	A[1] = -(Speed*Speed/Rperp) * (Ux[1]*CosTheta + Uy[1]*SinTheta);
	A[2] = -(Speed*Speed/Rperp) * (Ux[2]*CosTheta + Uy[2]*SinTheta);

      }
      Convert(X, V, A, PositionType, X, V, A, PT, t);
    }

    virtual char *GetType() {return "Constant Motion Type";}

};

#endif
