#ifndef PolyMotion_Objects
#define PolyMotion_Objects

#include "SpStdarg.H"
#include "SpMath.H"

#include "SpMatrixEqu.H"
#include "SpDynItem.H"
#include "SpNetTypes.H"

#define END_POLY_MOTION -1.123e20
#define N_MAX_COEFF_POLY_MOTION 11

class POLY_N_MOTION : public DYNAMIC_POSITION_ITEM {
  private:
  protected:

    NET_DOUBLE **A0;
    NET_DOUBLE **A1;
    NET_DOUBLE **A2;

    NET_DOUBLE *B0;
    NET_DOUBLE *B1;
    NET_DOUBLE *B2;

  public:

    POLY_N_MOTION() {
      A0 = A1 = A2 = NULL;
      B0 = B1 = B2 = NULL;
    }

    void Reset() {
      int i;
      int n = GetDim();

      if (A0 != NULL) {
	for (i=0; i<n; i++) delete A0[i];
	delete A0;
	A0 = NULL;
      }

      if (A1 != NULL) {
	for (i=0; i<n; i++) delete A1[i];
	delete A1;
	A1 = NULL;
      }

      if (A2 != NULL) {
	for (i=0; i<n; i++) delete A2[i];
	delete A2;
	A2 = NULL;
      }

      if (B0 != NULL) {delete B0; B0 = NULL;}
      if (B1 != NULL) {delete B1; B1 = NULL;}
      if (B2 != NULL) {delete B2; B2 = NULL;}

    }

    void AddPoints(double ts, double te,
		   double t, double v0, double v1, double v2,
		   double err2, ...);

    void MakePoly() {
      NET_DOUBLE X[N_MAX_COEFF_POLY_MOTION];
      NET_DOUBLE *c;
      //int status;
      int i;
      int n=GetDim();

// for Cx

      /*status = */matrix_equation(A0, B0, X, n);
      c = GetCx();
      for (i=0; i<n; i++) c[i] = X[i];

// for Cy

      /*status = */matrix_equation(A1, B1, X, n);
      c = GetCy();
      for (i=0; i<n; i++) c[i] = X[i];

// for Cz

      /*status = */matrix_equation(A2, B2, X, n);
      c = GetCz();
      for (i=0; i<n; i++) c[i] = X[i];

    }

/*
** Evaluate Methods (does single dimension - not all three at once)
*/

    void Evaluate(NET_DOUBLE *c, double t, int n, double &X) {
      BIG_FLOAT ti = 1;
      BIG_FLOAT X1;
      BIG_FLOAT T = t - StartTime;

      X1 = 0.0;

      for (int i=0; i<n; i++) {
	X1 += c[i]*ti;
	ti *= T;
      }
      X = X1;
    }

    void Evaluate(NET_DOUBLE *c, double t, int n, double &X, double &V) {
      BIG_FLOAT X1, V1;
      BIG_FLOAT ti = 1;
      BIG_FLOAT T = t - StartTime;

      X1 = 0.0;
      V1 = 0.0;

      for (int i=0; i<n; i++) {
	X1 += c[i]*ti;
	if (i<n-1) V1 += (i+1)*c[i+1]*ti;
	ti *= T;
      }
      X = X1;
      V = V1;
    }

    void Evaluate(NET_DOUBLE *c, double t, int n, double &X, double &V,
		  double &A) {
      BIG_FLOAT X1, V1, A1_local_variable;
      BIG_FLOAT ti = 1;
      BIG_FLOAT T = t - StartTime;

      X1 = 0.0;
      V1 = 0.0;
      A1_local_variable = 0.0;

      for (int i=0; i<n; i++) {
	X1 += c[i]*ti;
	if (i<n-1) V1 += (i+1)*c[i+1]*ti;
	if (i<n-2) A1_local_variable += (i+2)*(i+1)*c[i+2]*ti;
	ti *= T;
      }
      X = X1;
      V = V1;
      A = A1_local_variable;
    }

/*
** Overloaded operators to evaluate position
*/

    virtual void operator () (double t, int PT, double X[3]) {
      Evaluate(GetCx(), t, GetDim(), X[0]);
      Evaluate(GetCy(), t, GetDim(), X[1]);
      Evaluate(GetCz(), t, GetDim(), X[2]);
      Convert(X, PositionType, X, PT, t);
    }

    virtual void operator () (double t, int PT, double X[3], double V[3]) {
      Evaluate(GetCx(), t, GetDim(), X[0], V[0]);
      Evaluate(GetCy(), t, GetDim(), X[1], V[1]);
      Evaluate(GetCz(), t, GetDim(), X[2], V[2]);
      Convert(X, V, PositionType, X, V, PT, t);
    }

    virtual void operator () (double t, int PT, double X[3], double V[3],
			      double A[3]) {
      Evaluate(GetCx(), t, GetDim(), X[0], V[0], A[0]);
      Evaluate(GetCy(), t, GetDim(), X[1], V[1], A[1]);
      Evaluate(GetCz(), t, GetDim(), X[2], V[2], A[2]);
      Convert(X, V, A, PositionType, X, V, A, PT, t);
    }

/*
** Virtual functions that should be supplied by the derived child
*/

    virtual int GetDim() {return 0;}

    virtual NET_DOUBLE *GetCx() {return NULL;}
    virtual NET_DOUBLE *GetCy() {return NULL;}
    virtual NET_DOUBLE *GetCz() {return NULL;}

    virtual char *GetType() {return "Dynamic poly N";}
};

class POLY_1_MOTION : public POLY_N_MOTION {
  private:
    NET_DOUBLE Cx[2];
    NET_DOUBLE Cy[2];
    NET_DOUBLE Cz[2];
  protected:
  public:
    virtual int GetDim() {return 2;}
    virtual NET_DOUBLE *GetCx() {return Cx;}
    virtual NET_DOUBLE* GetCy() {return Cy;}
    virtual NET_DOUBLE* GetCz() {return Cz;}
    virtual char *GetType() {return "Dynamic poly 1";}
};

class POLY_2_MOTION : public POLY_N_MOTION {
  private:
    NET_DOUBLE Cx[3];
    NET_DOUBLE Cy[3];
    NET_DOUBLE Cz[3];
  protected:
  public:
    virtual int GetDim() {return 3;}
    virtual NET_DOUBLE* GetCx() {return Cx;}
    virtual NET_DOUBLE* GetCy() {return Cy;}
    virtual NET_DOUBLE* GetCz() {return Cz;}
    virtual char *GetType() {return "Dynamic poly 2";}
};

class POLY_3_MOTION : public POLY_N_MOTION {
  private:
    NET_DOUBLE Cx[4];
    NET_DOUBLE Cy[4];
    NET_DOUBLE Cz[4];
  protected:
  public:
    virtual int GetDim() {return 4;}
    virtual NET_DOUBLE* GetCx() {return Cx;}
    virtual NET_DOUBLE* GetCy() {return Cy;}
    virtual NET_DOUBLE* GetCz() {return Cz;}
    virtual char *GetType() {return "Dynamic poly 3";}
};

class POLY_4_MOTION : public POLY_N_MOTION {
  private:
    NET_DOUBLE Cx[5];
    NET_DOUBLE Cy[5];
    NET_DOUBLE Cz[5];
  protected:
  public:
    virtual int GetDim() {return 5;}
    virtual NET_DOUBLE* GetCx() {return Cx;}
    virtual NET_DOUBLE* GetCy() {return Cy;}
    virtual NET_DOUBLE* GetCz() {return Cz;}
    virtual char *GetType() {return "Dynamic poly 4";}
};

class POLY_5_MOTION : public POLY_N_MOTION {
  private:
    NET_DOUBLE Cx[6];
    NET_DOUBLE Cy[6];
    NET_DOUBLE Cz[6];
  protected:
  public:
    virtual int GetDim() {return 6;}
    virtual NET_DOUBLE* GetCx() {return Cx;}
    virtual NET_DOUBLE* GetCy() {return Cy;}
    virtual NET_DOUBLE* GetCz() {return Cz;}
    virtual char *GetType() {return "Dynamic poly 5";}
};

class POLY_6_MOTION : public POLY_N_MOTION {
  private:
    NET_DOUBLE Cx[7];
    NET_DOUBLE Cy[7];
    NET_DOUBLE Cz[7];
  protected:
  public:
    virtual int GetDim() {return 7;}
    virtual NET_DOUBLE* GetCx() {return Cx;}
    virtual NET_DOUBLE* GetCy() {return Cy;}
    virtual NET_DOUBLE* GetCz() {return Cz;}
    virtual char *GetType() {return "Dynamic poly 6";}
};

class POLY_7_MOTION : public POLY_N_MOTION {
  private:
    NET_DOUBLE Cx[8];
    NET_DOUBLE Cy[8];
    NET_DOUBLE Cz[8];
  protected:
  public:
    virtual int GetDim() {return 8;}
    virtual NET_DOUBLE* GetCx() {return Cx;}
    virtual NET_DOUBLE* GetCy() {return Cy;}
    virtual NET_DOUBLE* GetCz() {return Cz;}
    virtual char *GetType() {return "Dynamic poly 7";}
};

class POLY_8_MOTION : public POLY_N_MOTION {
  private:
    NET_DOUBLE Cx[9];
    NET_DOUBLE Cy[9];
    NET_DOUBLE Cz[9];
  protected:
  public:
    virtual int GetDim() {return 9;}
    virtual NET_DOUBLE* GetCx() {return Cx;}
    virtual NET_DOUBLE* GetCy() {return Cy;}
    virtual NET_DOUBLE* GetCz() {return Cz;}
    virtual char *GetType() {return "Dynamic poly 8";}
};

class POLY_9_MOTION : public POLY_N_MOTION {
  private:
    NET_DOUBLE Cx[10];
    NET_DOUBLE Cy[10];
    NET_DOUBLE Cz[10];
  protected:
  public:
    virtual int GetDim() {return 10;}
    virtual NET_DOUBLE* GetCx() {return Cx;}
    virtual NET_DOUBLE* GetCy() {return Cy;}
    virtual NET_DOUBLE* GetCz() {return Cz;}
    virtual char *GetType() {return "Dynamic poly 9";}
};

class POLY_10_MOTION : public POLY_N_MOTION {
  private:
    NET_DOUBLE Cx[11];
    NET_DOUBLE Cy[11];
    NET_DOUBLE Cz[11];
  protected:
  public:
    virtual int GetDim() {return 11;}
    virtual NET_DOUBLE* GetCx() {return Cx;}
    virtual NET_DOUBLE* GetCy() {return Cy;}
    virtual NET_DOUBLE* GetCz() {return Cz;}
    virtual char *GetType() {return "Dynamic poly 10";}
};

#endif
