// SpRhumbLine.C file

#include "SpRhumbLine.H"
#include "SpGeoFunctions.H"

static double signum (double x) {
  return ((x == 0) ? 1 :(x/fabs(x)));
} 

static double LAT_to_THETA(double LAT)
  { return ( PI / 2 - DEG_to_RAD(LAT) ); }

static double LNG_to_PHI(double LNG)
  { return ( DEG_to_RAD(LNG) + PI ); }

static double angular_abs_deg(double v0) {
  double r = fmod(v0, 360.0);

  if (r<0)
    r+=360.0;

  return r;
}

static double angular_abs_rad(double v0) {
  double r = fmod(v0,2*PI);
  
  if (r<0) 
    r+=2*PI;

  return r;
}

static double lng_abs(double lng) {
  lng = angular_abs_deg(lng);

  if (lng > 180)
    lng -= 360;

  return lng;
}

static double lat_abs(double lat) {
  lat = lng_abs(lat);
  
  if (lat > 90)
    lat = 180 - lat;
  
  if (lat < -90)
    lat = -180 - lat;

  return lat;
}

double RHUMB_LINE::init_Vconstant(
	   double LatInit, double LonInit, double AltInit, double TimeInit,
	   double LatFin, double LonFin, double AltFin,
	   double V) {

  double E1[3];
  double GroundDist, ActualDist;
  double Period;
  double R0;

  E0[0] = LatInit;
  E0[1] = LonInit;
  E0[2] = AltInit;

  E1[0] = LatFin;
  E1[1] = LonFin;
  E1[2] = AltFin;

  EARTH_to_ECR(E1, U1);
  Normalize(U1);

  Speed = V;

  StartTime = TimeInit;

/*
** DeltaAltitude is the difference between final and initial altitudes.
*/

  R0 = RE + AltInit;
  DeltaAltitude = AltFin - AltInit;

  double theta1 = LAT_to_THETA(lat_abs(LatInit));
  double phi1 = angular_abs_rad(LNG_to_PHI(LonInit));
  double theta2 = LAT_to_THETA(lat_abs(LatFin));
  double phi2 = angular_abs_rad(LNG_to_PHI(LonFin));

  double Dphi = phi2 - phi1;
  double e1 = signum(Dphi) * signum(PI - fabs(Dphi));
  if (fabs(Dphi) > PI) Dphi = e1* (2*PI - fabs(Dphi));
  
  // kludge: do great circle at the poles! rl and gc are the same there!

  if ((fabs(LatInit) == 90) || (fabs(LatFin) == 90)) {
    
    if (!G_C_Equiv)
      G_C_Equiv = new GREAT_CIRCLE;

    EndTime = G_C_Equiv->init_Vconstant(LatInit, LonInit, AltInit, TimeInit,
					LatFin, LonFin, AltFin, V);
  }
    
  else {

    if (theta1 == theta2) {
      GroundDist = RE * sin(theta1) * fabs(Dphi);
      Bearing = e1*PI/2;
    }

    else {
      double help = -1 * Dphi/log(fabs(tan(.5*theta1)/tan(.5*theta2)));
      GroundDist = RE * fabs(theta1-theta2)*sqrt(1 + help*help);
      Bearing = ((PI/2) * (1 + signum(theta2-theta1))) - atan(help);
    }

    if (Bearing < 0)
      Bearing += 2*PI;
    Bearing *= 180/PI;

    if (DeltaAltitude == 0.0)
      ActualDist = (R0/RE)*GroundDist;

    else {
      // not implemented
      ActualDist = (R0/RE)*GroundDist;
    }

    Period = ActualDist/V;
    EndTime = StartTime + Period;

    /*
    ** Velocity components (perpendicular and parallel to ground)
    */

    Vperp = DeltaAltitude/Period;
    Vpar = sqrt(Speed*Speed-Vperp*Vperp);
    Vground = GroundDist/Period;
  }

  return EndTime;
}

double RHUMB_LINE::init_Tconstant(
	   double LatInit, double LonInit, double AltInit, double TimeInit,
	   double LatFin, double LonFin, double AltFin, double TimeFin) {

  double E1[3];
  double GroundDist, ActualDist;
  double Period;
  double R0;

  E0[0] = LatInit;
  E0[1] = LonInit;
  E0[2] = AltInit;

  E1[0] = LatFin;
  E1[1] = LonFin;
  E1[2] = AltFin;

  EARTH_to_ECR(E1, U1);
  Normalize(U1);

  StartTime = TimeInit;
  EndTime = TimeFin;

/*
** DeltaAltitude is the difference between final and initial altitudes.
*/

  R0 = RE + AltInit;
  DeltaAltitude = AltFin - AltInit;
  Period = TimeInit - TimeFin;

  double theta1 = LAT_to_THETA(lat_abs(LatInit));
  double phi1 = angular_abs_rad(LNG_to_PHI(LonInit));
  double theta2 = LAT_to_THETA(lat_abs(LatFin));
  double phi2 = angular_abs_rad(LNG_to_PHI(LonFin));

  double Dphi = phi2 - phi1;
  double e1 = signum(Dphi) * signum(PI - fabs(Dphi));
  if (fabs(Dphi) > PI) Dphi = e1* (2*PI - fabs(Dphi));
  
  // kludge: do great circle at the poles! rl and gc are the same there!

  if ((fabs(LatInit) == 90) || (fabs(LatFin) == 90)) {

    if (!G_C_Equiv)
      G_C_Equiv = new GREAT_CIRCLE;

    return G_C_Equiv->init_Tconstant(LatInit, LonInit, AltInit, TimeInit,
				     LatFin, LonFin, AltFin, TimeFin);
  }

  if (theta1 == theta2)
    GroundDist = RE * sin(theta1) * fabs(Dphi);

  else {
    double help = -1 * Dphi/log(fabs(tan(.5*theta1)/tan(.5*theta2)));
    GroundDist = RE * fabs(theta1-theta2)*sqrt(1 + help*help);
  }

  if (DeltaAltitude == 0.0)
    ActualDist = (R0/RE)*GroundDist;

  else {
    // not implemented
    ActualDist = (R0/RE)*GroundDist;
  }

  Speed = ActualDist/Period;

/*
** Velocity components (perpendicular and parallel to ground)
*/

  Vperp = DeltaAltitude/Period;
  Vpar = sqrt(Speed*Speed-Vperp*Vperp);
  Vground = GroundDist/Period;

  return Speed;
}

/* If Vground is determined correctly (but it isn't currently for
 * changes in altitude), the following should work fine.
 */

void RHUMB_LINE::operator () (double t, int PT, double X[3]) {
  double E_New[3];
  double dt = t - StartTime;
  double Mag;

  if (G_C_Equiv) {

    (* G_C_Equiv)(t, PT, X);
    return;
  }

  if (t > EndTime)
    cout << "Warning (Rhumb Line): Time greater than end time" << endl;

/*
** Use elapsed time and ground speed to calculate intermediate
** ground distance. Use old bearing and new distance to calculate
** new position.
*/

  double NewDist = dt*Vground;

  double LatInit = lat_abs(E0[0]);
  double LonInit = lng_abs(E0[1]);

  double brg = angular_abs_rad(DEG_to_RAD(Bearing));
  double theta1 = PI/2 - LatInit*(PI/180);
  double phi1 = LonInit*(PI/180) + PI;

  double theta2, d_phi;
  
  if ((brg == PI/2) || (brg == 3*PI/2)) {

    theta2 = theta1;
    d_phi = signum(PI - brg)*(NewDist/RE)/sin(theta1);
  }

  else {
    int coeff = 1;

    if ((brg <= PI/2) || (brg >= 3*PI/2))
      coeff = -1;

    double help = tan((PI/2)*(1+coeff) - brg);

    theta2 = theta1 + 
      coeff*(NewDist/RE)/sqrt(1 + help*help);

    d_phi = help*log(fabs(tan(theta2/2)/tan(theta1/2)));
  }

  double x = theta2;
  double y = phi1 + d_phi;

  double LatNew = (PI/2 - x);
  double LonNew = (y - PI);
  
  if (fabs(LonNew) > PI)
    LonNew -= 2*PI * signum(LonNew);

  LatNew *= (180/PI);
  LonNew *= (180/PI); // degrees from radians;

  Mag = RE + E0[2] + DeltaAltitude*dt/(EndTime-StartTime);
  
  E_New[0] = LatNew;
  E_New[1] = LonNew;
  E_New[2] = Mag;

  EARTH_to_ECR(E_New, X);

  Convert(X, PositionType, X, PT, t);
}

/* The velocity vector parallel to the Earth should be oriented the
 * same as a vector from the current position to the final position.
 * (Contrasted to GreatCircle, where the velocity vector parallel to
 * the Earth should be oriented the same as a vector from the starting
 * position to the current position.)
 */

void RHUMB_LINE::operator () (double t, int PT, double X[3], double V[3]) {
  double U[3];
  double Temp1[3];
  double Temp2[3];

  if (G_C_Equiv) {

    (* G_C_Equiv)(t, PT, X, V);
    return;
  }

  (*this)(t, ECR, X);
  VectorCopy(U,X);
  Normalize(U);

  V[0] = U[0]*Vperp;
  V[1] = U[1]*Vperp;
  V[2] = U[2]*Vperp;

  CrossProduct(U,U1,Temp1);
  CrossProduct(Temp1,U,Temp2);
  Normalize(Temp2);

  V[0] += Temp2[0]*Vpar;
  V[1] += Temp2[1]*Vpar;
  V[2] += Temp2[2]*Vpar;

  Convert(X, V, PositionType, X, V, PT, t);
}

/* The following does not calculate the acceleration vector correctly. */

void RHUMB_LINE::operator () (double t, int PT, double X[3],
			      double V[3], double A[3]) {
  double U[3];
  double N[3];
  double Temp1;
  double Temp2;
  double R;

  if (G_C_Equiv) {

    (* G_C_Equiv)(t, PT, X, V, A);
    return;
  }

  (*this)(t, ECR, X, V);

  VectorCopy(U,X);
  Normalize(U);

  VectorCopy(N,V);
  Normalize(N);

  R = Magnitude(X);

  Temp1 = Vperp*Vpar/R;
  Temp2 = Vpar*Vpar/R;
  A[0] = Temp1*N[0] - Temp2*U[0];
  A[1] = Temp1*N[1] - Temp2*U[1];
  A[2] = Temp1*N[2] - Temp2*U[2];

  Convert(X, V, A, PositionType, X, V, A, PT, t);
}
