// C_SpRadar.H header file defines a radar component that works with
// precomputations

#ifndef RADAR_COMP_OBJ
#define RADAR_COMP_OBJ

#include "C_SpSensor.H"
#include "SpEventHandler.H"
#include "SpMath.H"

/*
** Three Event Handlers attached to the Radar Component:
*/
/******************************************************************************
* H_RADAR_BEGIN_PRECOMP - Enacted by a DeliverProxyPtr event (which signals that
* a new remote proxy has been added to this SimObj), this event handler
* first checks if the new proxy represents an object my sensor is interested
* in.  If so, it kicks off the precomp process.
******************************************************************************/
class H_RADAR_BEGIN_PRECOMP : public SpEventHandler {
  public:
    H_RADAR_BEGIN_PRECOMP();
    virtual void Process(char *eventName, SpSimTime &timeTag,
                       char *data, int dataBytes,
                       int externalId);
};

/******************************************************************************
* H_RADAR_END_PRECOMP - Enacted by a RemoveProxyPtr event (which signals
* that a remote proxy has been removed from this SimObj), this event handler 
* checks if there is a gate associated with this proxy.  If so, kill the
* precomp process, delete the gate, and do other cleanup.
******************************************************************************/
class H_RADAR_END_PRECOMP : public SpEventHandler {
  public:
    H_RADAR_END_PRECOMP();
    virtual void Process(char *eventName, SpSimTime &timeTag,
                       char *data, int dataBytes,
                       int externalId);
};

/******************************************************************************
* H_RADAR_REFLECT_ATTRIBUTES - Enacted by a TouchProxy event (which signals 
* that the object represented by this proxy has been modified), this event 
* handler checks if there is a gate associated with this proxy.  If so, set 
* that gate. 
******************************************************************************/
class H_RADAR_REFLECT_ATTRIBUTES  : public SpEventHandler {
  public:
    H_RADAR_REFLECT_ATTRIBUTES();
    virtual void Process(char *eventName, SpSimTime &timeTag,
                       char *data, int dataBytes,
                       int externalId);
};

/******************************************************************************
* H_RADAR_UPDATE_ATTRIBUTES - Enacted by a ScheduleUpdate event (which signals
* that this SimObj has been modified), this event handler opens all gates.
******************************************************************************/
class H_RADAR_UPDATE_ATTRIBUTES : public SpEventHandler {
  public:
    H_RADAR_UPDATE_ATTRIBUTES();
    virtual void Process(char *eventName, SpSimTime &timeTag,
                       char *data, int dataBytes,
                       int externalId);
};


/*
** Radar Component Class Declaration:
*/
/******************************************************************************
* C_SpRadar - handles the precomputation for radars
******************************************************************************/
class C_SpRadar : public C_SpSensor {

  private:
    RB_double MaxRange;
    RB_double PowerConstant;  // assumes radar equation of the following form: 
                              // signal = (PowerConstant * CrossSection)/R^4 
    H_RADAR_BEGIN_PRECOMP BeginPrecomp;
    H_RADAR_END_PRECOMP EndPrecomp;
    H_RADAR_REFLECT_ATTRIBUTES ReflectAttributes; // remote object changed
    H_RADAR_UPDATE_ATTRIBUTES UpdateAttributes;   // this object changed


  protected:
    void SetMaxRange(double maxRange) {MaxRange = maxRange;}
    double GetMaxRange() {return MaxRange;}
    double GetMaxRange(double crossSection)
    {double maxRange = MaxRange;
     if (PowerConstant) {
       maxRange = sqrt(sqrt(crossSection/PowerConstant));
       if (maxRange > MaxRange) maxRange = MaxRange;
     }  
     return maxRange;}
    void SetDetectionRange(double range, double crossSection)
      {PowerConstant = crossSection / (range*range*range*range);}
   

  public:  
    C_SpRadar();

};

typedef C_SpRadar RADAR_COMPONENT;

#endif
