#include "SpCoverageCalc.H"
#include "SpObjProxy.H"
#include "xpriority.H"

// tolerance in kilometers
#define TOLERANCE 0.0000000001
// width in seconds
#define ZERO_WIDTH 0.0000001
#define MAX_ITERS 100

// f(t) = X(t)*X(t) - R^2
static double compute_c_term(double X[3], double R) {

  return (DotProduct(X, X) - Squared(R));
}

// f'(t) = 2*X(t)*V(t)
static double compute_b_term(double X[3], double V[3]) {

  return (2 * DotProduct(X, V));
}

// (1/2) * f''(t) = V(t)*V(t) + X(t)*A(t)
static double compute_a_term(double X[3], double V[3], double A[3]) {

  return (DotProduct(V, V) + DotProduct(X, A));
}

int getSphericalCoverageState(SpObjProxy *sensor_plat_proxy,
			      SpObjProxy *target_plat_proxy,
			      char *position_ref,
			      double radius,
			      double time) {

  DYNAMIC_POSITION_ATTRIBUTE *sensor_position =
    (DYNAMIC_POSITION_ATTRIBUTE *) sensor_plat_proxy->Find(position_ref);

  DYNAMIC_POSITION_ATTRIBUTE *target_position =
    (DYNAMIC_POSITION_ATTRIBUTE *) target_plat_proxy->Find(position_ref);

  double X_s[3];
  double V_s[3];
  double X_t[3];
  double V_t[3];
  double X[3];
  double V[3];

  (*sensor_position)(time, ECR, X_s, V_s);
  (*target_position)(time, ECR, X_t, V_t);

  Difference(X_s, X_t, X);
  Difference(V_s, V_t, V);

  double c = compute_c_term(X, radius);

  if (fabs(c) <= TOLERANCE) {
    double b = compute_b_term(X, V);

    if (b < 0)
      return CC_ENTER;

    else if (b > 0)
      return CC_EXIT;
    
    else
      return CC_STABLE;
  }

  else if (c < 0)
    return CC_IN;

  else
    return CC_OUT;
}

// If f(t) = X*X - R^2, then f'(t) = 2*X*V and f''(t) = 2*[V*V + X*A].
// We want to solve f(t) == 0 with t = t0 + dt. Using a Taylor's series
// expansion to three terms we have that: 
// f(t) == 0 == f(t0) + dt*f'(t0) + ((dt^2)/2)*f''(t0)
// Solving for dt using the quadratic formula gives us 0, 1, or 2 choice
// for the next t0 to use in the iteration. (If we have 2 choices, the
// lowest next t0 is chosen that is within bounds.)

// The following funtion finds a zero for f(t) within the bounds
// given to within a certain tolerance. There may be some machine
// dependent, or other, reason that limits how small the tolerance can
// be set. If so, this needs to be given further thought. f(t) must be
// continuous and differentiable within the lower and upper bounds given.

static int steinman_iteration(DYNAMIC_POSITION_ATTRIBUTE *sensor_position,
			      DYNAMIC_POSITION_ATTRIBUTE *target_position,
			      double radius,
			      double lower_bound,
			      double upper_bound,
			      double t0,
			      double &next_t0) {

  double X_s[3];
  double V_s[3];
  double A_s[3];

  double X_t[3];
  double V_t[3];
  double A_t[3];

  double X[3];
  double V[3];
  double A[3];

  (*sensor_position)(t0, ECR, X_s, V_s, A_s);
  (*target_position)(t0, ECR, X_t, V_t, A_t);

  Difference(X_s, X_t, X);
  Difference(V_s, V_t, V);
  Difference(A_s, A_t, A);

  double c = compute_c_term(X, radius);

  // check for f(t0) == 0 (within tolerance)
  if (fabs(c) <= TOLERANCE) {

    next_t0 = t0;
    return CC_FOUND;
  }

  double b = compute_b_term(X, V);
  double a = compute_a_term(X, V, A);

  double disc = Squared(b) - 4 * a * c;

  if ((disc < 0.0) ||
      ((a == 0) && (b == 0))) {

    next_t0 = 0;
    return CC_NONE_FOUND;
  }

  double dt_plus, dt_minus;

  if (a == 0) {

    dt_plus = -c/b;
    dt_minus = -c/b;
  }

  else {

    dt_plus = (-b + sqrt(disc))/(2 * a);
    dt_minus = (-b - sqrt(disc))/(2 * a);
  }

  double t_plus = t0 + dt_plus;
  double t_minus = t0 + dt_minus;

  int t_plus_valid = 0;
  int t_minus_valid = 0;

  if ((t_plus > lower_bound) && (t_plus < upper_bound))
    t_plus_valid = 1;

  if ((t_minus > lower_bound) && (t_minus < upper_bound))
    t_minus_valid = 1;

  // picking lowest next_t0 
  if ((!t_plus_valid) && (!t_minus_valid)) {

    next_t0 = 0;
    return CC_NONE_FOUND;
  }

  if (!t_plus_valid) {

    next_t0 = t_minus;
    return CC_TRY_AGAIN;
  }

  if (!t_minus_valid) {

    next_t0 = t_plus;
    return CC_TRY_AGAIN;
  }

  if (t_plus > t_minus) {
    
    next_t0 = t_minus;
    return CC_TRY_AGAIN;
  }
  
  else {

    next_t0 = t_plus;
    return CC_TRY_AGAIN;
  }
}

// The following function attempts to find the smallest time t for which
// the value of f(t) is zero. f(t) must be continuous and differentiable
// between the start and stop times.

// Can [start, stop] be broken up into smaller intervals, each such
// interval having only 1 zero? This could be done if the times the
// function had maxs and mins could be found, because the zeros must lie
// between these values and the endpoints. Finding the times the function
// has maxs and mins can be done by finding the zeros of f'(t). But due
// to the variety of motion types in SPEEDES, finding the zeros of f'(t)
// is almost as difficult (if not as difficult) as finding the zeros of
// f(t). Therefore this approach seems to be a dead end.

// But, given this, possibly incorrect assumption, the following function
// uses a sound method to determine the lowest zero: given an initial t0,
// the steinman_iteration method will return the smallest of the 2 roots
// closest to the initial t0. By choosing the initial t0 at the midpoint
// of the interval, and then dividing the search area in half (or greater
// than half) each time, the lowest zero that is actually within the bounds
// can be found.

// Within any given time interval, there is no guessing how many zeros
// there may actually be. (That is, there is no easy way to determine upper
// limit on the number of zeros. If within the examined time span both
// the sensor platform and the target platform are following a GreatCircle
// motion, then there will be at most 2 zeros. But we can't assume
// GreatCircle motion.) But in a neighborhood around every actual zero in
// the function, all function values will be within tolerance. To prevent
// multiple iterations from returning "phantom" zeros, it becomes
// necessary to define a zero_width, the radius around a t such that f(t)
// lies within tolerance, where it is unecessary to search for further
// zeros. It seems reasonable to say that zero_width will shrink in
// proportion to tolerance.

int binary_breakdown_search(DYNAMIC_POSITION_ATTRIBUTE *sensor_position,
			    DYNAMIC_POSITION_ATTRIBUTE *target_position,
			    double radius,
			    double start,
			    double stop,
			    double &zero_time) {
  double lower_bound = start;
  double upper_bound = stop;
  int binary_count = 1;
  double initial_t0, t0, next_t0;
  int zero_found_p = 0;
  // compiler demands initialization
  double lowest_zero = 0;

  for (;;) {
    int itr_count = 1;

    initial_t0 = t0 = (lower_bound + upper_bound) / 2.0;

    int itr_result;
    for (;;) {

      itr_result = steinman_iteration(sensor_position,
				      target_position,
				      radius,
				      lower_bound,
				      upper_bound,
				      t0,
				      next_t0);

      if (itr_result == CC_TRY_AGAIN) {

	if (itr_count > MAX_ITERS) {

	  itr_result = CC_NONE_FOUND;
	  break;
	}
	
	else {

	  itr_count++;
	  t0 = next_t0;
	}
      }
      
      else
	break;
    }

    if (itr_result == CC_FOUND) {
      zero_found_p = 1;
      lowest_zero = next_t0;

      double zero_boundary = lowest_zero - ZERO_WIDTH;

      if (zero_boundary >= initial_t0)
	upper_bound = initial_t0;
      else
	upper_bound = zero_boundary;
    }

    if ((itr_result == CC_NONE_FOUND) || (lower_bound >= upper_bound)) {

      if (zero_found_p) {

	zero_time = lowest_zero;
	return CC_FOUND;
      }
      
      else
	return CC_NONE_FOUND;
    }

    binary_count++;
  }
}

// Each SPEEDES DYNAMIC_POSITION_ITEM should have a position function that
// is both continuous and differentiable within its StartTime and EndTime.
// But a DYNAMIC_POSITION_ATTRIBUTE, constructed from individual 
// DYNAMIC_POSITION_ITEMs, need not have a position function that is
// differentiable over its span. (If the DYNAMIC_POSITION_ATTRIBUTE's
// position function is not continuous, then it has been poorly constructed
// in its application. Error checking is not used to determine discontinuity.)
// A DYNAMIC_POSITION_ATTRIBUTE's position function may be undifferentiable 
// at each point in time where it has been spliced together from individual
// motion types. These points of undifferentiability can cause havoc with 
// the iteration methods above. But by determining these splice times we can
// search for a coverage zero (when the sensor and target are exactly radius
// km. from each other) using intervals upon which both position functions can
// be sure to be differentiable. This is what the following function does to
// return the first coverage zero in the time span (begin_time, end_time).

int findNextSphericalCoverageZero(SpObjProxy *sensor_plat_proxy,
				  SpObjProxy *target_plat_proxy,
				  char *position_ref,
				  double radius,
				  double begin_time,
				  double end_time,
				  double &zero_time) {
  double start;
  double stop;

  begin_time += ZERO_WIDTH;
 
  DYNAMIC_POSITION_ATTRIBUTE *sensor_position =
    (DYNAMIC_POSITION_ATTRIBUTE *) sensor_plat_proxy->Find(position_ref);

  DYNAMIC_POSITION_ATTRIBUTE *target_position =
    (DYNAMIC_POSITION_ATTRIBUTE *) target_plat_proxy->Find(position_ref);

  RB_priority splice_times;
  C_SQ_ITEM *splice_time;

  DYNAMIC_POSITION_ITEM *segment;
  segment = (DYNAMIC_POSITION_ITEM *) sensor_position->get_top();

  while (segment) {
    double seg_start = segment->GetStartTime();

    if ((seg_start > begin_time) && (seg_start < end_time)) {

      splice_time = new C_SQ_ITEM;
      splice_time->set_time_tag(seg_start);
      splice_times.insert(splice_time);
    }

    segment = (DYNAMIC_POSITION_ITEM *) (++(*sensor_position));
  }

  segment = (DYNAMIC_POSITION_ITEM *) target_position->get_top();

  while (segment) {
    double seg_start = segment->GetStartTime();

    if ((seg_start > begin_time) && (seg_start < end_time)) {

      splice_time = new C_SQ_ITEM;
      splice_time->set_time_tag(segment->GetStartTime());
      splice_times.insert(splice_time);
    }

    segment = (DYNAMIC_POSITION_ITEM *) (++(*target_position));
  }

  splice_time = new C_SQ_ITEM;
  splice_time->set_time_tag(end_time);
  splice_times.insert(splice_time);

  start = begin_time;
  int seg_count = 1;
  
  while ((splice_time = splice_times.pop())) {
    stop = splice_time->get_time_tag();
    
    delete splice_time;
    splice_time = NULL;

    if (stop <= start)
      continue;

    seg_count++;

    int ret_val = binary_breakdown_search(sensor_position,
					  target_position,
					  radius,
					  start,
					  stop,
					  zero_time);

    if (ret_val == CC_FOUND) {

      // delete the rest of the items in the pqueue
      while ((splice_time = splice_times.pop()))
	delete splice_time;

      return CC_FOUND;
    }

    start = stop;
  }

  return CC_NONE_FOUND;
}

