// SpDynItem.H

#ifndef DynamicItem_Object
#define DynamicItem_Object

#include "qitem.H"
#include "SpConvert.H"
#include "SpNetTypes.H"

#define DYNAMIC_OBJECT SpDynItem

/************************************************************************
 * Class: SpDynItem
 *
 * Summary: The base type of types that are stored in DYNAMIC_ATTRIBUTEs
 *
 * Description: A Dynamic attribute is essentially a list of
 *              SpDynItem's and children of this class.  For some
 *              of these objects, it will be obvious how they work.
 *              For others, such as the DYNAMIC_POSITION_ITEM, things
 *              are a bit more involved.  You are free to develop your
 *              own classes and then create DYNAMIC_ATTRIBUTEs for
 *              those classes.  Just make sure they work the way you
 *              expect them to work!
 *
 *              Note that even the direct descendents of this type are
 *              not used directly by SPEEDES.  We further derive
 *              classes (some described in SpDynObjs.H which are
 *              many of the classes that are actually used to build
 *              the DYNAMIC_ATTRIBUTEs. These classes are the
 *              DYNAMIC_FLOAT_CONSTANT, DYNAMIC_LOGICAL_CONSTANT and
 *              the DYNAMIC_LOGICAL_CONSTANT along with many specific
 *              types of dynamic motions.
 *
 * Examples: SpDynItem is used in the following classes such as
 *           DYNAMIC_INT_ATTRIBUTE and DYNAMIC_POSITION_ITEM
 *
************************************************************************/

class SpDynItem : public C_ITEM {

public:

  /*************************************************************************
   * Method: SpDynItem
   * Summary: default constructor for a SpDynItem
   * Description: The default constructor initializes the start time
   *              and end timeto -1.0e20 (number commonly used as
   *              INFINITY in SPEEDES), sets the CountId to Counter,
   *              and increments Counter.  Counter is a static member
   *              that keeps track of each SpDynItem that is
   *              created and how many of these items there are.
   * Return: A new SpDynItem.
   *************************************************************************/
  SpDynItem() {
    StartTime = -1.0e20;
    EndTime = -1.e20;
    CountId = Counter;
    Counter++;
  }


  /*************************************************************************
   * Method: SetStartTime(double)
   * Summary: Sets the start time for a dynamic item.
   * Description: Each dynamic item is valid for a given amount of
   *              time.  If you think of an item such as the inventory
   *              of a company, the number of items in their inventory
   *              changes and each value is valid only for a given
   *              interval of time (although this value may repeat).
   *              This method sets the initial time for the interval in
   *              which the attribute is valid.
   * Return: void
   *************************************************************************/
  virtual void SetStartTime(double s //In: Time for which this item becomes
			             //    valid.
			    ) {StartTime = s;}

  /*************************************************************************
   * Method: SetEndTime(double)
   * Summary: Sets the end time for this dynamic item
   * Description: Each dynamic item is valid for a given amount of
   *              time.  If you think of an item such as the inventory
   *              of a company, the number of items in their inventory
   *              changes and each value is valid only for a given
   *              interval of time (although this value may repeat).
   *              This method sets the end time for the interval in 
   *              which the attribute is valid. 
   * Return: void
   *************************************************************************/
  virtual void SetEndTime(double e//In: Time for which this item stops
			          //    being valid.
			  ) {EndTime = e;}

  /*************************************************************************
   * Method: GetStartTime()
   * Summary: Returns the time this attribute becomes valid
   * Description: Each dynamic item is valid for a given amount of
   *              time.  If you think of an item such as the inventory
   *              of a company, the number of items in their inventory
   *              changes and each value is valid only for a given
   *              interval of time (although this value may repeat).
   *              This method returns the start time for the interval
   *              in which the attribute is valid.
   * Return: The time at which this item becomes valid
   *************************************************************************/
  double GetStartTime() {return StartTime;}

  /*************************************************************************
   * Method: GetEndTime()
   * Summary: Returns the time at which this item stops being valid
   * Description: Each dynamic item is valid for a given amount of
   *              time.  If you think of an item such as the inventory
   *              of a company, the number of items in their inventory
   *              changes and each value is valid only for a given
   *              interval of time (although this value may repeat).
   *              This method returns the end time for the interval in 
   *              which the attribute is valid. 
   * Return: The time at which this item stops being valid
   *************************************************************************/
  double GetEndTime() {return EndTime;}


  /*************************************************************************
   * Method: GetCountId()
   * Summary: returns the value of the CountId for this object
   * Description: The value of the CountID is used within the object
   *              proxies for updating values.  Each object has a
   *              unique ID (assuming there are less than MAX_INT
   *              DYNAMIC_ITMEs) and this allows us to know which item
   *              needs to be removed from the list in the
   *              DYNAMIC_ATTRIBUTE class
   * Return: The value of the CountId for this item
   *************************************************************************/
  int GetCountId() {return CountId;}


  /*************************************************************************
   * Method: GoodTime(double)
   * Summary: Returns 1 if passed in time is valid for this item. O otherwise
   * Description: Each dynamic item is valid for a given amount of
   *              time.  If you think of an item such as the inventory
   *              of a company, the number of items in their inventory
   *              changes and each value is valid only for a given
   *              interval of time (although this value may repeat).
   *              This method is used to check if a given item is valid
   *              for a given point in time.
   * Return: Returns 1 if passed in time is valid for this item. O otherwise
   *************************************************************************/
  int GoodTime(double t) {
    int status = 1;
    if (t < StartTime) status = 0;
    if (t > EndTime) status = 0;
    return status;
  }


  /*************************************************************************
   * Method: GetSize()
   * Summary: Returns the size of a SpDynItem
   * Description: Since these types and their derived classes are
   *              generally used in object proxies, We need to be able
   *              to find their size so that we can send them from one
   *              node to another as an update for the other proxy.
   * Return: size of a SpDynItem
   *************************************************************************/
  virtual int GetSize() {return sizeof(SpDynItem);}

  /*************************************************************************
   * Method: GetType()
   * Summary: Returns a char * corresponding to the type of this item
   * Description: Returns a char * corresponding to the type of this
   *              item.  This string needs to be unique for each
   *              class so a good choice is the name of the class.
   * Return: Returns a char * corresponding to the type of this item
   *************************************************************************/
  virtual char *GetType() =0;


protected:
  NET_DOUBLE StartTime;
  NET_DOUBLE EndTime;

  NET_INT CountId;
  static int Counter;

private:
};

/************************************************************************
 * Class: DYNAMIC_INT_ITEM
 *
 * Summary: DYNAMIC_INT_ITEM is an intermediate class used to define a
 *          dynamic integer attribute
 *
 * Description: DYNAMIC_INT_ITEM is an intermediate class used to define a
 *              dynamic integer attribute
 *
 * Examples: DYNAMIC_INT_ITEM is used as a base class for the
 *           DYNAMIC_INT_CONSTANT class defined in SpDynObjs.H
 *
************************************************************************/
class DYNAMIC_INT_ITEM : public SpDynItem {

public:

  /*************************************************************************
   * Method: operater () (double)
   * Summary: Should not be called.
   * Description: Should be overridden to return value.
   * Return: int 
   *************************************************************************/
  virtual int operator () (double) = 0;

protected:

private:

};

/************************************************************************
 * Class: DYNAMIC_FLOAT_ITEM
 *
 * Summary: Intermediary class used in the development of the
 *          DYNAMIC_FLOAT_CONSTANT class
 *
 * Description: Intermediary class used in the development of the
 *              DYNAMIC_FLOAT_CONSTANT class.  This class is never
 *              directly used and only derived versions of this class
 *              are used by SPEEDES
 *
 * Examples:DYNAMIC_FLOAT_CONSTANT class in SpDynObjs.H
 *
************************************************************************/
class DYNAMIC_FLOAT_ITEM : public SpDynItem {

public:


  /*************************************************************************
   * Method: operator () (double)
   * Summary: Should not be called
   * Description: Should be overridden to return value.
   * Return: double
   *************************************************************************/
  virtual double operator () (double) = 0;


  /*************************************************************************
   * Method: operator () (double, double &)
   * Summary: Should not be called
   * Description: Should be overridden to return value.
   * Return: double
   *************************************************************************/
  virtual double operator () (double , double &) = 0;


  /*************************************************************************
   * Method: operator () (double, double &, double &)
   * Summary: Should not be called
   * Description: Should be overridden to return value.
   * Return: -1
   *************************************************************************/
  virtual double operator () (double, double &, double &) =0;


protected:

private:

};

/************************************************************************
 * Class: DYNAMIC_LOGICAL_ITEM
 *
 * Summary: The DYNAMIC_LOGICAL_ITEM type is an intermediary class for
 *          the DYNAMIC_LOGICAL_CONSTANT.
 *
 * Description: The DYNAMIC_LOGICAL_ITEM type is an intermediary class for
 *              the DYNAMIC_LOGICAL_CONSTANT
 *
 * Examples: DYNAMIC_LOGICAL_CONSTANT in SpDynObjs.H
 *
************************************************************************/
class DYNAMIC_LOGICAL_ITEM : public SpDynItem {

public:


  /*************************************************************************
   * Method: operator() (double)
   * Summary: Used to return hte value of the logical at a point in time
   * Description: Should not be called.
   * Return: -1
   *************************************************************************/
  virtual int operator () (double)=0;

protected:

private:

};

/************************************************************************
 * Class: DYNAMIC_POSITION_ITEM
 *
 * Summary: DYNAMIC_POSITION_ITEM's are essentially equations of motion 
 *
 * Description: DYNAMIC_POSITION_ITEM is used as an intermediate type
 *              for the various equations of motions provided by
 *              SPEEDES.  Each dynamic position type can return its
 *              position in any of EARTH, ECR, or ECI coordinates when
 *              requested.
 *
 * Examples: Several examples of the use of this type can be found in
 *           POLY_N_POTION within PolyyMotion.H, RHUMB_LINE in
 *           SpRhumbLine.H and SPLINE6_MOTION in SpSpline5Motion.H.
 *           Naturally, there are also examples in ObjctProxy.C and in
 *           Attribute.[CH].
 *
************************************************************************/
class DYNAMIC_POSITION_ITEM : public SpDynItem {

public:


  /*************************************************************************
   * Method: DYNAMIC_POSITION_ITEM()
   * Summary: Default constructor of a DYNAMIC_POSITION_ITEM
   * Description: The default constructor sets the Position type to -1
   *              (an invalid value)
   * Return: a new DYNAMIC_POSITION_ITEM
   *************************************************************************/
  DYNAMIC_POSITION_ITEM() {PositionType = -1;}

  /*************************************************************************
   * Method: SetEARTH()
   * Summary: Sets the coordinate system to EARTH
   * Description: This sets the default method for returning a
   *              position to using EARTH (lat, long, alt)
   *              coordinates.
   * Return: void
   *************************************************************************/
  void SetEARTH() {PositionType = EARTH;}

  /*************************************************************************
   * Method: SetECI()
   * Summary: Sets the coordinate system to ECI
   * Description: This sets the default method for returning a
   *              position to using EARTH (X, Y, Z in a non rotating
   *              earth) coordinates.
   * Return: void
   *************************************************************************/
  void SetECI() {PositionType = ECI;}

  /*************************************************************************
   * Method: SetECR()
   * Summary: Sets the coordinate system to ECR
   * Description: This sets the default method for returning a
   *              position to using EARTH (X, Y, Z in a rotating
   *              earth) coordinates
   * Return: void
   *************************************************************************/
  void SetECR() {PositionType = ECR;}


  /*************************************************************************
   * Method: CheckMotion(double)
   * Summary: Uses finite differences to check the derivatives at a point.
   * Description: Since SPEEDES often uses the derivatives (velocity,
   *              and acceleration) to compute nearest approach as
   *              well as first contact with a moving object, it is
   *              important that the derivative values are valid.
   *              This method will use finite differences to check if
   *              the derivative values are within reasonable
   *              tolerances.  If they are not, it prints an error
   *              message and returns a non-zero value.  It is
   *              important to use this method when you are first
   *              developing a new motion type.
   * Return: 1 if velocity is bad,
   *         2 of acceleration is bad,
   *         3 if both acceleration and velocity are bad
   *         0 if the derivatives are good.
   *************************************************************************/
  int CheckMotion(double t) {
    double x1[3], x2[3];
    double v1[3], v2[3];
    double a1[3], a2[3];
    double temp[3];
    double step = 0.1;
    double Err;
    int rc = 0;

    if (t+step > EndTime) step = -1.0; {
      if (t+step < StartTime) step = StartTime - t;
    }

    (*this)(t, ECR, x1, v1, a1);
    (*this)(t+step, ECR, x2, v2, a2);

    temp[0] = (v2[0] + v1[0]) / 2.0 - (x2[0] - x1[0]) / step;
    temp[1] = (v2[1] + v1[1]) / 2.0 - (x2[1] - x1[1]) / step;
    temp[2] = (v2[2] + v1[2]) / 2.0 - (x2[2] - x1[2]) / step;

    Err = 2.0 * Magnitude(temp) / (Magnitude(v2) + Magnitude(v1));
    //      cout << "Error in velocity is " << Err << endl;
    if (Err > 0.001) {
      cout << "Error in velocity is too large " << Err << endl;
      rc = 1;
    }

    temp[0] = (a2[0] + a1[0]) / 2.0 - (v2[0] - v1[0]) / step;
    temp[1] = (a2[1] + a1[1]) / 2.0 - (v2[1] - v1[1]) / step;
    temp[2] = (a2[2] + a1[2]) / 2.0 - (v2[2] - v1[2]) / step;

    Err = 2.0 * Magnitude(temp) / (Magnitude(a2) + Magnitude(a1));
    //      cout << "Error in acceleration is " << Err << endl;
    if (Err > 0.001) {
      cout << "Error in acceleration is too large " << Err << endl;
      rc += 2;
    }

    if (rc) cout << GetType() << endl;
    return rc;
  }


  /*************************************************************************
  * Method: CheckMotion()
  * Summary: Checks the motion for invalid derivatives at the start,
  *          middle, and end.
  * Description: Since SPEEDES often uses the derivatives (velocity,
  *              and acceleration) to compute nearest approach as
  *              well as first contact with a moving object, it is
  *              important that the derivative values are valid.
  *              This method will use finite differences to check if
  *              the derivative values are within reasonable
  *              tolerances.  If they are not, it prints an error
  *              message and returns a non-zero value.  It is
  *              important to use this method when you are first
  *              developing a new motion type.
  * Return: 1 if velocity is bad at StartTime,
  *         2 of acceleration is bad at StartTime,
  *         3 if both acceleration and velocity are bad at StartTime
  *         10 if velocity is bad at the middle,
  *         20 of acceleration is bad at the middle,
  *         30 if both acceleration and velocity are bad at the middle
  *       : 100 if velocity is bad at End,
  *         200 of acceleration is bad at EndTime,
  *         300 if both acceleration and velocity are bad at EndTime
  *         0 if the derivatives are good.
  *         any sum of the above.
  *************************************************************************/
  int CheckMotion() {
    int Status1;
    int Status2;
    int Status3;
    Status1 = CheckMotion(StartTime + 0.001*(EndTime-StartTime));
    Status2 = CheckMotion(StartTime + 0.5*(EndTime-StartTime));
    Status3 = CheckMotion(StartTime + 0.999*(EndTime-StartTime));
    return (Status1 + 10*Status2 + 100*Status3);
  }


  /*************************************************************************
   * Method: operator () (double, int, double[3])
   * Summary: should not be used.
   * Description: over ridden to evaluate positions
   * Return: void
   *************************************************************************/
  virtual void operator () (double, int, double [3]) =0;


  /*************************************************************************
   * Method: operator () (double, int, double[3], double[3])
   * Summary: should not be used
   * Description: over ridden to evaluate positions
   * Return: void
   *************************************************************************/
  virtual void operator () (double , int , double [3], double [3]) =0;


  /*************************************************************************
   * Method: operator () (double, int, double[3], double[3], double[3])
   * Summary: should not be used
   * Description: over ridden to evaluate positions
   * Return: void
   *************************************************************************/
  virtual void operator () (double , int, double [3], double [3],
			    double [3]) =0;


  int GetPositionType(){ return PositionType; }

protected:
  NET_INT PositionType;	// ECI, ECR, or EARTH flag

private:

};

typedef SpDynItem DYNAMIC_ITEM;

#endif
