// SpDynObjs.H

#ifndef DynObj_Objects
#define DynObj_Objects

#include "SpStdarg.H"
#include "SpMath.H"

#include "SpDynItem.H"
#include "SpMatrixEqu.H"
#include "SpNetTypes.H"

/*
** Define some constants if they are not already defined
*/

static const int N_MAX_COEFF=11;
static const double END_POLY=-1.123e20;

#ifndef PI
#define PI 3.14159265358979323846
#endif

#ifndef TWO_PI
#define TWO_PI (2.0*PI)
#endif

#ifndef EXP
#define EXP 2.71828182845904523536
#endif

/************************************************************************
 * Class: DYNAMIC_COMPLEX_EXPONENTIAL
 *
 * Summary: This is a dynamic position type in one variable.  The
 *          specific type is an complex exponential
 *
 * Description: This yeilds an equation of motion in one dimension that
 *              follows the path of a complex exponential.  A complex
 *              exponential is of the form:
 *
 *              y = a*e^(i*Omega(t-t0))
 *
 *              where a is an amplitude, omega is a parameter, t0 is
 *              the initial time, t is the value at which we are
 *              evaluating, and i and e are the standard matheamtical
 *              constants square root(-1) and 2.718281828...
 *
 *              An interesting result of this is that
 *
 *              a*e^(i*Omega(t-t0)) = a(cos(Omega(t-t0)) +isin(Omega(t-t0)))
 *
 *              so, if RealFlag is set through the method
 *              SetRealFlag(), we return the real portion of the
 *              equation a*cos(Omega(t-t0)) and, otherwise, we return
 *              the imaginary portion a*sin(Omega(t-t0).
 *
 *              A mathematical curiosity that arises from these
 *              equations is knows as Euler's formula: e^(i*pi)-1=0.
 *              this equation is significant since it involves the 5
 *              most important numbers in all of mathematics in a
 *              single short equation.
 *
 * Examples: 
 *
************************************************************************/
class DYNAMIC_COMPLEX_EXPONENTIAL : public DYNAMIC_FLOAT_ITEM {
  // X(t) = Amplitude * exp[i*Omega*(t-t0)]
  // If RealFlag is set:     X(t) = Amplitude * cos[Omega*(t-t0)]
  // If RealFlag is not set: X(t) = Amplitude * sin[Omega*(t-t0)]
public:

  /*************************************************************************
   * Method: DYNAMIC_COMPLEX_EXPONENTIAL
   * Summary: Default constructor for a DYNAMIC_COMPLEX_EXPONENTIAL
   * Description: The default constructor sets the RealFlag to 1,
   *              amplitude to 0, the angle parameter to 0, and the
   *              offset (t0) to 0
   * Return: a new DYNAMIC_COMPLEX_EXPONENTIAL
   *************************************************************************/
  DYNAMIC_COMPLEX_EXPONENTIAL() {
    RealFlag = 1;
    Amplitude=0;
    Omega = 0;
    t0 = 0;
  }


  /*************************************************************************
   * Method: Set(double, double, double)
   * Summary: Initializes the amplitude, angle parameter, and offset
   * Description: Initializes the amplitude, angle parameter, and offset
   * Return: void
   *************************************************************************/
  void Set(double a, //In: The new amplitude for this equation
	   double o, //In: The new angle paramter (Omega) for this equation.
	   double t  //In: The new offset (t0) for this equation
	   ) {
    Amplitude = a;
    Omega = o;
    t0 = t;
  }


  /*************************************************************************
   * Method: SetAmplitude(double)
   * Summary: Sets the amplitude to the passed in value
   * Description: Sets the amplitude to the passed in value
   * Return: void
   *************************************************************************/
  void SetAmplitude(double a //In: New value for the amplitude
		    ) {Amplitude = a;}

  /*************************************************************************
   * Method: SetOmega(double)
   * Summary: Sets the value of the angle parameter
   * Description: Sets the value of the angle parameter
   * Return: void
   *************************************************************************/
  void SetOmega(double o //In: New value for Omega
		) {Omega = o;}

  /*************************************************************************
   * Method: SetTimeOffset(double)
   * Summary: Sets the value for the time offset.
   * Description: Sets the value for the time offset.
   * Return: void
   *************************************************************************/
  void SetTimeOffset(double t //In: New value of the time offset.
		     ) {t0 = t;}

  /*************************************************************************
   * Method: SetPeriod(double)
   * Summary: Sets the angle parameter to match the passed in period
   * Description: Sets the angle parameter to match the passed in
   * Return: void
   *************************************************************************/
  void SetPeriod(double t //In: Desired period
		 ) {Omega = TWO_PI/t;}

  /*************************************************************************
   * Method: SetFrequency(double)
   * Summary: Sets the angle parameter to match the passed in frequency
   * Description: 
   * Return: 
   *************************************************************************/
  void SetFrequency(double f//In: Desired frequency
		    ) {Omega = TWO_PI*f;}


  /*************************************************************************
   * Method: GetAmplitude()
   * Summary: returns the amplitude
   * Description: returns the amplitude
   * Return: returns the amplitude
   *************************************************************************/
  double GetAmplitude() {return Amplitude;}

  /*************************************************************************
   * Method: GetOmega()
   * Summary: Returns the value of the angle parameter
   * Description: Returns the value of the angle parameter
   * Return: Returns the value of the angle parameter
   *************************************************************************/
  double GetOmega() {return Omega;}

  /*************************************************************************
   * Method: GetTimeOffset()
   * Summary: Returns the value of the time offset.
   * Description: Returns the value of the time offset.
   * Return: Returns the value of the time offset.
   *************************************************************************/
  double GetTimeOffset() {return t0;}

  /*************************************************************************
   * Method: GetPeriod()
   * Summary: Returns the period of the function
   * Description: Returns the period of the function
   * Return: Returns the period of the function
   *************************************************************************/
  double GetPeriod() {return TWO_PI/Omega;}

  /*************************************************************************
   * Method:GetFrequency() 
   * Summary: Returns the frequency of the function
   * Description: Returns the frequency of the function
   * Return: Returns the frequency of the function
   *************************************************************************/
  double GetFrequency() {return Omega/TWO_PI;}


  /*************************************************************************
   * Method: SetReal()
   * Summary: Sets the real flag to 1
   * Description: When the real flag is set, only the real portion of the
   *              result is returned
   * Return: void
   *************************************************************************/
  void SetReal() {RealFlag = 1;}

  /*************************************************************************
   * Method: SetImag()
   * Summary: Sets the realflag to 0
   * Description: When the real flag is unset, only the imgainary portion
   *              of the result is returned
   * Return: void
   *************************************************************************/
  void SetImag() {RealFlag = 0;}

  /*************************************************************************
   * Method: GetRealFlag()
   * Summary: Returns the state of the real flag
   * Description: Returns the state of the real flag
   * Return: Returns the state of the real flag (1==real, 0==imaginary)
   *************************************************************************/
  int GetRealFlag() {return RealFlag;}


  /*************************************************************************
   * Method: operator()(double)
   * Summary: returns the value of the equation at the passed in time
   * Description: If the real flag is set, this returns a(cos(omega(t-t0)))
   *              and returns a(sin(omega(t-t0))) otherwise.
   * Return: returns the value of the equation at the passed in time
   *************************************************************************/
  virtual double operator () (double t //In: Time at which to evaluate the
			               //    equation.
			      ) {
    double X;
    double dt = t-t0;
    if (RealFlag) {
      X = Amplitude*cos(dt*Omega);
    }else{
      X = Amplitude*sin(dt*Omega);
    }
    return X;
  }


  /*************************************************************************
   * Method: operator ()(double, double &)
   * Summary: returns the value of the equation at the passed in time
   * Description: If the real flag is set, this returns
   *              a(cos(omega(t-t0))) and returns a(sin(omega(t-t0)))
   *              otherwise. The second parameter is then filled with
   *              the value of the first derivative (the velocity) in
   *              the apropriate real or imaginary mode
   * Return: returns the value of the equation at the passed in time
   *************************************************************************/
  virtual double operator () (double t,//In:  Time at which to evaluate the
			               //     equation.
			      double &V//Out: Value of the derivative.
			      ) {
    double X;
    double dt = t-t0;
    if (RealFlag) {
      V = -Omega*Amplitude*sin(dt*Omega);
      X = Amplitude*cos(dt*Omega);
    }else{
      V = Omega*Amplitude*cos(dt*Omega);
      X = Amplitude*sin(dt*Omega);
    }
    return X;
  }


  /*************************************************************************
   * Method: operator ()(double, double &, double &)
   * Summary: returns the value of the equation at the passed in time
   * Description: If the real flag is set, this returns
   *              a(cos(omega(t-t0))) and returns a(sin(omega(t-t0)))
   *              otherwise. The second and third parameters are then
   *              filled with the value of the first dand second
   *              derivative (the velocity and acceleration) in the
   *              apropriate real or imaginary mode
   * Return: returns the value of the equation at the passed in time
   *************************************************************************/
  virtual double operator () (double t, //In:  Time at which to evaluate the
			                //     equation.
			      double &V,//Out: Value of the derivative
			      double &A //Out: Value of the second derivative
			      ) {
    double X;
    double dt = t-t0;
    if (RealFlag) {
      V = -Omega*Amplitude*sin(dt*Omega);
      X = Amplitude*cos(dt*Omega);
    }else{
      V = Omega*Amplitude*cos(dt*Omega);
      X = Amplitude*sin(dt*Omega);
    }
    A = -Omega*Omega*X;
    return X;
  }


  /*************************************************************************
   * Method: GetType()
   * Summary: Returns the type of this SpDynItem
   * Description: Returns the type of this SpDynItem
   * Return: Returns the type of this SpDynItem
   *************************************************************************/
  virtual char *GetType() {return "Dynamic complex exponential";}
protected:

private:
  NET_INT RealFlag;	// Defines value as real or imaginary part
  NET_DOUBLE Omega;	// two pi times frequency
  NET_DOUBLE Amplitude;   // real (not complex) amplitude
  NET_DOUBLE t0;		// time offset

};

/************************************************************************
 * Class: DYNAMIC_EXPONENTIAL
 *
 * Summary: A univariate exponential equation of motion.
 *
 * Description: This yields an equation of motion in one dimension
 *              that follows the path of an exponential.  The exact
 *              equation that is uses is y=a*e^(L*(t-StartTime)) where
 *              a, L, and StartTime are constants and e is the number
 *              2.718281828....  a is the amplitude, L is just a
 *              constant multiplier of the variable (it can be thought
 *              of as a horizontal stretch/shrinking of the graph),
 *              and t0 is a phase shift (horizontal shift) of the
 *              function
 *
 * Examples: 
 *
************************************************************************/
class DYNAMIC_EXPONENTIAL : public DYNAMIC_FLOAT_ITEM {

public:

  /*************************************************************************
   * Method: DYNAMIC_EXPONENTIAL()
   * Summary: Default constructor of a DYNAMIC_EXPONENTIAL
   * Description: The default constructor sets the amplitude to 0 and
   *              the multiplier to 1.  The amplitude being 0
   *              essentially means that this is now the zero function
   * Return: returns a new DYNAMIC_EXPONENTIAL equation
   *************************************************************************/
  DYNAMIC_EXPONENTIAL() {Amplitude=0; Lambda = 1.0;}


  //geb2:  new method
  /*************************************************************************
   * Method: Set(double, double)
   * Summary: Initializes the amplitude and Lambda
   * Description: Initializes the amplitude and Lambda
   * Return: void
   *************************************************************************/
  void Set(double amp,    //In: The new amplitude for this equation
	   double lambda  //In: The new lambda param
	   ) {
    Amplitude = amp;
    Lambda = lambda;
  }


  /*************************************************************************
   * Method: SetLambda(double)
   * Summary: Sets the value of the constant multiplier
   * Description: Sets the value of the constant multiplier.  Values
   *              >1 shrink the graph horizontally, values between 0
   *              and 1 expand the graph horizontally.  Values less
   *              than 0 have the same effect as their positive
   *              counterparts with the additional affect of flipping
   *              the graph about the y axis
   * Return: Sets the value of the constant multiplier
   *************************************************************************/
  void SetLambda(double l //In: Desired value for the constant multiplier.
		 ) {Lambda = l;}

  /*************************************************************************
   * Method: GetLambda()
   * Summary: Returns the current value of Lambda (the constant multiplier)
   * Description: Returns the current value of Lambda (the constant multiplier)
   * Return: Returns the current value of Lambda (the constant multiplier)
   *************************************************************************/
  double GetLambda() {return Lambda;}

  /*************************************************************************
   * Method: SetTimeConstant()
   * Summary: Sets lambda so that the value of the equation at the
   *          passed in time is the amplitude
   * Description: If you want to make sure that the graph passes
   *              through the point (t, Amplitude), pass in the
   *              desired value of t here.
   * Return: void
   *************************************************************************/
  void SetTimeConstant(double t //In: Desired time for which y=Amplitude
		       ) {Lambda = 1.0/t;}

  /*************************************************************************
   * Method: GetTimeConstant()
   * Summary: Returns the value of t for which the equation returns Amplitude
   * Description: If you want to find out when the graph crosses the
   *              horizonal line y=Amplitude, call this method.
   * Return: Returns the value of t for which the equation returns Amplitude
   *************************************************************************/
  double GetTimeConstant() {return 1.0/Lambda;}

  /*************************************************************************
   * Method: SetAmplitude(double)
   * Summary: Sets the amplitude to the passed in value
   * Description: Sets the amplitude to the passed in value
   * Return: Sets the amplitude to the passed in value
   *************************************************************************/
  void SetAmplitude(double a //In: desired value for the amplitude
		    ) {Amplitude = a;}

  /*************************************************************************
   * Method: GetAmplitude()
   * Summary: Returns the current value for the amplitude
   * Description: Returns the current value for the amplitude
   * Return: Returns the current value for the amplitude
   *************************************************************************/
  double GetAmplitude() {return Amplitude;}

  /*************************************************************************
   * Method: operator () (double)
   * Summary: Returns the value of the equation at the given time
   * Description: Returns the value of the equation at the given time
   * Return: Returns the value of the equation at the given time
   *************************************************************************/
  virtual double operator () (double t //In: Time at which the value of the
			               //    function is desired.
			      ) {
    double dt = t-StartTime;
    double X = Amplitude*exp(dt*Lambda);
    return X;
  }

  /*************************************************************************
   * Method: operator () (double, double &)
   * Summary: Returns the value of the equation at the given time
   * Description: In addition to the value of the function, the second
   *              parameter is filled in with the value of the first
   *              derivative (the velocity) of the function.
   * Return: Returns the value of the equation at the given time
   *************************************************************************/
  virtual double operator () (double t,//In:  Time at which the value of the
			               //     function is desired.
			      double &V//Out: Value of the derivative
			      ) {
    double dt = t-StartTime;
    double X = Amplitude*exp(dt*Lambda);
    V = Lambda*X;
    return X;
  }

  /*************************************************************************
   * Method: operator () (double, double &, double &)
   * Summary: Returns the value of the equation at the given time
   * Description: In addition to the value of the function, the second
   *              parameter is filled in with the value of the first
   *              derivative (the velocity) of the function and the
   *              third parameter is filled in with the second
   *              derivative (the acceleration) of the function
   * Return: Returns the value of the equation at the given time
   *************************************************************************/
  virtual double operator () (double t, //In: Time at which the value of the
			                //    function is desired.
			      double &V,//Out: Value of the derivative
			      double &A //Out: Value of the second derivative
			      ) {
    double dt = t-StartTime;
    double X = Amplitude*exp(dt*Lambda);
    V = Lambda*X;
    A = Lambda*V;
    return X;
  }

  /*************************************************************************
   * Method: GetType()
   * Summary: Returns a char * string of the type of this item
   * Description: Returns a char * string of the type of this item
   * Return: Returns a char * string of the type of this item
   *************************************************************************/
  virtual char *GetType() {return "Dynamic exponential";}

protected:

private:
  NET_DOUBLE Lambda;      // Note, Lambda = 1/TimeConstant
  NET_DOUBLE Amplitude;   // value at StartTime
};

/************************************************************************
 * Class: DYNAMIC_FLOAT_CONSTANT
 *
 * Summary: A dynamic float value
 *
 * Description: This could be very useful for functions that are
 *              constant over periods of time (such as a step
 *              function).  This could be imitated using other
 *              functions but the best way is just to use a
 *              DYNAMIC_FLOAT_CONSTANT.
 *
 * Examples: 
 *
************************************************************************/
class DYNAMIC_FLOAT_CONSTANT : public DYNAMIC_FLOAT_ITEM {
public:

  /*************************************************************************
   * Method: DYNAMIC_FLOAT_CONSTANT()
   * Summary: Default constructor for a DYNAMIC_FLOAT_CONSTANT
   * Description: Default constructor sets the value of the constant to 0
   * Return: a new DYNAMIC_FLOAT_CONSTANT
   *************************************************************************/
  DYNAMIC_FLOAT_CONSTANT() {a=0;}

  /*************************************************************************
   * Method: Set(double)
   * Summary: Sets the value of the constant to the passed in value
   * Description: Sets the value of the constant to the passed in value
   * Return: void
   *************************************************************************/
  void Set(double v //In: Desired value for the constant
	   ) {a = v;}

  /*************************************************************************
   * Method: operator () (double)
   * Summary: Returns the value of the constant
   * Description: Returns the value of the constant
   * Return: Returns the value of the constant
   *************************************************************************/
  virtual double operator () (double) {return a;}

  /*************************************************************************
   * Method: operator () (double, double&)
   * Summary:Returns the value of the constant 
   * Description: The first derivative is 0 so the second parameter is
   *              set to 0
   * Return: Returns the value of the constant
   *************************************************************************/
  virtual double operator () (double,
			      double &V //Out: Set to 0.
			      ) {V = 0.0; return a;}

  /*************************************************************************
   * Method: operator () (double, double&, double&)
   * Summary: Returns the value of the constant
   * Description: The first and second derivatives are 0 so the second
   *              and third parameters are set to 0
   * Return: Returns the value of the constant
   *************************************************************************/
  virtual double operator () (double,
			      double &V, //Out: Set to 0.
			      double &A  //Out: Set to 0.
			      ) {
    V = 0.0;
    A = 0.0;
    return a;
  }
  /*************************************************************************
   * Method: GetType()
   * Summary: Returns a char * string of the type of this item
   * Description: Returns a char * string of the type of this item
   * Return: Returns a char * string of the type of this item
   *************************************************************************/
  virtual char *GetType() {return "Dynamic float constant";}

protected:

private:
  NET_DOUBLE a;
};

/************************************************************************
 * Class: DYNAMIC_POLY_N
 *
 * Summary: An n-degree one dimensional polynomial that best fits the
 *          given points
 *

 * Description: If you provide more than n points, this class will
 *              construct the best fit polynomial through those
 *              points.  You are allowed to specify sigma-squared (the
 *              standard deviation squared) for each of the points to
 *              specify how close you want the fit to match those
 *              points.
 *
 *              While every possible precaution is taken, you should
 *              be warned that the system of equations that must be
 *              solved can be notoriously ill conditioned for larger
 *              values of n (greater than 8 or 9).  Every precaution
 *              is taken to try and solve this system as best as is
 *              possible but it still can give incorrect values.
 *
 *              Also, high degree polynomials often have wild
 *              occilations, even when your data appears to be smooth.
 *              For this reason, rather than using a high degree
 *              (greater than 6) polynomial, it maybe wiser to use a
 *              spline.
 *
 * Examples: 
 *
************************************************************************/
class DYNAMIC_POLY_N : public DYNAMIC_FLOAT_ITEM {
public:


  /*************************************************************************
   * Method: DYNAMIC_POLY_N()
   * Summary: Default constructor for an n-dimensional polynomial
   * Description: Default constructor for an n-dimensional polynomial.
   *              The work array A and the work vector B are set to
   *              NULL.  These work areas are used to find the best
   *              fit polynomial to the given points.
   * Return: a new DYNAMIC_POLY_N
   *************************************************************************/
  DYNAMIC_POLY_N() : A(NULL), B(NULL), NumPointsAdded(0) {}


  /*************************************************************************
   * Method: AddPoints()
   * Summary: Add points for which you wish to fit the equation.

   * Description: AddPoints can take any number of arguments that is 2
   *              plus a multiple of 3.  For example, if you wanted
   *              the line that went through the points (0, 1) and (2,
   *              7) from t=0 through 3 you would first use the child
   *              class DYNAMIC_POLY_1 and then would call
   *
   *              AddPoints(0, 3, 0, 1, 17, 2, 7, 17, END_POLY);
   *
   *              Notice the extra numbers '17' that I have in the
   *              equation.  those are supposed to be the standard
   *              deviation (squared) allowed for each point in the
   *              line fit.  Since we are giving the exact number of
   *              points required to fit a line, these values are
   *              ignored.
   *
   *              Suppose I wanted to find the best fit quadratic to
   *              the points (1, 2), (3, 7) (4, 12), and (8, 6) from
   *              t=0 to t=13. I also want the line to be a better fit
   *              for the first, third and, fourth numbers than for
   *              the second.  In this case, I would use the child
   *              class DYNAMIC_POLY_2 and would call the method
   *
   *              AddPoints(0,  13,
   *                        1,  2, .1,
   *                        3,  7, 10,
   *                        4, 12, .1,
   *                        8,  6, .1,
   *                        END_POLY)
   *
   *              Since the sigma squared is smaller in the first,
   *              third, and fourth points, the line will fit those
   *              points better while it will be a worse fit for the
   *              scond point.
   *
   *              Three important points of note.  First, you must
   *              always terminate the argument list with END_POLY.
   *              This is the signal to the computer that you have
   *              entered your last point.  Second, sigma squared
   *              (err2) must never be 0 or negative.  0 will result
   *              in either NaN's or a core dump and negative numbers
   *              can result in garbage results.
   *
   * Return: void
   *************************************************************************/
  void AddPoints(double ts,  //In: Start time for the polynomial
		 double te,  //In: End time for the polynomial
		 double t,   //In: Time value
		 double v,   //In: Position at that time.
		 double err2,//In: sigma squared for that position
		 ...         //In: Further (t,v,err2) triplets
		 );


  /*************************************************************************
   * Method: Reset()
   * Summary: Frees the work areas and removes all points that were added.
   * Description: Frees the work areas and removes all points that
   *              were added.  This does not undo the polynomial that
   *              you have created but allows you to recreate it with
   *              new points.
   * Return: void
   *************************************************************************/
  void Reset() {
    int i;
    int n = GetDim();
    if (A != NULL) {
      for (i=0; i<n; i++) delete [] A[i];
      delete []A;
      A = NULL;
    }
    if (B != NULL) {
      delete []B;
      B = NULL;
    }
    NumPointsAdded=0;
  }


  /*************************************************************************
   * Method: MakePoly()
   * Summary: Generates the polynomial for the points you entered.
   * Description: Generates the polynomial for the points you
   *              entered. If too few points have been added to
   *              generate the polynomial you desire, an error message
   *              is printed.
   * Return: void
   *************************************************************************/
  //geb3:
#if HETEROGENEOUS_DATA_TYPE

  void MakePoly() {
    //int status;
    int n=GetDim();
    if (NumPointsAdded >= n){
      BIG_FLOAT *coeffs = new BIG_FLOAT[n];

      matrix_equation(A, B, coeffs, n);

      // copy BIG_FLOATS back to NET_DOUBLES:
      NET_DOUBLE *ndCoeffs = GetCoeff();
      for(int i=0; i<n; i++) {
	ndCoeffs[i] = coeffs[i];
      }
      delete [] coeffs;
    }
    else{
      cout<<"You tried to generate a "<<n<<"th degree polynmial with too few"
	  <<" points:"<<NumPointsAdded<<endl;
    }
  }

  //geb3
#else
  void MakePoly() {
    //int status;
    int n=GetDim();
    if (NumPointsAdded >= n){
      /*status = */matrix_equation(A, B, GetCoeff(), n); //rvt
    }
    else{
      cout<<"You tried to generate a "<<n<<"th degree polynmial with too few"
	  <<" points:"<<NumPointsAdded<<endl;
    }
  }

#endif

  /*************************************************************************
   * Method: Evaluate(BIG_FLOAT*, double, int)
   * Summary: Evaluates at t, the nth degree polynomial whose
   *          coefficients are given by c

   * Description: A polynomial is uniquely determined by it's
   *              coefficients.  these coefficients are given in the
   *              array c (in decreasing order of exponent) and the
   *              degree is passed in the variable n.  The polynomial
   *              is computed for the interval [0, EndTime-StartTime]
   *              so we need to offset where we evaluate the
   *              polynomial to the t-StartTime. We then use Horner's
   *              rule to evaluate the polynomial efficiently.
   * Return: The value of the polynomial, evaluated at t.
   *************************************************************************/
  double Evaluate(NET_DOUBLE *c, //In: Coefficients of polynomial in decreasing
		                //    order of exponent
		  double t,     //In: Point at which to evaluate the polynmial
		  int n         //In: degree of the polynomial
		  ) {
    BIG_FLOAT value = 0;
    BIG_FLOAT ti = 1;
    for (int i=0; i<n; i++) {
      value += c[i]*ti;
      ti *= t-StartTime;
    }
    return double(value);
  }


  /*************************************************************************
   * Method: Evaluate(NET_DOUBLE*, double, int, double &)
   * Summary: Evaluates at t, the nth degree polynomial whose
   *          coefficients are given by c
   * Description: A polynomial is uniquely determined by it's
   *              coefficients.  these coefficients are given in the
   *              array c (in decreasing order of exponent) and the
   *              degree is passed in the variable n.  The polynomial
   *              is computed for the interval [0, EndTime-StartTime]
   *              so we need to offset where we evaluate the
   *              polynomial to the t-StartTime. We then use Horner's
   *              rule to evaluate the polynomial efficiently.  We
   *              also evaluate the first derivative of this
   *              polynomial and return that in the parameter V.  The
   *              derivative is also evaluated using a variant of
   *              Horner's rule.
   * Return: 
   *************************************************************************/
  double Evaluate(NET_DOUBLE *c, //In:  Coefficients of polynomial in decreasing
		                //     order of exponent
		  double t,     //In:  Point at which to evaluate the polynmial
		  int n,        //In:  degree of the polynomial
		  double &V     //Out: Value of the first derivative.
		  ) {
    BIG_FLOAT value = 0;
    BIG_FLOAT V1 = 0;
    BIG_FLOAT ti = 1;
    for (int i=0; i<n; i++) {
      value += c[i]*ti;
      if (i<n-1) V1 += (i+1)*c[i+1]*ti;
      ti *= t-StartTime;
    }
    V = V1;
    return double(value);
  }


  /*************************************************************************
   * Method: Evaluate(NET_DOUBLE*, double, int, double &, double &)
   * Summary: Evaluates at t, the nth degree polynomial whose
   *          coefficients are given by c
   * Description: A polynomial is uniquely determined by it's
   *              coefficients.  these coefficients are given in the
   *              array c (in decreasing order of exponent) and the
   *              degree is passed in the variable n.  The polynomial
   *              is computed for the interval [0, EndTime-StartTime]
   *              so we need to offset where we evaluate the
   *              polynomial to the t-StartTime. We then use Horner's
   *              rule to evaluate the polynomial efficiently.  We
   *              also evaluate the first and second derivatives of
   *              this polynomial and return that in the parameters V
   *              and A_out.  The first and second second derivatives
   *              also evaluated using a variant of Horner's rule.
   * Return: value of the polynomial at t
   *************************************************************************/
  double Evaluate(NET_DOUBLE *c, //In:  Coefficients of polynomial in decreasing
		                //     order of exponent
		  double t,     //In:  Point at which to evaluate the polynmi
		  int n,        //In:  degree of the polynomial
		  double &V,    //Out: Value of the first derivative.
		  double &A_out //Out: Value of the second derivative.
		  ) {
    BIG_FLOAT value = 0;
    BIG_FLOAT V1 = 0;
    BIG_FLOAT A1 = 0;
    BIG_FLOAT ti = 1;
    for (int i=0; i<n; i++) {
      value += c[i]*ti;
      if (i<n-1) V1 += (i+1)*c[i+1]*ti;
      if (i<n-2) A1 += (i+2)*(i+1)*c[i+2]*ti;
      ti *= t-StartTime;
    }
    V = V1;
    A_out = A1;
    return double(value);
  }


  /*************************************************************************
   * Method: operator (double)
   * Summary: Evaluates the polynomial at t where coefficients are given by c
   * Description: A polynomial is uniquely determined by it's
   *              coefficients.  these coefficients are given in the
   *              array c (in decreasing order of exponent) and the
   *              degree is passed in the variable n.  The polynomial
   *              is computed for the interval [0, EndTime-StartTime]
   *              so we need to offset where we evaluate the
   *              polynomial to the t-StartTime. We then use Horner's
   *              rule to evaluate the polynomial efficiently.
   * Return: value of the polynomial at t
   *************************************************************************/
  virtual double operator () (double t) {
    return Evaluate(GetCoeff(), t, GetDim()); //rvt
  }


  /*************************************************************************
   * Method: operator (double, double&)
   * Summary: Evaluates the polynomial at t where coefficients are given by c
   * Description: A polynomial is uniquely determined by it's
   *              coefficients.  these coefficients are given in the
   *              array c (in decreasing order of exponent) and the
   *              degree is passed in the variable n.  The polynomial
   *              is computed for the interval [0, EndTime-StartTime]
   *              so we need to offset where we evaluate the
   *              polynomial to the t-StartTime. We then use Horner's
   *              rule to evaluate the polynomial efficiently.  We
   *              also evaluate the first derivative of this
   *              polynomial and return that in the parameter V.  The
   *              derivative is also evaluated using a variant of
   *              Horner's rule.
   * Return: value of the polynomial at t
   *************************************************************************/
  virtual double operator () (double t, double &V) {
    return Evaluate(GetCoeff(), t, GetDim(), V); //rvt
  }


  /*************************************************************************
   * Method: operator (double, double&, double &)
   * Summary: Evaluates the polynomial at t where coefficients are given by c
   * Description: A polynomial is uniquely determined by it's
   *              coefficients.  these coefficients are given in the
   *              array c (in decreasing order of exponent) and the
   *              degree is passed in the variable n.  The polynomial
   *              is computed for the interval [0, EndTime-StartTime]
   *              so we need to offset where we evaluate the
   *              polynomial to the t-StartTime. We then use Horner's
   *              rule to evaluate the polynomial efficiently.  We
   *              also evaluate the first and second derivatives of
   *              this polynomial and return that in the parameters V
   *              and A_out.  The first and second second derivatives
   *              also evaluated using a variant of Horner's rule.
   * Return: value of the polynomial at t
   *************************************************************************/
  virtual double operator () (double t, double &V, double &A_output_arg) {
    return Evaluate(GetCoeff(), t, GetDim(), V, A_output_arg);
  }


  /*************************************************************************
   * Method: GetDim()
   * Summary: Returns the degree of the polynomial
   * Description: Returns the degree of the polynomial
   * Return: Returns the degree of the polynomial
   *************************************************************************/
  virtual int GetDim() {return 0;}

  /*************************************************************************
   * Method: GetCoeff()
   * Summary: Returns the coefficients of the polynomial
   * Description: Returns the coefficients of the polynomial
   * Return: Returns the coefficients of the polynomial
   *************************************************************************/
  virtual NET_DOUBLE *GetCoeff() {return NULL;}

  /*************************************************************************
   * Method: GetType()
   * Summary: Returns a char * string of the type of this item
   * Description: Returns a char * string of the type of this item
   * Return: Returns a char * string of the type of this item
   *************************************************************************/
  virtual char *GetType() {return "Dynamic poly N";}

protected:
  BIG_FLOAT **A; //rvt
  BIG_FLOAT *B; //rvt
  int NumPointsAdded; //rvt

private:
};

/************************************************************************
 * Class: DYNAMIC_POLY_1
 *
 * Summary: A dynamic motion type that is a line
 *
 * Description: This give a dynamic motion type that consists of
 *              simply a straight line.  You just use the AddPoints()
 *              method to add at least two points and then a best fit
 *              to the given data will be computed for you.
 *
 * Examples: 
 *
************************************************************************/
class DYNAMIC_POLY_1 : public DYNAMIC_POLY_N {
public:

  /*************************************************************************
   * Method: GetDim()
   * Summary: Returns the degree of the polynomial
   * Description: Returns the degree of the polynomial
   * Return: Returns the degree of the polynomial
   *************************************************************************/
  virtual int GetDim() {return 2;}

  /*************************************************************************
   * Method: GetCoeff()
   * Summary: Returns the coefficients of the polynomial
   * Description: Returns the coefficients of the polynomial
   * Return: Returns the coefficients of the polynomial
   *************************************************************************/
  virtual NET_DOUBLE *GetCoeff() {return c;}

  /*************************************************************************
   * Method: GetType()
   * Summary: Returns a char * string of the type of this item
   * Description: Returns a char * string of the type of this item
   * Return: Returns a char * string of the type of this item
   *************************************************************************/
  virtual char *GetType() {return "Dynamic poly 1";}
protected:
private:
  NET_DOUBLE c[2];
};

/************************************************************************
 * Class: DYNAMIC_POLY_2
 *
 * Summary: Dynamic motion of type quadratic polynomial
 *
 * Description: This give a dynamic motion type that consists of
 *              simply a quadratic (a parabola).  You just use the
 *              AddPoints() method to add at least three points and
 *              then a best fit to the given data will be computed for
 *              you.
 *
 * Examples: 
 *
************************************************************************/
class DYNAMIC_POLY_2 : public DYNAMIC_POLY_N {
public:

  /*************************************************************************
   * Method: GetDim()
   * Summary: Returns the degree of the polynomial
   * Description: Returns the degree of the polynomial
   * Return: Returns the degree of the polynomial
   *************************************************************************/
  virtual int GetDim() {return 3;}

  /*************************************************************************
   * Method: GetCoeff()
   * Summary: Returns the coefficients of the polynomial
   * Description: Returns the coefficients of the polynomial
   * Return: Returns the coefficients of the polynomial
   *************************************************************************/
  virtual NET_DOUBLE *GetCoeff() {return c;}

  /*************************************************************************
   * Method: GetType()
   * Summary: Returns a char * string of the type of this item
   * Description: Returns a char * string of the type of this item
   * Return: Returns a char * string of the type of this item
   *************************************************************************/
  virtual char *GetType() {return "Dynamic poly 2";}
protected:
private:
  NET_DOUBLE c[3];
};

/************************************************************************
 * Class: DYNAMIC_POLY_3
 *
 * Summary: Dynamic motion of type cubic polynomial
 *
 * Description: This give a dynamic motion type that consists of
 *              simply a cubic polynomial.  You just use the
 *              AddPoints() method to add at least four points and
 *              then a best fit to the given data will be computed for
 *              you.
 *
 * Examples: 
 *
************************************************************************/
class DYNAMIC_POLY_3 : public DYNAMIC_POLY_N {
public:

  /*************************************************************************
   * Method: GetDim()
   * Summary: Returns the degree of the polynomial
   * Description: Returns the degree of the polynomial
   * Return: Returns the degree of the polynomial
   *************************************************************************/
  virtual int GetDim() {return 4;}

  /*************************************************************************
   * Method: GetCoeff()
   * Summary: Returns the coefficients of the polynomial
   * Description: Returns the coefficients of the polynomial
   * Return: Returns the coefficients of the polynomial
   *************************************************************************/
  virtual NET_DOUBLE *GetCoeff() {return c;}

  /*************************************************************************
   * Method: GetType()
   * Summary: Returns a char * string of the type of this item
   * Description: Returns a char * string of the type of this item
   * Return: Returns a char * string of the type of this item
   *************************************************************************/
  virtual char *GetType() {return "Dynamic poly 3";}
protected:
private:
  NET_DOUBLE c[4];
};

/************************************************************************
 * Class: DYNAMIC_POLY_4
 *
 * Summary: A dynamic motion type that is a quadratic polynomial
 *
 * Description: This give a dynamic motion type that consists of
 *              simply a quadratic polynomial.  You just use the
 *              AddPoints() method to add at least five points and
 *              then a best fit to the given data will be computed for
 *              you.
 *
 * Examples: 
 *
************************************************************************/
class DYNAMIC_POLY_4 : public DYNAMIC_POLY_N {
public:

  /*************************************************************************
   * Method: GetDim()
   * Summary: Returns the degree of the polynomial
   * Description: Returns the degree of the polynomial
   * Return: Returns the degree of the polynomial
   *************************************************************************/
  virtual int GetDim() {return 5;}

  /*************************************************************************
   * Method: GetCoeff()
   * Summary: Returns the coefficients of the polynomial
   * Description: Returns the coefficients of the polynomial
   * Return: Returns the coefficients of the polynomial
   *************************************************************************/
  virtual NET_DOUBLE *GetCoeff() {return c;}

  /*************************************************************************
   * Method: GetType()
   * Summary: Returns a char * string of the type of this item
   * Description: Returns a char * string of the type of this item
   * Return: Returns a char * string of the type of this item
   *************************************************************************/
  virtual char *GetType() {return "Dynamic poly 4";}
protected:
private:
  NET_DOUBLE c[5];
};

/************************************************************************
 * Class: DYNAMIC_POLY_5
 *
 * Summary: A dynamic motion type that is a 5th order polynomial
 *
 * Description: This give a dynamic motion type that consists of
 *              simply a fifth order polynomial.  You just use the
 *              AddPoints() method to add at least six points and
 *              then a best fit to the given data will be computed for
 *              you.
 *
 * Examples: 
 *
************************************************************************/
class DYNAMIC_POLY_5 : public DYNAMIC_POLY_N {
public:

  /*************************************************************************
   * Method: GetDim()
   * Summary: Returns the degree of the polynomial
   * Description: Returns the degree of the polynomial
   * Return: Returns the degree of the polynomial
   *************************************************************************/
  virtual int GetDim() {return 6;}

  /*************************************************************************
   * Method: GetCoeff()
   * Summary: Returns the coefficients of the polynomial
   * Description: Returns the coefficients of the polynomial
   * Return: Returns the coefficients of the polynomial
   *************************************************************************/
  virtual NET_DOUBLE *GetCoeff() {return c;}

  /*************************************************************************
   * Method: GetType()
   * Summary: Returns a char * string of the type of this item
   * Description: Returns a char * string of the type of this item
   * Return: Returns a char * string of the type of this item
   *************************************************************************/
  virtual char *GetType() {return "Dynamic poly 5";} //geb3 (this was private!?)
protected:
private:
  NET_DOUBLE c[6];
};

/************************************************************************
 * Class: DYNAMIC_POLY_6
 *
 * Summary:A dynamic motion type that is a 6th order polynomial 
 *
 * Description: This give a dynamic motion type that consists of
 *              simply a sixth degree polynomial.  You just use the
 *              AddPoints() method to add at least seven points and then
 *              a best fit to the given data will be computed for you.
 *
 * Examples: 
 *
************************************************************************/
class DYNAMIC_POLY_6 : public DYNAMIC_POLY_N {
public:

  /*************************************************************************
   * Method: GetDim()
   * Summary: Returns the degree of the polynomial
   * Description: Returns the degree of the polynomial
   * Return: Returns the degree of the polynomial
   *************************************************************************/
  virtual int GetDim() {return 7;}

  /*************************************************************************
   * Method: GetCoeff()
   * Summary: Returns the coefficients of the polynomial
   * Description: Returns the coefficients of the polynomial
   * Return: Returns the coefficients of the polynomial
   *************************************************************************/
  virtual NET_DOUBLE *GetCoeff() {return c;}

  /*************************************************************************
   * Method: GetType()
   * Summary: Returns a char * string of the type of this item
   * Description: Returns a char * string of the type of this item
   * Return: Returns a char * string of the type of this item
   *************************************************************************/
  virtual char *GetType() {return "Dynamic poly 6";}
protected:
private:
  NET_DOUBLE c[7];
};

/************************************************************************
 * Class: DYNAMIC_POLY_7
 *
 * Summary: Dynamic motion of type seventh order polynomial
 *
 * Description: This give a dynamic motion type that consists of
 *              simply a seventh order polynomial.  You just use the
 *              AddPoints() method to add at least eight points and
 *              then a best fit to the given data will be computed for
 *              you.
 *
 * Examples: 
 *
************************************************************************/
class DYNAMIC_POLY_7 : public DYNAMIC_POLY_N {
public:

  /*************************************************************************
   * Method: GetDim()
   * Summary: Returns the degree of the polynomial
   * Description: Returns the degree of the polynomial
   * Return: Returns the degree of the polynomial
   *************************************************************************/
  virtual int GetDim() {return 8;}

  /*************************************************************************
   * Method: GetCoeff()
   * Summary: Returns the coefficients of the polynomial
   * Description: Returns the coefficients of the polynomial
   * Return: Returns the coefficients of the polynomial
   *************************************************************************/
  virtual NET_DOUBLE *GetCoeff() {return c;}

  /*************************************************************************
   * Method: GetType()
   * Summary: Returns a char * string of the type of this item
   * Description: Returns a char * string of the type of this item
   * Return: Returns a char * string of the type of this item
   *************************************************************************/
  virtual char *GetType() {return "Dynamic poly 7";}
protected:
private:
  NET_DOUBLE c[8];
};

/************************************************************************
 * Class: DYNAMIC_POLY_8
 *
 * Summary: Dynamic motion of type eighth degree polynomial
 *
 * Description: This give a dynamic motion type that consists of
 *              simply a eighth degree polynomial.  You just use the
 *              AddPoints() method to add at least nine points and
 *              then a best fit to the given data will be computed for
 *              you.
 *
 * Examples: 
 *
************************************************************************/
class DYNAMIC_POLY_8 : public DYNAMIC_POLY_N {
public:

  /*************************************************************************
   * Method: GetDim()
   * Summary: Returns the degree of the polynomial
   * Description: Returns the degree of the polynomial
   * Return: Returns the degree of the polynomial
   *************************************************************************/
  virtual int GetDim() {return 9;}

  /*************************************************************************
   * Method: GetCoeff()
   * Summary: Returns the coefficients of the polynomial
   * Description: Returns the coefficients of the polynomial
   * Return: Returns the coefficients of the polynomial
   *************************************************************************/
  virtual NET_DOUBLE *GetCoeff() {return c;}

  /*************************************************************************
   * Method: GetType()
   * Summary: Returns a char * string of the type of this item
   * Description: Returns a char * string of the type of this item
   * Return: Returns a char * string of the type of this item
   *************************************************************************/
  virtual char *GetType() {return "Dynamic poly 8";}
protected:
private:
  NET_DOUBLE c[9];
};

/************************************************************************
 * Class: DYNAMIC_POLY_9
 *
 * Summary: Dynamic motion of type ninth degree polynomial
 *
 * Description: This give a dynamic motion type that consists of
 *              simply a ninth degree polynomial.  You just use the
 *              AddPoints() method to add at least ten points and then
 *              a best fit to the given data will be computed for you.
 *
 * Examples: 
 *
************************************************************************/
class DYNAMIC_POLY_9 : public DYNAMIC_POLY_N {
public:

  /*************************************************************************
   * Method: GetDim()
   * Summary: Returns the degree of the polynomial
   * Description: Returns the degree of the polynomial
   * Return: Returns the degree of the polynomial
   *************************************************************************/
  virtual int GetDim() {return 10;}

  /*************************************************************************
   * Method: GetCoeff()
   * Summary: Returns the coefficients of the polynomial
   * Description: Returns the coefficients of the polynomial
   * Return: Returns the coefficients of the polynomial
   *************************************************************************/
  virtual NET_DOUBLE *GetCoeff() {return c;}

  /*************************************************************************
   * Method: GetType()
   * Summary: Returns a char * string of the type of this item
   * Description: Returns a char * string of the type of this item
   * Return: Returns a char * string of the type of this item
   *************************************************************************/
  virtual char *GetType() {return "Dynamic poly 9";}
protected:
private:
  NET_DOUBLE c[10];
};

/************************************************************************
 * Class: DYNAMIC_POLY_10
 *
 * Summary:Dynamic motion of type tenth polynomial 
 *
 * Description: This give a dynamic motion type that consists of
 *              simply a tenth degree polynomial.  You just use the
 *              AddPoints() method to add at least eleven points and
 *              then a best fit to the given data will be computed for
 *              you.
 *
 * Examples: 
 *
************************************************************************/
class DYNAMIC_POLY_10 : public DYNAMIC_POLY_N {
public:

  /*************************************************************************
   * Method: GetDim()
   * Summary: Returns the degree of the polynomial
   * Description: Returns the degree of the polynomial
   * Return: Returns the degree of the polynomial
   *************************************************************************/
  virtual int GetDim() {return 11;}

  /*************************************************************************
   * Method: GetCoeff()
   * Summary: Returns the coefficients of the polynomial
   * Description: Returns the coefficients of the polynomial
   * Return: Returns the coefficients of the polynomial
   *************************************************************************/
  virtual NET_DOUBLE *GetCoeff() {return c;}

  /*************************************************************************
   * Method: GetType()
   * Summary: Returns a char * string of the type of this item
   * Description: Returns a char * string of the type of this item
   * Return: Returns a char * string of the type of this item
   *************************************************************************/
  virtual char *GetType() {return "Dynamic poly 10";}
protected:
private:
  NET_DOUBLE c[11];
};


/************************************************************************
 * Class: DYNAMIC_SPLINE_3
 *
 * Summary: Creates a cubic polynomial with specified position and
 *          velocity over the given interval.
 *
 * Description: Despite the name, this does not create what is
 *              normally thought of as a cubic spline.  You construct
 *              one of this spline pieces by specifying the position
 *              and velocity at both the start time and the end time.
 *              We then compute the cubic polynomial (also known as a
 *              Hermite interpolant) for the given information.
 *
 * Examples: 
 *
************************************************************************/
class DYNAMIC_SPLINE_3 : public DYNAMIC_FLOAT_ITEM {
public:

  /*************************************************************************
   * Method: init(double, double, double, double, double, double)
   * Summary: Initializes (creates) the cubic interpolant
   * Description: when you provide the values for position and
   *              velocity at the initial point and at the end point,
   *              this will find the unique cubic polynomial (a
   *              trivial case of Hermite interpolation) that
   *              satisfies those conditions.
   * Return: void
   ************************************************************************/
  void init(double t0, //In: Start time
	    double x0, //In: Initial position
	    double v0, //In: Initial velocity
	    double t1, //In: end time
	    double x1, //In: End position
	    double v1  //In: End velocity
	    ) {
    double dt1 = t1-t0;
    double dt2 = dt1*dt1;
    double dt3 = dt2*dt1;

    StartTime = t0;
    EndTime = t1;
    X0 = x0; V0 = v0;

    a = -(v1+2*v0)/dt1 + 3.0*(x1-x0)/dt2;
    b = (v1+v0)/dt2 - 2.0*(x1-x0)/dt3;

  }


  /*************************************************************************
   * Method: operator () (double)
   * Summary: returns the position at the given point in time.
   * Description: This method computes the position of the dynamic
   *              motion at the passed in time.
   * Return: the position at the given point in time
   *************************************************************************/
  virtual double operator () (double t //In: Time at which position is
			               //    desired
			      ) {
    double dt1 = t-StartTime;
    double dt2 = dt1*dt1;
    double dt3 = dt2*dt1;
    return (X0 + V0*dt1 + a*dt2 + b*dt3);
  }


  /*************************************************************************
   * Method: operator () (double, double&)
   * Summary: returns the position at the given point in time.
   * Description: This method computes the position of the dynamic
   *              motion at the passed in time.  It also computes the
   *              first derivative (velocity) and returns that in the
   *              second parameter
   * Return: the position at the given point in time
   *************************************************************************/
  virtual double operator() (double t, //In:  Time at which position is
			               //     desired
			     double &V //Out: Value for the velocity
			     ) {
    double dt1 = t-StartTime;
    double dt2 = dt1*dt1;
    double dt3 = dt2*dt1;
    V = V0 + 2.0*a*dt1 + 3.0*b*dt2;
    return (X0 + V0*dt1 + a*dt2 + b*dt3);
  }


  /*************************************************************************
   * Method: operator () (double, double&, double&)
   * Summary: returns the position at the given point in time.
   * Description: This method computes the position of the dynamic
   *              motion at the passed in time.  It also computes the
   *              first derivative (velocity) and the second
   *              derivative (acceleration) and returns them in the
   *              second and third parameters.
   * Return: the position at the given point in time
   *************************************************************************/
  virtual double operator() (double t, //In:  Time at which position is
			               //     desired
			     double &V,//Out: Value for the velocity
			     double &A //Out: Value for the acceleration
			     ) {
    double dt1 = t-StartTime;
    double dt2 = dt1*dt1;
    double dt3 = dt2*dt1;
    V = V0 + 2.0*a*dt1 + 3.0*b*dt2;
    A = 2.0*a + 6.0*b*dt1;
    return (X0 + V0*dt1 + a*dt2 + b*dt3);
  }


  /*************************************************************************
   * Method: GetType()
   * Summary: Returns a char * string of the type of this item
   * Description: Returns a char * string of the type of this item
   * Return: Returns a char * string of the type of this item
   *************************************************************************/
  virtual char *GetType() {return "Dynamic spline 3";}
protected:

private:
  NET_DOUBLE X0;
  NET_DOUBLE V0;
  NET_DOUBLE a;
  NET_DOUBLE b;
};


/************************************************************************
 * Class: DYNAMIC_SPLINE_6
 *
 * Summary: Creates a 6th order polynomial matching position,
 *          velocity, and acceleration at the end points
 *
 * Description: This method does not create a spline through a number
 *              of points but, rather, creates a single polynomial
 *              through two points and makes sure that it matches both
 *              first and second derivatives at those points
 *
 * Examples: 
 *
************************************************************************/
class DYNAMIC_SPLINE_6 : public DYNAMIC_FLOAT_ITEM {
public:


  /*************************************************************************
   * Method: init(double, double, double, double,
   *              double, double, double, double)
   * Summary: Creates the polynomial matching the passed in values
   * Description: When you pass in an initial and final position,
   *              velocity, and acceleration, this will create the
   *              polynomial that matches those values.
   * Return: 
   *************************************************************************/
  void init(double t0, //In: Initial time
	    double x0, //In: Initial position
	    double v0, //In: Initial velocity
	    double a0, //In: Initial acceleration
	    double t1, //In: End time
	    double x1, //In: End position
	    double v1, //In: End velocity
	    double a1  //In: End acceleration
	    ) {
    double dt1 = t1-t0;
    double dt2 = dt1*dt1;
    double dt3 = dt2*dt1;

    StartTime = t0;
    EndTime = t1;
    X0 = x0; V0 = v0; A0 = a0;

    double alpha = x1 - X0 - V0*dt1 - A0*dt2/2.0;
    double beta = v1 - V0 - A0*dt1;
    double gamma = a1 - A0;

    a = (alpha*10.0 - beta*4.0*dt1 + gamma*dt2/2.0) / dt3;
    b = (-alpha*15.0/dt1 + beta*7.0 - gamma*dt1) / dt3;
    c = (alpha*6.0/dt2 - beta*3.0/dt1 + gamma/2.0) / dt3;
  }


  /*************************************************************************
   * Method: operator () (double)
   * Summary: Returns the position at a given point in time
   * Description: Returns the position at a given point in time
   * Return: Returns the position at a given point in time
   *************************************************************************/
  virtual double operator () (double t //In: Time at which position is
			               //    desired
			      ) {
    double dt1 = t-StartTime;
    double dt2 = dt1*dt1;
    double dt3 = dt2*dt1;
    double dt4 = dt3*dt1;
    double dt5 = dt4*dt1;
    return (X0 + V0*dt1 + A0*dt2/2.0 + a*dt3 + b*dt4 + c*dt5);
  }


  /*************************************************************************
   * Method: operator () (double, double&)
   * Summary: Returns the position at a given point in time plus the first
   *          derivative
   * Description: Returns the position at a given point in time plus the first
   *              derivative
   * Return: Returns the position at a given point in time
   *************************************************************************/
  virtual double operator () (double t, double &V) {
    double dt1 = t-StartTime;
    double dt2 = dt1*dt1;
    double dt3 = dt2*dt1;
    double dt4 = dt3*dt1;
    double dt5 = dt4*dt1;
    V = V0 + A0*dt1 + 3.0*a*dt2 + 4.0*b*dt3 + 5.0*c*dt4;
    return (X0 + V0*dt1 + A0*dt2/2.0 + a*dt3 + b*dt4 + c*dt5);
  }


  /*************************************************************************
   * Method:operator () (double, double&, double &) 
   * Summary: Returns the position at a given point in timeplus the first
   *          and second derivatives
   * Description: Returns the position at a given point in timeplus the first
   *              and second derivatives
   * Return: Returns the position at a given point in time
   *************************************************************************/
  virtual double operator () (double t, double &V, double &A) {
    double dt1 = t-StartTime;
    double dt2 = dt1*dt1;
    double dt3 = dt2*dt1;
    double dt4 = dt3*dt1;
    double dt5 = dt4*dt1;
    V = V0 + A0*dt1 + 3.0*a*dt2 + 4.0*b*dt3 + 5.0*c*dt4;
    A = A0 + 6.0*a*dt1 + 12.0*b*dt2 + 20.0*c*dt3;
    return (X0 + V0*dt1 + A0*dt2/2.0 + a*dt3 + b*dt4 + c*dt5);
  }


  /*************************************************************************
   * Method: GetType()
   * Summary: Returns a char * string of the type of this item
   * Description: Returns a char * string of the type of this item
   * Return: Returns a char * string of the type of this item
   *************************************************************************/
  virtual char *GetType() {return "Dynamic spline 6";}

protected:

private:
  NET_DOUBLE X0;
  NET_DOUBLE V0;
  NET_DOUBLE A0;
  NET_DOUBLE a;
  NET_DOUBLE b;
  NET_DOUBLE c;
};

/************************************************************************
 * Class: DYNAMIC_INT_CONSTANT
 *
 * Summary: A dynamic integer value
 *
 * Description: If you have integer values that you know are going to
 *              change over time and you know those values and times
 *              in advance, this provides the perfect data type for
 *              storing those values.
 *
 * Examples: 
 *
************************************************************************/
class DYNAMIC_INT_CONSTANT : public DYNAMIC_INT_ITEM {
public:

  /*************************************************************************
   * Method: DYNAMIC_INT_CONSTANT()
   * Summary: Default constructor for a DYNAMIC_INT_CONSTANT
   * Description: Default constructor sets the value to 0.
   * Return: 
   *************************************************************************/
  DYNAMIC_INT_CONSTANT() {a=0;}

  /*************************************************************************
   * Method: operator = (int)
   * Summary: overloaded assignment operator that simply assigns a value
   * Description: The equals operator has been overloaded so that you can
   *              treat this type just like a normal integer.
   * Return: the new value
   *************************************************************************/
  int operator = (int v //In: New value for the DYNAMIC_INT_CONSTANT
		  ) {return (a = v);}


  /*************************************************************************
   * Method: Set()
   * Summary: Assigns the DYNAMIC_INT_CONSTANT a new value
   * Description: If you don't like overator overloading, you can use the
   *              Set method rather than just assignment
   * Return: void
   *************************************************************************/
  void Set(int v//In: New value for the DYNAMIC_INT_CONSTANT
	   ) {a = v;}

  /*************************************************************************
   * Method: operator () (double)
   * Summary: Normally used to return value at a specific time but
   *          this is a constant.
   * Description: Normally, one can ask for the value of a DYNAMIC_*
   *              at a specific point in time.  Since a
   *              DYNAMIC_INT_CONSTANT is constant, this simply
   *              returns the value.
   * Return: The value of the DYNAMIC_INT_CONSTANT
   *************************************************************************/
  virtual int operator () (double) {return a;}

  /*************************************************************************
   * Method: GetType()
   * Summary: Returns a char * string of the type of this item
   * Description: Returns a char * string of the type of this item
   * Return: Returns a char * string of the type of this item
   *************************************************************************/
  virtual char *GetType() {return "Dynamic int constant";}
protected:
private:
  NET_INT a;
};

/************************************************************************
 * Class: DYNAMIC_LOGICAL_CONSTANT
 *
 * Summary: A dynamically changing logical constant.
 *
 * Description: If you know your object will have predetermined values
 *              of true or false, you could put all of them in a big
 *              list and then use this to hold each of their values
 *
 * Examples: 
 *
************************************************************************/
class DYNAMIC_LOGICAL_CONSTANT : public DYNAMIC_LOGICAL_ITEM {
public:

  /*************************************************************************
   * Method: DYNAMIC_LOGICAL_CONSTANT()
   * Summary: default constructor
   * Description: The default constructor initializes the value to 0 (FALSE)
   * Return: a new DYNAMIC_LOGICAL_CONSTANT
   *************************************************************************/
  DYNAMIC_LOGICAL_CONSTANT() {a=0;}

  /*************************************************************************
   * Method: operator = (int)
   * Summary: overloaded assignment assigns a new value
   * Description: This makes a DYNAMIC_LOGICAL_CONSTANT act just like
   *              a logical for assignment since you can just assign
   *              it an integer value
   * Return: The new value of the DYNAMIC_LOGICAL_CONSTANT
   *************************************************************************/
  int operator = (int v//In: New value for the DYNAMIC_LOGICAL_CONSTANT
		  ) {return (a = v);}


  /*************************************************************************
   * Method: Set(int)
   * Summary: Assigns a new value to this DYNAMIC_LOGICAL_CONSTANT
   * Description: If you don't like operator overloading (using the =
   *               operator, you can use this method to assign a new
   *               value.
   * Return: void
   *************************************************************************/
  void Set(int v//In: New value for the DYNAMIC_LOGICAL_CONSTANT
	   ) {a = v;}

  /*************************************************************************
   * Method: operator () (double)
   * Summary: Normally used to return value at a specific time but
   *          this is a constant.
   * Description: Normally, one can ask for the value of a DYNAMIC_*
   *              at a specific point in time.  Since a
   *              DYNAMIC_LOGICAL_CONSTANT is constant, this simply
   *              returns the value.
   * Return: The value of the DYNAMIC_LOGICAL_CONSTANT
   *************************************************************************/
  virtual int operator () (double) {return a;}

  /*************************************************************************
   * Method: GetType()
   * Summary: Returns a char * string of the type of this item
   * Description: Returns a char * string of the type of this item
   * Return: Returns a char * string of the type of this item
   *************************************************************************/
  virtual char *GetType() {
    if (a) {
      return "Dynamic logical constant = LOGICAL_TRUE";
    }else{
      return "Dynamic logical constant = LOGICAL_FALSE";
    }
  }
protected:
private:
  NET_INT a;
};

#endif
