// File : SpProxySubscriber.C

#include "SpProxySubscriber.H"
#include "SpSubscriberHandle.H"
#include "RB_Types.H"

#include "RB_SpBinaryTree.H"
#include "RB_SpDefineClass.H"

RB_DEFINE_CLASS(SpSubscriberHandle);
RB_DEFINE_CLASS(RB_SpBinaryTree);

#include "SpEvent.H"
extern SpEvent *CurrentEvent;

int SpProxySubscriber::DistListId = SpProxySubscriber::NUM_DEFAULT_DIST_LIST;

int 
SpProxySubscriber::GetNewDistListId(void){
  return (DistListId++);
}

SpProxySubscriber::SpProxySubscriber()
  : node(new RB_SpBinaryTree) {
  node->SetBalancedTreeMode();
}

SpProxySubscriber::~SpProxySubscriber() {
  RB_SpBinaryTree *t;
  SpSubscriberHandle *h;

  while ((t = (RB_SpBinaryTree *) node->RemoveFirstElement())){
    while ((h = (SpSubscriberHandle *) t->RemoveFirstElement()))
      delete h;
    delete t;
  }
  delete node;
}

void
SpProxySubscriber::debug(ostream& os){
  RB_SpBinaryTree *t;
  SpSubscriberHandle *h;

  t = (RB_SpBinaryTree *) node->GetFirstElement();
  while (t){
    h = (SpSubscriberHandle *) t->GetFirstElement();
    os << "Subscriber on node " << node->GetCurrentKey() << " : ";
    h = (SpSubscriberHandle *) t->GetFirstElement();
    while (h){
      os << h->GetGlobalId() << ", "; 
      h = (SpSubscriberHandle *) t->GetNextElement();
    }
    os << endl;
    t = (RB_SpBinaryTree *) node->GetNextElement();
  }
}

void
SpProxySubscriber::Add(int dist_list_id,
                   int _node, int _type, int _local_id, int _global_id,
                   int &deliver,          // TRUE  if need to deliver proxy to the subscriber's node
                                          // FALSE if proxy has already been delivered to the
                                          //       subscriber node, only need to notify subscriber 
                   int &duplicate,        // TRUE  if proxy has already been delivered/notified
                   SpSubscriberHandle* &subscriber){
  RB_SpBinaryTree *node_tree;
  SpSubscriberHandle *subscriber_obj = (SpSubscriberHandle *) 0;

  duplicate  = 0;
  deliver    = 0;
  subscriber = (SpSubscriberHandle *) 0;

  node_tree = (RB_SpBinaryTree *) node->Find(_node);
  if (node_tree)
    subscriber_obj = (SpSubscriberHandle *) node_tree->Find(_global_id);

  if (!subscriber_obj){
    subscriber_obj = new SpSubscriberHandle(_local_id, _node, _type, _global_id);
    subscriber_obj = RB_NEW_SpSubscriberHandle(subscriber_obj);
  }
  subscriber_obj->incrNumSubscribe(dist_list_id);

  if (node_tree){
    if (subscriber_obj->getNumSubscribe()>1){
      duplicate      = 1;
      // DM note : The subscriber has already subscribed to the ancestor/descendant of 
      // this HLA.  Thus, proxy has already been delivered to the subscriber.
      // DO NOT decrement number of subscriber - the user must unsubscribe to all 
      // ancestor/descendant inorder to have the proxy undelivered.
    }
    else{
      // proxy has already been delivered to the subscriber's node,
      // notify subscriber of proxy
      node_tree->Insert(subscriber_obj, _global_id);
    }
  }
  else {
    node_tree = RB_NEW_RB_SpBinaryTree();
    node_tree->Insert(subscriber_obj, _global_id);
    node->Insert(node_tree, _node);
    deliver = 1;
    // "deliver" proxy to the subscriber's node and notify subscriber of proxy
  }

  subscriber = subscriber_obj;
}
   
int	// TRUE - if need to un-deliver the proxy
SpProxySubscriber::Remove(int dist_list_id, int _node, int _global_id){
  RB_SpBinaryTree *node_tree;
  SpSubscriberHandle *subscribe_obj;
  int undeliver = 0;

  node_tree = (RB_SpBinaryTree *) node->Find(_node);
  if (!node_tree) abort();

  subscribe_obj = (SpSubscriberHandle *) node_tree->Find(_global_id);
  if (!subscribe_obj) abort();

  subscribe_obj->decrNumSubscribe(dist_list_id);

  if (subscribe_obj->getNumSubscribe()>0) return (undeliver);

  undeliver = 1;

  subscribe_obj = (SpSubscriberHandle *) node_tree->Remove(_global_id);
  if (subscribe_obj)
    RB_DELETE_SpSubscriberHandle(subscribe_obj);

  if (node_tree->GetNumElements()<=0)
    RB_DELETE_RB_SpBinaryTree((RB_SpBinaryTree *) node->Remove(_node));

  return (undeliver);
}

void
SpProxySubscriber::Remove(int dist_list_id, BINARY_TREE &subscriber_nodes){ 
  RB_SpBinaryTree *node_tree;
  SpSubscriberHandle *subscriber_obj;
  BINARY_TREE remove_subscribers;
  int num_empty_nodes, *empty_nodes;
  SpBinaryTree *current_node;

  if (node->GetNumElements()<=0) return;

  num_empty_nodes = 0;
  empty_nodes     = new int [node->GetNumElements()];

  node_tree = (RB_SpBinaryTree *) node->GetFirstElement();  
  while (node_tree){

    subscriber_obj = (SpSubscriberHandle *) node_tree->GetFirstElement();  
    while (subscriber_obj){
      subscriber_obj->resetNumSubscribe(dist_list_id);
      if (subscriber_obj->getNumSubscribe()<=0)
        remove_subscribers.Insert(subscriber_obj, subscriber_obj->GetGlobalId());
      subscriber_obj = (SpSubscriberHandle *) node_tree->GetNextElement();
    }

    current_node = new SpBinaryTree();
    subscriber_nodes.Insert((void *) current_node, int(node->GetCurrentKey()));

    while ((subscriber_obj = (SpSubscriberHandle *) remove_subscribers.RemoveFirstElement())){
      current_node->Insert((void *) new SpProxyObjHandle(subscriber_obj->GetLocalId(),
                                                         subscriber_obj->GetNode(),
                                                         subscriber_obj->GetObjType(),
                                                         subscriber_obj->GetGlobalId()), 
                           subscriber_obj->GetGlobalId());
      subscriber_obj = (SpSubscriberHandle *) node_tree->Remove(subscriber_obj->GetGlobalId());
      if (subscriber_obj)
        RB_DELETE_SpSubscriberHandle(subscriber_obj);
    }

    if (node_tree->GetNumElements()<=0)
      empty_nodes[num_empty_nodes++] = int(node->GetCurrentKey());
 
    node_tree = (RB_SpBinaryTree *) node->GetNextElement();
  }// node_tree

  for (int i=0; i<num_empty_nodes; i++)
    RB_DELETE_RB_SpBinaryTree((RB_SpBinaryTree *) node->Remove(empty_nodes[i]));

  delete [] empty_nodes;
}

void
SpProxySubscriber::SetDestination(DESTINATION &destination){
  RB_SpBinaryTree *node_tree;

  destination.Reset();

  node_tree = (RB_SpBinaryTree *) node->GetFirstElement();
  while (node_tree){
    destination.Set(int(node->GetCurrentKey()));
    node_tree = (RB_SpBinaryTree *) node->GetNextElement();
  }
}
