
 /* The basic MPI comm library.  IF messages must be sent to a different
   LocalBoxNumber, you might need to use the global node ID.  I have not 
   tested any of the network functions, although the stubs for these
   functions are in place, for compatibility with Metron's original design.  

   In the MPI systems I have available for testing, every MPI process 
   has a unique rank SpComm_LocalNode, and LocalBoxNumber = 0 for all 
   participating processes.  No messages are ever sent or received to
   or from a different LocalBoxNumber, even when the application is run 
   on a network of workstations.

   If we migrate to a multiprocessor system where SpComm_LocalNode is
   local to the LocalBoxNumber -- then the network functions would
   have to be implemented.  In particular, it will be necessary to
   translate from the local process ID (i.e. SpComm_LocalNode) to a 
   global node ID as a function of the LocalBoxNumber prior to sending
   a message to a different box.

   If a message is received from outside LocalBoxNumber, the global node ID
   would have to be be translated to the local process ID (SpComm_LocalNode)
   as a function of the LocalBoxNumber of the recipient processor.   

   Roger Hillson
   Code 5583
   Naval Research Laboratory
   Washington, DC 20375-5337
   
   Initial Version:  13 March 1998

   Revised:          25 September 1998

   Revised:          28 March 1999

   Major Changes (25 September 1998):

   -  Incorporate latest SPEEDES command line parser, etc.  
   -  Eliminate use of MPI_Bsend;  use MPI_Send instead.
   -  Eliminate limit on message size, at the cost of
      removing tag incrementing to prevent backmasking.
      (Non-overtaking property of MPI messages is sufficient.)
   -  Fix build_Simtime() function to reflect changes
      in the SPEEDES class Simtime.
   -  Revise Fuzzy Barrier functions.
   -  Add functions SpComm_CleanupOther() and SpComm_CleanupShm().

   Major Changes (28 March 1999)
   -  Reinsert MPI_Bsend to avoid live-lock for large message sizes

*/

#include <stdlib.h>
#include <unistd.h>
#include <sys/wait.h>
#include <sys/shm.h>
#include <errno.h>
#include <signal.h>
#include <iostream.h>
#include <mpi.h>

#include "SpeedesComm.H"
#include "SpComm_Net.H"
#include "SpeedesCommPar.H"
#include "MPI_Strucs.H"
#include "MPI_Globals.H"
#include "Queue.H"
#include "DataParser.H"

#define IMPORTANT_DEBUG_SP_COMM
#define SAFE_RECEIVE

//  Define buffer size for circular queue
#define QueueBufferSize N_MESSAGE_TYPES + MPI_MESSAGE_TYPES


int Startup = 0;

/*
** These variable correspond to node information.
*/

int *GlobalNodeIDs;           // Global and local node id's so that we
int *LocalNodeIDs;            // can reroute message appropriately
int *LocalNodeZeros;          /* This is an array of the GIDs of the local node
				 zero's */

int SpComm_Nboxes = 0;         /* The total number of physical computers.
                                  If one, there is ONE physical system.  
                                  Value is returned by
                                  SpComm_Pre_Fork_Net_Get_IDs().        */
int *NodeBoxNumber;            /* A map from GlobalNodeIDs to which physical
				  computer a node resides on. */
int LocalBoxNumber = 0;       // The current node's box number
int SpComm_Node = 0;          // Our global node (process) number
int SpComm_Nnodes = 1;        // The number of global nodes
int SpComm_Nlocal;            // The number of local nodes
int SpComm_LocalNode;         // Our local node number (i.e. MPI process rank)
int GroupId = 1;              // Our group ID
static int Parent = 1;        // 1 if we are the parent local node, 0 o/w
int SpCounter=0;

//  Begin MPI global block for SpeedesCommMPI.C

int MPISize;  // global variable for total number of MPI processes
int MPIRank;  // global variable for MPI rank of local process


MPI_Datatype MPI_SIMTIME; /*  MPI datatype which will be created
                              for passing Speedes SIMTIME objects. 
                              This ptr is initialized by a call to
                              build_SIMTIME in StartUp().       */

MPIst MPISimTime;     // MPIst      is typdefed in MPI_Strucs.H

extern MPI_Group  World_Group, Make_Group;

//  Declare the three communicators required by Speedes
extern MPI_Comm   MPI_Fuzzy_Comm;  // MPI_Comm for fuzzy barrier operations
extern MPI_Comm   MPI_PTP_Comm;    // MPI_Comm for point-to-point operations
extern MPI_Comm   MPI_Sync_Comm;   // MPI_Comm for synchronous sends
extern MPI_Comm   MPI_Bcast_Comm;  // MPI_Comm for asynchronous broadcasts

//  Declare the buffer and buffer size for buffered sends and receives
//  The buffer is attached in  SpComm_StartUp()

char  MPI_Buffer[MPI_BUFFER_SIZE];
  
/*  This file contains the MPI functions for sending and receiving
    blocking, synchronous, and multicast messages.                    */

int MsgInOutCount = 0;  // Used for MPI blocking sends

/*  Declare the queue object for saving the MPI-received BUFFERED messages. */

//  N_MESSAGE_TYPES is defined in SpeedesComm.

BUFFER_QUEUE MPI_InQueue[QueueBufferSize];

static int FuzzyBarrierCounter = 0; 

/*  Prototype for build_SIMTIME, which constructs an MPI data type for
    passing objects of class SIMTIME.                                  */

void build_SIMTIME(MPIst *simtime, 
		   MPI_Datatype *ptr_simtime);

// User-defined minimization functions for finding global SIMTIME minimum
void SimMinFun(void *in, void *inout, int *len, MPI_Datatype *datatype);
void SimMaxFun(void *in, void *inout, int *len, MPI_Datatype *datatype);


//////////////////////////////////////////////////////////////////////

/* How to look up a local node ID from a global node ID */

#define LocalNodeID(GlobalNodeID) (LocalNodeIDs[GlobalNodeID])
#define IsLocalID(GlobalNodeID) (NodeBoxNumber[GlobalNodeID] == LocalBoxNumber)
#define LocalNodeZero(GlobalNodeID) (LocalNodeZeros[GlobalNodeID])

int GlobalNodeID(int LocalNodeID) {

  int i;
  for(i=0; ((NodeBoxNumber[i] != LocalBoxNumber) &&
	    (LocalNodeIDs[i] != LocalNodeID)); ++i);
  return i;
}

//////////////////////////////////////////////////////////////////////

//  This function is required by qnet.

void SpComm_CleanupShm()
{
return;
}

//////////////////////////////////////////////////////////////////////
/*
** At exit function that cleans up the MPI functions
*/

void SpComm_CleanupOther() {
  //  This routine is called at exit.
  //  Do not detach from the MPI buffer.  This causes a fault.

  SpComm_Net_Terminate();
  if (!Parent) return;

  if (GlobalNodeIDs)  delete [] GlobalNodeIDs;
  if (LocalNodeIDs)   delete [] LocalNodeIDs;
  if (LocalNodeZeros) delete [] LocalNodeZeros;
}

//////////////////////////////////////////////////////////////////////

/*
void sigintfunc(int) {
  if (SpComm_Node == 0) cout << endl;
  exit(1);
}
*/

//////////////////////////////////////////////////////////////////////

const char *CmdLineArgs[]={
  "nnodes",
  "lnodes",
  "group",
  "pipe",
  "help",
  NULL
};
//////////////////////////////////////////////////////////////////////

enum{
  NNODES,
  LNODES,
  GROUP,
  PIPE,
  HELP
};

//////////////////////////////////////////////////////////////////////

void CommandLineHelp()
{
  cout<<"Command line arguments are:"<<endl
      <<endl
      <<" -nnodes <total number of nodes>"<<endl
      <<" -lnodes <total number of local nodes>"<<endl
      <<" -group <your group ID"<<endl
      <<" -help  Displays this message"<<endl
      <<endl
      <<" If nnodes is not given and lnodes is given, that number is"<<endl
      <<" taken to be the total number of nodes "<<endl
      <<" If none of the arguments are given, the proper values are"<<endl
      <<" read from the speedes.par file"<<endl;
  exit(0);
}

//////////////////////////////////////////////////////////////////////

int MyStrcmpOnLength(const char * str1, const char * str2, const int length){
  int rc = 0;
  for (int i=0; (i < length) && (!rc); ++i){
    if (*str1 != *str2){
      if       (*str1 < *str2) rc = -1;
      else if  (*str1 > *str2) rc = 1;
    }
    ++str1;
    ++str2;
  }
  return rc;
}
//////////////////////////////////////////////////////////////////////

void BadArgument(const char * arg, const char * errorMessage = NULL)
{
  cout<<endl<<"In bad argument"<<endl
      <<"Bad command line argument: "<<arg<<endl;
  if (errorMessage!= NULL){
    cout<<errorMessage<<endl;
  }
  CommandLineHelp();
  exit(-1);
}

//////////////////////////////////////////////////////////////////////

/*
  findverb is the workhorse of the whole process, it returns two things:
  
  1) a pointer to the first non-whitespace character after a verb is
  found, or NULL if it can't parse a verb from the string.
  
  2) the number of the verb in the verbs[] table. this should be ignored
     if the pointer returned is NULL.
*/
int FindVerb(char *p)
{
  int i = 0;
  int verb = -1;

  if (*p != '-'){
    BadArgument(p);
  }
  else{
    ++p; //Skip over the '-'
    while (CmdLineArgs[i] != NULL){
      if (MyStrcmpOnLength(CmdLineArgs[i], p, strlen(p)) == 0){
	/* Found a match */
	if (verb == -1){
	  verb = i;
	}
	else{			/* Found a second match!! Bad!!! */
	  char matches[256];
	  sprintf(matches, "At least two matches!!!!: %s %s",
		  CmdLineArgs[i], CmdLineArgs[verb]);
	  BadArgument(p, matches);
	}
      }
      ++i;
    }
  }
  return verb;
}
//////////////////////////////////////////////////////////////////////

void HandleCommandLine(int argc,
		       char **argv,
		       int &nTot,
		       int &nLoc,
		       int &group,
		       char ** pInfo)
{
  int verb;
  char s[50];
  char * pipeInfo = NULL;
  nLoc = 0;
  nTot = 0;
  int i = 1;
  group = -1;
  int NumMPIProcs, MPIRank;

  //  Initialize MPI -- no MPI calls can preceed this statement -rh
MPI_Init(&argc, &argv);

  /*  If we are not MPI process 0, return immediately.  The command
      line parameters will be handled by process 0, and broadcast to
      the remaining processes.                                    */

MPI_Comm_rank(MPI_COMM_WORLD, &MPIRank);
if(MPIRank != 0) return;


  if (argc > 1){


    //First handle old stype command line aruments
    if (('0' <= argv[1][0]) && ('9' >=  argv[1][0])){
      nTot = atoi(argv[1]);
      char junkString[256];
      sprintf(junkString, "%d",nTot);
      if (strcmp(junkString, argv[1])){
	cout<<"Bad command line "<<argv[1]<<endl;
	exit(17);
      }
      i = 2;
      if (argc > 2){
	if (('0' <= argv[2][0]) && ('9' >=  argv[2][0])){ 
	  nLoc = atoi(argv[2]); 
	  char junkString[256]; 
	  sprintf(junkString, "%d",nLoc); 
	  if (strcmp(junkString, argv[2])){ 
	    cout<<"Bad command line "<<argv[2]<<endl; 
	    exit(17);
	  }
	  i = 3;
	  if (argc > 3){
	    if (('0' <= argv[3][0]) && ('9' >=  argv[3][0])){
	      group = atoi(argv[2]);
	      char junkString[256];
	      sprintf(junkString, "%d",group);
	      if (strcmp(junkString, argv[3])){
		cout<<"Bad command line "<<argv[3]<<endl;
		exit(17);
	      }
	      i = 4;
	    }
	  }
	}
      }
    }


    for (; i < argc; ++i){
      int nTotGivenTwice = 0;
      int nLocGivenTwice=0;
      verb = FindVerb(argv[i]);
      switch(verb){
      case NNODES:

	if (nTot != 0){
	  cout<<"Cannot use both old and new command line arguments!!!"
	      <<endl;
	  exit(17);
	}
	if (nTotGivenTwice){
	  cout<<"Total number of nodes can only be given once"<<endl;
	  exit(17);
	}
	nTotGivenTwice=1;
	nTot = atoi(argv[++i]);
	sprintf(s, "%d",nTot);
	if (strcmp(s, argv[i])){
	  sprintf(s, "-nnodes %d",nTot);
	  BadArgument(s);
	}
	break;
      case LNODES:
        if (nLoc != 0){ 
          cout<<"Cannot use both old and new command line arguments!!!" 
              <<endl; 
	  exit(17);
        } 
        if (nLocGivenTwice){ 
          cout<<"Total number of nodes can only be given once"<<endl; 
          exit(17); 
        } 
        nLocGivenTwice=1; 
	nLoc = atoi(argv[++i]);
	sprintf(s, "%d", nLoc);
	if (strcmp(s, argv[i])){
	  sprintf(s, "-nnodes %d", nLoc);
	  BadArgument(s);
	}
	break;
      case GROUP:
	group = atoi(argv[++i]);
	sprintf(s, "%d", group);
	if (strcmp(s, argv[i])){
	  sprintf(s, "-nnodes %d", group);
	    BadArgument(s);
	}
	break;
      case PIPE:
	*pInfo = argv[++i];
	break;
      case HELP:
	CommandLineHelp();
	break;
      }
    }
    if (nTot == 0){
      nTot = nLoc;
    }
    else if (nLoc == 0){
      nLoc = nTot;
    }
    else if ((nTot == 0) && (nLoc == 0)){
      cout<<"Cannot run SPEEDES with no nodes!"<<endl;
      exit(-1);
    }
    if ((nLoc > nTot) && (nLoc > 0) && (nTot > 0)){
      cout<<"Cannot have more local nodes than global node"<<endl;
      exit(-1);
    }
    if ((pipeInfo == NULL) && (nTot > 1)){ 
/*Need to store argv[0] for spawn()*/
      *pInfo = argv[0];
    }
    if (group == -1){
      DATA_PARSER parser("speedes.par");
      parser.GoTo("SpeedesServer", NULL);
      group = parser.GetInt("Group");
    }
  }
  else{				/* No arguments given so go to the
				   speedes.par file for all the info */
    DATA_PARSER *parser = new DATA_PARSER("speedes.par");
    parser->GoTo("SpeedesServer", NULL);
    group = parser->GetInt("Group");
    parser->GoTo("parameters", NULL);
    nTot = parser->GetInt("n_nodes");
    nLoc = nTot;
  }
  cout.flush();
}
//////////////////////////////////////////////////////////////////////

int SpComm_StartUp(int nTot, int nLoc, int group, char * pipeInfo) 
{

  //  char* pipeInfo is not used by the MPI libraries
  int i;
  char shmSegmentName[8192]; //Should not be any longer than this
   LocalBoxNumber = 0;
  int parentProcessId = -1;
  int NumMPIProcs;
 
  MPI_Comm_rank(MPI_COMM_WORLD,&MPIRank);
  MPI_Comm_size(MPI_COMM_WORLD,&MPISize);

  /*  Assume that SpComm_LocalNode := MPIRank.  In effect, we are
      assuming that even if there are multiple boxes, there is only
      one ordering of the MPI nodes, and this ordering spans the boxes.   */

  SpComm_LocalNode = MPIRank;

  //  Announce startup ...
  cout<<"Start node "<<SpComm_LocalNode<<endl;
  system("hostname");

  //  It will do 'initially' to assume that the rank 0 process always
  //  supports stdinput and stdoutput.  Production system should verify
  //  this assumption -- it's not part of the MPI standard.

      //  Check to verify that the number of processes spawned by
      //  MPI correspond to the number of processes requested by
      //  the user!

if (SpComm_LocalNode == 0)
  {

  if (nTot < 0){
    SpComm_Nnodes = -nTot;
    SpComm_Nlocal = -nLoc;
  }
  else{

    SpComm_Nnodes = nTot;
    SpComm_Nlocal = nLoc;
  }

  GroupId = group;

        if(MPISize != SpComm_Nnodes)
	{
	  perror("Number of MPI processes does not equal number of user-requested processes!");
	  return ERROR_SP_COMM;
	}
      if(SpComm_Nlocal < MPISize)
	{
	  cout <<"Warning: " << SpComm_Nlocal <<" local processes < "<< NumMPIProcs << " total MPI processes." << endl; 
	}
    } // matches 'if (SpCommLocal_Node == 0)'

#ifdef MPI_DEBUG
  cout << "begin parameter broadcast: " << SpComm_LocalNode << endl;
#endif

  //  Check to verify that the number of spawned process <= MAX_PROC
  if(MPISize > MAX_PROC)
    {
      cout<<" Fatal error:  Number of MPI Processes "<<MPISize<< 
	" exceeds MAX_PROC "<<MAX_PROC<<" ."<<endl;
      exit(1);
    }

  //  Broadcast SpComm_Nnodes, SpComm_Nlocal, and GroupId to everyone. -rh
  MPI_Bcast(&SpComm_Nnodes, 1, MPI_INT, 0, MPI_COMM_WORLD); 
  MPI_Bcast(&SpComm_Nlocal, 1, MPI_INT, 0, MPI_COMM_WORLD); 
  MPI_Bcast(&GroupId,       1, MPI_INT, 0, MPI_COMM_WORLD);

#ifdef MPI_DEBUG 
  cout << "end broadcast: " << SpComm_LocalNode << endl;
#endif

  /* Set up GIDs and LIDs */
  GlobalNodeIDs = new int[SpComm_Nnodes];
  LocalNodeIDs = new int[SpComm_Nnodes];
  for (i=0; i < SpComm_Nnodes; ++i){
    GlobalNodeIDs[i] = i;
    LocalNodeIDs[i] = -1; //Initially, have no routing
  }
  LocalBoxNumber = 0; //Needes to be changed according to comm...
  if (SpComm_Net_Initialize()) return ERROR_SP_COMM;

  if(SpComm_LocalNode != 0) Parent = 0;

  SpComm_Post_Fork_Net_GetIDs(&LocalNodeIDs, SpComm_Nboxes, 
			      &LocalNodeZeros, &NodeBoxNumber, LocalBoxNumber);

  for (i = 0; i < SpComm_Nnodes; ++i){
    if ((LocalNodeIDs[i] == SpComm_LocalNode) &&
	(NodeBoxNumber[i] == LocalBoxNumber)){
      SpComm_Node = i;

    }
  }

  //  Build the MPI data type for sending and receiving SIMTIME objects. 

//  Attach the buffer for buffered MPI Sends
MPI_Buffer_attach(MPI_Buffer, MPI_BUFFER_SIZE);
   
  build_SIMTIME(&MPISimTime, &MPI_SIMTIME);

  //  Create communicators for different classes of communication
  //  MPI_Fuzzy_Comm => communicator for fuzzy barrier functions
  //  MPI_PTP_Comm   => communicator for point-to-point communications
  //  MPI_Sync_Comm  => communicator for synchronous communications
  //  MPI_Bcast_Comm => communicator for asynchronous broadcasts

  MPI_Comm_group( MPI_COMM_WORLD, &World_Group);
  MPI_Comm_create(MPI_COMM_WORLD, World_Group, &MPI_Fuzzy_Comm);
  MPI_Comm_create(MPI_COMM_WORLD, World_Group, &MPI_PTP_Comm);
  MPI_Comm_create(MPI_COMM_WORLD, World_Group, &MPI_Sync_Comm);
  MPI_Comm_create(MPI_COMM_WORLD, World_Group, &MPI_Bcast_Comm);

  /*
  ** Set the Node and Nnodes values
  */

  Startup = 1;

  SpComm_BarrierSync();
  return GOOD_SP_COMM;

} // end StartUp

//////////////////////////////////////////////////////////////////////

int SpComm_GetNumNodes() {
  if (!SpComm_Nnodes) {
    cout<<"Error GetNumNodes, number of nodes not yet set in Startup - Exiting"
	<< endl;
    exit(1);
  }
  return SpComm_Nnodes;
}

//////////////////////////////////////////////////////////////////////

int SpComm_GetNodeId() {
  if (!Startup) {
    cout << "Error GetNodeId, SpComm_StartUp must be called first - Exiting"
	 << endl;
    exit(1);
  }
  return SpComm_Node;
}
//  Begin MPI insertion

//////////////////////////////////////////////////////////////////////

void SpComm_Send(int destination, int type, int nbytes, void *buff)
{
  //  Node-to-node message
   char *buffer_ptr = (char *) buff; // cast because *buff is type void

  /*  Use a buffered send to send the message in char buff[] to destination.
      However, these messages are of variable length nbytes.  On the receive
      side we will test the message size prior to receiving.  There is no
      point in using an immediate send here, because 'buff' could be
      overwritten as soon as the call returns.                          */

  MPI_Bsend(buffer_ptr, nbytes, MPI_CHAR, destination, type, MPI_PTP_Comm);

}//  End MPI_SpComm_Send


//////////////////////////////////////////////////////////////////////

void SpComm_Send(DESTINATION *dest, int type, int nbytes, void *buff){

  /* This function is a rewrite of Metron's original SpComm_Send
     function.  MPI_Send() replaces SpComm_Shmem_Send();  with a 'for' 
     loop added to loop through all destinations.  It is difficult to use
     MPI_Bcast() here because this would require that all processes 
     compute a new communicator reflecting the broadcast group computed
     herein.  --rh                                                     */

  int Dest;
  static int **DEST;                // class defined in Destination.H
  char *buffer_ptr = (char *) buff; // cast because *buff is type void
  static int FirstTime = 1;
  int i, j;
  int PresBox;

static int BCounter  = 1;

  if (FirstTime){
    DEST = new int *[SpComm_Nboxes];
    for (i=0; i < SpComm_Nboxes; i++){
      DEST[i] = new int[SpComm_Nnodes+1]; //should really be (#machines/box)+1
    }
    FirstTime = 0;
  }
  for (i=0; i < SpComm_Nboxes; DEST[i++][0] = 0);
  Dest = dest->GetFirstNode();
  while (Dest != -1){
    PresBox=NodeBoxNumber[Dest];
    //  Count the destinations for Machine# 'PresBox' in DEST[PresBox][j=0] 
    //  Put address j in DEST[PresBox][j+1] where j = 0, 1, MPISIZE - 1
    DEST[PresBox][0] = DEST[PresBox][0] + 1;
    DEST[PresBox][DEST[PresBox][0]] = Dest;
    //cout<<"Dest: "<<Dest<<endl;
    Dest = dest->GetNextNode(Dest);
  }
  /* Now we have built up arrays of the destinations, send them on to
     their respective local node zero's for redistribution */

  SpComm_Receive();  //  Attempt a receive before sending.

  //  SpComm_Nboxes should be set to ONE by SpComm_Pre_Fork(..) call
  for (i=0; i < SpComm_Nboxes; i++){
    //cout<<"DEST["<<i<<"][0] "<<DEST[i][0]<<endl;
    if ((DEST[i][0])){ // If we have messages for box number i
      if (i == LocalBoxNumber)  // If local box, use buffered MPI send
	{

	  for (j=0; j < DEST[i][0]; ++j)
	    { // cout << "index j "<< j <<" "<< DEST[i][j + 1] << endl;
	      // cout << "DEST[i][j+1] "<<DEST[i][j+1]<<endl;
	      //MPI_Send(buff, nbytes, MPI_CHAR, DEST[i][j + 1], type, MPI_PTP_Comm);}

         MPI_Bsend(buffer_ptr, nbytes, MPI_CHAR, DEST[i][j + 1], type, MPI_PTP_Comm);}

	  /*  In DEST[i][j], i => machine#, j + 1 => destination.  Since this 
	      is multicast, neither the message or message type will change 
	      within the loop.  */

	} // end 'if (i == LocalBoxNumber)' 
      else
	SpComm_Net_Send(DEST[i], type, nbytes, buff);
    }
  }// matches 'for (i=0; i < SpComm_Nboxes; ++i)'  

  //SpComm_Receive();

}

//////////////////////////////////////////////////////////////////////

void SpComm_Send(int type, int nbytes, void *buff){
  //rh1
  /* Conventional broadcast of O(N*N*#msgs) if all processes participate. */

  /* This is a broadcast to everyone in MPI_PTP_Comm.  We do not
     have an array of destinations.  Don't send the message to 
     yourself at SpComm_LocalNode.                              */

char *buffer_ptr = (char *) buff; // cast because *buff is type void
int box, dest, counter = 0;
  

  SpComm_Receive();  // Attempt a receive before sending
 
  for (box = 0; box < SpComm_Nboxes; box++){
    if(box == LocalBoxNumber){
      for (dest=0; dest < SpComm_Nlocal; ++dest) //  Broadcast to LOCAL nodes.
	{
	  if(dest != SpComm_LocalNode){   // Don't send to yourself

	    // cout<<SpComm_LocalNode<<" dest "<<dest<<" tag "<< type <<" nbyte//s "<<nbytes<<endl;
	    //MPI_Ibsend(buffer_ptr, nbytes, MPI_CHAR, dest, type, MPI_PTP_Comm, &requests[counter++]);
 MPI_Bsend(buffer_ptr, nbytes, MPI_CHAR, dest, type, MPI_PTP_Comm);
 //cout<<SpComm_LocalNode<<" complete send dest "<<dest<<" tag "<< type <<" nbyte//s "<<nbytes<<endl;
	  } //if(dest != SpComm_LocalNode
	} // matches 'for (dest=0; dest < SpComm_Nlocal; ++dest)'
 
    } // end 'if (box == LocalBoxNumber)' 
    else
      /*  Send the message to each box other than your own.  */
      cout<<"Trying to broadcast to a box "<< box << endl;
    SpComm_Net_Send(type, nbytes, buff);
  } // matches 'for (box = 0; box < SpComm_Nboxes; i++)'

}

//////////////////////////////////////////////////////////////////////

char *SpComm_Receive(int type, int &nbytes)
{
  //  Overloaded Receive() function for blocking (synching) sends
  //  Called from MPI_SpComm_BlockingReceive(int &nbytes)
  //  which is called by MPI_SpComm_send_Sync()
  if(SpComm_GetPendingMessage(type,nbytes)==NULL){
  SpComm_Receive();
  return SpComm_GetPendingMessage(type, nbytes);}
}

//////////////////////////////////////////////////////////////////////

char *SpComm_GetPendingMessage(int type, int &nbytes)
{
  char * retval = NULL;
  int dest, // unused parameter to the call to QUEUE::Remove();
      Type; // unused parameter to the call to QUEUE::Remove();

/* if((type < 0)||(type > QueueBufferSize - 1))
	   { perror("Out of Range Message Type for MPI_InQueue [2].");
	   exit;} */

  if (MPI_InQueue[type].Count())
    {
      retval = MPI_InQueue[type].Remove(nbytes, Type, dest);
      //      cout << SpComm_LocalNode <<" GetPendingMessage: retval "<< retval << "Type "<<type<<" nbytes "<< nbytes << endl;
    }
  else  
    {nbytes = 0;}
  return retval;
};

//////////////////////////////////////////////////////////////////

extern int InFuzzyBarrier;

void SpComm_CoordinatedSend(int dest, int nbytes, void *buff){

static int BCounter  = 1;

  //  Node-to-node message
  if (InFuzzyBarrier)
    cout<<"Coordinated send while in fuzzy barrier 4"<<endl;
  ++MsgInOutCount;
  SpComm_Send(dest, MPI_SPCOMM_MESSAGE_BLOCKING_SEND, nbytes, buff);
}

//////////////////////////////////////////////////////////////////////

void SpComm_CoordinatedSend(DESTINATION *dest, int nbytes, void *buff){
  //  Node-to-multiple nodes

static int BCounter  = 1;
SpComm_Receive(); 

  if (InFuzzyBarrier)
    cout<<"Coordinated send while in fuzzy barrier 5"<<endl;
  int Node = dest->GetFirstNode();
  while (Node != -1){++MsgInOutCount; Node = dest->GetNextNode(Node);}
  // dest is a pointer to an object of type DESTINATION
  SpComm_Send(dest, MPI_SPCOMM_MESSAGE_BLOCKING_SEND, nbytes, buff);



}

//////////////////////////////////////////////////////////////////////

void SpComm_CoordinatedSend(int nbytes, void *buff){

static int BCounter  = 1;

  if (InFuzzyBarrier)
    cout<<"Coordinated send while in fuzzy barrier 6"<<endl;
  MsgInOutCount += (SpComm_GetNumNodes() - 1);
  SpComm_Send(MPI_SPCOMM_MESSAGE_BLOCKING_SEND, nbytes, buff);

  /*  Do a receive after every DO_RECEIVE sends to prevent queue of 
      buffered messages from becoming unacceptably large.      */
  //      if(!((BCounter++) % DO_RECEIVE)) {SpComm_Receive();  BCounter = 1;}
}
//////////////////////////////////////////////////////////////////////

#define NEW_BEHAVIOUR 0

char *SpComm_CoordinatedReceive(int &nbytes){

  void * RcvBuff = SpComm_GetPendingMessage(MPI_SPCOMM_MESSAGE_BLOCKING_SEND,
					    nbytes);
  //static int EnteredFuzzy = 0;
  //static int ExitedFuzzy = 0;
  if (RcvBuff){
    --MsgInOutCount;
    return (char*) RcvBuff;
  }

#if (NEW_BEHAVIOUR)
  if (!EnteredFuzzy){
    SpComm_EnterFuzzyBarrier();
    EnteredFuzzy = 1;
  }
  if (!ExitedFuzzy){
    while (!SpComm_ExitFuzzyBarrier()){
      RcvBuff = SpComm_Receive(MPI_SPCOMM_MESSAGE_BLOCKING_SEND, nbytes);
      if (RcvBuff != NULL) {
	--MsgInOutCount;
	return RcvBuff; 
      }
    }
  }
  ExitedFuzzy = 1;
#endif
  /* Now we have exited fuzzy barrier.  This means everyone thinks
     they have received all of their messages.  Now we start checking
     if we are done.  */
  while (SpComm_GlobalSum(MsgInOutCount)){
    RcvBuff = SpComm_Receive(MPI_SPCOMM_MESSAGE_BLOCKING_SEND, nbytes);
    if (RcvBuff != NULL) {
      --MsgInOutCount;
      return (char*) RcvBuff; 
    }
  }
  //EnteredFuzzy = 0;
  //ExitedFuzzy = 0;
  MsgInOutCount = 0;
  return NULL;
}


/*  This next section contains the MPI global reduction and fuzzy barrier
    functions.                                                   */

//////////////////////////////////////////////////////////////////////

void build_SIMTIME(MPIst *simtime, \
                   MPI_Datatype *ptr_simtime)
{
  /*  Build an MPI Datatype for the SPEEDES class SIMTIME  */

  /*  This procedure returns a ptr ptr_simtime to a new
      data type consisting of one double variable
      and four integer variables.  The input *simtime is a
      pointer to a variable of type struct (or class)
      defined in a main program.  Note that *ptr_simtime 
      only has to be defined ONCE, and can then be used 
      repeatedly.                                           */

  /*  The number of intrinsic elements in each "block" of
      the new type:  1 double, 1 int, 1 int, 1 int, 1 int.   */

  int block_lengths[5];

  /*  Displacement of each element from the start of the new 
      type.  */
  MPI_Aint displacements[5];

  /*  MPI data types for the successive elements.     */
  MPI_Datatype typelist[5];

  /*  Use for calculating displacements               */
  MPI_Aint start_address;
  MPI_Aint address;

/*  If one of the dataypes was a vector, it would have 
    multiple elements per block.  For 'triplet', there
    is only one element per block.  */

  block_lengths[0] = block_lengths[1]   
    = block_lengths[2] = block_lengths[3] = block_lengths[4] = 1;
  /*for(i=0; i<=2; i++)
    printf("block_length[%d] = %d\n",i,block_lengths[i]);*/

/*  Define the typelist */
  typelist[0] = MPI_DOUBLE;
  typelist[1] = MPI_INT;
  typelist[2] = MPI_INT;
  typelist[3] = MPI_INT;
  typelist[4] = MPI_INT;

  /*  Calculate the displacements                */

/*  I.    The displacement of ptr_src is 0       */
  displacements[0] = 0;

  /*  II.   Calculate ptr_src's starting address */
  MPI_Address(&simtime->Time, &start_address);

  /*  III.  Calculate the four remaining displacements
            by subtracting start_address         */
  MPI_Address(&simtime->Priority1, &address);
  displacements[1] = address - start_address;
  /*printf(" disp[1] %i \n",displacements[1]);*/  
  
  MPI_Address(&simtime->Priority2, &address);
  displacements[2] = address - start_address;
  /*printf(" disp[2] %i \n",displacements[2]);*/
   
  MPI_Address(&simtime->counter, &address);
  displacements[3] = address - start_address;
  /*printf(" disp[3] %i \n",displacements[3]);*/

 
  MPI_Address(&simtime->uniqueId, &address);
  displacements[4] = address - start_address;
  /*printf(" disp[4] %i \n",displacements[4]);*/

  /*  Finally, we build and commit the derived dataype */
  MPI_Type_struct(5, block_lengths, displacements,
		  typelist, ptr_simtime);
  MPI_Type_commit(ptr_simtime);

} //  matches build_SIMTIME()


//////////////////////////////////////////////////////////////////////

void SimMinFun(void *in, void *inout, int *len, MPI_Datatype *datatype)
{
  //  User-defined MPI function for finding minimum value of two SIMTIME 
  //  variables.  Operator must be used with MPI_Reduce or MPI_Allreduce.
  //  The '<' and '=' operators are overloaded SIMTIME operators.

  //  Operator can be applied to an array of length n, but in SPEEDES
  //  len would normally be 1.

  //  NB:  MPI_Op_create() requires that the user defined function has void
  //  pointers as arguments for the input and output vectors.  C++ requires
  //  explicit conversions from void pointers to other pointers.

  SIMTIME *invec, *inoutvec;
  invec    = (SIMTIME *) in;
  inoutvec = (SIMTIME *) inout;

  int i;
  int flag = 0;

  //invec->print();
  //inoutvec->print();

  for (i=0;  i < *len; i++) { 
    if(*invec < *inoutvec) *inoutvec = *invec;

    //  Increment the array pointers.

    invec++;
    inoutvec++;

  } // End For loop
} // End SimMinFun()

//////////////////////////////////////////////////////////////////////

void SimMaxFun(void *in, void *inout, int *len, MPI_Datatype *datatype)
{
  //  User-defined MPI function for finding minimum value of two SIMTIME 
  //  variables.  Operator must be used with MPI_Reduce or MPI_Allreduce.
  //  The '>' and '=' operators are overloaded SIMTIME operators.

  //  Operator can be applied to an array of length n, but in SPEEDES
  //  len would normally be 1.

  //  NB:  MPI_Op_create() requires that the user defined function has void
  //  pointers as arguments for the input and output vectors.  C++ requires
  //  explicit conversions from void pointers to other pointers.

  SIMTIME *invec, *inoutvec;
  invec    = (SIMTIME *) in;
  inoutvec = (SIMTIME *) inout;

  int i;
  int flag = 0;

  //invec->print();
  //inoutvec->print();

  for (i=0;  i < *len; i++) { 
    if(*invec > *inoutvec) *inoutvec = *invec;

    //  Increment the array pointers.

    invec++;
    inoutvec++;

  } // End For loop
} // End SimMaxFun()
//////////////////////////////////////////////////////////////////////
int InFuzzyBarrier = 0;

void SpComm_EnterFuzzyBarrier()
{
  /*  When a process enters the fuzzy barrier, it increments the
      fuzzy barrier counter.  When all processes have incremented
      the counter, everyone has entered the fuzzy barrier, and 
      SpComm_ExitFuzzyBarrier() will execute a barrier synchronization
      and return true.                                              */

 FuzzyBarrierCounter++;
  InFuzzyBarrier = 1;

return;

}//  END void  MPI_SpComm_EnterFuzzyBarrier()return;


//////////////////////////////////////////////////////////////////////

int SpComm_ExitFuzzyBarrier()
{

int AllSame = 0;

MPI_Allreduce(&FuzzyBarrierCounter,&AllSame,1,MPI_INT,MPI_LAND,MPI_Fuzzy_Comm);

if(AllSame)
  {
    //  IF we are here, then everyone has incremented FuzzyBarrierCounter,
    //	 and all processes have entered the fuzzy barrier.               
    MPI_Barrier(MPI_Fuzzy_Comm);
    InFuzzyBarrier = 0;
    return 1;  //  synchronization complete!
  }
else
  return 0;

} 
//  END int MPI_SpComm_ExitFuzzyBarrier()


//////////////////////////////////////////////////////////////////////

void SpComm_BarrierSync()
{
  //  Blocking barrier synchronization function
  MPI_Barrier(MPI_COMM_WORLD);
}
//////////////////////////////////////////////////////////////////////

//  Overload function SpComm_MPIGlobalSum(T value)

int SpComm_GlobalSum(int value)
{
  int result;
  //  Return the global sum for an integer variable.
  MPI_Allreduce(&value,&result,1,MPI_INT,MPI_SUM,MPI_COMM_WORLD);
  return result;
}
///////////////////////////////////////////////////////////////

double SpComm_GlobalSum(double value)
{
  double result;
  //  Return the global minimum for an integer variable.
  MPI_Allreduce(&value,&result,1,MPI_DOUBLE,MPI_SUM,MPI_COMM_WORLD);
  return result;
}

///////////////////////////////////////////////////////////////

//  Overload function SpComm_MPIGlobalMax(T value)

int SpComm_GlobalMax(int value)
{
  int result;
  //  Return the global maximum for an integer variable.
  MPI_Allreduce(&value,&result,1,MPI_INT,MPI_MAX,MPI_COMM_WORLD);
  return result;
}

//////////////////////////////////////////////////////////////////////

double SpComm_GlobalMax(double value)
{
  double result;
  //  Return the global maximum for an integer variable.
  MPI_Allreduce(&value,&result,1,MPI_DOUBLE,MPI_MAX,MPI_COMM_WORLD);
  return result;
}

//////////////////////////////////////////////////////////////////////
//  Overload function SpComm_MPIGlobalMin(T value)

int SpComm_GlobalMin(int value)
{
  int result;
  //  Return the global minimum for an integer variable.
  MPI_Allreduce(&value,&result,1,MPI_INT,MPI_MIN,MPI_COMM_WORLD);
  return result;
}

//////////////////////////////////////////////////////////////////////

double SpComm_GlobalMin(double value)
{
  double result;
  //  Return the global minimum for an integer variable.
  MPI_Allreduce(&value,&result,1,MPI_DOUBLE,MPI_MIN,MPI_COMM_WORLD);
  return result;
}

//  Define MPI functions for SpComm_MPI_GlobalMin(SimTime) and
//  SpComm_MPI_GlobalMax(SimTime)


//////////////////////////////////////////////////////////////////////

SIMTIME &SpComm_GlobalMin(SIMTIME &SimTime)
{
  /*  Find the global minimum for a SPEEDES variable of type
      SIMTIME.  SIMTIME is a class, and simtime.h includes
      overloaded operators for relational comparisons.  Return
      a reference to a variable of type SIMTIME.            */

  extern int MPIRank;
  extern int MPISize;
  extern MPI_Datatype MPI_SIMTIME;
 
  static int FirstTime = 1;
  static MPI_Op SimMinOp;
  SIMTIME input_class[1], output_class[1];

  // Return a handle to user-specified reduction operator

if(FirstTime)
  {
  MPI_Op_create(SimMinFun, 1, &SimMinOp);
  FirstTime = 0;
  }

  //  C-based MPI routines cannot handle references!
  //  So set a local variable equal in value.

  input_class[0] = SimTime;

  MPI_Allreduce(input_class, output_class, 1, MPI_SIMTIME, \
		SimMinOp, MPI_COMM_WORLD);

  SimTime = output_class[0];
  return SimTime; 

} // End SpComm_MPIUDF_GlobalMin(SimTime)

//////////////////////////////////////////////////////////////////////

SIMTIME &SpComm_GlobalMax(SIMTIME &SimTime)
{
  /*  Find the global maximum for a SPEEDES variable of type
      SIMTIME.  SIMTIME is a class, and simtime.h includes
      overloaded operators for relational comparisons.  Return
      a reference to a variable of type SIMTIME.            */

  extern int MPIRank;
  extern int MPISize;
  extern MPI_Datatype MPI_SIMTIME;
 
static int FirstTime = 1;  
static MPI_Op SimMaxOp;
SIMTIME input_class[1], output_class[1];

  // Return a handle to user-specified reduction operator

if(FirstTime)
  {
  MPI_Op_create(SimMaxFun, 1, &SimMaxOp);
  FirstTime = 0;
  }

  //  C-based MPI routines cannot handle references!
  //  So set a local variable equal in value.

  input_class[0] = SimTime;

  MPI_Allreduce(input_class, output_class, 1, MPI_SIMTIME, \
		SimMaxOp, MPI_COMM_WORLD);

  SimTime = output_class[0];
  return SimTime; 

} // End SpComm_MPIUDF_GlobalMax(SimTime)

//  End SpeedesComm.C


//////////////////////////////////////////////////////////////////

void SpComm_Net_Terminate()
{
  static int TerminatedAlready =0;

  if (!TerminatedAlready){
    cout<<SpComm_LocalNode<<" call MPI_Finalize "<<endl;
    system("hostname");
    MPI_Finalize();
    ++TerminatedAlready;
  }
}

//////////////////////////////////////////////////////////////////

int SpComm_Net_Initialize()
{
  return 0;
}

//////////////////////////////////////////////////////////////////

extern int SpComm_Nlocal;

/*  The pre-fork, post-fork distinction for the following calls probably
    doesn't make much sense in the MPI communication library, since MPI
    will order all processes in some absolute sense.  Also, there is
    no fork in the code, although there is code which is pre-MPI_Init()
    and post-MPI_Init().  But this distinction is retained for the sake
    of upward compatibility.                                         */

void SpComm_Pre_Fork_Net_GetIDs(int **LocalNodeIDs, int &Nboxes,
		       int ** LocalNodeZeros, int **NodeBoxNumber,
                       int &BoxNumber)
{
  *LocalNodeZeros = new int[SpComm_Nlocal];
  *LocalNodeIDs =   new int[SpComm_Nlocal];
  *NodeBoxNumber =  new int[SpComm_Nlocal];
  for (int i=0; i < SpComm_Nlocal; ++i){
    LocalNodeIDs[0][i] = i;
    LocalNodeZeros[0][i] = NodeBoxNumber[0][i] = 0;
  }
  Nboxes = 1;
  BoxNumber = 0; // Under MPI, there is one 'virtual' box
  return;
}
//////////////////////////////////////////////////////////////////

void SpComm_Post_Fork_Net_GetIDs(int **LocalNodeIDs, int &Nboxes,
		       int ** LocalNodeZeros, int **NodeBoxNumber,
                       int &BoxNumber)
{
  *LocalNodeZeros = new int[SpComm_Nlocal];
  *LocalNodeIDs  = new int[SpComm_Nlocal];
  *NodeBoxNumber = new int[SpComm_Nlocal];
  for (int i=0; i < SpComm_Nlocal; ++i){
    LocalNodeIDs[0][i] = i;
    LocalNodeZeros[0][i] = NodeBoxNumber[0][i] = 0;
  }
  Nboxes = 1;
  BoxNumber = 0; // Under MPI, there is one 'virtual' box
  return;
}

//////////////////////////////////////////////////////////////////

void SpComm_Net_Redop()
{
  return;
}

//////////////////////////////////////////////////////////////////

void SpComm_Net_Max(int * ival,double * dval,SIMTIME * Stime,int Compath)
{
  return;
}

//////////////////////////////////////////////////////////////////

void SpComm_Net_Sum(int * ival,double * dval,SIMTIME * Stime,int Compath)
{
  return;
}

//////////////////////////////////////////////////////////////////

void SpComm_Net_Min(int * ival,double * dval,SIMTIME * Stime,int Compath)
{
  return;
}

//////////////////////////////////////////////////////////////////


void SpComm_Net_Enter_nbRedop()
{
  return;
}

//////////////////////////////////////////////////////////////////

int SpComm_Net_nbRedop()
{
  return 1;
}
//////////////////////////////////////////////////////////////////

void SpComm_Net_Receive(int blocking)
{
  SpCounter++;

/*  SpCounter is defined as an external variable in TestSpComm.C.  It is
    required here to support the MPI Comm Library build.  SpCounter is
    also defined in SpComm_TCPIP.C;  this file is built and linked 
    when the SPEEDES shared memory communication library is built.  */
 	       
  return;
}

//////////////////////////////////////////////////////////////////

void SpComm_Net_Send(int dest, int type, int nbytes, void * buff)
{
  return;
}

//////////////////////////////////////////////////////////////////

void SpComm_Net_Send(DESTINATION *dest, int type, int nbytes, void *buff)
{
  return;
}

//////////////////////////////////////////////////////////////////

/* Assume int * dest is all on one box with dest[0] having the number of dest*/
void SpComm_Net_Send(int *dest, int type, int nbytes, void *buff)
{
  return;
}

//////////////////////////////////////////////////////////////////

/* send to everyone but self */
void SpComm_Net_Send(int type, int nbytes, void *buff)
{
  return;
}

//////////////////////////////////////////////////////////////////////

void SpComm_Receive()
{
  //rh1
  // For non-tree broadcast algorithm.
  // Send and receive standard and synching messages.
  // Do not modify without full regression testing.

  static char* Buffer; // Retain the buffer address between calls
  static int size = 1;
  

  static int first = 0;
  if(first == 0)
    {//  Go here to intialize receive buffer for the first time.

  Buffer =  new char [size];
  first = 1;
    }

  int flag;
  int len;   // length (bytes/chars) of incoming message

  MPI_Status FirstStatus, StatusProbe;

  while(1)
    {

      MPI_Iprobe(MPI_ANY_SOURCE, MPI_ANY_TAG, \
		 MPI_PTP_Comm, &flag, &StatusProbe);

      if(flag != 0)
	{// incoming message!

	  //  cout<<SpComm_LocalNode<<" Probed: src "<<StatusProbe.MPI_SOURCE<<" tag "<<StatusProbe.MPI_TAG<<" flag "<<flag<<endl;

	  // StatusProbe.MPI_TAG is the message type.

	  //  Increase the size of the receive buffer if necessary.
MPI_Get_count(&StatusProbe, MPI_CHAR, &len);
if( len > size )
  {
    //   cout << "Increase buffer from "<<sizeof(Buffer)<<" to "<<len<<" on process "<<SpComm_LocalNode<<endl;
delete [] Buffer;
size = len;
Buffer = new char [size];
  }

//  Now that the buffer size is adequate, receive the message.
	  MPI_Recv(Buffer, len, MPI_CHAR,  StatusProbe.MPI_SOURCE, StatusProbe.MPI_TAG, MPI_PTP_Comm, &FirstStatus);
	  
	  //  Having received a message, add to the tail of the circular queue.

 /*	  if((StatusProbe.MPI_TAG < 0)||(StatusProbe.MPI_TAG > QueueBufferSize - 1))
  { perror("Out of Range Speedes Message Type for MPI_InQueue [1].");
	   exit;}   */
	  
	  //  Having received a message, add to the tail of the circular queue.
	  MPI_InQueue[StatusProbe.MPI_TAG].AddToTail((char *)Buffer,len,StatusProbe.MPI_TAG, SpComm_LocalNode);
   
	}// matches 'if(flag != 0)'
      else
      	break;
	
    } // matches 'while(flag)'

} // End void MPI_SpComm_Receive()

//////////////////////////////////////////////////////////////////////
