 // TestSpComm.C - This file tests all of the SPEEDES Communications and
// measures performance.

#include <iostream.h>
#include <stdlib.h>
#ifndef __SPEEDES_NT__
#include <unistd.h>
#endif
#include <time.h>

#include "SpeedesComm.H"
#include "StopWatch.H"
#include "DataParser.H"
#include <mpi.h>

#define N_ITERATIONS 10

#if defined (__CYGWIN32__) || defined(__SPEEDES_NT__)
inline int lrand48(){return rand();}
inline void srand48(int seed){srand(seed);}
#endif

unsigned int Rand(unsigned int RandMax) {return (lrand48()%RandMax); } //Gives random number from 0-RandMax-1

extern int SpComm_LocalNode;
int TestDEBUG = 0;
int Node = 0;
int Nnodes = 1;
int CommandLineTest = 0;

double TestBarrierSync(int Nits);
double TestFuzzyBarrier(int Nits);

double TestGlobalSumInt(int Nits);
double TestGlobalMinInt(int Nits);
double TestGlobalMaxInt(int Nits);

double TestGlobalSumDouble(int Nits);
double TestGlobalMinDouble(int Nits);
double TestGlobalMaxDouble(int Nits);

double TestGlobalMinSimTime(int Nits);
double TestGlobalMaxSimTime(int Nits);

double SpComm_send_receive_single_messages(int Nits, int MsgSize);
double SpComm_send_Sync(int Nits, int MsgLen);
double TestMultiSend(int Nits, int MsgLen);
double TestBroadCast(int Nits, int MsgLen);

double TestCoordinatedMultiSend(int Nits, int MsgLen);
double TestCoordinatedBroadCast(int Nits, int MsgLen);

double PingTest(int nits, int msglen);
double BandwidthTest(int nits, int msglen);

void RandomTests(int Nits, int MsgLen);
void TotallyRandomTests(int Nits, int MsgLen);


double debugtest(int nits = 0);

enum {
  ALL_TESTS = 1,
  ALL_SYNC_TESTS,
  BARRIER_SYNC,
  FUZZY_BARRIER,
  INTEGER_GLOBAL_SUM,
  INTEGER_GLOBAL_MIN,
  INTEGER_GLOBAL_MAX,
  DOUBLE_GLOBAL_SUM,
  DOUBLE_GLOBAL_MIN,
  DOUBLE_GLOBAL_MAX,
  SIMTIME_GLOBAL_MIN,
  SIMTIME_GLOBAL_MAX,
  SEND_RECEIVE_TEST,
  SEND_SYNC_TEST,
  MULTISEND,
  BROADCAST,
  BLOCKING_MULTISEND,
  BLOCKING_BROADCAST,
  PING_TEST,
  BANDWIDTH_TEST,
  RANDOMTEST,
  TOTALLYRANDOMTEST,
  DEBUGTEST,
  EXIT_COMMAND,
  N_TESTS
};

void HandleCommandLine(int argc,
		       char **argv,
		       int &nTot,
		       int &nLoc,
		       int &group,
		       char ** pInfo);

extern int SpCounter;

int main(int argc, char **argv) {
  char InputOption[256];
  int Option = EXIT_COMMAND;
  int Nits = 0;
  int MsgLen = 0;
  int RandSeed = 0;
  int CommandLineTest = 0;

/*
** First, we start up the communications and make sure that there are no
** problems.
*/

  int nNode=-1, nLoc=-1, group=-1;
  char * pipeInfo = NULL;
  HandleCommandLine(argc, argv, nNode, nLoc, group, &pipeInfo);

  if (SpComm_StartUp(nNode, nLoc, group, pipeInfo) == ERROR_SP_COMM) {
    cout << "Error in starting up Speedes Communications - exiting" << endl;
    exit(1);
  }

  Node = SpComm_GetNodeId();
  Nnodes = SpComm_GetNumNodes();
  srand48(1-Node);

  if (CommandLineTest){
    TotallyRandomTests(-100000, -1000);
  }
/*
** Loop until the user enters the exit command.
*/


  




  while (1) {

/*
** Node 0 gets the test number from the user (the other nodes set the test to
** zero. All of the nodes find out what the test is by doing a global sum of
** the test number across nodes. The answer, of course, is the value that node
** 0 supplied. Then, we run the test.
*/

    if (Node == 0) {

      cout << endl;
      cout << "Commands" << endl;
      cout << "--------" << endl;
      cout << ALL_TESTS          <<"  --> Run all tests" << endl;
      cout << ALL_SYNC_TESTS     <<"  --> Run all Synching tests" << endl;
      cout << BARRIER_SYNC       <<"  --> Barrier Sync" << endl;
      cout << FUZZY_BARRIER      <<"  --> Fuzzy Barrier" << endl;
      cout << INTEGER_GLOBAL_SUM <<"  --> Integer Global Sum" << endl;
      cout << INTEGER_GLOBAL_MIN <<"  --> Integer Global Minimum" << endl;
      cout << INTEGER_GLOBAL_MAX <<"  --> Integer Global Maximum" << endl;
      cout << DOUBLE_GLOBAL_SUM  <<"  --> Double Global Sum" << endl;
      cout << DOUBLE_GLOBAL_MIN  <<"  --> Double Global Minimum" << endl;
      cout << DOUBLE_GLOBAL_MAX  <<" --> Double Global Maximum" << endl;
      cout << SIMTIME_GLOBAL_MIN <<" --> SimTime Global Minimum" << endl;
      cout << SIMTIME_GLOBAL_MAX <<" --> SimTime Global Maximum" << endl;
      cout << SEND_RECEIVE_TEST  <<" --> Send() Receive() test" << endl;
      cout << SEND_SYNC_TEST     <<" --> Syncronous send/receive test" << endl;
      cout << MULTISEND          <<" --> Multi-send test" << endl;
      cout << BROADCAST          <<" --> Broadcast test" << endl;
      cout << BLOCKING_MULTISEND <<" --> Coordinated Multi-cast test" << endl;
      cout << BLOCKING_BROADCAST <<" --> Coordinated Broadcast test" << endl;
      cout << PING_TEST          <<" --> Latency test" << endl;
      cout << BANDWIDTH_TEST     <<" --> Bandwith test" << endl;
      cout << RANDOMTEST         <<" --> Random test loop" << endl;
      cout << TOTALLYRANDOMTEST  <<" --> Totally Random test loop" << endl;
      cout << DEBUGTEST          <<" --> debugging test" << endl;
      cout << EXIT_COMMAND       <<" --> Exit" << endl;
      cout << endl;

      //     cout << "Enter command [1-" << N_TESTS-1 << "] ";
      //     cout.flush();
      //  To fix Janeway bug, add endl to cout statement.
      cout << "Enter command [1-" << N_TESTS-1 << "] " << endl;
      cin >> InputOption;
      Option = atoi(InputOption);
      if ((Option != EXIT_COMMAND) &&
	  (Option > 0) &&
	  (Option < (N_TESTS-1))) {
	//  cout << "Enter number of iterations: ";
	//  cout.flush();
	//  To fix Janeway bug, add endl to cout statement.
	cout << "Enter number of iterations: "<<endl;
	cin >> InputOption;
	Nits = atoi(InputOption);
	switch (Option)
	  {
	  case ALL_TESTS:
	  case SEND_RECEIVE_TEST:
	  case SEND_SYNC_TEST:
	  case MULTISEND:
	  case BROADCAST:
	  case BLOCKING_MULTISEND:
	  case BLOCKING_BROADCAST:
	  case RANDOMTEST:
	  case TOTALLYRANDOMTEST:
	  case PING_TEST:
	  case BANDWIDTH_TEST:
	    cout<<" Enter the size of messages to be sent.  A negative"
		<<endl<<" number indicates sending random sizes with"
		  <<endl<<" the maximum size given as -value"<<endl;
	    cin>>InputOption;
	    MsgLen = atoi(InputOption);
	    break;
	  default:
	    break;
	  }
	  
	}else{
	  Nits = 0;
	}
    }else{
      
      MsgLen = 0;
      Option = 0;
      Nits = 0;
    }

/*
** Make sure everybody has the command and the number of iterations
*/

    Option = SpComm_GlobalSum(Option);
    Nits = SpComm_GlobalSum(Nits);
    MsgLen = SpComm_GlobalSum(MsgLen);
    srand48(RandSeed);
    RandSeed = (int) lrand48();
    

/*
** Run the test that was selected
*/

    switch(Option) {
      
    case 0:
      if (Node == 0) {
	cout << "Error, commands must be between [1-"
	     << N_TESTS-1 << "] and you entered a "<<Option << endl;
	break;
      }
      
    case ALL_TESTS:
      TestBarrierSync(Nits);
      TestFuzzyBarrier(Nits);
      TestGlobalSumInt(Nits);
      TestGlobalMinInt(Nits);
      TestGlobalMaxInt(Nits);
      TestGlobalSumDouble(Nits);
      TestGlobalMinDouble(Nits);
      TestGlobalMaxDouble(Nits);
      TestGlobalMinSimTime(Nits);
      TestGlobalMaxSimTime(Nits);
      SpComm_send_receive_single_messages(Nits, MsgLen);
      SpComm_send_Sync(Nits, MsgLen);
      TestMultiSend(Nits, MsgLen);
      TestBroadCast(Nits, MsgLen);
      TestCoordinatedMultiSend(Nits, MsgLen);
      TestCoordinatedBroadCast(Nits, MsgLen);
      PingTest(Nits, MsgLen);
      //BandwidthTest(Nits, MsgLen);
      break;
      
    case ALL_SYNC_TESTS:
      TestBarrierSync(Nits);
      TestFuzzyBarrier(Nits);
      TestGlobalSumInt(Nits);
      TestGlobalMinInt(Nits);
      TestGlobalMaxInt(Nits);
      TestGlobalSumDouble(Nits);
      TestGlobalMinDouble(Nits);
      TestGlobalMaxDouble(Nits);
      TestGlobalMinSimTime(Nits);
      TestGlobalMaxSimTime(Nits);
      break;
      
    case BARRIER_SYNC:
      TestBarrierSync(Nits);
      break;
      
    case FUZZY_BARRIER:
      TestFuzzyBarrier(Nits);
      break;
      
    case INTEGER_GLOBAL_SUM:
      TestGlobalSumInt(Nits);
      break;
      
    case INTEGER_GLOBAL_MIN:
      TestGlobalMinInt(Nits);
      break;
      
    case INTEGER_GLOBAL_MAX:
      TestGlobalMaxInt(Nits);
      break;
      
    case DOUBLE_GLOBAL_SUM:
      TestGlobalSumDouble(Nits);
      break;
      
    case DOUBLE_GLOBAL_MIN:
      TestGlobalMinDouble(Nits);
      break;
      
    case DOUBLE_GLOBAL_MAX:
      TestGlobalMaxDouble(Nits);
      break;
      
    case SIMTIME_GLOBAL_MIN:
      TestGlobalMinSimTime(Nits);
      break;
      
    case SIMTIME_GLOBAL_MAX:
      TestGlobalMaxSimTime(Nits);
      break;
      
    case SEND_RECEIVE_TEST:
      SpComm_send_receive_single_messages(Nits, MsgLen);
      break;

    case SEND_SYNC_TEST:
      SpComm_send_Sync(Nits, MsgLen);
      break;

    case MULTISEND:
      TestMultiSend(Nits, MsgLen);
      break;

    case BROADCAST:
      TestBroadCast(Nits, MsgLen);
      break;

    case BLOCKING_MULTISEND:
      TestCoordinatedMultiSend(Nits, MsgLen);
      break;

    case BLOCKING_BROADCAST:
      TestCoordinatedBroadCast(Nits, MsgLen);
      break;

    case PING_TEST:
      PingTest(Nits, MsgLen);
      break;

    case BANDWIDTH_TEST:
      BandwidthTest(Nits, MsgLen);
      break;

    case RANDOMTEST:
      RandomTests(Nits, MsgLen);
      break;

    case TOTALLYRANDOMTEST:
      TotallyRandomTests(Nits, MsgLen);
      break;

    case EXIT_COMMAND:
      MPI_Finalize();
	exit(1);

    case DEBUGTEST:
      debugtest(Nits);
      break;
	
    default:
      if (Node == 0) {
	  cout << "Error, commands must be between [1-"
	       << N_TESTS-1 << "] and you entered a "<<Option << endl;
      }
    }
  }
}



/*
** Test SpComm_BarrierSync
*/

double TestBarrierSync(int Nits) {
  double time;
  int i;
  STOPWATCH StopWatch("Barrier Sync");

  StopWatch.StartTimer();
  for (i=0; i<Nits; i++) {
    SpComm_BarrierSync();
  }
  StopWatch.StopTimer();

  time = StopWatch.get_CumulativeTime() / Nits;
  if (Node == 0) {
    cout << "Time for SpComm_BarrierSync " << time << endl;
  }
  SpComm_BarrierSync();
  return time;
}


/*
** Test SpComm_EnterFuzzyBarrier and SpComm_ExitFuzzyBarrier
*/

double TestFuzzyBarrier(int Nits) {
  double time;
  int i;
  STOPWATCH StopWatch("Barrier Sync");

  StopWatch.StartTimer();
  for (i=0; i<Nits; i++) {
    SpComm_EnterFuzzyBarrier();
    while (1) {
      if (SpComm_ExitFuzzyBarrier()) break;
    }
  }
  StopWatch.StopTimer();

  time = StopWatch.get_CumulativeTime() / Nits;
  if (Node == 0) {
    cout << "Time for SpComm_FuzzyBarrier " << time << endl;
  }
  SpComm_BarrierSync();
  return time;
}

/*
** Test SpComm_GlobalSum for integers
*/

double TestGlobalSumInt(int Nits) {
  double time;
  int i;
  int TotNodes = SpComm_GetNumNodes();
  int Expected = (TotNodes-1)*TotNodes/2;
  int Result = -1;
  STOPWATCH StopWatch("GlobalSum(int)");

  StopWatch.StartTimer();
  for (i=0; i<Nits; i++) {
    Result = SpComm_GlobalSum(Node);
    if (Result != Expected) {
      cerr << "Error in GlobalSumInt: Got " << Result
	   << ", Expected " << Expected << endl<<"!!!!!!!!!!!"<<SpCounter<< endl;
	abort();
    }
  }
  StopWatch.StopTimer();

  time = StopWatch.get_CumulativeTime() / Nits;
  if (Node == 0) {
    cout << "Time for SpComm_GlobalSum(int) " << time << endl;
  }
  SpComm_BarrierSync();
  return time;
}

/*
** Test SpComm_GlobalMin for integers
*/

double TestGlobalMinInt(int Nits) {
  double time;
  int i;
  int Expected = 12345;
  int Result = -1;
  STOPWATCH StopWatch("GlobalMin(int)");

  StopWatch.StartTimer();
  for (i=0; i<Nits; i++) {
    Result = SpComm_GlobalMin(Node+12345);
    if (Result != Expected) {
      cerr << "Error in GlobalMinInt: Got " << Result
	   << ", Expected " << Expected <<endl<<"!!!!!!!!!!!"<<SpCounter << endl;
	abort();
    }
  }
  StopWatch.StopTimer();

  time = StopWatch.get_CumulativeTime() / Nits;
  if (Node == 0) {
    cout << "Time for SpComm_GlobalMin(int) " << time << endl;
  }
  SpComm_BarrierSync();
  return time;
}

/*
** Test SpComm_GlobalMin for integers
*/

double TestGlobalMaxInt(int Nits) {
  double time;
  int i;
  int Expected = 12345;
  int Result = -1;
  STOPWATCH StopWatch("GlobalMax(int)");

  StopWatch.StartTimer();
  for (i=0; i<Nits; i++) {
    Result = SpComm_GlobalMax(12345-Node);
    if (Result != Expected) {
      cerr << "Error in GlobalMaxInt: Got " << Result
	   << ", Expected " << Expected <<endl<<"!!!!!!!!!!!"<<SpCounter<< endl;
	abort();
    }
  }
  StopWatch.StopTimer();

  time = StopWatch.get_CumulativeTime() / Nits;
  if (Node == 0) {
    cout << "Time for SpComm_GlobalMax(int) " << time << endl;
  }
  SpComm_BarrierSync();
  return time;
}

/*
** Test SpComm_GlobalSum for doubles
*/

double TestGlobalSumDouble(int Nits) {
  double time;
  int i;
  double TotNodes = SpComm_GetNumNodes();
  double Expected = (TotNodes-1)*TotNodes/2;
  double Result = -1;
  STOPWATCH StopWatch("GlobalSum(double)");

  StopWatch.StartTimer();
  for (i=0; i<Nits; i++) {
    Result = SpComm_GlobalSum(double(Node));
    if (Result != Expected) {
      cerr << "Error in GlobalSumDouble: Got " << Result
	   << ", Expected " << Expected <<endl<<"!!!!!!!!!!!"<<SpCounter<< endl;
	abort();
    }
  }
  StopWatch.StopTimer();

  time = StopWatch.get_CumulativeTime() / Nits;
  if (Node == 0) {
    cout << "Time for SpComm_GlobalSum(double) " << time << endl;
  }
  SpComm_BarrierSync();
  return time;
}

/*
** Test SpComm_GlobalMin for doubles
*/

double TestGlobalMinDouble(int Nits) {
  double time;
  int i;
  double Expected = 12345.5;
  double Result = -1;
  STOPWATCH StopWatch("GlobalMin(double)");

  StopWatch.StartTimer();
  for (i=0; i<Nits; i++) {
    Result = SpComm_GlobalMin(double(Node)+12345.5);
    if (Result != Expected) {
      cerr << "Error in GlobalMinDouble: Got " << Result
	   << ", Expected " << Expected <<endl<<"!!!!!!!!!!!"<<SpCounter<< endl;
	abort();
    }
  }
  StopWatch.StopTimer();

  time = StopWatch.get_CumulativeTime() / Nits;
  if (Node == 0) {
    cout << "Time for SpComm_GlobalMin(double) " << time << endl;
  }
  SpComm_BarrierSync();
  return time;
}

/*
** Test SpComm_GlobalMin for doubles
*/

double TestGlobalMaxDouble(int Nits) {
  double time;
  int i;
  double Expected = 12345.5;
  double Result = -1;
  STOPWATCH StopWatch("GlobalMax(double)");

  StopWatch.StartTimer();
  for (i=0; i<Nits; i++) {
    Result = SpComm_GlobalMax(12345.5-double(Node));
    if (Result != Expected) {
      cerr << "Error in GlobalMaxDouble: Got " << Result
	   << ", Expected " << Expected <<endl<<"!!!!!!!!!!!"<<SpCounter<< endl;
	abort();
    }
  }
  StopWatch.StopTimer();

  time = StopWatch.get_CumulativeTime() / Nits;
  if (Node == 0) {
    cout << "Time for SpComm_GlobalMax(double) " << time << endl;
  }
  SpComm_BarrierSync();
  return time;
}

/*
** Test SpComm_GlobalMin for SIMTIME
*/

double TestGlobalMinSimTime(int Nits) {
  double time;
  int i;
  SIMTIME SimTime;
  SIMTIME Expected;
  SIMTIME Result;
  STOPWATCH StopWatch("GlobalMin(SIMTIME)");

  SimTime = 12345.5;
  SimTime.SetPriority1(123);
  SimTime.SetPriority2(Node+1000);

  Expected = SimTime;
  Expected.SetPriority2(1000);

  StopWatch.StartTimer();
  for (i=0; i<Nits; i++) {
    Result = SpComm_GlobalMin(SimTime);
    if (Result != Expected) {
      cerr << "Error in GlobalMinSimTime: Got " << Result
	   << ", Expected " << Expected <<endl<<"!!!!!!!!!!!"<<SpCounter<< endl;
	abort();
    }
  }
  StopWatch.StopTimer();

  time = StopWatch.get_CumulativeTime() / Nits;
  if (Node == 0) {
    cout << "Time for SpComm_GlobalMin(SIMTIME) " << time << endl;
  }
  SpComm_BarrierSync();
  return time;
}


/*
** Test SpComm_GlobalMin for SIMTIME
*/

double TestGlobalMaxSimTime(int Nits) {
  double time;
  int i;
  SIMTIME SimTime;
  SIMTIME Expected;
  SIMTIME Result;
  STOPWATCH StopWatch("GlobalMax(SIMTIME)");

  SimTime = 12345.5;
  SimTime.SetPriority1(123);
  SimTime.SetPriority2(1000 - Node);

  Expected = SimTime;
  Expected.SetPriority2(1000);

  StopWatch.StartTimer();
  for (i=0; i<Nits; i++) {
    Result = SpComm_GlobalMax(SimTime);
    if (Result != Expected) {
      cerr << "Error in GlobalMaxSimTime: Got " <<endl << Result
	   << ", Expected " <<endl << Expected <<endl<<"!!!!!!!!!!!"<<SpCounter<< endl;
	abort();
    }
  }
  StopWatch.StopTimer();

  time = StopWatch.get_CumulativeTime() / Nits;
  if (Node == 0) {
    cout << "Time for SpComm_GlobalMax(SIMTIME) " << time << endl;
  }
  SpComm_BarrierSync();
  return time;
}




int GetMessageLen(int MsgLen)
{
  if (MsgLen >= 0) return MsgLen;
  else return Rand(-MsgLen);
}




int ProcessMessages()
{
  int MsgsReceived = 0;
  int i, nbytes;
  unsigned char *InMsg;
  int old_MsgsReceived = 1;

  while (old_MsgsReceived != MsgsReceived){
    old_MsgsReceived = MsgsReceived;
    for (i=0; i < N_MESSAGE_TYPES; ++i){
      while ((InMsg = (unsigned char *) SpComm_GetPendingMessage(i, nbytes))
	     != NULL){
	if (nbytes && (InMsg[0] != InMsg[nbytes-1])){
	  cerr<<"NonBlocking Message Error"<<" node "<<SpComm_GetNodeId()
	      <<" Type "<<i
	      <<" Nbytes = "<<nbytes<<" InMsg[0] = "<<(int)InMsg[0]
	      <<" InMsg[nbytes-1] = "<<(int)InMsg[nbytes-1]<<endl;
	  if (CommandLineTest){
	    abort();
	  }
	}
	else ++MsgsReceived;
	/*
	  if (nbytes){
	  cout<<SpComm_GetNodeId()<<" Receiving type="<<i
	  <<" nbytes="<<nbytes
	  <<" payload="<<(int)InMsg[0]<<", "<<(int)InMsg[nbytes-1]<<endl;
	  }
	  else{
	  cout<<SpComm_GetNodeId()<<" Receiving type="<<i
	  <<" nbytes="<<nbytes
	  <<" payload="<<(int)InMsg[0]<<endl;
	  }
	*/
	delete [] InMsg;
      }
    }
    SpComm_Receive();
  }
  return MsgsReceived;
}



void ProcessSyncMessages()
{
  int nbytes;
  char *InMsg;

  while ((InMsg = (char *) SpComm_CoordinatedReceive(nbytes)) != NULL){
    if ((InMsg[0] != InMsg[nbytes-1]) && nbytes){
      cerr<<"Coordinated Message Error "<<SpComm_GetNodeId()
	  <<" InMsg[0] = "<<(int)InMsg[0]<<" InMsg[nbytes-1] = "
	  <<(int)InMsg[nbytes-1]<<endl;
      if (CommandLineTest){
	abort();
      }
    }
    delete [] InMsg;
  }
  SpComm_Receive();
}


double SpComm_send_receive_single_messages(int Nits, int MaxMsgLen)
{
  int MLen = (MaxMsgLen >= 0) ? MaxMsgLen : -MaxMsgLen;
  char *msg = new char[(MLen)];
  int Receiver;
  int Nnodes = SpComm_GetNumNodes();
  int CountInOut = 0;
  int i;
  STOPWATCH StopWatch("SpComm_Send() and SpComm_Receive()");
  double time;
  int Type;
  int TotBytesSent = 0;
  int TotBytes = 128*Nits;
  int MsgLen;

  StopWatch.StartTimer();
  for (i=0; i < MLen; msg[i++]=127);
  while (TotBytesSent != TotBytes){
    MsgLen = GetMessageLen(MaxMsgLen);
    if (TotBytesSent + MsgLen > TotBytes){
      MsgLen = TotBytes - TotBytesSent;
    }
    TotBytesSent += MsgLen;
    msg[0] = Receiver = Rand(Nnodes);
    if (MsgLen){
      msg[MsgLen-1] = Receiver;
    }
    Type = Rand(N_MESSAGE_TYPES);
    ++CountInOut;
    SpComm_Send(Receiver, Type,  MsgLen, msg);
    if (!(i%10))SpComm_Receive();// We want to receive after every 10 msgs sent
    //Don't let too many incoming messages stack up...
    if (!Rand(100)){
      CountInOut -= ProcessMessages();
    }
    if (MsgLen){
      msg[MsgLen-1] = 127;
    }
  }
  SpComm_EnterFuzzyBarrier();
  while (!SpComm_ExitFuzzyBarrier());
  while (SpComm_GlobalSum(CountInOut)){
    SpComm_Receive();
    CountInOut -= ProcessMessages();
  }
  StopWatch.StopTimer();
  
  time = StopWatch.get_CumulativeTime() / Nits / Nnodes;
  if (Node == 0)
    cout << "Time for SpComm_send_receive_single_message() " << time << endl;
  delete [] msg;
  return time;
}



double SpComm_send_Sync(int Nits, int MaxMsgLen)
{
  int MLen = (MaxMsgLen >= 0) ? MaxMsgLen : -MaxMsgLen;
  char *msg = new char[MLen];
  int Receiver;
  int Nnodes = SpComm_GetNumNodes();
  int i;
  STOPWATCH StopWatch("SpComm_Send() and SpComm_Receive()");
  double time;
  int TotBytesSent = 0;
  int TotBytes = 128*Nits;
  int MsgLen;
  
  StopWatch.StartTimer();
  for (i=0; i < MLen; msg[i++]=127);

  i = 0;
  while (TotBytesSent != TotBytes){
    ++i;
    MsgLen = GetMessageLen(MaxMsgLen);
    if (TotBytesSent + MsgLen > TotBytes){
      MsgLen = TotBytes - TotBytesSent;
    }
    TotBytesSent += MsgLen;
    msg[0] = Receiver = Rand(Nnodes);
    if (MsgLen){
      msg[MsgLen-1] = Receiver;
    }
    SpComm_CoordinatedSend(Receiver, MsgLen, msg);
    // We want to receive after every 50 msgs sent
    if (!(i%10)) SpComm_Receive();
    if (!(i%50)) ProcessSyncMessages();
    if (MsgLen){
      msg[MsgLen-1] = 127;
    }
  }
  ProcessSyncMessages();
  StopWatch.StopTimer();

  time = StopWatch.get_CumulativeTime() / Nits / Nnodes;
  if (Node == 0) {
    cout << "Time for SpComm_send_sync_message() " << time << endl;
  }
  
  delete [] msg;
  return time;
}


double TestMultiSend(int Nits, int MsgLen)
{
  int MLen = (MsgLen >= 0) ? MsgLen : -MsgLen;
  char *msg = new char[MLen];
  int Nnodes = SpComm_GetNumNodes();
  int Mynode = SpComm_GetNodeId();
  int i;
  STOPWATCH StopWatch("SpComm_Send() and SpComm_Receive()");
  double time;
  DESTINATION *Dest;
  int TotalUnSent = Nits-1;
  int CountInOut = 0;
  int Type = 0;

  StopWatch.StartTimer();
  Dest = new DESTINATION(Nnodes);
  for (i=0; i < MLen; msg[i++]=127);
  while (TotalUnSent){
    Dest->Reset();
    int j;
    MLen = GetMessageLen(MsgLen);
    //This actually favors sending to the lower numbered nodes slightly...
    for (j=0; (j < Nnodes) && TotalUnSent; ++j){
      if (Rand(10)>1){ //Give about a 80% chance of sending to that node
	Dest->Set(j);
	--TotalUnSent;
	++CountInOut;
      }
    }
    Type = Rand(254);
    msg[0] = msg[MLen-1] = j;
    msg[1] = Mynode;
    SpComm_Send(Dest, Type, MLen, msg);
    if (!Rand(8)) SpComm_Receive();
    if (!Rand(100)) CountInOut -= ProcessMessages();
  }
  CountInOut -= ProcessMessages();

  delete Dest;
  SpComm_EnterFuzzyBarrier();
  while (!SpComm_ExitFuzzyBarrier());
  while (SpComm_GlobalSum(CountInOut)){
    CountInOut -= ProcessMessages();
  }
  StopWatch.StopTimer();
  
  time = StopWatch.get_CumulativeTime() / Nits / Nnodes;
  if (Node == 0) {
    cout << "Time for Multi-send/receive() " << time << endl;
  }
  SpComm_BarrierSync();
  delete [] msg;
  return time;
}




double TestBroadCast(int Nits, int MaxMsgLen)
{
  int MLen = (MaxMsgLen >= 0) ? MaxMsgLen : -MaxMsgLen;
  unsigned char * msg = new unsigned char[MLen];
  int Nnodes = SpComm_GetNumNodes();
  int Mynode = SpComm_GetNodeId();
  STOPWATCH StopWatch("SpComm_Send() and SpComm_Receive()");
  double time;
  int CountInOut = 0;
  int BroadCastCount = Nits;
  int BroadCastThisTime = 0;
  int MsgLen;
  int Type;
  int i = 0;

  StopWatch.StartTimer();
  for (int j=0; j < MLen; msg[j++]=127);
  while (BroadCastCount){
    ++i;
    MsgLen = GetMessageLen(MaxMsgLen);
    if (Rand(Nnodes) == 0){
      --BroadCastCount;
      CountInOut += (Nnodes - 1);
      BroadCastThisTime = 1;
    }
    msg[0] = Mynode;
    if (MsgLen){
      msg[MsgLen-1] = Mynode;
    }
    if (BroadCastThisTime){
      Type = Rand(254);
      SpComm_Send(Type, MsgLen, msg);
      BroadCastThisTime = 0;
    }
    if (!Rand(8)){
      SpComm_Receive();
    }
    if (!Rand(100)){
      CountInOut -= ProcessMessages();
    }
    if (MsgLen){
      msg[MsgLen-1] = 127;
    }
  }
  CountInOut -= ProcessMessages();

  SpComm_EnterFuzzyBarrier();
  while (!SpComm_ExitFuzzyBarrier());
  while (SpComm_GlobalSum(CountInOut)){
    CountInOut -= ProcessMessages();
   }
  StopWatch.StopTimer();
  
  time = StopWatch.get_CumulativeTime() / Nits / Nnodes;
  if (Node == 0) {
    cout << "Time for TestBroadCast() " << time << endl;
  }
  SpComm_BarrierSync();
  delete [] msg;
  return time;
}


double TestCoordinatedMultiSend(int Nits, int MsgLen)
{
  int MLen = (MsgLen >= 0) ? MsgLen : -MsgLen;
  char *msg  = new char[MLen];
  int Nnodes = SpComm_GetNumNodes();
  int Mynode = SpComm_GetNodeId();
  int i;
  STOPWATCH StopWatch("SpComm_Send() and SpComm_Receive()");
  double time;
  DESTINATION *Dest;
  int TotalUnSent = Nits-1;
  int ProcessMessageCount = Nits/Nnodes;

  StopWatch.StartTimer();
  Dest = new DESTINATION(Nnodes);
  for (i=0; i < MLen; msg[i++]=127);
  i = 0;
  while (TotalUnSent){
    Dest->Reset();
    MLen = GetMessageLen(MsgLen);
    for (int j=0; (j < Nnodes) && TotalUnSent; ++j)
      if (Rand(10)>1){ //Give about a 80% chance of sending to that node
	Dest->Set(j);
	--TotalUnSent;
      }
    
    msg[0] = msg[MLen-1] = Mynode;
    SpComm_CoordinatedSend(Dest, MLen, msg);
    if (ProcessMessageCount){
      ProcessSyncMessages();
      --ProcessMessageCount;
    }
  }
  delete Dest;
  while (ProcessMessageCount){
    ProcessSyncMessages();
    --ProcessMessageCount;
  }
  StopWatch.StopTimer();
  
  time = StopWatch.get_CumulativeTime() / Nits / Nnodes;
  if (Node == 0) {
    cout << "Time for Coordinated Multi-send/receive() " << time << endl;
  }
  SpComm_BarrierSync();
  delete [] msg;
  return time;
}




double TestCoordinatedBroadCast(int Nits, int MaxMsgLen)
{

  int MLen = (MaxMsgLen >= 0) ? MaxMsgLen : -MaxMsgLen;
  unsigned char * msg = new unsigned char[MLen];
  int Nnodes = SpComm_GetNumNodes();
  int Mynode = SpComm_GetNodeId();
  STOPWATCH StopWatch("SpComm_Send() and SpComm_Receive()");
  double time;
  int BroadCastCount = Nits;
  int BroadCastThisTime = 0;
  int MsgLen;
  int i = 0;
  int ProcessMessageCount = Nits/Nnodes;

  StopWatch.StartTimer();
  for (int j=0; j < MLen; msg[j++]=127);
  i = 0;
  while (BroadCastCount){
    ++i;
    MsgLen = GetMessageLen(MaxMsgLen);
    if (Rand(Nnodes) == 0){
      --BroadCastCount;
      BroadCastThisTime = 1;
    }
    msg[0] = Mynode;
    if (MsgLen){
      msg[MsgLen-1] = Mynode;
    }
    if (BroadCastThisTime){
      //      if(Mynode == 0)cout<< "i: "<<i<<" SpComm_CoordinatedSend."<<endl;
      SpComm_CoordinatedSend(MsgLen, msg);
      BroadCastThisTime = 0;
    }
    if (!(i%8)) SpComm_Receive();
    if (!(BroadCastCount%25) && ProcessMessageCount){
      // if(Mynode == 0)cout<<" ProcessSyncMsgs - 1"<<endl;
      ProcessSyncMessages();
      --ProcessMessageCount;
    }
    if (MsgLen){
      msg[MsgLen-1] = 127;
    }
  }
  while (ProcessMessageCount){
    //   if(Mynode == 0)cout<<" ProcessSyncMsgs - 2"<<endl;
    ProcessSyncMessages();
    --ProcessMessageCount;
  }
  StopWatch.StopTimer();
  
  time = StopWatch.get_CumulativeTime() / Nits / Nnodes;
  if (Node == 0) {
    cout << "Time for TestCoordinatedBroadCast() " << time << endl;
  }
  SpComm_BarrierSync();
  delete [] msg;
  return time;
}

///

double PingTest(int nits, int len)
{
  int i;
  char * msg = new char[len];
  double AvgTime = 0;
  STOPWATCH StopWatch("Ping Test");
  char * RecvdMsg;

  if ((Nnodes > 1) && ((Node == 1) || (Node == 0))){
    StopWatch.StartTimer();
    for (i=0; i < nits; ++i){
      int nbytes;
      if (!Node){
	while ((RecvdMsg = (char *)SpComm_Receive(0, nbytes)) == NULL);
	delete [] RecvdMsg;
	SpComm_Send(1, 0, len, msg);
      }
      else{
	SpComm_Send(0, 0, len, msg);
	while ((RecvdMsg = (char *)SpComm_Receive(0, nbytes)) == NULL);
	delete [] RecvdMsg;
      }
    }
    StopWatch.StopTimer();
    AvgTime = StopWatch.get_CumulativeTime()/nits/2;
  }
  delete [] msg;

  if (Node == 0){
    cout<<"Latency is "<<AvgTime<<" for a message of length "<<len<<endl;
  }
  SpComm_BarrierSync();

  return AvgTime;
}


double BandwidthTest(int nits, int len)
{
  int i;
  double AvgTime = 0;
  STOPWATCH StopWatch("Bandwidth Test");
  char * RecvdMsg;
  int RecvdCount = 0;
  int nbytes;
  char * msg = new char[len];

  if ((Nnodes > 1) && ((Node == 1) || (Node == 0))){
    StopWatch.StartTimer();
    if (Node){
      for (i=0; i < nits; ++i){
	SpComm_Send(0, 0, len, msg);
      }
      while ((RecvdMsg = (char *)SpComm_Receive(0, nbytes)) != NULL);
      delete [] RecvdMsg;
    }
    else{
      while (RecvdCount != nits){
	SpComm_Receive();
	while ((RecvdMsg = (char *)SpComm_Receive(0, nbytes)) != NULL){
	  ++RecvdCount;
	  delete [] RecvdMsg;
	}
	SpComm_Send(1, 0, 1, msg);//Send a 1 byte message back
      }
    }
  }
  
  delete [] msg;
  SpComm_BarrierSync();
  StopWatch.StopTimer();

  double BitsPerSecond=(double)len*(double)nits/StopWatch.get_CumulativeTime();

  if (Node == 0){
    cout<<"Throughput is "<<BitsPerSecond
	<<" bits/second for a message of length "
	<<len<<endl;
  }

  return AvgTime;
}












double debugtest(int /*nits*/)
{
  return 0.0;
}


void RandomTests(int Nits, int MsgLen)
{
  int Option, Ntests = N_TESTS-7;
  while(1){
    if (!SpComm_GetNodeId()) Option = 3+Rand(Ntests);
    else Option = 0;
    Option = SpComm_GlobalSum(Option);
    switch(Option) {
    case BARRIER_SYNC:
      if (!SpComm_LocalNode) cout<<"Barrier sync: ";
      TestBarrierSync(Nits);
      break;
    case FUZZY_BARRIER:
      if (!SpComm_LocalNode) cout<<"Fuzzy Barrier: ";
      TestFuzzyBarrier(Nits);
      break;
    case INTEGER_GLOBAL_SUM:
      if (!SpComm_LocalNode) cout<<"integer global sum: ";
      TestGlobalSumInt(Nits);
      break;
    case INTEGER_GLOBAL_MIN:
      if (!SpComm_LocalNode) cout<<"integer global min: ";
      TestGlobalMinInt(Nits);
      break;
    case INTEGER_GLOBAL_MAX:
      if (!SpComm_LocalNode) cout<<"integer global max: ";
      TestGlobalMaxInt(Nits);
      break;
    case DOUBLE_GLOBAL_SUM:
      if (!SpComm_LocalNode) cout<<"double global sum: ";
      TestGlobalSumDouble(Nits);
      break;
    case DOUBLE_GLOBAL_MIN:
      if (!SpComm_LocalNode) cout<<"double global min: ";
      TestGlobalMinDouble(Nits);
      break;
    case DOUBLE_GLOBAL_MAX:
      if (!SpComm_LocalNode) cout<<"double global max: ";
      TestGlobalMaxDouble(Nits);
      break;
    case SIMTIME_GLOBAL_MIN:
      if (!SpComm_LocalNode) cout<<"simtime global min: ";
      TestGlobalMinSimTime(Nits);
      break;
    case SIMTIME_GLOBAL_MAX:
      if (!SpComm_LocalNode) cout<<"simtime global max: ";
      TestGlobalMaxSimTime(Nits);
      break;
    case SEND_RECEIVE_TEST:
      if (!SpComm_LocalNode) cout<<"send/receive: ";
      SpComm_send_receive_single_messages(Nits, MsgLen);
      break;
    case SEND_SYNC_TEST:
      if (!SpComm_LocalNode) cout<<"send sync: ";
      SpComm_send_Sync(Nits, MsgLen);
      break;
    case MULTISEND:
      if (!SpComm_LocalNode) cout<<"multi send: ";
      TestMultiSend(Nits, MsgLen);
      break;
    case BROADCAST:
      if (!SpComm_LocalNode) cout<<"broadcast: ";
      TestBroadCast(Nits, MsgLen);
      break;
    case BLOCKING_MULTISEND:
      if (!SpComm_LocalNode) cout<<"blocking multi-send: ";
      TestCoordinatedMultiSend(Nits, MsgLen);
      break;
    case BLOCKING_BROADCAST:
      if (!SpComm_LocalNode) cout<<"blocking broadcast: ";
      TestCoordinatedBroadCast(Nits, MsgLen);
    default:
      if (!SpComm_LocalNode) cout<<"Option "<<Option<<" chosen"<<endl;
      break;
    }
    if (SpComm_GetNodeId()) if (!SpComm_LocalNode) cout<<endl;
  }
}

enum {
  RANDOM_BARRIER_SYNC = 0,
  RANDOM_FUZZY_BARRIER,
  RANDOM_INTEGER_GLOBAL_SUM,
  RANDOM_INTEGER_GLOBAL_MIN,
  RANDOM_INTEGER_GLOBAL_MAX,
  RANDOM_DOUBLE_GLOBAL_SUM,
  RANDOM_DOUBLE_GLOBAL_MIN,
  RANDOM_DOUBLE_GLOBAL_MAX,
  RANDOM_SIMTIME_GLOBAL_MIN,
  RANDOM_SIMTIME_GLOBAL_MAX,
  RANDOM_SEND,
  RANDOM_PING_TEST,
  RANDOM_BANDWIDTH_TEST,
  N_RANDOM_TESTS
};

enum {
  NB_UNICAST=0,
  NB_MULTICAST,
  NB_BROADCAST,
  SYNC_UNICAST,
  SYNC_MULTICAST,
  SYNC_BROADCAST,
  N_SEND
};



void SendTest(int Nits, int msgLen){
  int option;
  static int myNode = SpComm_GetNodeId();
  static int Nnodes = SpComm_GetNumNodes();
  int i, j;
  int nbMsgInOutCount = 0;
  int msgOutCount = 0;
  int maxMsgLen = (msgLen < 0)? -msgLen  : msgLen;
  char * msg = new char[maxMsgLen];
  static DESTINATION * dest = new DESTINATION(Nnodes);

  for (i = 0; i < maxMsgLen; ++i){
    msg[i] = 127;
  }
  while (msgOutCount < Nits){
    option = Rand(N_SEND);
    int mLen = GetMessageLen(msgLen);
    int recv = Rand(Nnodes);
    int type = Rand(N_MESSAGE_TYPES);
    int NumMultiCast = 0;
    if (mLen){
      msg[0] = msg[mLen-1] = recv;
    }
    switch(option)
      {
      case NB_UNICAST:
      case SYNC_UNICAST:
	++msgOutCount;
	nbMsgInOutCount += (option==NB_UNICAST) ? 1 : 0;
	if (option == NB_UNICAST){SpComm_Send(recv, type, mLen, msg);}
	else                     {SpComm_CoordinatedSend(recv, mLen, msg);}
	break;
      case NB_MULTICAST:
      case SYNC_MULTICAST:
	dest->Reset();
	for (j=0; j < Nnodes; ++j){
	  if (Rand(10)>1){ //Give about a 80% chance of sending to that node
	    dest->Set(j);
	    nbMsgInOutCount += (option==NB_MULTICAST) ? 1 : 0;
	    ++msgOutCount;
	    ++NumMultiCast;
	  }
	}
	if (option == NB_MULTICAST){SpComm_Send(dest, type, mLen, msg);}
	else                       {SpComm_CoordinatedSend(dest, mLen, msg);}
	break;
      case NB_BROADCAST:
      case SYNC_BROADCAST:
	msgOutCount += Nnodes;
	nbMsgInOutCount += (option==NB_BROADCAST) ? Nnodes - 1 : 0;
	if (option==NB_BROADCAST){SpComm_Send(type, mLen, msg);}
	else                     {SpComm_CoordinatedSend(mLen, msg);}
	break;
      }
    if (!Rand(10)){
      nbMsgInOutCount -= ProcessMessages();
    }
    if (mLen){
      msg[mLen-1] = 127;
    }
  }
  while (SpComm_GlobalSum(nbMsgInOutCount)){
    SpComm_Receive();
    nbMsgInOutCount -= ProcessMessages();
  }
  ProcessSyncMessages();
  delete [] msg;
}




void TotallyRandomTests(int Nits, int MsgLen)
{
  int Option;
  int MyNode = SpComm_GetNodeId();
  int RandSeed1 = 0;
  int RandSeed2 = MyNode;
  int absMsgLen = (MsgLen < 0) ? -MsgLen : MsgLen;
  unsigned int count = 0;
  int countLimit = -1;

  if (Nits < 0){
    countLimit = -Nits;
    Nits = 1;
  }
  while(1){
    if (countLimit == count++){
      exit(0);
    }
    srand48(RandSeed1); //Make sure we all select the same test
    Option = Rand(N_RANDOM_TESTS);
    RandSeed1 = (int) lrand48();
    srand48(RandSeed2); /* Each test should use a different seed
			 (particularly the send) */
    RandSeed2 = (int) lrand48();
    switch(Option) {
    case RANDOM_BARRIER_SYNC:
      if (!SpComm_LocalNode) cout<<"Barrier sync: ";
      cout.flush();
      TestBarrierSync(Nits);
      break;
    case RANDOM_FUZZY_BARRIER:
      if (!SpComm_LocalNode) cout<<"Fuzzy Barrier: ";
      cout.flush();
      TestFuzzyBarrier(Nits);
      break;
    case RANDOM_INTEGER_GLOBAL_SUM:
      if (!SpComm_LocalNode) cout<<"integer global sum: ";
      cout.flush();
      TestGlobalSumInt(Nits);
      break;
    case RANDOM_INTEGER_GLOBAL_MIN:
      if (!SpComm_LocalNode) cout<<"integer global min: ";
      cout.flush();
      TestGlobalMinInt(Nits);
      break;
    case RANDOM_INTEGER_GLOBAL_MAX:
      if (!SpComm_LocalNode) cout<<"integer global max: ";
      cout.flush();
      TestGlobalMaxInt(Nits);
      break;
    case RANDOM_DOUBLE_GLOBAL_SUM:
      if (!SpComm_LocalNode) cout<<"double global sum: ";
      cout.flush();
      TestGlobalSumDouble(Nits);
      break;
    case RANDOM_DOUBLE_GLOBAL_MIN:
      if (!SpComm_LocalNode) cout<<"double global min: ";
      cout.flush();
      TestGlobalMinDouble(Nits);
      break;
    case RANDOM_DOUBLE_GLOBAL_MAX:
      if (!SpComm_LocalNode) cout<<"double global max: ";
      cout.flush();
      TestGlobalMaxDouble(Nits);
      break;
    case RANDOM_SIMTIME_GLOBAL_MIN:
      if (!SpComm_LocalNode) cout<<"simtime global min: ";
      cout.flush();
      TestGlobalMinSimTime(Nits);
      break;
    case RANDOM_SIMTIME_GLOBAL_MAX:
      if (!SpComm_LocalNode) cout<<"simtime global max: ";
      cout.flush();
      TestGlobalMaxSimTime(Nits);
      break;
    case RANDOM_SEND:
      if (!SpComm_LocalNode) cout<<"Send test"<<endl;
      cout.flush();
      SendTest(6*Nits, MsgLen);
      break;
    case RANDOM_PING_TEST:
      if (!SpComm_LocalNode) cout<<"Ping test: ";
      cout.flush();
      PingTest(Nits, absMsgLen);
      break;
    case RANDOM_BANDWIDTH_TEST:
      if (!SpComm_LocalNode) cout<<"Bandwidth: ";
      cout.flush();
      //BandwidthTest(Nits, absMsgLen);
      break;
    default:
      cout<<"Chose option "<<Option<<endl;
      break;
    }
    if (!SpComm_LocalNode && SpComm_GetNodeId() && (Option != RANDOM_SEND)){
      cout<<endl;
    }
  }
}
