/* The basic MPI comm library.  IF messages must be sent to a different
   LocalBoxNumber, you might need to use the global node ID.  I have not 
   tested any of the network functions, although the stubs for these
   functions are in place, for compatibility with Metron's original design.  

   In the MPI systems I have available for testing, every MPI process 
   has a unique rank SpComm_LocalNode, and LocalBoxNumber = 0 for all 
   participating processes.  No messages are ever sent or received to
   or from a different LocalBoxNumber.

   If we migrate to a multiprocessor system where SpComm_LocalNode is
   local to the LocalBoxNumber -- then the network functions would
   have to be implemented.  In particular, it will be necessary to
   translate from the local process ID (i.e. SpComm_LocalNode) to a 
   global node ID as a function of the LocalBoxNumber prior to sending
   a message to a different box.

   If a message is received from outside LocalBoxNumber, the global node ID
   would have to be be translated to the local process ID (SpComm_LocalNode)
   as a function of the LocalBoxNumber of the recipient processor.   

   - Roger Hillson
   13 March 1998                                                           */

#include "SpStdlib.H"
#include <unistd.h>
#include <sys/wait.h>
#include <sys/shm.h>
#include "SpErrno.H"
#include "SpSignal.H"
#include "SpIostream.H"
#include "SpFstream.H"
#include "SpStrstream.H"
#include "SpIomanip.H"
#include <mpi.h>

#include "SpComm.H"
#include "SpCommNet.H"
#include "SpCommPar.H"
#include "MPI_SpStructs.H"
#include "MPI_SpGlobals.H"
#include "SpShMemQueue.H"

#define IMPORTANT_DEBUG_SP_COMM
#define SAFE_RECEIVE

int Startup = 0;

/*
** These variable correspond to node information.
*/

int *GlobalNodeIDs;           // Global and local node id's so that we
int *LocalNodeIDs;            // can reroute message appropriately
int *LocalNodeZeros;          /* This is an array of the GIDs of the local node
				 zero's */

int SpComm_Nboxes = 0;         /* The total number of physical computers.
                                  If one, there is ONE physical system.  
                                  Value is returned by
                                  SpComm_Pre_Fork_Net_Get_IDs().        */
int *NodeBoxNumber;            /* A map from GlobalNodeIDs to which physical
				  computer a node resides on. */
int LocalBoxNumber = 0;       // The current node's box number
int SpComm_Node = 0;          // Our global node (process) number
int SpComm_Nnodes = 1;        // The number of global nodes
int SpComm_Nlocal;            // The number of local nodes
int SpComm_LocalNode;         // Our local node number (i.e. MPI process rank)
int GroupId = 1;              // Our group ID
static int Parent = 1;        // 1 if we are the parent local node, 0 o/w

//  Begin MPI global block for SpeedesCommMPI.C

int MPISize;  // global variable for total number of MPI processes
int MPIRank;  // global variable for MPI rank of local process

//  The key to recover the tag attibute from MPI_PTP_Comm
int TAG_COUNTER_ARRAY_KEY_PTP; 

int MsgInOutCount = 0;  //Used for the blocking sends

MPI_Datatype MPI_SIMTIME; /*  MPI datatype which will be created
                              for passing Speedes SpSimTime objects. 
                              This ptr is initialized by a call to
                              build_SIMTIME in StartUp().       */

MPI_Datatype MsgType; /* MsgType is a ptr initialized by a call
                         to build_SpeedesMsg() called in StartUp(). */
SpeedesMsg MsgMap;    // SpeedesMsg is typedefed in MPI_SpStructs.H
MPIst MPISimTime;     // MPIst      is typedefed in MPI_SpStructs.H

extern MPI_Group  World_Group, Make_Group;

//  Declare the three communicators required by Speedes
extern MPI_Comm   MPI_Fuzzy_Comm;  // MPI_Comm for fuzzy barrier operations
extern MPI_Comm   MPI_PTP_Comm;    // MPI_Comm for point-to-point operations
extern MPI_Comm   MPI_Sync_Comm;   // MPI_Comm for synchronous sends

//  Declare the buffer and buffer size for buffered sends and receives
//  The buffer is attached in  SpComm_StartUp()
extern char  MPI_Buffer[MPI_BUFFER_SIZE];
  
/*  This file contains the MPI functions for sending and receiving
    blocking, synchronous, and multicast messages.                    */

int MPI_MsgInOutCount = 0;  // Used for MPI blocking sends

/*  Declare the queue object for saving the MPI-received BUFFERED messages. */

BUFFER_QUEUE MPI_InQueue[N_MESSAGE_TYPES + MPI_MESSAGE_TYPES];

/*  in_mesg and out_mesg have to be global variables for  
    MPI_SpComm_EnterFuzzyBarrier() and MPI_SpComm_ExitFuzzyBarrier(). 
    Otherwise the send buffers go out of scope when MPI_nb_sync returns
    PRIOR to receiving the messages in MPI_nb_check(). */
 
int  in_mesg[MAX_PROC];
int out_mesg[MAX_PROC];

/*  Likewise send_request and recv_request arrays must be global.
    The sends and receives are split between MPI_SpComm_EnterFuzzyBarrier()
    and MPI_SpComm_ExitFuzzyBarrier(), and the send and receive requests 
    must be maintained globally between function calls.  It seems prudent,
    if not absolutely necessary, to make the status arrays global 
    as well.  These could also be cached as communicator attributes
    but posting the arrays globally might be a bit faster.*/

MPI_Request send_request[MAX_PROC]; 
MPI_Request recv_request[MAX_PROC];
MPI_Status   send_status[MAX_PROC];
MPI_Status   recv_status[MAX_PROC]; 




/*  Prototype for build_SIMTIME, which constructs an MPI data type for
    passing objects of class SpSimTime.
                                */

void build_SIMTIME(MPIst *simtime, \
		   MPI_Datatype *ptr_simtime);
void build_SpeedesMsg(SpeedesMsg *mesg_map, \
                   MPI_Datatype *ptr_message);

// User-defined minimization functions for finding global SpSimTime minimum
void SimMinFun(void *in, void *inout, int *len, MPI_Datatype *datatype);
void SimMaxFun(void *in, void *inout, int *len, MPI_Datatype *datatype);

void create_tag_array(MPI_Comm* communicator, /* In */ \
		      int* TAG_COUNTER_KEY   /* Out */ );
 



//////////////////////////////////////////////////////////////////////

/* How to look up a local node ID from a global node ID */

#define LocalNodeID(GlobalNodeID) (LocalNodeIDs[GlobalNodeID])
#define IsLocalID(GlobalNodeID) (NodeBoxNumber[GlobalNodeID] == LocalBoxNumber)
#define LocalNodeZero(GlobalNodeID) (LocalNodeZeros[GlobalNodeID])

int GlobalNodeID(int LocalNodeID) {

  int i;
  for(i=0; ((NodeBoxNumber[i] != LocalBoxNumber) &&
	    (LocalNodeIDs[i] != LocalNodeID)); ++i);
  return i;
}

//////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////
/*
** At exit function that cleans up the MPI functions
*/

void SpComm_AtExit() {
  //  This routine is called at exit.

  SpComm_Net_Terminate();
  if (!Parent) return;

  if (GlobalNodeIDs)  delete [] GlobalNodeIDs;
  if (LocalNodeIDs)   delete [] LocalNodeIDs;
  if (LocalNodeZeros) delete [] LocalNodeZeros;
}

void sigintfunc(int) {
  if (SpComm_Node == 0) cout << endl;
  exit(1);
}


int HandleCommandLine(int argc, char **argv){

  int temp;    // Temp int used for reading in the input.
  char s[256]; // A temporary string used to verify that the input is valid

  cout << "Starting Up the SPEEDES Communications" << endl;
  atexit(SpComm_AtExit);  //Register SpComm_AtExit
  signal(SIGINT, sigintfunc);

  /*
  ** Read in the command line inputs to determine:
  ** 	1.	Number of local nodes
  **	2.	Number of total nodes
  **	3.	Group Id for the application
  */

  temp = -1;
  if (argc >= 2) {
    sscanf(argv[1], "%d", &temp);
    sprintf(s,"%d", temp);
    if (strcmp(s,argv[1])) {
      cout<<"Error SpComm_StartUp bad command line input 1"<<endl;
      return ERROR_SP_COMM;
    }
    else if (temp <= 0){
      cout<<"Error SpComm_StartUp: Must have at least one computer running"
	  <<endl;
      return ERROR_SP_COMM;
    }
    else{
      SpComm_Nnodes = temp;
      SpComm_Nlocal = temp;
    }
  }

  temp = -1;
  if (argc >= 3) {
    sscanf(argv[2], "%d", &temp);
    sprintf(s,"%d", temp);
    if (strcmp(s,argv[2])) {
      cout<<"Error SpComm_StartUp: bad command line input 2"<<endl;
      return ERROR_SP_COMM;
    }else{
      SpComm_Nnodes = temp;
    }
  }

  if (SpComm_Nlocal > SpComm_Nnodes) {
    cout << "Command line error: local nodes " << SpComm_Nlocal
	 << " < total nodes " << SpComm_Nnodes << endl;
    return ERROR_SP_COMM;
  }

  temp = -1;
  if (argc >= 4) {
    sscanf(argv[3], "%d", &temp);
    sprintf(s,"%d", temp);
    if (strcmp(s,argv[3])) {
      cout<<"Error SpComm_StartUp: bad command line input 3"<<endl;
      return ERROR_SP_COMM;
    }else{
      GroupId = temp;
    }
  }

  cout << "Number of total nodes = " << SpComm_Nnodes << endl;
  cout << "Number of local nodes = " << SpComm_Nlocal << endl;
  cout << "Group Id = " << GroupId << endl;

  return GOOD_SP_COMM;
}

int SpComm_StartUp(int argc, char **argv) {
  int i, NumMPIProcs;

  //  Initialize MPI -- no MPI calls can preceed this statement -rh
  MPI_Init(&argc, &argv);

  MPI_Comm_rank(MPI_COMM_WORLD,&MPIRank);
  MPI_Comm_size(MPI_COMM_WORLD,&MPISize);


  /*  Assume that SpComm_LocalNode := MPIRank.  In effect, we are
      assuming that even if there are multiple boxes, there is only
      one ordering of the MPI nodes, and this ordering spans the boxes.   */

  SpComm_LocalNode = MPIRank;

  //  It will do 'initially' to assume that the rank 0 process always
  //  supports stdinput and stdoutput.  Production system should verify
  //  this assumption -- it's not part of the MPI standard.

  //  Who am I?  Only process 0 should read
  //  command line.
  if(SpComm_LocalNode == 0)
    {

      //  Read the command line

      int rc = HandleCommandLine(argc, argv);

      if (rc == ERROR_SP_COMM) return rc;


      //  Check to verify that the number of processes spawned by
      //  MPI correspond to the number of processes requested by
      //  the user!

      if(MPISize != SpComm_Nnodes)
	{
	  perror("Number of MPI processes does not equal number of user-requested processes!");
	  return ERROR_SP_COMM;
	} 
      if(SpComm_Nlocal < MPISize)
	{
	  cout <<"Warning: " << SpComm_Nlocal <<" local processes < "<< NumMPIProcs << " total MPI processes." << endl; 
	}
    } // matches 'if (SpCommLocal_Node == 0)'

#ifdef MPI_DEBUG
  cout << "begin parameter broadcast: " << SpComm_LocalNode << endl;
#endif

  //  Check to verify that the number of spawned process <= MAX_PROC
  if(MPISize > MAX_PROC)
    {
      cout<<" Fatal error:  Number of MPI Processes "<<MPISize<< \
	" exceeds MAX_PROC "<<MAX_PROC<<" ."<<endl;
      exit(1);
    }
 
  //  Broadcast SpComm_Nnodes, SpComm_Nlocal, and GroupId to everyone. -rh
  MPI_Bcast(&SpComm_Nnodes, 1, MPI_INT, 0, MPI_COMM_WORLD); 
  MPI_Bcast(&SpComm_Nlocal, 1, MPI_INT, 0, MPI_COMM_WORLD); 
  MPI_Bcast(&GroupId,       1, MPI_INT, 0, MPI_COMM_WORLD);

#ifdef MPI_DEBUG 
  cout << "end broadcast: " << SpComm_LocalNode << endl;
#endif

  /* Set up GIDs and LIDs */
  GlobalNodeIDs = new int[SpComm_Nnodes];
  LocalNodeIDs = new int[SpComm_Nnodes];
  for (i=0; i < SpComm_Nnodes; ++i){
    GlobalNodeIDs[i] = i;
    LocalNodeIDs[i] = -1; //Initially, have no routing
  }
  LocalBoxNumber = 0; //Needes to be changed according to comm...
  if (SpComm_Net_Initialize()) return ERROR_SP_COMM;

  if(SpComm_LocalNode != 0) Parent = 0;

  // This call is 'post fork' in the sense that it follows MPI_Init()
  SpComm_Post_Fork_Net_GetIDs(&LocalNodeIDs, SpComm_Nboxes, 
			      &LocalNodeZeros, &NodeBoxNumber, LocalBoxNumber);

  for (i = 0; i < SpComm_Nnodes; ++i){
    if ((LocalNodeIDs[i] == SpComm_LocalNode) &&
	(NodeBoxNumber[i] == LocalBoxNumber)){
      SpComm_Node = i;

    }
  }
  //  Attach the buffer for buffered MPI Sends
  MPI_Buffer_attach(MPI_Buffer, MPI_Buffer_Size);

  //  build the datatype required for sending Speedes Messages
  build_SpeedesMsg(&MsgMap, &MsgType);
              
  build_SIMTIME(&MPISimTime, &MPI_SIMTIME);

  //  Create communicators for different classes of communication
  //  MPI_Fuzzy_Comm => communicator for fuzzy barrier functions
  //  MPI_PTP_Comm   => communicator for point-to-point communications
  //  MPI_Sync_Comm  => communicator for synchronous communications 

  MPI_Comm_group( MPI_COMM_WORLD, &World_Group);
  MPI_Comm_create(MPI_COMM_WORLD, World_Group, &MPI_Fuzzy_Comm);
  MPI_Comm_create(MPI_COMM_WORLD, World_Group, &MPI_PTP_Comm);
  MPI_Comm_create(MPI_COMM_WORLD, World_Group, &MPI_Sync_Comm);

  //  Create and cache the control tag array for comm (MPI_PTP_Comm
  create_tag_array(&MPI_PTP_Comm, &TAG_COUNTER_ARRAY_KEY_PTP); 

  /*
  ** Set the Node and Nnodes values
  */

  Startup = 1;

  SpComm_BarrierSync();
  return GOOD_SP_COMM;

} // end StartUp

//////////////////////////////////////////////////////////////////////

int SpComm_GetNumNodes() {
  if (!SpComm_Nnodes) {
    cout<<"Error GetNumNodes, number of nodes not yet set in Startup - Exiting"
	<< endl;
    exit(1);
  }
  return SpComm_Nnodes;
}

//////////////////////////////////////////////////////////////////////

int SpComm_GetNodeId() {
  if (!Startup) {
    cout << "Error GetNodeId, SpComm_StartUp must be called first - Exiting"
	 << endl;
    exit(1);
  }
  return SpComm_Node;
}
//  Begin MPI insertion

//////////////////////////////////////////////////////////////////////

void SpComm_Send(int destination, int type, int nbytes, void *buff)
{

  //  Instantiate a struct for Speedes messages
  SpeedesMsg Message;
  int i, j, dest, flag;
  char *buffer_ptr = (char *) buff; // cast because *buff is type void
  int *tag_vec;

  //  Uncache the attribute pointer tag_vec.  tag_vec points to
  //  the counter array, which is on the heap.

  MPI_Attr_get(MPI_PTP_Comm, TAG_COUNTER_ARRAY_KEY_PTP, &tag_vec, &flag);
  if(flag == 0) 
    {perror("Error in recovering attribute tag_vec In SpComm_Send(1) ");
    exit(1);} 

  if((nbytes <=0)||(nbytes>MAX_MESG_SIZE))
    {perror("Message size is out of range!");
    exit(1);}

  for(i=0;i<nbytes;i++)
    Message.buffer[i] = buffer_ptr[i];
  Message.size=nbytes;
  Message.type=type;

  /*  Use a buffered send to send the message in char buff[] to destination.
      However, these messages are of variable length nbytes.  On the receive
      side we will test the message size prior to receiving.           */

  //MPI_Bsend(buff, nbytes, MPI_CHAR, destination, tag, MPI_PTP_Comm);
  MPI_Bsend(&Message, 1, MsgType, destination, \
	    *(tag_vec + destination), MPI_PTP_Comm);

  //  increment the destination counter
  *(tag_vec + destination) += 1;

}//  End MPI_SpComm_Send


//////////////////////////////////////////////////////////////////////

void SpComm_Send(SpDestination *dest, int type, int nbytes, void *buff){

  /* This function is a rewrite of Metron's original SpComm_Send
     function.  MPI_Bsend() replaces SpComm_Shmem_Send();  with a for 
     loop added to loop through all destinations.  It is difficult to use
     MPI_Bcast() here because this would require that all processes 
     compute a new communicator reflecting the broadcast group computed
     herein.  --rh                                                     */

  int Dest;
  static int **DEST;                // class defined in SpDestination.H
  char *buffer_ptr = (char *) buff; // cast because *buff is type void
  static int firsttime = 1;
  int i, j, flag;
  int PresBox;
  SpeedesMsg Message;  // declare the MPI struct for Speedes messages
  int *tag_vec;  // the attribute pointer for the tag counter

  //  Retrieve the attribute for the tag counter
  MPI_Attr_get(MPI_PTP_Comm, TAG_COUNTER_ARRAY_KEY_PTP, &tag_vec, &flag);
  if(flag == 0) 
    {perror("Error in recovering attribute tag_vec ");
    exit(1);} 

  if((nbytes <=0)||(nbytes>MAX_MESG_SIZE))
    {perror("Message size is out of range!");
    exit(1);}

  //  Build the MPI message
  for(i=0;i<nbytes;i++)
    Message.buffer[i] = buffer_ptr[i];
  Message.size=nbytes;
  Message.type=type;

  if (firsttime){
    DEST = new int *[SpComm_Nboxes];
    for (i=0; i < SpComm_Nboxes; i++){
      DEST[i] = new int[SpComm_Nnodes+1]; //should really be (#machines/box)+1
    }
    firsttime = 0;
  }
  for (i=0; i < SpComm_Nboxes; DEST[i++][0] = 0);
  Dest = dest->GetFirstNode();
  while (Dest != -1){
    PresBox=NodeBoxNumber[Dest];
    //  Count the destinations for Machine# 'PresBox' in DEST[PresBox][j=0] 
    //  Put address j in DEST[PresBox][j+1] where j = 0, 1, MPISIZE - 1
    DEST[PresBox][0] = DEST[PresBox][0] + 1;
    DEST[PresBox][DEST[PresBox][0]] = Dest;
    //cout<<"Dest: "<<Dest<<endl;
    Dest = dest->GetNextNode(Dest);
  }
  /* Now we have built up arrays of the destinations, send them on to
     their respective local node zero's for redistribution */

  //  SpComm_Nboxes should be set to ONE by SpComm_Pre_Fork(..) call
  for (i=0; i < SpComm_Nboxes; i++){
    //cout<<"DEST["<<i<<"][0] "<<DEST[i][0]<<endl;
    if ((DEST[i][0])){ // If we have messages for box number i
      if (i == LocalBoxNumber)  // If local box, use buffered MPI send
	{

	  for (j=0; j < DEST[i][0]; ++j)
	    { // cout << "index j "<< j <<" "<< DEST[i][j + 1] << endl;
	      // cout << "DEST[i][j+1] "<<DEST[i][j+1]<<endl;
	      //MPI_Bsend(buff, nbytes, MPI_CHAR, DEST[i][j + 1], type, MPI_PTP_Comm);}
	      MPI_Bsend(&Message, 1, MsgType, DEST[i][j + 1], \
			*(tag_vec + DEST[i][j + 1]), MPI_PTP_Comm);

	      //  increment the destination counter
	      *(tag_vec + DEST[i][j + 1]) += 1;}

	  /*  In DEST[i][j], i => machine#, j + 1 => destination.  Since this 
	      is multicast, neither the message or message type will change 
	      within the loop.  */

	} // end 'if (i == LocalBoxNumber)' 
      else
	SpComm_Net_Send(DEST[i], type, nbytes, buff);
    }
  }// matches 'for (i=0; i < SpComm_Nboxes; ++i)'  
}

//////////////////////////////////////////////////////////////////////

void SpComm_Send(int type, int nbytes, void *buff){

  /* This is a broadcast to everyone in MPI_PTP_Comm.  We do not
     have an array of destinations.  Don't send the message to 
     yourself at SpComm_LocalNode.                              */

  SpeedesMsg Message;

  int i, j, flag;
  char *buffer_ptr = (char *) buff; // cast because *buff is type void
  int box, dest;
  int *tag_vec;
  //cout<<"SpComm_Send: ## msg size "<<nbytes<<endl;
  //  Uncache the attribute pointer tag_vec.  tag_vec points to
  //  the counter array, which is on the heap.

  MPI_Attr_get(MPI_PTP_Comm, TAG_COUNTER_ARRAY_KEY_PTP, &tag_vec, &flag);
  if(flag == 0) 
    {perror("Error in recovering attribute tag_vec ");
    exit(1);} 

  for (box = 0; box < SpComm_Nboxes; box++){
    if(box == LocalBoxNumber){
      for (dest=0; dest < SpComm_Nlocal; ++dest) //  Broadcast to LOCAL nodes.
	{
	  if(dest != SpComm_LocalNode){   // Don't send to yourself

	    if((nbytes <=0)||(nbytes>MAX_MESG_SIZE))
	      {perror("Message size is out of range!");
	      exit(1);}

	    for(j=0;j<nbytes;++j)
	      Message.buffer[j] = buffer_ptr[j];
	    Message.size=nbytes;
	    Message.type=type;

	    //cout<<SpComm_LocalNode<<" dest "<<dest<<" tag "<< *(tag_vec + dest) <<" nbyte//s "<<nbytes<<endl;

	    MPI_Bsend(&Message, 1, MsgType, dest, *(tag_vec + dest), MPI_PTP_Comm);

	    //  increment the counter as a function of destination
	    *(tag_vec + dest) += 1;

	  } //if(dest != SpComm_LocalNode
	} // matches 'for (dest=0; dest < SpComm_Nlocal; ++dest)'
 
    } // end 'if (box == LocalBoxNumber)' 
    else
      /*  Send the message to each box other than your own.  */
      cout<<"Trying to broadcast to a box "<< box << endl;
    SpComm_Net_Send(type, nbytes, buff);
  } // matches 'for (box = 0; box < SpComm_Nboxes; i++)' 

 
}


//////////////////////////////////////////////////////////////////////

char *SpComm_Receive(int type, int &nbytes)
{
  //  Overloaded Receive() function for blocking (synching) sends
  //  Called from MPI_SpComm_BlockingReceive(int &nbytes)
  //  which is called by MPI_SpComm_send_Sync()

  SpComm_Receive();
  return SpComm_GetPendingMessage(type, nbytes);
}

//////////////////////////////////////////////////////////////////////

void SpComm_Receive()
{
  // Send and receive buffered and synching messages.
  // Do not modify without full regression testing.

  int i;

  SpeedesMsg Message;
  char* BufferPtr = &Message.buffer[0];
  int* tag_vec;

  int flag, AttrFlag;
  int type;  // SPEEDES Message type
  int len;   // length (bytes/chars) of incoming message

  MPI_Status FirstStatus, StatusProbe;

//  Uncache the tag_counter_array
  MPI_Attr_get(MPI_PTP_Comm, TAG_COUNTER_ARRAY_KEY_PTP, &tag_vec, &AttrFlag);
  if(AttrFlag == 0) 
    {perror("Error in recovering attribute pointer tag_vec ");
    exit(1);}

  while(1)
    {

      MPI_Iprobe(MPI_ANY_SOURCE, MPI_ANY_TAG, \
		 MPI_PTP_Comm, &flag, &StatusProbe);


      if(flag != 0)
	{// incoming message!

	  //cout<<SpComm_LocalNode<<" Probed: src "<<StatusProbe.MPI_SOURCE<<" //tag "<<StatusProbe.MPI_TAG<<" flag "<<flag<<endl;
	  //cout<<SpComm_LocalNode<<" Recv tag requested "<<*(tag_vec + MAX_PRO//C + StatusProbe.MPI_SOURCE)<<endl;

	  MPI_Recv(&Message, 1, MsgType,  StatusProbe.MPI_SOURCE,   \
		   *(tag_vec + MAX_PROC + StatusProbe.MPI_SOURCE ), \
		   MPI_PTP_Comm, &FirstStatus);
	  
	  *(tag_vec + MAX_PROC + StatusProbe.MPI_SOURCE) += 1;

	  //  Having received a message, add to the tail of the circular queue.
	  MPI_InQueue[Message.type].AddToTail((char *)&Message.buffer[0],\
					      Message.size, Message.type, SpComm_LocalNode);
   
	}// matches 'if(flag != 0)'
      else
	break;

    } // matches 'while(flag)'

} // End void MPI_SpComm_Receive()

//////////////////////////////////////////////////////////////////////

char *SpComm_GetPendingMessage(int type, int &nbytes)
{
  char * retval = NULL;
  int dest = SpComm_LocalNode;
  if (MPI_InQueue[type].Count())
    {
      retval = MPI_InQueue[type].Remove(nbytes, type, dest);
      //cout<<"GetPendingMessage: retval "<<retval<< "Type "<<type<<endl;
    }
  else  nbytes = 0;
  return retval;
};

//////////////////////////////////////////////////////////////////

extern int InFuzzyBarrier;

void SpComm_CoordinatedSend(int dest, int nbytes, void *buff){

  if (InFuzzyBarrier)
    cout<<"Coordinated send while in fuzzy barrier 4"<<endl;
  ++MsgInOutCount;
  SpComm_Send(dest, MPI_SPCOMM_MESSAGE_BLOCKING_SEND, nbytes, buff);
}

void SpComm_CoordinatedSend(SpDestination *dest, int nbytes, void *buff){

  if (InFuzzyBarrier)
    cout<<"Coordinated send while in fuzzy barrier 5"<<endl;
  int Node = dest->GetFirstNode();
  while (Node != -1){++MsgInOutCount; Node = dest->GetNextNode(Node);}
  SpComm_Send(dest, MPI_SPCOMM_MESSAGE_BLOCKING_SEND, nbytes, buff);
}

void SpComm_CoordinatedSend(int nbytes, void *buff){

  if (InFuzzyBarrier)
    cout<<"Coordinated send while in fuzzy barrier 6"<<endl;
  MsgInOutCount += (SpComm_GetNumNodes() - 1);
  SpComm_Send(MPI_SPCOMM_MESSAGE_BLOCKING_SEND, nbytes, buff);
}

#define NEW_BEHAVIOUR 0

char *SpComm_CoordinatedReceive(int &nbytes){

  char * RcvBuff = SpComm_GetPendingMessage(MPI_SPCOMM_MESSAGE_BLOCKING_SEND,
                                            nbytes);
  static int EnteredFuzzy = 0;
  static int ExitedFuzzy = 0;
  if (RcvBuff){
    --MsgInOutCount;
    return RcvBuff;
  }

#if (NEW_BEHAVIOUR)
  if (!EnteredFuzzy){
    SpComm_EnterFuzzyBarrier();
    EnteredFuzzy = 1;
  }
  if (!ExitedFuzzy){
    while (!SpComm_ExitFuzzyBarrier()){
      RcvBuff = SpComm_Receive(MPI_SPCOMM_MESSAGE_BLOCKING_SEND, nbytes);
      if (RcvBuff != NULL) {
        --MsgInOutCount;
        return RcvBuff; 
      }
    }
  }
  ExitedFuzzy = 1;
#endif
  /* Now we have exited fuzzy barrier.  This means everyone thinks
     they have received all of their messages.  Now we start checking
     if we are done.  */
  if (InFuzzyBarrier) cout<<"Coordinated messages while in fuzzy barrier 7"\
			  <<endl;
  while (SpComm_GlobalSum(MsgInOutCount)){
    if (InFuzzyBarrier)
      cout<<"Coordinated messsages while in fuzzy barrier 8"<<endl;
    RcvBuff = SpComm_Receive(MPI_SPCOMM_MESSAGE_BLOCKING_SEND, nbytes);
    if (RcvBuff != NULL) {
      --MsgInOutCount;
      return RcvBuff; 
    }
  }
  EnteredFuzzy = 0;
  ExitedFuzzy = 0;
  MsgInOutCount = 0;
  return NULL;
}


/*  This next section contains the MPI global reduction and fuzzy barrier
    functions.                                                   */

//////////////////////////////////////////////////////////////////////

void build_SIMTIME(MPIst *simtime, \
                   MPI_Datatype *ptr_simtime)
{
  /*  Build an MPI Datatype for the SPEEDES class SpSimTime  */

  /*  This procedure returns a ptr ptr_simtime to a new
      data type consisting of one double variable
      and two integer variables.  The input *simtime is a
      pointer to a variable of type struct (or class)
      defined in a main program.  Note that *ptr_simtime 
      only has to be defined ONCE, and can then be used 
      repeatedly.                                           */

  /*  The number of intrinsic elements in each "block" of
      the new type:  1 double, 1 int, 1 int.           */

  int block_lengths[3];

  /*  Displacement of each element from the start of the new 
      type.  */
  MPI_Aint displacements[3];

  /*  MPI data types for the successive elements.     */
  MPI_Datatype typelist[3];

  /*  Use for calculating displacements               */
  MPI_Aint start_address;
  MPI_Aint address;

/*  If one of the dataypes was a vector, it would have 
    multiple elements per block.  For 'triplet', there
    is only one element per block.  */

  block_lengths[0] = block_lengths[1]   
    = block_lengths[2] = 1;
  /*for(i=0; i<=2; i++)
    printf("block_length[%d] = %d\n",i,block_lengths[i]);*/

/*  Define the typelist */
  typelist[0] = MPI_DOUBLE;
  typelist[1] = MPI_INT;
  typelist[2] = MPI_INT;

  /*  Calculate the displacements                */

/*  I.    The displacement of ptr_src is 0     */
  displacements[0] = 0;

  /*  II.   Calculate ptr_src's starting address */
  MPI_Address(&simtime->Time, &start_address);

  /*  III.  Calculate the two remaining displacements
            by subtracting start_address         */
  MPI_Address(&simtime->Priority1, &address);
  displacements[1] = address - start_address;
  /*printf(" disp[1] %i \n",displacements[1]);*/  
  
  MPI_Address(&simtime->Priority2, &address);
  displacements[2] = address - start_address;
  /*printf(" disp[2] %i \n",displacements[2]);*/
  
  /*  Finally, we build and commit the derived dataype */
  MPI_Type_struct(3, block_lengths, displacements,
		  typelist, ptr_simtime);
  MPI_Type_commit(ptr_simtime);

} //  matches build_SIMTIME()


//////////////////////////////////////////////////////////////////////

void build_SpeedesMsg(SpeedesMsg *mesg, \
		      MPI_Datatype *ptr_mesg)
{
  /*  Build an MPI Datatype for SPEEDES message passing */

  /*  The number of intrinsic elements in each "block" of
      the new type:  char array[MAX_MESG_SIZE], 1 int, 1 int. */

  int block_lengths[3];

  /*  Displacement of each element from the start of the new 
      type.  */
  MPI_Aint displacements[3];

  /*  MPI data types for the successive elements.     */
  MPI_Datatype typelist[3];

  /*  Use for calculating displacements               */
  MPI_Aint start_address;
  MPI_Aint address;

  block_lengths[0] = MAX_MESG_SIZE;

  block_lengths[1] = block_lengths[2] = 1;
  /*for(i=0; i<=2; i++)
    printf("block_length[%d] = %d\n",i,block_lengths[i]);*/

/*  Define the typelist */
  typelist[0] = MPI_CHAR;
  typelist[1] = MPI_INT;
  typelist[2] = MPI_INT;

  /*  Calculate the displacements                */

/*  I.    The displacement of ptr_src is 0     */
  displacements[0] = 0;

  /*  II.   Calculate ptr_src's starting address */
  MPI_Address(&mesg->buffer, &start_address);

  /*  III.  Calculate the two remaining displacements
            by subtracting start_address         */
  MPI_Address(&mesg->size, &address);
  displacements[1] = address - start_address;
  /*printf(" disp[1] %i \n",displacements[1]);*/  
  
  MPI_Address(&mesg->type, &address);
  displacements[2] = address - start_address;
  /*printf(" disp[2] %i \n",displacements[2]);*/
  
  /*  Finally, we build and commit the derived dataype */
  MPI_Type_struct(3, block_lengths, displacements,
		  typelist, ptr_mesg);
  MPI_Type_commit(ptr_mesg);

} //  matches build_mesg()

//////////////////////////////////////////////////////////////////////

void SimMinFun(void *in, void *inout, int *len, MPI_Datatype *datatype)
{
  //  User-defined MPI function for finding minimum value of two SpSimTime 
  //  variables.  Operator must be used with MPI_Reduce or MPI_Allreduce.
  //  The '<' and '=' operators are overloaded SpSimTime operators.

  //  Operator can be applied to an array of length n, but in SPEEDES
  //  len would normally be 1.

  //  NB:  MPI_Op_create() requires that the user defined function has void
  //  pointers as arguments for the input and output vectors.  C++ requires
  //  explicit conversions from void pointers to other pointers.

  SpSimTime *invec, *inoutvec;
  invec    = (SpSimTime *) in;
  inoutvec = (SpSimTime *) inout;

  int i;
  int flag = 0;

  //invec->print();
  //inoutvec->print();

  for (i=0;  i < *len; i++) { 
    if(invec < inoutvec) inoutvec = invec;

    //  Increment the array pointers.

    invec++;
    inoutvec++;

  } // End For loop
} // End SimMinFun()

//////////////////////////////////////////////////////////////////////

void SimMaxFun(void *in, void *inout, int *len, MPI_Datatype *datatype)
{
  //  User-defined MPI function for finding minimum value of two SpSimTime 
  //  variables.  Operator must be used with MPI_Reduce or MPI_Allreduce.
  //  The '>' and '=' operators are overloaded SpSimTime operators.

  //  Operator can be applied to an array of length n, but in SPEEDES
  //  len would normally be 1.

  //  NB:  MPI_Op_create() requires that the user defined function has void
  //  pointers as arguments for the input and output vectors.  C++ requires
  //  explicit conversions from void pointers to other pointers.

  SpSimTime *invec, *inoutvec;
  invec    = (SpSimTime *) in;
  inoutvec = (SpSimTime *) inout;

  int i;
  int flag = 0;

  //invec->print();
  //inoutvec->print();

  for (i=0;  i < *len; i++) { 
    if(invec > inoutvec) inoutvec = invec;

    //  Increment the array pointers.

    invec++;
    inoutvec++;

  } // End For loop
} // End SimMaxFun()

//////////////////////////////////////////////////////////////////////
int InFuzzyBarrier = 0;

void SpComm_EnterFuzzyBarrier()
{
  /*  This is used with MPI_nb_check().  Each process which
      calls MPI_nb_sync makes a nonblocking send to process rank 0,
      including the rank 0 process itself.  MPI_nb_check() will
      check to see if all nproc sends have been received by process
      0, and if so, broadcast a return to all sending processes.  */  

  extern MPIRank;
  extern MPISize;
  int i, src;
  int tag  = 0;
  int dest = 0;    // all sends are to process 0


  /*  We could also use a blocking or buffered send here, and it may
      be useful to obtain the timings relative to using MPI_Isend.
      Both calls unblock when the data has been cleared from the 
      send buffer.  See Pacheco, p. 299.  Improvement will be greatest
      on a distributed memory machine.                           */


//cout<<"Entering MPI_SpComm_EnterFuzzyBarrier - rank: "<< rank << endl;

  dest = (MPIRank + 1)%MPISize;
  tag  = MPIRank;
  out_mesg[MPIRank] = 1;

  /*  Each process sends a message to root (MPIRank +1)%MPISize:  
    1 message per process.
    No one can write to the send buffer until the send is complete.
    Also, the send buffer must be retained when this function goes
    out of scope, because the receive process is completed in
    MPI_SpComm_ExitFuzzyBarrier() */

   
  MPI_Isend(&out_mesg[MPIRank], 1, MPI_INT, dest, tag, \
	    MPI_Fuzzy_Comm, &send_request[MPIRank]);
  //cout << "MPI_Isend complete: "<< rank << endl;

//  Post the nonblocking receives IF rank == 0


//  Initiate the Ireceives ...

  src  = MPIRank - 1;
  if(src < 0) src = MPISize - 1;
  tag  = src;

  MPI_Irecv(&in_mesg[MPIRank], 1, MPI_INT, src, tag, MPI_Fuzzy_Comm, &recv_request[MPIRank]);

  InFuzzyBarrier = 1;

  //cout <<"Zero rank MPI_Irecv complete: "<< rank << endl;

  //  We will test for completion of the sends and receives in the 
  //  MPI_SpComm_ExitFuzzyBarrier routine.

}//  END void  MPI_SpComm_EnterFuzzyBarrier()

//////////////////////////////////////////////////////////////////////

int SpComm_ExitFuzzyBarrier()
{
  /*  This function is used with (i.e. following) a call by the process
      to MPI_SpComm_EnterFuzzyBarrier().

      MPI_SpComm_ExitFuzzyBarrieIt returns a '0' when all processes have 
      called MPI_SpComm_EnterFuzzyBarrier() and a '1' otherwise.                                     */
  extern MPIRank;
  extern MPISize;
  int tag, flag;

  //cout << "About to call MPI_Test in MPI_SpComm_ExitFuzzyBarrier" << endl;

//MPI_Test returns a nonzero value (TRUE) if local send has been completed.
  MPI_Test(&send_request[MPIRank], &flag, &send_status[MPIRank]);
  if(!flag) return 0;  // Exit with FALSE (0) if send is not complete
  
//  If flag != 0, local send has completed.  This is not a
//  global indicator -- sends from other processes  may not have 
//  completed.

//cout << "Rank " << rank << " completed MPI_Test send_request in nb_check." << endl;

  MPI_Test(&recv_request[MPIRank], &flag, &recv_status[MPIRank]);
  if(!flag) return 0; 

//  If I'm here, message has been cleared from send buffer and 
//  received from neighbor.

  MPI_Barrier(MPI_Fuzzy_Comm);
  InFuzzyBarrier = 0;
  return 1;  //  synchronization complete!
  
}//  END int MPI_SpComm_ExitFuzzyBarrier()

///////////////////////////////////////////////////////////////

void SpComm_BarrierSync()
{
  //  Blocking barrier synchronization function
  MPI_Barrier(MPI_COMM_WORLD);
}
///////////////////////////////////////////////////////////////

//  Overload function SpComm_MPIGlobalSum(T value)

int SpComm_GlobalSum(int value)
{
  int result;
  //  Return the global sum for an integer variable.
  MPI_Allreduce(&value,&result,1,MPI_INT,MPI_SUM,MPI_COMM_WORLD);
  return result;
}
///////////////////////////////////////////////////////////////

double SpComm_GlobalSum(double value)
{
  double result;
  //  Return the global minimum for an integer variable.
  MPI_Allreduce(&value,&result,1,MPI_DOUBLE,MPI_SUM,MPI_COMM_WORLD);
  return result;
}

///////////////////////////////////////////////////////////////

//  Overload function SpComm_MPIGlobalMax(T value)

int SpComm_GlobalMax(int value)
{
  int result;
  //  Return the global maximum for an integer variable.
  MPI_Allreduce(&value,&result,1,MPI_INT,MPI_MAX,MPI_COMM_WORLD);
  return result;
}

//////////////////////////////////////////////////////////////////////

double SpComm_GlobalMax(double value)
{
  double result;
  //  Return the global maximum for an integer variable.
  MPI_Allreduce(&value,&result,1,MPI_DOUBLE,MPI_MAX,MPI_COMM_WORLD);
  return result;
}

//////////////////////////////////////////////////////////////////////
//  Overload function SpComm_MPIGlobalMin(T value)

int SpComm_GlobalMin(int value)
{
  int result;
  //  Return the global minimum for an integer variable.
  MPI_Allreduce(&value,&result,1,MPI_INT,MPI_MIN,MPI_COMM_WORLD);
  return result;
}

//////////////////////////////////////////////////////////////////////

double SpComm_GlobalMin(double value)
{
  double result;
  //  Return the global minimum for an integer variable.
  MPI_Allreduce(&value,&result,1,MPI_DOUBLE,MPI_MIN,MPI_COMM_WORLD);
  return result;
}

//  Define MPI functions for SpComm_MPI_GlobalMin(SimTime) and
//  SpComm_MPI_GlobalMax(SimTime)

//////////////////////////////////////////////////////////////////////

SpSimTime &SpComm_GlobalMin(SpSimTime &SimTime)
{
  /*  Find the global minimum for a SPEEDES variable of type
      SpSimTime.  SpSimTime is a class, and simtime.h includes
      overloaded operators for relational comparisons.  Return
      a reference to a variable of type SpSimTime.             */

  extern int MPIRank;
  extern int MPISize;
  extern MPI_Datatype MPI_SIMTIME;

  int i, j;
  SpSimTime local_time, temp;
  //  Create the array to hold the gathered SpSimTime variables

  SpSimTime *global_time;
  global_time = new SpSimTime[MPISize];

//  C-based MPI routines cannot handle references!
//  So set a local variable equal in value.

  local_time = SimTime;

  //local_time.print();

//  Gather the local SimTime in an array on process 0 
  MPI_Gather(&local_time,1,MPI_SIMTIME,global_time,1, \
	     MPI_SIMTIME,0,MPI_COMM_WORLD);

  if(MPIRank == 0)
    {
      //  Find the minimum global SpSimTime
      temp = SimTime;
      for(i=0;i<MPISize;i++)
	if(global_time[i] < temp) temp = global_time[i];
    }

  //  Broacast the minimum SimTime to all processes

  MPI_Bcast(&temp,1,MPI_SIMTIME,0,MPI_COMM_WORLD);
  SimTime = temp;
  delete[] global_time;

  return SimTime;
 
} // End SpComm_MPI_GlobalMin(SimTime)

//  Define MPI functions for SpComm_MPI_GlobalMin(SimTime) and
//  SpComm_MPI_GlobalMax(SimTime)

//////////////////////////////////////////////////////////////////////

SpSimTime &SpComm_GlobalMax(SpSimTime &SimTime)
{
  /*  Find the global maximum for a SPEEDES variable of type
      SpSimTime.  SpSimTime is a class, and simtime.h includes
      overloaded operators for relational comparisons.  Return
      a reference to a variable of type SpSimTime.             */

  extern int MPIRank;
  extern int MPISize;
  extern MPI_Datatype MPI_SIMTIME;

  int i, j;
  const int GSIZE = MPISize;
  SpSimTime local_time, temp;
  //  Create the array to hold the gathered SpSimTime variables
  SpSimTime *global_time;
  global_time = new SpSimTime[MPISize];

  //  C-based MPI routines cannot handle references!
  //  So set a local variable equal in value.

  local_time = SimTime;

//local_time.print();

//  Gather the local SimTime in an array on process 0 
  MPI_Gather(&local_time,1,MPI_SIMTIME,global_time,1, \
	     MPI_SIMTIME,0,MPI_COMM_WORLD);

  if(MPIRank == 0)
    {
      //  Find the maximum global SpSimTime
      temp = SimTime;
      for(i=0;i<MPISize;i++)
	if(global_time[i] > temp) temp = global_time[i];
    }

  //  Broacast the maximum SimTime to all processes

  MPI_Bcast(&temp,1,MPI_SIMTIME,0,MPI_COMM_WORLD);
  SimTime = temp;
  delete[] global_time;
  return SimTime;
 
} // End SpComm_MPI_GlobalMax(SimTime)

//////////////////////////////////////////////////////////////////////

SpSimTime &SpComm_UDF_GlobalMin(SpSimTime &SimTime)
{
  /*  Find the global minimum for a SPEEDES variable of type
      SpSimTime.  SpSimTime is a class, and simtime.h includes
      overloaded operators for relational comparisons.  Return
      a reference to a variable of type SpSimTime.            */

  extern int MPIRank;
  extern int MPISize;
  extern MPI_Datatype MPI_SIMTIME;
 
  MPI_Op SimMinOp;
  SpSimTime input_class[1], output_class[1];

  // Return a handle to user-specified reduction operator

  MPI_Op_create(SimMinFun, 1, &SimMinOp);

  //  C-based MPI routines cannot handle references!
  //  So set a local variable equal in value.

  input_class[0] = SimTime;

  MPI_Allreduce(input_class, output_class, 1, MPI_SIMTIME, \
		SimMinOp, MPI_COMM_WORLD);

  SimTime = output_class[0];
  return SimTime; 

} // End SpComm_MPIUDF_GlobalMin(SimTime)

//////////////////////////////////////////////////////////////////////

SpSimTime &SpComm_UDF_GlobalMax(SpSimTime &SimTime)
{
  /*  Find the global minimum for a SPEEDES variable of type
      SpSimTime.  SpSimTime is a class, and simtime.h includes
      overloaded operators for relational comparisons.  Return
      a reference to a variable of type SpSimTime.            */

  extern int MPIRank;
  extern int MPISize;
  extern MPI_Datatype MPI_SIMTIME;
 
  MPI_Op SimMaxOp;
  SpSimTime input_class[1], output_class[1];

  // Return a handle to user-specified reduction operator

  MPI_Op_create(SimMaxFun, 1, &SimMaxOp);

  //  C-based MPI routines cannot handle references!
  //  So set a local variable equal in value.

  input_class[0] = SimTime;

  MPI_Allreduce(input_class, output_class, 1, MPI_SIMTIME, \
		SimMaxOp, MPI_COMM_WORLD);

  SimTime = output_class[0];
  return SimTime; 

} // End SpComm_MPIUDF_GlobalMax(SimTime)

//  End SpComm.C

//////////////////////////////////////////////////////////////////////

void create_tag_array(MPI_Comm* communicator,        /* In  */
		      int* TAG_COUNTER_ARRAY_KEY    /* Out */ )
{
  /*   This procedure creates a tag counter array and caches it
       to the input communicator.                          */

/*  'communicator' is the MPI communicator which will have the tag counter
     as an attribute.

     TAG_COUNTER_ARRAY_KEY is the key to retrieve tag counter
     array  */

  static int first = 0;
  if(first == 0)
    {
      //  Go here if cache has never been initialized
      first = 1;
      int i, flag;
      int* tag_vec;    // ptr to tag counter array
      void* extra_arg = 0;

      tag_vec = new int[2*MAX_PROC];
      if(tag_vec == 0)
	{
	  perror("Error in create_tag_array():  memory not allocated for tag_vec.");
	  exit(1);
	}

      //  Create the attribute key
      MPI_Keyval_create(MPI_DUP_FN, MPI_NULL_DELETE_FN, \
			TAG_COUNTER_ARRAY_KEY, extra_arg);

      //  Initialize the tag counter array.

      for(i=0;i<2*MAX_PROC;i++)
	*(tag_vec + i) = 0;

//  Cache the array to the input communicator
      MPI_Attr_put(*communicator, *TAG_COUNTER_ARRAY_KEY, \
		   tag_vec);

    }
}//  End of create_tag_array()

//////////////////////////////////////////////////////////////////
//  END SpComm.C

void SpComm_Net_Terminate()
{
  static int TerminatedAlready =0;

  if (!TerminatedAlready){
    //  Detach from the MPI buffer
    int Buffer_Size = MPI_Buffer_Size;
    MPI_Buffer_detach(&MPI_Buffer, &Buffer_Size);
    
    //  Delete the cached attribute for MPI_PTP_Comm
    MPI_Attr_delete(MPI_PTP_Comm, TAG_COUNTER_ARRAY_KEY_PTP);
    
    MPI_Finalize();
    ++TerminatedAlready;
  }

}

int SpComm_Net_Initialize()
{
  return 0;
}

extern int SpComm_Nlocal;

/*  The pre-fork, post-fork distinction for the following calls probably
    doesn't make much sense in the MPI communication library, since MPI
    will order all processes in some absolute sense.  Also, there is
    no fork in the code, although there is code which is pre-MPI_Init()
    and post-MPI_Init().  But this distinction is retained for the sake
    of upward compatibility.                                         */

void SpComm_Pre_Fork_Net_GetIDs(int **LocalNodeIDs, int &Nboxes,
		       int ** LocalNodeZeros, int **NodeBoxNumber,
                       int &BoxNumber)
{
  *LocalNodeZeros = new int[SpComm_Nlocal];
  *LocalNodeIDs =   new int[SpComm_Nlocal];
  *NodeBoxNumber =  new int[SpComm_Nlocal];
  for (int i=0; i < SpComm_Nlocal; ++i){
    LocalNodeIDs[0][i] = i;
    LocalNodeZeros[0][i] = NodeBoxNumber[0][i] = 0;
  }
  Nboxes = 1;
  BoxNumber = 0; // Under MPI, there is one 'virtual' box
  return;
}

void SpComm_Post_Fork_Net_GetIDs(int **LocalNodeIDs, int &Nboxes,
		       int ** LocalNodeZeros, int **NodeBoxNumber,
                       int &BoxNumber)
{
  *LocalNodeZeros = new int[SpComm_Nlocal];
  *LocalNodeIDs  = new int[SpComm_Nlocal];
  *NodeBoxNumber = new int[SpComm_Nlocal];
  for (int i=0; i < SpComm_Nlocal; ++i){
    LocalNodeIDs[0][i] = i;
    LocalNodeZeros[0][i] = NodeBoxNumber[0][i] = 0;
  }
  Nboxes = 1;
  BoxNumber = 0; // Under MPI, there is one 'virtual' box
  return;
}




void SpComm_Net_Redop()
{
  return;
}

void SpComm_Net_Max(int * ival,double * dval,SpSimTime * Stime,int Compath)
{
  return;
}


void SpComm_Net_Sum(int * ival,double * dval,SpSimTime * Stime,int Compath)
{
  return;
}


void SpComm_Net_Min(int * ival,double * dval,SpSimTime * Stime,int Compath)
{
  return;
}




void SpComm_Net_Enter_nbRedop()
{
  return;
}


int SpComm_Net_nbRedop()
{
  return 1;
}

void SpComm_Net_Receive()
{
  return;
}


void SpComm_Net_Send(int dest, int type, int nbytes, void * buff)
{
  return;
}



void SpComm_Net_Send(SpDestination *dest, int type, int nbytes, void *buff)
{
  return;
}

/* Assume int * dest is all on one box with dest[0] having the number of dest*/
void SpComm_Net_Send(int *dest, int type, int nbytes, void *buff)
{
  return;
}

/* send to everyone but self */
void SpComm_Net_Send(int type, int nbytes, void *buff)
{
  return;
}



