#ifndef SPCOMM_NET_H
#define SPCOMM_NET_H

class SpSimTime;
class SpDestination;

/*************************************************************************
 * Function: SpComm_Net_Initialize
 * Summary: Initialize the network for communication
 * Description: Initialized the network interface. In the normal
 *              SpComm.C, this is done before each box has
 *              fork()'ed off each of it's children.  This does not
 *              need to be the case and a different SpComm.C
 *              could be written.
 * Return: 0 on success, 1 on failure
 *************************************************************************/


int SpComm_Net_Initialize();


/*************************************************************************
 * Function: SpComm_Pre_Fork_Net_GetIDs
 * Summary: A call before the fork() to get the network ID's
 * Description: Determines the local node id's, the number of boxes,
 *              the global node id's of the local node zero's for each
 *              node, and  the number of the physical box on which
 *              each node resides.  For  example, if we had two
 *              boxes, one with 2 CPU's and the other with  3, these
 *              structures could look like:
 *
 *
 *              GID    LID       LNZ      NBN
 *               0      0         0        0
 *               1      1         0        0
 *               2      0         2        1
 *               3      2         0        0
 *               4      1         2        1
 *
 *
 *              In this case, the box with 3 CPU's got assigned global
 *              node ID's 0,  1, and 3. while the other box was
 *              assigned ID's 2 and 4.   LocalNodeZero is not
 *              necessarily needed in each situation but with  FM, we
 *              need to know who is routing the messages for that box.
 *
 *              There are two versions of the calls that you must
 *              provide.  One is called before the child processes are
 *              forked off and the other is called after.  Both are
 *              provided since you may have a need for one or the
 *              other or both.
 * Return: void
 *************************************************************************/
void SpComm_Pre_Fork_Net_GetIDs(int **LocalNodeIDs, //Out: *LocalNodeIDs as
				                     //     in the description
				 int &Nboxes,        //Out: Number of boxes
				                     //     participating
				 int **LocalNodeZeros, //Out: *LocalNodeZeros
				                       //     as in the
				                       //     description
				 int **NodeBoxNumber, //Out: *NodeBoxNumber
				                       //     as in the
				                       //     description
				 int &BoxNumber //Out: What box number is the
				                //     box calling this
				                //     procedure
				 );


/*************************************************************************
 * Function: SpComm_Post_Fork_Net_GetIDs
 * Summary: A pre-fork call to set up the node tables.
 * Description: Determines the local node id's, the number of boxes,
 *              the global node id's of the local node zero's for each
 *              node, and  the number of the physical box on which
 *              each node resides.  For  example, if we had two
 *              boxes, one with 2 CPU's and the other with  3, these
 *              structures could look like:
 *
 *
 *              GID    LID       LNZ      NBN
 *               0      0         0        0
 *               1      1         0        0
 *               2      0         2        1
 *               3      2         0        0
 *               4      1         2        1
 *
 *
 *              In this case, the box with 3 CPU's got assigned global
 *              node ID's 0,  1, and 3. while the other box was
 *              assigned ID's 2 and 4.   LocalNodeZero is not
 *              necessarily needed in each situation but with  FM, we
 *              need to know who is routing the messages for that box.
 *
 *              There are two versions of the calls that you must
 *              provide.  One is called before the child processes are
 *              forked off and the other is called after.  Both are
 *              provided since you may have a need for one or the
 *              other or both.
 * Return: void
 *************************************************************************/
void SpComm_Post_Fork_Net_GetIDs(int **LocalNodeIDs, //Out: *LocalNodeIDs as
				                     //     in the description
				 int &Nboxes,        //Out: Number of boxes
				                     //     participating
				 int **LocalNodeZeros, //Out: *LocalNodeZeros
				                       //     as in the
				                       //     description
				 int **NodeBoxNumber, //Out: *NodeBoxNumber
				                       //     as in the
				                       //     description
				 int &BoxNumber //Out: What box number is the
				                //     box calling this
				                //     procedure
				 );



/*************************************************************************
 * Function: SpComm_Net_Terminate
 * Summary:  Terminate the network communications
 * Description: This call should clean up anything that the network
 *           leaves hanging. That is, this should consist of anything
 *           necessary to shut down the communications.
 * Return: void
 *************************************************************************/
void SpComm_Net_Terminate();



/*************************************************************************
* Function: SpComm_Net_Enter_nbRedop
* Summary: Enter the non-blocking reduction
* Description: Once all the nodes on the box agree that the non-blocking
*              reduction has been competed, we check with the network to
*              see if the rest of the boxes agree
* Return: void
*************************************************************************/
void SpComm_Net_Enter_nbRedop();



/*************************************************************************
 * Function: SpComm_Net_nb_Redop
 * Summary: check to see if it is ok to exit the non-blocking reduction
 * Description: After one has entered the non-blocking reduction with
 *              SpComm_Net_Enter_nbRedop(), sucessive calls to
 *              SpComm_Net_nbRedop will indicate if one can exit
 *              this "fuzzy" barrier This non-blocking reduction is
 *              first finished through shared memory.  Once we have
 *              completed the shared memory reduction, one node (local
 *              node 0) checks the network to see if the rest of the
 *              boxes agree that the non-blocking reduction can be
 *              completed.
 * Return: 1 if we can exit the non-blocking reduction.  0 otherwise
 *************************************************************************/
int SpComm_Net_nbRedop();



/*************************************************************************
 * Function: SpComm_Net_Redop
 * Summary: Performs a blocking reduction
 * Description: A blocking reduction is first performed on each box.  Once
 *              this is completed, Node 0 then goes to the network to
 *              see if it can finish the blocking reduction.  As soon
 *              as the blocking reduction on the network finishes, this
 *              call returns.
 * Return: void
 *************************************************************************/
void SpComm_Net_Redop();



/*************************************************************************
 * Function: SpComm_Net_Sum
 * Summary: returns the global sum of ival, dval, and SimTime for all nodes
 * Description: Each node finds the global sum of ival, dval or
 *              SimTime.  These are then passed into SpComm_Net_Sum and
 *              the sum over all the boxes is computed.  The values are
 *              then returned in ival, dval and SimTime.
 * Return: void
 *************************************************************************/
void SpComm_Net_Sum(int * ival, //INOUT: Pass in the value of ival, return
		                //       the global sum of all ivals
		    double * dval, //INOUT: Pass in the value of dval, return
		                   //       the global sum of all dvals
		    SpSimTime *SimTime, //INOUT: Pass in the value of SimTime,
		                      //       return the global sum of all
		                      //       SimTimes
		    int ComPath //In: An extra parameter in case a special
		                //    compath is needed for communication
		    );



/*************************************************************************
 * Function: SpComm_Net_Min
 * Summary: returns the global Min of ival, dval, and SimTime for all nodes
 * Description: Each node finds the global Min of ival, dval or
 *              SimTime.  These are then passed into SpComm_Net_Min and
 *              the min over all the boxes is computed.  The values are
 *              then returned in ival, dval and SimTime.
 * Return: void
 *************************************************************************/
void SpComm_Net_Min(int * ival, //INOUT: Pass in the value of ival, return
		                //       the global sum of all ivals
		    double * dval, //INOUT: Pass in the value of dval, return
		                   //       the global sum of all dvals
		    SpSimTime *SimTime, //INOUT: Pass in the value of SimTime,
		                      //       return the global sum of all
		                      //       SimTimes
		    int ComPath //In: An extra parameter in case a special
		                //    compath is needed for communication
		    );



/*************************************************************************
 * Function: SpComm_Net_Max
 * Summary: returns the global Max of ival, dval, and SimTime for all nodes
 * Description: Each node finds the global Max of ival, dval or
 *              SimTime.  These are then passed into SpComm_Net_Max and
 *              the max over all the boxes is computed.  The values are
 *              then returned in ival, dval and SimTime.
 * Return: void
 *************************************************************************/
void SpComm_Net_Max(int * ival, //INOUT: Pass in the value of ival, return
		                //       the global sum of all ivals
		    double * dval, //INOUT: Pass in the value of dval, return
		                   //       the global sum of all dvals
		    SpSimTime *SimTime, //INOUT: Pass in the value of SimTime,
		                      //       return the global sum of all
		                      //       SimTimes
		    int ComPath //In: An extra parameter in case a special
		                //    compath is needed for communication
		    );



/*************************************************************************
 * Function: SpComm_Net_Send
 * Summary: send a single message
 * Description: send a single message to a node that is not on our box.
 *              This is a non-blocking call so once we sent it, we can
 *              assume that it is going to reach it's destination without
 *              any further intervention from the user.
 * Return: void
 *************************************************************************/
void SpComm_Net_Send(int destination, //In: Destination of the message
		     int type,        //In: Type of the message
		     int nbytes,      //In: Size of the message
		     void * buff      //In: message
		     );



/*************************************************************************
 * Function: SpComm_Net_Sent
 * Summary: Sends a message to many destinations
 * Description: dest is a class that contains all of the destinations
 *              of this message.  We can assume that all the
 *              destinations are on the same box.  As with other
 *              Net_Send calls, this is non-blocking so we can assume
 *              that it is going to reach it's destination without any
 *              further intervention from the user.
 * Return: void
 *************************************************************************/
void SpComm_Net_Send(SpDestination *dest, //In: Destinations of the message
		     int type,          //In: Type of the message
		     int nbytes,        //In: Size of the message
		     void *buff         //In: message
		     );



/*************************************************************************
 * Function: SpComm_Net_Send
 * Summary: Sends a message to multiple destinations
 * Description: *dest is an array containing all hte destinations
 *              which are all on the same box. dest[0] having the
 *              number of destinations.  Then, dest[1] through
 *              dest[dest[0]+1] contain each of the destinations.  As
 *              with other Net_Send calls, this is non-blocking so we
 *              can assume that it is going to reach it's destination
 *              without any further intervention from the user.
 * Return: void
 *************************************************************************/
void SpComm_Net_Send(int *dest,      //In: Destinations of the message
		     int type,       //In: Type of the message
		     int nbytes,     //In: Size of the message
		     void *buff      //In: message
		     );



/*************************************************************************
 * Function: SpComm_Net_Send
 * Summary: Broadcast to all nodes not on our box
 * Description: We broadcast to each node on our box first with shared
 *              memory and then we call this net_send to sent the
 *              message to everyone that is not on our box. As with
 *              other Net_Send calls, this is non-blocking so we can
 *              assume that it is going to reach it's destination
 *              without any further intervention from the user.
 * Return: void
 *************************************************************************/
void SpComm_Net_Send(int type,   //In: Type of the message
		     int nbytes, //In: Size of the message
		     void *buff  //In: message
		     ); 



/*************************************************************************
 * Function: SpComm_Net_Receive
 * Summary: Receive all messages possible from the network
 * Description: Since receiving is sometimes costly, we want to just
 *              call receive to try and get all messages at once.
 *              This call should receive all the messages from the
 *              network that it can and then route them to the
 *              apropriate destinations, usually through shared
 *              memory.  It defaults to non-blocking and simply
 *              returns when there are no other messages it can pick
 *              up at that time.
 * Return: void
 *************************************************************************/
void SpComm_Net_Receive(int blocking = 0 //Default to not blocking on a receive
			);

#endif
