#ifndef Client_Sock_Class
#define Client_Sock_Class

#include "SpStreamSock.H"
#include "SpConnection.H"

class GENERIC_MSG_HEADER;

/*
** status values for connections (which can be checked in the
** ConnectionStatus() method
*/

enum {
  BAD_CONNECTION_STATUS,
  GOOD_CONNECTION_STATUS
};

class CLIENT_SOCK : public STREAM_SOCK {
 public:
   // Construct a new CLIENT_SOCK and connect to the server running on machine
   // serverMachineName via port server_port;  if fail to connect, keep trying for
   // waitSecs seconds.
   // After the CLIENT_SOCK has been instantiated, you are ready to read/write msgs
   // from/to the server (assuming no error making connection;  see ConnectionStatus).
   /*****************************************************************
   * Method: CLIENT_SOCK constructor
   * Summary: Creates the obj and connects to server running on
   *          serverMachineName at serverPort.
   * Description: Construct a new CLIENT_SOCK and connect to the
   *              server running on machine serverMachineName at
   *              serverPort.  If fail to connect, keep trying for
   *              waitSecs seconds. After the CLIENT_SOCK has been
   *              instantiated, you are ready to read/write msgs
   *              from/to the server (assuming no error making
   *              connection; see ConnectionStatus)
   * Return: NONE
   ******************************************************************/
   CLIENT_SOCK(char *serverMachineName,
	       int serverPort=DEFAULT_PORT,
	       int waitSecs = 150);

   ~CLIENT_SOCK(void);
    
   // Connect to server on given machine @ port;  keep trying for 150 secs
   // before exiting
   int connectToServer(char *serverMachineName, int serverPort=DEFAULT_PORT,
		       int waitSecs = 150);
   
   int connectToServer(struct sockaddr_in *serverAddr, int waitSecs);

   /******************************************************************
   * Method: ConnectionStatus
   * Summary: Tells whether we're connected to server.
   * Description: Returns GOOD_CONNECTION_STATUS if properly connected
   *              to the server; otherwise returns
   *              BAD_CONNECTION_STATUS.
   * Return: GOOD_CONNECTION_STATUS or BAD_CONNECTION_STATUS
   ******************************************************************/
   int ConnectionStatus() {
     int ConnStatus = BAD_CONNECTION_STATUS;
     if (ServerConn) ConnStatus = GOOD_CONNECTION_STATUS;
     return ConnStatus;
   }



   /* NOTE:  these are the new, PREFERRED methods for reading/writing messages and
             are intended to replace getMessage, getMessage_blk, 
	     sendMsg, sendMsg_blk,
             (which remain for backward compatibility).  The main difference is that
             they return more informative status/error codes (see enum Socket_Status_Code
	     in SpStreamSock.H for status codes).
   ***/

   /******************************************************************
   * Method: readMsg
   * Summary: Read a message from server (non-blocking); return status
   *          code.
   * Description: Does a non-blocking read of the next msg (if any)
   *              from the server into the SockMsgType pointed to by
   *              arg, msg (basically just a char* and #bytes).
   *              Returns a status code indicating what happened. If
   *              you only get part of a message, repeatedly call
   *              readMsg until you've got the whole thing.
   * Return: returns one of the following STREAM_SOCK:: enums
   *         (see SpStreamSock.H):
   *         WHOLE_MSG_OK  :   entire msg read;  everything is OK
   *         PARTIAL_MSG_OK:   part msg read;  everything OK
   *         ZERO_BYTES_READ: read 0 bytes; if select says sock is 
   *           readable but we get 0 bytes, this means that the
   *           connection has been closed by the peer.
   *         MSG_TOO_BIG: incoming msg bigger than MAX_MSG_BYTES 
   *           (100 MEG!);  probably a mangled msg.
   *         TEMP_SOCK_PROBLEM: temporary problem (eg full buffer, 
   *           interupted call) that _may_ fix itself
   *         BAD_SOCK:  Bad file descriptor for connection (probably
   *           closed conn)
   *         BAD_PTR_OR_ARG:  Bad internal pointer
   *         NOT_YET_CONNECTED: Possibly socket is ok, but still 
   *           doing connection internal work
   *         CLOSED_CONN: Connection is closed
   *         UNKNOWN_SOCK_ERROR:  Don't know what is wrong;  better 
   *           bail out.
   ******************************************************************/
   int readMsg(SockMsgType *msg) {
      return ServerConn->readMsg(msg);
   }

   /*********************************************************************
   * Method: readMsg_blk
   * Summary: Block until next msg read from server; return status code.
   * Description: Does a blocking read of the next msg 
   *              from the server into the SockMsgType pointed to by
   *              arg, msg (basically just a char* and #bytes).
   *              Returns a status code indicating what happened. 
   *              Will either get the entire message, or else
   *              return an error code.
   * Return: returns one of the following STREAM_SOCK:: enums
   *         (see SpStreamSock.H):
   *         WHOLE_MSG_OK  :   entire msg read;  everything is OK
   *         MSG_TOO_BIG: incoming msg bigger than MAX_MSG_BYTES 
   *           (100 MEG!);  probably a mangled msg.
   *         TEMP_SOCK_PROBLEM: temporary problem (eg full buffer, 
   *           interupted call) that _may_ fix itself
   *         BAD_SOCK:  Bad file descriptor for connection (probably
   *           closed conn)
   *         BAD_PTR_OR_ARG:  Bad internal pointer
   *         NOT_YET_CONNECTED: Possibly socket is ok, but still 
   *           doing connection internal work
   *         CLOSED_CONN: Connection is closed
   *         UNKNOWN_SOCK_ERROR:  Don't know what is wrong;  better 
   *           bail out.
   ******************************************************************/
   int readMsg_blk(SockMsgType *msg) {
      return ServerConn->readMsg_blk(msg);
   }

   /******************************************************************
   * Method: readMsgIntoBuff
   * Summary: Read next message (non-blocking) from server into 
   *          user-provided buffer; return status code.
   * Description: Read msg (non-blocking) into user-provided buffer,
   *              permanentMsgBuff, and set *msgBytes to be the number
   *              of bytes stored in permanentMsgBuff; return a status
   *              code.  Arg buffLen is the size of permanentMsgBuff
   *              (in bytes); if the incoming msg is too big, return
   *              STREAM_SOCK::USER_BUFF_TOO_SMALL, and set *msgBytes
   *              equal to the size of the incoming msg.  Then, the
   *              user can retrieve this msg by passing in a bigger
   *              buffer, or else by calling readMsg (which will
   *              allocate a new char[N] buffer).  If this returns
   *              PARTIAL_MSG_OK, do 1 or more additional calls until
   *              you've got the entire msg (return == WHOLE_MSG_OK).
   *              The provided buffer is called "permanentMsgBuff"
   *              because it must remain intact in case there is a
   *              partial receive that will be completed later; this
   *              means it should NOT be stack storage (unless you're
   *              certain you'll complete the read before exiting the
   *              function).
   * Return: returns one of the following STREAM_SOCK:: enums
   *         (see SpStreamSock.H):
   *         WHOLE_MSG_OK  :   entire msg read;  everything is OK
   *         PARTIAL_MSG_OK:   part msg read;  everything OK
   *         USER_BUFF_TOO_SMALL: buffer is too small to hold in-
   *           coming msg;  redo the call with a bigger buffer or
   *           call readMsg (which will allocate one for you).
   *         ZERO_BYTES_READ: read 0 bytes; if select says sock is 
   *           readable but we get 0 bytes, this means that the
   *           connection has been closed by the peer.
   *         MSG_TOO_BIG: incoming msg bigger than MAX_MSG_BYTES 
   *           (100 MEG!);  probably a mangled msg.
   *         TEMP_SOCK_PROBLEM: temporary problem (eg full buffer, 
   *           interupted call) that _may_ fix itself
   *         BAD_SOCK:  Bad file descriptor for connection (probably
   *           closed conn)
   *         BAD_PTR_OR_ARG:  Bad msg buff pointer
   *         NOT_YET_CONNECTED: Possibly socket is ok, but still 
   *           doing connection internal work
   *         CLOSED_CONN: Connection is closed
   *         UNKNOWN_SOCK_ERROR:  Don't know what is wrong;  better 
   *           bail out.
   ******************************************************************/

   int readMsgIntoBuff(GENERIC_MSG_HEADER *permanentMsgBuff, int buffLen, long int *msgBytes) {
      return ServerConn->readMsgIntoBuff((char *) permanentMsgBuff, buffLen, msgBytes);
   }
		       


   /******************************************************************
   * Method: readMsgIntoBuff_blk
   * Summary: Read next message (blocking) from server into 
   *          user-provided buffer; return status code.
   * Description: Read msg (blocking) into user-provided buffer,
   *              permanentMsgBuff, and set *msgBytes to be the number
   *              of bytes stored in permanentMsgBuff; return a status
   *              code.  Arg buffLen is the size of permanentMsgBuff
   *              (in bytes); if the incoming msg is too big, return
   *              STREAM_SOCK::USER_BUFF_TOO_SMALL, and set *msgBytes
   *              equal to the size of the incoming msg.  Then, the
   *              user can retrieve this msg by passing in a bigger
   *              buffer, or else by calling readMsg (which will
   *              allocate a new char[N] buffer).  If this returns
   *              PARTIAL_MSG_OK, do 1 or more additional calls until
   *              you've got the entire msg (return == WHOLE_MSG_OK).
   *              The provided buffer is called "permanentMsgBuff"
   *              because it must remain intact in case there is a
   *              partial receive that will be completed later; this
   *              means it should NOT be stack storage (unless you're
   *              certain you'll complete the read before exiting the
   *              function).
   * Return: returns one of the following STREAM_SOCK:: enums
   *         (see SpStreamSock.H):
   *         WHOLE_MSG_OK  :   entire msg read;  everything is OK
   *         USER_BUFF_TOO_SMALL: buffer is too small to hold in-
   *           coming msg;  redo the call with a bigger buffer or
   *           call readMsg (which will allocate one for you).
   *         ZERO_BYTES_READ: read 0 bytes; if select says sock is 
   *           readable but we get 0 bytes, this means that the
   *           connection has been closed by the peer.
   *         MSG_TOO_BIG: incoming msg bigger than MAX_MSG_BYTES 
   *           (100 MEG!);  probably a mangled msg.
   *         TEMP_SOCK_PROBLEM: temporary problem (eg full buffer, 
   *           interupted call) that _may_ fix itself
   *         BAD_SOCK:  Bad file descriptor for connection (probably
   *           closed conn)
   *         BAD_PTR_OR_ARG:  Bad msg buff pointer
   *         NOT_YET_CONNECTED: Possibly socket is ok, but still 
   *           doing connection internal work
   *         CLOSED_CONN: Connection is closed
   *         UNKNOWN_SOCK_ERROR:  Don't know what is wrong;  better 
   *           bail out.
   ******************************************************************/
   int readMsgIntoBuff_blk(GENERIC_MSG_HEADER *permanentMsgBuff, 
			   int buffLen, long int *msgBytes) {
      return ServerConn->readMsgIntoBuff_blk((char *) permanentMsgBuff, buffLen, msgBytes);
   }
		       

   /******************************************************************
   * Method: writeMsg
   * Summary: Send len bytes pointed to by arg, msgBuff, to server
   *          (non-blocking); return status code.
   * Description: Send len bytes pointed to by arg, msgBuff, to
   *              server (non-blocking); return status code.  If you
   *              get a partial write (returns PARTIAL_MSG_OK), the
   *              state of the operation is kept internally and must
   *              be completed before beginning another write; call
   *              resumePartialwrite 1 or more times to finish a
   *              partial write.
   * Return: returns one of the following STREAM_SOCK:: enums
   *         (see SpStreamSock.H):
   * WHOLE_MSG_OK  :  entire msg sent;  everything is OK
   * PARTIAL_MSG_OK:  part msg sent;  everything OK. Complete this
   *                  operation with calls to resumePartialwrite.
   * TEMP_SOCK_PROBLEM: temporary problem (eg full buffer, interupted 
   *                    call) that _may_ fix itself
   * BAD_SOCK:  Bad file descriptor for connection (probably closed 
   *            connection)
   * BAD_PTR_OR_ARG:  Bad pointer or msg length arg
   * NOT_YET_CONNECTED: Possibly socket is ok, but still doing 
   *                    connection internal work
   * CLOSED_CONN: Connection is closed
   * UNKNOWN_SOCK_ERROR:  Don't know what is wrong;  better quit
   ******************************************************************/
   int writeMsg(GENERIC_MSG_HEADER *msgBuff, int len) {
      return ServerConn->writeMsg((char*) msgBuff, len);
   }

   /******************************************************************
   * Method: writeMsg_blk
   * Summary: Send len bytes pointed to by arg, msgBuff, to server
   *          (blocking); return status code.
   * Description: Send len bytes pointed to by arg, msgBuff, to
   *              server (blocking); return status code.
   * Return: returns one of the following STREAM_SOCK:: enums
   *         (see SpStreamSock.H):
   * WHOLE_MSG_OK  :  entire msg sent;  everything is OK
   * TEMP_SOCK_PROBLEM: temporary problem (eg full buffer, interupted 
   *                    call) that _may_ fix itself
   * BAD_SOCK:  Bad file descriptor for connection (probably closed 
   *            connection)
   * BAD_PTR_OR_ARG:  Bad pointer or msg length arg
   * NOT_YET_CONNECTED: Possibly socket is ok, but still doing 
   *                    connection internal work
   * CLOSED_CONN: Connection is closed
   * UNKNOWN_SOCK_ERROR:  Don't know what is wrong;  better quit
   ******************************************************************/
   int writeMsg_blk(GENERIC_MSG_HEADER *msgBuff, int len) {
      int rc = ServerConn->writeMsg_blk((char*) msgBuff, len);
      return rc;
   }


   /******************************************************************
   * Method: writeMsgNoCopy
   * Summary: Send len bytes pointed to by arg, permMsgBuff, to server
   *          (non-blocking); return status code.  In case of partial
   *          write, do NOT copy remainder from permMsgBuff.
   * Description: Send len bytes pointed to by arg, msgBuff, to
   *              server (non-blocking); return status code.  If you
   *              get a partial write (returns PARTIAL_MSG_OK), the
   *              state of the operation is kept internally; in this
   *              case we will NOT copy the remaining bytes, but will
   *              just keep a pointer into permMsgBuff until the
   *              write can be finished (with 1 or more calls to
   *              resumePartialWrite), so make sure permMsgBuff points
   *              to permanent storage.  The arg, deletePermMsg_p, if
   *              true, means that this method will delete the storage
   *              when done with it using :
   *              delete[] (char *) permMsgBuff;
   *              So permMsgBuff must be allocated as new char[N] if
   *              you want to use this option;  otherwise delete it
   *              yourself.
   * Return: returns one of the following STREAM_SOCK:: enums
   *         (see SpStreamSock.H):
   * WHOLE_MSG_OK  :  entire msg sent;  everything is OK
   * PARTIAL_MSG_OK:  part msg sent;  everything OK. Complete this
   *                  operation with calls to resumePartialWrite.
   * TEMP_SOCK_PROBLEM: temporary problem (eg full buffer, interupted 
   *                    call) that _may_ fix itself
   * BAD_SOCK:  Bad file descriptor for connection (probably closed 
   *            connection)
   * BAD_PTR_OR_ARG:  Bad pointer or msg length arg
   * NOT_YET_CONNECTED: Possibly socket is ok, but still doing 
   *                    connection internal work
   * CLOSED_CONN: Connection is closed
   * UNKNOWN_SOCK_ERROR:  Don't know what is wrong;  better quit
   ******************************************************************/
   int writeMsgNoCopy(GENERIC_MSG_HEADER *permMsgBuff, int len, int deletePermMsg_p) {
      return ServerConn->writeMsgNoCopy((char*) permMsgBuff, len, deletePermMsg_p);
   }




   /******************************************************************
   * Method: writeMsgNoCopy_blk
   * Summary: Send len bytes pointed to by arg, permMsgBuff, to server
   *          (blocking); return status code.  In case of partial
   *          write, do NOT copy remainder from permMsgBuff.
   * Description: Send len bytes pointed to by arg, msgBuff, to
   *              server (non-blocking); return status code. Although
   *              a blocking call cannot return a partial write, this
   *              can happen internally. In such a case,
   *              we will NOT copy the remaining bytes, but will
   *              just keep a pointer into permMsgBuff until the
   *              write can be finished (with 1 or more calls to
   *              resumePartialWrite), so make sure permMsgBuff points
   *              to permanent storage.  The arg, deletePermMsg_p, if
   *              true, means that this method will delete the storage
   *              when done with it using :
   *              delete[] (char *) permMsgBuff;
   *              So permMsgBuff must be allocated as new char[N] if
   *              you want to use this option;  otherwise delete it
   *              yourself.
   * Return: returns one of the following STREAM_SOCK:: enums
   *         (see SpStreamSock.H):
   * WHOLE_MSG_OK  :  entire msg sent;  everything is OK
   * TEMP_SOCK_PROBLEM: temporary problem (eg full buffer, interupted 
   *                    call) that _may_ fix itself
   * BAD_SOCK:  Bad file descriptor for connection (probably closed 
   *            connection)
   * BAD_PTR_OR_ARG:  Bad pointer or msg length arg
   * NOT_YET_CONNECTED: Possibly socket is ok, but still doing 
   *                    connection internal work
   * CLOSED_CONN: Connection is closed
   * UNKNOWN_SOCK_ERROR:  Don't know what is wrong;  better quit
   ******************************************************************/
   int writeMsgNoCopy_blk(GENERIC_MSG_HEADER *permMsgBuff, int len, int deletePermMsg_p) {
      return ServerConn->writeMsgNoCopy_blk((char*) permMsgBuff, len, deletePermMsg_p);
   }


   /******************************************************************
   * Method: resumePartialWrite
   * Summary: Try to finish writing the last message (which resulted
   *          in a partial write);  return status code.
   * Description: resumePartialWrite attempts to complete the last
   *              partial write operation, and returns a status code.
   *              A call to this method may itself result in a part-
   *              ial send, so it may take more than 1 call to get
   *              the whole message out.  Since the state of a partial
   *              send is recorded internally, no args are needed.
   *              Note the asymmetry between the reads and the 
   *              writes in that there is NO analogous "resumeRead"
   *              method.  This method was necessary to avoid compli-
   *              cations resulting from trying to write a new msg
   *              when there is a pending partial write.
   * Return: returns one of the following STREAM_SOCK:: enums
   *         (see SpStreamSock.H):
   * WHOLE_MSG_OK  :  entire msg sent;  everything is OK
   * PARTIAL_MSG_OK:  part msg sent;  everything OK. Complete this
   *                  operation with calls to resumePartialwrite.
   * TEMP_SOCK_PROBLEM: temporary problem (eg full buffer, interupted 
   *                    call) that _may_ fix itself
   * BAD_SOCK:  Bad file descriptor for connection (probably closed 
   *            connection)
   * BAD_PTR_OR_ARG:  Bad pointer or msg length arg
   * NOT_YET_CONNECTED: Possibly socket is ok, but still doing 
   *                    connection internal work
   * CLOSED_CONN: Connection is closed
   * UNKNOWN_SOCK_ERROR:  Don't know what is wrong;  better quit
   ******************************************************************/
   int resumePartialWrite(void) {
      return ServerConn->resumePartialWrite();
   }


   /******************************************************************
   * Method: havePartialWrite_p
   * Summary: Returns true (1) if there is a pending partial write;
   *          otherwise returns false (0).
   * Description: Returns true (1) if there is a pending partial
   *              write; otherwise returns false (0). If true, you 
   *              must finish the partial write with 1 or more calls 
   *              to resumePartialWrite before starting a new write.
   *         
   * Return: 0 or 1.
   ******************************************************************/
   int havePartialWrite_p(void) {
      return ServerConn->havePartialWrite_p();
   }


   /******************************************************************
   * Method: havePartialRead_p
   * Summary: Returns true (1) if there is a pending partial read;
   *          otherwise returns false (0).
   * Description: Returns true (1) if there is a pending partial
   *              read; otherwise returns false (0). If true, you 
   *              must finish the partial read with 1 or more calls 
   *              to the various read methods.
   * Return: 0 or 1.
   ******************************************************************/
   int havePartialRead_p(void) {
      return ServerConn->havePartialRecv_p();
   }


   int dataReadyForRead_p(void) { 
      return ServerConn->dataReadyForRead_p();
   }


   int sendMsg(GENERIC_MSG_HEADER *msgBuff, int len) {
      return ServerConn->sendMessage((char*) msgBuff, len);
   }

   int sendMsg_blk(GENERIC_MSG_HEADER *msgBuff, int len) {
      int rc = ServerConn->sendMessage_blk((char*) msgBuff, len);
      return rc;
   }

   // Try to complete sending of last partial send.   Use this fcn
   // repeatedly until entire msg is sent.  Returns status codes:
   // STREAM_SOCK::WHOLE_MSG_OK, PARTIAL_MSG_OK, and error codes.
   int resumePartialMsgSend(void) {
      return ServerConn->resumePartialSend();
   }


   int getMessage(SockMsgType *msg) {
      return ServerConn->getMsg(msg);
   }

   int getMessage_blk(SockMsgType *msg) {
      return ServerConn->getMsg_blk(msg);
   }

   int send_data(char *buff, int length, int flags=0) {
      return STREAM_SOCK::send_data(getfd(), buff, length, flags);
   }

   int receive_data(char *buff, int length, int flags=0) {
      return STREAM_SOCK::receive_data(getfd(), buff, length, flags);
   }

  GENERIC_MSG_HEADER *getMsg_blk();
  
  GENERIC_MSG_HEADER *getMsg(void) {
    if (ServerConn->dataReadyForRead_p()){
      return getMsg_blk();
    }
    else{
      return NULL;
    }
  }

 protected:


 private:
   char *ServerName;        // machine name server is on
   int ServerPort;          // port server is bound to

   CONNECTION *ServerConn;  // CONNECTION obj thru which read/writes occur

};

#endif
