#include "SpAssert.H"
#include <sys/types.h>
#ifndef __SPEEDES_NT__
#include <netdb.h>
#else
#include <windows.h>
#endif

#include "SpStdlib.H"
#include "SpErrno.H"
#include "SpString.H"
#include "SpConnection.H"


const int BOGUS_PORT = -3333;
const int BOGUS_FD = -7531;
const int BOGUS_ID = -1234;
const int BOGUS_MSG_LENGTH = -5555;
const int MAX_SOCK_BUFF = 54250;  // This is for HP;  what about others?


CONNECTION::CONNECTION(int fd, int port, int idnum, SERVICE_GROUP *cgrp)
  : STREAM_SOCK(fd, port), host_name(NULL)
{
  initializeUnix_msghdr();
  PartialSendMemory = NULL;
  ConsecutiveTempErrors = 0;

  id = idnum;
  setPartialSend((msghdr*)NULL);
  setConnGroup(cgrp);
  setNoPartialRecv();
  setSocketStatus(CONNECTED);
  if (fd != BOGUS_FD) setMaxUnixBuffers();  // max Unix internal send/recv buffers

  init_host_name();
  setNonBlocking(); // non-blocking mode is default
}


CONNECTION::~CONNECTION() {
  if (host_name) free(host_name);
  Sendq.clear();

  if (PartialSendMemory!=NULL) delete[] PartialSendMemory;
}


void CONNECTION::closeConnection(void){
  cout << "CONNECTION::closeConnection" << endl;
#ifndef __SPEEDES_NT__
  int ret=close(getfd());
#else
  int ret=closesocket(getfd());
#endif
  if (ret < 0) perror("closing socket:");
}


void CONNECTION::init_host_name()
{
#ifdef I_DO_NOT_THINK_WE_REALLY_NEED_THE_HOST_NAME_OF_THE_PEER
  static struct sockaddr connectorAddr;
  int addrBytesInOut = sizeof connectorAddr;
  
  int failed_p = getpeername(getfd(),
			     (sockaddr*)
			     &connectorAddr,
			     &addrBytesInOut);
  assert(!failed_p);
  assert(addrBytesInOut == sizeof connectorAddr);

  const struct hostent * const hn
    = gethostbyaddr((char *)(&(connectorAddr.sin_addr)),
		     sizeof(connectorAddr.sin_addr), AF_INET);
  assert(hn != NULL);

  free(host_name);
  host_name = strdup(hn->h_name);
#endif
}


int CONNECTION::readMsgIntoBuff(char *permanentMsgBuff, int buffLen, long int *msgBytes) {
   int rcvStatus, fd = getfd();
   SockMsgType *partialMsg, msg;

   if (havePartialRecv_p()) {    // in middle of receiving message
      partialMsg = getPartialRecv();
      rcvStatus = STREAM_SOCK::resumeReadMsgIntoBuff(fd, partialMsg,
						     permanentMsgBuff, buffLen, msgBytes);
      if (rcvStatus == WHOLE_MSG_OK) {
	 setNoPartialRecv(); // mark data member as empty
      }
   }
   else {   // Brand new message:
      rcvStatus = STREAM_SOCK::readMsgIntoBuff(fd, &msg, permanentMsgBuff, buffLen, msgBytes);

      if (rcvStatus == PARTIAL_MSG_OK) {
	 setPartialRecv(&msg);
	 //cout << " --> getMsg part recv on 1st try; #bytes = " 
	 //     << msg->getBytesRecvd() << endl;
      }
   }

   return rcvStatus;
}
  

int CONNECTION::readMsg(SockMsgType *msg, int flags) {
   int rcvStatus, fd = getfd();
   SockMsgType *partialMsg;

   if (havePartialRecv_p()) {    // in middle of receiving message
      partialMsg = getPartialRecv();
      rcvStatus = resumeRecvMsg(fd, partialMsg, flags);
      if (rcvStatus == WHOLE_MSG_OK) {
	 *msg = *partialMsg;
	 setNoPartialRecv(); // mark data member as empty
      }
   }
   else {   // Brand new message:
      rcvStatus = receiveMsg(fd, msg, flags);
      if (rcvStatus == PARTIAL_MSG_OK) {
	 setPartialRecv(msg);
	 //cout << " --> getMsg part recv on 1st try; #bytes = " 
	 //     << msg->getBytesRecvd() << endl;
      }
   }

   return rcvStatus;
}
int CONNECTION::readMsg_blk(SockMsgType *msg, int flags) {
   int ret;
   int blockp = isBlocking_p();  // is connection blocking?
   if (!blockp) setBlocking();   // block until connection
     
   do {
      ret = readMsg(msg, flags);
   } while ( (ret!=WHOLE_MSG_OK) && (!FatalErrorStatus(ret)) );

   if (!blockp) setNonBlocking();  // restore non-blocking status if nec
   return ret;
}



int CONNECTION::readMsgIntoBuff_blk(char *permanentMsgBuff, int buffLen, long int *msgBytes) {
   int ret;
   int blockp = isBlocking_p();  // is connection blocking?
   if (!blockp) setBlocking();   // block until connection
     
   do {
      ret = readMsgIntoBuff(permanentMsgBuff, buffLen, msgBytes);
   } while ( (ret!=WHOLE_MSG_OK) && (!FatalErrorStatus(ret))
	     && (ret!=USER_BUFF_TOO_SMALL) );

   if (!blockp) setNonBlocking();  // restore non-blocking status if nec
   return ret;
}


inline int CONNECTION::convertToOldStatusCode(int newCode) {
   int oldCode;
   switch (newCode) {
      case ZERO_BYTES_READ:
      case MSG_TOO_BIG:
	 oldCode = NO_MSG_AVAIL;
	 break;
      default:
	 oldCode = newCode;
	 break;
   }
   
   return oldCode;
}


int CONNECTION::getMsg(SockMsgType *msg, int flags) {
   int newStatusCode = readMsg(msg, flags);
   return( convertToOldStatusCode(newStatusCode) );
}
   

int CONNECTION::getMsg_blk(SockMsgType *msg, int flags) {
   int newStatusCode = readMsg_blk(msg, flags);
   return( convertToOldStatusCode(newStatusCode) );
}



//                    CONNECTION::writeMsg
// Send len bytes pointed to by msg down this connection.  Return a
// status code indicating what happened (these enums are in STREAM_SOCK
// so you may have to write STREAM_SOCK::WHOLE_MSG_OK, etc):
// WHOLE_MSG_OK  :   entire msg sent;  everything is OK
// PARTIAL_MSG_OK:   part msg sent;  everything OK (see [2] below)
// TEMP_SOCK_PROBLEM: temporary problem (eg full buffer, interupted call) that _may_ fix itself
//               Fatal codes (see [1] below):
// BAD_SOCK:  Bad file descriptor for connection (probably closed conn)
// BAD_PTR_OR_ARG:  Bad ptr in sendmsg msghdr or bad msg size arg to sendmsg
// NOT_YET_CONNECTED: Possibly socket is ok, but still doing connection internal work
// CLOSED_CONN: Connection is closed
// UNKNOWN_SOCK_ERROR:  Don't know what the hell is wrong;  better bail out.

// [1]: Fatal Error codes:  STREAM_SOCK::FatalErrorStatus(status) returns true iff your
//      status is fatal;  probably all you can do is close the connection.
//
// [2]: Partial send: finish writing w/1 or more calls to 
// resumePartialWrite(). The state of a partial write is kept internally and MUST be 
// completed before initiating new writes


int CONNECTION::writeMsg(char *msg, int len) {
   if (getPartialSend() != NULL) {
      cerr << "CONNECTION::writeMsg ERROR.  Attempt to send msg before completing old one." << endl;
      exit(1);
   }

   int fd = getfd();

   int status = sendMsg(fd, msg, len,  &MsgUnixFormat, InternalHdrSpace);


   if (ErrorStatus(status)) {
     cout << "Error sending msg in CONNECTION::writeMsg." << endl;
     cout << "Status = " << status << endl;
   }
   else if (status == PARTIAL_MSG_OK) {
      // copy the remaining msg bytes into perm storage:
      int numBytesUnsent = MsgUnixFormat.msg_iov[1].iov_len;
      char *msgRemaining = (char *) MsgUnixFormat.msg_iov[1].iov_base;
      char *msgCopy = new char[numBytesUnsent];

      memcpy(msgCopy, msgRemaining, numBytesUnsent);
      MsgUnixFormat.msg_iov[1].iov_base = msgCopy;

      PartialSendMemory = msgCopy;   // save ptr in data member for later delete
      setPartialSend(&MsgUnixFormat); // keep partial state of send so we can finish later
   }

   return status;
}


int CONNECTION::writeMsgNoCopy(char *permMsgBuffer, int len, int deleteMsgBuff_p) {
   if (getPartialSend() != NULL) {
      cerr << "CONNECTION::writeMsg ERROR.  Attempt to send msg before completing old one." << endl;
      exit(1);
   }

   int fd = getfd();

   int status = sendMsg(fd, permMsgBuffer, len,  &MsgUnixFormat, InternalHdrSpace);


   if (ErrorStatus(status)) {
     cout << "Error sending msg in CONNECTION::writeMsg." << endl;
     cout << "Status = " << status << endl;
   }
   else if (status == WHOLE_MSG_OK) {
      if (deleteMsgBuff_p) {
	 delete[] permMsgBuffer;
      }
   }
   else if (status == PARTIAL_MSG_OK) {
      setPartialSend(&MsgUnixFormat);       // keep partial state of send so we can finish later
      if (deleteMsgBuff_p) {
	 PartialSendMemory = permMsgBuffer; // we'll need to delete the entire block later
      }
   }

   return status;

}

int CONNECTION::writeMsg_blk(char *msg, int len) {
   int ret;
   int blockp = isBlocking_p();  // is connection blocking?
   if (!blockp){
     setBlocking();  // block until connection
   }
     
   ret = writeMsg(msg, len);

   // Keep trying if we've got a partial send:
   while (ret == PARTIAL_MSG_OK) {
      ret = resumePartialWrite();
   }

   if (!blockp){
     setNonBlocking();  // restore non-blocking status if nec
   }
   return ret;
}




int CONNECTION::writeMsgNoCopy_blk(char *permMsgBuff, int len, int deleteMsgBuff_p) {
   int ret;
   int blockp = isBlocking_p();  // is connection blocking?
   if (!blockp) setBlocking();  // block until connection
     
   ret = writeMsgNoCopy(permMsgBuff, len, deleteMsgBuff_p);

   // Keep trying if we've got a partial send:
   while (ret == PARTIAL_MSG_OK) {
      ret = resumePartialWrite();
   }

   if (!blockp) setNonBlocking();  // restore non-blocking status if nec
   return ret;
}



// Keep this around temporarily for backward compatibility
// Returns TRUE iff were able to send entire msg.
int CONNECTION::sendMessage(char *msg, int len) {
   int status = writeMsg(msg, len);
   return( status == WHOLE_MSG_OK );
}


int CONNECTION::sendMessage_blk(char *msg, int len) {
  int status = writeMsg_blk(msg, len);
  return( status == WHOLE_MSG_OK );
}


// Pull msgs off Sendq and send down connection;  return #msgs fully sent.
int CONNECTION::sendQueuedMsgs(void) {
   int numMsgs = Sendq.length(), i, status = NO_MSGS_SENT;
   
   if (getPartialSend() != NULL){
      status = resumePartialSend();
      if (status != WHOLE_MSG_OK){
	 return status;
      }
   }
   SockMsgType *msg = getNextSend();//RVI (SockMsgType *) Sendq.pop_top();
   for(i=0; i<numMsgs; i++) {
      status = writeMsgNoCopy(msg->getMsgBuff(), msg->getMsgLength(),
			      msg->getDeleteBuff_p());
      if (status == WHOLE_MSG_OK) { /*able to send whole msg */
	 delete msg;
	 msg = getNextSend();//RVI (SockMsgType *) Sendq.pop_top();
      }
      else {
	 //cout << "sendQedMsgs return i = " << i << ";  numMsgs = " << numMsgs << endl;
	 return status;
      }
   }
   ////cout << "sendQedMsgs return numMsgs = " << numMsgs << endl;

   
   return status;
}

int CONNECTION::sendQueuedMsgs_blk(void) {
   int blockp = isBlocking_p();  // is connection blocking?
   if (!blockp) setBlocking();  // block until connection
     
   int ret = sendQueuedMsgs();

   if (!blockp) setNonBlocking();  // restore non-blocking status if nec
   return ret;
}


// Note: socket has 2 buffers, one for sending msgs, and one for
// receiving Pass arg SO_RCVBUF to get receive buffer size; SO_SNDBUF
// for send buffer
int CONNECTION::getConnBuffsize(int fd, int sendOrRecvBuff) {
   int ret, buffsize, len=sizeof buffsize;

   if ((ret=GETSOCKOPT(fd, SOL_SOCKET, sendOrRecvBuff, (char*) &buffsize, &len)) < 0) {
      perror("getsockopt failure");
      buffsize = ret;
   }

   return buffsize;
}


int CONNECTION::setConnBuffsize(int fd, int sendOrRecvBuff, int sizeInBytes) {
   int ret, len=sizeof sizeInBytes;

   if (sizeInBytes > MAX_SOCK_BUFF) {
      cerr << "setSockBuffsize:  size > MAX_SOCK_BUFF;  reducing to MAX."  << endl;
      sizeInBytes = MAX_SOCK_BUFF;
   }

   if ((ret=setsockopt(fd, SOL_SOCKET, sendOrRecvBuff, (char *) &sizeInBytes, len)) < 0) {
      perror("setsockopt failure");
   }

   return ret;
}


void CONNECTION::scheduleWrite(char *msg, int len) {
  if (outMsgsWaiting_p()) {
    char * NewMsg = new char[len];
    memcpy(NewMsg, msg, len);

    const int deleteIt_p = 1; // set flag to delete[] NewMsg later:
    SockMsgType *msgObj = new SockMsgType(NewMsg, len, deleteIt_p);
    enqueueSend(msgObj);
  }
  else
    writeMsg(msg, len);
}


void CONNECTION::scheduleWriteNoCopy(char *permMsgBuff, int len, int deleteMsg_p) {
  if (outMsgsWaiting_p()) {
    SockMsgType *msgObj = new SockMsgType(permMsgBuff, len, deleteMsg_p);
    enqueueSend(msgObj);
  }
  else
    writeMsgNoCopy(permMsgBuff, len, deleteMsg_p);
}
