#ifndef __connection__
#define __connection__

#ifndef __SPEEDES_NT__
#include <unistd.h>     // needed by netinet/in.h on HPUX
#include <sys/types.h>
#include <netinet/in.h> // needed by sys/socket.h on SunOS
#include <sys/socket.h> // needed for msghdr object in class CONNECTION
#else
#include <windows.h>
#endif


#include "SpStdlib.H"     // for exit call

#include "SpStreamSock.H"
#include "queue.H"
#include "SpOS.H"

class SERVICE_GROUP;

extern const int BOGUS_PORT;
extern const int BOGUS_FD;
extern const int BOGUS_ID;
extern const int BOGUS_MSG_LENGTH;
extern const int MAX_SOCK_BUFF;  // This is for HP;  what about others?


class CONNECTION : public STREAM_SOCK, public C_ITEM {

  public:

   /******************************************************************************
   * Method: CONNECTION constructor
   * Summary: Create CONNECTION object to mediate comms on connected
   *          socket.
   * Description: After making or accepting a connection, use this to
   *              build a CONNECTION object. Arg fd is the file des-
   *              criptor of the TCP/IP connection; port is the port
   *              number; idnum is an arbitrary int with which to
   *              identify this connection, and cgrp points to the
   *              SERVICE_GROUP that this obj belongs to (there need
   *              not be any).
   * Return: None
   ******************************************************************************/
    CONNECTION(int fd=BOGUS_FD, int port=BOGUS_PORT, int idnum=BOGUS_ID, SERVICE_GROUP *cgrp=NULL);

   /******************************************************************
   * Method: CONNECTION destructor
   * Summary: Destruct the CONNECTION.
   * Description: Returns some allocated storage.
   * Return: None
   ******************************************************************/
    ~CONNECTION();


   /*************       6/98 NOTE
     I've created a bunch of new methods whose names begin with "read" or "write" for
     receiving or sending messages, respectively.
     These are the new, PREFERRED methods for reading/writing messages and
     are intended to replace getMsg, getMsg_blk, sendMessage, sendMessage_blk,
     (which remain for backward compatibility).  The main difference is that
     they return more informative status/error codes (see enum Socket_Status_Code
     in SpStreamSock.H for status codes).

     Basically, what you want to do with a CONNECTION is read or write msgs.  Because
     of the various options available (explained below), there are several permutations:
     Blocking/Non-blocking:  does method stop (block) until read/write completes, or 
                             immediately return a result?
     Copy/NoCopy:  (for writing msgs) if we can write only part of your msg, should we
                   copy the remaining portion or just keep a ptr into your msg storage?
     Buffer Provided/Not:  some read methods will allocate buffers with "new char[msgLen]";
                           alternately, you can pass in a buffer to hold the recvd msg.
   *************/


   /******************************************************************
   * Method: readMsg
   * Summary: Read a message from connection (non-blocking); return
   *          status code.
   * Description: Does a non-blocking read of the next msg (if any)
   *              from the connection into the SockMsgType pointed to
   *              by arg, msg (basically just a char* and #bytes).
   *              Returns a status code indicating what happened. If
   *              you only get part of a message, repeatedly call
   *              readMsg until you've got the whole thing.
   * Return: returns one of the following STREAM_SOCK:: enums
   *         (see SpStreamSock.H):
   *         WHOLE_MSG_OK  :   entire msg read;  everything is OK
   *         PARTIAL_MSG_OK:   part msg read;  everything OK
   *         ZERO_BYTES_READ: read 0 bytes; if select says sock is 
   *           readable but we get 0 bytes, this means that the
   *           connection has been closed by the peer.
   *         MSG_TOO_BIG: incoming msg bigger than MAX_MSG_BYTES 
   *           (100 MEG!);  probably a mangled msg.
   *         TEMP_SOCK_PROBLEM: temporary problem (eg full buffer, 
   *           interupted call) that _may_ fix itself
   *         BAD_SOCK:  Bad file descriptor for connection (probably
   *           closed conn)
   *         BAD_PTR_OR_ARG:  Bad internal pointer
   *         NOT_YET_CONNECTED: Possibly socket is ok, but still 
   *           doing connection internal work
   *         CLOSED_CONN: Connection is closed
   *         UNKNOWN_SOCK_ERROR:  Don't know what is wrong;  better 
   *           bail out.
   ******************************************************************/
   int readMsg(SockMsgType *msg, int flags=0);


   /*********************************************************************
   * Method: readMsg_blk
   * Summary: Block until next msg read from connection; return status
   *          code.
   * Description: Does a blocking read of the next msg 
   *              from the connection into the SockMsgType pointed to
   *              by arg, msg (basically just a char* and #bytes).
   *              Returns a status code indicating what happened. 
   *              Will either get the entire message, or else
   *              return an error code.
   * Return: returns one of the following STREAM_SOCK:: enums
   *         (see SpStreamSock.H):
   *         WHOLE_MSG_OK  :   entire msg read;  everything is OK
   *         MSG_TOO_BIG: incoming msg bigger than MAX_MSG_BYTES 
   *           (100 MEG!);  probably a mangled msg.
   *         TEMP_SOCK_PROBLEM: temporary problem (eg full buffer, 
   *           interupted call) that _may_ fix itself
   *         BAD_SOCK:  Bad file descriptor for connection (probably
   *           closed conn)
   *         BAD_PTR_OR_ARG:  Bad internal pointer
   *         NOT_YET_CONNECTED: Possibly socket is ok, but still 
   *           doing connection internal work
   *         CLOSED_CONN: Connection is closed
   *         UNKNOWN_SOCK_ERROR:  Don't know what is wrong;  better 
   *           bail out.
   ******************************************************************/
   int readMsg_blk(SockMsgType *msg, int flags=0);

   // Read msg into user-provided buffer, permanentMsgBuff, return the usual status
   // codes (see readMsg), and set *msgBytes to be #bytes stored in permanentMsgBuff.
   // The provided buffer is called "permanentMsgBuff" because it should be non-stack
   // storage (in case we only get a partial read and have to finish later).
   // Arg buffLen is the size of permanentMsgBuff (in bytes);  if the incoming msg
   // is too big, return STREAM_SOCK::USER_BUFF_TOO_SMALL, and set *msgBytes equal to the
   // size of the incoming msg.  Then, the user can retrieve this msg by passing in a
   // bigger buffer, or else by calling readMsg (which will allocate a new char[N]
   // buffer).  If this returns PARTIAL_MSG_OK, 1 or more additional calls will
   // complete the operation (and return WHOLE_MSG_OK).

   /******************************************************************
   * Method: readMsgIntoBuff
   * Summary: Read next message (non-blocking) from connection into
   *          user-provided buffer; return status code.
   * Description: Read msg (non-blocking) into user-provided buffer,
   *              permanentMsgBuff, and set *msgBytes to be the number
   *              of bytes stored in permanentMsgBuff; return a status
   *              code.  Arg buffLen is the size of permanentMsgBuff
   *              (in bytes); if the incoming msg is too big, return
   *              STREAM_SOCK::USER_BUFF_TOO_SMALL, and set *msgBytes
   *              equal to the size of the incoming msg.  Then, the
   *              user can retrieve this msg by passing in a bigger
   *              buffer, or else by calling readMsg (which will
   *              allocate a new char[N] buffer).  If this returns
   *              PARTIAL_MSG_OK, do 1 or more additional calls until
   *              you've got the entire msg (return == WHOLE_MSG_OK).
   *              The provided buffer is called "permanentMsgBuff"
   *              because it must remain intact in case there is a
   *              partial receive that will be completed later; this
   *              means it should NOT be stack storage (unless you're
   *              certain you'll complete the read before exiting the
   *              function).
   * Return: returns one of the following STREAM_SOCK:: enums
   *         (see SpStreamSock.H):
   *         WHOLE_MSG_OK  :   entire msg read;  everything is OK
   *         PARTIAL_MSG_OK:   part msg read;  everything OK
   *         USER_BUFF_TOO_SMALL: buffer is too small to hold in-
   *           coming msg;  redo the call with a bigger buffer or
   *           call readMsg (which will allocate one for you).
   *         ZERO_BYTES_READ: read 0 bytes; if select says sock is 
   *           readable but we get 0 bytes, this means that the
   *           connection has been closed by the peer.
   *         MSG_TOO_BIG: incoming msg bigger than MAX_MSG_BYTES 
   *           (100 MEG!);  probably a mangled msg.
   *         TEMP_SOCK_PROBLEM: temporary problem (eg full buffer, 
   *           interupted call) that _may_ fix itself
   *         BAD_SOCK:  Bad file descriptor for connection (probably
   *           closed conn)
   *         BAD_PTR_OR_ARG:  Bad msg buff pointer
   *         NOT_YET_CONNECTED: Possibly socket is ok, but still 
   *           doing connection internal work
   *         CLOSED_CONN: Connection is closed
   *         UNKNOWN_SOCK_ERROR:  Don't know what is wrong;  better 
   *           bail out.
   ******************************************************************/
   int readMsgIntoBuff(char *permanentMsgBuff, int buffLen, long int *msgBytes);



   /******************************************************************
   * Method: readMsgIntoBuff_blk
   * Summary: Read next message (blocking) from connection into 
   *          user-provided buffer; return status code.
   * Description: Read msg (blocking) into user-provided buffer,
   *              permanentMsgBuff, and set *msgBytes to be the number
   *              of bytes stored in permanentMsgBuff; return a status
   *              code.  Arg buffLen is the size of permanentMsgBuff
   *              (in bytes); if the incoming msg is too big, return
   *              STREAM_SOCK::USER_BUFF_TOO_SMALL, and set *msgBytes
   *              equal to the size of the incoming msg.  Then, the
   *              user can retrieve this msg by passing in a bigger
   *              buffer, or else by calling readMsg (which will
   *              allocate a new char[N] buffer).  If this returns
   *              PARTIAL_MSG_OK, do 1 or more additional calls until
   *              you've got the entire msg (return == WHOLE_MSG_OK).
   *              The provided buffer is called "permanentMsgBuff"
   *              because it must remain intact in case there is a
   *              partial receive that will be completed later; this
   *              means it should NOT be stack storage (unless you're
   *              certain you'll complete the read before exiting the
   *              function).
   * Return: returns one of the following STREAM_SOCK:: enums
   *         (see SpStreamSock.H):
   *         WHOLE_MSG_OK  :   entire msg read;  everything is OK
   *         USER_BUFF_TOO_SMALL: buffer is too small to hold in-
   *           coming msg;  redo the call with a bigger buffer or
   *           call readMsg (which will allocate one for you).
   *         ZERO_BYTES_READ: read 0 bytes; if select says sock is 
   *           readable but we get 0 bytes, this means that the
   *           connection has been closed by the peer.
   *         MSG_TOO_BIG: incoming msg bigger than MAX_MSG_BYTES 
   *           (100 MEG!);  probably a mangled msg.
   *         TEMP_SOCK_PROBLEM: temporary problem (eg full buffer, 
   *           interupted call) that _may_ fix itself
   *         BAD_SOCK:  Bad file descriptor for connection (probably
   *           closed conn)
   *         BAD_PTR_OR_ARG:  Bad msg buff pointer
   *         NOT_YET_CONNECTED: Possibly socket is ok, but still 
   *           doing connection internal work
   *         CLOSED_CONN: Connection is closed
   *         UNKNOWN_SOCK_ERROR:  Don't know what is wrong;  better 
   *           bail out.
   ******************************************************************/
   int readMsgIntoBuff_blk(char *permanentMsgBuff, int buffLen, long int *msgBytes);



   /******************************************************************
   * Method: writeMsg
   * Summary: Send len bytes pointed to by arg, msgBuff, to peer
   *          (non-blocking); return status code.
   * Description: Send len bytes pointed to by arg, msgBuff, down
   *              connection (non-blocking); return status code.  If
   *              you get a partial write (returns PARTIAL_MSG_OK),
   *              the state of the operation is kept internally and
   *              must be completed before beginning another write;
   *              call resumePartialwrite 1 or more times to finish a
   *              partial write.
   * Return: returns one of the following STREAM_SOCK:: enums
   *         (see SpStreamSock.H):
   * WHOLE_MSG_OK  :  entire msg sent;  everything is OK
   * PARTIAL_MSG_OK:  part msg sent;  everything OK. Complete this
   *                  operation with calls to resumePartialwrite.
   * TEMP_SOCK_PROBLEM: temporary problem (eg full buffer, interupted 
   *                    call) that _may_ fix itself
   * BAD_SOCK:  Bad file descriptor for connection (probably closed 
   *            connection)
   * BAD_PTR_OR_ARG:  Bad pointer or msg length arg
   * NOT_YET_CONNECTED: Possibly socket is ok, but still doing 
   *                    connection internal work
   * CLOSED_CONN: Connection is closed
   * UNKNOWN_SOCK_ERROR:  Don't know what is wrong;  better quit
   ******************************************************************/
   int writeMsg(char *msg, int len);

   /******************************************************************
   * Method: writeMsgNoCopy
   * Summary: Send len bytes pointed to by arg, permMsgBuff, to peer
   *          (non-blocking); return status code.  In case of partial
   *          write, do NOT copy remainder from permMsgBuff.
   * Description: Send len bytes pointed to by arg, msgBuff, to
   *              peer (non-blocking); return status code.  If you
   *              get a partial write (returns PARTIAL_MSG_OK), the
   *              state of the operation is kept internally; in this
   *              case we will NOT copy the remaining bytes, but will
   *              just keep a pointer into permMsgBuff until the
   *              write can be finished (with 1 or more calls to
   *              resumePartialWrite), so make sure permMsgBuff points
   *              to permanent storage.  The arg, deletePermMsg_p, if
   *              true, means that this method will delete the storage
   *              when done with it using :
   *              delete[] (char *) permMsgBuff;
   *              So permMsgBuff must be allocated as new char[N] if
   *              you want to use this option;  otherwise delete it
   *              yourself.
   * Return: returns one of the following STREAM_SOCK:: enums
   *         (see SpStreamSock.H):
   * WHOLE_MSG_OK  :  entire msg sent;  everything is OK
   * PARTIAL_MSG_OK:  part msg sent;  everything OK. Complete this
   *                  operation with calls to resumePartialWrite.
   * TEMP_SOCK_PROBLEM: temporary problem (eg full buffer, interupted 
   *                    call) that _may_ fix itself
   * BAD_SOCK:  Bad file descriptor for connection (probably closed 
   *            connection)
   * BAD_PTR_OR_ARG:  Bad pointer or msg length arg
   * NOT_YET_CONNECTED: Possibly socket is ok, but still doing 
   *                    connection internal work
   * CLOSED_CONN: Connection is closed
   * UNKNOWN_SOCK_ERROR:  Don't know what is wrong;  better quit
   ******************************************************************/
   int writeMsgNoCopy(char *permMsgBuff, int len, int deletePermMsg_p);




   /******************************************************************
   * Method: writeMsg_blk
   * Summary: Send len bytes pointed to by arg, msgBuff, to peer
   *          (blocking); return status code.
   * Description: Send len bytes pointed to by arg, msgBuff, to
   *              peer (blocking); return status code.
   * Return: returns one of the following STREAM_SOCK:: enums
   *         (see SpStreamSock.H):
   * WHOLE_MSG_OK  :  entire msg sent;  everything is OK
   * TEMP_SOCK_PROBLEM: temporary problem (eg full buffer, interupted 
   *                    call) that _may_ fix itself
   * BAD_SOCK:  Bad file descriptor for connection (probably closed 
   *            connection)
   * BAD_PTR_OR_ARG:  Bad pointer or msg length arg
   * NOT_YET_CONNECTED: Possibly socket is ok, but still doing 
   *                    connection internal work
   * CLOSED_CONN: Connection is closed
   * UNKNOWN_SOCK_ERROR:  Don't know what is wrong;  better quit
   ******************************************************************/
   int writeMsg_blk(char *msg, int len);

   /******************************************************************
   * Method: writeMsgNoCopy_blk
   * Summary: Send len bytes pointed to by arg, permMsgBuff, to peer
   *          (blocking); return status code.  In case of partial
   *          write, do NOT copy remainder from permMsgBuff.
   * Description: Send len bytes pointed to by arg, msgBuff, to
   *              peer (non-blocking); return status code. Although
   *              a blocking call cannot return a partial write, this
   *              can happen internally. In such a case,
   *              we will NOT copy the remaining bytes, but will
   *              just keep a pointer into permMsgBuff until the
   *              write can be finished (with 1 or more calls to
   *              resumePartialWrite), so make sure permMsgBuff points
   *              to permanent storage.  The arg, deletePermMsg_p, if
   *              true, means that this method will delete the storage
   *              when done with it using :
   *              delete[] (char *) permMsgBuff;
   *              So permMsgBuff must be allocated as new char[N] if
   *              you want to use this option;  otherwise delete it
   *              yourself.
   * Return: returns one of the following STREAM_SOCK:: enums
   *         (see SpStreamSock.H):
   * WHOLE_MSG_OK  :  entire msg sent;  everything is OK
   * TEMP_SOCK_PROBLEM: temporary problem (eg full buffer, interupted 
   *                    call) that _may_ fix itself
   * BAD_SOCK:  Bad file descriptor for connection (probably closed 
   *            connection)
   * BAD_PTR_OR_ARG:  Bad pointer or msg length arg
   * NOT_YET_CONNECTED: Possibly socket is ok, but still doing 
   *                    connection internal work
   * CLOSED_CONN: Connection is closed
   * UNKNOWN_SOCK_ERROR:  Don't know what is wrong;  better quit
   ******************************************************************/
   int writeMsgNoCopy_blk(char *permMsgBuff, int len, int deletePermMsg_p);



   /******************************************************************
   * Method: resumePartialWrite
   * Summary: Try to finish writing the last message (which resulted
   *          in a partial write);  return status code.
   * Description: resumePartialWrite attempts to complete the last
   *              partial write operation, and returns a status code.
   *              A call to this method may itself result in a part-
   *              ial send, so it may take more than 1 call to get
   *              the whole message out.  Since the state of a partial
   *              send is recorded internally, no args are needed.
   *              Note the asymmetry between the reads and the 
   *              writes in that there is NO analogous "resumeRead"
   *              method.  This method was necessary to avoid compli-
   *              cations resulting from trying to write a new msg
   *              when there is a pending partial write.
   * Return: returns one of the following STREAM_SOCK:: enums
   *         (see SpStreamSock.H):
   * WHOLE_MSG_OK  :  entire msg sent;  everything is OK
   * PARTIAL_MSG_OK:  part msg sent;  everything OK. Complete this
   *                  operation with calls to resumePartialwrite.
   * TEMP_SOCK_PROBLEM: temporary problem (eg full buffer, interupted 
   *                    call) that _may_ fix itself
   * BAD_SOCK:  Bad file descriptor for connection (probably closed 
   *            connection)
   * BAD_PTR_OR_ARG:  Bad pointer or msg length arg
   * NOT_YET_CONNECTED: Possibly socket is ok, but still doing 
   *                    connection internal work
   * CLOSED_CONN: Connection is closed
   * UNKNOWN_SOCK_ERROR:  Don't know what is wrong;  better quit
   ******************************************************************/
   int resumePartialWrite(void){
      msghdr *mh = getPartialSend();
      if (mh==NULL) {
	 cerr << "Attempt to resume partial send when no partial state stored." << endl;
	 exit(1);
      }

      int ret = STREAM_SOCK::resumePartialSend(getfd(), mh);
      if (ret == WHOLE_MSG_OK) {
	 setPartialSend((msghdr*)NULL);
	 if (PartialSendMemory!=NULL) {
	    delete[] PartialSendMemory;
	    PartialSendMemory = NULL;
	 }
      }

      return ret;
   }



   /******************************************************************
   * Method: resumePartialWrite_blk
   * Summary: Block until finished writing the last, partially-sent
   *          message;  return status code.
   * Description: resumePartialWrite_blk attempts to complete the last
   *              partial write operation, and returns a status code.
   *              Since the state of a partial
   *              send is recorded internally, no args are needed.
   * Return: returns one of the following STREAM_SOCK:: enums
   *         (see SpStreamSock.H):
   * WHOLE_MSG_OK  :  entire msg sent;  everything is OK
   * TEMP_SOCK_PROBLEM: temporary problem (eg full buffer, interupted 
   *                    call) that _may_ fix itself
   * BAD_SOCK:  Bad file descriptor for connection (probably closed 
   *            connection)
   * BAD_PTR_OR_ARG:  Bad pointer or msg length arg
   * NOT_YET_CONNECTED: Possibly socket is ok, but still doing 
   *                    connection internal work
   * CLOSED_CONN: Connection is closed
   * UNKNOWN_SOCK_ERROR:  Don't know what is wrong;  better quit
   ******************************************************************/
  int resumePartialWrite_blk(void) {
    int ret;
    int blockp = isBlocking_p();  // is connection blocking?
    if (!blockp) setBlocking();   // set to block

    do {
      ret = resumePartialWrite();
    } while (ret==PARTIAL_MSG_OK);

    if (!blockp) setNonBlocking(); // resume non-blocking, if nec
    return ret;
  }





   /******************************************************************
   * Method: scheduleWrite
   * Summary: Put msg on a queue to be sent out later using method
   *          sendQueuedMsgs;  if nothing waiting, send immediately.
   * Description: Use this method to queue-up outgoing messages until
   *              it is convenient to send them with sendQueuedMsgs.
   *              If the queue is empty and there is no pending
   *              partial write, just send msg immediately.  Queued
   *              msgs are COPIED so you may delete msg right after
   *              the call. To avoid copying, see scheduleWriteNoCopy.
   * Return: None.
   ******************************************************************/
   void scheduleWrite(char *msg, int len);


   /******************************************************************
   * Method: scheduleWriteNoCopy
   * Summary: Put msg on a queue to be sent out later using method
   *          sendQueuedMsgs;  if nothing waiting, send immediately.
   * Description: Use this method to queue-up outgoing messages until
   *              it is convenient to send them with sendQueuedMsgs.
   *              If the queue is empty and there is no pending
   *              partial write, just send msg immediately.  Queued
   *              msgs are NOT COPIED so make sure permMsgBuff
   *              points to "permanent" storage (ie will not get
   *              trashed before sendQueuedMsgs is called).
   *              The arg, deleteMsg_p, when true, means that your
   *              msg will be deleted when we're done with it (so
   *              make sure permMsgBuff was allocated as
   *              permMsgBuff = new char[len]).  Otherwise, you must
   *              delete it yourself.
   * Return: None.
   ******************************************************************/
   void scheduleWriteNoCopy(char *permMsgBuff, int len, int deleteMsg_p);


   /******************************************************************
   * Method: scheduleSend
   * Summary: Same as scheduleWrite
   * Description: Old name for scheduleWrite (kept around for back-
   *              wards compatibility).
   * Return: None
   ******************************************************************/
   void scheduleSend(char *msg, int len) {
      scheduleWrite(msg, len);
   }


   /******************************************************************
   * Method: sendQueuedMsgs
   * Summary: Finish partial write (if nec) and try to send all queued
   *          msgs (non-blocking);  return status code.
   * Description: First tries to complete a partial write, if there is
   *              one, and then attempts to write all enqueued msgs.
   *              Since this is non-blocking, it will return
   *              immediately if a resource is blocked (eg full unix
   *              buffer), and therefore not necessarily send out all
   *              queued msgs.
   * Return: returns one of the following STREAM_SOCK:: enums
   *         (see SpStreamSock.H):
   * WHOLE_MSG_OK  :  entire _queue_ sent;  everything is OK
   * PARTIAL_MSG_OK:  part of _queue_ sent;  everything OK.
   * TEMP_SOCK_PROBLEM: temporary problem (eg full buffer, interupted 
   *                    call) that _may_ fix itself
   * BAD_SOCK:  Bad file descriptor for connection (probably closed 
   *            connection)
   * BAD_PTR_OR_ARG:  Bad pointer or msg length arg
   * NOT_YET_CONNECTED: Possibly socket is ok, but still doing 
   *                    connection internal work
   * CLOSED_CONN: Connection is closed
   * UNKNOWN_SOCK_ERROR:  Don't know what is wrong;  better quit
   ******************************************************************/
   int sendQueuedMsgs(void);


   /******************************************************************
   * Method: sendQueuedMsgs_blk
   * Summary: Finish partial write (if nec) and try to send all queued
   *          msgs (blocking);  return status code.
   * Description: First tries to complete a partial write, if there is
   *              one, and then attempts to write all enqueued msgs.
   *              Since this is non-blocking, it will either
   *              send the entire queue or else return and error
   *              code.
   * Return: returns one of the following STREAM_SOCK:: enums
   *         (see SpStreamSock.H):
   * WHOLE_MSG_OK  :  entire _queue_ sent;  everything is OK
   * TEMP_SOCK_PROBLEM: temporary problem (eg full buffer, interupted 
   *                    call) that _may_ fix itself
   * BAD_SOCK:  Bad file descriptor for connection (probably closed 
   *            connection)
   * BAD_PTR_OR_ARG:  Bad pointer or msg length arg
   * NOT_YET_CONNECTED: Possibly socket is ok, but still doing 
   *                    connection internal work
   * CLOSED_CONN: Connection is closed
   * UNKNOWN_SOCK_ERROR:  Don't know what is wrong;  better quit
   ******************************************************************/
   int sendQueuedMsgs_blk(void);


   /******************************************************************
   * Method: outMsgsWaiting_p
   * Summary: Return true if any queued msgs waiting or if there is a
   *          partial write pending;  otherwise return false.
   * Description: Return true if any queued msgs waiting or if there
   *              is a partial write pending;  otherwise return false.
   * Return: true or false.
   ******************************************************************/
   int outMsgsWaiting_p(void) {
      return( (getPartialSend() != NULL) || (Sendq.get_length() > 0) );
   }



   // Read next message from this connection and return thru arg, msg, if able
   // to read whole thing.  The returned val indicates the status of the call:
   // WHOLE_MSG_OK  :  read entire msg;
   // PARTIAL_MSG_OK:  read part of msg (rest will be read in subsequent getMsg call(s);
   // NO_MSG_AVAIL  :  couldn't even get whole internal header (usually means 0 bytes ready at
   //                  socket so it may indicate a problem if msg expected)
   // CLOSED_CONN   :  other side has closed the connection.

   int getMsg(SockMsgType *msg, int flags=0);  

   // Blocking version of getMsg:

   /******************************************************************
   * Method: 
   * Summary: 
   * Description: 
   * Return: 
   ******************************************************************/
   int getMsg_blk(SockMsgType *msg, int flags=0);



   /******************************************************************
   * Method: 
   * Summary: 
   * Description: 
   * Return: 
   ******************************************************************/

   // Send len bytes pointed to by msg down this connection.  Return true if entire
   // msg sent;  otherwise false.  If the entire msg has not been sent, the partial
   // state is stored internally and can be sent using resumePartialSend (in this case
   // the unsent part of msg will be copied, so msg can be safely deleted or allocated
   // on the stack).  It is
   // an error to attempt another sendMessage before completing the last one.
   int sendMessage(char *msg, int len);

   // Blocking version of sendMessage:
   int sendMessage_blk(char *msg, int len);


   // Try to complete sending of last partial send.  Return true if rest of msg has
   // been sent;  false otherwise.  Use this fcn repeatedly until entire msg is sent.
   int resumePartialSend(void) {
      return resumePartialWrite();
   }





   void setPartialSend(msghdr *partMsg) { PartialSend = partMsg; }
   msghdr *getPartialSend(void) { return PartialSend; }

   int havePartialWrite_p(void) {
      return( PartialSend != NULL );
   }


   void setPartialRecv(SockMsgType *partMsg) { PartialRecv = *partMsg; }
   SockMsgType *getPartialRecv(void) { return &PartialRecv; }

   SERVICE_GROUP *getConnGroup(void) { return MyGroup; }
   void setConnGroup(SERVICE_GROUP *cgroup) { MyGroup = cgroup; }

   void setNoPartialRecv(void) {
      SockMsgType *partRecv = getPartialRecv();
      partRecv->setMsgLength(BOGUS_MSG_LENGTH);
   }

   int havePartialRecv_p(void) {
      SockMsgType *partRecv = getPartialRecv();
      return( partRecv->getMsgLength() != BOGUS_MSG_LENGTH );
   }

   void resetConsecutiveTempErrors(void) {
      ConsecutiveTempErrors = 0;
   }
   
   int getConsecutiveTempErrors(void) {
      return ConsecutiveTempErrors;
   }

   int incrementConsecutiveTempErrors(void) {
      ConsecutiveTempErrors++;
      return ConsecutiveTempErrors;
   }
      

   SockMsgType *getNextSend(void) { return (SockMsgType *) Sendq.pop_top(); }

   void enqueueSend(SockMsgType *msg) { Sendq.push_bot(msg); }



  int NumQueuedMsgs(void){
    return ((getPartialSend() != NULL) || (Sendq.length()));
  }

   void closeConnection(void);


   int getSendBuffsize(void) {
      return getConnBuffsize(getfd(), SO_SNDBUF);
   }

   int getRecvBuffsize(void) {
      return getConnBuffsize(getfd(), SO_RCVBUF);
   }


   //Set unix internal send buff;  return 0 if ok
   int setSendBuffsize(int sizeInBytes) {
      return setConnBuffsize(getfd(), SO_SNDBUF, sizeInBytes);
   }

   //Set unix internal recv buff;  return 0 if ok
   int setRecvBuffsize(int sizeInBytes) {
      return setConnBuffsize(getfd(), SO_RCVBUF, sizeInBytes);
   }

   //Set unix send/recv buffers to max;  return 0 if ok
   int setMaxUnixBuffers(void) {
      return( setSendBuffsize(MAX_SOCK_BUFF)||setRecvBuffsize(MAX_SOCK_BUFF) );
   }

   const char *get_host_name() const { return host_name; }

   void markSocketAsOtherHungUp(){ setSocketStatus(OTHER_HUNGUP);};


  private:
   char *host_name;
   C_QUEUE Sendq;

   msghdr MsgUnixFormat;   // Perm storage of Unix msg struct (avoid new and delete)
   iovec PairVec[2];       // list of (char* length) pairs for MsgUnixFormat
   long InternalHdrSpace[2]; // place for internal msg header

   void initializeUnix_msghdr(void) {  // way to initialize MsgUnixFormat
      MsgUnixFormat.msg_name = NULL;
      MsgUnixFormat.msg_namelen = 0;
#ifdef Linux
      MsgUnixFormat.msg_control = NULL;
      MsgUnixFormat.msg_controllen = 0;
      MsgUnixFormat.msg_flags = 0;
#else
      MsgUnixFormat.msg_accrights = NULL;
      MsgUnixFormat.msg_accrightslen = 0;
#endif

      // Since a connection can send only ONE msg at a time, we only need space
      // for one iovec and one internal header string.  Thus, we only need to allocate once:
      PairVec[0].iov_base = (char *) InternalHdrSpace;
      PairVec[0].iov_len = HEADER_SIZE;

      MsgUnixFormat.msg_iovlen = 2;
      MsgUnixFormat.msg_iov = PairVec;
   }

   
   msghdr *PartialSend;       // Place to keep state of a partial send
   SockMsgType PartialRecv;   // Place to keep state of a partial receive


   // If we have to copy the remainder of a partial send, store the pointer here
   // so we can delete the memory later:
   char *PartialSendMemory;

   int ConsecutiveTempErrors;

   SERVICE_GROUP *MyGroup; // circular reference--group containing this connection

   int getConnBuffsize(int fd, int sendOrRecvBuff);
   int setConnBuffsize(int fd, int sendOrRecvBuff, int sizeInBytes);

  //virtual void setfd(int in) { STREAM_SOCK::setfd(in); init_host_name(); }
 
   void init_host_name();

   // Covert new status code to old one for backward compatiblity:
   int convertToOldStatusCode(int newCode);



};

#endif
