#include "SpAssert.H"
#ifndef __SPEEDES_NT__
#include <termios.h>
#else
#include <windows.h>
#endif
#include "SpStdlib.H"
#include "SpErrno.H"
#include "SpIostream.H"
#include "SpFstream.H"
#include "SpStrstream.H"
#include "SpIomanip.H"
#include "SpMsgHeader.H"
#include "SpMsgServer.H"
#include "SpSignal.H"


// max # consecutive temporary errors before closing connection
#define MAX_CONSECUTIVE_TEMP_ERRORS (3)

//#define DEBUG_SEND_DELAY

#ifdef DEBUG_SEND_DELAY

#include <sys/time.h>
#include <unistd.h>
#include "SpStdio.H"

inline void PrintMessageSendTime(int MsgID){
  static struct timeval tm;
  static struct timezone tz;
  static FILE * DebugSendData = fopen("/tmp/ProxDebugSendData_Server.txt","w");

  if (MsgID){
    gettimeofday(&tm, &tz);
    double Time = double(tm.tv_sec)+double(tm.tv_usec)/1000000;
    fprintf(DebugSendData, "%d %40.60e\n", MsgID, Time);
  }
}

#endif

const int MESSAGE_SERVER::DEFAULT_GROUP_ID = 0;

// Arg ignoreSIGPIPE: if you don't know what a SIGPIPE signal is, just
// let this arg default.  The arg, ignoreSIGPIPE, when true, causes
// Unix SIGPIPE signals to be ignored (this is the default action).
// This means that attempting to write on a connection whose peer has
// died will NOT crash your program.  If you want your program to
// crash, or want to insert your own handler, pass 0 in when you
// construct a

MESSAGE_SERVER::MESSAGE_SERVER(int num_server_types, int port, int ignoreSIGPIPE)
  : NumServerTypes(num_server_types),
    SrvMakerArr(new SERVICE_OBJ_MAKER[num_server_types]),
    DefaultServiceGrp(DEFAULT_GROUP_ID),
    SERVER_SOCK(port)
{
  // "ignore" unix signal (ie don't get killed) that occurs if we
  //  attempt to send down a connection whose other end has died.
  //  We'll get an (errno) error code instead and handle it.
#ifndef __SPEEDES_NT__
  if (ignoreSIGPIPE)
    signal(SIGPIPE, SIG_IGN);
#endif
}

MESSAGE_SERVER::~MESSAGE_SERVER()
{
  delete [] SrvMakerArr;
}


void MESSAGE_SERVER::handleBadRecv(int status, CONNECTION *conn) {
  switch (status) {
     case TEMP_SOCK_PROBLEM:
     case BAD_PTR_OR_ARG:
     case NOT_YET_CONNECTED:
	perror("MESSAGE_SERVER::handleBadRecv");
	if (conn->incrementConsecutiveTempErrors() > MAX_CONSECUTIVE_TEMP_ERRORS) {
	   cout << "handleBadRecv:  too many temp errors;  closing connection. (status = "
		<< status << ")" << endl;
	   closeConnection(conn);
	}
	break;

     case CLOSED_CONN:
	closeConnection(conn);
	break;

     case BAD_INTERNAL_MSG:
     case BAD_SOCK:
     case ZERO_BYTES_READ:
     case UNKNOWN_SOCK_ERROR:
     case MSG_TOO_BIG:
	cout << "handleBadRecv has fatal error;  have to close connection. (status="
	     << status << ")" << endl;
	cout << "connectionID = " << conn->get_id() << endl;
	perror("MESSAGE_SERVER::handleBadRecv");
	closeConnection(conn);
	break;

     default:
	cout << "Unknown bad status = " << status << endl;
	break;
  }
}
	

void MESSAGE_SERVER::handleBadSend(int status, CONNECTION *conn) {
  switch (status) {
     case TEMP_SOCK_PROBLEM:
     case BAD_PTR_OR_ARG:
     case NOT_YET_CONNECTED:
	perror("MESSAGE_SERVER::handleBadSend");
	if (conn->incrementConsecutiveTempErrors() > MAX_CONSECUTIVE_TEMP_ERRORS) {
	   cout << "handleBadSend:  too many temp errors;  closing connection. (status = "
		<< status << ")" << endl;
	   closeConnection(conn);
	}
	break;

     case CLOSED_CONN:
	closeConnection(conn);
	break;

     case BAD_SOCK:
     case ZERO_BYTES_READ:
     case UNKNOWN_SOCK_ERROR:
     case MSG_TOO_BIG:
	cout << "handleBadSend has fatal error;  have to close connection. (status="
	     << status << ")" << endl;
	cout << "connectionID = " << conn->get_id() << endl;
	perror("MESSAGE_SERVER::handleBadSend");
	closeConnection(conn);
	break;

     default:
	cout << "Unknown bad status = " << status << endl;
	break;
  }
}

void MESSAGE_SERVER::closeConnection(CONNECTION *conn) {
   // Clear out bits in fd sets so select won't check for closed connection:
   SERVER_SOCK::closeConnection(conn->getfd());

   SERVICE_GROUP *grp = conn->getConnGroup();
   SERVICE_GROUP::HandlerStatus status = grp->ConnectionHasBeenClosed(conn);

   if (status == SERVICE_GROUP::delete_this_group) {
      dequeueGroup(grp);
      delete grp;
   }
}
	   
void MESSAGE_SERVER::clearConnInFdSets(CONNECTION *conn)
{
  int fd = conn->getfd();

  FD_CLR(fd, &Readfds);
  FD_CLR(fd, &Writefds);
  FD_CLR(fd, &Exceptionfds);
}

void MESSAGE_SERVER::dequeueGroup(SERVICE_GROUP *grp)
{
  for(CONNECTION *conn = grp->getFirstConnection();
      (conn != NULL);
      conn = grp->getNextConnection(conn)) {
    // Clear out bits in fd sets so select won't check for closed connection:
    SERVER_SOCK::closeConnection(conn->getfd());
  }


  SrvGroupList.deque(grp);
}


SERVICE_GROUP *MESSAGE_SERVER::moveConnIntoGroup(CONNECTION *conn, SockMsgType& msg) {
   GENERIC_MSG_HEADER *hdr = (GENERIC_MSG_HEADER *) msg.getMsgBuff();
   SERVICE_GROUP *oldgrp = conn->getConnGroup();
   oldgrp->dequeueConnection(conn);

   int serviceID = hdr->getSrvType();
   int groupID = hdr->getGroupNum();

   SERVICE_GROUP *newgrp = findGroup(serviceID, groupID);


   // Create a new group if this one does not exist:
   if (newgrp == NULL) {
      int srvType = hdr->getSrvType();
      newgrp = makeNewServiceGroup(srvType, groupID);
      addNewGroup(newgrp);
      cout << "Added new service group id = " << groupID << endl;
   }

   newgrp->addConnection(conn);
   return newgrp;
}

SERVICE_GROUP *MESSAGE_SERVER::makeNewServiceGroup(int serviceType, int groupID) {
   if ((serviceType < 0) || (serviceType >= NumServerTypes)) {
      cerr << "Bad service type# in msg; serviceType, groupID = " << serviceType
	   << ", " << groupID << endl;
      exit(1);
      return NULL; // just to stop compiler warning
   }
   else {
      SERVICE_OBJ_MAKER srvMaker = getSrvMaker(serviceType);
      SERVICE_GROUP *srvGrp = srvMaker();

      srvGrp->ServiceTypeID = serviceType;
      srvGrp->GroupID = groupID;

      return srvGrp;
   }
}
   

SERVICE_GROUP *MESSAGE_SERVER::findGroup(int serviceID, int groupID) {
   SERVICE_GROUP *group = getFirstGroup();
   while (group != NULL) {
      if ((group->ServiceTypeID==serviceID) && (group->GroupID==groupID))
	return group;
      else
	group = getNextGroup(group);
   }

   return NULL;
}

inline void MESSAGE_SERVER::HandleMsgsFromAllGroups()
{
  SockMsgType msg;
  SERVICE_GROUP *group = getFirstGroup();
  while (group != NULL) {
    CONNECTION *conn = group->getFirstConnection();
    while (conn != NULL) {
       if (readyToRecv(conn)) {
	int status = conn->readMsg(&msg);
	if (OKStatus(status)) {
	   if (status == WHOLE_MSG_OK) {
#ifdef DEBUG_SEND_DELAY
	      PrintMessageSendTime(((COMM_SERVER_HEADER*)msg.getMsgBuff())
				   ->getUnused());
#endif
	      group->callHandler(conn, msg);
	      if (conn->NumQueuedMsgs()){
		 FD_SET(conn->getfd(), &Writefds);
	      }
	   }
	   conn->resetConsecutiveTempErrors(); //#consec temporary sock errs = 0
	}
	else {  // ERROR STATUS:
	   handleBadRecv(status, conn);
	   return;
	}
      }
      conn = group->getNextConnection(conn);
    }
    group = getNextGroup(group);
  }
}

inline void MESSAGE_SERVER::HandleMsgsFromNewConnections()
{
  // Handle msgs from new connections (which are stored in a default
  // connection group until a msg arrives that tells which group it
  // belongs in):
  CONNECTION *conn = getDefaultServiceGrp()->getFirstConnection();
  while (conn != NULL) {
    CONNECTION *nextConn = getDefaultServiceGrp()->getNextConnection(conn);// follow link b4 we move conn @ ** below
    if (readyToRecv(conn)) {
      SockMsgType msg;
      int status = conn->readMsg(&msg);
      if (status == WHOLE_MSG_OK) {
	// put conn into group;  if group doesn't exist, create it:
	SERVICE_GROUP *group = moveConnIntoGroup(conn, msg);  //<-- move conn to new list **
	group->callHandler(conn, msg);
      }
      else if (ErrorStatus(status)) {
	handleBadRecv(status, conn);
      }
    }
    conn = nextConn;
  }
}


inline void MESSAGE_SERVER::SetWritefds(void)
{
  SERVICE_GROUP *group = getFirstGroup();
  CONNECTION * conn;
  while (group != NULL) {
    conn = group->getFirstConnection();
    while (conn != NULL){
      if (conn->NumQueuedMsgs()){
	FD_SET(conn->getfd(), &Writefds);
      }
      conn = group->getNextConnection(conn);
    }
    group = getNextGroup(group);
  }
}



void MESSAGE_SERVER::serviceMessages(void) {
   SERVICE_GROUP *group;//, *defaultGroup =
   getDefaultServiceGrp();
   CONNECTION *conn;
   //float timeoutSecs = 2.0;
   int numEvents;

   int failed_p;
   int TimeOut, status;
#ifndef __SPEEDES_NT__
   termios termios_p;


   failed_p = tcgetattr(STDIN_FILENO, &termios_p);
   if (!failed_p)
   {
     termios_p.c_lflag &= ~ICANON & ~ECHO;
     termios_p.c_cc[VMIN] = 0;
     termios_p.c_cc[VTIME] = 0;
  
     failed_p = tcsetattr(STDIN_FILENO, TCSANOW, &termios_p);
     assert(!failed_p);
 
     FD_SET(STDIN_FILENO, &Readfds);
   }
   cout << "Press a key to exit." << endl;
#else
   cout << "Press control-C to exit." << endl;
#endif

   for (;;)
   {
      TimeOut = 3600; // timeout set at an hour
      numEvents = checkForConnectionReadsAndWrites(1, TimeOut);

      if ((numEvents > 0)) {
#ifndef __SPEEDES_NT__
	if (FD_ISSET(STDIN_FILENO, &Copy_readfds))
	{
	  char c;
	  //const int num_read =
	  read(STDIN_FILENO, &c, 1);
	  // assert(num_read == 1);
	  return;
	}
#endif
	
	int numConnects=getAllWaitingConnections(); // accept connects & insert into Default grp
	if (numConnects != 0){
	  cout << "Just got " << numConnects << " connections..." << endl;
	}

	HandleMsgsFromAllGroups();

	HandleMsgsFromNewConnections();

	SetWritefds();
	
	 // Send msgs from all groups:
	 group = getFirstGroup();
	 while (group != NULL) {
	    conn = group->getFirstConnection();
	    while (conn != NULL) {
	       if (readyToSend(conn)) {
		 status = conn->sendQueuedMsgs();
		 if (ErrorStatus(status)) {
		    handleBadSend(status, conn);
		 }

		 if (!conn->outMsgsWaiting_p()) {
		    dontCheckIfWritableInSelect(conn);
		 }
	       }
	       conn = group->getNextConnection(conn);
	    }
	    group = getNextGroup(group);
	 }
      }

   }

}

