#ifndef Message_Server_Class
#define Message_Server_Class

#include "SpServerSock.H"
#include "SpServiceGroup.H"

#include <sys/types.h>

#ifndef __SPEEDES_NT__
#include <unistd.h> // needed by netinet/in.h on HPUX
#include <netinet/in.h> // needed by bindSock's default argument
#else
#include <windows.h>
#endif
#include "SpTime.H"

typedef SERVICE_GROUP* (*SERVICE_OBJ_MAKER)(void);



class MESSAGE_SERVER : public SERVER_SOCK {
 private:
   const int NumServerTypes;
   SERVICE_OBJ_MAKER * const SrvMakerArr;

   // List of connection-groups.  The 1st msg from a new connection will specify
   // goup# and connection# within the group:
   C_QUEUE SrvGroupList; // List of SERVICE_GROUP *

   // Put new connections here, unless otherwise specified:
   SERVICE_GROUP DefaultServiceGrp;

   inline void HandleMsgsFromAllGroups();
   inline void HandleMsgsFromNewConnections();
    
 public:
   static const int DEFAULT_GROUP_ID;

   // Arg ignoreSIGPIPE:  if you don't know what a SIGPIPE signal is, just let this arg
   // default.
   // The arg, ignoreSIGPIPE, when true, causes Unix SIGPIPE signals to be ignored (this
   // is the default action).  This means that attempting to write on a connection whose
   // peer has died will NOT crash your program.  If you want your program to crash, or
   // want to insert your own handler, pass 0 in when you construct a MESSAGE_SERVER.
   MESSAGE_SERVER(int num_server_types, int port, int ignoreSIGPIPE=1);

   ~MESSAGE_SERVER();

   SERVICE_OBJ_MAKER getSrvMaker(int srvType) {
      return SrvMakerArr[srvType];
   }

   void saveSrvMaker(int srvType, SERVICE_OBJ_MAKER srvMaker) {
      SrvMakerArr[srvType] = srvMaker;
   }

   SERVICE_GROUP *makeNewServiceGroup(int serviceType, int groupID);


   SERVICE_GROUP *getDefaultServiceGrp(void) { return &DefaultServiceGrp; }

   int acceptConnectionIntoGroup(SERVICE_GROUP *cgroup = NULL) {
      if (cgroup==NULL) cgroup = getDefaultServiceGrp(); // <-- use default group if none given

      int connectFD = acceptConnection();
      if (connectFD < 0)
	 cerr << "Error accepting connection into group."  << endl;
      else
	 cgroup->addNewConnection(connectFD, PortNum);

      return connectFD;
   }


   // Accept all pending connects and insert them into *cgroup:
   int getAllWaitingConnections(SERVICE_GROUP *cgroup) {
      int connFd, numConns=0;
      while (connectIsWaiting_p()) {
	 if ((connFd=acceptConnectionIntoGroup(cgroup)) < 0) {
	    break;
	 }
	 else {
	    // UN-set Unix select write bit for this connection;  it will be set ONLY when
	    // there are pending outgoing msgs.  This removes spurious select events--ie
	    // select telling us that connections are writeable when there is nothing
	    // ready to be written for these connections.
	    dontCheckIfWritableInSelect(connFd);
	    numConns++;
	 }
      }

      return numConns;
   }

   // Accept all pending connects and insert them into DefaultServiceGrp (member):
   int getAllWaitingConnections(void) {
      return getAllWaitingConnections( getDefaultServiceGrp() );
   }

   int readyToRecv(CONNECTION *conn) {
      return FD_ISSET(conn->getfd(), &Copy_readfds);
   }

   int readyToSend(CONNECTION *conn) {
      return FD_ISSET(conn->getfd(), &Copy_writefds);
   }

   int exceptionOccurring(CONNECTION *conn) {
      return FD_ISSET(conn->getfd(), &Copy_exceptionfds);
   }


   // set Writefds bit corresponding to this conn so that next SELECT call
   // will check to see if connection is writable
   void checkIfWritableInSelect(CONNECTION *conn) {
      FD_SET(conn->getfd(), &Writefds);
   }

   // CLEAR Writefds bit corresponding to this conn so that next SELECT call
   // will check to see if connection is writable
   void dontCheckIfWritableInSelect(CONNECTION *conn) {
      FD_CLR(conn->getfd(), &Writefds);
   }

   void dontCheckIfWritableInSelect(int connFd) {
      FD_CLR(connFd, &Writefds);
   }


   int getNumGroups(void) {
      return SrvGroupList.length();
   }

   void addNewGroup(SERVICE_GROUP *grp) {
      SrvGroupList.push_bot(grp);
   }

   void dequeueGroup(SERVICE_GROUP *grp);

   void clearConnInFdSets(CONNECTION *conn);

   SERVICE_GROUP *getFirstGroup(void) { return (SERVICE_GROUP *) SrvGroupList.ptop(); }

   SERVICE_GROUP *getNextGroup(SERVICE_GROUP *group) {
      SERVICE_GROUP *nextGrp;
      
      if (group == SrvGroupList.get_bot())
	nextGrp = NULL;
      else
	nextGrp = (SERVICE_GROUP *) group->get_link();

      return nextGrp;
   }



   // Move connection into service group denoted by groupID in msg.  If group does not
   // exist, create it.  Return ptr to service group.
   SERVICE_GROUP *moveConnIntoGroup(CONNECTION *conn, SockMsgType& msg);

   // A SERVICE_GROUP is uniquely identified by the PAIR (serviceID, groupID)
   SERVICE_GROUP *findGroup(int serviceID, int groupID);


    //int recvNextMsg(SockMsgType *msg, SERVICE_GROUP *connGrp, int flags=0);

   /**********
   void handleMsg(SockMsgType *msg) {
      GENERIC_MSG_HEADER *msgBuff = (GENERIC_MSG_HEADER *) msg->getMsgBuff();
      MsgHandlerObj->callHandler(msgBuff);
   }
   ***********/
   void handleBadRecv(int status, CONNECTION *conn);

   void handleBadSend(int status, CONNECTION *conn);

   void closeConnection(CONNECTION *conn);


   void serviceMessages(void);


  void SetWritefds(void);


};


#endif
