#include "SpStdlib.H"
#include <sys/types.h>

#ifndef __SPEEDES_NT__
#include <netinet/in.h>
#include <sys/socket.h>
#include <unistd.h>
#else
#include <windows.h>
#endif

#include "SpString.H"
#include "SpServerSock.H"

static const struct timeval ZeroWait = { 0, 0 };     // zero timeout used in select calls


SERVER_SOCK::SERVER_SOCK(int port, int accept_qsize) : STREAM_SOCK(port) {
   NumConnections = 0;
   MaxFD = getfd();

   // Zero-out descriptor bit-arrays since there are not connections yet:
   FD_ZERO(&Readfds); FD_ZERO(&Copy_readfds);
   FD_ZERO(&Writefds); FD_ZERO(&Copy_writefds);
   FD_ZERO(&Exceptionfds); FD_ZERO(&Copy_exceptionfds);

   // Include ourself in read, and exception sets (but NOT the write
   // set since it causes spurious select events):
   FD_SET(getfd(), &Readfds );
   FD_SET(getfd(), &Exceptionfds );

   setSocketStatus(SERVER);

   if (port == -1){
     port = 10000 + 
#ifndef __SPEEDES_NT__
       int(getuid());
#else
       0;
#endif
   }
   bindSock(port);
   setNonBlocking();

   listenForConnections(accept_qsize);
}

void SERVER_SOCK::listenForConnections(int acceptQsize) {
  if (listen(getfd(), acceptQsize) < 0) {
    perror("Listen Error");
    exit(1);
  }
}


int SERVER_SOCK::acceptConnection(void) {
   int connectFD;
   static sockaddr connectorAddr;
   int addrBytesInOut = sizeof connectorAddr;

   int blockp;  // is socket blocking?
   if (!(blockp = isBlocking_p()))
     setBlocking();  // block until connection
     

   // addr of connector is returned, and #bytes of that address:
   connectFD = ACCEPT(getfd(), &connectorAddr, &addrBytesInOut);
   if (connectFD < 0) {
      perror( "accept" );
   }
   else {
      NumConnections++;
       // keep track of max file descriptor (for select):
      if (connectFD > MaxFD) MaxFD = connectFD;

      /******* note removal:
      These are now done in the CONNECTION constructor
      set_nonblocking(connectFD);
      set_nodelay(connectFD); // <-- Ron's fix for Irix bug
      *******/


      // Add this connection into the read, write, and exception sets
      // Note:  setting the write bit can generate "spurious" select events
      //        (ie when connection is writeable but there is nothing ready
      //        to be written, you may not want to select on this)
      FD_SET(connectFD, &Readfds );
      FD_SET(connectFD, &Writefds );
      FD_SET(connectFD, &Exceptionfds );
   }

   if (!blockp) setNonBlocking();  // restore non-blocking status if nec

   return connectFD;
}



/* checkForAllConnectionEvents Checks for read, write, and exception
   events on all file descriptors denoted by Readfds, Writefds, and
   Exceptionfds. Returns the # of such events from select call.  The
   descriptors w/events are returned in the data members Copy_readfds,
   Copy_writefds, and Copy_exceptionfds.  To check if file descriptor
   fd0 has a pending READ, use macro FD_ISSET(fd0, Copy_readfds) which
   is True iff the Copy_readfds bitmask indicates a pending read at
   fd0 (likewise for write, exception).  The args default to a
   BLOCKING select call (which will block until 1 or more bits are
   set); to poll (ie wait 0 secs before timing out) use
   checkForAllConnectionEvents(0); otherwise use the call:
   checkForAllConnectionEvents(0, timeOutSecs) where timeOutSecs=FLOAT
   # secs before timing out */

int SERVER_SOCK::checkForAllConnectionEvents(int, float waitSecs) {
   struct timeval timeOut;
   struct timeval *waitTime;
   //if (blockp)
   if (0)
     waitTime = NULL;
   else if (waitSecs < 0)
   {
     timeOut = ZeroWait;
     waitTime = &timeOut;
   }
   else {
      timeOut.tv_sec = int(waitSecs);  //integer secs
      timeOut.tv_usec = long(1E6 * (waitSecs - timeOut.tv_sec));  // millionths of a sec
      waitTime = &timeOut;
   }
       

   memcpy(	(char *)&Copy_readfds,
		(char *)&Readfds,
		sizeof(Readfds)	);
   

   memcpy(	(char *)&Copy_writefds,
		(char *)&Writefds,
		sizeof(Writefds)	);

   memcpy(	(char *)&Copy_exceptionfds,
		(char *)&Exceptionfds,
		sizeof(Exceptionfds)	);


   int numEvents = SELECT(MaxFD+1, // check fd's 0..MaxFD only (which is MaxFD+1 fds)
			  &Copy_readfds,
			  &Copy_writefds, 
			  &Copy_exceptionfds,
			  waitTime);
     
   if (numEvents < 0) {
      perror("Select Error");
      abort();
   }

   return numEvents;
}

// Same as checkForAllConnectionEvents but without exceptions
int SERVER_SOCK::checkForConnectionReadsAndWrites(int blockp, float waitSecs) {
   struct timeval timeOut;
   struct timeval *waitTime;
   if (blockp)
     waitTime = NULL;
   else if (waitSecs < 0)
   {
     timeOut = ZeroWait;
     waitTime = &timeOut;
   }
   else {
      timeOut.tv_sec = int(waitSecs);  //integer secs
      timeOut.tv_usec = long(1E6 * (waitSecs - timeOut.tv_sec));  // millionths of a sec
      waitTime = &timeOut;
   }
       
   memcpy(	(char *)&Copy_readfds,
		(char *)&Readfds,
		sizeof(Readfds)	);


   memcpy(	(char *)&Copy_writefds,
		(char *)&Writefds,
		sizeof(Writefds)	);



   int numEvents = SELECT(MaxFD+1, // check fd's 0..MaxFD only (which is MaxFD+1 fds)
			  &Copy_readfds, 
			  &Copy_writefds,
			  NULL,     // <-- don't check exceptions
			  waitTime);
     
   if (numEvents < 0) {
      perror("Select Error"); exit(1);
   }



   return numEvents;
}


   
// Same as checkForAllConnectionEvents, except only checks for reads.
int SERVER_SOCK::checkForConnectionReads(int blockp, float waitSecs) {
   struct timeval timeOut;
   struct timeval *waitTime;
   if (blockp)
     waitTime = NULL;
   else if (waitSecs < 0)
   {
     timeOut = ZeroWait;
     waitTime = &timeOut;
   }
   else {
      timeOut.tv_sec = int(waitSecs);  //integer secs
      timeOut.tv_usec = long(1E6 * (waitSecs - timeOut.tv_sec));  // millionths of a sec
      waitTime = &timeOut;
   }
       
   memcpy(	(char *)&Copy_readfds,
		(char *)&Readfds,
		sizeof(Readfds)	);



   int numReads = SELECT(MaxFD+1, // check fd's 0..MaxFD only (which is MaxFD+1 fds)
			  &Copy_readfds, NULL, NULL, waitTime);
     
   if (numReads < 0) {
      perror("Select Error"); exit(1);
   }


   return numReads;
}

// Is the socket ready for a "read"--ie connection pending?
int SERVER_SOCK::connectIsWaiting_p(void) {
   fd_set readfds;

   FD_ZERO(&readfds);
   FD_SET(getfd(), &readfds);

   // Only check bits 0..getfd() since getfd() is rep by bit #getfd():
   int numSelectBits = getfd() + 1;  // THIS IS STANDARD ON ALL UNIX
   
   timeval temp = ZeroWait;
   int numReads = SELECT(numSelectBits, 
			  &readfds, NULL, NULL, &temp);
     
   if (numReads < 0) {
      perror("Select Error"); 
      exit(1);
   }
   return( numReads > 0 );
}



   
