#ifndef Service_Group_Class
#define Service_Group_Class

#include "SpConnection.H"
#include "SpMsgHeader.H"

#define NO_ID (-333)

#define STORE_MSG_HANDLER(hdlNum, className, methodName)\
   storeHandler(hdlNum, (MsgMethodPtr) &className::methodName)

class GENERIC_MSG_HEADER;

class SERVICE_GROUP : public C_ITEM {
 public:
   // Which service grp a msg gets routed to is determined by the PAIR
   // (ServiceTypeID, GroupID); eg 3 sims request comms service so their
   // associated sets of nodes will be put into 3 SERVICE_GROUP objs, each
   // having a different GroupID.
   int ServiceTypeID;
   int GroupID;

   enum HandlerStatus { handler_ok, delete_this_group };

 private:
   C_QUEUE Connlist;
   CONNECTION *CurrentReceiveConn;
   CONNECTION *CurrentSendConn;

   int ConnCount;  // connection count--just to give new connection an ID
   int NumHandlers;  // # msg handlers stored in MsgHandlerArr


 protected:
   // MsgMethodPtr is a pointer to a SERVICE_GROUP method that takes
   // args CONNECTION * and GENERIC_MSG_HEADER *, and doesn't return a
   // val:
  typedef HandlerStatus (SERVICE_GROUP::*MsgMethodPtr)(CONNECTION *, GENERIC_MSG_HEADER *);

  HandlerStatus ErrorMethod(CONNECTION *conn, GENERIC_MSG_HEADER *hdr);
  MsgMethodPtr *MsgHandlerArr; // array of ptrs to methods for handling msgs

   void allocateHandlerStorage(int numHandlers);
   
   void storeHandler(int methNumber, MsgMethodPtr methptr);


 public:
   void callHandler(CONNECTION *conn, SockMsgType& msg) {
      GENERIC_MSG_HEADER *mh = (GENERIC_MSG_HEADER *) msg.getMsgBuff();
      int hnum = mh->getHandlerNum();
      (this->*MsgHandlerArr[hnum])(conn, mh);
   }

    SERVICE_GROUP(int idnum=NO_ID) : MsgHandlerArr(NULL) {
      id=idnum;
      ConnCount=0;
      CurrentReceiveConn=NULL;
      CurrentSendConn=NULL;
   }

   ~SERVICE_GROUP();
    
   void addNewConnection(int connectFD, int port) {
      CONNECTION *conn = new CONNECTION(connectFD, port, ConnCount++, this);
      Connlist.push_bot((C_ITEM *)conn);
   }

   void addConnection(CONNECTION *conn) {
      conn->setConnGroup(this);
      Connlist.push_bot((C_ITEM *)conn);
   }

   void dequeueConnection(CONNECTION *conn) {
      Connlist.deque(conn);
      conn->set_link((C_ITEM*)NULL); // for safety (queue obj doesn't do this)
   }

   int numConnections(void) {
      return Connlist.length();
   }

// This method can be provided by the user to capture the fact that the
// connection has been closed. Applications should not delete the connection
// since this is automatically done later for you.

    virtual void CloseConnection(CONNECTION *) {;}

// This method is called when a connection is closed. It calls the
// CloseConnection() virtual function provided by the user, cleans up its
// own internal storage of the connection, and then deletes the connection
// class.

    // This function is called when the operating system reports that
    // the passed-in connection is closed.  Descendents of this class
    // may wish to keep additional information associated with each
    // CONNECTION in Connlist, by creating a list of a new class which
    // holds a CONNECTION* and the other info, or descendants may wish
    // to store the CONNECTIONs in Connlist in some other data
    // structure which makes lookups faster for them.  If they do
    // this, then they should override this function and make it
    // remove the CONNECTION * from their own list.  Then they should
    // call next method.  I.e. they call the following definition so
    // that the CONNECTION is also removed from this class'es
    // Connlist.  Or if their code is not robust enough to handle
    // closed connection, then they may wish to define this function
    // to return TRUE, which means that the msgServer will delete this
    // serviceGroup which in turn closes all of the groups
    // connections.
    HandlerStatus ConnectionHasBeenClosed(CONNECTION *conn) {
      CloseConnection(conn);
      dequeueConnection(conn);
      delete conn;
      return( numConnections() == 0 ? delete_this_group : handler_ok);
    }
   
   CONNECTION *getCurrRecvConn(void) { return CurrentReceiveConn; }
   void setCurrRecvConn(CONNECTION *conn) { CurrentReceiveConn = conn; }

   CONNECTION *nextRecvConn(void) {
      CONNECTION *nextRecv = getNextConnection( getCurrRecvConn() );
      setCurrSendConn(nextRecv);
      return nextRecv;
   }

   CONNECTION *getCurrSendConn(void) { return CurrentSendConn; }
   void setCurrSendConn(CONNECTION *conn) { CurrentSendConn = conn; }

   CONNECTION *nextSendConn(void) {
      CONNECTION *nextSend = getNextConnection( getCurrSendConn() );
      setCurrSendConn(nextSend);
      return nextSend;
   }

   CONNECTION *getFirstConnection(void) {
      return (CONNECTION *) Connlist.get_top();
   }

   CONNECTION *getNextConnection(CONNECTION *conn) {
      CONNECTION *nextConn;

      // C_QUEUE is not NULL-terminated: 
      if (conn == Connlist.get_bot())
	nextConn = NULL;
      else
	nextConn = (CONNECTION *) conn->get_link();

      return nextConn;
   }

};



/************* EXAMPLE of use:

class BogusServer : public SERVICE_GROUP {
 public:
   enum { F0, F1, F2, F3, F4 };

   int Knt;
   void f0(CONNECTION *conn, GENERIC_MSG_HEADER *mh) {
      cout << "f0 (i+1) = " << mh->data+1 << endl;
   }

   void f1(CONNECTION *conn, GENERIC_MSG_HEADER *mh) {
      cout << "f1 (i*2) = " << mh->data*2 << endl;
   }

   void f2(CONNECTION *conn, GENERIC_MSG_HEADER *mh) {
      cout << "f2 (i*3) = " << mh->data*3 << endl;
   }

   void f3(CONNECTION *conn, GENERIC_MSG_HEADER *mh) {
      int i = mh->data;
      cout << "f3 (i*i) = " << i*i << endl;
   }


   void f4(CONNECTION *conn, GENERIC_MSG_HEADER *mh) {
      Knt += mh->data;
      cout << "f4 (Sum i's) = " << Knt << endl;
   }

   BogusServer(void) {
      Knt = 0;
      allocateHandlerStorage(NUMHANDLERS);

      STORE_MSG_HANDLER(F0, BogusServer, f0);
      STORE_MSG_HANDLER(F1, BogusServer, f1);
      STORE_MSG_HANDLER(F2, BogusServer, f2);
      STORE_MSG_HANDLER(F3, BogusServer, f3);
      STORE_MSG_HANDLER(F4, BogusServer, f4);
   }
};


BogusServer *makeBogusServer(void) {
   return( new BogusServer );
}



*******************/


#endif
