#include <sys/types.h>

#ifndef __SPEEDES_NT__
#include <sys/uio.h>
#include <sys/socket.h>
#else
#include <windows.h>
#endif

#include "SpLimits.H"
#include "SpAssert.H"
#include "SpErrno.H"
#include <fcntl.h>
#include "SpStdlib.H"
#include "SpString.H"
#include "SpOS.H"

#include "SpStreamSock.H"
#include "SpMsgHeader.H"
#include "SpConnection.H"

static const int NUM_HDR_LONGS = 2;
const int MAX_MSG_BYTES = 100000000; // 100 million bytes max!
const int NO_MSG_AVAIL = -1234;      // Old status code

const int DEFAULT_PORT = 3729;  // Arbitrary port# > 1024
const int ACCEPT_Q_LENGTH = 20; // Q size of guys waiting to connect
const int NO_FD = -127;         // not a file descriptor

const SockMsgType STREAM_SOCK::NullMsg(NULL, 0);

const int MAX_ZERO_TRIES = 2;  // Max # attempts to send data on socket when 0 bytes written
const int MAX_SEND_TRIES = 2;  // Max # attempts to send msg
const int MAX_READ_TRIES = 2;  // # times to try low-level read; often don't get all N bytes w/1 try

enum SOCK_ERROR_SOURCE { SENDING, RECEIVING };



STREAM_SOCK::STREAM_SOCK(int port)
   : Sockfd(socket( AF_INET, SOCK_STREAM, 0 )) {
     
#ifdef __SPEEDES_NT__
  if (Sockfd < 0){
    if (WSAGetLastError() == WSANOTINITIALISED){
      WORD wVersionRequested = MAKEWORD(2,0);
      WSADATA wsaData;
      int err = WSAStartup(wVersionRequested, &wsaData);
      if (err){
	cout<<"Unable to initialize the windows sockets libraries."
	    <<" WinSock must be installed.\a"<<endl;
	exit(0);
      }
      if ((LOBYTE(wsaData.wVersion) != 2) || (HIBYTE(wsaData.wVersion) != 0)) {
	cout<<"Could not load winsock version 2.0 or lower.\a"<<endl;
	exit(0);
      }
      cout<<"Winsocket libraries initialized!"<<endl;
      Sockfd = socket( AF_INET, SOCK_STREAM, 0 );
    }
  }
#endif
  if (Sockfd < 0 ) {
    perror( "Couldn't open socket stream 1" );
    exit(-1);
  }
 
  PortNum = port;
  setSocketStatus(NO_INFO);

  updateStatusFlags(); // set data member copy of Unix fcntl status flags
  setNoDelay();

  // Cannot do setNonBlocking() here so mark this as un-set:
  Blocking_p = -1;
}

STREAM_SOCK::~STREAM_SOCK(void) {
  int sockStatus = getSocketStatus();
 
  // Tell my partner that I'm closing down:
  if (sockStatus == CONNECTED) sendCloseConnectMsg();
 
  // The close call was done by the CLIENT_SOCK's CONNECTION obj
  // (which has the same file descriptor):
  if (sockStatus != CLIENT) {
#ifndef __SPEEDES_NT__
    int ret=close(Sockfd);
#else
int ret = closesocket(Sockfd);
#endif
    if (ret < 0) perror("closing socket:");
  }
 
}

int STREAM_SOCK::isBlocking_p(int sock){
#ifndef __SPEEDES_NT__
  int fcntlarg;
  if ((fcntlarg = fcntl(sock, F_GETFL, 0)) < 0) {
    perror("isBlocking_p");
    cerr << "Warning:  isBlocking_p returned FALSE but doesn't know"
         <<" blocking status"<< endl;
    return 0;
  }
  else
    return !(fcntlarg & O_NDELAY);
#else
  cout<<"Should not ever be calling STREAM_SOCK::isBlocking_p under NT"
      <<endl;
  return 17;
#endif
}

void STREAM_SOCK::sendCloseConnectMsg(void) {
  internalMsgHdr hdr;
  hdr.makeCloseConnectMsg();
  sendInternalMsg(hdr);
}

int STREAM_SOCK::closeSock(int fd) {
  sendCloseConnectMsg();

#ifndef __SPEEDES_NT__
  int ret=close(fd);
#else
  int ret=closesocket(fd);
#endif
  if (ret < 0) perror("closing socket:");

  return ret;
}


// Args default as follows: ( u_short portNum=DEFAULT_PORT, u_long
// inetAddr=INADDR_ANY)
void STREAM_SOCK::bindSock(u_short portNum, u_long inetAddr) {
  sockaddr_in addr;
  int i = 0;
  memset(&addr, 0, sizeof(addr));
     
  addr.sin_family = AF_INET;
  addr.sin_addr.s_addr = inetAddr;
  addr.sin_port = htons(portNum);

  while (bind(Sockfd, (sockaddr *) &addr, sizeof(addr)) < 0) {
    cout<<"Couldn't bind main socket stream "
	<<"press <ctrl>-c to abort: "<<i++<<endl;
    SpeedesSleep(5);
  }
}




    

/************************************************************************
 * set the socket to no delay (faster but not as efficient--sends out
 * messages immediately instead of gathering small msgs into packets).
 ************************************************************************/
int STREAM_SOCK::set_nodelay(int sock) {
  int nodelayVal = 1;  // value assigned by setsockopt to nodelay boolean
   
  if (setsockopt(sock,
		 IPPROTO_TCP,
		 TCP_NODELAY,
		 (char *)&nodelayVal,
		 sizeof(int)) < 0) {
    perror("set_nodelay");
    return -1;
  }

  return 0;

}


/************************************************************************
 * UNset the socket to no delay (slower but more efficient--gathers small
 * msgs into packets instead of sending immediately.
 ************************************************************************/
int STREAM_SOCK::unset_nodelay(int sock) {
  int nodelayVal = 0;  // value assigned by setsockopt to nodelay boolean
   
  if (setsockopt(sock,
		 IPPROTO_TCP,
		 TCP_NODELAY,
		 (char *)&nodelayVal,
		 sizeof(int)) < 0) {
    perror("unset_nodelay");
    return -1;
  }

  return 0;

}


/************************************************************************
*	set a socket to be non-blocking					*
************************************************************************/
int STREAM_SOCK::set_nonblocking(int sock) {
  if (sock == getfd()) {
    return setNonBlocking();
  }

#ifndef __SPEEDES_NT__
  int fcntlarg;

  if ((fcntlarg = fcntl(sock, F_GETFL, 0)) < 0) {
    perror("set_nonblocking F_GETFL call:");
    return -1;
  }

  fcntlarg |= O_NDELAY;
  if (fcntl(sock, F_SETFL, fcntlarg) < 0) {
    perror("set_nonblocking F_SETFL call:");
    return -1;
  }
#else
  static unsigned long int junk = 1;
  ioctlsocket(sock, FIONBIO, &junk);
#endif
  return 0;

}

/************************************************************************
*	set a socket to be blocking					*
************************************************************************/
int STREAM_SOCK::set_blocking(int sock) {
  if (sock == getfd()) {
    return setBlocking();
  }

#ifndef __SPEEDES_NT__
  int fcntlarg;

  if ((fcntlarg = fcntl(sock, F_GETFL, 0)) < 0) {
    perror("set_blocking: get status call");
    return 0;
  }

  fcntlarg &= ~O_NDELAY;
  if (fcntl(sock, F_SETFL, fcntlarg) < 0) {
    perror("set_blocking: setting status");
    return 0;
  }

#else
  static unsigned long int junk = 0;
  int rc = ioctlsocket(sock, FIONBIO, &junk);
#endif
  return 1;

}




//                              sendMsg
// Returns one of the following:
// WHOLE_MSG_OK  :   entire msg sent;  everything is OK
// PARTIAL_MSG_OK:   part msg sent;  everything OK
// TEMP_SOCK_PROBLEM: temporary problem (eg full buffer, interupted call) that _may_ fix itself
// BAD_SOCK:  Bad file descriptor for connection (probably closed conn)
// BAD_PTR_OR_ARG:  Bad ptr in sendmsg msghdr or bad msg size arg to sendmsg
// NOT_YET_CONNECTED: Possibly socket is ok, but still doing connection internal work
// CLOSED_CONN: Connection is closed
// UNKNOWN_SOCK_ERROR:  Don't know what the hell is wrong;  better bail out.

int STREAM_SOCK::sendMsg(int connfd, char *msg, int len,
			 msghdr *mh, long *hdrSpace, int flags) {

  // "pair vec"--vector of len, char* pairs (NOTE: allocated in CONNECTION constructor)

  //A bit of a hack here to make sure that the number of bytes is always set
  //in the msg header.
  GENERIC_MSG_HEADER* genHdr = (GENERIC_MSG_HEADER*)msg;
  genHdr->SetTotalBytes(len);

  iovec *prvec = mh->msg_iov;
  assert(prvec != NULL);

  // Build internal header and pack into a char array:
  internalMsgHdr hdr(len);
  hdr.packMsgHeader(hdrSpace);

  // We want to insert a header BEFORE the msg to be sent out (w/out
  // copying), so we use unix sendmsg.  It takes a group of messages
  // and writes them out to the socket buffer.  The group is passed
  // in as a vector of (char*, length) pairs.

  prvec[0].iov_base = (char*) hdrSpace;
  prvec[0].iov_len = HEADER_SIZE;

  prvec[1].iov_base = (char*) msg;
  prvec[1].iov_len = len;


  int status = sendUnixHdr(connfd, mh, flags);
  return status;

}


//                              sendUnixHdr
// Returns one of the following:
// WHOLE_MSG_OK  :   entire msg sent;  everything is OK
// PARTIAL_MSG_OK:   part msg sent;  everything OK
// TEMP_SOCK_PROBLEM: temporary problem (eg full buffer, interupted call) that _may_ fix itself
// BAD_SOCK:  Bad file descriptor for connection (probably closed conn)
// BAD_PTR_OR_ARG:  Bad ptr in sendmsg msghdr or bad msg size arg to sendmsg
// NOT_YET_CONNECTED: Possibly socket is ok, but still doing connection internal work
// CLOSED_CONN: Connection is closed
// UNKNOWN_SOCK_ERROR:  Don't know what the hell is wrong;  better bail out.
int STREAM_SOCK::sendUnixHdr(int connfd,  msghdr *mh, int flags) {
  int bytes;
  int hdrBytesUnsent = mh->msg_iov[0].iov_len;
  int msgBytesUnsent = mh->msg_iov[1].iov_len;
  int totBytesUnsent = hdrBytesUnsent + msgBytesUnsent;

  for(int i=0; (i < MAX_SEND_TRIES) && (totBytesUnsent > 0); i++) {
    bytes = sendmsg(connfd, mh, flags);


    if (bytes < 0) {
#ifndef __SPEEDES_NT__
       int status = classifySockError(errno, SENDING);
#else
       int status = classifySockError(WSAGetLastError(), SENDING);
#endif
       if (status == TEMP_SOCK_PROBLEM) // not really an error (eg send buffer full temporarily)
	 break;  // quit for loop; 
       else
	 return status;
    }

    if (bytes > 0) {
      totBytesUnsent -= bytes;
      if (totBytesUnsent > 0) {
	// Handle cases: 1) only part of header sent, 2) entire header and part of msg
	if (hdrBytesUnsent > 0) {
	  if (bytes < hdrBytesUnsent) { // partial send of header only
	    hdrBytesUnsent -= bytes;
	    mh->msg_iov[0].iov_len = hdrBytesUnsent;
	    char * base = ((char *) mh->msg_iov[0].iov_base)+bytes;
	    mh->msg_iov[0].iov_base = base;
	  }
	  else { // entire header sent
	    bytes -= hdrBytesUnsent;
	    mh->msg_iov[0].iov_len = hdrBytesUnsent = 0;
	  }
	}
	    
	if (hdrBytesUnsent == 0) {    // entire header sent
	  if (bytes < msgBytesUnsent) {  // partial message sent
	    msgBytesUnsent -= bytes;
	    mh->msg_iov[1].iov_len = msgBytesUnsent;
	    char * base = ((char *) mh->msg_iov[1].iov_base)+bytes;
	    mh->msg_iov[1].iov_base = base;
	  }
	  else
	    msgBytesUnsent = 0;
	}
      }
    }

  }// while


  if (totBytesUnsent == 0){  // sent entire msg
    return WHOLE_MSG_OK;
  }
  else{
    return PARTIAL_MSG_OK;
  }
}


/**********  Good for debugging:
void showMsg(msghdr *mh) {
   internalMsgHdr xx;

   xx.unpackHeader( (long *) mh->msg_iov[0].iov_base);

   cout << "--------- SHOW MSG ---------" << endl;
   cout << "[0].iov_len = " << mh->msg_iov[0].iov_len << endl;
   cout << "IntrlHDR: NumBytes = " << xx.NumBytes << ";  MsgType= " << xx.MsgType << endl;
	  
   char *str = (char *) mh->msg_iov[1].iov_base;
   cout << "[0].iov_len = " << mh->msg_iov[1].iov_len << endl;
   cout << "Msg str: |" << str << "|" << endl;
   cout << "----------------------------\n" << endl;
}
***********/


// return INTERNAL_MSG_DONE iff we can do everything now; otherwise,
// return a val that will be passed back up the call chain:
int STREAM_SOCK::doInternalMsg(internalMsgHdr *hdr) {
  switch (hdr->MsgType) {
  case (internalMsgHdr::CLOSE_CONN):
    setSocketStatus(OTHER_HUNGUP);
    return CLOSED_CONN;
  default:
    cout << "Bad Internal msg type = " << hdr->MsgType << endl;
    return BAD_INTERNAL_MSG;
  }
}



// Try to read entire internal msg header, and return via arg msgHdr.  If
// we can only get part of it, return the bytes via args head and bytesRead.
// This same fcn can resume reading a partial hdr using the same 2 args.
int STREAM_SOCK::getInternalHdr(int connfd, internalMsgHdr *msgHdr,
				long *head, int *bytesRead) {
   // We'll read into buffer provided by head, offset by *bytesRead (if
   // *byteRead==0 [typical case], we're reading a new header):
   char *hptr = ( (char *)head ) + (*bytesRead);
   int hdrBytesLeft = HEADER_SIZE - (*bytesRead);

   int ret;

   int internal_p = 1;
   do {
      ret = nb_receive_data(connfd, hptr, hdrBytesLeft);
      if (ret != hdrBytesLeft) {
	 if (ErrorStatus(ret)) {
	    //cout << "getIntrnl errstatus = " << ret << endl;
	    return ret;
	 }
	 else if (ret==0) {
	    //cout << "getIntrnl 0 bytes!?" << endl;
	    return ZERO_BYTES_READ;
	 }
	 else {
	    *bytesRead += ret;
	    //cout << "@@@@@@@@@@@====> getInternal PARTIAL got #bytes = " << ret 
	    // << "; totBytes = " << *bytesRead << endl;
	    return PARTIAL_INTERNAL_HDR_OK;
	 }
      }
      else {  // got entire internal header
	 *bytesRead = 0; // prepare for next internal hdr
	 //cout << "@@@@@@@@@@@====> getInternal success getting hdr!" 
	 //   << "(got #bytes=" << ret << ")" << endl;
	 
	 msgHdr->unpackHeader(head);  // pull out items from char* and insert into obj
	 //cout << "finish unpaking;  msgType= " << msgHdr->MsgType << ", #bytes= "
	 //   << msgHdr->NumBytes << endl;

	 if (msgHdr->internalSockMsg_p()) {
	    ret = doInternalMsg(msgHdr); // handle internal msgs (eg close socket msg)
	    if (ret != INTERNAL_MSG_DONE)
	      return ret;
	 }
	 else {
	    internal_p = 0;
	 }
      }
   } while (internal_p);

  return INTERNAL_HEADER_OK;
}



// receiveMsg:  receive msg from connection whose file descriptor is connfd into
//              msg and return a status code.
// Returns one of the following:
// WHOLE_MSG_OK  :   entire msg sent;  everything is OK
// PARTIAL_MSG_OK:   part msg sent;  everything OK
// ZERO_BYTES_READ: read 0 bytes; if select says sock is readable but we get 0 bytes,
//                  this means that the connection has been closed by the peer.
// MSG_TOO_BIG: incoming msg bigger than MAX_MSG_BYTES (100 MEG!);  probably mangle msg
// TEMP_SOCK_PROBLEM: temporary problem (eg full buffer, interupted call) that _may_ fix itself
// BAD_SOCK:  Bad file descriptor for connection (probably closed conn)
// BAD_PTR_OR_ARG:  Bad ptr in sendmsg msghdr or bad msg size arg to sendmsg
// NOT_YET_CONNECTED: Possibly socket is ok, but still doing connection internal work
// CLOSED_CONN: Connection is closed
// UNKNOWN_SOCK_ERROR:  Don't know what the hell is wrong;  better bail out.
int STREAM_SOCK::receiveMsg(int connfd, SockMsgType *msg, int flags) {
  long hdr[NUM_HDR_LONGS];

  internalMsgHdr msgHdr;
  int ret;



  int hdrBytesRead = 0;
  ret = getInternalHdr(connfd, &msgHdr, hdr, &hdrBytesRead);
  if (ErrorStatus(ret)) {
     return ret;
  }
  // Store the partial hdr in msg so we can resume later:
  else if (ret == PARTIAL_INTERNAL_HDR_OK) {
     long *head = new long[NUM_HDR_LONGS];  // long* for alignment purposes
     memcpy((char *) head, hdr, hdrBytesRead);
     msg->savePartialHdr(head, HEADER_SIZE, hdrBytesRead);
     return PARTIAL_MSG_OK;
  }

  // Got entire internal msg header at this point

  long int msgLen = msgHdr.NumBytes;
  
  if (msgLen > MAX_MSG_BYTES) {
    return MSG_TOO_BIG;
  }

  //char *buff = (char*) malloc(int(msgLen) * sizeof(char));
  char *buff = new char[int(msgLen)];

  msg->setMsgBuff(buff);
  msg->setMsgLength(msgLen);

  int bytes=0;
  int bytesOrStatus;  // error status or #bytes returned from nb_receive_data

  for(int j=0; (j<MAX_READ_TRIES) && (bytes<msgLen); j++) {
     bytesOrStatus = nb_receive_data(connfd, buff+bytes, msgLen-bytes, flags);
     if (ErrorStatus(bytesOrStatus)) {
	if (FatalErrorStatus(bytesOrStatus)) {
	   msg->setBytesRecvd(bytes);
	   return bytesOrStatus;
	}
	else { // non-fatal error (eg read 0 bytes or sock temp busy)
	   break;
	}
     }
     else {
	bytes += bytesOrStatus;
     }
  }


  msg->setBytesRecvd(bytes);


  // Note: at this point we've already read the internal msg hdr, so even if
  //       bytes==0, we've read part of the whole msg (ie don't return ZERO_BYTES_READ!)
  if (bytes < msgLen)
    return PARTIAL_MSG_OK;
  else
    return WHOLE_MSG_OK;

}

int STREAM_SOCK::recvMsg_blk(int connfd, SockMsgType *msg, int flags) {
  int blockp = isBlocking_p(connfd);  // is socket blocking?
  if (!blockp) set_blocking(connfd);  // block until connection
     
  int ret = receiveMsg(connfd, msg, flags);

  if (!blockp) set_nonblocking(connfd);  // restore non-blocking status if nec

  return ret;
}


// Returns one of the following:
// WHOLE_MSG_OK  :   entire msg sent;  everything is OK
// PARTIAL_MSG_OK:   part msg sent;  everything OK
// ZERO_BYTES_READ: read 0 bytes; if select says sock is readable but we get 0 bytes,
//                  this means that the connection has been closed by the peer.
// TEMP_SOCK_PROBLEM: temporary problem (eg full buffer, interupted call) that _may_ fix itself
// BAD_SOCK:  Bad file descriptor for connection (probably closed conn)
// BAD_PTR_OR_ARG:  Bad ptr in sendmsg msghdr or bad msg size arg to sendmsg
// NOT_YET_CONNECTED: Possibly socket is ok, but still doing connection internal work
// CLOSED_CONN: Connection is closed
// UNKNOWN_SOCK_ERROR:  Don't know what the hell is wrong;  better bail out.
int STREAM_SOCK::resumeRecvMsg(int connfd, SockMsgType *partMsg, int flags) {
   internalMsgHdr msgHdr;
   int ret, gotSomeHdrBytes=0;
   long int msgLen, msgBytesRecvd;
   char *mbuff;


   if (partMsg->havePartialHdr_p()) {
      int hdrBytesRead = partMsg->getHdrBytesRecvd();
      long *internalHdr = (long *)partMsg->getMsgBuff();
      ret = getInternalHdr(connfd, &msgHdr, internalHdr, &hdrBytesRead);
      if (ErrorStatus(ret)) {
	 return ret;
      }
      else if (ret == PARTIAL_INTERNAL_HDR_OK) {
	 partMsg->setHdrBytesRecvd(hdrBytesRead);
	 return PARTIAL_MSG_OK;
      }
      //else got entire internal header:
      gotSomeHdrBytes = 1;

      msgLen = msgHdr.NumBytes;

      long *hdrBuff = (long *) partMsg->getMsgBuff();
      delete[] hdrBuff;
  
      if (msgLen > MAX_MSG_BYTES) {
	 return MSG_TOO_BIG;
      }

      mbuff = new char[int(msgLen)];

      partMsg->setMsgBuff(mbuff);
      partMsg->setMsgLength(msgLen);
      partMsg->setBytesRecvd(0);

      msgBytesRecvd = 0;
   }
   else { // no partial internal hdr
      msgBytesRecvd = partMsg->getBytesRecvd();
      mbuff =  partMsg->getMsgBuff() + msgBytesRecvd;
      msgLen = partMsg->getMsgLength() - msgBytesRecvd;
   }

  int bytes = 0;
  int bytesOrStatus;  // #bytes from nb_receive_data or error status

  for(int j=0; (j<MAX_READ_TRIES) && (bytes<msgLen); j++) {
    bytesOrStatus = nb_receive_data(connfd, mbuff+bytes, msgLen-bytes, flags);

    if (ErrorStatus(bytesOrStatus)) {
       if (FatalErrorStatus(bytesOrStatus)) {
	  partMsg->setBytesRecvd(msgBytesRecvd+bytes);
	  return bytesOrStatus;
       }
       else { // non-fatal error (eg read 0 bytes or sock temp busy)
	  break;
       }
    }
    else {
       bytes += bytesOrStatus;
    }
  }





  partMsg->setBytesRecvd(msgBytesRecvd+bytes);

  if ((bytes==0) && (!gotSomeHdrBytes))
    return ZERO_BYTES_READ;
  else if (bytes < msgLen)
    return PARTIAL_MSG_OK;
  else
    return WHOLE_MSG_OK;

}




int STREAM_SOCK::readMsgIntoBuff(int connfd, SockMsgType *msg, 
				 char *buff, int buffSize, long int *msgBytes) {
  long hdr[NUM_HDR_LONGS];

  internalMsgHdr msgHdr;
  int ret;

  *msgBytes = -1;

  //cout << "readMsgInBuff: msgLen = " << msg->getMsgLength()
  //     << ", bytesRcvd= " << msg->getBytesRecvd() << endl;


  int hdrBytesRead = 0;
  ret = getInternalHdr(connfd, &msgHdr, hdr, &hdrBytesRead);
  if (ErrorStatus(ret)) {
     return ret;
  }
  // Store the partial hdr in msg so we can resume later:
  else if (ret == PARTIAL_INTERNAL_HDR_OK) {
     long *head = new long[NUM_HDR_LONGS];  // long* for alignment purposes
     memcpy((char *) head, hdr, hdrBytesRead);
     msg->savePartialHdr(head, HEADER_SIZE, hdrBytesRead);
     return PARTIAL_MSG_OK;
  }

  // Got entire internal msg header at this point

  long int msgLen = msgHdr.NumBytes;
  *msgBytes = msgLen;
  
  if (msgLen > buffSize) {
     if (msgLen > MAX_MSG_BYTES)
       return MSG_TOO_BIG;     // something is VERY wrong (msgLen>100 MEG!??)
     else
       return USER_BUFF_TOO_SMALL;  // not nec too serious
  }

  ///  char *buff = new char[int(msgLen)];

  msg->setMsgBuff(buff);
  msg->setMsgLength(msgLen);

  int bytes=0, flags=0;
  int bytesOrStatus;  // error status or #bytes returned from nb_receive_data

  for(int j=0; (j<MAX_READ_TRIES) && (bytes<msgLen); j++) {
     bytesOrStatus = nb_receive_data(connfd, buff+bytes, msgLen-bytes, flags);
     if (ErrorStatus(bytesOrStatus)) {
	if (FatalErrorStatus(bytesOrStatus)) {
	   msg->setBytesRecvd(bytes);
	   *msgBytes = bytes;
	   return bytesOrStatus;
	}
	else { // non-fatal error (eg read 0 bytes or sock temp busy)
	   break;
	}
     }
     else {
	bytes += bytesOrStatus;
     }
  }


  msg->setBytesRecvd(bytes);
  *msgBytes = bytes;

  //   cout << "Exit receiveMsg with #bytes = " << msgLen << endl;

  // Note: at this point we've already read the internal msg hdr, so even if
  //       bytes==0, we've read part of the whole msg (ie don't return ZERO_BYTES_READ!)
  if (bytes < msgLen)
    return PARTIAL_MSG_OK;
  else
    return WHOLE_MSG_OK;

}


int STREAM_SOCK::resumeReadMsgIntoBuff(int connfd, SockMsgType *partMsg,
				       char *msgBuff, int buffSize, long int *msgBytes) {
   internalMsgHdr msgHdr;
   int ret, gotSomeHdrBytes=0;
   long int msgLen, msgBytesRecvd;
   char *mbuff = msgBuff;


   //   cout << "resumeReadInBuff: msgLen = " << partMsg->getMsgLength()
   //	<< ", bytesRcvd= " << partMsg->getBytesRecvd() << endl;

   if (partMsg->havePartialHdr_p()) {
      *msgBytes = -1; // indicate msg len is unknown
      // cout << "resumeReadMsgInBuff:  partial hdr" << endl;
      int hdrBytesRead = partMsg->getHdrBytesRecvd();
      long *internalHdr = (long *)partMsg->getMsgBuff();
      ret = getInternalHdr(connfd, &msgHdr, internalHdr, &hdrBytesRead);
      if (ErrorStatus(ret)) {
	 return ret;
      }
      else if (ret == PARTIAL_INTERNAL_HDR_OK) {
	 partMsg->setHdrBytesRecvd(hdrBytesRead);
	 return PARTIAL_MSG_OK;
      }
      //else got entire internal header:
      gotSomeHdrBytes = 1;

      msgLen = msgHdr.NumBytes;
      *msgBytes = msgLen;

      long *hdrBuff = (long *) partMsg->getMsgBuff();
      delete[] hdrBuff;
  
      if (msgLen > buffSize) {
	 if (msgLen > MAX_MSG_BYTES)
	   return MSG_TOO_BIG;     // something is VERY wrong (msgLen>100 MEG!??)
	 else
	   return USER_BUFF_TOO_SMALL;  // not nec too serious
      }
      // mbuff = new char[int(msgLen)];

      partMsg->setMsgBuff(mbuff);
      partMsg->setMsgLength(msgLen);
      partMsg->setBytesRecvd(0);

      msgBytesRecvd = 0;
   }
   else { // no partial internal hdr
      cout << "resumeReadMsgInBuff: no partial hdr" << endl;
      msgBytesRecvd = partMsg->getBytesRecvd();
      *msgBytes = msgBytesRecvd;
      mbuff =  partMsg->getMsgBuff() + msgBytesRecvd;
      msgLen = partMsg->getMsgLength() - msgBytesRecvd;
   }

  int bytes = 0, flags = 0;
  int bytesOrStatus;  // #bytes from nb_receive_data or error status

  for(int j=0; (j<MAX_READ_TRIES) && (bytes<msgLen); j++) {
    bytesOrStatus = nb_receive_data(connfd, mbuff+bytes, msgLen-bytes, flags);

    if (ErrorStatus(bytesOrStatus)) {
       if (FatalErrorStatus(bytesOrStatus)) {
	  *msgBytes = msgBytesRecvd + bytes;
	  partMsg->setBytesRecvd(*msgBytes);
	  return bytesOrStatus;
       }
       else { // non-fatal error (eg read 0 bytes or sock temp busy)
	  break;
       }
    }
    else {
       bytes += bytesOrStatus;
    }
  }

  *msgBytes = msgBytesRecvd + bytes;
  partMsg->setBytesRecvd(*msgBytes);

  if ((bytes==0) && (!gotSomeHdrBytes))
    return ZERO_BYTES_READ;
  else if (bytes < msgLen)
    return PARTIAL_MSG_OK;
  else
    return WHOLE_MSG_OK;

}



// Send length bytes from buffer out thru socket.  Let flags arg default to zero
// (included for poss future use--see send  man page)
int STREAM_SOCK::send_data(int socket, char *buffer, int length, int flags) {
  int zeroCount = 0; // # calls to write that write 0 bytes

  int bytes, totBytesSent = 0;

  while (totBytesSent < length) {
    bytes = send(socket, buffer+totBytesSent, length-totBytesSent, flags);

    if (bytes <= 0) {
      //  EINVAL is the offending HP error code--a couple tries
      // will get around the error somehow (possibly socket is in
      // "reset sequence"--see EINVAL error description in "send"
      // man page).
      if ( (bytes == 0) && ((errno==0) || (errno==EINVAL)) ) {
	if (++zeroCount>MAX_ZERO_TRIES) break;
      }
      else if (errno == EWOULDBLOCK) {
	break;
      } else if (errno == EINTR) {
	//cout << "send_data:  interrupted write;  trying again" << endl;
	bytes = 0;
      }else{
	perror("write_data err msg");
	return -1;
      }
    }
      
    totBytesSent += bytes;
  }

#ifdef NOISY
  if (zeroCount > MAX_ZERO_TRIES) {
    cout << "send_data: partial msg sent because too many 0-byte writes.\n"
      "send_data:  bytesSent = " << totBytesSent << ", msg len = " << length
      ", #attempts = " << zeroCount << endl;
  }
#endif

  return totBytesSent;
}
#ifndef __SPEEDES_NT__
int STREAM_SOCK::classifySockError(int unixErrorNumber, int errorSrc) {
  switch (unixErrorNumber) {
  case EAGAIN:
#ifdef HPUX0905
  case EWOULDBLOCK:  // Only on HP is EWOULDBLOCK != EAGAIN
#endif
  case EDEADLK:
  case EINTR:
  case ETIMEDOUT:
  case ENOBUFS:
    return TEMP_SOCK_PROBLEM;
    
  case EBADF:
  case ENOTSOCK:
    return BAD_SOCK;
    
  case EINVAL: 
#ifdef HPUX0905
    // hpux sometimes returns EINVAL instead of EPIPE (before OS knows peer is dead?)
    if (errorSrc==SENDING) return CLOSED_CONN;
#endif
  case EMSGSIZE:
  case EFAULT:
    return BAD_PTR_OR_ARG;
    
    
  case ENOTCONN:
    // broken conn or sock hasn't completed "connect sequence"
    return NOT_YET_CONNECTED;
    
  case EPIPE:      // err we get instead of ignored SIGPIPE signal (sending on conn after peer has died)
    
  case ECONNRESET:
  case ENETUNREACH:
  case ENETDOWN:
    return CLOSED_CONN;
    
  default:
	    // The following line is here so that the compiler will not
	    // give a warning saying that errorSrc is not used.
	    errorSrc = errorSrc;

    cout << "classifySockSendError:  unknown socket errno = " 
	 << unixErrorNumber << endl;
    return UNKNOWN_SOCK_ERROR;
  }
}
#else
int STREAM_SOCK::classifySockError(int unixErrorNumber, int) {
  switch (unixErrorNumber) {
  case ERROR_RESOURCE_TYPE_NOT_FOUND:
  case WSAETIMEDOUT:
  case WSAEWOULDBLOCK:
    return TEMP_SOCK_PROBLEM;
    
  case WSAENETDOWN:
  case WSAENETRESET:
  case WSAECONNRESET:
    return CLOSED_CONN;
    
  case WSAENOTSOCK:
  case WSAESHUTDOWN:
    return BAD_SOCK;
    
  case WSAEINVAL:
  case WSAEFAULT:
    return BAD_PTR_OR_ARG;
  default:
    cout << "classifySockSendError:  unknown socket GetLastError = " 
	 << unixErrorNumber << endl;
    return UNKNOWN_SOCK_ERROR;
  }
}
#endif



int STREAM_SOCK::nb_send_data(int sock, char *buff, int len, int flags) {
  int bytes;

  bytes = send(sock, buff, len, flags);

  if (bytes < 0) {
    if (errno == EWOULDBLOCK)
      bytes = 0;
    else
      perror("nb_socket_send");
  }

  return bytes;

}




int STREAM_SOCK::nb_receive_data(int sockOrConnect, char *buff, long int len, int flags) {
  long int bytes;

  bytes = recv(sockOrConnect, buff, (int)len, flags);
  //cout << "nb_receive_data got #bytes = " << bytes << " out of #wanted= " << len << endl;
  //if (errno!=0) perror("nb_receive_data error");

  if (bytes >= 0) {
    return bytes;
  }
  else {// bytes < 0 means error condition
    //perror("nb_receive_data");
#ifdef __SPEEDES_NT__
     int myErrorCode =  classifySockError(WSAGetLastError(), 0);
#else
    int myErrorCode =  classifySockError(errno, 0);
#endif
    return myErrorCode;
  }
}


int STREAM_SOCK::sendInternalMsg(internalMsgHdr& hdr) {
   long head[NUM_HDR_LONGS];
   hdr.packMsgHeader(head);  // packs hdr into head

   msghdr mh;
   iovec prvec[2];

   mh.msg_name = NULL;
   mh.msg_namelen = 0;
#ifdef Linux
   mh.msg_control = NULL;
   mh.msg_controllen = 0;
   mh.msg_flags = 0;
#else
   mh.msg_accrights = NULL;
   mh.msg_accrightslen = 0;
#endif

   // Internal msg is just an internal hdr and a NULL msg:
   prvec[0].iov_base = (char *) head;
   prvec[0].iov_len = HEADER_SIZE;

   // Need to fill these in so that sendUnixHdr works:
   prvec[1].iov_base = (char*) NULL;
   prvec[1].iov_len = 0;

   mh.msg_iovlen = 1;
   mh.msg_iov = prvec;
   
   int blockp = isBlocking_p();  // are we blocking?
   if (!blockp) setBlocking();   // do blocking send below
     
   int status = sendUnixHdr(getfd(), &mh);

   if (!blockp) setNonBlocking();  // restore non-blocking status if nec

   return status;
}
