#ifndef Stream_Socket_Class
#define Stream_Socket_Class

#include <sys/types.h>
// Following is to include <netinet/tcp.h> which lacks #ifndef gate on SGI:
#include "SpOS.H"
#include "SpErrno.H"

#ifndef __SPEEDES_NT__
#include <unistd.h> // needed by netinet/in.h on HPUX
#include <netinet/in.h> // needed by bindSock's default argument
#include <sys/uio.h>
#include <fcntl.h>
#include <sys/socket.h>
#else
#include <windows.h>
#endif

#include "SpInternalMsgHdr.H"
#include "SpSockMsgType.H" // needed by STREAM_SOCK::NullMsg


extern const int MAX_MSG_BYTES;     // 100 million bytes max!
extern const int NO_MSG_AVAIL;      // Old status code

extern const int DEFAULT_PORT;      // Arbitrary port# > 1024
extern const int ACCEPT_Q_LENGTH;   // Q size of guys waiting to connect
extern const int NO_FD;		    // not a file descriptor


class msghdr;


class STREAM_SOCK {
 private:
   int SocketStatus; // one of these enums: NO_INFO, CLIENT, SERVER,  CONNECTED, I_HUNGUP, OTHER_HUNGUP
   int Sockfd;       // socket file descriptor

   int Blocking_p;  // is this socket in blocking mode?
   int StatusFlags; // status bits gotten/set by fcntl unix call

   int sendUnixHdr(int fd, msghdr *mh, int flags=0);

   static inline int classifySockError(int unixErrorNumber, int sending_p);

   int getInternalHdr(int connfd, internalMsgHdr *msgHdr,
		      long *head, int *bytesRead);

 protected:
   enum { NO_INFO, CLIENT, SERVER,  CONNECTED, I_HUNGUP, OTHER_HUNGUP };  // status states
   int PortNum;    // connect via this port#
  
   static const SockMsgType NullMsg; // buffer, length, bytesRecvd
  
   void setSocketStatus(int status) { SocketStatus = status; }
   int getSocketStatus(void) { return SocketStatus; }
   void setConnected(void) { setSocketStatus(CONNECTED); }
  
   void setfd(int in) { Sockfd = in; }

   void updateStatusFlags(void) {
#ifndef __SPEEDES_NT__
      if ((StatusFlags = fcntl(Sockfd, F_GETFL, 0)) < 0) {
	 perror("setStatusFlags F_GETFL call:");
      }
#endif
   }

   int getStatusFlags(void) {
      if (StatusFlags < 0) {
	 updateStatusFlags();
      }
      return StatusFlags;
   }

   void setStatusFlags(int statFlgs) {
      StatusFlags = statFlgs;
   }

  
 public:
   // Make THIS streamSock non-blocking (set_nonblocking(sock) makes ANY 
   // sock file descriptor non-blocking)
   int setNonBlocking(void) {
      int statusFlags, ret=0;

#ifndef __SPEEDES_NT__
      if ((statusFlags = getStatusFlags()) >= 0) {
	 statusFlags |= O_NDELAY;
	 if (fcntl(Sockfd, F_SETFL, statusFlags) < 0) {
	    perror("setNonblocking F_SETFL call:");
	 }
	 else {
	    Blocking_p = 0;
	    setStatusFlags(statusFlags);
	    ret = 1;
	 }
      }
#else
      static unsigned long int junk = 1;
      if (ioctlsocket(Sockfd, FIONBIO, &junk) == SOCKET_ERROR){
	switch(WSAGetLastError()){
	case WSANOTINITIALISED:
	  cout<<"Did not startup the winsock libraries"<<endl;
	  break;
	case WSAENETDOWN:
	  cout<<"Network subsystem has failed"<<endl;
	  break;
	case WSAEINPROGRESS:
	  cout<<"A blocking Windows Sockets 1.1 call is in progress"<<endl;
	  break;
	case WSAENOTSOCK:
	  cout<<"Sockfd "<<Sockfd<<" is not a socket"<<endl;
	  break;
	case WSAEFAULT:
	  cout<<"argp argument not valid part of the user address space"<<endl;
	  break;
	}
	ret = 0;
      }
      else{
	Blocking_p = 0;
	ret = 1;
      }
#endif

      return ret;
   }


   // Make THIS streamSock blocking (set_blocking(sock) makes ANY 
   // sock file descriptor blocking)
   int setBlocking(void) {
     int statusFlags, ret=0;
     
#ifndef __SPEEDES_NT__
     if ((statusFlags = getStatusFlags()) >= 0) {
       statusFlags &= ~O_NDELAY;
       if (fcntl(Sockfd, F_SETFL, statusFlags) < 0) {
	 perror("setBlocking: setting status");
       }
       else {
	 Blocking_p = 1;
	 setStatusFlags(statusFlags);
	 ret = 1;
       }
     }
#else
     static unsigned long int junk = 0;
     int returncode;
     if ((returncode=ioctlsocket(Sockfd, FIONBIO, &junk)) == SOCKET_ERROR){
       switch((returncode = WSAGetLastError())){
       case WSANOTINITIALISED:
	 cout<<"Did not startup the winsock libraries"<<endl;
	 break;
       case WSAENETDOWN:
	 cout<<"Network subsystem has failed"<<endl;
	 break;
       case WSAEINPROGRESS:
	 cout<<"A blocking Windows Sockets 1.1 call is in progress"<<endl;
	 break;
       case WSAENOTSOCK:
	 cout<<"Sockfd "<<Sockfd<<" is not a socket"<<endl;
	 break;
       case WSAEFAULT:
	 cout<<"argp argument not valid part of the user address space"<<endl;
	 break;
       default:
	 cout<<"Unknown error in setting blocking mode "<<returncode<<endl;
       }
     }
     else{
       Blocking_p = 1;
     }
#endif
     
     return ret;
   }

  int setNoDelay(void) {
    int ret = 1;
    int nodelayVal = 1;  // value assigned by setsockopt to nodelay boolean
   
    if (setsockopt(Sockfd,
		   IPPROTO_TCP,
		   TCP_NODELAY,
		   (char *)&nodelayVal,
		   sizeof(int)) < 0) {
      perror("setNoDelay");
      ret = 0;
    }
    return ret;
  }


   STREAM_SOCK(int port = DEFAULT_PORT);
  
  
  
   // Kludgestructor used by CONNECTION so that we don't automatically
   // open up a new socket as in the 1-arg constructor above:
   STREAM_SOCK(int fd, int port) : Sockfd(fd) {
      PortNum = port;
      // Cannot do setNonBlocking() here so mark this as un-set:
      Blocking_p = -1;
      updateStatusFlags(); // set data member copy of Unix fcntl status flags
      setNoDelay();
   }
  
   ~STREAM_SOCK(void);


   // Status codes.  Note the ERR_OKCODE_SEPARATOR;  all errors MUST be less than
   // this and all other "OK" codes MUST be greater (see ErrorStatus, OKStatus methods)
   // A similar separation has been done for fatal/non-fatal errors
   enum Socket_Status_Code { 
      CLOSED_CONN = -100,      // Connection is closed
      BAD_SOCK,                // Bad file descriptor for connection (probably closed conn)
      BAD_PTR_OR_ARG,          // Bad ptr in sendmsg msghdr or bad msg size arg to sendmsg
      NOT_YET_CONNECTED,       // Possibly socket is ok, but still doing connection internal work
      MSG_TOO_BIG,             // Received msg internal hdr mangled (#bytes > MAX_MSG_BYTES!)
      UNKNOWN_SOCK_ERROR,      // Don't know what the hell is wrong;  better bail out.
      BAD_INTERNAL_MSG,        // Unrecognized internal msg (probably mangled)
      
      FATAL_NONFATAL_SEPARATOR,// ****** separates fatal errs from non-fatal ******

      ZERO_BYTES_READ,         // Read 0 bytes (means closed conn is SELECT said ready-to-read)
      TEMP_SOCK_PROBLEM,       // eg Unix buffer full;  temporary problem that MAY go away
      USER_BUFF_TOO_SMALL,     // User buff too small to hold incoming msg
                               // in readMsgIntoBuff or resumeReadMsgIntoBuff

      ERR_OKCODE_SEPARATOR = 0,// ****** separates errors from other status codes ******

      WHOLE_MSG_OK,            // read/wrote entire msg
      PARTIAL_MSG_OK,          // read/wrote partial msg
      INTERNAL_HEADER_OK,      // read/wrote internal msg header
      PARTIAL_INTERNAL_HDR_OK, // read/wrote partial msg header
      INTERNAL_MSG_DONE,       // read & executed internal msg
      NO_MSGS_SENT             // haven't sent any messages
   };

   static int ErrorStatus(int status) {
      return( status < ERR_OKCODE_SEPARATOR );
   }

   static int FatalErrorStatus(int status) {
      return( status < FATAL_NONFATAL_SEPARATOR );
   }

   static int NonFatalErrorStatus(int status) {
      return( (status < ERR_OKCODE_SEPARATOR) &&
	      (status > FATAL_NONFATAL_SEPARATOR) );
   }

   static int OKStatus(int status) {
      return( status > ERR_OKCODE_SEPARATOR );
   }

   static int PartialMsg(int status) {
      return( (status>WHOLE_MSG_OK) && (status<INTERNAL_MSG_DONE) );
   }



   // Bind (port, ipAddr) to socket.  This pair uniquely specifies a
   // location for clients to connect to.  The default INADDR_ANY
   // means "accept connections via ANY of the network inter- faces on
   // this machine"
   void bindSock(u_short portNum=DEFAULT_PORT, u_long inetAddr=INADDR_ANY);

   int closeSock(int fd);
    
   int closeSock(void) {
      return closeSock( Sockfd );
   }

   int getfd(void) {return Sockfd; }

   int set_nodelay(int sock);

   int unset_nodelay(int sock);

   int set_nonblocking(int sock); // not static anymore

   int set_blocking(int sock);  // not static anymore

   // Is ANY sock file descriptor blocking?
   int isBlocking_p(int sock);

   // Is THIS STREAM_SOCK blocking?
   int isBlocking_p(void) {
      return Blocking_p;
   }


   //senMsg: send msg thru connection denoted by connfd.  The msg is
   //defined by args msg and len.  The arg mh is an OUT param which is
   //used in case of a partial send.  It contains all the info needed
   //to resume the send when the write buffer clears (see
   //resumePartailSend).
   // The arg, "hdrSpace," is a permanent buffer that will be used to store the
   // internal msg header.  The reason it's an arg is because it needs to be "permanent"
   // in case of a partial send, and this is more efficient than doing a "new" and "delete"
   // 99% of the time.  Make sure it has at least HEADER_SIZE bytes.
  // Returns one of the following:
  // WHOLE_MSG_OK  :   entire msg sent;  everything is OK
  // PARTIAL_MSG_OK:   part msg sent;  everything OK
  // TEMP_SOCK_PROBLEM: temporary problem (eg full buffer, interupted call) that _may_ fix itself
  // BAD_SOCK:  Bad file descriptor for connection (probably closed conn)
  // BAD_PTR_OR_ARG:  Bad ptr in sendmsg msghdr or bad msg size arg to sendmsg
  // NOT_YET_CONNECTED: Possibly socket is ok, but still doing connection internal work
  // CLOSED_CONN: Connection is closed
  // UNKNOWN_SOCK_ERROR:  Don't know what the hell is wrong;  better bail out.
  int sendMsg(int connfd, char *msg, int len,
	      msghdr *mh, long *hdrSpace, int flags=0);

  // resumePartialSend: continue the partial send via connfd.  mh was
  // set previously by a sendMsg call and contains the state of the
  // suspended send.
  // Return vals:  same as sendMsg
  int resumePartialSend(int connfd, msghdr *mh, int flags=0) {
    return sendUnixHdr(connfd, mh, flags);
  }

  static int getMsgHeader(int connfd, internalMsgHdr *mhdr);

   // Jacob thinks no one calls this: SockMsgType receiveMsg(int
   // connfd, int flags=0);

  int recvMsg_blk(int connfd, SockMsgType *msg, int flags);

   //            receiveMsg
   // Returns one of the following:
   // WHOLE_MSG_OK  :   entire msg sent;  everything is OK
   // PARTIAL_MSG_OK:   part msg sent;  everything OK
   // ZERO_BYTES_READ: read 0 bytes; if select says sock is readable but we get 0 bytes,
   //                  this means that the connection has been closed by the peer.
   // MSG_TOO_BIG: incoming msg bigger than MAX_MSG_BYTES (100 MEG!);  probably mangled msg
   // TEMP_SOCK_PROBLEM: temporary problem (eg full buffer, interupted call) that _may_ fix itself
   // BAD_SOCK:  Bad file descriptor for connection (probably closed conn)
   // BAD_PTR_OR_ARG:  Bad ptr in sendmsg msghdr or bad msg size arg to sendmsg
   // NOT_YET_CONNECTED: Possibly socket is ok, but still doing connection internal work
   // CLOSED_CONN: Connection is closed
   // UNKNOWN_SOCK_ERROR:  Don't know what the hell is wrong;  better bail out.
  int receiveMsg(int connfd, SockMsgType *msg, int flags=0);

   //               resumeRecvMsg
   // Resume receiving partial msg;  return TRUE if got rest of msg
   // Returns one of the following:
   // WHOLE_MSG_OK  :   entire msg sent;  everything is OK
   // PARTIAL_MSG_OK:   part msg sent;  everything OK
   // ZERO_BYTES_READ: read 0 bytes; if select says sock is readable but we get 0 bytes,
   //                  this means that the connection has been closed by the peer.
   // TEMP_SOCK_PROBLEM: temporary problem (eg full buffer, interupted call) that _may_ fix itself
   // BAD_SOCK:  Bad file descriptor for connection (probably closed conn)
   // BAD_PTR_OR_ARG:  Bad ptr in sendmsg msghdr or bad msg size arg to sendmsg
   // NOT_YET_CONNECTED: Possibly socket is ok, but still doing connection internal work
   // CLOSED_CONN: Connection is closed
   // UNKNOWN_SOCK_ERROR:  Don't know what the hell is wrong;  better bail out.
  int resumeRecvMsg(int connfd, SockMsgType *partMsg, int flags=0);



   //               readMsgIntoBuff
   // Same as receiveMsg except the user must provide a PERMANENT (non-stack) buffer
   // into which the msg will be read.  The arg, buffSize, gives the size in bytes
   // of buff;  if the incoming msg is too big, returns USER_BUFF_TOO_SMALL and
   // sets *msgBytes to be the size of the msg.  At this point, the user can get a
   // bigger buffer and call resumeReadMsgIntoBuff (since the internal header has
   // already been read, we must "resume" the receive), or else call resumeRecvMsg
   // (which will allocate a sufficient buffer by itself).
   // If there is a partial read (return PARTIAL_MSG_OK), *msgBytes will be set to
   // the number of bytes read thus far.  If we cannot even read the internal hdr
   // (and therefore do not know the msg length yet), *msgBytes is set to -1.
   int readMsgIntoBuff(int connfd, SockMsgType *msg, 
		       char *buff, int buffSize, long int *msgBytes);

   //            resumeReadMsgIntoBuff
   // This is the partner to readMsgIntoBuff.  It attempts to finish reading a
   // partially-received msg (which may require 2 or more calls).  Note that this
   // fcn can also return USER_BUFF_TOO_SMALL (this will occur when readMsgIntoBuff
   // cannot even get an entire internal msg hdr and so resumeReadMsgIntoBuff
   // first determines that there is not enough space in the user buffer).
   int resumeReadMsgIntoBuff(int connfd, SockMsgType *partMsg,
			     char *mbuff, int buffSize, long int *msgBytes);


  int send_data(int socket, char *buffer, int length, int flags=0);

  int nb_send_data(int sock, char *buff, int bytes, int flags=0);

  int receive_data(int socket, char *buffer, long int length, int flags=0);

  static int nb_receive_data(int sock,char *buff, long int bytes, int flags=0);

  static int peekAtReceive(int sock, char *buff, int bytes) {
     return nb_receive_data(sock, buff, bytes, MSG_PEEK);
  }

  int doInternalMsg(internalMsgHdr *hdr);

  void sendCloseConnectMsg(void);
  int sendInternalMsg(internalMsgHdr& hdr);

  int dataReadyForRead_p(void) {
    internalMsgHdr head;
    
    int bytes = peekAtReceive(Sockfd, (char *) &head, HEADER_SIZE);
    
    return (bytes > 0);
  }

};


#endif
