#ifndef CIRCBUFF_H
#define CIRCBUFF_H



/*************************************************************************
 * Class: CIRC_BUFF_HEADER
 * Summary: This is a header for messages that will be inserted into the 
 *          circular buffer.
 * Description: Circ buff headers consist of several pieces.  They
 *               have a ReadFlag which indicates if the message has been
 *               read or not, an Offset to where the message actually
 *               starts, the number of bytes of the message along with
 *               the message type.  It is important that the header
 *               is of the same size as the machine so that it alligns
 *               on (n) bytes boundaries.  Presently, is should be
 *               8 bytes alligned.
 * Examples: 
*************************************************************************/

class CIRC_BUFF_HEADER {
private:
  int ReadFlag;
  int Offset;
  int Nbytes;
  int Type;
protected:
public:
    CIRC_BUFF_HEADER() : ReadFlag(0), Offset(0), Nbytes(0), Type(0) {}

  /*************************************************************************
   * Method: init
   * Summary: set the header members to the default values
   * Description: A header defaults to having everything set to zero.  That
   *              is, it has 0 offset, is unread, has no bytes and is of
   *              type 0.
   * Return: void
   *************************************************************************/
  void init() {
    Offset = ReadFlag = Nbytes = Type = 0;
  }
  
  /*************************************************************************
   * Method: init
   * Summary: set the header members to specific values
   * Description: This sets the number of bytes, the offset, and the type to
   *              the passed in parameters.  the read flag is set to unread.
   * Return: void
   *************************************************************************/
  void init(const int n, //IN: The number of bytes of the message
	    const int o, //IN: The offset to the message
	    const int t  //IN: The type of the message
	    ) {
    Nbytes = n;
    Offset = o;
    Type = t;
    ReadFlag = 0;
  }
  
  /*************************************************************************
   * Method: SetReadFlag
   * Summary: Set the read flag to indicate the message has been read
   * Description: 
   * Return: void
   *************************************************************************/
  void SetReadFlag() {ReadFlag = 1;}

  /*************************************************************************
   * Method: ResetReadFlag
   * Summary: Set the read flag to indicate the message has not been read
   * Description: 
   * Return: void
   *************************************************************************/
  void ResetReadFlag() {ReadFlag = 0;}

  /*************************************************************************
   * Method: GetReadFlag
   * Summary: return the value of the ReadFlag
   * Description: 
   * Return: value of the ReadFlag (0 means unread, 1 means read)
   *************************************************************************/
  int GetReadFlag() {return ReadFlag;}
  
  /*************************************************************************
   * Method: SetOffSet
   * Summary: Set the offset to the message to the passed in value
   * Description: A message is found at (char *)this + offset.  This simply
   *              sets the offset value so that we may find the message.
   * Return: void
   *************************************************************************/
  void SetOffset(const int o //IN: The number of bytes to the message.
		 ) {Offset = o;}

  /*************************************************************************
   * Method: GetOffSet
   * Summary: return the value of the offset
   * Description: A message is found at (char *)this + offset.  This simply
   *              sets the offset value so that we may find the message.
   * Return: number of bytes to the message.
   *************************************************************************/
  int GetOffset() {return Offset;}

  /*************************************************************************
   * Method: SetType
   * Summary: Sets the type of the message to the passed in value
   * Description: 
   * Return: void
   *************************************************************************/
  void SetType(const int t //IN: Message type
	       ) {Type = t;}

  /*************************************************************************
   * Method: GetType
   * Summary: Return the type of the message.
   * Description: 
   * Return: message type.
   *************************************************************************/
  int GetType() {return Type;}
  
  /*************************************************************************
   * Method: SetNbytes
   * Summary: Sets the number of bytes (size) of the message to the passed
   *          in value.
   * Description: 
   * Return: void
   *************************************************************************/
  void SetNbytes(const int n //IN: Number of bytes of the message
		 ) {Nbytes = n;}

  /*************************************************************************
   * Method: GetNbytes
   * Summary: Returns the size (number of bytes) of the message.
   * Description: 
   * Return: size of the message
   *************************************************************************/
  int GetNbytes() {return Nbytes;}
  
};

#define CBHeaderSize ((int)sizeof(CIRC_BUFF_HEADER))

/*************************************************************************
 * Class: CIRCULAR_BUFFER
 * Summary: This is a circular buffer for messages that allows reading
 *          writing and examining of messages.
 * Description: A circular buffer maintains a "contiguous" space of
 *              messages.  Messages are inserted at the head and then
 *              deleted from the tail as they are read.  If there is not
 *              enough space for a message at the end of a buffer (the
 *              head is ahead of the tail), then a "fuzzy end" to the
 *              buffer is created and the message is inserted at the
 *              start of the buffer.
 *
 *              Finally, if there is no room available, garbage collect
 *              is called and we attempt to clear out any messages that
 *              have been read.  Once this is done, we try again to
 *              insert the message.  If this fails, then we return that
 *              there is no room for the message.
 *              
 *   x = message/data
 *   0 = CIRC_BUFF_HEADER
 *   ? = empty space since the next message would not fit.
 *   
 *   **   [0xxxxxx____________________0xxxxxxxxxxx00xxxxxxxxx?????????????]
 *                ^                   ^                      ^
 *               Head                 Tail              BufferFuzzyEnd
 *   
 *    First write the message size, then write the message.  No message
 *    ever crosses over the end of the CIRCULAR_BUFFER.  If there is
 *    not enough room to write the piece at the end but there is room
 *    at the begining, the BufferFuzzyEnd is set to the present Head
 *    value and Head is set to zero with the message written there.
 *    
 *    Note: If Head == Tail, we have a completely empty buffer so tail
 *          is always behind head if there is any filled in space in
 *          the buffer.
 *
 * Examples: 
*************************************************************************/
class CIRCULAR_BUFFER {

private:
  char * Buff;
  int Head;
  int Tail;
  int BuffSize;
  int I_Created;
  int BufferFuzzyEnd;
  int Recheck;     /* Do we need to recheck for space or have we
		    * already done that */
protected:
  /*************************************************************************
   * Method: Allign8Bytes
   * Summary: returns a value greater than the parameter but divisible by 8
   * Description: This makes sure that a number is divisible by 8.  It
   *              will return the next number greater than val which is
   *              divisible by 8.  This also means that if a number which
   *              is already divisible by 8 is passed in, then 8 is added
   *              to that number and returned.
   * Return: integer divisible by 8 but greater than the parameter val
   *************************************************************************/
   int Allign8Bytes(int val //In: Number of bytes
		   )
  {return val + (8 - val%8);}

  /*************************************************************************
   * Method: GarbageCollect
   * Summary: Free's circular buffer space that has already been read.
   * Description: Garbage collect attempts to advance the position of the
   *              member tail.  It checks messages at the tail to see if they
   *              have been read.  If they have, it then advances tail by the
   *              size of the CIRC_BUFF_HEADER.  If everyone has read this
   *              message, then tail is advance also by the size of the
   *              message.  If tail reaches the buffer's fuzzy end, it is
   *              reset to zero and continues.  If it reaches the head,
   *              it returns since there are no more messages in the circular
   *              buffer
   * Return: void
   *************************************************************************/
  void GarbageCollect();
  
  
  /*************************************************************************
   * Method: CheckSpace
   * Summary: Checks to see if there is sufficient space in the circular
   *          buffer to write the given message.
   * Description: CheckSpace looks at the buffer and checks to see if there
   *              is space for the message plus as many circ_buff_headers
   *              as is required.  We first check to see if the tail has
   *              caught up with the head and, if so, we simply return
   *              the head.
   *              
   *              If not, we then check to see if the tail is following
   *              the head.  If so, return  the head if there is room at
   *              the end of the buffer.  Otherwise, set the fuzzy end and
   *              attempt to put the message at the start.  If all of this
   *              fails, garbage collect and try once more.
   *
   *              If the tail is ahead of the head, try to put the message
   *              at the head but before the tail.  If no room, garbage
   *              collect and try again.
   *
   * Return: integer giving location of where one can put the buffer or
   *         a -1 if there is no room for the message
   *************************************************************************/
  int CheckSpace(const int Givenbytes, //In: number of bytes of the message
		 const int n=1         //In: Number of destinations of message
		 );
  
public:
  
  /*************************************************************************
   * Method: CIRCULAR_BUFFER()
   * Summary: Constructor with a given size
   * Description: Creates a circular buffer of size sz.
   * Return: CIRCULAR_BUFFER
   *************************************************************************/
  CIRCULAR_BUFFER(const int sz //In: Size of the circular buffer to be created
		  );


  /*************************************************************************
   * Method: CIRCULAR_BUFFER()
   * Summary: Constructor with a given size and a pre-created buffer 
   * Description: Creates a circular buffer out of the buffer Buffer with
   *              size sz.  Buffer can be in shared memory which would allow
   *              many processes to access it
   * Return: CIRCULAR_BUFFER
   *************************************************************************/
  CIRCULAR_BUFFER(const int sz, //In: Size of the circular buffer to be created
		  char * Buffer //Buffer in which the circular buffer will
		                //reside.
		  );
  
  /*************************************************************************
   * Method: ~CIRCULAR_BUFFER
   * Summary: Destructor for a circular buffer
   * Description: If we were not passed in a buffer for creation, then 
   *              delete the buffer.  Otherwise, just return.
   * Return: 
   *************************************************************************/
  ~CIRCULAR_BUFFER(){
    if (I_Created) delete [] Buff;
  }
  
  /*************************************************************************
   * Method: Write
   * Summary: Places a message into the circular buffer
   * Description: This method first checks for space in the circular buffer.
   *              If it finds the space, it puts in n headers, followed by
   *              the message itself and returns the index of the first
   *              CIRC_BUFF_HEADER.  If it cannot find space, it returns -1
   * Return: Index of the first CIRC_BUFF_HEADER on success or -1 on failure
   *************************************************************************/
  int Write(const int bytes,   //In: Size of message in bytes
	    const char * buff, //In: Message in this buff
	    const int type,    //In: Type of message
	    const int n=1      //In: Number of recipients
	    );
  
  /*************************************************************************
   * Method: SetReadFlag
   * Summary: Sets the read flag at the given index
   * Description: Once one has read a message and no longer needs it in the
   *              buffer, the read flag must be set so that garbage collect
   *              can free the memory
   * Return: void
   *************************************************************************/
  void SetReadFlag(const int index //In: Index at which we will set the read
		                   //    flag
		   ) {
    CIRC_BUFF_HEADER *CircBuffHeader = (CIRC_BUFF_HEADER *)&Buff[index];
    CircBuffHeader->SetReadFlag();
  }

  /*************************************************************************
   * Method: GetOffSet
   * Summary: Returns the offset to the message
   * Description: The message occurs at a given space forward from the
   *              index of the CIRC_BUFF_HEADER.  This returns the index
   *              of where the message can be found.
   * Return: number of bytes to the message.
   *************************************************************************/
  int GetOffset(const int index //In: Index of the HEADER for which we want
		                //    the message
		) {
    CIRC_BUFF_HEADER *CircBuffHeader = (CIRC_BUFF_HEADER *)&Buff[index];
    return CircBuffHeader->GetOffset();
  }
  
  /*************************************************************************
   * Method: GetBuff
   * Summary: Returns a pointer to the message at a given index.
   * Description: GetBuff finds the message that is indicated by the header.
   *              We first go to the header, add on GetOffset() bytes, and 
   *              then return the char * at that point.  If we have already
   *              read this message (should not happen), we will print a
   *              warning message.
   * Return:  pointer to message which is indicated by the header
   *************************************************************************/
  char * GetBuff(const int index, //In: Index of the header
		 int &Type,       //Out: Type of the message pointed to by the
		                  //     header
		 int &Nbytes      //Out: Size of the message pointed to by the
		                  //     header
		 );
  
  /*************************************************************************
   * Method: ReadBuff
   * Summary: Returns a pointer to the message at a given index.
   * Description: GetBuff finds the message that is indicated by the header.
   *              We first go to the header, add on GetOffset() bytes, and 
   *              then return the char * at that point.  If we have already
   *              read this message (should not happen), we will print a
   *              warning message.  The message is also marked as read.
   * Return:  pointer to message which is indicated by the header but the
   *          memory is allocated by new (unlike GetBuff) and the message
   *************************************************************************/
  char *ReadBuff(const int index, //In: Index of the header
		 int &Type,       //Out: Type of the message pointed to by the
		                  //     header
		 int &Nbytes      //Out: Size of the message pointed to by the
		                  //     header
		 );
  
};

#endif // CIRCBUFF_H
