// SpCommShMem.H header file defines all of the special classes used in the
// shared memory SPEEDES Comm.

#ifndef SP_COMM_SHMEM
#define SP_COMM_SHMEM

#define MESSAGE_SENT 0
#define MESSAGE_FAILED 1
#define MESSAGE_POSTPONED 2

#include "SpCircularBuff.H"
#include "SpShMemQueue.H"
#include "SpComm.H"


/*************************************************************************
 * Class: SHMEM_REDOP
 * Summary: Creates a structure for the shared memory reductions
 * Description: Each shared memory reduction stage requires that certain
 *              values be updated.  These include the Toggle, the Ivalue
 *              Dvalue, and SimTime.
 * Examples: Look at the shared memory reduction operations.
 *************************************************************************/

class SHMEM_REDOP {
private:
protected:
public:
  SHMEM_REDOP(){init();}

  int Toggle;  // Indicates if this value has been set yet.  A global
               // variable indicates the previous state of the toggle
               // and we can check this value against that.
  int Ivalue;  // Integer value used for integer valued reductions
  double Dvalue; // Double value used for double valued reductions
  SpSimTime SimTime; // SpSimTime value used for SpSimTime valued reductions
  
  /*************************************************************************
   * Method: init
   * Summary: initialize redop values to 0
   * Description: Before the reductions are started, we want the values
   *              of Toggle, Ivalue, Dvalue, SimTime to be initialized
   *              to known values.  This sets them all to zero.
   * Return: void
   *************************************************************************/
  void init() {
    Toggle = 0;
    Ivalue = 0;
    Dvalue = 0.0;
    SimTime = 0.0;
  }
};

/*************************************************************************
 * enum: 
 * Summary: These are some return codes for certain types of message
 *          sendings as well as special shared memory message types.
 * Description: Shared memory needs some special return codes as well as
 *              some special message types for internal messages.
 *              SpeedesComm uses only 256 message types (0-255) so we use
 *              integers greater than 255 to enumerate any special messages
 *              we may need to take care of.
*************************************************************************/
enum{
  MSG_READ=-5,                 // Indicates a shared memory message was read
  SHMEM_MULTISEND_ALL=0,       // Was able to send all of multi/broadcast
  SHMEM_MULTISEND_PARTIAL=1,   // At least got the buffer for multi/broad
  SHMEM_MULTISEND_NONE=2,      // Did not send any of the multi/broadcast

  SHMEM_MESSAGE_TYPES=7,            // Number of shared memory types
  SPCOMM_MESSAGE_REROUTE=256,       // Reroute message type
  SPCOMM_MESSAGE_BROADCAST=257,     // Broadcast message type
  SPCOMM_MESSAGE_MULTICAST=258,     // Multicast message type
  SPCOMM_MESSAGE_BLOCKING_SEND=259, // Blocking send message type
  SPCOMM_MESSAGE_ANNOUNCE=260,      // Announce message type
  SPCOMM_MESSAGE_PARTIAL=261,       // A message is following in pieces
  SPCOMM_MESSAGE_PARTIAL_END=262,   // A partial message send has completed
  SPCOMM_MESSAGE_PARTIAL_DESTINATION=263,// DESTINATION of a partial send
  SPCOMM_MESSAGE_PARTIAL_MULTICAST_BEGIN=264,
  SPCOMM_MESSAGE_PARTIAL_BROADCAST_BEGIN=265
};



class SHMEM_MESSAGE;
extern SHMEM_MESSAGE **SpComm_ShmemMessage;



/*************************************************************************
 * Class: SHMEM_MESSGE
 * Summary: A class for managing messages within shared memory
 * Description: SHMEM_MESSAGE takes care of the reading, writing, and
 *              announcing of messages.  In order delivery is guaranteed
 *              and each message is guaranteed to be delivered.
 * Examples: Look at SpCommShMem.C
*************************************************************************/
class SHMEM_MESSAGE {
  
private:
  
  int AlreadyDeleted;
  char * SharedMemoryAddress; //Store pointer for later freeing
  int OutputBufferSize;
  int MaxMessages;
  int Nbytes;
  int ShmemId;
  int MyLocalNode;
  int NlocalNodes;
  int ErrorStatus;

  int SendingPartialMessage;
  int PartialMessageType;
  char * OutputBuffer;
  char ** PartialMsg;
  char ** PartialMsgPtr;
  int PartialMsgByteCount[1000];
  char * PartialSendMsg;
  SpDestination PartialSendDestination;
  BUFFER_QUEUE InQueue[N_MESSAGE_TYPES+SHMEM_MESSAGE_TYPES]; 
  BUFFER_QUEUE TempSendQueue;
  BUFFER_QUEUE AnnounceQueue;

  int *ReadIndex;
  
  
  /* These point to shared memory.  NEVER TOUCH POINTERS EXCEPT TO
  INITIALIZE/DELETE */

  int *LastReadIndex;
  int *LastWriteIndex; /* This keeps track of where we last let the
			  destination know that we put a message */
  int **InputArray;  /* This is where we look to see if we have any
			messages from node i in InputArray[i][LastReadIndex] */
  
  int MsgInOutCount;

protected:
  int FlushSendQueue();
  int TrySend(const int type, const int nbytes, const void * buff);
  int TrySend(SpDestination *dest, const int type, 
	      const int nbytes, const void * buff);
  int TrySend(int *dest, const int type,
		    const int nbytes, const void * buff);
  int TrySend(const int destination, const int type,
	      const int nbytes, const void *buff);

  int Announce(const int dest, int *Index){
    int NextWriteIndex = (LastWriteIndex[dest]+1)%MaxMessages;
    int RC = 0;
    if (dest < 0){
      cout<<"ACK!!!!!!!"<<endl;
      strcpy (NULL, "sadfasd");
    }
    
    if(MSG_READ ==
       SpComm_ShmemMessage[dest]->InputArray[MyLocalNode][NextWriteIndex])
      {
	SpComm_ShmemMessage[dest]->InputArray[MyLocalNode][NextWriteIndex]
	  = *Index;
	LastWriteIndex[dest] = NextWriteIndex;
	*Index = NextWriteIndex;
      }
    else RC = 1; // if we don't have a write index
    return RC;
  }

  void EnQueuePartialMsg(char * Msg,
			 int Dest,
			 int MsgLen,
			 int type,
			 int MaxMsgLen);

  void EnQueuePartialMsg(char * Msg,
			 SpDestination *Dest,
			 int MsgLen,
			 int type,
			 int MaxMsgLen);

  void EnQueuePartialMsg(char * Msg,
			 int MsgLen,
			 int type,
			 int MaxMsgLen);

  /*
   * Initialize should be called after the fork() that each process can
   * call shmat on their own.  The reason it should be done after the fork()
   * is because some operating systems (non-UNIX platforms specifically) may
   * not make SHMATed pointers valid unless it is called for each process
   * on it's own.
   */
public:

  int GetShmemId(){return ShmemId;}
  void SetShmemId(int id){ShmemId = id;}
  
  
  CIRCULAR_BUFFER *OutgoingMessageBuffer;  // This is where all messages are
                                           // when they are sent to another
                                           // node on the same box.

  
  /*************************************************************************
   * Method: SHMEM_MESSAGE
   * Summary: constructs the shared memory structure.
   * Description: The constructor sets the member variables to the
   *              values passed in.  It creates a circular buffer of size
   *              OutbufferSize and an array of size InArraySize in
   *              shared memory.  These are then initialized and cleared.
   * Return: the constructed SHMEM_MESSAGE
   *************************************************************************/
  SHMEM_MESSAGE(const int OutBufferSize, //In: size of the output circular
		                         //    buffer
		const int InArraySize,   //In: size of the array of writes
		const int Nloc,          //In: number of local nodes
		const int LocId,         //In: local node ID for this shared
		                         //    memory segment
		const int shmemID = -1
		);

  /*************************************************************************
   * Method: Initialize
   * Summary: Sets up the pointers in the shared memory segment
   * Description: This is not called in the constructor since we need to
   *              create the shared memory segment (SHMGET()) before the
   *              fork() but we need to do a SHMAT() after the fork().
   *              Therefore, the initialize call was added to do the
   *              SHMAT() after the fork() while the constructor does the
   *              SHMGET() before the fork.
   * Return: void
   *************************************************************************/
  void Initialize(const int OutBufferSize  //In: size of the output circular
		                           //    buffer
		  );
  
  /*************************************************************************
   * Method: ~SHMEM_MESSAGE
   * Summary: destructs a SHMEM_MESSAGE
   * Description: this deletes much of the memory allocated by the
   *              constructor. It also releases the shared memory segment(s)
   *              that were allocated by the constructor.  It should only
   *              be called by one node (likely node 0)
   * Return: void
   *************************************************************************/
  ~SHMEM_MESSAGE();
  
  /*************************************************************************
   * Method: GetMyLocalNode
   * Summary: returns the local node id of the shared memory segment
   * Description: returns the local node id of the shared memory segment
   * Return: returns the local node id of the shared memory segment
   *************************************************************************/
  int GetMyLocalNode() {return MyLocalNode;}
  
  /*************************************************************************
   * Method: GetErrorStatus
   * Summary: returns the error status that was set in the constructor
   * Description: returns the error status that was set in the constructor
   * Return: returns the error status that was set in the constructor
   *************************************************************************/
  int GetErrorStatus() {return ErrorStatus;}
  
  /*************************************************************************
   * Method: ClearBuffers
   * Summary: clears the InputArray buffers.
   * Description: This initializes much of the shared memory segments.  It
   *              sets each value of the InputArray to the value MSG_READ
   *              to indicate that each message has been read.  It also
   *              sets LastWriteIndex and LastReadIndex to -1 for each node.
   *              This says that they are both looking at the same place
   *              for the next message.
   * Return: void
   *************************************************************************/
  void ClearBuffers() {
    for (int i=0; i<NlocalNodes; i++) {
      for (int j=0; j < MaxMessages; j++) InputArray[i][j] = MSG_READ;
      LastWriteIndex[i] = -1;      // Next message to read is at 0
      LastReadIndex[i] = -1;       // Next message to write is at 0
    }
  }

  /*************************************************************************
   * Method: Send
   * Summary: Sends a message to a single destination.
   * Description: First, we check to see if we need to flush any messages
   *              from the send queue.  If we do, then we add this message
   *              to the queue of messages.  If we don't, then we try to
   *              send the message.  If we are unable to do so, we again add
   *              the message to the send queue.
   * Return: void
   *************************************************************************/
  void Send(const int dest,  //In: The destination of the message
	    const int type,  //In: The type of the message
	    const int nbytes,//In: The number of bytes of the message
	    const void *buff //In: The message itself in a void* array
	    )
  {
    int AddToQueue = 0; // Flag to indicate if we need to add this message
                        // to the TempSendQueue.
    if (dest == MyLocalNode){ //If send to self
      InQueue[type].AddToTail((char *)buff, nbytes, type);
    }
    else if (nbytes < OutputBufferSize - (int)sizeof(CIRC_BUFF_HEADER)){
      if (!FlushSendQueue()){
	if (TrySend(dest, type, nbytes, buff) == MESSAGE_FAILED){
	  AddToQueue = 1;
	}
      }
      else{
	AddToQueue = 1;
      }
      if (AddToQueue){
	TempSendQueue.AddToTail((char *)buff, nbytes, type, dest);
      }
    }
    else{
      EnQueuePartialMsg((char *)buff, dest, nbytes, type, OutputBufferSize-8);
    }
  }

  /*************************************************************************
   * Method: Send
   * Summary: Broadcasts a message to everyone on this box
   * Description: A broadcast message is sent to everyone on the box except
   *              for the node originating the broadcast message.  Again,
   *              if we are unable to flush the send queue, we add to the
   *              queue.  If we can flush the queue, we try to send it.
   *              If we cannot send it, we add it to the queue with the
   *              internal type of SPCOMM_MESSAGE_BROADCAST and we store the
   *              real type in the unused dest parameter.
   * Return: void
   *************************************************************************/
  void Send(const int type, //In: The type of the message
	    const int nbytes, //In: The number of bytes of the message
	    const void * buff //In: The message in a void * array.
	    )
  {
    int AddToQueue = 0; // Flag to indicate if we need to add this message
                        // to the TempSendQueue.
    if (nbytes < OutputBufferSize - NlocalNodes*(int)sizeof(CIRC_BUFF_HEADER)){
      if (!FlushSendQueue()){
	if (TrySend(type, nbytes, buff) != SHMEM_MULTISEND_NONE);
	else{
	  AddToQueue = 1;
	}
      }
      else{
	AddToQueue = 1;
      }
      if (AddToQueue){
	TempSendQueue.AddToTail((char*)buff, nbytes,
				SPCOMM_MESSAGE_BROADCAST, type);
      }
    }
    else{
      EnQueuePartialMsg((char *)buff, nbytes, type, OutputBufferSize-8);
    }
  }
  
  
  
  /*************************************************************************
   * Method: Send
   * Summary: Sends a message to a selection of destionations
   * Description: This multicasts a message to the destinations represented
   *              in the class SpDestination.  Similar to the others, we try
   *              to flush the send queue and queue message on failure. Then
   *              we try to send the message and queue it up on failure
   * Return: void
   *************************************************************************/
  void Send(SpDestination * DEST,
	    const int type,
	    const int nbytes,
	    const void * buff
	    )
  {
    int AddToQueue = 0; // Flag to indicate if we need to add this message
                        // to the TempSendQueue.
    /* If we cannot even write the buffer, we need to queue up the
       message along with the destinations.  We create a new message
       with the first portion containing the destinations and the
       second part contains the actual message */

    cout<<"May not have taken care of 8 byte allignment for this call and the"
	<<endl<<"integer array that is put at the head of the queued message"
	<<endl;


    if (!FlushSendQueue()){
      if (TrySend(DEST, type, nbytes, buff) != SHMEM_MULTISEND_NONE);
      else{
	AddToQueue = 1;
      }
    }
    else{
      AddToQueue = 1;
    }

    if (AddToQueue){
      int DestCount = 0;
      int Node = DEST->GetFirstNode();
      while (Node != -1){++DestCount; Node = DEST->GetNextNode(Node);}
      int *DestArray = new int[DestCount+1];
      int D = DEST->GetFirstNode();
      int *Dest = DestArray;
      *(Dest++) = DestCount;
      while (D != -1){*(Dest++) = D; D = DEST->GetNextNode(D);}
      
      TempSendQueue.AddToTail(DestArray, (char *) buff, nbytes,
			      SPCOMM_MESSAGE_MULTICAST, type);
    }
    else{
      EnQueuePartialMsg((char *)buff, DEST, nbytes, type, OutputBufferSize-8);
    }
  }
  
  /*************************************************************************
   * Method: Send
   * Summary: Sends to a multitude of destionations represented by int*Dest
   * Description: Similar to the other multicast, this sends messages to
   *              all the destinations represented in DEST. 
   * Return: void
   *************************************************************************/
  void Send(int * DEST,  //In: Array of destinations.  *DEST is the number
	                 //    of destinations and then DEST[1] through
	                 //    DEST[*DEST] contain the addresses
	    const int type, //In: The type of the message
	    const int nbytes, //In: The number of bytes of the message
	    const void * buff //In: a void * buffer containing the message
	    )
  {
    /********************************************************************
     * If we cannot even write the buffer, we need to queue up the
     * message along with the destinations.  We create a new message
     * with the first portion containing the destinations and the
     * second part contains the actual message.  The type is stored in
     * the unused dest parameter. 
     ********************************************************************/

    int AddToQueue = 0;// Flag to indicate if we need to add this message
                        // to the TempSendQueue.
    if (nbytes < OutputBufferSize - NlocalNodes*(int)sizeof(CIRC_BUFF_HEADER)){
      if (!FlushSendQueue()){
	if (TrySend(DEST, type, nbytes, buff) != SHMEM_MULTISEND_NONE);
	else{
	  AddToQueue = 1;
	}
      }
      else{
	AddToQueue = 1;
      }
      if (AddToQueue){
	TempSendQueue.AddToTail(DEST, (char *) buff, nbytes,
				SPCOMM_MESSAGE_MULTICAST, type);
      }
    }
    else{
      SpDestination Dest(NlocalNodes);
      int * pDEST = &DEST[1];
      for (int i=0; i < DEST[0]; ++i){
	if (*pDEST == MyLocalNode){ //If also multicast to self
	  InQueue[type].AddToTail((char *)buff, nbytes, type);
	}
	else{
	  Dest.Set(*pDEST);
	}
	++pDEST;
      }
      int tDest = Dest.GetFirstNode();
      int j = 0;
      while (tDest != -1){
	++j;
	tDest = Dest.GetNextNode(tDest);
      }
      if (j){
	cout<<"Queueing up a multi-send"<<endl;
	EnQueuePartialMsg((char *)buff, &Dest, nbytes,type,OutputBufferSize-8);
      }
    }
  }
  
  
  /*************************************************************************
   * Method: GetPendingMessage
   * Summary: returns a message of certain type if available
   * Description: If we have a message of type type, we return it.
   *              Otherwise, we return NULL;
   * Return: Pointer to the message, allocated by new.  Be sure you free
   *         the memory after you use it!
   *************************************************************************/
  void *GetPendingMessage(const int type, //In: Type of the message requested
			  int &nbytes //Out: Number of bytes of message
			              //     returned
			  )
  {
    void * retval = NULL; // message to be returned
    int dest, // unused parameter to the call to QUEUE::Remove();
      Type; // unused parameter to the call to QUEUE::Remove();
    if (InQueue[type].Count()){
      retval = InQueue[type].Remove(nbytes, Type, dest);
    }
    else{
      nbytes = 0;
    }
    return retval;
  };

  /*************************************************************************
   * Method: Receive
   * Summary: Attempts to return a message of a given type.
   * Description: This is similar to GetPendingMessage except it also tries
   *              to receive all messages presently available.  It then
   *              checks to see if we have a message of that type to return.
   * Return: a message of the given type or NULL if none available.
   *************************************************************************/
  void *Receive(const int type, //In: type of requested message
		int &nbytes //Out: Number of bytes of returned message
		)
  {
    void * retval = NULL; //Messgae to be returned.
    if (InQueue[type].Count()){
      retval = GetPendingMessage(type, nbytes);
    }
    else {
      Receive();
      retval = GetPendingMessage(type, nbytes);
    }
    return retval;
  }
  

  /*************************************************************************
   * Method: ShmemFlushSendQueue
   * Summary: Attempts to flush all messages out of send queue.
   * Description: When we are unable to send or announce messages, we queue
   *              up unsent messages in TempSendQueue and unannounced
   *              messages in AnnounceQueue.  This tries to announce all
   *              messages first and, if successful, then tries to send all
   *              messages that it can.  It returns 1 if it was unable to
   *              flush all the messages.  Otherwise, it returns 0 indicating
   *              that all queued message were sent.
   * Return: 0 if all queued messages/announcements were delivered. 1
   *         otherwise.
   *************************************************************************/
  friend int ShmemFlushSendQueue();

  /*************************************************************************
   * Method: Receive
   * Summary: Attempts to receive all messages from all local nodes
   * Description: When there are a large number of local nodes communicating
   *              through shared memory, receiving messages can be
   *              relatively time consuming.  This method attempts to
   *              receive as many messages as are waiting for us and then
   *              returns when it cannot receive any more messages
   * Return: void
   *************************************************************************/
  void Receive();

  void DeleteShm();
};

#ifndef SPCOMM_SHMEM_C

extern int SpComm_LocalNode;

  /*************************************************************************
   * Function: SpComm_Shmem_Receive
   * Summary: Receive all messages for this node
   * Description: This simply implements the method call Receive()
   *              for the shared memory segment for the local node.  This
   *              is here so that it is implemented as an inline function
   *              to reduce a small amount of overhead.
   * Return: void
   *************************************************************************/
static inline void SpComm_Shmem_Receive()
{ SpComm_ShmemMessage[SpComm_LocalNode]->Receive(); }

  /*************************************************************************
   * Function: SpComm_Shmem_Receive
   * Summary: Receive a message of a specific type for this node
   * Description: This simply implements the method call Receive
   *              for the shared memory segment for the local node.  This
   *              is here so that it is implemented as an inline function
   *              to reduce a small amount of overhead.
   * Return: pointer to message allocated by new.  Be sure to free this
   *         memory once you process the message
   *************************************************************************/
static inline void *SpComm_Shmem_Receive(const int type, //In: Message type
					 int &Nbytes     //Out: Number of bytes
					                 //     of the message
					 )
{ return SpComm_ShmemMessage[SpComm_LocalNode]->Receive(type, Nbytes);}

  /*************************************************************************
   * Function: SpComm_Shmem_Receive
   * Summary: Get a pending message of type type for this node.
   * Description: This simply implements the method call GetPendingMessage
   *              for the shared memory segment for the local node.  This
   *              is here so that it is implemented as an inline function
   *              to reduce a small amount of overhead.
   * Return: pointer to message allocated by new.  Be sure to free this
   *         memory once you process the message
   *************************************************************************/
static inline void *SpComm_Shmem_GetPendingMessage(const int type, //In: Msg
						                   //    type
						   int &Nbytes//Out: # of bytes
					                      //     of the
						              //     message
						   )
{return SpComm_ShmemMessage[SpComm_LocalNode]->GetPendingMessage(type,Nbytes);}

  /*************************************************************************
   * Function: SpComm_Shmem_Receive
   * Summary: Receive all messages for this node
   * Description: This simply implements the method call send
   *              for the shared memory segment for the local node.  This
   *              is here so that it is implemented as an inline function
   *              to reduce a small amount of overhead.
   * Return: void
   *************************************************************************/
static inline void SpComm_Shmem_Send(const int dest,   //In: Message
				                       //    destination
				     const int type,   //In: Message type
				     const int nbytes, //In: Message size
				     const void *buff  //In: Message
				     )
{ SpComm_ShmemMessage[SpComm_LocalNode]->Send(dest, type, nbytes, buff);}

  /*************************************************************************
   * Function: SpComm_Shmem_Receive
   * Summary: Receive all messages for this node
   * Description: This simply implements the method call send
   *              for the shared memory segment for the local node.  This
   *              is here so that it is implemented as an inline function
   *              to reduce a small amount of overhead.
   * Return: void
   *************************************************************************/
static inline void SpComm_Shmem_Send(SpDestination *dest, //In: Destinations of
				                        //    message
				     const int type,    //In: Message type
				     const int nbytes,  //In: Message size
				     const void *buff   //In: Message
				     )
{ SpComm_ShmemMessage[SpComm_LocalNode]->Send(dest, type, nbytes, buff);}

  /*************************************************************************
   * Function: SpComm_Shmem_Receive
   * Summary: Receive all messages for this node
   * Description: This simply implements the method call send
   *              for the shared memory segment for the local node.  This
   *              is here so that it is implemented as an inline function
   *              to reduce a small amount of overhead.
   * Return: void
   *************************************************************************/
static inline void SpComm_Shmem_Send(int *dest,        //In: array of
				                       //    destinations
				     const int type,   //In: Message type
				     const int nbytes, //In Message size
				     const void *buff  //In: Message
				     )
{ SpComm_ShmemMessage[SpComm_LocalNode]->Send(dest, type, nbytes, buff);}

  /*************************************************************************
   * Function: SpComm_Shmem_Receive
   * Summary: Receive all messages for this node
   * Description: This simply implements the method call send
   *              for the shared memory segment for the local node.  This
   *              is here so that it is implemented as an inline function
   *              to reduce a small amount of overhead.
   * Return: void
   *************************************************************************/
static inline void SpComm_Shmem_Send(const int type,   //In: Message type
				     const int nbytes, //In: Message size
				     const void *buff  //In: Message
				     )
{ SpComm_ShmemMessage[SpComm_LocalNode]->Send(type, nbytes, buff);}

#endif //SPCOMM_SHMEM_C

void PerrorLocation(char *message, char *file, int line);
#define PERROR_LOCATION(s) PerrorLocation(s, __FILE__, __LINE__)

#endif
