#ifndef Queue_h
#define Queue_h

#include "SpString.H"
#include "SpIostream.H"



/*************************************************************************
 * Class: BUFFER_NODE
 * Summary: The nodes for the BUFFER_QUEUE data type
 * Description: This is a special node for the BUFFER_QUEUE data type.
 *              BUFFER_QUEUE's are a special doubly linked queue just for
 *              the shared memory library.
 * Examples: See BUFFER_QUEUE
*************************************************************************/
class BUFFER_NODE{
private:
  char * Val;
  int Len;
  int Type;
  int Destination;
  int EightByteAllignedGarbage;
protected:

public:
  BUFFER_NODE *Prev; // A pointer to the previous node
  BUFFER_NODE *Next; // A pointer to the next node

  /*************************************************************************
   * Method: BUFFER_NODE
   * Summary: constructor
   * Description: Creates a buffer node with a given Val, Len, Type,
   *              and Destination  If AllocateMemory is set to 0, then
   *              memory for v is not created.  If it is 1, then we
   *              create the memory and do a memcpy.
   * Return: BUFFER_NODE
   *************************************************************************/
  BUFFER_NODE(char * v,                  //In: The message
	      const int len,             //In: The length of the message
	      const int type,            //In: The type of the message
	      const int Dest,            //In: The destination of the message
	      const int AllocateMemory=1 //In: Should we allocate memory for
	                                 //    the message or not?
	      ){
    if (AllocateMemory){
      Val = new char[len];
      memcpy(Val, v, len);
    }
    else{
      Val = v;
    }
    Type = type;
    Len = len;
    Destination = Dest;
    Next = Prev = (BUFFER_NODE *)NULL;
  }

  /*************************************************************************
   * Method: Allign8Bytes
   * Summary: returns a value greater than the parameter but divisible by 8
   * Description: This makes sure that a number is divisible by 8.  It
   *              will return the next number greater than val which is
   *              divisible by 8.  This also means that if a number which
   *              is already divisible by 8 is passed in, then 8 is added
   *              to that number and returned.
   * Return: integer divisible by 8 but greater than the parameter val
   *************************************************************************/
  int Allign8Bytes(int val//In: Number of bytes
		   )
  {return val + (8 - val%8);}

  /*************************************************************************
   * Method: BUFFER_NODE
   * Summary: constructor
   * Description: Creates a buffer node with a given Val, Len, Type,
   *              and Destination  If AllocateMemory is set to 0, then
   *              memory for v is not created.  If it is 1, then we
   *              create the memory and do a memcpy.  The additional
   *              parameter d represents an array of destinations for this
   *              message.  *d is how many destinations there are and then
   *              the rest of the destinations are in d[1] through d[(*d)+1]
   * Return: BUFFER_NODE
   *************************************************************************/
  BUFFER_NODE(const int * d,  //In: Destinations for this message
	      const char * v, //In: The message
	      const int len,  //In: The length of the message
	      const int type, //In: The type of the message
	      const int Dest  //In: The destination of the message (unused)
	      ){
    Val = new char[len+ Allign8Bytes(*d+1)*sizeof(int)];
    memcpy(Val, d, (*d+1)*sizeof(int));
    memcpy(Val+ Allign8Bytes(*d+1)*sizeof(int), v, len);
    Type = type;
    Len = len;
    Destination = Dest;
    Next = Prev = (BUFFER_NODE *)NULL;
  }

  /*************************************************************************
   * Method: ~BUFFER_NODE
   * Summary: BUFFER_NODE destructor
   * Description:  Destroys a buffer node.  Presently a null function
   * Return:
   *************************************************************************/
  ~BUFFER_NODE(){ }

  /*************************************************************************
   * Method: BUFFER_NODE()
   * Summary: Default constructor
   * Description: Default constructor that simply sets several members to
   *              preset values
   * Return: BUFFER_NODE
   *************************************************************************/
  BUFFER_NODE()
  {
    Val=(char *)NULL,
      Len= -1;
    Next = Prev = (BUFFER_NODE *)NULL;
  }

  /*************************************************************************
   * Method: Value
   * Summary: Returns the message stored in Val
   * Description: Returns the message stored in Val
   * Return: Returns the message stored in Val
   *************************************************************************/
  char * Value(){return Val;}

  /*************************************************************************
   * Method: GetType
   * Summary: Returns the type of the message
   * Description: Returns the type of the message
   * Return: Returns the type of the message
   *************************************************************************/
  int GetType(){return Type;}

  /*************************************************************************
   * Method: GetDest
   * Summary: Returns the Destination of the message
   * Description: Returns the Destination of the message
   * Return: Returns the Destination of the message
   *************************************************************************/
  int GetDest(){return Destination;}

  /*************************************************************************
   * Method: Length()
   * Summary: Returns the length of the message
   * Description: Returns the length of the message
   * Return: Returns the length of the message
   *************************************************************************/
  int Length(){return Len;}
};


/*************************************************************************
 * Class: BUFFER_QUEUE
 * Summary: This is a special type of doubly linked queue for the shared
 *          memory message passing.
 * Description: The head serves as both the head and tail of this circular
 *              buffer.  The head follows the tail so head->Next is the last
 *              item in the buffer.  head also leads the front of the queue so
 *              head->Prev is the first item in the queue.  count keeps track
 *              of how many items are in the queue.  Message can be inserted
 *              at the head or tail of the queue and can be either removed
 *              or examined at the front of the queue.
 *
 * Examples: See the shared memory library.  This is used to queue up
 *           messages that could either not be sent or not be announced.
 *************************************************************************/

class BUFFER_QUEUE : public BUFFER_NODE{

private:
  BUFFER_NODE *head;
  int count;

protected:
public:

  /*************************************************************************
   * Method: BUFFER_QUEUE()
   * Summary: Default constructor
   * Description: The constructor allocates a new node for the head of the
   *              queue and then sets the head->Prev and head->next to itself
   *              It also initalizes the number of elements to zero.
   * Return:
   *************************************************************************/
  BUFFER_QUEUE(){
    head = new BUFFER_NODE();
    head->Prev = head->Next = head;
    count = 0;
  }

  ~BUFFER_QUEUE(){
    delete head;
  }
    
  /*************************************************************************
   * Method: Print()
   * Summary: Print the elements of the queue (assumes each val is null
   *          Terminated string.
   * Description: Useful mostly only for debugging
   * Return: void
   *************************************************************************/
  void Print(const char * Str //In: string to be printed before each "message"
	     ){
    BUFFER_NODE *temp = head; // Temporary pointer to use to walk along the
                              // queue.
    cout<<Str<< " and Count = "<<count<<endl;
    for (int i=0; i < count; ++i){
      temp = temp->Prev;
      cout<<temp->Value()<<endl;}
    cout<<endl;
  }


  /*************************************************************************
   * Method: AddToTail
   * Summary: Adds a message to the tail of the queue.
   * Description: Adds a messge to the tail of the queue.
   * Return: void
   *************************************************************************/
  void AddToTail(char * B,           //In: The message
		 const int len,      //In: The messages length
		 const int NodeType, //In: The messages type
		 const int Dest = -1,//In: The messages destination
		 const int allocate=1
		 )
    {
      BUFFER_NODE *node = //Node to be added
	new BUFFER_NODE(B, len, NodeType, Dest, allocate);
      if (count){
	node->Next = head->Next;
	node->Prev = head;
	(head->Next)->Prev = node;
	head->Next = node;
      }
      else{
	node->Next = node->Prev = head;
	head->Next = head->Prev = node;
      }
      ++count;
    }

  /*************************************************************************
   * Method: AddToTail
   * Summary: Adds a message with multiple destinations to the tail of 
   *          the queue
   * Description: This is slightly different from the other AddToTail
   *              method since it adds two buffers.  dest is an array of 
   *              (*dest)+1 integers, the first is the number of integers
   *              and then the rest of them follow, each being a different
   *              destination for the message
   * Return: void
   *************************************************************************/
  void AddToTail(int * dest,         // Array of destinations
		 const char * B,     // Message to be sent
		 const int len,      // Length of the message
		 const int NodeType, // Type of the message
		 const int Dest = -1 // Destination of the message
		 )
    // Adds an item to the end (tail) of the queue
    {
      BUFFER_NODE *node = //Node to be added
	new BUFFER_NODE(dest, B, len, NodeType, Dest);
      if (NodeType == 261){
	cout<<NodeType<<endl;
      }
      if (count){
	node->Next = head->Next;
	node->Prev = head;
	(head->Next)->Prev = node;
	head->Next = node;
      }
      else{
	node->Next = node->Prev = head;
	head->Next = head->Prev = node;
      }
      ++count;
    }

  /*************************************************************************
   * Method: AddToFront
   * Summary: Adds a message to the front (head) of the queue
   * Description: Adds a message to the front (head) of the queue.  This is
   *              often used when a message needs to be sent before the rest
   *              of the messages in the queue.  The reason for this may be
   *              that we were able to write the message in the outgoing
   *              buffer but need to queue up the announcement.
   * Return: void
   *************************************************************************/
  void AddToFront(char * B,
		  const int len,
		  const int T,
		  const int Dest = -1,
		  const int allocate = 1
		  )
    {
      BUFFER_NODE *node = // node to be added.
	new BUFFER_NODE(B, len, T, Dest, allocate);
      if (count){
	(head->Prev)->Next = node;
	node->Prev = head->Prev;
	node->Next = head;
	head->Prev = node;
      }
      else{
	node->Next = node->Prev = head;
	head->Next = head->Prev = node;
      }
      ++count;
    }
  
  
  /*************************************************************************
   * Method: Remove
   * Summary: Removes the next queue item and returns it.
   * Description: This method pulls the next item from the queue and then
   *              returns the message, type, nbytes, and destination.
   * Return: message buffer
   *************************************************************************/
  char * Remove(int &nbytes,   //Out: Number of bytes of message
		int &ItemType, //Out: Type of message
		int &Dest      //Out: Destination of message
		){
    char * Retval = NULL;         //Char * to hold returned message
    BUFFER_NODE * Tnode = NULL;   //Temp node to place hold while removing.

    if (count > 0){
      --count;
      Tnode = head->Prev;
      Retval = Tnode->Value();
      if (count){
	Tnode->Prev->Next = head;
	head->Prev = Tnode->Prev;
      }
      else {
	head->Next = head->Prev = (BUFFER_NODE *)NULL;
      }
      nbytes = Tnode->Length();
      ItemType = Tnode->GetType();
      Dest = Tnode->GetDest();
      delete Tnode;
    }
    else{ ItemType = Dest = -1;}
    return Retval;
  }

  /*************************************************************************
   * Method: NextValue
   * Summary: returns the value of the next queue item without removing it
   * Description:  Sometimes you want to examine the next item in the queue
   *               before you decide to remove it or not.  This allows one
   *               to do that.
   * Return: char * buffer pointing to the next message in the queueu
   *************************************************************************/
  char * NextValue(int &nbytes,   // Number of bytes of the message
		   int &NodeType, // Type of the message
		   int &Dest      // Destination of the message
		   ){
    char * Retval = NULL;         //Char * to hold returned message
    BUFFER_NODE * Tnode = NULL;   //Temp node to place hold while removing.
    
    if (count > 0){
      Tnode = head->Prev;
      Retval = Tnode->Value();
      nbytes = Tnode->Length();
      NodeType = Tnode->GetType();
      Dest = Tnode->GetDest();
    }
    else{ NodeType = Dest = -1;}
    return Retval;
  }
  
  /*************************************************************************
   * Method: Empty
   * Summary: Returns non-zero if queue not empty, 0 if empty
   * Description: Returns non-zero if queue not empty, 0 if empty
   * Return:Returns non-zero if queue not empty, 0 if empty
   *************************************************************************/
  int Empty(){return !count;}

  /*************************************************************************
   * Method: Count
   * Summary: Returns number of items in the queue
   * Description: Returns number of items in the queue
   * Return: Returns number of items in the queue
   *************************************************************************/
  int Count(){return count;}

};






#endif
