//#include <values.h>
#include "SpMath.H"
#include "SpIostream.H"
#include "SpFstream.H"
#include "SpStrstream.H"
#include "SpIomanip.H"

#include "SpNetTypes.H"
#include "SpServiceGroup.H"
#include "SpCommServerMsgs.H"
#include "SpCommServer.H"
#include "SpMsgServer.H"
#include "SpDataParser.H"
#include "SpCommServerHeader.H"

/************ EXAMPLE SERVICE GROUP CLASS **************** 

  This is an example of how to define a service group class, an
  instance of which will be created by the MESSAGE_SERVER and will
  respond to a set of arriving messages.  The essential idea is this:
  you must write a method capable of handling each possible message
  type.  When the server receives a message intended for your object,
  it will read an int off the message header and this will tell it
  which one of the methods to invoke; so each handler-method must be
  identified by an associated int (see [1] below).

  You may have as many other methods or data members as you
  want--that's the idea of encapsulating this all in an object.  After
  you handler methods have been defined, you must write a constructor
  that will "register" these handlers so that the MESSAGE_SERVER can
  call them (by putting pointers to the methods into an array).  This
  requires two steps--one to allocate storage, and one to store a
  pointer for each method (see [2] and [3] below).

  Finally, you must write a global function to create an instance of
  your class (see [4]).  This is so the MESSAGE_SERVER can
  instantinate your obj.  It may have other SERVICE_GROUP class defs
  like yours.  When it receives a message addressed to a
  SERVICE_GROUP, it does the following: if the group exists, it
  invokes the appropriate method to handle the message; if the group
  does NOT exist, it creates an instance of the right kind of group
  (determined by the serviceType field in the msg header), and then
  calls the handler as above.

**********************/


//#define NOISY

//CONNECTION** SpComm_Sock= NULL;

#define LocalNodeZero(GlobalNodeID) (LocalNodeZeros[GlobalNodeID])
#define BoxNumber(GlobalNodeId) (BoxNumbers[GlobalNodeId])
void CommServer::Message(CONNECTION *, COMM_SERVER_HEADER *mh) {
  /* Simply bounce the message to the correct destination */

  int node = mh->getNode();
  int bytes = mh->getBytes() + sizeof(COMM_SERVER_HEADER);
  if (GatherStatistics && (mh->getType()<NumMsgTypes) && (bytes < MaxMsgSize)){
    if (MessageStatistics[mh->getType()] == NULL){
      MessageStatistics[mh->getType()] = new double[MaxMsgSize];
      for (int i=0; i < MaxMsgSize; ++i){
	MessageStatistics[mh->getType()][i]=0;
      }
    }
    ++MessageStatistics[mh->getType()][bytes];
  }

#ifdef NOISY
  static int jj = 0;
  if (mh->getType() == 259)
    cout<<"Sending message "<<++jj<<" of type "<<mh->getType()
	<<" and length "<<mh->getBytes()<<" to "<<node<<endl;
#endif
  SendMessage(node, (char *)mh, bytes);
  delete [] (char *) mh;
}

//static inline int Allign8Bytes(int val) {return val + (8 - val%8);}

void CommServer::Multicast(CONNECTION *, COMM_SERVER_HEADER *mh) {

  int TotBytes = mh->getBytes() + sizeof(COMM_SERVER_HEADER);
  NET_INT * Dest = ((NET_INT *)&mh[1]), *TmpPtr;
  int NumDest = *Dest;
  int i;
  int TrueBytes = mh->getBytes() - Allign8Bytes(((NumDest)+1)*sizeof(NET_INT));
  char * TrueMsg = ((char *)Dest) +Allign8Bytes(((NumDest)+1)*sizeof(NET_INT));

  ++Dest;
  for (i = 0; i < NumDest; ++i){ /* Create a message for each box that
				    is on the destination list */    
    int PresBox = BoxNumber(*Dest); //Which box contains the destination
    if (MultiSendMessages[PresBox] == NULL){
      MultiSendMessages[PresBox] = new char[TotBytes];
      TmpPtr = (NET_INT*)(MultiSendMessages[PresBox] +
			  sizeof(COMM_SERVER_HEADER));
      *TmpPtr = 0; // Initially, we have no destinations for this box
    }
    TmpPtr = (NET_INT*)(MultiSendMessages[PresBox] +
			sizeof(COMM_SERVER_HEADER));
    ++*TmpPtr; // Increment number of destination for this box
    TmpPtr[*TmpPtr] = *Dest; // Add present destination to this box's list
    ++Dest; // Move on to the next destination.
  }

  /* Now, all of the destinations are on the same box so we send it to
     the node zero of that box */
  for (i = 0; i < NumBoxes; ++i){
    if (MultiSendMessages[i] != NULL){
      COMM_SERVER_HEADER* hdr = (COMM_SERVER_HEADER*)MultiSendMessages[i];

      /* Prepare the message for this node.  ie fill in the actual
	 data  and fill in the header.*/
      memcpy(hdr, mh, sizeof(COMM_SERVER_HEADER));
      NumDest = ((NET_INT *)&hdr[1]) [0];
      char* ptr = ((char *)&(hdr[1])) +
	Allign8Bytes(sizeof(NET_INT)*(NumDest+1));
      memcpy(ptr, TrueMsg, TrueBytes);

      int node = LocalNodeZero(  ((NET_INT *)&hdr[1]) [1] );
#ifdef NOISY
      static int jj = 0;
      if (mh->getType() == 259)
	cout<<"Sending Multi-cast message "<<++jj<<" of type "<<mh->getType()
	    <<" and length "<<mh->getBytes()<<" to "<<node<<endl;
#endif
      int bytes = sizeof(COMM_SERVER_HEADER) + TrueBytes +
	Allign8Bytes(sizeof(NET_INT)*(NumDest+1));
      mh->setBytes(bytes - sizeof(COMM_SERVER_HEADER));
      SendMessage(node, (char *)mh, bytes);
      
      delete [] (char *) hdr;
      MultiSendMessages[i] = NULL;
    }
  }
  delete [] (char *) mh;
}


void CommServer::Broadcast(CONNECTION *, COMM_SERVER_HEADER *mh) {
  /* send a message to the local node zero's of each box other than
     the sender and then they will broadcast to their own box */

  int LocalNodeZeroOfFromNode = LocalNodeZero(mh->getNode());
  int bytes = mh->getBytes() + sizeof(COMM_SERVER_HEADER);

  for (int i=0; i < NumBoxes; ++i){
    if (LocalNodeZeroOfFromNode != LNZ[i]){//Don't broadcast to sender's box
#ifdef NOISY
  static int jj = 0;
  if (mh->getType() == 259)
  cout<<"Sending broadcast message "<<++jj<<" of type "<<mh->getType()
      <<" and length "<<mh->getBytes()<<" to "<<i<<endl;
#endif
      SendMessage(LNZ[i], (char *)mh, bytes);
    }
  }
  delete [] (char *)mh;
}


/* This single call Reduction() handles all of the reductions.  There
   are a number of switch statements used to differentiate between the
   differnt reductions but they are all handled similarly.  */
#define DEBUG_FUZZY

void CommServer::Reduction(CONNECTION *, COMM_SERVER_HEADER *mh) {
  REDOP_INFO *Rinf;
  int itemp;
  double dtemp;
  SpSimTime stemp;
  int hnum = mh->getHandlerNum();

  switch (hnum){
  case GLOBAL_SUM:
  case GLOBAL_MIN:
  case GLOBAL_MAX:
    Rinf = (REDOP_INFO *) &(mh[1]);
    if (NumCheckedIn == 0){
      Ival = Rinf->GetInt();
      Dval = Rinf->GetDouble();
      Sval = Rinf->GetSimTime();
    }
    else {
      itemp = Rinf->GetInt();
      dtemp = Rinf->GetDouble();
      stemp = Rinf->GetSimTime();
    }
    break;
  case NON_BLOCKING_SYNCHRONIZATION:
  case BLOCKING_SYNCHRONIZATION:
  default:
    break;
  }
  if (NumCheckedIn){ //If we are not the first one to check in
    switch(hnum){
    case GLOBAL_MIN:
      if (Ival > itemp){ Ival = itemp;}
      if (Dval > dtemp){ Dval = dtemp;}
      if (Sval > stemp){ Sval = stemp;}
      break;
    case GLOBAL_MAX:
      if (Ival < itemp){ Ival = itemp;}
      if (Dval < dtemp){ Dval = dtemp;}
      if (Sval < stemp){ Sval = stemp;}
      break;
    case GLOBAL_SUM: //Do not sum the SpSimTime.
      Ival += itemp;
      Dval += dtemp;
      break;
    case NON_BLOCKING_SYNCHRONIZATION:
    case BLOCKING_SYNCHRONIZATION:
    default:
      break;
    }
  }
  ++NumCheckedIn;

  if (NumCheckedIn == NumBoxes){ 
    /* report back that the reduction is done Notice that we send back
       mh.  This is because it already has the proper setup of the
       header along with the proper size.  We just initialize the
       redopinfo portion of the message if this is a sum, min or max,
       and send the message back */

    int nbytes;
    
    NumCheckedIn = 0;
    switch (hnum){
    case GLOBAL_SUM:     // These messages return redop info along with the
      ++GlobalSumCount;  // header
    case GLOBAL_MIN:
      if (hnum == GLOBAL_MIN) ++GlobalMinCount;
    case GLOBAL_MAX:
      if (hnum == GLOBAL_MAX) ++GlobalMaxCount;
      Rinf->Init(Ival, Dval, Sval);
      nbytes = sizeof(COMM_SERVER_HEADER) + sizeof(REDOP_INFO);
      break;
    case NON_BLOCKING_SYNCHRONIZATION:
      ++FuzzyCount;
    case BLOCKING_SYNCHRONIZATION:
      if (hnum == BLOCKING_SYNCHRONIZATION) ++BlockingCount;
    default:
      nbytes = sizeof(COMM_SERVER_HEADER);
      break;
    }
    for (int i = 0; i < NumBoxes; ++i){
      SendMessage(LNZ[i], (char *)mh, nbytes);
    }
  }
  delete [] (char *)mh;
}





void CommServer::Startup(CONNECTION *conn, COMM_SERVER_HEADER *mh) {

  /* Node zero of each box will check in and request a certain number
     of local nodes.  Set them up and then, when all nodes have
     checked in, report back the tables */

  int LocalNodeID = mh->getNode();
  int BoxNumber = mh->getType(); //Store the box number in type.
  int i;
  char * BoxListMessage;
  char * Msg = (char *)&mh[1];

  if (!TablesCreated){
    NumInGroup = *(NET_INT *)Msg;
    GatherStatistics = ((NET_INT *)Msg)[1];
    int NumLocal = mh->getNode();
    cout<<"Box Number "<<NumBoxes
	<<" reporting in with nlocal = "<<NumLocal
	<<" and num in group "<<NumInGroup<<endl;
  
    if (NumCheckedIn == 0){
      /* This is the first box to check in so allocate the memory for the
	 structures */
      SpComm_Sock =    new CONNECTION*[NumInGroup];
      LocalNodeIDs =   new NET_INT[NumInGroup];
      BoxNumbers =     new NET_INT[NumInGroup];
      LocalNodeZeros = new NET_INT[NumInGroup];
      LNZ =            new NET_INT[NumInGroup]; //Really should be # of boxes
      for (i=0; i <NumInGroup; ++i){
	SpComm_Sock[i] = NULL;
      }
    }
  
    ++NumBoxes;
    SpComm_Sock[NumCheckedIn] = conn;
    if (NumCheckedIn > NumInGroup){
      cout<<"Tried to connect in more than in the group."<<endl
	  <<"NumCheckedIn = "<<NumCheckedIn<<" and num in group = "
	  <<NumInGroup<<endl;
      return;
    }
    LNZ[NextBoxNumber] = NumCheckedIn; /* NumCheckedIn ==node # of
					  local node zero */
    if (NumCheckedIn > NumInGroup){
      cout<<"Tried to connect in more than in the group."<<endl
	  <<"NextBoxNumber = "<<NextBoxNumber<<" and num in group = "
	  <<NumInGroup<<endl;
      return;
    }
    for (i=0; i < NumLocal; ++i){
      if (NumCheckedIn > NumInGroup){
	cout<<"Tried to connect in more than in the group."<<endl
	    <<"i+NumCheckedIn = "<<i+NumCheckedIn<<" and num in group = "
	    <<NumInGroup<<endl;
	return;
      }
      LocalNodeIDs[i+NumCheckedIn] = i;
      BoxNumbers[i+NumCheckedIn] = NextBoxNumber;
      LocalNodeZeros[i+NumCheckedIn] = NumCheckedIn; //NumCheckedIn==localnode0
    }

    ++NextBoxNumber;
    NumCheckedIn += NumLocal;

    if (NumCheckedIn == NumInGroup){ //Every node 0 has checked in
      MultiSendMessages = new char*[NumBoxes];
      for (int zz=0; zz < NumBoxes; ++zz){
	MultiSendMessages[zz] = NULL;
      }
      //Then send a message back to everyone, letting them know what the
      //tables are.

      TablesCreated = 1;
      int Nbytes = sizeof(COMM_SERVER_HEADER) +
	Allign8Bytes(sizeof(NET_INT)*NumInGroup)*3 +
	Allign8Bytes(sizeof(NET_INT)) +
	Allign8Bytes(sizeof(NET_INT));
      BoxListMessage = new char [Nbytes];
      char * tmp = BoxListMessage + sizeof(COMM_SERVER_HEADER);

      memcpy(tmp, LocalNodeIDs, sizeof(NET_INT)*NumInGroup);
      tmp += Allign8Bytes(sizeof(NET_INT)*NumInGroup);

      memcpy(tmp, LocalNodeZeros, sizeof(NET_INT)*NumInGroup);
      tmp += Allign8Bytes(sizeof(NET_INT)*NumInGroup);

      memcpy(tmp, BoxNumbers, sizeof(NET_INT)*NumInGroup);
      tmp += Allign8Bytes(sizeof(NET_INT)*NumInGroup);

      NET_INT NumBoxesNetInt = NumBoxes;
      memcpy(tmp, &NumBoxesNetInt, sizeof(NET_INT));
      tmp += Allign8Bytes(sizeof(NET_INT));

      for (i = 0; i < NumInGroup; ++i){
	if (LocalNodeIDs[i] == 0){ //Only send to the local node zero's
	  memcpy(tmp, &BoxNumbers[i],sizeof(NET_INT));
	  SendMessage(i, BoxListMessage, Nbytes);
	}
      }
      delete [] BoxListMessage;
      NumCheckedIn = 0;
    }
  }
  else{
    if (mh->getNode() != 0){ //Node 0's already have their conn assigned.
      SpComm_Sock[mh->getNode()] = conn;
    }
  }
  delete [] (char *)mh;
}

CommServer::~CommServer(){
  cout<<"Deleting a comm server"<<endl;
  NumCheckedIn = 0;
  NextBoxNumber = 0;
  TablesCreated = 0;
  NumBoxes = 0;
  if (SpComm_Sock) delete [] SpComm_Sock;
  if (LocalNodeIDs) delete [] LocalNodeIDs;
  if (BoxNumbers) delete [] BoxNumbers;
  if (LocalNodeZeros) delete [] LocalNodeZeros;
  if (LNZ) delete [] LNZ;
  if (MultiSendMessages){
    for (int jj=0; jj < NumBoxes; ++jj){
      if (MultiSendMessages[jj]) delete [] MultiSendMessages[jj];
    }
    delete [] MultiSendMessages;
  }
  if (GatherStatistics){
    double sum;
    double avg;
    double count;
    int i, j;

    for (i=0; i < NumMsgTypes; ++i){
      count = sum = 0;
      if (MessageStatistics[i] != NULL){
	for (j=0; j < MaxMsgSize; ++j){
	  count += MessageStatistics[i][j];
	  sum += MessageStatistics[i][j]*j;
	}
	if (count){
	  avg = sum/count;
	  sum = 0;
	  for (j=0; j < MaxMsgSize; ++j){
	    sum += MessageStatistics[i][j]*(j - avg)*(j - avg);
	    MessageStatistics[i][j] = 0;
	  }
	  delete [] MessageStatistics[i];
	  MessageStatistics[i] = NULL;
	  double stddev = sqrt(sum/count);
	  cout<<"Msg Type "<<i<<" Count = "<<count<<" avg size = "<<avg
	      <<" Standard dev = "<<stddev<<endl;
	}
      }
    }
    if (GlobalSumCount){
      cout<<"There were "<<GlobalSumCount<<" Global sums"<<endl;
      GlobalSumCount = 0;
    }
    if (GlobalMinCount){
      cout<<"There were "<<GlobalMinCount<<" Global mins"<<endl;
      GlobalMinCount = 0;
    }
    if (GlobalMaxCount){
      cout<<"There were "<<GlobalMaxCount<<" Global maxs"<<endl;
      GlobalMaxCount = 0;
    }
    if (FuzzyCount){
      cout<<"There were "<<FuzzyCount<<" non-blocking reductions"<<endl;
      FuzzyCount = 0;
    }
    if (BlockingCount){
      cout<<"There were "<<BlockingCount<<" blocking reductions"<<endl;
      BlockingCount = 0;
    } 
  }
  delete [] MessageStatistics;
}


void CommServer::SendMessage(int whichConn, char * msg, int nBytes)
{
  if (SpComm_Sock[whichConn]){
    SpComm_Sock[whichConn]->scheduleSend(msg, nBytes);
  }
}

void CommServer::CloseConnection(CONNECTION * connectionThatIsClosing){

  int i;
  for(i=0; i < NumInGroup; ++i){
    if (SpComm_Sock[i] != NULL){
      if (SpComm_Sock[i]->getfd() == connectionThatIsClosing->getfd()){
	SpComm_Sock[i] = NULL;
	cout<<"Closing the connection to node "<<i<<endl;
	i = -1;
	break;
      }
    }
  }
  if (i != -1){
    cout<<"Got a CommServer::CloseConnection but could not find who"
	<<" was closing"<<endl;
  }
}

CommServer::CommServer(void) :
    NumCheckedIn(0),
    NextBoxNumber(0),
    TablesCreated(0),
    NumInGroup(0),
    LNZ(NULL),
    NumBoxes(0),
    SendSocket(NULL),
    LocalNodeIDs(NULL),
    BoxNumbers(NULL),
    LocalNodeZeros(NULL),
    GatherStatistics(0),
    MultiSendMessages(NULL),
    SpComm_Sock(NULL),
    Ival(0),
    Dval(0),
    Sval(0),
    GlobalSumCount(0),
    GlobalMinCount(0),
    GlobalMaxCount(0),
    BlockingCount(0),
    FuzzyCount(0)
{
  MessageStatistics = new double*[NumMsgTypes];
  for (int i=0; i < NumMsgTypes; ++i){
    MessageStatistics[i] = NULL;
  }
    allocateHandlerStorage(NUMHANDLERS);/* [2] Allocate storage for
					   your msg handlers */
    
      // [3] Use this macro to store pointer, and associate an int
      // with the pointer (just puts pointer in Nth slot in an array):
    STORE_MSG_HANDLER(DUMMY_CONN_MESSAGE, CommServer, DummyConnectionMessage);
    STORE_MSG_HANDLER(NORMAL_MESSAGE, CommServer, Message);
    STORE_MSG_HANDLER(MESSAGE_BROADCAST, CommServer, Broadcast);
    STORE_MSG_HANDLER(MESSAGE_MULTICAST, CommServer, Multicast);
    STORE_MSG_HANDLER(GLOBAL_SUM, CommServer, Reduction);
    STORE_MSG_HANDLER(GLOBAL_MIN, CommServer, Reduction);
    STORE_MSG_HANDLER(GLOBAL_MAX, CommServer, Reduction);
    STORE_MSG_HANDLER(NON_BLOCKING_SYNCHRONIZATION, CommServer, Reduction);
    STORE_MSG_HANDLER(BLOCKING_SYNCHRONIZATION, CommServer, Reduction);
    STORE_MSG_HANDLER(TCPIP_STARTUP, CommServer, Startup);
  }
