// SpSubscriptionItem.H header file defines a subscription by name and then by
// object specifiers or wild cards.

#ifndef SUBSCRIPTION_ITEM_OBJ
#define SUBSCRIPTION_ITEM_OBJ

#include "SpStdlib.H"
#include "SpIostream.H"
#include "SpFstream.H"
#include "SpStrstream.H"
#include "SpIomanip.H"
#include "SpString.H"
#include "qitem.H"

class SUBSCRIPTION_ITEM : public C_ITEM {

  private:

    char *Name;
    char *ObjectName;
    int ObjectType;
    int LocalId;
    int Node;
    int ExternalId;

  protected:

  public:

/*
** No-argument constructor. This should be used with the UnPackBuffer method
** for initializing the object.
*/

    SUBSCRIPTION_ITEM():Name(NULL),
      ObjectName(NULL),
      ObjectType(-1),
      LocalId(-1),
      Node(-1),
      ExternalId(-1){}

/*
** Constructor that passes in the name of the subscription and the name of
** the object being subscribed.
*/

    SUBSCRIPTION_ITEM(char *name, char *objectName) {
      Name = strdup(name);
      ObjectName = strdup(objectName);
      ObjectType = -1;
      LocalId = -1;
      Node = -1;
      ExternalId = -1;
    }

/*
** Constructor that passes in the name of the subscription and the integer
** values for object type, local id, and node. Note that if none of those
** parameters is passed in, we are subscribing to everyone.
*/

    SUBSCRIPTION_ITEM(char *name,
		      int objectType = -1,
		      int localId = -1,
		      int node = -1) {
      Name = strdup(name);
      ObjectName = strdup("None");
      ObjectType = objectType;
      LocalId = localId;
      Node = node;
      ExternalId = -1;
    }

/*
** Destructor for the subscription item
*/

    ~SUBSCRIPTION_ITEM() {
      if (Name) free(Name);
      if (ObjectName) free(ObjectName);
      Name = NULL;
      ObjectName = NULL;
    }

    virtual char *GetName() {return Name;}

    char *GetObjectName() {
      char *objectName = ObjectName;
      if (!strcmp(objectName, "None")) objectName = NULL;
      return objectName;
    }

    int GetObjectType() {return ObjectType;}
    int GetLocalId() { return LocalId;}
    int GetNode() {return Node;}

    int GetExternalId() {return ExternalId;}
    void SetExternalId(int externalId) {ExternalId = externalId;}

/*
** Sets the integer object id information
*/

    void SetObjectInfo(int objectType, int localId, int node) {
      ObjectType = objectType;
      LocalId = localId;
      Node = node;
    }

/*
** Compares if the input object has the same information. Returns 1 if so.
** Otherwise, returns 0. Note that this does not compare the ObjectNames
** since it is assumed that the ObjectType, LocalId, and Node info is
** equivalent.
*/

    int Compare(SUBSCRIPTION_ITEM *item) {
      int value = 1;

      for (;;) {

	if (item->GetExternalId() != GetExternalId()) {
	  value = 0;
	  break;
	}

	if (item->GetObjectType() != GetObjectType()) {
	  value = 0;
	  break;
	}

	if (item->GetLocalId() != GetLocalId()) {
	  value = 0;
	  break;
	}

	if (item->GetNode() != GetNode()) {
	  value = 0;
	  break;
	}

	if (strcmp(item->GetName(), GetName())) {
	  value = 0;
	  break;
	}

	break;	// If we get here, then we have a match

      }

      return value;

    }

/*
** Checks if an object is contained within the SubItem
*/ 

    int CheckContained(int objectType, int localId, int node) {
      int RetVal = 0;
      if (GetObjectType() == -1) {
	RetVal = 1;
      }else{
	if (GetObjectType() == objectType) {
	  if (GetLocalId() == -1) {
	    RetVal = 1;
	  }else{
	    if (GetLocalId() == localId) {
	      RetVal = (GetNode() == node);
	    }
	  }
	}
      }
      return RetVal;
    }

/*
** Gets the size of the buffer needed to send the subscription item as a
** message
*/

    int GetBufferSize() {
      int totalSize = 4 * sizeof(NET_INT) //geb3 fix (was sizeof(int))
	+ 2*sizeof(char *)
	+ strlen(Name) + 1
	+ strlen(ObjectName) + 1;
      return totalSize;
    }

/*
** Creates a buffer and packs it
*/

    char *CreateBuffer(int &size) {
      size = GetBufferSize();
      char *Buffer = new char[size];
      PackBuffer(Buffer);
      return Buffer;
    }

/*
** Packs the buffer with the information in this class
*/

  //geb3 fix: USE NET_INTS!
    void PackBuffer(char *Buffer) {
      NET_INT netVal;

      strcpy(Buffer, Name);
      Buffer += strlen(Name) + 1;

      strcpy(Buffer, ObjectName);
      Buffer += strlen(ObjectName) + 1;

      netVal = ObjectType;
      memcpy(Buffer, (char *)&netVal, sizeof(NET_INT));
      Buffer += sizeof(NET_INT);

      netVal = LocalId;
      memcpy(Buffer, (char *)&netVal, sizeof(NET_INT));
      Buffer += sizeof(NET_INT);

      netVal = Node;
      memcpy(Buffer, (char *)&netVal, sizeof(NET_INT));
      Buffer += sizeof(NET_INT);

      netVal = ExternalId;
      memcpy(Buffer, (char *)&netVal, sizeof(NET_INT));
    }

  /*********geb3 fix: old version:
    void PackBuffer(char *Buffer) {
      strcpy(Buffer, Name);
      Buffer += strlen(Name) + 1;

      strcpy(Buffer, ObjectName);
      Buffer += strlen(ObjectName) + 1;

      memcpy(Buffer, (char *)&ObjectType, sizeof(int));
      Buffer += sizeof(int);

      memcpy(Buffer, (char *)&LocalId, sizeof(int));
      Buffer += sizeof(int);

      memcpy(Buffer, (char *)&Node, sizeof(int));
      Buffer += sizeof(int);

      memcpy(Buffer, (char *)&ExternalId, sizeof(int));
      Buffer += sizeof(int);
    }
  ***********/

/*
** Unpacks the buffer
*/

  //geb3 fix:  unpack NET_INTS!
    void UnPackBuffer(char *Buffer) {
      NET_INT netVal;

      if (Name) free(Name);
      if (ObjectName) free(ObjectName);

      Name = strdup(Buffer);
      Buffer += strlen(Name) + 1;

      ObjectName = strdup(Buffer);
      Buffer += strlen(ObjectName) + 1;

      memcpy((char *)&netVal, Buffer, sizeof(NET_INT));
      ObjectType = netVal;
      Buffer += sizeof(NET_INT);

      memcpy((char *)&netVal, Buffer, sizeof(NET_INT));
      LocalId = netVal;
      Buffer += sizeof(NET_INT);

      memcpy((char *)&netVal, Buffer, sizeof(NET_INT));
      Node = netVal;
      Buffer += sizeof(NET_INT);

      memcpy((char *)&netVal, Buffer, sizeof(NET_INT));
      ExternalId = netVal;
    }

  /******* geb3 fix: old version
    void UnPackBuffer(char *Buffer) {

      if (Name) free(Name);
      if (ObjectName) free(ObjectName);

      Name = strdup(Buffer);
      Buffer += strlen(Name) + 1;

      ObjectName = strdup(Buffer);
      Buffer += strlen(ObjectName) + 1;

      memcpy((char *)&ObjectType, Buffer, sizeof(int));
      Buffer += sizeof(int);

      memcpy((char *)&LocalId, Buffer, sizeof(int));
      Buffer += sizeof(int);

      memcpy((char *)&Node, Buffer, sizeof(int));
      Buffer += sizeof(int);

      memcpy((char *)&ExternalId, Buffer, sizeof(int));
      Buffer += sizeof(int);
    }
  ************/

/*
** Prints information
*/

    void Print(ostream &out=cout) {
      char* oName = GetObjectName();
      if (oName == NULL){
	oName = "NONE";
      }
      out << "Subscription Name = " << GetName() << endl;
      out << "Object Name = " << oName << endl;
      out << "Object Type = " << GetObjectType() << endl;
      out << "Local Id = " << GetLocalId() << endl;
      out << "Node = " << GetNode() << endl;
      out << "ExternalId = " << GetExternalId() << endl;
    }

};

#endif
