// TwoWayCommand.C file

#include "SpIostream.H"
#include "SpFstream.H"
#include "SpStrstream.H"
#include "SpIomanip.H"
#include "SpStdlib.H"
#include "SpString.H"
#include "SpSignal.H"

#include "SpEmHostUser.H"

SpEmHostUser *EmHostUser = NULL;

/*
** Handler routine for when (if) the user hits control C
*/

void sigintfunc(int) {
  cout << "Interupt" << endl;
  signal(SIGINT, sigintfunc);
  if (EmHostUser) delete EmHostUser;
  EmHostUser = NULL;
  exit(1);
}

/*
** General error printing routing
*/

void PrintError() {
  cout << 
    "TwoWayCommand <CommandName, ObjectName, DataString, BarrierMode, TimeLag>"
       << endl;
  cout <<
    "BarrierMode = AGGRESSIVE_BARRIER, SYNCHRONIZED_BARRIER, or NO_BARRIER"
       << endl;
}

/*
*****************************************************************************
* TwoWayCommand exchanges two-way commands with SPEEDES. The data is
* repeatedly sent as a string argument. Similarly, the data received is
* also a string value (note, that in general, the data could be anything).
* The basic EmHostUser calls are:
*
*     EmHostUser->SendTwoWayCommand(char *CommandName,
*                                   char *ObjectName,
*				    char *Data,
*				    int Bytes,
*				    int BarrierMode,
*				    double TimeLag);
*     where:
*           CommandName is the name of the event or event handler in SPEEDES
*           ObjectName is the name of the object that receives the event
*           Data is the data that is sent as part of the event
*           Bytes is the number of bytes for the data
*           BarrierMode can be:
*	        AGGRESSIVE_BARRIER: EventTime = gvt, barrier = gvt + TimeLag
*               SYNCHRONIZED_BARRIER: EventTime = barrier, barrier += TimeLag
*               NO_BARRIER: EventTime = gvt, no barrier established
*           TimeLag is the time interval used for establishing barriers
*
*     Note: the first EventTime is set to gvt.
*
* After this call, the application is expected to make calls to receive data
* in a manner similar to the following example:
*
*     while (1) {
*	char *Data = EmHostUser->ReceiveData(&Time, int &Bytes);
*       if (Data) break;
*     }
*
*     where:
*           Data is a returned buffer (can be arbitrarily casted to anything)
*           Time is the EventTime that the data was sent from SPEEDES
*           Bytes is the number of bytes in the data
*
*****************************************************************************
*/

int main(int argc, char **argv) {
  int BarrierMode;
  double TimeLag;
  int i;
  char *Data;
  int Bytes;
  double Time;
  int GlobalId;

  signal(SIGINT, sigintfunc);

/*
** Make sure that the number of arguments are correct and then convert them
** to the right form.
*/

  if (argc == 6) {

    BarrierMode = -1;

    if (!strcmp(argv[4], "AGGRESSIVE_BARRIER")) {
      BarrierMode = AGGRESSIVE_BARRIER;
    }

    if (!strcmp(argv[4], "SYNCHRONIZED_BARRIER")) {
      BarrierMode = SYNCHRONIZED_BARRIER;
    }

    if (!strcmp(argv[4], "NO_BARRIER")) {
      BarrierMode = NO_BARRIER;
    }

    if (BarrierMode == -1) {
      PrintError();
      exit(1);
    }

    TimeLag = atof(argv[5]);
    if (TimeLag <= 0) {
      PrintError();
      exit(1);
    }

/*
** Create the Host User
*/

    EmHostUser = new SpEmHostUser();

/*
** Loop 10 times to test the two-way command
*/

    for (i=0; i<10; i++) {

/*
** First, send the command out
*/

      EmHostUser->SendTwoWayCommand(argv[1], argv[2], argv[3],
				    strlen(argv[3]) + 1,
				    BarrierMode, TimeLag);

/*
** Then, try to read the message back (while doing this, check if SPEEDES
** has been killed).
*/

      while (1) {

	Data = EmHostUser->ReceiveData(Time, Bytes, GlobalId);
	if (Data) break;

	if (!EmHostUser->SpeedesExecuting()) {
	  cout << "SPEEDES Killed - exiting" << endl;
	  delete EmHostUser;
	  exit(1);
	}

      }

/*
** Print out the data that was sent back (assumed to be a string for this test)
*/

      cout << Data << " at time " << Time << ", with " << Bytes << " bytes"
	   << endl;

/*
** Delete the data now that we are done with it.
*/

      delete [] Data;

    }

/*
** delete the Host User since we are preparing to exit
*/

    delete EmHostUser;

  }else{

/*
** Print out an error message and exit since the arguments are not right
*/
    PrintError();

  }

  return 0;
}
