// SpStats.C method file

#include "SpMath.H"
#include "SpErrno.H"
#include <fcntl.h>
#include "SpAssert.H"
#include <sys/types.h>
#ifndef __SPEEDES_NT__
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/resource.h>
#else
#include <windows.h>
#include <io.h>
#endif
#include "SpStdio.H"
#include "SpString.H"
#include "SpStdlib.H"
#include "SpStats.H"
#include "SpSimTime.H"
#include "StatsMessage.H"
#include "speedes_comm.H"
#include "SpOS.H"

#include "SpIostream.H"
#include "SpFstream.H"
#include "SpStrstream.H"
#include "SpIomanip.H"
ofstream *GvtStatOut = NULL;	// file (stream) for GVT statistics

#ifndef INFINITY
#define INFINITY 1.0e20
#endif


extern void WriteGvtToGUI(double);

#ifdef __SPEEDES_NT__
const static int RUSAGE_SELF = 0;
struct  rusage {
        struct timeval ru_utime;
        struct timeval ru_stime;
};
  
inline int getrusage(int who, struct rusage * rusage)
{
  const static unsigned int myMaxInt = 0 - 1;
  FILETIME creationTime, exitTime, kernelTime, userTime;
  const static double numNanoSeconds = 10000000.0;
  const static double numMicoSeconds = 1000000.0;
  GetProcessTimes(GetCurrentProcess(),
		  &creationTime,
		  &exitTime,
		  &kernelTime,
		  &userTime);
  
  double userTemp = ((double)userTime.dwHighDateTime)*(double)myMaxInt +
    userTime.dwLowDateTime;
  double kernTemp = ((double)kernelTime.dwHighDateTime)*(double)myMaxInt +
    kernelTime.dwLowDateTime;

  userTemp /= numNanoSeconds;
  kernTemp /= numNanoSeconds;

  rusage->ru_stime.tv_sec = (int)floor(kernTemp);
  kernTemp -= rusage->ru_stime.tv_sec;
  rusage->ru_stime.tv_usec = (int)(kernTemp*numMicoSeconds);

  rusage->ru_utime.tv_sec = (int)floor(userTemp);
  userTemp -= rusage->ru_utime.tv_sec;
  rusage->ru_utime.tv_usec = (int)(userTemp*numMicoSeconds);

  return 0;
}

const static int STDERR_FILENO = 17;
static FILE * outputFile = NULL;
#endif

#ifdef hpux
#include <sys/syscall.h>
extern "C" {int syscall(int, int, void *);} //Don't know if parameters correct
#define getrusage(a, b)  syscall(SYS_GETRUSAGE, a, b)
#endif /* hpux */

#ifdef sun
  extern "C" {int getrusage(int who, struct rusage *rusage);}
#endif /* sun */

/************************************************************************
* STATISTICS : STATISTICS constructor					*
************************************************************************/
SpStats::SpStats(SpDataParser *parser, int n_nodes_arg, int node_arg)
  : n_nodes(n_nodes_arg), node(node_arg)
{

//...... initialize some quantities

  gvt = -INFINITY;
  cpu = 0.0;
  wall = 0.0;
  cycle = 0;
  phase1 = 0;
  phase2 = 0;
  events = -1;
  nevtgvt = 0;
  rollbacks = 0;
  messages = 0;
  antimessages = 0;
  exmessages = 0;
  star = 0;
  oldgvt = -1.0e20;
  oldwall = 0.0;
  TotalProcessingTime = 0.0;
  TotalCommittedTime = 0.0;
  LocalTime = -1e20;

  PhaseBTB = 0.0;
  PhaseFBen = 0.0;
  PhaseFBex = 0.0;
  PhaseGVT = 0.0;
  PhaseCMT = 0.0;

//...... set up the printout ReportTime

  parser->GoTo("statistics", NULL);
  ReportTime = parser->GetInt("ReportTime");
  if (node == 0)
  {
    const char * const Output_Method = parser->GetString("Output_Method");
    if (strcasecmp(Output_Method, "stderr") == 0)
      output_fd = STDERR_FILENO;
    else if (strcasecmp(Output_Method, "File") == 0)
    {
      const char * const Stat_Output_Filename
	= parser->GetString("Stat_Output_Filename");
      //output_fd = open(Stat_Output_Filename, O_WRONLY | O_CREAT);
#ifdef __SPEEDES_NT__
      outputFile = fopen(Stat_Output_Filename, "w");
#else
      output_fd = creat(Stat_Output_Filename, 0600);
#endif
      if (output_fd <= 0){
	//exit_all_other_nodes(); //SpComm
	flush(cout);
	flush(cerr);
	fflush(stdout);
	fflush(stderr);
	cerr << "\nError: Could not open Stat_Output_Filename \""
	     << Stat_Output_Filename << '"' << endl;
	perror("");
	exit(1);
      }
    }
    else if (strcasecmp(Output_Method, "Socket") == 0)
    {
      /* Create socket */
      output_fd = socket(AF_INET, SOCK_STREAM, 0);
      if (output_fd < 0) {
	perror("opening stream socket");
	exit(1);
      }
     
      const char * const Stat_Displayer = parser->GetString("Stat_Displayer");
      struct hostent *hp = gethostbyname(Stat_Displayer);
      if (hp == NULL) {
	fprintf(stderr, "%s: unknown Stat_Displayer", Stat_Displayer);
	exit(2);
      }
     
      struct sockaddr_in server;
      server.sin_family = AF_INET;
      int port = parser->GetInt("Port");
      server.sin_port = htons(port);
      memmove(&server.sin_addr, hp->h_addr, hp->h_length);

      if (connect(output_fd, (sockaddr*) &server, sizeof(server)) < 0)
      {
	//exit_all_other_nodes();
	flush(cout);
	flush(cerr);
	fflush(stdout);
	fflush(stderr);
	cerr << "\nError: Could not connect to Stat_Displayer "
	     << Stat_Displayer << " on port " << port << endl;
	perror("");
        // This exit doesn't work when running with
        // speedes_comm_server.  It will print the above cerr line but
        // will not actually exit.  But it does work when not using
        // speedes_comm_server.
	exit(1);
      }
    }
    else
    {
      //exit_all_other_nodes();
      flush(cout);
      flush(cerr);
      fflush(stdout);
      fflush(stderr);
      cerr << "\nError: You specified \"" << Output_Method
	   << "\" for the Output_Method in\nthe statistics section of the"
	" speedes.par file, but the three\nallowable output methods are"
	" stderr, File, and Socket." << endl;
      exit(1);
    }
  }
  
  tprint = 0.0;

//...... set up the print flag

  print_flag = 0;
  if (parser->GetLogical("CYCLE"))		print_flag += CYCLE;
  if (parser->GetLogical("GVT"))		print_flag += GVT;
  if (parser->GetLogical("WALL"))		print_flag += WALL;
  if (parser->GetLogical("CPU"))		print_flag += CPU;
  if (parser->GetLogical("STAR"))		print_flag += STAR;
  if (parser->GetLogical("PHASE1"))		print_flag += PHASE1;
  if (parser->GetLogical("PHASE2"))		print_flag += PHASE2;
  if (parser->GetLogical("EVENTS"))		print_flag += EVENTS;
  if (parser->GetLogical("EVTGVT"))		print_flag += EVTGVT;
  if (parser->GetLogical("EVENTSCYCLE"))	print_flag += EVENTSCYCLE;
  if (parser->GetLogical("ROLLBACKS"))		print_flag += ROLLBACKS;
  if (parser->GetLogical("MESSAGES"))		print_flag += MESSAGES;
  if (parser->GetLogical("ANTIMESSAGES"))	print_flag += ANTIMESSAGES;
  if (parser->GetLogical("CANCELS"))		print_flag += CANCELS;
  if (parser->GetLogical("EXMESSAGES"))		print_flag += EXMESSAGES;
  if (parser->GetLogical("PROC"))		print_flag += PROCESSED;
  if (parser->GetLogical("COMMIT"))		print_flag += COMMIT;
  if (parser->GetLogical("PROCEFF"))		print_flag += PROCEFF;
  if (parser->GetLogical("LVT"))		print_flag += LVT;
  if (parser->GetLogical("BTW"))		print_flag += BTW;
  StatsByNode = parser->GetLogical("STATS_BY_NODE");

  if (parser->GetLogical("WriteGvtStatistics")) {
    char s[120];
    char *GvtFileName = parser->GetString("FileName");

    for (int i=0; i<120; i++) s[i] = 0;
    sprintf(s,"%s.%d", GvtFileName, node);

    GvtStatOut = new ofstream();
    GvtStatOut->open(s, ios::out);
  }

}

/************************************************************************
* print : print out the statistics to standard out			*
************************************************************************/
void SpStats::print() {
  char string[256];
  int itemp;
  int index;
  double GlobTproc = 0;
  double GlobTcmt = 0;
  struct rusage Usage;
  static StatsMessage *SMsg = NULL;
  static double cpu_offset = -1.0;

  if (cpu_offset == -1) {
    getrusage(RUSAGE_SELF, &Usage);
    int MyCPUsec = (int) (Usage.ru_utime.tv_sec + Usage.ru_stime.tv_sec);
    int MyCPUusec = (int)(Usage.ru_utime.tv_usec + Usage.ru_stime.tv_usec);
    cpu_offset = MyCPUsec + double(MyCPUusec)/1000000.0;
  }

  if (SMsg == NULL){
    SMsg = new StatsMessage[n_nodes];
  }

//...... check if we need to do any work

  if (print_flag == 0) return;

  if (node == 0) {
    if ((wall - tprint) > ReportTime) {
      tprint = wall;
      itemp = 1;
    }else{
      itemp = 0;
    }
  }else{
    itemp = 0;
  }

  if (n_nodes > 1) {
    itemp = SpComm_GlobalSum(itemp);
  }

  if (itemp == 0) return;


#ifdef DPAT
//...... a KLUDGE function added to support the DPAT GUI

  if (node == 0) WriteGvtToGUI(gvt);

#endif


//...... set up the index to be 0

  index = 0;

//...... cycle

  if (print_flag & CYCLE) {
      SMsg[node].SetWhichCycle(cycle);
    sprintf(&string[index],"%d)  ",cycle);
    index += strlen(&string[index]);
  }

//...... gvt

  if (print_flag & GVT) {
    SMsg[node].SetGVT(gvt);
    sprintf(&string[index],"GVT=%g  ",gvt);
    index += strlen(&string[index]);
  }

//...... cpu

  if (print_flag & CPU) {
    getrusage(RUSAGE_SELF, &Usage);
    int MyCPUsec = (int) (Usage.ru_utime.tv_sec + Usage.ru_stime.tv_sec);
    int MyCPUusec = (int)(Usage.ru_utime.tv_usec + Usage.ru_stime.tv_usec);
    cpu = MyCPUsec + double(MyCPUusec)/1000000.0 - cpu_offset;
    SMsg[node].SetNodeCPU(cpu);
    sprintf(&string[index],"cpu=%g  ", SpComm_GlobalSum(cpu));
    index += strlen(&string[index]);
  }

//...... wall

  if (print_flag & WALL) {
    SMsg[node].SetWall(wall);
    sprintf(&string[index],"wall=%g  ",wall);
    index += strlen(&string[index]);
  }

//...... star

  if (print_flag & STAR) {
    if (oldgvt == -1.0e20) {
      star = 0;
    }else{
      star = (gvt - oldgvt) / (wall - oldwall + 1e-20);
    }
    oldgvt = gvt;
    oldwall = wall;
    SMsg[node].SetSTAR(star);
    sprintf(&string[index],"STAR=%g  ",star);
    index += strlen(&string[index]);
  }

//...... Tproc

  if (print_flag & PROCESSED) {
    SMsg[node].SetTproc(TotalProcessingTime);
    GlobTproc = SpComm_GlobalSum(TotalProcessingTime);
    sprintf(&string[index],"Tproc=%g  ",GlobTproc);
    index += strlen(&string[index]);
  }

//...... Tcommit

  if (print_flag & COMMIT) {
    SMsg[node].SetTcmt(TotalCommittedTime);
    GlobTcmt = SpComm_GlobalSum(TotalCommittedTime);
    sprintf(&string[index],"Tcmt=%g  ",GlobTcmt);
    index += strlen(&string[index]);
  }

//...... Eff

  if (print_flag & PROCEFF) {
    SMsg[node].SetEff(TotalCommittedTime
                   / (1e-20 + TotalProcessingTime));
    sprintf(&string[index],"Eff=%g  ",GlobTcmt / (1e-20 + GlobTproc));
    index += strlen(&string[index]);
  }

//...... phase 1

  if (print_flag & PHASE1) {
    SMsg[node].SetPhase1(phase1);
    sprintf(&string[index],"Nproc=%d  ",SpComm_GlobalSum(phase1));
    index += strlen(&string[index]);
  }

//...... phase 2

  if (print_flag & PHASE2) {
    SMsg[node].SetPhase2(phase2);
    sprintf(&string[index],"Ncmt=%d  ",SpComm_GlobalSum(phase2));
    index += strlen(&string[index]);
  }

//...... events

  if (print_flag & EVENTS) {
    int totevents = SpComm_GlobalSum(events);
    SMsg[node].SetEvents(events);
    sprintf(&string[index],"e=%d  ",totevents);
    index += strlen(&string[index]);
    
    if (print_flag & EVENTSCYCLE) {
	SMsg[node].SetEventsPerCycle(int (double(SMsg[node].GetEvents())/
				       (double(n_nodes)*double(cycle)) + 0.5));
      sprintf(&string[index],"e/c=%d  ",
	      int(double(totevents)/(double(n_nodes)*double(cycle)) + 0.5));
      index += strlen(&string[index]);
    }
    
  }

//...... events processed during gvt update

  if (print_flag & EVTGVT) {
    SMsg[node].SetEvtGVT(nevtgvt);
    sprintf(&string[index],"eg=%d  ", SpComm_GlobalSum(nevtgvt));
    index += strlen(&string[index]);
  }

//...... rollbacks

  if (n_nodes > 1) {

    if (print_flag & ROLLBACKS) {
      SMsg[node].SetNumRollBacks(rollbacks);
      sprintf(&string[index],"r=%d  ", SpComm_GlobalSum(rollbacks));
      index += strlen(&string[index]);
    }

//...... messages

    if (print_flag & MESSAGES) {
      SMsg[node].SetNumMessages(messages);
      sprintf(&string[index],"m=%d  ", SpComm_GlobalSum(messages));
      index += strlen(&string[index]);
    }

//...... antimessages

    if (print_flag & ANTIMESSAGES) {
      SMsg[node].SetNumAntiMessages(antimessages);
      sprintf(&string[index],"a=%d  ", SpComm_GlobalSum(antimessages));
      index += strlen(&string[index]);
    }

  }

//...... canceled events

  if (print_flag & CANCELS) {
    SMsg[node].SetNumCancels(cancels);
    sprintf(&string[index],"c=%d  ", SpComm_GlobalSum(cancels));
    index += strlen(&string[index]);
  }

//...... external messages

  if (print_flag & EXMESSAGES) {
    SMsg[node].SetNumExternalMessages(exmessages);
    sprintf(&string[index],"x=%d  ", SpComm_GlobalSum(exmessages));
    index += strlen(&string[index]);
  }

/*
** Others in the message...
*/

  SMsg[node].SetLocalTime(LocalTime);
  SMsg[node].SetPhaseBTB(PhaseBTB);
  SMsg[node].SetPhaseFBen(PhaseFBen);
  SMsg[node].SetPhaseFBex(PhaseFBex);
  SMsg[node].SetPhaseGVT(PhaseGVT);
  SMsg[node].SetPhaseCMT(PhaseCMT);

//...... Collect and print out the statistics

  if (node != 0){
    SMsg[node].SetNode(node);
    SpComm_Send(0, 234, sizeof(StatsMessage), (void *)&(SMsg[node]));
  }
  else{
    int count = 0;
    int Nnodes = SpComm_GetNumNodes();
    while (count != Nnodes-1){
      char * msg;
      int nbytes;
      if ( (msg = (char *)SpComm_Receive(234, nbytes)) != NULL){
	StatsMessage* OtherSmsg = (StatsMessage*) msg;
	memcpy(&SMsg[OtherSmsg->GetNode()], msg, sizeof(StatsMessage));
	delete [] ((char*) msg);
	++count;
      }
    }
    sprintf(&string[index],"\n");
    index += strlen(&string[index]);
    
#ifdef __SPEEDES_NT__
    if (output_fd == STDERR_FILENO){
      cerr<<string;
    }
    else if ( outputFile != NULL){
      fwrite(string, sizeof(char), strlen(string),outputFile);
    }
    else{
      if (write(output_fd, string, index) < 0)
	{
	  perror("writing on stream socket");
	  assert(0);
	}
    }
#else
    for (;;)
    {
      int s = write(output_fd, string, index);
      if (s >= 0)
      {
	if (s != index)
	  cerr << "Warning: Could not write all of the requested statistical information." << endl;
	break;
      }
      else
      { 
	if (errno == EINTR)
	  continue;
	else
	{
	  perror("Warning: Could not write the requested statistical information.");
	  assert(0);
	}
      } 
    }
#endif
    
    if (StatsByNode)
    {
      FILE * StatOutput;
      while ((StatOutput = fopen("./StatOutput", "w")) == NULL);
      fwrite(SMsg, sizeof(StatsMessage), n_nodes, StatOutput);
      fclose(StatOutput);
    }
  }

/*
** Write out gvt statistics by node to a file (if turned on)
*/

  WriteNodeStatistics();

}

/************************************************************************
* fast_print : fast print out the statistics to standard out		*
************************************************************************/
void SpStats::fast_print(double g) {

  if (node != 0) return;

  if ((wall - tprint) > ReportTime) {
    tprint = wall;
  }else{
    cycle++;
    return;
  }

  gvt = g;
  fprintf(stderr,"%d)  gvt=%g  wall=%g\n", cycle, gvt, wall);
  cycle++;

}

/************************************************************************
* WriteNodeStatistics : Writes gvt statistics by node to a file		*
************************************************************************/
void SpStats::WriteNodeStatistics() {
  if (!GvtStatOut) return;

#define PRINT_SEP endl

  if (print_flag & CYCLE) {
    *GvtStatOut << "Cycle=" << cycle << PRINT_SEP;
  }

  if (print_flag & GVT) {
    *GvtStatOut << "GVT=" << gvt << PRINT_SEP;
  }

  if (print_flag & LVT) {
    *GvtStatOut << "LVT=" << LocalTime << PRINT_SEP;
  }

  if (print_flag & BTW) {
    if (PhaseCMT > 0.0) {
      *GvtStatOut << "Risk=" << PhaseBTB << PRINT_SEP;
      *GvtStatOut << "EnterFB=" << PhaseFBen << PRINT_SEP;
      *GvtStatOut << "ExitFB=" << PhaseFBex << PRINT_SEP;
      *GvtStatOut << "UpdateGVT=" << PhaseGVT << PRINT_SEP;
      *GvtStatOut << "Commit=" << PhaseCMT << PRINT_SEP;
    }
  }

  if (print_flag & CPU) {
    *GvtStatOut << "CPU=" << cpu << PRINT_SEP;
  }

  if (print_flag & WALL) {
    *GvtStatOut << "Wall=" << wall << PRINT_SEP;
  }

  if (print_flag & STAR) {
    *GvtStatOut << "STAR=" << star << PRINT_SEP;
  }

  if (print_flag & PROCESSED) {
    *GvtStatOut << "Tproc=" << TotalProcessingTime << PRINT_SEP;
  }

  if (print_flag & COMMIT) {
    *GvtStatOut << "Tcmt=" << TotalCommittedTime << PRINT_SEP;
  }

  if (print_flag & PROCEFF) {
    *GvtStatOut << "Eff=" << TotalCommittedTime
                               / (1e-20 + TotalProcessingTime) << PRINT_SEP;
  }

  if (print_flag & PHASE1) {
    *GvtStatOut << "Nproc=" << phase1 << PRINT_SEP;
  }

  if (print_flag & PHASE2) {
    *GvtStatOut << "Ncmt=" << phase2 << PRINT_SEP;
  }

  if (print_flag & EVENTS) {
    *GvtStatOut << "Evts=" << events << PRINT_SEP;
  }

  if (print_flag & EVENTSCYCLE) {
    *GvtStatOut << "E/Cyc=" << int(0.5 + double(events)/double(cycle))
		<< PRINT_SEP;
  }

  if (print_flag & EVTGVT) {
    *GvtStatOut << "E/gvt=" << nevtgvt << PRINT_SEP;
  }

  if (print_flag & ROLLBACKS) {
    *GvtStatOut << "RBs=" << rollbacks << PRINT_SEP;
  }

  if (print_flag & MESSAGES) {
    *GvtStatOut << "Msgs=" << messages << PRINT_SEP;
  }

  if (print_flag & ANTIMESSAGES) {
    *GvtStatOut << "Anti=" << antimessages << PRINT_SEP;
  }

  if (print_flag & CANCELS) {
    *GvtStatOut << "Cancel=" << cancels << PRINT_SEP;
  }

  if (print_flag & EXMESSAGES) {
    *GvtStatOut << "Xmsgs=" << exmessages << PRINT_SEP;
  }

  *GvtStatOut << endl;

}
