// O_SpBlockDecomp.C source file

//#define DEBUG_BLOCK_DECOMP

#include "O_SpBlockDecomp.H"

extern SpSimObj *CurrentSimobj;

//-----------------------------------------------------------------------------
// O_SpBlockDecomp constructor
//-----------------------------------------------------------------------------
O_SpBlockDecomp::O_SpBlockDecomp() {}

//-----------------------------------------------------------------------------
// CreateSimObjs - creates all of the SimObjs using a block decomposition
// strategy.
//-----------------------------------------------------------------------------
void O_SpBlockDecomp::CreateSimObjs(int numObjs, int replication) {
  int currentId;
  int i;
  int seed;
  SpSimObj *simObj;
  char *name;
  char s[120];

  SimObjMgrNames->Insert(this, GetSimObjMgrName());
  SimObjMgrTypes->Insert(this, OBJECT_TYPE);

/*
** Find out how many nodes are in our cluster along with our cluster node.
*/

  GetClusterInfo(GetClusterName(), NumClusterNodes, ClusterNode);

  N_TOT = numObjs;
  DecompInfo(N_TOT, NumClusterNodes, ClusterNode,
	     FirstClusterNode, N_LOC, Offset);

/*
** Reserve room for N_LOC objects and then create them in a loop.
*/

  reserve_n_objects(N_LOC);

  currentId = Offset;
  for (i=0; i<N_LOC; i++) {

/*
** Create the simObj
*/

    simObj = CreateSimObj();

/*
** Generate a unique seed for the simObj
*/

    seed = 1 + NumTotCreated + currentId + (N_TOT*replication);
    simObj->set_SEED(seed);

/*
** Pass in the Number Id (e.g., 0,1,2,...,N_TOT-1) for the object
*/

  simObj->SetNumId(currentId);

/*
** Generate the name of the simObj
*/
    if ((name = GetSimObjMgrName()) && (Parser==NULL)) {
      sprintf(s, "%s %d", name, currentId);
      simObj->set_NAME(strdup(s));
    }

/*
** Pass the object into SPEEDES and then increment the currentId
*/

    define_object(simObj);
    currentId++;

  }

  TOTAL_OBJECTS += N_TOT;
  NumTotCreated += N_TOT;

/*
** Initialize the simObj
*/

  SpSimObj *tempSimObj = CurrentSimobj;
  if (Parser == NULL) {
    for (i=0; i<N_LOC; i++) {
      simObj = get_obj(i);
      CurrentSimobj = simObj;
      simObj->Init();
    }
  }
  CurrentSimobj = tempSimObj;

}

//-----------------------------------------------------------------------------
// CreateSimObjs - creates all of the SimObjs using a card dealing
// strategy.
//-----------------------------------------------------------------------------
void O_SpBlockDecomp::CreateSimObjs(char *parFileName, int replication) {
  int i, j;
  SpSimObj *simObj;
  void *Mark;

  Parser = new SpDataParser(parFileName);
  Mark = Parser->GoTo("Objects", NULL);
  N_TOT = Parser->GetNumReferences(Mark, "OBJECT");

  CreateSimObjs(N_TOT, replication);

  Parser->GoToFirstReference(Mark, "OBJECT");
  j = 0;
  for (i=0; i<N_TOT; i++) {

    int *num = new int;
    *num = i;
    SimObjNames.Insert(num, Parser->GetReferenceName());

    SpSimObj *tempSimObj = CurrentSimobj;
    if (Offset <= i) {
      if (i < Offset+N_LOC) {
	simObj = get_obj(j);
	simObj->set_NAME(Parser->GetReferenceName());
	CurrentSimobj = simObj;
	simObj->InitParser(Parser);
	j++;
      }
    }
    CurrentSimobj = tempSimObj;
    Parser->GoToNextReference(Mark, "OBJECT");
  }

}

//-----------------------------------------------------------------------------
// DecompInfo - returns information needed for decomposing the objects to
// different cluster nodes. In this case, the algorithm is card dealing.
//-----------------------------------------------------------------------------
void O_SpBlockDecomp::DecompInfo(int numTotalObjs,
			       int numClusterNodes,
			       int clusterNode,
			       int &firstClusterNode,
			       int &numLoc,
			       int &offset) {

  firstClusterNode = NumTotCreated % numClusterNodes;

/*
** Determine my equivalent node
*/

  int myNode = (numClusterNodes + clusterNode
		- firstClusterNode) % numClusterNodes;

  int n1 = numTotalObjs / numClusterNodes;
  int n2 = numTotalObjs % numClusterNodes;

  if (myNode < n2) {
    offset = myNode * (n1+1);
    numLoc = n1+1;
  }else{
    offset = n2 * (n1+1) + (myNode-n2) * n1;;
    numLoc = n1;
  }

}

//-----------------------------------------------------------------------------
// GetObjHandle - returns the object handle for an object who's numerical
// Id is passed in. In this case, the  card dealing algorithm is used to do
// the computation.
//-----------------------------------------------------------------------------
SpObjHandle O_SpBlockDecomp::GetObjHandle(int numId) {
  int locId;
  int nodeId;

  int n1 = N_TOT / NumClusterNodes;
  int n2 = N_TOT % NumClusterNodes;
  int n3 = (n1+1) * n2;

  if (numId < n3) {
    nodeId = numId / (n1+1);
    locId = numId % (n1+1);
  }else{
    nodeId = n2 + (numId-n3) / n1;
    locId = (numId-n3) % n1;
  }

  int cn = (nodeId + FirstClusterNode) % NumClusterNodes;

  SpObjHandle objHandle;
  objHandle.NodeId = GetNode(GetClusterName(), cn);
  objHandle.SimObjId = locId;
  objHandle.ObjManId = OBJECT_TYPE;

  return objHandle;

}
