// SpComm.H header file defines all of the communications services that
// are used by SPEEDES.

#ifndef SpeedesComm_Services
#define SpeedesComm_Services

/*
** Include files required by the SPEEDES Communications library
*/

#include "SpSimTime.H"		// Abstract representation of simulation time
#include "SpDestination.H"	// Supports multiple-node destinations

/*
** Starting up the communications service. This routine must be done by the
** application at the beginning of the program before any of the other
** communications services are invoked. The command line arguments are
** passed into the StartUp routine and have the following meaning.
**
** 0 arguments: Number of nodes and group ID specified in speedes.par file.
**              All nodes run on same machine
**
** 1 argument:  Value = Number of nodes (all on same machine).
**              Group ID specified in speedes.par file
**
** 2 arguments: First value = Number of nodes on that machine
**              Second value = number of total nodes in the simulation
**              Group ID specified in the speedes.par file
**
** 3 arguments: First value = Number of nodes on that machine
**              Second value = number of total nodes in the simulation
**              Third value = Group ID
**
** Note, if there are any problems in starting up the communications, this
** routine should return ERROR_SP_COMM, otherwise it returns GOOD_SP_COMM.
*/

enum {
  GOOD_SP_COMM,
  ERROR_SP_COMM
};

enum {
  MsgsSent,
  MsgsQueued,
  MsgsError
};

int SpComm_StartUp(int nNodes, int nLoc, int group, char * MiscInfo);

// If we are at the end of the program, then we do not need to delete
// memory or close sockets since the program is about to exit and the
// OS will delete that memory and close the sockets for us.  In fact,
// trying to do it ourselves could be bad since to do this may require
// communication with other nodes, and if something is going wrong,
// then we may wait forever for a response from those other nodes.
// This would cause us to never exit.  However even though the OS will
// clean up dynamic memory and sockets, it will NOT clean up shared
// memory, and it is easy to run out of shared memory.  Therefore it
// is good to at least try to clean up shared memory since that
// doesn't require communication with other nodes, but as mentioned
// above, it is bad to try to clean up other memory and sockets.
//
// However, it would still be good to clean this stuff up so that
// error-checking programs like Insure++ and purify do not complain
// about memory leaks.  Therefore we have a compromise by supplying
// two cleanup functions.  SpComm_CleanupShm just cleans up shared
// memory (and some very small additional dynamic memory used to keep
// track of this shared memory which can only be cleaned up after the
// shared memory is cleaned up).  And SpComm_CleanupOther cleans up
// other memory and sockets.  Users of SpComm can pass a pointer to
// SpComm_CleanupShm to the signal function and to the atexit
// function.  And users can call SpComm_CleanupOther at the end of
// their program to clean up other things (sockets and dynamic memory
// from new and malloc).  This way, if we are exiting because of some
// error like writing to a NULL pointer, then because of the earlier
// call to signal, the SpComm_CleanupShm function will immediately
// clean up shared memory which is necessary since the OS will not do
// that for us.  However if there is no serious error like writing to
// NULL, then we will make it to the end of the user's program where
// the call to SpComm_CleanupOther will run and clean up sockets and
// dynamic memory.  This will allow someone who is using Insure++ to
// get useful info about leaks he needs to clean up.  Then the user's
// program exits and the atexit function of SpComm_CleanupShm is
// called to cleanup the shared memory.  At the end of the user's
// program, when it calls SpComm_CleanupOther it will need
// communication with other nodes, but that's OK, because in this
// example there are no serious problems, which means the other nodes
// will be able to respond and everything will be OK.

// Clean up shared memory.
void SpComm_CleanupShm();

// Clean up things other than shared memory.
void SpComm_CleanupOther();


/*
** Returns the number of nodes and the local node id.
*/

int SpComm_GetNumNodes();	// returns the number of nodes
int SpComm_GetNodeId();	       	// returns the node Id [0 through NumNodes-1]

/*
** Barrier synchronizations. The BarrierSync function blocks until all of
** the nodes have made this call. The EnterFuzzyBarrier function works
** together with the ExitFuzzyBarrier function. Nodes call the non-blocking
** EnterFuzzyBarrier function and then continue their processing. The
** ExitFuzzyBarrier function is then polled to determine when all of the
** nodes have entered the fuzzy barrier.
*/

void SpComm_BarrierSync();	 // returns when all nodes make this call
void SpComm_EnterFuzzyBarrier(); // returns right away (non-blocking)
int SpComm_ExitFuzzyBarrier();   // returns 0 or 1 (1 means exit fuzzy barrier)

/*
** Global Reductions combine values from each node. The operations include
** global sums, global minimums, and global maximums.The result is returned.
*/

int SpComm_GlobalSum(int value);	     // sum of everyone's value
double SpComm_GlobalSum(double value);	     // sum of everyone's value

int SpComm_GlobalMin(int value);	     // min of everyone's value
double SpComm_GlobalMin(double value);	     // min of everyone's value
SpSimTime &SpComm_GlobalMin(SpSimTime &SimTime); // min of everyone's SimTime

int SpComm_GlobalMax(int value);	     // max of everyone's value
double SpComm_GlobalMax(double value);	     // max of everyone's value
SpSimTime &SpComm_GlobalMax(SpSimTime &SimTime); // max of everyone's SimTime


/* 
**    The following send functions read the data at buff, but after
** the send functions return they will never read at this buff
** again. Therefore it is safe to do things such as delete this buff
** if allocated with new or to change the data at buff after the send
** functions return.
*/ 


/* 
**    The application must delete the memory for a received message
** when it is no longer needed.  This delete must be done to a pointer
** whose type is char* and square brackets must be used on this
** delete.
*/ 


/*
** Asynchronous (non-blocking) message sending. These are general
** purpose send and receive routines for asynchronous message
** sending. The three versions of send are (1) node-to-node, (2)
** node-to-multiple-nodes, and (3)
** node-to-every-other-node-but-self. Invoking the receive routine
** without any arguments reads incoming messages and queues them up by
** their type (types range from 0 to N_MESSAGE_TYPES-1). Nothing is
** returned to the application. The receive function with arguments
** reads incoming messages and then returns a message of the specified
** type if one is available (NULL is returned if there are no
** available messages of the specified type. The GetPendingMessage
** routine does not try to read an incoming messages but returns a
** previously queued message of the specified type if one is available
** (again, NULL is returned if there are no messages of the specified
** type).
*/

#define N_MESSAGE_TYPES 256

void SpComm_Send(int destination, int type, int nbytes, void *buff);
void SpComm_Send(SpDestination *dest, int type, int nbytes, void *buff);
void SpComm_Send(int type, int nbytes, void *buff);

void SpComm_Receive();
char *SpComm_Receive(int type, int &nbytes);
char *SpComm_GetPendingMessage(int type, int &nbytes);

/*
** Coordinated message sending. These routines are used differently by
** an application from the asynchronous services. It is assumed that
** each node has a number of messages that it wants to send to
** different nodes. Again, there are three ways for defining the
** destination (1) node-to-node, (2) node-to-multiple-nodes, and (3)
** node-to-every-other-node-but-self
**
** Each node first makes calls to the coordinated send routine for
** each message that it wants to send. After all of the outgoing
** messages have been passed to the coordinated send routines, the
** application is expected to make calls to the coordinated receive
** routine to receive messages from other nodes. The coordinated
** receive routine returns a NULL pointer when there are no further
** messages to be received by that node. The coordinated message
** passing services guarantee that all coordinated messages are
** received before the NULL pointer is returned. Note that this call
** may block if it has no messages to return and all messages have
** not yet been received.
**
** It should be noted that it is a logic error to mix these calls with
** any of the reductions.  That is, once one has sent a coordinated
** message, one should not make a call to any synchronization until
** the coordinated receive has returned NULL.
*/

void SpComm_CoordinatedSend(int destination, int nbytes, void *buff);
void SpComm_CoordinatedSend(SpDestination *dest, int nbytes, void *buff);
void SpComm_CoordinatedSend(int nbytes, void *buff);

char *SpComm_CoordinatedReceive(int &nbytes);

#endif

