// SpComponent.H header file defines the base classe that supports adding
// reusable components with associated event handlers to simobjs

#ifndef COMP_OBJ
#define COMP_OBJ

#include "SpString.H"
#include "xlist.H"
#include "SpComponent.H"
#include "RB_Types.H"
#include "SpEventHandlerMethod.H"

class SpSimObj;

/******************************************************************************
* Class: SpComponent
* Summary: A generic reusable object with associated event handlers
*          for plugging in and plugging out of SpSimObjs during
*          compile time and/or runtime. NOTE: ROLLBACK
*              NOT IMPLEMENTED YET.
* Description: Components allow the decoupling of subobjects in
*              SpSimObjs so that such subobjects, if they inherit from
*              SpComponent, can be plugged into various different
*              SpSimObjs and rollbackably removed from those SpSimObjs
*              during runtime. When SpComponents are added and removed
*              to and from SpSimObjs, any event handlers associated
*              with the SpComponents are also added and removed. To
*              associate event handlers with SpComponentS, use the
*              SpComponent AddEventHandler and RemoveEventHandler
*              methods, which are also rollbackable.
******************************************************************************/
class SpComponent {

 public:

/******************************************************************************
* Constructor Method: SpComponent
******************************************************************************/
   SpComponent();

/******************************************************************************
* Constructor Method: SpComponent
* Summary: Constructors of SpComponents require the associated SpSimObj pointer
* Description: Use the following syntax for descendents of SpComponent:
*              SpComponent_CHILD (SpSimObj *simobj): SpComponent (simobj)
*              { ... }
******************************************************************************/
   SpComponent(SpSimObj *simobj // pointer to associated SpSimObj
	     );

/******************************************************************************
* Method: PluggedIn
* Summary: Returns whether this component is currently plugged in to
*          the SpSimObj
* Description: 
* Return: Whether this component is currently plugged in to the SpSimObj
******************************************************************************/
   int PluggedIn() { return AddedToSimobj; } 

 protected:

/******************************************************************************
* Method: AddEventHandler
* Summary: Add an event handler to this SpComponent
* Description: When this SpComponent is added to or removed from a
*              SpSimObj, the event handlers previously added to or
*              removed from this SpComponent using AddEventHandler(),
*              AddEventHandlerMethod(), and RemoveEventHandler()
*              methods will automatically be added to or removed from
*              that SpSimObj as well. This method can be used
*              rollbackably during runtime.
* Return: None
******************************************************************************/
   void AddEventHandler(char *eventName, // name of event   
			SpEventHandler *eventHandler // event handler
			);

/******************************************************************************
* Method: AddEventHandlerMethod
* Summary: Add an event handler method to this SpComponent
* Description: Pass the name of the method to this method's second
*              argument. When this SpComponent is added to or removed
*              from a SpSimObj, the event handlers previously added to
*              or removed from this SpComponent using AddEventHandler(),
*              AddEventHandlerMethod(), and RemoveEventHandler()
*              methods will automatically be added to or removed from
*              that SpSimObj as well. This method can be used
*              rollbackably during runtime.
* Return: Pointer to the event handler created by calling
*         AddEventHandlerMethod(). This is an optional handle if needed
*         to subsequently remove this event handler using
*         RemoveEventHandler().
******************************************************************************/
   SpEventHandler *AddEventHandlerMethod (char *eventName,
					 // name of event
					 EVENT_HANDLER_METHOD_FNC_PTR method,
					 // method name to be invoked
					 // as the associated event
					 // handler.
					 // (implementation-wise, it
					 // is a pointer created by
					 // the EVENT_HANDLER_METHOD_H
					 // macro)
					 void *objPtr
					 // pointer to the method's
					 // object (will be the 'this'
					 // pointer if the method is
					 // contained directly in the
					 // SpSimObj or SpComponent.
					 );
   
/******************************************************************************
* Method: RemoveEventHandler
* Summary: Removes the event handler from the SpComponent
* Description: Removes the event handler from the SpComponent's set of
*              event handlers. When this SpComponent is added to or
*              removed from a SpSimObj, the event handlers previously
*              added to or removed from this SpComponent using
*              AddEventHandler(), AddEventHandlerMethod(), and
*              RemoveEventHandler() methods will automatically be
*              added to or removed from that SpSimObj as well. This
*              method can be used rollbackably during runtime.
* Return: None
******************************************************************************/
   void RemoveEventHandler(char *eventName,             // name of event
			   SpEventHandler *eventHandler  // pointer to
			                                // SpEventHandler
			   );

 public:

   virtual ~SpComponent();
   friend class SpSimObj;
   
 protected:

   SpSimObj *SimObj;       // pointer to this component's SpSimObj
   int ObjectType;         // pointer to this component's SpSimObj's object
                           // type
   int LocalId;            // pointer to this component's SpSimObj's local ID
   int GlobalId;           // pointer to this component's SpSimObj's global ID
   int Node;               // pointer to this component's SpSimObj's node

 private:
   
   XLIST EventHandlerList; // list of event handlers
   RB_int AddedToSimobj;   // whether this component is presently plugged
                           // into a SpSimObj
   int LoadedSimobjData;   // whether this component has had simobj data loaded
   void LoadSimobjData(SpSimObj *simobj);
                           // load Simobj data (the protected variables)
};

/******************************************************************************
* Internal Struct: EventHandlerElement
* Summary: elements for the EventHandlerList XLIST
* Description: 
******************************************************************************/
struct EventHandlerElement
{
  EventHandlerElement():eventName(NULL){;} //Added just for persistence
   EventHandlerElement (char *name, SpEventHandler *handler):
      eventHandler(handler)
   {
      eventName = new char[strlen(name)+1];
      memcpy(eventName, name, strlen(name)+1);
   }
   ~EventHandlerElement ()
   {
      delete [] eventName;
   }
   char *eventName;
   SpEventHandler *eventHandler;
};

//RB_DEFINE_CLASS(EventHandlerElement); // Change back to WITH_NO_DEFAULT_CONSTRUCTOR
				// when persistence supports it.

typedef SpComponent COMPONENT;

#endif
