// SpHostUser.C file that implements the Host User class

#include "SpIostream.H"
#include "SpFstream.H"
#include "SpStrstream.H"
#include "SpIomanip.H"
#include "SpStdlib.H"

#include "SpHostUser.H"
#include "SpServer.H"
#include "SpHostRouterMsgs.H"
#include "SpNameHandle.H"
#include "SpMsgHolder.H"
#include "SpClientSock.H"
#include "SpEndianNature.H"
#include "RB_Types.H"

/*
*****************************************************************************
* SpHostUser no-argument constructor
*****************************************************************************
*/

SpHostUser::SpHostUser() {
  SpEndianNature::DetermineEndianNature();
  SpDataParser Parser("speedes.par");

  Parser.GoTo("SpeedesServer", NULL);
  GroupId = Parser.GetInt("Group");

  Parser.GoTo("SpeedesServer", "HostRouter", NULL);

  if (Parser.GetLogical("UseDefaults")) {
    Parser.GoTo("SpeedesServer", NULL);
    ClientSock = new CLIENT_SOCK(Parser.GetString("DefaultMachineName"),
				 Parser.GetInt("DefaultPort"),
				 0);
  }else{
    ClientSock = new CLIENT_SOCK(Parser.GetString("MachineName"),
				 Parser.GetInt("Port"),
				 0);
  }

  InputFileFlag = 0;
  OutputFileFlag = 0;
  FlushFlag = 0;
}

/*
*****************************************************************************
* SpHostUser constructor with Data Parser as argument
*****************************************************************************
*/

SpHostUser::SpHostUser(SpDataParser *Parser) {

  SpEndianNature::DetermineEndianNature();

  Parser->GoTo("SpeedesServer", NULL);
  GroupId = Parser->GetInt("Group");

  Parser->GoTo("SpeedesServer", "HostRouter", NULL);

  if (Parser->GetLogical("UseDefaults")) {
    Parser->GoTo("SpeedesServer", NULL);
    ClientSock = new CLIENT_SOCK(Parser->GetString("DefaultMachineName"),
				 Parser->GetInt("DefaultPort"),
				 0);
  }else{
    ClientSock = new CLIENT_SOCK(Parser->GetString("MachineName"),
				 Parser->GetInt("Port"),
				 0);
  }

  InputFileFlag = 0;
  OutputFileFlag = 0;
  FlushFlag = 0;

}

/*
*****************************************************************************
* SpHostUser constructor with Machine Name and Port as arguments
*****************************************************************************
*/

SpHostUser::SpHostUser(char *MachineName, int Port, int Group) {
  SpEndianNature::DetermineEndianNature();
  GroupId = Group;
  ClientSock = new CLIENT_SOCK(MachineName, Port, 0);
  InputFileFlag = 0;
  OutputFileFlag = 0;
  ExternalGVT = -1.0e20;
  FlushFlag = 0;
}

/*
*****************************************************************************
* SpHostUser constructor with File Name as argument
*****************************************************************************
*/

SpHostUser::SpHostUser(char *FileName) {
  SpEndianNature::DetermineEndianNature();
  GroupId = -1;
  fp_input = fopen(FileName, "rb");
  InputFileFlag = 1;
  OutputFileFlag = 0;
  FlushFlag = 0;
}

/*
*****************************************************************************
* SpHostUser destructor
*****************************************************************************
*/

SpHostUser::~SpHostUser() {
  delete ClientSock;
  if (InputFileFlag) fclose(fp_input);
  if (OutputFileFlag) fclose(fp_output);
}

/*
*****************************************************************************
* SpHostUser::GetMessage() method with no arguments returns a message from
* the host router if there is one. Otherwise, NULL is returned.
*****************************************************************************
*/

void SpHostUser::ReadMessages() {
  SpMsgHolder *MsgHolder;
  GENERIC_MSG_HEADER *Message;

/*
** Return if we are not connected
*/

  if (ClientSock->ConnectionStatus() == BAD_CONNECTION_STATUS) return;

/*
** Read all incoming messages until there are no more
*/

  for (;;) {
    Message = ClientSock->getMsg();
    if (Message) {
      MsgHolder = new SpMsgHolder(Message);
      MessageQueue.push_bot(MsgHolder);
    }else{
      break;
    }
  }

}

/*
*****************************************************************************
* SpHostUser::GetMessage() method with no arguments returns a message from
* the host router if there is one. Otherwise, NULL is returned.
*****************************************************************************
*/

GENERIC_MSG_HEADER *SpHostUser::GetMessage() {
  SpMsgHolder *MsgHolder;
  GENERIC_MSG_HEADER *Message;

/*
** Return if we are not connected
*/

  if (ClientSock->ConnectionStatus() == BAD_CONNECTION_STATUS) return NULL;

/*
** First see if there are any messages queued up. Return the first one if
** are any messages queued.
*/

  if (MessageQueue.get_length()) {
    MsgHolder = (SpMsgHolder *)MessageQueue.pop_top();
    Message = MsgHolder->GetHolderMessage();
    delete MsgHolder;
    return Message;
  }

/*
** Try to read a message and return it if there is one
*/

  Message = ClientSock->getMsg();
  return Message;

}

/*
*****************************************************************************
* SpHostUser::GetMessage() method with no arguments returns a message from
* the host router if there is one. Otherwise, NULL is returned.
*****************************************************************************
*/

GENERIC_MSG_HEADER *SpHostUser::GetMessage(int Type) {
  SpMsgHolder *MsgHolder;
  GENERIC_MSG_HEADER *Message;
  int i;
  int len;
  C_QUEUE Q;

/*
** Return if we are not connected
*/

  if (ClientSock->ConnectionStatus() == BAD_CONNECTION_STATUS) return NULL;

/*
** First look through already received messages (if any) and return the first
** one that matches the same Type.
*/

  len = MessageQueue.get_length();
  for (i=0; i<len; i++) {
    MsgHolder = (SpMsgHolder *)MessageQueue.pop_top();
    if (MsgHolder->CompareType(Type)) {
      Q.concat(&MessageQueue);
      MessageQueue = Q;
      Message = MsgHolder->GetHolderMessage();
      delete MsgHolder;
      return Message;
    }else{
      Q.push_bot(MsgHolder);
    }
  }

/*
** Now try to read incoming messages. Queue up the messages if they are not
** of the right Type. Return once one of the right type has been read, or if
** there are no more messages.
*/

  for (;;) {
    Message = ClientSock->getMsg();
    if (Message) {
      if (Message->handlerNum == Type) {
	return Message;
      }else{
	MsgHolder = new SpMsgHolder(Message);
	MessageQueue.push_bot(MsgHolder);
      }
    }else{
      return NULL;
    }
  }

}

/*
*****************************************************************************
* SpHostUser::SpeedesConnection() method must be called by SPEEDES to set up
* various Name List and Node information inside the host router. 
*****************************************************************************
*/

void SpHostUser::SpeedesConnection(int Nnodes, int Node, C_QUEUE *NameList) {
  SPEEDES_CONNECTION_MESSAGE *SpeedesConnectionMessage;
  int i;
  int len;
  int size;
  int total_size;
  char *buff;
  char *buff1;
  SpNameHandle *NameHandle;

/*
** Return if we are not connected
*/

  if (ClientSock->ConnectionStatus() == BAD_CONNECTION_STATUS) return;

/*
** Send the first message to connect to the HostRouter
*/

  size = 0;
  len = NameList->length();
  NameHandle = (SpNameHandle *)NameList->get_top();
  for (i=0; i<len; i++) {
    size += NameHandle->GetBufferSize();
    NameHandle = (SpNameHandle *)NameHandle->get_link();
  }

  total_size = size + sizeof(SPEEDES_CONNECTION_MESSAGE);

  if (!InputFileFlag) {

    buff = new char[total_size];

    SpeedesConnectionMessage = (SPEEDES_CONNECTION_MESSAGE *)buff;
    memset(buff, 0, total_size);
    SpeedesConnectionMessage->Init();
    SpeedesConnectionMessage->serviceType = HOST_ROUTER_ID;
    SpeedesConnectionMessage->group = GroupId;
    SpeedesConnectionMessage->handlerNum = SPEEDES_CONNECTION_MESSAGE_ID;
    SpeedesConnectionMessage->bytes = total_size;
    SpeedesConnectionMessage->Nnodes = Nnodes;
    SpeedesConnectionMessage->Node = Node;

    buff1 = (char *)&SpeedesConnectionMessage[1];
    NameHandle = (SpNameHandle *)NameList->get_top();
    for (i=0; i<len; i++) {
      buff1 = NameHandle->FillBuffer(buff1);
      NameHandle = (SpNameHandle *)NameHandle->get_link();
    }

    if (buff1 - buff - total_size) {
      cout << "Error in host user SpeedesConnection: Bad Buffer Size "
	   << (buff1-buff) << " should be " << total_size << endl;
    }

    SpeedesConnectionMessage->setBytes(total_size);
    ClientSock->sendMsg_blk(SpeedesConnectionMessage, total_size);

    delete [] buff;

  }
}

/*
*****************************************************************************
* SpHostUser::Flush() method sends a message to the host router and receives
* a message back. It basically pings the host router, but in doing so, it
* guarantees that the host router has read all of the messages previously
* sent.
*****************************************************************************
*/

void SpHostUser::Flush() {
  FLUSH_MESSAGE FlushMessage;
  SpMsgHolder *MsgHolder;

/*
** Return if we are not connected
*/

  if (ClientSock->ConnectionStatus() == BAD_CONNECTION_STATUS) return;

/*
** Return if there is nothing to flush
*/

  if (FlushFlag == 0) return;
  FlushFlag = 0;

/*
** Do the flush (i.e., send a message to the HostRouter and wait for a flush
** message to be returned;
*/

  if (!InputFileFlag) {
    FlushMessage.serviceType = HOST_ROUTER_ID;
    FlushMessage.group = GroupId;
    FlushMessage.handlerNum = FLUSH_MESSAGE_ID;
    FlushMessage.bytes = sizeof(FLUSH_MESSAGE);
    ClientSock->sendMsg_blk(&FlushMessage, sizeof(FLUSH_MESSAGE));

    for (;;) {
      FLUSH_MESSAGE *ReturnMessage = (FLUSH_MESSAGE *)ClientSock->getMsg_blk();
      if (ReturnMessage->handlerNum == FLUSH_MESSAGE_ID) {
	delete [] (char *)ReturnMessage;
	break;
      }else{
	MsgHolder = new SpMsgHolder(ReturnMessage);
	MessageQueue.push_bot(MsgHolder);
      }
    }

  }

}

/*
*****************************************************************************
* SpHostUser::QueryReply() method returns NameValueList data to the
* external query request.
*****************************************************************************
*/

void SpHostUser::QueryReply(SpNameValueList *NameValueList, int ExternalId) {
  QUERY_REPLY_MESSAGE *QueryReplyMessage;

  FlushFlag = 1;

  int Size = NameValueList->GetSizeOfBuffer();
  int TotalSize = Size + sizeof(QUERY_REPLY_MESSAGE);

  char *Buff = new char[TotalSize];
  QueryReplyMessage = (QUERY_REPLY_MESSAGE *)Buff;
  QueryReplyMessage->Init();
  NameValueList->PackBuffer((char *)&QueryReplyMessage[1], Size);

  QueryReplyMessage->serviceType = HOST_ROUTER_ID;
  QueryReplyMessage->group = GroupId;
  QueryReplyMessage->handlerNum = QUERY_REPLY_MESSAGE_ID;
  QueryReplyMessage->bytes = Size;

  QueryReplyMessage->ExternalId = ExternalId;

  ClientSock->sendMsg_blk(QueryReplyMessage, TotalSize);
  delete [] Buff;

}

/*
*****************************************************************************
* SpHostUser::TimeReply() method returns gvt to the external module making
* the Time() request.
*****************************************************************************
*/

void SpHostUser::TimeReply(double Time, int ExternalId) {
  TIME_REPLY_MESSAGE TimeReplyMessage;

  FlushFlag = 1;

  TimeReplyMessage.serviceType = HOST_ROUTER_ID;
  TimeReplyMessage.group = GroupId;
  TimeReplyMessage.handlerNum = TIME_REPLY_MESSAGE_ID;
  TimeReplyMessage.bytes = sizeof(TIME_REPLY_MESSAGE);
  TimeReplyMessage.ExternalId = ExternalId;

  TimeReplyMessage.Time = Time;
  ClientSock->sendMsg_blk(&TimeReplyMessage, sizeof(TIME_REPLY_MESSAGE));

}

/*
*****************************************************************************
* SpHostUser::MinMsgTimeReply() method returns the time value of the minimum
* outgoing messages to the external module making the MinMsgTime() request.
*****************************************************************************
*/

void SpHostUser::MinMsgTimeReply(SpSimTime Time, int ExternalId) {
  MIN_MSG_TIME_REPLY_MESSAGE MinMsgTimeReplyMessage;

  FlushFlag = 1;

  MinMsgTimeReplyMessage.serviceType = HOST_ROUTER_ID;
  MinMsgTimeReplyMessage.group = GroupId;
  MinMsgTimeReplyMessage.handlerNum = MIN_MSG_TIME_REPLY_ID;
  MinMsgTimeReplyMessage.bytes = sizeof(MIN_MSG_TIME_REPLY_MESSAGE);
  MinMsgTimeReplyMessage.ExternalId = ExternalId;

  MinMsgTimeReplyMessage.Time = Time;
  ClientSock->sendMsg_blk(&MinMsgTimeReplyMessage,
			  sizeof(MIN_MSG_TIME_REPLY_MESSAGE));

}


/*
*****************************************************************************
* SpHostUser::CommandReply() method returns time and status to external
* user making a command.
*****************************************************************************
*/

void SpHostUser::CommandReply(double Time, int ExternalId, 
			      int CommandStatus) {

  COMMAND_REPLY_MESSAGE CommandReplyMessage;

  FlushFlag = 1;

  CommandReplyMessage.serviceType = HOST_ROUTER_ID;
  CommandReplyMessage.group = GroupId;
  CommandReplyMessage.handlerNum = COMMAND_REPLY_MESSAGE_ID;

  CommandReplyMessage.ExternalId = ExternalId;
  CommandReplyMessage.Time = Time;
  CommandReplyMessage.CommandStatus = CommandStatus;
  CommandReplyMessage.bytes = sizeof(COMMAND_REPLY_MESSAGE);

  ClientSock->sendMsg_blk(&CommandReplyMessage, sizeof(COMMAND_REPLY_MESSAGE));

}

/*
*****************************************************************************
* SpHostUser::TimeAdvanceGrant() method returns gvt to the external module
* making the TimeAdvanceRequest() call.
*****************************************************************************
*/

void SpHostUser::TimeAdvanceGrant(SpSimTime Time, int ExternalId) {
  TIME_ADVANCE_GRANT_MESSAGE TimeMessage;

  FlushFlag = 1;

  TimeMessage.serviceType = HOST_ROUTER_ID;
  TimeMessage.group = GroupId;
  TimeMessage.handlerNum = TIME_ADVANCE_GRANT_MESSAGE_ID;
  TimeMessage.bytes = sizeof(TIME_ADVANCE_GRANT_MESSAGE);

  TimeMessage.ExternalId = ExternalId;
  TimeMessage.Time = Time;

  ClientSock->sendMsg_blk(&TimeMessage, sizeof(TIME_ADVANCE_GRANT_MESSAGE));

}


/*
*****************************************************************************
* SpHostUser::SendData() method sends a message to an external module
* with Id given by ExternalId.
*****************************************************************************
*/

void SpHostUser::SendData(SpSimTime Time, int ExternalId,
			  char *Data, int Bytes, int GlobalId) {

  SPEEDES_TO_EXTERNAL_MODULE_MESSAGE *SpEmMessage;
  int TotalBytes;

  if (ExternalId < 0) {
    cout << "Warning, trying to send data to a bad External Id" << endl;
    return;
  }

  TotalBytes = sizeof(SPEEDES_TO_EXTERNAL_MODULE_MESSAGE) + Bytes;
  SpEmMessage = (SPEEDES_TO_EXTERNAL_MODULE_MESSAGE *) new char[TotalBytes];
  SpEmMessage->Init();
  memcpy((char *)&SpEmMessage[1], Data, Bytes);

  FlushFlag = 1;

  SpEmMessage->serviceType = HOST_ROUTER_ID;
  SpEmMessage->group = GroupId;
  SpEmMessage->handlerNum = SPEEDES_TO_EXTERNAL_MODULE_MESSAGE_ID;
  SpEmMessage->bytes = Bytes;

  SpEmMessage->GlobalId = GlobalId;
  SpEmMessage->ExternalId = ExternalId;
  SpEmMessage->Time = Time;

  ClientSock->sendMsg_blk(SpEmMessage, TotalBytes);

  delete [] (char *)SpEmMessage;
}

/*
*****************************************************************************
* SpHostUser::RB_SendData() method sends a message to an external module
* with Id given by ExternalId.
*****************************************************************************
*/
void SpHostUser::RB_SendData(SpSimTime Time,
			     int ExternalId,
			     char *Data, 
			     int Bytes,
			     int GlobalId) {

  RBSendData(Time, ExternalId, Data, Bytes, GlobalId, this);

}

/*
*****************************************************************************
* SpHostUser::SendNamedData() method sends a named message to an external
* module with Id given by ExternalId. THIS NEEDS TO BE CHANGED TO USE A NEW
* MESSAGE TYPE IN THE HOST ROUTER.
*****************************************************************************
*/
void SpHostUser::SendNamedData(char *Name,
			       SpSimTime Time,
			       int ExternalId,
			       char *Data,
			       int DataBytes,
			       int GlobalId) {

    int len = strlen(Name)+1;
    char *buff = new char[len + DataBytes];
    memcpy(buff, Name, len);
    memcpy(&buff[len], Data, DataBytes);
    SendData(Time, ExternalId, buff, len+DataBytes, GlobalId);
    delete [] buff;

  }

/*
*****************************************************************************
* SpHostUser::RB_SendNamedData() method sends a named message to an external
* module with Id given by ExternalId. THIS NEEDS TO BE CHANGED TO USE A NEW
* MESSAGE TYPE IN THE HOST ROUTER.
*****************************************************************************
*/
void SpHostUser::RB_SendNamedData(char *Name,
				  SpSimTime Time,
				  int ExternalId,
				  char *Data, 
				  int DataBytes,
				  int GlobalId) {

    int len = strlen(Name)+1;
    char *buff = new char[len + DataBytes];
    memcpy(buff, Name, len);
    memcpy(&buff[len], Data, DataBytes);
    RBSendData(Time, ExternalId, buff, len+DataBytes, GlobalId, this);
    delete [] buff;

}


/*
*****************************************************************************
* SpHostUser::RB_SendNamedData() method sends a named message to an external
* module with Id given by ExternalId. THIS NEEDS TO BE CHANGED TO USE A NEW
* MESSAGE TYPE IN THE HOST ROUTER.
*****************************************************************************
*/
void SpHostUser::RB_SendSubscribedData(double Time,
				       char *Name,
				       char *Data,
				       int Bytes,
				       int GlobalId) {

    RBSendSubscribedData(Time, Name, Data, Bytes, GlobalId, this);
}


/*
*****************************************************************************
* SpHostUser::SendSubscribedData() method sends a message to external
* modules using the Name of the subscription.
*****************************************************************************
*/

void SpHostUser::SendSubscribedData(double Time, char *Name,
				    char *Data, int Bytes, int GlobalId) {
  SUBSCRIBED_DATA_MESSAGE *SpEmMessage;
  int TotalBytes;
  char *Buff;

  TotalBytes = sizeof(SUBSCRIBED_DATA_MESSAGE)
    + Bytes + strlen(Name) + 1;

  SpEmMessage = (SUBSCRIBED_DATA_MESSAGE *) new char[TotalBytes];
  SpEmMessage->Init();
  Buff = (char *)&SpEmMessage[1];
  strcpy(Buff, Name);
  Buff += strlen(Name) + 1;

  memcpy(Buff, Data, Bytes);

  FlushFlag = 1;

  SpEmMessage->serviceType = HOST_ROUTER_ID;
  SpEmMessage->group = GroupId;
  SpEmMessage->handlerNum = SUBSCRIBED_DATA_MESSAGE_ID;
  SpEmMessage->bytes = Bytes;

  SpEmMessage->GlobalId = GlobalId;
  SpEmMessage->Time = Time;

  SpEmMessage->bytes = Bytes;
  ClientSock->sendMsg_blk(SpEmMessage, TotalBytes);

  delete [] (char *)SpEmMessage;
}



/*
*****************************************************************************
* SpHostUser::GrantGoToTime() method returns gvt to the external module
* making the GoToTime() call.
*****************************************************************************
*/

void SpHostUser::GrantGoToTime(SpSimTime Time, int ExternalId) {
  GO_TO_TIME_GRANT_MESSAGE TimeMessage;

  FlushFlag = 1;

  TimeMessage.serviceType = HOST_ROUTER_ID;
  TimeMessage.group = GroupId;
  TimeMessage.handlerNum = GO_TO_TIME_GRANT_MESSAGE_ID;

  TimeMessage.ExternalId = ExternalId;
  TimeMessage.Time = Time;

  TimeMessage.bytes = sizeof(GO_TO_TIME_GRANT_MESSAGE);
  ClientSock->sendMsg_blk(&TimeMessage, sizeof(GO_TO_TIME_GRANT_MESSAGE));
}

  

void SpHostUser::ScheduleEventReply(int errorCode,
				    double time,
				    int externalId,
				    int cancelHandle) {

  SCHEDULE_EVENT_REPLY_MESSAGE msg;

  msg.Time = time;
  msg.ExternalId = externalId;
  msg.ErrorCode = errorCode;
  msg.CancelHandle = cancelHandle;
  msg.handlerNum = SCHEDULE_EVENT_REPLY_ID;
  msg.bytes = sizeof(SCHEDULE_EVENT_REPLY_ID);

  ClientSock->sendMsg_blk(&msg, sizeof(SCHEDULE_EVENT_REPLY_MESSAGE));
}


void SpHostUser::CancelReply(int externalId, int errorCode){

  CANCEL_EVENT_REPLY_MESSAGE msg;

  msg.serviceType = HOST_ROUTER_ID;
  msg.group = GroupId;
  msg.ExternalId = externalId;
  msg.ErrorCode = errorCode;
  msg.handlerNum = CANCEL_EVENT_REPLY_ID;
  msg.bytes = sizeof(CANCEL_EVENT_REPLY_MESSAGE);

  ClientSock->sendMsg_blk(&msg, sizeof(CANCEL_EVENT_REPLY_MESSAGE));

}

int SpHostUser::ConnectionStatus() {
  return ClientSock->ConnectionStatus();
}
