#ifndef SpHostUser_OBJECT
#define SpHostUser_OBJECT

#include "SpStdio.H"

#include "SpDataParser.H"
#include "queue.H"
#include "SpNameValueList.H"
#include "GoToTime.H"

class GENERIC_MSG_HEADER;
class CLIENT_SOCK;

/*************************************************************************
 * Class: SpHostUser
 *
 * Summary: The SpHostUser class is the internal SPEEDES interface
 *          to the Host Router which allows communication with the
 *          world outside of SPEEDES
 *
 * Description: From external modules, the only way to get information
 *              from inside of SPEEDES is to use the SpEmHostUser
 *              class which simply gets its information from the
 *              HostRouter (Contained within the SpeedesServer
 *              executable).  The way this information is sent to the
 *              Host Router from SPEEDES is through the use of the
 *              SpHostUser interface.  This interface allows one to
 *              get information out of SPEEDES as well as get
 *              information into SPEEDES.
 *
 *              This interface includes several generic GetMessage
 *              calls to get messages from the outside world as well
 *              as generic methods for sending data, query replies,
 *              subscribed data, time synchronization calls and other.
 *
 * Examples: There are several examples of how to use some of these
 *           calls in the demo BmdSim which communicates with the
 *           SpStateMgr (the OPSM which uses the SpEmHostUser to
 *           get data from SPEEDES).  There is an example of how to
 *           reply to a query message within the default Query event
 *           in src/engine/builtInEvents/E_SpQuery.C.
 *
 *           One overiding rule that you should always try to follow
 *           when sending data to the outside world is to only do this
 *           only in commit().  You may notice that the default query
 *           event in E_SpQuery.C sends its data in process().  This is
 *           because we can guarantee that query events will only be
 *           processed at GVT which means that they cannot be rolled
 *           back.  Please understand that this is an exception to the
 *           rule and that you should, in general, only send data to
 *           the outside world in commit().
 *
 *           Even following this rule may lead to unexpected results.
 *           The reason is that when you process() your event, the
 *           object may have one state but when you call commit(),
 *           that state may have changed.  The reason is that other
 *           events may "sneak" in between your process and commit()
 *           (commit() is called at GVT updates) and change the state
 *           of the object.  If you need to ensure this does not
 *           happen, you need to create data that is only modifiable
 *           by your event and use this data in commit() when you send
 *           your message out.
 *
 ************************************************************************/

class SpHostUser {
public:


/************************************************************************
 * Method: SpHostUser()
 * Summary: default constructor for a SpHostUser
 * Description: The default constructor searches the current directory
 *              for the file speedes.par.  It then goes to the section
 *              on the SpeedesServer and finds the appropriate machine
 *              name, port and group for the HostRouter and attempts
 *              to connect to that HostRouter.  If it cannot connect,
 *              it simply returns a NULL for that SpHostUser.
 * Return: A new SpHostUser
*************************************************************************/
  SpHostUser();


/************************************************************************
 * Method: SpHostUser(SpDataParser *)
 * Summary: Constructor of a SpHostUser from an already opened par file
 * Description: If you have already opened the par file from which you
 *              want to read the port, machine, and group number for
 *              your SPEEDES process, this is the constructor you want
 *              to use.
 * Return: A new SpHostUser
*************************************************************************/
  SpHostUser(SpDataParser *Parser //In: parser that contains the machine
	                           //    name, port and group number for
	                           //    the host router and this simulation
	       );


/************************************************************************
 * Method: SpHostUser(char *, int, int)
 * Summary: Constructor of a SpHostUser for a specific machine, port,
 *          and group
 * Description: If, for some reason, you want to specify exactly what
 *              machine, port and group you want to use for this
 *              simulation, feel free to do so.  But this is NOT THE
 *              BEST WAY TO DO THINGS!!!.  The prefered method is to
 *              use the default constructor and just read this data
 *              from the file speedes.par
 * Return: A new SpHostUser
*************************************************************************/
  SpHostUser(char *MachineName, //In: Machine name where the HostRouter
	                          //    (SpeedesServer) resides
	       int Port,          //In: Port of the Host Router
	       int Group          //In: Group ID for your simulation.
	       );


/************************************************************************
 * Method: SpHostUser(char *)
 * Summary: Construcs a SpHostUser using the data from a pre-recorded
 *          simulation.
 * Description: One can record all messages from the outside world and
 *              then pipe them back into a simulation using this call.
 *              This can be very useful for debugging your
 *              applications
 * Return: A new SpHostUser
*************************************************************************/
  SpHostUser(char *FileName //In: File name of the pre-recorded simulation
	       );


/************************************************************************
 * Method: ~SpHostUser()
 * Summary: destructor of a SpHostUser
 * Description: This is an internal call and should not be used by end
 *              usersdestroys a SpHostUser and disconnects from the
 *              HOST_ROUTER.  Users should never make this call.
 *              Presently, the destruction of a SpHostUser does not
 *              remove any barriers that external modules raised.
 *              This could lead to locking SPEEDES since nothing would
 *              then ever remove those barrers.
 * Return: 
*************************************************************************/
  ~SpHostUser();



/************************************************************************
 * Method: ReadMessages()
 * Summary: Attempts to read all messages from the HostRouter and queue them
 * Description: 
 * Return: void
*************************************************************************/
  void ReadMessages();


/************************************************************************
 * Method: GetMessage()
 * Summary: returns a message from the host router if there is one,
 *          otherwise NULL.
 * Description: This method first checks to see if any messages have
 *              been queued up from a prior ReadMessages() call.  If
 *              so, it just grabs the first one it can find and
 *              returns it.  If it does not have any queued messages,
 *              it attempts to read one from the Host Router and
 *              return it.  If it still cannot find any messages, it
 *              returns null
 * Return: a char * message or NULL if none are available.
*************************************************************************/
  GENERIC_MSG_HEADER *GetMessage();


/************************************************************************
 * Method: GetMessage(int)
 * Summary: Attempts to return a message of a certain type.  NULL otherwise
 * Description: If you are waiting for a message of a certain type,
 *              this call can be used to try and receive that message.
 *              It will first look to see if a message of type Type
 *              has been previously received by a ReadMessages call
 *              and return that message if possible.  If not, it will
 *              try to receive and queue messages until it gets one of
 *              the appropriate type.  If it cannot find one, then it
 *              returns NULL
 * Return: 
*************************************************************************/
  GENERIC_MSG_HEADER *GetMessage(int Type //In: the type of message you wish
				          //    to return.
				 );



/************************************************************************
 * Method: SpeedesConnection(int, int, C_QUEUE*)
 * Summary: Sets up various Name list and Node information within the
 *          Host Router.
 * Description: This call is used by SPEEDES internally.  It should
 *              never be made by users.  This method is called by each
 *              node after all the objects have been created to tell
 *              the HostRouter what objects are on this node and what
 *              their names are.
 * Return: 
*************************************************************************/
  void SpeedesConnection(int Nnodes,        //In: Total number of nodes
			 int Node,          //In: Node number of this box
			 C_QUEUE *NameList  //In: A list of all the objects
			                    //    on this node and their names
			 );



/************************************************************************
 * Method: Flush()
 * Summary: Sends a flush message to the Host Router
 * Description: When one want to make sure that the Host Router has
 *              received any previously sent data, one can issue a
 *              Flush() call which acts similar to a ping.  It sends a
 *              message to the Host Router and then waits for the
 *              return.  This assumes that TCP/IP (or a similar
 *              transport mechanism) is being using that guarantees in
 *              order delivery of messages.
 * Return: void
*************************************************************************/
  void Flush();



/************************************************************************
 * Method: QueryReply(SpNameValueList*, int)
 * Summary: Sends the SpNameValueList in response to a query message
 * Description: Upon receiving a query, one can either use the built
 *              in query response (found in src/engine/builtInEvents/E_SpQuery.C) or
 *              one can write his/her own query response.  This
 *              response can contain anything you wish in the
 *              NameValueList.  An example of how to fill out the
 *              NameValueList can be found in src/speedes/SpSimObj.C.
 *              Take a look at SpSimObj::query() to see how one can
 *              filll out this list.
 * Return: void
*************************************************************************/
  void QueryReply(SpNameValueList *NameValueList, //In: The information to be
		                                  //    sent in response to a
		                                  //    query event.
		  int ExternalId                  //In: ID of the external
		                                  //    module that scheduled
		                                  //    query.
		  );



/************************************************************************
 * Method: CommandReply(double, int, int)
 * Summary: returns the time and status to an external user making a command
 * Description: This is an internal SPEEDES call.  SPEEDES
 *              automatically calls this method in response to any
 *              command that was received from an external module.  It
 *              replies to let it know what time the command was
 *              processed as well as a status of the command.  This
 *              status is, essentially, whether or not SPEEDES was able
 *              to find a matching command or not.
 * Return: void
*************************************************************************/
  void CommandReply(double Time,   //In: Time at which command was processed
		    int ExternalId,//In: ID of module that scheduled command
		    int CommandStatus     //In: Status of the command
		    );



/************************************************************************
 * Method: TimeReply(double, int)
 * Summary: Sends the time to an external module.
 * Description: This is another internal SPEEDES method call.
 *              Whenever SPEEDES receives a request for its time, it
 *              does a GVT update and then sends this message out with
 *              the correct simulation time.
 * Return: void
*************************************************************************/
  void TimeReply(double Time,   //In: GVT of the Simulation
		 int ExternalId //In: ID of module requesting GVT
		 );


/************************************************************************
 * Method: MinMsgTimeReply(double, int)
 * Summary: Sends the time to an external module.
 * Description: This is another internal SPEEDES method call.
 *              Whenever SPEEDES receives a request for its MinMsgtime, it
 *              does a GVT update and then sends this message out with
 *              the correct MinMsgTime value.
 * Return: void
*************************************************************************/
  void MinMsgTimeReply(SpSimTime Time,   //In: MinMsgTime of the Simulation
		       int ExternalId //In: ID of module requesting GVT
		       );


/************************************************************************
 * Method: TimeAdvanceGrant(SpSimTime, int)
 * Summary: Returns GVT to an external module while setting a barrier
 * Description: Another internal SPEEDES method call for the Host
 *              Router If an external module makes a
 *              TimeAdvanceRequest, it makes this request with a given
 *              timelag.  SPEEDES then creates a barrier at
 *              GVT+timelag, returns GVT in this message and, upon
 *              reaching the barrier, blocks until or unless another
 *              TimeAdvanceRequest message was received or the
 *              external module has disconnected.
 * Return: void
*************************************************************************/
  void TimeAdvanceGrant(SpSimTime Time,   //In: GVT of the simulation
			int ExternalId    //In: ID of module requesting GVT
			);



/************************************************************************
 * Method: GoToTimeGrant(SpSimTime, int)
 * Summary: Returns GVT to an external module while setting a barrier
 * Description: Another internal SPEEDES method call for the Host
 *              Router If an external module makes a
 *              TimeAdvanceRequest, it makes this request with a given
 *              timelag.  SPEEDES then creates a barrier at
 *              GVT+timelag, returns GVT in this message and, upon
 *              reaching the barrier, blocks until or unless another
 *              TimeAdvanceRequest message was received or the
 *              external module has disconnected.
 * Return: void
*************************************************************************/
  void GrantGoToTime(SpSimTime Time,   //In: GVT of the simulation
		     int ExternalId    //In: ID of module requesting GVT
		     );



/************************************************************************
 * Method: SendData(SpSimTime, int, char*, int, int)
 * Summary: Sends data to an external module
 * Description: This method sends data to an external module.  The
 *              external module can then process it in any way it
 *              desires.  You should be certain that you only make
 *              these calls when you are committing the event because,
 *              otherwise, the event could be rolled back and this
 *              would lead to multiple sendings of possibly
 *              conflicting information.  We hope to create a roll
 *              backabe SpHostUser that will allow you to send the
 *              data in process() (similar to how you can do a RB_COUT
 *              instead of cout) but this is not available yet.
 *
 *              Warning, don't use this with the OPSM (SpStateMgr).
 *              Only useSendSubscribedData to send data back to the
 *              OSPM Also note that your object state data may change
 *              between process() and commit().  Other events can
 *              change the state of the object which could lead to
 *              different data being sent out to the outside world
 *              than you anticipated unless the Data is constructed
 *              during process.
 * Return: 
*************************************************************************/
  void SendData(SpSimTime Time,   //In: Time at which data was created.
		int ExternalId,   //In: ID of module to receive this data
		char *Data,       //In: Actual data as a char array
		int Bytes,        //In: Size of data in bytes
		int GlobalId      //In: Global ID of object creating the data
		);

  /* Roll backable version of SendData() */
  void RB_SendData(SpSimTime Time,
		   int ExternalId,
		   char *Data,
		   int DataBytes,
		   int GlobalId);

/*
** Send named data (NEEDS TO BE CHANGED)
*/

  void SendNamedData(char *Name,
		     SpSimTime Time,
		     int ExternalId,
		     char *Data,
		     int DataBytes,
		     int GlobalId);

  /* Roll backable version of SendNamedData() */
  void RB_SendNamedData(char *Name,
			SpSimTime Time,
			int ExternalId,
			char *Data,
			int DataBytes,
			int GlobalId);

/************************************************************************
 * Method: SendSubscribedData(double, char*, char*, int, int)
 * Summary: Sends data with a special subscription name to an external module
 * Description: This is the proper way to send data back to an OPSM.
 *              The OPSM should subscribe to nameed data with the name
 *              Name for this object.  As with any other communication
 *              with the outside world, this call should only be made
 *              during commit. Also note that your object state data
 *              may change between process() and commit().  Other
 *              events can change the state of the object which could
 *              lead to different data being sent out to the outside
 *              world than you anticipated unless the Data is
 *              constructed during process.
 * Return: void
*************************************************************************/
  void SendSubscribedData(double Time,  //In: Time at which data was created.
			  char *Name,   //In: Name of the subcription
			  char *Data,   //In: Actual data as a char array
			  int Bytes,    //In: Size of data in bytes
			  int GlobalId  //In: Global ID of object creating
			                //    the data
			  );


  void RB_SendSubscribedData(double Time, //In: Time at which data was created.
			     char *Name,  //In: Name of the subcription
			     char *Data,  //In: Actual data as a char array
			     int Bytes,   //In: Size of data in bytes
			     int GlobalId //In: Global ID of object creating
			                  //    the data
			     );




/************************************************************************
 * Method: ScheduleTimeAdvanceGrant(double, int)
 * Summary: Schedules a time advance request.
 * Description: This is an internal method used for handing time advance
 *              requests from external modules.
 * Return: void
*************************************************************************/
  void ScheduleTimeAdvanceGrant(double TimeLag, //In: Timelag for module
				                //    requesting time advance
				int ExternalId  //In: ID of the module
				                //    requesting time advance
				) {
    C_ITEM *item = new C_ITEM();
    item->set_time_tag(TimeLag);
    item->set_id(ExternalId);
    TimeAdvanceGrantQueue.push_bot(item);
  }



/************************************************************************
 * Method: UnScheduleTimeAdvanceGrant(int)
 * Summary: Unschedule a time advance grant for an external module
 * Description: This is an internal method used for removing a
 *              scheduled time advance grant when an external module
 *              removes barriers or exits.
 * Return: 
*************************************************************************/
  void UnScheduleTimeAdvanceGrant(int ExternalId //In: ID of the module
				                //     requesting time advance
				  ) {
    C_ITEM *item = TimeAdvanceGrantQueue.deque(ExternalId);
    if (item) delete item;
  }



/************************************************************************
 * Method: GetScheduledTimeAdvanceGrant(double&, int&)
 * Summary: gets value of timelag and external ID for the next time
 *          advance grant.
 * Description: This is another internal method call used to iterate
 *              through all scheduled time advance requests so that we
 *              can send out all the time advance's that have been
 *              requested.
 * Return: Status of the operation.  1 if we had anything scheduled
 *         and 0 otherwise
*************************************************************************/
  int GetScheduledTimeAdvanceGrant(double &TimeLag, //Out: Time lag of module
				                    //     requesting time
				                    //     advance
				   int &ExternalId  //Out: ID of the module
				                    //     requesting time
				                    //     advance
				   ) {
    int FoundStatus = 0;
    C_ITEM *item = TimeAdvanceGrantQueue.pop_top();
    if (item) {
      TimeLag = item->get_time_tag();
      ExternalId = item->get_id();
      delete item;
      FoundStatus = 1;
    }
    return FoundStatus;
  }



  int GetNumGoToTimeGrants(){
    return GoToTimeGrantQueue.get_length();
  }
  
  void ScheduleGoToTimeGrant(double BlockTime,
			     double LagTime,
			     int ExternalId) {
    GO_TO_TIME *item = new GO_TO_TIME;
    item->SetGoToTime(BlockTime);
    item->set_time_tag(LagTime);
    item->set_id(ExternalId);
    GoToTimeGrantQueue.push_bot(item);
  }
  
  void UnScheduleGoToTimeGrant(int ExternalId) {
    GO_TO_TIME *item = (GO_TO_TIME *)GoToTimeGrantQueue.deque(ExternalId);
    if (item) delete item;
  }
  
  int GetScheduledGoToTimeGrant(double& BlockTime,
				double& LagTime,
				int& ExternalId){
    int FoundStatus = 0;
    GO_TO_TIME *item = (GO_TO_TIME *) GoToTimeGrantQueue.pop_top();
    if (item != NULL) {
      BlockTime = item->GetGoToTime();
      LagTime = item->get_time_tag();
      ExternalId = item->get_id();
      delete item;
      FoundStatus = 1;
    }
    return FoundStatus;
  }

  int FindScheduledGoToTimeGrant(double& BlockTime,
				 double& LagTime,
				 int& ExternalId //InOut
				 ){
    int FoundStatus = 0;
    GO_TO_TIME *item = (GO_TO_TIME *) GoToTimeGrantQueue.get_item(ExternalId);
    if (item != NULL){
      BlockTime = item->GetGoToTime();
      LagTime = item->get_time_tag();
      ExternalId = item->get_id();
      FoundStatus = 1;
    }
    return FoundStatus;
  }

  void ScheduleEventReply(int errorCode,
			  double time,
			  int ExternalId,
			  int handle);


  void CancelReply(int externalId,
		   int returnCode);


  int ConnectionStatus();


protected:
private:
  CLIENT_SOCK *ClientSock;
  FILE *fp_output;
  FILE *fp_input;
  int InputFileFlag;
  int OutputFileFlag;
  int GroupId;
  int FlushFlag;

  C_QUEUE MessageQueue;
  C_QUEUE TimeAdvanceGrantQueue;
  C_QUEUE GoToTimeGrantQueue;
  double ExternalGVT;

};

typedef SpHostUser SP_HOST_USER;

#endif
