// SpReportStatus.C file supports the status reporting that is done at the
// end of a SPEEDES simulation.

#include "SpReportStatus.H"
#include "SpBinaryTree.H"
#include "SpComm.H"

char **SpReportStatusRecord::ReportStatusString = NULL;
SpBinaryTree *GlobalStatusTree = NULL;

SpReportStatusRecord::SpReportStatusRecord() {
  TestName = NULL;
  for (int i=0; i<SP_NUM_REPORT_STATUS; i++) StatusCounter[i] = 0;
  if (ReportStatusString == NULL) {
    ReportStatusString = new char *[SP_NUM_REPORT_STATUS];
    ReportStatusString[SP_REPORT_STATUS_PASSED] = "Passed";
    ReportStatusString[SP_REPORT_STATUS_INCONCLUSIVE] = "Inconclusive";
    ReportStatusString[SP_REPORT_STATUS_FAILED] = "Failed";
  }
}

SpReportStatusRecord::SpReportStatusRecord(char *testName) {
  TestName = strdup(testName);
  for (int i=0; i<SP_NUM_REPORT_STATUS; i++) StatusCounter[i] = 0;
  if (ReportStatusString == NULL) {
    ReportStatusString = new char *[SP_NUM_REPORT_STATUS];
    ReportStatusString[SP_REPORT_STATUS_PASSED] = "Passed";
    ReportStatusString[SP_REPORT_STATUS_INCONCLUSIVE] = "Inconclusive";
    ReportStatusString[SP_REPORT_STATUS_FAILED] = "Failed";
  }
}

//----------------------------------------------------------------------------
// SpReportStatus() is called by applications or regression tests during
// termination to report status (passed, inconclusive, or failed). SPEEDES
// automatically collects the information across nodes by calling the
// SpGlobalReportStatus() function and then prints out the results on node 0
// using the SpPrintStatus() method.
//----------------------------------------------------------------------------
void SpReportStatus(char *testName, SP_REPORT_STATUS status) {
  SpReportStatusRecord *reportStatusRecord;

  if (GlobalStatusTree == NULL) GlobalStatusTree = new SpBinaryTree();

  reportStatusRecord = (SpReportStatusRecord *)
    GlobalStatusTree->Find(testName);

  if (reportStatusRecord == NULL) {
    reportStatusRecord = new SpReportStatusRecord(testName);
    GlobalStatusTree->Insert(reportStatusRecord, testName);
  }

  reportStatusRecord->UpdateStatusCounter(status);

}

//----------------------------------------------------------------------------
// SpGlobalReportStatus() is called by.SPEEDES to automatically collect
// status information across nodes.
//----------------------------------------------------------------------------
void SpGlobalReportStatus() {
  SpReportStatusRecord *reportStatusRecord;
  int buffSize;
  char *buff;
  char *b;

  if (GlobalStatusTree == NULL) {
    GlobalStatusTree = new SpBinaryTree();
  }

  if (SpComm_GlobalSum(GlobalStatusTree->GetNumElements()) == 0) {
    return;
  }

/*
** If we are not node 0, then form a message to send to node 0 that contains
** all of my local status reports and then return. Node zero collects all of
** the information.
*/

  if (SpComm_GetNodeId() > 0) {

/*
** First, deterimine the buffer size
*/

    buffSize = 0;

    reportStatusRecord = (SpReportStatusRecord *)
      GlobalStatusTree->GetFirstElement();

    while (reportStatusRecord) {

      buffSize += reportStatusRecord->GetBuffSize();

      reportStatusRecord = (SpReportStatusRecord *)
	GlobalStatusTree->GetNextElement();
    }

/*
** Create the buffer
*/

    if (buffSize) {
      buff = new char[buffSize];
    }else{
      buff = NULL;
    }

/*
** Now pack the buffer
*/

    b = buff;

    reportStatusRecord = (SpReportStatusRecord *)
      GlobalStatusTree->GetFirstElement();

    while (reportStatusRecord) {

      b = reportStatusRecord->PackBuff(b);

      reportStatusRecord = (SpReportStatusRecord *)
	GlobalStatusTree->GetNextElement();
    }

    if (buff) {
      SpComm_CoordinatedSend(0, buffSize, buff);
      delete [] buff;
    }

    while ((buff = SpComm_CoordinatedReceive(buffSize))) {
      cout << "Error, SpReportStatus should only receive buffs on node 0"
	   << endl;
      delete [] buff;
    }

    return;

  }

/*
** Node 0 collects all the buffers from the other nodes and updates its
** records.
*/

  while ((buff = SpComm_CoordinatedReceive(buffSize))) {
    b = buff;
    while (b-buff != buffSize) {
      reportStatusRecord = (SpReportStatusRecord *)GlobalStatusTree->Find(b);
      if (reportStatusRecord == NULL) {
	reportStatusRecord = new SpReportStatusRecord(b);
	GlobalStatusTree->Insert(reportStatusRecord, b);
      }
      b = reportStatusRecord->UnpackBuff(b);
    }
    delete [] buff;
  }

}


//----------------------------------------------------------------------------
// SpPrintReportStatus() is called by.SPEEDES on node 0 to print out the
// final status of the simulation..
//----------------------------------------------------------------------------
void SpPrintReportStatus(ostream &out) {
  int i, j;
  int maxLen;
  SpReportStatusRecord *reportStatusRecord;

  if (SpComm_GetNodeId() != 0) return;
  if (GlobalStatusTree->GetNumElements() == 0) return;

/*
** Find the maximum TestName string length for alignment
*/

  reportStatusRecord = (SpReportStatusRecord *)
    GlobalStatusTree->GetFirstElement();

  maxLen = 0;
  while (reportStatusRecord) {

    i = strlen(reportStatusRecord->GetTestName());
    if (maxLen < i ) maxLen = i;

    reportStatusRecord = (SpReportStatusRecord *)
      GlobalStatusTree->GetNextElement();
  }

/*
** Print out the first line of the report status
*/

  out << endl;
  out << "Begin Report Status ------------------------------------------------"
      << endl;

/*
** Loop through each record and print them out
**/

  reportStatusRecord = (SpReportStatusRecord *)
    GlobalStatusTree->GetFirstElement();

  while (reportStatusRecord) {

    out << "  " << reportStatusRecord->GetTestName() << " ";

    j = maxLen - strlen(reportStatusRecord->GetTestName()) + 3;
    for (i=0; i<j; i++) out << ".";
    out << " " << reportStatusRecord->GetStatusString() << " ";

    out << "(P="
	<< reportStatusRecord->GetStatusCounter(SP_REPORT_STATUS_PASSED)
	<< ", I="
	<< reportStatusRecord->GetStatusCounter(SP_REPORT_STATUS_INCONCLUSIVE)
	<< ", F="
	<< reportStatusRecord->GetStatusCounter(SP_REPORT_STATUS_FAILED)
	<< ")"
	<< endl;

    reportStatusRecord = (SpReportStatusRecord *)
      GlobalStatusTree->GetNextElement();
  }

/*
** Print out the last line of the report status
*/

  out << "End of Report Status -----------------------------------------------"
      << endl;
  out << endl;

}
