// SpScheduleEvent.H header file that defines the overloaded schedule
// functions. These functions are not inherited from events, but call
// the schedule methods for the current event object (which is where
// the messages are stored).

#ifndef SCHEDULE_FUNCS_H
#define SCHEDULE_FUNCS_H

#include "SpExecute.H"
#include "SpMsg.H"
#include "SpEventHandlerTypes.H"
#include "E_SpEventHandler.H"
// #include "SpEventHandler.H"
#include "SpEvent.H"
class SpString80;
#include "SpMasterSimObjMgr.H"
#include "SpSimObjMgr.H"
#include "SpEventTypes.H"
#include "SpAntiMsg.H"
#include "M_SpProc.H"
#include "SpObjHandle.H"

#include "SpMemory.H"
#include "SpParmSet.H"

#include "O_SpBuiltIn.H"

extern SpEvent *CurrentEvent;
extern SpSimObj *CurrentSimobj;
extern SpEventTypes *EventTypes;
extern SpHostUser *HostUser;
extern SpMasterSimObjMgr *MasterSimObjMgr;

/*
** Some common functions that access global data
*/

static inline SpEvent *GetEvent() {return CurrentEvent;}

static inline SpMasterSimObjMgr *GetMasterSimObjMgr() {return MasterSimObjMgr;}

static inline SpSimObjMgr *GetSimObjMgr(int mgrId) {
  return (SpSimObjMgr *)GetMasterSimObjMgr()->get_objman(mgrId);
}

static inline SpSimObjMgr *GetSimObjMgr(char *mgrName) {
  int mgrId = GetEvent()->object_type(mgrName);
  return GetSimObjMgr(mgrId);
}

static inline SpSimObj *GetSimObj() {return CurrentSimobj;}

static inline SpSimObj *GetSimObj(SpObjHandle &objHandle) {
  SpSimObjMgr *objMgr = GetSimObjMgr(objHandle.ObjManId);
  return objMgr->get_obj(objHandle.SimObjId);
}

static inline SpSimTime &GetTime() {return GetEvent()->get_time_tag();}
static inline SpSimTime &GetSimTime() {return GetEvent()->get_time_tag();}
static inline SpMsg *GetMsg() {return GetEvent()->get_MESSAGE();}
static inline char *GetData() {return GetEvent()->get_DATA();}
static inline int GetDataBytes() {return GetMsg()->bytes;}
static inline int GetNode() {return GetSimObj()->get_NODE();}
static inline int GetNumNodes() {return GetSimObj()->get_N_NODES();}
static inline int GetMgrId() {return GetSimObj()->get_OBJECT_TYPE();}
static inline int GetLocalId() {return GetSimObj()->get_LOCAL_ID();}
static inline int GetUniqueId() {return GetSimObj()->get_GLOBAL_ID();}
static inline int GetGlobalId() {return GetSimObj()->get_GLOBAL_ID();}
static inline int GetEnumId() {return GetSimObj()->get_NUM_ID();}
static inline int GetEventType() {return GetEvent()->get_EVENT_TYPE();}
static inline SpHostUser *GetSpHostUser() {return HostUser;}

static inline SpObjHandle GetObjHandle() {
  SpObjHandle objHandle;
  objHandle.NodeId = GetNode();
  objHandle.ObjManId = GetMgrId();
  objHandle.SimObjId = GetLocalId();
  return objHandle;
}

static inline SpObjHandle GetObjHandle(int mgrId, int enumId)
{
   return ((O_SpBuiltIn *) SpGetSimObjMgr(mgrId))->GetObjHandle(enumId);
}

static inline SpObjHandle GetObjHandle(char *mgrName, int enumId)
{
   return ((O_SpBuiltIn *) SpGetSimObjMgr(mgrName))->GetObjHandle(enumId);
}

/******************************************************************************
* Function: ScheduleEvent
* Summary: Schedules an event to be processed for a simobj at a specified
*          time. Events must never be scheduled in the past of the current
*          event.
* Description: This function can be called from anywhere within
*              SPEEDES to schedule an event. An event must have a time
*              tag, event type, simulation object type, simulation
*              object local Id, and node number to uniquely identify
*              the event and the simobj that it acts on. Optional
*              arguments are the message header (users must set
*              specified data members in the descendent class of
*              SpMsg (note that descendents MUST NOT have
*              multiple inheritance) before making the schedule call),
*              and a variable data buffer followed by the number of
*              bytes. If the variable data buffer is not provided, the
*              SPEEDES internal free list mechanisms are used to
*              reduce overheads.  SPEEDES will not delete header or
*              data.
* Return: A cancel handle for the event which can be later used to retract
*         the event before it is processed. Applications are not required to
*         use this returned value.
******************************************************************************/
SpCancelHandle ScheduleEvent(
			     const SpSimTime &time, // IN: Logical time
			     int eventType,	     // IN: Event type Id
			     int simobjType,      // IN: Simulation object
			     //     type
			     int simobjLocalId,   // IN: Simulation object
			     //     local Id
			     int simobjNode,      // IN: Node that the
			     //     simulation object
			     //     is on
			     const SpMsg *header = NULL, // IN: Optional
			     //     message for
			     //     scheduling the
			     //     event
			     const char *data = NULL, // IN: Variable length 
			     //     data that gets
			     //     passed with the
			     //     event (optional)
			     int dataBytes = 0    // IN: Number of bytes for
			     //     variable length
                                                     //     data
			     );

/******************************************************************************
* Function: ScheduleEvent
* Summary: Schedules an event to be processed for a simobj at a specified
*          time. Events must never be scheduled in the past of the current
*          event. Here, users pass in a reference to the message header.
* Description: This function can be called from anywhere within
*              SPEEDES to schedule an event. An event must have a time
*              tag, event type, simulation object type, simulation
*              object local Id, and node number to uniquely identify
*              the event and the simobj that it acts on. Optional
*              arguments are the message header (users must set
*              specified data members in the descendent class of
*              SpMsg (note that descendents MUST NOT have
*              multiple inheritance) before making the schedule call),
*              and a variable data buffer followed by the number of
*              bytes. If the variable data buffer is not provided, the
*              SPEEDES internal free list mechanisms are used to
*              reduce overheads.  SPEEDES will not delete header or
*              data.
* Return: A cancel handle for the event which can be later used to retract
*         the event before it is processed. Applications are not required to
*         use this returned value.
******************************************************************************/
static inline SpCancelHandle ScheduleEvent(
				const SpSimTime &time, // IN: Logical time
				int eventType,	     // IN: Event type Id
				int simobjType,      // IN: Simulation object
                                                     //     type
				int simobjLocalId,   // IN: Simulation object
                                                     //     local Id
				int simobjNode,      // IN: Node that the
                                                     //     simulation object
                                                     //     is on
				const SpMsg &header, // IN: 
                                                         //     message for
                                                         //     scheduling the
                                                         //     event
				const char *data = NULL,   // IN: Variable length 
                                                     //     data that gets
                                                     //     passed with the
                                                     //     event (optional)
				int dataBytes = 0    // IN: Number of bytes for
                                                     //     variable length
                                                     //     data
				){
  return ScheduleEvent(time, eventType, simobjType, simobjLocalId,
		       simobjNode, &header, data, dataBytes);
}

/******************************************************************************
* Function: ScheduleEvent
* Summary: Schedules an event to be processed for a simobj at a specified
*          time. Events must never be scheduled in the past of the current
*          event. Here, users pass in an object handler reference.
* Description: This function can be called from anywhere within
*              SPEEDES to schedule an event. An event must have a time
*              tag, event type simulation object type, simulation
*              object local Id, and node number to uniquely identify
*              the event and the simobj that it acts on. Optional
*              arguments are the message header (users must set
*              specified data members in the descendent class of
*              SpMsg (note that descendents MUST NOT have
*              multiple inheritance) before making the schedule call),
*              and a variable data buffer followed by the number of
*              bytes. If the variable data buffer is not provided, the
*              SPEEDES internal free list mechanisms are used to
*              reduce overheads.  SPEEDES will not delete header or
*              data.
* Return: A cancel handle for the event which can be later used to retract
*         the event before it is processed. Applications are not required to
*         use this returned value.
******************************************************************************/
SpCancelHandle ScheduleEvent(
			     const SpSimTime &time, // IN: Logical time
			     int eventType,	     // IN: Event type Id
			     const SpObjHandle &simobj, // IN: Simulation
			     //     object handle
			     const SpMsg *header = NULL, // IN: Optional
			     //     message for
			     //     scheduling the
			     //     event
			     const char *data = NULL, // IN: Variable length 
			     //     data that gets
			     //     passed with the
			     //     event (optional)
			     int dataBytes = 0    // IN: Number of bytes for
			     //     variable length
			     //     data
			     );

static inline SpCancelHandle ScheduleEvent(
				const SpSimTime &time, // IN: Logical time
				int eventType,	     // IN: Event type Id
				const SpObjHandle &simobj, // IN: Simulation
				                         //     object handle
				const SpMsg &header, // IN: Optional
                                                         //     message for
                                                         //     scheduling the
                                                         //     event
				const char *data = NULL, // IN: Variable length 
                                                     //     data that gets
                                                     //     passed with the
                                                     //     event (optional)
				int dataBytes = 0    // IN: Number of bytes for
                                                     //     variable length
                                                     //     data
				){
  return ScheduleEvent(time, eventType, simobj, &header, data, dataBytes);
}

/******************************************************************************
* Function: ReScheduleEvent
* Summary: Re-schedules the same event to be processed for the same simobj at
*          a future time.
* Description: This function can be called from anywhere within SPEEDES to
*              reschedule the same event for the same simobj for processing
*              at a later time.
* Return: A cancel handle for the event which can be later used to retract
*         the event before it is processed. Applications are not required to
*         use this returned value.
******************************************************************************/
static inline SpCancelHandle ReScheduleEvent(
				  const SpSimTime &time, // IN: Logical time
				  const SpMsg *header = NULL,// IN: Optional
                                                          //     message for
                                                          //     scheduling the
                                                          //     event
				  const char *data = NULL,   // IN: Variable length 
                                                       //     data that gets
                                                       //     passed with the
                                                       //     event (optional)
				  int dataBytes = 0    // IN: Number of bytes
                                                       //     for variable
                                                       //     length data
				  ){
  return ScheduleEvent(time,
		       CurrentEvent->get_EVENT_TYPE(),
		       CurrentEvent->get_OBJECT_TYPE(),
		       CurrentEvent->get_LOCAL_ID(),
		       CurrentEvent->get_NODE(),
		       header,
		       data,
		       dataBytes);
}

/******************************************************************************
* Function: ReScheduleEvent
* Summary: Re-schedules the same event to be processed for the same simobj at
*          a future time.
* Description: This function can be called from anywhere within SPEEDES to
*              reschedule the same event for the same simobj for processing
*              at a later time.
* Return: A cancel handle for the event which can be later used to retract
*         the event before it is processed. Applications are not required to
*         use this returned value.
******************************************************************************/
static inline SpCancelHandle ReScheduleEvent(
				  const SpSimTime &time, // IN: Logical time
				  const SpMsg &header,// IN: Optional
                                                          //     message for
                                                          //     scheduling the
                                                          //     event
				  const char *data = NULL,   // IN: Variable length 
                                                       //     data that gets
                                                       //     passed with the
                                                       //     event (optional)
				  int dataBytes = 0    // IN: Number of bytes
                                                       //     for variable
                                                       //     length data
				  ){
  return ScheduleEvent(time,
		       CurrentEvent->get_EVENT_TYPE(),
		       CurrentEvent->get_OBJECT_TYPE(),
		       CurrentEvent->get_LOCAL_ID(),
		       CurrentEvent->get_NODE(),
		       header,
		       data,
		       dataBytes);

}



/******************************************************************************
* Function: ScheduleProcess
* Summary: Schedules a process to be started for a simobj at a specified
*          time. Processes must never be scheduled in the past of the current
*          event or process.
* Description: This function can be called from anywhere within SPEEDES to
*              schedule a process. An proces must have a time tag, event type
*              simulation object type, simulation object local Id, and node
*              number to uniquely identify the process and the simobj that it
*              acts on. Optional arguments are the message header (users must
*              set specified data members before making the schedule call), and
*              a variable data buffer followed by the number of bytes. If the
*              variable data buffer is not provided, the SPEEDES internal
*              free list mechanisms are used to reduce overheads.
* Return: A cancel handle for the event which can be later used to retract
*         the event before it is processed. Applications are not required to
*         use this returned value.
******************************************************************************/
SpCancelHandle ScheduleProcess(
			       const SpSimTime &time, // IN: Logical time
			       int eventType,	     // IN: Event type Id
			       int simobjType,      // IN: Simulation object
				//     type
			       int simobjLocalId,   // IN: Simulation object
			       //     local Id
			       int simobjNode,      // IN: Node that the
			       //     simulation object
			       //     is on
			       const M_SpProc *message = NULL, // IN: Optional
			       //     message for
			       //     scheduling
			       //     the event
			       const char *data = NULL,   // IN: Variable length 
			       //     data that gets
			       //     passed with the
			       //     event (optional)
			       int dataBytes = 0    // IN: Number of bytes for
			       //     variable length
			       //     data
				);

/******************************************************************************
* Function: ScheduleProcess
* Summary: Schedules a process to be started for a simobj at a specified
*          time. Processes must never be scheduled in the past of the current
*          event or process.
* Description: This function can be called from anywhere within SPEEDES to
*              schedule a process. An proces must have a time tag, event type
*              simulation object type, simulation object local Id, and node
*              number to uniquely identify the process and the simobj that it
*              acts on. Optional arguments are the message header (users must
*              set specified data members before making the schedule call), and
*              a variable data buffer followed by the number of bytes. If the
*              variable data buffer is not provided, the SPEEDES internal
*              free list mechanisms are used to reduce overheads.
* Return: A cancel handle for the event which can be later used to retract
*         the event before it is processed. Applications are not required to
*         use this returned value.
******************************************************************************/
static inline SpCancelHandle ScheduleProcess(
				const SpSimTime &time, // IN: Logical time
				int eventType,	     // IN: Event type Id
				const SpObjHandle &simobj, // IN: simobj Object
                                                     //     Handle
				const M_SpProc *message = NULL, // IN: Optional
                                                            //     message for
                                                            //     scheduling
                                                            //     the event
				const char *data = NULL,   // IN: Variable length 
                                                     //     data that gets
                                                     //     passed with the
                                                     //     event (optional)
				int dataBytes = 0    // IN: Number of bytes for
                                                     //     variable length
                                                     //     data
				){
   return ScheduleProcess(time, eventType, simobj.ObjManId, simobj.SimObjId, simobj.NodeId, message, data,
			dataBytes);
}

static inline SpCancelHandle ScheduleProcess(
				const SpSimTime &time, // IN: Logical time
				int eventType,	     // IN: Event type Id
				const SpObjHandle &simobj, // IN: simobj Object
                                                     //     Handle
				const M_SpProc &message,    // IN:
                                                            //     message for
                                                            //     scheduling
                                                            //     the event
				const char *data = NULL,   // IN: Variable length 
                                                     //     data that gets
                                                     //     passed with the
                                                     //     event (optional)
				int dataBytes = 0    // IN: Number of bytes for
                                                     //     variable length
                                                     //     data
				){
   return ScheduleProcess(time, eventType, simobj.ObjManId, simobj.SimObjId, simobj.NodeId, &message, data,
			dataBytes);
}

/******************************************************************************
* Function: ReScheduleProcess
* Summary: Re-schedules the same process to be started for the same simobj at
*          a future time.
* Description: This function can be called from anywhere within SPEEDES to
*              reschedule the same proces for the same simobj for starting
*              at a later time.
* Return: A cancel handle for the process which can be later used to retract
*         the process before it is started. Applications are not required to
*         use this returned value.
******************************************************************************/
static inline SpCancelHandle ReScheduleProcess(
				  const SpSimTime &time, // IN: Logical time
				  M_SpProc *message = NULL,// IN: Optional
                                                             //     message for
                                                             //     scheduling the
                                                             //     process
				  const char *data = NULL,   // IN: Variable length 
                                                       //     data that gets
                                                       //     passed with the
                                                       //     process (optional)
				  int dataBytes = 0    // IN: Number of bytes
                                                       //     for variable
                                                       //     length data
				  ){
  return ScheduleProcess(time,
		       CurrentEvent->get_EVENT_TYPE(),
		       CurrentEvent->get_OBJECT_TYPE(),
		       CurrentEvent->get_LOCAL_ID(),
		       CurrentEvent->get_NODE(),
		       message,
		       data,
		       dataBytes);
}

/******************************************************************************
* Function: CancelEvent
* Summary: Cancels an event (or event handler or process) that was previously
*          scheduled.
* Description: This function may be called from anywhere in a SPEEDES
*              application to retract an event that was previously scheduled.
* Return: void
******************************************************************************/
static inline void CancelEvent(SpCancelHandle *cancelHandle  // IN: Cancel handle
		 ) {
  CurrentEvent->cancel(cancelHandle);
}

/******************************************************************************
* Function: CreateSimObj
* Summary: Schedules the event that creates a SimObj.
* Description: This function may be called from anywhere in a SPEEDES
*              application to schedule the creation of a SimObj.
* Return: SpCancelHandle
******************************************************************************/
SpCancelHandle CreateSimObj(const SpSimTime &time,
			    int objManId,
			    int nodeId,
			    SpObjHandle &objHandle,
			    char *data = NULL,
			    int dataBytes = 0);

/******************************************************************************
* Function: DestroySimObj
* Summary: Schedules the event that deletes a SimObj.
* Description: This function may be called from anywhere in a SPEEDES
*              application to schedule the deletion of a SimObj.
* Return: SpCancelHandle
******************************************************************************/
SpCancelHandle DestroySimObj(const SpSimTime &time,
			     const SpObjHandle objHandle);

#ifndef __IN_BUILTIN_EVENTS__
DECLARE_3_ARG(ScheduleUndirectedEventHandlers,
	      SpString80,
	      NET_INT,
	      NET_INT
	      );


/******************************************************************************
* Overloaded Function: ScheduleEventHandler
* Summary: Schedule an event handler for the present or future time for any
*          simulation object (SpSimObj) in the simulation. The data that
*          is passed in gets delivered to the event handler when it is
*          processed. The cancel handle is returned to the application and can
*          later be used to retract the event. If dest is NULL (the default), then all event handlers that have been explicitly subscribed (and only those) are invoked. 
* Description: This method may be called from anywhere in a SPEEDES
*              application and schedules an event handler for any simulation
*	       object. The eventName is the name of the event handler(s) that
*              get invoked when the event is processed. The data passed in
*              is what is delivered to the event handler. The cancel
*              handle is returned to the user. This allows the event to later
*              be retracted using the Cancel() method.
******************************************************************************/
SpCancelHandle ScheduleEventHandler(
                                 const SpSimTime& time, // IN: Simulated time 
			         const char *eventName,     // IN: String name of
                                                      //     the event handler
				 SpObjHandle *dest=NULL,    // IN: object to schedule for  
	                         const char *data = NULL,   // IN: Variable length
                                                      //     data that is
                                                      //     passed to the
                                                      //     event handler(s)
	                         int dataBytes = 0, // IN: Number of bytes
			                              //     in the data
				 enum SpEventHandlerType // category of eh
				 ehType=LEGACYEVENTHANDLER
			       );

/******************************************************************************
* Function: ScheduleInteraction
* Summary: Schedules directed or undirected interaction handlers
******************************************************************************/

static inline SpCancelHandle ScheduleInteraction(
					  const SpSimTime &time,
					  const char *eventName,
					  SpParmSet &paramSet,
					  SpObjHandle *oh = NULL
					  )
{
  int size;
  char *buff = paramSet.GenerateBuff(size);
  return ScheduleEventHandler(time,eventName,oh,buff,size,
			      INTERACTIONHANDLER);
}

/******************************************************************************
* Function: ScheduleEventHandler
* Summary: This function is provided for backwards compatibility.
* It schedules only legacy event handlers.
******************************************************************************/

static inline SpCancelHandle ScheduleEventHandler(
                                 const SpSimTime& time, // IN: Simulated time 
			         const char *eventName,     // IN: String name of
                                                      //     the event handler
			         int simobjType,      // IN: Simulation object
			                              //     type
			         int simobjLocalId,   // IN: LOCAL_ID of the
			                              //     simulation object
			         int simobjNode,      // IN: NODE that the
			                              //     simulation object
			                              //     is on
	                         const char *data = NULL,   // IN: Variable length
                                                      //     data that is
                                                      //     passed to the
                                                      //     event handler(s)
	                         int dataBytes = 0    // IN: Number of bytes
			                              //     in the data
			       ) {
  SpObjHandle dest;
  dest.SimObjId = simobjLocalId;
  dest.NodeId = simobjNode;
  dest.ObjManId = simobjType;
  return ScheduleEventHandler(time,eventName,&dest,data,dataBytes,
			      LEGACYEVENTHANDLER);

}


/******************************************************************************
* Overloaded Function: ReScheduleEventHandler
* Summary: Re-schedules the same event handler to be processed for the same
*          simobj at a future time.  The data that
*          is passed in gets delivered to the event handler when it is
*          processed. The cancel handle is returned to the application and can
*          later be used to retract the event.
* Description: This method may be called from anywhere in a SPEEDES
*              application. The eventName is the name of the event handler(s)
*              that get invoked when the event is processed. The data passed in
*              is what is delivered to the event handler. The cancel
*              handle is returned to the user. This allows the event to later
*              be retracted using the Cancel() method.
* Return: void
******************************************************************************/
static inline SpCancelHandle ReScheduleEventHandler(
                                 const SpSimTime& time, // IN: Simulated time 
	                         const char *data = NULL, // IN: Variable length
                                                      //     data that is
                                                      //     passed to the
                                                      //     event handler(s)
	                         int dataBytes = 0    // IN: Number of bytes
			                              //     in the data
			       ) {
  return ScheduleEventHandler(time,
			      CurrentEvent->get_eventHandlerName(),
			      CurrentEvent->get_OBJECT_TYPE(),
			      CurrentEvent->get_LOCAL_ID(),
			      CurrentEvent->get_NODE(),
			      data,
			      dataBytes);
}
#endif //#ifndef __IN_BUILTIN_EVENTS__
#endif
