// Filename: SpSimObj.H

#define SEED Random

#ifndef simobj_object
#define simobj_object

#include "SpDataParser.H"
class RB_SpBinaryTree;
class SpDataParser;
#include "xqueue.H"
#include "xhash.H"
#include "xlist.H"
#include "dbl_queue.H"
#include "SpPriorityQueueHeap.H"
class C_RANDOM;
#include "RB_SpRandom.H"
#include "SpSimObjMgrTypes.H"
#include "SpSimTime.H"
class SpNameValueList;
class SpEventHandler;
#include "SpEventHandlerTypes.H"
class SpEventHandlerHolder;
#include "SpObjHandle.H"
#include "SpPersistenceBaseClass.H"
#include "PO.H"
class SpStringBasedEventHandler;
class SpComponent;
class C_SQ_ITEM;
typedef void (*EVENT_HANDLER_METHOD_FNC_PTR) (void *, char *, SpSimTime &, char *, int , int);
#include "SpMath.H"

/******************************************************************************
* Class: SpSimObj
* Summary: A generic SPEEDES simulation object
* Description: This is the fundamental modeling unit in SPEEDES applications on
*              which SpEvents act. SpSimObjs must be designed independent of
*              other SpSimObjs so as to permit each SpSimObj to temporarily
*              exist at different logical (simulated) times as controlled by
*              the operative time management algorithm. SpSimObjs are permitted
*              to schedule  vents for other SpSimObjs, but this is generally
*              discouraged in favor of maximizing SpSimObj reuse. To achieve
*              greater reuse, SpSimObjs should use events attached to
*              themselves for scheduling events for other SpSimObjs.
******************************************************************************/
class SpSimObj : public SpPersistenceBaseClass {

  friend class SpMasterSimObjMgr;

 public:

/******************************************************************************
* Virtual Method: Query
* Summary: Query this simulation object for a name-value list of information
* Description: Simulation objects can be queried during run-time for real-time
*              updates of their state. The state information queried is defined
*              by the simulation object at compile time by overridding this
*              virtual function.
* Return: the size, in bytes, of the returned buffer, buff
******************************************************************************/
   virtual void Query(double time_tag, // IN: simulated time for the query
		      SpNameValueList &NameValueList // OUT: query list
		      );

    virtual void query(double time_tag, SpNameValueList &NameValueList) {
      Query(time_tag, NameValueList);
    }


/******************************************************************************
* Virtual Method: NamedQuery
* Summary: Query this simulation object for a name-value list of information
* Description: Simulation objects can be queried during run-time for real-time
*              updates of their state. The state information queried is defined
*              by the simulation object at compile time by overridding this
*              virtual function. This form of the query passes a query name
*	       to allow specific types of queries
* Return: the size, in bytes, of the returned buffer, buff
******************************************************************************/
   virtual void NamedQuery(double time_tag, // IN: simulated time for the query
		      SpNameValueList &NameValueList, // OUT: query list
		      char *queryName
		      );

/******************************************************************************
* Virtual Method: print
* Summary: User-overridable function to print something about this simulation
*          object during run-time
* Description:
* Return: None
******************************************************************************/
  virtual void print();

/******************************************************************************
* Virtual Method: terminate
* Summary: User-overridable function for operations (such as printing
*          statistics) associated with this simulation object to be performed
*          when the simulation completes
* Description: 
* Return: None
******************************************************************************/
  virtual void terminate(double   // the simulation passes the end time of
          		          // the simulation here (tend). Using this
			          // value is of course optional.
			 );

/******************************************************************************
* Overloaded Method: schedule
* Summary: Schedule an event at the present or a future simulated time for any
*          simulation object (SpSimObj) in the simulation
* Description: For reusability, prefer scheduling events in the process user
*              code of events (SpEvents). See the schedule methods in SpEvent.
* Return: the message for this event
******************************************************************************/
  void *schedule(const SpSimTime& time, // IN: simulated time 
		 int event_type,      // IN: enumerated or integer type of
	                              // event to schedule
		 int object_type,     // IN: enumerated or integer type of
	                              // object for which to schedule the
	                              // event
	                              // (must be the one type of object
	                              // required for the event_type) 
	         int object_id,       // IN: the LOCAL_ID of the object for
	                              // which to schedule the event
	         int object_node      // IN: the NODE of the object for which
	         );

/******************************************************************************
* Overloaded Method: schedule
* Summary: Schedule an event at the present or a future simulated time for any
*          simulation object (SpSimObj) in the simulation
* Description: For reusability, prefer scheduling events in the process user
*              code of events (SpEvents). See the schedule methods in SpEvent.
* Return: the message for this event
******************************************************************************/
  void *schedule(const SpSimTime& time, // IN: simulated time 
		 int event_type,      // IN: enumerated or integer type of
	                              // event to schedule
		 int object_type,     // IN: enumerated or integer type of
	                              // object for which to schedule the
	                              // event
	                              // (must be the one type of object
	                              // required for the event_type) 
	         int object_id,       // IN: the LOCAL_ID of the object for
	                              // which to schedule the event
	         int object_node,     // IN: the NODE of the object for which
	                              // to schedule the event
	         int size,            // IN: (OPTIONAL) number of bytes of
	                              // the variable-length part of the
	                              // event message
	         char *&buf           // IN: (USED IF AND ONLY IF "size" IS
	                              // USED) the variable-length part of
	                              // the event message 
	         );

/******************************************************************************
* Method: get_LOCAL_ID
* Summary: Get the ID of this simulation object
* Description: Each object manager has its own set of SpSimObjs, each with
*              their own LOCAL_ID. Thus LOCAL_ID only needs to be unique for a
*              given object manager (SpSimObjMgr), not unique to a given node or
*              to a given simulation
* Return: the ID of this simulation object
******************************************************************************/
  int get_LOCAL_ID() {return LOCAL_ID;}

  int get_NUM_ID() {return NUM_ID;}
  void set_NUM_ID(int numId) {NUM_ID = numId;}


// better names for the get and set methods

    int GetLocalId() {return LOCAL_ID;}
    void SetLocalId(int localId) {LOCAL_ID = localId;}

    int GetGlobalId() {return GLOBAL_ID;}
    void SetGlobalId(int globalId) {GLOBAL_ID = globalId;}

    int GetNumId() {return NUM_ID;}
    void SetNumId(int numId) {NUM_ID = numId;}

    int GetObjType() {return OBJECT_TYPE;}
    void SetObjType(int objType) {OBJECT_TYPE = objType;}

    int GetNode() {return NODE;}
    void SetNode(int node) {NODE = node;}

// This initializes the object using the built-in object managers

    virtual void InitParser(SpDataParser *) {;}
    virtual void InitData(char *) {;}
    virtual void Init(void) {;}

/******************************************************************************
* Method: get_OBJECT_TYPE
* Summary: Get the ID (or enumerated type) of the type of this simulation
*          object
* Description: A unique OBJECT_TYPE is required for each SpSimObj class.  Each
*              SpSimObj class is managed by one instantiation of
*              SpSimObjMgr. Thus, the OBJECT_TYPE also uniquely defines the type
*              of SpSimObj that each instantiation of SpSimObjMgr manages. Each
*              node has a single instantiation of SpMasterSimObjMgr, which in turn
*              manages a list of SpSimObjMgrs. SpMasterSimObjMgr refers to each SpSimObjMgr
*              by the OBJECT_TYPE it manages.
* Return: the OBJECT_TYPE, ie, the ID of the type of this simulation type
******************************************************************************/
  int get_OBJECT_TYPE() {return OBJECT_TYPE;}

/******************************************************************************
* Method: get_NODE
* Summary: Get the node (process number on local machine) that this SpSimObj is
*          instantialted on.
* Description:
* Return: the node (process number on local machine) that this SpSimObj
*         is instantialted on.
******************************************************************************/
  int get_NODE() const {return NODE;}
  
/******************************************************************************
* Method: get_N_NODES
* Summary: Get the total number of nodes (process number on local machine) in
*          the entire simulation (including all other machines' process)
* Description: 
* Return: total number of nodes
******************************************************************************/
  int get_N_NODES() {return N_NODES;}

/******************************************************************************
* Method: object_type
* Summary: Retrieves a simulation object's OBJECT_TYPE (see set_OBJECT_TYPE)
*          based on its string NAME
* Description: 
* Return: the object type of this simulation object
******************************************************************************/
  int object_type(char *c     // IN: NAME of a simulation object
		   ) {return obman->get_obid(c);}

/******************************************************************************
* Method: event_type 
* Summary: Get the enumerated ID for an event type, specified by the event
*          type's string name
* Description: 
* Return: the enumerated ID for an event type
******************************************************************************/
  int event_type(char *c   // IN: NAME of an event
		 );

/******************************************************************************
* Method: get_NAME
* Summary: Get the name of this object (for referring to this SpSimObj by name)
* Description: 
* Return: name of this object
******************************************************************************/
  char *get_NAME() {return NAME;}

/******************************************************************************
* Method: get_RANDOM
* Summary: Get the random number generator. There is exactly one C_RANDOM
*          object per node, set automatically by the framework.
* Description: see C_RANDOM
* Return: the random number generator
******************************************************************************/
  C_RANDOM *get_RANDOM() {return RANDOM;}

/******************************************************************************
* Method: get_SEED
* Summary: Get the seed for the random number generator. Must be saved and
*          recalled for repeatability due to rollbacks.
* Description: 
* Return: the random number generator seed
******************************************************************************/
  int get_SEED() {return SEED;}

/******************************************************************************
* Method: GetRandom
* Summary: Get the rollbackable random number generator.
* Description: 
* Return: the random number generator
******************************************************************************/
  RB_SpRandom *GetRandom() {return &Random;}

/******************************************************************************
* Method: get_monitor
* Summary: Get the number of external monitors connected to this simulation
*          object
* Description: 
* Return: the number of external monitors connected to this simulation
*         object
******************************************************************************/
  int get_monitor() {return qmon.length();}

/******************************************************************************
* Method: get_qmon
* Summary: Get the list of external monitors connected to this simulation
*          object
* Description: 
* Return: the list of external monitors connected to this simulation object
******************************************************************************/
  C_QUEUE *get_qmon() {return &qmon;}

/******************************************************************************
* Method: get_GLOBAL_ID
* Summary: Get the unique ID, simulaiton-wide, of this object. GLOBAL_ID
*          generated automatically, but can be overridden by the user.
* Description: 
* Return: the unique ID of this simulation object
******************************************************************************/
  int get_GLOBAL_ID() {return GLOBAL_ID;}

/******************************************************************************
* Method: set_GLOBAL_ID
* Summary: Set the unique ID, simulaiton-wide, of this object. GLOBAL_ID
*          generated automatically, but can be overridden by the user with this
*          method.
* Description: 
* Return: None
******************************************************************************/
  void set_GLOBAL_ID(int i) {GLOBAL_ID= i;}

/******************************************************************************
* Method: get_N_OBJECT_TYPES
* Summary: Get the total number of OBJECT_TYPEs (see above) in the current
*          simulation
* Description: 
* Return: the total number of OBJECT_TYPEs (see above) in the current
*         simulation
******************************************************************************/
  int get_N_OBJECT_TYPES() {return N_OBJECT_TYPES;}

/******************************************************************************
* Method: get_TOTAL_OBJECTS
* Summary: Get the total number of objects in the simulation. That is, the sum
*          of each object of each object type on each node.
* Description: 
* Return: the total number of objects in the simulation
******************************************************************************/
  int get_TOTAL_OBJECTS() {return TOTAL_OBJECTS;}

/******************************************************************************
* Method: get_manager
* Summary: Get the object manager for another simulation object, specified by
*          its OBJECT_TYPE
* Description: 
* Return: an object manager, casted to (void *)
******************************************************************************/
  void *get_manager(int om = -2 // IN: OBJECT_TYPE of the desired object manager
		                //     Default is this simobj's object manager
		    );

/******************************************************************************
* Method: EXCHANGE_SEED
* Summary: Exchange the SEED of this SpSimObj for rollback purposes
* Description: 
* Return: None
******************************************************************************/
  void EXCHANGE_SEED(int &seed  // INOUT: SEED to exchange with this SpSimObj's
		               // SEED
		     ) { Random.ExchangeSeed (seed); }

/******************************************************************************
* Method: get_EXTSUBS
* Summary: Gets the list of external subscriptions
* Description: Returns the EXTSUBS C_XQUEUE
* Return: Returns the pointer to the EXTSUBS object (of type C_XQUEUE)
******************************************************************************/
  C_XQUEUE *get_EXTSUBS() {return &EXTSUBS;}

/******************************************************************************
* Method: CheckSubType
* Summary: Checks if an external module (specified by the externalId input)
*          has subscribed to a specified type (represented as the string Type)
* Description: Searches through the EXTSUBS list for the pair <Type and
*              externalId> and returns 1 if found (otherwise, returns 0).
* Return: Returns 1 if the subscription type for that externalId is in
*         the EXTSUBS list. Otherwise, returns 0.
******************************************************************************/
  int CheckSubType(char *type, int externalId);

/******************************************************************************
* Method: CheckSubscription
* Summary: Checks if a subscription type specified by the input argument Name
*          is in the EXTSUBS list.
* Description: Searches through the EXTSUBS list for the named subscription.
* Return: Returns 1 if the subscription name is in the EXTSUBS list.
*         Otherwise, returns 0.
******************************************************************************/
  int CheckSubscription(char *name);

/******************************************************************************
* Method: AddComponent
* Summary: Adds the component to this SpSimObj
* Description: When a SpComponent is added to or removed from this
*              SpSimObj using AddComponent() and RemoveComponent(),
*              the event handlers associated with that SpComponent will
*              automatically be added to or removed from this SpSimObj
*              as well. This method can be used rollbackably during
*              runtime.
* Return: None
******************************************************************************/
  void AddComponent(SpComponent *component  // component to add
		    );

/******************************************************************************
* Method: RemoveComponent
* Summary: Removes the component from this SpSimObj
* Description: When a SpComponent is added to or removed from this
*              SpSimObj using AddComponent() and RemoveComponent(),
*              the event handlers associated with that SpComponent will
*              automatically be added to or removed from this SpSimObj
*              as well. This method can be used rollbackably during
*              runtime.
* Return: None
******************************************************************************/
  void RemoveComponent(SpComponent *component  // component to remove
		       );
   
/******************************************************************************
* Method: AddEventHandler
* Summary: Adds the event handler to the SpSimObj
* Description: If the hash table has not been created, it
*              automatically creates the hash table with a nominal
*              size.Adds the event handler to the simobj's set of
*              event handlers.  Note that it is possible to add more
*              than one event handler for a specified eventName. Each
*              eventName is managed in an item that is stored in the
*              EventHandlerHash table as a group.
* Return: None
******************************************************************************/
  void AddEventHandler(char *eventName,              // name of event
		       SpStringBasedEventHandler *eventHandler   // pointer to the
		                                     // associated
		                                     // event handler
		       );
/******************************************************************************
* Method: AddSubscribeEventHandler
* Summary: Adds the event handler to the SpSimObj, and subscribes
* to undirected interactions that go by eventName
* Description: Same as previous, except that the event handler
* will also receive undirected interactions             
* Return: None
******************************************************************************/
  void AddSubscribeEventHandler(char *eventName,              // name of event
		       SpStringBasedEventHandler *eventHandler   // pointer to the
		                                     // associated
		                                     // event handler
		       );


/******************************************************************************
* Method: AddEventHandlerMethod
* Summary: Adds the event handler method to the SpSimObj
* Description: Pass the name of the method to this method's second
*              argument. If the event handler hash table has not been
*              created, it automatically creates the hash table with a
*              nominal size. Adds the event handler to the simobj's
*              set of event handlers.  Note that it is possible to add
*              more than one event handler for a specified
*              eventName. Each eventName is managed in an item that is
*              stored in the EventHandlerHash table as a group.
* Return: Pointer to the event handler created by calling
*         AddEventHandlerMethod(). This is an optional handle if needed
*         to subsequently remove this event handler using
*         RemoveEventHandler().
******************************************************************************/
SpEventHandler *          AddEventHandlerMethod(char *eventName,
					       // name of event
					       EVENT_HANDLER_METHOD_FNC_PTR method,
					       // method name to be
					       // invoked as the
					       // associated event handler
					       // (implementation-wise, it is
					       // a pointer created by
					       // the EVENT_HANDLER_METHOD_H
					       // macro)
					       void *objPtr
					       // pointer to the
					       // method's object
					       // (will be the 'this'
					       // pointer if the
					       // method is contained
					       // directly in the
					       // SpSimObj.
					       );

/******************************************************************************
* Method: RemoveEventHandler
* Summary: Removes the event handler from the simobj. Note that the event
*          name item that is in the EventHandlerHash table is never deleted.
* Description: Removes the event handler from the simobj's set of event
*              handlers.
* Return: None
******************************************************************************/
  void RemoveEventHandler(char *eventName,             // name of event
			  SpStringBasedEventHandler *eventHandler,  // pointer to
			                               // SpEventHandler
			  enum SpEventHandlerMode 
			  flag=DONTCARE // optionally specifies
			  // which instances to remove, directed or
			  // undirected
			  );

/******************************************************************************
* Method: ProcessEventHandlers
* Summary: Processes all of the event handlers with a specified name.
* Description: Removes the event handler from the simobj's set of event
*              handlers.
* Return: None
******************************************************************************/
  void ProcessEventHandlers(char *eventName,  // IN: name of the event type
			    SpSimTime &timeTag, // IN: simulation time
			    char *data,       // IN: data passed to the
                                              // IN: event handler
			    int dataBytes,    // IN: number of bytes for data
			    int externalId,  // IN: id of external module
                                              //     if applicable (-1 means
                                              //     event not generated by
                                              //     external module)
			    int ehType=LEGACYEVENTHANDLER,    // IN: type of event handlers
			    //                                //     to process
			    int Undirected=0); // if undirected, only subscribed ehs will be
  //                                           // processed

/******************************************************************************
* Method: CheckEventHandler
* Summary: Checks to see if any event handlers with a specified name exist
*          for the simobj.
* Description: Searches for event handlers with a specified name and returns
*              1 if found (otherwise, returns 0).
* Return: 1 if event handler is found, otherwise returns 0.
******************************************************************************/
   int CheckEventHandler(char *eventName // string name of event to check
			);
   
/******************************************************************************
* Method: GetObjectHandle
* Summary: Returns a reference to this simobj's object handle
* Description: 
* Return: A reference to this simobj's object handle
******************************************************************************/
   SpObjHandle *GetObjectHandle();
   
/******************************************************************************
* Method: GetEarliestNextEventTime
* Summary: Returns a simtime that is identical to the current simtime, except
*          that its counter is increased to the max of the current time's
*          counter and the simobj's event counter, plus the length of the
*          outgoing message queue plus one.
* Description: 
* Return: A simtime
******************************************************************************/
  SpSimTime GetEarliestNextEventTime ();

/******************************************************************************
* Internal Method: print_event_memory
* Summary: print out free list event memory statistics
* Description: 
* Return: None
******************************************************************************/
  void print_event_memory();

/******************************************************************************
* Internal Method: set_NAME
* Summary: Set the name of this object (for referring to this SpSimObj by name)
* Description: 
* Return: None
******************************************************************************/
  virtual void set_NAME(char *n) {NAME = n;}

/******************************************************************************
* Internal Method: set_NODE
* Summary: Set the node (process number on local machine) that this SpSimObj
*          shall be instantialted on.
* Description:
* Return: None
******************************************************************************/
  void set_NODE(int nd) {NODE = nd;}

/******************************************************************************
* Internal Method: set_N_NODES
* Summary: Set the total number of nodes (process number on local machine) in
*          the entire simulation (including all other machines' process)
* Description: 
* Return: None
******************************************************************************/
    void set_N_NODES(int nds) {N_NODES = nds;}

/******************************************************************************
* Internal Method: set_LOCAL_ID
* Summary: Set the ID of this simulation object
* Description: Each object manager has its own set of SpSimObjs, each with
*              their own LOCAL_ID. Thus LOCAL_ID only needs to be unique for a
*              given object manager (SpSimObjMgr), not unique to a given node or
*              to a given simulation
* Return: None
******************************************************************************/
  void set_LOCAL_ID(int i) {LOCAL_ID = i;}

/******************************************************************************
* Internal Method: set_OBJECT_TYPE
* Summary: Set the ID (or enumerated type) of the type of this simulation
*          object
* Description: A unique OBJECT_TYPE is required for each SpSimObj class.  Each
*              SpSimObj class is managed by one instantiation of
*              SpSimObjMgr. Thus, the OBJECT_TYPE also uniquely defines the type
*              of SpSimObj that each instantiation of SpSimObjMgr manages. Each
*              node has a single instantiation of SpMasterSimObjMgr, which in turn
*              manages a list of SpSimObjMgrs. SpMasterSimObjMgr refers to each SpSimObjMgr
*              by the OBJECT_TYPE it manages.
* Return: None
******************************************************************************/
  void set_OBJECT_TYPE(int obtyp) {OBJECT_TYPE = obtyp;}

/******************************************************************************
* Internal Method: set_N_OBJECT_TYPES
* Summary: Set the total number of OBJECT_TYPEs (see above) in the current
*          simulation
* Description: 
* Return: None
******************************************************************************/
  void set_N_OBJECT_TYPES(int nt) {N_OBJECT_TYPES = nt;}

/******************************************************************************
* Internal Method: set_RANDOM
* Summary: Set the random number generator. There is exactly one C_RANDOM
*          object per node, set automatically by the framework.  User should
*          only use get.
* Description: see C_RANDOM
* Return: None
******************************************************************************/
  void set_RANDOM(C_RANDOM *r) {RANDOM = r;}

/******************************************************************************
* Internal Method: set_SEED
* Summary: Set the seed for the random number generator. Must be saved and
*          recalled for repeatability due to rollbacks.
* Description: 
* Return: None
******************************************************************************/
  void set_SEED(int sd) {SEED = sd;}

/******************************************************************************
* Internal Method: set_TOTAL_OBJECTS
* Summary: Set the total number of objects in the simulation. That is, the sum
*          of each object of each object type on each node.
* Description: 
* Return: None
******************************************************************************/
  void set_TOTAL_OBJECTS(int to) {TOTAL_OBJECTS = to;}

/******************************************************************************
* Internal Method: set_rq_holder
* Summary: Set the rollback queue holder for this simulation object (see
*          C_RQ_HOLDER)
* Description: 
* Return: None
******************************************************************************/
  void set_rq_holder(C_RQ_HOLDER *rqh) {rq_holder = rqh;}

/******************************************************************************
* Internal Method: set_obman
* Summary: Set this simulation object's object manager constructor
* Description: 
* Return: None
******************************************************************************/
  void set_obman(SpSimObjMgrTypes *ob) {obman = ob;}

/******************************************************************************
* Internal Method: set_TIME
* Summary: Set the time of this simulation object. This is not necessarily the
*          time of the current event.
* Description:  
* Return: None
*****************************************************************************/
  void set_TIME(SpSimTime& tt) {TIME = tt;}

/******************************************************************************
* Internal Method: set_uid_seed
* Summary: Set the seed to calculate GLOBAL_ID
* Description: 
* Return: None
******************************************************************************/
  void set_uid_seed(int sd) {uid_seed = sd;}

/******************************************************************************
* Internal Method: set_last_etype
* Summary: Set the previous event type to act on this simulation object
* Description: 
* Return: None
******************************************************************************/
  void set_last_etype(int e) {last_etype = e;}

/******************************************************************************
* Internal Method: set_rqueue
* Summary: Attach this node's event queue to this simulation object
* Description: 
* Return: None
******************************************************************************/
  void set_rqueue (SpPriorityQueueHeap *rq) {rqueue = rq;}

/******************************************************************************
* Internal Method: set_init_event_queue 
* Summary: Set the event queue for this node to its initial queue
* Description: 
* Return: None
******************************************************************************/
  void set_init_event_queue(C_QUEUE *qi) {init_events = qi;}

/******************************************************************************
* Internal Method: set_evtype
* Summary: Set the event manager object for this node (type SpEventTypes)
* Description: 
* Return: None
******************************************************************************/
  void set_evtype(void *e) {evtype = e;}

/******************************************************************************
* Internal Method: get_TIME
* Summary: Get the time of this simulation object. This is not necessarily the
*          time of the current event.
* Description:  
* Return: time of this simulaiton object
*****************************************************************************/
  SpSimTime& get_TIME() {return TIME;}

/******************************************************************************
* Internal Method: get_uid_seed
* Summary: Get the seed to calculate GLOBAL_ID
* Description: 
* Return: seed for calculating GLOBAL_ID
******************************************************************************/
  int get_uid_seed() {return uid_seed;}

/******************************************************************************
* Internal Method: get_last_etype
* Summary: Get the previous event type to act on this simulation object
* Description: 
* Return: the previous event type to act on this simulation object
******************************************************************************/
  int get_last_etype() {return last_etype;}

/******************************************************************************
* Internal Method: get_local_rqueue
* Summary: Retrieve this simulation object's queue of unprocessed events
* Description: 
* Return: this simulation object's queue of unprocessed events
******************************************************************************/
  SpPriorityQueueHeap *get_local_rqueue () {return &local_rqueue;}

/******************************************************************************
* Internal Method: get_rq_holder
* Summary: Get the rollback queue holder for this simulation object (see
*          C_RQ_HOLDER)
* Description: 
* Return: the rollback queue holder for this simulation object
******************************************************************************/
  C_RQ_HOLDER *get_rq_holder() {return rq_holder;}

/******************************************************************************
* Internal Method: reset_nlobjs
* Summary: Set the number of local objects to zero
* Description: 
* Return: None
******************************************************************************/
    void reset_nlobjs() {nlobjs = 0;}
    void set_nlobjs(int n) {nlobjs = n;}

/******************************************************************************
* Internal Method: init_objs
* Summary: Initialization routine for simulation objects
* Description: 
* Return: None
******************************************************************************/
  void init_objs() {uid_seed = NODE; TOTAL_OBJECTS = 0;}

/******************************************************************************
* Internal Method: combine_total_objects
* Summary: Recalculate the total number of objects on all nodes
* Description: 
* Return: None
******************************************************************************/
  void combine_total_objects();

/******************************************************************************
* Internal Method: insert_event
* Summary: Insert an event into this simulation object's list of pending events
* Description: 
* Return: None
******************************************************************************/
  void insert_event(C_SQ_ITEM *event);

/******************************************************************************
* Internal Method: get_evtq
* Summary: Get the event queue on this node
* Description: 
* Return: the event queue on this node
******************************************************************************/
  C_DBL_QUEUE *get_evtq() {return &evtq;}

/******************************************************************************
* Internal Method: get_qblock
* Summary: Get the queue of blocking times for this node
* Description: 
* Return: the queue of blocking times for this node
******************************************************************************/
  static C_QUEUE *get_qblock() {return qblock;}

/******************************************************************************
* Internal Method: get_blocking
* Summary: Get the number of blocking times on this node
* Description: 
* Return: the number of blocking times on this node
******************************************************************************/
  int get_blocking() {return qblock->length();}

/******************************************************************************
* Internal Method: get_tblock
* Summary: Get the simulated time tag of the earliest blocking event
* Description: 
* Return: None
******************************************************************************/
  SpSimTime &get_tblock();

/******************************************************************************
* Internal Method: add_block
* Summary: Add a blocking time
* Description: 
* Return: None
******************************************************************************/
  static void add_block(C_ITEM *it  // IN: a C_ITEM whose TIMETAG is
			            // a blocking time
			) {qblock->insert_after(it);}

/******************************************************************************
* Internal Method: add_unblock
* Summary: Insert the current event as an unblocking event after a specified
*          event in this node's list of unblocking events
* Description: 
* Return: None
******************************************************************************/
static void add_unblock(C_ITEM *it  // IN: the event after which the
                                    // current event is to be inserted
                       ) {qunblock->insert_after(it);}

/******************************************************************************
* Internal Method: remove_block
* Summary: Remove a time from this node's list of blocking times
* Description: 
* Return: None
******************************************************************************/
void remove_block(C_ITEM *it   // IN: a C_ITEM with whose TIME_TAG is
                               // a blocking time to remove from this
                               // node's list of blocking times
		  ) {qblock->deque(it);}

/******************************************************************************
* Internal Method: unblock
* Summary: unblock a socket
* Description: 
* Return: None
******************************************************************************/
  void unblock(int soc  // IN: the socket ID to unblock
	       );

/******************************************************************************
* Internal Method: update_block
* Summary: update the blocking and unblocking event lists for this node
* Description: 
* Return: None
******************************************************************************/
  void update_block();

/******************************************************************************
* Internal Method: GetCancelTree
* Summary: Get the binary tree of cancelled events for this simulation object
* Description: 
* Return: the binary tree of cancelled events for this simulation object
******************************************************************************/
  RB_SpBinaryTree *GetCancelTree() {return CancelTree;}
  int GetNumCanceled();
  void AttachCancelTreePointer();

/******************************************************************************
* Internal Method: get_obman
* Summary: Get this simulation object's object manager constructor
* Description: 
* Return: this simulation object's object manager constructor
******************************************************************************/
  SpSimObjMgrTypes *get_obman() {return obman;}

/******************************************************************************
* Internal Method: get_obname (FOR BACKWARD COMPATIBILITY)
* Summary: Get the NAME of this simulation object
* Description: 
* Return: the NAME of this simulation object
******************************************************************************/
  char *get_obname(int i) {return obman->get_obname(i);}

  int GetEvCounter() {return EvCounter;}
  void SetEvCounter(int evCounter) {EvCounter = evCounter;}

  void IncrementProcessingTime(double CPU_Time) {
    ProcessingTime += CPU_Time;
  }

  double GetProcessingTime() {return ProcessingTime;}

  void IncrementCriticalPathTime(double earliestStartTime, double CPU_Time) {
    if (earliestStartTime < CriticalPathTime) {
      CriticalPathTime += CPU_Time;
    }else{
      CriticalPathTime = earliestStartTime + CPU_Time;
    }
  }

  double GetCriticalPathTime() {return CriticalPathTime;}

  void IncrementTotalTime(double CPU_Time) {TotalTime += CPU_Time;}
  double GetTotalTime() {return TotalTime;}

  void IncrementCommittedTime(double CPU_Time) {CommittedTime += CPU_Time;}
  double GetCommittedTime() {return CommittedTime;}

  void IncrementTotalEvents() {TotalEvents++;}
  int GetTotalEvents() {return TotalEvents;}

  void IncrementCommittedEvents() {CommittedEvents++;}
  int GetCommittedEvents() {return CommittedEvents;}

  virtual SpSimTime GetMinGatewayMsgTime() {
    SpSimTime time;
    time = 1.0e20;
    return time;
  }

    SpSimObj();

    SpSimObj(void*):EventHandlerHash(1){}

    virtual ~SpSimObj();

  private:
    friend class SpProcStackElement; // to call the following function:
    int RegisterProc() { return ++ProcCounter; }

    friend class SpSimObjMgr; // to call the following two functions:
    void SetObjectHandle(int nodeId, int objManId, int simObjId);
    void GetObjectHandleAttributes(const SpObjHandle& simobj, int& nodeId,
				   int& objManId, int& simObjId);

    SpObjHandle ObjectHandle;       // object handle for this simobj
    static int done;			// done flag (first simobj created)
   
    int last_etype;			// last event type of object
    C_QUEUE qmon;			// queue of external monitors
    C_DBL_QUEUE evtq;			// event queue for object
    RB_SpBinaryTree* CancelTree;	// tree of canceled event cancel Ids

    SpPriorityQueueHeap local_rqueue;	// speedes queue for this object
    static SpPriorityQueueHeap *rqueue;	// speedes queue for local objects

    C_RQ_HOLDER *rq_holder;		// rqueue holder for the object
    // RVI: Needs to be set to the correct value

    C_XHASH EventHandlerHash;		// manages event handlers
    XLIST ComponentList;		// manages components

    RB_int EvCounter;                   // Counter for tie-breaking
    RB_double CriticalPathTime;		// Time spent on the critical path
    RB_double ProcessingTime;		// Total time processing events
    double TotalTime;			// Total CPU time (includes rollbacks)
    double CommittedTime;		// Total CPU time committed
    int TotalEvents;			// Total events processed
    int CommittedEvents;		// Total events committed

    static void *evtype;		// evtype manages event creation

    static SpSimObjMgrTypes *obman;	// object manager constructor
    static C_QUEUE *qblock;		// queue of blocking times
    static C_QUEUE *qunblock;		// queue of unblocking events
    static C_QUEUE *init_events;	// initial event queue
    static int nlobjs;			// number of local objects
    static int uid_seed;		// unique id for this object

    SpEventHandlerHolder *SetUpEventHandlerHolder (char *eventName);

    SpSimTime TIME;		        // time of object GEB3
    RB_int ProcCounter;                 // counter for this simobj's processes,
                                        // used for giving processes their ProcId
  void ResetRqueueHolder(){rq_holder = NULL;}

  virtual void RegisterWithPersistence(){};
  virtual void ResetObjectForRestart(){};

 protected:


    char *NAME;				// name of object
  //RVI Need a smart pointer here



    int OBJECT_TYPE;			// object type
    int LOCAL_ID;			// local object id
    int GLOBAL_ID;			// unique id for simulation
    int NUM_ID;				// id ranges 0..N_TOT-1

//    STATE_INT SEED;                   // random number SEED
    RB_SpRandom Random;			// random number generator and SEED

    C_XQUEUE EXTSUBS;    		// requested external subscriptions

    static RB_int TOTAL_OBJECTS;	// total number of objects created
    static int N_OBJECT_TYPES;		// number of object types
    static int N_NODES;			// the total number of nodes
    static int NODE;			// node id

    static C_RANDOM *RANDOM;		// shared random number generator

};

typedef SpSimObj C_SIMOBJ;

#endif
