// SpSimObjMgr.H header file

#ifndef objman_object
#define objman_object

#include "random.H"
#include "SpSimObj.H"
#include "queue.H"
#include "defunc.H"
#include "SpDynPtrArray.H"
class RB_SpHashTree;

class SpDataParser;

#define MAX_NAME_LENGTH 40

// for card dealing:
typedef SpSimObj * (*SIMOBJ_CONSTRUCTOR_FUNCTION)(SpDataParser &dp);

/*****************************************************************************
* Class: C_GLOBAL_NAME
* Summary: Structure for a hash table of global names
* Description: Used only by SpSimObjMgr
*****************************************************************************/
class C_GLOBAL_NAME : public C_ITEM {
  public:
    NET_INT node;
    NET_INT lid;
    char name[MAX_NAME_LENGTH];
};

/*****************************************************************************
* Class: SpSimObjMgr
* Summary: Manages (creates and distributes) the SpSimObjs of a given type on
*          this node. Also, since this class inherits from SpSimObj itself, it
*          can handle events scheduled for itself (LOCAL_ID = -1).
* Description: Descendants of this class should set the variables
*              SpSimObjMgr::N_LOC and SpSimObjMgr::N_TOT, call reserve_n_objects ()
*              and define_object (), and increment TOTAL_OBJECTS. To card deal
*              this object type, use deal_me () or CardDeal().
*****************************************************************************/
class SpSimObjMgr : public SpSimObj {

 public:

/******************************************************************************
* Virtual Method: init_events
* Summary: Method in which to schedule initial events.
* Description: This method is called just before the beginning of the
*              simulation.  (I.e., just prior to simulated time = zero). This
*              is done so users can schedule evetns for simulated time =
*              zero. Users schedule initial events in each SpSimObjMgr on each
*              node by overriding this method.
* Return: None
******************************************************************************/
  virtual void init_events();

/******************************************************************************
* Virtual Method: reserve_n_objects
* Summary: Allocate memory for an array of n pointers to the SpSimObjs to be
*          created by this SpSimObjMgr
* Description: By default,
* Return: None
******************************************************************************/
  virtual void reserve_n_objects(int numObjs // number of objects to reserve
				 );

/******************************************************************************
* Virtual Method: define_object
* Summary: Put a SpSimObj into the array of SpSimObjs managed by this SpSimObjMgr
* Description: 
* Return: None
******************************************************************************/
  virtual void define_object(SpSimObj *simObj // SpSimObj to define
			     );

/******************************************************************************
* Virtual Method: get_obj
* Summary: Get the SpSimObj from this SpSimObjMgr (ie, of this SpSimObjMgr's
*          OBJECT_TYPE) specified by LOCAL_ID
* Description: 
* Return: The SpSimObj from this SpSimObjMgr (ie, of this SpSimObjMgr's OBJECT_TYPE)
*         specified by LOCAL_ID
******************************************************************************/
  virtual SpSimObj *get_obj(int localId  // LOCAL_ID of returned SpSimObj
			    );

/******************************************************************************
* Virtual Method: get_oid
* Summary: Get the GLOBAL_ID of the SpSimObj specified by NODE and LOCAL_ID
* Description: 
* Return: None
******************************************************************************/
  virtual void get_oid(int &oi, // OUT: GLOBAL_ID of SpSimObj of interest
		       int n,   // IN: NODE of SpSimObj of interest
		       int i    // IN: LOCAL_ID of SpSimObj of interest
		       );  

/******************************************************************************
* Virtual Method: create_object
* Summary: Dynamically (and rollbackably) create a SpSimObj of this SpSimObjMgr's
*          OBJECT_TYPE
* Description: Defaults to returning NULL (must be overridden)
* Return: pointer to a dynamically-created SpSimObj of this SpSimObjMgr's
*         OBJECT_TYPE
******************************************************************************/
  virtual SpSimObj *create_object() {return NULL;}


    virtual SpSimObj *CreateSimObj(){return NULL;}
    virtual void DestroySimObj(SpSimObj *) {;}


/******************************************************************************
* Virtual Method: destroy_object
* Summary: Dynamically (and rollbackably) destroy a SpSimObj of this SpSimObjMgr's
*          OBJECT_TYPE
* Description: Defaults to a pass-through method (must be overridden)
* Return: None
******************************************************************************/
  virtual void destroy_object(SpSimObj * // SpSimObj to destroy
			      ) {;}

/******************************************************************************
* Virtual Method: init_object
* Summary: Initialize a particular SpSimObj managed by this SpSimObjMgr 
* Description: Defaults to a pass-through method (must be overridden)
* Return: None
******************************************************************************/
  virtual void init_object(SpSimObj *, // intended as SpSimObj to initialize
			   void *      // intended as initialization data
			   ) {;}

/******************************************************************************
* Method: set_global_names
* Summary: Creates a hash table of the NAMEs of all of the SpSimObjs in the
*          simulation for easy lookups using get_nodid (char *name, int &nd,
*          int &lid)
* Description: The purpose of this method is to ease the burden on the user
*              when scheduling events. This method uses global reduction
*              techniques to create a local data base of each SpSimObj in the
*              simulation that has the NAME attribute defined (regardless of
*              what node they reside on). We recommend calling this method in
*              this SpSimObjMgr's constructor. Then, get_nodid (char *name, int
*              &nd, int &lid) and print_global_names() can be called any time
*              during the simulation and access the database created at that
*              point. However, the database is not updated as SpSimObjs are
*              created and destroyed.
* Return: None
******************************************************************************/
  void set_global_names();

/******************************************************************************
* Method: print_global_names
* Summary: Print the global name database to cout (see set_global_names())
* Description: 
* Return: None
******************************************************************************/
  void print_global_names();

/******************************************************************************
* Overloaded Method: get_nodid
* Summary: Get the NODE and LOCAL_ID of the SpSimObj specified by its NAME
* Description: 
* Return: None
******************************************************************************/
  void get_nodid(char *name, // IN: name of SpSimObj of interest
		 int &nd,    // OUT: The NODE of the SpSimObj of interest
		 int &lid);  // OUT: The LOCAL_ID of the SpSimObj of
			     // interest

/******************************************************************************
* Overloaded Virtual Method: get_nodid
* Summary: Get the NODE and LOCAL_ID of the SpSimObj specified by its GLOBAL_ID
* Description: 
* Return: None
******************************************************************************/
  virtual void get_nodid(int oi, // IN: The GLOBAL_ID of the SpSimObj of
			         // interest
			 int &n, // OUT: The NODE of the SpSimObj of interest
			 int &i  // OUT: The LOCAL_ID of the SpSimObj of
			         // interest
			 );

/******************************************************************************
* Method: set_N_TOT
* Summary: Set N_TOT, which is the total number of this SpSimObjMgr's type of
*          SpSimObj in the simulation
* Description: Users are responsible to set this variable in this SpSimObjMgr's
*              constructor
* Return: None
******************************************************************************/
  void set_N_TOT(int n // N_TOT
		 ) {N_TOT = n;}

/******************************************************************************
* Method: get_N_TOT
* Summary: Get N_TOT, which is the total number of this SpSimObjMgr's type of
*          SpSimObj in the simulation
* Description: 
* Return: N_TOT
******************************************************************************/
  int get_N_TOT() {return N_TOT;}

/******************************************************************************
* Overloaded Method: set_N_LOC
* Summary: Set N_LOC, which is the total number of this SpSimObjMgr's type of
*          SpSimObj on this node
* Description: Users are responsible to set this variable in this SpSimObjMgr's
*              constructor
* Return: None
******************************************************************************/
  void set_N_LOC(int n // N_LOC
		 ) {N_LOC = n;}

/******************************************************************************
* Method: set_N_LOC
* Summary: Set N_LOC, which is the total number of this SpSimObjMgr's type of
*          SpSimObj on this node
* Description: Users are responsible to set this variable in this SpSimObjMgr's
*              constructor. This version of this overloaded method uses N_TOT
*              (which must be set previously by the user) and N_NODES to
*              determine N_LOC, using the deal_me () algorithm.
* Return: None
******************************************************************************/
  void set_N_LOC();

/******************************************************************************
* Method: get_N_LOC
* Summary: Get N_LOC, which is the total number of this SpSimObjMgr's type of
*          SpSimObj on this node
* Description: 
* Return: N_LOC
******************************************************************************/
  int get_N_LOC() {return N_LOC;}

/*
** GenerateNewLocalId - generates a new unique local id for dynamic object
** creation.
*/

  int GenerateNewLocalId() {
    int newLocalId = N_TOT + CreationOffset + NODE;
    CreationOffset += N_NODES;
    return newLocalId;
  }


/******************************************************************************
* Internal Method: add_object
* Summary: Add a SpSimObj dynamically to the simulation
* Description: This function is not currently supported
* Return: None
******************************************************************************/
  void add_object(SpSimObj *simobj // Dynamically created SpSimObj to add to
		                   // the simulation
		  );

/******************************************************************************
* Internal Method: remove_object
* Summary: Remove a SpSimObj dynamically to the simulation
* Description: This function is not currently supported
* Return: None
******************************************************************************/
  void remove_object(SpSimObj *simobj // SpSimObj to remove from the simulation
		     );

/******************************************************************************
* Internal Method: name_hash
* Summary: 
* Description: 
* Return: 
******************************************************************************/
  int name_hash(char *name);

/******************************************************************************
* Internal Method: get_allids
* Summary: 
* Description: 
* Return: 
******************************************************************************/
  char *get_allids();

/******************************************************************************
* Internal Method: get_allids_size
* Summary: 
* Description: 
* Return: 
******************************************************************************/
  int get_allids_size() {return (interact*(3*sizeof(int) + 40)*N_LOC);}

/******************************************************************************
* Internal Method: set_interact
* Summary: 
* Description: 
* Return: None
******************************************************************************/
  void set_interact() {interact = 1;}

/******************************************************************************
* Internal Method: reset_interact
* Summary: 
* Description: 
* Return: None
******************************************************************************/
  void reset_interact() {interact = 0;}

/******************************************************************************
* Internal Method: term
* Summary: 
* Description: 
* Return: None
******************************************************************************/
  void term(double tend);

/******************************************************************************
* Internal Method: CriticalPath
* Summary: 
* Description: 
* Return: None
******************************************************************************/
  void ComputeCriticalPath(double &CPU_Time, double &criticalPathTime);


    void IncrementCombinedTotalEvents() {CombinedTotalEvents++;}
    int GetCombinedTotalEvents() {return CombinedTotalEvents;}

    void IncrementCombinedCommittedEvents() {CombinedCommittedEvents++;}
    int GetCombinedCommittedEvents() {return CombinedCommittedEvents;}

    void IncrementCombinedTotalTime(double CPU_Time) {
      CombinedTotalTime += CPU_Time;
    }
    double GetCombinedTotalTime() {return CombinedTotalTime;}

    void IncrementCombinedCommittedTime(double CPU_Time) {
      CombinedCommittedTime += CPU_Time;
    }
    double GetCombinedCommittedTime() {return CombinedCommittedTime;}



  SpSimObjMgr();
  void RegisterSimObjsWithPersistence();
  void ResetLocalRqueues();
  void ResetObjectForRestart();
  
  protected:

    int N_TOT;			        // total number of objects of my type
    int N_LOC;			        // local number of objects of my type
    RB_int CreationOffset;		// offset for dynamic object creation

   void SetObjectHandle(int nodeId, int objManId, int simObjId);

   void CardDeal(int totalSimObj,
		 SpDynPtrArray &locObj,
		 SIMOBJ_CONSTRUCTOR_FUNCTION simobjConstFunc,
		 SpDataParser &parser,
		 int replication = 0,
		 char *objTypeName = "<name not available>",
		 int useCodeGenParFilePattern = 0);

   void CodeGenCardDeal(SpDynPtrArray &locObj,
			SIMOBJ_CONSTRUCTOR_FUNCTION simobjConstFunc,
			SpDataParser &parser,
			int replication,
			char *objTypeName = "<name not available>")
   {
      CardDeal(-1, locObj, simobjConstFunc, parser, replication, objTypeName,
	       1);
   }
   
/**************************************************************************
* Method: deal_me
* Summary: card deal objects of this simulation object type for this node

* Description: this function is called by this simulation type's object manager
*              constructor when the distribution strategy for this class of
*              simulation objects is to "card deal" all the simulation objects
*              of this type to all of the nodes in the simulation. nobjs is the
*              total number of objects that the calling manager wants
*              distributed to *all* nodes (i.e. the total number of objects
*              that need to be managed collectively by all the nodes).  This
*              function returns the number of those objects that this node
*              should manage (the other nodes will manage the other objects).
*              The calling manager manages objects with ids in the calling
*              manager of N_NODES * j, where j is a whole number.
* Return: N_LOC
**************************************************************************/
  int deal_me(int nobjs // IN: total number of objects of this type to be
	                // managed collectively by all SpSimObjMgrs of this
	                // type on all nodes (normally N_TOT)
	      );

/**************************************************************************
* Method: deal_me
* Summary: card deal objects of this simulation object type for this node
* Description: This function is called by this simulation type's object manager
*              constructor when the distribution strategy for this class of
*              simulation objects is to "card deal" all the simulation objects
*              of this type to all of the nodes in the simulation. nobjs is the
*              total number of objects of this type to be managed collectively
*              by all SpSimObjMgrs of this type on all nodes (normally N_TOT is
*              passed as the argument for this parameter). This function
*              returns the number of those objects that this node should manage
*              (the other nodes will manage the other objects).  This SpSimObjMgr
*              manages objects with ids in the calling manager of offset +
*              N_NODES * j, where j is a whole number. Or, another way to state
*              this is that offset is used to determine which SpSimObjs this
*              SpSimObjMgr should manage. Namely, this SpSimObjMgr manages all
*              C_SIMOJs of LOCAL_ID == i where ((i-offset+N_NODES)%N_NODES) ==
*              0 and (0 <= i < N_TOT).
* Return: N_LOC
**************************************************************************/
  int deal_me(int nobjs,  // IN: total number of objects of this type to be
	                  // managed collectively by all SpSimObjMgrs of this
	                  // type on all nodes (normally N_TOT)
	      int &offset // OUT: node offset
	      );


  private:

    static int done;		// pointer to the objects
    static int seed_update;	// seed update
    static int ntobj;		// number of total objects

    int CombinedTotalEvents;	// total number of events for my simobjs
    int CombinedCommittedEvents;// number of committed events for my simobjs
    double CombinedTotalTime;	// total processing time for my simobjs
    double CombinedCommittedTime; // total committed time for my simobjs

    int interact;		// interaction flag
    int ndef;			// number of local object defined so far
//    SpSimObj **gobj;		// pointer to the objects
    RB_SpHashTree* LocalSimObjs;	// container of local SimObjs
    C_GLOBAL_NAME *global_name;	// array of global names
    C_QUEUE *name_queue;	// array of queues of global names
};

typedef SpSimObjMgr C_OBJMAN;

#endif
