// Filename: SpSimTime.H

#ifndef simtime_object
#define simtime_object

#include "SpStdio.H"
#include "SpIostream.H"
#include "SpFstream.H"
#include "SpStrstream.H"
#include "SpIomanip.H"
#include "SpNetTypes.H"

#ifdef __SPEEDES_NT__
extern const double DefaultTime;
extern const int DefaultPriority1;
extern const int DefaultPriority2;
extern const int DefaultCounter;
extern const int DefaultUniqueId;

extern const int MaxPriority1;
extern const int MaxPriority2;
#else
static const double DefaultTime = 0.0;
static const int DefaultPriority1 = 0;
static const int DefaultPriority2 = 0;
static const int DefaultCounter = 0;
static const int DefaultUniqueId = 0;

static const int MaxPriority1 = 0x7fffffff;
static const int MaxPriority2 = 0x7fffffff;
#endif
/*****************************************************************************
* Class: SpSimTime
* Summary: An abstract representation of simulated time
* Description: SpSimTime uses a double for representing the physical time and
*              then two integer priority fields for breaking physical time
*              ties. This is for repeatability when on multiple nodes and
*              rollbacks are involved. Since rollbacks are not repeatable, when
*              events are rescheduled, there must be a way to order them the
*              same way each time they are reprocessed. Since physical times
*              are often the same, priorities are needed to indicate the order
*              in which events which have the same physical time are to be
*              processed. The first priority field has precedence over the
*              second priority field. The combination of the physical time and
*              the two priority fields is called logical time. Repeatable
*              rollbackable simulations are assured only if events always
*              schedule new events in future logical time. In other words,
*              scheduling must either have physical time non-zero lookahead, or
*              physical time zero lookahead with higher priority values
*              (meaning later logical time) than the priorities of the current
*              event.
*****************************************************************************/

class SpSimTime {

 public:

/******************************************************************************
* Constructor: SpSimTime (same as SetValues ())
* Summary: Set the three values in the SpSimTime object
* Description: The three values are 1. physical time, 2. first priority
*              field, 3. second priority field.
******************************************************************************/
  SpSimTime(double t=DefaultTime,    // physical time
	  int p1=DefaultPriority1, // first priority field
	  int p2=DefaultPriority2,  // second priority field
	  int counter=DefaultCounter,   // counter priority field
	  int uniqueId=DefaultUniqueId // unique id tie-breaking field
	  ) {
#ifdef __INSURE__
    // Jacob added to get rid of a LEAK_SCOPE error from insure.  I do
    // not know why setting this causes the error to go away.
    // You might think that it wastes time to set it, but the ifdef
    // means it is only used if insure is used, and since insure
    // already drastically slows things down, it does not matter to
    // slow it down only a tiny bit more with this line.
    Time = 0;
#endif
    SetValues(t,p1,p2,counter,uniqueId);
  }

/*****************************************************************************
 * Constructor: SpSimTime ()
 * Summary: Copy constructor
 * Description: Explicit copy constructors are required for the APEX
 *              compiler from Rational since it creates a faulty
 *              default copy constructor
******************************************************************************/ 
  
  SpSimTime(const SpSimTime &x //In: SpSimTime which we will copy into *this
	  ) :
    Time(x.Time),
    Priority1(x.Priority1),
    Priority2(x.Priority2),
    Counter(x.Counter),
    UniqueId(x.UniqueId) {}

/******************************************************************************
* Method: SetValues
* Summary: Set the five values in the SpSimTime object
* Description: The five values are 1. physical time, 2. first priority
*              field, 3. second priority field, 4. the automatic tie-breaking
*              counter, and 5. the unique id for breaking counter ties
* Return: None
******************************************************************************/
  void SetValues(double t=DefaultTime,    // physical time
		 int p1=DefaultPriority1, // first priority field
		 int p2=DefaultPriority2,  // second priority field
		 int counter=DefaultCounter,  // counter field
		 int uniqueId=DefaultUniqueId // unique id field
		 ) {
    Time = t;
    Priority1 = p1;
    Priority2 = p2;
    Counter = counter;
    UniqueId = uniqueId;
  }

/******************************************************************************
* Method: print
* Summary: Print the SpSimTime object's values
* Description: 
* Return: None
******************************************************************************/
  void print(ostream &out = cout  // ostream object to which to print
	     ) {
    out << "SpSimTime: Time = " << Time
	<< ", Priority1 = " << Priority1
	<< ", Priority2 = " << Priority2
	<< ", Counter = "   << Counter
	<< ", UniqueId = "  << UniqueId
	<< endl;
  }

/******************************************************************************
* Method: IncrementPriority1
* Summary: Add a specified amount to the first priority field (default=1)
* Description: 
* Return: A reference to this SpSimTime object
******************************************************************************/
  SpSimTime &IncrementPriority1(int value = 1 // amount to add to first priority field
			  ) {Priority1 += value; return *this;}

/******************************************************************************
* Method: IncrementPriority2
* Summary: Add a specified amount to the second priority field (default=1)
* Description: 
* Return: A reference to this SpSimTime object
******************************************************************************/
  SpSimTime &IncrementPriority2(int value = 1 // amount to add to first priority field
			  ) {Priority2 += value; return *this;}

/******************************************************************************
* Method: DecrementPriority1
* Summary: Subtract a specified amount from the first priority field (default=1)
* Description: 
* Return: A reference to this SpSimTime object
******************************************************************************/
  SpSimTime &DecrementPriority1(int value = 1 // amount to subtract from first
			                // priority field
			  ) {Priority1 -= value; return *this;}

/******************************************************************************
* Method: DecrementPriority2
* Summary: Subtract a specified amount from the second priority field (default=1)
* Description: 
* Return: A reference to this SpSimTime object
******************************************************************************/
  SpSimTime &DecrementPriority2(int value = 1 // amount to subtract from second
			                // priority field
			  ) {Priority2 -= value; return *this;}

/******************************************************************************
* Method: GetTime
* Summary: Get the physical time
* Description: 
* Return: The physical time
******************************************************************************/
  double GetTime() const {return Time;}

/******************************************************************************
* Method: SetTime
* Summary: Set the physical time
* Description: 
* Return: None
******************************************************************************/
  void SetTime(double t  // physical time
	       ) {Time=t;}

/******************************************************************************
* Overloaded Method: SetPriority (same as SetPriority1)
* Summary: Set the first priority field
* Description: 
* Return: None
******************************************************************************/
  void SetPriority(int p1 // first priority field value
		   ) {Priority1 = p1;}

/******************************************************************************
* Overloaded Method: SetPriority
* Summary:  Set the first and second priority fields
* Description: 
* Return: None
******************************************************************************/
  void SetPriority(int p1, // first priority field value
		   int p2  // second priority field value
		   ) {Priority1 = p1; Priority2 = p2;}

/******************************************************************************
* Method: GetPriority1
* Summary: Get the first priority field
* Description: 
* Return: The first priority field value
******************************************************************************/
  int GetPriority1() const {return Priority1;}

/******************************************************************************
* Method: SetPriority1 (same as SetPriority)
* Summary: Set the first priority field
* Description: 
* Return: None
******************************************************************************/
  void SetPriority1(int p // first priority field value
		    ) {Priority1 = p;}

/******************************************************************************
* Method: GetPriority2
* Summary: Get the second priority field
* Description: 
* Return: The second priority field 
******************************************************************************/
  int GetPriority2() const {return Priority2;}

/******************************************************************************
* Method: SetPriority2
* Summary: Set the second priority field
* Description: 
* Return: None
******************************************************************************/
  void SetPriority2(int p // second priority field value
		    ) {Priority2 = p;}

/******************************************************************************
* Method: GetCounter
* Summary: Get the counter field
* Description: 
* Return: The counter field 
******************************************************************************/
  int GetCounter() const {return Counter;}

/******************************************************************************
* Method: SetCounter
* Summary: Set the counter field
* Description: 
* Return: None
******************************************************************************/
  void SetCounter(int counter // counter field value
		    ) {if (Counter!=0x7fffffff) Counter = counter;}

/******************************************************************************
* Method: GetUniqueId
* Summary: Get the unique id field
* Description: 
* Return: The unique id field
******************************************************************************/
  int GetUniqueId() const {return UniqueId;}

/******************************************************************************
* Method: SetUniqueId
* Summary: Set the unique id field
* Description: 
* Return: None
******************************************************************************/
  void SetUniqueId(int uniqueId // counter field value
		    ) {if (UniqueId!=0x7fffffff) UniqueId = uniqueId;}

/******************************************************************************
* Method: operator double
* Summary: When used as a primitive, SpSimTime automatically converts to a double
*          representing its physical time
* Description: 
* Return: The physical time
******************************************************************************/
  operator double() const {return Time;}

/******************************************************************************
* Method: operator =
* Summary: When assigned to a double, SpSimTime assigns that double to its
*          physical time WITHOUT CHANGING ITS PRIORITIES
* Description: 
* Return: A reference to this SpSimTime object
******************************************************************************/
  SpSimTime& operator = (double t) {
    Time = t;
    return(*this);
  }

/******************************************************************************
* Method: operator +=
* Summary: When add-assigned to a double, SpSimTime adds that double to its
*          physical time WITHOUT CHANGING ITS PRIORITIES
* Description: 
* Return: A reference to this SpSimTime object
******************************************************************************/
  SpSimTime& operator += (double t) {
    Time += t;
    return(*this);
  }

/******************************************************************************
* Method: operator -=
* Summary: When subtract-assigned to a double, SpSimTime subtracts that double
*          from its physical time WITHOUT CHANGING ITS PRIORITIES
* Description: 
* Return: A reference to this SpSimTime object
******************************************************************************/
  SpSimTime& operator -= (double t) {
    Time -= t;
    return(*this);
  }

/******************************************************************************
* Method: operator =
* Summary: When assigned to another SpSimTime, this SpSimTime undergoes an
*          element-by-element copy, INCLUDING CHANGING ITS PRIORITIES
* Description: 
* Return: A reference to this SpSimTime object
******************************************************************************/
  SpSimTime& operator = (const SpSimTime &tt) {
    Time = tt.GetTime();
    Priority1 = tt.GetPriority1();
    Priority2 = tt.GetPriority2();
    Counter = tt.GetCounter();
    UniqueId = tt.GetUniqueId();
    return(*this);
  }

/******************************************************************************
* Method: operator +=
* Summary: When add-assigned to another SpSimTime, SpSimTime adds that double to
*          its physical time INCLUDING CHANGING ITS PRIORITIES
* Description: 
* Return: A reference to this SpSimTime object
******************************************************************************/
  SpSimTime& operator += (SpSimTime &tt) {
    Time += tt.GetTime();
    return(*this);
  }

/******************************************************************************
* Method: operator -=
* Summary: When substract-assigned to another SpSimTime, SpSimTime subtracts that
*          double from its physical time INCLUDING CHANGING ITS PRIORITIES
* Description: 
* Return: A reference to this SpSimTime object
******************************************************************************/
  SpSimTime& operator -= (SpSimTime &tt) {
    Time -= tt.GetTime();
    return(*this);
  }

/******************************************************************************
* Method: operator <
* Summary: Perform a less-than logical time comparison between this SpSimTime and
*          another SpSimTime
* Description: That this is a logical time comparison means that the physical
*              time is compared and if they are equal, the first priority
*              fields are compared and if they are equal, the second priority
*              fields are compared
* Return: Whether this SpSimTime is earlier than the specified SpSimTime in
*         logical time
******************************************************************************/
  int operator < (const SpSimTime& tt) const {
    int retVal = 1;

    for (;;) {

      if (Time < tt.GetTime()) break;
      if (Time > tt.GetTime()) {
	retVal = 0;
	break;
      }

      if (Priority1 < tt.GetPriority1()) break;
      if (Priority1 > tt.GetPriority1()) {
	retVal = 0;
	break;
      }

      if (Priority2 < tt.GetPriority2()) break;
      if (Priority2 > tt.GetPriority2()) {
	retVal = 0;
	break;
      }

      if (Counter < tt.GetCounter()) break;
      if (Counter > tt.GetCounter()) {
	retVal = 0;
	break;
      }

      if (UniqueId < tt.GetUniqueId()) break;
      if (UniqueId > tt.GetUniqueId()) {
	retVal = 0;
	break;
      }

      retVal = 0;

      break;
    }

    return retVal;
  }

/******************************************************************************
* Method: operator >
* Summary: Perform a greater-than logical time comparison between this SpSimTime
*          and another SpSimTime
* Description: That this is a logical time comparison means that the physical
*              time is compared and if they are equal, the first priority
*              fields are compared and if they are equal, the second priority
*              fields are compared
* Return: Whether this SpSimTime is later than the specified SpSimTime in
*         logical time
******************************************************************************/
  int operator > (const SpSimTime& tt) const {
    int retVal = 1;

    for (;;) {

      if (Time > tt.GetTime()) break;
      if (Time < tt.GetTime()) {
	retVal = 0;
	break;
      }

      if (Priority1 > tt.GetPriority1()) break;
      if (Priority1 < tt.GetPriority1()) {
	retVal = 0;
	break;
      }

      if (Priority2 > tt.GetPriority2()) break;
      if (Priority2 < tt.GetPriority2()) {
	retVal = 0;
	break;
      }

      if (Counter > tt.GetCounter()) break;
      if (Counter < tt.GetCounter()) {
	retVal = 0;
	break;
      }

      if (UniqueId > tt.GetUniqueId()) break;
      if (UniqueId < tt.GetUniqueId()) {
	retVal = 0;
	break;
      }

      retVal = 0;

      break;
    }

    return retVal;

  }

/******************************************************************************
* Method: operator <=
* Summary: Perform a less-than-or-equal logical time comparison between this
*          SpSimTime and another SpSimTime
* Description: That this is a logical time comparison means that the physical
*              time is compared and if they are equal, the first priority
*              fields are compared and if they are equal, the second priority
*              fields are compared
* Return: Whether this SpSimTime is earlier or equal to the specified SpSimTime in
*         logical time
******************************************************************************/
  int operator <= (const SpSimTime& tt) const { return( !(*this > tt) ); }

/******************************************************************************
* Method: operator >=
* Summary: Perform a greater-than-or-equal logical time comparison between this
*          SpSimTime and another SpSimTime
* Description: That this is a logical time comparison means that the physical
*              time is compared and if they are equal, the first priority
*              fields are compared and if they are equal, the second priority
*              fields are compared
* Return: Whether this SpSimTime is later or equal to the specified SpSimTime in
*         logical time
******************************************************************************/
  int operator >= (const SpSimTime& tt) const { return( !(*this < tt) ); }

/******************************************************************************
* Method: operator ==
* Summary: Perform an equality logical time comparison between this SpSimTime and
*          another SpSimTime
* Description: That this is a logical time comparison means that the physical
*              time is compared and if they are equal, the first priority
*              fields are compared and if they are equal, the second priority
*              fields are compared
* Return: Whether this SpSimTime is equal to the specified SpSimTime in logical
*         time
******************************************************************************/
  int operator == (const SpSimTime& tt) const {

    int retVal = 1;

    if (Time != tt.GetTime()) {
      retVal = 0;
    }else{
      if (Priority1 != tt.GetPriority1()) {
	retVal = 0;
      }else{
	if (Priority2 != tt.GetPriority2()) {
	  retVal = 0;
	}else{
	  if (Counter != tt.GetCounter()) {
	    retVal = 0;
	  }else{
	    if (UniqueId != tt.GetUniqueId()) retVal = 0;
	  }
	}
      }
    }

    return retVal;

  }

  void DisableAutoTieBreaking(void) {
    Counter  = 0x7fffffff;
    UniqueId = 0x7fffffff;
  }


/******************************************************************************
* Method: operator !=
* Summary: Perform a not-equal logical time comparison between this SpSimTime and
*          another SpSimTime
* Description: That this is a logical time comparison means that the physical
*              time is compared and if they are equal, the first priority
*              fields are compared and if they are equal, the second priority
*              fields are compared
* Return: Whether this SpSimTime is not equal to the specified SpSimTime in logical
*         time
******************************************************************************/
  int operator != (const SpSimTime& tt) const {return !(*this == tt); }

/*
** Comparison operators between a SpSimTime and a double
*/

    int operator < (double t) const {return (Time < t);}
    int operator > (double t) const {return (Time > t);}
    int operator <= (double t) const {return (Time <= t);}
    int operator >= (double t) const {return (Time >= t);}
    int operator == (double t) const {return (Time == t);}
    int operator != (double t) const {return (Time != t);}

 protected:

  NET_DOUBLE Time;	// physical time
  NET_INT Priority1;	// First tie breaker
  NET_INT Priority2;	// Second tie breaker
  NET_INT Counter;      // Counter for automatic tie-breaking
  NET_INT UniqueId;     // Unique id for breaking Counter ties

};

/******************************************************************************
* Method: operator <<
* Summary: Print the comtents of this SpSimTime to an ostream
* Description: 
* Return: The specified ostream
******************************************************************************/
inline ostream& operator << (ostream &out,     // the ostream to which to print
			     const SpSimTime& tt // SpSimTime to print
			     ) {

  out << "Time = {" << tt.GetTime()
      << "," << tt.GetPriority1()
      << "," << tt.GetPriority2()
      << "," << tt.GetCounter()
      << "," << tt.GetUniqueId()
      << "}";

  return out;

}

/*
** Operators when the double is on the left side
*/

inline int operator < (double t, const SpSimTime& tt) { return (tt > t);}
inline int operator <= (double t, const SpSimTime& tt) { return (tt >= t);}
inline int operator > (double t, const SpSimTime& tt) { return (tt < t);}
inline int operator >= (double t, const SpSimTime& tt) { return (tt <= t);}
inline int operator == (double t, const SpSimTime& tt) { return (tt == t);}
inline int operator != (double t, const SpSimTime& tt) { return (tt != t);}

typedef SpSimTime SIMTIME;

#endif
