// Filename: SpEvent.H

#ifndef event_object
#define event_object

class C_HOST_USER;

#include "SpEventStatus.H"
#ifdef USING_FAST_FORWARD_DECLARATIONS

#include "sq_item.H"
#include "SpAltMgr.H"

class C_DBL_ITEM;
class C_DBL_QUEUE;
class SpMasterSimObjMgr;
class C_RANDOM;
class SpSimObj;

#else

#include "SpMasterSimObjMgr.H"

#endif

class SpEventTypes;
#include "SpMsg.H"
#include "lptr.H"
#include "SpTag.H"

/*****************************************************************************
* Class: SpEvent
* Summary: A generic simulation event
* Description: This is the fundamental active building block in SPEEDES
*              applications that acts on SpSimObjs. It is also the way SPEEDES
*              passes simulated time. SpEvents are discrete events are designed
*              to act on a single class (type) of simulation object, ie, a
*              particular descendent of SpSimObj. Events are generally also
*              responsible for scheduling other events. To achieve maximum
*              reuse, SpEvents should provide all of the actions and
*              associations between SpSimObjs.
******************************************************************************/
class SpEvent : public C_SQ_ITEM {

 public:

/******************************************************************************
* Virtual Method: init 
* Summary: Initialize this event
* Description: This is for the user to write code that gets executed in this
*              event's constructor. That is, code in this method will not occur
*              in the process virtual method phase, and thus will not get
*              rolled back.
* Return: None
******************************************************************************/
  virtual void init(SpMsg *   // the framework passes in the message for the
		                 // event here, and its use is optional
                    ) {}

/******************************************************************************
* Virtual Method: exchange 
* Summary: This is for the user to write exchange code for fast rollbacks.
* Description: This is usually not required, since rollbackable items in the
*              process virtual method essentially undergo exchanges
*              automatically. But if rollbackable items are not available for
*              certain state data, this method allows the user to write
*              low-level code to exchange a value with a saved previous
*              value. The SPEEDES framework calls this method whenever
*              performing a rollback.
* Return: None 
******************************************************************************/
  virtual void exchange() {;}

/******************************************************************************
* Virtual Method: temp_process 
* Summary: The main process user code block (same as process())
* Description: This method is essentially always overridden for every SpEvent.
*              That is because it is the place where the body of user code is
*              written for the event. It is the only virtual method wherein
*              rollbackable operations (such as using STATE variables, etc.)
*              are permitted. State information that is not rollbackable should
*              not be used in this method, since this method is run more than
*              once whenever it is rolled back and reprocessed. Also, calls to
*              outside of the SPEEDES framework should not go here for the same
*              reason, and should instead occur in the commit virtual function.
* Return: None
******************************************************************************/
  virtual void temp_process() {process();}

/******************************************************************************
* Virtual Method: process 
* Summary: The main process user code block (same as temp_process())
* Description: The user overrides this method for essentially every SpEvent.
*              That is because it is the place where the body of user code is
*              written for the event. It is the only virtual method wherein
*              rollbackable operations (such as using STATE variables, etc.)
*              are permitted. State information that is not rollbackable should
*              not be used in this method, since this method is run more than
*              once whenever it is rolled back and reprocessed. Also, calls to
*              outside of the SPEEDES framework should not go here for the same
*              reason, and should instead occur in the commit virtual function.
* Return: None
******************************************************************************/
  virtual void process() {;}

/******************************************************************************
* Virtual Method: perm_process
* Summary: The part of the user process code that is run after the event is
*          committed (same as commit())
* Description: The user should override this method whenever they want code to
*              be guaranteed not to rollback. This is achieved by the SPEEDES
*              framework by calling this method after the event has been
*              committed (ie, after it has passed global virtual time and thus
*              won't ever again rollback). This method is overridden whenever
*              the user needs to call code outside of the SPEEDES framework and
*              calling it more than once is undesired (such as calling screen
*              graphics). An alternative to using commit() is to extend the
*              SPEEDES framework with rollbackable items that can thus be used
*              within the process() method.
* Return: None
******************************************************************************/
  virtual void perm_process() {commit();}

/******************************************************************************
* Virtual Method: commit 
* Summary: The part of the user process code that is run after the event is
*          committed (same as perm_process())
* Description: The user should override this method whenever they want code to
*              be guaranteed not to rollback. This is achieved by the SPEEDES
*              framework by calling this method after the event has been
*              committed (ie, after it has passed global virtual time and thus
*              won't ever again rollback). This method is overridden whenever
*              the user needs to call code outside of the SPEEDES framework and
*              calling it more than once is undesired (such as calling screen
*              graphics). An alternative to using commit() is to extend the
*              SPEEDES framework with rollbackable items that can thus be used
*              within the process() method.
* Return: None
******************************************************************************/
  virtual void commit() {;}

/******************************************************************************
* Virtual Method: lazy 
* Summary: User code written to determine, each time this event rolls back,
*          whether to roll forward (and thus avoid the overhead of sending
*          antimessages and reprocessing) or to undergo the normal rollback
*          operations.
* Description: Often when an event gets rolled back and reproccessed, the
*              results of the simulation are the same (or nearly the same) as
*              if the framework didn't roll it back at all even though it was
*              processed out of simulated time order. This is user code to
*              determine whether this event should send out antimessages when
*              it gets rolled back (like a regular rollback) or whether, after
*              rolling back, tests show that nothing critical would change by
*              sending out antimessages and thus it can simply roll forward
*              back to where it was before it got rolled back.
* Return: flag for whether the lazy test passed (non-zero = pass). If it
*         passes, and if lazy cancellation is enabled, this event is rolled
*         forward without sending out antimessages. If it fails, it is treated
*         just like any other rollback.
******************************************************************************/
  virtual int lazy() {return 0;}

/******************************************************************************
* Virtual Method: cleanup 
* Summary: User-written code called in the event destructor
* Description: 
* Return: None
******************************************************************************/
  virtual void cleanup() {;}

/******************************************************************************
* Method: event_type
* Summary: Get the enumerated ID for an event type, specified by the event
*          type's string name
* Description: 
* Return: the enumerated ID for an event type
******************************************************************************/
  int event_type(char *c);

/******************************************************************************
* Method: object_type
* Summary: Retrieves a simulation object's OBJECT_TYPE (see set_OBJECT_TYPE)
*          based on its string NAME
* Description: 
* Return: the object type of this simulation object
******************************************************************************/
  int object_type(char *c);

/******************************************************************************
* Method: get_NODE 
* Summary: Get the node (process number on local machine) that this SpSimObj is
*          instantialted on. 
* Description: 
* Return: the node (process number on local machine) that this SpSimObj
*         is instantialted on.
******************************************************************************/
  int get_NODE() {return NODE;}

/******************************************************************************
* Method: get_N_NODES
* Summary: Get the total number of nodes (process number on local machine) in
*          the entire simulation (including all other machines' process)
* Description: 
* Return: total number of nodes
******************************************************************************/
  int get_N_NODES() {return N_NODES;}

/******************************************************************************
* Method: get_SIMOBJ
* Summary: Get the simulation object (SpSimObj) associated with this event
* Description: This is a fundamental method for SpEvent, since events are
*              designed to act upon one type of simulation object, ie one class
*              that inherits from SpSimObj. Use this method to read and write
*              to the simulation object on which the event is acting.
* Return: the simulation object associated with this event
******************************************************************************/
  SpSimObj *get_SIMOBJ() {return SIMOBJ;}

/******************************************************************************
* Method: get_MESSAGE 
* Summary: Get the message used to enact this event
* Description: This is a fundamental method for SpEvent, since events are
*              enacted by a message, and that message often contains
*              information intended to be extracted by its effected event. Use
*              this method get and then read the contents of the message.
* Return: the message used to enact this event
******************************************************************************/
  SpMsg *get_MESSAGE() {return MESSAGE;}

/******************************************************************************
* Method: get_DATA 
* Summary: Get the data used to enact this event
* Description: This is the variable length data.  When this event was
*              scheduled, if the schedule or ScheduleEvent function was
*              passed a char * buffer, then this function will return a
*              pointer to a copy of that buffer, else it returns NULL.
* Return: the data used to enact this event, or NULL if no such data exists.
******************************************************************************/
  char *get_DATA() {return DATA;}


/******************************************************************************
* Method: get_OBJECT_TYPE 
* Summary: Get the ID (or enumerated type) of the type of the simulation
*          object associated with this event
* Description: A unique OBJECT_TYPE is required for each SpSimObj class.  Each
*              SpSimObj class is managed by one instantiation of
*              SpSimObjMgr. Thus, the OBJECT_TYPE also uniquely defines the type
*              of SpSimObj that each instantiation of SpSimObjMgr manages. Each
*              node has a single instantiation of SpMasterSimObjMgr, which in turn
*              manages a list of SpSimObjMgrs. SpMasterSimObjMgr refers to each SpSimObjMgr
*              by the OBJECT_TYPE it manages.
* Return: the OBJECT_TYPE, ie, the ID of the type of this simulation type
******************************************************************************/
  int get_OBJECT_TYPE() {return OBJECT_TYPE;}


/******************************************************************************
* Method: get_EVENT_TYPE 
* Summary: Get the ID (or enumerated type) of the type of this event
* Description: A unique EVENT_TYPE is required for each SpEvent class, and is
*              found in event_types.C
* Return: the OBJECT_TYPE, ie, the ID of the type of this simulation type
******************************************************************************/
  int get_EVENT_TYPE() {return EVENT_TYPE;}

/******************************************************************************
* Method: get_GLOBAL_ID 
* Summary: Get the unique ID, simulaiton-wide, of the object associated with
*          this event. GLOBAL_ID generated automatically, but can be overridden
*          by the user.
* Description: 
* Return: the unique ID of this simulation object
******************************************************************************/
  int get_GLOBAL_ID() {return GLOBAL_ID;}

/******************************************************************************
* Method: get_LOCAL_ID 
* Summary: Get the ID of the simulation object associated with this event
* Description: Each object manager has its own set of SpSimObjs, each with
*              their own LOCAL_ID. Thus LOCAL_ID only needs to be unique for a
*              given object manager (SpSimObjMgr), not unique to a given node or
*              to a given simulation
* Return: the ID of this simulation object
******************************************************************************/
  int get_LOCAL_ID() {return LOCAL_ID;}

/******************************************************************************
* Method: get_RANDOM 
* Summary: Get the random number generator. There is exactly one C_RANDOM
*          object per node, set automatically by the framework.
* Description: see C_RANDOM
* Return: the random number generator
******************************************************************************/
  C_RANDOM *get_RANDOM() {return RANDOM;}

/******************************************************************************
* Method: get_manager 
* Summary: Get the object manager for another simulation object, specified by
*          its OBJECT_TYPE
* Description: 
* Return: an object manager, casted to (void *)
******************************************************************************/
  void *get_manager(int om = -2 // IN: OBJECT_TYPE of the desired object
		                //     manager. Default is this event's
		                //     simobj's object manager.
		    );

/******************************************************************************
* Method: memory_print 
* Summary: print out event free list memory statistics
* Description: 
* Return: None
******************************************************************************/
  void memory_print();

/******************************************************************************
* Method: print_event_memory 
* Summary: print out event free list memory statistics
* Description: 
* Return: None
******************************************************************************/
  void print_event_memory();

/******************************************************************************
* Overloaded Method: create_object
* Summary: Schedule for dynamically creating a simulation object (SpSimObj)
* Description: This is simply a schedule call for a CREATE event. Like all
*              schedule calls, it returns the associated message.
* Return: the message for dynamically created simulation object (SpSimObj)
******************************************************************************/
  SpMsg *create_object(SpSimTime& time,   // IN: simulated time to create the
		                       // object
		      int object_type, // IN: type of object to create 
		      int object_node  // IN: node on which to create the
		                       // object
		      );
/******************************************************************************
* Overloaded Method: create_object
* Summary: Schedule for dynamically creating a simulation object (SpSimObj)
* Description: This is simply a schedule call for a CREATE event. Like all
*              schedule calls, it returns the associated message.
* Return: the message for dynamically created simulation object (SpSimObj)
******************************************************************************/
  SpMsg *create_object(void *ch,        // cancel handle (see schedule method)
		      SpSimTime& time,   // IN: simulated time to create the
		                       // object
		      int object_type, // IN: type of object to create 
		      int object_node  // IN: node on which to create the
		                       // object
		      );

/******************************************************************************
* Overloaded Method: create_object
* Summary: Schedule for dynamically creating a simulation object (SpSimObj)
* Description: This is simply a schedule call for a CREATE event. Like all
*              schedule calls, it returns the associated message.
* Return: the message for dynamically created simulation object (SpSimObj)
******************************************************************************/
  SpMsg *create_object(SpSimTime& time,   // IN: simulated time to create the
		                       // object
		      int object_type, // IN: type of object to create 
		      int object_node, // IN: node on which to create the
		                       // object
		      int mess_size,   // IN: (OPTIONAL) number of bytes of
	                               // the variable-length part of the
	                               // event message
		      char *&buf       // IN: (USED IF AND ONLY IF "mess_size"
	                               // IS USED) the variable-length part of
	                               // the event message 
		      );

/******************************************************************************
* Overloaded Method: create_object
* Summary: Schedule for dynamically creating a simulation object (SpSimObj)
* Description: This is simply a schedule call for a CREATE event. Like all
*              schedule calls, it returns the associated message.
* Return: the message for dynamically created simulation object (SpSimObj)
******************************************************************************/
  SpMsg *create_object(void *ch,        // cancel handle (see schedule method)
		      SpSimTime& time,   // IN: simulated time to create the
		                       // object
		      int object_type, // IN: type of object to create 
		      int object_node, // IN: node on which to create the
		                       // object
		      int mess_size,   // IN: (OPTIONAL) number of bytes of
	                               // the variable-length part of the
	                               // event message
		      char *&buf       // IN: (USED IF AND ONLY IF "mess_size"
	                               // IS USED) the variable-length part of
	                               // the event message 
		      );

/******************************************************************************
* Overloaded Method: destroy_object 
* Summary: Schedule for dynamically destroying (deleting) a simulation object
*          (SpSimObj)
* Description: This is simply a schedule call for a DESTROY event. Like all
*              schedule calls, it returns the associated message.
* Return: the message for dynamically destroyed simulation object (SpSimObj)
******************************************************************************/
  SpMsg *destroy_object(SpSimTime& time,   // IN: simulated time to destroy the
		                        // object
		       int object_type, // IN: type of object to destroy
		       int object_id,   // IN: the LOCAL_ID of the object to
		                        // destroy
		       int object_node  // IN: node on which to destroy the
		                        // object
		       );

/******************************************************************************
* Overloaded Method: destroy object
* Summary: Schedule for dynamically destroying (deleting) a simulation object
*          (SpSimObj)
* Description: This is simply a schedule call for a DESTROY event. Like all
*              schedule calls, it returns the associated message.
* Return: the message for dynamically destroyed simulation object (SpSimObj)
******************************************************************************/
  SpMsg *destroy_object(void *ch,        // cancel handle (see schedule method)
		       SpSimTime& time,   // IN: simulated time to destroy the
		                        // object
		       int object_type, // IN: type of object to destroy
		       int object_id,   // IN: the LOCAL_ID of the object to
		                        // destroy
		       int object_node  // IN: node on which to destroy the
		                        // object
		       );

/******************************************************************************
* Overloaded Method: schedule
* Summary: Schedule an event at the present or a future simulated time for any
*          simulation object (SpSimObj) in the simulation. This version of
*          schedule () allows the event's message to contain a variable-length
*          character buffer as well as the variables contained in the inherited
*          SpMsg.
* Description: This is a fundamental method for SpEvent, since events
*              are generally responsible for scheduling future
*              events. The message for this event schedule is returned
*              (SpMsg*), and should be subsequently filled in. This
*              is OK, since the message isn't sent until after the
*              event user code (the process() virtual method)
*              completes.  The buf argument passed in need not be
*              initialized.  It is a return by reference argument.
*              After schedule returns, the buf argument will point to
*              a buffer of size mess_size, and you may then write
*              data into it.  Write into it AFTER schedule returns!
* Return: the message for this scheduled event (SpMsg*)
******************************************************************************/
  SpMsg *schedule(const SpSimTime& time, // IN: simulated time 
		 int event_type,      // IN: enumerated or integer type of
	                              // event to schedule
		 int object_type,     // IN: enumerated or integer type of
	                              // object for which to schedule the
	                              // event
	                              // (must be the one type of object
	                              // required for the event_type) 
	         int object_id,       // IN: the LOCAL_ID of the object for
	                              // which to schedule the event
	         int object_node     // IN: the NODE of the object for which
	         );

/******************************************************************************
* Overloaded Method: schedule
* Summary: Schedule an event at the present or a future simulated time for any
*          simulation object (SpSimObj) in the simulation. This version of
*          schedule () allows the event's message to contain a variable-length
*          character buffer as well as the variables contained in the inherited
*          SpMsg.
* Description: This is a fundamental method for SpEvent, since events
*              are generally responsible for scheduling future
*              events. The message for this event schedule is returned
*              (SpMsg*), and should be subsequently filled in. This
*              is OK, since the message isn't sent until after the
*              event user code (the process() virtual method)
*              completes.  The buf argument passed in need not be
*              initialized.  It is a return by reference argument.
*              After schedule returns, the buf argument will point to
*              a buffer of size mess_size, and you may then write
*              data into it.  Write into it AFTER schedule returns!
* Return: the message for this scheduled event (SpMsg*)
******************************************************************************/
  SpMsg *schedule(const SpSimTime& time, // IN: simulated time 
		 int event_type,      // IN: enumerated or integer type of
	                              // event to schedule
		 int object_type,     // IN: enumerated or integer type of
	                              // object for which to schedule the
	                              // event
	                              // (must be the one type of object
	                              // required for the event_type) 
	         int object_id,       // IN: the LOCAL_ID of the object for
	                              // which to schedule the event
	         int object_node,     // IN: the NODE of the object for which
	                              // to schedule the event
	         int mess_size,       // IN: (OPTIONAL) number of bytes of
	                              // the variable-length part of the
	                              // event message
	         char *&buf           // IN: (USED IF AND ONLY IF "size" IS
	                              // USED) the variable-length part of
	                              // the event message 
	         );

/******************************************************************************
* Overloaded Method: schedule
* Summary: Schedule a cancelable event at the present or a future simulated
*          time for any simulation object (SpSimObj) in the simulation. This
*          version of schedule () allows the event's message to contain a
*          variable-length character buffer as well as the variables contained
*          in the inherited SpMsg.
* Description: This is a fundamental method for SpEvent, since events are
*              generally responsible for scheduling future events. The message
*              for this event schedule is returned (inherited from SpMsg),
*              and should be subsequently filled in. This is OK, since the
*              message isn't sent until after the event user code (the
*              process() virtual method) completes. The cancel handle (first
*              argument) simply must be an uninitialized object (ie, the user
*              need only create it, not fill it in) of type SpCancelHandle (or
*              equivalently, SpAntiMsg, which is what it is typedef'd
*              with). This function returns the handle filled-in with all of
*              the information to use this handle later in the cancel () call,
*              to later cancel the event now being scheduled. The cancel ()
*              call, if later called by the user, will cancel the event now
*              being scheduled by scheduling a CANCEL event. When calling
*              cancel (), the event to cancel must be in the simulated time
*              future.
* Return: The message for this scheduled event (inherited from SpMsg)
******************************************************************************/
  SpMsg *schedule(void *ch,            // IN: cancel handle
		 const SpSimTime& time, // IN: simulated time 
		 int event_type,      // IN: enumerated or integer type of
	                              // event to schedule
		 int object_type,     // IN: enumerated or integer type of
	                              // object for which to schedule the
	                              // event
	                              // (must be the one type of object
	                              // required for the event_type) 
	         int object_id,       // IN: the LOCAL_ID of the object for
	                              // which to schedule the event
	         int object_node     // IN: the NODE of the object for which
	         );

/******************************************************************************
* Overloaded Method: schedule
* Summary: Schedule a cancelable event at the present or a future simulated
*          time for any simulation object (SpSimObj) in the simulation. This
*          version of schedule () allows the event's message to contain a
*          variable-length character buffer as well as the variables contained
*          in the inherited SpMsg.
* Description: This is a fundamental method for SpEvent, since events are
*              generally responsible for scheduling future events. The message
*              for this event schedule is returned (inherited from SpMsg),
*              and should be subsequently filled in. This is OK, since the
*              message isn't sent until after the event user code (the
*              process() virtual method) completes. The cancel handle (first
*              argument) simply must be an uninitialized object (ie, the user
*              need only create it, not fill it in) of type SpCancelHandle (or
*              equivalently, SpAntiMsg, which is what it is typedef'd
*              with). This function returns the handle filled-in with all of
*              the information to use this handle later in the cancel () call,
*              to later cancel the event now being scheduled. The cancel ()
*              call, if later called by the user, will cancel the event now
*              being scheduled by scheduling a CANCEL event. When calling
*              cancel (), the event to cancel must be in the simulated time
*              future.
* Return: The message for this scheduled event (inherited from SpMsg)
******************************************************************************/
  SpMsg *schedule(void *ch,            // IN: cancel handle
		 const SpSimTime& time, // IN: simulated time 
		 int event_type,      // IN: enumerated or integer type of
	                              // event to schedule
		 int object_type,     // IN: enumerated or integer type of
	                              // object for which to schedule the
	                              // event
	                              // (must be the one type of object
	                              // required for the event_type) 
	         int object_id,       // IN: the LOCAL_ID of the object for
	                              // which to schedule the event
	         int object_node,     // IN: the NODE of the object for which
	                              // to schedule the event
	         int mess_size,       // IN: (OPTIONAL) number of bytes of
	                              // the variable-length part of the
	                              // event message
	         char *&buf           // IN: (USED IF AND ONLY IF "size" IS
	                              // USED) the variable-length part of
	                              // the event message 
	         );

/******************************************************************************
* Method: cancel 
* Summary: Cancel the event specified by a SpCancelHandle
* Description: The event to be cancelled must have been scheduled with the
*              cancel handle used here (SpCancelHandle), and the event to
*              cancel must be in the simulated time future.
* Return: None.
******************************************************************************/
  void cancel(void *ch  // IN: the cancel handle (SpCancelHandle) of the
                        // event to cancel, which was the first argument in 
                        // that event's schedule () call
	      );

/******************************************************************************
* Method: trace 
* Summary: Trace this event
* Description: Provide statistics about this event to a trace file during run
*              time
* Return: None
******************************************************************************/
  void trace();

/******************************************************************************
* Method: set_trace 
* Summary: Enable tracing this event
* Description: 
* Return: None
******************************************************************************/
  void set_trace() {traceflag = 1;}

/******************************************************************************
* Method: reset_trace 
* Summary: Disable tracing this event
* Description: 
* Return: None
******************************************************************************/
  void reset_trace() {traceflag = 0;}

/******************************************************************************
* Method: open_tracefile 
* Summary: Open a file into which to output trace information (see trace())
* Description: 
* Return: None
******************************************************************************/
  void open_tracefile(char *filename  // filename for trace output
		      );

/******************************************************************************
* Method: WriteLookaheadfile 
* Summary: Write to the lookahead file, if enabled
* Description: 
* Return: None
******************************************************************************/
  void WriteLookaheadfile(const SpSimTime& t2,  // lookahead time
			  int ty2,            // second object EVENT_TYPE
			  int ot2,            // second object OBJECT_TYPE
			  int id2,            // second object LOCAL_ID
			  int nd              // second object NODE
			  );

/******************************************************************************
* Method: set_WriteLookahead 
* Summary: Enable write lookahead
* Description: 
* Return: None
******************************************************************************/
  void set_WriteLookahead() {WriteLookahead = 1;}

/******************************************************************************
* Method: reset_WriteLookahead 
* Summary: Disable write lookahead
* Description: 
* Return: None
******************************************************************************/
  void reset_WriteLookahead() {WriteLookahead = 0;}

/******************************************************************************
* Method: open_Lookaheadfile 
* Summary: Open a file into which to output lookahead information
* Description: 
* Return: None
******************************************************************************/
  void open_Lookaheadfile(char *filename);

/******************************************************************************
* Internal Method: enable_lazy 
* Summary: Enable lazy cancellation for the entire node
* Description: This is done at the start of the simulation if the speedes.par
*              file has lazy enablement specified
* Return: None
******************************************************************************/
  static void enable_lazy() {EnableLazy = 1;}

/******************************************************************************
* Method: set_lazy 
* Summary: Enable lazy cancellation
* Description: 
* Return: None
******************************************************************************/
  void set_lazy() {lazy_flag = 1;}

/******************************************************************************
* Method: reset_lazy 
* Summary: Disable lazy cancellation
* Description: 
* Return: None
******************************************************************************/
  void reset_lazy() {lazy_flag = 0;}

/******************************************************************************
* Method: GetObjectHandle
* Summary: Returns a reference to the current simobj's object handle
* Description: 
* Return: A reference to this simobj's object handle
******************************************************************************/
   SpObjHandle *GetObjectHandle();

/******************************************************************************
* Internal Method: get_lazy 
* Summary: Determine whether lazy cancellation is enabled
* Description: 
* Return: Whether lazy cancellation is enabled (1=yes)
******************************************************************************/
  int get_lazy() {return (lazy_flag & EnableLazy);}

/******************************************************************************
* Internal Method: lazy_rollback 
* Summary: Late roll back this event from lazy cancellation
* Description: This method does the required processing for an event that has
*              failed the lazy test. It has already undergone a "shallow"
*              rollback (meaning a rollback with no antimessages), so this
*              method handles the antimessage part.
* Return: Current number of rollbacks
******************************************************************************/
  int lazy_rollback();

/******************************************************************************
* Internal Method: reset_lazy_proc 
* Summary: Reset the flag for whether this event has been lazy-processed
* Description: 
* Return: None
******************************************************************************/
  void reset_lazy_proc() {lazy_proc = 0;}

/******************************************************************************
* Internal Method: get_lazy_proc 
* Summary: Get the flag for whether this event has been lazy-processed
* Description: Time management algorithms, when pulling events off of the event
*              queue for processing, must know whether the event has already
*              been processed, rolled back without sending antimessages, and is
*              waiting for a lazy test (meaning they have been
*              lazy-processed). In this case the algorithm must handle the
*              event differently than if the event is simply waiting to be
*              processed (the usual case). This method determines which case
*              the event falls under.
* Return: Flag for whether this event has been lazy-processed
******************************************************************************/
  int get_lazy_proc() {return lazy_proc;}

/******************************************************************************
* Internal Method: check_proc 
* Summary: Check if the event is in the processed events queue
* Description: This is used for lazy cancellation
* Return: Flag for whether the event is in the processed events queue
******************************************************************************/
  int check_proc(int flag);

/******************************************************************************
* Internal Method: check_clear 
* Summary: Check if the event is in a state appropriate to be processed. If 
*          not, this is a run-time error.
* Description: 
* Return: Flag for whether it is safe to process this event
******************************************************************************/
  int check_clear();

/******************************************************************************
* Internal Method: set_MESSAGE 
* Summary: Set the message used to enact this event
* Description: 
* Return: None
******************************************************************************/
  void set_MESSAGE(SpMsg *h) {MESSAGE=h;}
 
/******************************************************************************
* Internal Method: set_DATA 
* Summary: Set the data used to enact this event
* Description: 
* Return: None
******************************************************************************/
  void set_DATA(char *c) {DATA=c;}

/******************************************************************************
* Internal Method: set_OBJECT_TYPE
* Summary: Set the ID (or enumerated type) of the type of the simulation
*          object associated with this event
* Description: 
* Return: None
******************************************************************************/
  void set_OBJECT_TYPE(int ot) {OBJECT_TYPE = ot;}

/******************************************************************************
* Internal Method: set_EVENT_TYPE 
* Summary: Set the ID (or enumerated type) of the type of this event
* Description: 
* Return: None
******************************************************************************/
  void set_EVENT_TYPE(int et) {EVENT_TYPE = et;}

/******************************************************************************
* Internal Method: set_GLOBAL_ID 
* Summary: Set the unique ID, simulaiton-wide, of the object associated with
*          this event. GLOBAL_ID generated automatically, but can be overridden
*          by the user.
* Description: 
* Return: None
******************************************************************************/
  void set_GLOBAL_ID(int g) {GLOBAL_ID = g;}

/******************************************************************************
* Internal Method: set_LOCAL_ID 
* Summary: Set the ID of the simulation object associated with this event
* Description: 
* Return: None
******************************************************************************/
  void set_LOCAL_ID(int l) {LOCAL_ID = l;}

/******************************************************************************
* Internal Method: set_RANDOM 
* Summary: Set the random number generator. There is exactly one C_RANDOM
*          object per node, set automatically by the framework with this method.
* Description: 
* Return: None
******************************************************************************/
  void set_RANDOM(C_RANDOM *ran) {RANDOM = ran;}

/******************************************************************************
* Internal Method: set_HOST_USER 
* Summary: Set the host user
* Description: 
* Return: None
******************************************************************************/
  void set_HOST_USER(C_HOST_USER *h) {HOST_USER = h;}

/******************************************************************************
* Internal Method: set_objects 
* Summary: Set the SpMasterSimObjMgr object for this node
* Description: 
* Return: None
******************************************************************************/
  void set_objects(SpMasterSimObjMgr *obj) {objects = obj;}

/******************************************************************************
* Internal Method: set_evtype 
* Summary: Set the SpEventTypes object for this node
* Description: 
* Return: None
******************************************************************************/
  void set_evtype(SpEventTypes *et) {evtype = et;}

/******************************************************************************
* Internal Method: set_free 
* Summary: Set the state of this object to free, meaning it is a freed object 
*          in the free list, able to be reused at a later time (see FREELST)
* Description: Used only by freelist management
* Return: None
******************************************************************************/
  void set_free() {free_flag = 1;}

/******************************************************************************
* Internal Method: set_processed 
* Summary: Set the state of this event to processed, meaning it has been
*          processed (the user method process() has been called) by the
*          framework
* Description: 
* Return: None
******************************************************************************/
  void set_processed();

/******************************************************************************
* Internal Method: set_rbid 
* Summary: Set the roll back unique ID
* Description: 
* Return: None
******************************************************************************/
  void set_rbid(int r) {rbid = r;}

/******************************************************************************
* Internal Method: set_cancel_id 
* Summary: Set the cancel event ID
* Description: 
* Return: None
******************************************************************************/
  void set_cancel_id(int cid) {cancel_id = cid;}

/******************************************************************************
* Internal Method: set_valid_opt 
* Summary: Set the state of the event to valid optimistic
* Description: 
* Return: None
******************************************************************************/
  void set_valid_opt() {valid = 2;}

/******************************************************************************
* Internal Method: set_valid 
* Summary: Set the state of the event to valid non-optimistic
* Description: 
* Return: None
******************************************************************************/
  void set_valid() {valid = 1;}

/******************************************************************************
* Internal Method: set_event_canceled 
* Summary: Set the state of this event to canceled, meaning when the framework
*          pulls it off the queue, it will discard it rather than process it
* Description: 
* Return: None
******************************************************************************/
  void set_event_canceled() {event_canceled = 1;}

/******************************************************************************
* Internal Method: set_lazy_proc 
* Summary: Set the state of this event to lazy-processed
* Description: Time management algorithms, when pulling events off of the event
*              queue for processing, must know whether the event has already
*              been processed, rolled back without sending antimessages, and is
*              waiting for a lazy test (meaning they have been
*              lazy-processed). In this case the algorithm must
*              handle the event differently than if the event is simply waiting
*              to be processed (the usual case).
* Return: None
******************************************************************************/
  void set_lazy_proc() {lazy_proc = 1;}

/******************************************************************************
* Internal Internal Method: set_NODE 
* Summary: Set the node (process number on local machine) that this SpSimObj is
*          instantialted on.
* Description: 
* Return: None
******************************************************************************/
  void set_NODE(int nd) {NODE = nd; gencid = NODE;}

/******************************************************************************
* Internal Internal Method: set_N_NODES 
* Summary: Set the total number of nodes (process number on local machine) in
*          the entire simulation (including all other machines' process)
* Description: 
* Return: None
******************************************************************************/
  void set_N_NODES(int nds) {N_NODES = nds;}

/******************************************************************************
* Internal Internal Method: set_SIMOBJ 
* Summary: Set the simulation object (SpSimObj) associated with this event
* Description: This is a fundamental method for SpEvent, since events are
*              designed to act upon one type of simulation object, ie one class
*              that inherits from SpSimObj.
* Return: None
******************************************************************************/
  void set_SIMOBJ(SpSimObj *g) {SIMOBJ=g;}

/******************************************************************************
* Internal Method: set_qproc 
* Summary: Set the queue of processed events
* Description: 
* Return: None
******************************************************************************/
  static void set_qproc(C_DBL_QUEUE *q) {dbl_qproc = q;}

/******************************************************************************
* Internal Method: void set_one_node_flag 
* Summary: Set the one-node flag, ie whether the simulation is on one node (ie
*          sequential) or not (ie parallel)
* Description: 
* Return: None
******************************************************************************/
  static void set_one_node_flag(int onf) {one_node_flag = onf;}

/******************************************************************************
* Internal Method: get_objects 
* Summary: Get the SpMasterSimObjMgr object on this node
* Description: 
* Return: None
******************************************************************************/
  SpMasterSimObjMgr *get_objects() {return objects;}

/******************************************************************************
* Internal Method: void set_twosmess 
* Summary: Set the SpOptMessager object for this node
* Description: 
* Return: None
******************************************************************************/
  static void set_twosmess(void *tm);

/******************************************************************************
* Internal Method: insert
* Summary: Attach this event to its simulation object (SpSimObj)
* Description: 
* Return: None
******************************************************************************/
  void insert() {SIMOBJ->insert_event(this);}

/******************************************************************************
* Internal Method: get_free 
* Summary: Get the free flag (used for event list management)
* Description: 
* Return: The free flag
******************************************************************************/
  int get_free() {return free_flag;}

/******************************************************************************
* Internal Method: get_from_SpComm 
* Summary: Get the from_SpComm member which says if this event came from
*          SpComm.  It would be false if it came from this same node.
* Description: 
* Return: from_SpComm
******************************************************************************/
  int get_from_SpComm() {return from_SpComm;}

/******************************************************************************
* Method: set_from_SpComm
* Summary: Record the fact that this event came from SpComm.  Therefore
*          the proper way to deallocate this event is to use delete [] on
*          a char* type.
* Description: 
* Return: None
******************************************************************************/
  void set_from_SpComm() {from_SpComm = 1;}

/******************************************************************************
* Method: reset_from_SpComm
* Summary: Record the fact that this event DID NOT come from SpComm.
* Description: 
* Return: None
******************************************************************************/
  void reset_from_SpComm() {from_SpComm = 0;}

/******************************************************************************
* Internal Method: get_tmin
* Summary: Get the minimum time of a generated message
* Description: 
* Return: The minimum time of a generated message
******************************************************************************/
  SpSimTime &get_tmin() {return tmin;}

/******************************************************************************
* Internal Method: calculate_tmin
* Summary: Calculate the minimum time of a generated message
* Description: 
* Return: None
******************************************************************************/
  void calculate_tmin();

/******************************************************************************
* Internal Method: get_processed 
* Summary: Get whether this event has been processed
* Description: 
* Return: Flag for whether this event has been processed
******************************************************************************/
  int get_processed()
    {
      int rc = 0;
      if (!one_node_flag){
	rc = objevt.get_processed();
      }
      return rc;
    }
  
/******************************************************************************
* Internal Method: reset_processed 
* Summary: Tell this event that it has not been processed
* Description: 
* Return: None
******************************************************************************/
  void reset_processed() {objevt.reset_processed();}

/******************************************************************************
* Internal Method: pop_roll_back 
* Summary: Pop this event off the roll back queue
* Description: 
* Return: None
******************************************************************************/
  void pop_roll_back();

/******************************************************************************
* Internal Method: local_messages 
* Summary: Send out the local messages
* Description: 
* Return: Current number of rollbacks
******************************************************************************/
  int local_messages();

/******************************************************************************
* Internal Method: optimistic_insert 
* Summary: Optimistically insert all of this node's events into the event list
* Description: 
* Return: Current number of rollbacks
******************************************************************************/
  int optimistic_insert();

/******************************************************************************
* Internal Method: rollback 
* Summary: Roll back each event on this node if necessary
* Description: 
* Return: Current number of rollbacks
******************************************************************************/
  int rollback();

/******************************************************************************
* Internal Method: rollback_event 
* Summary: Roll back this event
* Description: 
* Return: Current number of rollbacks
******************************************************************************/
  int rollback_event();

/******************************************************************************
* Internal Method: cancel_event 
* Summary: Cancel this event
* Description: 
* Return: Current number of rollbacks
******************************************************************************/
  int cancel_event();

/******************************************************************************
* Internal Method: get_dbl_item 
* Summary: Get this event's processed events list item pointer
* Description: 
* Return: This event's processed events list item pointer
******************************************************************************/
  C_DBL_ITEM *get_dbl_item() {return &dbl_item;}

/******************************************************************************
* Internal Method: get_rbid 
* Summary: Get this event's rollback ID
* Description: 
* Return: This event's rollback ID
******************************************************************************/
  int get_rbid() {return rbid;}

/******************************************************************************
* Internal Method: get_cancel_id 
* Summary: Get this event's cancel ID
* Description: 
* Return: This event's cancel ID
******************************************************************************/
  int get_cancel_id() {return cancel_id;}

/******************************************************************************
* Internal Method: get_nopt 
* Summary: Get the length of the queue of optimistically generated events
* Description: 
* Return: The length of the queue of optimistically generated events
******************************************************************************/
  int get_nopt() {return optevts.length();}

/******************************************************************************
* Internal Method: get_evtptr 
* Summary: Get this event's optimistic event object (C_LPTR)
* Description: 
* Return: This event's optimistic event object (C_LPTR)
******************************************************************************/
  C_LPTR *get_evtptr() {return &evtptr;}

/******************************************************************************
* Internal Method: reset_valid 
* Summary: Tell this event that it is not valid
* Description: 
* Return: None
******************************************************************************/
  void reset_valid() {valid = 0;}

/******************************************************************************
* Internal Method: get_valid 
* Summary: Get whether this event is valid
* Description: 
* Return: Flag as to whether this event is valid
******************************************************************************/
  int get_valid() {return valid;}

/******************************************************************************
* Internal Method: clear_messages 
* Summary: Clear all of the messages out of this event
* Description: 
* Return: None
******************************************************************************/
  void clear_messages();

/******************************************************************************
* Internal Method: sendmess 
* Summary: Send messages out to the appropriate nodes
* Description: 
* Return: None
******************************************************************************/
  void sendmess(C_QUEUE *q1nd  // IN: the messages to send out
		);

/******************************************************************************
* Internal Method: getmess
* Summary: Get messages coming in
* Description: 
* Return: None
******************************************************************************/
  void getmess(C_QUEUE *q1nd     // OUT: the messages coming in
	       ) {q1nd->concat(&qmessout);}

/******************************************************************************
* Internal Method: get_mque 
* Summary: Get this event's queue of outgoing messages
* Description: 
* Return: This event's queue of outgoing messages
******************************************************************************/
  C_QUEUE *get_mque() {return &qmessout;}

/******************************************************************************
* Internal Method: get_amque 
* Summary: Get this event's queue of outgoing antimessages
* Description: 
* Return: This event's queue of outgoing antimessages
******************************************************************************/
  C_QUEUE *get_amque() {return &qanti;}

/******************************************************************************
* Internal Method: attach_object 
* Summary: Attach this event to its associated simulation object (SpSimObj)
* Description: 
* Return: None
******************************************************************************/
  void attach_object();

/******************************************************************************
* Internal Method: spin_loop 
* Summary: Loop to waste time
* Description: Waits until a specific point in time
* Return: None
******************************************************************************/
  void spin_loop(double twait  // amount of time to add to the desired point 
		               // in time to wait
		 );

/******************************************************************************
* Internal Method: check_canceled 
* Summary: Check if this event has been cancelled
* Description: 
* Return: The new number of rollbacks (ie, the same variable as was passed in)
******************************************************************************/
  int check_canceled(int &n_roll  // number of rollbacks
		     );

/******************************************************************************
* Internal Method: reset_event_canceled 
* Summary: Uncancel an event
* Description: 
* Return: None
******************************************************************************/
  void reset_event_canceled() {event_canceled = 0;}

/******************************************************************************
* Internal Method: get_event_canceled 
* Summary: Get cancel flag
* Description: 
* Return: cancel flag (true if cancelled)
******************************************************************************/
  int get_event_canceled() {return event_canceled;}

/******************************************************************************
* Internal Method: void_event 
* Summary: Get cancel flag (same as get_event_canceled)
* Description: 
* Return: cancel flag (true if cancelled)
******************************************************************************/
  virtual int void_event() {return event_canceled;}

/******************************************************************************
* Internal Method: clear_data 
* Summary: Clear the data for this event
* Description: 
* Return: None
******************************************************************************/
  void clear_data();

/******************************************************************************
* Internal Method: delete_MESSAGE 
* Summary: delete the MESSAGE associated with this event (if it exists)
* Description: 
* Return: None
******************************************************************************/
  void delete_MESSAGE();

/******************************************************************************
* Internal Method: init_qroll_backs 
* Summary: Place this event's rollback queue into the rollback queue manager
* Description: This allows the rollback queue manager (SpAltMgr) to perform the
*              rollbacks for this event
* Return: None
******************************************************************************/
  void init_qroll_backs() {SpCurrentAltMgr = &AltMgr;}

/******************************************************************************
* Internal Method: get_HOST_USER
* Summary: Get the host user on this node (C_HOST_USER)
* Description: see HOST_USER
* Return: the host user
******************************************************************************/
  C_HOST_USER *get_HOST_USER() {return HOST_USER;}

/******************************************************************************
* Internal Method: set_eventHandlerName
* Summary: Set the current event handler event name (set by
*          ScheduleEventHandler() to be possibly later used by
*          ReScheduleEventHandler())
* Description: 
* Return: None
******************************************************************************/
  void set_eventHandlerName(char *name // current event handler event name
			    ) { strcpy (eventHandlerName, name); }

/******************************************************************************
* Internal Method: get_eventHandlerName
* Summary: Get the lastprocessed event handler event name (used by
*          ReScheduleEventHandler())
* Description: 
* Return: the last-processed event handler event name
******************************************************************************/
  char *get_eventHandlerName() {return eventHandlerName;}

  void TieBreakGeneratedMessages();

  virtual void CriticalPathMessages(double earliestStartTime);

  void PrintTimeTag();

  void SetCPU_Time(double cpu_Time) {CPU_Time = cpu_Time;}
  double GetCPU_Time() {return CPU_Time;}

  SpTag *GetTag() {return &Tag;}

  double GetProcessingTime() {return ProcessingTime;}
  double GetCriticalPathTime() {return CriticalPathTime;}
  double GetTotalTime() {return TotalTime;}

  void SetProcessingTime(double processingTime) {
    ProcessingTime = processingTime;
  }

  void SetCriticalPathTime(double criticalPathTime) {
    CriticalPathTime = criticalPathTime;
  }

  void SetTotalTime(double totalTime) {
    TotalTime = totalTime;
  }

  void PrintScheduledEvents();

  SpEvent();

 private:

  static int EnableLazy;	        // lazy cancellation enabled flag
  static int one_node_flag;	        // one node flag
  static int gencid;		        // generated cancel id
  static SpMasterSimObjMgr *objects;	        // object manager
  static SpEventTypes *evtype;	        // event and message memory manager
  static C_DBL_QUEUE *dbl_qproc;	// queue of processed events
  static double spin_clock;          	// clocked spin time
  static double spin_time;	        // desired spin time
  static int traceflag;	                // trace flag (1 means on, 0 means off)
  static char *tracefile;      	        // name of trace file
  static int WriteLookahead;	        // write lookahead file (1=on, 0=off)
  static FILE *LookaheadFP;	        // file pointer for lookahead file
  static char *Lookaheadfile;	        // name of lookahead file
  
  SpEventStatus objevt;		        // object event status and link
  SpAltMgr AltMgr;	                // queue of roll backs
  C_QUEUE qmessout;		        // queue of outgoing messages
  C_QUEUE qanti;		        // queue of outgoing antimessages
  C_QUEUE optevts;		        // queue of optimistically generated
                                        // events
  C_LPTR evtptr;		        // optimistic event object
  C_DBL_ITEM dbl_item;	                // processed event item
  SpSimTime tmin;		                // minimum time of a generated message
  int valid;			        // valid event flag
  int rbid;			        // roll back unique id
  int cancel_id;		        // cancel event id
  int free_flag;		        // flag set only by free list
                                        // management
  int lazy_flag;		        // flag for lazy cancellation
  int lazy_proc;		        // flag for lazy processed
  int event_canceled;		        // event has canceled flag

  double CPU_Time;			// CPU time processing this event

  double CriticalPathTime;		// Time spent on the critical path
  double ProcessingTime;		// Total time processing events
  double TotalTime;			// Total CPU time (includes rollbacks)

  int from_SpComm;                      // non-zero iff event came from SpComm
    
 protected:

  SpTag Tag;				// user tag used for trace files

  SpSimObj *SIMOBJ;			// pointer to this event's simobj
  SpMsg *MESSAGE;			// pointer to event message
  char *DATA;			        // pointer to variable length data
  int OBJECT_TYPE;			// object type
  int EVENT_TYPE;			// event type
  int GLOBAL_ID;			// global id for simobj
   
  static int NODE;			// node
  static int N_NODES;			// number of nodes
  
  static C_RANDOM *RANDOM;		// random number generator
  static C_HOST_USER *HOST_USER;	// host user object
  
  char eventHandlerName[128];           // event handler event name

};

typedef SpEvent C_EVENT;

#endif
