// Filename: qitem.H

#ifndef dbl_item_object
#define dbl_item_object

#include "SpIostream.H"
#include "SpFstream.H"
#include "SpStrstream.H"
#include "SpIomanip.H"
#include "SpSimTime.H" // GEB2

/*****************************************************************************
* Class: C_DBL_ITEM
* Summary: Base class for classes whose objects can be non-rollbackably
*          inserted into doubly linked list data structures of type
*          C_DBL_QUEUE
* Description: Classes must inherit from this class to be inserted into
*              C_DBL_QUEUEs. This class does not have any rollback support, and
*              is thus simply a generic doubly-linked queue item that can be
*              sorted according to simulated time tags by a C_DBL_QUEUE. For
*              doubly-linked rollbackable queues, use C_SQ_ITEMs and
*              C_XQUEUEs. Objects of this type can only be inserted into
*              C_DBL_QUEUEs (ie, not C_QUEUEs or C_XQUEUEs). They can also be
*              used in a C_DBL_QUEUE used as a stack. That is, when
*              consistantly popping from the top of the queue, pushing can be
*              to the top or the bottom of the queue. When consistantly pushing
*              to the top, the C_QUEUE acts as a stack. When consistantly
*              pushing to the bottom, the C_QUEUE acts as a queue. Pushing
*              effects are vice-versa when consistantly popping from the bottom
*              (rather than the top) of the queue. Sorting and inserting
*              operations always put the low time value on top so that
*              consecutive pops off the top of the sorted queue or stack occurs
*              in simulated time order.
*****************************************************************************/
class C_DBL_ITEM {

 public:

/******************************************************************************
* Virtual Method: print 
* Summary: By default, this method prints the ID and simulation time to stderr,
*          but is user-overridable.
* Description: 
* Return: None
******************************************************************************/
  virtual void print();

/******************************************************************************
* Method: get_next
* Summary: Gets the forward link from this C_DBL_ITEM to the next C_DBL_ITEM
* Description: Use this to get the C_DBL_ITEM linked as the next C_DBL_ITEM in
*              the doubly-linked C_DBL_QUEUE data structure in which this
*              C_DBL_ITEM is contained
* Return: The C_DBL_ITEM forwardly linked from this C_DBL_ITEM
******************************************************************************/
  C_DBL_ITEM *get_next() {return next;}

/******************************************************************************
* Method: get_previous
* Summary: Gets the backward link from this C_DBL_ITEM to the previous
*          C_DBL_ITEM
* Description: Use this to get the C_DBL_ITEM linked as the previous C_DBL_ITEM
*              in the doubly-linked C_DBL_QUEUE data structure in which this
*              C_DBL_ITEM is contained
* Return: The C_DBL_ITEM backwardly linked from this C_DBL_ITEM
******************************************************************************/
  C_DBL_ITEM *get_previous() {return previous;}

/******************************************************************************
* Method: set_valu
* Summary: Sets the simulated time of this item
* Description: 
* Return: None
******************************************************************************/
  void set_valu(const SpSimTime& v //the simulated time of this item
		) {valu = v;}

/******************************************************************************
* Method: get_valu
* Summary: Gets the simulated time of this item
* Description: 
* Return: The simulated time of this item
******************************************************************************/
  SpSimTime& get_valu() {return valu;}

/******************************************************************************
* Method: set_id
* Summary: Sets the ID of this item
* Description: 
* Return: None
******************************************************************************/
  void set_id(int i // IN: the ID of this item
	      ) {id = i;}

/******************************************************************************
* Method: get_id
* Summary: Gets the ID of this item
* Description: 
* Return: The ID of this item
******************************************************************************/
  int get_id() {return id;}

/******************************************************************************
* Method: set_ptr
* Summary: Sets the pointer of this item to its data 
* Description: 
* Return: None
******************************************************************************/
  void set_ptr(void *p // IN: the pointer to this item's data
	       ) {ptr = p;}

/******************************************************************************
* Method: get_ptr
* Summary: Gets the pointer of this item to its data
* Description: 
* Return: The pointer to this item's data
******************************************************************************/
  void *get_ptr() {return ptr;}
  
/******************************************************************************
* Internal Method: set_next
* Summary: Sets the forward link from this C_DBL_ITEM to the next C_DBL_ITEM
* Description: Use this to set the link to the next C_DBL_ITEM in the
*              doubly-linked C_DBL_QUEUE data structure in which this
*              C_DBL_ITEM is contained
* Return: None
******************************************************************************/
  void set_next(C_DBL_ITEM *nxt // IN: C_ITEM to follow this one in the
		                // C_DBL_QUEUE this is currently contained in
		) {next = nxt;}

/******************************************************************************
* Internal Method: set_previous
* Summary: Sets the backward link from this C_DBL_ITEM to the C_DBL_ITEM to
*          precede this one
* Description: Use this to set the link to the previous C_DBL_ITEM in the
*              doubly-linked C_DBL_QUEUE data structure in which this
*              C_DBL_ITEM is contained
* Return: None
******************************************************************************/
  void set_previous(C_DBL_ITEM *prv // IN: C_ITEM to preceed this one in the
		                    // C_DBL_QUEUE this is currently
		                    // contained in
		) {previous = prv;}

  C_DBL_ITEM() {next = previous = NULL;}

 private:

  C_DBL_ITEM *previous;	// link to previous item
  C_DBL_ITEM *next;		// link to next item

protected:

  int id;			// id of the item
  SpSimTime valu;		// value of the item
  void *ptr;			// pointer to some data
  
};

#endif

