// dbl_queue.H header file

#ifndef dbl_queue_object
#define dbl_queue_object

#include "dbl_item.H"

/*****************************************************************************
* Class: C_DBL_QUEUE
* Summary: Doubly-linked list non-rollbackable data structure
* Description: This class is simply a non-rollbackable generic dounly-linked
*              queue that sorts C_DBL_ITEMs according to simulated time
*              tags. It can also be used as a stack. That is, when consistantly
*              popping from the top of the queue, pushing can be to the top or
*              the bottom of the queue. When consistantly pushing to the top,
*              the C_DBL_QUEUE acts as a stack. When consistantly pushing to
*              the bottom, the C_DBL_QUEUE acts as a queue. Pushing effects are
*              vice-versa when consistantly popping from the bottom (rather
*              than the top) of the queue (for C_XQUEUEs or C_DBL_QUEUEs
*              only). Sorting and inserting operations always put the low time
*              value on top so that consecutive pops off the top of the sorted
*              queue or stack occurs in simulated time order.
*****************************************************************************/
class C_DBL_QUEUE {

 public:

/******************************************************************************
* Method: get_length
* Summary: Get the number of items in the queue
* Description: 
* Return: The number of items in the queue
******************************************************************************/
  int get_length() {return length;}

/******************************************************************************
* Method: get_top
* Summary: Get a pointer to the item at the top of the queue without removing
*          it from the queue.
* Description: 
* Return: A pointer to the item at the top of the queue
******************************************************************************/
  C_DBL_ITEM *get_top() {return top;}

/******************************************************************************
* Method: get_bot
* Summary: Get a pointer to the item at the bottom of the queue without
*          removing it from the queue.
* Description: 
* Return: A pointer to the item at the bottom of the queue
******************************************************************************/
  C_DBL_ITEM *get_bot() {return bot;}

/******************************************************************************
* Method: push_top
* Summary: Pushes a specified item onto (ie, inserts a specified item at) the
*          top of the queue
* Description: This method moves the top pointer in this queue to the specified
*              item, which then points to the old top item
* Return: None
******************************************************************************/
  void push_top(C_DBL_ITEM *dbl_item // IN: specified item to insert
		);

/******************************************************************************
* Method: push_bot
* Summary: Pushes a specified item onto (ie, inserts a specified item at) the
*          bottom of the queue
* Description: 
* Return: None
******************************************************************************/
  void push_bot(C_DBL_ITEM *dbl_item // IN: specified item to insert
		);

/******************************************************************************
* Method: pop_top
* Summary: Pops a specified item off of (ie, removes a specified item from) the
*          top of the queue and returns it
* Description: 
* Return: Item that has been popped off of (removed from) the top of the queue
******************************************************************************/
  C_DBL_ITEM *pop_top();

/******************************************************************************
* Method: pop_bot
* Summary: Pops a specified item off of (ie, removes a specified item from) the
*          bottom of the queue and returns it
* Description: 
* Return: Item that has been popped off of (removed from) the bottom of the
*         queue
******************************************************************************/
  C_DBL_ITEM *pop_bot();

/******************************************************************************
* Method: insert
* Summary: Inserts a specified item into the queue in time order (earlier time
*          on top)
* Description: 
* Return: None
******************************************************************************/
  void insert(C_DBL_ITEM *dbl_item // IN: specified item to insert
	      );

/******************************************************************************
* Method: remove
* Summary: Remove (delete) an item from the queue
* Description: 
* Return: None
******************************************************************************/
  void remove(C_DBL_ITEM *dbl_item // IN: item to remove from the queue
	      );

/******************************************************************************
* Internal Debugging Method: check_order
* Summary: Checks if the items are in time order, from earliest to latestS
* Description: 
* Return: None
******************************************************************************/
  void check_order();

  C_DBL_QUEUE();

 private:

  int length;			// length of the queue
  C_DBL_ITEM *top;		// pointer to the top of the queue
  C_DBL_ITEM *bot;		// pointer to the bottom of the queue

};

inline C_DBL_ITEM *C_DBL_QUEUE::pop_top() {
  C_DBL_ITEM *dbl_item = NULL;

  if (length != 0) {
    dbl_item = top;
    top = top->get_next();
    length--;
  }
  return dbl_item;
}


#endif
