// parser.C method file

#include "SpStdio.H"
#include "SpStdlib.H"
#include "SpBasetype.H"
#include "gentype.H"
#include "genint.H"
#include "genfloat.H"
#include "genstring.H"
#include "genenum.H"
#include "genlogical.H"
#include "parser.H"

/************************************************************************
* PARSER : PARSER constructor						*
************************************************************************/
C_PARSER::C_PARSER(char *file) {
  int enm;
  SpBasetype *basetype;
  C_ITEM *item;

  name = strdup(file);

  fp = fopen(name,"r");
  if (fp == NULL) {
    fprintf(stderr,"PARSER file %s does not exist\n",name);
    return;
  }

  while(fscanf(fp,"%s %s\n",s1,s2) != EOF) {

    basetype = new SpBasetype(s1);
    types.push_bot(basetype);

    while (1) {

      fscanf(fp,"%s\n",s1);
      if (!strcmp(s1,"}")) break;

      fscanf(fp,"%s\n",s2);

      item = NULL;

      if (!strcmp(s1,"type")) {

	item = new C_GENTYPE(s2);
        basetype->add_type(item);

      }else{

	if (!strcmp("enum",s1)) {

	  enm = basetype->get_nenum();
	  item = new C_GENENUM(s2,enm);
          basetype->add(item);
	  basetype->set_enum(enm+1);

	}else{

          fscanf(fp,"%s\n",s3);

	  if (!strcmp("int",s1)) item = new C_GENINT(s2,s3);
	  if (!strcmp("float",s1)) item = new C_GENFLOAT(s2,s3);
	  if (!strcmp("string",s1)) item = new C_GENSTRING(s2,s3);
	  if (!strcmp("logical",s1)) item = new C_GENLOGICAL(s2,s3);

          if (item != NULL) {
            basetype->add(item);
          }else{
	    fprintf(stderr,"Error, %s not a defined type\n",s1);
	    exit(1);
          }

	}

      }

    }


  }

  fclose(fp);

  check();

}

C_PARSER::~C_PARSER()
{
  types.clear();
  free(name);
} 

/************************************************************************
* check : check that everything is ok with the file			*
************************************************************************/
int C_PARSER::check() {
  int i,j,k,len,nt,ne,status;
  char *type_name;
  char *elename;
  C_QUEUE *elements;
  C_GENTYPE *gentype;
  SpBasetype *basetype1;
  SpBasetype *basetype2;
  C_NAMIT *namit1;
  C_NAMIT *namit2;

  status = 1;
  len = types.length();
  basetype1 = (SpBasetype *)types.ptop();

  for (i=0; i<len; i++) {

//...... check that there are no duplicate names for type definitions

    type_name = basetype1->get_name();
    basetype2 = (SpBasetype *)basetype1->get_link();
    for (j=i+1; j<len; j++) {
      if (!strcmp(type_name, basetype2->get_name())) {
	status = 0;
	fprintf(stderr,"Error, duplicate %s\n", type_name);
      }
      basetype2 = (SpBasetype *)basetype2->get_link();
    }

//...... check that there are no undefined type references

    elements = basetype1->get_elements();
    nt = basetype1->get_ntypes();
    gentype = (C_GENTYPE *)elements->ptop();

    for (j=0; j<nt; j++) {

      type_name = gentype->get_name();
      basetype2 = get_basetype(type_name);

      if (basetype2 != NULL) {
	gentype->set_basetype(basetype2);
      }else{
	fprintf(stderr,"Error, %s referenced by %s does not exist\n",
		type_name, basetype1->get_name());
	status = 0;
      }

      gentype = (C_GENTYPE *)gentype->get_link();

    }

//...... check that there are no duplicate names within a type

    namit1 = (C_NAMIT *)gentype;
    ne = elements->length() - nt;

    for (j=0; j<ne; j++) {
      elename = namit1->get_name();

      namit2 = (C_NAMIT *)namit1->get_link();
      for (k=j+1; k<ne; k++) {
	if (!strcmp(elename,namit2->get_name())) {
	  status = 0;
	  fprintf(stderr,"Error, duplicate %s in %s\n", elename, type_name);
	}
	namit2 = (C_NAMIT *)namit2->get_link();
      }

      namit1 = (C_NAMIT *)namit1->get_link();
    }

//...... get the next basetype

    basetype1 = (SpBasetype *)basetype1->get_link();
  }

  return status;

}

/************************************************************************
* add_type : add a type to the list of types				*
************************************************************************/
void C_PARSER::add_type(char *s) {
  SpBasetype *basetype;

  basetype = get_basetype(s);
  if (basetype == NULL) {
    basetype = new SpBasetype(s);
    types.push_bot(basetype);
  }else{
    fprintf(stderr,"Error, type %s already exists\n",s);
  }

}

/************************************************************************
* backup : make a backup of the original file				*
************************************************************************/
void C_PARSER::backup() {
  int len;

  len = strlen(name);
  memcpy(s1,name,len);
  memcpy(&s1[len],".bak",4);
  s1[len+4] = '\0';

  fp = fopen(s1,"w");
  if (fp == NULL) {
    fprintf(stderr,"Error, can not store backup %s\n",s1);
    return;
  }

  printfile(fp);

  fclose(fp);

}

/************************************************************************
* list_types : list the names of the types				*
************************************************************************/
void C_PARSER::list_types() {
  int i,len;
  SpBasetype *basetype;

  len = types.length();
  basetype = (SpBasetype *)types.ptop();
  for (i=0; i<len; i++) {
    basetype->printname();
    basetype = (SpBasetype *)basetype->get_link();
  }

}

/************************************************************************
* print_type : print out the data for this type				*
************************************************************************/
void C_PARSER::print_type(char *s) {
  SpBasetype *basetype;

  basetype = get_basetype(s);
  if (basetype != NULL) {
    basetype->printfile(stdout);
  }else{
    fprintf(stderr,"Error, %s does not exist\n",s);
  }

}

/************************************************************************
* get_basetype : get the basetype if it exists				*
************************************************************************/
SpBasetype *C_PARSER::get_basetype(char *s) {
  int i,len;
  SpBasetype *basetype;

  len = types.length();
  basetype = (SpBasetype *)types.ptop();
  for (i=0; i<len; i++) {
    if (basetype->check_name(s)) return basetype;
    basetype = (SpBasetype *)basetype->get_link();
  }

  fprintf(stderr,"Error in Parser, could not find basetype %s\n",s);
  exit(1);
  return NULL;

}

/************************************************************************
* printfile : print out the data to a file (screen if f = stdout)	*
************************************************************************/
void C_PARSER::printfile(FILE *f) {
  int i,len;
  SpBasetype *basetype;

  len = types.length();
  basetype = (SpBasetype *)types.ptop();
  for (i=0; i<len; i++) {
    basetype->printfile(f);
    basetype = (SpBasetype *)basetype->get_link();
  }

}
