// Filename: qitem.H

#ifndef item_object
#define item_object

// #define valu time_tag
#define TIME_TAG time_tag
#define LOCAL_ID id
#define ID id

#include "SpSimTime.H"
#include "SpNetTypes.H"

/*****************************************************************************
* Internal Class: C_DUMMY_V
* Summary: Dummy class that puts the V table at the top of item objects
* Description: 
*****************************************************************************/
class C_DUMMY_V {
  private:
#ifndef KSR1
    NET_INT dummy_v_allign;		// dummy alignment
#endif
  protected:
  public:
    virtual void print();
};

/*****************************************************************************
* Class: C_ITEM
* Summary: Base class for classes whose objects can be
*          non-rollbackably inserted into singly linked list data
*          structures of type C_QUEUE
* Description: Classes must inherit from this class to be inserted
*              into C_QUEUEs. This class does not have any rollback
*              support, and is thus simply a generic queue item that
*              can be sorted according to simulated time tags by a
*              C_QUEUE. Objects of this type can only be inserted into
*              C_QUEUEs (ie, not C_XQUEUEs, which require rollback
*              support). They can also be used in a C_QUEUE used as a
*              stack. That is, popping must always be from the top of
*              the queue (since it is singly linked) but pushing can
*              be to the top or the bottom of the queue. When
*              consistantly pushing to the top, the C_QUEUE acts as a
*              stack. When consistantly pushing to the bottom, the
*              C_QUEUE acts as a queue. Sorting and inserting
*              operations always put the low time value on top so that
*              popping off the sorted queue or stack occurs in
*              simulated time order.
*
*              This class is also used as a base type for many of the
*              other structures within SPEEDES.  For example, it is
*              used in the class C_HASH (essentially a hash table of
*              C_QUEUE's) and in the SPEEDES event queue
*****************************************************************************/
class C_ITEM : public C_DUMMY_V {

 public:

/******************************************************************************
* Virtual Method: compare 
* Summary: By default, this method determines whether the IDs of this and
*          another C_ITEM are the same. It is user-overridable.
* Description:  
* Return: Flag for whether the IDs of this and the passed-in C_ITEM are the 
*         same
******************************************************************************/
    virtual int compare(C_ITEM *it // IN: the C_ITEM with which to compare
			) {
       int i = 0; if (id == it->get_id()) i=1; return i;}

/******************************************************************************
* Virtual Method: copy
* Summary: By default, this is not implemented, but is user overridable.
* Description: 
* Return: A C_ITEM
******************************************************************************/
    virtual C_ITEM *copy();

/******************************************************************************
* Virtual Method: print 
* Summary: By default, this method prints the ID and simulation time
*          to stderr, but is user-overridable.
* Description: 
* Return: None
******************************************************************************/
    virtual void print();

/******************************************************************************
* Virtual Method: printstream 
* Summary: By default, this method prints the ID and simulation time
*          to an ostream object, but is user-overridable.
* Description: 
* Return: None
******************************************************************************/
    virtual void printstream(ostream &out // IN: ostream object to print to
			     );

/******************************************************************************
* Virtual Method: printfile 
* Summary: By default, this is not implemented, but is
*          user-overridable, and is intended to print statistics about
*          this C_ITEM to a file.
* Description: 
* Return: None
******************************************************************************/
    virtual void printfile(FILE * // IN: file to print to
			   );

/******************************************************************************
* Virtual Method: GetName 
* Summary: By default, this is not implemented, but is
*          user-overridable, and is intended to return the name of
*          this C_ITEM.
* Description: 
* Return: A character pointer intended to return the name of this C_ITEM
******************************************************************************/
    virtual char *GetName();

/******************************************************************************
* Virtual Destructor: ~C_ITEM
* Summary: By default, this destructor does nothing, but is user-overridable
* Description: 
******************************************************************************/
    virtual ~C_ITEM();

/******************************************************************************
* Method: get_link 
* Summary: Gets the link for this C_ITEM (ie, the link in the singly-linked
*          C_QUEUE data structure)
* Description: 
* Return: The link for this C_ITEM (ie, the link in the singly-linke C_QUEUE
*         data structure)
******************************************************************************/
    C_ITEM *get_link() {return link;}

/******************************************************************************
* Method: get_time_tag 
* Summary: Gets the simulated time tag for this C_ITEM
* Description: 
* Return: The simulated time tag (SpSimTime)
******************************************************************************/
    SpSimTime&  get_time_tag() {return time_tag;}

/******************************************************************************
* Method: GetDoubleValue 
* Summary: Gets the simulated time tag for this C_ITEM without returning the
*          priority fields
* Description: 
* Return: The simulated time tag (as a double without returning the priority
*         fields)
******************************************************************************/
    double GetDoubleValue() {return time_tag.GetTime();}

/******************************************************************************
* Method: SetDoubleValue 
* Summary: Sets the simulated time tag for this C_ITEM without setting or
*          affecting the SpSimTime priority fields
* Description: 
* Return: None
******************************************************************************/
    void SetDoubleValue(double v) {time_tag.SetTime(v);}

/******************************************************************************
* Method: set_time_tag 
* Summary: Sets the simulated time tag for this C_ITEM
* Description:
* Return: None
******************************************************************************/
    void set_time_tag(const SpSimTime& t  // IN: simulated time tag
		      ) {
      time_tag = t;
    }

/******************************************************************************
* Overloaded Method: set_time_tag 
* Summary: Sets the simulated time tag for this C_ITEM, setting the SpSimTime
*          priority fields automatically to zero.
* Description:
* Return: None
******************************************************************************/
    void set_time_tag(double t  // IN: time tag
		      ) {time_tag = t;}

/******************************************************************************
* Overloaded Method: set_time_tag 
* Summary: Sets the simulated time tag for this C_ITEM, setting the
*          first SpSimTime priority field to a passed-in value, and
*          setting the second priority field automatically to zero.
* Description: 
* Return: None
******************************************************************************/
    void set_time_tag(double t,  // IN: time tag
		      int p1     // IN: first priority field 
		      ) {
       time_tag = t;
       time_tag.SetPriority1(p1);
    }

/******************************************************************************
* Overloaded Method: set_time_tag 
* Summary: Sets the simulated time tag for this C_ITEM, including both priority
*          fields
* Description: 
* Return: None
******************************************************************************/
    void set_time_tag(double t,  // IN: time tag
		      int p1,    // IN: first priority field
		      int p2     // IN: second priority field
		      ) {
       time_tag = t;
       time_tag.SetPriority1(p1);
       time_tag.SetPriority2(p2);
    }

/******************************************************************************
* Method: set_id 
* Summary: Set the ID of this C_ITEM
* Description: 
* Return: None
******************************************************************************/
    void set_id(int i) {id = i;}

/******************************************************************************
* Method: get_id 
* Summary: Get the ID of this C_ITEM
* Description: 
* Return: The ID of this C_ITEM
******************************************************************************/
    int get_id() const {return id;}

/******************************************************************************
* Internal Method: set_id 
* Summary: Set the free ID of this C_ITEM (for free list management)
* Description: 
* Return: None
******************************************************************************/
    void set_freeid(int i) {freeid = i;}

/******************************************************************************
* Internal Method: get_id 
* Summary: Get the free ID of this C_ITEM (for free list management)
* Description: 
* Return: The free ID of this C_ITEM
******************************************************************************/
    int get_freeid() {return freeid;}

/******************************************************************************
* Internal Method: put_link (same as set_link)
* Summary: Sets the link for this C_ITEM (ie, the link in the singly-linked
*          C_QUEUE data structure)
* Description: 
* Return: None
******************************************************************************/
    void put_link(C_ITEM *lnk // IN: next C_ITEM in the C_QUEUE
		  ) {link = lnk;}

/******************************************************************************
* Internal Method: set_link (same as put_link) 
* Summary: Sets the link for this C_ITEM (ie, the link in the singly-linked
*          C_QUEUE data structure)
* Description: 
* Return: None
******************************************************************************/
    void set_link(C_ITEM *lnk // IN: next C_ITEM in the C_QUEUE
		  ) {link = lnk;}

    C_ITEM() {link = NULL; freeid = -1;}

    C_ITEM *link;		// link to next item
#ifndef KSR1
    NET_INT item_allign;       	// dummy alignment
#endif
    NET_INT id;			// id
    NET_INT freeid;		// free list id
    SpSimTime time_tag;		// time tag of item for sorting
};


#endif

