// random.C method file

#include "SpStdio.H"
#include "SpMath.H"
#include "random.H"

#define IA 2416
#define IC 374441
#define IM 1771875

#define ONE_OVER_ROOT_TWO_PI 0.39894228

/************************************************************************
* C_RANDOM : construct a random object					*
************************************************************************/
C_RANDOM::C_RANDOM() {

//  printf("RANDOM object created\n");

  seed = 0x55555555;
  scaler = 0x7fffffff;
  lo = 0;
  hi = 0x7fffffff;
  flo = 0.0;
  fhi = 1.0;
  sigma = 1.0;
  mean = 0.0;
  gaussflag = 0;
  gaussvalue = 0.0;
  lambda = 1.0;
  fastmode = 0;
  beta_coeff_f = NULL;
  beta_coeff_F = NULL;

}

/************************************************************************
* get_random_bits : get random bits					*
************************************************************************/
int C_RANDOM::get_random_bits(int nbits) {
  int i,rn;

  rn = 0;
  for (i=0; i<nbits; i++) {
    rn <<= 1;
    rn |= get_random_bit();
  }

  return rn;

}

/************************************************************************
* get_random_int1 : get a random int number				*
************************************************************************/
int C_RANDOM::get_random_int1() {
  int i,rn;

  rn = 0;
  for (i=0; i<NBITS; i++) {
    rn <<= 1;
    rn |= get_random_bit();
  }

  return rn;

}

/************************************************************************
* get_random_int : get a random integer number between lo and hi	*
************************************************************************/
int C_RANDOM::get_random_int() {

  if (fastmode == 0) {

    int rn;
    unsigned int bin,nbm,nb;

    bin = 1+hi-lo;
    nbm = 0x80000000 / bin;

    do {
      rn = get_random_int1();
      nb = rn / bin;
    }   while (nb == nbm);

    rn = rn % bin + lo;
    return rn;

  }else{

    seed = (seed*IA+IC)%IM;
    return int(double(hi-lo+1)*(double(seed) / double(IM)) + lo);

  }

}

/************************************************************************
* get_random_float : get a random double between flo and fhi		*
************************************************************************/
double C_RANDOM::get_random_float() {

  if (fastmode == 0) {

    int rn;
    double frn;

    rn = get_random_int1();
    frn = rn / scaler;
    frn = frn * (fhi-flo) + flo;
    return frn;

  }else{

    seed = (seed*IA+IC)%IM;
    return (fhi-flo)*(double(seed) / double(IM)) + flo;

  }

}

/************************************************************************
* get_random_float : get a random double between flo and fhi		*
************************************************************************/
double C_RANDOM::get_random_float(int nbits) {
  int rn;
  double temp,frn;

  rn = get_random_bits(nbits);
  temp = (1<<nbits);
  frn = rn / temp;
  frn = frn * (fhi-flo) + flo;
  return frn;

}

/************************************************************************
* get_random_gauss : get a random gaussian value			*
************************************************************************/
double C_RANDOM::get_random_gauss() {
  double fac,r,v1,v2;

  if (gaussflag) {
    gaussflag = 0;
    return gaussvalue*sigma + mean;
  }else{
    gaussflag = 1;
    do {
      v1 = 2.0 * get_random_int1() / scaler - 1.0;
      v2 = 2.0 * get_random_int1() / scaler - 1.0;
      r = v1*v1 + v2*v2;
    } while (r >= 1.0);
    fac = sqrt(-2.0*log(r)/r);
    gaussvalue = v1*fac;
    return (v2*fac)*sigma + mean;
  }

}

/************************************************************************
* get_random_exp : get a random exponential value			*
************************************************************************/
double C_RANDOM::get_random_exp() {
  double ex;

  ex = (get_random_int1() / scaler);
  ex = -log(ex) / lambda;
  return ex;

}

/************************************************************************
* set_beta : initialize the beta density coefficients			*
************************************************************************/
void C_RANDOM::set_beta(int n1, int n2) {
  int i;
  double t;
  double f, F;
  double sumCF;

// set the n1 and n2 values

  beta_n1 = n1;
  beta_n2 = n2;

// create the new array of beta cumulative density coefficients

  if (beta_coeff_f != NULL) delete [] beta_coeff_f;
  beta_coeff_f = new double[n2+1];

  if (beta_coeff_F != NULL) delete [] beta_coeff_F;
  beta_coeff_F = new double[n2+1];

// set up the coefficients for the density function f

  beta_coeff_f[0] = fact(n1+n2+1) / (fact(n1)*fact(n2));

  for (i=0; i<n2; i++) {
    beta_coeff_f[i+1] = -beta_coeff_f[i]*double(n2-i)/double(i+1);
  }

// set up the coefficients for the cummulative density function F

  for (i=0; i<=n2; i++) {
    beta_coeff_F[i] = beta_coeff_f[i] / (n1+i+1);
  }

// make sure that the coefficients are correct

  sumCF = 0.0;
  for (i=0; i<=n2; i++) {
    sumCF += beta_coeff_F[i];
  }
  if (fabs(1.0-sumCF) > 0.001) {
    printf("Sum of F coeff = %f (should be 1)\n",sumCF);
  }

  for (t=0.0; t<=1.0; t+=0.1) {
    compute_beta_fF(t,f,F);
  }

}


/************************************************************************
* fact : factorial of a number						*
************************************************************************/
double C_RANDOM::fact(int n) {
  int i;
  double value;

  value = 1;
  for (i=2; i<=n; i++) value *= i;
  return value;

}

/************************************************************************
* compute_beta_fF : computes the beta density distributions f and F	*
************************************************************************/
void C_RANDOM::compute_beta_fF(double t, double &f, double &F) {
  int i;
  double tp;

  tp = 1.0;
  for (i=0; i<beta_n1; i++) tp*=t;

  f = 0.0;
  F = 0.0;
  for (i=0; i<=beta_n2; i++) {
    f += beta_coeff_f[i] * tp;
    tp *= t;
    F += beta_coeff_F[i] * tp;
  }

//  printf("Computed time %f, f(t) = %f, F(t) = %f\n",t,f,F);

}

/************************************************************************
* get_random_beta : get a random beta number				*
************************************************************************/
double C_RANDOM::get_random_beta() {
  double rvalue;
  double t, tlo, thi, tnew;
  double f, F;
  double error;
  int nits;

  rvalue = (get_random_float() - flo) / (fhi-flo);
//  printf("rvalue = %f\n",rvalue);

  t = 0.5;
  tlo = 0.0;
  thi = 1.0;
  error = 1.0;
  nits = 0;

  while ((error > 0.000001) && (nits < 100)) {
    nits++;
    compute_beta_fF(t, f, F);
    tnew = t - (F-rvalue)/f;


    if (tnew < tlo) {
      thi = t;
      tnew = tlo + (thi-tlo)/2.0;
    }
    if (tnew > thi) {
      tlo = t;
      tnew = tlo + (thi-tlo)/2.0;
    }
    error = fabs(tnew-t);
    t = tnew;
  }

//  printf("done after %d iterations\n",nits);

  return t;

}


/************************************************************************
* get_random_unit_vector : get a random unit vector			*
************************************************************************/
void C_RANDOM::get_random_unit_vector(double vec[3]) {
  double costheta, sintheta, phi, cosphi, sinphi;

  set_float_limits(-1.0, 1.0);
  costheta = get_random_float();
  sintheta = sqrt(1.0 - costheta*costheta);

  set_float_limits(0.0, 6.283185307);
  phi = get_random_float();
  cosphi = cos(phi);
  sinphi = sin(phi);

  vec[0] = sintheta*cosphi;
  vec[1] = sintheta*sinphi;
  vec[2] = costheta;

}

/************************************************************************
* get_random_mag_vector : get a random magnitude vector			*
************************************************************************/
void C_RANDOM::get_random_mag_vector(double mag, double vec[3]) {
  double ma;

  get_random_unit_vector(vec);
  set_gauss(1, 0);
  ma = mag * get_random_gauss();

  vec[0] *= ma;
  vec[1] *= ma;
  vec[2] *= ma;

}

/************************************************************************
* erf : compute the error function					*
************************************************************************/
double C_RANDOM::erf(double x) {
  int i;
  double x2;
  double X;
  double Xlast;
  double Xret;
  double a;

  if (x>6) return 0.5;

  i=0;
  X = x;
  Xlast = 0.0;
  x2 = x*x;
  a = X;

  while (fabs(X-Xlast) > 0.000001) {
    Xlast = X;
    a *= (-x2*double(2*i+1)) / double(2*(i+1)*(2*i+3));
    X += a;
    i++;
  }

  Xret = ONE_OVER_ROOT_TWO_PI * X;
  if (Xret <= 0.5) return Xret;
  return 0.5;
}

/************************************************************************
* inverse_erf : compute the inverse error function			*
************************************************************************/
double C_RANDOM::inverse_erf(double x) {
  double xguess;
  double Xret;
  double lower;
  double upper;
  double dx;

  if ((x>0.5) || (x<0.0)) {
    printf("Error (inverse_erf): bad value %f\n",x);
  }
  if (x==0.5) return 1.0e20;
  if (x==0.0) return -1.0e20;

  while (1) {

    if (x < 0.1) {
      lower = 0.0;
      upper = 0.26;
      break;
    }

    if (x < 0.2) {
      lower = 0.25;
      upper = 0.53;
      break;
    }

    if (x < 0.3) {
      lower = 0.52;
      upper = 0.85;
      break;
    }

    if (x < 0.4) {
      lower = 0.84;
      upper = 1.29;
      break;
    }

    if (x < 0.45) {
      lower = 1.28;
      upper = 1.65;
      break;
    }

    if (x < 0.49) {
      lower = 1.64;
      upper = 2.33;
      break;
    }

    if (x >= 0.49) {
      lower = 2.32;
      upper = 5.1;
      break;
    }

  }

  Xret = (upper+lower) / 2.0;
  dx = 1.0;

  upper += 0.1;
  lower -= 0.1;
  if ( lower < 0.0) lower = 0.0;

  while (fabs(dx) > 0.000001) {

    xguess = erf(Xret);
    dx = (x-xguess) / (ONE_OVER_ROOT_TWO_PI*exp(-Xret*Xret/2.0));
    if (xguess < x) {
      lower = Xret-0.001;
      if (lower < 0.0) lower = 0.0;
    }else{
      upper = Xret+0.001;
    }

    Xret += dx;

    if (Xret > upper) Xret = (upper+lower) / 2.0;
    if (Xret < lower) Xret = (upper+lower) / 2.0;

  }

  return Xret;

}

/************************************************************************
* his_bit : make a histogram of random bits				*
************************************************************************/
void C_RANDOM::his_bit() {
  int his[2];
  int i,j;
  int rbit;

  printf("bit histogram\n");

  for (i=0; i<2; i++) his[i] = 0;

  do {
    rbit = get_random_bit();
    his[rbit]++;
  } while (his[rbit] < 99);

  for (i=0; i<2; i++) {
    printf("%2d [%2d] ",i,his[i]);
    for (j=0; j<his[i]; j++) printf("*");
    printf("\n");
  }
  printf("\n");

}

/************************************************************************
* his_int1 : make a histogram of random ints				*
************************************************************************/
void C_RANDOM::his_int1() {
  int his[20];
  int i,j;
  int rint;
  double bin;
  int index;

  printf("int histogram\n");

  for (i=0; i<20; i++) his[i] = 0;
  bin = 0x7fffffff / 20.0;

  do {
    rint = get_random_int1();
    index = int(rint / bin);
    if (index > 19) index = 19;
    his[index]++;
  } while (his[index] < 99);

  for (i=0; i<20; i++) {
    printf("%2d [%2d] ",i,his[i]);
    for (j=0; j<his[i]; j++) printf("*");
    printf("\n");
  }
  printf("\n");

}

/************************************************************************
* his_int : make a histogram of random ints				*
************************************************************************/
void C_RANDOM::his_int() {
  int his[20];
  int i,j;
  int rint;
  double bin;
  int index;

  printf("int histogram between %d and %d\n",lo,hi);

  for (i=0; i<20; i++) his[i] = 0;
  bin = (hi-lo) / 20.0;

  do {
    rint = get_random_int();
    index = int((rint - lo) / bin);
    if (index > 19) index = 19;
    his[index]++;
  } while (his[index] < 99);

  for (i=0; i<20; i++) {
    printf("%2d [%2d] ",i,his[i]);
    for (j=0; j<his[i]; j++) printf("*");
    printf("\n");
  }
  printf("\n");

}

/************************************************************************
* his_double : make a histogram of random doubles			*
************************************************************************/
void C_RANDOM::his_double() {
  int his[20];
  int i,j;
  double rdouble;
  double bin;
  int index;

  printf("double histogram between %f and %f\n",flo,fhi);

  for (i=0; i<20; i++) his[i] = 0;
  bin = (fhi-flo) / 20.0;

  do {
    rdouble = get_random_float();
    index = int((rdouble - flo) / bin);
    if (index > 19) index = 19;
    his[index]++;
  } while (his[index] < 99);

  for (i=0; i<20; i++) {
    printf("%2d [%2d] ",i,his[i]);
    for (j=0; j<his[i]; j++) printf("*");
    printf("\n");
  }
  printf("\n");

}

/************************************************************************
* his_gauss : make a histogram of random gaussians			*
************************************************************************/
void C_RANDOM::his_gauss() {
  int his[40];
  int i,j;
  double rgauss;
  double bin;
  int index;
  double glo;

  printf("gauss histogram with sigma = %f, mean = %f from %f to %f\n",
	sigma,mean,mean-3.0*sigma,mean+3.0*sigma);

  for (i=0; i<40; i++) his[i] = 0;
  bin = (6.0*sigma) / 40.0;
  glo = mean - 3.0*sigma;

  do {
    rgauss = get_random_gauss();
    index = int((rgauss - glo) / bin);
    if (index < 0) index = 0;
    if (index > 39) index = 39;
    his[index]++;
  } while (his[index] < 99);

  for (i=0; i<40; i++) {
    printf("%2d [%2d] ",i,his[i]);
    for (j=0; j<his[i]; j++) printf("*");
    printf("\n");
  }
  printf("\n");

}

/************************************************************************
* his_exp : make a histogram of random exponentials			*
************************************************************************/
void C_RANDOM::his_exp() {
  int his[40];
  int i,j;
  double rexp;
  double bin;
  int index;

  printf("exp histogram with lambda = %f from 0.0 to %f\n",lambda,5.0/lambda);

  for (i=0; i<40; i++) his[i] = 0;
  bin = (5.0 /lambda) / 40.0;

  do {
    rexp = get_random_exp();
    index = int(rexp / bin);
    if (index > 39) index = 39;
    his[index]++;
  } while (his[index] < 99);

  for (i=0; i<40; i++) {
    printf("%2d [%2d] ",i,his[i]);
    for (j=0; j<his[i]; j++) printf("*");
    printf("\n");
  }
  printf("\n");

}

/************************************************************************
* print_cycle : print the number of cycles for the generater		*
************************************************************************/
void C_RANDOM::print_cycle() {
  int i;
  int sed;

  printf("checking the cycle of the random number generator\n");

  i = 0;
  sed = seed;
  get_random_int();

  while(sed != int(seed)) {
    get_random_int();
    i++;
    if (i%10000 == 0) printf("%10dth new random number\n",i);
  }

  printf("cycle repeats after %d iterations\n",i);

}

/************************************************************************
* print : print random integer between lo and hi			*
************************************************************************/
void C_RANDOM::print() {
  int sed;
  int rn;
  double frn;

  sed = seed;
  rn = get_random_int();
  seed = sed;
  frn = get_random_float();

  printf("Iran = %d,\tFran = %f,\tseed = %08x\n",rn,frn,seed);

}

int C_RANDOM::get_random_bit() {
  unsigned int newbit;

  newbit = (seed & IB18) >> 17
	^  (seed & IB5) >> 4
	^  (seed & IB2) >> 1
	^  (seed & IB1);
  seed = (seed << 1) | newbit;
  return newbit;
}
