// Filename: random.H
					
#ifndef random_object
#define random_object

#define IB1	1
#define IB2	2
#define IB5	16
#define IB18	131072
#define NBITS	31

/*****************************************************************************
* Class: C_RANDOM
* Summary: Random number generator (one per node)
* Description: Use this class to generate random numbers. It keeps track of
*              its seed automatically, but not rollbackably. For rollbacks,
*              the seed must be saved by the simulation objects unless the
*              process model is being used, in which case rollbackable random
*              numbers is supported automatically.
*****************************************************************************/
class C_RANDOM {

 public:

/******************************************************************************
* Method: set_fast
* Summary: Set the generator to linear-congruential mode (far faster than
*          generate-random-bits mode, and far more cycles before the random
*          number cycle repeats, but far less even random number distribution)
* Description: 
* Return: None
******************************************************************************/
  void set_fast() {fastmode = 1;}

/******************************************************************************
* Method: set_slow
* Summary: Set the generator to generate-random-bits mode (far slower than
*          linear-congruential mode, and far less cycles before the random
*          number cycle repeats (256,000 cycles), but far more even random
*          number distribution)
* Description: 
* Return: None
******************************************************************************/
  void set_slow() {fastmode = 0;}

/******************************************************************************
* Method: set_seed
* Summary: Set the seed for this generator
* Description: 
* Return: None
******************************************************************************/
  void set_seed(int sd  // IN: seed
		) {seed = sd;}

/******************************************************************************
* Method: get_seed
* Summary: Get the seed for this generator
* Description: 
* Return: The seed for this generator
******************************************************************************/
  int get_seed() {return seed;}

/******************************************************************************
* Method: set_int_limits
* Summary: Set the low and high limits for random integers
* Description: These variables become state for this object. Since there is
*              only one of this object per node, this method should be called
*              at least once per event where random integers are generated.
* Return: None
******************************************************************************/
  void set_int_limits(int l, // IN: low limit for random integers
		      int h  // IN: high limit for random integers
		      ) {lo = l; hi = h;}

/******************************************************************************
* Method: set_float_limits
* Summary: Set the low and high limits for random floats (doubles)
* Description: These variables become state for this object. Since there is
*              only one of this object per node, this method should be called
*              at least once per event where random floats are generated.
* Return: None
******************************************************************************/
  void set_float_limits(double l, // IN: low limit for random floats
			double h  // IN: high limit for random floats
			) {flo = l; fhi = h;}

/******************************************************************************
* Method: get_random_int
* Summary: Generate a random integer between the low limit and high limit as
*          last set by set_int_limits ()
* Description: Uses the algorithm specified by the last call to either
*              set_fast () or set_slow ()
* Return: A random integer
******************************************************************************/
  int get_random_int();

/******************************************************************************
* Method: get_random_bits
* Summary: Generate a random integer with a specified number of bits
* Description: Uses the generate-random-bits (slower but more random) algorithm
* Return: A random integer with a specified number of bits
******************************************************************************/
  int get_random_bits(int nbits // IN: specified number of bits
		      );

/******************************************************************************
* Overloaded Method: get_random_float
* Summary: Generate a random float (double) between the low limit and high
*          limit as last set by set_float_limits ()
* Description: Uses the algorithm specified by the last call to either
*              set_fast () or set_slow ()
* Return: A random float (double)
******************************************************************************/
  double get_random_float();

/******************************************************************************
* Overloaded Method: get_random_float
* Summary: Generate a random integer with a specified number of bits
* Description: Uses the generate-random-bits (slower but more random) algorithm
* Return:  A random integer with a specified number of bits
******************************************************************************/
  double get_random_float(int nbits // IN: specified number of bits
			  );

/******************************************************************************
* Method: get_random_gauss
* Summary: Generate a random float (double) according to a gaussian
*          distribution with parameters set by the last call to set_gauss ()
* Description: 
* Return: A random float (double) according to a gaussian distribution
******************************************************************************/
  double get_random_gauss();

/******************************************************************************
* Method: get_random_exp
* Summary: Generate a random float (double) according to an exponential
*          distribution with parameters set by the last call to set_exp ()
* Description: 
* Return: A random float (double) according to an exponential distribution
******************************************************************************/
  double get_random_exp();

/******************************************************************************
* Method: get_random_unit_vector
* Summary: Generate a random unit vector
* Description: 
* Return: None
******************************************************************************/
  void get_random_unit_vector(double vec[3] // OUT: unit vector
			      );

/******************************************************************************
* Method: get_random_mag_vector
* Summary: Generate a random magnitude vector
* Description: This first generates a unit vector, then multiplies the
*              magnitude by a random gaussian with sigma=1 and mean=0, and then
*              multiplies that number by each of the components of the unit
*              vector
* Return: None
******************************************************************************/
  void get_random_mag_vector(double mag,   // IN: magnitude
			     double vec[3] // OUT: magnitude vector
			     );

/******************************************************************************
* Method: set_gauss
* Summary: Set sigma and mean for random gaussian distributions
* Description: 
* Return: None
******************************************************************************/
  void set_gauss(double s, // IN: sigma
		 double m  // IN: mean
		 ) {sigma = s; mean = m; gaussflag = 0;}

/******************************************************************************
* Method: set_exp
* Summary: Set lambda for random exponential distributions
* Description: 
* Return: None
******************************************************************************/
  void set_exp(double l // IN: lambda
	       ) {lambda = l;}

/******************************************************************************
* Method: set_beta
* Summary: Set n1 and n2 for beta distributions
* Description: 
* Return: None
******************************************************************************/
  void set_beta(int n1, // IN: n1
		int n2  // IN: n2
		);

/******************************************************************************
* Method: compute_beta_fF
* Summary: Computes the beta density distributions f and F
* Description: 
* Return: None
******************************************************************************/
  void compute_beta_fF(double tt,  // IN: t
		       double &ff, // OUT: beta density distributions f
		       double &FF  // OUT: beta density distributions F
		       );

/******************************************************************************
* Method: get_random_beta
* Summary: Get a random beta number
* Description: 
* Return: A random beta number
******************************************************************************/
  double get_random_beta();

/******************************************************************************
* Internal Method: print_cycle
* Summary: Print the number of calls to get_random_int () method before it
*          repeats the saem cycle of random integers
* Description: 
* Return: None
******************************************************************************/
  void print_cycle();

/******************************************************************************
* Internal Method: print
* Summary: Print a raondom integer and a random float
* Description: 
* Return: None
******************************************************************************/
  void print();

/******************************************************************************
* Internal Method: get_random_int1
* Summary: Get a random integer with no limits using generate-random-bits
*          (slower) method
* Description: 
* Return: A random integer with no limits
******************************************************************************/
  int get_random_int1();

/******************************************************************************
* Internal Method: fact
* Summary: Get the factorial of a specified integer
* Description: 
* Return: The factorial of a specified integer
******************************************************************************/
  double fact(int n // IN: specified integer
	      );

/******************************************************************************
* Internal Method: erf
* Summary: Compute the error function
* Description: 
* Return: 
******************************************************************************/
  double erf(double x);

/******************************************************************************
* Internal Method: inverse_erf
* Summary: Compute the inverse error function
* Description: 
* Return: 
******************************************************************************/
  double inverse_erf (double x);

/******************************************************************************
* Internal Method: get_random_bit
* Summary: Generate a random integer between 0 and 1
* Description: 
* Return: A random integer between 0 and 1
******************************************************************************/
  int get_random_bit();

/******************************************************************************
* Internal Method: his_bit
* Summary: Make a histogram of random 
* Description: 
* Return: None
******************************************************************************/
  void his_bit();

/******************************************************************************
* Internal Method: his_int1
* Summary: Make a histogram of random 
* Description: 
* Return: None
******************************************************************************/
  void his_int1();

/******************************************************************************
* Internal Method: his_int
* Summary: Make a histogram of random 
* Description: 
* Return: None
******************************************************************************/
  void his_int();

/******************************************************************************
* Internal Method: his_double
* Summary: Make a histogram of random 
* Description: 
* Return: None
******************************************************************************/
  void his_double();

/******************************************************************************
* Internal Method: his_gauss
* Summary: Make a histogram of random 
* Description: 
* Return: None
******************************************************************************/
  void his_gauss();

/******************************************************************************
* Internal Method: his_exp
* Summary: Make a histogram of random 
* Description: 
* Return: None
******************************************************************************/
  void his_exp();

  C_RANDOM();


  void reset(){beta_coeff_f = beta_coeff_F = 0;}

 private:

  int fastmode;		// flag for fast random number generation
  int seed;	        // random number seed (used by everthing)
  int lo;		// lower limit for integer random numbers
  int hi;		// upper limit for integer random numbers
  double flo;		// lower limit for float random numbers
  double fhi;		// upper limit for float random numbers
  double scaler;	// maximum integer random number
  double sigma;		// sigma for gaussian distributions
  double mean;		// mean for gaussian distributions
  int gaussflag;	// flag for an extra gaussian value available
  double gaussvalue;	// extra gaussian value
  double lambda;	// lambda for exponential distributions
  double *beta_coeff_f;	// coefficients for the beta function
  double *beta_coeff_F;	// coefficients for the cum. beta function
  double beta_n1;	// beta function n1
  double beta_n2;	// beta function n2;
};

#endif
