// sq_item.H header file

#ifndef sq_item_object
#define sq_item_object

#include "NetTypes.H"
#include "queue.H"
#include "qitem.H"

/*****************************************************************************
* Class: C_SQ_ITEM 

* Summary: Base class for classes whose objects can be rollbackably inserted
*          into doubly-linked list data structures of type C_XQUEUE, or
*          non-rollbackably inserted into singly-linked list data structures of
*          type C_QUEUE
* Description: Classes must inherit from this class to be inserted into
*              C_XQUEUEs or C_QUEUEs. This class is simply a generic queue item
*              that can be sorted according to simulated time tags by a
*              C_XQUEUE (doubly-linked) or C_QUEUE (singly-linked). All
*              operations on a C_SQ_ITEM in or by a C_XQUEUE are rollbackable,
*              while all operations on a C_SQ_ITEM in or by a C_QUEUE are
*              non-rollbackable. They can also be used in a C_XQUEUE or a
*              C_QUEUE used as a stack. That is, when consistantly popping from
*              the top of the queue (which is the only way to pop in a C_QUEUE,
*              since it is singly-linked), pushing can be to the top or the
*              bottom of the queue. When consistantly pushing to the top, the
*              C_XQUEUE or C_QUEUE acts as a stack. When consistantly pushing
*              to the bottom, the C_XQUEUE or C_QUEUE acts as a queue. Pushing
*              effects are vice-versa when consistantly popping from the bottom
*              (rather than the top) of the queue (for C_XQUEUEs or
*              C_DBL_QUEUEs only). Sorting and inserting operations always put
*              the low time value on top so that consecutive pops off the top
*              of the sorted queue or stack occurs in simulated time order.
*****************************************************************************/
class C_SQ_ITEM : public C_ITEM {

 public:

/******************************************************************************
* Method: get_backlink
* Summary: Gets the back link for this C_ITEM (ie, the link to the previous
*          item in the doubly-linked C_XQUEUE data structure)
* Description: 
* Return: The back link for this C_ITEM (ie, the link to the previous item in
*         the doubly-linked C_XQUEUE data structure)
******************************************************************************/
  C_SQ_ITEM *get_backlink() {return backlink;}

/******************************************************************************
* Internal Method: set_backlink
* Summary: Sets the back link for this C_ITEM (ie, the link to the previous
*          item in the doubly-linked C_XQUEUE data structure)
* Description:
* Return: None
******************************************************************************/
  void set_backlink(C_SQ_ITEM *bl) {backlink = bl;}

/******************************************************************************
* Internal Method: set_processed
* Summary: 
* Description: 
* Return: None
******************************************************************************/
  void set_processed() {processed = 1;}

/******************************************************************************
* Internal Method: reset_processed
* Summary: 
* Description: 
* Return: None
******************************************************************************/
  void reset_processed() {processed = 0;}

/******************************************************************************
* Internal Method: get_processed
* Summary: 
* Description: 
* Return: None
******************************************************************************/
  int get_processed() {return processed;}

/******************************************************************************
* Internal Method: set_xinsert
* Summary: 
* Description: 
* Return: None
******************************************************************************/
  void set_xinsert() {processed = 1;}

/******************************************************************************
* Internal Method: set_xremove
* Summary: 
* Description: 
* Return: None
******************************************************************************/
  void set_xremove() {processed = 0;}

/******************************************************************************
* Internal Method: get_queue
* Summary: 
* Description: 
* Return: None
******************************************************************************/
  C_QUEUE *get_queue() {return &queue;}

  C_SQ_ITEM() {processed = 0;}

  int processed;
  C_QUEUE queue;        // used for rqueues
  C_SQ_ITEM *backlink;	// link to previous item
#ifndef KSR1
    NET_INT sq_item_allign;		// allignment
#endif


};

#endif

