// xqueue.H header file

#ifndef xqueue_object
#define xqueue_object

#include "sq_item.H"
#include "RB_SpDefineClass.H"

/*****************************************************************************
* Class: C_XQUEUE
* Summary: Doubly-linked list rollbackable data structure
* Description: This class is simply a rollbackable generic queue that sorts
*              C_SQ_ITEMs according to simulated time tags. It can also be used
*              as a stack. That is, when consistantly popping from the top of
*              the queue, pushing can be to the top or the bottom of the
*              queue. When consistantly pushing to the top, the C_XQUEUE acts
*              as a stack. When consistantly pushing to the bottom, the
*              C_XQUEUE acts as a queue. Pushing effects are vice-versa when
*              consistantly popping from the bottom (rather than the top) of
*              the queue (for C_XQUEUEs or C_DBL_QUEUEs only). Sorting and
*              inserting operations always put the low time value on top so
*              that consecutive pops off the top of the sorted queue or stack
*              occurs in simulated time order.
*****************************************************************************/
class C_XQUEUE {

 public:

/******************************************************************************
* Method: reset
* Summary: Clear this queue without deleting the items
* Description: Sets top and bottom of this queue to NULL, and number of items
*              to zero
* Return: None
******************************************************************************/
  void reset() {length = 0; top = NULL; bot = NULL;}

/******************************************************************************
* Method: ClearAndDestroy
* Summary: Clear this queue and delete all of the items
* Description:
* Return: None
******************************************************************************/
  void ClearAndDestroy();

/******************************************************************************
* Method: get_length
* Summary: Get the number of items in the queue
* Description: 
* Return: The number of items in the queue
******************************************************************************/
  int get_length() {return length;}

/******************************************************************************
* Method: set_length
* Summary: Set the number of items in the queue (GET RID OF!)
* Description: 
* Return: None
******************************************************************************/
  void set_length(int l) {length = l;}

/******************************************************************************
* Method: get_top
* Summary: Get a pointer to the item at the top of the queue without removing
*          it from the queue.
* Description: 
* Return: A pointer to the item at the top of the queue
******************************************************************************/
  C_SQ_ITEM *get_top() {return top;}

/******************************************************************************
* Method: get_bot
* Summary: Get a pointer to the item at the bottom of the queue without
*          removing it from the queue.
* Description: 
* Return: A pointer to the item at the bottom of the queue
******************************************************************************/
  C_SQ_ITEM *get_bot() {return bot;}

/******************************************************************************
* Method: push_bot
* Summary: Pushes a specified item onto (ie, inserts a specified item at) the
*          bottom of the queue
* Description: 
* Return: None
******************************************************************************/
  void push_bot(C_SQ_ITEM *x_item // IN: specified item to insert
		  );

/******************************************************************************
* Method: push_top
* Summary: Pushes a specified item onto (ie, inserts a specified item at) the
*          top of the queue
* Description: This method moves the top pointer in this queue to the specified
*              item, which then points to the old top item
* Return: None
******************************************************************************/
  void push_top(C_SQ_ITEM *x_item // IN: specified item to insert
		);

/******************************************************************************
* Method: pop_top
* Summary: Pops a specified item off of (ie, removes a specified item from) the
*          top of the queue and returns it
* Description: 
* Return: Item that has been popped off of (removed from) the top of the queue
******************************************************************************/
  C_SQ_ITEM *pop_top();

/******************************************************************************
* Method: pop_bot
* Summary: Pops a specified item off of (ie, removes a specified item from) the
*          bottom of the queue and returns it
* Description: 
* Return: Item that has been popped off of (removed from) the bottom of the
*         queue
******************************************************************************/
  C_SQ_ITEM *pop_bot();

/******************************************************************************
* Method: remove
* Summary: Remove (but don't delete) an item from the queue
* Description: 
* Return: None
******************************************************************************/
  void remove(C_SQ_ITEM *x_item // INOUT: item to remove from the queue
		);
/******************************************************************************
* Method: find
* Summary: Find (without deleting) an item from the queue
* Description: 
* Return: An item from the queue
******************************************************************************/
  C_SQ_ITEM *find(int id // IN: item ID to find from the queue
		  );

/******************************************************************************
* Method: join
* Summary: Concatenate a specified queue to this queue, but do not reset 
*          the specified queue
* Description: 
* Return: None
******************************************************************************/
  void join(C_XQUEUE *q // IN: specified queue to concatenate with this queue
	    );

/******************************************************************************
* Method: print
* Summary: Print the items in this queue
* Description: 
* Return: None
******************************************************************************/
  void print();

  C_XQUEUE();

/******************************************************************************
* Internal Method: exchange
* Summary: Exchange an item from the queue
* Description: 
* Return: None
******************************************************************************/
  void exchange(C_SQ_ITEM *x_item);

/******************************************************************************
* Internal Debugging Method: check_integrity
* Summary: Make sure the variables of this class are consistent with one
*          another
* Description: 
* Return: None
******************************************************************************/
  int check_integrity();

 private:

    int length;			// length of the queue
#ifndef KSR1
    int xqueue_allign1;		// allignment
#endif

    C_SQ_ITEM *top;		// pointer to the top of the queue
#ifndef KSR1
    int xqueue_allign2;		// allignment
#endif

    C_SQ_ITEM *bot;		// pointer to the bottom of the queue
#ifndef KSR1
    int xqueue_allign3;		// allignment
#endif

};
//RB_DEFINE_CLASS(C_XQUEUE);

typedef C_XQUEUE RB_queue;

#endif
