// SpFedStateMgr.C source file implements the object proxy state manager

#include "SpFedStateMgr.H"
#include "SpAltMgr.H"
#include "SpObjFactory.H"


extern SpObjFactory *ObjFact;
//SpObjFactory *ObjFact = NULL;
SpFedStateMgr *GlobalFedStateMgr = NULL;

/*
** Include/define built-in events
*/

#include "SpCreateObjProxyEvent.H"
#include "SpReflectAttributesEvent.H"

DEFINE_STATE_MGR_EVENT(DiscoverObject, SpCreateObjProxyEvent);
DEFINE_STATE_MGR_EVENT(ReflectAttributes, SpReflectAttributesEvent);

#ifdef COMMENTOUT
// Test if we receive discover messages
class SpTestDiscover : public SpStateMgrEvent {
  private:
  protected:
  public:

    SpTestDiscover() {
      cout << "Creating SpTestDiscover event" << endl;
    }

    virtual void Process() {
      cout << Data << " at time " << time_tag << endl;
    }

};
DEFINE_STATE_MGR_EVENT(DiscoverObject, SpTestDiscover);
#endif

#ifdef COMMENTOUT
// Test if we receive reflect messages
class SpTestReflect : public SpStateMgrEvent {
  private:
  protected:
  public:

    SpTestReflect() {
      cout << "Creating SpTestReflect event" << endl;
    }

    virtual void Process() {
      cout << Data << " at time " << time_tag << endl;
    }

};
DEFINE_STATE_MGR_EVENT(ReflectAttributes, SpTestReflect);
#endif


/*
*****************************************************************************
* SpFedStateMgr constructors
*****************************************************************************
*/

SpFedStateMgr::SpFedStateMgr(char *machineName, int port, int groupId) {
  EmHostUser = new SpEmHostUser(machineName, port, groupId);
  Init();
}

SpFedStateMgr::SpFedStateMgr(SpDataParser *parser) {
  EmHostUser = new SpEmHostUser(parser);
  Init();
}

SpFedStateMgr::SpFedStateMgr(char *fileName) {
  EmHostUser = new SpEmHostUser(fileName);
  Init();
}

/*
*****************************************************************************
* Init - private method called by constructor
*****************************************************************************
*/

void SpFedStateMgr::Init() {

  SpEndianNature::DetermineEndianNature();

  ObjFact = new SpObjFactory("Objects.par");

  TimeLag = 0.0;
  TimeGranted = -1.0e20;
  TimeRequested = -1.0e20;
  TimeGrantedFlag = 1;

  ObjectTypes = NULL;
  ObjectInstances = NULL;

  SpStateMgrEvent *Event = new SpStateMgrEvent(-1.0e20);
  ProcessedEvents.push_bot(Event);
  CurrentEvent = Event;

  CurrentEvent->SetFutureEvents(&FutureEvents);
  CurrentEvent->SetProcessedEvents(&ProcessedEvents);

  ObjectProxyIdHash = new C_XHASH(1000);
  ObjectProxyNameHash = new C_XHASH(1000);
  ObjectProxyTypeHash = new C_XHASH(1000);

  CurrentEvent->SetObjectProxyIdHash(ObjectProxyIdHash);
  CurrentEvent->SetObjectProxyNameHash(ObjectProxyNameHash);
  CurrentEvent->SetObjectProxyTypeHash(ObjectProxyTypeHash);

  CurrentEvent->SetEmHostUser(EmHostUser);

  GlobalState = new SpGlobalState();
  CurrentEvent->SetGlobalState(GlobalState);

  SpCurrentAltMgr->SetOptimistic();
  GlobalFedStateMgr = this;

  PLUG_IN_STATE_MGR_EVENT(DiscoverObject);
  PLUG_IN_STATE_MGR_EVENT(ReflectAttributes);

}

/*
*****************************************************************************
* SpFedStateMgr destructor
*****************************************************************************
*/

SpFedStateMgr::~SpFedStateMgr() {

  delete EmHostUser;
  if(ObjectTypes)
    delete ObjectTypes;
  if(ObjectInstances)
    delete ObjectInstances;
}

/*
*****************************************************************************
* PlugInEvent method
*****************************************************************************
*/

void SpFedStateMgr::PlugInEvent(char *eventName, SpStateMgrEvent *(*f)()) {

  cout << "PlugIn StateMgrEvent: " << eventName << endl;

  SpEventConstructorsByName *eventConstructorsByName =
    (SpEventConstructorsByName *)EventConstructors.Find(eventName);

  if (!eventConstructorsByName) {
    eventConstructorsByName = new SpEventConstructorsByName();
    EventConstructors.Insert(eventConstructorsByName, eventName);
  }

  SpEventNameConstructor *eventNameConstructor = new SpEventNameConstructor();
  eventNameConstructor->f = f;
  eventConstructorsByName->
    EventNameConstructors.InsertBot(eventNameConstructor);

}

/*
*****************************************************************************
* Subscribe method
*****************************************************************************
*/

void SpFedStateMgr::Subscribe(char *className) {

  SpSimTime eventTime = TimeGranted;
  eventTime += TimeLag;
  EmHostUser->SendCommand("Subscribe Object",
			  NodeId,
			  MgrId,
			  LocalId,
			  eventTime,
			  className,
			  strlen(className) + 1);

}

/*
*****************************************************************************
* Subscribe Interaction method
*****************************************************************************
*/

void SpFedStateMgr::SubscribeInteraction(char *) {
  // This needs to be defined in the child classes (gateway/federate state
  // mgrs) so that they can plug in their own events to handle interactions
  // coming from SPEEDES.
}

/*
*****************************************************************************
* Send Interaction method
*****************************************************************************
*/

void SpFedStateMgr::SendInteraction(char *className, 
				    SpParmSet &paramSet,
				    SpSimTime eventTime) {

  int dataBytes = strlen(className) + 1 + paramSet.GetPackLength();
  char *data = new char[dataBytes];
  char *buff = data;

  strcpy(buff, className);
  buff += strlen(className) + 1;
  paramSet.PackBuff(buff);

  if (eventTime < TimeGranted)
    eventTime = TimeGranted;
  eventTime += TimeLag;

  EmHostUser->SendCommand("Send Interaction",
			  NodeId,
			  MgrId,
			  LocalId,
			  eventTime,
			  data,
			  dataBytes);

  delete [] data;
}

/*
*****************************************************************************
* SetTimeLag method
*****************************************************************************
*/

void SpFedStateMgr::SetTimeLag(double timeLag) {
  NET_DOUBLE tLag = timeLag;

  SpSimTime eventTime = TimeGranted;
  eventTime += TimeLag;
  EmHostUser->SendCommand("Set Time Lag",
			  NodeId,
			  MgrId,
			  LocalId,
			  eventTime,
			  (char *)&tLag,
			  sizeof(NET_DOUBLE));
  TimeLag = timeLag;
}

/*
*****************************************************************************
* Next event request method
*****************************************************************************
*/

void SpFedStateMgr::NextEventRequest(SpSimTime &time) {

  EmHostUser->SendCommand("Next Event Request",
			  NodeId,
			  MgrId,
			  LocalId,
			  TimeGranted,
			  (char *)&time,
			  sizeof(SpSimTime));
  TimeRequested = time;
  TimeGrantedFlag = 0;
}


void
SpFedStateMgr::ValidateGrant () {

  if (TimeGranted != TimeRequested) { // grant corresponds to an event
    // figure out if event has arrived yet
    if (TimeGranted != FutureEvents.get_time()) {
#if 0
      if (TimeGranted < FutureEvents.get_time()) { // event has not arrived
	cerr << "WAITING FOR EVENT AT " << TimeGranted << endl;
	ReadMessages();  // flush rest of msgs waiting to be proccessed
      }
      else { // TimeGranted > next event time
	// There are other events already scheduled to take place before
	// the event that corresponds to the grant.  Since there is no
	// way to iterate through the priority queue to examine the time
	// tags of those events, the only reliable solution is to force
	// the remaining msgs to be processed.
	cerr << "WAITING FOR EVENT AT " << TimeGranted << endl;
	ReadMessages();  // flush rest of msgs waiting to be proccessed
      }
#else  // revert to above case if some way is found to iterate thru event queue
      cerr << "WAITING FOR EVENT AT " << TimeGranted << endl;
      ReadMessages();  // flush rest of msgs waiting to be proccessed
#endif
    }
  }
}

/*
*****************************************************************************
* Time advance grant method
*****************************************************************************
*/

int SpFedStateMgr::TimeAdvanceGranted() {
  char *data;
  SpSimTime time;
  int dataBytes;

  if (TimeGrantedFlag) return 1;

  data = EmHostUser->ReceiveNamedData("TimeAdvanceGrant",
				      time,
				      dataBytes,
				      GlobalId);

  if (data) {
    SpSimTime grant;
    memcpy((char *)&grant, data, sizeof(SpSimTime));
    delete [] data;

    if ((grant > TimeGranted) ||
	((grant == TimeGranted) && (grant == TimeRequested))) {
      cout << "Received TimeAdvanceGrant at " << time 
	   << " with " << dataBytes << " bytes" << endl;
      TimeGranted = grant;
      TimeGrantedFlag = 1;
      ValidateGrant();
    }
    else {  // spurious grant; need to resend request
      cerr << "Grant of " << grant << " is in past, resending request\n";
      NextEventRequest(TimeRequested);
    }
  }

  return TimeGrantedFlag;

}

/*
*****************************************************************************
* Read messages
*****************************************************************************
*/

void SpFedStateMgr::ReadMessages() {
  char *data;
  SpSimTime dataTime;
  int dataBytes;
  int globalId;
  char *eventName;
  int priority;

  while (1) {

    data = EmHostUser->ReceiveAnyNamedData(eventName,
					   dataTime,
					   dataBytes,
					   globalId);

    if (data == NULL) break;
    cout << "Received " << eventName << " at " << dataTime
	 << " with " << dataBytes << " bytes"
	 << endl;

    if (!strcmp(eventName, "TimeAdvanceGrant")) {

      SpSimTime grant;
      memcpy((char *)&grant, data, sizeof(SpSimTime));
      delete [] data;

      if ((grant > TimeGranted) ||
	  ((grant == TimeGranted) && (grant == TimeRequested))) {
	TimeGranted = grant;
	TimeGrantedFlag = 1;
      }
      else {  // spurious grant; need to resend request
	cerr << "Grant of " << grant << " is in past, resending request\n";
	NextEventRequest(TimeRequested);
      }
    }else{

      SpEventConstructorsByName *eventConstructorsByName =
	(SpEventConstructorsByName *)EventConstructors.Find(eventName);

      if (eventConstructorsByName) {

	SpEventNameConstructor *eventNameConstructor =
	  (SpEventNameConstructor *)
	  eventConstructorsByName->EventNameConstructors.GetFirstElement();

	priority = 0;

	while (eventNameConstructor) {

	  cout << "Scheduling " << eventName
	       << " at time " << dataTime
	       << endl;

	  SpStateMgrEvent *stateMgrEvent = eventNameConstructor->f();
	  stateMgrEvent->Init(dataTime, data, dataBytes, globalId, eventName);
	  stateMgrEvent->SetPriority(stateMgrEvent->GetPriority() + priority);
	  priority++;

	  FutureEvents.insert(stateMgrEvent);

	  eventNameConstructor = (SpEventNameConstructor *)
	    eventConstructorsByName->EventNameConstructors.GetNextElement();
	}
      }

    }

  }

}

/*
*****************************************************************************
* GoToTime() method advances to the specified time
*****************************************************************************
*/

void SpFedStateMgr::GoToTime(SpSimTime time) {
  SpStateMgrEvent *NextEvent;
  SpStateMgrEvent *PreviousEvent;

/*
** Process forward (or backwards) to the requested time
*/

  if (time >= CurrentEvent->get_time_tag()) {

/*
** First go forward in the already processed event list (i.e., rollforward
** if this makes sense)
*/

    for (;;) {

      if (CurrentEvent == (SpStateMgrEvent *)ProcessedEvents.get_bot()) {
	break;
      }

      NextEvent = (SpStateMgrEvent *)CurrentEvent->get_link();

      if (time < NextEvent->get_time_tag()) {
	return;
      }else{
	NextEvent->RollForwardEvent();
	CurrentEvent = NextEvent;
      }

    }

/*
** Now process events that have not yet been processed
*/

    for (;;) {

      if (FutureEvents.get_length() == 0) {
	break;
      }

      if (time < FutureEvents.get_time()) {
	break;
      }

      CurrentEvent = (SpStateMgrEvent *)FutureEvents.pop();
      CurrentEvent->ProcessEvent();
      ProcessedEvents.push_bot(CurrentEvent);

    }

  }else{

/*
** Go backwards in time by rollback events
*/

    for (;;) {

      if (CurrentEvent == ProcessedEvents.get_top()) {
	break;
      }

      PreviousEvent = (SpStateMgrEvent *)CurrentEvent->get_backlink();

      if (CurrentEvent->get_time_tag() < time) {
	break;
      }

      CurrentEvent->RollBackEvent();
      CurrentEvent = PreviousEvent;

    }

  }

}


/*
*****************************************************************************
* GetNextProxy() - Gets the next proxy of a given type
*****************************************************************************
*/

SpObjProxy *SpFedStateMgr::GetNextProxy() {

  if (CurrentQ) {

    return GetNextProxyByType();

  }else{

    C_XQUEUE *Qall = ObjectProxyIdHash->get_all();

    if (CurrentProxyItem == Qall->get_bot()) {
      CurrentObjectProxy = NULL;
    }else{
      CurrentProxyItem = (F_SpProxyItem *)CurrentProxyItem->get_link();
      CurrentObjectProxy = CurrentProxyItem->GetObjectProxy();
    }

    return CurrentObjectProxy;

  }

}

/*
*****************************************************************************
* GetNextProxyByType() - Gets the next proxy of a given type
*****************************************************************************
*/

SpObjProxy *SpFedStateMgr::GetNextProxyByType() {
  char *type = CurrentObjectProxy->ObjProxyGetClassName();

  F_SpProxyItem *BotProxyItem = (F_SpProxyItem *)CurrentQ->get_bot();

  if (CurrentProxyItem == NULL) {
    cout << "Warning, trying to GetNextProxy before calling GetFirstProxy"
	 << endl;
    return NULL;
  }

  if (CurrentProxyItem == BotProxyItem) return NULL;

/*
** Loop over proxies until we find the next one with the right class name
*/

  for (;;) {

    CurrentProxyItem = (F_SpProxyItem *)CurrentProxyItem->get_link();
    CurrentObjectProxy = CurrentProxyItem->GetObjectProxy();

    if (!strcmp(type, CurrentObjectProxy->ObjProxyGetClassName())) {
      return CurrentObjectProxy;
    }

    if (CurrentProxyItem == BotProxyItem) break;

  }

  return NULL;

}


char*
SpFedStateMgr::GetObjectName (int instId) {
  if(ObjectInstances == NULL) {
    ObjectInstances = EmHostUser->ObjectNames();
  }
  int numInsts = ObjectInstances->GetNitems();
  for(int i=0; i < numInsts; i++) {
    if(instId == ObjectInstances->GetValue(i)) {
      return ObjectInstances->GetName(i);
    }
  }
  return NULL;
}

int
SpFedStateMgr::GetObjectId (const char *instName) {
  if(ObjectInstances == NULL) {
    ObjectInstances = EmHostUser->ObjectNames();
  }
  int instFound = 0;
  int instId = (int) ObjectInstances->GetValue((char*)instName, instFound);
  if(!instFound)
    return -1;
  return instId;
}

char*
SpFedStateMgr::GetObjectMgrName (int typeId) {
  if(ObjectTypes == NULL) {
    ObjectTypes = EmHostUser->ObjectManagerTypes();
  }
  int numTypes = ObjectTypes->GetNitems();
  for(int i=0; i < numTypes; i++) {
    if(typeId == ObjectTypes->GetValue(i)) {
      return ObjectTypes->GetName(i);
    }
  }
  return NULL;
}

int
SpFedStateMgr::GetObjectMgrId (const char *typeName) {
  if(ObjectTypes == NULL) {
    ObjectTypes = EmHostUser->ObjectManagerTypes();
  }
  int typeFound = 0;
  int typeId = (int) ObjectTypes->GetValue((char*)typeName, typeFound);
  if(!typeFound)
    return -1;
  return typeId;
}
