// SpStateMgr.C source file implements the object proxy state manager

#include "SpStateMgr.H"
#include "SpObjFactory.H"

#include "SpCreateObjProxyEvent.H"
#include "SpReflectAttributesEvent.H"

extern SpObjFactory *ObjFact;

/*
*****************************************************************************
* SpStateMgr constructors
*****************************************************************************
*/

SpStateMgr::SpStateMgr(char *machineName, int port, int groupId,
			     double timeLag) {

  EmHostUser = new SpEmHostUser(machineName, port, groupId);
  ObjFact = new SpObjFactory("Objects.par");

  TimeLag = timeLag/2;

  Init();

}

SpStateMgr::SpStateMgr(SpDataParser *parser, double timeLag) {

  EmHostUser = new SpEmHostUser(parser);
  ObjFact = new SpObjFactory("Objects.par");

  TimeLag = timeLag/2;

  Init();

}

SpStateMgr::SpStateMgr(char *fileName) {

  EmHostUser = new SpEmHostUser(fileName);
  ObjFact = new SpObjFactory("Objects.par");

  TimeLag = 0.0;

  Init();

}

/*
*****************************************************************************
* Init - private method called by constructor
*****************************************************************************
*/

extern int EXTERNAL_OBJECT_PROXY;

void SpStateMgr::Init() {
  EXTERNAL_OBJECT_PROXY = 1;

  Time = -1.0e20;
  TimeRequestFlag = 0;

  SpStateMgrEvent *Event = new SpStateMgrEvent(-1.0e20);
  ProcessedEvents.push_bot(Event);
  CurrentEvent = Event;

  CurrentEvent->SetFutureEvents(&FutureEvents);
  CurrentEvent->SetProcessedEvents(&ProcessedEvents);

  ObjectProxyIdHash = new C_XHASH(1000);
  ObjectProxyNameHash = new C_XHASH(1000);
  ObjectProxyTypeHash = new C_XHASH(1000);

  CurrentEvent->SetObjectProxyIdHash(ObjectProxyIdHash);
  CurrentEvent->SetObjectProxyNameHash(ObjectProxyNameHash);
  CurrentEvent->SetObjectProxyTypeHash(ObjectProxyTypeHash);

  CurrentEvent->SetEmHostUser(EmHostUser);

  GlobalState = new SpGlobalState();
  CurrentEvent->SetGlobalState(GlobalState);

  CurrentEvent->GetAltMgr()->SetOptimistic();

}

/*
*****************************************************************************
* SpStateMgr destructor
*****************************************************************************
*/

SpStateMgr::~SpStateMgr() {

  delete EmHostUser;

}

/*
*****************************************************************************
* Subscribe methods
*****************************************************************************
*/

void SpStateMgr::SubscribeAll() {

  EmHostUser->Subscribe("ObjectProxy");

}

void SpStateMgr::SubscribeType(char *type, ...) {
  va_list va;
  char *nextType;

  va_start(va, type);
  nextType = type;

  while (nextType) {
    cout << "Subscribing to " << nextType << endl;
    EmHostUser->SubscribeType("ObjectProxy", nextType);
    nextType = va_arg(va, char *);
  }

  va_end(va);
}

void SpStateMgr::SubscribeObject(char *objectName, ...) {
  va_list va;
  char *nextObjectName;

  va_start(va, objectName);
  nextObjectName = objectName;

  while (nextObjectName) {
    cout << "Subscribing to " << nextObjectName << endl;
    EmHostUser->SubscribeObject("ObjectProxy", nextObjectName);
    nextObjectName = va_arg(va, char *);
  }

  va_end(va);
}

/*
*****************************************************************************
* GoToTime() method advances to the specified time
*****************************************************************************
*/

void SpStateMgr::GoToTime(double time) {
  int i;
  SpStateMgrEvent *NextEvent;
  SpStateMgrEvent *PreviousEvent;
  char *Data;
  double DataTime;
  int DataBytes;
  int GlobalId;
  SpReflectAttributesEvent *ReflectAttributesEvent;
  SpCreateObjProxyEvent *CreateObjectProxyEvent;
  char *EventName;
  char **UserEventName = FreeUserEvents.GetNames();
  int NumberOfEvents = FreeUserEvents.GetNtypes();
  double CurrentTime = EmHostUser->Time;

  EmHostUser->Time = Time = time;

  EmHostUser->ReadMessages();

  if (!EmHostUser->SpeedesExecuting()) {
    exit(1);
  }

  //Make sure we send at least one message
  TimeAdvanceRequest(); 
  EmHostUser->Time = CurrentTime;
  EmHostUser->ReadMessages(); 


  while (GetGrantedTime() < time) {
    EmHostUser->Time = time;
    TimeAdvanceRequest();
    EmHostUser->Time = CurrentTime;
    EmHostUser->ReadMessages();

    if (!EmHostUser->SpeedesExecuting()) {
      exit(1);
    }

  }

/*
** Now receive all messages that have been queued up and turn them into events
*/

  EmHostUser->ReadMessages();

  for (;;) {

    Data = EmHostUser->ReceiveAnySubscribedData(EventName,
						DataTime,
						DataBytes,
						GlobalId);

    if (Data == NULL) break;

    cout << "Received subscribed event " << EventName
	 << " at time " << DataTime
	 << " with " << DataBytes << " bytes" << endl;

    if (!strcmp(EventName, "ObjectProxy")) {

      ReflectAttributesEvent = new SpReflectAttributesEvent();

      ReflectAttributesEvent->Init(DataTime,
				   Data,
				   DataBytes,
				   GlobalId,
				   EventName);

      FutureEvents.insert(ReflectAttributesEvent);
      EventName = "REFLECT_ATTRIBUTES";

    }

    for (i=0; i<NumberOfEvents; i++) {
      if (!strcmp(EventName, UserEventName[i])) {

	SpStateMgrEvent *StateManagerEvent = (SpStateMgrEvent *)
	  FreeUserEvents.new_object(i);

	StateManagerEvent->Init(DataTime,
				Data,
				DataBytes,
				GlobalId,
				EventName);

	StateManagerEvent->SetPriority(i+1);

	FutureEvents.insert(StateManagerEvent);
      }
    }

  }


  for (;;) {

    Data = EmHostUser->ReceiveData(DataTime, DataBytes, GlobalId);
    if (Data == NULL) break;

    cout << "Got data at time " << DataTime
	 << " with " << DataBytes << " bytes" << endl;

    CreateObjectProxyEvent = new SpCreateObjProxyEvent();
    CreateObjectProxyEvent->Init(DataTime, Data, DataBytes, GlobalId, NULL);

    FutureEvents.insert(CreateObjectProxyEvent);

    for (i=0; i<NumberOfEvents; i++) {
      if (!strcmp("DISCOVER_OBJECT", UserEventName[i])) {

	SpStateMgrEvent *StateManagerEvent = (SpStateMgrEvent *)
	  FreeUserEvents.new_object(i);

	StateManagerEvent->Init(DataTime,
				Data,
				DataBytes,
				GlobalId,
				"DISCOVER_OBJECT");

	StateManagerEvent->SetPriority(i+1);

	FutureEvents.insert(StateManagerEvent);
      }
    }
  }

/*
** Process forward (or backwards) to the requested time
*/


  if (time >= CurrentEvent->get_time_tag()) {

/*
** First go forward in the already processed event list (i.e., rollforward
** if this makes sense)
*/

    for (;;) {

      if (CurrentEvent == (SpStateMgrEvent *)ProcessedEvents.get_bot()) {
	break;
      }

      NextEvent = (SpStateMgrEvent *)CurrentEvent->get_link();

      if (time < NextEvent->get_time_tag()) {
	return;
      }else{
	NextEvent->RollForwardEvent();
	CurrentEvent = NextEvent;
      }

    }

/*
** Now process events that have not yet been processed
*/

    for (;;) {

      if (FutureEvents.get_length() == 0) {
	break;
      }

      if (time < FutureEvents.get_time()) {
	break;
      }

      CurrentEvent = (SpStateMgrEvent *)FutureEvents.pop();
      CurrentEvent->ProcessEvent();
      ProcessedEvents.push_bot(CurrentEvent);

    }

  }else{

/*
** Go backwards in time by rollback events
*/

    for (;;) {

      if (CurrentEvent == ProcessedEvents.get_top()) {
	break;
      }

      PreviousEvent = (SpStateMgrEvent *)CurrentEvent->get_backlink();

      if (CurrentEvent->get_time_tag() < time) {
	break;
      }

      CurrentEvent->RollBackEvent();
      CurrentEvent = PreviousEvent;

    }

  }

}


/*
*****************************************************************************
* AdvanceTime() - Tries to advance time to what is passed in if possible, but
* does a best effort job without blocking.
*****************************************************************************
*/
double SpStateMgr::AdvanceTime(double requestedTime) {
  double t;

  if ((GetGrantedTime() <= requestedTime) || (requestedTime == -1.0e20)) {
    TimeAdvanceRequest();
  }

/*
** This is a subtle time management consideration, but we cannot goto a time
** that is equal to the granted time since that would screw up messages
** that arrive at that time. In other words, we must go to a time that is
** less thant the granted time for everything to really work right. So,
** we multiply the granted time by LONE (i.e., Less that ONE = LONE). LONE
** was veriefied to be 0.999999999999999994 on the SGI O2. To be safe, we
** make it a bit smaller.
*/

  double LONE = 0.999999999999999;
  t = GetGrantedTime() * LONE;
  
  if (requestedTime != -1.0e20) {
    if (requestedTime < t) {
      GoToTime(requestedTime);
    }else{
      GoToTime(t);
    }
  }else{
    GoToTime(t);
  }

  return Time;
  
}

/*
*****************************************************************************
* GetNextProxy() - Gets the next proxy of a given type
*****************************************************************************
*/

SpObjProxy *SpStateMgr::GetNextProxy() {

  if (CurrentQ) {

    return GetNextProxyByType();

  }else{

    C_XQUEUE *Qall = ObjectProxyIdHash->get_all();

    if (CurrentProxyItem == Qall->get_bot()) {
      CurrentObjectProxy = NULL;
    }else{
      CurrentProxyItem = (F_SpProxyItem *)CurrentProxyItem->get_link();
      CurrentObjectProxy = CurrentProxyItem->GetObjectProxy();
    }

    return CurrentObjectProxy;

  }

}

/*
*****************************************************************************
* GetNextProxyByType() - Gets the next proxy of a given type
*****************************************************************************
*/

SpObjProxy *SpStateMgr::GetNextProxyByType() {
  char *type = CurrentObjectProxy->ObjProxyGetClassName();

  F_SpProxyItem *BotProxyItem = (F_SpProxyItem *)CurrentQ->get_bot();

  if (CurrentProxyItem == NULL) {
    cout << "Warning, trying to GetNextProxy before calling GetFirstProxy"
	 << endl;
    return NULL;
  }

  if (CurrentProxyItem == BotProxyItem) return NULL;

/*
** Loop over proxies until we find the next one with the right class name
*/

  for (;;) {

    CurrentProxyItem = (F_SpProxyItem *)CurrentProxyItem->get_link();
    CurrentObjectProxy = CurrentProxyItem->GetObjectProxy();

    if (!strcmp(type, CurrentObjectProxy->ObjProxyGetClassName())) {
      return CurrentObjectProxy;
    }

    if (CurrentProxyItem == BotProxyItem) break;

  }

  return NULL;

}
