// SpStateMgr.H header file

#ifndef DEF_SpStateMgr
#define DEF_SpStateMgr

#include "SpStdarg.H"
#include "SpStdlib.H"
#include "SpIostream.H"
#include "SpFstream.H"
#include "SpStrstream.H"
#include "SpIomanip.H"

#include "SpDataParser.H"
#include "SpEmHostUser.H"
#include "SpStateMgrEvent.H"
#include "SpObjProxy.H"
#include "SpPriorityQueueHeap.H"
#include "xqueue.H"
#include "xhash.H"
#include "SpGlobalState.H"
#include "F_SpProxyItem.H"
#include "free_list.H"


/*************************************************************************
 * Class: SpStateMgr
 * Summary: The SpStateMgr class is used to access the object proxies
 *          of SPEEDES in a time ordered fashion.
 *
 * Description: Many external modules would like to have regular
 *              updates to data that is published in the object
 *              proxies.  Examples of this include the HLA gateway,
 *              DIS gateways, graphics interfaces and others.  The
 *              Object Proxy State Manager (OPSM) provides this.  In
 *              addition, it contains an SpEmHostUser for additional
 *              interface to SPEEDES that the OPSM may not provide
 *              directly. The OPSM also provides the ability to
 *              process events that may be launched in response to
 *              user level messages that come from SPEEDES.  For
 *              example, SPEEDES may respond with subscribed data with
 *              the name "NUMBER_OF_STAGES" This could then spawn
 *              events within the OPSM to pump data to other
 *              interfaces or drive displays.
 *
 *              Naturally, these events are all roll-backable (it does
 *              not roll back the SPEEDES simulation) so that one can
 *              examine the state of the simulation at any point in
 *              the past.  This may be useful in graphical displays so
 *              that one can look back and see what was happening a
 *              few seconds or minutes ago and then return to the
 *              present to monitor the progress of the simulation.
 *
 * Examples: The present main example of using object proxies exists
 *           in the demos/BmdSim directory.  Examine the file OPSM.C
 *           there to see how this could best be used.  In the
 *           future, we will also have examples of this for the HLA
 *           gateway and perhaps some GUI displays of our own.
 *
*************************************************************************/


class SpStateMgr {

  public:

/************************************************************************
 * Method: SpStateMgr()
 * Summary: Constructor where the user provides all the data.
 * Description: Constructor requires the name of the host router
 *              machine, SPEEDES group Id, and maximum lag time (used
 *              to gurarantee that the object proxy state manager
 *              never falls more than a certain amount of simulation
 *              time behind the SPEEDES simulation).
 * Return: A new SpStateMgr
*************************************************************************/

  SpStateMgr(char *machineName, //In: The machine where the HostRouter is
		                   //    running
		int port,          //In: The port of the HostRouter
		int groupId,       //In: The group id of your simulation
		double timeLag     //In: The maximal amount of time you will
		                   //    allow the OPSM to fall behind SPEEDES.
		                   //    Setting this to 0 freezes the sim.
		);
/************************************************************************
 * Method: SpStateMgr()
 * Summary: Constructor where the Host Router data is read from a par file..
 * Description: Constructor that reads the Host Router data from a par file
 *              and maximum lag time is provided in the constructor (used
 *              to gurarantee that the object proxy state manager
 *              never falls more than a certain amount of simulation
 *              time behind the SPEEDES simulation).
 * Return: A new SpStateMgr
*************************************************************************/
  SpStateMgr(SpDataParser *parser, //In: Par file that contains all the
		                     //    data to connecct to the Host Router
		double timeLag       //In: The maximal amount of time you will
		                     //    allow the OPSM to fall behind
		                     //    SPEEDES. Setting this to 0 freezes
		                     //    the sim.
		);


/************************************************************************
 * Method: SpStateMgr()
 * Summary: Constructor from a file that recorded the output of a
 *          previous SpStateMgr
 * Description: Constructor that reads in the simulation from a file.
 *              This file is the saved state of a previous simulation
 *              and this is then used so that one can replay the
 *              simulation "from a tape."  The functionality to record
 *              a simulation to a file has not been implemented yet
 *              but this is how one will then read that file.
 * Return: A new SpStateMgr
*************************************************************************/
  SpStateMgr(char *FileName //In: File name of the previously recorded
		               //    simulation.
		);

/************************************************************************
 * Method: ~SpStateMgr()
 * Summary: SpStateMgr destructor
 * Description: Destroys a SpStateMgr. It is always a good thing to
 *              destroy an SpStateMgr when you are done with it
 *              since that allows the HostRouter to know that you have
 *              left the simulation
 * Return: none
*************************************************************************/
    ~SpStateMgr();


/************************************************************************
 * Method: SubscribeAll()
 * Summary: Subscribes to all objects
 * Description: This method allows you to subscribe to all object
 *              proxies in the simulation.  Use it with caution since
 *              you usually don't want all object but rather, a subset
 *              of those objects.  Subscribing to only those you want
 *              cuts down on the amount of work the SpStateMgr
 *              needs to perform as well as cuts down on the amount of
 *              communication between your SpStateMgr and the
 *              HostRouter.  DO NOT WASTE BANDWITH!!!!
 * Return: void
*************************************************************************/
    void SubscribeAll();


/************************************************************************
 * Method: SubscribeType()
 * Summary: Subscribes to all objects of a given (or list of) type(s)
 * Description: If you want to subscribe to all objects of a specific
 *              type, this is the method to call.  It takes a variable
 *              list of arguments (similar to printf()) so you can
 *              stack a bunch on the same call.  For instance, you can
 *              make a call SubscribeType("MISSILE", "TANK", "SBIRS",
 *              NULL); which will then subscribe to the three types
 *              MISSILE, TANK, and SBIRS. Notice that your list must
 *              be null terminated so that we can tell when you are
 *              done subscribing!
 * Return: void
*************************************************************************/
  void SubscribeType(char *type,//In: name of Type to which you wish
		                //    to subscribe.
		        ...);


/************************************************************************
 * Method: SubscribeObject()
 * Summary: Subsribes to a single (or list of) object(s)
 * Description: Similar to SubscribeType(), this will allow you to
 *              stack up a list of items that you would like to
 *              subscribe to.  Suppose you only wanted to subscribe to
 *              MISSILE_33, TANK_11, and SBIRS_2, SBIRS_3, and
 *              SBIRS_4.  Using this call, you could execute a
 *              SubscribeObject(MISSILE_33, TANK_11, SBIRS_2, SBIRS_3,
 *              SBIRS_4, NULL) and just subscribe to those objects
 *              rather than all missiles, tanks and sbirs.
 * Return: void
*************************************************************************/
  void SubscribeObject(char *objectName,//In: name of object to which you wish
		                        //    to subscribe.
		       ...);



/************************************************************************
 * Method: GoToTime()
 * Summary: Moves the SpStateMgr to a particular time.
 * Description: The SpStateMgr allows one to roll forward or
 *              backward in time.  This allows it to examine the
 *              values of the various objects at points in the past
 *              (perhaps a graphics package could use it) or to simply
 *              request he simulation to advance in time.  If the
 *              simulation has not yet reached the time you have
 *              requested, then GoToTime() will block until the
 *              simulation has reached that point.  If you try to go
 *              to a time in the past, the function returns
 *              immediately.
 * Return: void
*************************************************************************/
  void GoToTime(double time//In: Time at which you want to OPSM to exist
		);


/************************************************************************
 * Method: AdvanceTime()
 * Summary: Advances time as far as possible without blocking.
 * Description: If you simply want to catch up with the simulation,
 *              call this method.  It returns the time at which you
 *              arrived.
 * Return: Time to which the SpStateMgr was able to advance
*************************************************************************/
    double AdvanceTime(double requestedTime = -1.0e20);

/************************************************************************
 * Method: GetCurrentTime()
 * Summary: Returns the time at which the SpStateMgr is functioning
 * Description: Returns the time at which the SpStateMgr is functioning
 * Return: Returns the time at which the SpStateMgr is functioning
*************************************************************************/
    double GetCurrentTime(){return Time;}


/************************************************************************
 * Method:  TimeAdvanceRequest()
 * Summary: Requests SPEEDES to advance time.
 * Description: This method requests SPEEDES to advance time and keeps
 *              track internally whether a time advance request is
 *		outstanding. It works with the TimeAdvanceGranted and
 *		GetGrantedTime methods.
 *              
 * Return: nothing
*************************************************************************/
  void TimeAdvanceRequest() {
    if (TimeRequestFlag == 0) {
      EmHostUser->TimeAdvanceRequest(TimeLag);
      TimeRequestFlag = 1;
    }
  }

/************************************************************************
 * Method:  GetGrantedTime()
 * Summary: Returns the time that SPEEDES granted us.  This is GVT.
 * Description: When we do a AdvanceTime() call, this returns a
 *              message which we can grab with this call.  This
 *              message is the GVT which SPEEDES was willing to grant
 *              us.
 * Return: 
*************************************************************************/
  double GetGrantedTime() {
    if (EmHostUser->TimeAdvanceGranted()) TimeRequestFlag = 0;
    return EmHostUser->GetTimeGranted();
  }

/************************************************************************
 * Method: GetLastProcessedEventTime()
 * Summary: Returns the time of the last processed SpStateMgr event
 * Description: Returns the time of the last processed SpStateMgr event
 * Return: Returns the time of the last processed SpStateMgr event
*************************************************************************/
  double GetLastProcessedEventTime() {
    SpStateMgrEvent *LastProcessedEvent =
      (SpStateMgrEvent *)ProcessedEvents.get_bot();
    return LastProcessedEvent->get_time_tag();
  }

/************************************************************************
 * Method: SetTimeLag()
 * Summary: Sets the lag time for the SpStateMgr
 * Description: The SpStateMgr always follows behind the simulation
 *              in time.  This lag can be configured and one can be
 *              certain that the SpStateMgr will never be farther
 *              than timelag behind the simulation.  Warning: Setting
 *              this to zero will freeze the simulation!!!
 * Return: 
*************************************************************************/
  void SetTimeLag(double timeLag //In: Maximal time to lag the simulation.
		  ) {TimeLag = timeLag;}


/************************************************************************
 * Method: GetTimeLag()
 * Summary: Returns the current amount of time we are lagging the simulation
 * Description: Returns the current amt of time we are lagging the simulation
 * Return: Returns the current amount of time we are lagging the simulation
*************************************************************************/
    double GetTimeLag() {return TimeLag;}

/*
** Methods for iterating through all proxies. GetNproxies() returns the total
** number of object proxies. GetFirstProxy() gets the first proxy and
** GetNextProxy() gets subsequent proxies. The GetNextProxy() returns NULL
** when there are no more proxies.
*/


/************************************************************************
 * Method: GetNproxies()
 * Summary: Returns the number of proxies at your time.
 * Description: Remember, the number of proxies will change with time.
 *              This will return the number of proxies for your time
 * Return: Returns the number of proxies at your time.
*************************************************************************/
    int GetNproxies() {return ObjectProxyIdHash->get_length();}

/************************************************************************
 * Method: GetFirstProxy()
 * Summary: Returns the first object proxy.
 * Description: This goes to the list and then grabs the first object
 *              proxy, regardless of type.
 * Return: Returns the first object proxy.
*************************************************************************/
  SpObjProxy *GetFirstProxy() {
    CurrentQ = NULL;
    ObjectProxyIdHash->combine();
    C_XQUEUE *Qall = ObjectProxyIdHash->get_all();
    CurrentProxyItem = (F_SpProxyItem *)Qall->get_top();
    if (CurrentProxyItem) {
      CurrentObjectProxy = CurrentProxyItem->GetObjectProxy();
    }else{
      CurrentObjectProxy = NULL;
    }
    return CurrentObjectProxy;
  }
  

/************************************************************************
 * Method: GetNextProxy()
 * Summary: Returns the SpObjProxy after the one previously given.
 * Description: This will return the next SpObjProxy after the
 *              previous returned proxy.  If you have just called
 *              GetFirstProxy(), then it just returns the next proxy
 *              regardless of type.  IF, on the other hand, you had
 *              just called GetFirstProxy("MISSILE"), then
 *              GetNextProxy() would return the next missile proxy.
 *              If there are no more object proxies to be given, this
 *              function returns NULL.
 * Return: Returns the SpObjProxy after the one previously given.
*************************************************************************/
  SpObjProxy *GetNextProxy();



/************************************************************************
 * Method: GetNproxies(char *)
 * Summary: Returns the number of object proxies of a given type for your time
 * Description: This will return the number of object proxies of a
 *              given type at the current time for the SpStateMgr.
 *              Again, this will change throughout the simulation.
 * Return: Returns the nubmer of object proxies of a given type for your time
*************************************************************************/
  int GetNproxies(char *type //In: The type of proxy for which you want a count
		  ) {
    int counter = 0;
    SpObjProxy *ObjectProxy;
    C_XQUEUE *Q = ObjectProxyTypeHash->GetQ(type);
    F_SpProxyItem *ProxyItem = (F_SpProxyItem *)Q->get_top();
    for (int i=0; i<Q->get_length(); i++) {
      ObjectProxy = ProxyItem->GetObjectProxy();
      if (!strcmp(type, ObjectProxy->ObjProxyGetClassName())) counter++;
      ProxyItem = (F_SpProxyItem *)ProxyItem->get_link();
    }
    return counter;
  }
  
  
/************************************************************************
 * Method: GetFirstProxy(char *)
 * Summary: Returns the first SpObjProxy of a given type
 * Description: If you want to iterate through all the object proxies
 *              of a given type (such as MISSILE), you would first
 *              issue a GetFirstProxy("MISSILE") which would return
 *              the first missile object proxy.  You could then issue
 *              a GetNextProxy() which would return the next missile
 *              proxy and so on until you get the last object proxy of
 *              the missile type.
 * Return: Returns the first SpObjProxy of a given type
*************************************************************************/
  SpObjProxy *GetFirstProxy(char *type //In: Type for which you want the
			                 //    first SpObjProxy
			      ) {
    SpObjProxy *ObjectProxy;
    SpObjProxy *ReturnedObjectProxy = NULL;
    CurrentQ = ObjectProxyTypeHash->GetQ(type);
    F_SpProxyItem *ProxyItem = (F_SpProxyItem *)CurrentQ->get_top();
    for (int i=0; i<CurrentQ->get_length(); i++) {
      ObjectProxy = ProxyItem->GetObjectProxy();
      if (!strcmp(type, ObjectProxy->ObjProxyGetClassName())) {
	CurrentObjectProxy = ObjectProxy;
	CurrentProxyItem = ProxyItem;
	ReturnedObjectProxy = ObjectProxy;
	break;
      }
      ProxyItem = (F_SpProxyItem *)ProxyItem->get_link();
    }
    return ReturnedObjectProxy;
  }
  

/************************************************************************
 * Method: GetProxy(char *)
 * Summary: Gets an object proxy by instance name.
 * Description: If you know the exact name of the proxy you want, this
 *              method allows one to obtain that proxy without
 *              iterating through all of the proxies of that type.
 *              For instance, if you wanted the proxy for MISSILE_33,
 *              you could do a GetFirstProxy("MISSILE") and then start
 *              iterating through until you got to the 33rd misile.
 *              Or, more simply, you could just use this method call
 *              and give a GetProxy("MISSILE_33") which would save
 *              some time and some effort.
 * Return: the SpObjProxy with the passed in name
*************************************************************************/
    SpObjProxy *GetProxy(char *objectName) {
      SpObjProxy *ObjectProxy;
      F_SpProxyItem *ProxyItem = (F_SpProxyItem *)
	ObjectProxyNameHash->find(objectName);
      if (ProxyItem) {
	ObjectProxy = ProxyItem->GetObjectProxy();
      }else{
	ObjectProxy = NULL;
      }
      return ObjectProxy;
    }



/************************************************************************
 * Method: GetEmHostUser()
 * Summary: Returns the SpEmHostUser for this SpStateMgr
 * Description: Sometimes, you want more specific functionality that
 *              the SpStateMgr provides.  In order to obtain a
 *              finer grained approach, you can grab the SpEmHostUser
 *              and give commands directly to the HostRouter.  This
 *              will allow you to control time better, to pause the
 *              simulation, resume it, kill speedes or use any of the
 *              other commands that are important to your simulation.
 * Return:Returns the SpEmHostUser for this SpStateMgr 
*************************************************************************/
    SpEmHostUser *GetEmHostUser() {return EmHostUser;}


/************************************************************************
 * Method: ScheduleEvent(SpStateMgrEvent *, double)
 * Summary: Schedules an event on the OPSMSTATE_MANAGER's event queue
 *          for processing
 * Description: the SpStateMgr maintains an event queue and
 *              processes these events when their time is
 *              encountered. This does NOT schedule events in the
 *              SPEEDES system (use the SpEmHostUser method
 *              SendCommand to accomplish that goal) but only within
 *              the StateManager.  Similar to SPEEDES events, these
 *              are roll backable and roll forwardable.  You need to
 *              use DefineNumberOfEvents in order to give the
 *              SpStateMgr knowlege of how many events are in the
 *              SpStateMgr and then you use DefineEvent to actually
 *              inform the SpStateMgr of what a particular event
 *              is.  Your own events will be classes which derive from
 *              the class SpStateMgrEvent and these events must
 *              implement the virtual function Process() which is how
 *              the event is processed.
 * Return: void
*************************************************************************/
  void ScheduleEvent(SpStateMgrEvent *event, //In: The event you want to
		                                 //    schedule
		     double time //In: Time at which the event should
		                 //    be scheduled
		     ) {
    if (event->get_link()) {
      cout << "Warning (StateManager): Posting event twice?" << endl;
    }else{
      event->set_time_tag(time);
      SpStateMgrEvent *LastProcessedEvent =
	(SpStateMgrEvent *)ProcessedEvents.get_bot();
      if (time < LastProcessedEvent->get_time_tag()) {
	cout << "Warning (StateManager): Posting an event in the past"
	     << endl;
      }else{
	FutureEvents.insert(event);
      }
    }
  }
  


/************************************************************************
 * Method: DefineNumberOfEvents(int)
 * Summary: Defines the number of events in the SpStateMgr.
 * Description: The SpStateMgr must know how many events you may
 *              want to process.  This allows it to set up some arrays
 *              for processing and must be called before you call
 *              ScheduleEvent or DefineEvent.  A handy way to define a
 *              number for this is to use enums:
 *
 *              enum{
 *                User_Event_0,
 *                User_Event_1,
 *                ...
 *                Last_User_Event,
 *                N_USER_EVENTS
 *              }
 *
 *              Then you can call DefineNumberOfEvents(N_USER_EVENTS)
 *              and always have the correct number, even if you add in
 *              events later!
 * Return: void
*************************************************************************/
  void DefineNumberOfEvents(int numberOfEvents //In: Number of events in the
			                       //    SpStateMgr
			    ) {
    FreeUserEvents.set_ntypes(numberOfEvents);
  }


/************************************************************************
 * Method: DefineEvent(char *, int, void*(int), int, int)
 * Summary: Defines the event for a given message.
 * Description: SPEEDES can send messages directly to a SpStateMgr.
 *              Two of the messages that SPEEDES already sends are
 *              "DISCOVER_OBJECT" and "REFLECT_ATTRIBUTES".  The first
 *              message is sent the very first time that the
 *              SpStateMgr "sees" or discovers an object proxy to
 *              which it has subscribed.  You are not told which
 *              object proxy this is, only that there is a new one.
 *              The second message is sent whenever an object proxy,
 *              to which the SpStateMgr is subscribing, changes.
 *              Again, you are not told which one changed, only that
 *              one of them is now different.
 *
 *              You can also have your own messages sent to the
 *              SpStateMgr through the use of the SpEmHostUser
 *              methods SendTwoWayCommand() or SubscribeNamedData().
 *              These can be created so that they send a message to
 *              the SpStateMgr with a special named string (you can
 *              use anything you want) and then the STATE_MANGER, upon
 *              recipt of this string, will launch the event you have
 *              defined.
 *
 *              For example, you could have some event
 *              NUMBER_OF_STAGES_EVENT with id of
 *              NUMBER_OF_STAGES_EVENT_ID, a constructor
 *              CreateNUMBER_OF_STAGES() that returns n of these
 *              events, and you subscribe to some data within SPEEDES
 *              that sends a message with the string name
 *              "NUMBER_OF_STAGES".  Within the STATE_MANGER, you
 *              could call
 *
 *              DefineEvent("NUMBER_OF_STAGES",
 *                          NUMBER_OF_STAGES_EVENT_ID,
 *                          CreateNUMBER_OF_STAGES,
 *                          sizeof(NUMBER_OF_STAGES_EVENT),
 *                          30)
 *
 *              The number '30' simply means that you want to maintain
 *              30 of these events on freelists so that the
 *              SpStateMgr is not continually calling new and
 *              delete on NUMBER_OF_STAGES events.  If you anticipate
 *              having a large number of these events at any given
 *              point in time, set this number fairly high.  If you
 *              expect only a few, set the number low.  It is a trade
 *              off in memory versus execution time so know how your
 *              problem will behave.
 *
 *              When your event gets executed, there will be several
 *              pieces of local data you can grab that are provided
 *              for you.  These variables are double Time, int Bytes,
 *              char * Data, and int GlobalId.  They represent the
 *              time the data was sent, the number of bytes contained
 *              in the data, and the global id of the sim object that
 *              sent the data.  These are automatically filled in for
 *              you when the event is called.  You should not delete
 *              the data since you may want that data if you roll the
 *              event back and then roll it forward again.
 *
 * Return: void
*************************************************************************/
  void DefineEvent(char *nameOfEvent, //In: Named message that will trigger
		                      //    this event
		   int idOfEvent,     //In: integer id of the event to trigger
		   void *(*ConstructFuncForEvent)(int),//In: function that
		                                       //    will return
		                                       //    numberOfEventsToCreate
		                                       //    of this event type
		   int sizeOfEvent,   //In: Size of the event to be created
		   int numberOfEventsToCreate //In: Number of events to create
		                              //    on the free list.
		   ) {
    
    FreeUserEvents.set_type(nameOfEvent,
			    idOfEvent,
			    ConstructFuncForEvent,
			    sizeOfEvent,
			    numberOfEventsToCreate);
    
  }
  


/************************************************************************
 * Method: SetUserState(void *)
 * Summary: Sets user defined state for the SpStateMgr
 * Description: Notice that we use a void * pointer here.  This is
 *              because this information is only used and created by
 *              the user.  The STATE_MANGER has no understanding of
 *              what is contained in this void * and it must be
 *              completely handled by the user.  It could be used for
 *              implementing your own state variables (you must take
 *              care of updating them with time) or for any other
 *              purpose you desire.
 * Return: void
*************************************************************************/
  void SetUserState(void *state //In: new value for the user state.
		    ) {UserState = state;}


/************************************************************************
 * Method: GetUserState()
 * Summary: Returns the current value of the user state
 * Description: Again, the SpStateMgr washes its hands of the user
 *              defined state.  You can put whatever you want in this
 *              and modify it however you choose but the SpStateMgr
 *              will only store it for you, it will not modify
 *              anything you put here.
 * Return: void * pointer to the user state
*************************************************************************/
  void *GetUserState(){ return UserState; }
  
protected:
  
  
private:
  
  double TimeLag;		      // Max time lag with SPEEDES
  double Time;		    	      // Time of GoToTime method invokation
  int TimeRequestFlag;		      // Time request made - not granted yet
  
  void *UserState;                    // User-defined state variables
  C_XHASH *ObjectProxyIdHash;         // Rollbackable hash table of proxies
  C_XHASH *ObjectProxyNameHash;       // Rollbackable named hash table
  C_XHASH *ObjectProxyTypeHash;       // Rollbackable typed hash table
  C_XQUEUE *CurrentQ;                 // Current Xqueue in hash table
  SpObjProxy *CurrentObjectProxy;   // Current Object Proxy
  F_SpProxyItem *CurrentProxyItem;     // Current proxy item holds a proxy
  
  SpPriorityQueueHeap FutureEvents;	      // Future unprocessed events
  C_XQUEUE ProcessedEvents;	      // List of processed events
  SpStateMgrEvent *CurrentEvent;  // Last event processed
  C_FREE_LIST FreeUserEvents;         // Free list of user defined events
  SpEmHostUser *EmHostUser;	      // Host User for external modules
  SpGlobalState *GlobalState;          // Global state shared by events

  void Init(); // called by the State Manager constructors
  SpObjProxy *GetNextProxyByType();
  
  
};

typedef SpStateMgr STATE_MANAGER;

#endif

