// SpStateMgrEvent.H header file

/*
** This is the base class for events in the object proxy state manager.
** Through inheritance, specific event types are defined and processed
** through virtual functions.
*/

#include "SpAltMgr.H"
#include "sq_item.H"
#include "SpEmHostUser.H"
#include "xhash.H"
#include "RB_SpFrameworkFuncs.H"
#include "SpGlobalState.H"
#include "SpPriorityQueueHeap.H"

#ifndef DEF_SM_EVENT
#define DEF_SM_EVENT

/*
** This is used by the ProcessStateFlag
*/

enum {
  EVENT_UNPROCESSED,
  EVENT_PROCESSED,
  EVENT_ROLLEDBACK
};

class SpStateMgrEvent : public C_SQ_ITEM {

  private:

  SpAltMgr AltMgr;
    static C_QUEUE **UserEvents;
    int ProcessedFlag;

  protected:

    char *Data;
    int DataBytes;
    int GlobalId;
    char *EventName;

    static void *UserState;
    static C_XHASH *ObjectProxyIdHash;
    static C_XHASH *ObjectProxyNameHash;
    static C_XHASH *ObjectProxyTypeHash;
    static SpEmHostUser *EmHostUser;
    static SpGlobalState *GlobalState;
    static SpPriorityQueueHeap *FutureEvents;
    static C_XQUEUE *ProcessedEvents;

  public:

/*
** Constructor
*/

    SpStateMgrEvent() {
      time_tag.SetValues(-1.0e20, 0, 0);
      ProcessedFlag = EVENT_UNPROCESSED;
    }

    SpStateMgrEvent(SpSimTime time) {
      time_tag = time;
      ProcessedFlag = EVENT_UNPROCESSED;
    }

/*
** The Init method feeds the data that has been passed from SPEEDES to the
** event in a common way.
*/

    void Init(SpSimTime time,
	      char *data,
	      int dataBytes,
	      int globalId,
	      char *eventName) {

      time_tag = time;

      Data = data;
      DataBytes = dataBytes;
      GlobalId = globalId;
      EventName = eventName;
    }

/*
** The Init method feeds the data that has been passed from SPEEDES to the
** event in a common way.
*/

    void SetPriority(int priority) {time_tag.SetUniqueId(priority);}
    int GetPriority() {return time_tag.GetUniqueId();}

/*
** Static member functions that provide information to state manager events
*/

    static void SetUserState(void *userState) {UserState = userState;}

    static void SetObjectProxyIdHash(C_XHASH *objectProxyIdHash) {
      ObjectProxyIdHash = objectProxyIdHash;
    }

    static void SetObjectProxyNameHash(C_XHASH *objectProxyNameHash) {
      ObjectProxyNameHash = objectProxyNameHash;
    }

    static void SetObjectProxyTypeHash(C_XHASH *objectProxyTypeHash) {
      ObjectProxyTypeHash = objectProxyTypeHash;
    }

    static void SetEmHostUser(SpEmHostUser *emHostUser) {
      EmHostUser = emHostUser;
    }

    SpAltMgr* GetAltMgr() {return &AltMgr;}

    static void SetGlobalState(SpGlobalState *globalState) {
      GlobalState = globalState;
    }

    static void SetFutureEvents(SpPriorityQueueHeap *futureEvents) {
      FutureEvents = futureEvents;
    }

    static void SetProcessedEvents(C_XQUEUE *processedEvents) {
      ProcessedEvents = processedEvents;
    }

/*
** Set the time tag of the event. Events are managed by the state manager so
** that they are processed in correct time order.
*/

    void SetTimeTag(SpSimTime){;}

/*
** Roll the event backward and forward. Events are only processed once, but
** then are reversible through the rollback queue mechanism. Rollbacks and
** rollforwards only affect state variables. They do not cancel previously\
** scheduled events, etc..
*/

    void ProcessEvent();
    void RollBackEvent();
    void RollForwardEvent();

    void InitRollBackQueue() {SpCurrentAltMgr = &AltMgr;}

/*
** Add a user defined event to a given event type. This allows users to add
** processing to parent events of a given type. After an event is processed,
** its list of UserEvents are automatically processed, which may process
** other user events after they are processed, etc.. In this manner, event
** processing is hierarchical with arbitrary numbers of levels.
*/

    static void AddUserEvent(int /*ParentEventType*/, int /*UserEventType*/){;}

/*
** Schedule an event for processing in the future. This method inserts a new
** event into the event queue for later processing. Applications should
** create a new event. It is an error to try to reschedule an already existing
** event.
*/

    void ScheduleEvent(SpSimTime time) {
      if (get_link()) {
	cout << "Warning (StateManagerEvent): Posting event twice?" << endl;
      }else{
        time_tag = time;
	SpStateMgrEvent *LastProcessedEvent =
	  (SpStateMgrEvent *)ProcessedEvents->get_bot();
	if (time < LastProcessedEvent->get_time_tag()) {
	  cout << "Warning (StateManagerEvent): Posting an event in the past"
	       << endl;
	}else{
	  FutureEvents->insert(this);
	}
      }
    }

/*
** The StateManager calls this method to process an event. Through
** inheritance, specific processing for different event types is achieved.
** The ProcessUserEvents() method is called by the object proxy state manager
** after Process() completes.
*/

    virtual void Process(){;}
    virtual void ProcessUserEvents(){;}

};

typedef SpStateMgrEvent STATE_MANAGER_EVENT;

class SpFedStateMgr;

#define DEFINE_STATE_MGR_EVENT(eventName, eventClassName)		\
inline SpStateMgrEvent *New##eventName() {return new eventClassName();}	\
extern SpFedStateMgr *GlobalFedStateMgr 				\

#define PLUG_IN_STATE_MGR_EVENT(eventName)				\
GlobalFedStateMgr->PlugInEvent(#eventName, New##eventName)		\

#endif

