// BinaryTree.C method file

#include "SpIostream.H"
#include "SpFstream.H"
#include "SpStrstream.H"
#include "SpIomanip.H"
#include "SpBinaryTree.H"

const double TREE_BALANCE_RATIO = 20.0;
const int TREE_CUTOFF = 40;
const int N_SPLAY = 100;

/*
** Insert an element into the binary tree
*/

#ifdef USE_FREE_LISTS
inline SpKey* SpBinaryTree::Clone(SpKey* key){
  SpKey* rc = NULL;
  SpStringKey* tempStringRc;
  SpDoubleKey* tempDoubleRc;
  SpIntervalKey* tempIntervalRc;
  switch (key->GetMode()){
  case DOUBLE_KEY_MODE:
    tempDoubleRc = ALLOCATE_SpDoubleKey();
    *tempDoubleRc = *key;
    rc = tempDoubleRc;
    break;
  case STRING_KEY_MODE:
    tempStringRc = ALLOCATE_SpStringKey();
    *tempStringRc = *key;
    rc = tempStringRc;
  case INTERVAL_KEY_MODE:
    tempIntervalRc = ALLOCATE_SpIntervalKey();
    *tempIntervalRc = *key;
    rc = tempIntervalRc;
  }
  return rc;
}
#else

inline SpKey* SpBinaryTree::Clone(SpKey* key){
  return key->Clone();
}

#endif

void SpBinaryTree::Insert(void *element, SpKey * key, int tieBreaker) {
  TREE_ITEM *node;
  TREE_ITEM *left;
  TREE_ITEM *right;
  TREE_ITEM *item;

  if (element == NULL) {
    cout << "Warning, tring to insert a NULL element into a SpBinaryTree"
	 << endl;
    return;
  }

  if ((KeyMode != key->GetMode()) || (KeyMode == NO_KEY_MODE)){
    KeyMode = key->GetMode();
  }

  item = NewTreeItem();
  item->Element = element;
  item->NodeKey = Clone(key);

  if (tieBreaker == -1) {
    item->TieBreaker = TieBreaker;
    LastTieBreaker = TieBreaker;
    TieBreaker++;
  }else{
    item->TieBreaker = tieBreaker;
    LastTieBreaker = tieBreaker;
  }

  CurrentNode = item;

  if (FirstNode) {
    if (*item < *FirstNode) FirstNode = item;
  }else{
    FirstNode = item;
  }

  if (LastNode) {
    if (*item >= *LastNode) LastNode = item;
  }else{
    LastNode = item;
  }

  NumElements++;

  if (Root == NULL) {
    Root = item;
    return;
  }

  node = Root;

//  int InsertCounter = 0;

  while (1) {

//    InsertCounter++;

    if (Mode == BALANCED_TREE_MODE) {
      node = RoughBalance(node);
    }

    if (*item < *node) {
      node->Nleft++;
      left = node->LeftChild;
      if (left) {
	node = left;
      }else{
	node->LeftChild = item;
	item->Parent = node;

	if (Mode == SPLAY_TREE_MODE) {
	  Splay(item);
	}

	break;
      }
    }else{
      node->Nright++;
      right = node->RightChild;
      if (right) {
	node = right;
      }else{
	node->RightChild = item;
	item->Parent = node;

	if (Mode == SPLAY_TREE_MODE) {
	  Splay(item);
	}

	break;
      }
    }

  }

//  cout << "InsertCounter = " << InsertCounter << endl;

}

/*
** Find an element
*/

void *SpBinaryTree::Find(const SpKey * key) {
  #ifdef __INSURE__
  // Without this definition of a ostream object, there will be a
  // crash when using INSURE++.  This must be a bug in INSURE++.  The
  // ifdef prevents this code from slowing down the program when you
  // are NOT compiling with INSURE++.  Note that I only define this
  // variable, but do not use it, since defining it was sufficient to
  // stop INSURE++ from crashing.
  ostrstream os;
  #endif

  TREE_ITEM *node;
  static TREE_ITEM keyItem; //Static so that the destructor is not called
  keyItem.TieBreaker = -1;
  TREE_ITEM *lChildFind;
  void* rc = NULL;

  if ((KeyMode != key->GetMode()) && (KeyMode != NO_KEY_MODE)){
    cout<<"Trying to do a find with the wrong key type"<<endl;
  }
  else{
    keyItem.NodeKey = (SpKey*)key;
    
    node = Find(Root, keyItem);
    if (node) {
      
      if ((lChildFind = Find(node->LeftChild, keyItem))){
	node = lChildFind;
      }
      if (Mode == SPLAY_TREE_MODE) {
	Splay(node);
      }
      CurrentNode = node;
      rc = node->Element;
    }
  }
  return rc;
}


TREE_ITEM *SpBinaryTree::Find(TREE_ITEM *subTree,
			      const TREE_ITEM& keyItem) {
  TREE_ITEM *node;
  TREE_ITEM *lChildFind;
  TREE_ITEM* rc = NULL;

  if (subTree != NULL){
    node = subTree;
    while (node) {
      
      if (Mode == BALANCED_TREE_MODE) {
	node = RoughBalance(node);
      }
      
      if (*node == keyItem) {
	if ((lChildFind = Find(node->LeftChild, keyItem))){
	  node = lChildFind;
	}
	rc = node;
	break;
      }
      
      if (keyItem < *node) {
	node = node->LeftChild;
	if (node == NULL){
	  break;
	}
      }else{
	node = node->RightChild;
	if (node == NULL){
	  break;
	}
      }
    }
  }
  return rc;
}

/*
** Remove an element
*/

void SpBinaryTree::RemoveNode(TREE_ITEM *node) {
  TREE_ITEM *left;
  TREE_ITEM *right;
  TREE_ITEM *parent;

  UpdateFirstLastNode(node);

  if (CurrentNode == node) GetNextElement();
  
  if (node == Root) {

//	cout << "Removing root" << endl;

    if (node->LeftChild) {
      Root = node->LeftChild;
      Root->Parent = NULL;
      InsertSubTree(Root, node->RightChild);
    }else{
      if (node->RightChild) {
	Root = node->RightChild;
	Root->Parent = NULL;
      }else{
	Root = NULL;
      }
    }

  }else{

    parent = node->Parent;
    left = node->LeftChild;
    right = node->RightChild;
    
    if (parent == NULL) {
      cout << "Bad parent..." << endl;
    }

    if (parent->LeftChild == node) {
      parent->LeftChild = left;
      if (left) {
	parent->Nleft = 1 + left->Nleft + left->Nright;
	left->Parent = parent;
      }else{
	parent->Nleft = 0;
      }
      InsertSubTree(parent, right);
    }else{
      if (parent->RightChild != node) cout << "Bad children" << endl;
      parent->RightChild = right;
      if (right) {
	parent->Nright = 1 + right->Nleft + right->Nright;
	right->Parent = parent;
      }else{
	parent->Nright = 0;
      }	  
      InsertSubTree(parent, left);
    }
    
  }
  
  NumElements--;
  
  if (NumElements == 0) {
    FirstNode = NULL;
    LastNode = NULL;
  }
  
  DeleteTreeItem(node);
  
}

/*
** Check the integrity of the tree
*/

int SpBinaryTree::CheckIntegrity() {
  void *element;
  int i;
  TREE_ITEM *keyItem;
  int retVal = 1;
  int nl, nr;
  TREE_ITEM *left;
  TREE_ITEM *right;

  GetFirstElement();
  if (CurrentNode != FirstNode) {
    retVal = 0;
    cout << "Error, bad FirstNode" << endl;
  }

  GetLastElement();
  if (CurrentNode != LastNode) {
    retVal = 0;
    cout << "Error, bad LastNode" << endl;
  }

  element = GetFirstElement();
  keyItem = GetCurrentNode();
  i = 0;
  while (element) {

    nl = 0;
    left = CurrentNode->LeftChild;
    if (left) nl = left->Nleft + left->Nright + 1;
    if (CurrentNode->Nleft != nl) {
      cout << "Bad right/left" << endl;
      retVal = 0;
    }

    nr = 0;
    right = CurrentNode->RightChild;
    if (right) nr = right->Nleft + right->Nright + 1;
    if (CurrentNode->Nright != nr) {
      cout << "Bad right/left " << endl;
      retVal = 0;
    }

    if (*keyItem > *GetCurrentNode()) {
      cout << "Tree not sorted" << endl;
      retVal = 0;
    }
    keyItem = GetCurrentNode();
    element = GetNextElement();
    i++;
  }

  if (i != NumElements) {
    cout << "Bad integrity: wrong number of elements ("
	 << NumElements << ", " << i << ")" << endl;
    retVal = 0;
    return retVal;
  }else{
    return retVal;
  }

}

/*
** Get the first element
*/

void SpBinaryTree::InsertSubTree(TREE_ITEM *startingNode, TREE_ITEM *subTree) {
  TREE_ITEM *node;
  TREE_ITEM *left;
  TREE_ITEM *right;

  if (startingNode == NULL) {
    cout << "Error, starting node is NULL" << endl;
    return;
  }
  if (subTree == NULL) {
/*
    cout << "subTree = NULL: "
	 << "Nleft = " << startingNode->Nleft
	 << ", Nright = " << startingNode->Nright << endl;
*/
    CorrectNchildrenUp(startingNode);
    return;
  }

  node = startingNode;

  while (1) {
    if (*subTree < *node) {
      left = node->LeftChild;
      if (left) {
	node = left;
      }else{
	node->LeftChild = subTree;
	subTree->Parent = node;
	node->Nleft = 1 + subTree->Nleft + subTree->Nright;
	CorrectNchildrenUp(node);
/*
	if (Mode == SPLAY_TREE_MODE) {
	  Splay(node);
	}
*/
	break;
      }
    }else{
      right = node->RightChild;
      if (right) {
	node = right;
      }else{
	node->RightChild = subTree;
	subTree->Parent = node;
	node->Nright = 1 + subTree->Nleft + subTree->Nright;
	CorrectNchildrenUp(node);
/*
	if (Mode == SPLAY_TREE_MODE) {
	  Splay(node);
	}
*/
	break;

      }
    }
  }

}

/*
** Correct the Nleft and Nright values when adding a subTree to a node
*/
void SpBinaryTree::CorrectNchildrenUp(TREE_ITEM *it) {
  TREE_ITEM *node = it;
  TREE_ITEM *parent;

  while (node != Root) {

    parent = node->Parent;

    if (parent->LeftChild == node) {
      parent->Nleft = 1 + node->Nleft + node->Nright;
    }else{
      parent->Nright = 1 + node->Nleft + node->Nright;
    }

    node = parent;
  }

}

/*
** Get the first element
*/

void *SpBinaryTree::GetFirstElement() {

  CurrentNode = Root;
  if (CurrentNode == NULL) return NULL;

  while (CurrentNode->LeftChild) CurrentNode = CurrentNode->LeftChild;
  return CurrentNode->Element;

}

/*
** Get the last element
*/

void *SpBinaryTree::GetLastElement() {

  CurrentNode = Root;
  if (CurrentNode == NULL) return NULL;

  while (CurrentNode->RightChild) CurrentNode = CurrentNode->RightChild;
  return CurrentNode->Element;

}

/*
** Get the next element
*/

void *SpBinaryTree::GetNextElement() {
  TREE_ITEM *parent;

  if (CurrentNode == NULL) return NULL;

  if (CurrentNode == Root) {
    CurrentNode = CurrentNode->RightChild;
    if (CurrentNode == NULL) return NULL;
    while (CurrentNode->LeftChild) CurrentNode = CurrentNode->LeftChild;
    return CurrentNode->Element;
  }

  if (CurrentNode->RightChild) {
    CurrentNode = CurrentNode->RightChild;
    while (CurrentNode->LeftChild) CurrentNode = CurrentNode->LeftChild;
    return CurrentNode->Element;
  }

  while (1) {

    if (CurrentNode == Root) {
      CurrentNode = NULL;
      return NULL;
    }

    parent = CurrentNode->Parent;
    if (parent->LeftChild == CurrentNode) {
      CurrentNode = parent;
      return CurrentNode->Element;
    }

    CurrentNode = parent;

  }

}

/*
** Get the previous element
*/

void *SpBinaryTree::GetPreviousElement() {
  TREE_ITEM *parent;

  if (CurrentNode == NULL) return NULL;

  if (CurrentNode == Root) {
    CurrentNode = CurrentNode->LeftChild;
    if (CurrentNode == NULL) return NULL;
    while (CurrentNode->RightChild) CurrentNode = CurrentNode->RightChild;
    return CurrentNode->Element;
  }

  if (CurrentNode->LeftChild) {
    CurrentNode = CurrentNode->LeftChild;
    while (CurrentNode->RightChild) CurrentNode = CurrentNode->RightChild;
    return CurrentNode->Element;
  }

  while (1) {

    if (CurrentNode == Root) {
      CurrentNode = NULL;
      return NULL;
    }

    parent = CurrentNode->Parent;
    if (parent->RightChild == CurrentNode) {
      CurrentNode = parent;
      return CurrentNode->Element;
    }

    CurrentNode = parent;

  }

}

/*
** Rotate this node left
*/

TREE_ITEM *SpBinaryTree::RotateLeft(TREE_ITEM *item) {
  TREE_ITEM *X = item;
  TREE_ITEM *Y;
  TREE_ITEM *B;
  TREE_ITEM *parent;

//  cout << "Rotating Left" << endl;

  if (X == NULL) {
    cout << "Error, X = NULL in RotateLeft" << endl;
    return NULL;
  }

  Y = X->RightChild;

  if (Y) {

    parent = X->Parent;

    B = Y->LeftChild;

    Y->LeftChild = X;
    X->RightChild = B;

    if (B) {
      B->Parent = X;
      X->Nright = 1 + B->Nleft + B->Nright;
    }else{
      X->Nright = 0;
    }

    Y->Nleft = 1 + X->Nleft + X->Nright;

    Y->Parent = parent;

    if (item != Root) {
      if (parent->RightChild == X) {
	parent->RightChild = Y;
      }else{
	parent->LeftChild = Y;
      }
    }else{
      Root = Y;
    }

    X->Parent = Y;
    Y->LeftChild = X;

  }else{

    return X;

  }

  return Y;

}

/*
** Rotate this node right
*/

TREE_ITEM *SpBinaryTree::RotateRight(TREE_ITEM *item) {
  TREE_ITEM *Y = item;
  TREE_ITEM *X;
  TREE_ITEM *B;
  TREE_ITEM *parent;

//  cout << "Rotating Right" << endl;

  if (Y == NULL) {
    cout << "Error, Y = NULL in RotateRight" << endl;
    return NULL;
  }

  X = Y->LeftChild;

  if (X) {

    parent = Y->Parent;

    B = X->RightChild;

    X->RightChild = Y;
    Y->LeftChild = B;

    if (B) {
      B->Parent = Y;
      Y->Nleft = 1 + B->Nleft + B->Nright;
    }else{
      Y->Nleft = 0;
    }

    X->Nright = 1 + Y->Nleft + Y->Nright;

    X->Parent = parent;

    if (item != Root) {
      if (parent->LeftChild == Y) {
	parent->LeftChild = X;
      }else{
	parent->RightChild = X;
      }
    }else{
      Root = X;
    }

    Y->Parent = X;
    X->RightChild = Y;

  }else{

    return Y;

  }

  return X;

}

/*
** Perform the splay operation
*/

void SpBinaryTree::Splay(TREE_ITEM *node) {
  TREE_ITEM *parent;
  TREE_ITEM *grandParent;
  int nSplay = 0;
  int parentFlag;
  int grandParentFlag;

  while(node != Root) {

    parent = node->Parent;

    if (parent == Root) {

      if (parent->LeftChild == node) {
        RotateRight(parent);
      }else{
        RotateLeft(parent);
      }

      break;

    }else{

      if (parent->LeftChild == node) {
        parentFlag = 0;
      }else{
        parentFlag = 1;
      }

      grandParent = parent->Parent;
      if (grandParent->LeftChild == parent) {
        grandParentFlag = 0;
      }else{
        grandParentFlag = 1;
      }

      if ((parentFlag + grandParentFlag) == 1) {

//...... Zig Zag

        if (parentFlag == 0) {
          RotateRight(parent);
          RotateLeft(grandParent);
        }else{
          RotateLeft(parent);
          RotateRight(grandParent);
        }

      }else{

//...... Zig Zig

        if (parentFlag == 0) {
          RotateRight(grandParent);
          RotateRight(parent);
        }else{
          RotateLeft(grandParent);
          RotateLeft(parent);
        }

      }

    }

    if (nSplay >= N_SPLAY) break;
    nSplay++;

  }

}



#ifdef USE_FREE_LISTS
PLUGIN_MEMPOOL( TREE_ITEM );
PLUGIN_MEMPOOL( SpStringKey );
PLUGIN_MEMPOOL( SpDoubleKey );
PLUGIN_MEMPOOL( SpIntervalKey );
#endif
