// splay_tree.H header file

#ifndef Binary_Tree_Object
#define Binary_Tree_Object

#include "SpString.H"
#include "SpKey.H"
#include "SpIostream.H"
#include "SpAlt.H"
#include "SpTreeMode.H"

extern const double TREE_BALANCE_RATIO;
extern const int TREE_CUTOFF;
extern const int N_SPLAY;

class TREE_ITEM {

  friend class SpBinaryTree;
  friend class SpHashTree;
  friend class RB_SpBinaryTree;
  friend class RB_SpPriorityTree;
  

  private:

    void *Element;

    TREE_ITEM *Parent;
    TREE_ITEM *LeftChild;
    TREE_ITEM *RightChild;
    int Nleft;
    int Nright;
    SpKey * NodeKey;
    int TieBreaker;

    TREE_ITEM() {Init();}

    void Init() {
      Element = NULL;
      Parent = NULL;
      LeftChild = NULL;
      RightChild = NULL;
      Nleft = 0;
      Nright = 0;
      NodeKey = NULL;
      TieBreaker = 0;
    }

    void Cleanup(){ delete NodeKey;}

    int operator == (const TREE_ITEM &item) const {
      return (*NodeKey == *item.NodeKey);
    }

    int operator != (const TREE_ITEM &item) const {
      return !(*NodeKey == *item.NodeKey);
    }

  int operator < (const TREE_ITEM &item) const {
    int rc = (*NodeKey < *item.NodeKey);
    if (rc == 0){
      if (*NodeKey == *item.NodeKey){
	rc = (TieBreaker < item.TieBreaker);
      }
    }
    return rc;
  }

  int operator <= (const TREE_ITEM &item) const {
    int rc = (*NodeKey <= *item.NodeKey);
    if (rc == 1){
      if (*NodeKey == *item.NodeKey){
	rc = (TieBreaker < item.TieBreaker);
      }
    }
    return rc;
  }

  int operator > (const TREE_ITEM &item) const {
    int rc = (*NodeKey > *item.NodeKey);
    if (rc == 0){
      if (*NodeKey == *item.NodeKey){
	rc = (TieBreaker > item.TieBreaker);
      }
    }
    return rc;
  }

  int operator >= (const TREE_ITEM &item) const {
    int rc = (*NodeKey >= *item.NodeKey);
    if (rc == 1){
      if (*NodeKey == *item.NodeKey){
	rc = (TieBreaker > item.TieBreaker);
      }
    }
    return rc;
  }
public:
  double DoubleVal(){
    return (double)*(SpDoubleKey*)NodeKey;
  }
};

//DEFINE_MEMPOOL( TREE_ITEM, ALT_CHUNK_SIZE );

class SpBinaryTree {

  friend class SpPriorityTree;
  friend class SpAltTreeChange;
  friend class SpHashTree;
 

  private:

  protected:

    KEY_MODE KeyMode;

    TREE_ITEM *Root;
    TREE_ITEM *CurrentNode;
    TREE_ITEM *FirstNode;
    TREE_ITEM *LastNode;
    int NumElements;
    int Mode;
    int TieBreaker;
    int LastTieBreaker;

    TREE_ITEM *RotateLeft(TREE_ITEM *it);
    TREE_ITEM *RotateRight(TREE_ITEM *it);

    void InsertSubTree(TREE_ITEM *startingNode, TREE_ITEM *subTree);

    void *Find(const SpKey* key);

    TREE_ITEM *Find (TREE_ITEM* subTree, const TREE_ITEM& key);

    void RemoveNode(TREE_ITEM *node);

    TREE_ITEM *GetCurrentNode() {return CurrentNode;}
    void SetCurrentNode(TREE_ITEM *currentNode) {CurrentNode = currentNode;}

    TREE_ITEM *GetFirstNode() {return FirstNode;}
    TREE_ITEM *GetLastNode() {return LastNode;}

    TREE_ITEM *NewTreeItem() {
      //TREE_ITEM *item = ALLOCATE_TREE_ITEM();
      TREE_ITEM *item = new TREE_ITEM();
      item->Init();
      return item;
    }

    void DeleteTreeItem(TREE_ITEM *item) {
      item->Cleanup();
      delete item;
      //DEALLOCATE_TREE_ITEM(item);
    }

    void CorrectNchildrenUp(TREE_ITEM *it);

    void UpdateFirstLastNode(TREE_ITEM *node) {
      if (NumElements == 0) {
	FirstNode = NULL;
	LastNode = NULL;
      }
      if (node == FirstNode) {
	CurrentNode = FirstNode;
	GetNextElement();
	FirstNode = CurrentNode;
      }
      if (node == LastNode) {
	CurrentNode = LastNode;
	GetPreviousElement();
	LastNode = CurrentNode;
      }
    }

    TREE_ITEM *RoughBalance(TREE_ITEM *it) {
      TREE_ITEM *node = it;
      int Nleft = node->Nleft;
      int Nright = node->Nright;
      if (Nright+Nleft > TREE_CUTOFF) {
	if (Nright < Nleft) {
      	  if (Nleft/(Nright+1) > TREE_BALANCE_RATIO) node = RotateRight(node);
	}else{
     	  if (Nright/(Nleft+1) > TREE_BALANCE_RATIO) node = RotateLeft(node);
	}
      }
      return node;
    }

    void Splay(TREE_ITEM *node);

    void Insert(void *element, SpKey* key, int tieBreaker=-1);

    void *Remove(const SpKey* key) {
      void * rc = NULL;
      if (KeyMode == key->GetMode()){
	void *element = Find(key);
	if (element) {
	  RemoveNode(CurrentNode);
	}
	rc = element;
      }
      else{
	if (KeyMode != NO_KEY_MODE){
	  cout<<"Tried to remove when in the wrong mode"<<endl;
	}
      }
      if (NumElements == 0){
	KeyMode = NO_KEY_MODE;
      }
      return rc;
    }

    int GetLastTieBreaker() {return LastTieBreaker;}

    void Insert(void *element, double key, int tieBreaker) {

      if (KeyMode == NO_KEY_MODE){
	KeyMode = DOUBLE_KEY_MODE;
      }
      if (KeyMode == DOUBLE_KEY_MODE){
	SpDoubleKey newKey(key);
	Insert(element, &newKey, tieBreaker);
      }
      else{
	cout<<"Tried to insert a double key when not in double key mode. "
	    <<"Ignoring your request"<<endl;
      }
    }

    void Insert(void *element, char* keyName, int tieBreaker) {
      if (KeyMode == NO_KEY_MODE){
	KeyMode = STRING_KEY_MODE;
      }
      if (KeyMode == STRING_KEY_MODE){
	SpStringKey newKey(keyName);
	Insert(element, &newKey, tieBreaker);
      }
      else{
	cout<<"Tried to insert a string key when not in string key mode. "
	    <<"Ignoring your request"<<endl;
      }
    }


    void Insert(void* element, double key, char* keyName, int tieBreaker = -1){
      if (keyName){
	Insert(element, keyName, tieBreaker);
      }
      else{
	Insert(element, key, tieBreaker);
      }
    }

    SpKey* Clone(SpKey* key);

  public:

    SpBinaryTree():
      KeyMode(NO_KEY_MODE),
      Root(NULL),
      CurrentNode(NULL),
      FirstNode(NULL),
      LastNode(NULL),
      NumElements(0),
      Mode(NORMAL_TREE_MODE),
      TieBreaker(0)
    {}

    int GetNumElements() {return NumElements;}

    // Removes all keys and values but does not delete them.
    void Reset() {
      while (RemoveFirstElement())
	continue;
    }
    
    double GetCurrentKey() {
      double retVal = -1;
      if (CurrentNode){
	if (KeyMode == DOUBLE_KEY_MODE){
	  retVal = *((SpDoubleKey*)CurrentNode->NodeKey);
	}
	else{
	  cout<<"Called GetCurrentKey when not in double key mode"<<endl;
	}
      }
      return retVal;
    }

    char *GetCurrentKeyName() {
      char *retVal = NULL;
      if (CurrentNode){
	if (KeyMode == STRING_KEY_MODE){
	  retVal = (char *)(*((SpStringKey*)(CurrentNode->NodeKey)));
	}
	else{
	  cout<<"Not in string key mode"<<endl;
	}
      }
      return retVal;
    }

    void *GetCurrentElement() {
      void *e=NULL;
      if (CurrentNode) e = CurrentNode->Element;
      return e;
    }

    int CheckIntegrity();

    void SetNormalTreeMode() {Mode = NORMAL_TREE_MODE;}
    void SetBalancedTreeMode() {Mode = BALANCED_TREE_MODE;}
    void SetSplayTreeMode() {Mode = SPLAY_TREE_MODE;}

    int IsNormalTreeMode() {return (Mode == NORMAL_TREE_MODE);}
    int IsBalancedTreeMode() {return (Mode == BALANCED_TREE_MODE);}
    int IsSplayTreeMode() {return (Mode == SPLAY_TREE_MODE);}

    void Insert(void *element, SpKey& key){
      Insert(element, &key);
    }

    void Insert(void *element, double key) {
      Insert(element, key, -1);
    }

    void Insert(void *element, char* keyName) {
      Insert(element, keyName, -1);
    }

    void * Remove(const SpKey& key){ return Remove(&key);}

    void *Find(const SpKey& key){ return Find(&key);}

    void *Find(const double key) 
    {
      void * rc = NULL;
      if (KeyMode == DOUBLE_KEY_MODE){
	SpDoubleKey newKey(key);
	rc = Find(&newKey);
      }
      else if (KeyMode != NO_KEY_MODE) {
	cout<<"Tried to find a double key when not in double key mode."<<endl;
      }
      return rc;
    }
    void *Find(const char * keyName) 
    {
      void * rc = NULL;
      if (KeyMode == STRING_KEY_MODE){
	SpStringKey newKey(keyName);
	rc = Find(&newKey);
      }
      else if (KeyMode != NO_KEY_MODE) {
	cout<<"Tried to find a string key when not in string key mode."<<endl;
      }
      return rc;
    }

    // Find the next element whose key matches the key of the last
    // returned element.  Return NULL if there are no such elements
    // left.
    void *FindNext() {
      SpKey *lastKey = CurrentNode->NodeKey->Clone();
      void *rc = GetNextElement();
      if (rc != NULL && !(*lastKey == *CurrentNode->NodeKey))
	rc = NULL;
      delete lastKey;
      return rc;
    } 

    void *Remove(double key)
    {
      void * rc = NULL;
      if (KeyMode == DOUBLE_KEY_MODE){
	SpDoubleKey newKey(key);
	rc = Remove(&newKey);
      }
      else if (KeyMode != NO_KEY_MODE) {
	cout<<"Tried to remove a double key when not in double key mode."
	    <<endl;
      }
      return rc;
    }

    void *Remove(char *keyName) {
      void * rc = NULL;
      if (KeyMode == STRING_KEY_MODE){
	SpStringKey newKey(keyName);
	rc = Remove(&newKey);
      }
      else if (KeyMode != NO_KEY_MODE) {
	cout<<"Tried to remove a string key when not in string key mode."
	    <<endl;
      }
      return rc;
    }

    void* Remove(double val, char* keyName){
      void* rc;
      if (keyName){
	rc = Remove(keyName);
      }
      else{
	rc = Remove(val);
      }
      return rc;
    }

    void *RemoveFirstElement() {
      void *element = NULL;
      if (FirstNode) {
	element = FirstNode->Element;
	RemoveNode(FirstNode);
      }
      return element;
    }

    void *RemoveFirstElement(double &key) {
      void *element = NULL;
      key = 1.0e20;
      if (FirstNode) {
	element = FirstNode->Element;
	key = *((SpDoubleKey*)FirstNode->NodeKey);
	RemoveNode(FirstNode);
      }
      return element;
    }

    void *GetFirstElement();
    void *GetLastElement();

    // This function returns elements in increasing order.
    void *GetNextElement();

    // This function returns elements in decreasing order.
    void *GetPreviousElement();

    void *operator ++ () {return GetNextElement();}
    void *operator ++ (int) {
      void *e = GetCurrentElement();
      GetNextElement();
      return e;
    }

    void *operator -- () {return GetPreviousElement();}
    void *operator -- (int) {
      void *e = GetCurrentElement();
      GetPreviousElement();
      return e;
    }

    void operator += (void *ptr) {
      Insert(ptr, int(ptr));
    }

    void *operator -= (void *ptr) {
      return Remove(int(ptr));
    }

};

typedef SpBinaryTree BINARY_TREE;
#endif
