// SpHashTree.H header file defines a hash table that is constructed as an
// array of trees. The array index is used for hashing to keep the trees
// fairly small. This should guarantee log(n) behavior. Note that the
// balancing heuristic is used in the Binary Tree.

#ifndef SpHashTree_OBJ
#define SpHashTree_OBJ

#include "SpBinaryTree.H"
#ifndef __SPEEDES_NT__
#include <values.h>
#else
#define INTBITS 32
#define MININT      (1 << (INTBITS - 1))
#define MAXINT      (~MININT)
#endif
#include "SpMath.H"
#include "SpAssert.H"

class SpHashTree {

  private:

    int ArraySize;
    SpBinaryTree *BinaryTree;

    int NumElements;
    int CurrentArrayIndex;

  protected:

  public:

    ~SpHashTree();

    SpHashTree(int arraySize = 10);

  protected:
    void SetArraySize(int arraySize) {
      ArraySize = arraySize;
      BinaryTree = new SpBinaryTree[arraySize];
      for (int i=0; i<arraySize; i++) BinaryTree[i].SetBalancedTreeMode();
    }

  public:
    void SetNormalTreeMode() {
      for (int i=0; i<ArraySize; i++) {
	BinaryTree[i].SetNormalTreeMode();
      }
    }

    void SetBalancedTreeMode() {
      for (int i=0; i<ArraySize; i++) {
	BinaryTree[i].SetBalancedTreeMode();
      }
    }

    void SetSplayTreeMode() {
      for (int i=0; i<ArraySize; i++) {
	BinaryTree[i].SetSplayTreeMode();
      }
    }

    int IsNormalTreeMode() {return BinaryTree[0].IsNormalTreeMode();}
    int IsBalancedTreeMode() {return BinaryTree[0].IsBalancedTreeMode();}
    int IsSplayTreeMode() {return BinaryTree[0].IsSplayTreeMode();}

    int GetNumElements() {return NumElements;}
    int GetArraySize() {return ArraySize;}

  protected:
    SpBinaryTree *GetBinaryTree() {return BinaryTree;}
    SpBinaryTree *GetBinaryTree(int arrayIndex) {
      return &BinaryTree[arrayIndex];
    }

    // This class used to NOT be able to handle negative keys, but
    // this function contains logic which does handle negative keys.
    int GetHashId(double key) {
      if (key < 0)
	key = -key;
      int hash_id;
      if (key > double(MAXINT)) {
	hash_id = int(fmod(key, ArraySize));
      } else {
	hash_id = int(key) % ArraySize;
      }
      return hash_id;
    }

    int GetHashId(char *keyName) {
      int hashId = 0;
      int len = strlen(keyName);
      for (int i=0; i<len; i++) {
	hashId ^= keyName[i] << ((i*8)%32);
      }
      return (hashId & 0x7fffffff);
    }

  public:


    // Removes all keys and values but does not delete them.
    void Reset() {
      for (int i=0; i<ArraySize; i++)
	BinaryTree[i].Reset();
      NumElements = 0;
      CurrentArrayIndex = 0;
    }
    
    int CheckIntegrity() {
      int retVal = 0;
      int numElements = 0;
      for (int i=0; i<ArraySize; i++) {
	numElements += BinaryTree[i].GetNumElements();
	retVal |= BinaryTree[i].CheckIntegrity();
      }
      if (numElements != NumElements) {
	cout << "SpHashTree error: Bad integrity NumElements "
	     << numElements << " not equal to " << NumElements
	     << endl;
      }
      return retVal;
    }

    void Insert(void *element, double key) {
      BinaryTree[GetHashId(key)].Insert(element, key);
      NumElements++;
    }

    void Insert(void *element, char *keyName) {
      int hashId = GetHashId(keyName);
      BinaryTree[hashId%ArraySize].Insert(element, keyName);
      NumElements++;
    }

    void *Find(double key) {
      CurrentArrayIndex = GetHashId(key);
      return BinaryTree[CurrentArrayIndex].Find(key);
    }

    void *Find(char *keyName) {
      CurrentArrayIndex = GetHashId(keyName)%ArraySize;
      return BinaryTree[CurrentArrayIndex].Find(keyName);
    }

    // Find the next element whose key matches the key of the last
    // returned element.  Return NULL if there are no such elements
    // left.
    void *FindNext() {
      TREE_ITEM *cn = BinaryTree[CurrentArrayIndex].CurrentNode;
      assert(cn != NULL);
      SpKey *lastKey = cn->NodeKey->Clone();
      void *rc = GetNextElement();
      if (rc != NULL &&
	  !(*lastKey == *BinaryTree[CurrentArrayIndex].CurrentNode->NodeKey))
	rc = NULL;
      delete lastKey;
      return rc;
    } 

    void *Remove(double key) {
      void *retVal = BinaryTree[GetHashId(key)].Remove(key);
      if (retVal) NumElements--;
      return retVal;
    }

    void *Remove(char *keyName) {
      int hashId = GetHashId(keyName);
      void *retVal = BinaryTree[hashId%ArraySize].Remove(keyName);
      if (retVal) NumElements--;
      return retVal;
    }

    // This class has a certain order, but that order is usually NOT
    // by order of increasing or decreasing key.  It is usually some
    // other order which is undefined by the class'es interface.  If
    // you repeatedly call GetNextElement then it returns elements in
    // this undefined order.  But if you repeatedly call
    // GetPreviousElement, then it returns elements in the exact reverse
    // order of the order used by GetNextElement.

    void *GetFirstElement() {
      void *element;
      for (int i=0; i<ArraySize; i++) {
	element = BinaryTree[i].GetFirstElement();
	if (element) {
	  CurrentArrayIndex = i;
	  break;
	}
      }
      return element;
    }

    void *GetLastElement() {
      void *element;
      for (int i=ArraySize-1; i>=0; i--) {
	element = BinaryTree[i].GetLastElement();
	if (element) {
	  CurrentArrayIndex = i;
	  break;
	}
      }
      return element;
    }

    void *GetNextElement() {
      void *element = NULL;
      for (int i=CurrentArrayIndex; i<ArraySize; i++) {
	element = BinaryTree[i].GetNextElement();
	if (element) {
	  CurrentArrayIndex = i;
	  break;
	}
	if (i < ArraySize-1) element = BinaryTree[i+1].GetFirstElement();
	if (element) {
	  CurrentArrayIndex = i+1;
	  break;
	}
      }
      return element;
    }

    void *GetPreviousElement() {
      void *element = NULL;
      for (int i=CurrentArrayIndex; i>=0; i--) {
	element = BinaryTree[i].GetPreviousElement();
	if (element) {
	  CurrentArrayIndex = i;
	  break;
	}
	if (i > 0) element = BinaryTree[i-1].GetLastElement();
	if (element) {
	  CurrentArrayIndex = i-1;
	  break;
	}
      }
      return element;
    }

    void *GetCurrentElement() {
      return BinaryTree[CurrentArrayIndex].GetCurrentElement();
    }

    double GetCurrentKey() {
      return BinaryTree[CurrentArrayIndex].GetCurrentKey();
    } 

    char* GetCurrentKeyName() {
      return BinaryTree[CurrentArrayIndex].GetCurrentKeyName();
    } 

    void *operator ++ () {return GetNextElement();}
    void *operator ++ (int) {
      void *element = GetCurrentElement();
      GetNextElement();
      return element;
    }

    void *operator -- () {return GetPreviousElement();}
    void *operator -- (int) {
      void *element = GetCurrentElement();
      GetPreviousElement();
      return element;
    }

    void operator += (void *ptr) {
      Insert(ptr, int(ptr));
    }

    void *operator -= (void *ptr) {
      return Remove(int(ptr));
    }

};

typedef SpHashTree HASH_TREE;

#endif
