#ifndef __SpHiGridRegionVec__
#define __SpHiGridRegionVec__

#include "SpAssert.H"
#include "SpStdlib.H"

class MAIN_PUBSUB_ITEM;

// An object of this class describes a number of specific grids and
// the region in each grid which is covered by a lookup item, and
// whether or not the entire grid is covered.
class SpHiGridRegionVec
{
  private:
    SpHiGridRegionVec() : NumDimRegions(-1) { abort(); }
    SpHiGridRegionVec(SpHiGridRegionVec&) : NumDimRegions(-1) { abort(); }
    void operator=(SpHiGridRegionVec&) { abort(); }
    
    int MaxNum; // allocated size of IdCoverage
    int NextIndexToSet;

    class GridIdCoverage
    {
      private:
	//   GridCovered_p is TRUE iff a call to HiGrid_Lookup passed in a
	// lookup item that completely covered the grid specified by the
	// variable Id.  Note that in intermediate stages the Id might
	// refer to an id in a dimension or it might refer to an id in a
	// SpSpaceSubSet, etc.  This id can change during the various steps
	// through the HiGrid_Lookup functions.  For example, it will have
	// other ids added to it, but with some ids first being multiplied
	// by a number large enough that the previous id info is still
	// retrievable.  The GridCovered_p will also change.  For example,
	// it might first be set to TRUE, because one dimension completely
	// covers the grid, but when it is later discovered that another
	// dimension does NOT completely cover the grid, then it will be
	// changed to FALSE.
	//   Since this class is used so much it might be worth it to
	// convert the following two integers into only one integer, and
	// then use the convention that if the integer is negative, it is
	// covered, and if it is positive it is not fully covered, and
	// then the Id accesser function would return the absolute value
	// of the member variable, etc.  But what about id of 0?  I think
	// it may be too kludgey to rely on the feature of the
	// architecture which stores the sign of 0.  So we might do some
	// less kludgey thing, like use an unsigned integer, and then have
	// all high numbers mean covered, and all low numbers mean not
	// fully covered, and we have to subtract a huge constant from the
	// high numbers before returning the id, etc.
	int Id, GridCovered_p;

      public:
	GridIdCoverage() {}

	friend class SpHiGridRegionVecIterator;
	friend class SpHiGridRegionVec;
    } *IdCoverage;

    class DimRegion
    {
      public:
	DimRegion() {}
	
	void operator=(const DimRegion &right)
	    { Lo = right.Lo; Hi = right.Hi; Res = right.Res; }

	double get_Lo () const { return Lo;  }
	double get_Hi () const { return Hi;  }
	double get_Res() const { return Res; }

	DimRegion &set_Lo (double lo)  { Lo  = lo;  return *this; }
	DimRegion &set_Hi (double hi)  { Hi  = hi;  return *this; }
	DimRegion &set_Res(double res) { Res = res; return *this; }

      private:
	DimRegion(DimRegion&) { abort(); }
	
	double Lo, Hi, Res;
    } *Region;

    // The number of DimRegions per GridIdCoverage.  This number may be more
    // than the number of dimensions listed in the .par file since
    // some descendants of class SpDim, like SpLatLngDim will
    // have two DimRegions, one for latitude and one for longitude
    // that together correspond to the single SpLatLngDim object.
    const int NumDimRegions;

    friend class SpHiGridRegionVecIterator;
    friend class SpHiGridRegionVecReader;
    friend class SpHiGridRegionVecModifier;

    DimRegion &get_DimRegion_internal(int IdCoverage_index, int dim_index)
      const {
      assert(IdCoverage_index < MaxNum);
      assert(dim_index < NumDimRegions);
      return Region[IdCoverage_index * NumDimRegions + dim_index]; }

    DimRegion &get_DimRegion(int IdCoverage_index, int dim_index)
	{ return get_DimRegion_internal(IdCoverage_index, dim_index); }
	
    const DimRegion &get_DimRegion(int IdCoverage_index, int dim_index) const
	{ return get_DimRegion_internal(IdCoverage_index, dim_index); }

    void CopyRegion
      (int &target_index, const SpHiGridRegionVecReader &source);
    
    void Init_PubSub_Item
      (MAIN_PUBSUB_ITEM * const MainPubSubItem, const int CurrentIndex) const;

    void alloc();
    
  public:
    //   Pass in the number of grids about which you anticipate you
    // will need to store information.  And pass the number of HiGrid
    // dimensions corresponding to each grid.  Then call set_next() to
    // tell this class that you're ready to set the id, lo, hi,
    // etc. for the first grid.  Then go ahead and set them.  Then
    // call set_next() again and set_Lo, set_Hi, etc. to set stuff in
    // the next grid, etc.
    //   Remember that if you need to set multiple fields that you
    // should not do it like this:
    // 
    // SpHiGridRegionVec v(3, 2);
    // v.set_next().set_field1(...);
    // v.set_next().set_field2(...);
    // 
    //   This will set field1 of the first grid and set field2 of the
    // *NEXT* grid, which means you have NOT set field2 in the first
    // grid.  What you probably want is this:
    // 
    // SpHiGridRegionVec v(3);
    // v.set_next().set_field1(...).set_field2(...);
    SpHiGridRegionVec(const int num_grids, const int numDimRegions);

    ~SpHiGridRegionVec();
    
    SpHiGridRegionVec &set_next();

    int get_NumDimRegions() const { return NumDimRegions; }
    
    SpHiGridRegionVec &set_id(int id)
	{ assert(NextIndexToSet >= 0);
	IdCoverage[NextIndexToSet].Id = id; return *this; }

    SpHiGridRegionVec &set_gridCovered_p(int gridCovered_p)
	{ assert(NextIndexToSet >= 0);
	IdCoverage[NextIndexToSet].GridCovered_p = gridCovered_p;
	return *this; }

    SpHiGridRegionVec &set_DimRegion
      (int dim_index, double Lo, double Hi, double Res)
	{
	  assert(NextIndexToSet >= 0);
	  get_DimRegion(NextIndexToSet, dim_index)
	    .set_Lo(Lo).set_Hi(Hi).set_Res(Res);
	  return *this;
	}

    SpHiGridRegionVec &set_Lo(int dim_index, double Lo)
	{ assert(NextIndexToSet >= 0);
	get_DimRegion(NextIndexToSet, dim_index).set_Lo(Lo); return *this; }

    SpHiGridRegionVec &set_Hi(int dim_index, double Hi)
	{ assert(NextIndexToSet >= 0);
	get_DimRegion(NextIndexToSet, dim_index).set_Hi(Hi); return *this; }

    SpHiGridRegionVec &SetRegionsToConcatenationOf
      (const SpHiGridRegionVecReader &reader1,
       const SpHiGridRegionVecReader &reader2);
    
    SpHiGridRegionVec &copy_grid(const SpHiGridRegionVecReader &reader);
    
    SpHiGridRegionVec &clear()
	{ NextIndexToSet = -1; return *this; }
    
    int entries() const { return NextIndexToSet + 1; }
};

class SpHiGridRegionVecIterator
{
  protected:
    SpHiGridRegionVecIterator() : CurrentIndex(-1) {}

    int plus_plus(const SpHiGridRegionVec *v)
	{ return ((++CurrentIndex) <= v->NextIndexToSet); }

    int Id(const SpHiGridRegionVec *v) const
	{ assert_valid(); return v->IdCoverage[CurrentIndex].Id; }

    int GridCovered_p(const SpHiGridRegionVec *v) const
	{ assert_valid();
	return v->IdCoverage[CurrentIndex].GridCovered_p; }
    
    void assert_valid() const { assert(CurrentIndex >= 0); }

    int get_CurrentIndex() const { return CurrentIndex; }
    
  private:
    SpHiGridRegionVecIterator(SpHiGridRegionVecIterator&) { abort(); }
    void operator=(SpHiGridRegionVecIterator&) { abort(); }

    int CurrentIndex;
};

// The current position is undefined immediately after construction of
// these two classes.  Use operator++() to define the position at the
// beginning.  Then most of the other functions like the set and get
// functions operate on the grid at the current position.  After you're
// done with this grid, call operator++() to move to the next grid.
class SpHiGridRegionVecReader : public SpHiGridRegionVecIterator
{
  public:
    // Pass in the SpHiGridRegionVec that you would like to read.
    SpHiGridRegionVecReader(const SpHiGridRegionVec *hGRVector)
      : HGRVector(hGRVector) { assert(hGRVector != NULL); }

    // Advances the iterator and returns TRUE if the position is
    // valid, or FALSE if the iterator advanced over all of the grids.
    int operator++() { return plus_plus(HGRVector); }

    int id() const { return Id(HGRVector); }
    int gridCovered_p() const { return GridCovered_p(HGRVector); }
    void Init_PubSub_Item(MAIN_PUBSUB_ITEM *MainPubSubItem) const
	{ HGRVector->Init_PubSub_Item(MainPubSubItem, get_CurrentIndex()); }
    
    const SpHiGridRegionVec *get_HGRVector() const { return HGRVector; }

    const SpHiGridRegionVec::DimRegion *get_currentDimRegion()
	const { return &(HGRVector->get_DimRegion(get_CurrentIndex(), 0)); }

  private:
    const SpHiGridRegionVec * const HGRVector;
};

class SpHiGridRegionVecModifier : public SpHiGridRegionVecIterator
{
  public:
    SpHiGridRegionVecModifier(SpHiGridRegionVec *hGRVector)
      : HGRVector(hGRVector) { assert(hGRVector != NULL); }
    
    int operator++() { return plus_plus(HGRVector); }

    SpHiGridRegionVecModifier &set_DimRegion
      (int dim_index, double Lo, double Hi, double Res)
	{
	  assert_valid();
	  HGRVector->get_DimRegion(get_CurrentIndex(), dim_index)
	    .set_Lo(Lo).set_Hi(Hi).set_Res(Res);
	  return *this;
	}

    SpHiGridRegionVecModifier &set_Res(int dim_index, double Res)
	{
	  assert_valid();
	  HGRVector->get_DimRegion(get_CurrentIndex(), dim_index).set_Res(Res);
	  return *this;
	}

  private:
    SpHiGridRegionVec * const HGRVector;
};

typedef SpHiGridRegionVec HiGridRegionVector;

#endif
