#ifndef SpHolderItem_H
#define SpHolderItem_H

#include "SpMempool.H"
#include "SpString.H"

class SpHolderItem : public SpMempool{
public:
  void SetValue(void* value);
  void* GetValue();
private:
  void* Value;
};

inline void SpHolderItem::SetValue(void* value) {Value = value;}

inline void* SpHolderItem::GetValue(){ return Value; }
  
DEFINE_MEMPOOL(SpHolderItem, 1000); //Create 1000 at a shot


// The SpPoMempool is similar to the normal SpMempool in that it has
// RemoveItem and ReturnItem methods.  The difference is that this
// uses an array of free lists and requires the user to new chunks of
// pool data for the free list rather than having the free list
// automatically allocate blocks of items.
//
// The typical way this would be used is something like this:
//
// ChunkFreeId = -1;
//
// chunk* GetChunkFromFreeList()
// {
//   extern SpPoMempool* freeList;
//   void* chunkRc = freeList->RemoveItem(ChunkFreeId);
//   if (chunkRc == NULL){
//     ChunkFreeId = freeList->AddChunk(new chunk[100],
//                                      ChunkFreeId,
//                                      100,
//                                      sizeof(chunk));
//     chunkRc = freeList->RemoveItem(ChunkFreeId);
//   }
//   return chunkRc;
// }
//
// void ReturnChunkToFreeList(chunk* c)
// {
//   extern SpPoMempool* freeList;
//   if (ChunkFreeId == -1){
//     ChunkFreeId = freeList->AddChunk(new chunk[100],
//                                      ChunkFreeId,
//                                      100,
//                                      sizeof(chunk));
//   }
//   freeList->ReturnItem(ChunkFreeId, c);
// }
//
// One optimization would be to make sure that the free id has been
// initialized before returning chunks to the free list so that one
// does not have to have the additional if statement in that code.

class SpPoMempool {
public:
  int AddChunk(void* chunk, int freeId, int numElements, int sizeOfElement);
  void* RemoveItem(int freeId);
  void ReturnItem(int freeId, void* item);

  SpPoMempool();
  ~SpPoMempool(); //Note: destroying a SpPoMemppol looses all the
		  //memory that is being managed by this free list...
protected:
  void GetMoreLists();

private:
  //Following class defined by the DEFINE_MEMPOOL macro above
  SpHolderItemMempoolItemStack* PoolList;
  int NextId;
  int NumLists;

};

inline void* SpPoMempool::RemoveItem(int freeId)
{
  void* rc = NULL;
  if (freeId != -1){
    if (freeId < NextId){ //In case someone initializes freeId to some
			  //large value (maybe in the case of restart
			  //after a checkpoint)
      SpHolderItemMempool* holder =
	(SpHolderItemMempool*)PoolList[freeId].Remove();
      if (holder){
	rc = holder->GetValue();
	DEALLOCATE_SpHolderItem(holder);
      }
    }
  }
  return rc;
}

inline void SpPoMempool::ReturnItem(int freeId, void* item)
{
  //Following class defined by the DEFINE_MEMPOOL macro above
  SpHolderItemMempool* holder= (SpHolderItemMempool*)ALLOCATE_SpHolderItem();
  holder->SetValue(item);
  PoolList[freeId].Insert(holder);
}


#endif

extern SpPoMempool* GlobalSpPoMempool;
