// SpParmSet source file

#include "SpParmSet.H"

const char SpParmSet::MAGIC = (char)(
'S' + 'p' + 'P' + 'a' + 'r' + 'm' + 'S' + 'e' + 't'
);

SpParmSet::SpParmSet (char* buff, int size) {
  int offset = 0;
  NET_INT netint;
  int normint, numelem = 0;

  Elements.SetBalancedTreeMode();

  // Check magic byte
  if (*(buff+offset) != SpParmSet::MAGIC) {
    cerr << "SpParmSet Error: this buffer wasn't returned by ";
    cerr << "SpParmSet::GenerateBuff()!" << endl;
    abort();
  } 
  offset++;

  // decode header data
  memcpy((char*)&netint, buff+offset, sizeof(NET_INT));  // packing method
  offset += sizeof(NET_INT);
  Pack = (SpParmSetPackMethod)(int) netint;

  switch(Pack) {
  case PACK_NET_TYPES:
    memcpy((char*)&netint, buff+offset, sizeof(NET_INT));  // type of parm ids
    offset += sizeof(NET_INT);
    IdType = (SpParmSetIdType)(int) netint;
    memcpy((char*)&netint, buff+offset, sizeof(NET_INT));  // # elements
    offset += sizeof(NET_INT);
    numelem = netint;
    break;
  case PACK_NATIVE:
    memcpy((char*)&normint, buff+offset, sizeof(int));  // type of parm ids
    offset += sizeof(int);
    IdType = (SpParmSetIdType) normint;
    memcpy((char*)&numelem, buff+offset, sizeof(int));  // # elements
    offset += sizeof(int);
    break;
  case PACK_XDR:
#ifdef SPPARMSET_USE_XDR
    // initialize the XDR handle
    xdrmem_create(&XDRHandle, buff+offset, size-offset, XDR_DECODE);
    // start decoding data
    xdr_enum(&XDRHandle, (int*)&IdType);  // type of parm ids
    xdr_int(&XDRHandle, &numelem);        // # elements
#endif
    break;
  }

  // decode parameter id/value data
  for(int i=0; i < numelem; i++) {
    // create a holder for the parameter value
    SpParmSetElement* newElem = new SpParmSetElement();

    // decode parameter id & insert pair into b-tree
    if(IdType == ID_TYPE_INT) {
      int id;

      switch(Pack) {
      case PACK_NET_TYPES:
	memcpy((char*)&netint, buff+offset, sizeof(NET_INT));
	offset += sizeof(NET_INT);
	id = netint;
	break;
      case PACK_NATIVE:
	memcpy((char*)&id, buff+offset, sizeof(int));
	offset += sizeof(int);
	break;
      case PACK_XDR:
#ifdef SPPARMSET_USE_XDR
	xdr_int(&XDRHandle, &id);
#endif
	break;
      }

      Elements.Insert(newElem, id);
    }
    else if(IdType == ID_TYPE_STRING) {
      char* id;
      int strlength;

      switch(Pack) {
      case PACK_NET_TYPES:
	id = buff+offset;
	strlength = strlen(id)+1;
	offset += strlength;
	break;
      case PACK_NATIVE:
	id = buff+offset;
	strlength = strlen(id)+1;
	offset += strlength;
	break;
      case PACK_XDR:
#ifdef SPPARMSET_USE_XDR
	xdr_int(&XDRHandle, &strlength);
	id = new char[strlength];
	xdr_string(&XDRHandle, &id, strlength);
#endif
	break;
      }

      Elements.Insert(newElem, id);
    }

    // decode parameter value
#ifdef SPPARMSET_USE_XDR
    newElem->UnpackBuffer(Pack, buff, offset, &XDRHandle);
#else
    newElem->UnpackBuffer(Pack, buff, offset);
#endif
  }
}

SpParmSet::SpParmSet(SpParmSet& set) {
  Pack = set.Pack;
  IdType = set.IdType;

  Elements.SetBalancedTreeMode();
  SpParmSetElement* iter;
  for(iter = (SpParmSetElement*) set.Elements.GetFirstElement();
      iter;
      iter = (SpParmSetElement*) set.Elements.GetNextElement()) {

    SpParmSetElement* insert = new SpParmSetElement(*iter);
    if(IdType == ID_TYPE_INT)
      Elements.Insert(insert, set.Elements.GetCurrentKey());
    else
      Elements.Insert(insert, set.Elements.GetCurrentKeyName());
  }
}

SpParmSet::~SpParmSet() {
  // iterate through the tree, destroying the elements as we go
  SpParmSetElement* iter;
  while((iter = (SpParmSetElement*) Elements.RemoveFirstElement())) {
    delete iter;
  }
}

int
SpParmSet::RemoveInt (int parmId) {
  SpParmSetElement* value;
  if((value = (SpParmSetElement*)Elements.Remove(parmId)) == NULL)
    return 0;
  int ret = value->GetInt();
  delete value;
  return ret;
}

int
SpParmSet::RemoveInt (char* parmName) {
  SpParmSetElement* value;
  if((value = (SpParmSetElement*)Elements.Remove(parmName)) == NULL)
    return 0;
  int ret = value->GetInt();
  delete value;
  return ret;
}

float
SpParmSet::RemoveFloat (int parmId) {
  SpParmSetElement* value;
  if((value = (SpParmSetElement*)Elements.Remove(parmId)) == NULL)
    return 0;
  float ret = value->GetFloat();
  delete value;
  return ret;
}

float
SpParmSet::RemoveFloat (char* parmName) {
  SpParmSetElement* value;
  if((value = (SpParmSetElement*)Elements.Remove(parmName)) == NULL)
    return 0;
  float ret = value->GetFloat();
  delete value;
  return ret;
}

double
SpParmSet::RemoveDouble (int parmId) {
  SpParmSetElement* value;
  if((value = (SpParmSetElement*)Elements.Remove(parmId)) == NULL)
    return 0;
  double ret = value->GetDouble();
  delete value;
  return ret;
}

double
SpParmSet::RemoveDouble (char* parmName) {
  SpParmSetElement* value;
  if((value = (SpParmSetElement*)Elements.Remove(parmName)) == NULL)
    return 0;
  double ret = value->GetDouble();
  delete value;
  return ret;
}

char*
SpParmSet::RemoveString (int parmId) {
  SpParmSetElement* value;
  if((value = (SpParmSetElement*)Elements.Remove(parmId)) == NULL)
    return NULL;
  char* ret = strdup(value->GetString());
  delete value;
  return ret;
}

char*
SpParmSet::RemoveString (char* parmName) {
  SpParmSetElement* value;
  if((value = (SpParmSetElement*)Elements.Remove(parmName)) == NULL)
    return NULL;
  char* ret = strdup(value->GetString());
  delete value;
  return ret;
}

char*
SpParmSet::RemoveBuffer (int parmId, int& size) {
  SpParmSetElement* value;
  if((value = (SpParmSetElement*)Elements.Remove(parmId)) == NULL)
    return NULL;
  char* val = value->GetBuffer(size);
  char* ret = new char[size];
  memcpy(ret, val, size);
  delete value;
  return ret;
}

char*
SpParmSet::RemoveBuffer (char* parmName, int& size) {
  SpParmSetElement* value;
  if((value = (SpParmSetElement*)Elements.Remove(parmName)) == NULL)
    return NULL;
  char* val = value->GetBuffer(size);
  char* ret = new char[size];
  memcpy(ret, val, size);
  delete value;
  return ret;
}


char*
SpParmSet::GenerateBuff(int& size) {
  size = GetPackLength();
  char* buff = new char[size];

  this->PackBuff(buff);
  return buff;
}

void
SpParmSet::PackBuff(char *buff) {
  int size = GetPackLength();
  int i = 0;

  // If we are using XDR encoding, the XDR handle must be initialized.
  // The header information starts off with a NET_INT declaring the
  // packing method used.  We do not want XDR to overwrite that NET_INT.
#ifdef SPPARMSET_USE_XDR
  if(Pack == PACK_XDR)
    xdrmem_create(&XDRHandle, buff+sizeof(NET_INT), size-sizeof(NET_INT),
		  XDR_ENCODE);
#endif
  // encode header data
  GenerateHeader(buff, i);

  NET_INT netnum;

  // encode element data
  SpParmSetElement* iter = NULL;
  for(iter = (SpParmSetElement*)Elements.GetFirstElement();
      iter;
      iter = (SpParmSetElement*)Elements.GetNextElement()) {

    if(IdType == ID_TYPE_INT) {
      int id = (int)Elements.GetCurrentKey();
      switch(Pack) {
      case PACK_NET_TYPES:
	netnum = id;
	memcpy(buff+i, (char*)&netnum, sizeof(NET_INT));
	i += sizeof(NET_INT);
	break;
      case PACK_NATIVE:
	memcpy(buff+i, (char*)&id, sizeof(int));
	i += sizeof(int);
	break;
      case PACK_XDR:
#ifdef SPPARMSET_USE_XDR
	xdr_int(&XDRHandle, &id);
#endif
	break;
      }
#ifdef SPPARMSET_USE_XDR
      iter->PackBuffer(Pack, buff, i, &XDRHandle);
#else
      iter->PackBuffer(Pack, buff, i);
#endif
    }
    else if(IdType == ID_TYPE_STRING) {
      char* id = Elements.GetCurrentKeyName();
      int strlength = strlen(id)+1;

      switch(Pack) {
      case PACK_NET_TYPES:
	memcpy(buff+i, id, strlength);
	i += strlength;
	break;
      case PACK_NATIVE:
	memcpy(buff+i, id, strlength);
	i += strlength;
	break;
      case PACK_XDR:
#ifdef SPPARMSET_USE_XDR
	xdr_int(&XDRHandle, &strlength);
	xdr_string(&XDRHandle, &id, strlength);
#endif
	break;
      }
#ifdef SPPARMSET_USE_XDR
      iter->PackBuffer(Pack, buff, i, &XDRHandle);
#else
      iter->PackBuffer(Pack, buff, i);
#endif
    }
  }
}

int
SpParmSet::GetPackLength() {
  SpParmSetElement* iter = NULL;
  int length = 0;

  // compute header length
  length += GetHeaderLength();

  // compute set values length
  for(iter = (SpParmSetElement*)Elements.GetFirstElement();
      iter;
      iter = (SpParmSetElement*)Elements.GetNextElement()) {

    if(IdType == ID_TYPE_INT) {
      switch(Pack) {
      case PACK_NET_TYPES:
	length += sizeof(NET_INT);
	break;
      case PACK_NATIVE:
	length += sizeof(int);
	break;
      case PACK_XDR:
	length += sizeof(int);
	break;
      }
      length += iter->GetPackLength(Pack);
    }
    else if(IdType == ID_TYPE_STRING) {
      char* id = Elements.GetCurrentKeyName();
      int strlength = strlen(id)+1;

      switch(Pack) {
      case PACK_NET_TYPES:
	length += strlength;
	break;
      case PACK_NATIVE:
	length += strlength;
	break;
      case PACK_XDR:
#ifdef SPPARMSET_USE_XDR
	length += sizeof(int) + strlength;
#endif
	break;
      }
      length += iter->GetPackLength(Pack);
    }
  }
  length++; // Increment to account for magic byte at beginning
  return length;
}

int
SpParmSet::GetHeaderLength() {
  int length = sizeof(NET_INT);
  length += 1;  // for MAGIC byte

  switch(Pack) {
  case PACK_NET_TYPES:
    length += sizeof(NET_INT) * 2;
    break;
  case PACK_NATIVE:
    length += sizeof(int) * 2;
    break;
  case PACK_XDR:
    length += sizeof(int) * 2;
    break;
  }

  return length;
}

void
SpParmSet::GenerateHeader(char* buff, int& offset) {
  NET_INT netint;
  int normint;

  buff[offset] = MAGIC;
  offset++;

  netint = (int)Pack;                   // byte packing method
  memcpy(buff+offset, (char*)&netint, sizeof(NET_INT));
  offset += sizeof(NET_INT);

  switch(Pack) {
  case PACK_NET_TYPES:
    netint = (int)IdType;                 // int or str ids
    memcpy(buff+offset, (char*)&netint, sizeof(NET_INT));
    offset += sizeof(NET_INT);
    netint = Elements.GetNumElements();   // # of elements
    memcpy(buff+offset, (char*)&netint, sizeof(NET_INT));
    offset += sizeof(NET_INT);
    break;
  case PACK_NATIVE:
    normint = (int)IdType;                 // int or str ids
    memcpy(buff+offset, (char*)&normint, sizeof(int));
    offset += sizeof(int);
    normint = Elements.GetNumElements();  // # of elements
    memcpy(buff+offset, (char*)&normint, sizeof(int));
    offset += sizeof(int);
    break;
  case PACK_XDR:
#ifdef SPPARMSET_USE_XDR
    xdr_enum(&XDRHandle, (int*)&IdType);  // int or str ids
    normint = Elements.GetNumElements();  // # of elements
    xdr_int(&XDRHandle, &normint);
#endif
    break;
  }
}


