// SpParmSet header file

#ifndef _SpParmSet_H
#define _SpParmSet_H

#ifndef __SPEEDES_NT__
#define SPPARMSET_USE_XDR
#endif

#include "SpParmSetElement.H"
#include "SpBinaryTree.H"

typedef enum {
  ID_TYPE_NONE,
  ID_TYPE_INT,
  ID_TYPE_STRING
} SpParmSetIdType;

/**************************************************************************
 * Class: SpParmSet
 * Summary: A container class that stores values of certain known types
 *          that are referenced by either an int or string key.  Mixing
 *          the two types of keys in the same set is not permitted.  
 *          The currently supported value types are int, float, double,
 *          string, and buffer (a byte array of arbitrary length).
 *
 *          The GenerateBuff method will cause the set to pack itself 
 *          a byte array.  The manner in which the set packs itself is
 *          determined by the user with the SetNET_TYPES, SetNative,
 *          and SetXDR methods.
 *************************************************************************/

class SpParmSet {
public:
  SpParmSet() {                       // basic ctor
    Elements.SetBalancedTreeMode();
    Pack = PACK_NET_TYPES;
    IdType = ID_TYPE_NONE;
  }
  SpParmSet(SpParmSet& set);         // copy ctor
  SpParmSet(char* buff, int size);   // unpack set from buffer
  ~SpParmSet();                      // frees memory stored in binary tree

  /*************************************************************************
   * Method: Insert<TYPE>
   * Summary: Insert a value of type <TYPE> into the set with the specified
   *          key.  Keys can either be ints or strings, but once a value is
   *          inserted into the set with a particular key type, the other
   *          key type may not be used to insert new values into the set.
   ************************************************************************/

  void InsertInt (int parmId, int value) {
    if(IdType == ID_TYPE_NONE)  IdType = ID_TYPE_INT;
    Elements.Insert((new SpParmSetElement(value)), parmId);
  }
  void InsertInt (char* parmName, int value) {
    if(IdType == ID_TYPE_NONE)  IdType = ID_TYPE_STRING;
    Elements.Insert((new SpParmSetElement(value)), parmName);
  }
  void InsertFloat (int parmId, float value) {
    if(IdType == ID_TYPE_NONE)  IdType = ID_TYPE_INT;
    Elements.Insert((new SpParmSetElement(value)), parmId);
  }
  void InsertFloat (char* parmName, float value) {
    if(IdType == ID_TYPE_NONE)  IdType = ID_TYPE_STRING;
    Elements.Insert((new SpParmSetElement(value)), parmName);
  }
  void InsertDouble (int parmId, double value) {
    if(IdType == ID_TYPE_NONE)  IdType = ID_TYPE_INT;
    Elements.Insert((new SpParmSetElement(value)), parmId);
  }
  void InsertDouble (char* parmName, double value) {
    if(IdType == ID_TYPE_NONE)  IdType = ID_TYPE_STRING;
    Elements.Insert((new SpParmSetElement(value)), parmName);
  }
  void InsertString (int parmId, char* value) {
    if(IdType == ID_TYPE_NONE)  IdType = ID_TYPE_INT;
    Elements.Insert((new SpParmSetElement(value)), parmId);
  }
  void InsertString (char* parmName, char* value) {
    if(IdType == ID_TYPE_NONE)  IdType = ID_TYPE_STRING;
    Elements.Insert((new SpParmSetElement(value)), parmName);
  }
  void InsertBuffer (int parmId, char* buff, int size) {
    if(IdType == ID_TYPE_NONE)  IdType = ID_TYPE_INT;
    Elements.Insert((new SpParmSetElement(buff, size)), parmId);
  }
  void InsertBuffer (char* parmName, char* buff, int size) {
    if(IdType == ID_TYPE_NONE)  IdType = ID_TYPE_STRING;
    Elements.Insert((new SpParmSetElement(buff, size)), parmName);
  }

  /*************************************************************************
   * Method: Get<TYPE>
   * Summary: Retrieves a value of type <TYPE> that was inserted into the
   *          set with the specified key.  Zero or NULL will be returned
   *          if the value could not be found.  This is not ideal in the
   *          case of the numerical values, so watch out.  If the value
   *          was inserted as one type and is being retrieved as another
   *          (inserted as double, retrieved as int for example) the value
   *          will be cast to the retrieved type.  This only holds true
   *          for the numerical value types (you cannot get a string value
   *          as an int, or vice versa).
   ************************************************************************/

  int GetInt (int parmId) {
    SpParmSetElement* value;
    return ((value = (SpParmSetElement*)Elements.Find(parmId)) ? 
	    (value->GetInt()) : 0);
  }
  int GetInt (char* parmName) {
    SpParmSetElement* value;
    return ((value = (SpParmSetElement*)Elements.Find(parmName)) ?
	    (value->GetInt()) : 0);
  }
  float GetFloat (int parmId) {
    SpParmSetElement* value;
    return ((value = (SpParmSetElement*)Elements.Find(parmId)) ?
	    (value->GetFloat()) : 0);
  }
  float GetFloat (char* parmName) {
    SpParmSetElement* value;
    return ((value = (SpParmSetElement*)Elements.Find(parmName)) ?
	    (value->GetFloat()) : 0);
  }
  double GetDouble (int parmId) {
    SpParmSetElement* value;
    return ((value = (SpParmSetElement*)Elements.Find(parmId)) ?
	    (value->GetDouble()) : 0);
  }
  double GetDouble (char* parmName) {
    SpParmSetElement* value;
    return ((value = (SpParmSetElement*)Elements.Find(parmName)) ?
	    (value->GetDouble()) : 0);
  }
  char* GetString (int parmId) {
    SpParmSetElement* value;
    return ((value = (SpParmSetElement*)Elements.Find(parmId)) ?
	    (value->GetString()) : 0);
  }
  char* GetString (char* parmName) {
    SpParmSetElement* value;
    return ((value = (SpParmSetElement*)Elements.Find(parmName)) ?
	    (value->GetString()) : 0);
  }
  char* GetBuffer (int parmId, int& size) {
    SpParmSetElement* value = (SpParmSetElement*)Elements.Find(parmId);
    if(value)
      return value->GetBuffer(size);
    return NULL;
  }    
  char* GetBuffer (char* parmName, int& size) {
    SpParmSetElement* value = (SpParmSetElement*)Elements.Find(parmName);
    if(value)
      return value->GetBuffer(size);
    return NULL;
  }

  /*************************************************************************
   * Method: Remove<TYPE>
   * Summary: See the comment for Get<TYPE> above.  The value is removed
   *          from the set before it is retured to the caller.
   * Side Effects: In the case of RemoveString and RemoveBuffer, the
   *               value returned must be deleted by the caller.
   ************************************************************************/

  int RemoveInt (int parmId);
  int RemoveInt (char* parmName);
  float RemoveFloat (int parmId);
  float RemoveFloat (char* parmName);
  double RemoveDouble (int parmId);
  double RemoveDouble (char* parmName);
  char* RemoveString (int parmId);
  char* RemoveString (char* parmName);
  char* RemoveBuffer (int parmId, int& size);
  char* RemoveBuffer (char* parmName, int& size);

  /*************************************************************************
   * Method: GetNumElements
   * Summary: Returns the number of elements in this set.
   ************************************************************************/

  int GetNumElements() { return Elements.GetNumElements(); }

  /*************************************************************************
   * Method: GetParmIdType
   * Summary: Returns the type of keys used in this set.
   ************************************************************************/

  SpParmSetIdType GetParmIdType() { return IdType; }

  /*************************************************************************
   * Method: GetFirst<TYPE>ParmId
   *         GetNext<TYPE>ParmId
   * Summary: Used to iterate through the keys of the set.
   ************************************************************************/

  int* GetFirstIntParmId() {
    void* ptr = Elements.GetFirstElement();
    if(ptr)  { keyHolder = (int)Elements.GetCurrentKey(); return &keyHolder; }
    return NULL;
  }
  int* GetNextIntParmId() {
    void* ptr = Elements.GetNextElement();
    if(ptr)  { keyHolder = (int)Elements.GetCurrentKey(); return &keyHolder; }
    return NULL;
  }
  char* GetFirstStringParmId() {
    void* ptr = Elements.GetFirstElement();
    return (ptr ? Elements.GetCurrentKeyName() : (char*)NULL);
  }
  char* GetNextStringParmId() {
    void* ptr = Elements.GetNextElement();
    return (ptr ? Elements.GetCurrentKeyName() : (char*)NULL);
  }

  /*************************************************************************
   * Method: operator []
   * Summary: Using the [] operator is very similar to the Get<TYPE> 
   *          methods above.  The caller passes in an int or string key
   *          (whichever is appropriate for the set), and a reference to
   *          the corresponding SpParmSetElement is returned.  The 
   *          SpParmSetElement class has conversion methods for all the
   *          known value types.  So the following two statements are
   *          identical:
   *
   *                int value1 = set.GetInt("key");
   *                int value2 = set["key"];
   *
   *          Alternatively, the caller can use the [] operator to insert
   *          values into the set.  If an unknown key is passed in, a new
   *          SpParmSetElement is created and inserted into the set.  The
   *          SpParmSetElement class has the = operator overloaded for all
   *          the known value types.  So the following two statements are
   *          identical:
   *
   *                set.InsertInt("key", 5);
   *                set["key"] = 5;
   *
   *          Values of type buffer are a special case since they need
   *          an extra size parameter.  In this case you must use the
   *          GetBuffer and SetBuffer methods from the SpParmSetElement
   *          class:
   *
   *                set["key"].SetBuffer(buff, size);
   *                buff = set["key"].GetBuffer(size);
   *
   ************************************************************************/

  SpParmSetElement& operator [] (int parmId) {
    SpParmSetElement* value;
    if((value = (SpParmSetElement*)Elements.Find(parmId)) == NULL) {
      if(IdType == ID_TYPE_NONE)  IdType = ID_TYPE_INT;
      Elements.Insert((value=new SpParmSetElement()), parmId);
    }
    return *value;
  }
  SpParmSetElement& operator [] (char* parmName) {
    SpParmSetElement* value;
    if((value = (SpParmSetElement*)Elements.Find(parmName)) == NULL) {
      if(IdType == ID_TYPE_NONE)  IdType = ID_TYPE_STRING;
      Elements.Insert((value=new SpParmSetElement()), parmName);
    }
    return *value;
  }

  /*************************************************************************
   * Method: SetNET_TYPES
   *         SetNative
   *         SetXDR
   * Summary: These methods allow the caller to choose which packing method
   *          the SpParmSet will use when it encodes itself for transport
   *          over a network.  NET_TYPES tells the set to use SPEEDES 
   *          NET_INTs and NET_DOUBLEs when packing numerical values.
   *          Native means that the numerical values are packed as is, and
   *          this method should only be used when operating on a
   *          homogeneous network.  XDR tells the set to use SUN's XDR
   *          encoding protocol.
   ************************************************************************/

  inline void SetNET_TYPES() {
#if HETEROGENEOUS_DATA_TYPE
    Pack = PACK_NET_TYPES;
#else
    Pack = PACK_NATIVE;
#endif
  }
  inline void SetNative() { Pack = PACK_NATIVE; }
  inline void SetXDR() {
#ifdef SPPARMSET_USE_XDR
    Pack = PACK_XDR;
#else
    cerr << "SpParmSet: must define SPPARMSET_USE_XDR to use XDR encoding.\n";
#endif
  }

  /*************************************************************************
   * Method: GenerateBuff
   * Summary: Causes the set to pack itself into a byte array that is
   *          returned to the caller.  The size of the array is stored in
   *          the 'size' parameter.
   * Side Effects: The returned buffer must be deleted by the caller.
   ************************************************************************/

  char* GenerateBuff(int& size);

  /*************************************************************************
   * Method: GetPackLength
   * Summary: Returns the number of bytes that will be needed to hold this
   *          set if it is compacted into a buffer.
   *
   * Method: PackBuff
   * Summary: Causes the set to pack itself into a user-supplied byte
   *          array.  The byte array must have enough space to hold the
   *          set, the size of which can be determined with the 
   *          GetPackLength method.
   ************************************************************************/

  int GetPackLength();
  void PackBuff(char *buff);

private:
  // determines how many bytes the header of the packed set will occupy
  int GetHeaderLength();
  // packs the set's header information into a byte array
  void GenerateHeader(char* buff, int& offset);

  static const char MAGIC;   // helps detect corrupt flattened parmsets
  SpBinaryTree Elements;     // where keys/values are stored
  SpParmSetIdType IdType;    // set is using int or string keys
  SpParmSetPackMethod Pack;  // packing method - defaults to NET_TYPES
#ifdef SPPARMSET_USE_XDR
  XDR XDRHandle;  // for XDR encoding/decoding
#endif
  int keyHolder;  // needed so we can return an int* on iteration methods
};

#endif
