// SpParmSetElement header file

#ifndef _SpParmSetElement_H
#define _SpParmSetElement_H

#include "SpString.H"
#include "SpIostream.H"

#ifdef SPPARMSET_USE_XDR
#ifdef __cplusplus
extern "C" {
#endif
#include <rpc/types.h>
#include <rpc/xdr.h>
#ifdef __cplusplus
}
#endif
#endif

#include "SpNetTypes.H"

typedef enum {
  PACK_NET_TYPES,
  PACK_NATIVE,
  PACK_XDR
} SpParmSetPackMethod;

typedef enum {
  NONE_DATA,
  INT_DATA,
  FLOAT_DATA,
  DOUBLE_DATA,
  STRING_DATA,
  BUFFER_DATA
} SpParmSetDataType;

/**************************************************************************
 * Class: SpParmSetDataItem
 * Summary: Abstract base class that defines the accessor methods for any
 *          value type that can be stored in an SpParmSet.  Each instance
 *          of SpParmSetElement (see below) keeps a pointer to a DataItem
 *          that defines the value of that set element.
 *************************************************************************/

class SpParmSetDataItem {
public:
  virtual ~SpParmSetDataItem() { ; }        // virtual dtor

  /*************************************************************************
   * Method: Clone
   * Summary: Return a pointer to a fresh copy of this data item.  This new
   *          data item must be deleted by the caller.
   ************************************************************************/

  virtual SpParmSetDataItem* Clone() = 0;

  /*************************************************************************
   * Method: Set<TYPE>
   * Summary: Assign the passed in value to the value of this data item.
   *          This operation may not be possible if the two values are not
   *          of similar types (for instance, you cannot call "SetInt" on
   *          a string data item).
   ************************************************************************/

  virtual void SetInt (int val) = 0;
  virtual void SetFloat (float val) = 0;
  virtual void SetDouble (double val) = 0;
  virtual void SetString (char* val) = 0;
  virtual void SetBuffer (char* val, int size) = 0;

  /*************************************************************************
   * Method: Get<TYPE>
   * Summary: Return the value of this data item.  If the data item type
   *          and the return type do not match, the data item value is
   *          cast to the appropriate type.  This operation may not be
   *          possible if the two values are not of similar types (for 
   *          instance, you cannot call "GetInt" on a string data item).
   ************************************************************************/

  virtual int GetInt () = 0;
  virtual float GetFloat () = 0;
  virtual double GetDouble () = 0;
  virtual char* GetString () = 0;
  virtual char* GetBuffer (int& size) = 0;

  /*************************************************************************
   * Method: GetPackLength
   * Summary: Return the number of bytes this data item would occupy if
   *          it was packed into a byte array with the specified packing
   *          method.
   ************************************************************************/

  virtual int GetPackLength (SpParmSetPackMethod pack) = 0;

  /*************************************************************************
   * Method: PackBuffer
   * Summary: Pack this data item into the specified buffer using the
   *          specified packing method.
   ************************************************************************/
#ifdef SPPARMSET_USE_XDR
  virtual void PackBuffer (SpParmSetPackMethod pack, 
			   char* buff, int& offset,
			   XDR* xdrHandle) = 0;
#else
  virtual void PackBuffer (SpParmSetPackMethod pack, 
			   char* buff, int& offset) = 0;
#endif
  /*************************************************************************
   * Method: UnpackBuffer
   * Summary: Return a new data item, unpacked from the supplied buffer
   *          with the specified packing method.
   * Side Effects: The new data item must be deleted by the caller.
   ************************************************************************/
#ifdef SPPARMSET_USE_XDR
  static SpParmSetDataItem* UnpackBuffer (SpParmSetPackMethod pack, 
					  char* buff, int& offset,
					  XDR* xdrHandle);
#else
  static SpParmSetDataItem* UnpackBuffer (SpParmSetPackMethod pack, 
					  char* buff, int& offset);
#endif
};

/**************************************************************************
 * Class: SpParmSetIntData
 * Summary: A data item that holds an integer value.
 *************************************************************************/

class SpParmSetIntData : public SpParmSetDataItem {
public:
  SpParmSetIntData(int val) { Value = val; }
#ifdef SPPARMSET_USE_XDR
  SpParmSetIntData(SpParmSetPackMethod pack, char* buff, int& offset,
		   XDR* xdrHandle);
#else
  SpParmSetIntData(SpParmSetPackMethod pack, char* buff, int& offset);
#endif
  ~SpParmSetIntData() { ; }

  SpParmSetDataItem* Clone() { return new SpParmSetIntData(Value); }

  void SetInt (int val) { Value = val; }
  void SetFloat (float val) { Value = (int) val; }
  void SetDouble (double val) { Value = (int) val; }
  void SetString (char*) { 
    cerr << "Cannot assign string to int data.\n"; }
  void SetBuffer (char*, int) {
    cerr << "Cannot assign buffer to int data.\n"; }

  int GetInt () { return Value; }
  float GetFloat () { return (float) Value; }
  double GetDouble () { return (double) Value; }
  char* GetString () { 
    cerr << "Cannot return int data as string.\n"; return NULL; }
  char* GetBuffer (int& size) {
    cerr << "Cannot return int data as buffer.\n"; size=size; return NULL; }

  int GetPackLength (SpParmSetPackMethod pack);
#ifdef SPPARMSET_USE_XDR
  void PackBuffer (SpParmSetPackMethod pack, char* buff, int& offset,
		   XDR* xdrHandle);
#else
  void PackBuffer (SpParmSetPackMethod pack, char* buff, int& offset);
#endif

private:
  int Value;
};

/**************************************************************************
 * Class: SpParmSetFloatData
 * Summary: A data item that holds a float value.
 *************************************************************************/

class SpParmSetFloatData : public SpParmSetDataItem {
public:
  SpParmSetFloatData(float val) { Value = val; }
#ifdef SPPARMSET_USE_XDR
  SpParmSetFloatData(SpParmSetPackMethod pack, char* buff, int& offset,
		     XDR* xdrHandle);
#else
  SpParmSetFloatData(SpParmSetPackMethod pack, char* buff, int& offset);
#endif
  ~SpParmSetFloatData() { ; }

  SpParmSetDataItem* Clone() { return new SpParmSetFloatData(Value); }

  void SetInt (int val) { Value = (float) val; }
  void SetFloat (float val) { Value = val; }
  void SetDouble (double val) { Value = (float) val; }
  void SetString (char*) {
    cerr << "Cannot assign string to float data.\n"; }
  void SetBuffer (char*, int) {
    cerr << "Cannot assign buffer to float data.\n"; }

  int GetInt () { return (int) Value; }
  float GetFloat () { return Value; }
  double GetDouble () { return (double) Value; }
  char* GetString () {
    cerr << "Cannot return float data as string.\n"; return NULL; }
  char* GetBuffer (int&) {
    cerr << "Cannot return float data as buffer.\n"; return NULL; }

  int GetPackLength (SpParmSetPackMethod pack);
#ifdef SPPARMSET_USE_XDR
  void PackBuffer (SpParmSetPackMethod pack, char* buff, int& offset,
		   XDR* xdrHandle);
#else
  void PackBuffer (SpParmSetPackMethod pack, char* buff, int& offset);
#endif

private:
  float Value;
};

/**************************************************************************
 * Class: SpParmSetDoubleData
 * Summary: A data item that holds a double value.
 *************************************************************************/

class SpParmSetDoubleData : public SpParmSetDataItem {
public:
  SpParmSetDoubleData(double val) { Value = val; }
#ifdef SPPARMSET_USE_XDR
  SpParmSetDoubleData(SpParmSetPackMethod pack, char* buff, int& offset,
		      XDR* xdrHandle);
#else
  SpParmSetDoubleData(SpParmSetPackMethod pack, char* buff, int& offset);
#endif
  ~SpParmSetDoubleData() { ; }

  SpParmSetDataItem* Clone() { return new SpParmSetDoubleData(Value); }

  void SetInt (int val) { Value = (double) val; }
  void SetFloat (float val) { Value = (double) val; }
  void SetDouble (double val) { Value = val; }
  void SetString (char*) {
    cerr << "Cannot assign string to double data.\n"; }
  void SetBuffer (char*, int) {
    cerr << "Cannot assign buffer to double data.\n"; }

  int GetInt () { return (int) Value; }
  float GetFloat () { return (float) Value; }
  double GetDouble () { return Value; }
  char* GetString () {
    cerr << "Cannot return double data as string.\n"; return NULL; }
  char* GetBuffer (int&) {
    cerr << "Cannot return double data as buffer.\n"; return NULL; }

  int GetPackLength (SpParmSetPackMethod pack);
#ifdef SPPARMSET_USE_XDR
  void PackBuffer (SpParmSetPackMethod pack, char* buff, int& offset,
		   XDR* xdrHandle);
#else
  void PackBuffer (SpParmSetPackMethod pack, char* buff, int& offset);
#endif

private:
  double Value;
};

/**************************************************************************
 * Class: SpParmSetStringData
 * Summary: A data item that holds a string value.
 *************************************************************************/

class SpParmSetStringData : public SpParmSetDataItem {
public:
  SpParmSetStringData(char* val) {
    Value = new char[strlen(val)+1];
    strcpy(Value, val);
  }
#ifdef SPPARMSET_USE_XDR
  SpParmSetStringData(SpParmSetPackMethod pack, char* buff, int& offset,
		      XDR* xdrHandle);
#else
  SpParmSetStringData(SpParmSetPackMethod pack, char* buff, int& offset);
#endif
  ~SpParmSetStringData() { delete [] Value; }

  SpParmSetDataItem* Clone() { return new SpParmSetStringData(Value); }

  void SetInt (int) { 
    cerr << "Cannot assign int to string data.\n"; }
  void SetFloat (float) {
    cerr << "Cannot assign float to string data.\n"; }
  void SetDouble (double) {
    cerr << "Cannot assign double to string data.\n"; }
  void SetString (char* val) {
    if(strlen(val) > strlen(Value)) {
      delete [] Value;
      Value = new char[strlen(val)+1];
    }
    strcpy(Value, val);
  }
  void SetBuffer (char*, int) {
    cerr << "Cannot assign buffer to string data.\n"; }

  int GetInt () { cerr << "Cannot return string data as int.\n"; return 0; }
  float GetFloat () { 
    cerr << "Cannot return string data as float.\n"; return 0; }
  double GetDouble () { 
    cerr << "Cannot return string data as double.\n"; return 0; }
  char* GetString () { return Value; }
  char* GetBuffer (int& size) {
    cerr << "Cannot return string data as buffer.\n"; size=size; return NULL; }

  int GetPackLength (SpParmSetPackMethod pack);
#ifdef SPPARMSET_USE_XDR
  void PackBuffer (SpParmSetPackMethod pack, char* buff, int& offset,
		   XDR* xdrHandle);
#else
  void PackBuffer (SpParmSetPackMethod pack, char* buff, int& offset);
#endif

private:
  char* Value;
};

/**************************************************************************
 * Class: SpParmSetBufferData
 * Summary: A data item that holds a byte array value of arbitrary length.
 *************************************************************************/

class SpParmSetBufferData : public SpParmSetDataItem {
public:
  SpParmSetBufferData(char* buff, int size) {
    Size = size;
    Value = new char[Size];
    memcpy(Value, buff, Size);
  }
#ifdef SPPARMSET_USE_XDR
  SpParmSetBufferData(SpParmSetPackMethod pack, char* buff, int& offset,
		      XDR* xdrHandle);
#else
  SpParmSetBufferData(SpParmSetPackMethod pack, char* buff, int& offset);
#endif
  ~SpParmSetBufferData() { delete [] Value; }

  SpParmSetDataItem* Clone() { return new SpParmSetBufferData(Value, Size); }

  void SetInt (int) { 
    cerr << "Cannot assign int to buffer data.\n"; }
  void SetFloat (float) { 
    cerr << "Cannot assign float to buffer data.\n"; }
  void SetDouble (double) {
    cerr << "Cannot assign double to buffer data.\n"; }
  void SetString (char*) {
    cerr << "Cannot assign string to buffer data.\n"; }
  void SetBuffer (char* val, int size) {
    if((u_int)size > Size) {
      delete [] Value;
      Value = new char[size];
    }
    Size = size;
    memcpy(Value, val, Size);
  }

  int GetInt () { cerr << "Cannot return buffer data as int.\n"; return 0; }
  float GetFloat () {
    cerr << "Cannot return buffer data as float.\n"; return 0; }
  double GetDouble () {
    cerr << "Cannot return buffer data as double.\n"; return 0; }
  char* GetString () {
    cerr << "Cannot return buffer data as string.\n"; return NULL; }
  char* GetBuffer (int& size) {
    size = Size;
    return Value;
  }

  int GetPackLength (SpParmSetPackMethod pack);
#ifdef SPPARMSET_USE_XDR
  void PackBuffer (SpParmSetPackMethod pack, char* buff, int& offset,
		   XDR* xdrHandle);
#else
  void PackBuffer (SpParmSetPackMethod pack, char* buff, int& offset);
#endif

private:
  char* Value;
  unsigned int Size;
};


/**************************************************************************
 * Class: SpParmSetElement
 * Summary: An element in an SpParmSet.  It holds a pointer to some
 *          generic SpParmSetDataItem (see above) and actually doesn't
 *          know what type of value it holds.  Nor does it need to.
 *************************************************************************/

class SpParmSetElement {
public:
  SpParmSetElement ();                       // type none
  SpParmSetElement (int val);                // type int
  SpParmSetElement (float val);              // type float
  SpParmSetElement (double val);             // type double
  SpParmSetElement (char* val);              // type string
  SpParmSetElement (char* buff, int size);   // type buffer
  SpParmSetElement (const SpParmSetElement& elem);  // copy ctor
  ~SpParmSetElement ();

  /*************************************************************************
   * Method: Set<TYPE>
   * Summary: Assign the specified value to the value of the data item.
   *          If this element has not been initialized yet, a new data
   *          item is created and set to the specified value.  If this
   *          element has been initialized, this operation may not be
   *          possible, depending on the type of the specified value and
   *          the type of the data item (for instance, you cannot call 
   *          "SetInt" on a string data item).
   ************************************************************************/

  void SetInt (int val) {
    if(!Data)  Data = new SpParmSetIntData(val);
    else       Data->SetInt(val);
  }
  void SetFloat (float val) {
    if(!Data)  Data = new SpParmSetFloatData(val);
    else       Data->SetFloat(val);
  }
  void SetDouble (double val) {
    if(!Data)  Data = new SpParmSetDoubleData(val);
    else       Data->SetDouble(val);
  }
  void SetString (char* val) {
    if(!Data)  Data = new SpParmSetStringData(val);
    else       Data->SetString(val);
  }
  void SetBuffer (char* val, int size) {
    if(!Data)  Data = new SpParmSetBufferData(val, size);
    else       Data->SetBuffer(val, size);
  }

  /*************************************************************************
   * Method: Get<TYPE>
   * Summary: Return the value of the data item.  If the data item type
   *          and the return type do not match, the data item value is
   *          cast to the appropriate type.  This operation may not be
   *          possible if the two values are not of similar types (for 
   *          instance, you cannot call "GetInt" on a string data item).
   ************************************************************************/

  inline int GetInt () { if(Data) return Data->GetInt();  return 0; }
  inline float GetFloat () { if(Data) return Data->GetFloat();  return 0; }
  inline double GetDouble () { if(Data) return Data->GetDouble();  return 0; }
  inline char* GetString () { if(Data) return Data->GetString();  
                              return NULL; }
  inline char* GetBuffer (int& size) { if(Data) return Data->GetBuffer(size);  
                                       return NULL; }

  /*************************************************************************
   * Method: operator =
   * Summary: Allows the value of the data item to be set in a more
   *          intuitive fashion.
   ************************************************************************/

  inline void operator = (int val) { this->SetInt(val); }
  inline void operator = (float val) { this->SetFloat(val); }
  inline void operator = (double val) { this->SetDouble(val); }
  void operator = (char* val) { this->SetString(val); }

  /*************************************************************************
   * Method: operator <TYPE>
   * Summary: Allows the value of the data item to be retrieved in a more
   *          intuitive fashion.
   ************************************************************************/

  inline operator int () { return this->GetInt(); }
  inline operator float () { return this->GetFloat(); }
  inline operator double () { return this->GetDouble(); }
  inline operator char* () { return this->GetString(); }

  //  inline SpParmSetDataItem* GetData () { return Data; }

  /*************************************************************************
   * Method: GetPackLength
   * Summary: Returns the number of bytes it would take to hold this
   *          element (including its data item) using the specified packing
   *          method.
   ************************************************************************/

  int GetPackLength (SpParmSetPackMethod pack);

  /*************************************************************************
   * Method: PackBuffer
   * Summary: Packs this element (including its data item) into the 
   *          specified buffer using the specified packing method.
   ************************************************************************/
#ifdef SPPARMSET_USE_XDR
  void PackBuffer (SpParmSetPackMethod pack, char* buff, int& offset,
		   XDR* xdrHandle);
#else
  void PackBuffer (SpParmSetPackMethod pack, char* buff, int& offset);
#endif

  /*************************************************************************
   * Method: UnpackBuffer
   * Summary: Initializes this element from the specified buffer using
   *          the specified packing method.
   ************************************************************************/
#ifdef SPPARMSET_USE_XDR
  void UnpackBuffer (SpParmSetPackMethod pack, char* buff, int& offset,
		     XDR* xdrHandle);
#else
  void UnpackBuffer (SpParmSetPackMethod pack, char* buff, int& offset);
#endif

private:
  SpParmSetDataItem* Data;  // the data item
};

#endif
