// SpSet.C
// Implementation of SpSet subclass methods

#include "SpStdlib.H"
#include "SpCtype.H"
#include "SpIostream.H"
#include "SpAssert.H"
#define __SpSet_Internal__
#include "SpSet.H"
#define __STRICT_ANSI__
#include "SpNetTypes.H"
#undef __STRICT_ANSI__

SpSet *SpSet::Embed(char *name) 
{
  if (Namespace->Find( name)) {
    // If an element of the same name already exists, give up
    // and return NULL
    return NULL;
  }
    
  SpSet *retval = new SpSet(name, this);
  
  return retval;
}

SpSetElement *SpSet::GetElement(char *name)
{
  // Search first for name in the general namespace of this set
  SpSetElement *ep;
  if (!(NULL == (ep = (SpSetElement *)Namespace->Find(name)))) {

    // Resolve unresolved references if necessary
    if (ep->GetType() == UNRES_ET) {

      // The following will actually create a reference
      // element to the first element in the scope of this
      // set, and return a pointer to it, or NULL if no
      // such name lies in scope
      SpSetElement *reference = createreference(name);
      if (reference) {
	delete ep;
	Namespace->Remove(name);
	ep = reference;	
	Namespace->Insert(ep,name);
      }
    }
    return (SpSetElement *)ep;
  }
  return NULL;
}  

// The following method will return the element corresponding to the
// specified name, or, if that cannot be found, a newly created 
// element of the specified type

SpSetElement &SpSet::Element(char *name, ElementType type,
			     int arrayLength) 
{
  SpSetElement *ep = GetElement(name);
  if (ep != NULL && ep->GetType() != type) {
    cerr << "Element " << name << " is of type ";
    cerr << SpSetElementName[(int)ep->GetType()] << ", ";
    cerr << "not " << SpSetElementName[(int)type] << "!" << endl;
    abort();
  }
  else {
    if (arrayLength <= 0) {
      switch(type) {
      case INT_ET:
	ep = new SpSetInt;
	break;
      case FLOAT_ET:
	ep = new SpSetFloat;
	break;
      case STRING_ET:
	ep = new SpSetString;
	break;
      case LOGICAL_ET:
	ep = new SpSetLogical;
	break;
      case SET_ET:
	ep = new SpSet(name);
	break;
      case UNRES_ET:
	ep = new SpSetUnres(name);
	break;
      default:
	cerr << "Unknown type \"" << SpSetElementName[type] << "\" at " << __FILE__
	     << ", " << __LINE__ << endl;
	break;
      }
    }
    else {
      ep = new SpSetArray(type,arrayLength);
    }
    Nelements++;
    Namespace->Insert((void *)ep,name);
  }
  
  return *ep;
}

SpSet *SpSet::GetSet(char *name)
{
  SpSetElement *baseelement = GetElement(name);
  if (baseelement) {
    if (baseelement->GetType() == SET_ET) {
      return (SpSet *)baseelement;
    }
  }
  return NULL;
}

int SpSet::GetInt(char *key)
{
  SpSetElement *baseelement = GetElement(key);
  if (baseelement) {
    if (baseelement->GetType() == INT_ET) {
      return (int) *baseelement;
    }
  }
  cerr << "Integer named " << key << " not found ";
  cerr << "in set " << Name << endl;
  abort();
  return 0;
}

int SpSet::GetInt(char *key, int defaultValue, int &status)
{
  SpSetElement *baseelement = GetElement(key);
  if (baseelement) {
    if (baseelement->GetType() == INT_ET) {
      status = 0;
      return (int) *baseelement;
    }
  }
  status = 1;
  return defaultValue;
} 

double SpSet::GetFloat(char *key)
{
  SpSetElement *baseelement = GetElement(key);
  if (baseelement) {
    if (baseelement->GetType() == FLOAT_ET) {
      double tmp = (double) *baseelement;
      // return (float) *baseelement;
      return tmp;
    }
  }
  cerr << "float named " << key << " not found ";
  cerr << "in set " << Name << endl;
  abort();
  return 0.;
}

double SpSet::GetFloat(char *key, double defaultValue, int &status)
{
  SpSetElement *baseelement = GetElement(key);
  if (baseelement) {
    if (baseelement->GetType() == FLOAT_ET) {
      status = 0;
      return (double) *baseelement;
    }
  }
  status = 1;
  return defaultValue;
} 

// Note: the user owns the data behind all pointers returned,
// and the user is responsible for freeing them

char *SpSet::GetString(char *key) 
{
  SpSetElement *baseelement = GetElement(key);
  if (baseelement) {
    if (baseelement->GetType() == STRING_ET) {
      return ((char *) *baseelement); 
    }
  }
  return NULL;
}

bool SpSet::GetLogical(char *key) 
{
  SpSetElement *baseelement = GetElement(key);
  if (baseelement) {
    if (baseelement->GetType() == LOGICAL_ET) {
      return (bool)*(SpSetLogical *)baseelement;
    }
  }
  cerr << "logical named " << key << " not found ";
  cerr << "in set " << Name << endl;
  abort();
  return false;
}

bool SpSet::GetLogical(char *key, bool defaultValue,
			    int &status)
{
  SpSetElement *baseelement = GetElement(key);
  if (baseelement) {
    if (baseelement->GetType() == LOGICAL_ET) {
      status = 0;
      return (bool)*(SpSetLogical *)baseelement;
    }
  }
  status = 1;
  return defaultValue;
}


SpSetArray *SpSet::GetArray(char *key)
{
  SpSetElement *baseelement = GetElement(key);
  if (baseelement) {
    if (baseelement->GetType() == ARRAY_ET) {
      return (SpSetArray *)baseelement;
    }
  }
  return NULL;
}

int SpSet::GetEnumIntID( char *key,  char *enumString)
{
  SpSetElement *baseelement = GetElement(key);
  if (baseelement) {
    if (baseelement->GetType() == ENUM_ET) {
      SpSetEnum &enumarray = *(SpSetEnum *)baseelement;
      int retval = enumarray(enumString);
      return retval; 
    }
  }
  cerr << "enumerated type named " << key << " not found ";
  cerr << "in set " << Name << endl;
  abort();
  return 0;
}

int SpSet::GetEnumIntID( char *key,  char *enumString,
		   int defaultValue, int &status)
{
  SpSetElement *baseelement = GetElement(key);
  if (baseelement) {
    if (baseelement->GetType() == ENUM_ET) {
      status = 0;
      return (*(SpSetEnum *)baseelement)(enumString);
    }
  }
  status = 1;
  return defaultValue;
}  

int *SpSet::GetIntArray( char *key)
{
  SpSetElement *baseelement = GetElement(key);
  if (baseelement) {
    if (baseelement->GetType() == ARRAY_ET) {
      if (((SpSetArray *)baseelement)->GetArrayType() == INT_ET) {
	return (int *) *(SpSetArray *)&baseelement;
      }
    }
  }
  return (int *)NULL;
}

int *SpSet::GetIntArray( char *key, int &length)
{
  SpSetElement *baseelement = GetElement(key);
  if (baseelement) {
    if (baseelement->GetType() == ARRAY_ET) {
      if (((SpSetArray *)baseelement)->GetArrayType() == INT_ET) {
	length = ((SpSetArray *)baseelement)->GetLength();
	return (int *) *(SpSetArray *)&baseelement;
      }
    }
  }
  return (int *)NULL;
}

double *SpSet::GetFloatArray( char *key)
{
  SpSetElement *baseelement = GetElement(key);
  if (baseelement) {
    if (baseelement->GetType() == ARRAY_ET) {
      if (((SpSetArray *)baseelement)->GetArrayType() == FLOAT_ET) {
	return (double *) *(SpSetArray *)&baseelement;
      }
    }
  }
  return (double *) NULL;
}

double *SpSet::GetFloatArray( char *key, int &length)
{
  SpSetElement *baseelement = GetElement(key);
  if (baseelement) {
    if (baseelement->GetType() == ARRAY_ET) {
      if (((SpSetArray *)baseelement)->GetArrayType() == FLOAT_ET) {
	length = ((SpSetArray *)baseelement)->GetLength();
	return (double *) *(SpSetArray *)&baseelement;
      }
    }
  }
  return (double *) NULL;
}

char **SpSet::GetStringArray( char *key)
{
  SpSetElement *baseelement = GetElement(key);
  if (baseelement) {
    if (baseelement->GetType() == ARRAY_ET) {
      if (((SpSetArray *)baseelement)->GetArrayType() == STRING_ET) {
	return (char **) *(SpSetArray *)&baseelement;
      }
    }
  }
  return (char **)NULL;
}

char **SpSet::GetStringArray( char *key, int &length)
{
  SpSetElement *baseelement = GetElement(key);
  if (baseelement) {
    if (baseelement->GetType() == ARRAY_ET) {
      if (((SpSetArray *)baseelement)->GetArrayType() == STRING_ET) {
	length = ((SpSetArray *)baseelement)->GetLength();
	return (char **) *(SpSetArray *)&baseelement;
      }
    }
  }
  return (char **)NULL;
}

bool *SpSet::GetLogicalArray( char *key)
{
  SpSetElement *baseelement = GetElement(key);
  if (baseelement) {
    if (baseelement->GetType() == ARRAY_ET) {
      if (((SpSetArray *)baseelement)->GetArrayType() == LOGICAL_ET) {
	return (bool *)*(SpSetArray *)&baseelement;
      }
    }
  }
  return (bool *)NULL;
}

bool *SpSet::GetLogicalArray( char *key, int &length)
{
  SpSetElement *baseelement = GetElement(key);
  if (baseelement) {
    if (baseelement->GetType() == ARRAY_ET) {
      if (((SpSetArray *)baseelement)->GetArrayType() == LOGICAL_ET) {
	length = ((SpSetArray *)baseelement)->GetLength();
	return (bool *)*(SpSetArray *)&baseelement;
      }
    }
  }
  return (bool *)NULL;
}

void *SpSet::GetVoidPtr( char *key) 
{
  SpSetElement *baseelement = GetElement(key);
  if (baseelement) {
    if (baseelement->GetType() == VOIDPTR_ET) {
      return (void *)*(SpSetVoidPtr *)&baseelement;
    }
  }
  return (void *)NULL;
}

SpSet *SpSet::GetAncestor(char *setName)
{
  SpSet *currentset = this;
  while (currentset->Parent != NULL && 
	 strcmp(currentset->Parent->Name,setName)) {
    currentset = currentset->Parent;
  }
  return currentset->Parent;
}

SpSet *SpSet::GetAncestor(int numLevels)
{
  SpSet *currentset = this;
  while(numLevels > 0) {
    currentset = currentset->Parent;
    numLevels--;
  }
  return currentset;
}

SpSet *SpSet::Root() 
{
  SpSet *currentset = this;
  while (currentset->Parent != NULL) {
    currentset = currentset->Parent;
  }
  return currentset;
}

SpSetElement &SpSet::Element( char *name)
{
  SpSetElement *ep = GetElement(name);

  // If you find the named element, return it. Otherwise,
  // create a new, empty element of type NONE by that name

  if (ep != NULL) {
    return *ep;
  }
  else {
    SpSetElement *newelement = new SpSetNone(name,this);
    Namespace->Insert(newelement,name);
    return *newelement;
  }
}

SpSet *SpSet::GetFirstSet() 
{
  SpSetElement *element = GetFirstElement();
  while(element) {
    if (element->GetType() == SET_ET) {
      CurrentElement = element;
      return (SpSet *)element;
    }
    element = GetNextElement();
  }
  return (SpSet *)NULL;
}

SpSet *SpSet::GetNextSet()
{
  SpSetElement *element = GetNextElement();
  while(element) {
    if (element->GetType() == SET_ET) {
      CurrentElement = element;
      return (SpSet *)element;
    }
    element = GetNextElement();
  }
  return (SpSet *)NULL;
}

// Add a reference to the first element identified by
// key in the scope of this set 
void SpSet::AddRef(char *key) 
{
  Element(key,UNRES_ET);
}

void SpSet::Remove(char *name=NULL) {
  // Remove element from this set; by default, destroy self
  if (name==NULL) {
    this->~SpSet();
  }
  else {
    SpSetElement *element = (SpSetElement *)Namespace->Find(name);
    if (element != NULL) {
      // Remove from the namespace of this set
      Namespace->Remove(name);
      // Destroy the element itself
      delete element;
    }
  }
}

SpSet::~SpSet()
{
  // This destructor will get invoked for both "original" sets and 
  // set references. In the event that the set in question is just a
  // reference (i.e., created by invoking the copy constructor, which
  // does a shallow copy), don't delete all the private member data.
  // The original will take care of this.

  SpSetElement *element;
  if (this == Original) { // We are the original 
    // Delete all contents of this set
    while((element = (SpSetElement *)Namespace->RemoveFirstElement())) {
      delete element;
    }
    delete Namespace;
    
    // Update all referers to this set
    removereferences();

    // Free the name string
    if (Name) {
      free(Name);
    }
    
  }
  else {
    // I am reference. Update my original
    References->Remove((int)this);
  }
}

// This method checks if refSet is visible within the scope of 
// this set If so, a reference to set refSet is inserted in this 
// set, and a pointer to it is returned. Otherwise, NULL is returned
SpSet *SpSet::Insert(SpSet *refSet) 
{
  char *name = refSet->Name;
  SpSet *orig;
  SpSet *newreference;

  // Is refSet itself a reference? If so, use the pointer to its 
  // original 
  orig = (SpSet *)refSet->Original == refSet? refSet : (SpSet *)refSet->Original;
    

  if (inscope((SpSetElement *)orig,refSet->Name)) {
    // Make a new reference if refSet is an original set, or
    // it's parent is different than it's original
    if (refSet->Original == refSet ||
	refSet->Parent != ((SpSet *)refSet->Original)->Parent) {
      newreference = new SpSet(*refSet);
    }
    else {
      newreference = refSet;
    }
      

    // Had this element already had the name of an existing element
    // that wasn't precisely the same, inscope would have returned
    // FALSE. So, it's safe to insert this into the namespace.
    Namespace->Insert((void *)newreference,name);
    Nelements++;
    newreference->Parent = this;
    return newreference;
  }
  
  return (SpSet *)NULL   ; // Couldn't find a set by this name in scope
  //                     // (or, there was one, but not this one, or, there
  //                     // was an element of the same name, but it wasn't
  //                     // even a set)
}

void SpSet::Insert(void *value, char *key, int length=0)
{
  *(SpSetVoidPtr *)&Element(key,VOIDPTR_ET) = value;
  ((SpSetVoidPtr *)&Element(key,VOIDPTR_ET))->SetLength(length);
}

void SpSet::Insert(int value, char *key)
{
  *(SpSetInt *)&Element(key,INT_ET) = value;

}

void SpSet::Insert(char *value, char *key)
{
  *(SpSetString *)&Element(key,STRING_ET) = value;

}

void SpSet::Insert(double value, char *key)
{
  *(SpSetFloat *)&Element(key,FLOAT_ET) = value;

}

void SpSet::Insert(bool value, char *key)
{
  *(SpSetLogical *)&Element(key,LOGICAL_ET) = value;
}

// void Insert(SpSetUserDefined &value, char *key)
// {

// }

SpSetArray &SpSet::AddArray(char *key,
			     ElementType type,int arrayLength)
{
  return *(SpSetArray *) &Element(key,type,arrayLength);
}

// Form the union of this set with otherset 
SpSet *SpSet::Union(SpSet *otherSet) 
{

  SpSetElement *ep = (SpSetElement *) otherSet->Namespace->GetFirstElement();
  while(ep) {
    char *name = otherSet->Namespace->GetCurrentKeyName();

  // The strategy for treating references is to do a deep copy of
  // any references not lying under the tree of whose root is 
  // otherSet, in the event that otherSet is not a root set
  // The packwhich() method determines this for us

    if (GetElement(name)) {
      cerr << "Element named " << name << " already occurs in ";
      cerr << "target set. Omitting from merge" << endl;
    }
    else {
    
      SpSetElement *thingtocopy = ep->packwhich(name,otherSet);
      SpSetElement *newelement;
      if (thingtocopy->GetType() == SET_ET && !thingtocopy->IsRef()) {
	newelement = Embed(name);
	((SpSet *)newelement)->Union((SpSet *)thingtocopy);
      }
      else {
	
	newelement = thingtocopy->Clone();
	Nelements++;
      }
      Namespace->Insert((void *) newelement, name);
    }
    ep = (SpSetElement *) otherSet->Namespace->GetNextElement();
  }
  return this;
}

SpSetElement *SpSet::Clone()
{
  SpSet *retval;
  if (IsRef()) {
    // shallow copy of shallow copy = shallow copy of original
    retval = new SpSet(*this); 
  }
  else {
    retval = new SpSet();
    retval->Union(this);
  }
  return retval;
}

// Reconstruct an SpSet from a buffer
SpSet::SpSet(char *buffer): SpSetElement(SET_ET,true), Nelements(0),
  Name(NULL), References(NULL)
  
{
  Namespace = new SpBinaryTree;
  Namespace->SetBalancedTreeMode();
  if (!Restore(buffer)) {
    cerr << "Couldn't restore buffer. ";
    cerr << "Wasn't returned from Pack() method" << endl;
    abort();
  }
}



int SpSet::Size(SpSet *top) 
{
  // The argument to Size indicates whether Size() is being called
  // by another set. If so, the size of a reference will depend on
  // on that reference's original is packed automatically when
  // top is packed. If not (e.g., the original is in the scope of top,
  // and not embedded under top), then Size() will return the size
  // of the Original. Otherwise, the size of the reference 

  int retval = 0;
  if (!IsRef()) {
    int header = sizeof (MAGIC) + sizeof(char) + sizeof(NET_INT);
    SpSetElement *ep = GetFirstElement();
    retval += header;
    while(ep) {
      // -----------------------------------
      // Fixed length section
      retval += sizeof(char) + sizeof(NET_INT);
      // -----------------------------------

      // --------------------------------------------------
      // Variable length section
      char *name = Namespace->GetCurrentKeyName();
      retval += strlen(name)+1;
      SpSetElement *thingtopack = ep->packwhich(name,top?top:this);
      if (thingtopack->GetType() == SET_ET) {
	retval += ((SpSet *)thingtopack)->Size(top?top:this);
      }
      else {
	retval += thingtopack->Size();
      }
      // --------------------------------------------------

      ep = GetNextElement();
    }
  }
  else {
    retval = sizeof(MAGIC) + sizeof(char) + sizeof(char);
  }
  return retval;
}

int SpSet::Size() 
{
  return Size(NULL);
}

char *SpSet::Pack(int &bufferLength,char *buffer,SpSet *top)
{

  // Layout of buffer returned by the Pack() methods:
  //
  //   For an embedded set:
  //   <buffer> := <header><series of records>
  //   <header> := <magic byte><flag><number of records in series>  
  //   <series of records> := <series of records><record>
  //   <record> := <fixed length section><variable length section>
  //   <fixed length section> := <char element type><size of element data>
  //   <char element type> := (char)ElementType
  //   <variable length section> := <name string><element data>
  //   <refflag> := [0,1]
  //
  //   For a reference:
  //   <buffer> := <magic byte><flag><char element type>
  //   
  //
  // During the restore() counterpart of this Pack() method, 
  // the element data of the variable length section of each record 
  // is dispatched to the restore() method of the appropriate 
  // element type.
  // 
  // The flag is one byte. It is used to indicate metainformation
  // to the restore() method, such as whether this element is a
  // reference element.

  SpEndianNature::DetermineEndianNature();

  char flags = 0;

  bufferLength = Size(top?top:this);

  if (!buffer) {
    buffer = (char *)malloc(bufferLength);
  }

  // Am I an original, or a reference? If original, do the usual pack
  // If a reference, do only minimal pack
  if (!IsRef()) {
    // Do regular pack
    SpSetElement *ep = GetFirstElement();
    
    char *offset = buffer;
    
    // ===================================
    // Write header

    // Maybe Insure does not understand this commented-out line.  Maybe
    // it will understand the line after the commented out line, and
    // thereby not give a read_uninitialized_memory error.

    //memset(offset,SpSet::MAGIC,sizeof(SpSet::MAGIC));
    offset[0] = SpSet::MAGIC;

    offset += sizeof(SpSet::MAGIC);
    
    memset(offset,flags,sizeof(flags));
    offset += sizeof(flags);
    
    NET_INT nelements = Nelements;
    
    memcpy(offset,&nelements,sizeof(nelements));
    offset += sizeof(nelements);
    // ===================================
    
    // ===================================
    // Write series of records
    while(ep) {

      // -----------------------------------------------------
      // Fixed length section

      char *name = Namespace->GetCurrentKeyName();
      SpSetElement *thingtopack = ep->packwhich(name,top?top:this);

      char etype;
      // if ElementType is really a set reference, record this
      // in the buffer as UNRES_ET  

      if (ep == thingtopack && ep->IsRef()) {
	//      if (ep != thingtopack) {
	etype = (char) UNRES_ET;
      }
      else {
	etype = (char) thingtopack->GetType();
      }
      memcpy(offset,&etype,sizeof(etype));
      offset += sizeof(etype);


      
      NET_INT datasize;
      if (etype == (char) SET_ET) {
	datasize = ((SpSet *)thingtopack)->Size(top?top:this);
      }
      else {
	datasize = thingtopack->Size();
      }
      memcpy(offset,&datasize,sizeof(datasize));
      offset += sizeof(datasize);
      // ----------------------------------------------------
      
      // ----------------------------------------------------
      // Variable length section 
      memcpy(offset,name,1+strlen(name));
      offset += 1+strlen(name);
      if (etype == (char) SET_ET) {
	((SpSet *)thingtopack)->Pack(offset,top?top:this);
      }
      else {
	thingtopack->Pack(offset);
      }
      // ----------------------------------------------------

      offset += datasize;
      ep = GetNextElement();
    }
    // ===================================
    
  }
  else {
    flags |= SPSETREFFLAG;
    char *offset = buffer;
    memset(offset,MAGIC,sizeof(MAGIC));
    offset += sizeof(MAGIC);    
    memset(offset,flags,sizeof(flags));
    offset += sizeof(flags);
    char etype = (char) SET_ET;
    memset(offset,etype,sizeof(etype));
  }
  return buffer;
}




// This method is the complement to the previous Pack() methods
// Restores an SpSet object from a buffer returned by Pack() 
SpSetElement *SpSet::Restore(char *buffer)
{

  SpEndianNature::DetermineEndianNature();

  char *offset = buffer;
  if (*buffer != MAGIC) {
    cerr << "Buffer wasn't returned by Pack(), or is corrupt" << endl;
    return NULL;
  }

  offset += sizeof(MAGIC);
  char flags;
  memcpy(&flags,offset,sizeof(flags));
  offset += sizeof(flags);

  if (flags & SPSETREFFLAG) {
    cerr << "Buffer contains a reference, not an original set!" << endl;
    cerr << "This Restore method should never return  a reference,";
    cerr << endl;
    return NULL;
  }

  
  NET_INT nelements;
  memcpy(&nelements,offset,sizeof(NET_INT));
  offset += sizeof(NET_INT);
  int n = nelements;
  for (int k=0; k<n; k++) {
    // -----------------------------------
    // Read fixed length section
    char et;
    memcpy(&et,offset,sizeof(et));
    offset += sizeof(et);
    NET_INT DATASIZE;
    memcpy(&DATASIZE,offset,sizeof(DATASIZE));
    offset += sizeof(DATASIZE);
    int datasize = DATASIZE;
    // -----------------------------------

    // -----------------------------------
    // Read variable length section
    char *name = offset;
    
    offset += 1+strlen(name);
    SpSetElement *ep;
    ElementType etype = (ElementType) et;
    int wasdefault;
    wasdefault = 0;
    switch (etype) {
    case SET_ET:
      ep = new SpSet(name,this);
      break;
    case INT_ET:
      ep = new SpSetInt;
      break;
    case FLOAT_ET:
      ep = new SpSetFloat;
      break;
    case STRING_ET:
      ep = new SpSetString;
      break;
    case LOGICAL_ET:
      ep = new SpSetLogical;
      break;
    case ARRAY_ET:
      ep = new SpSetArray;
      break;
    case ENUM_ET:
      ep = new SpSetEnum;
      break;
    case VOIDPTR_ET:
      ep = new SpSetVoidPtr;
      break;
    case UNRES_ET:
      ep = new SpSetUnres;
      break;
    default:
      cerr << "SpSet::Restore(): Don't know how to restore element ";
      cerr << "of type " << SpSetElementName[(int)etype] << endl;
      wasdefault = 1;
    }
    if (!wasdefault) {
      ep->Restore(offset);
      if (etype != SET_ET) {
	Namespace->Insert((void *)ep,name);
      }
    }
    // -----------------------------------
    offset += datasize;
  }
  Nelements = n;
  return this;
}

// Protected constructor called by Embed
SpSet::SpSet(char *name, SpSet *parent): SpSetElement(SET_ET), 
  Nelements(0), Parent(parent), References(NULL)
{
  if (name) {
    Name = strdup(name);
    parent->Namespace->Insert((void *)this, Name);
    parent->Nelements++;
  }
  else {
    cerr << "Name of set must be non-null for all sets below ";
    cerr << "the root set of a tree!" << endl;
    abort();
  }
  
  // Create a new binary tree for Namespace. 
  Namespace = new SpBinaryTree;
  Namespace->SetBalancedTreeMode();
}

// protected method for returning a reference to the first
// element named key in this set's scope
SpSetElement *SpSet::createreference(char *key)
{
  SpSet *forbear = Parent;
  SpSetElement *candidate;

  SpSetElement *ep = NULL;
  
  while(forbear != NULL) {
    if ((candidate = (SpSetElement *)(forbear->Namespace->Find(key)))) {
      switch(candidate->GetType()) {
      case INT_ET:
	ep = new SpSetInt(*(SpSetInt *)candidate);
	break;
      case FLOAT_ET:
	ep = new SpSetFloat(*(SpSetFloat *)candidate);
	break;
      case STRING_ET:
	ep = new SpSetString(*(SpSetString *)candidate);
	break;
      case LOGICAL_ET:
	ep = new SpSetLogical(*(SpSetLogical *)candidate);
	break;
      case ARRAY_ET:
	ep = new SpSetArray(*(SpSetArray *)candidate);
	break;
      case ENUM_ET:
	ep = new SpSetEnum(*(SpSetEnum *)candidate);
	break;
      case SET_ET:
	ep = new SpSet(*(SpSet *)candidate);
	break;
      case VOIDPTR_ET:
	ep = new SpSet(*(SpSet *)candidate);;
	break;
      default:
	cerr << "SpSet::createreference: Don't know how ";
	cerr << "to create reference to element of type ";
	cerr << SpSetElementName[(int)candidate->GetType()];
	break;
      }
      if (ep) return ep;
    }
    forbear = forbear->GetParent();
  }
  return ep;
}

// Finds the element named key in this set's scope
SpSetElement *SpSet::find_in_scope(char *key)
{
  SpSet *forbear = Parent;
  SpSetElement *candidate;

  while(forbear != NULL) {
    if ((candidate = (SpSetElement *)(forbear->Namespace->Find(key)))) {
      return candidate;
    }
    forbear = forbear->GetParent();
  }
  return NULL;
}


// Upon the death of this set, remove any references to this set
// in other sets
void SpSet::removereferences() 
{
  SpSet *referer;
  if (References) {
    while((referer = (SpSet *)References->RemoveFirstElement())) {
      referer->Parent->Remove(referer->Name);
    }
    delete References;
  }
}

bool SpSet::inscope(SpSetElement *element, char *name)
{
  SpSet *forbear = Parent;
  SpSetElement *candidate;

  while(forbear != NULL) {
    if ((candidate = forbear->GetElement(name))) {
      if (candidate == element) {
	return true; // The two SpSetElements point to same place 
	//          // in memory 
      } 
    }
    forbear = forbear->GetParent();
  }
  return false;
}

// The following two methods allow one to iterate through all
// elements in the set. They skip any unresolved references.
// Perhaps they should give a warning for such unresolved 
// references. 
SpSetElement *SpSet::GetFirstElement()
{
  SpSetElement *retval;
  SpSetElement *ep = (SpSetElement *) Namespace->GetFirstElement();
  retval = ep;
  if (ep) {
    if (ep->GetType() == UNRES_ET) {
      char *name = strdup(Namespace->GetCurrentKeyName());
      SpSetElement *nep = createreference(name);
      if (nep) {
	delete ep;
	Namespace->Remove(name);
	Namespace->Insert(nep,name);
	free(name);
	retval = nep;
      }
      else { // Skip references that can't be resolved
	retval = GetNextElement();
      }
    }
  }
  return retval;
}

SpSetElement *SpSet::GetNextElement()
{
  SpSetElement *retval;
  SpSetElement *ep = (SpSetElement *) Namespace->GetNextElement();
  retval = ep;
  if (ep) {
    if (ep->GetType() == UNRES_ET) {
      char *name = strdup(Namespace->GetCurrentKeyName());
      SpSetElement *nep = createreference(name);
      if (nep) {
	delete ep;
	Namespace->Remove(name);
	Namespace->Insert(nep,name);
	free(name);
	retval = nep;
      }
      else { // Skip references that can't be resolved
	retval = GetNextElement();
      }
    }
  }

  return retval;
} 
 
char *SpSet::GetCurrentKeyName()
{
  return Namespace->GetCurrentKeyName();
}

const char SpSet::MAGIC = 'S';
