//  Jim Brutocao
//  SpDataParser
//  9/13/98
//  Draft Number: 6
//
//
//
// TERMS:
//
// 1. data parser (SpDataParser) - An object that converts a *.par file to a
//                                 recursive data structure called a set.
//                                 This set is given the name "Top Level Set".
// 2. set (SpSet) - The fundamental recursive data structure, containing one
//                  or more elements.
// 3. element - A set, integer, integer array, real, real array, string,
//              string array, logical, logical array, enum, or user define.
//
// AN EXAMPLE *.par FILE:
//
// keyword Triple (3)
//
// Set_1
// {
//   Set_4
//   {
//     inherit Set_2 { int_1 6, Set_3.int_2 5 }
//     int_3 6
//   }
//   Set_5
//   {
//     Set_6
//     {
//       Set_4
//       Set_7
//       {
//         string_1 "hello"
//       }
//       Triple one two three
//       Triple "one" "two" "quotes are optional and allow for white space"
//     }
//     int_1 4
//     integer int_2 5 // explicit typing is optional
//     real_1 6.4
//     logical_1 T
//     int_array { 1 3 6 8 }
//     enum_1 { Green, Blue, Yellow }
//   }
// }
//
// Set_2
// {
//   int_1 4
//   Set_3
//   {
//     int_2 5
//   }
// }
//
// *.par FILE LANGUAGE OVERVIEW:
//  ----------------------------
//
//
//  SETS
//
//  Sets are the fundamental unit of data collection in *.par
//  files. In fact, a *.par file is simply a succession of keyword
//  declarations followed by succession of sets. Set syntax (see Set_*
//  above) consists of an identifier representing the set name
//  followed by a braced set of elements (or followed by nothing if a
//  reference). Elements that comprise sets include nested sets,
//  inherited sets, referenced sets, integers, integer arrays, floats,
//  float arrays, logicals (booleans), logical arrays, strings, string
//  arrays, enums, and user defines. Any element type, which includes
//  a name (or user-defined keyword) and a value, including a set, may
//  be dynamically added to any set or nested set, by simply
//  calling methods on that set.
//
//  ELEMENTS
//
//  Element syntax is a follows: the reserved word "inherit" if it is
//  an inherited set, followed by an optional element type (one of the
//  following reserved words: "set", "int", "float", "logical",
//  "string", "enum", and "user-define"), followed by an identifier
//  representing the element's name (or for a user-define, the
//  user-defined keyword), followed by the element's value, which may
//  be a single or a compound value. The element type may be omitted
//  (i.e., it is implicit) because the syntax for the value of each
//  element type is disambiguous. An integer value is a string of
//  numbers without a decimal point, a float value is a string of
//  numbers with a decimal point, a logical is an unquoted 'T' (true)
//  or 'F' (false), a string is a double-quoted string of characters,
//  an enum is a braced set of white-space-delimited unquoted strings
//  of characters, a user-define is a user-defined keyword followed by
//  a user-defined number of white-space-delimited unquoted strings of
//  characters, and a set is a braced collection of
//  white-space-delimited elements. User-defined keywords are
//  specified at the top of a *.par file outside of any sets with the
//  reserved word "keyword" followed by a parenthesized integer
//  representing the number of arguments for that user-defined
//  keyword. The data parser will type check each user-defined element
//  by assuring that each keyword is consistently followed by its
//  declared number of arguments. For expressing arrays (applicable to
//  ints, floats, logicals, and strings), simply follow the name of
//  the element by a braced white-space-delimited succession of values.
//
//  SET INHERITANCE
//
//  A set may inherit another set's elements and their values
//  (including all nested sets) by using the reserved keyword
//  "inherit" followed by the set's name to be inherited. As in
//  conventional object orientation, the elements in the inherited
//  set, apart from the inherited set in which they are originally
//  contained, are directly contained in the inheritor. This is
//  opposed to set references, where the referenced set itself is
//  contained in the referencing set (see next section). In addition,
//  any inherited element's value may be overridden by following an
//  inherited set's name by a braced section of comma-delimited
//  inherited value names followed by their overridden values. For
//  overridding nested values, use the "." delimiting operator to
//  specify the set path (see Set_4 contained in Set_1 the above
//  example). When new elements are dynamically added to the inherited
//  set, the inheritor will also have those elements and their values
//  added.
//
//  SET REFERENCES
//
//  Names of sets that match the names of other sets defined in the
//  same *.par file may be added to sets as contained set
//  references. All nested types and values from the referenced set
//  are included in the reference. Syntactically, this requires simply
//  including the referenced set's name in the referencing set without
//  a following braced section. When new elements are dynamically
//  added to the referenced set, the referencing set will also have
//  those elements and their values added. The difference between
//  referencing and inheriting sets (see above section) are twofold:
//  1) values of elements in referenced sets cannot be overridden,
//  while values of elements in inherited sets can, and 2) referenced
//  sets are included as containers of referenced elements in the
//  referencing set, while only elements contained in inherited sets
//  are included in inheriting sets.
//
//
// SOFTWARE DESIGN OVERVIEW:
// -------------------------
//
//
// There are three basic classes in this file: SpDataParser, SpSet,
// and SpSetElement (a base class representing many
// children). SpDataParser is responsible for reading a *.par file and
// converting the resulting stream into a recursive data structure,
// called SpSet, that contains of all of the data (nested sets) in the
// *.par file. SpSet simply a list of elements, all of which inherit
// from SpSetElement, with methods that allow users to extract the
// data contained in those elements. What makes SpSet recursive is
// that SpSet is a type of element, so SpSets can contain SpSets. The
// types of SpSetElements are: SpSetInt, SpSetArray,
// SpSetEnum, SpSetFloat, SpSetLogical, SpSetString, and SpSetUserDefined.
// 
// 


// declarations:


#ifndef _SpSet_H
#define _SpSet_H

#include "SpSetElement.H"
#include "SpBinaryTree.H"

#define SPSETREFFLAG 1

class SpSet: public SpSetElement
{
  friend class SpSetElement;
  friend class SpSetNone;
public:
  
  // Constructor for an original copy of a set. The set is a root 
  // set, the top of an independent set tree.
  SpSet(): SpSetElement(SET_ET,true), Nelements(0), Name(NULL), Parent(NULL),
    References(NULL) {
    Namespace = new SpBinaryTree;
    Namespace->SetBalancedTreeMode();
  }
    
  // Constructor for a "set reference." Accordingly, a copy
  // constructor (note that Original is set to address of the
  // set this one refers to)
  SpSet(SpSet &orig): SpSetElement(SET_ET) {
    *this = *(SpSet *)orig.Original;    // shallow copy
    Original = orig.Original;
    if (!orig.References) {
      orig.References = new SpBinaryTree;
      orig.References->SetBalancedTreeMode();
      References = orig.References;
    }
    References->Insert((void *)this,(int) this);
  }

  // Reconstruct a set from a character buffer returned from the
  // pack() method
  SpSet(char *buffer); 

  // Preferred method for embedding one set inside another
  SpSet *Embed(char *name); 



  // "Accessing" methods



  // The first of each of the following methods will return the value  // identified by key if present, and set the status flag to nonzero  // and return the default if not. The second of each of these 
  // methods will return the value identified by key if available, 
  // otherwise, it prints a message to stderr and aborts.

  int GetInt( char *key, int defaultValue, int &status);
  int GetInt( char *key);

  double GetFloat( char *key, double defaultValue, int &status);
  double GetFloat( char *key);

  // No default argument method. Returns NULL if no string 
  // corresponding to "key" found
  char *GetString( char *key);

  bool GetLogical( char *key);
  bool GetLogical( char *key, bool defaultValue,
		       int &status);

  SpSetArray *GetArray(char *key);

  // These methods return a pointer to private data for the 
  // sake of performance (TSK TSK!). DON'T GO FREEING IT!
  // The second of each of them returns also the length of
  // of the array. If no array corresponding to "key" exists,
  // each returns NULL 
  int *GetIntArray( char *key);
  int *GetIntArray( char *key, int &length);

  double *GetFloatArray( char *key);
  double *GetFloatArray( char *key, int &length);

  char **GetStringArray( char *key);
  char **GetStringArray( char *key, int &length);

  bool *GetLogicalArray( char *key);
  bool *GetLogicalArray( char *key, int &length);

  void *GetVoidPtr( char *key);

  
  
  // The following method is provided for  compatibility with other 
  // SPEEDES container classes. It can be used to access VOIDPTR
  // SpSetElements, in addition to the usual Get* methods. For all 
  // other SpSetElements, only the Get* methods are provided
  void *Find( char *key) {
    return GetVoidPtr(key);
  }    

  // Finds the element named key anywhere in the scope of this
  // set. Returns NULL if not found
  SpSetElement *find_in_scope(char *key);

  // Method for retrieving an embedded or referenced set
  SpSet *GetSet(char *name);


  // Methods for packing a set and all its dependencies
  // into a buffer suitable for sending in messages

  // If no buffer is passed, these method allocates and returns a 
  // pointer to the allocated space. bufferLength is set to the 
  // number of bytes allocated. Generally, you should never need
  // to pass an argument for top. It is only necessary if you
  // want the Pack() method to leave any references visible in
  // the scope of this set (which are therefore not defined in
  // this set but somewhere under top) undefined. The first method
  // returns the length (if buffer was allocated), the second doesn't
  char *Pack();
  char *Pack(int &bufferLength);
  char *Pack(int &bufferLength,char *buffer);
  char *Pack(char *buffer, SpSet *top);
  char *Pack(char *buffer);
  char *Pack(int &bufferLength,char *buffer, SpSet *top);






  // Gives the size in bytes of the buffer needed to pack this set
  // Can be used in conjunction witht the second pack() method
  // above, e.g. set->pack((char *)malloc(set->Size())); Gererally,
  // you should never need to pass an argument for top. It is only
  // necessary if you want the size to be calculated based on
  // packing up only unresolved references to elements not defined
  // under this set, but which are defined somewhere under top.
  int Size(SpSet *top);
  int Size();

  // This method gets called by SpSet(char *) to restore a set
  // from a buffer returned by the Pack() methods 
  SpSetElement *Restore(char *buffer);

  // "Editing" methods
  ~SpSet();

  void Insert(void *value               , char *key, int length);
  void Insert(int value                 , char *key);


  void Insert(char *value               , char *key);
  void Insert(double value               , char *key);

  void Insert(bool value                 , char *key);

  void Insert(SpSetUserDefined & /*value*/            , char * /*name*/) {}

  // insert a reference to a set visible within the scope of
  // this set
  SpSet *Insert(SpSet *value);                

  // Allocate an array element 
  SpSetArray &AddArray(char *key,ElementType type, int arrayLength);

  // Form the union of this set with otherSet
  SpSet *Union(SpSet *otherSet);
  
  // Clone this set as a new top-level set (references in
  // the scope of this set will get deep-copied)
  SpSetElement *Clone();

  // Remove the element named key. If NULL, destroy self
  void Remove(char *key);

  // Add a reference to the first element identified by key
  // in the scope of this set.
  void AddRef(char *key);
  

  // methods for getting pointers to ancestor sets:
  virtual SpSet *GetParent() { return Parent; }
  SpSet *GetAncestor(char *setName);
  SpSet *GetAncestor(int numLevels = 1);
  SpSet *Root(); // Returns the toplevel set in this tree



  // NOTE: Basic rule for looping through either sets, user-defines,
  // or all elements: only one loop can be operating at a time. This
  // means you cannot be calling GetFirstSet(), GetNextSet(),
  // etc. intermixed with calling GoToFirstUserDefine(),
  // GoToNextUserDefine(), etc. or intermixed with calling
  // GetFirstElement(), GetNextElement(), etc. This limitation will
  // be alleviated when we develop iterators for this class.

  // to loop through (and get the pointers to) all sets within this set:
  int GetNumSets();
  SpSet *GetFirstSet();
  SpSet *GetNextSet();

  // The following methods are for looping through a (or all)
  // user-define(s). keyword and argPosition are INPUT parameters.
  // NOTE: GetNextUserDefine() methods get the next instance of a
  // user-defined element of the keyword specified in the most recent
  // call to GoToFirstUserDefine(). GetNextUserDefine() does NOT get
  // the next arg of the currentElement.
  int    GetNumUserDefines(char *keyword = NULL); // NULL = all user defines
  void GoToFirstUserDefine(char *keyword = NULL); // NULL = all user defines
  void  GoToNextUserDefine();
  char   *GetUserDefineArg(int argNum);
  // methods for when the keyword or number of arguments are NOT known
  char *GetUserDefineKeyword();
  int   GetUserDefineNumArgs();

  // gets the enum element integer ID for which the string equivalent
  // of one of the enum elements is known
  int GetEnumIntID(char *key, char *enumString, 
		   int defaultValue, int &status);
  int GetEnumIntID(char *key, char *enumString);
  // gets the string equivalent for which the enum element integer ID
  // of one of the enum elements is known
  char *GetEnumString(char *name, int intID);

  // the following methods are for the unusual case where the
  // application does not know specifically what data is contained in
  // the *.file.  This is accomplished by simply looping through all
  // the elements contained in this set and having the element type,
  // name, and arrayLengh returned.
  //
  // SPECIAL CASES:
  // 1.) If the element is not an array, arrayLength
  //     returns 1.
  // 2.) If the element is user-defined, name returns the
  //     keyword, and arrayLength returns the number of arguments, and
  // 3.) If the element is an enum, arrayLength returns the number of
  //     enumerations.
  int GetNumElements() {return Nelements;};


protected:

  SpSetElement *GetFirstElement();
  SpSetElement *GetNextElement();
  char *GetCurrentKeyName();

  SpSetElement &Element(char *name); // Returns element named
    //                                                                         // name
  SpSetElement &Element(char *name,ElementType, int arrayLength=0); // Returns
  //                                                                                                       //  element
  //                                                                                                       //  named
  //                                                                                                       // name of
  //                                                                                                       // ElementType
  SpSetElement *GetElement(char *name);

  // Tells us whether element identified by key is in the 
  // scope of this set or not 
  bool inscope(SpSetElement *element, char *key);

  // Protected constructor (called by Embed(const char *))
  SpSet( char *name, SpSet *parent);

  void removereferences(); // Upon the death of this set,
                              // remove any references to this set
                              // in other sets
  
  // Returns pointer to a new SpSetElement referencing the
  // first element identified by key in the scope of this set
  // or NULL if no such key lies in this set's scope
  SpSetElement *createreference(char *key);

private:
  int            Nelements;   // Total number of elements in this set
  char          *Name;         // this set's name
  SpSet *Parent;             // element  in which this set is contained
  SpBinaryTree   *Namespace;      // elements contained in this set

  // We keep around a table of all references to this set, so they 
  // can be destroyed in case the original is destroyed  
  SpBinaryTree   *References;     



  SpSetElement *CurrentElement;  // used when looping through contained
                                 // elements. This includes sets, user
                                 // defines, or all elements

  char    *CurrentKeyword; // used when looping through user defines
  static const char MAGIC;
};



inline char *SpSet::Pack(char *buffer,SpSet *top)
{
  int bufferlength;
  return Pack(bufferlength,buffer,top);
}

inline char *SpSet::Pack(char *buffer)
{ 
  int bufferlength;
  return Pack(bufferlength,buffer,NULL);
}

inline char *SpSet::Pack(int &bufferLength,char *buffer)
{
  return Pack(bufferLength,buffer,NULL);
}

inline char *SpSet::Pack(int &bufferLength)
{
  return Pack(bufferLength,NULL,NULL);
}

inline char *SpSet::Pack()
{
  int bufferLength;
  return Pack(bufferLength,NULL,NULL);
}



#endif // #ifndef _SpSet_H
