// SpSetElement.H
// Defines the basic elements that can be contained in an SpSet

#ifndef _SpSetElement_H
#define _SpSetElement_H

#include "SpIostream.H"
#include "SpString.H"
#include "SpStdlib.H"
#include "SpAssert.H"

#if !defined(__GNUC__) && !defined(__SPEEDES_NT__)
#define bool short
#define false (short) 0
#define true (short) 1
#endif  // #ifndef __GNUC__


enum ElementType
{
  NONE_ET,
  UNRES_ET,
  SET_ET,
  USER_DEFINED_ET,
  ENUM_ET,
  INT_ET,
  FLOAT_ET,
  STRING_ET,
  LOGICAL_ET,
  ARRAY_ET,
  VOIDPTR_ET,
  N_ET
};

#define __SpSet_Internal__

#ifdef __SpSet_Internal__
static char *SpSetElementName[N_ET] = {
  "none",
  "unresolved reference",
  "set",
  "user-defined",
  "enum",
  "int",
  "float",
  "string",
  "logical",
  "array",
  "void ptr"
};
#endif // __ifdef __SpSet_Internal__

typedef enum ElementType ElementType;

class SpSet;
class SpSetUserDefined;

class SpSetElement
{
  friend class SpSet;
public:
  SpSetElement(ElementType type, SpSetElement *original): 
    Original(original), Init(true), Type(type)  {}
  SpSetElement(ElementType type=NONE_ET, bool init=true): Original(this), 
      Init(init), Type(type)
      {}
  virtual ~SpSetElement(void);
  virtual operator int() {
    cerr << "Attempt to cast non-integer element to integer!" << endl;
    abort();
    return 0;
  }
  virtual operator double() {
    cerr << "Attempt to cast non-float element to float!" << endl;
    abort();
    return 0;
  }
  virtual operator char*() {
    cerr << "Attempt to cast non-string element to string!" << endl;
    abort();
    return NULL;
  }

  virtual operator bool () {
    cerr << "Attempt to cast non-bool element to bool!" << endl;
    abort();
    return false;
  }
//  virtual operator int *() {
//    cerr << "Attempt to cast non-int array to int *!" << endl;
//    abort();
//    return NULL;
//  }
//  virtual operator double *() {
//    cerr << "Attempt to cast non float array to float *!" << endl;
//    abort();
//    return NULL;
//  }
//  virtual operator char **() {
//    cerr << "Attempt to cast non-string array to char **!" << endl;
//    abort();
//    return NULL;
//  }
//   virtual operator bool *() {
//     cerr << "Attempt to cast non bool array to bool *!" << endl;
//     abort();
//     return NULL;
//   }
  virtual operator void *() {
    cerr << "Attempt to cast non void pointer to void *!" << endl;
    abort();
    return NULL;
  }

  virtual SpSetElement &Element(char *) {
    cerr << "Attempt to extract an element from non-set ";
    cerr << "element!" << endl;
    abort();
    return *(new SpSetElement);
  }

  virtual char *Pack(char *) {
    cerr << "No pack() method implemented for element of type ";
    cerr << SpSetElementName[(int)Type] << endl;
    return NULL;
  }

  virtual SpSetElement *Restore(char *) {
    if (Type != UNRES_ET) {
      cerr << "No Restore() method implemented for element of type ";
      cerr << SpSetElementName[(int)Type] << endl;
    }
    return NULL;
  }

  virtual int Size() {
    cerr << "No Size() method implemented for element of type ";
    cerr << SpSetElementName[(int)Type] << endl;
    return 0;
  }

  // Make identical copy of this element
  virtual SpSetElement *Clone() {
    cerr << "SpSetElement: No Clone() method implemented for element of type ";
    cerr << SpSetElementName[(int)Type] << endl;
    return 0;
  }

//    virtual SpSetElement &operator=(int rval) {
//      cerr << "Attempt to assign integer to non-int element!" << endl;
//      abort();
//      return *(new SpSetElement);
//    }

//   virtual SpSetElement &operator=(char *rval) {
//     cerr << "Attempt to assign  char * to non-string element!" << endl;
//     abort();
//     return *(new SpSetElement);
//   }
//   virtual SpSetElement &operator=(double rval) {
//     cerr << "Attempt to assign float to non-float element!" << endl;
//     abort();
//     return *(new SpSetElement);
//   }
//   virtual SpSetElement &operator=(void *rval) {
//     cerr << "Attempt to assign void * to non-voidptr element!" << endl;
//     abort();
//     return *(new SpSetElement);
//   }

//   virtual SpSetElement &operator=(SpSetElement &element) {
//     cerr << "No method defined for assigning an element to this ";
//     cerr << "object!" << endl;
//     abort();
//     return *(new SpSetElement);
//   } 

  virtual SpSetElement &operator[](int) {
    cerr << "Attempted array dereference on non-array ";
    cerr << "element!" << endl;
    abort();
    return *(new SpSetElement);
  }
  
  virtual void Remove(char *) {
    delete this;
  }
  ElementType GetType() { return Type; }

  // A pack method for packing up each type of element into a
  // character buffer
  virtual char *pack(char *) {return NULL;} // take this out and make it pure
  virtual char *pack(int &) {return NULL;} // take this out and make it pure

  // Indicates whether this element is reference
  virtual inline bool IsRef() {
    bool retval = (bool)(Original != this);
    return retval;
  }

protected:
  SpSetElement *Original; // this, unless reference
  bool Init; // Indicates whether this element has been initialized

  // Returns correct element to pack, this element or Original
  SpSetElement *packwhich(char *name,SpSet *top);

private:
  ElementType Type;
};

class SpSetArray;

class SpSetUserDefined: public SpSetElement
{
public:
  SpSetUserDefined(const char *keyword, SpSetElement *argList, 
		   int numArgs): SpSetElement(USER_DEFINED_ET,true), 
    ArgList(argList), NumArgs(numArgs),Next(NULL) {
      if (keyword) {
	Keyword = strdup(keyword);
      }
  }
  SpSetUserDefined(const char * /*keyword*/, SpSetArray * /*array*/): 
    SpSetElement(USER_DEFINED_ET,true) {}
  
  SpSetElement &operator[](int index) {
    assert(index >=0 && index<NumArgs);
    return ArgList[index];
  }
  
  SpSetUserDefined *Append(SpSetUserDefined *next) {
    Next = next;
    next->Next = NULL;
    return Next;
  }
  
  SpSetUserDefined &GetNext() {
    return *Next;
  }

  char *pack (int &) {
    cerr << "Pack method not yet implemented for user defines";
    cerr << endl;
    abort();
    return NULL;
  }

  
  char *pack (char *) {
    cerr << "Pack method not yet implemented for user defines";
    cerr << endl;
    abort();
    return NULL;
  }

private:
  SpSetElement *ArgList;
  int NumArgs;
  SpSetUserDefined *Next;
  char *Keyword;
};


class SpSetInt: public SpSetElement
{
public:
  SpSetInt(): SpSetElement(INT_ET,false) {};
  SpSetInt(int val):
    SpSetElement(INT_ET,true), Val(val) {}
  SpSetInt(SpSetInt &orig); // Create reference to orig

  operator int() {
    return Val;
  }
  SpSetElement &operator=(int rval) {
    Init = true;
    Val = rval;
    return *this;
  }
  char *Pack(char *buffer);
  SpSetElement *Restore(char *buffer);
  int Size();
  SpSetElement *Clone();

private:
  int Val;
};

class SpSetEnum: public SpSetElement
{
public:
  SpSetEnum(char *, char **labelarray, int numlabels):
    SpSetElement(ENUM_ET,true), Label(labelarray), Nlabels(numlabels) {}
  SpSetEnum(SpSetEnum &orig); // Creates reference element
  SpSetEnum(): SpSetElement(ENUM_ET,false), Label(NULL), Nlabels(0) {}
  ~SpSetEnum();
  
  int operator()(const char *string) const; // Return index of  string
  char *Pack(char *buffer);

private:
  char **Label;
  int Nlabels;
};

class SpSetFloat: public SpSetElement
{
public:
  SpSetFloat(): SpSetElement(FLOAT_ET,false) {}
  SpSetFloat(SpSetFloat &orig); // Create reference to orig
  SpSetFloat(double val): SpSetElement(FLOAT_ET,true),
    Val(val) {}
  operator double() {
    double retval;
    if (Init) {
      retval = Val;
    }
    else {
      cerr << "Attempt to use unitialized SpSetFloat " << endl;
      abort();
    }
    return retval;
  }
  SpSetElement &operator=(double rval)
    {
      Val = rval;
      Init = true;
      return *this;
    }
  char *Pack(char *buffer);
  int Size();
  SpSetElement *Restore(char *buffer);
  SpSetElement *Clone();

private:
  double Val;
};

class SpSetLogical: public SpSetElement
{
public:
  SpSetLogical(): SpSetElement(LOGICAL_ET,false) {}
  SpSetLogical(SpSetLogical &orig); // Create reference to orig
  SpSetLogical(bool val):
    SpSetElement(LOGICAL_ET,true), Val(val) {}
  operator int() {  
    return Val==true?1:0;
  }
  operator bool() {
    return Val;
  }
  SpSetElement &operator=(bool l) 
    {
      Init = true;
      Val = l;
      return *this;
    };
  
  char *Pack(char *buffer);
  int Size();
  SpSetElement *Restore(char *buffer);
  SpSetElement *Clone();

private:
  bool Val;
};

class SpSetString: public SpSetElement
{
public:
  SpSetString(char *val);
  SpSetString(SpSetString &orig); // Create reference to orig
  SpSetString(): SpSetElement(STRING_ET,false),Val(NULL) {}

  ~SpSetString();

  virtual SpSetElement &operator=(char *rval) {
    Init = true;
    if (Val != NULL) free(Val);
    Val = strdup(rval);
    return *this;
  }
  operator char *();

  virtual char *Pack(char *buffer);
  int Size();
  SpSetElement *Restore(char *buffer);
  SpSetElement *Clone();

private:
  char *Val;
};

class SpSetArray: public SpSetElement
{
public: 
  SpSetArray(): SpSetElement(ARRAY_ET,false),Val(NULL), ArrayLength(0),
    TotalLength(0) {}
  SpSetArray(SpSetArray &orig); // Create reference to orig
  SpSetArray(ElementType type, int arrayLength);
  ~SpSetArray();
  ElementType GetArrayType() {
    return ArrayType;
  }

  int GetLength() { return ArrayLength; }

  // These operators give unfettered access to private data for
  // the sake of performance. Do NOT attempt to free the pointer
  // returned!

  operator int *();
  operator double *();
  operator char **();
  operator bool *();
  SpSetElement &operator[](int); 

  char *Pack(char *buffer);
  int Size();
  SpSetElement *Restore(char *buffer);
  SpSetArray *Push(int);
  SpSetArray *Push(bool);
  SpSetArray *Push(double);
  SpSetArray *Push(char *);
  SpSetArray *Push(SpSetElement &element);
//   SpSetElement &operator[](int index) {
//     SpSetElement *retval;
//     if (!(ArrayType == NONE_ET)) {
//       cerr << "SpSetArray: cannot apply operator[] ";
//       cerr << "to this type of array" << endl;
//       abort();
//     }
//     if (index < ArrayLength) {
//       retval = (SpSetElement *)((SpSetElement **) Val)[index];
//     }
//     else {
//       cerr << "SpSetArray: out-of-bounds array dereference" << endl;
//       abort();
//     }
//     return *retval;
//   }
    SpSetElement *Clone();

private:
  void *Val;
  int ArrayLength;
  int TotalLength;
  enum ElementType ArrayType;  
};

class SpSetVoidPtr: public SpSetElement 
{
public:
  SpSetVoidPtr(void *val): SpSetElement(VOIDPTR_ET), Val(val), Length(0) {}
  SpSetVoidPtr(void *val,int length);
  SpSetVoidPtr(SpSetVoidPtr &orig); // Create reference to orig
  SpSetVoidPtr(): SpSetElement(VOIDPTR_ET,false), Val(NULL), Length(0) {}
  operator void *() {
    return Val;
  }
  ~SpSetVoidPtr();
  SpSetElement &operator=(void *val) {
    Val = val;
    return *this;
  }
  void SetLength(int length) {Length = length;}
  int GetLength() {return Length;}
  char *Pack(char *buffer);
  int Size();
  SpSetElement *Restore(char *buffer);

private:
  void *Val;
  int Length;
};


// Objects of this class should never be created by users
// They will only be used by the SpSet class. 
class SpSetUnres: public SpSetElement
{
public:
  SpSetUnres(char *key): SpSetElement(UNRES_ET) {Name = strdup(key);}
  SpSetUnres(): SpSetElement(UNRES_ET,false), Name(NULL) {}
  ElementType Target() { return TargetType; }
private:
  char *Name;
  ElementType TargetType;
};


class SpSetNone: public SpSetElement
// The purpose of this class is to allow the Element(char *key)
// method to return an SpSetElement &, before actually knowing
// what type of element is to be added to the set, e.g.,
// 
// myset.Element("someelement") = 3;
//
// The Element() method will return an SpSetNone. The SpSetNone
// will actually replace itself with an SpSetInt with the 
// appropriate value (in this case, 3) 
{
 public:
  SpSetNone(char *key, SpSet *parent): SpSetElement(NONE_ET), Parent(parent) {
    Name = strdup(key);
  }
  
  SpSetElement &operator=(int rval);
  
  SpSetElement &operator=(double rval);
  
  SpSetElement &operator=(char *rval);  
  
  
  SpSetElement &operator=(bool rval);
  
 private:
  SpSet *Parent;
  char *Name;
};

#undef __SpSet_Internal__
#endif // #ifndef _SpSetElement_H
