// SpHiGrid.C

#ifndef __SPEEDES_NT__
#include <values.h>
#else
#define INTBITS 32
#define MININT      (1 << (INTBITS - 1))
#define MAXINT      (~MININT)
#endif
#include "SpIostream.H"
#include "SpIomanip.H"

#include "SpPubSubRecord.H"
#include "SpHiGrid.H"
#include "SpStopWatch.H"
#include "SpList.H"

#define PRUNE_HIGRID

// #define TEST_HIGRID
// #define HIGRID_TIMINGS
// #define DEBUG_HIGRID
// #define TEST_VA_INTERFACES
// #define HIGRID_ERROR_CHECK

int SpHiGrid::NhiGrids = 0;
int SpHiGrid::UniqueMark = 0;
C_QUEUE *SpHiGrid::QfreePSI = NULL;
C_QUEUE *SpHiGrid::QfreePSR = NULL;
int SpHiGrid::NcreatedPSI = 0;
int SpHiGrid::NcreatedPSR = 0;

/************************************************************************
* SpHiGrid - Constructor	doesn't do very much. Most of the		*
* initialization work is done in the init method or the InitChild	*
* method								*
************************************************************************/
SpHiGrid::SpHiGrid() {

/*
** Initialize some basic things here
*/

  H[SUBSCRIBER].NumAtOrBelowThisLevel = 0;
  H[PUBLISHER].NumAtOrBelowThisLevel = 0;
  Level = NULL;
  Offset = 0;

/*
** Establish the unique ID for this grid-cell
*/

  NhiGrids++;
  HiGridId = NhiGrids;

#ifdef DEBUG_HIGRID
  cout << "Constructing HiGrid " << HiGridId << endl;
#endif

  if (QfreePSI == NULL) {
    QfreePSI = new C_QUEUE();
    QfreePSR = new C_QUEUE();
  }

}

void SpHiGrid::UpdateNumAtOrBelowThisLevel(SpPubSub e) {
  H[e].NumAtOrBelowThisLevel++;
  if (Parent) Parent->UpdateNumAtOrBelowThisLevel(e);
}

/************************************************************************
* SpHiGrid - Destructor							*
************************************************************************/
SpHiGrid::~SpHiGrid() {
  int i;

#ifdef DEBUG_HIGRID
  cout << "Deleting HiGrid " << HiGridId << " Level:";
  for (i=0; i<Ndimensions+Offset; i++) {
    cout << " " << Level[i];
   }
  cout << endl;
#endif

  Delete(PUBLISHER);
  Delete(SUBSCRIBER);

/*
** Delete the children in a recursive manner before deleting the rest
** of the memory in this object.
*/

  for (i=0; i< Ndimensions; i++) {
    if (Children[i]) {
      delete [] Children[i];
    }
  }
  delete [] Children;
  delete [] Nchildren;

/*
** Delete the resolution information next. Remember that the root creates
** the top level stuff so it needs to delete all of its resolution arrays.
*/

  if (Parent == NULL) {
    for (i=0; i<Ndimensions; i++) {
      delete [] Resolution[i];
    }
  }
  delete [] Resolution;
  delete [] Nresolutions;

/*
** Delete the rest of the various arrays that were created
*/

  delete [] Level;
  delete [] Lo;
  delete [] Hi;

  if (referencing_parent_p) {
    delete [] DimNum;
    delete [] DimName;
  }
}



void SpHiGrid::init(char *SpName, int SpNum, int Ndims,
			 SpDimInfo *DimInfo) {
  int i,j;
  int Dnum;
  char *Dname;
  //double Res[MAX_RESOLUTIONS];

#ifdef DEBUG_HIGRID
  cout << endl << "Initializing HiGrid" << endl;
#endif

/*
** The SpaceName is assumed to be represented as a string that never gets
** deleted. It is shared by many grid-cells to save memory. Here, we just
** reference it for general information. We also save the ID of the space
** (i.e., the SpaceNum). The number of dimensions is saved but keep in mind
** that this is a recursive data strucuture where children of a dimension
** do not include the previous dimensions of the parent.Counting always
** starts at zero.
*/

  SpaceName = SpName;
  SpaceNum = SpNum;
  Ndimensions = Ndims;

/*
** The Level array keeps track of where we are in the hierarchical levels.
** This information is used just for documenting where we are and is not
** used for anything important.
*/

  Level = new int[Ndims];

/*
** Create arrays for the dimension names and numbers. Remember, that
** dimension names have an id associated with them but each HiGrid
** starts its dimension numbering with zero.
*/

  DimName = new char *[Ndims];
  DimNum = new int[Ndims];
  referencing_parent_p = 1;

/*
** Create the Lo and Hi arrays which describe the bounds for each dimension.
** These bounds collectively define the HiGrid cell in the multidimensional
** space.
*/


  Lo = new double[Ndims];
  Hi = new double[Ndims];

/*
** Resolutions are stored as an array of arrays for each dimension. This
** allows a set of resolutions for each dimension. For a given dimension,
** the resolutions must be decreasing in size (makes sense). When creating
** a child HiGrid object, the resolutions are shared from the parent except
** for the first one which is recreated (in terms of storage). The first
** resolution is (later in this routine) adjusted so that an exact number
** of cells can be defined (e.g., if Lo=2 and Hi=7, and the first resolution
** in the arrary of resolutions is 1.1, then we adjust it so that it is 1).
*/

  Nresolutions = new int[Ndims];
  Resolution = new double *[Ndims];

/*
** Create an array of pointers to Children. The Children are not actually
** created (to save memory) unless a publisher or subscriber item needs to
** be inserted at their level. This should provide a more effective use of
** memory. The Parent pointer is set to NULL since we are creating the root
** of the HiGrid in this routine (see InitChild for initializing Children 
** nodes).
*/

  Children = new SpHiGrid *[Ndims];
  Nchildren = new int[Ndims];

  Parent = NULL;
  BinNumber = 0;

/*
** Initialize some of the quantities
*/

  for (i=0; i<Ndims; i++) {
    Children[i] = NULL;
    Nchildren[i] = 0;
    DimNum[i] = -1;
    DimName[i] = NULL;
    Lo[i] = 0;
    Hi[i] = 0;
    Nresolutions[i] = 0;
    Resolution[i] = NULL;
    Level[i] = 0;
  }

/*
** Now get stuff out of the DimInfo object. This is similar to the VA stuff
** in the init method.
*/


  for (i=0; i<Ndims; i++) {
    Dname = DimInfo[i].GetDimName();
    Dnum = DimInfo[i].GetDimNum();
    DimNum[Dnum] = Dnum;
    DimName[Dnum] = Dname;

    Lo[Dnum] = DimInfo[i].GetLo();
    Hi[Dnum] = DimInfo[i].GetHi();

    Nresolutions[Dnum] = DimInfo[i].GetNresolutions();
    Resolution[Dnum] = new double[Nresolutions[Dnum]];
    for (j=0; j<Nresolutions[Dnum]; j++) {
      Resolution[Dnum][j] = DimInfo[i].GetResolution(j);
    }
  }

/*
** Adjust the resolutions so that they fit exactly into an even number of
** bins.
*/

  AdjustResolution();

/*
** Create new hash tables for looking up publisher and subscribers. This is
** only done in the root, not by any of the Children HiGrids. The hash tables
** are provided to the HiGrids during initialization of the Children.
*/

  H[PUBLISHER].ParentHash = new SpHashTree(N_HIGRID_HASH);
  H[SUBSCRIBER].ParentHash = new SpHashTree(N_HIGRID_HASH);


#ifdef DEBUG_HIGRID
  print();
  cout << "Done Initializing HiGrid" << endl << endl;
#endif

}


/************************************************************************
* InitChild - This method initializes a Hierarchical Grid object child. *
************************************************************************/
void SpHiGrid::InitChild(SpHiGrid *P, int Dim, int Bin) {
  int i;
  double **Pres;
  int *Pnres;
  int *Plev;
  double *TempRes;

/*
** Set the Parent pointer and the Bin Number
*/

  Parent = P;
  BinNumber = Bin;

/*
** Copy the space name, space number and number from the parent. The number
** of dimensions here is the parents number of dimensions - the dimension
** number passed in as an argument. This is done in order to keep the lowest
** dimension of the HiGrid indexed as 0.
*/

  SpaceName = Parent->get_SpaceName();
  SpaceNum = Parent->get_SpaceNum();
  Ndimensions = Parent->get_Ndimensions()-Dim;

#ifdef DEBUG_HIGRID
  cout << "Initializing HiGrid Child with "
       << Ndimensions << " dimensions" << endl;
#endif

/*
** Get the publisher and subscriber lookup hash tables from the parent
*/

  H[PUBLISHER].ParentHash = Parent->H[PUBLISHER].ParentHash;
  H[SUBSCRIBER].ParentHash = Parent->H[SUBSCRIBER].ParentHash;

/*
** Set up the Level information. The Level array keeps track of which
** level the current HiGrid is in.
*/

  Plev = Parent->get_Level();
  Offset = Dim + Parent->get_Offset();
  Level = new int[Ndimensions+Offset];
  memcpy(Level, Plev, (Ndimensions+Offset)*sizeof(int));
  Level[Offset]++;

/*
** Get the dimension names and dimension numbers from the parent.
*/

  DimName = Parent->get_DimNames(Dim);
  DimNum = Parent->get_DimNums(Dim);
  referencing_parent_p = 0;

/*
** Get the resolution information from the parent. We will use this later when
** we determine the Lo and Hi values. Also, we will use this to recalculate
** the first level resolutions to match the Lo and Hi values. Note that only
** the resolutions for the lowest dimension are recalculated (and use new
** memory as well). Then, we subtract one from the number of resultions of
** the lowest dimension and apply the next resolution to this HiGrid.
*/

  Pres = Parent->get_Resolutions(Dim);
  Pnres = Parent->get_Nresolutions(Dim);

  Lo = new double[Ndimensions];
  Hi = new double[Ndimensions];
  memcpy(Lo, Parent->get_Los(Dim), Ndimensions*sizeof(double));
  memcpy(Hi, Parent->get_His(Dim), Ndimensions*sizeof(double));
  Lo[0] += Pres[0][0]*Bin;
  Hi[0] = Lo[0] + Pres[0][0];

  Nresolutions = new int[Ndimensions];
  memcpy(Nresolutions, Pnres, Ndimensions*sizeof(int));
  Nresolutions[0]--;

#ifdef DEBUG_HIGRID
  cout << "Initializing HiGrid Child: Nresolutions[0] "
       << Nresolutions[0] << endl;
#endif

/*
** Create a new resolution array with one less for our lowest dimension and 
** then copy all but the parents first resolution into this array. Remember,
** the Child has now moved down one resolution so we need to reflect this.
** Then, we need to recalculate that resolution to make sure that we have
** an exact binning.
*/

  Resolution = new double *[Ndimensions];
  memcpy(Resolution, Pres, Ndimensions*sizeof(double **));
  TempRes = Resolution[0];

  if (Nresolutions[0] > 0) {
    Resolution[0] = &TempRes[1];
  }else{
    Resolution[0] = NULL;
  }

/*
** Finally, create the array of pointers to Children in case we need to ever
** create new Children.
*/

  Children = new SpHiGrid *[Ndimensions];
  Nchildren = new int[Ndimensions];
  for (i=0; i<Ndimensions; i++) {
    Children[i] = NULL;
    Nchildren[i] = 0;
  }

#ifdef DEBUG_HIGRID
  print();
  cout << "Done initializing HiGrid Child" << endl << endl;
#endif

}


/************************************************************************
* Prune - This method prunes the hierarchical grid tree structure if	*
* it can. It knows it can prune the tree if its number of Publishers	*
* and subscribers is zero						*
************************************************************************/
void SpHiGrid::Prune(int Pub, int Sub, int Dim=-1, int Nbelow=-1) {
  int i;
  int Index;
  int tot;
  int MyN;

/*
** First subtract from either the number of publishers or the number of
** subscribers. Then compute the total number of either. If the answer is
** zero, then that means that we might be able to do some cleanup.
*/

  H[PUBLISHER].NumAtOrBelowThisLevel -= Pub;
  H[SUBSCRIBER].NumAtOrBelowThisLevel -= Sub;

/*
** We can turn off the deleting of HiGrid objects with the PRUNE_HIGRID
** compiler flag (which is defined at the top of this file).
*/

#ifdef PRUNE_HIGRID

  MyN = H[PUBLISHER].NumAtOrBelowThisLevel + H[SUBSCRIBER].NumAtOrBelowThisLevel;

  if (Nbelow == 0) {

/*
** If there are no publishers or subscribers below me then I might be able to
** delete some children. Note, that we have to convert from Dim to an Index
** since we have to keep straight the fact that all children are referenced
** starting with zero so we have to use the offset.
*/

    Index = Dim-Offset;

    if (MyN == 0) {

/*
** If I also don't have any publishers or subscribers, then I know it is safe
** to delete the children below me (i.e., the Index from the Dim parameter
** that was passed in.
*/

      delete [] Children[Index];
      Children[Index] = NULL;
      Nchildren[Index] = 0;

    }else{

/*
** If MyN is not zero, it is still possible for my children for that Index
** to not have any publishers or subscribers so let's check them. If any of
** them have publishers or subscribers, then we forget it.
*/

      tot = 0;
      for (i=0; i<Nchildren[Index]; i++) {
	tot += Children[Index][i].get_Npubsub();
	if (tot) break;
      }
      if (tot == 0) {
	delete [] Children[Index];
	Children[Index] = NULL;
	Nchildren[Index] = 0;
      }

    }

  }

#endif

/*
** Here is where we recursively call ourselves again until we get to the
** root (i.e., where the Parent pointer is NULL.
*/

  if (Parent) Parent->Prune(Pub, Sub, Offset, MyN);

}


int SpHiGrid::CalculateNumChildren(const int dimension) const
{
  const double MORE_THAN_ONE = 1.0000000001;
  return int( MORE_THAN_ONE*(Hi[dimension] - Lo[dimension])
	      / Resolution[dimension][0]);
}

int SpHiGrid::Intersect
  (MAIN_PUBSUB_ITEM *item, const int starting_dimension)
{
  for (int i=0; i<Ndimensions; i++) {
    const int dim_with_offset = starting_dimension + i;
    if (item->GetRes(dim_with_offset) >= 0 &&
	((item->GetLo(dim_with_offset) > Hi[i]) ||
	 (item->GetHi(dim_with_offset) < Lo[i]))) {
      return 0;
      //cout << "Error (SpHiGrid::Add): Lo/Hi values out of range" << endl;
    }
  }
  return 1;
}

int intersect(MAIN_PUBSUB_ITEM *a, MAIN_PUBSUB_ITEM *b)
{
  assert(a->GetNdimensions() == b->GetNdimensions());
  for (int i=0; i<a->GetNdimensions(); i++) {
    if (a->GetRes(i) >= 0 &&
	b->GetRes(i) >= 0 &&
	((a->GetLo(i) > b->GetHi(i)) ||
	 (a->GetHi(i) < b->GetLo(i)))) {
      return 0;
    }
  }
  return 1;
}

/************************************************************************
* This method adds a new publisher or subscriber to the hierarchical	*
* grid using defined arguments. It does this in a recursive manner.	*
************************************************************************/
int SpHiGrid::AddInternal(SpPubSub pubOrSub, MAIN_PUBSUB_ITEM *MainPubSubItem,
			   const int starting_dimension) {
  int i,j;
  int Dim, LoDim, HiDim;
  int Bin, LoBin, HiBin;

  if (!Intersect(MainPubSubItem, starting_dimension))
      return 0;

/*
** Find out what the next resolution Dimension and Bin cell should be. If
** this routine returns values -1 for Dim and Bin, then we know we are in
** a HiGrid cell with the right resolution so we should just add the item
** to this grid. The dimension returned by these two calls should always be
** the same. If not, then we have an error.
*/

  GoNextResolution(&MAIN_PUBSUB_ITEM::GetLo, MainPubSubItem, starting_dimension,
		   LoDim, LoBin);
  GoNextResolution(&MAIN_PUBSUB_ITEM::GetHi, MainPubSubItem, starting_dimension,
		   HiDim, HiBin);

  if (LoDim != HiDim) {
    cout << "Error SpHiGrid LoDim (" << LoDim
	 << ") is not equal to HiDim (" << HiDim << ")" << endl;
    Dim = (LoDim + HiDim) / 2;	// This should never happen
  }else{
    Dim = LoDim;
  }

#ifdef DEBUG_HIGRID
  cout << "Dim " << Dim << ", LoBin " << LoBin << ", HiBin " << HiBin << endl;
  print();
  cout << endl;
#endif

/*
** Handle the case where we just need to add the item to the current
** level. We need to also add a record to the publishers or
** subscribers lookup hash table
*/

  if (LoDim == -1) {
    H[pubOrSub].ItemsStoredHere.Insert
                        (MainPubSubItem, MainPubSubItem->GetUniqueKey());
    H[pubOrSub].ParentHash->Insert(this, MainPubSubItem->GetUniqueKey());

#ifdef DEBUG_HIGRID
    cout << "SpHiGrid: Added PubSubItem "
	 << endl;
    print();
#endif

    UpdateNumAtOrBelowThisLevel(pubOrSub);

    return 1;
  }

/*
** Handle the case where we need to create the children first before adding
** the item. Each child must be initialized to get its proper resolutions,
** etc.
*/

  if (Children[Dim] == NULL) {
    j = CalculateNumChildren(Dim);
    Children[Dim] = new SpHiGrid[j];
    Nchildren[Dim] = j;
    for (i=0; i<j; i++) {
      Children[Dim][i].InitChild(this, Dim, i);
    }
  }

#ifdef HIGRID_ERROR_CHECK
  if (!CheckIntegrity()) {
    cout << "Bad children created" << endl;
    abort();
  }
#endif

/*
** Make sure that the Lo and Hi bins do not take us out of bounds in our
** Children array.
*/

  if (LoBin >= Nchildren[Dim]) {
    cout << "Error SpHiGrid::Add: Bad LoBin number " << LoBin
	 << " (max should be " << Nchildren[Dim] << ")" << endl;
    if (LoBin>0) LoBin = Nchildren[Dim]-1;
  }
  if (LoBin < 0) {
    cout << "Error SpHiGrid::Add: Negative LoBin number " << LoBin << endl;
    LoBin = 0;
  }

  if (HiBin >= Nchildren[Dim]) {
    cout << "Error SpHiGrid::Add: Bad HiBin number " << HiBin
	 << " (max should be " << Nchildren[Dim] << ")" << endl;
    if (HiBin>0) HiBin = Nchildren[Dim]-1;
  }
  if (HiBin < 0) {
    cout << "Error SpHiGrid::Add: Negative HiBin number " << HiBin << endl;
    HiBin = 0;
  }

/*
** Loop over the Lo and Hi dimensions that were returned from the
** GoNextResolution method. This will ensure that we add the publisher
** or subscriber to each cell that it needs to be added to. We have to
** be careful in this loop since if an item gets added to multiple
** cells, we have to clone the item into multiple objects just before
** we add them. We never add the actual item.
*/

  for (Bin = LoBin; Bin<=HiBin; Bin++) {
#ifdef ERROR_CHECK
    const int added_p = 
#endif
    Children[Dim][Bin].AddInternal
      (pubOrSub, MainPubSubItem, starting_dimension + Dim);
#ifdef ERROR_CHECK
    assert(added_p);
#endif
  } 
  return 1;
} 



SpMainPubSubItem *SpHiGrid::Remove(SpPubSub pubOrSub, int uniqueKey) {
  SpMainPubSubItem *MainPubSubItem = NULL;
  SpHiGrid *HiGrid;
 
#ifdef DEBUG_HIGRID
  cout << "Remove: Trying to remove " << i << endl;
#endif

  while ((HiGrid = (SpHiGrid *)H[pubOrSub].ParentHash->Remove(uniqueKey)))
  {
    SpMainPubSubItem *just_found = (SpMainPubSubItem *)
      HiGrid->H[pubOrSub].ItemsStoredHere.Remove(uniqueKey);
    if (MainPubSubItem) {
      if (MainPubSubItem != just_found) {
	cout << "Error, multiple MainPubSubItems - uniqueKey : " << uniqueKey << endl;
        abort();
      }
    }

    MainPubSubItem = just_found;

    if (pubOrSub == PUBLISHER)
      HiGrid->Prune(1, 0);
    else if (pubOrSub == SUBSCRIBER)
      HiGrid->Prune(0, 1);
    else
      abort();

  }

#ifdef ERROR_CHECK
  int extra = 0;
  void *v;
  for (v = H[pubOrSub].ParentHash->GetFirstElement(); v;
       v = H[pubOrSub].ParentHash->GetNextElement())
  {
    if (H[pubOrSub].ParentHash->GetCurrentKey() == uniqueKey)
      ++extra;
  }
  if (extra > 0)
  {
    cerr << "Found " << extra << " of uniqueKey " << uniqueKey << endl;
    assert(0);
  }
#endif

/*
** Return the MainPubSubItem back to the caller of this routine so that they
** can do whatever is necessary (like delete it, or whatever)
*/

  return MainPubSubItem;

}


/************************************************************************
* This method finds all of the publishers or subscribers that are	*
* in the same grids (or higher or lower) as the lookup. 		*
************************************************************************/
void SpHiGrid::GetInternal(SpPubSub pubOrSub, SpList *matchingPubsOrSubs,
			   MAIN_PUBSUB_ITEM *lookup,
			   const int starting_dimension) {
  int i,j;
  int Lindex, Hindex;
  SpMainPubSubItem *MainPubSubItem;

  if (!Intersect(lookup, starting_dimension))
      return;

/*
** First check if there are any publishers or subscribers in our level
** or below. If there are not any, then we do not need to look
** further.
*/

  if (!H[pubOrSub].NumAtOrBelowThisLevel) return;

/*
** Loop over all of the PubSub items in the publisher or subscriber
** list for this grid and put their MainPubSubItem in the matchingPubsorSubs list
** that is returned. Here, we have be careful that we do not put the
** same item in the list twice (it could happen if we are not careful
** since regions are defined for both subscribers and publishers). We
** handle this problem by setting a unique mark in the MainPubSubItem
** which we can check each time.
**
** I do not believe the reason given above makes sense.  The comment,
** written by someone else, says the reason we have to be careful is
** about regions being defined...  I think the real reason is that the
** top-level SpHiGrid is divided up into several smaller hiGrids, and a
** PubSubItem might be in more than one of those smaller HiGrids.
*/

  for (MainPubSubItem
	 = (SpMainPubSubItem *)H[pubOrSub].ItemsStoredHere.GetFirstElement();
       MainPubSubItem;
       MainPubSubItem = (SpMainPubSubItem *)(++(H[pubOrSub].ItemsStoredHere)))
  {
    if (MainPubSubItem->get_Mark() != UniqueMark) {
      MainPubSubItem->set_Mark(UniqueMark);
      if (intersect(MainPubSubItem, lookup))
	matchingPubsOrSubs->push_bot(MainPubSubItem);
    }
  }

/*
** Now we go down and look at the children to see if they have
** publishers or subscribers that might need to be included.
*/

  for (i=0; i<Ndimensions; i++) {
    const int dim_with_offset = starting_dimension + i;
    if (Nchildren[i]) {
      if (lookup->GetRes(dim_with_offset) < 0)
      {
	Lindex = 0;
	Hindex = Nchildren[i] - 1;
      }
      else
      {
	Lindex = GetChildIndex(i, lookup->GetLo(dim_with_offset));
	Hindex = GetChildIndex(i, lookup->GetHi(dim_with_offset));
      }
      
      for (j=Lindex; j<=Hindex; j++) {
	Children[i][j]
	  .GetInternal(pubOrSub, matchingPubsOrSubs,
		       lookup, dim_with_offset);
      }
    }
  }

}

void SpHiGrid::GetWithKey(SpPubSub pubOrSub, int uniqueKeyOfPubOrSub,
			  SpList *matchingPubsOrSubs)
{
  SpHiGrid *HiGrid
    = (SpHiGrid*) H[Opposite(pubOrSub)].ParentHash->Find(uniqueKeyOfPubOrSub);
  if (HiGrid != NULL) {
    SpMainPubSubItem *item = (SpMainPubSubItem*)
      HiGrid->H[Opposite(pubOrSub)].ItemsStoredHere.Find(uniqueKeyOfPubOrSub);
    assert(item);
    Get(pubOrSub, item, matchingPubsOrSubs);
  }
}

int SpHiGrid::GetChildIndex(int dimension, double lookupPoint)
{
  const int future_num_children = CalculateNumChildren(dimension);
  //assert(future_num_children >= 1);
  if (lookupPoint < Lo[dimension])
    lookupPoint = Lo[dimension];
  else if (lookupPoint > Hi[dimension])
    lookupPoint = Hi[dimension];
  
  double double_bin = (lookupPoint - Lo[dimension]) / Resolution[dimension][0];
  if (double(-MAXINT) >= double_bin || double_bin >= double(MAXINT))
  {
    cerr << "Overflow or underflow detected." << endl;
    assert(0);
  }
  int child_index = int(double_bin);
  if (child_index < 0) {
    cerr << "Internal SpHiGrid error" << endl;
    child_index = 0;
  }
  else if (child_index >= future_num_children)
  {
    if (child_index > future_num_children) {
      cerr << "Internal SpHiGrid error" << endl;
    }
    child_index = future_num_children - 1;
  }
  assert(child_index >= 0);
  assert(child_index < future_num_children);
  
  return child_index;
}

/************************************************************************
* AdjustResolution - This method adjusts the resolutions that were	*
* passed in during the initialization process and makes then exactly	*
* line up between lo and hi values.					*
************************************************************************/
void SpHiGrid::AdjustResolution() {
  int i,j,k;

  for (i=0; i<Ndimensions; i++) {
    assert(Nresolutions[i] >= 1);
    k = int(0.5 + (Hi[i]-Lo[i])/Resolution[i][0]);
    assert(k != 0);
    Resolution[i][0] = (Hi[i]-Lo[i]) / k;
    for (j=1; j<Nresolutions[i]; j++) {
      k = int(0.5 + (Resolution[i][j-1] / Resolution[i][j]));
      Resolution[i][j] = Resolution[i][j-1] / k;
    }
  }

}


/************************************************************************
* GoNextResolution - This method determines if we should go to the	*
* next resolution in the grid hierarchy or if the current resolution is	*
* fine.  Last two arguments are output arguments.			*
************************************************************************/
void SpHiGrid::GoNextResolution(GetLoOrHiFunc getLoOrHiFunc,
				MAIN_PUBSUB_ITEM *MainPubSubItem,
				int starting_dimension, int &Dim, int &Bin) {
  int i;
  double temp1;
  double temp2;

#ifdef DEBUG_HIGRID
  cout << "GoNextResolution Starting" << endl;
#endif

/*
** We loop over each dimension to see if we need to go down a level to
** achieve finer resolution. For each dimension, we first check if
** we care about the dimension (i.e., if Res[i] > 0) and check if further
** resolutions are possible (or are we already at the finest resolution).
** If so, then we check if it is better for us to stay at the current
** resolution or should we go down another level. If the requested resolution
** is larger than the current resolution in the HiGrid, then we certainly do
** not want to go down to a lower resolution. The final check is if the next
** resolution is too small. This is the case where the requested resolution
** is smaller than the current one but the next one might be too small. A
** simple rule is used that compares the ratios of the requested resolution
** to the current resolution to 1 / (the same ratio using the next resolution).
** This is another way of comparing the efficiency of the current resolution
** with one over the number of grid cells in the lower resolution.OK, it's a
** bit complicated but it seems to work.
*/

  for (i=0; i<Ndimensions; i++) {
    const int dim_with_offset = starting_dimension + i;
    if (MainPubSubItem->GetRes(dim_with_offset) >= 0) {
      if (Nresolutions[i] > 0) {
        temp1 = MainPubSubItem->GetRes(dim_with_offset) / (Hi[i]-Lo[i]);
        if (temp1 < 1) {
	  temp2 = MainPubSubItem->GetRes(dim_with_offset) / Resolution[i][0];
	  if ((1.0/temp1) > temp2) {
	    Dim = i;
	    Bin = GetChildIndex(i, ((*MainPubSubItem).*getLoOrHiFunc)
				(dim_with_offset));

#ifdef DEBUG_HIGRID
            cout << "GoNextResolution: Dim " << Dim << " Bin " << Bin << endl;
#endif

	    return;
	  }
	}
      }
    }
  }

/*
** If we get to this point, it means that all of the dimensions have passed
** their resolution tests for the current HiGrid and we should just use this
** resolution.
*/

  Dim = -1;
  Bin = -1;

#ifdef DEBUG_HIGRID
  cout << "GoNextResolution: Bin " << Bin << " Dim " << Dim << endl;
#endif

}


/************************************************************************
* CheckIntegrity - This method checks to make sure that the children	*
* are all within bounds of their parents				*
************************************************************************/
int SpHiGrid::CheckIntegrity() {
  int i, j, k;
  int status;
  double *l;
  double *h;
  status = 1;

/*
** First check that my children are ok
*/

  for (i=0; i<Ndimensions; i++) {
    for (j=0; j<Nchildren[i]; j++) {
      l = Children[i][j].get_Los(0);
      h = Children[i][j].get_His(0);
      for (k=i; k<Ndimensions; k++) {
	// I took out the multiplication by MORE_THAN_ONE or by
	// LESS_THAN_ONE since that code only works for positive Lo
	// and Hi values.
        if ((Lo[k] > l[k-i]) || (Hi[k] < h[k-i])) {
	  cout << "Error (CheckIntegrity): Bad grids,\nParent = ["
	       << setprecision(30) << Lo[k] << ", " << Hi[k] << "]\n Child = ["
	       << l[k-1] << ", " << h[k-1] << "]" << endl;
	  status = 0;
	  return status;
        }
      }
    }
  }

/*
** Then, recursively do the same for all of my children
*/

  for (i=0; i<Ndimensions; i++) {
    for (j=0; j<Nchildren[i]; j++) {
      status = Children[i][j].CheckIntegrity();
      if (status == 0) return status;
    }
  }

  return status;

}


/************************************************************************
* print - This method prints out the Hierarchical Grid object, but only	*
* at the current level (for printing out the entire tree structure,	*
* use the RecursivePrint method						*
************************************************************************/
void SpHiGrid::print(ostream &out) const {
  int i, j;

  if (Level == NULL) {
    out << "HiGrid " << HiGridId
	<< " has not been initialized... Aborting print" << endl;
    return;
  }

/*
** Print out the parent if there is one
*/

  if (Parent) {
    Parent->print(out);
  }

/*
** Print out header type information
** (1) Space Number and Name
** (2) HiGrid Unique Id
** (3) Level information
** (4) The Dimensional offset
*/

  out << "Space Name(" << SpaceNum << "): " << SpaceName << endl;
  out << "HiGrid Id: " << HiGridId << endl;
  out << "Level:";
  for (i=0; i<Ndimensions+Offset; i++) out << " " << Level[i];
  out << endl;
  out << "Offset: " << Offset << endl;

/*
** For each dimension, we print out the dimension name, Lo, Hi, and
** resolution information
*/

  for (i=0; i<Ndimensions; i++ ) {
     out << DimName[i] << "(" << i+Offset << "): Lo " << Lo[i]
	 << ", Hi " << Hi[i] << ", Resolution [";
    for (j=0; j<Nresolutions[i]; j++) {
      out << Resolution[i][j];
      if (j != Nresolutions[i]-1) out << ", ";
    }
    out << "]" << endl;
  }

/*
** Finally we print out the number of publishers and subscribers in our HiGrid
** which includes its children.
*/

  out << "H[PUBLISHER].NumAtOrBelowThisLevel " << H[PUBLISHER].NumAtOrBelowThisLevel
       << ", H[SUBSCRIBER].NumAtOrBelowThisLevel " << H[SUBSCRIBER].NumAtOrBelowThisLevel
       << endl;

}

void SpHiGrid::PrintItemsStoredHere(ostream &out, SpPubSub e)
{
  out << "Number of " << e
      << " is " << H[e].ItemsStoredHere.GetNumElements() << ":";
  for (SpMainPubSubItem *MainPubSubItem =
	 (SpMainPubSubItem *)H[e].ItemsStoredHere.GetFirstElement();
       MainPubSubItem;
       MainPubSubItem =
	 (SpMainPubSubItem *)H[e].ItemsStoredHere.GetNextElement())
    out << " " << MainPubSubItem->GetUniqueKey();
  out << '\n';
}

/************************************************************************
* RecursivePrint - This method recursively prints out an entire HiGrid	*
* tree									*
************************************************************************/
void SpHiGrid::RecursivePrint(ostream &out,
			     int L,
			     double *Lval,
			     double *Hval,
			     char **Dname) {
  if (L == 0)
    out << "Space Name: " << SpaceName << endl;

/*
** We don't go down any more levels if there are no publishers or subscribers
** below.
*/

  if (get_Npubsub() == 0) return;

  int i, j;
  char **DN;
  double *LV;
  double *HV;
  double *TLV;
  double *THV;


/*
** If this is the top level, then set up the low and high values as well as
** the dimension name for the rest of the recursive calls.
*/

  if (L == 0) {
    LV = new double[Ndimensions];
    HV = new double [Ndimensions];
    DN = DimName;
  }else{
    LV = Lval;
    HV = Hval;
    DN = Dname;
    TLV = new double[Ndimensions];
    THV = new double [Ndimensions];
  }

/*
** Readjust the low and high values for printing
*/

  for (i=0; i<Ndimensions; i++) {
    if (L != 0) {
      TLV[i] = LV[i+Offset];
      THV[i] = HV[i+Offset];
    }
    LV[i+Offset] = Lo[i];
    HV[i+Offset] = Hi[i];
  }

/*
** Only print stuff for our level if there are publishers or subscribers in
** the queues at this level. Note, there may be publishers and subscribers
** below, but we do not want to print anything here.
*/

  if (H[PUBLISHER ].ItemsStoredHere.GetNumElements()+
      H[SUBSCRIBER].ItemsStoredHere.GetNumElements()) {

    out << "this = " << this << ", Parent = " << Parent << endl;
    for (i=0; i<L; i++) out << "..";
    out << "Level:";
    for (i=0; i<Ndimensions+Offset; i++) out << " " << Level[i];

    out << '\n';
    for (i=0; i<Offset+Ndimensions; i++) {
      out << "  [" << DN[i]
	  << ": Lo " << LV[i]
	  << ", Hi " << HV[i]
          << "]\n";
    }


    PrintItemsStoredHere(out, SUBSCRIBER);
    PrintItemsStoredHere(out, PUBLISHER);
    
    out << endl;

  }

/*
** If there are children below, see if they need to be printed by calling
** routine in a recursive manner.
*/

  for (i=0; i<Ndimensions; i++ ) {
    for (j=0; j<Nchildren[i]; j++) {
      Children[i][j].RecursivePrint(out, L+1, LV, HV, DN);
    }
  }

/*
** Restore the low and high values
*/

  if (L != 0) {
    for (i=0; i<Ndimensions; i++) {
      LV[i+Offset] = TLV[i];
      HV[i+Offset] = THV[i];
    }

    delete [] TLV;
    delete [] THV;
  }

/*
** If we are at the top level, then clean up any memory that might have been
** created
*/

  if (L == 0) {
    delete [] LV;
    delete [] HV;
  }

}


/************************************************************************
* new_pubsub_item - This method creates a new pubsub item using free	*
* lists									*
************************************************************************/
SpPubSubItem * SpHiGrid::new_pubsub_item(SpPubSubItem *PSI) {
  SpPubSubItem *PubSubItem;
  int i;

/*
** Get the pubsub item from the free list if possible. Otherwise, we
** have to create it N_PSI items at a time. If N_PSI items could not
** be created, then try to create one. If even one cannot be created, then
** we report an error and exit the program. During the creation of new
** items, we keep track of how many have been created in order to do some
** memory management. If we are able to create N_PSI items, we return the
** first one and then put the rest in the free list for future use.
*/

  if (QfreePSI->get_length()) {
    PubSubItem = (SpPubSubItem *)QfreePSI->pop_top();
  }else{
    PubSubItem = new SpPubSubItem[N_PSI];
    if (PubSubItem == NULL) {
      PubSubItem = new SpPubSubItem();
      if (PubSubItem == NULL) {
        cout << "Error, could not create PubsubItem" << endl;
        abort();
      }else{
        NcreatedPSI++;
      }
    }else{
      NcreatedPSI += N_PSI;
      for (i=1; i<N_PSI; i++) QfreePSI->push_bot(&PubSubItem[i]);
    }
  }

/*
** If the PSI object is not NULL, then we want to initialize the new item
** using the PSI object.
*/

  if (PSI) PubSubItem->init(PSI);

/*
** Return the new item
*/

  return PubSubItem;

}


/************************************************************************
* new_pubsub_record - This method creates a new pubsub record using	*
* free lists   								*
************************************************************************/
SpPubSubRecord * SpHiGrid::new_pubsub_record() {
  SpPubSubRecord *PubSubRecord;
  int i;

/*
** Get the pubsub record from the free list if possible. Otherwise, we
** have to create it N_PSR records at a time. If N_PSR records could not
** be created, then try to create one. If even one cannot be created, then
** we report an error and exit the program. During the creation of new
** records, we keep track of how many have been created in order to do some
** memory management. If we are able to create N_PSR records, we return the
** first one and then put the rest in the free list for future use.
*/

  if (QfreePSR->get_length()) {
    PubSubRecord = (SpPubSubRecord *)QfreePSR->pop_top();
  }else{
    PubSubRecord = new SpPubSubRecord[N_PSR];
    if (PubSubRecord == NULL) {
      PubSubRecord = new SpPubSubRecord();
      if (PubSubRecord == NULL) {
        cout << "Error, could not create PubsubRecord" << endl;
        abort();
      }else{
        NcreatedPSR++;
      }
    }else{
      NcreatedPSR += N_PSR;
      for (i=1; i<N_PSR; i++) QfreePSR->push_bot(&PubSubRecord[i]);
    }
  }

/*
** Return the record
*/

  return PubSubRecord;

}

void SpHiGrid::Delete(SpPubSub pubOrSub)
{
/*
** Delete the actual publisher and subscriber queue items if they haven't
** already been deleted. This is really a warning situation since you
** should not delete any HiGrids that have publishers or subscribers in
** them.
*/

  if (H[pubOrSub].ItemsStoredHere.GetNumElements() > 0) {
    cout << "Warning: Deleting a HiGrid object with Subscribers or Publishers" << endl;
    H[pubOrSub].ItemsStoredHere.Reset();
  }

/*
** Delete the lookup hash tables for publishers and subscribers. Again, there
** should not be any publishers or subscribers in these hash tables when
** deleting a HiGrid object. Note, this only needs to be checked when trying
** to delete the root.
*/

  if (Parent == NULL) {
    //C_QUEUE *Qall;

    if (H[pubOrSub].ParentHash->GetNumElements() > 0) {
      H[pubOrSub].ParentHash->Reset();
      cout << "Warning: Deleting a HiGrid hash table with Publishers or Subscribers" << endl;
    }
    delete H[pubOrSub].ParentHash;
  }
}

int 
SpHiGrid::Add(SpPubSub pubOrSub, MAIN_PUBSUB_ITEM *MainPubSubItem) {

  if (H[pubOrSub].ParentHash->Find(MainPubSubItem->GetUniqueKey())){
    cout << "Error, multiple MainPubSubItems - uniqueKey : " 
         << MainPubSubItem->GetUniqueKey() << endl;
    abort();
  }

#ifdef ERROR_CHECK
      assert(H[pubOrSub].ParentHash->Find(MainPubSubItem->GetUniqueKey())
             == NULL);
#endif

      return AddInternal(pubOrSub, MainPubSubItem, 0);
}
