// SpHiGrid.H header file

#ifndef HiGrid_object
#define HiGrid_object

#include "SpStdlib.H"
#include "SpStdarg.H"
#include "SpIostream.H"
#include "SpAssert.H"

#include "SpBinaryTree.H"
#include "SpHashTree.H"
#include "SpPubSubItem.H"
#include "SpDimInfo.H"
#include "queue.H"

//#define END_DIM -1.0
//#define END_SPACE -1
#define MAX_RESOLUTIONS 5
#define MAX_DIMENSIONS 20
#define N_HIGRID_HASH 100
#define N_PSI 1000
#define N_PSR 1000

/*
** This class provides a hierarchical structure for storing items in a
** multidimensional and multiresolution grid space.
*/
class SpList;
class SpPubSubRecord;

typedef double (MAIN_PUBSUB_ITEM::*GetLoOrHiFunc)(int dimension);
  
class SpPubSubHolder {
    friend class SpHiGrid;

    // Collection of publishers or subscribers in this SpHiGrid.
    // I.e., they are not in the children's SpHiGrids.  It is used
    // with keys of UniqueKey and values of SpMainPubSubItem*.
    SpBinaryTree ItemsStoredHere;

    // Children (and grandchildren, etc.) SpHiGrids all have the same
    // address for this SpHashTree pointer:

    // This is used with keys of UniqueKey and values of pointers to
    // SpHiGrids whose ItemsStoredHere has SpMainPubSubItems with the
    // same UniqueKey as was used to key this ParentHash.
    SpHashTree *ParentHash;

    int NumAtOrBelowThisLevel;
};

class SpHiGrid {
  public:
    enum SpPubSub { PUBLISHER=0, SUBSCRIBER=1 };

  private:
    int *Level;			// Array of level values for this subspace
    int Offset;			// Dimensional offset for this subspace
    int BinNumber;		// Bin number (related to Parent)

    char *SpaceName;		// Name of this grid-space
    int SpaceNum;		// Id of this grid-space
    int Ndimensions;		// Number of dimensions for subspace
    char **DimName;		// Name of each dimension for this grid-space

    // The following variable is set, but the only place it is read is
    // to make a copy, but the copy is never read, so the contents of
    // this array can never affect anything.  I assume that Jeff S. just
    // forgot to put code that reads it.
    int *DimNum;		// Id for each dimension in this grid-space
    int referencing_parent_p;

    double *Lo;			// Lo grid-boundary values for each dimension
    double *Hi;			// Hi grid-boundary values for each dimension
    int *Nresolutions;		// Number of resolutions for each dimension
    double **Resolution;	// Array of resolutions for each dimension

    SpHiGrid *Parent;		// Pointer to my parent (root is NULL)
    SpHiGrid **Children;		// Children array for each dimension
    int *Nchildren;		// Number of children for each dimension

    SpPubSubHolder H[2];

    int HiGridId;		// Unique local Id for HiGrid
    static int NhiGrids;	// Number of HiGrids created

    static int UniqueMark;	// identifies a unique mark to be used

    static C_QUEUE *QfreePSI;	// Free list for pubsub items
    static int NcreatedPSI;	// Number of pubsub items created
    static C_QUEUE *QfreePSR;	// Free list for pubsub records
    static int NcreatedPSR;	// Number of pubsub records created

    // This function returns the index of the child that contains the
    // lookupPoint.  int(x), where x is a huge floating-point value,
    // causes overflow.  The K & R C book says that the action taken
    // on overflow is machine-dependent.  Underflow is also
    // machine-dependent.  This class used to have code where overflow
    // occurred, but I wrote this function to avoid overflow and
    // underflow.  The caller should make sure to pass in values which
    // will not cause overflow and underflow.  This function will
    // crash, printing an error message if it detects any violations of
    // this rule.
    int GetChildIndex(int dimension, double lookupPoint);

    void Delete(SpPubSub);

    void PrintItemsStoredHere(ostream &, SpPubSub);
    
    // This function works even if the children have not been created
    // yet.  It will just calculate how many SHOULD be created.
    int CalculateNumChildren(const int dimension) const;
    
  protected:

    SpPubSubItem *new_pubsub_item(SpPubSubItem *PSI = NULL);
    void delete_pubsub_item(SpPubSubItem *PSI) {QfreePSI->push_bot(PSI);}
    SpPubSubRecord *new_pubsub_record();
    void delete_pubsub_record(C_ITEM *PSR) {QfreePSR->push_bot(PSR);}

    void InitChild(SpHiGrid *P, int Dim, int Bin);
    void Prune(int Pub, int Sub, int Dim, int Nbelow);

    // This function returns zero iff the item does not intersect the SpHiGrid.
    int Intersect(MAIN_PUBSUB_ITEM *item, const int starting_dimension);
    
    void GoNextResolution(GetLoOrHiFunc, MAIN_PUBSUB_ITEM *,
			  int starting_dimension, int &Dim, int &Bin);
    void AdjustResolution();

    int *get_Level() {return Level;}
    int get_Offset() {return Offset;}

    void UpdateUniqueMark() {UniqueMark++;}
    void ResetUniqueMark() {UniqueMark = 0;}

    SpHiGrid *get_Parent() {return Parent;}
    SpHiGrid **get_Children() {return Children;}
    int *get_Nchildren() {return Nchildren;}

    void UpdateNumAtOrBelowThisLevel(SpPubSub);

    // returns true iff the it actually added something.  I.e., returns
    // true iff the passed in region intersects this SpHiGrid.
    int AddInternal(SpPubSub, MAIN_PUBSUB_ITEM *MainPubSubItem,
		    const int starting_dimension);

    void GetInternal(SpPubSub, SpList *matchingPubsOrSubs,
		     MAIN_PUBSUB_ITEM *MainPubSubItem,
		     const int starting_dimension);

    // These functions have not yet been written:
    void JoinPublishers(C_QUEUE *Qpub, int UpFlag);
    void JoinSubscribers(C_QUEUE *Qsub, int UpFlag);

  public:

    SpHiGrid();
    ~SpHiGrid();

    // It might be very difficult to write a correct version of the
    // following functions, since you have to avoid merely copying
    // member pointers.  If you did that then the destructor would
    // run on both the original SpHiGrid and the copy of that original,
    // and then it would delete the same address twice, since that
    // address was merely copied.  To write these functions correctly, you
    // would probably have to copy the data to which the member
    // pointers point.
  private:
    SpHiGrid       (SpHiGrid&) {abort();}
    void operator=(SpHiGrid&) {abort();}
    
  public:
    
    // I think you need to call one of these two functions after
    // creating a new HiGrid.  Maybe a public constructor should call
    // them.  For children pointed to by another HiGrid, maybe those
    // children could be created with some private constructor.
    void init(char *Name, int SpNum, int Ndims, SpDimInfo *DimInfo);

    // These Add functions store the address passed in for the
    // MainPubSubItem argument and later they dereference that address
    // and read and write certain data in the passed in object.  For
    // example, if you call AddSubscriber, and pass the address of a
    // local stack MAIN_PUBSUB_ITEM object, and then the caller of
    // AddSubscriber returns, then the stack is no longer valid memory,
    // but the SpHiGrid class will probably access this memory later.
    // Therefore, you should probably only pass these functions
    // something returned by new.
    //
    // If the MAIN_PUBSUB_ITEM does NOT intersect this SpHiGrid, then
    // we just return 0 having changed nothing.  But if it does
    // intersect then we return 1.
    //
    // You may NOT pass a item with a UniqueKey which is the same as
    // the UniqueKey of another item in this SpHiGrid.
    int AddPublisher(MAIN_PUBSUB_ITEM *MainPubSubItem) {
      return Add(PUBLISHER, MainPubSubItem);
    }

    int AddSubscriber(MAIN_PUBSUB_ITEM *MainPubSubItem) {
      return Add(SUBSCRIBER, MainPubSubItem);
    }
 
    int Add(SpPubSub pubOrSub, MAIN_PUBSUB_ITEM *MainPubSubItem);

    // The following three versions of Add functions differ from the
    // above versions in that they return a list by reference.  The
    // AddPublisher function returns a list of all subscribers that
    // intersect the passed-in region.  The AddSubscriber function
    // returns a list of all publishers that intersect the passed-in
    // region.  And the Add function's first argument is the type of
    // thing you are adding, and it returns the opposite of the type
    // indicated by the first argument.  Note that the the
    // MAIN_PUBSUB_ITEM and Splist self-documenting variable names
    // contain opposite types where type is subscriber or publisher.
    int AddPublisherAndGet
      (MAIN_PUBSUB_ITEM *publication_region, SpList *subscibers) {
      return AddAndGet(PUBLISHER, publication_region, subscibers);
    }

    int AddSubscriberAndGet
      (MAIN_PUBSUB_ITEM *subscription_region, SpList *publishers) {
      return AddAndGet(SUBSCRIBER, subscription_region, publishers);
    }
 
    int AddAndGet(SpPubSub pubOrSub, MAIN_PUBSUB_ITEM *region,
	    SpList *matchingPubsOrSubs)
    {
      const int ret = Add(pubOrSub, region);
      Get(Opposite(pubOrSub), region, matchingPubsOrSubs);
      return ret;
    }


    // If the uniqueKey is not found, then these functions return
    // NULL, else they return the SpMainPubSubItem having the passed
    // in uniqueKey.
    SpMainPubSubItem *RemovePublisher(int uniqueKey)
    { return Remove(PUBLISHER, uniqueKey); }
    SpMainPubSubItem *RemoveSubscriber(int uniqueKey)
    { return Remove(SUBSCRIBER, uniqueKey); }
    SpMainPubSubItem *Remove(SpPubSub, int uniqueKey);

    // If you want a dimension to be a don't care, then you should set
    // Res to be negative and you should NOT set hi or lo.  If you do
    // care, then you should set Res to be 0 or positive, and set hi
    // and lo to the region in which you care.
    //
    // These 3 functions add stuff to the SpList passed in.  If the
    // SpList has stuff before the call, then it will have that same stuff
    // after the call.
    //
    // If the MAIN_PUBSUB_ITEM does NOT intersect this SpHiGrid, then
    // we just return without adding anything to the SpList output argument.
    void GetPublishers(MAIN_PUBSUB_ITEM *MainPubSubItem, SpList *Qpub) {
      Get(PUBLISHER, MainPubSubItem, Qpub);
    }

    void GetSubscribers(MAIN_PUBSUB_ITEM *MainPubSubItem, SpList *Qsub) {
      Get(SUBSCRIBER, MainPubSubItem, Qsub);
    }

    void Get(SpPubSub pubOrSub, MAIN_PUBSUB_ITEM *MainPubSubItem,
	     SpList *matchingPubsOrSubs)
    {
      UniqueMark++;
      GetInternal(pubOrSub, matchingPubsOrSubs, MainPubSubItem, 0);
    }

    // If the uniqueKey is not in this SpHiGrid, then the function
    // just returns without adding anything to the SpList output
    // argument.
    //
    // The GetPublishers function returns a list of all publishers
    // that intersect the passed in subscriber.  The GetSubscribers
    // function returns a list of all subscribers that intersect the
    // passed in publisher.  Note that the the int and Splist
    // self-documenting variable names contain opposite types where
    // type is subscriber or publisher.  And the Get function's first
    // argument is the type of thing you are geting, and it takes a key
    // of the opposite of the type indicated by the first argument.
    // Its second argument is of the opposite type as its third argument.
    void GetPublishersWithKey(int uniqueKeyOfSubscriber, SpList *publishers) {
      GetWithKey(PUBLISHER, uniqueKeyOfSubscriber, publishers);
    }

    void GetSubscribersWithKey(int uniqueKeyOfPublisher, SpList *subscribers) {
      GetWithKey(SUBSCRIBER, uniqueKeyOfPublisher, subscribers);
    }

    void GetWithKey(SpPubSub pubOrSub, int uniqueKeyOfPubOrSub,
	     SpList *matchingPubsOrSubs);


    // Return the opposite of what is passed in.
    static inline SpPubSub Opposite(SpPubSub e) {
      return e == SpHiGrid::PUBLISHER
	? SpHiGrid::SUBSCRIBER
	: SpHiGrid::PUBLISHER;
    } 

    char *get_SpaceName() {return SpaceName;}
    int get_SpaceNum() {return SpaceNum;}
    int get_Ndimensions() {return Ndimensions;}
    char **get_DimNames(int Dim) {return &DimName[Dim];}
    char *get_DimName(int Dim) {return DimName[Dim];}
    int *get_DimNums(int Dim) {return &DimNum[Dim];}
    double *get_Los(int Dim) {return &Lo[Dim];}
    double *get_His(int Dim) {return &Hi[Dim];}
    double get_Lo(int Dim) {return Lo[Dim];}
    double get_Hi(int Dim) {return Hi[Dim];}
    double **get_Resolutions(int Dim) {return &Resolution[Dim];}
    double *get_Resolution(int Dim) {return Resolution[Dim];}
    int *get_Nresolutions(int Dim) {return &Nresolutions[Dim];}
    int get_Nresolution(int Dim) {return Nresolutions[Dim];}
    int get_Npubsub() {return  H[PUBLISHER].NumAtOrBelowThisLevel+H[SUBSCRIBER].NumAtOrBelowThisLevel;}

    void print(ostream &out=cout) const;
    void RecursivePrint(ostream &out=cout,
			int L=0,
			double *Lval=NULL,
			double *Hval=NULL,
			char **Dname=NULL);

    int CheckIntegrity();

};

inline ostream & operator<<(ostream &os, const SpHiGrid &HiGrid)
{
  HiGrid.print(os);
  return os;
}

inline ostream & operator<<(ostream &os, const SpHiGrid::SpPubSub e)
{
  if (e == SpHiGrid::PUBLISHER)
    os << "SpHiGrid::PUBLISHER";
  else if (e == SpHiGrid::SUBSCRIBER)
    os << "SpHiGrid::SUBSCRIBER";
  else
    assert(0);

  return os; 
} 

typedef SpHiGrid HI_GRID;

#endif
