// Filename: SpDataParser.H

#ifndef DataParserObject
#define DataParserObject

#include "SpString.H"
#include "SpIostream.H"
#include "SpFstream.H"
#include "SpStrstream.H"
#include "SpIomanip.H"
#include "SpStdarg.H"

#include "queue.H"
#include "SpDataParserTypes.H"

class BASE_DATA_TYPE;
class DEFINE_DATA_TYPE;

/*****************************************************************************
* Class: SpDataParser
* Summary:  Object that reads in an input file and provides data for 
*           applications.
* Description: This SpDataParser is designed to parse files of the format
*              exemplified by:
*
*              // begin example *.par file
*
*              Set_1
*              {
*                int i_1 14
*                reference Group_1 Set_2
*                reference Group_1 Set_3
*                Group_1 Set_4
*                {
*                  int i 40
*                }
*                Set_5
*                {
*                  int 1 50
*                }
*                int i_2 16
*                enum e_1
*                string s_1 happy
*                float f_1 1.4
*                logical l_1 T
*              }
*             
*              Set_2
*              {
*                int i 20
*              }
*              
*              Set_3
*              {
*                int i 30
*              }
*
*              // end example *.par file
*
*              DEFINITIONS:
*              -Top-Level Set: A named, braced collection of primitives and/or
*                                embedded sets that is not embedded in another
*                                set
*              -Embedded Set:  A named, braced collection of primitives and/or
*                                embedded sets that is embedded in another set
*              -Reference Set: An alias of (ie, a specifier to include) a
*                                top-level set embedded in either a top-level
*                                set or an embedded set
*              -Set Group:     A collection of embedded sets and/or reference
*                                sets nested one level into a top-level set or
*                                an embedded set unified by a common group
*                                name for the purpose of accessing each set in
*                                the group in turn via a loop control structure
*              -Group Name:    The string tag that unifies a set group
*              -Primitive:     A named value of one of the fundamental units
*                                of data extractable by the SpDataParser,
*                                consisting of one of the following types:
*                                integer, enumeration, logical, float, string,
*                                or define
*
*              KEYWORDS:
*
*              -"reference":   specifies a set as a reference set, followed
*                                by a group name, followed by a top-level set
*                                name
*              -"int":         specifies a integer primitive, followed by a
*                                variable name, followed by a value 
*              -"enum":        specifies a enumeration primitive, followed by
*                                a variable name, followed by a value 
*              -"logical":     specifies a logical primitive, followed by a
*                                variable name, followed by a value 
*              -"float":       specifies a floating point primitive, followed
*                                by a variable name, followed by a value 
*              -"string":      specifies a string primitive, followed by a
*                                variable name, followed by a value 
*              -"define":      specifies a define primitive, followed by a
*                                variable name, followed by a value 
*
*              OVERVIEW:
*              Methods are provided to move in and out of top-level
*              sets (such as "Set_1" above), embedded sets (such as
*              "Set_4" above), and reference sets (such as "Set_2"
*              nested in "Set_1" above). The set to which the user has
*              most recently moved is called the current set. Get*()
*              methods are provided to extract primitive data (such as
*              ints, floats, etc.) from the current set. Sets are
*              reference sets when preceded by two words: the key word
*              "reference" and a group name. Reference sets are nested
*              sets that are actually an inline inclusion of a
*              top-level set of the same set name. Sets may be grouped
*              together for the user to loop through them by having a
*              common group name (such as "Group_1" in the above
*              example). Sets that are members of a group may be
*              reference sets (such as "Set_2" nested in "Set_1"
*              above) or embedded sets (such as "Set_4" above). When
*              looping through a group (call it Group_1) nested one
*              level into a set (call it Set_1), the current set
*              changes each iteration of the loop to the nested set in
*              Group_1 currently arrived at by the loop. This set be
*              inline or a "reference" to a top-level set. At any
*              point during set_a's loop through Group_1, the user may
*              MoveIn () to the current set in Group_1, perform any
*              Get*()'s, loop through groups, perform nested
*              MoveIn()'s and MoveOut()'s, and then finally MoveOut ()
*              back to Set_1 and continue where Set_1 left off in its
*              loop through Group_1.
* Implementation: The SpDataParser handles such nested navigation
*                 through a *.par file through internally keeping
*                 track of a stack of pointers to sets. It "pushes" a
*                 set pointer onto this stack when the user calls
*                 MoveIn() or MoveToFirstSetInGroup(), "pops" set
*                 pointer(s) off this stack when the user calls
*                 MoveOut () or MoveOutCompletely (), and both "pops"
*                 and "pushes" a set pointer when the user calls
*                 MoveToNextSetInGroup ().
*****************************************************************************/
class SpDataParser {

 public:

/******************************************************************************
* Constructor Method: SpDataParser
* Summary: Construct a SpDataParser with the file it is to parse
* Description: The data parser must always be constructed with the file it is
*              to parse. Thus each SpDataParser is designed to parse one and
*              only one file. To parse multiple files, simply create multiple
*              DATA_PARSERs.
******************************************************************************/
   SpDataParser(const char *fileName // file to be parsed
	       );

/******************************************************************************
* Method: MoveIn
* Summary: Move into a top-level set (if not currently in a set) or
*          into a set nested in the current set. The set moved to is
*          called the current set, in which subsequant data extraction
*          is possible.
* Description: MoveIn () is used to move to a set (with or without a
*              reference name or the keyword "reference") that is
*              nested in the current set (ie, the set to which the
*              user last moved). MoveIn () must be called at least
*              once (ie, to move to a top-level set and then move into
*              nested sets as desired) before any data extraction is
*              possible. MoveIn (), MoveOut (), MoveOutCompletely (),
*              MoveToFirstSetInGroup (), and MoveToNextSetInGroup () are
*              the five recommended methods to use for navigating
*              through the nested data structures in a *.par file.
* Return: Success flag
******************************************************************************/
   int MoveIn (char *setName // the name of the set into which to move
	       );

/******************************************************************************
* Method: MoveOut
* Summary: Move out of a set to the set in which it is nested,
*          optionally multiple times
* Description: MoveOut () changes the current set to be the set in
*              which the cureent set is nested. MoveIn (), MoveOut (),
*              MoveOutCompletely (), MoveToFirstSetInGroup (), and
*              MoveToNextSetInGroup () are the five recommended methods
*              to use for navigating through the nested data
*              structures in a *.par file.
* Return: None
******************************************************************************/
   void MoveOut (int numLevels = 1 // the number of MoveOuts to execute
		 );

/******************************************************************************
* Method: MoveOutCompletely
* Summary: Jump out of the arbitrarily deeply nested current set, such
*          that the next call to MoveIn () moves to a top-level set.
* Description: MoveOutCompletely () is equivalent to calling MoveOut
*              () multiple times until no longer in any sets (ie,
*              reseting the SpDataParser to the initial condition where
*              there is no current set). Thus, the user must call
*              MoveIn () with a top-level set as the argument just
*              after calling MoveOutCompletely () in order to do any
*              subsequent data extraction. MoveIn (), MoveOut (),
*              MoveOutCompletely (), MoveToFirstSetInGroup (), and
*              MoveToNextSetInGroup () are the five recommended methods
*              to use for navigating through the nested data
*              structures in a *.par file.
* Return: None
******************************************************************************/
   void MoveOutCompletely ();

/******************************************************************************
* Method: GetNumSetsInGroup
* Summary: Get the number of sets belonging to the specified set group
*          nested one level in from the current set. This method does
*          not change which set is the current set.
* Description: This method is normally used just before looping
*              through a set group so as to know how many sets to loop
*              through.
* Return: The number of sets belonging to a reference name group nested one
*         level in from the current set
******************************************************************************/
   int GetNumSetsInGroup(char *groupName // group name of the set group
			                 // of interest
			);

/******************************************************************************
* Method: MoveToFirstSetInGroup
* Summary: Moves into the first set of the specified set group. The
*          set to which the SpDataParser moves becomes the current set.
* Description: If the first set in the specified set group is an
*              inline set, the SpDataParser will move one nested level
*              into that set, which becomes the current set. On the
*              other hand, if the first set in the specified set group
*              is a reference, the SpDataParser will actually move to
*              the top-level set referenced and make that set the
*              current set. MoveIn (), MoveOut (), MoveOutCompletely
*              (), MoveToFirstSetInGroup (), and MoveToNextSetInGroup
*              () are the five recommended methods to use for
*              navigating through the nested data structures in a
*              *.par file.
* Return: 1=success, 0=fail
******************************************************************************/
   int MoveToFirstSetInGroup(char *groupName // group name of the set group
			                     // of interest
			     );

/******************************************************************************
* Method: MoveToNextSetInGroup
* Summary: Moves into the next set of the set group specified by the
*          last call to MoveToFirstSetInGroup () at this level in the
*          nested structure. The set to which the SpDataParser moves
*          becomes the current set.
* Description: If the next set in the set group specified by the last
*              call to MoveToFirstSetInGroup () is an inline set, the
*              SpDataParser will move out of the current set and then
*              into that inline set, which becomes the current set. On
*              the other hand, if the next set is a reference, the
*              SpDataParser will move out of the current set and then
*              actually move to the top-level set referenced and make
*              that set the current set. MoveIn (), MoveOut (),
*              MoveOutCompletely (), MoveToFirstSetInGroup (), and
*              MoveToNextSetInGroup () are the five recommended
*              methods to use for navigating through the nested data
*              structures in a *.par file.
* Return: 1=success, 0=fail (no more sets in group)
******************************************************************************/
   int MoveToNextSetInGroup();

/******************************************************************************
* Method: GetSetName
* Summary: Gets the name of the current set
* Description: 
* Return: The name of the current set
******************************************************************************/
    char *GetSetName();
    int CheckStack();

/******************************************************************************
* Method: GetArraySize
* Summary: Gets the size of the array of a primitive, specified by its
*          name, nested one level inside of the current set, without
*          changing which set is the current set
* Description: If the specified primitive is not an array, this results
*              in a terminal error
* Return: The size of the array of a primitive, specified by its name, nested
*         one level inside of the current set
******************************************************************************/
   int GetArraySize(char *Name);

/******************************************************************************
* Method: GetInt
* Summary: Gets the value of an integer primitive, specified by its
*          name, nested one level inside of the current set, without
*          changing which set is the current set
* Description: If the specified primitive is not an integer, this results
*              in a terminal error
* Return: The value of an integer primitive, specified by its name, nested
*         one level inside of the current set
******************************************************************************/
   int GetInt(char *name, int Index=-1);

/******************************************************************************
* Method: GetEnum
* Summary: Gets the value of an enumeration primitive, specified by
*          its name, nested one level inside of the current set,
*          without changing which set is the current set
* Description: If the specified primitive is not an enumeration, this results
*              in a terminal error
* Return: The value of an enumeration primitive (the integral order of this
*         primitve starting at 0 in the current set of enumerations),
*         specified by its name, nested one level inside of the current set
******************************************************************************/
   int GetEnum(char *name, int Index=-1);

/******************************************************************************
* Method: GetLogical
* Summary: Gets the value of a logical primitive, specified by its
*          name, nested one level inside of the current set, without
*          changing which set is the current set
* Description: If the specified primitive is not a logical, this results
*              in a terminal error
* Return: The value of a logical primitive (0 for f or false, 1 for t or
*         true), specified by its name, nested one level inside of the
*         current set
******************************************************************************/
   int GetLogical(char *name, int Index=-1);

/******************************************************************************
* Method: GetFloat
* Summary: Gets the value of a float (floating point) primitive,
*          specified by its name, nested one level inside of the
*          current set, without changing which set is the current set
* Description: If the specified primitive is not a float, this results
*              in a terminal error
* Return: The value of a float primitive, specified by its name, nested
*         one level inside of the current set
******************************************************************************/
   double GetFloat(char *name, int Index=-1);

/******************************************************************************
* Method: GetString
* Summary: Gets the value of a string primitive, specified by its
*          name, nested one level inside of the current set, without
*          changing which set is the current set
* Description: If the specified primitive is not a string, this results
*              in a terminal error
* Return: The value of a string primitive, specified by its name, nested
*         one level inside of the current set
******************************************************************************/
   char *GetString(char *name, int Index=-1);

/******************************************************************************
* Method: GetDefine
* Summary: Gets the value of a define primitive, specified by its
*          name, nested one level inside of the current set, without
*          changing which set is the current set
* Description: If the specified primitive is not a define, this results
*              in a terminal error
* Return: The value of a define primitive (the string following the define
*         name), specified by the define name, nested one level inside of the
*         current set
******************************************************************************/
   char *GetDefine(char *name, int Index=-1);

/******************************************************************************
* Method: print
* Summary: Print this SpDataParser's *.par file with appropriate indentation
* Description: 
* Return: None
******************************************************************************/
   void print(ostream &out = cout);

/******************************************************************************
* Internal Method: *GoToType
* Summary: 
* Description: 
* Return: None
******************************************************************************/
   void *GoToType(char *TypeName)
   {
     return GoTo(TypeName, NULL);
   }

/******************************************************************************
* Internal Method: GoTo
* Summary: 
* Description: 
* Return: None
******************************************************************************/
   void *GoTo(char *FirstName, ...);

/******************************************************************************
* Internal Method: GetMark
* Summary: 
* Description: 
* Return: None
******************************************************************************/
   void *GetMark() {return GoToBDT;}

/******************************************************************************
* Internal Method: SetMark
* Summary: 
* Description: 
* Return: None
******************************************************************************/
   void SetMark(void *Mark) {GoToBDT = (BASE_DATA_TYPE *)Mark;}

/******************************************************************************
* Internal Method: GetNtypes
* Summary: 
* Description: 
* Return: None
******************************************************************************/
   int GetNtypes() {return Qtypes.get_length();}

/******************************************************************************
* Internal Method: GetFirstType
* Summary: 
* Description: 
* Return: None
******************************************************************************/
   void *GetFirstType() 
   {
     GoToBDT = (BASE_DATA_TYPE *)Qtypes.get_top();
     return GoToBDT;
   }

/******************************************************************************
* Internal Method: GetNextType
* Summary: 
* Description: 
* Return: None
******************************************************************************/
   void *GetNextType(void *current) 
   {
     if (current == Qtypes.get_bot()) {
       GoToBDT = NULL;
     }else{
       GoToBDT = (BASE_DATA_TYPE *) ((C_ITEM *)current)->get_link();
     }
     return GoToBDT;
   }

/******************************************************************************
* Internal Method: GetNumElements
* Summary: 
* Description: 
* Return: None
******************************************************************************/
   int GetNumElements(void *Mark, char *ReferenceName=NULL);

/******************************************************************************
* Internal Method: GoToFirstElement
* Summary: 
* Description: 
* Return: None
******************************************************************************/
   void *GoToFirstElement (void *Mark, char *ReferenceName=NULL);

/******************************************************************************
* Internal Method: GoToNextElement
* Summary: 
* Description: 
* Return: None
******************************************************************************/
   void  *GoToNextElement (void *Mark, char *ReferenceName=NULL);

/******************************************************************************
* Internal Method: Find
* Summary: 
* Description: 
* Return: None
******************************************************************************/
   void *Find(void *Mark, char *ReferenceName);

/******************************************************************************
* Internal Method: GetNumReferences
* Summary: 
* Description: 
* Return: None
******************************************************************************/
   int GetNumReferences(void *Mark, char *ReferenceName)
   {
     return GetNumElements(Mark, ReferenceName);
   }

/******************************************************************************
* Internal Method: GoToFirstReference
* Summary: 
* Description: 
* Return: None
******************************************************************************/
   void *GoToFirstReference(void *Mark, char *ReferenceName);

/******************************************************************************
* Internal Method: GoToNextReference
* Summary: 
* Description: 
* Return: None
******************************************************************************/
   void *GoToNextReference(void *Mark, char *ReferenceName);

/******************************************************************************
* Internal Method: GetReferenceName
* Summary: 
* Description: 
* Return: None
******************************************************************************/
   char *GetReferenceName();

/******************************************************************************
* Internal Method: GetType
* Summary: 
* Description: 
* Return: None
******************************************************************************/
   char *GetType();

/******************************************************************************
* Internal Method: GetNdefines
* Summary: 
* Description: 
* Return: None
******************************************************************************/
   int GetNdefines() {return GetNtypes("define");}

/******************************************************************************
* Internal Method: GetFirstDefine
* Summary: 
* Description: 
* Return: None
******************************************************************************/
   void GetFirstDefine(char *&name, char *&value);

/******************************************************************************
* Internal Method: GetNextDefine
* Summary: 
* Description: 
* Return: None
******************************************************************************/
   void GetNextDefine(char *&name, char *&value);

/******************************************************************************
* Internal Method: GetNtypes
* Summary: 
* Description: 
* Return: None
******************************************************************************/
   int GetNtypes(char *Type);

   ~SpDataParser();

 private:

   char *FileName;
   C_QUEUE Qtypes;
   C_QUEUE Qatoms;
   BASE_DATA_TYPE *GoToBDT;  // Current Mark
   DEFINE_DATA_TYPE *DefPtr;
   C_QUEUE markStack;
   void *setMark; // Current set in which references are being looped
   char setMarkRefName[128]; // Current set's reference name being looped
};

typedef SpDataParser DATA_PARSER;

#endif
