// SpEmHostUser.C file that implements the Host User class for external modules

#include "SpIostream.H"
#include "SpFstream.H"
#include "SpStrstream.H"
#include "SpIomanip.H"
#include "SpStdlib.H"
#include "SpFloat.H"

#include "SpEmHostUser.H"
#include "SpServer.H"
#include "SpNameHandle.H"
#include "SpMsgHolder.H"
#include "SpSubscriptionItem.H"
#include "SpEndianNature.H"

/*
*****************************************************************************
* SpEmHostUser no-argument constructor
*****************************************************************************
*/

SpEmHostUser::SpEmHostUser() {
  SpEndianNature::DetermineEndianNature();
  SpDataParser Parser("speedes.par");

  Parser.GoTo("SpeedesServer", NULL);
  GroupId = Parser.GetInt("Group");

  Parser.GoTo("SpeedesServer", "HostRouter", NULL);

  if (Parser.GetLogical("UseDefaults")) {
    Parser.GoTo("SpeedesServer", NULL);
    ClientSock = new CLIENT_SOCK(Parser.GetString("DefaultMachineName"),
				 Parser.GetInt("DefaultPort"));
  }else{
    ClientSock = new CLIENT_SOCK(Parser.GetString("MachineName"),
				 Parser.GetInt("Port"));
  }

  InputFileFlag = 0;
  OutputFileFlag = 0;
  Time = -1.0e20;
  DisconnectBarrierFlag = 0;
  SpeedesExecutingFlag = 1;

  ExternalModuleConnection();
}

/*
*****************************************************************************
* SpEmHostUser constructor with Data Parser as argument
*****************************************************************************
*/

SpEmHostUser::SpEmHostUser(SpDataParser *Parser) {
  SpEndianNature::DetermineEndianNature();

  Parser->GoTo("SpeedesServer", NULL);
  GroupId = Parser->GetInt("Group");

  Parser->GoTo("SpeedesServer", "HostRouter", NULL);

  if (Parser->GetLogical("UseDefaults")) {
    Parser->GoTo("SpeedesServer", NULL);
    ClientSock = new CLIENT_SOCK(Parser->GetString("DefaultMachineName"),
				 Parser->GetInt("DefaultPort"));
  }else{
    ClientSock = new CLIENT_SOCK(Parser->GetString("MachineName"),
				 Parser->GetInt("Port"));
  }

  InputFileFlag = 0;
  OutputFileFlag = 0;
  Time = -1.0e20;
  DisconnectBarrierFlag = 0;
  SpeedesExecutingFlag = 1;

  ExternalModuleConnection();
}

/*
*****************************************************************************
* SpEmHostUser constructor with Machine Name and Port as arguments
*****************************************************************************
*/

SpEmHostUser::SpEmHostUser(char *MachineName, int Port, int Group) {
  SpEndianNature::DetermineEndianNature();
  GroupId = Group;
  ClientSock = new CLIENT_SOCK(MachineName, Port);
  InputFileFlag = 0;
  OutputFileFlag = 0;
  Time = -1.0e20;
  DisconnectBarrierFlag = 0;
  SpeedesExecutingFlag = 1;

  ExternalModuleConnection();
}

/*
*****************************************************************************
* SpEmHostUser constructor with File Name as argument
*****************************************************************************
*/

SpEmHostUser::SpEmHostUser(char *FileName) {
  SpEndianNature::DetermineEndianNature();
  GroupId = -1;
  fp_input = fopen(FileName, "rb");
  InputFileFlag = 1;
  OutputFileFlag = 0;
  Time = -1.0e20;
  DisconnectBarrierFlag = 0;
  SpeedesExecutingFlag = 1;
}

/*
*****************************************************************************
* SpEmHostUser destructor
*****************************************************************************
*/

SpEmHostUser::~SpEmHostUser() {
  DisconnectBarrier();
  delete ClientSock;
  if (InputFileFlag) fclose(fp_input);
  if (OutputFileFlag) fclose(fp_output);
}

/*
*****************************************************************************
* SpEmHostUser::RecordInputMessages() method allows received messages to
* automatically be saved to file. They can later be played back using the
* File Name HOST_USER constructor.
*****************************************************************************
*/

void SpEmHostUser::RecordInputMessages(char *FileName) {
  fp_output = fopen(FileName, "wb");
  OutputFileFlag = 1;
}

/*
*****************************************************************************
* SpEmHostUser::GetMessage() method with no arguments returns a message from
* the host router if there is one. Otherwise, NULL is returned.
*****************************************************************************
*/

void SpEmHostUser::ReadMessages() {
  SpMsgHolder *MsgHolder;
  GENERIC_MSG_HEADER *Message;

/*
** Read all incoming messages until there are no more
*/

  for (;;) {
    if (InputFileFlag == 1){
      Message = GetSpeedesMessageFromFile();
    }
    else{
      Message = ClientSock->getMsg();
      if (OutputFileFlag && Message){
	WriteSpeedesMessageToFile(Message);
      }
    }
    if (Message) {

      if (Message->handlerNum == KILL_SPEEDES_MESSAGE_ID) {
	SpeedesExecutingFlag = 0;
      }

      MsgHolder = new SpMsgHolder(Message);
      MessageQueue.push_bot(MsgHolder);
    }else{
      break;
    }
  }

}

/*
*****************************************************************************
* SpEmHostUser::GetMessage() method with no arguments returns a message from
* the host router if there is one. Otherwise, NULL is returned.
*****************************************************************************
*/

GENERIC_MSG_HEADER *SpEmHostUser::GetMessage() {
  SpMsgHolder *MsgHolder;
  GENERIC_MSG_HEADER *Message;

/*
** First see if there are any messages queued up. Return the first one if
** are any messages queued.
*/

  if (MessageQueue.get_length()) {

    MsgHolder = (SpMsgHolder *)MessageQueue.pop_top();
    Message = MsgHolder->GetHolderMessage();
    delete MsgHolder;

    if (Message->handlerNum == KILL_SPEEDES_MESSAGE_ID) {
      SpeedesExecutingFlag = 0;
    }

    return Message;
  }

/*
** Try to read a message and return it if there is one
*/

  if (InputFileFlag == 1){
    Message = GetSpeedesMessageFromFile();
  }
  else{
    Message = ClientSock->getMsg();
    if (OutputFileFlag && Message){
      WriteSpeedesMessageToFile(Message);
    }
  }

  if (Message->handlerNum == KILL_SPEEDES_MESSAGE_ID) {
    SpeedesExecutingFlag = 0;
  }

  return Message;

}

/*
*****************************************************************************
* SpEmHostUser::GetMessage() method with no arguments returns a message from
* the host router if there is one. Otherwise, NULL is returned.
*****************************************************************************
*/

GENERIC_MSG_HEADER *SpEmHostUser::GetMessage(int Type) {
  SpMsgHolder *MsgHolder;
  GENERIC_MSG_HEADER *Message;
  int i;
  int len;
  C_QUEUE Q;

/*
** First look through already received messages (if any) and return the first
** one that matches the same Type.
*/

  len = MessageQueue.get_length();
  for (i=0; i<len; i++) {
    MsgHolder = (SpMsgHolder *)MessageQueue.pop_top();
    if (MsgHolder->CompareType(Type)) {
      Q.concat(&MessageQueue);
      MessageQueue = Q;
      Message = MsgHolder->GetHolderMessage();
      delete MsgHolder;

      if (Message->handlerNum == KILL_SPEEDES_MESSAGE_ID) {
	SpeedesExecutingFlag = 0;
      }

      return Message;
    }else{
      Q.push_bot(MsgHolder);
    }
  }

  MessageQueue = Q;

/*
** Now try to read incoming messages. Queue up the messages if they are not
** of the right Type. Return once one of the right type has been read, or if
** there are no more messages.
*/

  for (;;) {
    if (InputFileFlag == 1){
      Message = GetSpeedesMessageFromFile();
    }
    else{
      Message = ClientSock->getMsg();
      if (OutputFileFlag && Message){
	WriteSpeedesMessageToFile(Message);
      }
    }
    if (Message) {
      if (Message->handlerNum == Type) {

	if (Message->handlerNum == KILL_SPEEDES_MESSAGE_ID) {
	  SpeedesExecutingFlag = 0;
	}

	return Message;
      }else{
	MsgHolder = new SpMsgHolder(Message);
	MessageQueue.push_bot(MsgHolder);
      }
    }else{
      return NULL;
    }
  }

}

GENERIC_MSG_HEADER *SpEmHostUser::GetMessageBlocking() {
  GENERIC_MSG_HEADER * retval;
  if (InputFileFlag == 1){
    retval = GetSpeedesMessageFromFile();
  }
  else{
    retval = ClientSock->getMsg_blk();
    if (OutputFileFlag && retval){
      WriteSpeedesMessageToFile(retval);
    }
  }
  return retval;
}

/*
*****************************************************************************
* SpEmHostUser::ReceiveData() method returns data that was sent from SPEEDES
* using the SendData method of the HostUser.
*****************************************************************************
*/

char *SpEmHostUser::ReceiveData(double &time, int &bytes, int &globalId) {

  SPEEDES_TO_EXTERNAL_MODULE_MESSAGE *Message;
  char *Data;

  ReadMessages();

  Message = (SPEEDES_TO_EXTERNAL_MODULE_MESSAGE *)
    GetMessage(SPEEDES_TO_EXTERNAL_MODULE_MESSAGE_ID);

  if (Message) {
    int numBytes = Message->bytes;
    Data = new char[numBytes];
    memcpy(Data, (char *)&Message[1], Message->bytes);
    time = Message->Time;
    bytes = Message->bytes;
    globalId = Message->GlobalId;
    delete [] (char *)Message;
    return Data;
  }else{
    globalId = 0;
    bytes = 0;
    time = 0.0;
    return NULL;
  }

}

/*
*****************************************************************************
* SpEmHostUser::ExternalModuleConnection() method must be called by external
* modules to establish hybrid synchronization methods
*****************************************************************************
*/

void SpEmHostUser::ExternalModuleConnection() {
  EXTERNAL_MODULE_CONNECTION_MESSAGE ConnectionMessage;

  ConnectionMessage.serviceType = HOST_ROUTER_ID;
  ConnectionMessage.group = GroupId;
  ConnectionMessage.handlerNum = EXTERNAL_MODULE_CONNECTION_MESSAGE_ID;
  ConnectionMessage.bytes = sizeof(EXTERNAL_MODULE_CONNECTION_MESSAGE);

  ClientSock->sendMsg_blk(&ConnectionMessage,
			  sizeof(EXTERNAL_MODULE_CONNECTION_MESSAGE));

  EXTERNAL_MODULE_CONNECTION_MESSAGE* ReturnMessage =
    (EXTERNAL_MODULE_CONNECTION_MESSAGE *)GetMessageBlocking();

  if (ReturnMessage->handlerNum == KILL_SPEEDES_MESSAGE_ID) {
    SpeedesExecutingFlag = 0;
  }

  if (ReturnMessage->handlerNum == EXTERNAL_MODULE_CONNECTION_MESSAGE_ID) {
    delete [] ReturnMessage;
  }else{
    cout << "Error, bad external module connection return message" << endl;
  }

}

/*
*****************************************************************************
* SpEmHostUser::DisconnectBarrier() method for disconnecting barriers that
* may have been set up in SPEEDES
*****************************************************************************
*/

void SpEmHostUser::DisconnectBarrier() {
  DISCONNECT_BARRIER_MESSAGE DisconnectBarrierMessage;

  if (DisconnectBarrierFlag) {
    DisconnectBarrierFlag = 0;
    DisconnectBarrierMessage.serviceType = HOST_ROUTER_ID;
    DisconnectBarrierMessage.group = GroupId;
    DisconnectBarrierMessage.handlerNum = DISCONNECT_BARRIER_MESSAGE_ID;

    if (InputFileFlag != 1){
      DisconnectBarrierMessage.bytes = sizeof(DISCONNECT_BARRIER_MESSAGE);
      ClientSock->sendMsg_blk(&DisconnectBarrierMessage,
			      sizeof(DISCONNECT_BARRIER_MESSAGE));
    }
  }

}

/*
*****************************************************************************
* SpEmHostUser::NameLookup() method takes a string name argument and finds out
* the global Id of the object. The global Id can be used for directing
* messages to objects in SPEEDES.
*****************************************************************************
*/

int SpEmHostUser::NameLookup(char *ObjectName) {
  SpMsgHolder *MsgHolder;
  int GlobalId = -1;
  int size = sizeof(NAME_LOOKUP_MESSAGE) + strlen(ObjectName) + 1;
  char *buff = new char[size];
  NAME_LOOKUP_MESSAGE *NameLookupMessage = (NAME_LOOKUP_MESSAGE *)buff;
  char *buff1 = (char *)&NameLookupMessage[1];
  memcpy(buff1, ObjectName, strlen(ObjectName) + 1);

  if (!InputFileFlag) {
    NameLookupMessage->serviceType = HOST_ROUTER_ID;
    NameLookupMessage->group = GroupId;
    NameLookupMessage->handlerNum = NAME_LOOKUP_MESSAGE_ID;
    NameLookupMessage->bytes = size;

    if (InputFileFlag != 1){
      ClientSock->sendMsg_blk(NameLookupMessage, size);
    }

    for (;;) {
      NAME_LOOKUP_MESSAGE *ReturnMessage =
	(NAME_LOOKUP_MESSAGE *)GetMessageBlocking();

      if (ReturnMessage->handlerNum == KILL_SPEEDES_MESSAGE_ID) {
	SpeedesExecutingFlag = 0;
      }

      if (ReturnMessage->handlerNum == NAME_LOOKUP_MESSAGE_ID) {
	GlobalId = ReturnMessage->GlobalId;
	delete [] (char *)ReturnMessage;
	break;
      }else{
	MsgHolder = new SpMsgHolder(ReturnMessage);
	MessageQueue.push_bot(MsgHolder);
      }
    }
  }

  delete [] buff;
  return GlobalId;

}

/*
*****************************************************************************
* SpEmHostUser::ObjectTypes() method returns a list of the names of object
* types in a SPEEDES simulation.
*****************************************************************************
*/

SpNameValueList *SpEmHostUser::ObjectManagerTypes() {
  OBJECT_TYPE_MESSAGE ObjectTypeMessage;
  SpNameValueList *NameValueList;
  SpMsgHolder *MsgHolder;

  cout<<"ObjectManagerTypes()"<<endl;
  ObjectTypeMessage.serviceType = HOST_ROUTER_ID;
  ObjectTypeMessage.group = GroupId;
  ObjectTypeMessage.handlerNum = OBJECT_TYPE_MESSAGE_ID;
  ObjectTypeMessage.bytes = sizeof(OBJECT_TYPE_MESSAGE);

  if (!InputFileFlag) {
    ClientSock->sendMsg_blk(&ObjectTypeMessage, sizeof(OBJECT_TYPE_MESSAGE));
  }
  for (;;) {

    OBJECT_TYPE_MESSAGE *ReturnMessage =
      (OBJECT_TYPE_MESSAGE *)GetMessageBlocking();

    if (ReturnMessage->handlerNum == KILL_SPEEDES_MESSAGE_ID) {
      SpeedesExecutingFlag = 0;
    }

    if (ReturnMessage->handlerNum == OBJECT_TYPE_MESSAGE_ID) {

      if (ReturnMessage->bytes < 0) {
	cout << "Error, bad object type response" << endl;
	delete [] (char *)ReturnMessage;
	break;
      }else{
	NameValueList = new SpNameValueList((char *)&ReturnMessage[1]);
	delete [] (char *)ReturnMessage;
	NameValueList->Sort();
	return NameValueList;
      }

    }else{

      MsgHolder = new SpMsgHolder(ReturnMessage);
      MessageQueue.push_bot(MsgHolder);

    }
  }
  return NULL;
}

/*
*****************************************************************************
* SpEmHostUser::ObjectType() method returns the id of an object manager type
* in SPEEDES.
*****************************************************************************
*/

int SpEmHostUser::ObjectManagerType(char *Type) {
  int NameFound;

  SpNameValueList *NameValueList = ObjectManagerTypes();

  int Value = int(NameValueList->GetValue(Type, NameFound));
  if (!NameFound) Value = -1;

  delete NameValueList;
  return Value;

}


/*
*****************************************************************************
* SpEmHostUser::GetNumNodes() method returns the number of nodes in SPEEDES.
*****************************************************************************
*/

int SpEmHostUser::GetNumNodes() {
  NUM_NODES_MESSAGE NumNodesMessage;
  SpMsgHolder *MsgHolder;

  NumNodesMessage.serviceType = HOST_ROUTER_ID;
  NumNodesMessage.group = GroupId;
  NumNodesMessage.handlerNum = NUM_NODES_MESSAGE_ID;
  NumNodesMessage.bytes = sizeof(NUM_NODES_MESSAGE);

  if (!InputFileFlag) {
    ClientSock->sendMsg_blk(&NumNodesMessage, sizeof(NUM_NODES_MESSAGE));
  }

  for (;;) {

    NUM_NODES_MESSAGE *ReturnMessage =
      (NUM_NODES_MESSAGE *)GetMessageBlocking();

    if (ReturnMessage->handlerNum == KILL_SPEEDES_MESSAGE_ID) {
      SpeedesExecutingFlag = 0;
      return 0;
    }

    if (ReturnMessage->handlerNum == NUM_NODES_MESSAGE_ID) {

      if (ReturnMessage->bytes < 0) {
	cout << "Error, bad GetNumNodes response" << endl;
	delete [] (char *)ReturnMessage;
	return 0;
      }else{
	int NumNodes = ReturnMessage->NumNodes;
	delete [] (char *)ReturnMessage;
	return NumNodes;
      }

    }else{

      MsgHolder = new SpMsgHolder(ReturnMessage);
      MessageQueue.push_bot(MsgHolder);

    }

  }

}

/*
*****************************************************************************
* SpEmHostUser::GetNumNodes() method returns the number of nodes in SPEEDES.
*****************************************************************************
*/

int SpEmHostUser::GetNextNode() {
  NEXT_NODE_MESSAGE NextNodeMessage;
  SpMsgHolder *MsgHolder;

  NextNodeMessage.serviceType = HOST_ROUTER_ID;
  NextNodeMessage.group = GroupId;
  NextNodeMessage.handlerNum = NEXT_NODE_MESSAGE_ID;
  NextNodeMessage.bytes = sizeof(NEXT_NODE_MESSAGE);

  if (!InputFileFlag) {
    ClientSock->sendMsg_blk(&NextNodeMessage, sizeof(NEXT_NODE_MESSAGE));
  }

  for (;;) {

    NEXT_NODE_MESSAGE *ReturnMessage =
      (NEXT_NODE_MESSAGE *)GetMessageBlocking();

    if (ReturnMessage->handlerNum == KILL_SPEEDES_MESSAGE_ID) {
      SpeedesExecutingFlag = 0;
      return 0;
    }

    if (ReturnMessage->handlerNum == NEXT_NODE_MESSAGE_ID) {

      if (ReturnMessage->bytes < 0) {
	cout << "Error, bad GetNextNode response" << endl;
	delete [] (char *)ReturnMessage;
	return 0;
      }else{
	int NextNode = ReturnMessage->NextNode;
	delete [] (char *)ReturnMessage;
	return NextNode;
      }

    }else{

      MsgHolder = new SpMsgHolder(ReturnMessage);
      MessageQueue.push_bot(MsgHolder);

    }

  }

}

/*
*****************************************************************************
* SpEmHostUser::ObjectNames() method takes a string Type argument and returns
* a list of object names in SPEEDES of that type.
*****************************************************************************
*/

SpNameValueList *SpEmHostUser::ObjectNames(int Type) {
  OBJECT_NAME_MESSAGE ObjectNameMessage;
  SpNameValueList *NameValueList;
  SpMsgHolder *MsgHolder;

  if (!InputFileFlag) {
    ObjectNameMessage.serviceType = HOST_ROUTER_ID;
    ObjectNameMessage.group = GroupId;
    ObjectNameMessage.handlerNum = OBJECT_NAME_MESSAGE_ID;

    ObjectNameMessage.ObjectType = Type;
    ObjectNameMessage.NameFlag = 0;
    ObjectNameMessage.bytes = sizeof(OBJECT_NAME_MESSAGE);

    if (InputFileFlag != 1){
      ClientSock->sendMsg_blk(&ObjectNameMessage, sizeof(OBJECT_NAME_MESSAGE));
    }

    for (;;) {

      OBJECT_NAME_MESSAGE *ReturnMessage =
	(OBJECT_NAME_MESSAGE *)GetMessageBlocking();

      if (ReturnMessage->handlerNum == KILL_SPEEDES_MESSAGE_ID) {
	SpeedesExecutingFlag = 0;
      }

      if (ReturnMessage->handlerNum == OBJECT_NAME_MESSAGE_ID) {

	if (ReturnMessage->bytes < 0) {
	  cout << "Error, bad object type response" << endl;
	  delete [] (char *)ReturnMessage;
	  return NULL;
	}else{
	  NameValueList = new SpNameValueList((char *)&ReturnMessage[1]);
	  delete [] (char *)ReturnMessage;
	  NameValueList->Sort();
	  return NameValueList;
	}

      }else{

	MsgHolder = new SpMsgHolder(ReturnMessage);
	MessageQueue.push_bot(MsgHolder);

      }
    }

  }

  return NULL;

}

/*
*****************************************************************************
* SpEmHostUser::ObjectNames() method takes a string Type argument and returns
* a list of object names in SPEEDES of that type.
*****************************************************************************
*/

SpNameValueList *SpEmHostUser::ObjectNames(char *Type) {
  OBJECT_NAME_MESSAGE *ObjectNameMessage;
  SpNameValueList *NameValueList;
  SpMsgHolder *MsgHolder;
  char *buff;
  int Nbytes;

  if (!InputFileFlag) {

    if (Type) {
      Nbytes = sizeof(OBJECT_NAME_MESSAGE) + strlen(Type) + 1;
      buff = new char[Nbytes];
      ObjectNameMessage = (OBJECT_NAME_MESSAGE *)buff;
      strcpy((char *)&ObjectNameMessage[1], Type);
    }else{
      Nbytes = sizeof(OBJECT_NAME_MESSAGE);
      buff = new char[Nbytes];
      ObjectNameMessage = (OBJECT_NAME_MESSAGE *)buff;
    }

    ObjectNameMessage->serviceType = HOST_ROUTER_ID;
    ObjectNameMessage->group = GroupId;
    ObjectNameMessage->handlerNum = OBJECT_NAME_MESSAGE_ID;

    ObjectNameMessage->ObjectType = -1;
    ObjectNameMessage->NameFlag = 1;
    ObjectNameMessage->bytes = Nbytes;

    if (InputFileFlag != 1){
      ClientSock->sendMsg_blk(ObjectNameMessage, Nbytes);
    }
    delete [] buff;

    for (;;) {

      OBJECT_NAME_MESSAGE *ReturnMessage =
	(OBJECT_NAME_MESSAGE *)GetMessageBlocking();

      if (ReturnMessage->handlerNum == KILL_SPEEDES_MESSAGE_ID) {
	SpeedesExecutingFlag = 0;
      }

      if (ReturnMessage->handlerNum == OBJECT_NAME_MESSAGE_ID) {

	if (ReturnMessage->bytes < 0) {
	  cout << "Error, bad object name response" << endl;
	  delete [] (char *)ReturnMessage;
	  return NULL;
	}else{
	  NameValueList = new SpNameValueList((char *)&ReturnMessage[1]);
	  delete [] (char *)ReturnMessage;
	  NameValueList->Sort();
	  return NameValueList;
	}

      }else{

	MsgHolder = new SpMsgHolder(ReturnMessage);
	MessageQueue.push_bot(MsgHolder);

      }
    }

  }

  return NULL;

}

/*
*****************************************************************************
* SpEmHostUser::Flush() method sends a message to the host router and receives
* a message back. It basically pings the host router, but in doing so, it
* guarantees that the host router has read all of the messages previously
* sent.
*****************************************************************************
*/

void SpEmHostUser::Flush() {
  FLUSH_MESSAGE FlushMessage;
  SpMsgHolder *MsgHolder;

  if (!InputFileFlag) {
    FlushMessage.serviceType = HOST_ROUTER_ID;
    FlushMessage.group = GroupId;
    FlushMessage.handlerNum = FLUSH_MESSAGE_ID;
    FlushMessage.bytes = sizeof(FLUSH_MESSAGE);

    if (InputFileFlag != 1){
      ClientSock->sendMsg_blk(&FlushMessage, sizeof(FLUSH_MESSAGE));
    }

    for (;;) {
      FLUSH_MESSAGE *ReturnMessage = (FLUSH_MESSAGE *)GetMessageBlocking();


      if (ReturnMessage->handlerNum == KILL_SPEEDES_MESSAGE_ID) {
	SpeedesExecutingFlag = 0;
      }

      if (ReturnMessage->handlerNum == FLUSH_MESSAGE_ID) {
	delete [] (char *)ReturnMessage;
	break;
      }else{
	MsgHolder = new SpMsgHolder(ReturnMessage);
	MessageQueue.push_bot(MsgHolder);
      }
    }

  }

}

/*
*****************************************************************************
* SpEmHostUser::Pause() method sends a message to SPEEDES to pause the
* simulation. This does not stop processing - it stops gvt from advancing.
*****************************************************************************
*/

void SpEmHostUser::Pause() {
  PAUSE_MESSAGE PauseMessage;

  if (!InputFileFlag) {
    PauseMessage.serviceType = HOST_ROUTER_ID;
    PauseMessage.group = GroupId;
    PauseMessage.handlerNum = PAUSE_MESSAGE_ID;
    PauseMessage.bytes = sizeof(PAUSE_MESSAGE);

    ClientSock->sendMsg_blk(&PauseMessage, sizeof(PAUSE_MESSAGE));
  }

}

/*
*****************************************************************************
* SpEmHostUser::Resume() method sends a message to SPEEDES to resume the
* simulation after it has been paused. This allows gvt to once again advance.
*****************************************************************************
*/

void SpEmHostUser::Resume() {
  RESUME_MESSAGE ResumeMessage;

  if (!InputFileFlag) {
    ResumeMessage.serviceType = HOST_ROUTER_ID;
    ResumeMessage.group = GroupId;
    ResumeMessage.handlerNum = RESUME_MESSAGE_ID;
    ResumeMessage.bytes = sizeof(RESUME_MESSAGE);
    ClientSock->sendMsg_blk(&ResumeMessage, sizeof(RESUME_MESSAGE));
  }

}

/*
*****************************************************************************
* SpEmHostUser::Pause() method sends a message to SPEEDES to pause the
* simulation. This does not stop processing - it stops gvt from advancing.
*****************************************************************************
*/

void SpEmHostUser::Pause(char *pauseName, double time) {
  NAMED_PAUSE_MESSAGE *NamedPauseMessage;
  int bytes;

  bytes = sizeof(NAMED_PAUSE_MESSAGE) + strlen(pauseName) + 1;
  NamedPauseMessage = (NAMED_PAUSE_MESSAGE *)(new char[bytes]);

  if (!InputFileFlag) {
    NamedPauseMessage->Time = time;
    strcpy((char *)&NamedPauseMessage[1], pauseName);
    NamedPauseMessage->serviceType = HOST_ROUTER_ID;
    NamedPauseMessage->group = GroupId;
    NamedPauseMessage->handlerNum = NAMED_PAUSE_MESSAGE_ID;
    NamedPauseMessage->bytes = bytes;

    ClientSock->sendMsg_blk(NamedPauseMessage, bytes);
  }

  delete [] (char *)NamedPauseMessage;

}

/*
*****************************************************************************
* SpEmHostUser::Resume() method sends a message to SPEEDES to resume the
* simulation after it has been paused. This allows gvt to once again advance.
*****************************************************************************
*/

void SpEmHostUser::Resume(char *resumeName) {
  NAMED_RESUME_MESSAGE *NamedResumeMessage;
  int bytes;

  bytes = sizeof(NAMED_RESUME_MESSAGE) + strlen(resumeName) + 1;
  NamedResumeMessage = (NAMED_RESUME_MESSAGE *)(new char[bytes]);

  if (!InputFileFlag) {
    strcpy((char *)&NamedResumeMessage[1], resumeName);
    NamedResumeMessage->serviceType = HOST_ROUTER_ID;
    NamedResumeMessage->group = GroupId;
    NamedResumeMessage->handlerNum = NAMED_RESUME_MESSAGE_ID;
    NamedResumeMessage->bytes = bytes;

    ClientSock->sendMsg_blk(NamedResumeMessage, bytes);
  }

  delete [] (char *)NamedResumeMessage;

}


/*
*****************************************************************************
* SpEmHostUser::KillSPEEDES() method sends a message to SPEEDES that kills
* the execution.
*****************************************************************************
*/

void SpEmHostUser::KillSPEEDES() {
  KILL_SPEEDES_MESSAGE KillSPEEDESMessage;

  if (!InputFileFlag) {
    KillSPEEDESMessage.serviceType = HOST_ROUTER_ID;
    KillSPEEDESMessage.group = GroupId;
    KillSPEEDESMessage.handlerNum = KILL_SPEEDES_MESSAGE_ID;
    KillSPEEDESMessage.bytes = sizeof(KILL_SPEEDES_MESSAGE);

    ClientSock->sendMsg_blk(&KillSPEEDESMessage, sizeof(KILL_SPEEDES_MESSAGE));
  }

}

/*
*****************************************************************************
* SpEmHostUser::Query() method sends a query message to an object in SPEEDES
* and returns a SpNameValueList back.
*****************************************************************************
*/

SpNameValueList *SpEmHostUser::Query(int GlobalId, char *queryName) {
  QUERY_MESSAGE *QueryMessage;
  SpMsgHolder *MsgHolder;
  SpNameValueList *NameValueList;
  int bytes;
  char *Buff;

  if (!InputFileFlag) {

    bytes = strlen(queryName) + 1;
    Buff = new char[sizeof(QUERY_MESSAGE) + bytes];
    QueryMessage = (QUERY_MESSAGE *)Buff;

    QueryMessage->serviceType = HOST_ROUTER_ID;
    QueryMessage->group = GroupId;
    QueryMessage->handlerNum = QUERY_MESSAGE_ID;
    QueryMessage->bytes = bytes;

    QueryMessage->GlobalId = GlobalId;

    Buff = (char *)&QueryMessage[1];
    strcpy(Buff, queryName);

    ClientSock->sendMsg_blk(QueryMessage, sizeof(QUERY_MESSAGE) + bytes);

    Buff = (char *)QueryMessage;
    delete [] Buff;

    for (;;) {

      QUERY_REPLY_MESSAGE *ReturnMessage =
	(QUERY_REPLY_MESSAGE *)GetMessageBlocking();

      if (ReturnMessage->handlerNum == KILL_SPEEDES_MESSAGE_ID) {
	SpeedesExecutingFlag = 0;
      }

      if (ReturnMessage->handlerNum == QUERY_REPLY_MESSAGE_ID) {

	if (ReturnMessage->bytes < 0) {
	  cout << "Error, bad query" << endl;
	  delete [] (char *)ReturnMessage;
	  return NULL;
	}else{
	  NameValueList = new SpNameValueList((char *)&ReturnMessage[1]);
	  delete [] (char *)ReturnMessage;
	  return NameValueList;
	}

      }else{

	MsgHolder = new SpMsgHolder(ReturnMessage);
	MessageQueue.push_bot(MsgHolder);

	if (SpeedesExecutingFlag == 0) {
	  cout << "Aborting command - SPEEDES exiting" << endl;
	  return NULL;
	}

      }
    }

  }

  return NULL;

}


/*
*****************************************************************************
* SpEmHostUser::Query() method sends a query message to an object in SPEEDES
* and returns a SpNameValueList back.
*****************************************************************************
*/

SpNameValueList *SpEmHostUser::Query(char *ObjectName, char *queryName) {
  QUERY_MESSAGE *QueryMessage;
  SpMsgHolder *MsgHolder;
  char *Buff;
  SpNameValueList *NameValueList;
  int bytes1;
  int bytes2;

  if (!InputFileFlag) {

    bytes1 = strlen(ObjectName) + 1;
    bytes2 = strlen(queryName) + 1;

    Buff = new char[sizeof(QUERY_MESSAGE) + bytes1 + bytes2];

    QueryMessage = (QUERY_MESSAGE *)Buff;

    QueryMessage->serviceType = HOST_ROUTER_ID;
    QueryMessage->group = GroupId;
    QueryMessage->handlerNum = QUERY_MESSAGE_ID;
    QueryMessage->bytes = bytes1 + bytes2;
    QueryMessage->GlobalId = -1;

    Buff = (char *)&QueryMessage[1];
    strcpy(Buff, ObjectName);

    Buff += bytes1;
    strcpy(Buff, queryName);

    ClientSock->sendMsg_blk(QueryMessage,
			    sizeof(QUERY_MESSAGE) + bytes1 + bytes2);

    Buff = (char *)QueryMessage;
    delete [] Buff;

    for (;;) {
      QUERY_REPLY_MESSAGE *ReturnMessage = (QUERY_REPLY_MESSAGE *)
	GetMessageBlocking();

      if (ReturnMessage->handlerNum == KILL_SPEEDES_MESSAGE_ID) {
	SpeedesExecutingFlag = 0;
      }

      if (ReturnMessage->handlerNum == QUERY_REPLY_MESSAGE_ID) {

	if (ReturnMessage->bytes == 0x7fffffff) {
	  cout << "Error, bad query" << endl;
	  delete [] (char *)ReturnMessage;
	  return NULL;
	}else{
	  NameValueList = new SpNameValueList((char *)&ReturnMessage[1]);
	  delete [] (char *)ReturnMessage;
	  return NameValueList;
	}

      }else{

	MsgHolder = new SpMsgHolder(ReturnMessage);
	MessageQueue.push_bot(MsgHolder);

	if (SpeedesExecutingFlag == 0) {
	  cout << "Aborting command - SPEEDES exiting" << endl;
	  return NULL;
	}

      }
    }

  }

  return NULL;

}



/*
*****************************************************************************
* SpEmHostUser::SendCommand() method sends a command to SPEEDES. In SPEEDES,
* the message will be appended to the SpMsg class.
*****************************************************************************
*/

int SpEmHostUser::SendCommand(char *CommandName,
			      int Node,
			      int ObjectType,
			      int LocalId,
			      SpSimTime EventTime,
			      char *Message,
			      int Bytes,
			      int ReturnStatus) {

  int TotalBytes = sizeof(COMMAND_MESSAGE)
		       + strlen(CommandName) + 1
		       + Bytes;

  char *Buff = new char[TotalBytes];
  COMMAND_MESSAGE *CommandMessage = (COMMAND_MESSAGE *)Buff;

  char *Buff1 = (char *)&CommandMessage[1];

  strcpy(Buff1, CommandName);
  Buff1 += strlen(CommandName) + 1;

  memcpy(Buff1, Message, Bytes);
  Buff1 += Bytes;

  if (Buff1 - Buff != TotalBytes) {
    cout << "Error, bad direct Command Buffer" << endl;
  }

  if (!InputFileFlag) {
    CommandMessage->serviceType = HOST_ROUTER_ID;
    CommandMessage->group = GroupId;
    CommandMessage->handlerNum = COMMAND_MESSAGE_ID;
    CommandMessage->bytes = TotalBytes;

    CommandMessage->Node = Node;
    CommandMessage->ObjectType = ObjectType;
    CommandMessage->LocalId = LocalId;
    CommandMessage->Time = EventTime;
    CommandMessage->GlobalId = -1;
    CommandMessage->ReturnStatus = ReturnStatus;

    ClientSock->sendMsg_blk(CommandMessage, TotalBytes);

    if (ReturnStatus) {
      return ReceiveCommandStatus();
    }

  }

  return COMMAND_STATUS_GOOD;

}

/*
*****************************************************************************
* SpEmHostUser::SendCommand() method sends a command to SPEEDES. In SPEEDES,
* the message will be appended to the SpMsg class.
*****************************************************************************
*/

int SpEmHostUser::SendCommand(char *CommandName,
			      int Node,
			      int ObjectType,
			      int LocalId,
			      char *Message,
			      int Bytes,
			      int ReturnStatus) {

  return this->SendCommand(CommandName, Node, ObjectType, LocalId,
			   -1e20, Message, Bytes, ReturnStatus);
}


/*
*****************************************************************************
* SpEmHostUser::SendCommand() method sends a command to SPEEDES. In SPEEDES,
* the message will be appended to the SpMsg class.
*****************************************************************************
*/

int SpEmHostUser::SendCommand(char *CommandName, int GlobalId,
			       char *Message, int Bytes, int ReturnStatus) {

  int TotalBytes = sizeof(COMMAND_MESSAGE)
		       + strlen(CommandName) + 1
		       + Bytes;

  char *Buff = new char[TotalBytes];
  COMMAND_MESSAGE *CommandMessage = (COMMAND_MESSAGE *)Buff;

  char *Buff1 = (char *)&CommandMessage[1];

  strcpy(Buff1, CommandName);
  Buff1 += strlen(CommandName) + 1;

  memcpy(Buff1, Message, Bytes);
  Buff1 += Bytes;

  if (Buff1 - Buff != TotalBytes) {
    cout << "Error, bad Command Buffer" << endl;
  }

  if (!InputFileFlag) {
    CommandMessage->serviceType = HOST_ROUTER_ID;
    CommandMessage->group = GroupId;
    CommandMessage->handlerNum = COMMAND_MESSAGE_ID;
    CommandMessage->bytes = TotalBytes;
    CommandMessage->ReturnStatus = ReturnStatus;
    CommandMessage->Node = -1;
    CommandMessage->ObjectType = -1;
    CommandMessage->LocalId = -1;

    CommandMessage->GlobalId = GlobalId;

    //geb3 fix
    //CommandMessage->Time = -1e20;

    ClientSock->sendMsg_blk(CommandMessage, TotalBytes);

    if (ReturnStatus) {
      return ReceiveCommandStatus();
    }

  }

  return COMMAND_STATUS_GOOD;

}

/*
*****************************************************************************
* SpEmHostUser::SendCommand() method sends a command to SPEEDES. In SPEEDES,
* the message will be appended to the SpMsg class.
*****************************************************************************
*/

int SpEmHostUser::SendCommand(char *CommandName, char *ObjectName,
			      char *Message, int Bytes, int ReturnStatus) {

  int TotalBytes = sizeof(COMMAND_MESSAGE)
		       + strlen(CommandName) + 1
		       + strlen(ObjectName) + 1
		       + Bytes;

  char *Buff = new char[TotalBytes];
  COMMAND_MESSAGE *CommandMessage = (COMMAND_MESSAGE *)Buff;

  char *Buff1 = (char *)&CommandMessage[1];

  strcpy(Buff1, ObjectName);
  Buff1 += strlen(ObjectName) + 1;

  strcpy(Buff1, CommandName);
  Buff1 += strlen(CommandName) + 1;

  memcpy(Buff1, Message, Bytes);
  Buff1 += Bytes;

  if (Buff1 - Buff != TotalBytes) {
    cout << "Error, bad Command Buffer" << endl;
  }

  if (!InputFileFlag) {
    CommandMessage->serviceType = HOST_ROUTER_ID;
    CommandMessage->group = GroupId;
    CommandMessage->handlerNum = COMMAND_MESSAGE_ID;
    CommandMessage->bytes = TotalBytes;
    CommandMessage->ReturnStatus = ReturnStatus;

    CommandMessage->GlobalId = -1;
    CommandMessage->Node = -1;
    CommandMessage->ObjectType = -1;
    CommandMessage->LocalId = -1;

    //geb3 fix
    //CommandMessage->Time = -1e20;

    ClientSock->sendMsg_blk(CommandMessage, TotalBytes);

    if (ReturnStatus) {
      return ReceiveCommandStatus();
    }

  }

  return COMMAND_STATUS_GOOD;

}


int SpEmHostUser::ReceiveCommandStatus() {
  SpMsgHolder *MsgHolder;

  for (;;) {

    COMMAND_REPLY_MESSAGE *ReturnMessage = (COMMAND_REPLY_MESSAGE *)
      GetMessageBlocking();

    if (ReturnMessage->handlerNum == KILL_SPEEDES_MESSAGE_ID) {
      SpeedesExecutingFlag = 0;
      cout << "Aborting command - SPEEDES exiting" << endl;
      return COMMAND_STATUS_ABORT;
    }

    if (ReturnMessage->handlerNum == COMMAND_REPLY_MESSAGE_ID) {

      int CommandStatus = ReturnMessage->CommandStatus;

      if (CommandStatus == COMMAND_STATUS_BAD_OBJECT) {
	cout << "COMMAND_STATUS_BAD_OBJECT" << endl;
      }

      if (CommandStatus == COMMAND_STATUS_BAD_EVENT) {
	cout << "COMMAND_STATUS_BAD_EVENT" << endl;
      }

      return CommandStatus;

    }else{

      MsgHolder = new SpMsgHolder(ReturnMessage);
      MessageQueue.push_bot(MsgHolder);

      if (SpeedesExecutingFlag == 0) {
	cout << "Aborting command - SPEEDES exiting" << endl;
	return COMMAND_STATUS_ABORT;
      }

    }
  }

}


/*
*****************************************************************************
* SpEmHostUser::SendTwoWayCommand() method sends a command to SPEEDES. In
* SPEEDES, the message will be appended to the SpMsg class.
*****************************************************************************
*/

int SpEmHostUser::SendTwoWayCommand(char *CommandName, int GlobalId,
				     char *Message, int Bytes,
				     int BarrierMode, double TimeLag) {
  SpMsgHolder *MsgHolder;

  DisconnectBarrierFlag = 1;

  int TotalBytes = sizeof(TWO_WAY_COMMAND_MESSAGE)
    + strlen(CommandName) + 1
    + Bytes;

  char *Buff = new char[TotalBytes];
  TWO_WAY_COMMAND_MESSAGE *CommandMessage = (TWO_WAY_COMMAND_MESSAGE *)Buff;

  char *Buff1 = (char *)&CommandMessage[1];

  strcpy(Buff1, CommandName);
  Buff1 += strlen(CommandName) + 1;

  memcpy(Buff1, Message, Bytes);
  Buff1 += Bytes;

  if (Buff1 - Buff != TotalBytes) {
    cout << "Error, bad two-way command buffer" << endl;
  }

  if (!InputFileFlag) {
    CommandMessage->serviceType = HOST_ROUTER_ID;
    CommandMessage->group = GroupId;
    CommandMessage->handlerNum = TWO_WAY_COMMAND_MESSAGE_ID;
    CommandMessage->bytes = TotalBytes;

    CommandMessage->GlobalId = GlobalId;

    CommandMessage->BarrierMode = BarrierMode;
    CommandMessage->TimeLag = TimeLag;

    ClientSock->sendMsg_blk(CommandMessage, TotalBytes);

    for (;;) {
      COMMAND_REPLY_MESSAGE *ReturnMessage = (COMMAND_REPLY_MESSAGE *)
	GetMessageBlocking();

      if (ReturnMessage->handlerNum == KILL_SPEEDES_MESSAGE_ID) {
	SpeedesExecutingFlag = 0;
      }

      if (ReturnMessage->handlerNum == COMMAND_REPLY_MESSAGE_ID) {

	int CommandStatus = ReturnMessage->CommandStatus;

	if (CommandStatus == COMMAND_STATUS_BAD_OBJECT) {
	  cout << "COMMAND_STATUS_BAD_OBJECT" << endl;
	}

	if (CommandStatus == COMMAND_STATUS_BAD_EVENT) {
	  cout << "COMMAND_STATUS_BAD_EVENT" << endl;
	}

	return CommandStatus;

      }else{

	MsgHolder = new SpMsgHolder(ReturnMessage);
	MessageQueue.push_bot(MsgHolder);

	if (SpeedesExecutingFlag == 0) {
	  cout << "Aborting command - SPEEDES exiting" << endl;
	  return COMMAND_STATUS_ABORT;
	}

      }
    }

  }
  return COMMAND_STATUS_ABORT;
}

/*
*****************************************************************************
* SpEmHostUser::SendTwoWayCommand() method sends a command to SPEEDES. In
* SPEEDES, the message will be appended to the SpMsg class.
*****************************************************************************
*/

int SpEmHostUser::SendTwoWayCommand(char *CommandName, char *ObjectName,
				     char *Message, int Bytes,
				     int BarrierMode, double TimeLag) {
  SpMsgHolder *MsgHolder;

  DisconnectBarrierFlag = 1;

  int TotalBytes = sizeof(TWO_WAY_COMMAND_MESSAGE)
    + strlen(CommandName) + 1
    + strlen(ObjectName) + 1
    + Bytes;

  char *Buff = new char[TotalBytes];
  TWO_WAY_COMMAND_MESSAGE *CommandMessage = (TWO_WAY_COMMAND_MESSAGE *)Buff;

  char *Buff1 = (char *)&CommandMessage[1];

  strcpy(Buff1, ObjectName);
  Buff1 += strlen(ObjectName) + 1;

  strcpy(Buff1, CommandName);
  Buff1 += strlen(CommandName) + 1;

  memcpy(Buff1, Message, Bytes);
  Buff1 += Bytes;

  if (Buff1 - Buff != TotalBytes) {
    cout << "Error, bad two-way command Buffer" << endl;
  }

  if (!InputFileFlag) {
    CommandMessage->serviceType = HOST_ROUTER_ID;
    CommandMessage->group = GroupId;
    CommandMessage->handlerNum = TWO_WAY_COMMAND_MESSAGE_ID;
    CommandMessage->bytes = TotalBytes;

    CommandMessage->GlobalId = -1;

    CommandMessage->BarrierMode = BarrierMode;
    CommandMessage->TimeLag = TimeLag;

    ClientSock->sendMsg_blk(CommandMessage, TotalBytes);

    for (;;) {
      COMMAND_REPLY_MESSAGE *ReturnMessage = (COMMAND_REPLY_MESSAGE *)
	GetMessageBlocking();

      if (ReturnMessage->handlerNum == KILL_SPEEDES_MESSAGE_ID) {
	SpeedesExecutingFlag = 0;
      }

      if (ReturnMessage->handlerNum == COMMAND_REPLY_MESSAGE_ID) {

	int CommandStatus = ReturnMessage->CommandStatus;

	if (CommandStatus == COMMAND_STATUS_BAD_OBJECT) {
	  cout << "COMMAND_STATUS_BAD_OBJECT" << endl;
	}

	if (CommandStatus == COMMAND_STATUS_BAD_EVENT) {
	  cout << "COMMAND_STATUS_BAD_EVENT" << endl;
	}

	return CommandStatus;

      }else{

	MsgHolder = new SpMsgHolder(ReturnMessage);
	MessageQueue.push_bot(MsgHolder);

	if (SpeedesExecutingFlag == 0) {
	  cout << "Aborting command - SPEEDES exiting" << endl;
	  return COMMAND_STATUS_ABORT;
	}

      }
    }

  }
  return COMMAND_STATUS_ABORT;
}

/*
*****************************************************************************
* SpEmHostUser::GetTime() method sends a message to SPEEDES to return the
* current gvt value. This is a blocking call.
*****************************************************************************
*/

double SpEmHostUser::GetTime() {
  TIME_MESSAGE TimeMessage;

  if (!InputFileFlag) {

    if (!SpeedesExecutingFlag){
      return 1e20;
    }

    TimeMessage.serviceType = HOST_ROUTER_ID;
    TimeMessage.group = GroupId;
    TimeMessage.handlerNum = TIME_MESSAGE_ID;
    TimeMessage.bytes = sizeof(TIME_MESSAGE);

    ClientSock->sendMsg_blk(&TimeMessage, sizeof(TIME_MESSAGE));

    double return_value;
    for (;;) {

      TIME_REPLY_MESSAGE *ReturnMessage = (TIME_REPLY_MESSAGE *)
	GetMessageBlocking();

      if (ReturnMessage->handlerNum == KILL_SPEEDES_MESSAGE_ID) {
	SpeedesExecutingFlag = 0;
      }

      if (ReturnMessage->handlerNum == TIME_REPLY_MESSAGE_ID) {
	Time = ReturnMessage->Time;
	return_value = Time;
	delete [] (char*)ReturnMessage;
	break;
      }else{
	SpMsgHolder *MsgHolder = new SpMsgHolder(ReturnMessage);
	MessageQueue.push_bot(MsgHolder);

	if (!SpeedesExecutingFlag) {
	  cout << "Aborting command - SPEEDES exiting" << endl;
	  return_value = 1.0e20;
	  break;
	}

      }
    }
    return return_value;

  }
  else{ //RVI, notice that the else is on the "incorrect" line  :-)
    return -DBL_MAX;
  }

}

/*
*****************************************************************************
* SpEmHostUser::GetTime() method sends a message to SPEEDES to return the
* current gvt value. This is a blocking call.
*****************************************************************************
*/

SpSimTime &SpEmHostUser::GetMinMsgTime(void) {
  MIN_MSG_TIME_MESSAGE MinMsgTimeMessage;

  if (!InputFileFlag) {

    if (!SpeedesExecutingFlag){
      HlaMinMsgTime = 1.0e20;
      return HlaMinMsgTime;
    }

    MinMsgTimeMessage.serviceType = HOST_ROUTER_ID;
    MinMsgTimeMessage.group = GroupId;
    MinMsgTimeMessage.handlerNum = MIN_MSG_TIME_ID;
    MinMsgTimeMessage.bytes = sizeof(MIN_MSG_TIME_MESSAGE);

    ClientSock->sendMsg_blk(&MinMsgTimeMessage, sizeof(MIN_MSG_TIME_MESSAGE));

    for (;;) {

      MIN_MSG_TIME_REPLY_MESSAGE *ReturnMessage =
	(MIN_MSG_TIME_REPLY_MESSAGE *)GetMessageBlocking();

      if (ReturnMessage->handlerNum == KILL_SPEEDES_MESSAGE_ID) {
	SpeedesExecutingFlag = 0;
      }

      if (ReturnMessage->handlerNum == MIN_MSG_TIME_REPLY_ID) {
	HlaMinMsgTime = ReturnMessage->Time;
	delete [] (char*)ReturnMessage;
	break;
      }else{
	SpMsgHolder *MsgHolder = new SpMsgHolder(ReturnMessage);
	MessageQueue.push_bot(MsgHolder);

	if (!SpeedesExecutingFlag) {
	  cout << "Aborting command - SPEEDES exiting" << endl;
	  HlaMinMsgTime = 1.0e20;
	  break;
	}

      }
    }
    return HlaMinMsgTime;

  }else{ //JS, notice that the else is really on the "correct" line
    HlaMinMsgTime = -DBL_MAX;
    return HlaMinMsgTime;
  }

}

/*
*****************************************************************************
* SpEmHostUser::TimeAdvanceRequest() method requests a time update from
* SPEEDES. What actually happens is a message is sent to SPEEDES that returns
* gvt in the TimeAdvanceGrant() method. Each time SPEEDES receives this
* message, a barrier (gvt + TimeLag) is established. The old barrier, (unless
* it is the first time) is automatically removed. Barriers are automatically
* removed when the host user disconnects from the host router (or when the
* external module program exits).
*****************************************************************************
*/

void SpEmHostUser::TimeAdvanceRequest(double TimeLag) {
  TIME_ADVANCE_REQUEST_MESSAGE TimeMessage;

  if (!InputFileFlag) {
    DisconnectBarrierFlag = 1;
    TimeMessage.serviceType = HOST_ROUTER_ID;
    TimeMessage.group = GroupId;
    TimeMessage.handlerNum = TIME_ADVANCE_REQUEST_MESSAGE_ID;
    TimeMessage.TimeLag = TimeLag;
    TimeMessage.OldTime = Time;
    TimeMessage.bytes = sizeof(TIME_ADVANCE_REQUEST_MESSAGE);
    ClientSock->sendMsg_blk(&TimeMessage,
			    sizeof(TIME_ADVANCE_REQUEST_MESSAGE));
  }

}

/******************************************************************************
* SpEmHostUser::GoToTimeRequest() method requests SPEEDES to advance
* time to newTime and send a GoToTimeGrant message at that time.
* SPEEDES will also set up a barrier at newTime+TimeLag once newTime
* has been reached. What actually happens is a message is sent to
* SPEEDES that returns gvt in the TimeAdvanceGrant() method.  Suppose
* SPEEDES was at time 100 and one made the call GoToTimeRequest(200,
* 0) and then immediately made the call TimeAdvanceRequest(120).
* Suppose further that SPEEDES received both of these calls at time
* 100.  SPEEDES would then set up a barrier at time 200, advance to
* time 200, send back both a TimeAdvanceGrant and a GoToTimeGrant (in
* that order), and set up another barrier at time 200.  In essense,
* the TimeAdvanceRequest will be ignored.
******************************************************************************/

void SpEmHostUser::GoToTimeRequest(double newTime, double timeLag) {
  GO_TO_TIME_REQUEST_MESSAGE TimeMessage;

  if (!InputFileFlag) {
    DisconnectBarrierFlag = 1;
    TimeMessage.serviceType = HOST_ROUTER_ID;
    TimeMessage.group = GroupId;
    TimeMessage.handlerNum = GO_TO_TIME_REQUEST_MESSAGE_ID;
    TimeMessage.TimeLag = timeLag;
    TimeMessage.NewTime = newTime;
    TimeMessage.bytes = sizeof(GO_TO_TIME_REQUEST_MESSAGE);

    ClientSock->sendMsg_blk(&TimeMessage,sizeof(GO_TO_TIME_REQUEST_MESSAGE));
  }

}

/*
*****************************************************************************
* SpEmHostUser::GoToTime() method requests that SPEEDES advance time
* to newTime and blocks until SPEEDES reaches that point.  Once
* SPEEDES reaches that point, it sets up a barrier at gvt+timeLag.  If
* SPEEDES receives this request with newTime < gvt, SPEEDES simply
* returns gvt.
******************************************************************************/

void SpEmHostUser::GoToTime(double newTime, double timeLag) {

  GoToTimeRequest(newTime, timeLag);
  while (!GoToTimeGranted()){
    if (!SpeedesExecuting()){
      exit(0);
    }
  }
}

/*
*****************************************************************************
* SpEmHostUser::TimeAdvanceGranted() method tries to read time advance grant
* messages from the host router. It returns 0 if time has not yet advanced.
* It returns 1 when time advances along with the new time value. The time
* advance grant mechanism works with the TimeAdvanceRequest() method.
*****************************************************************************
*/

int SpEmHostUser::TimeAdvanceGranted() {
  TIME_ADVANCE_GRANT_MESSAGE *TimeMessage;

  TimeMessage = (TIME_ADVANCE_GRANT_MESSAGE *)
    GetMessage(TIME_ADVANCE_GRANT_MESSAGE_ID);

  if (TimeMessage) {
    Time = TimeMessage->Time;
    delete [] (char *)TimeMessage;
    return 1;
  }

  return 0;

}

/*
*****************************************************************************
* SpEmHostUser::GoToTimeGranted() method tries to read time advance grant
* messages from the host router. It returns 0 if time has not yet advanced.
* It returns 1 when time advances along with the new time value. The time
* advance grant mechanism works with the TimeAdvanceRequest() method.
*****************************************************************************
*/

int SpEmHostUser::GoToTimeGranted() {
  GO_TO_TIME_GRANT_MESSAGE *TimeMessage;

  TimeMessage = (GO_TO_TIME_GRANT_MESSAGE *)
    GetMessage(GO_TO_TIME_GRANT_MESSAGE_ID);

  if (TimeMessage) {
    Time = TimeMessage->Time;
    delete [] (char *)TimeMessage;
    return 1;
  }

  return 0;

}

/*
*****************************************************************************
* SpEmHostUser::Subscribe() method subscribes to a named subscription type
* by object name.
*****************************************************************************
*/

int SpEmHostUser::SendSubscribe(SUBSCRIPTION_ITEM *SubItem) {
  SUBSCRIBE_MESSAGE *SubscribeMessage;
  int Size = SubItem->GetBufferSize();

  SubscribeMessage = (SUBSCRIBE_MESSAGE *)
    new char[sizeof(SUBSCRIBE_MESSAGE) + Size];

  SubItem->PackBuffer((char *)&SubscribeMessage[1]);

  if (!InputFileFlag) {
    DisconnectBarrierFlag = 1;
    SubscribeMessage->serviceType = HOST_ROUTER_ID;
    SubscribeMessage->group = GroupId;
    SubscribeMessage->handlerNum = SUBSCRIBE_MESSAGE_ID;
    SubscribeMessage->bytes = Size;
    ClientSock->sendMsg_blk(SubscribeMessage,
			    sizeof(SUBSCRIBE_MESSAGE) + Size);
  }

  delete [] (char*) SubscribeMessage;
  return SUBSCRIBE_STATUS_GOOD;

}


/*
*****************************************************************************
* SpEmHostUser::Subscribe() method subscribes to a named subscription type
* by object name.
*****************************************************************************
*/

int SpEmHostUser::SendUnSubscribe(SUBSCRIPTION_ITEM *SubItem) {
  UNSUBSCRIBE_MESSAGE *UnSubscribeMessage;
  int Size = SubItem->GetBufferSize();

  UnSubscribeMessage = (UNSUBSCRIBE_MESSAGE *)
    new char[sizeof(UNSUBSCRIBE_MESSAGE) + Size];

  SubItem->PackBuffer((char *)&UnSubscribeMessage[1]);

  if (!InputFileFlag) {
    DisconnectBarrierFlag = 1;
    UnSubscribeMessage->serviceType = HOST_ROUTER_ID;
    UnSubscribeMessage->group = GroupId;
    UnSubscribeMessage->handlerNum = UNSUBSCRIBE_MESSAGE_ID;
    UnSubscribeMessage->bytes = Size;
    ClientSock->sendMsg_blk(UnSubscribeMessage,
			    sizeof(UNSUBSCRIBE_MESSAGE) + Size);
  }

  delete [] (char*)UnSubscribeMessage;
  return UNSUBSCRIBE_STATUS_GOOD;

}

/*
*****************************************************************************
* SpEmHostUser::Subscribe() method subscribes to a named subscription type
* by object name.
*****************************************************************************
*/

int SpEmHostUser::Subscribe(char *Name) {
  SUBSCRIPTION_ITEM SubItem(Name);
  return SendSubscribe(&SubItem);
}

/*
*****************************************************************************
* SpEmHostUser::SubscribeObject() method subscribes to a named subscription
* type by object name.
*****************************************************************************
*/

int SpEmHostUser::SubscribeObject(char *Name, char *ObjectName) {
  SUBSCRIPTION_ITEM SubItem(Name, ObjectName);
  return SendSubscribe(&SubItem);
}

/*
*****************************************************************************
* SpEmHostUser::SubscribeObject() method subscribes to a named subscription
* type by object Ids (which can be wild-carded).
*****************************************************************************
*/

int SpEmHostUser::SubscribeObject(char *Name,
				   int ObjectType,
				   int LocalId,
				   int Node) {
  SUBSCRIPTION_ITEM SubItem(Name, ObjectType, LocalId, Node);
  return SendSubscribe(&SubItem);
}

/*
*****************************************************************************
* SpEmHostUser::SubscribeType() method subscribes to a named subscription type
* by object Ids (which can be wild-carded).
*****************************************************************************
*/

int SpEmHostUser::SubscribeType(char *Name, char *Type) {
  int NameFound;
  SpNameValueList *NameValueList = ObjectManagerTypes();

  int Value = int(NameValueList->GetValue(Type, NameFound));
  delete NameValueList;
  if (!NameFound) {
    return SUBSCRIBE_STATUS_BAD_TYPE;
  }

  SUBSCRIPTION_ITEM SubItem(Name, Value);
  return SendSubscribe(&SubItem);

}

/*
*****************************************************************************
* SpEmHostUser::UnSubscribe() method unsubscribes from a named subscription
* type by object name.
*****************************************************************************
*/

int SpEmHostUser::UnSubscribe(char *Name) {
  SUBSCRIPTION_ITEM SubItem(Name);
  return SendUnSubscribe(&SubItem);
}

/*
*****************************************************************************
* SpEmHostUser::UnSubscribeObject() method unsubscribes from a named
* subscription type by object name.
*****************************************************************************
*/

int SpEmHostUser::UnSubscribeObject(char *Name, char *ObjectName) {
  SUBSCRIPTION_ITEM SubItem(Name, ObjectName);
  return SendUnSubscribe(&SubItem);
}

/*
*****************************************************************************
* SpEmHostUser::UnSubscribeObject() method unsubscribes from a named
* subscription type by object Ids (which can be wild-carded).
*****************************************************************************
*/

int SpEmHostUser::UnSubscribeObject(char *Name,
				   int ObjectType,
				   int LocalId,
				   int Node) {
  SUBSCRIPTION_ITEM SubItem(Name, ObjectType, LocalId, Node);
  return SendUnSubscribe(&SubItem);
}

/*
*****************************************************************************
* SpEmHostUser::UnSubscribeType() method unsubscribes from a named subscription
* type by object Ids (which can be wild-carded).
*****************************************************************************
*/

int SpEmHostUser::UnSubscribeType(char *Name, char *Type) {
  int NameFound;
  SpNameValueList *NameValueList = ObjectManagerTypes();

  int Value = int(NameValueList->GetValue(Type, NameFound));
  delete NameValueList;
  if (!NameFound) {
    return UNSUBSCRIBE_STATUS_BAD_TYPE;
  }

  SUBSCRIPTION_ITEM SubItem(Name, Value);
  return SendUnSubscribe(&SubItem);

}

/*
*****************************************************************************
* SpEmHostUser::ReceiveSubscribedData() method returns data that was sent
* from SPEEDES using the SendSubscribedData method of the SpHostUser.
*****************************************************************************
*/

char *SpEmHostUser::ReceiveSubscribedData(char *Name, double &time,
					  int &bytes, int &globalId) {
  SUBSCRIBED_DATA_MESSAGE *Message;
  SpMsgHolder *MsgHolder;
  int i;
  int len;
  C_QUEUE Q;
  char *Data;
  char *Buff;

  ReadMessages();

  len = MessageQueue.get_length();
  for (i=0; i<len; i++) {

    MsgHolder = (SpMsgHolder *)MessageQueue.pop_top();

    if (MsgHolder->CompareType(SUBSCRIBED_DATA_MESSAGE_ID)) {

      Message = (SUBSCRIBED_DATA_MESSAGE *)MsgHolder->GetHolderMessage();

      if (!strcmp((char *)&Message[1], Name)) {

	Q.concat(&MessageQueue);
	MessageQueue = Q;
	delete MsgHolder;

	int numBytes = Message->bytes;
	Data = new char[numBytes];
	Buff = (char *)&Message[1];
	Buff += strlen(Buff) + 1;

	memcpy(Data, Buff, Message->bytes);
	bytes = Message->bytes;
	time = Message->Time;
	globalId = Message->GlobalId;

	delete [] (char *)Message;
	return Data;

      }else{

	Q.push_bot(MsgHolder);

      }

    }else{

      Q.push_bot(MsgHolder);

    }

  }

  MessageQueue = Q;
  return NULL;

}

/*
*****************************************************************************
* SpEmHostUser::ReceiveNamedData() method returns data that was sent from
* SPEEDES using the SendNamedData method of the SpHostUser.
*****************************************************************************
*/
char *SpEmHostUser::ReceiveNamedData(char *Name,
				     SpSimTime &time,
				     int &DataBytes,
				     int &GlobalId) {
  SPEEDES_TO_EXTERNAL_MODULE_MESSAGE *Message;
  SpMsgHolder *MsgHolder;
  int i;
  int len;
  C_QUEUE Q;
  char *Data;
  char *Buff;

  ReadMessages();

  len = MessageQueue.get_length();
  for (i=0; i<len; i++) {

    MsgHolder = (SpMsgHolder *)MessageQueue.pop_top();

    if (MsgHolder->CompareType(SPEEDES_TO_EXTERNAL_MODULE_MESSAGE_ID)) {

      Message = (SPEEDES_TO_EXTERNAL_MODULE_MESSAGE *)
	MsgHolder->GetHolderMessage();

      if (!strcmp((char *)&Message[1], Name)) {

	Q.concat(&MessageQueue);
	MessageQueue = Q;
	delete MsgHolder;

	int nameLen = strlen((char *)&Message[1]) + 1;
	DataBytes = Message->bytes - nameLen;
	Data = new char[DataBytes];
	Buff = (char *)&Message[1];
	Buff += nameLen;

	memcpy(Data, Buff, DataBytes);
	time = Message->Time;
	GlobalId = Message->GlobalId;

	delete [] (char *)Message;
	return Data;

      }else{

	Q.push_bot(MsgHolder);

      }

    }else{

      Q.push_bot(MsgHolder);

    }

  }

  MessageQueue = Q;
  return NULL;
}

/*
*****************************************************************************
* SpEmHostUser::ReceiveData() method returns data that was sent from SPEEDES
* using the SendData method of the SpHostUser.
*****************************************************************************
*/

char *SpEmHostUser::ReceiveAnyNamedData(char *&Name, SpSimTime &time,
					int &DataBytes, int &GlobalId) {
  SPEEDES_TO_EXTERNAL_MODULE_MESSAGE *Message;
  SpMsgHolder *MsgHolder;
  int i;
  int len;
  C_QUEUE Q;
  char *Data;
  char *Buff;

  ReadMessages();

  len = MessageQueue.get_length();
  for (i=0; i<len; i++) {

    MsgHolder = (SpMsgHolder *)MessageQueue.pop_top();

    if (MsgHolder->CompareType(SPEEDES_TO_EXTERNAL_MODULE_MESSAGE_ID)) {

      Message = (SPEEDES_TO_EXTERNAL_MODULE_MESSAGE *)
	MsgHolder->GetHolderMessage();

      Name = strdup((char *)&Message[1]);

      Q.concat(&MessageQueue);
      MessageQueue = Q;
      delete MsgHolder;

      int nameLen = strlen((char *)&Message[1]) + 1;
      DataBytes = Message->bytes - nameLen;
      Data = new char[DataBytes];
      Buff = (char *)&Message[1];
      Buff += nameLen;

      memcpy(Data, Buff, DataBytes);
      time = Message->Time;
      GlobalId = Message->GlobalId;

      delete [] (char *)Message;
      return Data;

    }else{

      Q.push_bot(MsgHolder);

    }

  }

  MessageQueue = Q;
  return NULL;

}

/*
*****************************************************************************
* SpEmHostUser::ReceiveData() method returns data that was sent from SPEEDES
* using the SendData method of the SpHostUser.
*****************************************************************************
*/

char *SpEmHostUser::ReceiveAnySubscribedData(char *&Name, double &time,
					     int &bytes, int &globalId) {
  SUBSCRIBED_DATA_MESSAGE *Message;
  SpMsgHolder *MsgHolder;
  int i;
  int len;
  C_QUEUE Q;
  char *Data;
  char *Buff;

  ReadMessages();

  len = MessageQueue.get_length();
  for (i=0; i<len; i++) {

    MsgHolder = (SpMsgHolder *)MessageQueue.pop_top();

    if (MsgHolder->CompareType(SUBSCRIBED_DATA_MESSAGE_ID)) {

      Message = (SUBSCRIBED_DATA_MESSAGE *)MsgHolder->GetHolderMessage();

      Name = strdup((char *)&Message[1]);

      Q.concat(&MessageQueue);
      MessageQueue = Q;
      delete MsgHolder;

      int numBytes = Message->bytes;
      Data = new char[numBytes];
      Buff = (char *)&Message[1];
      Buff += strlen(Buff) + 1;

      memcpy(Data, Buff, Message->bytes);
      bytes = Message->bytes;
      time = Message->Time;
      globalId = Message->GlobalId;

      delete [] (char *)Message;
      return Data;

    }else{

      Q.push_bot(MsgHolder);

    }

  }

  MessageQueue = Q;
  return NULL;

}

SCHEDULE_EVENT_REPLY_MESSAGE* SpEmHostUser::GetScheduleReply()
{
  SCHEDULE_EVENT_REPLY_MESSAGE *rc = NULL;
  SpMsgHolder *MsgHolder;

  for (;;){
    SCHEDULE_EVENT_REPLY_MESSAGE *ReturnMessage = 
      (SCHEDULE_EVENT_REPLY_MESSAGE *)GetMessageBlocking();

    if (ReturnMessage->handlerNum == KILL_SPEEDES_MESSAGE_ID) {
      break;
    }
    
    if (ReturnMessage->handlerNum == SCHEDULE_EVENT_REPLY_ID) {
      rc = new SCHEDULE_EVENT_REPLY_MESSAGE;
      memcpy (rc, ReturnMessage, sizeof(SCHEDULE_EVENT_REPLY_MESSAGE));
      delete [] (char *)ReturnMessage;
      break;
    }else{
      MsgHolder = new SpMsgHolder(ReturnMessage);
      MessageQueue.push_bot(MsgHolder);
    }
  }
  return rc;
}

SCHEDULE_EVENT_REPLY_MESSAGE* SpEmHostUser::ScheduleEvent(double time,
							  int globalId,
							  char * eventName,
							  char * msg,
							  int msgBytes,
							  char * data,
							  int dataBytes,
							  TimeMode tmMode,
							  int CancelHandleRequested)
{
  int nBytes = sizeof(SCHEDULE_EVENT_MESSAGE) + msgBytes +
    dataBytes + strlen(eventName)+1;
  char * buff = new char[nBytes];
  char * tPtr;
  SCHEDULE_EVENT_MESSAGE *eventMsg = (SCHEDULE_EVENT_MESSAGE*)buff;
  SCHEDULE_EVENT_REPLY_MESSAGE *rc = NULL;

  eventMsg->serviceType = HOST_ROUTER_ID;
  eventMsg->group = GroupId;
  eventMsg->bytes = nBytes;
  eventMsg->GlobalId = globalId;
  eventMsg->MessageBytes = msgBytes;
  eventMsg->DataBytes = dataBytes;
  eventMsg->Time = time;
  eventMsg->handlerNum = SCHEDULE_EVENT_ID;
  eventMsg->CancelHandleRequested = CancelHandleRequested;

  eventMsg->ScheduleMode = tmMode;

  tPtr = (char *)&eventMsg[1];
  strcpy(tPtr, eventName);
  tPtr += (strlen(eventName)+1);
  memcpy(tPtr, msg, msgBytes);
  tPtr += msgBytes;
  memcpy(tPtr, data, dataBytes);

  ClientSock->sendMsg_blk(eventMsg, nBytes);

  delete [] buff;
  if (CancelHandleRequested){
    rc = GetScheduleReply();
  }
  return rc;
}


SCHEDULE_EVENT_REPLY_MESSAGE*  SpEmHostUser::ScheduleEvent(double time,
							   char* objectName,
							   char * eventName,
							   char * msg,
							   int msgBytes,
							   char * data,
							   int dataBytes,
							   TimeMode tmMode,
							   int CancelHandleRequested)
{
  int nBytes = sizeof(SCHEDULE_EVENT_MESSAGE) + msgBytes + dataBytes
    + strlen(eventName) + 1 + strlen(objectName)+1;
  char * buff = new char[nBytes];
  char * tPtr;
  SCHEDULE_EVENT_MESSAGE *eventMsg = (SCHEDULE_EVENT_MESSAGE*)buff;
  SCHEDULE_EVENT_REPLY_MESSAGE *rc = NULL;

  eventMsg->serviceType = HOST_ROUTER_ID;
  eventMsg->group = GroupId;
  eventMsg->bytes = nBytes;
  eventMsg->GlobalId = -1;
  eventMsg->MessageBytes = msgBytes;
  eventMsg->DataBytes = dataBytes;
  eventMsg->Time = time;
  eventMsg->handlerNum = SCHEDULE_EVENT_ID;
  eventMsg->ScheduleMode = tmMode;
  eventMsg->CancelHandleRequested = CancelHandleRequested;
  tPtr = (char *)&eventMsg[1];
  strcpy(tPtr, eventName);
  tPtr += (strlen(eventName)+1);
  memcpy(tPtr, msg, msgBytes);
  tPtr += msgBytes;
  memcpy(tPtr, data, dataBytes);
  tPtr += dataBytes;
  strcpy(tPtr, objectName);
  ClientSock->sendMsg_blk(eventMsg, nBytes);
  if (CancelHandleRequested){
    rc = GetScheduleReply();
  }

  delete [] buff;
  return rc;
}


CANCEL_EVENT_REPLY_MESSAGE* SpEmHostUser::GetCancelReply()
{
  CANCEL_EVENT_REPLY_MESSAGE * rc = NULL; 
  SpMsgHolder *MsgHolder;

  for (;;){
    CANCEL_EVENT_REPLY_MESSAGE*ReturnMessage = 
      (CANCEL_EVENT_REPLY_MESSAGE *)GetMessageBlocking();
    
    if (ReturnMessage->handlerNum == KILL_SPEEDES_MESSAGE_ID) {
      break;
    }
    
    if (ReturnMessage->handlerNum == CANCEL_EVENT_REPLY_ID) {
      rc = new CANCEL_EVENT_REPLY_MESSAGE;
      memcpy (rc, ReturnMessage, sizeof(CANCEL_EVENT_REPLY_MESSAGE));
      delete [] (char *)ReturnMessage;
      break;
    }else{
      MsgHolder = new SpMsgHolder(ReturnMessage);
      MessageQueue.push_bot(MsgHolder);
    }
  }
  return rc;
}

CANCEL_EVENT_REPLY_MESSAGE* SpEmHostUser::CancelEvent(int cancelHandle,
						      int replyRequested)
{
  CANCEL_EVENT_MESSAGE msg;
  CANCEL_EVENT_REPLY_MESSAGE * rc = NULL;

  msg.CancelHandle = cancelHandle;
  msg.handlerNum = CANCEL_EVENT_ID;
  msg.ReplyRequested = replyRequested;
  msg.serviceType = HOST_ROUTER_ID;
  msg.group = GroupId;
  msg.bytes = sizeof(CANCEL_EVENT_MESSAGE);
  ClientSock->sendMsg_blk(&msg, sizeof(CANCEL_EVENT_MESSAGE));
  if (replyRequested){
    rc = GetCancelReply();
  }
  return rc;
}


void SpEmHostUser::WriteSpeedesMessageToFile(GENERIC_MSG_HEADER* msg)
{
  fwrite(msg, sizeof(char), msg->TotalBytes, fp_output);
}


GENERIC_MSG_HEADER* SpEmHostUser::GetSpeedesMessageFromFile()
{
  static GENERIC_MSG_HEADER hdr;
  char * retval = NULL;
  char * tPtr;
  unsigned int bytesRead;

  while (retval == NULL){
    bytesRead = fread(&hdr, sizeof(char), sizeof(GENERIC_MSG_HEADER),fp_input);
    if (bytesRead == sizeof(GENERIC_MSG_HEADER)){
      int numBytes = hdr.TotalBytes;
      retval = new char[numBytes];
      memcpy(retval, &hdr, sizeof(GENERIC_MSG_HEADER));
      tPtr = retval + sizeof(GENERIC_MSG_HEADER);
      bytesRead = fread(tPtr,
			sizeof(char),
			hdr.TotalBytes - sizeof(GENERIC_MSG_HEADER),
			fp_input);
      if ((bytesRead) == hdr.TotalBytes - sizeof(GENERIC_MSG_HEADER)){
	
	switch(hdr.handlerNum){
	case KILL_SPEEDES_MESSAGE_ID: //Add in any cases we want to ignore here
	  delete [] retval;
	  retval = NULL;
	  break;
	default:
	  break;
	}
      }
      else{
	delete [] retval;
	retval = NULL;
	return NULL;
      }
    }
    else{
      return NULL;
    }
  }
  return (GENERIC_MSG_HEADER*) retval;
}
