#ifndef SpEmHostUser_OBJECT
#define SpEmHostUser_OBJECT

#include "SpStdio.H"

#include "SpClientSock.H"
#include "SpDataParser.H"
#include "queue.H"
#include "SpNameValueList.H"
#include "SpSubscriptionItem.H"
#include "SpHostRouterMsgs.H"

/*************************************************************************
 * Class: SpEmHostUser
 * Summary: The SpEmHostUser is the external module interface to the
 *          Host Router (and therefore, into SPEEDES)
 *
 * Description: In order to have access to SPEEDES from outside of the
 *              simulation one must go through the host router.  In
 *              order to use the HostRouter, one needs to use the
 *              SpEmHostUser class which is described below.  This is
 *              the only way to get information in and out of SPEEDES
 *              as well as control the simulation.  In fact, the
 *              SpStateMgr (OPSM) is essentially a wrapper for the
 *              SpEmHostUser with different interfaces.
 *
 *              We speak quite a bit about barriers within this class.
 *              a barrier simply means that (when SPEEDES is operating
 *              with optimistic time management) SPEEDES simply does
 *              not update GVT any farther than that barrier.  SPEEDES
 *              still continues to optimistically process events but
 *              these events are just not processed if the occur after
 *              the barrier.
 *
 *              Also, each external module can have only one barrier.
 *              If a method is called that modifies that barrier, then
 *              the only barrier for that module is just updated.  The
 *              one exception to this rule is the GoToTimeRequest()
 *              call.  If you make a GoToTimeRequest() call and the
 *              tiem you request occurs before a pre-defined barrier,
 *              then the barrier is moved up to the time requested. If
 *              the requested time occurs after a pre-defined barrier
 *              then two barriers will be in effect until the
 *              requested time is reached, after which, there will be
 *              only one barrier again which will be at the requested
 *              time + the time lag of the GoToTimeRequest call..
 *
 * Examples: We have provided several examples of how one may use the
 *           SpEmHostUser class in speedes/CommandLineUtilites.  These
 *           utilities demonstrate the use of many of these basic
 *           interfaces from the command line and also demonstrate
 *           some more advanced uses through combining several of the
 *           calls.  The SpStateMgr class also demonstrates a use
 *           of the SpEmHostUser class.
 *
*************************************************************************/

class SpEmHostUser {
public:
  

/************************************************************************
 * Method: SpEmHostUser()
 * Summary: standard constructor for a SpEmHostUser()
 * Description: If you construct an SpEmHostUser with no arguments, it
 *              attempts to load a speedes.par file from the local
 *              directory and then reads the port number, group number
 *              and machine name of the host router from that par
 *              file.
 * Return: a new SpEmHostUser.
*************************************************************************/
  SpEmHostUser();

  

/************************************************************************
 * Method: SpEmHostUser(SpDataParser *)
 * Summary: constructs a SpEmHostUser from an already opened par file.
 * Description: 
 * Return: a new SpEmHostUser
*************************************************************************/
  SpEmHostUser(SpDataParser *Parser //In: already opened par file in Parser
	       );
  

/************************************************************************
 * Method: SpEmHostUser(char *, int, int)
 * Summary: constructs a SpEmHostUser using a specific machine, port,
 *          and group
 * Description: 
 * Return: a new SpEmHostUser
*************************************************************************/
  SpEmHostUser(char *MachineName, //In: Machine where the HostRouter is running
	       int Port,          //In: Port of HostRouter on that machine
	       int Group          //In: Unique group number of our simulation
	       );
  

/************************************************************************
 * Method: SpEmHostUser(char *)
 * Summary: Constructs a SpEmHostUser from a pre-recorded simulation
 * Description: NOT IMPLEMENTED YET.  When available, one will be able
 *              to run a simulation, record all messages coming out
 *              from that simulation and then restart the SpEmHostUser
 *              from that recording.  This could be especially useful
 *              for trying to debug external modules that only monitor
 *              the simulation rather than trying to interact with it
 *              (like graphics routines).
 * Return: a new SpEmHostUser
*************************************************************************/
  SpEmHostUser(char *FileName //In: Filename of previously recorded simulation.
	       );
  

/************************************************************************
 * Method: ~SpEmHostUser()
 * Summary: destructor of an SpEmHostUser
 * Description: Please try and delete your host user.  It helps the
 *              Host Router know who to disconnect which makes the
 *              simulation run a bit more smoothly.
 * Return: none
*************************************************************************/
  virtual ~SpEmHostUser();
  

/************************************************************************
 * Method: RecordInputMesages(char *)
 * Summary: Tells the SpEmHostUser to record received messages in a file
 * Description: NOT IMPLEMENTED YET
 * Return: none
*************************************************************************/
  void RecordInputMessages(char *FileName //In: File to which we want to
			                  //    record input messages
			   );
  
/************************************************************************
 * Method: SpeedesExecuting()
 * Summary: Returns 1 if SPEEDES is still executing, 0 otherwise
 * Description: 
 * Return: Returns 1 if SPEEDES is still executing, 0 otherwise
*************************************************************************/
  int SpeedesExecuting() {ReadMessages();return SpeedesExecutingFlag;}
  
  

/************************************************************************
 * Method: DisconnectBarrier()
 * Summary: Removes any barriers that this SpEmHostUser has extablished.
 * Description: 
 * Return: void
*************************************************************************/
  void DisconnectBarrier();
  
  

/************************************************************************
 * Method: ReadMessages()
 * Summary: Attempts to read all messages and queue them for processesing
 * Description: 
 * Return: void
*************************************************************************/
  void ReadMessages();
  
  

/************************************************************************
 * Method: GetMessage()
 * Summary: Returns a message that has been queued up (NULL if no messages
 *          are available.
 * Description: 
 * Return: a GENERIC_MSG_HEADER pointer which is a queued up message
*************************************************************************/
  GENERIC_MSG_HEADER *GetMessage();
  

/************************************************************************
 * Method: GetMessage(int)
 * Summary: Returns a queued message of a given type, NULL if no messages
 *          of that type are available.
 * Description: 
 * Return: a GENERIC_MSG_HEADER pointer which is a queued up message of the
 *         requested type.
*************************************************************************/
  GENERIC_MSG_HEADER *GetMessage(int Type //In: Type of message desired.
				 );
  
  

/************************************************************************
 * Method: ReceiveData(double&, int&, int&)
 * Summary: Returns data sent by SPEEDES using the SendData call from
 *          the SPEEDES side of the HostRouter (SpHostUser).
 * Description: When a call in SPEEDES is made to
 *              SpHostUser->SendData(), this data can then be
 *              received in the SpEmHostUser through this ReceiveData
 *              call.  Time contains the time at which the data was
 *              created, Bytes is the size of the data, and GlobalId
 *              is the global id of the SPEEDES object that created
 *              the data.  It is returned as a char string that should
 *              be deleted as soon as it is no longer needed.
 * Return: Data sent from SPEEDES using HostUser->SendData()
*************************************************************************/
  char *ReceiveData(double &Time, //Out: Time at which data was created
		    int &Bytes,   //Out: Size of data in bytes
		    int &GlobalId //Out: Global id of object that created
		                  //     the data.
		    );
  
  char *ReceiveNamedData(char *Name,
			 SpSimTime &Time,
			 int &DataBytes,
			 int &GlobalId);

  char *ReceiveAnyNamedData(char *&Name,
			    SpSimTime &Time,
			    int &DataBytes,
			    int &GlobalId);

/************************************************************************
 * Method: NameLookup(char *)
 * Summary: Given the name of an object, this returns the global ID.
 * Description: 
 * Return: The global ID of the passed in name
*************************************************************************/
  int NameLookup(char *ObjectName //In: The object name for which one
		                  //    desires the global ID.
		 );
  
  

/************************************************************************
 * Method: ObjectManagerTypes()
 * Summary: Returns a list of the types of object managers in SPEEDES
 * Description: Returns a SpNameValueList of all object managers in
 *              SPEEDES.  The SpNameValueList can then be traversed to
 *              find the types within SPEEDES so that one can
 *              subscribe to them if you want.
 * Return: SpNameValueList of all object managers in SPEEDES
*************************************************************************/
  SpNameValueList *ObjectManagerTypes();
  
  

/************************************************************************
 * Method: ObjectManagerType(char *)
 * Summary: Returns the id of an object manager type in SPEEDES
 * Description: The global ID of an object manager could be useful if
 *              you wanted to dynamically create or destroy objects. It
 *              can also be used in the ObjectNames() method call so
 *              you have the 
 * Return: ID of an object manager type in SPEEDES
*************************************************************************/
  int ObjectManagerType(char *Type //In: name of type for which you want
			           //    the global ID.
			);
  
  

/************************************************************************
 * Method: ObjectNames(int)
 * Summary: Returns a list of the names of objects (optionally by Type)
 *          in SPEEDES.
 * Description: 
 * Return: Returns a SpNameValueList of the names of objects
 *         (optionally by Type) in SPEEDES.
 *************************************************************************/
  SpNameValueList *ObjectNames(int Type = -1 //In: Type for which you desire
			                     //    the list of names.  If it
			                     //    is not provided, return
			                     //    names for all types.
			       );
  
  

/************************************************************************
 * Method: ObjectNames(char *)
 * Summary: Returns a list of the names of objects in SPEEDES
 * Description: Given a type name, this returns a list of all names of
 *              that type within SPEEDES.
 * Return: Returns a SpNameValueList of the names of objects of a given
 *         type in SPEEDES.
*************************************************************************/
  SpNameValueList *ObjectNames(char *Type//In: Type for which you desire
			                 //    the list of names.
			       );
  
 

/************************************************************************
 * Method: GetNumNodes
 * Summary: Gets the number of SPEEDES nodes
 * Description: Sends a message to the host router to get the number of nodes.
 * Return: Returns the number of nodes.
*************************************************************************/
    int GetNumNodes();
 
/************************************************************************
 * Method: GetNextNode
 * Summary: Gets the next SPEEDES node in a round-robin manner
 * Description: Sends a message to the host router to get the next node
 * 		if this is called again, the next node increases by one
 *		(using modular arithmetic).
 * Return: Returns the next node.
*************************************************************************/
    int GetNextNode();

/************************************************************************
 * Method: Flush()
 * Summary: Sends a message to the HostRouter and waits for a reply
 *          like ping.
 * Description: This is used to try and ensure that all messages that
 *              we have sent to the host router have been received.
 *              Since we are using TCP/IP, we are guaranteed in order
 *              delivery. This will be the last message sent so, if it
 *              is received, we know everything before it has been
 *              received i.e. flushed.
 * Return: none
*************************************************************************/
  void Flush();
  
  

/************************************************************************
 * Method: Pause()
 * Summary: Pauses the SPEEDES simulation
 * Description: Pause increments a counter in SPEEDES of how many
 *              pauses have been received.  An equal number of
 *              Resume() calls must be made before SPEEDES is resumed.
 *              Notice that this only causes SPEEDES to stop updating
 *              GVT.  SPEEDES still continues to optimistically
 *              process events (if an optimistic time management
 *              algorithm is chosen).
 * Return: none
*************************************************************************/
  void Pause();
  

/************************************************************************
 * Method: Resume()
 * Summary: Resumes the SPEEDES simulation
 * Description: Each Resume() decrements a counter within SPEEDES (to
 *              a minimum of 0).  An equal number of Resume() calls as
 *              Pause() calls must be made before SPEEDES begins
 *              updating GVT.
 * Return: none
*************************************************************************/
  void Resume();
  

  /************************************************************************
 * Method: Pause()
 * Summary: Pauses the SPEEDES simulation
 * Description: Sends a named pause command to SPEEDES that can be resumed
 *              using the NamedResume() command. This allows external users
 *  	        to identify a pause with a string name. The named pause does
 *              not stop processing in SPEEDES. It simply prohibits gvt from
 *              advancing. Note that you can set named pauses in the
 *              speedes.par file for establishing initial named pauses.
 *              An optional argument is the time that you want to pause at.
 *              If gvt is greater than the time, it pauses at gvt instead.
 * Return: none
*************************************************************************/
  void Pause(char *pauseName, double time=0);
  

/************************************************************************
 * Method: Resume()
 * Summary: Resumes the SPEEDES simulation
 * Description: Resumes the simulation using a string name to identify the
 *              pause. Note that you can set named pauses in the speedes.par
 *              file for establishing initial named pauses.
 * Return: none
*************************************************************************/
  void Resume(char *resumeName);
  

/************************************************************************
 * Method: KillSPEEDES()
 * Summary: Kills the SPEEDES execution
 * Description: Kills SPEEDES.  Presently, a kill message is not sent
 *              to external modules.
 * Return: none
*************************************************************************/
  void KillSPEEDES();
  
  

/************************************************************************
 * Method: Query(int)
 * Summary: Queries an object by GlobalId
 * Description: When SPEEDES receives a query about an object, it will
 *              respond with ta SpNameValueList that is tagged at GVT.
 *		The query event that is generated by this method calls the
 *		Query virtual method on the simobj which fills in the name
 *		value list that is returned to the external user. The
 *		queryName string is optional, but it can be used to request
 *		specific types of application specific data.
 * Return: SpNameValueList which is filled in by the SPEEDES application
*************************************************************************/
  SpNameValueList *Query(int GlobalId, //In: Global ID of the object you want
			              //    to query.
			 char *queryName="");
  
  

/************************************************************************
 * Method: Query(char *)
 * Summary: Queries an object by ObjectName
 * Description: When SPEEDES receives a query about an object, it will
 *              respond with ta SpNameValueList that is tagged at GVT.
 *		The query event that is generated by this method calls the
 *		Query virtual method on the simobj which fills in the name
 *		value list that is returned to the external user. The
 *		queryName string is optional, but it can be used to request
 *		specific types of application specific data.
 * Return: SpNameValueList which is filled in by SPEEDES.
*************************************************************************/
  SpNameValueList *Query(char *ObjectName, //In: Name of object you wish to
			                  //    query.
			 char *queryName="");


/************************************************************************
 * Method: SendCommand(char *, int, int, int, char*, int, int)
 * Summary: Sends a command to an object by the triplet
 *          <Node, Type, LocalId>.
 * Return: none.
*************************************************************************/
  int SendCommand(char *CommandName, //In: Name of command to be scheduled
		  int Node,	     //In: Global id of SimObj
		  int Type,	     //In: Object manager type of SimObj
		  int LocalId,	     //In: Local id of SimObj
		  char *Message,     //In: The message/data for the command
		  int Bytes,         //In: The size of the message in bytes
		  int ReturnStatus=0 //In: Set to 1 to enable error checking
		  );

/************************************************************************
 * Method: SendCommand(char *, int, int, int, SpSimTime, char*, int, int)
 * Summary: Sends a command to an object defined by the triplet
 *          <Node, Type, LocalId>.  The command will be processed
 *          at the specified simtime.
 * Return: none.
*************************************************************************/
  int SendCommand(char *CommandName, //In: Name of command to be scheduled
		  int Node,	     //In: Global id of SimObj
		  int Type,	     //In: Object manager type of SimObj
		  int LocalId,	     //In: Local id of SimObj
		  SpSimTime EventTime,//In: Time that command takes place
		  char *Message,     //In: The message/data for the command
		  int Bytes,         //In: The size of the message in bytes
		  int ReturnStatus=0 //In: Set to 1 to enable error checking
		  );


/************************************************************************
 * Method: SendCommand(char *, int, char*, int, int)
 * Summary: Sends a command to an object by GlobalId and returns status
 * Description: Within SPEEDES, objects can respond to named commands
 *              similar to how the STATE_MANGER (OPSM) can respond to
 *              named commands.  This is how you can schedule events
 *              for a given object within SPEEDES and pass it some
 *              data it may need to act upon.  A status is returned
 *              that can be any one of
 *
 *              COMMAND_STATUS_BAD_OBJECT
 *              COMMAND_STATUS_BAD_EVENT
 *              COMMAND_STATUS_ABORT
 *
 *              The first two, naturally, mean that you schedules an
 *              event for a non-existant object or you scheduled a
 *              non-existent or bad event for a given object.  The
 *              third return code means that you scheduled a command
 *              for SPEEDES but SPEEDES had exited.
 * Return: Status of command as given above.
*************************************************************************/
  int SendCommand(char *CommandName, //In: Name of command to be scheduled
		  int GlobalId,      //In: Global id of Object for which you
		                     //    are scheduling the command
		  char *Message,     //In: The message/data for the command
		  int Bytes,         //In: The size of the message in bytes
		  int ReturnStatus=0 //In: Set to 1 to enable error checking
		  );
  

/************************************************************************
 * Method: SendCommand(char *, char *, char *, int, int)
 * Summary: 
 * Description: Within SPEEDES, objects can respond to named commands
 *              similar to how the STATE_MANGER (OPSM) can respond to
 *              named commands.  This is how you can schedule events
 *              for a given object within SPEEDES and pass it some
 *              data it may need to act upon.  A status is returned
 *              that can be any one of
 *
 *              COMMAND_STATUS_BAD_OBJECT
 *              COMMAND_STATUS_BAD_EVENT
 *              COMMAND_STATUS_ABORT
 *
 *              The first two, naturally, mean that you schedules an
 *              event for a non-existant object or you scheduled a
 *              non-existent or bad event for a given object.  The
 *              third return code means that you scheduled a command
 *              for SPEEDES but SPEEDES had exited. 
 * Return: Status of command as described in the Description.
*************************************************************************/
  int SendCommand(char *CommandName, //In: Name of command to be scheduled
		  char *ObjectName,  //In: Name of object for which you are
		                     //    scheduling the command
		  char *Message,     //In: Message or data for command
		  int Bytes,          //In: Size of mesage/data in bytes
		  int ReturnStatus=0 //In: Set to 1 to enable error checking
		  );
  

    int ReceiveCommandStatus();


/************************************************************************
 * Method: SendTwoWayCommand(char *, int, char*, int, int, double)
 * Summary: Sends a command and then waits for a reply
 * Description: SendTwoWayCommand has several options with the 5th
 *              parameter, BarrierMode.  This parameter determines how
 *              tightly the command will be handled with respect to
 *              when it is received.
 *
 *              If BarrierMode is NO_BARRIER, then SPEEDES simply tags
 *              the event at GVT and it is processed.  If BarrierMode
 *              is SYNCHRONIZED_BARRIER, then SPEEDES schedules the
 *              event at the time the message is recieved + TimeLag.
 *              If BarrierMode is AGGRESSIVE_BARRIER, then the event
 *              is scheduled at gvt+TimeLag.  Synchronized and
 *              Aggressive both also set up barriers within SPEEDES
 *              that are lifted when the event is processed and the
 *              reply is sent back to the HostRouter???  IS THIS
 *              RIGHT???
 * Return: Status of command.  See description of SendCommand() for details.
*************************************************************************/
  int SendTwoWayCommand(char *CommandName, //In: Name of command to be
			                   //    scheduled
			int GlobalId,      //In: Global id of object to
			                   //    receive the command
			char *Message,     //In: Message/data for command
			int Bytes,         //In: Size of message in bytes
			int BarrierMode,   //In: barrier mode as described
			                   //    above
			double TimeLag     //In: Amount of time lag for
			                   //    Synchronized and Aggressive
			                   //    barriers.
			);
  

/************************************************************************
 * Method: SendTwoWayCommand(char *, char *, char*, int, int, double)
 * Summary: Sends a command and then waits for a reply
 * Description: SendTwoWayCommand has several options with the 5th
 *              parameter, BarrierMode.  This parameter determines how
 *              tightly the command will be handled with respect to
 *              when it is received.
 *
 *              If BarrierMode is NO_BARRIER, then SPEEDES simply
 *              schedules the event at GVT and it is processed.  If
 *              BarrierMode is SYNCHRONIZED_BARRIER, then SPEEDES
 *              schedules the event at the current scheduled barrier
 *              (if there is none, then at GVT) and constructs another
 *              barrier at the time the event was schedules to
 *              occur+TimeLag.  If BarrierMode is AGGRESSIVE_BARRIER,
 *              then the event is scheduled at GVT and a barrier is
 *              constructed at GVT+TimeLag
 * Return: Status of command.  See description of SendCommand() for details.
*************************************************************************/
  int SendTwoWayCommand(char *CommandName, //In: Name of command to be
			                   //    scheduled
			char *ObjectName,  //In: Name of object to
			                   //    receive the command
			char *Message,      //In: Global id of object to
			                   //    receive the command
			int Bytes,         //In: Size of message in bytes
			int BarrierMode,   //In: barrier mode as described
			                   //    above
			double TimeLag     //In: Amount of time lag for
			                   //    Synchronized and Aggressive
			                   //    barriers.
			);
  

/************************************************************************
 * Method: GetTime()
 * Summary: Queries SPEEDES for the current GVT
 * Description: 
 * Return: Returns GVT.
*************************************************************************/
  double GetTime();


  /************************************************************************
 * Method: GetMinMsgTime()
 * Summary: Queries SPEEDES for the minimum time of an outgoing message
 * Description: 
 * Return: Returns MinMsgTime.
*************************************************************************/
  SpSimTime &GetMinMsgTime(void);


/************************************************************************
 * Method: TimeAdvanceRequest(double)
 * Summary: Requests that SPEEDES advance time by TimeLag
 * Description: SPEEDES will attempt to advance its GVT by the passed
 *              in timelag.  It will then block until the same
 *              EmHostUser calls TimeAdvanceGranted().
 * Return: none
*************************************************************************/
  void TimeAdvanceRequest(double TimeLag //In: Amount of time SPEEDES should
			                 //    advance GVT.
			  );
  

  
/************************************************************************
 * Method: TimeAdvanceGranted() 
 * Summary: Returns 1 if the grant has been received, 0 otherwise.  Removes
 *          the barrier raised by TimeAdvanceRequest().
 * Description:
 * Return: 1 if grant received, 0 otherwise.
*************************************************************************/
  int TimeAdvanceGranted();

  
  
/************************************************************************
 * Method: GetTimeGranted()
 * Summary: Returns the time granted by SPEEDES.
 * Description: 
 * Return: time granted by SPEEDES
*************************************************************************/
  SpSimTime GetTimeGranted() {return Time;}
  
/************************************************************************
 * Method: Subscribe (char *)
 * Summary: Subscribe to a named subscription type for all objects
 * Description: If every object has a named subscription called
 *              PercentDamaged, then calling
 *              Subscribe("PercentDamaged") will subscribe to that
 *              named data for every object.
 * Return: status of the operation  Either SUBSCRIBE_STATUS_GOOD or
 *         SUBSCRIBE_STATUS_BAD_TYPE
*************************************************************************/
  int Subscribe(char *Name //In: Named subscription type to subscribe to
		);
  
/************************************************************************
 * Method: SubscribeObject(char *, char *)
 * Summary: Subscribe to a named subscription type for a specified object.
 * Description: If you only want to subscribe to subscribable data for
 *              a single object, this allows you to do that.  Simply
 *              plug in the name of the subscription type and the
 *              object name and you will be signed up.
 * Return: status of the operation  Either SUBSCRIBE_STATUS_GOOD or
 *         SUBSCRIBE_STATUS_BAD_TYPE
*************************************************************************/
  int SubscribeObject(char *Name,      //In: Name of subscription type
		      char *ObjectName //In: name of object to whome you
		                       //    want to subscribe
		      );
  

/************************************************************************
 * Method: SubscribeObject(char *, int, int, int)
 * Summary: Subscribe to named subscription type for a specified object
 *          by <ObjectType, LocalId, Node>
 * Description: 
 * Return: status of the operation: either SUBSCRIBE_STATUS_GOOD or
 *         SUBSCRIBE_STATUS_BAD_TYPE 
*************************************************************************/
  int SubscribeObject(char *Name,     //In: Name of subscription type
		      int ObjectType, //In: Object type to subscribe to
		      int LocalId,    //In: Local ID of that object
		      int Node        //In: SPEEDES node number of the object
		      );
  

  

/************************************************************************
 * Method: SubscribeType(char *, char *)
 * Summary: Subscribe to all objects of a specified object manager type
 * Description: This method will allow you to subscribe to named data
 *              for all objects of a given type.
 * Return: status of the operation: either SUBSCRIBE_STATUS_GOOD or
 *         SUBSCRIBE_STATUS_BAD_TYPE
*************************************************************************/
  int SubscribeType(char *Name, //In: Name of the subscription type
		    char *Type  //In: Name of the object type to subscribe to
		    );
  

/************************************************************************
 * Method: UnSubscribe(char*)
 * Summary: UnSubscribe from a named subscription type for all objects.
 * Description: 
 * Return: status of the operation: UNSUBSCRIBE_STATUS_GOOD
*************************************************************************/
  int UnSubscribe(char *Name //In: Name of teh subscription type
		  );
  
  

/************************************************************************
 * Method: UnsubscribeObject(char *, char *)
 * Summary: UnSubscribe from a named subscription type for a specified
 *          object by name
 * Description: 
 * Return: status of the operation: UNSUBSCRIBE_STATUS_GOOD
*************************************************************************/
  int UnSubscribeObject(char *Name,      //In: Name of the subscription type
			char *ObjectName //In: Name of the object for whom you
			                 //    want to unsubscribe
			);
  
  

/************************************************************************
 * Method: UnSubscribeObject(char *, int, int, int)
 * Summary: UnSubscribe from a named subscription type for a specified
 *          object by <ObjectType, LocalId, Node>.
 * Description: 
 * Return: status of the operation: UNSUBSCRIBE_STATUS_GOOD
*************************************************************************/
  int UnSubscribeObject(char *Name,
			int ObjectType,
			int LocalId,
			int Node);
  
  

/************************************************************************
 * Method: UnSubscrieType(char *, char *)
 * Summary: Unsubscribe from all objects of a specified object manager type
 * Description: 
 * Return: status of the operation: UNSUBSCRIBE_STATUS_GOOD
*************************************************************************/
  int UnSubscribeType(char *Name, //In: Name of the subscription type
		      char *Type  //In: Name of type you want to unsubscribe
		      );
  
  

/************************************************************************
 * Method: ReceiveSubscribedData(char*, double&, int&, int&)
 * Summary: Returns specifically named subscribed data sent by SPEEDES
 *          using the SendSubscribedData call in the SpHostUser
 * Description: Within SPEEDES, one can issue
 *              HostUser->SendSubscribedData() which will send the
 *              data to the host router.  ReceiveSubscribedData will
 *              receive the data within the EmHostUser module.  The
 *              values Time, Bytes, and GlobalId will be filled in
 *              with the time the data was created, the number of
 *              bytes of the data, and the global id of the object
 *              that created that data. Once you are done with the
 *              data, you can delete it with delete [].
 * Return: Subscribed data
*************************************************************************/
  char *ReceiveSubscribedData(char *Name,   //In: Name of subscribed data
			      double &Time, //Out: Time subscribed data was
			                    //     created
			      int &Bytes,   //Out: Number of bytes of
			                    //     subscribed data
			      int &GlobalId //Out: Global ID of the object that
			                    //     created this data.
			      );
  
  

/************************************************************************
 * Method:ReceiveAnySubscribedData(char*&, double&, int&, int&) 
 * Summary:  Returns any subscribed data send by SPEEDES.
 * Description: As with ReceiveSubscribedData(), SPEEDES can call
 *              HostUser->SendSubscribedData() and that data can be
 *              received with this EmHostUser call.  The name of the
 *              data subscribed to is put in the value Name and is
 *              allocated with malloc (you need to delete it with
 *              free, not delete []!).  The data can then be processed
 *              and deleted with delete [] when you no longer need it.
 * Return: Subscribed data.
*************************************************************************/
  char *ReceiveAnySubscribedData(char *&Name, double &Time, int &Bytes,
				 int &GlobalId);



/************************************************************************
 * Method: GoToTimeRequest(double, double)
 * Summary:  Requests SPEEDES advance time and then construct another barrier
 * Description: GoToTimeRequest(newTime, timelag) requests that
 *              SPEEDES advance time to newTime, reply with a
 *              GoToTimeGranted call and then create another barrier
 *              at newTime+timeLag.  If SPEEDES has already passed
 *              newTime, then a reply is immediatly sent with gvt as
 *              the time and a barrier is constructed at gvt+timeLag.
 * Return: void
*************************************************************************/
  void GoToTimeRequest(double newTime, //In: Time to which SPEEDES should
		                       //    advance
		       double timeLag  //In: How far after newTime should
		                       //    SPEEDES construct another
		                       //    barrier?
		       );



/************************************************************************
 * Method: GoToTime(double, double);
 * Summary:  Requests SPEEDES advance to newTime and blocks until it is reached
 * Description: This method is identical to GoToTimeRequest except that
 *              this method blocks until newTime is reached.
 * Return: void
*************************************************************************/
    void GoToTime(double newTime, //In: Time to which SPEEDES should advance
		  double timeLag  //In: How far after newTime should SPEEDES
		                  //    construct another barrier?
		  );



/************************************************************************
 * Method: GoToTimeGranted()
 * Summary:  Returns 1 if a GoToTimeRequest has been granted, 0 otherwise
 * Description: Returns 1 if a GoToTimeRequest has been granted, 0 otherwise
 * Return: Returns 1 if a GoToTimeRequest has been granted, 0 otherwise
*************************************************************************/
  int GoToTimeGranted();


/************************************************************************ 
   * Method: ScheduleEvent(double, int, char*, char*, int, char*, int, TimeMode, int)
   * Summary: Schedules an event at a specified time, returning a
   *          cancel handle
   * Description: You can schedule an event at any point in time from
   *              GVT onwards.  This will return a
   *              SCHEDULE_EVENT_REPLY_MESSAGE* which you will need to
   *              delete when you are finished with it (the structure
   *              of the class SCHEDULE_EVENT_REPLY_MESSAGE can be
   *              found in SpHostRouterMsgs.H).  You can pass in the
   *              global id of the object along with the msg (the M_*
   *              class) and its size along with any data and its size
   *              for that event.  This will return a cancel handle in
   *              SCHEDULE_EVENT_REPLY_MESSAGE if requested and you
   *              can then use CancelEvent() to cancel this event.
   * Return: a SCHEDULE_EVENT_REPLY_MESSAGE if a CancelHandle is
   *         requested.  This includes the cancel handle along with
   *         any error code for the cancel.  These codes can be found
   *         in SpHostRouterMsgs.H
 *************************************************************************/ 

  SCHEDULE_EVENT_REPLY_MESSAGE*
  ScheduleEvent(double Time,	//In: Time at which the event will be scheduled
		int globalId,	//In: Global ID for which this event is
				//    scheduled
		char * eventName,// In: Name of the event to be scheduled
		char * msg,	// In: The message for the event (the M_ class)
		int msgBytes,	// In: The size of the message for the event
		char * data,	// In: Any data for this event
		int dataBytes,	// In: The number of bytes of data
		TimeMode = IF_IN_PAST_IGNORE, //Iin: How to schedule the event
				               // If it is equal to
					       // IF_IN_PAST_IGNORE, then the
					       // event will be ignored if it
					       // arrives in the past. If it is
					       // equal to the value
					       // IF_IN_PAST_SCHEDULE_AT_GVT
					       // then the event will be
					       // scheduled at GVT if the time
					       // it is scheduled is before
					       // GVT
		int CancelHandleRequested=1    // In: If equal to 1, a cancel
				               // is returned.  Otherwise,
					       // NULL is returned.
		);
/************************************************************************ 
   * Method: ScheduleEvent(double, char*, char*, char*, int, char*, int, TimeMode, int)
   * Summary: Schedules an event at a specified time, returning a
   *          cancel handle
   * Description: You can schedule an event at any point in time from
   *              GVT onwards.  This will return a
   *              SCHEDULE_EVENT_REPLY_MESSAGE* which you will need to
   *              delete when you are finished with it (the structure
   *              of the class SCHEDULE_EVENT_REPLY_MESSAGE can be
   *              found in SpHostRouterMsgs.H).  You can pass in the
   *              global id of the object along with the msg (the M_*
   *              class) and its size along with any data and its size
   *              for that event.  This will return a cancel handle in
   *              SCHEDULE_EVENT_REPLY_MESSAGE if requested and you
   *              can then use CancelEvent() to cancel this event.
   * Return: a SCHEDULE_EVENT_REPLY_MESSAGE if a CancelHandle is
   *         requested.  This includes the cancel handle along with
   *         any error code for the cancel.  These codes can be found
   *         in SpHostRouterMsgs.H
 *************************************************************************/ 

  SCHEDULE_EVENT_REPLY_MESSAGE*
  ScheduleEvent(double Time,	//In: Time at which the event will be scheduled
		char* objName,	//In: Name ofthe object for which this event is
				//    scheduled
		char * eventName,// In: Name of the event to be scheduled
		char * msg,	// In: The message for the event (the M_ class)
		int msgBytes,	// In: The size of the message for the event
		char * data,	// In: Any data for this event
		int dataBytes,	// In: The number of bytes of data
		TimeMode = IF_IN_PAST_IGNORE, //Iin: How to schedule the event
				              // If it is equal to
					      // IF_IN_PAST_IGNORE, then the
					      // event will be ignored if it
					      // arrives in the past. If it is
					      // equal to the value
					      // IF_IN_PAST_SCHEDULE_AT_GVT
					      // then the event will be
					      // scheduled at GVT if the time
					      // it is scheduled is before
					      // GVT
		int CancelHandleRequested=1   // In: If equal to 1, a cancel
				              // is returned.  Otherwise,
					      // NULL is returned.
		);

/************************************************************************ 
 * Method: CancelEvent(int, int)
 * Summary:  cancels an event that was previous scheduled through SchedulEvent
 * Description: This method will cancel an event that was previously
 *              scheduled externally using SchedulEvent() and a cancel
 *              handle was requested.  The handle that was returned
 *              from this other process can then be used to cancel with
 *              this method.  An error condition can be returned if
 *              desired but the default is to not send an error
* Return: a CANCEL_EVENT_REPLY_MESSAGE if requested and NULL otherwise.
*************************************************************************/ 
  

  CANCEL_EVENT_REPLY_MESSAGE* CancelEvent(int handle, //In: Cancel handle of
			                              // the event to be
						      // canceled
					  int replyRequested=0 //In: 1 if you
				                               //    want a
							       //    reply, 0
							       //    otherwise
					  );  
  

  //Needed so that SpStateMgr::GoToTime() can get access to
  //SpStateMgr::Time
  friend class SpStateMgr;

protected:

  GENERIC_MSG_HEADER* GetMessageBlocking();
  void WriteSpeedesMessageToFile(GENERIC_MSG_HEADER* msg);
  GENERIC_MSG_HEADER* GetSpeedesMessageFromFile();

  CANCEL_EVENT_REPLY_MESSAGE* GetCancelReply();
  SCHEDULE_EVENT_REPLY_MESSAGE* GetScheduleReply();  

private:
  CLIENT_SOCK *ClientSock;	// Socket connection to host router
  FILE *fp_output;		// Output file pointer
  FILE *fp_input;		// Input file pointer
  int InputFileFlag;		// Flag indicating if using input file
  int OutputFileFlag;		// Flag indicating if using output file
  int GroupId;		        // Group Id for connecting to Host Router
  
  double Time;		        // Time granted or received by SPEEDES
  SpSimTime HlaMinMsgTime;      // Min Msg TimeTime received from SPEEDES
  int DisconnectBarrierFlag;	// Flag indicating need for Disconnection
  int SpeedesExecutingFlag;	// Flag indicating if SPEEDES is Executing
    C_QUEUE MessageQueue;	// Messages that are queued for reading
  
  void ExternalModuleConnection();
  int SendSubscribe(SUBSCRIPTION_ITEM *SubItem);
  int SendUnSubscribe(SUBSCRIPTION_ITEM *SubItem);
  
};

typedef SpEmHostUser EM_HOST_USER;

#endif
