// free_list.H header file

#ifndef free_list_object
#define free_list_object

#include "SpString.H"
#include "SpIostream.H"
#include "SpFstream.H"
#include "SpStrstream.H"
#include "SpIomanip.H"

#include "queue.H"

/*****************************************************************************
* Class: C_FREE_LIST
* Summary: Manager of all of the freelists on a node
* Description: This class can manage multiple freelist object types (ie, manage
*              many freelists, each with a separate freelist type ID
*              corresponding to a C++ class). new_object () and delete_object()
*              are the fundamental calls for the C_FREE_LIST class. They replace
*              calling new and delete directly by saving no-longer-needed
*              objects when calling delete_object() and then returning pointers
*              to those no-longer-needed objects when calling new_object()
*              (which is much faster than dynamically allocating a new one) if
*              possible. If a freelist object type has no available old copies,
*              then calling new_object() dynamically creates a new one and
*              returns it. The name "freelist" refers to the list of currently
*              no-longer-needed, but not deleted, objects that are available
*              for future use.  Objects that go into free lists must descend
*              from C_ITEM, and NOT have multiple inheritance.
*****************************************************************************/

typedef void * (*FREE_OBJECT_CONSTRUCTOR_FUNCTION)(int n);

/*****************************************************************************
* Internal Class: ARRAY_OF_CHILDREN
* Summary: Holds an array of children.
* Description: If a C_QUEUE having an object of this class is cleared,
*              then the array of children in this class will be
*              deleted correctly.
*****************************************************************************/
class ARRAY_OF_CHILDREN : public C_ITEM {
  public:
/******************************************************************************
* Method: GetArray
* Summary: 
* Description: descendants of ARRAY_OF_CHILDREN must override this to
*              return their data member.  That data member is an array.
* Return: An array of objects of a class descending from C_ITEM
******************************************************************************/
    virtual C_ITEM *GetArray()=0;
};

typedef ARRAY_OF_CHILDREN * (*TO_BE_DELETED_FUNCTION)(int n);

/******************************************************************************
* Macro: FREE_DEFINE_CLASS
* Summary: Declares that a class will be used in a free list.
* Description: This macro creates a function having the name of the
*              first macro argument.  You may pass that function to
*              C_FREE_LIST::set_type.  The second argument is the name
*              of the class that you want to be able to have in a free
*              list.
******************************************************************************/
#define FREE_DEFINE_CLASS(globalFunction, /* func to pass to set_type      */ \
			  type            /* class that goes into free list*/ \
			  )                                                   \
class type ##_ARRAY_OF_CHILDREN : public ARRAY_OF_CHILDREN {                  \
  public:                                                                     \
    type ##_ARRAY_OF_CHILDREN(int n) : A(new type[n]) {}                      \
    ~type ##_ARRAY_OF_CHILDREN() { delete [] A; }                             \
  private:                                                                    \
    C_ITEM *GetArray() { return A; }                                          \
    type *A;                                                                  \
};                                                                            \
                                                                              \
ARRAY_OF_CHILDREN *globalFunction(int n)                                      \
{return new type ##_ARRAY_OF_CHILDREN(n);}


class C_FREE_LIST {

 public:

/******************************************************************************
* Method: set_ntypes
* Summary: Set (and create) n types of freelist objects
* Description: When called the first time, this function allocates an
*              array of n elements of size, neach, ncreated, name,
*              constructor, ToBeDeleteds, qfree. These are the
*              necessary attributes for each freelist object
*              type. When called after the first time, it moves these
*              arrays to new places in memory and adds or subtracts
*              elements to the new array.
* Return: None
******************************************************************************/
  void set_ntypes(int n  // number of freelist types
		  );

/******************************************************************************
* Method: set_type
* Summary: Sets the name, constructor, size, and neach for a specified freelist
*          object type
* Description: There are two versions of this function.  Use the first
*              one for backwards compatibility with old SPEEDES
*              applications.  Use the second one if you also used the
*              FREE_DEFINE_CLASS macro.
* Return: None
******************************************************************************/
  void set_type(char *nam,        //
		int type,         //
		FREE_OBJECT_CONSTRUCTOR_FUNCTION f, //
		int sz,           //
		int n             //
		);
  void set_type(char *nam,        //
		int type,         //
		TO_BE_DELETED_FUNCTION f, //
		int sz,           //
		int n             //
		);
/******************************************************************************
* Overloaded Method: get_size
* Summary: Get the size of a freelist object type
* Description: 
* Return: The size of the specified freelist object type
******************************************************************************/
  int get_size(int type  // freelist object type ID
	       ) {return size[type];}

/******************************************************************************
* Overloaded Method: get_size
* Summary: Get the size of a freelist object type
* Description: The specified object must inherit from C_ITEM, or SPEEDES has
*              undefined behavior. If it is a C_ITEM but not part of a
*              freelist, -1 is returned.
* Return: The size of the freelist object type of the specified object, or -1
*         if this object is not part of a freelist
******************************************************************************/
  int get_size(C_ITEM *object  // pointer to an object
	       );

/******************************************************************************
* Method: GetNames
* Summary: Get a pointer to the first element of the array of names of
*          freelists
* Description: 
* Return: Freelist name array
******************************************************************************/
  char **GetNames() {return name;}

/******************************************************************************
* Method: GetName
* Summary: Get the name of the i'th freelist
* Description: 
* Return: The name of the i'th freelist
******************************************************************************/
  char *GetName(int i // element ordering of the desired freelist name
		) {return name[i];}

/******************************************************************************
* Method: GetNtypes
* Summary: Get the number of type IDs of freelists
* Description: 
* Return: The number of type IDs of freelists (ntypes)
******************************************************************************/
  int GetNtypes() {return ntypes;}

/******************************************************************************
* Method: CopyType
* Summary: Create a new freelist type with a new name by copying the
*          constructor, ToBeDeleteds, size and neach attributes from
*          an old (already existing) freelist type
* Description: 
* Return: None
******************************************************************************/
  void CopyType(char *NewName, // name of new freelist type
		int NewType,   // type ID of new type
		int OldType);   // type ID of old type

/******************************************************************************
* Method: new_object
* Summary: Get or construct an object of an existing freelist object type ID
* Description: This and delete_object() are the fundamental calls for the
*              C_FREELIST class. They replace calling new and delete directly
*              by saving no-longer-needed objects when calling delete_object()
*              and then returning pointers to those no-longer-needed objects
*              when calling new_object() (which is much faster than dynamically
*              allocating a new one) if possible. If a freelist object type has
*              no available old copies, then calling new_object() dynamically
*              creates a new one and returns it. The name "freelist" refers to
*              the list of currently no-longer-needed, but not deleted, objects
*              that are available for future use.  Returned objects are casted
*              to (void *) and are not necessarily in their initialized state.
* Return: An object of the specified type
******************************************************************************/
  void *new_object(int type // an existing freelist object type ID
		   );

/******************************************************************************
* Overloaded Method: delete_object
* Summary: Put a no longer needed object, whose state currently contains a
*          freelist object type ID, into its freelist
* Description: This and new_object() are the fundamental calls for the
*              C_FREELIST class. They replace calling new and delete directly
*              by saving no-longer-needed objects when calling delete_object()
*              and then returning pointers to those no-longer-needed objects
*              when calling new_object() (which is much faster than dynamically
*              allocating a new one) if possible. If a freelist object type has
*              no available old copies, then calling new_object() dynamically
*              creates a new one and returns it. The name "freelist" refers to
*              the list of currently no-longer-needed, but not deleted, objects
*              that are available for future use.  Returned objects are casted
*              to (void *) and are not necessarily in their initialized
*              state. Use this version of this method when the object already
*              has a freelist object type ID. If this object has used the other
*              version of this method before, then this version can be used
*              thereafter. Or, if its freelist object type ID has been set by
*              some other means, this version can be used.
* Return: None
******************************************************************************/
  void delete_object(void *object  // object to put into its freelist, casted
		                   // to (void *)
		     );

/******************************************************************************
* Overloaded Method: delete_object
* Summary: Put a no longer needed object, which has possibly never participated
*          in freelists before, into a freelist specified by freelist object
*          type ID
* Description: This and new_object() are the fundamental calls for the
*              C_FREELIST class. They replace calling new and delete directly
*              by saving no-longer-needed objects when calling delete_object()
*              and then returning pointers to those no-longer-needed objects
*              when calling new_object() (which is much faster than dynamically
*              allocating a new one) if possible. If a freelist object type has
*              no available old copies, then calling new_object() dynamically
*              creates a new one and returns it. The name "freelist" refers to
*              the list of currently no-longer-needed, but not deleted, objects
*              that are available for future use.  Returned objects are casted
*              to (void *) and are not necessarily in their initialized
*              state. Use this version of this method when the object does not
*              yet have (or may not have) a freelist object type ID, but is the
*              type associated with the specified freelist object type ID. If
*              this object has never set its freelist object type ID and has
*              never been the calling argument in this version of this method
*              before, then this version must be used, and then the other
*              version of this method can be used thereafter.
* Return: None
******************************************************************************/
  void delete_object(void *object,  // an object of the type associated with
		                    // the freelist object type ID specified
		                    // by this method's second argument,
		                    // casted to (*void)
		     int type);     // the freelist's freelist object type ID
                                    // into which this object is to be placed 

/******************************************************************************
* Method: delete_list
* Summary: 
* Description: 
* Return: None
******************************************************************************/
  void delete_list(C_QUEUE *q) {delete_list(q,-1);}

/******************************************************************************
* Method: delete_list
* Summary: Add a list of objects back into a free list
* Description: 
* Return: None
******************************************************************************/
  void delete_list(C_QUEUE *q, // object list to delete
		   int type);  // objec\t type ID to delete

/******************************************************************************
* Method: generate
* Summary: Generate a free list object from a buffer
* Description: This takes a character buffer that represents the memory buffer
*              for an object that inherited from C_ITEM, and memcopies the
*              buffer over the proper free list object type and returns it. The
*              buffer is normally from an event message. C_ITEMs inherit from
*              C_DUMMY_V, which is a dummy class that puts the V table at the
*              top of C_ITEM objects. This way, objects that have virtual
*              functions can still be constructed with this function.
* Return: Pointer to an object of unknown type casted to (void *)
******************************************************************************/
  void *generate(char *buff // buffer from which to generate an object
		 );

/******************************************************************************
* Method: print
* Summary: Print out status information for the free list
* Description: 
* Return: None
******************************************************************************/
  void print(ostream &out = cout  // ostream to which to print
	     );
  
  C_FREE_LIST();
  ~C_FREE_LIST();

 private:
    void *f(int type, int n);

 C_QUEUE ToBeCleared;
 C_QUEUE *qfree;
 int *size;
 int *neach;
 FREE_OBJECT_CONSTRUCTOR_FUNCTION *constructor;
 TO_BE_DELETED_FUNCTION *ToBeDeleteds;
 int *ncreated;
 int ntypes;
 char **name;

};


#endif
