#include "SpGeoFunctions.H"

double get_delta_rad_lat_of_sphere
  (const double sphere_radius,
   const double distance_from_center_of_sphere_to_center_of_earth)
{
  //                Center of sphere
  //                |\
  //                | \
  //                |  \
  //  sphere_radius |   \ distance_from_center_of_sphere_to_center_of_earth
  //                |    \
  //                |     \
  //                |    A \
  //                R-------Center of earth
  //                    s
  //
  //   In this diagram the north pole would be near the top of the
  // diagram and the south pole would be near the bottom of the
  // diagram.
  //   A planar triangle is shown.
  //   The center of the sphere is labeled.  This sphere of
  // radius sphere_radius intersects vertex R since vertex R is a distance
  // of sphere_radius from the center of the sphere.
  //   Let side s be tangent to the sphere.  That means that angle A
  // represents the maximum number of latitude radians that the
  // sphere covers in the southern direction.  There is a geometric
  // theorem which says that a line of tangency is perpendicular to a
  // radius intersecting the point of tangency.  Because side s is
  // tangent to the sphere and because side sphere_radius is a radius of
  // that sphere, side s must be perpendicular to side sphere_radius, which
  // means that angle R is a right angle.  Therefore the triangle is
  // a right triangle which means the following formula applies:
  //
  // sin(A) = sphere_radius
  //        / distance_from_center_of_sphere_to_center_of_earth    
  //
  //   If delta_rad_lat = A, then the following expression is delta_rad_lat:
    
  return
    asin(sphere_radius / distance_from_center_of_sphere_to_center_of_earth);
}

double get_delta_rad_lng_of_sphere
  (const double sphere_radius,
   const double distance_from_center_of_sphere_to_axis_of_earth)
{
  //                Center of sphere
  //                |\
  //                | \
  //                |  \
  //  sphere_radius |   \ distance_from_center_of_sphere_to_axis_of_earth
  //                |    \
  //                |     \
  //                |    A \
  //                R-------Point on the Earth's axis of rotation
  //                    s
  //
  //   In this diagram the Earth's axis of rotation is viewed on end
  // which means it looks like just a point.  I.e., you are above the
  // North pole looking down.
  //   A planar triangle is shown.  Side s and side
  // distance_from_center_of_sphere_to_axis_of_earth are both
  // perpendicular to the Earth's axis of rotation.
  //   The center of the sphere is labeled.  This sphere of
  // radius sphere_radius intersects vertex R since vertex R is a
  // distance of sphere_radius from the center of the sphere.
  //   Let side s be tangent to the sphere.  That means that angle A
  // represents the maximum number of longitude radians that the sphere
  // covers in the eastern direction.  There is a geometric theorem
  // which says that a line of tangency is perpendicular to a radius
  // intersecting the point of tangency.  Because side s is tangent to
  // the sphere and because side sphere_radius is a radius of that
  // sphere, side s must be perpendicular to side sphere_radius, which
  // means that angle R is a right angle.  Therefore the triangle is a
  // right triangle which means the following formula applies:
  //
  // sin(A) = sphere_radius
  //        / distance_from_center_of_sphere_to_axis_of_earth    
  //
  //   If delta_rad_lng = A, then the following expression is delta_rad_lng:
    
  return
    asin(sphere_radius / distance_from_center_of_sphere_to_axis_of_earth);
}

double get_distance_from_axis
  (const double rad_lat,
   const double distance_from_input_point_to_center_of_earth)
{
  //                                                    s
  //                                      input point-------R
  //                                                 \lat   |
  //                                                  \     |
  //      distance_from_input_point_to_center_of_earth \    |
  //                                                    \   |Earth's axis
  //                                                     \  |
  //                                                      \ |
  //                                                   lat \|
  //                                                  ----- Center of earth
  //
  //   In this diagram the north pole would be near the top of the
  // diagram and the south pole would be near the bottom of the
  // digram.
  //   A planar triangle is shown.  Side s is perpendicular to the
  // Earth's axis, which means the angle at R is a right angle
  // which means it is a right triangle.
  //   There are two angles labeled lat.  The horizontal line
  // leaving the Center of earth is perpendicular to the Earth's
  // axis, which means that the bottom most angle lat is the
  // latitude of the input point.
  //   Side s and the line leaving the input point are
  // parallel since they are both perpendicular to the Earth's
  // axis.  Side distance_from_input_point_to_center_of_earth
  // contains a line intersecting these two parallel lines which
  // means that a geometric theorem says that alternating angles
  // are congruent.  The other angle labeled lat is an
  // alternating angle so therefore the two angles labeled lat are
  // congruent.
  //   Since the triangle is a right triangle the following
  // formula applies:
  //
  // cos(lat) = s / distance_from_input_point_to_center_of_earth
  //
  //   If s = distance_from_axis_to_input_point
  // then the following formula applies:

  const double distance_from_axis_to_input_point
    = cos(rad_lat) * distance_from_input_point_to_center_of_earth;

  return distance_from_axis_to_input_point;
}

void get_delta_rad_angles_of_sphere
  (const double rad_lat,
   const double sphere_radius,
   const double distance_from_center_of_sphere_to_center_of_earth,
   double &delta_rad_lat,
   double &delta_rad_lng)
{
  if (sphere_radius >= distance_from_center_of_sphere_to_center_of_earth)
  {
    // The center of the earth is within the sphere.  All latitudes
    // and all longitudes exist at the center of the earth.
    // Therefore, this sphere can see all latitudes and all
    // longitudes.
    delta_rad_lat = PI;
    delta_rad_lng = PI;
  }
  else
  {
    delta_rad_lat = get_delta_rad_lat_of_sphere
      (sphere_radius, distance_from_center_of_sphere_to_center_of_earth);
    
    double distance_from_axis_of_earth_to_center_of_sphere
      = get_distance_from_axis
      (rad_lat, distance_from_center_of_sphere_to_center_of_earth);
    
    if (distance_from_axis_of_earth_to_center_of_sphere == 0
	|| sphere_radius >= distance_from_axis_of_earth_to_center_of_sphere)
    {
      // The axis of the earth is within the sphere.  All longitudes
      // exist at the axis of the earth.  Therefore, this sphere can
      // see all longitudes.
      delta_rad_lng = PI;
    }
    else
    {
      delta_rad_lng = get_delta_rad_lng_of_sphere
	(sphere_radius, distance_from_axis_of_earth_to_center_of_sphere);
    }
  }
}

