// random.C method file

#include "SpRandom.H"

static const double SP_PI = 3.141592653589793;
static const double SP_TWO_PI = SP_PI * 2.0;
static const double SP_ONE_OVER_ROOT_TWO_PI = 0.39894228;

static const int SP_IB1            = 1;
static const int SP_IB2            = 2;
static const int SP_IB5            = 16;
static const int SP_IB18           = 131072;

//----------------------------------------------------------------------------
// Generate a random bit
//----------------------------------------------------------------------------
int SpBaseRandom::GenerateBit() {
  int seed = GetSeed();
  unsigned int newbit;

  newbit = (seed & SP_IB18) >> 17
        ^  (seed & SP_IB5) >> 4
        ^  (seed & SP_IB2) >> 1
	^  (seed & SP_IB1);
  seed = (seed << 1) | newbit;
  SetSeed(seed);
  return newbit;
}

//----------------------------------------------------------------------------
// Generate random bits that are packed into a returned integer. This method
// can be used to generate random integers of a specified bit length.
//----------------------------------------------------------------------------
int SpBaseRandom::GenerateBits(int numBits) {
  int i;
  int ranBits = 0;
  for (i=0; i<numBits; i++) {
    ranBits <<= 1;
    ranBits |= GenerateBit();
  }
  return ranBits;
}

//----------------------------------------------------------------------------
// Generate a random gaussian number with correct sigma and mean
//----------------------------------------------------------------------------
double SpBaseRandom::GenerateGaussian(double mean, double sigma) {
  double fac,r,v1,v2;

  do {
    v1 = 2.0 * GenerateUniform() - 1.0;
    v2 = 2.0 * GenerateUniform() - 1.0;
    r = v1*v1 + v2*v2;
  } while (r >= 1.0);
  fac = sqrt(-2.0*log(r)/r);
  return (v2*fac)*sigma + mean;
}

//----------------------------------------------------------------------------
// Compute the error function which is the integrated area under a normal
// distribution curve from 0 to x where the mean is assumed to be 0 and the
// standard deviation is assumed to be 1. Note, 0 <= Erf(x) <= 0.5.
//----------------------------------------------------------------------------
double SpBaseRandom::Erf(double x) {
  int i;
  double x2;
  double X;
  double Xlast;
  double Xret;
  double a;

  if (x>6) return 0.5;

  i=0;
  X = x;
  Xlast = 0.0;
  x2 = x*x;
  a = X;

  while (fabs(X-Xlast) > 0.000001) {
    Xlast = X;
    a *= (-x2*double(2*i+1)) / double(2*(i+1)*(2*i+3));
    X += a;
    i++;
  }

  Xret = SP_ONE_OVER_ROOT_TWO_PI * X;
  if (Xret <= 0.5) return Xret;
  return 0.5;
}

//----------------------------------------------------------------------------
// Compute the inverse error function (InvErf(Erf(x)) = x). This function
// is useful when you know a probability, but want to determine how many
// standard deviations it is from the mean. Note that 0 <= x <= 0.5.
//----------------------------------------------------------------------------
double SpBaseRandom::InvErf(double x) {
  double xguess;
  double Xret;
  double lower;
  double upper;
  double dx;

  if ((x>0.5) || (x<0.0)) {
    cout << "Error (InvErf): bad value " << x << endl;
  }

  if (x==0.5) return 1.0e20;
  if (x==0.0) return -1.0e20;

  for (;;) {

    if (x < 0.1) {
      lower = 0.0;
      upper = 0.26;
      break;
    }

    if (x < 0.2) {
      lower = 0.25;
      upper = 0.53;
      break;
    }

    if (x < 0.3) {
      lower = 0.52;
      upper = 0.85;
      break;
    }

    if (x < 0.4) {
      lower = 0.84;
      upper = 1.29;
      break;
    }

    if (x < 0.45) {
      lower = 1.28;
      upper = 1.65;
      break;
    }

    if (x < 0.49) {
      lower = 1.64;
      upper = 2.33;
      break;
    }

    if (x >= 0.49) {
      lower = 2.32;
      upper = 5.1;
      break;
    }

  }

  Xret = (upper+lower) / 2.0;
  dx = 1.0;

  upper += 0.1;
  lower -= 0.1;
  if ( lower < 0.0) lower = 0.0;

  while (fabs(dx) > 0.000001) {

    xguess = Erf(Xret);
    dx = (x-xguess) / (SP_ONE_OVER_ROOT_TWO_PI*exp(-Xret*Xret/2.0));
    if (xguess < x) {
      lower = Xret-0.001;
      if (lower < 0.0) lower = 0.0;
    }else{
      upper = Xret+0.001;
    }

    Xret += dx;

    if (Xret > upper) Xret = (upper+lower) / 2.0;
    if (Xret < lower) Xret = (upper+lower) / 2.0;

  }

  return Xret;

}

//----------------------------------------------------------------------------
// Factorial method computes n*(n-1)* ... 2*1
//----------------------------------------------------------------------------
double SpBaseRandom::Factorial(int n) {
  int i;
  double value;

  value = 1;
  for (i=2; i<=n; i++) value *= i;
  return value;

}

//----------------------------------------------------------------------------
// Generate a random vector
//----------------------------------------------------------------------------
void SpBaseRandom::GenerateVector(double vector[3], double magnitude) {
  double costheta, sintheta, phi, cosphi, sinphi;

  costheta = GenerateDouble(-1.0, 1.0);
  sintheta = sqrt(1.0-costheta*costheta);

  phi = GenerateDouble(0.0, SP_TWO_PI);
  cosphi = cos(phi);
  sinphi = sin(phi);

  vector[0] = magnitude*sintheta*cosphi;
  vector[1] = magnitude*sintheta*sinphi;
  vector[2] = magnitude*costheta;

}

//----------------------------------------------------------------------------
// Generate a random random beta distribution
//----------------------------------------------------------------------------
double SpBaseRandom::GenerateBeta(int n1, int n2) {
  int i;
  double t;
  double f, F;
  double sumCF;
  double rvalue;
  double tlo, thi, tnew;
  double error;
  int nits;
  static double *beta_coeff_f = 0;
  static double *beta_coeff_F = 0;
  static double beta_n1=-1;
  static double beta_n2=-1;

  if ((n1 != beta_n1) || (n2 != beta_n2)) {

// set the n1 and n2 values

    beta_n1 = n1;
    beta_n2 = n2;

// create the new array of beta cumulative density coefficients

    if (beta_coeff_f != 0) delete [] beta_coeff_f;
    beta_coeff_f = new double[n2+1];

    if (beta_coeff_F != 0) delete [] beta_coeff_F;
    beta_coeff_F = new double[n2+1];

// set up the coefficients for the density function f

    beta_coeff_f[0] = Factorial(n1+n2+1) / (Factorial(n1)*Factorial(n2));

    for (i=0; i<n2; i++) {
      beta_coeff_f[i+1] = -beta_coeff_f[i]*double(n2-i)/double(i+1);
    }

// set up the coefficients for the cummulative density function F

    for (i=0; i<=n2; i++) {
      beta_coeff_F[i] = beta_coeff_f[i] / (n1+i+1);
    }

// make sure that the coefficients are correct

    sumCF = 0.0;
    for (i=0; i<=n2; i++) {
      sumCF += beta_coeff_F[i];
    }
    if (fabs(1.0-sumCF) > 0.001) {
      cout << "Sum of F coeff = " << sumCF << " (should be 1)" << endl;
    }

    for (t=0.0; t<=1.0; t+=0.1) {
      ComputeBetatfF(t,f,F,beta_coeff_f,beta_coeff_F, beta_n1, beta_n2);
    }

  }

// now actually compute the random beta function

  rvalue = GenerateUniform();;

  t = 0.5;
  tlo = 0.0;
  thi = 1.0;
  error = 1.0;
  nits = 0;

  while ((error > 0.000001) && (nits < 100)) {
    nits++;
    ComputeBetatfF(t,f,F,beta_coeff_f,beta_coeff_F, beta_n1, beta_n2);
    tnew = t - (F-rvalue)/f;

    if (tnew < tlo) {
      thi = t;
      tnew = tlo + (thi-tlo)/2.0;
    }
    if (tnew > thi) {
      tlo = t;
      tnew = tlo + (thi-tlo)/2.0;
    }
    error = fabs(tnew-t);
    t = tnew;
  }

  return t;

}

//----------------------------------------------------------------------------
// Compute beta coeficients for f and F as a function of time
//----------------------------------------------------------------------------
void SpBaseRandom::ComputeBetatfF(double t, double &f, double &F,
				  double *beta_coeff_f,
				  double *beta_coeff_F,
				  double beta_n1,
				  double beta_n2) {
  int i;
  double tp;

  tp = 1.0;
  for (i=0; i<beta_n1; i++) tp*=t;

  f = 0.0;
  F = 0.0;
  for (i=0; i<=beta_n2; i++) {
    f += beta_coeff_f[i] * tp;
    tp *= t;
    F += beta_coeff_F[i] * tp;
  }

}

//----------------------------------------------------------------------------
// Generate a random number using the rejection method for an arbitrary
// density function.
//----------------------------------------------------------------------------
double SpBaseRandom::GenerateDensityFunction(SpDensityFunction *
					     densityFunction) {
  double x;
  double value;

  for (;;) {
    x = GenerateDouble(densityFunction->GetLoLimit(),
		       densityFunction->GetHiLimit());
    value = GenerateDouble(0.0, densityFunction->GetMaxAmplitude());
    if (value <= (*densityFunction)(x)) break;
  }

  return value;

}


// f(x) = (alpha/(2*pi)) * [1/(alpha^2 + x^2)]
double SpBaseRandom::GenerateCauchy(double alpha) {
  return alpha * tan(SP_TWO_PI * GenerateUniform() - SP_PI);}
