// Filename: Random.H
		
#ifndef RandomGate
#define RandomGate

#include "SpMath.H"
#include "SpIostream.H"
#include "SpFstream.H"
#include "SpStrstream.H"
#include "SpIomanip.H"

/*
** Constants used for generating random bits
*/


/*
** Constants used in Linear Congruential Generator
*/

static const int SP_IA = 2416;
static const int SP_IC = 374441;
static const int SP_IM = 1771875;

/*
** other constants
*/

static const double SP_SQRT_E = 1.648721271;

//----------------------------------------------------------------------------
// Class: SpDensityFunction
//
// This base class defines the virtual functions required for the rejection
// method used for generating some random numbers.
//----------------------------------------------------------------------------
class SpDensityFunction {

  private:
  protected:
  public:

    virtual double f(double x)=0;
    virtual double F(double x)=0;
    virtual double operator () (double x) {return f(x);}
    virtual double GetMaxAmplitude()=0;
    virtual double GetLoLimit()=0;
    virtual double GetHiLimit()=0;
};

class SpRayleighDensityFunction : public SpDensityFunction {

  private:

    double Alpha;
    double HiLimit;

  protected:
  public:

    SpRayleighDensityFunction(double alpha=1) {Alpha = alpha;}
    void SetAlpha(double alpha) {Alpha = alpha;}
    void SetHiLimit(double hiLimit) {HiLimit = hiLimit;}

    virtual double f(double x) {
      double value;
      if (0 < x) {
	double alpha2 = Alpha*Alpha;
	value = (x/alpha2)*exp(-(x*x)/(2*alpha2));
      }else{
	value = 0;
      }
      return value;
    }

    virtual double F(double) {return 0;}

    virtual double GetMaxAmplitude() {return 1.0/(Alpha*SP_SQRT_E);}
    virtual double GetLoLimit() {return 0;}
    virtual double GetHiLimit() {return HiLimit;}

};

//----------------------------------------------------------------------------
// Class: SpBaseRandom
//
// SpBaseRandom class provides all of the APIs for generating various random
// numbers. The method for generating uniform random numbers between [0,1]
// is supported through virtual functions in child classes. Also, the getting
// and setting of seeds is provided through virtual functions to support
// rollbackable versions of the random number generator. There is no state
// associated with SpBaseRandom, although it does support the generation of
// random bits.
//----------------------------------------------------------------------------
class SpBaseRandom {

  private:

    void ComputeBetatfF(double time,
			double &f,
			double &F,
			double *beta_coeff_f,
			double *beta_coeff_F,
			double beta_n1,
			double beta_n2);

  protected:

  public:

    virtual int GetSeed()=0;
    virtual void SetSeed(int seed)=0;
    virtual void ExchangeSeed(int &seed)=0;

    virtual operator int ()=0;

    int operator = (int seed) {SetSeed(seed); return seed;}

    int operator ++ () {
      int s = GetSeed() + 1;
      SetSeed(s);
      return s;
    }

    int operator ++ (int) {
      int s = GetSeed();
      SetSeed(s + 1);
      return s;
    }

    int operator -- () {
      int s = GetSeed() - 1;
      SetSeed(s);
      return s;
    }

    int operator -- (int) {
      int s = GetSeed();
      SetSeed(s - 1);
      return s;
    }

    int operator += (int seed) {
      int s = GetSeed() + seed;
      SetSeed(s);
      return s;
    }

    int operator -= (int seed) {
      int s = GetSeed() - seed;
      SetSeed(s);
      return s;
    }

    int operator |= (int seed) {
      int s = GetSeed() | seed;
      SetSeed(s);
      return s;
    }

    int operator &= (int seed) {
      int s = GetSeed() & seed;
      SetSeed(s);
      return s;
    }

    int operator ^= (int seed) {
      int s = GetSeed() ^ seed;
      SetSeed(s);
      return s;
    }

    int operator *= (int seed) {
      int s = GetSeed() * seed;
      SetSeed(s);
      return s;
    }

    int operator /= (int seed) {
      int s = GetSeed() / seed;
      SetSeed(s);
      return s;
    }

    int operator <<= (int seed) {
      int s = GetSeed() << seed;
      SetSeed(s);
      return s;
    }

    int operator >>= (int seed) {
      int s = GetSeed() >> seed;
      SetSeed(s);
      return s;
    }

// Flat between [0,1]
    virtual double GenerateUniform()=0;

// Random Bit methods
    int GenerateBit();
    int GenerateBits(int numBits);

// Random flat integer between [lo, hi]
    int GenerateInt(int lo, int hi) {
      int r = int((hi-lo+1) * GenerateUniform() + lo);
      if (r > hi)
      	r = hi;
      return r;
    }

// Random flat double between [lo, hi)
    double GenerateDouble(double lo=0.0, double hi=1.0) {
      return ((hi-lo) * GenerateUniform() + lo);
    }

// Random exponential distribution f(t) = timeConstant * exp(-t / timeConstant)
    double GenerateExponential(double timeConstant) {
      return - timeConstant * log(GenerateUniform());
    }

// Random Laplace distribution is same as exponential distribution but can be
// negative or positive
    double GenerateLaplace(double timeConstant) {
      double rv = GenerateDouble(-1.0, 1.0);
      double sign = 1.0;
      if (rv<0) sign = -1.0;
      rv = rv*sign;
      return -sign * timeConstant * log(rv);
    }
 
// f(x) = (x/alpha^2)exp(-x^2/(2*alpha^2) where 0<=x
    double GenerateRayleigh(double alpha) {
      return (sqrt(- 2.0 * alpha * alpha * log(GenerateUniform())));
    }

// f(x) = (power+1)*x^power where 0<=x<=1
    double GeneratePower(double power) {
      return pow(GenerateUniform(), 1.0/(power+1));
    }

// f(x) = 1 - (power+1)*x^power where 0<=x<=1
    double GenerateReversePower(double power) {
      return (1 - pow(GenerateUniform(), 1.0/(power+1)));
    }

// f(x) = 2x where 0<=x<=1
    double GenerateTriangleUp() {
      return GeneratePower(1);
    }

// f(x) = 2(1-x)
    double GenerateTriangleDown() {
      return (1.0 - GeneratePower(1));
    }

// f(x) = (n1+n2+1)/(n1!*n2!) * [t^n1 * (1-t^n2)]
    double GenerateBeta(int n1, int n2);

// f(x) = 1/(sqrt(2*pi) * sigma^2) * [exp(-(x-mean)^2/(2*sigma^2))]
    double GenerateGaussian(double mean, double sigma);

// f(x) is provided by the user (rejection method used)
    double GenerateDensityFunction(SpDensityFunction *densityFunction);

// Random vector with given magnitude (default, random unit vector)
    void GenerateVector(double vector[3], double magnitude=1.0);

// random vector with random magnitude given by mean and sigma magnitudes
    void GenerateVector(double vector[3],
			double meanMagnitude,
			double sigmaMagnitude) {
      GenerateVector(vector, GenerateGaussian(meanMagnitude, sigmaMagnitude));
    }
    double GenerateCauchy(double alpha);

// Error function = area under gaussian curve from 0 to x
// note, Erf(infinity = 0.5
    double Erf(double x);

// Inverse error function determines x given Erf(x)
    double InvErf(double erfx);

// Factorial = n * (n-1) * (*n-2) * ... * 1
    double Factorial(int n);

};

//----------------------------------------------------------------------------
// Class: SpUniformLCG
//
// Linear Congruential Generator provides very fast random number generation
// but not of high quality. Applications should never directly use this class
// since it requires a child to provide the representation of the seed (which
// may be a rollbackable integer or a normal integer).
//----------------------------------------------------------------------------
class SpUniformLCG : public SpBaseRandom {

  private:

  protected:

  public:

    virtual double GenerateUniform() {
      int seed = GetSeed();
      seed = (seed*SP_IA+SP_IC)%SP_IM;
      SetSeed(seed);
      return fabs(double(seed) / double(SP_IM));
    }

};

//----------------------------------------------------------------------------
// Class: SpUniformRandBits
//
// Random Bit Generator provides very high quality random numbers, but it
// is slow. Applications should never directly use this class since it
// requires a child to provide the representation of the seed (which may be
// a rollbackable integer or a normal integer)
//----------------------------------------------------------------------------
class SpUniformRandBits : public SpBaseRandom {

  private:

  enum{
    SP_NBITS = 31,
    SP_MAX_32_BIT_INT = 0x7fffffff
  };

  protected:

  public:

    virtual double GenerateUniform() {
      int ranBits = GenerateBits(SP_NBITS);
      return (double(ranBits) / (1.0+double(SP_MAX_32_BIT_INT)));
    }

};

//----------------------------------------------------------------------------
// SpRandom is the normal way people should use the random number generator
// when rollback support is not required. It is slow, but very good. If
// rollback support is needed, you should use the RB_SpRandom class which is
// identical to this class except that it uses an RB_int instead of an int
// to represent its seed.
//----------------------------------------------------------------------------
class SpRandom : public SpUniformRandBits {

  private:

    int Seed;

  protected:

  public:

    SpRandom(int seed=12345) {Seed = seed;}

    virtual int GetSeed() {return Seed;}
    virtual void SetSeed(int seed) {Seed = seed;}
    virtual void ExchangeSeed(int &seed) {
      int i = Seed;
      Seed = seed;
      seed = i;
    }
    virtual operator int () {return Seed;}

};

//----------------------------------------------------------------------------
// SpFastRandom is a very fast random number generator but it is not of
// high quality. Use RB_FastRandom when rollback support is needed.
//----------------------------------------------------------------------------
class SpFastRandom : public SpUniformLCG {

  private:

    int Seed;

  protected:
  public:

    SpFastRandom(int seed=1) {Seed = seed;}

    virtual int GetSeed() {return Seed;}
    virtual void SetSeed(int seed) {Seed = seed;}
    virtual void ExchangeSeed(int &seed) {
      int i = Seed;
      Seed = seed;
      seed = i;
    }
    virtual operator int () {return Seed;}

};

#endif
