// Filename: SpStopWatch.H

#ifndef StopWatchObject
#define StopWatchObject

#include "SpIostream.H"
#include "SpFstream.H"
#include "SpStrstream.H"
#include "SpIomanip.H"
#include "SpTime.H"
#include <sys/types.h>
#ifndef __SPEEDES_NT__
#include <sys/time.h>
#else
#include <windows.h>
#endif
#include "SpString.H"
#include "SpStdlib.H"

/*****************************************************************************
* Class: SpStopWatch
* Summary: Class for single and/or cumulative timing of any time intervals in
*          the simulation. Timing always includes both wall (actual/real/
*          total) time and CPU (computer processing) time.
* Description: Use objects of this class to time any aspects of the
*              simulation. Objects can be named during construction, reset, and
*              repeatedly started and stopped. If not reset, they only keep
*              track of cumulative time. Either CPU or wall time can be set as
*              the default of what is returned and/or printed, but STOPWATCHes
*              always keep track of both. Thus the flag for CPU or wall time
*              returning and printing can be changed at any time ad infinitum
*              during the cumulative timing.
*****************************************************************************/
class SpStopWatch {

 public:

/******************************************************************************
* Method: UseWallTime
* Summary: 
* Description: 
* Return: None
******************************************************************************/
  void UseWallTime() {ClockFlag = 0;}

/******************************************************************************
* Method: UseClockTime
* Summary: 
* Description:
* Return: None
******************************************************************************/
  void UseClockTime() {ClockFlag = 1;}

/******************************************************************************
* Method: StartTimer
* Summary: Begin timing a new time interval for either wall time or CPU time,
*          and continue cumulative wall time and CPU time
* Description: Objects of this type will return the time interval beginning
*              with this call when the next StopTimer() is called. The time
*              value returned depends on the setting of the Clock Flag (the
*              default is wall time) when StopTimer is called, which can be set
*              before or after calling StartTimer() with the methods
*              UseWallTime() and UseClockTime(). The first time this method is
*              called after object construction or after calling Reset(), this
*              method begins cumulative timing from zero for both wall clock
*              timing and CPU timing. Otherwise it continues accumulating both
*              wall clock and CPU time from past StartTimer()/StopTimer
*              intervals. Regardless of the Clock Flag, the class cumulatively
*              stores both types of time.
* Return: None
******************************************************************************/
  void StartTimer();

/******************************************************************************
* Method: StopTimer
* Summary: Stop timing the interval since the last StartTimer() call, stop
*          cumulative wall time and CPU time, and return the time value of the
*          time interval since the last StartTimer() call
* Description: This returns the time interval beginning with the most recent
*              StartTimer() call. The time value returned depends on the
*              current setting of the Clock Flag (the default is wall time),
*              which must be set previously, but can be set before or after the
*              most recent StartTimer() call with the methods UseWallTime() and
*              UseClockTime(). The first time StartTimer() and StopTimer() are
*              called after object construction or after calling Reset(), this
*              class begins cumulative timing from zero for both wall clock
*              timing and CPU timing. Otherwise it continues accumulating both
*              wall clock and CPU time from past StartTimer()/StopTimer
*              intervals. Regardless of the Clock Flag, the class cumulatively
*              stores both types of time.
* Return: Either wall time or CPU time in seconds, depending on the Clock Flag,
*         which is set by using either UseWallTime() or UseClockTime()
******************************************************************************/
  double StopTimer();

/******************************************************************************
* Method: Reset
* Summary: Reset cumulative wall time and and CPU time
* Description: The first time StartTimer() and StopTimer() are called after
*              object construction or after calling Reset(), this class begins
*              cumulative timing from zero for both wall clock timing and CPU
*              timing. Otherwise it continues accumulating both wall clock and
*              CPU time from past StartTimer()/StopTimer intervals. Regardless
*              of the Clock Flag, the class cumulatively stores both types of
*              time.
* Return: None
******************************************************************************/
  void Reset()
  {
    CumulativeTime = 0.0;
    CumulativeClockTime = 0.0;
  }

/******************************************************************************
* Method: get_CumulativeTime
* Summary: Get cumulative wall time
* Description: The first time StartTimer() and StopTimer() are called after
*              object construction or after calling Reset(), this class begins
*              cumulative timing from zero for both wall clock timing and CPU
*              timing. Otherwise it continues accumulating both wall clock and
*              CPU time from past StartTimer()/StopTimer intervals. Regardless
*              of the Clock Flag, the class cumulatively stores both types of
*              time.
* Return: Cumulative wall time in seconds
******************************************************************************/
  double get_CumulativeTime() const {return CumulativeTime;}

/******************************************************************************
* Method: get_CumulativeClockTime
* Summary: Get cumulative CPU time
* Description: The first time StartTimer() and StopTimer() are called after
*              object construction or after calling Reset(), this class begins
*              cumulative timing from zero for both wall clock timing and CPU
*              timing. Otherwise it continues accumulating both wall clock and
*              CPU time from past StartTimer()/StopTimer intervals. Regardless
*              of the Clock Flag, the class cumulatively stores both types of
*              time.
* Return: Cumulative CPU time in seconds
******************************************************************************/
  double get_CumulativeClockTime() const {return CumulativeClockTime;}

/******************************************************************************
* Method: get_name
* Summary: Get the name of this Stopwatch
* Description: This is set in the constructor of this class
* Return: The name of this Stopwatch
******************************************************************************/
  const char *get_name() const {return name;}

/******************************************************************************
* Overloaded Method: print
* Summary: Print the name of this stopwatch and the cumulative wall and CPU
*          times to cout
* Description: 
* Return: None
******************************************************************************/
  void print() const {print(cout);}

/******************************************************************************
* Overloaded Method: print
* Summary: Print the name of this stopwatch and the cumulative wall and CPU
*          times
* Description: 
* Return: None
******************************************************************************/
  void print(ostream &out // IN: ostream to which to print
	     ) const;   

/******************************************************************************
* Constructor:
* Summary: Constructs an unnamed stop watch object
* Description: 
******************************************************************************/
  SpStopWatch();

/******************************************************************************
* Constructor: 
* Summary: Initializes the object and sets the name of the stopwatch (can be
*          NULL)
* Description: 
******************************************************************************/
  SpStopWatch(char *n // IN: name of stopwatch (can be NULL)
	    );

  ~SpStopWatch();

 private:

  double StartTime;
  double CumulativeTime;
  clock_t StartClockTime;
  double CumulativeClockTime;
  char *name;
  int ClockFlag;

};

typedef SpStopWatch STOPWATCH;

#endif



