// Yacc/Bison input file to implement grammar for SPEEDES parfile
// format 

// Needed for this grammar: 
//
//   SpSet::AddRef(char *key) -- to add references to a set
//   SpSetArray::Push(builtin) -- to add dynamically resizable arrays
//x  class SpSetStack -- to implement set_stack 
//   SpSetUserDefined::Append(SpSetElement *), to construct the 
// argument list for a user-defined element 

%{
#define YYDEBUG 1
#include "SpSet.H"
#include <FlexLexer.h>

#include <stdio.h>
#include "SpMath.H"

extern "C" {    
  char *strdup(const char*);    
}
#define YYERROR_VERBOSE  
void yyerror(const char *);

#define _NML_INTERNAL_
#include "SpParserCommon.H"
#include "SpParser.H"

static SpSetUserDefined *build_udef(const char *key,SpBinaryTree *kwdtable, 
			     SpSetArray *arglist);
static ElementType SpEtype(int);
static void cleanup_include_stack(SpSetStack &);
static void cleanup_intro_tree(SpBinaryTree *);
static void unify_includes(SpIncludeStack &);
#define YYPARSE_PARAM commondata
#define YYLEX_PARAM commondata
#define TOPLEV ((struct SpParserCommon *) commondata)->toplev
#define INTRO ((struct SpParserCommon *) commondata)->intro
#define STACK ((struct SpParserCommon *) commondata)->set_stack
#define ARRAY ((struct SpParserCommon *) commondata)->current_array 
#define INCLUDES ((struct SpParserCommon *) commondata)->include_statements

%}

%pure_parser

%union {
  double floatval;
  char *stringval;
  int intval;
  bool boolval;
  struct floatval_seq *fseq;
  struct intval_seq *iseq;
  struct boolval_seq *bseq;
  struct str_seq *sseq;
  void *voidptr;
  SpSet *setptr;
  SpSetElement *setelement;
  SpSetUserDefined *udef;
  SpBinaryTree *tree;
  struct kwd_struct *kwd;
  struct include_statement *include_stmt;
  SpSetArray *array;
}

%{
extern "C" {
  int yylex(YYSTYPE *lvalp,void *commondata);
}
  %}

%token <floatval> FLOAT
%token <stringval> STRING
%token <intval> INT
%token <intval> LEOF
%token <boolval> TOKBOOL
%token <stringval> LNAME
%token <stringval> REFNAME
%token <stringval> COMMENT
%token <stringval> UKWD
%token <intval> TYPENAME
%token <intval> BUILTINTYPE
%token <intval> OPENPROTO
%token <intval> CLOSEPROTO
%token <intval> OPENSET
%token <intval> CLOSESET
%token <stringval> INCLUDE

%type <tree> intro
%type <tree> kwd_statement_seq
%type <setptr> set_begin
%type <setptr> set_end
%type <setptr> set
%type <voidptr> set_member
%type <setptr> set_member_seq
%type <setptr> parfile
%type <array> builtin_type_seq
%type <kwd> kwd_statement_or_comment
%type <array> kwd_val_seq
%type <setelement> kwd_val
%type <array> array_start
%type <array> array_member
%type <include_stmt> include

%%
parfile: intro set_member_seq LEOF {
  $$ = $2;
  TOPLEV = $$;
  // Handle include statements
  unify_includes(INCLUDES);
  // Clean up anything needing cleanup
  //  cleanup_include_stack(INCLUDES);
  cleanup_intro_tree(INTRO);
  YYACCEPT;
}

intro: kwd_statement_seq intro_extensions {
  $$ = $1;
  INTRO = $$;
}

// Leave for future extensions 
intro_extensions: {
}

// Construct table of user keywords. This is a lexical tie-in.
// The scanner will use this to generate UKWD tokens as necessary
// during the set_member_seq phase

kwd_statement_seq: /* Nothing */ {
  $$ = new SpBinaryTree;
  $$->SetBalancedTreeMode();
} | kwd_statement_or_comment {
  $$ = new SpBinaryTree;
  $$->SetBalancedTreeMode();
  if ($1) {
    $$->Insert((void *)$1,$1->name);
  }
} | kwd_statement_seq kwd_statement_or_comment {
  if ($2) {
    $1->Insert((void *)$2,$2->name);
  }
  $$ = $1;
}

// Construct kwd_struct for each keyword. 

kwd_statement_or_comment: UKWD builtin_type_seq {
  $$ = new kwd_struct;
  $$->name = strdup($1);
  $$->typelist = $2;
} | comment {
  $$ = NULL;
}

builtin_type_seq: /* Nothing */ {
  //  $$ = new struct intval_seq;
  //  $$->length = 0;
  //  $$->val = NULL;
  $$ = new SpSetArray(INT_ET,0);
} | BUILTINTYPE {
  // $$ = new struct intval_seq;
  //  $$->length = 1;
  //  $$->val = (int *) malloc(sizeof(int));
  //  $$->val[$$->length-1] = SpEtype($1);
  $$ = new SpSetArray(INT_ET,0);
  $$->Push(SpEtype($1));
} | builtin_type_seq BUILTINTYPE {
  $$ = $1;

  // This is inefficient and should be changed
  //  realloc($$->val,(1+$$->length) * sizeof(int));
  //  $$->val[$$->length] = SpEtype($2);
  //  $$->length++;
  $$->Push(SpEtype($2));
}

set: set_begin set_member_seq set_end {
  $$ = $3;
}

set_begin: LNAME OPENSET {
  if (!STACK.Top()) {
    STACK.Push(new SpSet);
  }
  $$ = STACK.Top()->Embed($1);
  $$->Insert(0,"_ncomments");
  STACK.Push($$);
}

set_end: CLOSESET {
  $$ = STACK.Pop();
}

set_member_seq: /* Nothing */ {$$=STACK.Pop();} 
| set_member {$$ = STACK.Top();} 
| set_member_seq set_member {$$ = STACK.Top();} 


set_member: set {$$=0;}
| prim_member {$$=0;}
| userdef {$$=0;}
| array_member {$$=0;}
| ref_member {$$=0;}
| comment {$$=0;}
| include {$$=0;}

prim_member: LNAME INT {STACK.Top()->Insert($2,$1);}
| LNAME FLOAT {STACK.Top()->Insert($2,$1);}
| LNAME TOKBOOL {STACK.Top()->Insert($2,$1);}
| LNAME STRING {STACK.Top()->Insert($2,$1);}


userdef: UKWD kwd_val_seq {
  STACK.Top()->Insert(*build_udef($1,INTRO,$2),$1);
}

kwd_val: INT {
  $$ = new SpSetInt($1);
} | FLOAT {
  $$ = new SpSetFloat($1);
} | STRING {
  $$ = new SpSetString($1);
} | TOKBOOL {
  $$ = new SpSetLogical($1);
} | REFNAME {
  $$ = new SpSetUnres($1);
}

kwd_val_seq: kwd_val {
  $$ = new SpSetArray;
  $$->Push(*$1);
} | kwd_val_seq kwd_val {
  $$ = $1->Push(*$2);
}

array_start: LNAME INT INT {
  $$ = &STACK.Top()->AddArray($1,INT_ET,2);
  ((int *)(*$$))[0] = $2;
  ((int *)(*$$))[1] = $3;
  ARRAY = $$;
} | LNAME FLOAT FLOAT {
  $$ = &STACK.Top()->AddArray($1,FLOAT_ET,2);
  ((double *)(*$$))[0] = $2;
  ((double *)(*$$))[1] = $3;
  ARRAY = $$;
} | LNAME TOKBOOL TOKBOOL {
  $$ = &STACK.Top()->AddArray($1,LOGICAL_ET,2);
  ((bool *)(*$$))[0] = $2;
  ((bool *)(*$$))[1] = $3;
  ARRAY = $$;
} | LNAME STRING STRING {
  $$ = &STACK.Top()->AddArray($1,STRING_ET,2);
  ((char **)(*$$))[0] = $2;
  ((char **)(*$$))[1] = $3;
  ARRAY = $$;
}

array_member: array_start builtin_val_seq {
  ARRAY = NULL;
}

builtin_val_seq: int_seq | float_seq | bool_seq | string_seq
;

int_seq: /* Nothing */
| INT {
  if (!ARRAY) YYABORT;
  ARRAY->Push($1);
} | int_seq INT {
  if (!ARRAY) YYABORT;
  ARRAY->Push($2);
}

float_seq: /* Nothing */
| FLOAT {
  if (!ARRAY) YYABORT;
  ARRAY->Push($1);
} | float_seq FLOAT {
  if (!ARRAY) YYABORT;
  ARRAY->Push($2);
}

string_seq: /* Nothing */
| STRING {
  if (!ARRAY) YYABORT;
  ARRAY->Push($1);
} | string_seq STRING {
  if (!ARRAY) YYABORT;
  ARRAY->Push($2);
}

bool_seq: /* Nothing */
| TOKBOOL {
  if (!ARRAY) YYABORT;
  ARRAY->Push($1);
} | bool_seq TOKBOOL {
  if (!ARRAY) YYABORT;
  ARRAY->Push($2);
}

ref_member: REFNAME {
  STACK.Top()->AddRef($1);
}

comment: COMMENT {
  int commentstatus;
  STACK.Top()->GetInt("_ncomments",0,commentstatus);
  if (commentstatus) STACK.Top()->Insert(0,"_ncomments");
  char *commentname = (char *) malloc(sizeof("_comment")+
			     int(1+log10(1+STACK.Top()->
					 GetInt("_ncomments"))));
  sprintf(commentname,"_comment%d",
	  STACK.Top()->GetInt("_ncomments"));
  STACK.Top()->Insert($1,commentname);
  STACK.Top()->Insert(STACK.Top()->GetInt("_ncomments")+1,
					    "_ncomments");
  free(commentname);
}

include: INCLUDE {
  $$ = new include_statement;
  $$->include_dest = STACK.Top();
  $$->include_target = strdup($1);
  INCLUDES.Push($$);
}

%%

void yyerror(const char *message) 
{
  cerr << message << endl;
}

SpParser::SpParser(char *filename)
{
  input = new ifstream(filename);
}

SpParser::SpParser(istream *infile) 
{
  input = infile;
}

#define __DONT_INCLUDE_y_tab_h__
#include "SpParserFlexLexer.H"

SpSet *SpParser::Parse(SpSet *set)
{
  if (set) {
    commondata.set_stack.Push(set);
  }
  else {
    commondata.set_stack.Push(new SpSet);
  }

  commondata.scanner = new SpParserFlexLexer(input);
  yyparse((void *) &commondata);
  return commondata.toplev;
}

static SpSetUserDefined *build_udef(const char *key,SpBinaryTree *kwdtable, 
			     SpSetArray *arglistp) 
{
  int k;
  SpSetUserDefined *retval = NULL;
  SpSetArray arglist = *arglistp;
  struct kwd_struct *kwd = (struct kwd_struct *) kwdtable->Find(key);
  if (kwd->typelist->GetLength() == arglist.GetLength()) {
    if (kwd) {
      for(k=0;k<arglist.GetLength();k++) {
	if (arglist[k].GetType() == STRING_ET && 
	   (int) (*(kwd->typelist))[k] == UNRES_ET) {
	  char *refname = (char *) arglist[k];
	  SpSetElement *ep = &arglist[k];
	  delete (SpSetString *) ep;
	  ep = new SpSetUnres(refname);
	}
	if ((int) (*(kwd->typelist))[k] != arglist[k].GetType()) {
	  cerr << "SpParser: userdefined typelist mismatch ";
	  cerr << "on keyword " << key << "!" << endl;
	  return NULL;
	}
      }
      retval = new SpSetUserDefined(key,arglistp);
    }
  }
  return retval;
}

// This function handles the include statements that have
// appeared in the par file, if necessary trying to read the
// target set from a file on disk. It also cleans up each 
// include_statement as it goes (obviating the need for the
// next function, cleanup_include_stack())
static void unify_includes(SpIncludeStack &includestack)
{
  struct include_statement *inc;
  SpSetElement *targetset;
  if (includestack.GetNumElements()) {
    while((inc = includestack.Pop())) {
      // Try to find the target set in the scope of the destination
      // set. If not, then try opening a file named 
      // inc->include_target
      if ((targetset = inc->include_dest->
	   find_in_scope(inc->include_target))) {
	if (targetset->GetType() == SET_ET) {
	  inc->include_dest->Union((SpSet *) targetset);
	}
      }
      else {
	istream *includefile;
	includefile = new ifstream(inc->include_target);
	if (includefile->bad()) {
	  cerr << "SpParser: Warning: couldn't find or open ";
	  cerr << "include file or set named " << inc->include_target;
	  cerr << endl;
	}
	else {
	  SpParser includefile_parser(includefile);
	  SpSet *new_includeset = includefile_parser.Parse();
	  inc->include_dest->Union(new_includeset);
	  delete new_includeset;
	}
      }
      free(inc->include_target);
      delete inc;
    }
  }
}

static void cleanup_include_stack(SpIncludeStack &stack)
{
  struct include_statement *inc;
  if (stack.GetNumElements()) {
    while((inc = stack.Pop())) {
      free(inc->include_target);
      delete inc;
    }
  }
}

static void cleanup_intro_tree(SpBinaryTree *tree)
{
  struct kwd_struct *kwd;
  if (tree) {
    while((kwd = (struct kwd_struct *) tree->RemoveFirstElement())) {
      delete kwd;
    }
    delete tree;
  }
}

static ElementType SpEtype(int token)
{
  ElementType retval;
  switch(token) {
  case TOKBOOL:
    retval = LOGICAL_ET;
    break;
  case INT:
    retval = INT_ET;
    break;
  case FLOAT:
    retval = FLOAT_ET;
    break;
  case STRING:
    retval = STRING_ET;
    break;
  case REFNAME:
    retval = UNRES_ET;
    break;
  default:
    retval = NONE_ET;
  }
  return retval;
}

#ifdef YY_MAIN
main(int argc,char **argv) 
{
  istream *inputfile;
  yydebug=1;
  if (argc>1) {
    inputfile = new ifstream(argv[1]);
    if (inputfile->bad()) {
      inputfile = &cin;
    }
  }
  else {
    inputfile = &cin;
  }

  SpParser parser(inputfile);
  SpSet *inputset = parser.Parse();
  int nelements = inputset->GetNumElements();
  cout << "Number of elements in inputset = " << nelements << endl;
  SpSet *parameterset = inputset->GetSet("parameters");
  if (parameterset) {
    cout << "FOUND set named \"parameters\" in top level set" << endl;
    int getintstatus;
    int retval = parameterset->GetInt("multiroute_thresh",23,
				      getintstatus);
    if (getintstatus) {
      cout << "Couldn't find integer \"multiroute_thresh\" ";
      cout << "in set \"parameters\"" << endl;
    }
    else {
      cout << "Value of integer \"multiroute_thresh\" ";
      cout << "in set \"parameters\" = " << retval << endl;
    }

    SpSetArray *intarray = parameterset->GetArray("intarray");
    if (intarray) {
      if ((int)(*intarray)[10] == 320) {
	cout << "intarray[10] = 320, as expected" << endl;
      }
      else {
	cout << "Unexpected value " <<  (int)(*intarray)[10] << " for ";
	cout << "intarray[10]" << endl;
      }
    }
    SpSet *gvtref = parameterset->GetSet("gvt_parameters");
    if (gvtref) {
      cout << "Retrieved gvt_parameters reference from ";
      cout << "parameters" << endl;
      int refstatus;
      int ngvt = gvtref->GetInt("Ngvt",5,refstatus);
      if (refstatus) {
	cout << "Failed to access Ngvt member from reference ";
	cout << "to set gvt_parameters" << endl;
      }
      else {
	if (ngvt == 1000) {
	  cout << "Ngvt in refset had expected value " << ngvt << endl;
	}
	else {
	  cout << "Ngvt in refset had unexpected value " << ngvt << endl;
	}
      }
      int hlaint = gvtref->GetInt("hlaint",42,refstatus);
      if (refstatus) {
	cout << "Failed to access hlaint member from reference ";
	cout << "to set gvt_parameters (which in turn came ";
	cout << "from included set HLA" << endl;
      }
      else {
	if (hlaint == 53401) {
	  cout << "hlaint in refset (from included set HLA) ";
	  cout << "had expected value " << hlaint << endl;
	}
	else {
	  cout << "hlaint in refset (from included set HLA) had ";
	  cout << "unexpected value " << hlaint << endl;
	}
      }
    }
  }
  else {
    cout << "COULDN'T find set named \"parameters\" in top level set";
    cout << endl;
  }
  
  int gotthreelevel=0;
  int thirdleveli;
  SpSet *SpeedesServerSet = inputset->GetSet("SpeedesServer");
  if (SpeedesServerSet) {
    SpSet *HostRouterSet = SpeedesServerSet->GetSet("HostRouter");
    if (HostRouterSet) {
      SpSet *thirdlevelset = HostRouterSet->GetSet("threelevel");
      if (thirdlevelset) {
	gotthreelevel = 1;
	thirdleveli = thirdlevelset->GetInt("i");
      }
    }
  }

  if (gotthreelevel) {
    cout << "Got threelevel set" << endl;
    cout << "value of integer i in threelevel set = " << thirdleveli << endl;
    cout << "\n\nPassed: SpParser test\n" << endl;
  }
  else {
    cout << "Couldn't find threelevel set" << endl;
  }
  
  
}
#endif // #ifdef YY_MAIN
